//#**************************************************************
//# filename:             ANCFBeamBE2D.cpp
//#
//# author:               Gerstmayr, Vetyukov
//#
//# generated:						
//# description:          
//# comments:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#include "element.h"
#include "material.h"
#include "Node.h"
#include "body2d.h"
#include "ANCFBeamBE2D.h"


ANCFBeamBE2D::~ANCFBeamBE2D(void) {};

void ANCFBeamBE2D::ElementDefaultConstructor() 
{
	//set number of shape functions: default is 4 (for nodal dofs r and r'), enhanced element works with continuous curvature, i.e. additional dof is r'' ==> ns = 6
	ns = 4;
	
	// set nodes
	this->nodes.SetLen(2);
	for (int i=1; i<=NNodes(); i++)
	{
		this->nodes(i) = i;
	}

	//set reference configuration
	q0.SetLen(2*SOS());
	q0.SetAll(0.);
	x_init.SetLen(2*SOS());
	
	//set material number
	materialnum = 1;

	//set size
	size = Vector3D(1,0.1,0.1);
}

//default set function, use for C^1 (ns=4) elements only 
void ANCFBeamBE2D::SetANCFBeamBE2D(int n1, int n2, int material_num, const Vector3D& color)
{
	ElementDefaultConstructor();
	
	// set color and material number
	col = color;
	SetMaterialNum(material_num);
	
	// set nodes
	this->nodes(1) = n1;
	this->nodes(2) = n2;

	// set reference configuration
	x_init = Vector(2*SOS());
	int count=0;
	for (int nn=1; nn<=NNodes(); nn++)
	{
		const ANCFNodeS1_2D &node = dynamic_cast<const ANCFNodeS1_2D&> (GetNode(nn));
		q0(++count) = node.GetRefPos2D()(1);
		q0(++count) = node.GetRefPos2D()(2);
		q0(++count) = node.GetRefSlope1()(1);
		q0(++count) = node.GetRefSlope1()(2);
		for (int i=1; i<=node.SOS(); i++)
		{
			x_init((nn-1)*node.SOS()+i) = node.X_Init()(i);
			x_init(SOS()+(nn-1)*node.SOS()+i) = node.X_Init()(node.SOS()+i);
		}
	}
	
	//set size
	double element_length = CalculateElementLength();
	Beam2DProperties bp = GetBeam2DProperties();
	assert(bp.IsRectangularCrossSection() && "ERROR: ANCFBeamBE2D implemented for rectangular cross section only!");
	size = Vector3D(element_length, bp.GetBeamThicknessY(), bp.GetBeamThicknessZ());
}

//$ PG 2014-6-10: [ deprecated: old type of set functions
void ANCFBeamBE2D::SetANCFBeamBE2D(int bodyind, const Vector& xc1, const Vector& xc2, int n1, int n2, int material_num, const Vector3D& size, const Vector3D& color)
{
	SetFiniteElementGenericBeam2D(bodyind, xc1, xc2, n1, n2, material_num, size, color);
};
//$ PG 2014-6-10: ]

double ANCFBeamBE2D::CalculateElementLength() const
{
	//
	// calculation of the length:
	// --------------------------
	//
	// length = \int_{-1}^{1} 1/2 |r'_0(x)| dx,   where x is in [-1,1]   (which is the interval of Gauss points for numerical integration)
	//
	//
	// representation of r'_0(x):
	// ----------------------------
	//
	// by definition, the beam element axis r_0(x) (r_0 is the axis in reference configuration, and x is in [-1,1])
	// is a cubic spline defined by the nodes a and b, i.e., the nodal positions r_0^a and r_0^b, and the nodal slopes r'_0^a and r'_0^b, respectively.
	// let us fix the i-th component (in space), then the i-th component of the beam axis is defined by 
	// (r_0(x))_i = a_0 + a_1 x_0 + a_2 x_0^2 + a_3 x_0^3, where the coefficients a_0,..,a_3 are yet unknown.
	// the coefficients a_0,..,a_3 are found by matching the boundary conditions, 
	// (r_0(-1))_i=(r_0^a)_i, (r_0(1))_i=(r_0^b)_i, (r'_0(-1))_i=(r'_0^a)_i, (r'_0(1))_i=(r'_0^b)_i
	// which leads to solving the linear system
	// [ (r_0^a)_i  ]   [ 1 -1  1 -1 ] [ a_0 ]
	// [ (r'_0^a)_i ] = [ 0  1 -2  3 ] [ a_1 ]
	// [ (r_0^b)_i  ]   [ 1  1  1  1 ] [ a_2 ]
	// [ (r'_0^b)_i ]   [ 0  1  2  3 ] [ a_3 ]
	// or, in short:
	// q0_i = \AMat \aVec.
	// hence, the coefficients read
	// \aVec = \AInv q0_i, where \AInv = \AMat^{-1},
	// and we end up with the representation as an inner product
	// (r'_0(x_0))_i = [  0  1  2(x_0)  3(x_0)^2] \AInv q0_i              (*)
	// 

	double length = 0;

	Vector2D r0a(q0(1),q0(2));
	Vector2D r0b(q0(1+4),q0(2+4));
	Vector2D r0a_prime(q0(3),q0(4));
	Vector2D r0b_prime(q0(3+4),q0(4+4));
	r0a_prime.Normalize();
	r0b_prime.Normalize();
	
	// calculate some geometrical quantities
	Vector2D vector_AB(r0b-r0a);                 // vector from node a to node b
	double distance_AB = vector_AB.Norm();       // distance between node a and node b
	
	// "full circle" beam elements or elements of length 0 are not allowed
	assert(distance_AB > 0);                     
	
	double alpha = acos(max(-1.,min(1.,(r0a_prime*vector_AB)/distance_AB)));    // acos returns value between [0,pi]; alpha is angle between vector_AB and slope in node a (which in an arc is equal to the angle between vector_AB and slope in node b) (r0a_prime is already normalized!)
	double sin_alpha = sin(alpha);   // is non-negative, since alpha is non-negative
	
	if (alpha < 1e-14) // if this is a straight beam element
	{
			return distance_AB; 
	}
	else if (alpha > MY_PI - 1e-14)  // if this is an almost full circle
	{
		assert (0 && "full circle beam elements elements are not allowed");
	}
	else // general curved beam element
	{
		distance_AB *= alpha/sin_alpha;    // arc_length = radius * theta,  where theta is the angle between the two sekants of the arc; there holds  radius = distance_AB / (2*sin(alpha)), and theta = 2*alpha
	}

	//multiply slopes at nodes with arclength and divide by two, since x is in [-1,1]
	r0a_prime *= distance_AB/2;
	r0b_prime *= distance_AB/2;
	
	// Matrix containing nodal positions and nodal slope vectors of arc-length divided by two (size of interval [-1,1])
	Matrix Q(4,3);
	
	// fill columns of Q
	for (int j=1; j<=3; j++)
	{
		Q(1,j)=r0a(j);
		Q(2,j)=r0a_prime(j);
		Q(3,j)=r0b(j);
		Q(4,j)=r0b_prime(j);
	}

	// first row (would be nr. 0) can be neglected due to multiplication with 0 in representation of (r'_0(x_0))_i   (see *)
	Matrix AInv(3,4);
	AInv(1,1) = -3;
	AInv(1,2) = -1;
	AInv(1,3) = 3;
	AInv(1,4) = -1;
	AInv(2,1) = 0;
	AInv(2,2) = -1;
	AInv(2,3) = 0;
	AInv(2,4) = 1;
	AInv(3,1) = 1;
	AInv(3,2) = 1;
	AInv(3,3) = -1;
	AInv(3,4) = 1;

	IntegrationRule ruleAxial;
	IntegrationRule::DefineIntegrationRuleLine(ruleAxial, 9);

	for (IntegrationPointsIterator ip(&ruleAxial); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point().X();
		Vector r0_prime_at_xj = Vector(1.,2.*x,3.*Sqr(x))*AInv*Q;
		//mbs->UO() << "r'r0_prime(" << x << ") = " << r0_prime_at_xj << "\n";
		length += ip.Weight()*r0_prime_at_xj.GetNorm();
	}	
	length /= 4.; // factor of AInv: 1/4

	return length;
}

// get direction of of action of i-th DOF
// AH: this is for drawing cones that represent the coordinate constraints
Vector3D ANCFBeamBE2D::GetDOFDirD(int idof) const
{
	if (idof > NS()) 
		idof -= NS();

	if (idof == 1) return ToP3D(Vector2D(1.,0.));
	else if (idof == 2) return ToP3D(Vector2D(0.,1.));

	return Vector3D(0.,0.,0.);
};

Vector3D ANCFBeamBE2D::GetDOFPosD(int idof) const //returns postion of i-th DOF
{
	if (idof <= 4)
		return ToP3D(GetPos2D(-1., TCD_draw));
	else
		return ToP3D(GetPos2D(1., TCD_draw));
};

void ANCFBeamBE2D::EvalF2(Vector& f, double t)
{
	Body2D::EvalF2(f, t);
	TMStartTimer(22);

	// do the computation
	ConstVector<ANCFBeamBE2DmaxDOF> fadd(2*NS());	// temporary storage vector
	ConstVector<ANCFBeamBE2DmaxDOF> delta_eps(2*NS()), delta_kappa(2*NS());
	
	double EI = GetBeam2DProperties().BeamEIy();
	double EA = GetBeam2DProperties().BeamEA();

#if 0 //$ PG 2014-6-23: default, as implemented by AH and AP

	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point2D().X();
		double eps = GetGamma1(x, TCD_compute);
		double kappa = GetKappa(x, TCD_compute);
		GetDeltaGamma1(x, delta_eps); 
		GetDeltaKappa(x, delta_kappa);

		fadd += ((EA*eps)*delta_eps + (EI*kappa)*delta_kappa) * (ip.Weight() * GetElementLength()*0.5);
	}	

#else  //$ PG 2014-6-23: testing reduced integration (bending: order 5 Gauss, axial def.: order 9 Gauss, both as in ANCFBeam3DTorsion)

	IntegrationRule ruleAxial, ruleCurvature;
	IntegrationRule::DefineIntegrationRuleLine(ruleCurvature, 5);
	IntegrationRule::DefineIntegrationRuleLine(ruleAxial, 9);

	for (IntegrationPointsIterator ip(&ruleCurvature); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point2D().X();
		double kappa = GetKappa(x, TCD_compute);
		GetDeltaKappa(x, delta_kappa);

		fadd += ((EI*kappa)*delta_kappa) * (ip.Weight() * GetElementLength()*0.5);
	}
	for (IntegrationPointsIterator ip(&ruleAxial); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point2D().X();
		double eps = GetGamma1(x, TCD_compute);
		GetDeltaGamma1(x, delta_eps); 


		fadd += ((EA*eps)*delta_eps) * (ip.Weight() * GetElementLength()*0.5);
	}	

#endif

	f -= fadd;

	/*ConstVector<ANCFBeamBE2DmaxDOF> fadd(2*NS());	// temporary storage vector
	ConstVector<ANCFBeamBE2DmaxDOF> delta_eps(2*NS()), delta_kappa(2*NS());
	
	double EI = GetBeamEIy();
	double EA = GetBeamEA();

	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point2D().X();

		//Vector2D rx = GetPosx2D(x);
		//Vector2D rxx = GetPosxx2D(x);

		Vector2D rx, rxx;
		for (int i = 1; i <= NS(); i++)
		{
			//double sx = ip.GetVectorData(1)->Get(i) * 2./lx;
			//double sxx = ip.GetVectorData(2)->Get(i) * 4./(lx*lx);

			double sx = GetS0x(x, i) * 2./GetElementLength();
			double sxx = GetS0xx(x, i) * 4./(GetElementLength()*GetElementLength());
			double d1 = q0(2*i-1) + XG(2*i-1);
			double d2 = q0(2*i  ) + XG(2*i  );

			rx(1) += sx * d1;
			rx(2) += sx * d2;
			rxx(1) += sxx * d1;
			rxx(2) += sxx * d2;
		}

		double rxn2 = rx.Norm2();

		double eps = rx.Norm() - 1.0;
		double kappa = (rx.Cross(rxx))/rxn2;

		double d = 1./rx.Norm();

		double kdux = -2.*rx(1)*kappa / rxn2 + rxx(2) / rxn2;
		double kduxx = -rx(2) / rxn2;
		double kdwx = -2.*rx(2)*kappa / rxn2 - rxx(1) / rxn2;
		double kdwxx = rx(1) / rxn2;

		for (int j = 1; j <= NS(); j++)
		{
			double sx = GetS0x(x, j) * 2./GetElementLength();
			double sxx = GetS0xx(x, j) * 4./(GetElementLength()*GetElementLength());
			//double sx = ip.GetVectorData(1)->Get(j) * 2./lx;
			//double sxx = ip.GetVectorData(2)->Get(j) * 4./(lx*lx);

			delta_eps(2*j-1) = d * rx(1) * sx;
			delta_eps(2*j  ) = d * rx(2) * sx;

			delta_kappa(2*j-1) = (kdux * sx + kduxx * sxx);
			delta_kappa(2*j  ) = (kdwx * sx + kdwxx * sxx);
		}

		fadd += ((EA*eps)*delta_eps + (EI*kappa)*delta_kappa) * (ip.Weight() * GetElementLength()*0.5);
	}	
	f -= fadd;*/

	TMStopTimer(22);
};

void ANCFBeamBE2D::EvalM(Matrix& m, double t)
{
	if (this->massmatrix.Getrows() != 0)	// mass matrix already computed
	{
		m = massmatrix;
		return;
	}
	
	Matrix H(SOS(),Dim());
	massmatrix.SetSize(SOS(), SOS());
	massmatrix.SetAll(0.);

	double rhoA = GetBeam2DProperties().BeamRhoA();
	
	for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
	{
		for (int i = 1; i <= NS(); i++)
		{
			H(2*i-1, 1) = GetS0(ip.Point2D().X(), i);
			H(2*i, 2) = H(2*i-1, 1);
		}
		massmatrix += (0.5*GetElementLength()*rhoA*ip.Weight()) * (H * H.GetTp());
	}
	m = massmatrix;

	//UO() << "Mass_ANCFBeamBE2D = " << m << "\n";
};

void ANCFBeamBE2D::GetH(Matrix& H)
{
	if (Hmatrix.Getrows() == SOS())
	{
		H = Hmatrix;
		return;
	}
	else
	{
		double A = this->GetBeam2DProperties().BeamRhoA() / this->GetBeam2DProperties().Density();

		H.SetSize(SOS(), Dim());
		H.SetAll(0);

		for (IntegrationPointsIterator ip(integrationRuleLoad); !ip.IsEnd(); ++ip)
		{
			double x = ip.Point2D().X();

			// jacobi determinant
			//Vector2D rx0 = GetRefPosx2D(x);
			double det = 0.5*GetElementLength(); //*rx0.Norm();

			double d = A * det * ip.Weight();

			for (int i = 1; i <= NS(); i++)
			{
				double s = GetS0(x, i);
				H(2*i-1, 1) += d * s;
				H(2*i,   2) += d * s; //H(2*i-1, 1);
			}	
		}
		Hmatrix = H;
	}
};

void ANCFBeamBE2D::GetdPosdqT(const Vector2D& p_loc, Matrix& dpdqi)
{
	//p = S(p.x,p.y,p.z)*q; d(p)/dq
	dpdqi.SetSize(SOS(),Dim());
	dpdqi.FillWithZeros();
	//d = S + ...
	for (int i = 1; i <= NS(); i++)
	{
		double s = GetS0(p_loc.X(), i);
		dpdqi((i-1)*Dim()+1,1) = s;
		dpdqi((i-1)*Dim()+2,2) = s;
	}
	if (p_loc.Y() != 0)
	{
		double y = p_loc.Y();
		Vector2D rx = GetPosx2D(p_loc.X(), TCD_compute);
		Vector2D n(-rx.X(), rx.Y());
		n /= rx.Norm();

		for (int i = 1; i <= NS(); i++)
		{
			double sx = GetS0x(p_loc.X(), i) * 2./GetElementLength();
			//y/|n|*dn/dq
			dpdqi((i-1)*Dim()+1,2) +=  y*sx;
			dpdqi((i-1)*Dim()+2,1) += -y*sx;

			//y*n/|n|*(r_x1*S_x1 + r_x2*S_x2)
			dpdqi((i-1)*Dim()+1,1) +=  y*n.X()*(rx.X()*sx);
			dpdqi((i-1)*Dim()+1,2) +=  y*n.Y()*(rx.X()*sx);
			dpdqi((i-1)*Dim()+2,1) +=  y*n.X()*(rx.Y()*sx);
			dpdqi((i-1)*Dim()+2,2) +=  y*n.Y()*(rx.Y()*sx);
		}
	}
};

//$ PG 2014-6-24: added for application of moments or rotational constraints (analogous to ANCFCable2D, but scaled with Lx/2)
void ANCFBeamBE2D::GetdAngle2DdqT(const Vector2D& ploc, Matrix& d)
{
	d.SetSize(NS()*Dim(),1);
	d.FillWithZeros();

	Vector2D rx = GetPosx2D(ploc.X(), TCD_compute);
	double rx2 = rx.Norm2()*GetElementLength()/2;

	for (int i = 1; i <= NS(); i++)
	{
		// d theta/d q = -rx.y*Sx.x + rx.x*Sx.y / (rx.x^2+rx.y^2)
		d((i-1)*Dim()+1,1) =-(rx.Y()*GetS0x(ploc,i))/rx2;  // Sx.x
		d((i-1)*Dim()+2,1) = (rx.X()*GetS0x(ploc,i))/rx2;  // Sx.y
	}
}

Vector2D ANCFBeamBE2D::GetE22D(const double& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D rx = GetPosx2D(p_loc, flag);
	rx.Normalize();
	return Vector2D(-rx(2), rx(1));
};

Vector2D ANCFBeamBE2D::GetE2P2D(const double& p_loc) const
{
	Vector2D rx = GetPosx2D(p_loc, TCD_compute);
	Vector2D vx = GetVelx2D(p_loc);
	
	double rxn = rx.Norm();
	double rxnp = 1./rxn*(rx(1)*vx(1) + rx(2)*vx(2));
	return 1./rxn * (Vector2D(-vx(2),vx(1)) - rxnp/rxn * Vector2D(-rx(2),rx(1)));
};

/*Vector2D ANCFBeamBE2D::GetRefInplaneUnitVector2D(const double& p_loc) const
{
	Vector2D rx0 = GetPosx2D_dc(p_loc, TCD_reference_configuration);
	rx0.Normalize();
	return Vector2D(-rx0(2), rx0(1));
};*/

double ANCFBeamBE2D::GetGamma1(const double& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D rx = GetPosx2D(p_loc, flag);
	Vector2D rx0 = GetPosx2D(p_loc, TCD_reference_configuration);
	return rx.Norm() - rx0.Norm();
};

double ANCFBeamBE2D::GetKappa(const double& p_loc, TComputeDrawInitFlag flag) const
{
	Vector2D rx = GetPosx2D(p_loc, flag);
	Vector2D rxx = GetPosxx2D(p_loc, flag);
	
	Vector2D rx0 = GetPosx2D(p_loc, TCD_reference_configuration);
	Vector2D rxx0 = GetPosxx2D(p_loc, TCD_reference_configuration);

	return (rx.Cross(rxx))/rx.Norm2() - (rx0.Cross(rxx0))/rx0.Norm2();
};

void ANCFBeamBE2D::GetDeltaGamma1(const double& p_loc, Vector& delta_eps)
{
	Vector2D rx = GetPosx2D(p_loc, TCD_compute);
	double d = 1./rx.Norm();
	for (int j = 1; j <= NS(); j++)
	{
		double sx = GetS0x(Vector2D(p_loc,0), j) * 2./GetElementLength();
		delta_eps(2*j-1) = d * rx(1) * sx;
		delta_eps(2*j  ) = d * rx(2) * sx;
	}
};

void ANCFBeamBE2D::GetDeltaKappa(const double& p_loc, Vector& delta_kappa)
{
	Vector2D rx = GetPosx2D(p_loc, TCD_compute);
	Vector2D rxx = GetPosxx2D(p_loc, TCD_compute);
	double rxn2 = rx.Norm2();
	double kappa = (rx.Cross(rxx))/rxn2;

	//AH: compute deltakappa
	double kdux = (-2.*rx(1)*kappa + rxx(2)) / rxn2;
	double kduxx = -rx(2) / rxn2;
	double kdwx = (-2.*rx(2)*kappa - rxx(1)) / rxn2;
	double kdwxx = rx(1) / rxn2;

	for (int j = 1; j <= NS(); j++)
	{
		double sx = GetS0x(p_loc, j) * 2./GetElementLength();
		double sxx = GetS0xx(p_loc, j) * 4./(GetElementLength()*GetElementLength());
		delta_kappa(2*j-1) = (kdux * sx + kduxx * sxx);
		delta_kappa(2*j  ) = (kdwx * sx + kdwxx * sxx);
	}
};

double ANCFBeamBE2D::GetStretch1(const double& p_loc, TComputeDrawInitFlag flag) const
{
	/*Vector2D rx = (flag == TCD_compute) ? GetPosx2D(p_loc) : GetPosx2DD(p_loc);*/
	Vector2D rx = GetPosx2D(p_loc, flag);
	return rx.Norm();
}

int ANCFBeamBE2D::GetAvailableSpecialValues(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables)
{
	FiniteElementGenericBeam2D::GetAvailableSpecialValues(available_variables);
	
	available_variables.Add(ReadWriteElementDataVariableType("Internal.kinetic_energy", 1, 0., 0., "kinetic_energy", TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Internal.potential_energy", 1, 0., 0., "potential_energy", TRWElementDataRead));
	available_variables.Add(ReadWriteElementDataVariableType("Internal.total_energy", 1, 0., 0., "total_energy", TRWElementDataRead));
	return 0;
};

double ANCFBeamBE2D::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D& p_loc, TComputeDrawInitFlag flag)
{
	// displacements
	if(fvd.VariableType() == FieldVariableDescriptor::FVT_displacement)   //$ PG 2017-2-5: overwritten from base class, displacement in thin elements is constant in cross section!
		return fvd.GetComponent(GetDisplacement2D(p_loc.X(), flag));

	return FiniteElementGenericBeam2D::GetFieldVariableValue(fvd, p_loc, flag);
}

int ANCFBeamBE2D::ReadSingleElementData(ReadWriteElementDataVariableType& RWdata)
{
	if(strcmp(RWdata.variable_name.c_str(), "Internal.kinetic_energy") == 0)
	{
		RWdata.value = this->GetKineticEnergy(); 
		return 1; 
	}
	else if(strcmp(RWdata.variable_name.c_str(), "Internal.potential_energy") == 0)
	{
		RWdata.value = this->GetPotentialEnergy(); 
		return 1; 
	}
	else if(strcmp(RWdata.variable_name.c_str(), "Internal.total_energy") == 0)
	{
		RWdata.value = this->GetPotentialEnergy() + this->GetKineticEnergy(); 
		return 1; 
	}

	return ReadSingleElementDataAuto(RWdata);
};

double ANCFBeamBE2D::GetKineticEnergy()
{
	Vector xgp(SOS()), temp(SOS());
	Matrix m;
	EvalM(m, 0.);

	for (int i = 1; i <= SOS(); i++)
	{
		xgp(i) = XGP(i);
	}

	Mult(xgp, m, temp);
	double kin_energy = 0.5*temp*xgp;
	return kin_energy;
};

double ANCFBeamBE2D::GetPotentialEnergy()
{
	double EI = GetBeam2DProperties().BeamEIy();
	double EA = GetBeam2DProperties().BeamEA();

	double energy_bending = 0.;
	double energy_extension = 0.;

	double total_energy;

	double length = GetElementLength();

	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
	{
		double x = ip.Point2D().X();
		double eps = GetGamma1(x, TCD_compute);
		double kappa = GetKappa(x, TCD_compute);
		energy_bending += 0.5*(EI*kappa)*kappa * (ip.Weight() * length*0.5);
		energy_extension += 0.5*(EA*eps)*eps * (ip.Weight() * length*0.5);
	}	
	total_energy = energy_bending + energy_extension;

	Vector f(SOS());
	f.SetAll(0.);
	EvalF2(f, mbs->GetTime());
	
	double strain = 0.;
	for (int i = 1; i <= SOS(); i++)
	{
		strain -= 0.5*f(i) * XG(i);
	}
	return total_energy - strain;
};

#pragma region shapefunction routines
double ANCFBeamBE2D::GetS0(const Vector2D& p_loc, int shape) const
{
	double x = p_loc.X();
	if (ns == 4)
	{
		switch (shape)
		{
			case 1: return 0.25 *( 2.	- 3.*x				+ 1.*x*x*x);
			case 2: return 0.25	*( 1.	- 1.*x	- 1.*x*x	+ 1.*x*x*x) * GetElementLength() * 0.5;		
			case 3: return 0.25 *( 2.	+ 3.*x				- 1.*x*x*x);		
			case 4: return 0.25	*(-1.	- 1.*x	+ 1.*x*x	+ 1.*x*x*x) * GetElementLength() * 0.5;				
			default: mbs->UO() << "ANCFBeamBE2D::GetS0 Error: Shape function " << shape << " not available!\n"; return 0;
		}
	}
	else
	{
		switch (shape)
		{
			case 1: return 0.125*( 4.	- 7.5*x				+ 5.*x*x*x					- 1.5*x*x*x*x*x);
			case 2: return 0.125*( 2.5	- 3.5*x - 3.*x*x	+ 5.*x*x*x	+ 0.5*x*x*x*x	- 1.5*x*x*x*x*x);		
			case 3: return 0.125*( 0.5	- 0.5*x	- 1.*x*x	+ 1.*x*x*x	+ 0.5*x*x*x*x	- 0.5*x*x*x*x*x);		
			case 4: return 0.125*( 4.	+ 7.5*x				- 5.*x*x*x					+ 1.5*x*x*x*x*x);				
			case 5: return 0.125*(-2.5	- 3.5*x	+ 3.*x*x	+ 5.*x*x*x	- 0.5*x*x*x*x	- 1.5*x*x*x*x*x);	
			case 6: return 0.125*( 0.5	+ 0.5*x	- 1.*x*x	- 1.*x*x*x	+ 0.5*x*x*x*x	+ 0.5*x*x*x*x*x);
			default: mbs->UO() << "ANCFBeamBE2D::GetS0 Error: Shape function " << shape << " not available!\n"; return 0;
		}
	}
};

double ANCFBeamBE2D::GetS0x(const Vector2D& p_loc, int shape) const
{
	double x = p_loc.X();

	if (ns == 4)
	{
		switch (shape)
		{
		case 1: return 0.25 *(-3.			+ 3.*x*x);
		case 2: return 0.25	*(-1.	- 2.*x	+ 3.*x*x) * GetElementLength() * 0.5;		
		case 3: return 0.25 *( 3.			- 3.*x*x);		
		case 4: return 0.25	*(-1.	+ 2.*x	+ 3.*x*x) * GetElementLength() * 0.5;				
		default: mbs->UO() << "ANCFBeamBE2D::GetS0x Error: Shape function " << shape << " not available!\n"; return 0;
		}
	}
	else
	{
		switch (shape)
		{
		case 1:	return 0.125*(-7.5			+ 15.*x*x				- 7.5*x*x*x*x);
		case 2:	return 0.125*(-3.5	- 6.*x	+ 15.*x*x	+ 2.*x*x*x	- 7.5*x*x*x*x);	
		case 3:	return 0.125*(-0.5	- 2.*x	+  3.*x*x	+ 2.*x*x*x	- 2.5*x*x*x*x);
		case 4:	return 0.125*( 7.5			- 15.*x*x				+ 7.5*x*x*x*x);
		case 5:	return 0.125*(-3.5	+ 6.*x	+ 15.*x*x	- 2.*x*x*x	- 7.5*x*x*x*x);
		case 6:	return 0.125*( 0.5	- 2.*x	-  3.*x*x	+ 2.*x*x*x	+ 2.5*x*x*x*x);
		default: mbs->UO() << "ANCFBeamBE2D::GetS0x Error: Shape function " << shape << " not available!\n";	return 0;
		}
	}
};

double ANCFBeamBE2D::GetS0xx(const Vector2D& p_loc, int shape) const
{
	double x = p_loc.X();
	if (ns == 4)
	{
		switch (shape)
		{
		case 1: return 1.5*(		  1.*x);
		case 2: return 0.5*(-1.		+ 3.*x) * GetElementLength() * 0.5;		
		case 3: return 1.5*(		- 1.*x);		
		case 4: return 0.5*( 1.		+ 3.*x) * GetElementLength() * 0.5;				
		default: mbs->UO() << "ANCFBeamBE2D::GetS0xx Error: Shape function " << shape << " not available!\n"; return 0;
		}
	}
	else
	{
		switch (shape)
		{
		case 1: return 0.25*(		+ 15.*x				- 15.*x*x*x);
		case 2:	return 0.25*(-3.	+ 15.*x	+ 3.*x*x	- 15.*x*x*x);
		case 3:	return 0.25*(-1.	+  3.*x	+ 3.*x*x	-  5.*x*x*x);
		case 4:	return 0.25*(		- 15.*x				+ 15.*x*x*x);
		case 5:	return 0.25*( 3.	+ 15.*x	- 3.*x*x	- 15.*x*x*x);
		case 6:	return 0.25*(-1.	-  3.*x	+ 3.*x*x	+  5.*x*x*x);
		default: mbs->UO() << "ANCFBeamBE2D::GetS0xx Error: Shape function " << shape << " not available!\n";	return 0;
		}
	}
};
#pragma endregion

void ANCFBeamBE2D::DefineIntegrationRule(IntegrationRule& integrationRule)
{
	assert(integrationRule.settings.elementType == TFE_ThinBeam2D);

	int ruleOrder = 0;

	if (integrationRule.settings.integratedValueType == IntegrationRule::IVT_Load)
		ruleOrder = 5;
	else if (integrationRule.settings.integratedValueType == IntegrationRule::IVT_Mass)
		ruleOrder = 6;  // same as in Cable2D
	//ruleOrder = 4;
	else
		ruleOrder = 10;

	IntegrationRule::DefineIntegrationRuleLine(integrationRule, ruleOrder);
};