//#**************************************************************
//# filename:             FiniteElementGenericBeam2D.h
//#
//# author:               Gerstmayr, Vetyukov
//#
//# generated:						
//# description:          
//# comments:
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 

#pragma once
#include "FiniteElementGeneric.h"
#include "XGProvider.h"


class FiniteElementGenericBeam2D : public FiniteElementGeneric<Body2D> //$EDC$[beginclass,classname=FiniteElementGenericBeam2D]
{
public:
enum {MAX_PLAST_GRIDP_Y = 50}; // maximal number of plastic grid points in y direction
enum {MAX_PLASTIC_QUANTITIES = 8}; // maximal number of plastic quantities
enum {MAX_IP = 10}; // maximum number of integration points in one direction
enum {MAX_NS = 6}; // maximum number of shape functions
enum {DIM = 2};

	FiniteElementGenericBeam2D(MBS* mbs) : FiniteElementGeneric<Body2D>(mbs), plasticip_x(0), plasticip_y(0) 
	{
		this->geometricNonlinearityStatus = GNS_NonlinearLargeStrain;

		//xgReferenceState.SetLen(FlexDOF());
		//xgInit.SetXGProvider(this, &xgReferenceState);
		//xgCompute.SetXGProvider(this, &xgReferenceState);
		//xgDraw.SetXGProvider(this, &xgReferenceState);
	}	// default constructor
	FiniteElementGenericBeam2D(const FiniteElementGenericBeam2D& e) : FiniteElementGeneric<Body2D>(e.mbs) 
	{	
		//xgReferenceState.SetLen(FlexDOF());
		//xgInit.SetXGProvider(this, &xgReferenceState);
		//xgCompute.SetXGProvider(this, &xgReferenceState);
		//xgDraw.SetXGProvider(this, &xgReferenceState);

		CopyFrom(e);	
	}
	
	~FiniteElementGenericBeam2D(void);

	virtual void ElementDefaultConstructor() = 0;
	
	void SetFiniteElementGenericBeam2D(
		int bodyind,
		const Vector& xc1, 
		const Vector& xc2, 
		int n1, 
		int n2, 
		int material_num,
		const Vector3D& size,
		const Vector3D& color);

	virtual void CopyFrom(const Element& e);
	virtual void PreAssemble();

	virtual mystr GetElementSpecification() //$EDC$[funcaccess,EDCvarname="element_type",tooltiptext="specification of element type. Once the element is added to the mbs, you MUST NOT change this type anymore!"]
	{
		mystr type = GetElementSpec();	//$ DR this is necessary for the skript language
		return type;
	}	

	virtual const Beam2DProperties& GetBeam2DProperties() const { return static_cast<const Beam2DProperties&>(GetMBS()->GetMaterial(materialnum)); }
	virtual int CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
	{
		if (dynamic_cast<Beam2DProperties*>(GetMBS()->GetMaterialPtr(materialnum)) == NULL)
		{
			errorstr += "Material of wrong type, Beam2DProperties is required.\n";
			return 2;
		}

		return Body2D::CheckConsistency(errorstr);
	}

	virtual double GetElementLength() const { return size.X(); }
	virtual int Dim() const { return 2; }
	virtual Vector3D GetDOFDirD(int idof) const = 0;

	virtual int UseSparseM() const { return 0; }

	virtual double GetStretch1(const double& p_loc, TComputeDrawInitFlag flag) const = 0;  // length of rx
	virtual double GetKappa(const double& p_loc, TComputeDrawInitFlag flag) const = 0;   // curvatoric strain measure
	virtual double GetGamma1(const double& p_loc, TComputeDrawInitFlag flag) const = 0;  // axial strain measure
	virtual double GetGamma2(const double& p_loc, TComputeDrawInitFlag flag) const {return 0.;} // shear strain measure
	virtual double GetNormalForce(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetBendingMoment(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual double GetShearForce(const double& p_loc, TComputeDrawInitFlag flag) const;

	virtual void ComputeStrainMatrix2D(Vector2D ploc, Matrix2D& strain, TComputeDrawInitFlag flag) const;

	virtual double GetJacInvDS(const Vector3D& p_loc, Matrix& jacinvDS) const { assert(0); return 0; }


	// unit vector in direction of cross-section
	virtual Vector2D GetE22D(const double& p_loc, TComputeDrawInitFlag flag) const = 0;
	virtual Vector2D GetE2P2D(const double& p_loc) const = 0;





#pragma region position and displacement 

// with one dimensional local position (in axial direction) 'p_loc' in interval (-1,1)  

	virtual Vector2D GetPos2D(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetPosx2D(const double& p_loc, TComputeDrawInitFlag flag) const;
	virtual Vector2D GetPosxx2D(const double& p_loc, TComputeDrawInitFlag flag) const;
	
	virtual Vector2D GetDisplacement2D(const double& p_loc, TComputeDrawInitFlag flag) const
	{
		return GetPos2D(p_loc, flag) - GetPos2D(p_loc, TCD_reference_configuration);
	}
	
	virtual Vector2D GetVel2D(const double& p_loc) const;
	virtual Vector2D GetVelx2D(const double& p_loc) const;

	
// with two dimensional local position 'p_loc' in interval (-1,1) x (-1,1) (used for drawing purpose only, may be overwritten in derived (particularly shear deformable) beam classes, computation therein)
	virtual Vector2D GetPos2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const
	{
		assert (flag == TComputeDrawInitFlag::TCD_draw || flag == TComputeDrawInitFlag::TCD_draw_magnified);
		return GetPos2D(p_loc.X(), flag)	+	0.5*GetBeam2DProperties().GetBeamThicknessY()*p_loc.Y()*GetE22D(p_loc.X(), flag);
	}
	
	//$ PG 2014-10-15: [ called by drawing routines of loads, sensors, etc.
	virtual Vector2D GetPos2DD(const Vector2D& p_loc)	const	{	return GetPos2D(p_loc, TCD_draw);	}
	virtual Vector2D GetPos2DD(const Vector2D& p_loc, double def_scale) const {	return GetPos2D(p_loc, TCD_draw_magnified); }
	//$ PG 2014-10-15: ]
	
	// AH: this method is called by mbsload drawing routine, implementation in rigid2d returns zero in case of displacement dofs
	virtual Vector3D GetRefPosD() const	{	return ToP3D(GetPos2D(0., TCD_draw)); }
	
	// AH: Needed for drawing loads
	virtual Vector2D GetRefPos2DD(const Vector2D& p_loc) const	{	return GetPos2D(p_loc, TCD_reference_configuration);	}
	
	virtual Vector2D GetDisplacement2D(const Vector2D& p_loc, TComputeDrawInitFlag flag) const		
	{	
		return GetPos2D(p_loc, flag)	- GetPos2D(p_loc, TCD_reference_configuration);	
	}
#pragma endregion


	// keep here for now, since q_ref still needs to be implemented in element
	virtual const double XG_dc(int iloc, TComputeDrawInitFlag flag) const 
	{
		switch (flag)
		{
			case TCD_compute:
				return q0(iloc) + GetXact(ltg.Get(iloc));
			case TCD_draw:
				return q0(iloc) + GetDrawValue(ltg.Get(iloc));
			case TCD_draw_magnified:
			{
				double scaling = mbs->GetDOption(105);
				return q0(iloc) + scaling*GetDrawValue(ltg.Get(iloc));
			}
			case TCD_reference_configuration:
				return q0(iloc);
			case TCD_initial_values:
				return q0(iloc) + GetXInit(iloc);
			case TCD_cached:
				return xg_cached(iloc);			
			default: 
			{
				assert("XG_dc called with illegal flag" && 0);
				return 0.;
			}
		}
	}

	virtual const double XGP_dc(int iloc, TComputeDrawInitFlag flag) const 
	{
		return XG_dc(iloc+SOS(), flag);
	}
			
#pragma region shape-functions
	// TODO AH: the shape functions are so different, so do we really want to declare them here? probably not... the problem, however, is that
	// the functions getdisplacement, getdisplacementx, etc. rely on them
	virtual double GetS0(const Vector2D& p_loc, int shape) const = 0;
	virtual double GetS0x(const Vector2D& p_loc, int shape) const = 0;
	virtual double GetS0xx(const Vector2D& p_loc, int shape) const = 0;
#pragma endregion

	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, TComputeDrawInitFlag flag);
	//$ PG 2014-10-8: [ old call, hook to new one
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, bool flagD)
	{
		return GetFieldVariableValue(fvd, local_position, flagD ? TCD_draw : TCD_compute);
	}
	//$ PG 2014-10-8: alex
	//virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector2D & local_position, bool flagD, TComputeDrawInitFlag flag = TCD_compute);
	//$ PG 2014-10-8: ]

	virtual void DrawElement();

	
#pragma region plasticity
	virtual void SetPlasticStrainGridsize(int ip_x, int ip_y) {
		plasticip_x = ip_x; plasticip_y = ip_y; 
		Vector datainit(DataS()); //automatically initialized with zeros
		SetDataInit(datainit); //initialize data variables with zero = initial inelastic strain == zero
	}
	// get plastic variables at gridpoint
	virtual void DataToPlasticVariables(Vector & plastic_variables, int ip_x, int ip_y) const;
	virtual void DataToPlasticVariablesLastTimeStep(Vector& plastic_variables, int ip_x, int ip_y) const;
	virtual void DataToPlasticVariablesD(Vector& plastic_variables, int ip_x, int ip_y) const;
	// interpolate plastic variables to ploc
	virtual void DataToPlasticVariables(Vector& plastic_variables, Vector2D& ploc);
	virtual void DataToPlasticVariablesD(Vector& plastic_variables, const Vector2D ploc);
	// interpolate plastic strain only, not hardening parameter and yield function
	virtual void DataToPlasticStrain(Vector& plasticstrain, Vector2D& ploc);
	// set plastic variables to internal XData
	virtual void PlasticVariablesToData(const Vector & plastic_variables, int ip_x, int ip_y);
	virtual void PlasticVariablesToDataD(const Vector & plastic_variables, int ip_x, int ip_y);
	// stored are maximal 5 quantities per gridpoint (like ANCFThinPlate3D):
	// in case of 1D-Plasticity: 
	// plasticstrain_xx, hardeningparameter, actual value of yield function
	// plasticstrain_yy and plasticstrain_xy are always zero
	// in case of 2D-Plasticity:
	// plasticstrain_xx, plasticstrain_yy, plasticstrain_xy,
	// hardeningparameter, actual value of yield function
	virtual int DataS() const { return NPlasticParameters() * NPlasticGridpoints(); }
	virtual int NPlasticParameters() const 
	{ 
		if (DimensionOfPlasticMaterialLaw()==1) { return 3; } 
		else if (DimensionOfPlasticMaterialLaw()==2) { return 5; } 
		else return 0;
	}
	virtual int NPlasticGridpoints() const { return plasticip_x * plasticip_y; }
	// dimension of plastic material law:
	// return 1 for 1D-Plasticity as in cable element
	// return 2 for 2D-Plasticity as in shear-deformable beam
	virtual int DimensionOfPlasticMaterialLaw() const { return 1;};

	// link XData corresponding to the plastic quantity quantitynr with Matrix mat
	// 1 .. epsp_xx
	// 2 .. epsp_yy
	// 3 .. epsp_xy
	// 4 .. hardeningparameter
	// 5 .. yield function
	virtual void GetPlasticQuantityMatrix(int quantitynr, const Matrix& mat, TComputeDrawInitFlag flag) const;
	virtual void GetPlasticQuantityMatrix(int quantitynr, Matrix& mat, TComputeDrawInitFlag flag);

	// routines for retrieving integral values of plastic strain
	// plastic axial strain = 1/h int_{-h/2}^h/2 epsp_xx dy
	virtual double GetPlasticAxialStrain(double x0, TComputeDrawInitFlag flag=TCD_compute);
	// plastic material curvature = 12/h^3 int_{-h/2}^h/2 y epsp_xx dy
	virtual double GetPlasticKappa(double x0, TComputeDrawInitFlag flag=TCD_compute);
	// plastic shear strain = 1/h int_{-h/2}^h/2 epsp_xy dy
	virtual double GetPlasticShearStrain(double x0, TComputeDrawInitFlag flag=TCD_compute);
	// plastic thickness strain = 1/h int_{-h/2}^h/2 epsp_yy dy
	virtual double GetPlasticThicknessStrain(double x0, TComputeDrawInitFlag flag=TCD_compute);

	// help routine to integrate some plastic strain over height
	// return 1/h int_{-h/2}^h/2 epsp dy
	double IntegralMean(double x0, Matrix& plasticstrain_mat);
	// help routine to evaluate some plastic strain on center line
	// return plastic strain evaluated at y=0
	double CenterLineValue(double x0, Matrix& plasticstrain_mat);

	virtual double PostNewtonStep(double t);

#pragma endregion

	virtual void GetElementDataAuto(ElementDataContainer& edc); 		
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); 

protected:

	// Each derived beam class must provide its own method for calculating an
	// element's length in reference configuration, based on its nodal positions 
	// and the respective geometrical description of the beam.
	// Numerical integration has to be performed by same rule as used in EvalF2
	// for the axial stiffness term.
	// The method (in contrast to GetElementLength()) is protected, and shall be
	// called out of a set-function (such as SetANCFBeamShearFE2DQuadratic) in 
	// order to set size.X().
	virtual double CalculateElementLength() const = 0;

	Vector q0;	// a vector for the initial positions and velocities is needed as x_init is the initial diplacement
	            // one could maybe use x_init of the node instead (see yuri's ancfthinplate3d)
	
	// number of plasticstrain-gridpoints
	int plasticip_x; //$EDC$[varaccess,EDCvarname="number_plastic_intpoints_x",EDCfolder="Geometry",tooltiptext="Number of integration points for plastic variable in x direction"]
	int plasticip_y; //$EDC$[varaccess,EDCvarname="number_plastic_intpoints_y",EDCfolder="Geometry",tooltiptext="Number of integration points for plastic variable in y direction"]

	// EDC access for elnum
	//EDC	int elnum; //$EDC$[varaccess,EDCvarname="element_number",EDCfolder="",readonly,tooltiptext="number of the element in the mbs"] //number in MBS-system

	// EDC access for name of the element
	//EDC	mystr elementname; //$EDC$[varaccess,EDCvarname="name",EDCfolder="",tooltiptext="name of the element"]

	// EDC access for nodes array
	//EDC TArray<int> nodes; //$EDC$[varaccess,EDCvarname="node_numbers",EDCfolder="Geometry",tooltiptext="Global node numbers"]

	// EDC access for mat_num
	//EDC int materialnum; //$EDC$[varaccess,EDCvarname="material_number",EDCfolder="Physics",tooltiptext="material number which contains the main material properties of the body or element"] //material number in MBS which contains material information

	//EDC Vector3D col; //$EDC$[varaccess,EDCvarname="RGB_color",EDCfolder="Graphics",tooltiptext="[red, green, blue] color of element, range = 0..1, use default color:[-1,-1,-1]"] //color for body

	// EDC access for loads
	//EDC	TArray<int> loads;  //$EDC$[varaccess,EDCvarname="loads",variable_length_vector,condition=!IsType(TConstraint),tooltiptext="Set loads attached to this element: 'nr_load1, nr_load2, ...' or empty"]

	// EDC access for sensors
	//EDC	TArray<int> sensors; //$EDC$[varaccess,EDCvarname="sensors",EDCfolder="Info",variable_length_vector,tooltiptext="attached sensors",readonly]//if the element equations of motion are dependent on sensor values (e.g. as an input), then add the sensor numbers here

	//Vector xgReferenceState;
	//XGProviderInit xgInit;					// initial positions and velocities
	//XGProviderCompute xgCompute;		// "compute" positions and velocities
	//XGProviderDraw xgDraw;					// "drawing" positions and velocities

	// for XG_dc
	Vector /*q_ref,*/ xg_cached;

}; //$EDC$[endclass,FiniteElementGenericBeam2D]
