//#**************************************************************
//#
//# filename:             iofmu.cpp
//#
//# author:               Dorninger Alexander
//#
//# generated:            November 2016
//# description:          An IO-element to load FMUs (functional mockup units)
//# remarks:				
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


//#include "mbs_interface.h"

#include "element.h"
#include "constraint.h"
#include "sensors.h"
#include "control.h"
#include "solversettings_auto.h"
#include "elementdataaccess.h"
#include "graphicsconstants.h"
#include "rendercontext.h"

#include "fmilib.h"
#include "IOFMU.h"
#include <JM/jm_portability.h>

// FMU/FMI specific code snippets taken from FMI_import_test.c ,...
#define BUFFER 1000

void importlogger(jm_callbacks* c, jm_string module, jm_log_level_enu_t log_level, jm_string message)
{
        printf("module = %s, log level = %d: %s\n", module, log_level, message);
}
static void fmi2logger(fmi2_component_environment_t env, fmi2_string_t instanceName, fmi2_status_t status, fmi2_string_t category, fmi2_string_t message, ...)
{
#ifdef _WIN64
	printf("FMU can not be used with 64 bit (yet)");
	return;
#else // win32

    int len;
	char msg[BUFFER];
	va_list argp;	
	va_start(argp, message);
	//len = jm_vsnprintf(msg, BUFFER, message, argp);
	printf("fmiStatus = %s;  %s (%s): %s\n", fmi2_status_to_string(status), instanceName, category, msg);
#endif
}
static void stepFinished(fmi2_component_environment_t env, fmi2_status_t status) {
#ifdef _WIN64
	printf("FMU can not be used with 64 bit (yet)");
	return;
#else // win32
	printf("stepFinished is called wiht fmiStatus = %s\n", fmi2_status_to_string(status));
#endif
}



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
const char* IOFMU::SymbolText() const 
{
	return "FMU";
}

// import the FMU from the file specified in source
int3 IOFMU::ImportFromFMU()
{
#ifdef _WIN64
	mbs->UO(UO_LVL_err) << "FMU can not be used with 64 bit (yet)";
	return int3(0,0,0);
#else // win32

	edcFMU.Reset();
	mystr unzippath = this->GetMBS()->GetTOption(126); // use sensor output path as temp path to unzip fmu

	jm_callbacks callbacks = Callbacks();

// general information at this level
	fmi_import_context_t* context = fmi_import_allocate_context(&callbacks);
	version = fmi_import_get_fmi_version(context, this->sourcefile.c_str(), unzippath.c_str());
	ElementData ed;
	ed.SetText(sourcefile,"SourceFile"); ed.SetLocked(1); edcFMU.Add(ed);
	ed.SetText(mystr(version),"FMUVersion"); ed.SetLocked(1); edcFMU.Add(ed);
	mbs->UO(UO_LVL_dbg1) << "identified FMU version" << version;
	
	int3 rv;
// process further depending on fmu version
	if(version == fmi_version_2_0_enu) 
	{
		rv = ImportFMUV2(context, unzippath);	
	}
	else if (version == fmi_version_1_enu)
	{
		mbs->UO(UO_LVL_err) << "Only FMI version 2.0 are supported so far\n" ; // not planned in first step
	}
	else
	{
		mbs->UO(UO_LVL_err) << "Only FMI version 2.0 are supported so far\n" ; // genreal error msg
	}
	
	fmi_import_free_context(context);
	
	return rv;
#endif
}

// FMU v2 swpecific import routine
int3 IOFMU::ImportFMUV2(fmi_import_context_t* context, mystr unzippath)
{
#ifdef _WIN64
	mbs->UO(UO_LVL_err) << "FMU can not be used with 64 bit (yet)";
	return int3(0,0,0);
#else // win32

	jm_status_enu_t status;

// parse the interface from XML
	fmi2_callback_functions_t callBackFunctions = CallbackFunctions();
	fmi2_import_t* fmu = fmi2_import_parse_xml(context, unzippath.c_str(), 0);
	if(!fmu) { mbs->UO(UO_LVL_err) << "Error parsing XML\n"; return int3(0,0,0); }

	const char* modelName = fmi2_import_get_model_name(fmu);
	const char* GUID = fmi2_import_get_GUID(fmu);

	ElementData ed;
	ed.SetText(modelName,"ModelName"); ed.SetLocked(1); edcFMU.Add(ed);
	ed.SetText(GUID,"ModelGUID"); ed.SetLocked(1); edcFMU.Add(ed);
	mbs->UO(UO_LVL_dbg1) << "Model name: " << modelName << "\n";
	mbs->UO(UO_LVL_dbg1) << "Model GUID: " << GUID << "\n";

////  fmi2_fmu_kind_unknown = 0,
////  fmi2_fmu_kind_me = 1,
////  fmi2_fmu_kind_cs = 2,
////  fmi2_fmu_kind_me_and_cs = 3
	if(fmi2_import_get_fmu_kind(fmu) & fmi2_fmu_kind_me)
	{
		const char* modelIdentifier = fmi2_import_get_model_identifier_ME(fmu);
		ed.SetText("Model Exchange","kind"); ed.SetLocked(1); edcFMU.Add(ed);
		ed.SetText(modelIdentifier,"ModelIdentifier"); ed.SetLocked(1); edcFMU.Add(ed);
		mbs->UO(UO_LVL_dbg1) << "Model identified for ModelExchange: " << modelIdentifier <<"\n";
		fmukind = fmi2_fmu_kind_me;
		
	} 
	else if(fmi2_import_get_fmu_kind(fmu) & fmi2_fmu_kind_cs)
	{
		const char* modelIdentifier = fmi2_import_get_model_identifier_CS(fmu);
		ed.SetText("CoSimulation","kind"); ed.SetLocked(1); edcFMU.Add(ed);
		ed.SetText(modelIdentifier,"ModelIdentifier"); ed.SetLocked(1); edcFMU.Add(ed);
		mbs->UO(UO_LVL_dbg1) << "Model identified for CoSimulation: " << modelIdentifier <<"\n";
		fmukind = fmi2_fmu_kind_cs;
	}
	else { mbs->UO(UO_LVL_err) << "Unxepected FMU kind\n"; return int3(0,0,0); }


	fmi2_import_variable_list_t* allVariables = fmi2_import_get_variable_list(fmu,2); // 2nd parameter sortingorder = 2 (by variable refernece)
	int nVariables = fmi2_import_get_variable_list_size(allVariables);

	for(int i=0; i<nVariables; i++)
	{
		mystr subTreeName = mystr("Variable_") + mystr(i);

		// variable name
		fmi2_import_variable_t* pVar = fmi2_import_get_variable(allVariables,i);
		const char* pVarName = fmi2_import_get_variable_name(pVar);
		ed.SetText(pVarName,"VariableName"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);

		// variable data type
		fmi2_import_variable_typedef_t* pVarTypeDecl = fmi2_import_get_variable_declared_type(pVar);
		fmi2_base_type_enu_t varTypeBase = fmi2_import_get_variable_base_type(pVar);
		if (varTypeBase == fmi2_base_type_real)
		{
			fmi2_import_real_variable_t* pVarReal = fmi2_import_get_variable_as_real(pVar);
			fmi2_real_t init = fmi2_import_get_real_variable_start(pVarReal);
			ed.SetText("real","VariableType"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			ed.SetDouble(init,"InitialValue"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
		else if (varTypeBase == fmi2_base_type_int)
		{
			fmi2_import_integer_variable_t* pVarInt = fmi2_import_get_variable_as_integer(pVar);
			fmi2_integer_t init = fmi2_import_get_integer_variable_start(pVarInt);
			ed.SetText("integer","VariableType"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			ed.SetInt(init,"InitialValue"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
		else if(varTypeBase == fmi2_base_type_bool)
		{
			fmi2_import_bool_variable_t* pVarInt = fmi2_import_get_variable_as_boolean(pVar);
			fmi2_boolean_t init = fmi2_import_get_boolean_variable_start(pVarInt);
			ed.SetText("boolean","VariableType"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			ed.SetBool(init,"InitialValue"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
		else if(varTypeBase == fmi2_base_type_str)
		{
			fmi2_import_string_variable_t* pVarInt = fmi2_import_get_variable_as_string(pVar);
			fmi2_string_t init = fmi2_import_get_string_variable_start(pVarInt);
			ed.SetText("string","VariableType"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			ed.SetText(init,"InitialValue"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
		else
		{
			ed.SetText("Not Implemented","VariableType"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);		
		}

		// variable causality
		fmi2_causality_enu_t varCausality = fmi2_import_get_causality(pVar);
		if(varCausality == fmi2_causality_enu_input)
		{
			ed.SetText("input","Causality"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			inputnames.Add(pVarName);
			inputVarNrs.Add(i);
		}
		else if(varCausality == fmi2_causality_enu_output)
		{
			ed.SetText("output","Causality"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			outputnames.Add(pVarName);
			outputVarNrs.Add(i);
		}
        else if(varCausality == fmi2_causality_enu_parameter)
		{
			ed.SetText("parameter","Causality"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
			parameternames.Add(pVarName);
			parameterVarNrs.Add(i);
		}
		else if(varCausality == fmi2_causality_enu_local)
		{
			ed.SetText("local","Causality"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
		else
		{
			ed.SetText("Not Implemented","Causality"); ed.SetLocked(1); edcFMU.TreeAdd(subTreeName,ed);
		}
	}
	fmi2_import_free_variable_list(allVariables);

// set default values for IOBlock 
	int nInputsAndParameters = inputVarNrs.Length()+parameterVarNrs.Length();
	inputs = NaturalNumbers(nInputsAndParameters);							// default input element numbers 1..n
	input_types.SetLen(nInputsAndParameters); input_types.SetAll(1);		// default input element types all to IOElement
	input_localnum.SetLen(nInputsAndParameters); input_localnum.SetAll(1);	// default input element types all to IOElement


	
	
	
// DLL
	status = fmi2_import_create_dllfmu(fmu, fmukind, &callBackFunctions);
	if(status == jm_status_error) {	mbs->UO(UO_LVL_err) << "Could not create the DLL loading mechanism(C-API).\n"; return int3(0,0,0); }

	mbs->UO(UO_LVL_dbg1) << "Version returned from FMU:" << fmi2_import_get_version(fmu) << "\n";

	fmi2_import_destroy_dllfmu(fmu);

	fmi2_import_free(fmu);

	return int3(0,0,0);
#endif
}

jm_callbacks IOFMU::Callbacks()
{
	jm_callbacks callBacks;

	callBacks.malloc = malloc;
	callBacks.calloc = calloc;
	callBacks.realloc = realloc;
	callBacks.free = free;
	callBacks.logger = importlogger;
	callBacks.log_level = jm_log_level_all;
	callBacks.context = 0;
	
	return callBacks;
}

fmi2_callback_functions_t IOFMU::CallbackFunctions()
{
	fmi2_callback_functions_t callBackFunctions;

	callBackFunctions.logger = fmi2logger;
	callBackFunctions.allocateMemory = calloc;
	callBackFunctions.freeMemory = free;
	callBackFunctions.stepFinished = stepFinished;
	callBackFunctions.componentEnvironment = 0;

	return callBackFunctions;
}

Element* IOFMU::GetCopy()
{
	Element* ec = new IOFMU(mbs);
	ec->CopyFrom(*this);

	return ec;
}

void IOFMU::CopyFrom(const Element& e)
{
	InputOutputElement::CopyFrom(e);
	
	const IOFMU& ce = (const IOFMU&)e;
	sourcefile = ce.sourcefile;
	version = ce.version;
	importSuccessful = ce.importSuccessful;

	fmu = NULL; // DO NOT COPY - this is crreated in the Initialize step
	initialized = 0;

	edcFMU.CopyFrom(ce.edcFMU);
}

void IOFMU::ElementDefaultConstructor()
{
	InputOutputElement::ElementDefaultConstructor();

	sourcefile = mystr("");
	version = fmi_version_unknown_enu;
	importSuccessful = 0;
	
	fmu = NULL;
	initialized = 0;

	edcFMU.Reset();
}

void IOFMU::Initialize()
{
#ifdef _WIN64
	mbs->UO(UO_LVL_err) << "FMU can not be used with 64 bit (yet)";
	return;
#else // win32

// remove any previous instance
	if(fmu!=NULL)
	{
		fmi2_import_destroy_dllfmu(fmu);
		fmi2_import_free(fmu);
	}

	jm_callbacks callBacks = Callbacks();
	fmi2_callback_functions_t callBackFunctions = CallbackFunctions();
	mystr unzippath = this->GetMBS()->GetTOption(126); // use sensor output path as temp path to unzip fmu

	fmi_import_context_t* context = fmi_import_allocate_context(&callBacks);
	//fmi_version_enu_t version = fmi_import_get_fmi_version(context, this->sourcefile.c_str(), unzippath.c_str())


	fmi_import_free_context(context); // ??? is this allowed right now  - otherweise make this also a member variable ???

	jm_status_enu_t status;
	//status = 
#endif
}


int IOFMU::CheckConsistency(mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = InputOutputElement::CheckConsistency(errorstr);
	if(rv==2){return rv;}

// TODO: class specific

	return rv;
}

void IOFMU::GetElementData(ElementDataContainer &edc)
{
	InputOutputElement::GetElementData(edc);

	ElementData ed;
	ed.SetText(sourcefile,"SourceFile"); ed.SetLocked(1); edc.Add(ed); // MUST BE same name as in CallCompFunction 40 
	ed.SetEDC(&edcFMU,"FMUInfo"); edc.Add(ed);

	// fill the IOBlock entries
	int nInputsAndParameters = inputVarNrs.Length()+parameterVarNrs.Length();

	mystr labels;
	for(int i=1; i<=inputVarNrs.Length(); i++)
	{
		labels += mystr("(") + mystr(i) + mystr(") ") + inputnames(i) + mystr(";  "); 
	}
	for(int i=1; i<=parameterVarNrs.Length(); i++)
	{
		labels += mystr("(") + mystr(i+inputVarNrs.Length()) + mystr(") ") + parameternames(i) + mystr(";  "); 
	}
	ed.SetText(labels,"inputIdentifiers"); ed.SetLocked(1); edc.TreeAdd("IOBlock",ed);

}

int IOFMU::SetElementData(ElementDataContainer &edc)
{
	int rv = InputOutputElement::SetElementData(edc);
	// do not read any of the FMU entries
	ElementData ed;
	return rv;
}

double IOFMU::GetOutput(double t, int i) const
{
	return 1.0;
}

void IOFMU::DrawBlockSymbol()
{
	return InputOutputElement::DrawBlockSymbol();
}