//#*****************************************************************************************
//#
//# filename:             ThermoConstraints.cpp
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          In this file all constraints for temperature elements are
//#                       collected.
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "constraint.h"
#include "FiniteElement3D.h"
#include "ThermoElementInterface.h"
#include "ThermoConstraints.h"
#include "node.h"
#include "..\UtilityLib\femathHelperFunctions.h"
#include "material.h"
#include "elementdataaccess.h"

//#############################################################
//
//               ThermalContact
//
//#############################################################
ThermalContact::ThermalContact(MBS *mbs) :
 Constraint(mbs,                      // pointer to MBS
            "ThermalContact",  // element name
       TArray<int>(),						 // element numbers
       Vector(),                 // data initialization
       Vector3D(-1.,0.,0.),      // drawing dimension
       0.,                       // spring stiffness
       0,                        // use local coordinate system
       1,                        // use penalty
			 Vector3D(0.8,0.3,0.3)),   // color
 mode(1),
 alpha(0.),
 wP_s(0.),
 f_M(0.5),
 sos(0),
 is(0),
 nr_of_master_elements(0),
 diskretization_x(1),
 diskretization_y(1),
 diskretization_z(1),
 draw_contact_points(0),
 consistent_mesh(false),
 is_initialized(true),
 slave_search_tree_initialized(true) // no elements by default, so nothing to initialize by default
{
	x_init = Vector();
}

int ThermalContact::CheckConsistency(mystr &errorstr)
{
	int rv = Constraint::CheckConsistency(errorstr);
	TArray<int> necessary_kinAccFuncs(0);

	for (int i = 1; i<= NE(); i++)
	{
		TKinematicsAccessFunctions tkaf_element = GetMBS()->GetElement(elements(i)).GetKinematicsAccessFunctions();
		bool element_suitable = (necessary_kinAccFuncs.Length() == 0);
		
		GetNecessaryKinematicAccessFunctions(necessary_kinAccFuncs, i);
		for (int k = 1; k <= necessary_kinAccFuncs.Length(); k++)
		{
			if ((tkaf_element | necessary_kinAccFuncs(k)) == tkaf_element)
			{
				element_suitable = true;
			}
		}
		if (!element_suitable)
		{
			errorstr += "Error in element " + mystr(GetOwnNum()) + " (ThermalContact): The constrained element " + mystr(elements(i));
			errorstr +=	" (" + mystr(GetMBS()->GetElement(elements(i)).GetElementSpec()) + ") is not suitable for this constraint\n";
			return max(1,rv);
		}

		if (dynamic_cast<const MaterialThermoMechanic*>(&GetMBS()->GetElementPtr(elements(i))->GetMaterial()) == NULL)
		{
			errorstr += "Material of wrong type, MaterialThermoMechanic is required.\n";
			rv = max(1, rv);
		}
	}

	if (consistent_mesh)
	{
		if (nr_of_master_elements*2 != NE())
		{
			errorstr += "Error in element " + mystr(GetOwnNum()) + " (ThermalContact): The number of slave elements must be equal the number of master elements if the flag consistent_mesh is set to true.\n";
			return max(1,rv);
		}
		// TODO: check if the mesh is really consistent
	}
	return rv;
}

void ThermalContact::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int /*numberOfKinematicPair*/)
{
	KinAccFunc.SetLen(0);
	int kaf = (int)(TKinematicsAccessFunctions(TKAF_temperature));
	KinAccFunc.Add(kaf);
}

void ThermalContact::CopyFrom(const ThermalContact &rhs)
{
	Constraint::CopyFrom(rhs);
	alpha = rhs.alpha;
	wP_s = rhs.wP_s;
	f_M = rhs.f_M;
	mode = rhs.mode;
	faces = rhs.faces;
	diskretization_x = rhs.diskretization_x;
	diskretization_y = rhs.diskretization_y;
	diskretization_z = rhs.diskretization_z;
	element_to_local_ltg_list = rhs.element_to_local_ltg_list;
	element_to_local_ltg_list_faces = rhs.element_to_local_ltg_list_faces;
	nr_of_master_elements = rhs.nr_of_master_elements;
	slave_search_tree = rhs.slave_search_tree;
	slave_search_tree_initialized = rhs.slave_search_tree_initialized;
	sos = rhs.sos;
	draw_contact_points = rhs.draw_contact_points;
	consistent_mesh = rhs.consistent_mesh;
	sos = rhs.sos;
	is = rhs.is;
	is_initialized = rhs.is_initialized;
	ltg_is = rhs.ltg_is;
	representative_elements = rhs.representative_elements;
	representative_master_local_coordinate_nrs = rhs.representative_master_local_coordinate_nrs;
	representative_slave_local_coordinate_nrs = rhs.representative_slave_local_coordinate_nrs;
}

Element* ThermalContact::GetCopy()
{
	ThermalContact *rv = new ThermalContact(*this);
	return rv;
}

const char* ThermalContact::GetElementSpec() const
{
	return "ThermalContact";
}


void ThermalContact::LinkToElements()
{
	// this is a special constraint since it mixes penalty (EvalF2) and Lagrange (EvalG and AddElementCqTLambda)
	BuildElementToLocalLTG(); //$ SW 2015-7-17: possible improvement: check if this was done before and no LinkToElementsPenalty was called.
	// in mode 2 there are lagrange parameters too
	if (mode==2)
	{
		for (int i = 1; i <= NE(); i++)
		{
			GetElem(i).AddConstraint(this,i);
		}
	}
}

int ThermalContact::SOS() const
{
	assert(is_initialized);
	return sos;
}

void ThermalContact::PreAssemble()
{
	// ensure that if the mbs is reassembled that initialize is called
	is_initialized = false;
}

void ThermalContact::Initialize()
{
	if (!is_initialized)
	{
		BuildElementToLocalLTG();
		if (consistent_mesh && (mode == 2 || mode == 3 ) )
		{
			BuildElementToLocalLTG_Faces();
			BuildRepresentingElementsList();
		}
		else
		{
			is = 0;
		}
		is_initialized = true;
	}
}

int ThermalContact::IS() const
{
	return is;
}

void ThermalContact::BuildRepresentingElementsList()
{
	assert(consistent_mesh);

	representative_elements.Flush();
	representative_master_local_coordinate_nrs.Flush();
	representative_slave_local_coordinate_nrs.Flush();

	IVector list_of_master_face_dofs;

	for (int ielem=1; ielem <= nr_of_master_elements; ielem++)
	{
		const FiniteElement3D & master = GetMasterElement(ielem);

		const FiniteElement3D & slave  = GetSlaveElement(ielem);

		FEFace masterFace = master.GetLocalFace(MasterFace(ielem));
		FEFace slaveFace = slave.GetLocalFace(SlaveFace(ielem));

		IVector permutation;
		GetNodePermutations(ielem, permutation);

		for (int i=1; i <= masterFace.NFaceNodes(); i++)
		{
			int master_node_nr = masterFace.Node(i);
			const ThermoNodeInterface& tempMasterNode = dynamic_cast<const ThermoNodeInterface &>(master.GetNode(master_node_nr));

			int global_master_index = master.GetNode(master_node_nr).LTG(tempMasterNode.GetTemperatureLocalIndex());
			int local_master_index = master.GetLTGArray().Find(global_master_index);
			
			
			assert(global_master_index == master.LTG(local_master_index));

			if (list_of_master_face_dofs.Find(global_master_index) != 0)
			{
				continue;
			}

			list_of_master_face_dofs.Add(global_master_index);

			int slave_node_nr = slaveFace.Node(permutation(i));
			const ThermoNodeInterface& tempSlaveNode = dynamic_cast<const ThermoNodeInterface &>(slave.GetNode(slave_node_nr));
			int global_slave_index = slave.GetNode(slave_node_nr).LTG(tempSlaveNode.GetTemperatureLocalIndex());

			int local_slave_index = slave.GetLTGArray().Find(global_slave_index);

			assert(global_slave_index == slave.LTG(local_slave_index));

			representative_elements.Add(ielem);
			representative_master_local_coordinate_nrs.Add(local_master_index);
			representative_slave_local_coordinate_nrs.Add(local_slave_index);
		}
	}
	if (mode == 2)
	{
		is = list_of_master_face_dofs.Length();
	}
	else
	{
		is = 0;
	}
}

void ThermalContact::AddLTG(int gi)
{
	ltg_is.Add(gi);
	// if the ltg list is already initialized we adapt the ltg list
	// otherwise this is done in the initialization
	if (is_initialized)
	{
		ltg.SetLen(2*SOS());
		ltg.Append(ltg_is);
	}
}

const IVector& ThermalContact::GetLagrangeDOFs() const
{
	return ltg_is;
}

void ThermalContact::BuildElementToLocalLTG()
{
	ltg.Flush();

	element_to_local_ltg_list.Flush();
	element_to_local_ltg_list.ReSize(NE());

	static const int positional_dof = 1;
	static const int velocity_dof = 2;

	// positional dofs = 1, velocity  dofs = 2 (first time derivative of positional dofs)
	for (int pos_or_vel=positional_dof; pos_or_vel <= velocity_dof; pos_or_vel++)
	{
		for (int k=1; k <= NE(); k++)
		{
			const FiniteElement3D & elem = static_cast<FiniteElement3D &>(GetElem(k));
			// in mode 1 only the DOFs at the contact face are constraint
			if ( mode == 1 )
			{
				FEFace face = elem.GetLocalFace(faces(k));
				for (int i=1; i <= face.NFaceNodes(); i++)
				{
					const ThermoNodeInterface& node = dynamic_cast<const ThermoNodeInterface &> (elem.GetNode(face.Node(i)));
					int global_index;
					if (pos_or_vel == positional_dof)
					{
						global_index = elem.GetNode(face.Node(i)).LTG(node.GetTemperatureLocalIndex());
					}
					else
					{
						global_index = elem.GetNode(face.Node(i)).LTG(node.GetTemperaturePLocalIndex());
					}
					int local_index = ltg.AddIfNotExists(global_index);
					element_to_local_ltg_list(k).Add(local_index);
				}
			}
			// in mode 2 all DOFs of the elements are constraint
			else if ( mode == 2 || mode == 3)
			{
				for (int i=1; i <= elem.NNodes(); i++)
				{
					const ThermoNodeInterface& node = dynamic_cast<const ThermoNodeInterface &>(elem.GetNode(i));
					int global_index;
					if (pos_or_vel == positional_dof)
					{
						global_index = elem.GetNode(i).LTG(node.GetTemperatureLocalIndex()); // first there are three positional DOF and then, the forth DOF, is the temperature constraint
					}
					else
					{
						global_index = elem.GetNode(i).LTG(node.GetTemperaturePLocalIndex()); // first there are three positional DOF and then, the forth DOF, is the temperature constraint
					}
					int local_index = ltg.AddIfNotExists(global_index);
					element_to_local_ltg_list(k).Add(local_index);
				}
			}
			else
			{
				assert (false && "Mode not implemented");
			}
		}
		if (pos_or_vel == positional_dof)
		{
			sos = ltg.Length();
		}
	}
	ltg.Append(ltg_is);
}

void ThermalContact::BuildElementToLocalLTG_Faces()
{
	if (mode == 1)
	{
		return;
	}

	element_to_local_ltg_list_faces.Flush();
	element_to_local_ltg_list_faces.ReSize(NE());

	for (int k=1; k <= NE(); k++)
	{
		const FiniteElement3D & elem =  static_cast<const FiniteElement3D &> (GetElem(k));
		// in mode 1 only the DOFs at the contact face are constraint

		const FEFace& face = elem.GetLocalFace(faces(k));
		for (int i=1; i <= face.NFaceNodes(); i++)
		{
			const Node & node = elem.GetNode(face.Node(i));
			const ThermoNodeInterface & tempNode = dynamic_cast<const ThermoNodeInterface &> (node);
			int global_index = node.LTG(tempNode.GetTemperatureLocalIndex());
			int local_index = ltg.Find(global_index);
			element_to_local_ltg_list_faces(k)(i) = local_index;
		}
	}
}



int ThermalContact::NE() const
{
	return elements.Length();
}

void ThermalContact::BuildGlobalSearchTree()
{
	if (slave_search_tree_initialized)
	{
		return;
	}
	Box3D global_box; // box over all slave elements

	int nr_of_slave_elements = NE() - nr_of_master_elements;
	for (int i = 1; i <= nr_of_slave_elements; i++)
	{
		const Element* elem = GetMBS()->GetElementPtr(SlaveElement(i));

		assert(elem != 0);

		global_box.Add(elem->GetElementBox());
	}

	global_box.Increase(EPS);
	
	slave_search_tree.ResetSearchTree(diskretization_x, diskretization_y, diskretization_z, global_box);

	for (int i = 1; i <= nr_of_slave_elements; i++)
	{
		const Element* elem = GetMBS()->GetElementPtr(SlaveElement(i));
		Box3D slave_elem_bounding_box = elem->GetBoundingBox();
		slave_elem_bounding_box.Increase(EPS);
		slave_search_tree.AddItem(slave_elem_bounding_box, i);
	}
	slave_search_tree_initialized = true;
}

IVector ThermalContact::GetSlaveElementsAt(const Vector3D& global_pos, TArrayDynamic<Vector3D>& local_positions)
{
	Vector3D local_pos;
	IVector possible_elements;
	if (slave_search_tree.isEmpty())
	{
		BuildGlobalSearchTree();
	}
	// Get the elements in the respective search tree box
	slave_search_tree.GetItemsOfBox(global_pos, possible_elements);

	// in order to avoid copying the array we move the indices of the elements
	// in which the global point may be included at the beginning and set then
	// then length of the vector accordingly. K denotes the running index of
	// the elements where the point is included.

	int k = 0;

	for (int i = 1; i <= possible_elements.Length(); i++)
	{
		int slave_element_nr = possible_elements(i);
		const FiniteElement3D& element = static_cast<const FiniteElement3D &> (GetMBS()->GetElement(SlaveElement(slave_element_nr)));

		Box3D element_bounding_box = element.GetBoundingBox();
		element_bounding_box.Increase(EPS);
		if ( element_bounding_box.IsIn(global_pos) )
		{
			// the center of the face is the starting point of the newton iteration
			// to find out the local position
			local_pos = element.GetFaceCenter(SlaveFace(slave_element_nr));
			if (element.GetLocPosOfGlobalPos(local_pos, global_pos, NEWTON_ITER))
			{
				possible_elements(++k) = slave_element_nr;
				local_positions(k) = local_pos;
			}
		}
	}
	possible_elements.SetLen(k);
	return possible_elements;
}

void ThermalContact::GetElementData(ElementDataContainer& edc)
{

	Constraint::GetElementData(edc);

	ElementData ed;
  Vector master_elements_vect(nr_of_master_elements);
	for(int i = 1; i <= nr_of_master_elements; i++)
	{
		master_elements_vect(i) = MasterElement(i);
	}
  ed.SetVector(master_elements_vect.GetVecPtr(), master_elements_vect.Length(), "elements"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("master element numbers"); edc.TreeAdd("Master_Elements",ed);

  Vector slave_elements_vect(NE() - nr_of_master_elements);
	for(int i = 1; i <= NE() - nr_of_master_elements; i++)
	{
		slave_elements_vect(i) = SlaveElement(i);
	}
  ed.SetVector(slave_elements_vect.GetVecPtr(),slave_elements_vect.Length(),"elements"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("slave element numbers"); edc.TreeAdd("Slave_Elements",ed);

	Vector master_faces_vect(nr_of_master_elements);
	for(int i = 1; i <= nr_of_master_elements; i++)
	{
		master_faces_vect(i) = MasterFace(i);
	}
	ed.SetVector(master_faces_vect.GetVecPtr(), master_faces_vect.Length(), "faces"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("numbers of master faces in contact: 1 ... neg. x-direction, 2 ... pos. x-direction, 3 ... neg. y direction, ... , 6 ... pos. z-direction"); edc.TreeAdd("Master_Elements",ed);

	Vector slave_faces_vect(NE() - nr_of_master_elements);
	for(int i = 1; i <= NE() - nr_of_master_elements; i++)
	{
		slave_faces_vect(i) = SlaveFace(i);
	}
  ed.SetVector(slave_faces_vect.GetVecPtr(), slave_faces_vect.Length(), "faces"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("numbers of slave faces in contact: 1 ... neg. x-direction, 2 ... pos. x-direction, 3 ... neg. y direction, ... , 6 ... pos. z-direction"); edc.TreeAdd("Slave_Elements",ed);

}

int ThermalContact::SetElementData(ElementDataContainer& edc)
{
	int rv = Constraint::SetElementData(edc);

	GetElemDataIVector(GetMBS(), edc, "Master_Elements.elements", elements, 1);
	nr_of_master_elements = elements.Length();

	IVector slave_elements;
	GetElemDataIVector(GetMBS(), edc, "Slave_Elements.elements", slave_elements, 1);
	elements.Append(slave_elements);

	GetElemDataIVector(GetMBS(), edc, "Master_Elements.faces", faces, 1);

	if (faces.Length() == 1)
	{
		int face_temp = faces(1);
		faces.SetLen(nr_of_master_elements);
		faces.SetAll(face_temp);
	}
	else
	{
		assert(faces.Length() == nr_of_master_elements);
	}

	IVector slave_faces;

	GetElemDataIVector(GetMBS(), edc, "Slave_Elements.faces", slave_faces, 1);

	if (slave_faces.Length() == 1)
	{
		int slave_faces_temp = slave_faces(1);
		slave_faces.SetLen(slave_elements.Length());
		slave_faces.SetAll(slave_faces_temp);
	}
	else
	{
		assert(slave_faces.Length() == slave_elements.Length());
	}
	faces.Append(slave_faces);

	if (wP_s.Length() == 1)
	{
		double wP_s_temp = wP_s(1);
		wP_s.SetLen(nr_of_master_elements);
		wP_s.SetAll(wP_s_temp);
	}
	
	is_initialized = false;
	Initialize();

	slave_search_tree_initialized=false;
	slave_search_tree.ResetSearchTree(0,0,0,Box3D());

	return rv;
}

void ThermalContact::DrawElement() 
{
	Constraint::DrawElement();
	double drawSize  = GetDrawSizeScalar();
	if (drawSize == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

	// Draw an arrow in the middle of the contact surfaces.
	for (int i = 1; i <= NE(); i++)
	{
		const FiniteElement3D &fe = static_cast<const FiniteElement3D &>(GetElem(i));
		const FEFace& face = fe.GetFace(faces(i));

		Vector3D center(0.);
	
		for (int k = 1; k<= face.NFaceNodes(); k++)
		{
			center += mbs->GetNode(face.Node(k)).GetPosD();
		}

		center /= face.NFaceNodes();

		Vector3D v1 = center - mbs->GetNode(face.Node(1)).GetPosD();
		Vector3D v2 = center - mbs->GetNode(face.Node(2)).GetPosD();

		Vector3D normal = v1.Cross(v2);
		normal *= drawSize / normal.Norm();

		mbs->Graphics3D()->DrawArrow(center, center + normal, drawSize/5., drawSize/3.);
	}
	
	// if the mesh is not consistent and mode 1 is selected then the heat flux
	// is prescribed on some points on the the master elements. These points are
	// drawn here. If the point on the master element is in contact with a slave
	// element, then the point is drawn red, otherwise if it is not in contact
	// with a slave element, it is drawn green.
	if (draw_contact_points)
	{
		if (mode==1 && consistent_mesh != 0)
		{
			for (int ielem = 1; ielem <= nr_of_master_elements; ++ielem)
			{
				//master elements ordered first
				const FiniteElement3D& master_element = GetMasterElement(ielem);
				//master_element.SetComputeTemperature(master_element);
				//get integration rule for the specific face of the master element
				IntegrationRule intRule_master(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
				master_element.DefineIntegrationRuleFace(MasterFace(ielem), intRule_master);

				//iterate over all ip's of master element face
				for (IntegrationPointsIterator ip(&intRule_master); !ip.IsEnd(); ++ip)
				{
					Vector3D ip_glob_pos = master_element.GetPosFaceD(MasterFace(ielem), ip.Point2D());
					TArrayDynamic<Vector3D> slave_loc_positions;
					IVector slave_elements = GetSlaveElementsAt(ip_glob_pos, slave_loc_positions); // should be something like GetSlaveElementAtD for drawing
					if (slave_elements.Length()>0)
					{
						// there exists a coorsponding slave element
						mbs->Graphics3D()->SetColor(Vector3D(1., 0., 0.));
					}
					else
					{
						// no corresponding slave element
						mbs->Graphics3D()->SetColor(Vector3D(0., 1., 0.));
					}
					mbs->Graphics3D()->DrawSphere(ip_glob_pos, drawSize*0.15);
				}
			}
		}
	}
}

void ThermalContact::EvalG(Vector &f, double t)
{
	if (mode==2 && consistent_mesh)
	{
		EvalG_ConsistentMesh_Mode2(f,t);
	}
}

void ThermalContact::EvalG_ConsistentMesh_Mode2(Vector &f, double t)
{

	assert(is_initialized);

	f.SetAll(0.);

	for (int i = 1; i <= is; i++)
	{
		int ielem = representative_elements(i);
		int local_master_index = representative_master_local_coordinate_nrs(i);
		int local_slave_index  = representative_slave_local_coordinate_nrs(i);

		const FiniteElement3D &master = GetMasterElement(ielem);
		const FiniteElement3D &slave  = GetSlaveElement(ielem);

		double master_temperature = master.XG(local_master_index);
		double slave_temperature  = slave.XG(local_slave_index);

		f(i) = master_temperature - slave_temperature;
	}
}

void ThermalContact::AddElementCqTLambda(double t, int locelemind, Vector &f)
{
	if (mode == 1 || mode == 3)
	{
		return;
	}
	if (mode == 2)
	{
		if (consistent_mesh)
		{
			assert(is_initialized);

			bool is_slave_element = false;
			int element_index = locelemind;
			if (locelemind > nr_of_master_elements)
			{
				is_slave_element = true;
				element_index = locelemind - nr_of_master_elements;
			}

			for (int i = 1; i <= is; i++)
			{
				if (representative_elements(i) == element_index)
				{
					if (is_slave_element)
					{
						int loccoord = representative_slave_local_coordinate_nrs(i);
						f(loccoord) += -XG(2*sos+i);
					}
					else
					{
						int loccoord = representative_master_local_coordinate_nrs(i);
						f(loccoord) += XG(2*sos+i);
					}
				}
			}
			return;
		}
		else // not consistent mesh
		{
			assert(false && "not implemented yet");
			return;
		}
	}
	assert(false && "unknown mode");
}

void ThermalContact::EvalF2(Vector& f, double t)
{
	if (consistent_mesh && mode == 1)
	{
		EvalF2_ConsistentMesh_Mode1(f, t);
	}
	else if (!consistent_mesh && mode == 1)
	{
		EvalF2_NotConsistentMesh_Mode1(f,t);
	}
	else if (consistent_mesh && mode == 2)
	{
		EvalF2_ConsistentMesh_Mode2(f,t);
	}
	else if (consistent_mesh && mode == 3)
	{
		EvalF2_ConsistentMesh_Mode2(f,t);
	}
	else
	{
		assert(false);
		GetMBS()->UO(UO_LVL_err) << "Mode " << mystr(mode) << " not implemented yet.\n";
	}
}

int ThermalContact::FastStiffnessMatrix() const
{
	if (consistent_mesh /*&& mode == 1*/)
	{
		return 2;
	}
	return 0;
}
//$ SW 2015-10-27: possible speed up: however tests showed that this is not a bottleneck

int ThermalContact::UseSparseK() const
{
	return 0;
	if (consistent_mesh && mode == 1)
	{
		return 1;
	}
	return 0;
}

void ThermalContact::AddKSparse_Static(SparseMatrix& K, double t)
{
	if (!consistent_mesh || mode != 1)
	{
		assert(false);
		return;
	}

	int sos = SOS();


	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face_master, s0Face_slave; 
	ConstVector<8> xgCache_master, xgCache_slave;

	ConstMatrix<2*20*2*20> partial_stiffness_matrix;
	IVector colref; // = rowref


	for (int ielem = 1; ielem <= nr_of_master_elements; ++ielem)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);
		const ThermoElementInterface& temp_master_element = dynamic_cast<const ThermoElementInterface &> (master_element);

		const FiniteElement3D& slave_element  = GetSlaveElement(ielem);
		const ThermoElementInterface& temp_slave_element = dynamic_cast<const ThermoElementInterface &> (slave_element);

		int nFaceMaster = MasterFace(ielem);
		int nFaceSlave  = SlaveFace(ielem);

		const MaterialGeneric& mat_master = master_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_master = static_cast<const MaterialThermoMechanic&>(mat_master);

		const MaterialGeneric& mat_slave = slave_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_slave = static_cast<const MaterialThermoMechanic&>(mat_slave);

		double specific_heat_capacity_master = tempmat_master.GetSpecificHeatCapacity();
		double specific_heat_capacity_slave  = tempmat_slave.GetSpecificHeatCapacity();

		double rho_master = tempmat_master.Density();
		double rho_slave  = tempmat_slave.Density();

		double fact_master = 1./(specific_heat_capacity_master * rho_master);
		double fact_slave  = 1./(specific_heat_capacity_slave  * rho_slave);

		//temp_master_element.SetComputeTemperatureOfFace(xgCache_master,nFaceMaster);		
		//temp_slave_element.SetComputeTemperatureOfFace(xgCache_slave,nFaceSlave);
		IVector permutation;
		GetNodePermutations(ielem, permutation);

		Matrix3D jac_master, jac_slave;

		int ns_master = master_element.NSFace(nFaceMaster);
		int ns_slave = slave_element.NSFace(nFaceSlave);

		partial_stiffness_matrix.SetSize(ns_master+ns_slave,ns_master+ns_slave);
		partial_stiffness_matrix.SetAll(0.);

		IntegrationRule intRuleMaster(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		IntegrationRule intRuleSlave(*slave_element.GetIntegrationRule(IntegrationRule::IVT_Mass));

		master_element.DefineIntegrationRuleFace(nFaceMaster, intRuleMaster);
		slave_element.DefineIntegrationRuleFace(nFaceSlave, intRuleSlave);

		IntegrationRule* intRule;
		if (intRuleMaster.GetNumberOfIntegrationPoints() < intRuleSlave.GetNumberOfIntegrationPoints())
		{ intRule = &intRuleSlave; }
		else
		{ intRule = &intRuleMaster; }

		double beta_master = alpha*fact_master;
		double beta_slave = alpha*fact_slave;

		for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
		{
		
			master_element.GetFaceRefJacobi(nFaceMaster, jac_master, ip.Point());

			// \lambda_M dT_m/dn =   \alpha * (T_m - T_w) + wP_s * f_M 
			// \int_{\partial \Omega} \alpha * (T_m - T_w) + wP_s * f_M  ds
			master_element.GetS0Face(nFaceMaster, ip.Point(), s0Face_master);
			//ip needs to be changed -> transformation to local coordinate system of slave element
			
			Vector2D ipPosSlave;
			if (ns_master == 4 || ns_master == 8)
			{
				ipPosSlave = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}

			slave_element.GetS0Face(nFaceSlave, ipPosSlave.MakeV3D(), s0Face_slave);

			Vector3D dS1(jac_master(1,1), jac_master(2,1), jac_master(3,1));
			Vector3D dS2(jac_master(1,2), jac_master(2,2), jac_master(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			for (int i = 1; i <= ns_master; i++)
			{
				for (int j = 1; j <= ns_master; j++)
				{
					partial_stiffness_matrix(i, j) -= beta_master*jacdetSurf*s0Face_master(i)*s0Face_master(j)*ip.Weight();
					//K(ltg(element_to_local_ltg_list(ielem)(i)), ltg(element_to_local_ltg_list(ielem)(j))) -= beta_master*jacdetSurf*s0Face_master(i)*s0Face_master(j)*ip.Weight();
				}
				for (int j = 1; j <= ns_slave; j++)
				{
					partial_stiffness_matrix(i, j + ns_master) += beta_master*jacdetSurf*s0Face_master(i)*s0Face_slave(j)*ip.Weight();
					//K(ltg(element_to_local_ltg_list(ielem)(i)), ltg(element_to_local_ltg_list(ielem+nr_of_master_elements)(j))) += beta_master*jacdetSurf*s0Face_master(i)*s0Face_slave(j)*ip.Weight();
				}
			}

			// integration over face of slave element
			slave_element.GetS0Face(nFaceSlave, ip.Point(), s0Face_slave);
			slave_element.GetFaceRefJacobi(nFaceSlave, jac_slave, ip.Point()); //?????????????????
			Vector2D ipPosMaster;
			if (ns_slave == 4 || ns_slave == 8)
			{
				ipPosMaster = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			master_element.GetS0Face(nFaceMaster, ipPosMaster.MakeV3D(), s0Face_master); 
			dS1 = Vector3D(jac_slave(1,1), jac_slave(2,1), jac_slave(3,1));
			dS2 = Vector3D(jac_slave(1,2), jac_slave(2,2), jac_slave(3,2));
			jacdetSurf = dS1.Cross(dS2).Norm();
			for (int i = 1; i <= ns_slave; i++)
			{
				for (int j = 1; j <= ns_slave; j++)
				{
					partial_stiffness_matrix(i+ns_master, j+ns_master) -= beta_slave*jacdetSurf*s0Face_slave(i)*s0Face_slave(j)*ip.Weight();
					//K(ltg(element_to_local_ltg_list(ielem + nr_of_master_elements)(i)), ltg(element_to_local_ltg_list(ielem + nr_of_master_elements)(j))) -= beta_slave*jacdetSurf*s0Face_slave(i)*s0Face_slave(j)*ip.Weight();
				}
				for (int j = 1; j <= ns_master; j++)
				{
					partial_stiffness_matrix(i+ns_master, j) += beta_slave*jacdetSurf*s0Face_slave(i)*s0Face_master(j)*ip.Weight();
					//K(ltg(element_to_local_ltg_list(ielem + nr_of_master_elements)(i)), ltg(element_to_local_ltg_list(ielem)(j))) += beta_slave*jacdetSurf*s0Face_slave(i)*s0Face_master(j)*ip.Weight();
				}
			}
		}
		for (int i = 1; i<=ns_master + ns_slave; i++)
		{
			colref(i) = ltg(element_to_local_ltg_list(ielem)(i));
		}

		K.AddMatrix(colref, colref, ns_master + ns_slave, ns_master + ns_slave, partial_stiffness_matrix);

	}	
}

void ThermalContact::StiffnessMatrix(Matrix &m)
{
	if (consistent_mesh && mode == 1)
	{
		StiffnessMatrix_ConsistentMesh_Mode1(m);
		return;
	}
	if (consistent_mesh && mode == 2)
	{
		StiffnessMatrix_ConsistentMesh_Mode2(m);
		return;
	}
	if (consistent_mesh && mode == 3)
	{
		StiffnessMatrix_ConsistentMesh_Mode2(m);
		return;
	}
	assert(false && "Stiffness matrix not implemented for this mode");
}

void ThermalContact::SetComputeTemperature(Vector & xgCache) const
{
	xgCache.SetLen(SOS());
	for (int i = 1; i <= xgCache.Length(); i++)
		xgCache(i) = XG(i);
}

void ThermalContact::EvalF2_NotConsistentMesh_Mode1(Vector& f, double t) 
{
	// In this mode the heat flux is proportional to the temperature difference between two elements.
	// Additionally a heat source at the surface can be defined, which brings 
	// \lambda_M dT_m/dn =   \alpha * (T_m - T_w) + wP_s * f_M
	// \lambda_W dT_m/dn = - \alpha * (T_m - T_w) + wP_s * (1 - f_M)

	int sos = SOS();
	Vector fadd(sos); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face_master, s0Face_slave; 

	
	// we iterate over the master elements and calculate the normal flux at the master elements
	// integration points. The same flux in the other direction is added to the corresponding 
	// slave element at this position.

	for (int ielem = 1; ielem <= nr_of_master_elements; ++ielem)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);
		const ThermoElementInterface& temp_master_element = dynamic_cast<const ThermoElementInterface &> (master_element);

		const MaterialContinuumMechanics& mat_master = master_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_master = (MaterialThermoMechanic&)(mat_master);

		double specific_heat_capacity_master = tempmat_master.GetSpecificHeatCapacity();
		double rho_master = tempmat_master.Density();
		//double fact_master = 1./(specific_heat_capacity_master * rho_master);

		//get integration rule for the specific face of the master element
		IntegrationRule intRule_master(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		master_element.DefineIntegrationRuleFace(MasterFace(ielem), intRule_master);

		double reference_face_area = 0.;
		for (IntegrationPointsIterator ip(&intRule_master); !ip.IsEnd(); ++ip)
		{
			reference_face_area += ip.Weight();
		}

		double master_face_area = master_element.GetFaceArea(MasterFace(ielem));

		//iterate over all ip's of master element face
		for (IntegrationPointsIterator ip(&intRule_master); !ip.IsEnd(); ++ip)
		{
			double master_temperature = temp_master_element.GetTemperature(master_element.GetVolumeCoordOfFaceCoord(MasterFace(ielem), ip.Point2D()));
			master_element.GetS0Face(MasterFace(ielem), ip.Point(), s0Face_master);

			Vector3D ip_glob_pos = master_element.GetPosFace(MasterFace(ielem), ip.Point2D());
			TArrayDynamic<Vector3D> slave_loc_positions;

			// Get all slave elements at the global position. Should be normally only one
			// but could be two or more if the integration point is at an edge between two
			// elements or on a vertex.
			IVector slave_elements = GetSlaveElementsAt(ip_glob_pos, slave_loc_positions);

			for (int k = 1; k <= slave_elements.Length(); k++)
			{
			
				Vector3D slave_loc_pos = slave_loc_positions(k);
				const FiniteElement3D& slave_element = GetSlaveElement(slave_elements(k));
				const ThermoElementInterface& temp_slave_element = dynamic_cast<const ThermoElementInterface &>(slave_element);

				assert(Vector3D(ip_glob_pos - slave_element.GetPos(slave_loc_pos)).Norm()<= 1e-9);

				const MaterialContinuumMechanics& mat_slave = slave_element.GetMaterial();
				const MaterialThermoMechanic& tempmat_slave = (MaterialThermoMechanic&)(mat_slave);

				double specific_heat_capacity_slave  = tempmat_slave.GetSpecificHeatCapacity();
				double rho_slave  = tempmat_slave.Density();

				double slave_temperature = temp_slave_element.GetTemperature(slave_loc_pos);
				Vector2D slave_face_pos = slave_element.GetFaceCoordOfVolumeCoord(SlaveFace(slave_elements(k)), slave_loc_pos);

				slave_element.GetS0Face(SlaveFace(slave_elements(k)), slave_face_pos.MakeV3D(), s0Face_slave);

				double factor = (master_face_area/reference_face_area)*ip.Weight() * (wP_s(ielem)*(1.-f_M) + alpha*(slave_temperature-master_temperature))/slave_elements.Length();
				for (int i = 1; i <= s0Face_master.Length(); i++)
				{
					fadd(element_to_local_ltg_list(ielem)(i)) += factor*s0Face_master(i);
				}

				for (int i = 1; i <= s0Face_slave.Length(); i++)
				{
					fadd(element_to_local_ltg_list(nr_of_master_elements + slave_elements(k))(i)) -= factor*s0Face_slave(i);
				}
			}
		}
	}

	f += fadd;

} 

void ThermalContact::EvalF2_ConsistentMesh_Mode1(Vector& f, double t)
{

	int sos = SOS();

	Vector fadd(sos); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face_master, s0Face_slave; 
	ConstVector<8> xgCache_master, xgCache_slave;

	for (int ielem = 1; ielem <= nr_of_master_elements; ++ielem)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);
		const ThermoElementInterface& temp_master_element = dynamic_cast<const ThermoElementInterface &> (master_element);

		const FiniteElement3D& slave_element  = GetSlaveElement(ielem);
		const ThermoElementInterface& temp_slave_element = dynamic_cast<const ThermoElementInterface &> (slave_element);

		int nFaceMaster = MasterFace(ielem);
		int nFaceSlave  = SlaveFace(ielem);

		temp_master_element.SetComputeTemperatureOfFace(xgCache_master,nFaceMaster);		
		temp_slave_element.SetComputeTemperatureOfFace(xgCache_slave,nFaceSlave);
		IVector permutation;
		GetNodePermutations(ielem, permutation);

		Matrix3D jac_master, jac_slave;

		int ns_master = master_element.NSFace(nFaceMaster);
		int ns_slave = slave_element.NSFace(nFaceSlave);

		IntegrationRule intRuleMaster(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		IntegrationRule intRuleSlave(*slave_element.GetIntegrationRule(IntegrationRule::IVT_Mass));

		master_element.DefineIntegrationRuleFace(nFaceMaster, intRuleMaster);
		slave_element.DefineIntegrationRuleFace(nFaceSlave, intRuleSlave);

		IntegrationRule* intRule;
		if (intRuleMaster.GetNumberOfIntegrationPoints() < intRuleSlave.GetNumberOfIntegrationPoints())
		{ intRule = &intRuleSlave; }
		else
		{ intRule = &intRuleMaster; }

		for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
		{
			master_element.GetFaceRefJacobi(nFaceMaster, jac_master, ip.Point());

			// \lambda_M dT_m/dn =   \alpha * (T_m - T_w) + wP_s * f_M 
			// \int_{\partial \Omega} \alpha * (T_m - T_w) + wP_s * f_M  ds
			master_element.GetS0Face(nFaceMaster, ip.Point(), s0Face_master);
			//ip needs to be changed -> transformation to local coordinate system of slave element
			
			Vector2D ipPosSlave;
			if (ns_master == 4 || ns_master == 8)
			{
				ipPosSlave = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}

			slave_element.GetS0Face(nFaceSlave, ipPosSlave.MakeV3D(), s0Face_slave);

			Vector3D dS1(jac_master(1,1), jac_master(2,1), jac_master(3,1));
			Vector3D dS2(jac_master(1,2), jac_master(2,2), jac_master(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();
			double temp;
			for (int i = 1; i <= ns_master; i++)
			{
				temp = 0.;
				for (int j = 1; j <= ns_master; j++)
				{
					temp += xgCache_master(j)*s0Face_master(j);
				}
				for (int j = 1; j <= ns_slave; j++)
				{
					temp -= xgCache_slave(j)*s0Face_slave(j);
				}
				fadd(element_to_local_ltg_list(ielem)(i)) += jacdetSurf * ip.Weight() * (wP_s(ielem)*f_M - alpha*temp) * s0Face_master(i);
			}

			// integration over face of slave element
			slave_element.GetS0Face(nFaceSlave, ip.Point(), s0Face_slave);
			slave_element.GetFaceRefJacobi(nFaceSlave, jac_slave, ip.Point()); //?????????????????
			Vector2D ipPosMaster;
			if (ns_slave == 4 || ns_slave == 8)
			{
				ipPosMaster = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			master_element.GetS0Face(nFaceMaster, ipPosMaster.MakeV3D(), s0Face_master); 
			dS1 = Vector3D(jac_slave(1,1), jac_slave(2,1), jac_slave(3,1));
			dS2 = Vector3D(jac_slave(1,2), jac_slave(2,2), jac_slave(3,2));
			jacdetSurf = dS1.Cross(dS2).Norm();
			for (int i = 1; i <= ns_slave; i++)
			{
				temp = 0.;
				for (int j = 1; j <= ns_slave; j++)
				{
					temp -= xgCache_slave(j)*s0Face_slave(j);
				}
				for (int j = 1; j <= ns_master; j++)
				{
					temp += xgCache_master(j)*s0Face_master(j);
				}
				fadd(element_to_local_ltg_list(nr_of_master_elements + ielem)(i)) += jacdetSurf * ip.Weight() * (wP_s(ielem)*(1.-f_M) + alpha*temp)*s0Face_slave(i);
			}
		}
	}
	f += fadd;
}

void ThermalContact::StiffnessMatrix_ConsistentMesh_Mode1(Matrix &K)
{
	K.SetAll(0.);

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face_master, s0Face_slave; 
	ConstVector<8> xgCache_master, xgCache_slave;

	for (int ielem = 1; ielem <= nr_of_master_elements; ++ielem)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);

		const FiniteElement3D& slave_element  = GetSlaveElement(ielem);

		int nFaceMaster = MasterFace(ielem);
		int nFaceSlave  = SlaveFace(ielem);

		IVector permutation;
		GetNodePermutations(ielem, permutation);

		Matrix3D jac_master, jac_slave;

		int ns_master = master_element.NSFace(nFaceMaster);
		int ns_slave = slave_element.NSFace(nFaceSlave);

		IntegrationRule intRuleMaster(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		IntegrationRule intRuleSlave(*slave_element.GetIntegrationRule(IntegrationRule::IVT_Mass));

		master_element.DefineIntegrationRuleFace(nFaceMaster, intRuleMaster);
		slave_element.DefineIntegrationRuleFace(nFaceSlave, intRuleSlave);

		IntegrationRule* intRule;
		if (intRuleMaster.GetNumberOfIntegrationPoints() < intRuleSlave.GetNumberOfIntegrationPoints())
		{ intRule = &intRuleSlave; }
		else
		{ intRule = &intRuleMaster; }

		for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
		{
		
			master_element.GetFaceRefJacobi(nFaceMaster, jac_master, ip.Point());

			// \lambda_M dT_m/dn =   \alpha * (T_m - T_w) + wP_s * f_M 
			// \int_{\partial \Omega} \alpha * (T_m - T_w) + wP_s * f_M  ds
			master_element.GetS0Face(nFaceMaster, ip.Point(), s0Face_master);
			//ip needs to be changed -> transformation to local coordinate system of slave element
			
			Vector2D ipPosSlave;
			if (ns_master == 4 || ns_master == 8)
			{
				ipPosSlave = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}

			slave_element.GetS0Face(nFaceSlave, ipPosSlave.MakeV3D(), s0Face_slave);

			Vector3D dS1(jac_master(1,1), jac_master(2,1), jac_master(3,1));
			Vector3D dS2(jac_master(1,2), jac_master(2,2), jac_master(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			for (int i = 1; i <= ns_master; i++)
			{
				for (int j = 1; j <= ns_master; j++)
				{
					K(element_to_local_ltg_list(ielem)(i), element_to_local_ltg_list(ielem)(j)) -= alpha*jacdetSurf*s0Face_master(i)*s0Face_master(j)*ip.Weight();
				}
				for (int j = 1; j <= ns_slave; j++)
				{
					K(element_to_local_ltg_list(ielem)(i), element_to_local_ltg_list(ielem+nr_of_master_elements)(j)) += alpha*jacdetSurf*s0Face_master(i)*s0Face_slave(j)*ip.Weight();
				}
			}

			// integration over face of slave element
			slave_element.GetS0Face(nFaceSlave, ip.Point(), s0Face_slave);
			slave_element.GetFaceRefJacobi(nFaceSlave, jac_slave, ip.Point()); //?????????????????
			Vector2D ipPosMaster;
			if (ns_slave == 4 || ns_slave == 8)
			{
				ipPosMaster = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			master_element.GetS0Face(nFaceMaster, ipPosMaster.MakeV3D(), s0Face_master); 
			dS1 = Vector3D(jac_slave(1,1), jac_slave(2,1), jac_slave(3,1));
			dS2 = Vector3D(jac_slave(1,2), jac_slave(2,2), jac_slave(3,2));
			jacdetSurf = dS1.Cross(dS2).Norm();
			for (int i = 1; i <= ns_slave; i++)
			{
				for (int j = 1; j <= ns_slave; j++)
				{
					K(element_to_local_ltg_list(ielem + nr_of_master_elements)(i), element_to_local_ltg_list(ielem + nr_of_master_elements)(j)) -= alpha*jacdetSurf*s0Face_slave(i)*s0Face_slave(j)*ip.Weight();
				}
				for (int j = 1; j <= ns_master; j++)
				{
					K(element_to_local_ltg_list(ielem + nr_of_master_elements)(i), element_to_local_ltg_list(ielem)(j)) += alpha*jacdetSurf*s0Face_slave(i)*s0Face_master(j)*ip.Weight();
				}
			}
		}
	}
}

void ThermalContact::EvalF2_ConsistentMesh_Mode2(Vector &f, double t)
{
	// Temperature continuous -> see EvalG and AddCqTLambdaElement
	// Heat flux continuous up to source term -> here

	assert(is_initialized);

	int sos = SOS();

	//Vector fadd(sos);
	ConstVector<20> xgCache_master, xgCache_slave; //number of nodes in a quadratic hexahedral -> maximum number of nodes an element can have

	ConstVector<8> s0Face_master, s0Face_slave;
	ConstMatrix<3*8> gradFace_master, gradFace_slave;

	for (int ielem = 1; ielem <= nr_of_master_elements; ielem++)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);
		const ThermoElementInterface& temp_master_element = dynamic_cast<const ThermoElementInterface &>(master_element);

		const FiniteElement3D& slave_element = GetSlaveElement(ielem);
		const ThermoElementInterface& temp_slave_element = dynamic_cast<const ThermoElementInterface &>(slave_element);

		int nFaceMaster = MasterFace(ielem);
		int nFaceSlave  = SlaveFace(ielem);

		temp_master_element.SetComputeTemperature(xgCache_master);
		temp_slave_element.SetComputeTemperature(xgCache_slave);

		IVector permutation;
		GetNodePermutations(ielem, permutation);

		Matrix3D jac_master, jac_slave;

		int ns_master = master_element.NSFace(nFaceMaster);
		int ns_slave = slave_element.NSFace(nFaceSlave);

		IntegrationRule intRule_master(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		IntegrationRule intRule_slave(*slave_element.GetIntegrationRule(IntegrationRule::IVT_Mass));

		master_element.DefineIntegrationRuleFace(nFaceMaster, intRule_master);
		slave_element.DefineIntegrationRuleFace(nFaceSlave, intRule_slave);
		IntegrationRule* intRule;
		if (intRule_master.GetNumberOfIntegrationPoints() < intRule_slave.GetNumberOfIntegrationPoints())
		{ intRule = &intRule_slave; }
		else
		{ intRule = &intRule_master; }

		Matrix agrad_master, agrad_slave;

		const MaterialContinuumMechanics& mat_master = master_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_master = (MaterialThermoMechanic&)(mat_master);

		const MaterialContinuumMechanics& mat_slave = slave_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_slave = (MaterialThermoMechanic&)(mat_slave);

		Matrix3D conductivity_master = tempmat_master.GetThermalConductivity();
		Matrix3D conductivity_slave = tempmat_slave.GetThermalConductivity();

		for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
		{

			Vector2D ipPosSlave;
			if (ns_master == 4 || ns_master == 8)
			{
				ipPosSlave = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			Vector3D ipPosSlave_3D = slave_element.GetVolumeCoordOfFaceCoord(nFaceSlave, ipPosSlave);
			slave_element.GetJacInvDS(ipPosSlave_3D, agrad_slave);

			Vector3D gradT_slave;
			slave_element.GradScal(xgCache_slave, agrad_slave, gradT_slave, 0);

			slave_element.GetFaceRefJacobi(nFaceMaster, jac_master, ip.Point());
			slave_element.GetFaceRefJacobi(nFaceSlave, jac_slave, ipPosSlave.MakeV3D());

			Vector3D dS1_master(jac_master(1,1), jac_master(2,1), jac_master(3,1));
			Vector3D dS2_master(jac_master(1,2), jac_master(2,2), jac_master(3,2));

			Vector3D dS1_slave(jac_slave(1,1), jac_slave(2,1), jac_slave(3,1));
			Vector3D dS2_slave(jac_slave(1,2), jac_slave(2,2), jac_slave(3,2));

			//normal vector on face of element 1 in global coordinates
			Vector3D normal_master = dS1_master.Cross(dS2_master);
			double jacdetSurf_master = normal_master.Norm();
			normal_master *= (1./jacdetSurf_master);

			//normal vector on face of element 2 in global coordinates
			Vector3D normal_slave = dS1_slave.Cross(dS2_slave); 
			double jacdetSurf_slave = normal_slave.Norm();
			normal_slave *= (1./jacdetSurf_slave);

			master_element.GetS0Face(nFaceMaster, ip.Point(), s0Face_master);

			// integral over face surface of element 1 
			for (int i = 1; i <= ns_master; i++)
			{
				int index = element_to_local_ltg_list_faces(ielem)(i);
				double val = jacdetSurf_master * ip.Weight() * (wP_s(ielem) - (gradT_slave*conductivity_slave*normal_slave))*s0Face_master(i);
				f(index) += val; // check position of conductivity matrix
			}

			// integral over face surface of element 2 
			Vector2D ipPosMaster;
			if (ns_slave == 4 || ns_slave == 8)
			{
				ipPosMaster = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			Vector3D ipPosMaster_3D = master_element.GetVolumeCoordOfFaceCoord(nFaceMaster, ipPosMaster);
			master_element.GetJacInvDS(ipPosMaster_3D, agrad_master);

			Vector3D gradT_master;
			master_element.GradScal(xgCache_master, agrad_master, gradT_master, 0);
			
			slave_element.GetS0Face(nFaceSlave, ip.Point(), s0Face_slave);
			// integral over face surface of element 1 
			for (int i = 1; i <= ns_slave; i++)
			{
				int index = element_to_local_ltg_list_faces(nr_of_master_elements + ielem)(i);
				double val = jacdetSurf_slave * ip.Weight() * (wP_s(ielem) - (gradT_master*conductivity_master*normal_master))*s0Face_slave(i);
				f(index) += val; // check position of conductivity matrix
			}
		}
	}

}


void ThermalContact::StiffnessMatrix_ConsistentMesh_Mode2(Matrix& K)
{
	// Temperature continuous -> see EvalG and AddCqTLambdaElement
	// Heat flux continuous up to source term -> here

	assert(is_initialized);
	K.FillWithZeros();
	
	int sos = SOS();

	ConstVector<20> xgCache_master, xgCache_slave; //number of nodes in a quadratic hexahedral -> maximum number of nodes an element can have

	ConstVector<8> s0Face_master, s0Face_slave;
	ConstMatrix<3*8> gradFace_master, gradFace_slave;

	for (int ielem = 1; ielem <= nr_of_master_elements; ielem++)
	{
		const FiniteElement3D& master_element = GetMasterElement(ielem);
		const ThermoElementInterface& temp_master_element = dynamic_cast<const ThermoElementInterface &>(master_element);

		const FiniteElement3D& slave_element = GetSlaveElement(ielem);
		const ThermoElementInterface& temp_slave_element = dynamic_cast<const ThermoElementInterface &>(slave_element);

		int nFaceMaster = MasterFace(ielem);
		int nFaceSlave  = SlaveFace(ielem);

		temp_master_element.SetComputeTemperature(xgCache_master);
		temp_slave_element.SetComputeTemperature(xgCache_slave);

		IVector permutation;
		GetNodePermutations(ielem, permutation);


		Matrix3D jac_master, jac_slave;

		int ns_master = master_element.NSFace(nFaceMaster);
		int ns_slave = slave_element.NSFace(nFaceSlave);

		IntegrationRule intRule_master(*master_element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		IntegrationRule intRule_slave(*slave_element.GetIntegrationRule(IntegrationRule::IVT_Mass));

		master_element.DefineIntegrationRuleFace(nFaceMaster, intRule_master);
		slave_element.DefineIntegrationRuleFace(nFaceSlave, intRule_slave);
		IntegrationRule* intRule;
		if (intRule_master.GetNumberOfIntegrationPoints() < intRule_slave.GetNumberOfIntegrationPoints())
		{ intRule = &intRule_slave; }
		else
		{ intRule = &intRule_master; }

		Matrix agrad_master, agrad_slave;

		const MaterialContinuumMechanics& mat_master = master_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_master = (MaterialThermoMechanic&)(mat_master);

		const MaterialContinuumMechanics& mat_slave = slave_element.GetMaterial();
		const MaterialThermoMechanic& tempmat_slave = (MaterialThermoMechanic&)(mat_slave);

		Matrix3D conductivity_master = tempmat_master.GetThermalConductivity();
		Matrix3D conductivity_slave = tempmat_slave.GetThermalConductivity();

		for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
		{

			Vector2D ipPosSlave;
			if (ns_master == 4 || ns_master == 8)
			{
				ipPosSlave = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			Vector3D ipPosSlave_3D = slave_element.GetVolumeCoordOfFaceCoord(nFaceSlave, ipPosSlave);
			slave_element.GetJacInvDS(ipPosSlave_3D, agrad_slave);


			slave_element.GetFaceRefJacobi(nFaceMaster, jac_master, ip.Point());
			slave_element.GetFaceRefJacobi(nFaceSlave, jac_slave, ipPosSlave.MakeV3D());

			Vector3D dS1_master(jac_master(1,1), jac_master(2,1), jac_master(3,1));
			Vector3D dS2_master(jac_master(1,2), jac_master(2,2), jac_master(3,2));

			Vector3D dS1_slave(jac_slave(1,1), jac_slave(2,1), jac_slave(3,1));
			Vector3D dS2_slave(jac_slave(1,2), jac_slave(2,2), jac_slave(3,2));

			//normal vector on face of element 1 in global coordinates
			Vector3D normal_master = dS1_master.Cross(dS2_master);
			double jacdetSurf_master = normal_master.Norm();
			normal_master *= (1./jacdetSurf_master);

			//Matrix ds_slave_0;
			//slave_element.GetDSMatrix0Face(nFaceSlave, ip.Point(), ds_slave_0);

			//normal vector on face of element 2 in global coordinates
			Vector3D normal_slave = dS1_slave.Cross(dS2_slave); 
			double jacdetSurf_slave = normal_slave.Norm();
			normal_slave *= (1./jacdetSurf_slave);

			master_element.GetS0Face(nFaceMaster, ip.Point(), s0Face_master);

			// integral over face surface of element 2 
			for (int i_eqn = 1; i_eqn <= ns_master; i_eqn++)
			{
				int index_eqn = element_to_local_ltg_list_faces(ielem)(i_eqn);
				for (int i_dof = 1; i_dof<= slave_element.NS(); i_dof++)
				{
					int index_dof = element_to_local_ltg_list(ielem+nr_of_master_elements)(i_dof);
					Vector3D test(agrad_slave(1,i_dof), agrad_slave(2,i_dof), agrad_slave(3,i_dof));
					K(index_eqn, index_dof) -= jacdetSurf_master * ip.Weight() * (test * conductivity_slave * normal_slave) * s0Face_master(i_eqn);
				}
			}

			// integral over face surface of element 2 
			Vector2D ipPosMaster;
			if (ns_slave == 4 || ns_slave == 8)
			{
				ipPosMaster = TransformLocalFacePositionQuad(ip.Point(), permutation);
			}
			else
			{
				assert(false && "not implemented yet for trigs");
			}
			Vector3D ipPosMaster_3D = master_element.GetVolumeCoordOfFaceCoord(nFaceMaster, ipPosMaster);
			master_element.GetJacInvDS(ipPosMaster_3D, agrad_master);
			
			slave_element.GetS0Face(nFaceSlave, ip.Point(), s0Face_slave);

			//Matrix ds_master_0;
			//master_element.GetDSMatrix0Face(nFaceMaster, ip.Point(), ds_master_0);

			// integral over face surface of element 1 
			for (int i_eqn = 1; i_eqn <= ns_slave; i_eqn++)
			{
				int index_eqn = element_to_local_ltg_list_faces(ielem+nr_of_master_elements)(i_eqn);
				for (int i_dof = 1; i_dof<= master_element.NS(); i_dof++)
				{
					int index_dof = element_to_local_ltg_list(ielem)(i_dof);
					Vector3D test(agrad_master(1,i_dof), agrad_master(2,i_dof), agrad_master(3,i_dof));
					K(index_eqn, index_dof) -= jacdetSurf_slave * ip.Weight() * (test * conductivity_master * normal_master) * s0Face_slave(i_eqn);
				}
			}
		}
	}
}


const double ThermalContact::EPS = 1e-10;

int ThermalContact::MasterElement(int master_element_nr) const
{
	assert(1 <= master_element_nr && master_element_nr <= nr_of_master_elements);
	return elements(master_element_nr);
}

int ThermalContact::SlaveElement(int slave_element_nr) const
{
	assert(1 <= slave_element_nr && slave_element_nr <= NE()-nr_of_master_elements);
	return elements(slave_element_nr + nr_of_master_elements);
}

int ThermalContact::MasterFace(int master_element_nr) const
{
	assert(1 <= master_element_nr && master_element_nr <= nr_of_master_elements);
	return faces(master_element_nr);
}

int ThermalContact::SlaveFace(int slave_element_nr) const
{
	assert(1 <= slave_element_nr && slave_element_nr <= NE()-nr_of_master_elements);
	return faces(slave_element_nr + nr_of_master_elements);
}

const FiniteElement3D& ThermalContact::GetSlaveElement(int slave_element_nr) const
{
	return (const FiniteElement3D&) GetMBS()->GetElement(SlaveElement(slave_element_nr));
}

const FiniteElement3D& ThermalContact::GetMasterElement(int master_element_nr) const
{
	return (const FiniteElement3D&) GetMBS()->GetElement(MasterElement(master_element_nr));
}

// if two faces are congruent, then the direction of the nodes must be reverse.
// 

//
void ThermalContact::GetNodePermutations(int element_nr, IVector& permutation) const
{
	const FiniteElement3D& master_element = GetMasterElement(element_nr);
	const FiniteElement3D& slave_element = GetSlaveElement(element_nr);
	FEFace& master_face = master_element.GetFace(MasterFace(element_nr));
	FEFace& slave_face = slave_element.GetFace(SlaveFace(element_nr));

	assert(master_face.NFaceNodes() == slave_face.NFaceNodes());

	permutation.SetLen(master_face.NFaceNodes());
	permutation.SetAll(0);


	Vector3D pos_slave_1 = mbs->GetNode(slave_face.Node(1)).RefConfPos();
	int shift = 0;
	for (int j = 1; j <= master_face.NFaceNodes(); ++j)
	{
		Vector3D pos_master = mbs->GetNode(master_face.Node(j)).RefConfPos();
		
		if((pos_slave_1 - pos_master).Norm() < EPS)
		{
			shift = j;
			break;
		}
	}
	
	if (shift == 0)
	{
		mbs->UO(UO_LVL_err) << "Error: Faces of elements " << mystr(element_nr) << " are not congruent!\n";
		assert(false);
	}

	// linear quad
	//  nodes of master element
	//  y     4 +------+ 3 
	//  ^       |      | 
	//  |       |      |
	//  |       |      |
	//  --->X 1 +------+ 2 
	//
	// nodes of slave element can be:
	//  y     2 +-----+ 3  3 +-----+ 4  4 +-----+ 1  1 +-----+ 2
	//  ^       |     |      |     |      |     |      |     | 
	//  |       |  1  |      |  2  |      |  3  |      |  4  |
	//  |       |     |      |     |      |     |      |     |
	//  --->X 1 +-----+ 4  2 +-----+ 1  3 +-----+ 2  4 +-----+ 3
	if (master_face.IsQuad() && master_face.NFaceNodes() == 4)
	{
		switch(shift)
		{
		case 1: permutation.Set4(1,4,3,2); return;
		case 2:	permutation.Set4(2,1,4,3); return;
		case 3: permutation.Set4(3,2,1,4); return;
		case 4: permutation.Set4(4,3,2,1); return;
		}
	}
	//   master nodes:
	//              7
	//   y     4 +--+--+ 3 
	//   ^       |     | 
	//   |     8 +     + 6
	//   |       |     |
	//   --->X 1 +--+--+ 2 
	//              5
	//
	//   slave nodes:                        
	//              6             7             8             5      
	//  y      2 +--+--+ 3   3 +--+--+ 4   4 +--+--+ 1   1 +--+--+ 2 
	//  ^        |     |       |     |       |     |       |     |   
	//  |      5 +  1  + 7   6 +  2  + 8   7 +  3  + 5   8 +  4  + 6 
	//  |        |     |       |     |       |     |       |     |   
	//  --->X  1 +--+--+ 4   2 +--+--+ 1   3 +--+--+ 2   4 +--+--+ 3 
	//              8             5             6             7      
	// 
	//              3             4             1             2    
	//         6 +--+--+ 7   7 +--+--+ 8   8 +--+--+ 5   5 +--+--+ 6
	//           |     |       |     |       |     |       |     | 
	//         2 +  5  + 4   3 +  6  + 1   4 +  7  + 2   1 +  8  + 3
	//           |     |       |     |       |     |       |     |
	//         5 +--+--+ 8   6 +--+--+ 5   7 +--+--+ 6   8 +--+--+ 7
	//              1             2             3             4
	else if (master_face.IsQuad() && master_face.NFaceNodes() == 8)
	{
		switch(shift)
		{
		case 1: permutation.SetXN(8,1,4,3,2,8,7,6,5); return;
		case 2: permutation.SetXN(8,2,1,4,3,5,8,7,6); return;
		case 3: permutation.SetXN(8,3,2,1,4,6,5,8,7); return;
		case 4: permutation.SetXN(8,4,3,2,1,7,6,5,8); return;
		case 5: permutation.SetXN(8,5,8,7,6,1,4,3,2); return;
		case 6: permutation.SetXN(8,6,5,8,7,2,1,4,3); return;
		case 7: permutation.SetXN(8,7,6,5,8,3,2,1,4); return;
		case 8: permutation.SetXN(8,8,7,6,5,4,3,2,1); return;
		}
	}
	// master face:
	//         2
	//         x
	//        / \
	//       /   \
	//      /     \
	//   3 x-------x 1
	//
	// slave face:
	//
	//         3               1                 2
	//         x               x                 x
	//        / \             / \               / \
	//       /   \           /   \             /   \
	//      /  1  \         /  2  \           /  3  \
	//   2 x-------x 1   3 x-------x 2     1 x-------x 3
	//
	else if (master_face.IsTrig() && master_face.NFaceNodes() == 3)
	{
		switch(shift)
		{
		case 1: permutation.Set3(1,3,2); return;
		case 2: permutation.Set3(2,1,3); return;
		case 3: permutation.Set3(3,2,1); return;
		}
	}
	// TODO: UP TO NOW QUADRATIC TETRAEDER RETURN LINEAR FACES. ADAPT THAT AND CHECK IF THE ORDER OF THE NODES IS CORRECT.
	// master face:
	//         2
	//         x
	//        / \
	//     5 x   x 4
	//      /     \
	//   3 x---x---x 1
	//         6
	//
	// slave face:
	//          3                1                 2
	//          x                x                 x     
	//         / \              / \               / \    
	//      5 x   x 6        6 x   x 4         4 x   x 5 
	//       /  1  \          /  2  \           /  3  \  
	//    2 x---x---x 1    3 x---x---x 2     1 x---x---x 3
	//          4                5                 6    
	//
	//          6                4                 5
	//          x                x                 x     
	//         / \              / \               / \    
	//      3 x   x 1        1 x   x 2         2 x   x 3 
	//       /  4  \          /  5  \           /  6  \  
	//    5 x---x---x 4    6 x---x---x 5     4 x---x---x 6
	//          2                3                 1    
	else if (master_face.IsTrig() && master_face.NFaceNodes() == 6)
	{
		switch(shift)
		{
		case 1: permutation.Set6(1,3,2,6,5,4); return;
		case 2: permutation.Set6(2,1,3,4,6,5); return;
		case 3: permutation.Set6(3,2,1,5,4,6); return;
		case 4: permutation.Set6(4,6,5,1,3,2); return;
		case 5: permutation.Set6(5,4,6,2,1,3); return;
		case 6: permutation.Set6(6,5,4,3,2,1); return;
		}
	}
	assert(false && "Unknown face type");
}

Vector2D ThermalContact::TransformLocalFacePositionQuad(const Vector3D& pos, const IVector& permutationFromSourceToTarget) const
{
	// X      D +---+ C   
 	// ^        |   |     
	// |        |   |     
	// +-->Y  A +---+ B  


	static Vector2D node_pos[4];
	node_pos[0] = Vector2D(-1,-1);
	node_pos[1] = Vector2D( 1,-1);
	node_pos[2] = Vector2D( 1, 1);
	node_pos[3] = Vector2D(-1, 1);
	double alpha = 0.5*(pos.X() + 1);
	double beta = 0.5*(pos.Y() + 1);

	Vector2D A = node_pos[permutationFromSourceToTarget(1)-1];
	Vector2D B = node_pos[permutationFromSourceToTarget(2)-1];
	Vector2D D = node_pos[permutationFromSourceToTarget(4)-1];
	return A + alpha*(B-A) + beta*(D-A);

}



//#############################################################
//
//               ThermalRadiation
//
//#############################################################

ThermalRadiation::ThermalRadiation(MBS *mbs) :
 Constraint(mbs,                      // pointer to MBS
            "ThermalRadiation",  // element name
       TArray<int>(),						 // element numbers
       Vector(),                 // data initialization
       Vector3D(-1.,0.,0.),      // drawing dimension
       0.,                       // spring stiffness
       0,                        // use local coordinate system
       1,                        // use penalty
			 Vector3D(0.8,0.3,0.3)),   // color
 sos(0),
 use_linearization(1),
 emissivity(0.),
 air_temperature(0.),
 faces(),
 is_initialized(true)
{
	x_init = Vector();
}

int ThermalRadiation::CheckConsistency(mystr &errorstr)
{
	int rv = Constraint::CheckConsistency(errorstr);
	TArray<int> necessary_kinAccFuncs(0);

	for (int i = 1; i<= NE(); i++)
	{
		TKinematicsAccessFunctions tkaf_element = GetMBS()->GetElement(elements(i)).GetKinematicsAccessFunctions();
		bool element_suitable = (necessary_kinAccFuncs.Length() == 0);
		
		GetNecessaryKinematicAccessFunctions(necessary_kinAccFuncs, i);
		for (int k = 1; k <= necessary_kinAccFuncs.Length(); k++)
		{
			if ((tkaf_element | necessary_kinAccFuncs(k)) == tkaf_element)
			{
				element_suitable = true;
			}
		}
		if (!element_suitable)
		{
			errorstr += "Error in element " + mystr(GetOwnNum()) + " (ThermalRadiation): The constrained element " + mystr(elements(i));
			errorstr +=	" (" + mystr(GetMBS()->GetElement(elements(i)).GetElementSpec()) + ") is not suitable for this constraint\n";
			return max(1,rv);
		}
	}

	return rv;
}

void ThermalRadiation::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int /*numberOfKinematicPair*/)
{
	KinAccFunc.SetLen(0);
	int kaf = (int)(TKinematicsAccessFunctions(TKAF_temperature));
	KinAccFunc.Add(kaf);
}

void ThermalRadiation::CopyFrom(const ThermalRadiation &rhs)
{
	Constraint::CopyFrom(rhs);
	emissivity = rhs.emissivity;
	use_linearization = rhs.use_linearization;
	air_temperature = rhs.air_temperature;
	faces = rhs.faces;
	sos = rhs.sos;
	element_to_local_ltg_list = rhs.element_to_local_ltg_list;
	is_initialized = rhs.is_initialized;
}

Element* ThermalRadiation::GetCopy()
{
	ThermalRadiation *rv = new ThermalRadiation(*this);
	return rv;
}

const char* ThermalRadiation::GetElementSpec() const
{
	return "ThermalRadiation";
}


void ThermalRadiation::LinkToElements()
{
	BuildElementToLocalLTG(); //$ SW 2015-7-17: possible improvement: check if this was done before and no LinkToElementsPenalty was called.
}

int ThermalRadiation::SOS() const
{
	return sos;
}

void ThermalRadiation::Initialize()
{
	if (!is_initialized)
	{
		BuildElementToLocalLTG();
		is_initialized = true;
	}
}

void ThermalRadiation::BuildElementToLocalLTG()
{
	LTGreset();

	element_to_local_ltg_list.Flush();
	element_to_local_ltg_list.ReSize(NE());

	static const int positional_dof = 1;
	static const int velocity_dof = 2;

	// positional dofs = 1, velocity  dofs = 2 (first time derivative of positional dofs)
	for (int pos_or_vel=positional_dof; pos_or_vel <= velocity_dof; pos_or_vel++)
	{
		for (int k=1; k <= NE(); k++)
		{
			const FiniteElement3D & elem = GetElement(k);
			
			FEFace face = elem.GetLocalFace(Face(k));
			for (int i=1; i <= face.NFaceNodes(); i++)
			{
				const ThermoNodeInterface* node = dynamic_cast<const ThermoNodeInterface*>(&elem.GetNode(i));
				int global_index;
				if (pos_or_vel == positional_dof)
				{
					global_index = elem.GetNode(face.Node(i)).LTG(node->GetTemperatureLocalIndex()); // first there are three positional DOF and then, the forth DOF, is the temperature constraint
				}
				else
				{
					global_index = elem.GetNode(face.Node(i)).LTG(node->GetTemperaturePLocalIndex()); // first there are three positional DOF and then, the forth DOF, is the temperature constraint
				}
				int local_index = ltg.AddIfNotExists(global_index);
				element_to_local_ltg_list(k).Add(local_index);
			}
			
		}
		if (pos_or_vel == positional_dof)
		{
			sos = ltg.Length();
		}
	}
}

int ThermalRadiation::NE() const
{
	return elements.Length();
}

void ThermalRadiation::GetElementData(ElementDataContainer& edc)
{

	Constraint::GetElementData(edc);

	ElementData ed;
  Vector elements_vect(NE());
	for(int i = 1; i <= NE(); i++)
	{
		elements_vect(i) = elements(i);
	}
  ed.SetVector(elements_vect.GetVecPtr(),elements.Length(), "elements"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("element numbers"); edc.TreeAdd("",ed);

	Vector faces_vect(NE());
	for(int i = 1; i <= NE(); i++)
	{
		faces_vect(i) = Face(i);
	}
	ed.SetVector(faces_vect.GetVecPtr(), faces.Length(), "faces"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("numbers of faces: 1 ... neg. x-direction, 2 ... pos. x-direction, 3 ... neg. y direction, ... , 6 ... pos. z-direction"); edc.TreeAdd("",ed);
}

int ThermalRadiation::SetElementData(ElementDataContainer& edc)
{
	int rv = Constraint::SetElementData(edc);

	GetElemDataIVector(GetMBS(), edc, "faces", faces, 1);
	GetElemDataIVector(GetMBS(), edc, "elements", elements, 1);

	if (faces.Length() == 1)
	{
		faces.SetLen(NE());
		faces.SetAll(faces(1));
	}
	else
	{
		assert(faces.Length() == NE());
	}
	
	is_initialized = false;
	//Initialize();

	return rv;
}

void ThermalRadiation::DrawElement() 
{
	Constraint::DrawElement();
	//
	double drawSize  = GetDrawSizeScalar();
	if (drawSize == 0) return;

	mbs->Graphics3D()->SetColor(GetCol());

	for (int i = 1; i <= NE(); i++)
	{
		const FiniteElement3D &element = GetElement(i);
		const FEFace& face = element.GetFace(faces(i));

		Vector3D center(0.);
	
		for (int k = 1; k<= face.NFaceNodes(); k++)
		{
			center += mbs->GetNode(face.Node(k)).GetPosD();
		}

		center /= face.NFaceNodes();

		Vector3D v1 = center - mbs->GetNode(face.Node(1)).GetPosD();
		Vector3D v2 = center - mbs->GetNode(face.Node(2)).GetPosD();

		Vector3D normal = v1.Cross(v2);
		normal *= drawSize / normal.Norm();
		mbs->Graphics3D()->DrawZyl(center, center + 0.15*normal,drawSize/5.);
		mbs->Graphics3D()->DrawZyl(center + 0.25*normal, center + 0.40*normal,drawSize/5.);
		mbs->Graphics3D()->DrawArrow(center + 0.5*normal,center + normal, drawSize/5., drawSize/6.);
	}
}



void ThermalRadiation::EvalF2(Vector& f, double t)
{
	if (use_linearization == 1)
	{
		EvalF2_Linearized(f, t);
	}
	else
	{
		EvalF2_Nonlinear(f, t);
	}
}

int ThermalRadiation::FastStiffnessMatrix() const
{
	return 2;
}

void ThermalRadiation::StiffnessMatrix(Matrix & m)
{
	if (use_linearization == 1)
	{
		StiffnessMatrix_Linearized(m);
	}
	else
	{
		StiffnessMatrix_Nonlinear(m);
	}
}

void ThermalRadiation::StiffnessMatrix_Linearized(Matrix& m)
{
	int sos = SOS();
	m.Resize(2*sos,2*sos);
	m.FillWithZeros(); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face;
	Matrix3D jac;
	
	for (int ielem = 1; ielem <= elements.Length(); ++ielem)
	{
		const ThermoElementInterface& tempElement = dynamic_cast<const ThermoElementInterface&>(GetElement(ielem));
		const FiniteElement3D& element = dynamic_cast<const FiniteElement3D&>(GetElement(ielem));

		const MaterialThermoMechanic& tempmat = (const MaterialThermoMechanic&)(element.GetMaterial());
		//double constraintFactor = emissivity/(tempmat.GetSpecificHeatCapacity()*tempmat.Density());

		//get integration rule for the specific face of the master element
		IntegrationRule intRule(*element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		element.DefineIntegrationRuleFace(Face(ielem), intRule);

		//just for info
		//double face_area = element.GetFaceArea(Face(ielem));

		//iterate over all ip's of element face
		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			double ref_temperature = tempElement.GetRefTemperature(element.GetVolumeCoordOfFaceCoord(Face(ielem), ip.Point2D()));
			element.GetS0Face(Face(ielem), ip.Point(), s0Face);

			
			element.GetFaceRefJacobi(Face(ielem), jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			double factor = SIGMA*4*pow(ref_temperature,3);

			//double diff = factor- factordbg;
			//double reldiff = diff/factor;

			factor *= ip.Weight()*jacdetSurf*emissivity;
			for (int i = 1; i <= s0Face.Length(); i++)
			{
				for (int j = 1; j<= s0Face.Length(); j++)
				{
					m(element_to_local_ltg_list(ielem)(i), element_to_local_ltg_list(ielem)(j)) -= factor*s0Face(i)*s0Face(j);
				}
			}
		}
	}
}

void ThermalRadiation::StiffnessMatrix_Nonlinear(Matrix & m)
{
	int sos = SOS();
	m.Resize(2*sos,2*sos); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face; 
	Matrix3D jac;
	
	// we iterate over the master elements and calculate the normal flux at the master elements
	// integration points. The same flux in the other direction is added to the corresponding 
	// slave element at this position.

	for (int ielem = 1; ielem <= elements.Length(); ++ielem)
	{
		const FiniteElement3D& element = GetElement(ielem);
		const ThermoElementInterface & temp_element = dynamic_cast<const ThermoElementInterface &>(element);


		const MaterialThermoMechanic& tempmat = (const MaterialThermoMechanic&)(element.GetMaterial());
		//double constraintFactor = emissivity/(tempmat.GetSpecificHeatCapacity()*tempmat.Density());

		//get integration rule for the specific face of the master element
		IntegrationRule intRule(*element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		element.DefineIntegrationRuleFace(Face(ielem), intRule);

		//just for info
		//double face_area = element.GetFaceArea(Face(ielem));

		//iterate over all ip's of master element face
		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			double temperature = temp_element.GetTemperature(element.GetVolumeCoordOfFaceCoord(Face(ielem), ip.Point2D()));
			element.GetS0Face(Face(ielem), ip.Point(), s0Face);

			element.GetFaceRefJacobi(Face(ielem), jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			double factor = 4*SIGMA*temperature*temperature*temperature;
			factor *= ip.Weight()*jacdetSurf*emissivity;
			for (int i = 1; i <= s0Face.Length(); i++)
			{
				for (int j = 1; j <= s0Face.Length(); j++)
				{
					m(element_to_local_ltg_list(ielem)(i),element_to_local_ltg_list(ielem)(j)) -= factor*s0Face(i)*s0Face(j);
				}
			}
		}
	}
}

void ThermalRadiation::SetComputeTemperature(Vector & xgCache) const
{
	xgCache.SetLen(SOS());
	for (int i = 1; i <= xgCache.Length(); i++)
	{
		xgCache(i) = XG(i);
	}
}

void ThermalRadiation::EvalF2_Linearized(Vector& f, double t) 
{

	int sos = SOS();
	Vector fadd(sos); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face;
	Matrix3D jac;
	
	double airpow4 = pow(air_temperature,4);


	for (int ielem = 1; ielem <= elements.Length(); ++ielem)
	{
		const ThermoElementInterface& tempElement = dynamic_cast<const ThermoElementInterface&>(GetElement(ielem));
		const FiniteElement3D& element = dynamic_cast<const FiniteElement3D&>(GetElement(ielem));

		//const MaterialThermoMechanic& tempmat = (const MaterialThermoMechanic&)(element.GetMaterial());
		//double constraintFactor = emissivity/(tempmat.GetSpecificHeatCapacity()*tempmat.Density());

		//get integration rule for the specific face of the master element
		IntegrationRule intRule(*element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		element.DefineIntegrationRuleFace(Face(ielem), intRule);

		//just for info
		//double face_area = element.GetFaceArea(Face(ielem));

		//iterate over all ip's of element face
		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			double temperature = tempElement.GetTemperature(element.GetVolumeCoordOfFaceCoord(Face(ielem), ip.Point2D()));
			double ref_temperature = tempElement.GetRefTemperature(element.GetVolumeCoordOfFaceCoord(Face(ielem), ip.Point2D()));
			element.GetS0Face(Face(ielem), ip.Point(), s0Face);

			
			element.GetFaceRefJacobi(Face(ielem), jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			//double factor1 = (ref_temperature - air_temperature)*(ref_temperature + air_temperature)*(ref_temperature*ref_temperature + air_temperature*air_temperature);
			//double factor2 = 4*pow(ref_temperature,3)*(temperature - ref_temperature);
			//double factordbg = SIGMA*(factor1+factor2);
			double factor = SIGMA*(pow(ref_temperature,4) + 4*pow(ref_temperature,3)*(temperature - ref_temperature) - airpow4);

			//double diff = factor- factordbg;
			//double reldiff = diff/factor;

			factor *= ip.Weight()*jacdetSurf*emissivity;
			for (int i = 1; i <= s0Face.Length(); i++)
			{
				fadd(element_to_local_ltg_list(ielem)(i)) += factor*s0Face(i);
			}
		}
	}

	f -= fadd;
} 

void ThermalRadiation::EvalF2_Nonlinear(Vector& f, double t)
{
	int sos = SOS();
	Vector fadd(sos); // initialized with zeros

	// temporary storage for the values of the face shape functions at a integration point,
	// size 8 because a quadrilateral quadratic face has 8 nodes and thus 8 shape functions.
	ConstVector<8> s0Face; 
	Matrix3D jac;
	
	// we iterate over the master elements and calculate the normal flux at the master elements
	// integration points. The same flux in the other direction is added to the corresponding 
	// slave element at this position.

	for (int ielem = 1; ielem <= elements.Length(); ++ielem)
	{
		const FiniteElement3D& element = GetElement(ielem);
		const ThermoElementInterface & temp_element = dynamic_cast<const ThermoElementInterface &>(element);


		//const MaterialThermoMechanic& tempmat = (const MaterialThermoMechanic&)(element.GetMaterial());
		//double constraintFactor = emissivity/(tempmat.GetSpecificHeatCapacity()*tempmat.Density());

		//get integration rule for the specific face of the master element
		IntegrationRule intRule(*element.GetIntegrationRule(IntegrationRule::IVT_Mass));
		element.DefineIntegrationRuleFace(Face(ielem), intRule);

		//just for info
		//double face_area = element.GetFaceArea(Face(ielem));

		//iterate over all ip's of master element face
		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			double temperature = temp_element.GetTemperature(element.GetVolumeCoordOfFaceCoord(Face(ielem), ip.Point2D()));
			element.GetS0Face(Face(ielem), ip.Point(), s0Face);

			element.GetFaceRefJacobi(Face(ielem), jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			double jacdetSurf = dS1.Cross(dS2).Norm();

			double factor = SIGMA*(temperature-air_temperature)*(temperature+air_temperature)*(sqr(temperature)+sqr(air_temperature));
			factor *= ip.Weight()*jacdetSurf*emissivity;
			for (int i = 1; i <= s0Face.Length(); i++)
			{
				fadd(element_to_local_ltg_list(ielem)(i)) += factor*s0Face(i);
			}
		}
	}

	f -= fadd;
}


const double ThermalRadiation::SIGMA = 5.67037321E-8;				// Stefan-Boltzman constant  [W / (m^{2} K^{4})]

	
int ThermalRadiation::Face(int local_element_nr) const
{
	assert(1 <= local_element_nr && local_element_nr <= elements.Length());
	return faces(local_element_nr);
}


const FiniteElement3D& ThermalRadiation::GetElement(int local_element_nr) const
{
	assert(1 <= local_element_nr && local_element_nr <= elements.Length());
	return (const FiniteElement3D&) GetMBS()->GetElement(elements(local_element_nr));
}




//#############################################################
//
//               VolumeHeatSource
//
//#############################################################

VolumeHeatSource::VolumeHeatSource(MBS *mbs) :
 Constraint(mbs,                      // pointer to MBS
            "VolumeHeatSource",  // element name
            TArray<int>(),						 // element numbers
            Vector(),                 // data initialization
            Vector3D(-1.,0.,0.),      // drawing dimension
            0.,                       // spring stiffness
            0,                        // use local coordinate system
            1,                        // use penalty
            Vector3D(0.8,0.3,0.3)),   // color
 sos(0),
// all_elements(0),
 is_initialized(true),
 power_density()
{
	x_init = Vector();
}

int VolumeHeatSource::CheckConsistency(mystr &errorstr)
{
	int rv = Constraint::CheckConsistency(errorstr);
	TArray<int> necessary_kinAccFuncs(0);

	for (int i = 1; i<= NE(); i++)
	{
		TKinematicsAccessFunctions tkaf_element = GetElem(i).GetKinematicsAccessFunctions();
		bool element_suitable = (necessary_kinAccFuncs.Length() == 0);
		
		GetNecessaryKinematicAccessFunctions(necessary_kinAccFuncs, i);
		for (int k = 1; k <= necessary_kinAccFuncs.Length(); k++)
		{
			if ((tkaf_element | necessary_kinAccFuncs(k)) == tkaf_element)
			{
				element_suitable = true;
			}
		}
		if (!element_suitable)
		{
			errorstr += "Error in element " + mystr(GetOwnNum()) + " (VolumeHeatSource): The constrained element " + mystr(GetElnum(i));
			errorstr +=	" (" + mystr(GetElem(i).GetElementSpec()) + ") is not suitable for this constraint\n";
			return max(1,rv);
		}
	}

	return rv;
}

void VolumeHeatSource::GetNecessaryKinematicAccessFunctions(TArray<int> &KinAccFunc, int /*numberOfKinematicPair*/)
{
	KinAccFunc.SetLen(0);
	int kaf = (int)(TKinematicsAccessFunctions(TKAF_temperature));
	KinAccFunc.Add(kaf);
}

void VolumeHeatSource::CopyFrom(const VolumeHeatSource &rhs)
{
	Constraint::CopyFrom(rhs);
	sos = rhs.sos;
	power_density = rhs.power_density;
	element_to_local_ltg_list = rhs.element_to_local_ltg_list;
	is_initialized = rhs.is_initialized;
}

Element* VolumeHeatSource::GetCopy()
{
	VolumeHeatSource *rv = new VolumeHeatSource(*this);
	return rv;
}

const char* VolumeHeatSource::GetElementSpec() const
{
	return "VolumeHeatSource";
}

void VolumeHeatSource::LinkToElements()
{
	BuildElementToLocalLTG(); //$ SW 2015-7-17: possible improvement: check if this was done before and no LinkToElementsPenalty was called.
}

int VolumeHeatSource::SOS() const
{
	return sos;
}

void VolumeHeatSource::Initialize()
{
	if (!is_initialized)
	{
		BuildElementToLocalLTG();
		is_initialized = true;
	}
}

void VolumeHeatSource::BuildElementToLocalLTG()
{
	LTGreset();

	element_to_local_ltg_list.Flush();
	element_to_local_ltg_list.ReSize(NE());

	static const int positional_dof = 1;
	static const int velocity_dof = 2;

	// positional dofs = 1, velocity  dofs = 2 (first time derivative of positional dofs)
	for (int pos_or_vel=positional_dof; pos_or_vel <= velocity_dof; pos_or_vel++)
	{
		for (int k=1; k <= NE(); k++)
		{
			const FiniteElement3D & elem = GetElement(k);
			for (int i=1; i <= elem.NNodes(); i++)
			{
				const ThermoNodeInterface & tempNode = dynamic_cast<const ThermoNodeInterface &>(elem.GetNode(i));
				int global_index;
				if (pos_or_vel == positional_dof)
				{
					global_index = elem.GetNode(i).LTG(tempNode.GetTemperatureLocalIndex());
				}
				else
				{
					global_index = elem.GetNode(i).LTG(tempNode.GetTemperaturePLocalIndex());
				}
				int local_index = ltg.AddIfNotExists(global_index);
				element_to_local_ltg_list(k).Add(local_index);
			}
		}
		if (pos_or_vel == positional_dof)
		{
			sos = ltg.Length();
		}
	}
}

int VolumeHeatSource::NE() const
{
	return elements.Length();
}

void VolumeHeatSource::GetElementData(ElementDataContainer& edc)
{

	Constraint::GetElementData(edc);

	ElementData ed;
  Vector elements_vect(NE());
	for(int i = 1; i <= NE(); i++)
	{
		elements_vect(i) = elements(i);
	}
  ed.SetVector(elements_vect.GetVecPtr(),elements.Length(), "elements"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("element numbers"); edc.TreeAdd("",ed);

	Vector power_density_vect(NE());
	for(int i = 1; i <= NE(); i++)
	{
		power_density_vect(i) = power_density(i);
	}
	
	ed.SetVector(power_density_vect.GetVecPtr(), power_density.Length(), "power_density"); ed.SetVariableLength(); ed.SetToolTipText("power_densities of the respective elements"); edc.TreeAdd("",ed);
}

int VolumeHeatSource::SetElementData(ElementDataContainer& edc)
{
	int rv = Constraint::SetElementData(edc);

	Vector tmp;
	GetElemDataVector(GetMBS(), edc, "power_density", tmp, 1);
	GetElemDataIVector(GetMBS(), edc, "elements", elements, 1);

	if (tmp.Length() == 1)
	{
		power_density.SetLen(NE());
		power_density.SetAll(tmp(1));
	}
	else
	{
		assert(tmp.Length() == NE());
		power_density = tmp;
	}
	
	is_initialized = false;
	Initialize();

	return rv;
}

void VolumeHeatSource::DrawElement() 
{
	Constraint::DrawElement();

	double drawSize  = GetDrawSizeScalar();
	if (drawSize == 0) return;

	double maxD = 1E-37;
	double minD = 1E+37;
	for (int i=1; i <= NE(); i++)
	{
		double d = power_density(i);
		if (d > maxD)
		{
			maxD = d;
		}
		if (d < minD)
		{
			minD = d;
		}
	}

	if (minD==maxD)
	{
		minD -= 1.;
	}

	Vector3D colormin(0.,0.,1.);
	Vector3D colormax(1.,0.,0.);

	for (int i = 1; i <= NE(); i++)
	{
		const FiniteElement3D &fe = GetElement(i);
		Vector3D center = fe.GetRefPos();

		double rel_color = (power_density(i) - minD)/(maxD - minD);
		mbs->Graphics3D()->SetColor(rel_color*colormax + (1-rel_color)*colormin);
		mbs->Graphics3D()->DrawSphere(center, drawSize/2.);
	}
}


void VolumeHeatSource::EvalF2(Vector& f, double t) 
{

	int sos = SOS();
	Vector fadd(sos); // initialized with zeros

	Matrix3D jac;
	double jacdet;

	int nelem = elements.Length();
	
	for (int ielem = 1; ielem <= nelem; ++ielem)
	{
		const FiniteElement3D& element = GetElement(ielem);

		const MaterialThermoMechanic& tempmat = (const MaterialThermoMechanic&)(element.GetMaterial());

		//iterate over all ip's of element face
		IntegrationRule intRule(*element.GetIntegrationRule(IntegrationRule::IVT_Load));
		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			element.GetJacobi(jac, ip.Point());
			jacdet = jac.Det();
			double factor = power_density.Get(ielem)*ip.Weight()*fabs(jacdet);

			for (int i = 1; i <= element.NS(); i++)
			{
				fadd(element_to_local_ltg_list.Get(ielem).Get(i)) += factor*element.GetS0(ip.Point(), i);
			}
		}
	}

	f += fadd;
} 

const FiniteElement3D& VolumeHeatSource::GetElement(int local_element_nr) const
{
	assert(1 <= local_element_nr && local_element_nr <= NE());
	return (const FiniteElement3D&) GetMBS()->GetElement(elements.Get(local_element_nr));
}