//#*****************************************************************************************
//#
//# filename:             ThermoConstraints.h
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          In this file all constraints for temperature elements are
//#	                      collected.
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#ifndef THERMOCONSTRAINTS__H
#define THERMOCONSTRAINTS__H

class FiniteElement3D;

//$EDC$[beginclass,classname=ThermalContact,parentclassname=Constraint,addelementtype=TAEconstraint+TAEspecial_connector+TAENotInRelease,addelementtypename=ThermalContact,
//texdescription="This element connects two sets of thermal elements with each other.\\
//  If the mesh of the two bodies are consistent, that is, the face of the $\text{n}^\text{th}$ element in body one is congruent with the face of the $\text{n}^\text{th}$ element in body two, then make sure that
//  the flag consistent\_mesh is set, since the computation will be much faster in this case.",
//texdescriptionEquations="In the following we will refer to the set of master elements as master body and to the set of slave elements as slave body.
// \subsubsection*{Mode 1}
//    In this mode the heat flux in normal direction on the contact surface of the two bodies is constrained
//    \begin{eqnarray*}
//    \lambda_{M}\frac{\partial T_{M}\left(\mathbf{x}\right)}{\partial n_{M}\left(\mathbf{x}\right)} & = & \phantom{{}-}\alpha\cdot\left(T_{M}\left(\mathbf{x}\right)-T_{W}\left(\mathbf{x}\right)\right)+w\cdot f_{M}\\
//    \lambda_{S}\frac{\partial T_{S}\left(\mathbf{x}\right)}{\partial n_{S}\left(\mathbf{x}\right)} & = & -\alpha\cdot\left(T_{M}\left(\mathbf{x}\right)-T_{S}\left(\mathbf{x}\right)\right)+w\cdot\left(1-f_{M}\right)\qquad\forall\mathbf{x}\in\Gamma
//    \end{eqnarray*}
//    Where \\
//    \begin{tabular}{ll}
//    $\lambda_{M}$ & thermal conductivity of master element\tabularnewline
//    $\lambda_{S}$ & thermal conductivity of slave element\tabularnewline
//    $T_{M}$ & temperature of the master element\tabularnewline
//    $T_{S}$ & temperature of the slave element\tabularnewline
//    $n_{M}$ & normal vector on master surface (directing outside)\tabularnewline
//    $n_{S}$ & normal vector on slave surface (directing outside)\tabularnewline
//    $\alpha$ & heat transfer coefficient\tabularnewline
//    $w$ & heat power density on surface (through friction etc.)\tabularnewline
//    $f_{M}$ & amount of the heat power density $w$ absorbed by the master element \tabularnewline
//    $\Gamma$ & contact surface \tabularnewline
//    \end{tabular}\\
//    If the mesh is not consistent, then the equation is evaluated at the integration points of the master element.
//    Thus the master body should have a finer diskretization to get an accurate result.
// \subsubsection*{Mode 2}
//    In this mode the temperature is continuous on the contact surface. The heat flux in normal direction is also continuous up to a given heat power density.\\
//    \begin{eqnarray*}
//    T_{M} & = & T_{W} \\
//    w & = & \lambda_{M}\frac{\partial T_{M}\left(\mathbf{x}\right)}{\partial n_{M}\left(\mathbf{x}\right)} + \lambda_{S}\frac{\partial T_{S}\left(\mathbf{x}\right)}{\partial n_{S}\left(\mathbf{x}\right)} \qquad\forall\mathbf{x}\in\Gamma
//    \end{eqnarray*}",
//texdescriptionLimitations="This constraint is only suitable for static computations"]
class ThermalContact : public Constraint
{
public:
	ThermalContact(MBS* mbs);

	// no pointers so we can use the implicitly generated copy constructor
	//ThermalContact(const ThermalContact& e); 

	virtual int CheckConsistency(mystr& errorstr);
	virtual void CopyFrom(const ThermalContact& rhs);
	virtual Element* GetCopy();
	virtual const char* GetElementSpec() const;
	virtual void Initialize();
	virtual int SOS() const;		// Attention: very slow until LinkToElementsPenalty was called
	virtual int SOSowned() const {return 0;}
	virtual int IS() const;
	virtual int NE() const;
	virtual void LinkToElements();
	virtual void SetGlobalInitConditions(Vector& x_glob) {return;}
	virtual void GetNecessaryKinematicAccessFunctions(TArray<int> &KAF, int numberOfKinematicPair);
	virtual void PreAssemble();
	//Function, which adds the constraint force and direction (action on the element DOF) to the element residual vector
	virtual void AddElementCqTLambda(double t, int locelemind, Vector& f);
	virtual void EvalG(Vector &f, double t);
	virtual void EvalF2(Vector& f, double t);
	virtual int FastStiffnessMatrix() const;
	virtual void StiffnessMatrix(Matrix &m);
	virtual void AddKSparse_Static(SparseMatrix& m, double t);
	virtual int UseSparseK() const;
	virtual void DrawElement();

	virtual void GetElementData(ElementDataContainer& edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);   //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter
	virtual void AddLTG(int gi);
	virtual const IVector& GetLagrangeDOFs() const;
	virtual void LTGreset()
	{
		ltg_is.Flush();
	}

	Vector& FrictionPowerDensity() {return wP_s;}

private:
	int MasterElement(int master_element_nr) const;
	int SlaveElement(int slave_element_nr) const;
	int MasterFace(int master_element_nr) const;
	int SlaveFace(int slave_element_nr) const;
	const FiniteElement3D& GetMasterElement(int master_element_nr) const;
	const FiniteElement3D& GetSlaveElement(int slave_element_nr) const;

	void BuildElementToLocalLTG();
	void BuildElementToLocalLTG_Faces();
	void BuildRepresentingElementsList();

	void BuildGlobalSearchTree();

	void GetNodePermutations(int element_nr, IVector& permutation) const;

	IVector GetSlaveElementsAt(const Vector3D& global_pos, TArrayDynamic<Vector3D>& local_positions);
	void EvalF2_ConsistentMesh_Mode1(Vector& f, double t);
	void EvalF2_ConsistentMesh_Mode2(Vector& f, double t);
	void EvalF2_NotConsistentMesh_Mode1(Vector& f, double t);
	void EvalG_ConsistentMesh_Mode2(Vector &f, double t);
	void StiffnessMatrix_ConsistentMesh_Mode1(Matrix &m);
	void StiffnessMatrix_ConsistentMesh_Mode2(Matrix &m);

	void SetComputeTemperature(Vector & xgCache) const;
	Vector2D TransformLocalFacePositionQuad(const Vector3D& pos, const IVector& permutationFromSourceToTarget) const;

	int mode;                               //$EDC$[varaccess,EDCvarname="mode",minval=1,EDCfolder="Physics",tooltiptext="1 .. heat flux given by difference of surface temperatures + heat source, 2 .. continuous temperature and heat flux at interface + heat source"]
	double alpha;                           //$EDC$[varaccess,EDCvarname="alpha",EDCfolder="Physics",tooltiptext="convective heat transfer coefficient (only used if mode = 1)"]
	Vector wP_s;                            //$EDC$[varaccess,variable_length_vector,EDCvarname="friction_power_density",EDCfolder="Physics",tooltiptext="friction power density"]
	double f_M;                             //$EDC$[varaccess,EDCvarname="f_1",EDCfolder="Physics",tooltiptext="amount of friction power absorbed in element 1 (only used if mode = 1)"]
	
	//EDC int use_penalty_formulation;      //$EDC$[varaccess,remove,EDCvarname="use_penalty_formulation",EDCfolder="Physics"]
	//EDC int spring_stiffness;             //$EDC$[varaccess,remove,EDCvarname="spring_stiffness",EDCfolder="Physics.Penalty"]
	//EDC int use_local_coordinate_system;  //$EDC$[varaccess,remove,EDCvarname="use_local_coordinate_system",EDCfolder="Geometry"]

	int nr_of_master_elements;

	IVector faces;

	int diskretization_x;										//$EDC$[varaccess,EDCvarname="diskretization_x",minval=1,EDCfolder="SearchTree",tooltiptext="diskretization in x-direction"]
	int diskretization_y;										//$EDC$[varaccess,EDCvarname="diskretization_y",minval=1,EDCfolder="SearchTree",tooltiptext="diskretization in y-direction"]
	int diskretization_z;										//$EDC$[varaccess,EDCvarname="diskretization_z",minval=1,EDCfolder="SearchTree",tooltiptext="diskretization in z-direction"]

	int draw_contact_points;                //$EDC$[varaccess,int_bool,EDCvarname="draw_contact_points",EDCfolder="Graphics",tooltiptext="flag if the contact points should be drawn or not (drawing can be slow)"]
	
	bool is_initialized; // flag if the following variables are set : [

	TArrayDynamic<IVector> element_to_local_ltg_list; // list which contains for every element a list whose nth entry is the index of the nth (used) element-DOF in the constraint DOF
	TArrayDynamic<IVector> element_to_local_ltg_list_faces;
	// The constraint forces in EvalG have to be applied by an element. Since the dofs are owned by nodes and elements
	// share the dofs with each other, we need for each constrained element a representing element and the appropriate
	// local coordinate number
	IVector representative_elements;                    // a list of master and slave element indices representing the dof
	IVector representative_master_local_coordinate_nrs; // the local coordinate numbers of the master elements
	IVector representative_slave_local_coordinate_nrs;  // the local coordinate numbers of the slave elements

	// Small positive number. The element bounding boxes are increased by this number, to prevent that a corresponding slave
	// element is not found due to numerical roundings.
	static const double EPS;

	bool slave_search_tree_initialized;
	SearchTree slave_search_tree;
	static const int NEWTON_ITER = 4;       // newton iterations for computing the local element coordinates from a global position

	int consistent_mesh;                   //$EDC$[varaccess,int_bool,EDCvarname="consistent_mesh",EDCfolder="",tooltiptext="flag if the mesh of the master and the slave are consistent, that is the face of the ith master element is congruent with the face of the ith slave element. This can save a lot of computation time."]

	int sos;
	int is;
	IVector ltg_is;
};
//$EDC$[endclass,ThermalContact]


//$EDC$[beginclass,classname=ThermalRadiation,parentclassname=Constraint,addelementtype=TAEconstraint+TAEspecial_connector+TAENotInRelease,addelementtypename=ThermalRadiation,
//texdescription="This element applies a radiation boundary condition to a surface (set of faces of elements); either linearized around a certain temperature or nonlinar ($T^4$).",
//texdescriptionEquations="In the following we will refer to the set of elements.
//
//  In this mode the heat flux in normal direction on the surface is set to be a (linearized or nonlinear) radiation boundary condition.
//  The radiation boundary condition at a certain boundary $\Gamma_R$ is defined by:
//  \begin{equation*}
//   \lambda\frac{\partial T}{\partial n}=\varepsilon\sigma T^{4}.
//  \end{equation*}
//  The constants $\varepsilon$ and $\sigma$ are the material- and surface-specific emissivity ($\sigma \in [0,1]$) and the Stefan\hbox{-}Boltzmann constant
//  $\sigma = 5.67037321\cdot10^{-8}\,W /(m^{2} K^{4})$.
//  The variational formulation is 
//  \begin{equation*}
//   \int_{\Gamma_{R}}\lambda\frac{\partial T}{\partial n}\cdot vdx= \varepsilon\sigma \int_{\Gamma_{R}} T^{4}\cdot vdx \qquad \forall v.
//  \end{equation*}
//  A linearization around a reference temperature $\bar{T}$ is as follows:
//  \begin{equation*}
//   \lambda\frac{\partial T}{\partial n}=\varepsilon\sigma T^{4} \approx \varepsilon\sigma\left(\bar{T}^{4}+4\bar{T}^{3}(T-\bar{T})\right),
//  \end{equation*}
//  which is equivalent to the variational form
//	\begin{equation*}
//  \int_{\Gamma_{R}}\lambda\frac{\partial T}{\partial n}\cdot vdx\approx \varepsilon\sigma \int_{\Gamma_{R}} \left(\bar{T}^{4}+4\bar{T}^{3}(T-\bar{T})\right)\cdot vdx \qquad \forall v.
//  \end{equation*}",
//texdescriptionLimitations="This constraint is so far only suitable for static computations"]
class ThermalRadiation : public Constraint
{
public:
	ThermalRadiation(MBS* mbs);

	// no pointers so we can use the implicitly generated copy constructor
	//ThermalContact(const ThermalContact& e); 

	virtual int CheckConsistency(mystr& errorstr);
	virtual void CopyFrom(const ThermalRadiation& rhs);
	virtual Element* GetCopy();
	virtual const char* GetElementSpec() const;
	virtual void Initialize();
	virtual int SOS() const;		// Attention: very slow until LinkToElementsPenalty was called
	virtual int SOSowned() const { return 0;}
	virtual int IS() const { return 0; }
	virtual int NE() const;
	virtual void LinkToElements();
	virtual void GetNecessaryKinematicAccessFunctions(TArray<int> &KAF, int numberOfKinematicPair);
	//Function, which adds the constraint force and direction (action on the element DOF) to the element residual vector
	virtual void EvalF2(Vector& f, double t);
	virtual int FastStiffnessMatrix() const;
	virtual void StiffnessMatrix(Matrix &m);
	virtual void DrawElement();
	virtual void GetElementData(ElementDataContainer& edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);   //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter
	virtual void PreAssemble() {is_initialized = false;}
	virtual int UseSparseK() const {return 0;}
	virtual void AddKSparse_Static(SparseMatrix& m, double t) { /* is kind of a load so no stiffness matrix */ }

private:
	//returns the face number of the nth element that is in contact with the 
	int Face(int local_element_nr) const;
	const FiniteElement3D& GetElement(int local_element_nr) const;

	void BuildElementToLocalLTG();

	void EvalF2_Nonlinear(Vector& f, double t);
	void EvalF2_Linearized(Vector& f, double t);
	void StiffnessMatrix_Nonlinear(Matrix &m);
	void StiffnessMatrix_Linearized(Matrix &m);
	
	void SetComputeTemperature(Vector & xgCache) const;

	int use_linearization;                  //$EDC$[varaccess,int_bool,EDCvarname="use_linearization",tooltiptext="1 .. heat radiation is linearized around reference_temperature; 0 ... nonlinear"]
	double emissivity;											//$EDC$[varaccess,EDCvarname="emissivity",minval=0,maxval=1,EDCfolder="Physics",tooltiptext="emissivity constant of the (body) surface, in [0,1]"]
	//double ref_temperature;									//$E D C$[varaccess,EDCvarname="ref_temperature",minval=0,EDCfolder="Physics",tooltiptext="in case of linearization: ref_temperature is used"]
	double air_temperature;									//$EDC$[varaccess,EDCvarname="air_temperature",minval=0,EDCfolder="Physics",tooltiptext="surrounding air_temperature"]

	//EDC int use_penalty_formulation;      //$EDC$[varaccess,remove,EDCvarname="use_penalty_formulation",EDCfolder="Physics"]
	//EDC int spring_stiffness;             //$EDC$[varaccess,remove,EDCvarname="spring_stiffness",EDCfolder="Physics.Penalty"]
	//EDC int use_local_coordinate_system;  //$EDC$[varaccess,remove,EDCvarname="use_local_coordinate_system",EDCfolder="Geometry"]

	IVector faces;

	bool is_initialized; // flag if the following variables are set : [

	TArrayDynamic<IVector> element_to_local_ltg_list; // list which contains for every element a list whose nth entry is the index of the nth (used) element-DOF in the constraint DOF;

	static const double SIGMA;			// Stefan-Boltzman constant 5.67037321E-8 [W / (m^{2} K^{4})]

	int sos;
};
//$EDC$[endclass,ThermalRadiation]


//$EDC$[beginclass,classname=VolumeHeatSource,parentclassname=Constraint,addelementtype=TAEconstraint+TAEspecial_connector+TAENotInRelease,addelementtypename=VolumeHeatSource,
//texdescription="This element applies a heat flux density to a domain (set of elements).",
//texdescriptionEquations="The right hand side $f$ of the heat equation
//
//   \begin{equation*}
//   \frac{\partial}{\partial t}T\left(\mathbf{x},t\right)-a\Delta T\left(\mathbf{x},t\right)=f\left(\mathbf{x},t\right)
//   \end{equation*}
//
//  is the quotient of the heat flux density (per volume) and the heat capacity per volume.\\
//
//  In variational formulation equation is multiplied with test functions w and integrated over the domain. The right hand term, which is added by this constraint, gets
//	\[
//		\int_{\Omega}f\cdot w\ = \int_{\Omega}\frac{p}{c\cdot\rho}\cdot w\,
//	\]
//
//	where $p$ is the heat flux density and $c$ is the heat capacity per mass and $\rho$ is the density of the material.\\
//
// ",
//texdescriptionLimitations="This constraint is so far only suitable for static computations ??HAS TO Clarified - EK??"]
class VolumeHeatSource : public Constraint
{
public:
	VolumeHeatSource(MBS* mbs);

	// no pointers so we can use the implicitly generated copy constructor
	//ThermalContact(const ThermalContact& e); 

	virtual int CheckConsistency(mystr& errorstr);
	virtual void CopyFrom(const VolumeHeatSource& rhs);
	virtual Element* GetCopy();
	virtual const char* GetElementSpec() const;
	virtual void Initialize();
	virtual int SOS() const;
	virtual int SOSowned() const { return 0;}
	virtual int IS() const { return 0; }
	virtual int NE() const;
	virtual void LinkToElements();
	virtual void GetNecessaryKinematicAccessFunctions(TArray<int> &KAF, int numberOfKinematicPair);
	virtual void EvalF2(Vector& f, double t);
	virtual int FastStiffnessMatrix() const {return 2;}
	virtual void StiffnessMatrix(Matrix& m) { m.FillWithZeros();}
	virtual void DrawElement();
	virtual void GetElementData(ElementDataContainer& edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);   //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int UseSparseK() const {return 0;}
	virtual void AddKSparse_Static(SparseMatrix& m, double t) { /* is kind of a load so no stiffness matrix */ }
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

private:
	const FiniteElement3D& GetElement(int local_element_nr) const;

	void BuildElementToLocalLTG();

	//EDC int use_penalty_formulation;      //$EDC$[varaccess,remove,EDCvarname="use_penalty_formulation",EDCfolder="Physics"]
	//EDC int spring_stiffness;             //$EDC$[varaccess,remove,EDCvarname="spring_stiffness",EDCfolder="Physics.Penalty"]
	//EDC int use_local_coordinate_system;  //$EDC$[varaccess,remove,EDCvarname="use_local_coordinate_system",EDCfolder="Geometry"]

	Vector power_density;

	bool is_initialized; // flag if the following variables are set : [

	TArrayDynamic<IVector> element_to_local_ltg_list; // list which contains for every element a list whose nth entry is the index of the nth (used) element-DOF in the constraint DOF;

	int sos;
};
//$EDC$[endclass,VolumeHeatSource]


#endif


