//#**************************************************************
//#
//# filename:             FE2DTriQuad.h
//#
//# author:               Gerstmayr Johannes, YV
//#
//# generated:						October 2010
//# description:          2D - triangle and quadrilateral
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#pragma once
#include "FiniteElement2D.h"
#include "vtk_interface.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// QUAD  
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// 4-node or 8-node quadrilateral
// base class is either FiniteElement2D or FiniteFacet2D (which is not initialized through node numbers but with vertex positions instead)
template<class FiniteElement2DBase> //$EDC$[beginclass,classname=QuadrilateralGeneric,templateclass=FiniteElement2DBase,specialization=FiniteElement2D,addelementtype=TAEBody+TAENotInRelease+TAE2D+TAEFE]
class QuadrilateralGeneric : public FiniteElement2DBase
{
public:
	QuadrilateralGeneric(MBS* mbsi) : FiniteElement2DBase(mbsi) {};
	QuadrilateralGeneric(const QuadrilateralGeneric& e) : FiniteElement2DBase(e.mbs) { CopyFrom(e); }

	void SetQuadrilateralGeneric(int bodyindi, const TArray<int>& nodelist, int material_num, double thickness, const Vector3D& coli);

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new QuadrilateralGeneric(*this);
		return ec;
	}

	virtual const char* GetElementSpec() const {return "QuadrilateralGeneric";}
	virtual TFiniteElementType GetElementType() const {return TFE_Quadrilateral;}

	virtual int NEdges() const {return 4;}
	virtual const FEEdge & GetLocalEdge(int i) const;
	virtual Vector2D GetEdgeNormal(int edge_nr) const;
	virtual Vector2D GetEdgeNormalD(int edge_nr) const;
	virtual Vector2D GetAreaCoordOfEdgeCoord(int edge, double p_loc_edge) const;


	virtual double GetS0(const Vector2D& ploc, int shape) const; 
	virtual void GetDSMatrix0(const Vector2D& ploc, Matrix& sm) const;
	virtual Vector2D GetNodeLocPos2D(int i) const;
	virtual void DrawElement();
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;

protected:
	virtual int GetActualInterpolationOrder() const;

	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const;
	virtual void DefineIntegrationRuleEdge(int /*nEdge*/, IntegrationRule & integrationRule) const
	{
		int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
		assert(ruleOrder != 0);
		IntegrationRule::DefineIntegrationRuleLine(integrationRule, ruleOrder);
	}

	//$EK 2017-02-14: thickness now editable through script ("ChangeProperties")
	//EDC double thickness; //$EDC$[varaccess,EDCvarname="thickness",EDCfolder="Geometry",tooltiptext="thickness of finite element (especially used for volumetric loads)"]
}; //$EDC$[endclass,QuadrilateralGeneric]


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// TRIG  
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// 3-node or 6-node triangle
// base class is FiniteElement2D
template<class FiniteElement2DBase> //$EDC$[beginclass,classname=TriangularGeneric,templateclass=FiniteElement2DBase,specialization=FiniteElement2D,addelementtype=TAEBody+TAENotInRelease+TAE2D+TAEFE]
class TriangularGeneric : public FiniteElement2DBase
{
public:
	TriangularGeneric(MBS* mbsi) : FiniteElement2DBase(mbsi, TArray<int>(3).Set3(1,2,3))
	{ };
	TriangularGeneric(const TriangularGeneric& e) : FiniteElement2DBase(e.mbs) { CopyFrom(e); }

	void SetTriangularGeneric(int bodyindi, const TArray<int>& nodelist, int material_num, double thickness, const Vector3D& coli);

	//To be overwritten in derived class:
	virtual Element* GetCopy()
	{
		Element* ec = new TriangularGeneric(*this);
		return ec;
	}

	void ElementDefaultConstructor()
	{
		massmatrix.Init();
		stiffnessmatrix.Init();
		Hmatrix.Init();
		SetGeometricNonlinearityStatus(GNS_NonlinearSmallStrain);

		nodes.Set3(1,2,3);
		bodyind=1;
		
		integrationRuleStiffness=NULL;
		integrationRuleMass=NULL;
		integrationRuleLoad=NULL;

		perform_initial_value_compuation = false;
	}

	virtual const char* GetElementSpec() const {return "TriangularGeneric";}
	virtual TFiniteElementType GetElementType() const {return TFE_Triangle;}

	virtual int NEdges() const {return 3;}
	virtual const FEEdge & GetLocalEdge(int i) const;
	virtual Vector2D GetEdgeNormal(int edge_nr) const;
	virtual Vector2D GetEdgeNormalD(int edge_nr) const;

	virtual Vector2D GetAreaCoordOfEdgeCoord(int edge, double p_loc_edge) const;

	virtual double GetS0(const Vector2D& ploc, int shape) const; 
	virtual void GetDSMatrix0(const Vector2D& ploc, Matrix& sm) const;
	virtual Vector2D GetNodeLocPos2D(int i) const;

	virtual void DrawElementPreProc();
	virtual void DrawElement();

	virtual Vector2D GetRefLocPos2D() const
	{
		return Vector2D(1. / 3, 1. / 3);
	}

	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	
	// return the information for the vtk output (node indices and element type)
	virtual vtk::vtkElementInfo GetVtkInfo() const;

protected:
	virtual int GetActualInterpolationOrder() const;

	// implementation of IntegrationRuleProvider
	virtual void DefineIntegrationRule(IntegrationRule & integrationRule);
	virtual int DefineIntegrationRuleOrder(IntegrationRule & integrationRule) const;
	virtual void DefineIntegrationRuleEdge(int /*nEdge*/, IntegrationRule & integrationRule) const
	{
		int ruleOrder = DefineIntegrationRuleOrder(integrationRule);
		assert(ruleOrder != 0);
		IntegrationRule::DefineIntegrationRuleLine(integrationRule, ruleOrder);
	}

	//$EK 2017-02-14: thickness now editable through script ("ChangeProperties")
	//EDC double thickness; //$EDC$[varaccess,EDCvarname="thickness",EDCfolder="Geometry",tooltiptext="thickness of finite element (especially used for volumetric loads)"]
}; //$EDC$[endclass,TriangularGeneric]

// these are the actual classes to be used in models
typedef QuadrilateralGeneric<FiniteElement2D> Quadrilateral;
//$EDC$[beginclass,classname=Quadrilateral,noautofunctions,addelementtype=TAEBody+TAENotInRelease+TAE2D+TAEFE,addelementtypename=Quadrilateral,texdescription="description for Quadrilateral"]
//$EDC$[endclass,Quadrilateral]

// these are the actual classes to be used in models
typedef TriangularGeneric<FiniteElement2D> Triangular;
//$EDC$[beginclass,classname=Triangular,noautofunctions,addelementtype=TAEBody+TAENotInRelease+TAE2D+TAEFE,addelementtypename=Triangular,texdescription="description for Triangular"]
//$EDC$[endclass,Triangular]

const char* Triangular::GetElementSpec() const
{
	return "Triangular";
}

const char* Quadrilateral::GetElementSpec() const
{
	return "Quadrilateral";
}