//#**************************************************************
//#
//# filename:             FiniteElement3D.cpp
//#
//# project:              
//#
//# author:               Gerstmayr Johannes, Aigner Larissa
//#
//# generated:						May 11, 2009
//# description:          general 3D FiniteElement
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#include "finiteElement3D.h"
#include "Material.h"
#include "node.h"
#include "solversettings_auto.h"
#include "femathhelperfunctions.h"
#include "rigid3d.h"
#include "rigid3dkardan.h"
#include "femeshinterface.h"
#include "referenceframe3d.h"
#include "basecmselement.h"

// this helper class will simplify browsing through all integration points for stiffness
// with the direct access to the associated data units
class IntPointsStiffnessIterator : public IntegrationPointsIterator
{
	FiniteElement3D * fe;
public:
	IntPointsStiffnessIterator(FiniteElement3D * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	FiniteElement3D::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};
// once again for accessing const data
class IntPointsStiffnessIteratorConst : public IntegrationPointsIterator
{
	const FiniteElement3D * fe;
public:
	IntPointsStiffnessIteratorConst(const FiniteElement3D * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	const FiniteElement3D::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};



double FEFace::GetS0Quad(const Vector3D& ploc, int shape) const
{
	double r = ploc.X();
	double s = ploc.Y();

	//order of nodes:
	//          
	// y     4 +------+ 3     
	// ^       |      | 
	// |       |      |
	// |       |      |
	// --->X 1 +------+ 2     
	//

	if(NFaceNodes() == 4)
	{
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)/4.0; break;
		case 2: return (1.0+r)*(1.0-s)/4.0; break;
		case 3: return (1.0+r)*(1.0+s)/4.0; break;
		case 4: return (1.0-r)*(1.0+s)/4.0; break;
		}
	}
	else if(NFaceNodes() == 8)
	{
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)*(-r-s-1.0)/4.0; break;
		case 2: return (1.0+r)*(1.0-s)*(r-s-1.0)/4.0; break;
		case 3: return (1.0+r)*(1.0+s)*(r+s-1.0)/4.0; break;
		case 4: return (1.0-r)*(1.0+s)*(-r+s-1.0)/4.0; break;
		case 5: return (1.0-r*r)*(1.0-s)/2.0; break;
		case 6: return (1.0-s*s)*(1.0+r)/2.0; break;
		case 7: return (1.0-r*r)*(1.0+s)/2.0; break;
		case 8: return (1.0-s*s)*(1.0-r)/2.0; break;
		}
	}
	assert(0 && "Wrong number of nodes or wrong index.");
	return 0;
};

double FEFace::GetS0Trig(const Vector3D& ploc, int shape) const
{
	double r = ploc.X();
	double s = ploc.Y();

	if (NFaceNodes() == 3)
	{
		switch(shape)
		{
		case 1: return 1.-r-s; break;
		case 2: return r; break;
		case 3: return s; break;
		}
	}
	else if (NFaceNodes() == 6)
	{
		switch(shape)
		{
		case 1: return (1.-r-s)*(2.*(1.-r-s)-1.); break; // 2t^2-t = 2r^2+2s^2+4rs-3r-3s+1
		case 2: return r*(2.*r-1.); break;               // 2r^2-r       
		case 3: return s*(2.*s-1.); break;               // 2s^2-s
		case 4: return 4*r*(1.-r-s); break;              //-4r^2-4rs+4
		case 5: return 4*r*s; break;                     // 4rs
		case 6: return 4*s*(1.-r-s); break;              //-4s^2-4rs+4
		}
	}
	assert(0 && "Wrong number of nodes or wrong index.");
	return 0;
};

void FEFace::GetDSMatrix0Quad(const Vector3D& ploc, Matrix& sm) const
{
	double r = ploc.X();
	double s = ploc.Y();
	sm.SetSize(2, NFaceNodes());

	if (NFaceNodes() == 4)
	{
		sm(1,1) = -(1.0-s)/4.0;
		sm(1,2) = (1.0-s)/4.0;
		sm(1,3) = (1.0+s)/4.0;
		sm(1,4) = -(1.0+s)/4.0;
		sm(2,1) = -(1.0-r)/4.0;
		sm(2,2) = -(1.0+r)/4.0;
		sm(2,3) = (1.0+r)/4.0;
		sm(2,4) = (1.0-r)/4.0;
	}
	else if (NFaceNodes() == 8)
	{
		sm(1,1) = -(1.0-s)*(-r-s-1.0)/4.0 - (1.0-r)*(1.0-s)/4.0;
		sm(1,2) =  (1.0-s)*( r-s-1.0)/4.0 + (1.0+r)*(1.0-s)/4.0;
		sm(1,3) =  (1.0+s)*( r+s-1.0)/4.0 + (1.0+r)*(1.0+s)/4.0;
		sm(1,4) = -(1.0+s)*(-r+s-1.0)/4.0 - (1.0-r)*(1.0+s)/4.0;
		sm(1,5) = -2*r*(1.0-s)/2.0;
		sm(1,6) =  (1.0-s*s)/2.0;
		sm(1,7) = -2*r*(1.0+s)/2.0;
		sm(1,8) = -(1.0-s*s)/2.0;  
		sm(2,1) = -(1.0-r)*(-r-s-1.0)/4.0 - (1.0-r)*(1.0-s)/4.0;
		sm(2,2) = -(1.0+r)*( r-s-1.0)/4.0 - (1.0+r)*(1.0-s)/4.0;
		sm(2,3) =  (1.0+r)*( r+s-1.0)/4.0 + (1.0+r)*(1.0+s)/4.0;
		sm(2,4) =  (1.0-r)*(-r+s-1.0)/4.0 + (1.0-r)*(1.0+s)/4.0;
		sm(2,5) = -(1.0-r*r)/2.0;
		sm(2,6) = -2*s*(1.0+r)/2.0;
		sm(2,7) = (1.0-r*r)/2.0;
		sm(2,8) = -2*s*(1.0-r)/2.0;
	}
	else 
	{
		assert(0 && "Wrong number of nodes or wrong index.");
	}
};

void FEFace::GetDSMatrix0Trig(const Vector3D& ploc, Matrix& sm) const
{
	double r = ploc.X();
	double s = ploc.Y();
	sm.SetSize(2, NFaceNodes());

	if (NFaceNodes() == 3)
	{
		sm(1,1) = -1.; //dS1/dr
		sm(2,1) = -1.; //dS1/ds
		sm(1,2) =  1.; 
		sm(2,2) =  0.;
		sm(1,3) =  0.;
		sm(2,3) =  1.;
	}
	else if(NFaceNodes() == 6)
	{
		//case 1: return (1.-r-s)*(2.*(1.-r-s)-1.); break; // 2t^2-t = 2r^2+2s^2+4rs-3r-3s+1
		//case 2: return r*(2.*r-1.); break;               // 2r^2-r       
		//case 3: return s*(2.*s-1.); break;               // 2s^2-s
		//case 4: return 4*r*(1.-r-s); break;              //-4r^2-4rs+4r
		//case 5: return 4*r*s; break;                     // 4rs
		//case 6: return 4*s*(1.-r-s); break;              //-4s^2-4rs+4s

// TODO: double check, tripple check, quadruple check,   still wrong results for area load testcase
		sm(1,1) = 4.*(r+s)-3.;      //dS1/dr
		sm(2,1) = 4.*(s+r)-3.;      //dS1/ds
		sm(1,2) = 4.*r-1.;         
		sm(2,2) = 0.;
		sm(1,3) = 0.;              
		sm(2,3) = 4.*s-1.;
		sm(1,4) =-8.*r-4.*s+4.;       
		sm(2,4) =-4.*r;
		sm(1,5) = 4.*s;            
		sm(2,5) = 4.*r;
		sm(1,6) =-4.*s;            
		sm(2,6) =-8.*s-4.*r+4.;			
	}
	else
	{
		assert(0 && "Wrong number of nodes or wrong index.");
	}
};







void FiniteElement3D::SetFiniteElement3D(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	FiniteElementGeneric<Body3D>::SetFiniteElementGeneric(bodyindi,nodelist,material_num,coli,CMSelementi);
	UpdateDependentVariables();
};

void FiniteElement3D::UpdateDependentVariables()
{
	assert(!GetMaterial().IsPlanarMaterial());
	ResetOuterFaceFlags();
	SetOuterFaces();
}

void FiniteElement3D::CopyFrom(const Element& e)
{
	FiniteElementGeneric<Body3D>::CopyFrom(e);
	const FiniteElement3D& ce = (const FiniteElement3D&)e;
	outer_face = ce.outer_face;

	this->rotor_angular_velocity = ce.rotor_angular_velocity;
	this->rotor_axis_position = ce.rotor_axis_position;
}

void FiniteElement3D::Initialize() 
{
	FiniteElementGeneric<Body3D>::Initialize();
	ComputeMass();
}

//compute element transformation matrix (jacobian)
void FiniteElement3D::GetJacobi(Matrix3D& jac, const Vector3D& ploc) const
{
	jac.SetSize(3,3);
	int ns = NS();
	int dim = Dim();
	jac.SetAll(0.);
	ConstMatrix<FEmaxDOF*FEmaxDOF> dsMatrix(FlexDOF(),FlexDOF());
	GetDSMatrix0(ploc,dsMatrix);
	// double aa[2][2] = {	{1,1}, {2,2} };		// *YV - commented out
	for (int j = 1; j <= dim; j++)
	{
		for (int k=1; k <= ns; k++)
		{ 
			Vector3D p0 = GetNodeRefPos(k);
			double sum = 0;
			for (int i = 1; i <= dim; i++)
			{ 
				jac(i,j) += dsMatrix(j,k)*p0(i);
			}
		}
	}
}

//compute DSMAtrix and apply element jacobian inverse; also compte determinant of jacobian
double FiniteElement3D::GetJacInvDS(const Vector3D& ploc, Matrix& jacinvDS) const
{
	//jacinvDS should have already Dim() x NS()
	jacinvDS.SetSize(Dim(), NS());

	Matrix3D jac0, jacinv;
	jacinv.SetSize(3,3);

	GetJacobi(jac0,ploc);

	jac0.GetInverse(jacinv);
	jacinv.TpYs();

	GetDSMatrix0(ploc,jacinvDS);
	for (int i=1; i <= NS(); i++)
	{
		Vector3D v(jacinvDS(1,i),jacinvDS(2,i),jacinvDS(3,i));
		v = jacinv*v;
		jacinvDS(1,i) = v(1);
		jacinvDS(2,i) = v(2);
		jacinvDS(3,i) = v(3);
	}
	//Mult(jacinv, DS, grad(i));

	return jac0.Det();
}

void FiniteElement3D::Gradu(const Vector& u, const Matrix& jacinvDS, Matrix3D& gradu) const
{
	gradu.SetAll(0);
	int dim = Dim();
	int l;
	for (int j = 1; j <= dim; j++) 
	{
		for (int i = 1; i <= NS(); i++)
		{
			l = (i-1)*dim+j;
			for (int k = 1; k <= dim; k++)
			{
				gradu(j,k) += jacinvDS(k,i)*u(l);
			}
		}
	}
}

void FiniteElement3D::Gradu(const Vector3D& ploc, const Vector& u, Matrix3D& gradu) const
{
	ConstMatrix<FEmaxDOF> jacinvDS(Dim(),NS());

	GetJacInvDS(ploc, jacinvDS);
	Gradu(u, jacinvDS, gradu);
}

//ploc -1 ... +1
void FiniteElement3D::GraduD(const Vector3D& ploc, Matrix3D& gradu) const
{
	ConstVector<FEmaxDOF> xgd(SOS(),1);		//*YV,JG
	
	for (int i=1; i<=SOS(); i++)
	{
		xgd(i) = XGD(i);
	}
	Gradu(ploc, xgd, gradu);
}

//$EK 2015-07: get all shape functions of a vector; assumes that shape is 
//already set to the correct size
void FiniteElement3D::GetS0(const Vector3D& ploc, Vector& shape) const
{
	for(int i = 1; i <= NS(); ++i)
	{
		shape(i) = GetS0(ploc, i);
	}
}


//set all outer faces active, which have no opposite face from an other element
void FiniteElement3D::SetOuterFaces()
{
	TArray<int> neighbours(20);
	GetNeighbourElements(neighbours);

	ResetOuterFaceFlags();

	for (int i=1; i<=NFaces(); i++)
	{
		FEFace face = GetFace(i);
		face.Invert();

		int hasface = 0;
		for (int j=1; j<=neighbours.Length(); j++)
		{
			int no_surfaceupdate = mbs->GetIOption(119);

			if (GetElement(neighbours(j)).IsFiniteElement() && GetElement(neighbours(j)).Dim()==3 && 
				(GetElement(neighbours(j)).DrawElementFlag()!=0 || no_surfaceupdate!=0) ) //Hex/Tet, drawbody flag is set if it matters
			{
				FiniteElement3D* fe = (FiniteElement3D*)GetElementPtr(neighbours(j));
				//const FiniteElement3D& h=(const FiniteElement3D&)(GetElement(neighbours(j)));
				if (fe->HasFace(face)) 
				{
					hasface = 1;
					break;
				}
			}
		}
		if (!hasface) SetOuterFaceFlag(i);
		//if (GetOuterFaceFlag(i) ) UO() << "face" << i << "=" << GetOuterFaceFlag(i) << "\n";
	}
}

//set all outer faces active, which have no opposite face from an other element (this version also checks visibility due to cutting plane)
void FiniteElement3D::SetOuterFacesCuttingPlane()
{
	TArray<int> neighbours(20);
	GetNeighbourElements(neighbours);

// check if any neighbor has drawflag set but is not drawn due to cutting plane
	int needrecalc = 0;
	for (int i=1; i<=neighbours.Length(); i++)
	{
		if(GetElement(neighbours(i)).DrawElementFlag() != 0)
		{
			if( !mbs->CuttingPlanesAllow(GetElement(neighbours(i)).GetRefPosD()))
			{
				// element has a neighbor that should be drawn due to flag, but is not drawn due to cutting plane
				needrecalc = 1;
				break;
			}
		}
	}
	if (!needrecalc)
		return;

	ResetOuterFaceFlags();

	for (int i=1; i<=NFaces(); i++)
	{
		FEFace face = GetFace(i);
		face.Invert();

		int hasface = 0;
		for (int j=1; j<=neighbours.Length(); j++)
		{
			if (GetElement(neighbours(j)).IsFiniteElement() && GetElement(neighbours(j)).Dim()==3 && 
				GetElement(neighbours(j)).DrawElementFlag()!=0 && mbs->CuttingPlanesAllow(GetElement(neighbours(j)).GetRefPosD())) //Hex/Tet, drawbody flag is set, neighbour is drawn
			{
				FiniteElement3D* fe = (FiniteElement3D*)GetElementPtr(neighbours(j));
				//const FiniteElement3D& h=(const FiniteElement3D&)(GetElement(neighbours(j)));
				if (fe->HasFace(face)) 
				{
					hasface = 1;
					break;
				}
			}
		}
		if (!hasface) SetOuterFaceFlag(i);
		//if (GetOuterFaceFlag(i) ) UO() << "face" << i << "=" << GetOuterFaceFlag(i) << "\n";
	}
}

//-1..+1 based!!!
//FFRF: deformed position relative to floating frame; else: absolute deformed position
Vector3D FiniteElement3D::GetPosRel(const Vector3D& p_loc) const
{
	Vector3D p; //initialized with zeros

	for (int j = 1; j <= NS(); j++)
	{
		double s0 = GetS0(p_loc,j);
		Vector3D p0 = GetNodeRefPos(j);
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += s0*(XG((j-1)*Dim()+i) + p0(i));
		}
	}
	return p;
};

//-1..+1 based!!!
Vector3D FiniteElement3D::GetDisplacementRel(const Vector3D& p_loc) const
{
	Vector3D p; //initialized with zeros

	for (int j = 1; j <= NS(); j++)
	{
		double s0 = GetS0(p_loc,j);
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += s0*XG((j-1)*Dim()+i);
		}
	}
	return p;
};

//-1..+1 based!!!
Vector3D FiniteElement3D::GetVelRel(const Vector3D& p_loc) const
{
	Vector3D p; //initialized with zeros

	for (int j = 1; j <= NS(); j++)
	{
		double s0 = GetS0(p_loc,j);
		for (int i = 1; i <= Dim(); i++)
		{
			double xgp = XGP((j-1)*Dim()+i);
			p(i) += s0*xgp;
		}
	}
	return p;
};

//-1..+1 based!!!
Vector3D FiniteElement3D::GetPosRelD(const Vector3D& p_loc, int use_magnification) const
{
	Vector3D p; //initialized with zeros
	double factor = GetMBS()->GetDOption(105);
	if (!use_magnification) factor = 1;

	for (int j = 1; j <= NS(); j++)
	{
		double s0 = GetS0(p_loc,j);
		Vector3D p0 = GetNodeRefPos(j);
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += s0*(XGD((j-1)*Dim()+i)*factor + p0(i));
		}
	}
	return p;
};

//-1..+1 based!!!
Vector3D FiniteElement3D::GetDisplacementRelD(const Vector3D& p_loc) const
{
	Vector3D p; //initialized with zeros

	for (int j = 1; j <= NS(); j++)
	{
		double s0 = GetS0(p_loc,j);
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += s0*XGD((j-1)*Dim()+i);
		}
	}
	return p;
};

//-1..+1 based!!!
Vector3D FiniteElement3D::GetVelRelD(const Vector3D& p_loc) const
{
	Vector3D p;

	for (int i = 1; i <= Dim(); i++)
	{
		for (int j = 1; j <= NS(); j++)
		{
			p(i) += GetS0(p_loc,j)*(XGPD((j-1)*Dim()+i));
		}
	}
	return p;
};

//-1..+1 based!!!
//relative position for FFRF elements, in reference configuration
Vector3D FiniteElement3D::GetRefConfPos(const Vector3D& p_loc) const
{
	Vector3D p;

	for (int j = 1; j <= NS(); j++)
	{
		Vector3D p0 = GetNodeRefPos(j);
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += GetS0(p_loc,j)*p0(i);
		}
	}
	return p;
}

Vector3D FiniteElement3D::GetNodeRefPos(int i) const //return nodal position of node i in reference configuration
{
	return GetNode(i).RefConfPos();
}

Vector3D FiniteElement3D::GetAngularVel(const Vector3D &ploc) const
{
	Vector3D omega;
	Vector3D ploc2 = ploc + Vector3D(0.5, 0.5, 0.5);
	Vector3D pglob = GetPos(ploc);
	Vector3D pglob2 = GetPos(ploc2);
	Vector3D r(0.), axis(0.);
	Vector3D v1, v2, v;
	double normr, invnormr;

	for (int i=1; i<=3; i++)
	{
		r = pglob2 - pglob;
		normr = r.Norm();
		invnormr = 1./normr;
		// axis is ith unit vector
		axis.Set(0.,0.,0.);
		axis(i) = 1.;
		int j=0;
		// find ploc2 such that pglob2+pglob is not parallel to axis
		while ( fabs(r*axis) > (0.55)*normr && j < 3 ) 
		{
			//global_uo << "not using " << axis << " and " << r << "\n";
			j++;
			Vector3D add(0.);
			add(j) = 1.;
			ploc2 = ploc + add;
			pglob = GetPos(ploc);
			pglob2 = GetPos(ploc2);
			r = pglob2 - pglob;
			normr = r.Norm();
			invnormr = 1./normr;
		}
		if (j==3)
		{
			UO() << "Error in FiniteElement3D::GetAngularVel: could not find direction not parallel to axis " << axis << "\n";
		}
		//double rax = r*axis;
		//r -= rax * axis;
		r(i) = 0;
		v1 = GetVel(ploc);
		v2 = GetVel(ploc2);
		v = v2-v1;
		//double vax = v*axis;
		//v -= vax*axis;
		v(i) = 0;
		double vr = -1./r.Norm2()*(v*r);
		v += vr * r;
		omega(i) = invnormr*v.Norm();
		//global_uo << "Pglob = " << pglob << ", axis = " << axis << " r = " << r << ", vel = " << v << "\n";
	}
	return omega;
}

void FiniteElement3D::GetNodedPosdqT(int node, Matrix& dpdqi)
{
	dpdqi.SetSize(SOS(),Dim());
	dpdqi.FillWithZeros();

	dpdqi((node-1)*3+1,1) = 1; //dnx/dnx
	dpdqi((node-1)*3+2,2) = 1; //dny/dny
	dpdqi((node-1)*3+3,3) = 1; //dny/dny
}

Vector3D FiniteElement3D::GetNodePos(int i) const
{
	return Vector3D(XG((i-1)*3+1), XG((i-1)*3+2), XG((i-1)*3+3)) + GetNodeRefPos(i);
}

Vector3D FiniteElement3D::GetNodePosD(int i) const 
{
	return Vector3D(XGD((i-1)*3+1), XGD((i-1)*3+2), XGD((i-1)*3+3)) + GetNodeRefPos(i);
}

Vector3D FiniteElement3D::GetNodeVel(int i) const
{
	return Vector3D(XGP((i-1)*3+1), XGP((i-1)*3+2), XGP((i-1)*3+3));
}

void FiniteElement3D::ComputeMass()
{
	mass = 0;
	Matrix3D jac;
	for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
	{
		GetJacobi(jac, ip.Point());
		double jacdet = jac.Det();
		mass += fabs(jacdet) * Rho() * ip.Weight();
	}
}

// fill only FlexibleDof entries of massmatrix, other entries remain unchanged
void FiniteElement3D::EvalMff(Matrix& m, double t)
{
	// Ioption 20 -> use elementwise precomputed mass matrix
	if (mbs->GetSolSet().store_FE_matrices && massmatrix.Getcols() == FlexDOF())
	{
		m.SetSubmatrix(massmatrix,1,1);
		return;
	}
	else
	{
		//computed only once and stored afterwards
		int dim = Dim(); 
		int ns = NS();
		int flexdof = FlexDOF();
		m.SetAll(0);

		ConstVector<FEmaxDOF> SF(ns);
		ConstMatrix<FEmaxDOF*FEmaxDOF> SST(flexdof,flexdof);
		Matrix3D jac;
		for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
		{
			GetJacobi(jac, ip.Point());
			double jacdet = jac.Det();
			// compute SF = vector containing shape functions
			for (int i=1; i<=ns; i++)
			{
				SF(i)=GetS0(ip.Point(), i);
			}

			// compute S * S^T  where S is the shape function matrix 
			//     [ SF(1) 0 0 SF(2) 0 0 ...
			// S = [ 0 SF(1) 0 0 SF(2) 0 ..
			//     [ 0 0 SF(1) 0 0 SF(2) ..
			SST.SetAll(0);
			for (int i=0; i<ns; i++)
				for (int j=0; j<ns; j++)
					for (int di=1; di<=dim; di++)
							SST(i*dim+di,j*dim+di) += SF(i+1)*SF(j+1);

			m.AddSubmatrix(SST,1,1,1,1,flexdof,flexdof,
				fabs(jacdet) * GetMaterial().Density() * ip.Weight());
			//massmatrix += fabs (jacdet) * GetMaterial().Density() * w(i1) * (HL*HL.GetTp());
		}

		// fill only FlexibleDof entries of massmatrix, other entries remain unchanged
		if (mbs->GetSolSet().store_FE_matrices) // if mass matrix is stored
		{
			massmatrix.SetSize(flexdof, flexdof);
			for (int i=1; i<=flexdof; i++)
				for (int j=1; j<=flexdof; j++)
					massmatrix(i,j) = m(i,j);
		}
	}
}

void FiniteElement3D::GetIntDuDqFCentrifugal(Matrix& H, const Vector3D& omega, const Vector3D& r0)
{	
	int dim = Dim();
	int ns = NS();
	int sos = SOS();

	H.SetSize(sos,dim);
	H.SetAll(0);
	Matrix3D jac;
	for (IntegrationPointsIterator ip(integrationRuleLoad); !ip.IsEnd(); ++ip)
	{
		GetJacobi(jac, ip.Point());
		double jacdet = jac.Det();
		double fact = fabs (jacdet) * ip.Weight();
		Vector3D r_rel = r0 - GetPos(ip.Point());
		Vector3D force = omega.Cross(omega.Cross(r_rel));

		for (int i = 0; i < ns; i++)
		{
			double combinedfactor = fact*GetS0(ip.Point(), i+1)*Rho();
			H(i*dim+1,1) += combinedfactor * force.X();
			H(i*dim+2,1) += combinedfactor * force.Y();
			H(i*dim+3,1) += combinedfactor * force.Z();
		}
	}
}

void FiniteElement3D::GetH(Matrix& H) 
{
	int ns = NS();
	int dim = Dim();
	if ( Hmatrix.Getrows() == ns*dim)
	{
		H = Hmatrix;
		return;
	}
	else
	{
		H.SetSize(ns*dim,dim);
		H.SetAll(0);
		ConstMatrix<(FEmaxDOF+FFRFsize)*3> dudq;
		Matrix3D jac;
		// flexible part Hf
		for (IntegrationPointsIterator ip(integrationRuleLoad); !ip.IsEnd(); ++ip)
		{
			GetJacobi(jac, ip.Point());
			double jacdet = jac.Det();
			double fact = fabs(jacdet) * ip.Weight();
			for (int i=0; i<ns; i++)
			{
				for (int j=1; j<=dim; j++)
				{
					H(i*dim+j,j) += fact * GetS0(ip.Point(), i+1);
				}
			}
		}
		Hmatrix = H;
	}
}


void FiniteElement3D::GetdPosdqT(const Vector3D& ploc, Matrix& d)
{
	d.SetSize(SOS(),Dim());
	d.FillWithZeros();

	for (int i = 1; i <= NS(); i++)
	{
		double sv =	GetS0(ploc, i);
		d((i-1)*Dim()+1,1) = sv;
		d((i-1)*Dim()+2,2) = sv;
		d((i-1)*Dim()+3,3) = sv;
	}
}


//fill in sos x sos components, m might be larger
//compute stiffness matrix
void FiniteElement3D::StiffnessMatrix(Matrix& m) 
{
	//linear stiffness matrix: for linear problems or component mode synthesis:
	int dim = Dim(); 
	int ns = NS();

	if (
		mbs->GetSolSet().store_FE_matrices &&
		GetGeometricNonlinearityStatus() == GNS_Linear &&
		!GetMaterial().IsInelasticMaterial() &&      //$ PG 2013-8-14: assemble every stiffness matrix in case of inelastic material
		stiffnessmatrix.Getcols() == FlexDOF()
		)
	{
		m.SetSubmatrix(stiffnessmatrix,1,1);
		return;
	}
	m.SetAll(0);

	// B = deps/dq in the linear case, in the nonlinear case B = dE/dq
	ConstMatrix<6*FEmaxDOF> B((dim-1)*3,dim*ns); //filled with zeros
	
	// intermediate quantities
	ConstMatrix<6*FEmaxDOF> CB((dim-1)*3,dim*ns); //filled with zeros
	ConstMatrix<FEmaxDOF*FEmaxDOF> BCB(dim*ns,dim*ns); //filled with zeros

	// Elasticity Matrix
	ConstMatrix<36> C;
	GetMaterial().ComputeElasticityMatrix(C);

	// needed only in geometrically linear case
	Matrix dShapei_dxj;   // shape function jacobians have already been precomputed per integration point (see GetGrad(..)), therefore Matrix instead of ConstMatrix<...>

	// needed only in geometrically nonlinear case:
	Matrix3D F, sigma, strain;  // deformation gradient F, second piola kirchhoff tensor, nonlinear strain tensor 1/2(F^T F - I)
	// dB_nl/dq^T sigma (B_nl is the nonlinear part of B in case of geometric nonlinear problem, see Skript p. 38
	ConstMatrix<FEmaxDOF*FEmaxDOF> dB_nldq_sigma; // Contains dB_nldq_sigma(3*(alpha-1)+j, 3*(beta-1)+k) = dB_nl(i, 3*(alpha-1)+j)/dq(3*(beta-1)+k)*stress(i)

	ConstVector<FEmaxDOF> xgCache;
	if (GetGeometricNonlinearityStatus() != GNS_Linear || GetMaterial().IsInelasticMaterial())
	{
		SetComputeCoordinates(xgCache);
	}

	// Integration routine
	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		//grad contains derivatives in compressed form:
		//grad(1,1..ns) = da/dx     where a = du/dqi or dv/dqi
		//grad(2,1..ns) = da/dy
		GetGrad(ip, dShapei_dxj);

		if (GetGeometricNonlinearityStatus() == GNS_Linear)  // calculation on reference volume with cauchy stress tensor and linearized strain tensor
		{
			B.SetAll(0);

			// Geometrically linearized B matrix
			for (int alpha = 1; alpha <= ns; alpha++)
			{
				B(1,3*(alpha-1)+1) = dShapei_dxj(1,alpha); //du/dx
				B(2,3*(alpha-1)+2) = dShapei_dxj(2,alpha); //dv/dy
				B(3,3*(alpha-1)+3) = dShapei_dxj(3,alpha); //dw/dz

				B(4,3*(alpha-1)+2) = dShapei_dxj(3,alpha); //dv/dz, eps_yz
				B(4,3*(alpha-1)+3) = dShapei_dxj(2,alpha); //dw/dy
				B(5,3*(alpha-1)+3) = dShapei_dxj(1,alpha); //dw/dx, eps_xz
				B(5,3*(alpha-1)+1) = dShapei_dxj(3,alpha); //du/dz
				B(6,3*(alpha-1)+1) = dShapei_dxj(2,alpha); //du/dy, eps_xy
				B(6,3*(alpha-1)+2) = dShapei_dxj(1,alpha); //dv/dx
			}

			if (GetMaterial().IsInelasticMaterial() && GetMaterial().GetInelasticitySolutionMethod() == Material::ISM_ReturnMapping && !perform_initial_value_compuation)
			{
				// compute linearized strain tensor
				ConstVector<6> strain_vector = (ConstVector<6>&) (B*xgCache);   // (eps_xx, eps_yy, eps_zz, 2 eps_yz, 2 eps_zx, 2 eps_xy)

				if(GetMaterial().HasInitialStrains())
				{
					Matrix3D initial_strain;
					GetMaterial().GetInitialStrain(initial_strain);
					ConstVector<6> initial_strain_vector;
					Matrix3DToStrainVector(initial_strain_vector, initial_strain);
					strain_vector -= initial_strain_vector;
				}

				// get inelastic variables
				ConstVector<MAXInelasticVariablesCount> inelastic_variables;
				DataLastStepToInelasticVariables(inelastic_variables, ip.GetIndex());   // (eps^p_xx, eps^p_yy, eps^p_zz, 2 eps^p_yz, 2 eps^p_zx, 2 eps^p_xy, hardening_parameter)

				// compute consistent tangent stiffness matrix
				GetMaterial().ComputeConsistentTangentStiffnessMatrix(C, strain_vector, inelastic_variables);  // C as input: elasticity tensor, as output: consistent tangent matrix C = C (I - 2 mu d actual_plastic_strain/d strain)
			}
		}
		else
		{
			B.SetAll(0.);
			dB_nldq_sigma.SetSize(dim*ns,dim*ns);  
			dB_nldq_sigma.SetAll(0);

			// compute actual displacement gradient
			Gradu(xgCache, dShapei_dxj, F);
			// add I matrix to obtain deformation gradient
			F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;

			// Green-Lagrange strain tensor
			//strain = 0.5 * (F.GetTp() * F);
			F.GetATA2(strain);
			strain(1,1) -= 0.5; strain(2,2) -= 0.5; strain(3,3) -= 0.5;

			if (IsInelasticMaterial())
			{
				mbs->UO(UO_LVL_warn) << "WARNING: As yet, inelastic material behavior has been implemented for linear geometry only!\n";
			}
			
			GetMaterial().ComputeStressFromStrain(strain, sigma);

			for (int alpha = 1; alpha <= ns; alpha++)
			{
				// nonlinear B matrix
				// Variation of nonlinear Green Lagrange strain tensor E
				for (int k=1; k<=dim; k++)
				{
					int alphak = 3*(alpha-1)+k;
					B(1,alphak) = F(k,1)*dShapei_dxj(1,alpha);   // duk/dx1 * dShape_alpha/dx1
					B(2,alphak) = F(k,2)*dShapei_dxj(2,alpha);   // duk/dx2 * dShape_alpha/dx2
					B(3,alphak) = F(k,3)*dShapei_dxj(3,alpha);   // duk/dx3 * dShape_alpha/dx3
					B(4,alphak) = F(k,2)*dShapei_dxj(3,alpha)+F(k,3)*dShapei_dxj(2,alpha);  // duk/dx2 * dShape_alpha/dx3 + duk/dx3 * dShape_alpha/dx2
					B(5,alphak) = F(k,1)*dShapei_dxj(3,alpha)+F(k,3)*dShapei_dxj(1,alpha);  
					B(6,alphak) = F(k,2)*dShapei_dxj(1,alpha)+F(k,1)*dShapei_dxj(2,alpha);  
				}

				//// entries of dB_nl/dq^T sigma
				// dB_nldq^T sigma (alpha j, beta j) = Sum_k dBnl(k, alphaj) / dq(betaj) * sigma(k) = add_dBdq_sigma is independent of j
				for (int beta=1; beta<=ns; beta++)
				{
					double add_dBdq_sigma = 0;
					add_dBdq_sigma += dShapei_dxj(1,beta)*dShapei_dxj(1,alpha) * sigma(1,1);
					add_dBdq_sigma += dShapei_dxj(2,beta)*dShapei_dxj(2,alpha) * sigma(2,2);
					add_dBdq_sigma += dShapei_dxj(3,beta)*dShapei_dxj(3,alpha) * sigma(3,3);
					add_dBdq_sigma += (dShapei_dxj(2,beta)*dShapei_dxj(3,alpha)+dShapei_dxj(3,beta)*dShapei_dxj(2,alpha)) * sigma(3,2);
					add_dBdq_sigma += (dShapei_dxj(3,beta)*dShapei_dxj(1,alpha)+dShapei_dxj(1,beta)*dShapei_dxj(3,alpha)) * sigma(1,3);
					add_dBdq_sigma += (dShapei_dxj(1,beta)*dShapei_dxj(2,alpha)+dShapei_dxj(2,beta)*dShapei_dxj(1,alpha)) * sigma(1,2);

					for (int j=1; j<=dim; j++)
					{
						int alphaj = 3*(alpha-1)+j;
						int betaj = 3*(beta-1)+j;
						dB_nldq_sigma(alphaj, betaj) += add_dBdq_sigma; 
					}
				}
			}
			if (GetMaterial().IsInelasticMaterial() && GetMaterial().GetInelasticitySolutionMethod() == Material::ISM_ReturnMapping)
			{
				GetMBS()->UO(UO_LVL_warn) << "WARNING: FiniteElement3D: Computation of Consistent Tangent Stiffness Matrix not implemented for large strain case!\n";
			}
		}

		// compute B^T C B
		Mult(C, B, CB);
		MultTp(B, CB, BCB);

		// Stiffness matrix linear:    K = B^T C B,
		//                  nonlinear: K = B_hat ^T C B_hat + dB_hat/dq sigma

		// integration weight factor
		double fact = -fabs(ip.Data().jacdet) * ip.Weight();

		// Add part  B^T C B 
		// AddSubmatrix(mat, offset rows submat, offset cols submat, offset rows mat, offset cols mat, # rows, # cols, factor)
		m.AddSubmatrix(BCB, 1, 1, 1, 1, ns*dim, ns*dim, fact);

		// geometrically nonlinear case:
		// Add part  dB/dq^T  sigma
		if (GetGeometricNonlinearityStatus() != GNS_Linear)
		{
			m.AddSubmatrix(dB_nldq_sigma, 1, 1, 1, 1, ns*dim, ns*dim, fact);
		}
	}

	// ++++++ centrifugal stiffness due to rotatoric motion with constant angular velocity omega ++++++
	if (rotor_angular_velocity.Norm() > 0.)
	{
		ConstMatrix<FEmaxDOF*FEmaxDOF> stiffness_centrifugal(ns*dim, ns*dim);
		StiffnessMatrixCentrifugal(stiffness_centrifugal);
		m.AddSubmatrix(stiffness_centrifugal, 1, 1, 1, 1, ns*dim, ns*dim, 1.);
	}

	if (
		mbs->GetSolSet().store_FE_matrices &&
		GetGeometricNonlinearityStatus() == GNS_Linear &&
		!(GetMaterial().IsInelasticMaterial() && GetMaterial().GetInelasticitySolutionMethod() == Material::ISM_ReturnMapping)
		) // if stiffness matrix is stored
	{
		stiffnessmatrix.SetSize(FlexDOF(), FlexDOF());
		for (int i = 1; i <= FlexDOF(); i++)
			for (int j = 1; j <= FlexDOF(); j++)
				stiffnessmatrix(i,j) = m(i,j);
	}
}

double FiniteElement3D::GetVolume() const
{
	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);
	ConstMatrix<FEmaxDOF> agrad(Dim(),NS());
	Matrix3D F;

	double volume = 0.;

	// Integration routine
	for (IntPointsStiffnessIteratorConst ip(this); !ip.IsEnd(); ++ip)
	{		
		GetGrad(ip, agrad);
		Gradu(xgCache, agrad, F);
		F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;
		volume += fabs(F.Det() * ip.Data().jacdet) * ip.Weight();
	}
	return fabs(volume);
}

double FiniteElement3D::GetRefVolume() const
{
	double volume = 0.;
	// Integration routine
	for (IntPointsStiffnessIteratorConst ip(this); !ip.IsEnd(); ++ip)
	{
		double fact = fabs(ip.Data().jacdet) * ip.Weight();
		volume += fact;
	}
	return fabs(volume);
}

// $EK 2013-03-21 compute center of mass via integration
void FiniteElement3D::GetRefCenterOfMass(Vector3D& center) const
{
	center.Set(0,0,0);
	// Integration routine
	for (IntPointsStiffnessIteratorConst ip(this); !ip.IsEnd(); ++ip)
	{
		double fact = fabs(ip.Data().jacdet) * ip.Weight();
		center += fact* GetPosRel(ip.Point());
		
	}
	center *= 1./this->GetRefVolume();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
bool FiniteElement3D::GetLocPosOfGlobalPos(Vector3D& p_loc, const Vector3D& p_glob, int niter) const
{
	static const double eps = 1e-7;

	Vector3D pos;
	Matrix3D jac, jacinv;

	for (int i = 1; i <= niter; ++i)
	{
		pos = GetPos(p_loc);
		GetJacobi(jac, p_loc);
		jac.GetInverse(jacinv);
		p_loc = p_loc - jacinv*(pos - p_glob);
	}
	return (abs(p_loc.X()) <= 1 + eps) && (abs(p_loc.Y()) <= 1 + eps) && (abs(p_loc.Z()) <= 1 + eps);
}

bool FiniteElement3D::GetLocPosOfGlobalPosPreAssemble(Vector3D& p_loc, const Vector3D& p_glob, int niter) const
{
	static const double eps = 1e-7;
//TODO: is this Hex-specific ?
	TArray<Vector3D> nodepositions;
	for(int k=1; k<=NNodes(); k++)
	{
		Vector3D npos = ((GenericNode&)GetNode(k)).GetRefPos();
		nodepositions.Add( npos );
	}

// compare to MeshElement::GetLocPosOfGlobalPos
	p_loc = Vector3D(0.,0.,0.);
	for(int i = 1; i <= niter; ++i)
	{
		
// compare to MeshElement::InterpolateVector3D
		Vector3D ptmp(0.,0.,0.);
		for(int k=1; k<=NNodes(); k++)
		{
			ptmp = ptmp + GetS0(p_loc,k)*nodepositions(k);
		}

// compare to MeshElement::GetJacobi
		Matrix3D jac(0.,0.,0., 0.,0.,0., 0.,0.,0.);
		Matrix dsmat;
		GetDSMatrix0(p_loc,dsmat);
		for(int j=1; j<=Dim(); j++)
		{
			for(int k=1; k<=NNodes(); k++)
			{
				Vector3D npos = nodepositions(k);
				for(int l=1; l<=Dim(); l++)
				{
					jac(l,j) += dsmat(j,k)*npos(l);
				}
			}
		}
		jac.Invert();
		p_loc = p_loc - /*inverted*/jac*(ptmp-p_glob);
	}

	return (abs(p_loc.X()) <= 1 + eps) && (abs(p_loc.Y()) <= 1 + eps) && (abs(p_loc.Z()) <= 1 + eps);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void FiniteElement3D::AddAreaLoad(Vector& f, double pressure, int facet) 
{
	IntegrationRule faceRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
	DefineIntegrationRuleFace(facet, faceRule);

	ConstVector<FEmaxDOF> xgCache;
	if (GetGeometricNonlinearityStatus() != GNS_Linear)
		SetComputeCoordinates(xgCache);

	for (IntegrationPointsIterator ip(&faceRule); !ip.IsEnd(); ++ip)
	{
	// surface normal
		Matrix3D jac;
		GetFaceRefJacobi(facet, jac, ip.Point());
		Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
		Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
		Vector3D n = dS1.Cross(dS2);
		Vector3D n0 = n;
		n0.Normalize();
		double absn = n.Norm();
/*
		Vector3D p_loc = ip.Point();
		FEFace face = GetLocalFace(facet);

		for (int i=1; i<=NSFace(facet); i++)
		{
			double fact = GetS0Face(facet, p_loc, i) * pressure * ip.Weight() * absn;

			int iel = face.Node(i); // node number of ith face node in element

			for(int j=1; j<=Dim(); j++)
			{
				f(Dim()*(iel-1)+j) += fact*n0(j);
			}
		}
*/

		double pressure_ip = pressure;
		if (GetGeometricNonlinearityStatus() != GNS_Linear)
		{
			//$ PG 2018-2-23: [
			// surface force acts on the body in current configuration and has to be integrated with surface measure from there;
			// however, we perform integration at reference configuration, whence surface force vector must be rescaled 
			//   f_ref = f_cur |det(F) F^{-T} N_ref|   
			// with F being the deformation gradient from reference to current configuration, and N_ref denoting the unit normal in reference configuration,
			// see e.g., "Mathematische Methoden in der Kontinuumsmechanik", Heinz Engl, eq.(1.3.87)))
			
			Matrix3D F, FInv;
			Matrix agrad;
			GetGrad(ip, agrad);
			Gradu(xgCache, agrad, F);
			F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;

			F.GetInverse(FInv); 
			Vector3D m = n*FInv;
			pressure_ip *= m.Norm()*fabs(F.Det());
			
			//$ PG 2018-2-23: ]
		}


		Vector2D p_loc_face = ip.Point2D();
		Vector3D p_loc_vol = GetVolumeCoordOfFaceCoord(facet, p_loc_face);

 		//FEFace face = GetLocalFace(facet);
		//for(int k=1 k<=NSFace(facet); k++)  // only the shape functions associated 
		//{
		//	int i = face.G

		for(int i=1; i<=NS(); i++)
		{
			double fact = 0.;
			if(1)
			{
				fact = GetS0(p_loc_vol,i) * pressure_ip * ip.Weight() * absn;

				for(int j=1; j<=Dim(); j++)
				{
					f(Dim()*(i-1)+j) += fact*n0(j);
				}
			}
		}
	}
}

void FiniteElement3D::AddAreaTraction(Vector& f, Vector3D fext, int local, int facet)
{
	IntegrationRule faceRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
	DefineIntegrationRuleFace(facet, faceRule);

	ConstVector<FEmaxDOF> xgCache;
	if (GetGeometricNonlinearityStatus() != GNS_Linear)
		SetComputeCoordinates(xgCache);

	for (IntegrationPointsIterator ip(&faceRule); !ip.IsEnd(); ++ip)
	{
	// surface normal - reference configuration
		Matrix3D jac;
		GetFaceRefJacobi(facet, jac, ip.Point());
		Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
		Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
		Vector3D n = dS1.Cross(dS2);
		double absn = n.Norm(); 


		Vector3D fglob = fext;
		if(local) 
		{		
			//$ PG 2018-2-23: [ follower force must be mapped onto parameterized surface of mesh block, not trivial! we for the moment forbid usage of follower loads!
			mbs->UO(UO_LVL_err) << "ERROR: follower surface load not allowed in FiniteElement3D!\n";
			
			//// convert external force fext ( local coordinates ) to fglob ( global coordinates )
			//Vector3D dS10 = dS1; dS10.Normalize();
			//Vector3D dS20 = dS2; dS20.Normalize();
			//Vector3D n0 = n; n0.Normalize();

			//fglob = dS10*fext.X() + dS20*fext.Y() + n0*fext.Z(); 
			//$ PG 2018-2-23: ]
		}
		if (GetGeometricNonlinearityStatus() != GNS_Linear)
		{
			//$ PG 2018-2-23: [
			// surface force acts on the body in current configuration and has to be integrated with surface measure from there;
			// however, we perform integration at reference configuration, whence surface force vector must be rescaled 
			//   f_ref = f_cur |det(F) F^{-T} N_ref|   
			// with F being the deformation gradient from reference to current configuration, and N_ref denoting the unit normal in reference configuration,
			// see e.g., "Mathematische Methoden in der Kontinuumsmechanik", Heinz Engl, eq.(1.3.87)))
			
			Matrix3D F, FInv;
			Matrix agrad;
			GetGrad(ip, agrad);
			Gradu(xgCache, agrad, F);
			F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;

			F.GetInverse(FInv); 
			Vector3D m = n*FInv;
			fglob *= m.Norm()*fabs(F.Det())/absn;
			
			//$ PG 2018-2-23: ]
		}
		
		Vector2D p_loc_face = ip.Point2D();
		Vector3D p_loc_vol = GetVolumeCoordOfFaceCoord(facet, p_loc_face);

		for(int i=1; i<=NS(); i++)
		{
			double fact = GetS0(p_loc_vol,i) * ip.Weight() * absn;

			if(local)
			{
				f(Dim()*(i-1)+1) += fglob.X()* fact;
				f(Dim()*(i-1)+2) += fglob.Y()* fact;
				if(Dim()==3) { f(Dim()*(i-1)+3) += fglob.Z()* fact; }
			}
			else
			{
				// force fext is already in global coordinate system
				f(Dim()*(i-1)+1) += fglob.X()* fact;
				f(Dim()*(i-1)+2) += fglob.Y()* fact;
				if(Dim()==3) { f(Dim()*(i-1)+3) += fglob.Z()* fact; }

			} // if local else
		} // for i
	} // for ip
}

void FiniteElement3D::AddElementSurfaceLoad(Vector& f, Vector3D fvec_perunit, int flag_pressure, int facet)
{
	if(flag_pressure)
	{
		AddAreaLoad(f, fvec_perunit.Z(), facet);
	}
	else
	{
		AddAreaTraction(f, fvec_perunit, 0, facet);
	}


}

Vector3D FiniteElement3D::GetSurfaceNormal(int facet)
{
	FEFace locface = GetLocalFace(facet);

	Vector3D p1D = GetNodePos(locface.Node(1)); 
	Vector3D p2D = GetNodePos(locface.Node(2)); 
	Vector3D pnD = GetNodePos(locface.Node(locface.NFaceNodes()));

	Vector3D v12 = p2D-p1D;
	Vector3D v1n = pnD-p1D;
	Vector3D n0 = v12.Cross(v1n);
	n0.Normalize();
	return n0;
}
Vector3D FiniteElement3D::GetSurfaceNormalD(int facet)
{
	FEFace locface = GetLocalFace(facet);

	Vector3D p1D = GetNodePosD(locface.Node(1)); 
	Vector3D p2D = GetNodePosD(locface.Node(2)); 
	Vector3D pnD = GetNodePosD(locface.Node(locface.NFaceNodes()));

	Vector3D v12 = p2D-p1D;
	Vector3D v1n = pnD-p1D;
	Vector3D n0 = v12.Cross(v1n);
	n0.Normalize();
	return n0;
}


//compute element transformation matrix (jacobian) from unit element to element in current (deformed) configuration
void FiniteElement3D::GetFaceJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const
{
	jac.SetSize(3,2);
	int ns = NSFace(nFace);
	int dim = Dim();
	jac.SetAll(0.);
	ConstMatrix<2*8> dsMatrix(2,ns); // size of matrix is at most 2 * 8 since there are no more than 8 nodes at a face
	GetDSMatrix0Face(nFace, ploc, dsMatrix);
	const FEFace& face = GetLocalFace(nFace);
	for (int j = 1; j <= 2; j++)
	{
		for (int k = 1; k <= ns; k++)
		{ 
			Vector3D p = GetNodePos(face.Node(k));
			for (int i = 1; i <= dim; i++)
			{ 
				jac(i,j) += dsMatrix(j,k)*p(i);
			}
		}
	}
}

//compute element transformation matrix (jacobian) from unit element to element in reference configuration
void FiniteElement3D::GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const
{
	jac.SetSize(3,2);
	int ns = NSFace(nFace);
	int dim = Dim();
	jac.SetAll(0.);
	ConstMatrix<2*8> dsMatrix(2,ns); // size of matrix is at most 2 * 8 since there are no more than 8 nodes at a face
	GetDSMatrix0Face(nFace, ploc, dsMatrix);
	const FEFace& face = GetLocalFace(nFace);
	for (int j = 1; j <= 2; j++)
	{
		for (int k = 1; k <= ns; k++)
		{ 
			Vector3D p0 = GetNodeRefPos(face.Node(k));
			for (int i = 1; i <= dim; i++)
			{ 
				jac(i,j) += dsMatrix(j,k)*p0(i);
			}
		}
	}
}
//void Finitelement3D::GetFaceJacobi_Def(int nFace, Matrix3D& jac, const Vector3D& ploc) const


void FiniteElement3D::GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const
{
	shape.SetLen(NSFace(nFace));
	for(int i = 1; i <= NSFace(nFace); ++i)
	{
		shape(i) = GetS0Face(nFace, ploc, i);
	}
}

//absolute deformed position of face ip
Vector3D FiniteElement3D::GetPosFace(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPos(locpos3D);
}

//absolute deformed position of face ip
Vector3D FiniteElement3D::GetPosFaceD(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPosD(locpos3D,1);
}

double FiniteElement3D::GetFaceArea(int face_nr) const
{
	IntegrationRule faceRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
	DefineIntegrationRuleFace(face_nr, faceRule);
	double area = 0.;
	Matrix3D jac;
	for (IntegrationPointsIterator ip(&faceRule); !ip.IsEnd(); ++ip)
	{
		GetFaceRefJacobi(face_nr, jac, ip.Point());
		Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
		Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
		area += ip.Weight()*dS1.Cross(dS2).Norm();
	}
	return area;
}

void FiniteElement3D::GradScal(const Vector& s, const Matrix& jacinvDS, 
																		 Vector3D& gradScal, int ) const
{
	gradScal.Set(0., 0., 0.);
	int dim = Dim();
	for (int i = 1; i <= NS(); i++)
	{
		for (int k = 1; k <= dim; k++)
		{
			gradScal(k) += jacinvDS(k,i)*s(i);
		}
	}
}

int FiniteElement3D::NSFace(int nFace) const
{
	return GetFace(nFace).NFaceNodes();
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//COMPUTE RESIDUAL       COMPUTE RESIDUAL       COMPUTE RESIDUAL       COMPUTE RESIDUAL       COMPUTE RESIDUAL       COMPUTE RESIDUAL
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void FiniteElement3D::EvalF2(Vector& f, double t) 
{
	//if (mbs->IsJacobianComputation())
	//{
	//	int a = 2;
	//}

	//add loads, constraint forces and other things in parent function:
	Body3D::EvalF2(f,t);
	TMStartTimer(22);

	int sos = SOS();

	ConstVector<FEmaxDOF> fadd;
	fadd.SetLen(sos);
	fadd.SetAll(0);

	if (GetGeometricNonlinearityStatus() == GNS_Linear || IsFFRF())
	{
		EvalF2GeomLin(fadd, t);
	}
	else
	{
		EvalF2GeomNonlin(fadd, t);
	}

	if (IsFFRF())
	{
		AddQuadraticVelocityVector(fadd, t);
	}	

	f -= fadd;
	TMStopTimer(22);


	// +++++++ damping: +++++++
	if (GetMaterial().DampingM() != 0)
	{
		// set velocities into a temporary vector, modified by YV:
		ConstVector<FEmaxDOF> velocities;
		velocities.SetLen(sos);
		for (int i = 1; i <= sos; i++)
			velocities(i) = XGP(i);

		ConstVector<FEmaxDOF> temp;
		temp.SetLen(sos);
		temp.SetAll(0.);

		if (mbs->GetSolSet().store_FE_matrices) // mass matrix is stored
		{
			if (massmatrix.Getcols() != sos) {UO() << "ERROR: finite element: mass matrix not built!\n"; return;}
			Mult(massmatrix,velocities,temp);
		}
		else // mass matrix is not stored
		{
			ConstMatrix<FEmaxDOF*FEmaxDOF> m(FlexDOF(),FlexDOF());
			EvalMff(m,t);
			Mult(m,velocities,temp);
		}
		temp *= GetMaterial().DampingM();
		f -= temp;
	}
}; 


void FiniteElement3D::EvalF2GeomNonlin(Vector& fadd, double t) 
{
	//add loads, constraint forces and other things in parent function:
	int sos = SOS();

	int ns = NS();
	int dim = Dim();

	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);

	Matrix3D strain, piola1, F;

	ConstVector<FEmaxDOF> temp;
	temp.SetLen(sos);
	temp.SetAll(0.);

	//compute deformation gradient, compute strain, compute stress and compute elastic forces:
	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		//int_V (P : d F / d q) dV
		//P = F * S
		//S = D : E
		//E = 1/2*(F^T F-I)

		Matrix agrad;
		GetGrad(ip, agrad);

		Gradu(xgCache, agrad, F);
		F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;

		// Green-Lagrange strain tensor
		//strain = 0.5 * (F.GetTp() * F);
		F.GetATA2(strain);
		strain(1,1) -= 0.5; strain(2,2) -= 0.5; strain(3,3) -= 0.5;
		
		ComputeStressFromStrain(piola1, strain, ip);  // returns cauchy stress instead of piola1

		piola1 = F * piola1;  //piola1 = F * piola2

		for (int j=1; j <= Dim(); j++)
		{
			for (int i = 1; i <= ns; i++)
			{
				temp(3*(i-1)+j) = agrad(1, i)*piola1(j,1)
					+ agrad(2, i)*piola1(j,2)
					+ agrad(3, i)*piola1(j,3);
			}
		}
		fadd.MultAdd(fabs(ip.Data().jacdet) * ip.Weight(), temp);
	}
}; 


void FiniteElement3D::EvalF2GeomLin(Vector& fadd, double t) 
{
	//add loads, constraint forces and other things in parent function:
	int sos = SOS();

	int ns = NS();
	int dim = Dim();

	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);

	//double u;

	Matrix3D strain, stress, gradu, F;

	ConstVector<FEmaxDOF> temp;
	temp.SetLen(sos);
	temp.SetAll(0.);

	//compute deformation gradient, compute strain, compute stress and compute elastic forces:
	Matrix agrad;
	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		//int_V (P : d grad u / d q) dV
		//P = D : E
		//E = 1/2*(grad u + grad u^T)
		
		GetGrad(ip, agrad);

		Gradu(xgCache, agrad, gradu);
		//EK 2012-06-27 is it necessary in the linear case to compute F ??? -> NO!
		//F = gradu;
		//F(1,1) += 1.; F(2,2) += 1.; F(3,3) += 1.;

		// Green-Lagrange strain tensor
		strain = 0.5 * (gradu.GetTp() + gradu);
		
		ComputeStressFromStrain(stress, strain, ip);

		for (int j=1; j <= dim; j++)
		{
			for (int i = 1; i <= ns; i++)
			{
				int k = dim*(i-1)+j;
				temp(k) = agrad(1, i)*stress(j,1)
					+ agrad(2, i)*stress(j,2)
					+ agrad(3, i)*stress(j,3);
				fadd(k) += fabs(ip.Data().jacdet) * ip.Weight() * temp(k);
			}
		}
	}
}; 

double FiniteElement3D::PostNewtonStep(double t)
{
	if (!IsInelasticMaterial()) return 0;

	double err = 0.; // error for nonlinear iteration
	int sos = FlexDOF();
	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);
	Matrix3D F,jac;

	//----------------------
	// loop over integration points
	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		// Jacobi-matrix and Jacobi-determinant for integration point p
		GetJacobi(jac,ip.Point());
		double jacdet = jac.Det();

		// compute F = grad u + I
		Matrix agrad;
		GetGrad(ip, agrad);
		Gradu(xgCache, agrad, F);
		F(1,1) += 1; F(2,2) += 1; F(3,3) += 1;

		// compute strain from F
		Matrix3D strain;
		if ( GetGeometricNonlinearityStatus() == GNS_Linear || IsFFRF())
		{
			strain = 0.5*(F.GetTp() + F) - Matrix3D(1.);
		}
		else
		{
			strain = 0.5*(F.GetTp()*F - Matrix3D(1.));
		}

		double ip_err = ComputeInelasticVariablesUpdateFromStrain(strain, ip);
		
		err += fabs(jacdet)*ip.Weight()*ip_err;
	}

	return err + FiniteElementGeneric<Body3D>::PostNewtonStep(t);
}

void FiniteElement3D::PostprocessingStep()
{
}


void FiniteElement3D::DrawElementPreProc() 
{
	if (!GetMBS()->GetIOption(118)) return;

	if (HasOuterFaces() || !GetMBS()->GetIOption(146)||GetMBS()->GetIOption(149)) //do not average stress at nodes if element does not have visible faces //$ DR 2012-09-20: bugfix, added GetMBS()->GetIOption(149)
	{
		FieldVariableDescriptor * fvd = GetMBS()->GetActualPostProcessingFieldVariable();
		if(fvd != NULL)
		{
			for (int i=1; i <= NNodes(); i++)
			{
				GetNode(i).GetDrawTemp() += GetFieldVariableValue(*fvd, (1./sqrt(3.))*GetNodeLocPos(i), true);
				GetNode(i).GetDrawTempCnt()++;
			}
		}
	}
};

void FiniteElement3D::DrawElement() 
{
	//$ YV 2012-12-11: the treatment of cutting planes and outer faces was placed here from mbs
	char backUpOfOuterFaces = Outer_Face();

	mbs->Graphics3D()->SetDrawOutline(1);
	int no_surfaceupdate = mbs->GetIOption(119);
	if (mbs->UseCuttingPlanes() && no_surfaceupdate == 0)
	{
		SetOuterFacesCuttingPlane();
	}
	//$ EK 2013-10-20: simplification
	//$ DR 2011-12-02:[ do not draw element if no surface at all
	//bool draw_this_element = 0;
	//if(!(GetMBS()->GetIOption(146))) 
	//{
	//	draw_this_element = 1;		// not only surface elements are drawn
	//}
	//else												// only surface elements are drawn
	//{

	//	//not necessary //EK only decisive if there is an outer flag
	//	/*for (int side=1; (side <= NFaces())&&(!draw_this_element) ; side++)
	//	{
	//		if (GetOuterFaceFlag(side)) 
	//		{
	//			draw_this_element = 1;
	//		}
	//	}*/
	//	draw_this_element = HasOuterFaces();	
	//}
	//$ DR 2011-12-02:] do not draw element if no surface at all
	//if(draw_this_element)
	//EK simpmlify
	if (!(GetMBS()->GetIOption(146)) || HasOuterFaces())
	{

		mbs->Graphics3D()->SetColor(col);

		//SetDrawCoordinates(); //for computation of GraduD; 12.4.9 ***not necessary any more!

		int linemode = 1; //0=no lines, 1=outline+color, 2=outline, 3=elementline+color, 4=elementline
		if (GetMBS()->GetIOption(110) && !GetMBS()->GetIOption(111))
		{
			linemode = 2;
		}
		if (!GetMBS()->GetIOption(110) && GetMBS()->GetIOption(111))
		{
			linemode = 0;
		}

		int colormode = 0;
		const FieldVariableDescriptor * fv = GetMBS()->GetActualPostProcessingFieldVariable();
		if (fv != NULL && HasFieldVariable(*fv))
		{
			colormode = 1;
		}

		//if(col==Vector3D(0.6,0.6,0.6)) colormode = 0;		//$!DR 2011-05-26: Hack: if body is grey: draw body color, for all the other bodies: draw mises or what ever else is defined

		int res = GetMBS()->GetDrawResolution()-1;
		if (res == -1) res = 0;

		double tilex = pow(2.,res);
		static TArray<Vector3D> points;
		points.SetLen((int)(tilex+1)*(int)(tilex+1));
		static TArray<double> vals;
		vals.SetLen((int)(tilex+1)*(int)(tilex+1));

		double v=0;
		Vector3D p0, vx, vy;

		double shrink = GetMBS()->GetDOption(106);//0.995;

		for (int side=1; side <= NFaces(); side++)
		{
			if (GetOuterFaceFlag(side) || !GetMBS()->GetIOption(146)) //draw surface elements only?
			{
				points.SetLen(0); vals.SetLen(0);
				Vector3D p0, vx, vy, vxy;

				GetLocalFaceCoordinates(side, p0, vx, vy, vxy);
				vx /= tilex;
				vy /= tilex;
				vxy /= tilex;

				if (shrink != 1) 
				{
					if (p0.Norm() != 0)
						p0 *= shrink; //hexahedral
					else
						p0 = Vector3D(0.5*(1.-shrink)); //tetrahedral

					vx *= shrink;
					vy *= shrink;
				}

				for (double iy = 0; iy <= tilex; iy++)
				{
					for (double ix = 0; ix <= tilex; ix++)
					{
						Vector3D ploc;
						if (GetElementType() == TFE_Tetrahedral)
						{
							ploc = (p0+ix*vx*(1.-iy/tilex)+iy*vy);
							if (side == 4) ploc = (p0+(1.-ix/tilex)*tilex*vx*(1.-iy/tilex)+ix*vxy*(1.-iy/tilex)+iy*vy);
						}
						//$ EK 2013-03-05 special rule for prism
						else if (GetElementType() == TFE_Prism)
						{
							if (side <= 3) // quadrilateral sides
								ploc = (p0+ix*vx+iy*vy);
							else // triangluar sides
								//ploc = (p0+ix*vx*(1.-iy/tilex)+iy*vy);
								ploc = (p0+ix*vx*(1.-iy/tilex)+iy*vy);
						}
						// $EK 2013-03-05 special rule for pyramid
						else if (GetElementType() == TFE_Pyramid)
						{
							if (side <= 1) // quadrilateral sides
							{	
								ploc = (p0+ix*vx+iy*vy); 
							}
							else // triangluar sides
							{	
								ploc = (p0+ix*vx*(1.-iy/tilex)+iy*vy);
							}
						}
						else
						{
							ploc = (p0+ix*vx+iy*vy);
						}
						points.Add(GetPosD(ploc));


						if (colormode)
						{
							if (!GetMBS()->GetIOption(118)) //compute stress at every local position
							{
								if (colormode)
									v = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), ploc, true);
							}
							else //stress interpolated from nodes!
							{
								v = 0;
								for (int i=1; i <= NNodes(); i++)
								{
									v += GetS0(ploc, i)*GetNode(i).GetDrawTemp();
								}
							}
						}
						vals.Add(v);
					}
				}

				FitPointsInDrawingRegion(points, (int)tilex+1, (int)tilex+1);		// in case the derived class wishes to control the drawing region

				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilex+1,(int)tilex+1,colormode,linemode);
			}
		}

	} // end: if (draw_this_element)

	Outer_Face() = backUpOfOuterFaces;
};

void FiniteElement3D::SetOuterFaceFlag(int face, int set /* = 1 */)
{
	if (set == 1)
	{
		outer_face = outer_face|GetCharBit(face);
	}
	else
	{
		if (GetOuterFaceFlag(face))
		{
			outer_face = (char)((int)outer_face - GetIntBit(face));
		}
	}
}

void FiniteElement3D::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	// add all the field variables of the parent class
	Body3D::GetAvailableFieldVariables(variables);
	//FVT_position,FVT_displacement, FVT_velocity: implemented in Element
	//AH: 01-2017: Velocity is not added in Element! (only for rigid bodies!)
	// possibility to remove entries does not exist yet, just do not use the Function of the parent class

	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_z, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress_cauchy, FieldVariableDescriptor::FVCI_z, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress_mises);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress_principal, FieldVariableDescriptor::FVCI_z, false);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_total_strain, FieldVariableDescriptor::FVCI_z, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_logarithmic_strain, FieldVariableDescriptor::FVCI_z, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_principal_strain, FieldVariableDescriptor::FVCI_z, false);


	if(GetMaterial().IsInelasticMaterial())
	{
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_inelastic_strain, FieldVariableDescriptor::FVCI_z, true);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_equivalent_plastic_strain);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_hardening_parameter);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_yield_function);
	}
}

double FiniteElement3D::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD)
{
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_displacement:
		{
			if(flagD)
				return fvd.GetComponent(GetDisplacementD(local_position));
			else
				return fvd.GetComponent(GetDisplacement(local_position));
		}
	case FieldVariableDescriptor::FVT_position:
		{
			if(flagD)
				return fvd.GetComponent(GetPosD(local_position));
			else
				return fvd.GetComponent(GetPos(local_position));
		}
	case FieldVariableDescriptor::FVT_velocity:
		{
			if(flagD)
				return fvd.GetComponent(GetVelD(local_position));
			else
				return fvd.GetComponent(GetVel(local_position));
		}
	case FieldVariableDescriptor::FVT_stress:
	case FieldVariableDescriptor::FVT_stress_cauchy: // \tau = (1/det(F)) F S F^T
	case FieldVariableDescriptor::FVT_stress_principal:
	case FieldVariableDescriptor::FVT_stress_mises:
		{

			ConstVector<MAXInelasticVariablesCount> inelastic_variables;
			Vector3D local_position_new = local_position;
			if(IsInelasticMaterial())
			{
				IntPointsStiffnessIterator ip(this);
				ip.GoClosestTo(local_position);
				local_position_new = ip.Point();
				if(flagD)
				{
					DataToInelasticVariablesD(inelastic_variables, ip.GetIndex());
				}
				else
				{
					DataToInelasticVariables(inelastic_variables, ip.GetIndex());
				}
			}

			Matrix3D gradu;
			if(flagD) 
				GraduD(local_position_new, gradu);
			else
			{
				ConstVector<FEmaxDOF> xgCache;
				SetComputeCoordinates(xgCache);
				Gradu(local_position_new, xgCache, gradu);
			}
			
			Matrix3D strain;
			GetStrain(strain, gradu);
			
			Matrix3D stress;
			GetMaterial().ComputeStressFromStrain(strain, stress, inelastic_variables);
			
			if(fvd.VariableType() == FieldVariableDescriptor::FVT_stress_mises)
				return stress.Mises();
			
			if(fvd.VariableType() == FieldVariableDescriptor::FVT_stress_cauchy || fvd.VariableType() == FieldVariableDescriptor::FVT_stress_principal)
			{
				gradu(1,1) += 1;  //  F = grad u + I 
				gradu(2,2) += 1;
				gradu(3,3) += 1;
				stress = (1./gradu.Det())*gradu*stress*gradu.GetTp();
			}

			if (fvd.VariableType() == FieldVariableDescriptor::FVT_stress_principal)
			{
				Vector3D principal_stresses;
				Matrix3D principal_directions;
				//the eigenvalues are sorted in ascending order, however here we want them in descending order, hence the double negation
				stress *= -1.;
				stress.EVP(principal_stresses, principal_directions);
				principal_stresses *= -1.;
				return fvd.GetComponent(principal_stresses);
			}
			
			return fvd.GetComponent(stress);
		}
	case FieldVariableDescriptor::FVT_total_strain:
	case FieldVariableDescriptor::FVT_logarithmic_strain:
	case FieldVariableDescriptor::FVT_principal_strain:
		{
			Matrix3D gradu;
			if(flagD) 
				GraduD(local_position, gradu);
			else
			{
				ConstVector<FEmaxDOF> xgCache;
				SetComputeCoordinates(xgCache);
				Gradu(local_position, xgCache, gradu);
			}
			if (fvd.VariableType() == FieldVariableDescriptor::FVT_logarithmic_strain)
			{
				double c1 = gradu(fvd.ComponentIndex1(),fvd.ComponentIndex2());
				double c2 = gradu(fvd.ComponentIndex2(),fvd.ComponentIndex1());
				return log(1.+(c1+c2)/2.);
			}
			Matrix3D strain;
			GetStrain(strain, gradu);
			if (fvd.VariableType() == FieldVariableDescriptor::FVT_principal_strain)
			{
				Vector3D principal_strains;
				Matrix3D principal_strain_directions;
				strain *= -1.;
				strain.EVP(principal_strains, principal_strain_directions);
				principal_strains *= -1.;
				return fvd.GetComponent(principal_strains);
			}


			return fvd.GetComponent(strain);
		}
	case FieldVariableDescriptor::FVT_yield_function:
		{
			if (!IsInelasticMaterial())
			{
				GetMBS()->UO(UO_LVL_err) << "ERROR: Inelastic field variables requested. Material is not declared to be inelastic, though!\n";
				return 0.;
			}

			Vector inelastic_variables;
			IntPointsStiffnessIterator ip(this);
			ip.GoClosestTo(local_position);
			if (flagD)
			{
				DataToInelasticVariablesD(inelastic_variables, ip.GetIndex());
			}
			else
			{
				DataToInelasticVariables(inelastic_variables, ip.GetIndex());
			}

			Vector3D new_local_position = ip.Point();
			Matrix3D gradu;
			if (flagD)
				GraduD(new_local_position, gradu);
			else
			{
				ConstVector<FEmaxDOF> xgCache;
				SetComputeCoordinates(xgCache);
				Gradu(new_local_position, xgCache, gradu);
			}
			Matrix3D strain;
			GetStrain(strain, gradu);
			Matrix3D initial_strain;
			GetInitialStrain(initial_strain, new_local_position);
			return GetMaterial().ComputeYieldFunction(strain-initial_strain, inelastic_variables);
		}
	}
	
	Vector inelastic_variables;
	if (IsInelasticMaterial())
	{
		IntPointsStiffnessIterator ip(this);
		ip.GoClosestTo(local_position);
		if (flagD)
		{
			DataToInelasticVariablesD(inelastic_variables, ip.GetIndex());
		}
		else
		{
			DataToInelasticVariables(inelastic_variables, ip.GetIndex());
		}
	}

	return GetMaterial().GetFieldVariableValue(fvd, inelastic_variables);
}

void FiniteElement3D::PreAssemble()
{
	FiniteElementGeneric<Body3D>::PreAssemble();
	SetOuterFaces();
}

//$ EK 2013-10-25: added in order to set the outer faces correctly (not done if 
//the outer faces are set in PreAssemble() due to missing LTG list)
//might reduce memory consumption for visualization
void FiniteElement3D::PostAssemble() 
{
	SetOuterFaces();
}


void FiniteElement3D::SetRotorAngularVelocity(const Vector3D& angular_velocity, const Vector3D& axis_position)
{
	rotor_angular_velocity = angular_velocity;
	rotor_axis_position = axis_position;
} 

void FiniteElement3D::GyroscopicMatrix(Matrix& gy)
{
	gy.SetSize(SOS(), SOS());
	gy.FillWithZeros();

	if (rotor_angular_velocity.Norm() > 0.)
	{
		int mode = 3; //mbs->GetModelDataContainer()->TreeGetInt("FE3D.gyroscopy_mode",0);
		if (mode > 3 || mode < 1)
			assert(0 && "ERROR: FE3D.gyroscopy_mode must be in {1,2,3}!\n");

		if (mode < 3) // Lagrangian Formulation
		{
			int dim = Dim(); 
			int ns = NS();
			int flexdof = FlexDOF();

			ConstVector<FEmaxDOF> sf(ns);
			ConstMatrix<FEmaxDOF*FEmaxDOF> S(dim,flexdof), gyfull(flexdof, flexdof);
			ConstMatrix<FEmaxDOF*3> temp(dim,flexdof);
			S.SetAll(0.);
			temp.SetAll(0.);
			sf.SetAll(0.);

			Matrix3D jac;

			Matrix3D Omega(
				0., -rotor_angular_velocity(3), rotor_angular_velocity(2),
				rotor_angular_velocity(3), 0., -rotor_angular_velocity(1),
				-rotor_angular_velocity(2), rotor_angular_velocity(1), 0.);

			double rho = this->GetMaterial().Density();

			for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
			{
				GetJacobi(jac, ip.Point());
				double jacdet = jac.Det();

				// compute sf = vector containing shape functions
				GetS0(ip.Point(), sf);

				for (int i=0; i<ns; i++)
				{
					double val = sf(i+1);
					int offset = 3*i;
					S(1, offset+1) = val;
					S(2, offset+2) = val;
					S(3, offset+3) = val;
				}

				double factor = 1;
				if (mbs->GetModelDataContainer()->TreeGetInt("FE3D.GyroscopicMatrix.use_transformation_factor", 0))
				{
					Vector3D p0 = GetRefConfPos(ip.Point());
					factor = (Sqr(p0.X()) + Sqr(p0.Y()))/Sqr(p0.Z())/2.0;
				}

				Mult(Omega, S, temp);
				MultTp(S, temp, gyfull);

				gy.AddSubmatrix(gyfull, 1, 1, 1, 1, flexdof, flexdof, 2. * rho * jac.Det() * ip.Weight() * factor);

				// compute S * S^T  where S is the shape function matrix 
				//     [ SF(1) 0 0 SF(2) 0 0 ...
				// S = [ 0 SF(1) 0 0 SF(2) 0 ..
				//     [ 0 0 SF(1) 0 0 SF(2) ..
				/*SST.SetAll(0);
				for (int i=0; i<ns; i++)
				for (int j=0; j<ns; j++)
				for (int di=1; di<=dim; di++)
				SST(i*dim+di,j*dim+di) += SF(i+1)*SF(j+1);

				m.AddSubmatrix(SST,1,1,1,1,flexdof,flexdof,
				fabs(jacdet) * GetMaterial().Density() * ip.Weight());*/
				//massmatrix += fabs (jacdet) * GetMaterial().Density() * w(i1) * (HL*HL.GetTp());
			}
		}
		else // Eulerian formulation   q^T Gy q = (sum_ij) q_i (x^T*Omega^T*dS_i_dx^T*S_j) q_j  (cf. eq(24) in B. Kirchgaessner, Finite Elements in Rotordynamics, Section 4)
		{
			int dim = Dim(); 
			int ns = NS();
			int flexdof = FlexDOF();

			ConstVector<FEmaxDOF> sf(ns);
			ConstMatrix<FEmaxDOF*3> DS(dim,ns);
			Matrix3D jac, jac_inv;
			Matrix3D Omega(
				0., -rotor_angular_velocity(3), rotor_angular_velocity(2),
				rotor_angular_velocity(3), 0., -rotor_angular_velocity(1),
				-rotor_angular_velocity(2), rotor_angular_velocity(1), 0.);

			double rho = this->GetMaterial().Density();

			for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
			{
				Vector3D ploc(ip.Point());

				// compute sf = vector containing shape functions
				GetS0(ploc, sf);
				
				// compute DS = (dS_i/dx_j)_ij matrix, containing spatial derivatives of the shape functions
				GetDSMatrix0(ploc, DS);
				
				// compute element transformation matrix
				GetJacobi(jac, ploc);
				double factor = -2.*fabs(jac.Det())*rho*ip.Weight();
				
				// compute rotational part of speed of this material point as a function of its position in actual configuration
				Vector3D r(GetPos(ploc) - rotor_axis_position);
				Vector3D v(Omega*r);
				
				// we have to multiply with inverse of Jacobian, i.e. with dXdx,
				// since we need to multiply v with dSdx (cf. eq(25) in B. Kirchgaessner), 
				// but we only get dSdX via GetDSMatrix0
				jac.GetInverse(jac_inv);	
				v = jac_inv*v;

				// it is of least computational effort to multiply with all scalar factors already at this point
				v *= factor;

				for (int i=0; i<ns; i++)
				{
					Vector3D dSidX(DS.Get0(0,i), DS.Get0(1,i), DS.Get0(2,i));
					double factor_i = (v*dSidX);
					int offset_i = i*dim;
					
					for (int j=0; j<ns; j++)
					{
						int offset_j = j*dim;
						double val = factor_i*sf.Get0(j);
						for (int d=1; d<=dim; d++)
							gy(offset_i + d, offset_j + d) += val;
					}
				}
			}
		}
	}
}



//stiffness matrix due to centrifugal loading by rotatoric motion around constant angular velocity omega
void FiniteElement3D::StiffnessMatrixCentrifugal(Matrix& Kc) const
{
	Kc.SetSize(SOS(), SOS());
	Kc.SetAll(0.);

	if (rotor_angular_velocity.Norm() > 0.) // Eulerian formulation   q^T Kc q = (sum_ij) q_i (x_{m} Omega_{lm} dS_{ki}/dx_{l} S_{kj}/dx_{h} Omega_{hn} x_n) q_j  (cf. eq(24) in B. Kirchgaessner, Finite Elements in Rotordynamics, Section 4)
	{
		int mode = 3; //mbs->GetModelDataContainer()->TreeGetInt("FE3D.gyroscopy_mode");
		if (mode > 3 || mode < 1)
			return;
			
		if (mode == 1)
		{
			int dim = Dim(); 
			int ns = NS();
			int flexdof = FlexDOF();

			ConstVector<FEmaxDOF> sf(ns);
			ConstMatrix<FEmaxDOF*FEmaxDOF> S(dim,flexdof), temp2(flexdof, flexdof);
			ConstMatrix<FEmaxDOF*3> temp(dim,flexdof);
			S.SetAll(0.);
			temp.SetAll(0.);

			Matrix3D jac;

			Matrix3D Omega(0., -rotor_angular_velocity(3), rotor_angular_velocity(2),
				rotor_angular_velocity(3), 0., -rotor_angular_velocity(1),
				-rotor_angular_velocity(2), rotor_angular_velocity(1), 0.);

			Matrix3D Omega2;
			Omega.GetATA2(Omega2);

			double rho = this->GetMaterial().Density();

			for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
			{
				GetJacobi(jac, ip.Point());
				double jacdet = jac.Det();

				// compute sf = vector containing shape functions
				GetS0(ip.Point(), sf);
				
				for (int i = 1; i <= ns; i++)
				{
					S(1, 3*i-2) = sf(i);
					S(2, 3*i-1) = sf(i);
					S(3, 3*i  ) = sf(i);
				}

				Mult(Omega2, S, temp);
				MultTp(S, temp, temp2);

				Kc.AddSubmatrix(temp2, 1, 1, 1, 1, flexdof, flexdof, fabs(jacdet) * rho * ip.Weight());
			}
		}
		else if (mode == 2)  // tangential component of centrifugal load only
		{
			int dim = Dim(); 
			int ns = NS();
			int flexdof = FlexDOF();

			ConstVector<FEmaxDOF> sf(ns);
			ConstMatrix<FEmaxDOF*3> S(dim,flexdof);
			ConstMatrix<FEmaxDOF*3> KS(dim,flexdof);
			ConstMatrix<FEmaxDOF*FEmaxDOF> S_KS(flexdof,flexdof);
			S.SetAll(0.);
			KS.SetAll(0.);
			S_KS.SetAll(0.);

			Matrix3D jac;
			Matrix3D Omega(
				0., -rotor_angular_velocity(3), rotor_angular_velocity(2),
				rotor_angular_velocity(3), 0., -rotor_angular_velocity(1),
				-rotor_angular_velocity(2), rotor_angular_velocity(1), 0.);
			double rho = this->GetMaterial().Density();

			for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
			{
				GetJacobi(jac, ip.Point());
				double jacdet = jac.Det();

				// compute sf = vector containing shape functions
				for (int i=1; i <= ns; i++)
				{
					sf(i) = GetS0(ip.Point(), i);
				}

				for (int i = 1; i <= ns; i++)
				{
					S(1, 3*i-2) = sf(i);
					S(2, 3*i-1) = sf(i);
					S(3, 3*i  ) = sf(i);
				}

				Vector3D p0 = GetRefConfPos(ip.Point());
				double norm_p0_sq = p0.Norm2();
				Vector3D F = (-1)*(Omega*Omega*p0);
				ConstVector<FEmaxDOF> v(ns); 
				Mult(S.GetTp(), F, v);
				for (int i=1; i<=dim; i++)
					for (int j=1; j<=flexdof; j++)
						KS(i,j) = p0(i)*v(j)/(-norm_p0_sq);

				KS += (-(F*p0)/norm_p0_sq)*S;
				MultTp(S, KS, S_KS);

				Kc.AddSubmatrix(S_KS, 1, 1, 1, 1, flexdof, flexdof, fabs(jacdet) * rho * ip.Weight());
			}
		}
		else if (mode == 3) // Eulerian formulation
		{
			int dim = Dim(); 
			int ns = NS();
			ConstMatrix<FEmaxDOF*3> DS(dim,ns);
			Matrix3D jac, jac_inv, Omega(
				0., -rotor_angular_velocity(3), rotor_angular_velocity(2),
				rotor_angular_velocity(3), 0., -rotor_angular_velocity(1),
				-rotor_angular_velocity(2), rotor_angular_velocity(1), 0.);
			double rho = this->GetMaterial().Density();

			for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
			{
				// compute DS = (dS_i/dx_j)_ij matrix, containing spatial derivatives of the shape functions
				GetDSMatrix0(ip.Point(), DS);

				GetJacobi(jac, ip.Point());
				double factor = fabs(jac.Det())*rho*ip.Weight();
				jac.GetInverse(jac_inv);
				Vector3D v = jac_inv*Omega*(GetPos(ip.Point()) - rotor_axis_position);
				v *= factor;

				for (int i=0; i<ns; i++)
				{
					int offset_i = i*dim;
					Vector3D dSi_dX(DS.Get0(0, i), DS.Get0(1, i), DS.Get0(2, i));
					double factor_i = v*dSi_dX;
					
					for (int j=0; j<ns; j++)
					{
						int offset_j = j*dim;
						Vector3D dSj_dX(DS.Get0(0, j), DS.Get0(1, j), DS.Get0(2, j));
						double val = factor_i*(v*dSj_dX);

						for (int d=1; d<=dim; d++)
							Kc(offset_i + d, offset_j + d) += val;
					}
				}
			}
		}
	}
}