//#**************************************************************
//#
//# filename:             FiniteElement3D.h
//#
//# author:               Gerstmayr Johannes, Aigner Larissa, YV
//#
//# generated:						March 2007
//# description:          3D - general finite element
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#pragma once

#include "body3d.h"
#include "FiniteElementGeneric.h"


/// Face of a spatial finite element.
/// A face of a finite element is defined by a set of node indices
/// and may have either 4 or 3 nodes.
class FEFace
{
public:
	/** @name Constructors */
	//@{
	FEFace() {};
	/// constructor for a face with four nodes
	FEFace(int v1, int v2, int v3, int v4)
	{
		f[0] = v1; f[1] = v2; f[2] = v3; f[3] = v4; 
		nfacenodes = 4;
	}
	/// constructor for a face with three nodes
	FEFace(int v1, int v2, int v3)
	{
		f[0] = v1; f[1] = v2; f[2] = v3;
		nfacenodes = 3;
	}
	/// constructor for a face with eight nodes
	FEFace(int v1, int v2, int v3, int v4, int v5, int v6, int v7, int v8)
	{
		f[0] = v1; 
		f[1] = v2; 
		f[2] = v3; 
		f[3] = v4; 
		f[4] = v5; 
		f[5] = v6;
		f[6] = v7; 
		f[7] = v8;
		nfacenodes = 8;
	}
	/// constructor for a face with six nodes
	FEFace(int v1, int v2, int v3, int v4, int v5, int v6)
	{
		f[0] = v1; 
		f[1] = v2; 
		f[2] = v3; 
		f[3] = v4; 
		f[4] = v5; 
		f[5] = v6;
		nfacenodes = 6;
	}
	/// copy constructor
	FEFace(const FEFace& face)
	{
		for (int i = 0; i < face.nfacenodes; ++i)
		{
			f[i] = face.f[i];
		}
		nfacenodes = face.nfacenodes;
	}
	//@

	/** @name Extracting information concerning the nodes */
	//@{
	virtual const int& Node(int i) const {return f[i-1];}
	virtual int& Node(int i) {return f[i-1];}
	virtual const int& NodeMod(int i) const {return f[(i-1)%NFaceNodes()];}
	virtual int& NodeMod(int i) {return f[(i-1)%NFaceNodes()];}
	//@}

	/** @name Obtaining general information concerning the face */
	//@{
	/// number of nodes
	virtual int NFaceNodes() const {return nfacenodes;}
	virtual int IsTrig() const {return (nfacenodes == 3) || (nfacenodes == 6);}
	virtual int IsQuad() const {return (nfacenodes == 4) || (nfacenodes == 8);}
	//@}

	/** @name Other general purpose functions */
	//@{

	/// inverts the order of nodes
	virtual void Invert()
	{
		if (NFaceNodes() == 4)
		{
			Swap(f[1],f[3]);
		}
		else if (NFaceNodes() == 3)
		{
			Swap(f[1],f[2]);
		}
		else if (NFaceNodes() == 8)
		{
			Swap(f[1],f[3]);
			Swap(f[4],f[7]);
			Swap(f[5],f[6]);
		}
		else if (NFaceNodes() == 6)
		{
			Swap(f[1],f[2]);
			Swap(f[3],f[5]); 
		}
	}

	/// returns 1 if both faces are exactly the same
	virtual int IsEqual(const FEFace& other) const
	{
		//if faces have different number of nodes, they can not be equal
		if (NFaceNodes() != other.NFaceNodes()) return 0;

		for (int i=0; i<NFaceNodes(); i++)
		{
			if (f[i] != other.f[i]) return 0;
		}
		return 1;
	}

	/// returns 1 if both faces are the same or if any cyclic permutation is the same
	virtual int IsCyclicEqual(const FEFace& other) const
	{
		//if faces have different number of nodes, they can not be equal
		if (NFaceNodes() != other.NFaceNodes()) return 0;

		for (int j=0; j<NFaceNodes(); j++)
		{
			int equal = 1;
			for (int i=1; i<=NFaceNodes(); i++)
			{
				if (NodeMod(i+j) != other.NodeMod(i)) 
				{
					equal = 0;
					break;
				}
			}
			if (equal) return 1;
		}
		return 0;
	}

	virtual double GetS0(const Vector3D& ploc, int shape) const
	{
		if(IsQuad()) return GetS0Quad(ploc, shape);
		if(IsTrig()) return GetS0Trig(ploc, shape);
		return 0.;
	}
	virtual double GetS0Quad(const Vector3D& ploc, int shape) const;
	virtual double GetS0Trig(const Vector3D& ploc, int shape) const;

	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const
	{
		if(IsQuad()) { GetDSMatrix0Quad(ploc, sm); return; }
		if(IsTrig()) { GetDSMatrix0Trig(ploc, sm); return; }
	}
	virtual void GetDSMatrix0Quad(const Vector3D& ploc, Matrix& sm) const;
	virtual void GetDSMatrix0Trig(const Vector3D& ploc, Matrix& sm) const;

	//@}

private:
	int nfacenodes;
	static const int FEMaxFaceNodes = 8; //$EK 2015-07-02 changed to 8 - in case of quadratic (shape functions) quadrilateral faces
	int f[FEMaxFaceNodes];
};

/// Base class for finite elements in 3D space.
/// Provides basic functionality of ANCF elements
/// and serves as a base class for FFRF and CMS elements.
/// Local coordinates in the element vary in the range -1...+1
class FiniteElement3D : public FiniteElementGeneric<Body3D> //$EDC$[beginclass, classname=FiniteElement3D, parentclassname=FiniteElementGeneric]
{
public:
	///@name Creation, initialization and copying of an element
	//@{
	/// Default constructor
	FiniteElement3D(MBS * mbs) : FiniteElementGeneric<Body3D>(mbs) 
	{
//		ElementDefaultConstructorInitialization();
	}
	/// Initialization
	void SetFiniteElement3D(
		int bodyindi,					///< [in] identifier of the body
		const TArray<int>& nodelist,	///< [in] list of the indices of the nodes of the element in the global array
		int material_num,				///< [in] index of the material of the element in the global array
		const Vector3D& coli,			///< [in] color of the graphical representation
		int CMSelnumi=0					///< [in] [optional] number of the corresponding CMS element
		);
	/// Filling in the data from an existing element
	virtual void CopyFrom(const Element& e);
	//@}

	///@name Late initialization
	//@{
	virtual void Initialize(); ///< After assembly and after initial conditions have been set or loaded
	//@}

	//$EK seems to be not necessary any more - 16.9.2014
	//void ElementDefaultConstructorInitialization()
	//{
	//	FiniteElementGeneric<Body3D>::InitConstructor();
	//}

	///@name FE system matrices and gradients for computation and drawing
	//@{
	/// Get the geometric finite element jacobian \b J for the point with given local coordinates
	virtual void GetJacobi(Matrix3D& jac, const Vector3D& ploc) const;
	/// Compute DS matrix and apply element jacobian inverse for the point with given local coordinates;\n
	/// also compute determinate of jacobian;\n
	/// jacinvDS should have already the dimensions Dim() x NS()
	///@return Determinant of the jacobian
	virtual double GetJacInvDS(const Vector3D& ploc, Matrix& jacinvDS) const;
	/// Compute displacement gradient gradu for given result of GetJacInvDS()
	virtual void Gradu(
		const Vector& u,		///< Nodal coordinates
		const Matrix& jacinvDS,	///< Pre-computed result of GetJacInvDS()
		Matrix3D& gradu			///< [out] result of computation
		) const;
	/// Compute displacement gradient at a point with given local coordinates
	virtual void Gradu(
		const Vector3D& ploc,   ///< Local coordinates
		const Vector& u,		///< Nodal coordinates
		Matrix3D& gradu) const;	///< [out] result of computation
	/// Compute displacement gradient at a point with given local coordinates
	/// for the actual "drawing" values of degrees of freedom of the element, see Element::XGD().
	virtual void GraduD(const Vector3D& ploc, Matrix3D& gradu) const;
	//@}

	//use sparse methods to write mass/stiffness matrix --> may be faster for sparse systems!
	
	//virtual int UseSparseK() const {return 0;}; //stiffness matrix is not calculated yet
	
	virtual int Dim() const { return 3; }

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// SHAPE FUNCTIONS
	virtual double GetS0(const Vector3D& ploc, int shape) const = 0;
	//$EK 2015-07: get all shape functions of a vector; assumes that shape is 
	//already set to the correct size
	void GetS0(const Vector3D& ploc, Vector& shape) const;

	//get derivative of shape functions w.r.t. all local coordinates:
	virtual void GetDSMatrix0(const Vector3D& ploc, Matrix& sm) const = 0;
	//get "dxj"-th derivative of "shape"-th shape function
	//virtual double GetDS0(const Vector3D& ploc, int shape, int dxj) const = 0;
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	virtual int FastStiffnessMatrix() const { return 2; }
	
	virtual void StiffnessMatrix(Matrix& m); //fill in sos x 2*sos components, m might be larger
	
	virtual void ComputeMass();
	// quadratic velocity vector - geometric stiffening effect is not present unless it is FFRF
	virtual void AddQuadraticVelocityVector(Vector& fadd, double t) { assert(0); }

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//floating frame of reference formulation: for FFRF elements
	//Position relative to frame, undeformed
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	// positions (determined by the own deformation of the element),
	// which are possibly relative to the reference frame:
	//-1..+1 based!!!
	virtual Vector3D GetDisplacementRel(const Vector3D& p_loc) const;
	virtual Vector3D GetDisplacementRelD(const Vector3D& p_loc) const;
	virtual Vector3D GetPosRel(const Vector3D& p_loc) const;
	virtual Vector3D GetVelRel(const Vector3D& p_loc) const;
	virtual Vector3D GetPosRelD(const Vector3D& p_loc, int use_magnification) const;
	virtual Vector3D GetVelRelD(const Vector3D& p_loc) const;

	//position in the reference configuration:
	//-1..+1 based!!!
	virtual Vector3D GetRefConfPos(const Vector3D& p_loc) const;
	virtual Vector3D GetPos(const Vector3D& p_loc) const { return GetPosRel(p_loc); }
	virtual Vector3D GetDisplacement(const Vector3D& p_loc) const { return GetDisplacementRel(p_loc); }
	virtual Vector3D GetVel(const Vector3D& p_loc) const { return GetVelRel(p_loc); }
	virtual Vector3D GetPosD(const Vector3D& p_loc, int use_magnification) const { return GetPosRelD(p_loc, use_magnification); }
	virtual Vector3D GetPosD(const Vector3D& p_loc) const { return GetPosD(p_loc, 1); }
	//virtual Vector3D GetPosD(const Vector3D& p_loc, double factor) const;
	virtual Vector3D GetDisplacementD(const Vector3D& p_loc) const { return GetDisplacementRelD(p_loc); }
	virtual Vector3D GetVelD(const Vector3D& p_loc) const { return GetVelRelD(p_loc); }

	// angular velocity of the finite element
	virtual Vector3D GetAngularVel(const Vector3D& ploc) const;
	//virtual double GetAngularVel(const Vector3D &ploc, const Vector3D &ploc2, const Vector3D& glob_axis) const;

	//assign a position to each DOF (for drawing of DOF-constraints)
	virtual Vector3D GetDOFPosD(int idof) const //returns postion of i-th DOF
	{
		int node = (idof-1)/3+1;
		return GetPosD(GetNodeLocPos(node));
	}
	//assign a direction of action to each DOF (for drawing of DOF-constraints)
	virtual Vector3D GetDOFDirD(int idof) const //returns direction of action of i-th DOF
	{
		int dir = (idof-1)%3;
		if (dir == 0) return Vector3D(1.,0.,0.);
		else if (dir == 1) return Vector3D(0.,1.,0.);
		else return Vector3D(0.,0.,1.);
	}

	//$ PG 2011-3-2: utilize Body3D::GetElementBox() and Body3D::GetElementBoxD()
	//return box in which the element fits
	virtual Box3D GetElementBox() const
	{
		//Box3D b = Body3D::GetElementBox();
		//b.Increase(2);

		/*for (int i=1;i <= NNodes(); i++)
		{
			b.Add(GetNodePos(i));
		}*/

		Box3D b;
		for (int i=1;i <= NNodes(); i++)
		{
			b.Add(GetNodePos(i));
		}
		return b;
	}

	virtual Box3D GetElementBoxD() const
	{
		//Box3D b = Body3D::GetElementBoxD();
		//$ PG 2013-12-9:[ scaled with radius of box, otherwise absolute value 2 increases bounding box non proportional to element size
		//b.Increase(2);
		//b.Increase(2*b.Radius());
		//$ PG 2013-12-9:]
//		b.Add(GetNodePosD(1));
		/*for (int i=1;i <= NNodes(); i++)
		{
			b.Add(GetNodePosD(i));
		}*/
		Box3D b;
		for (int i=1;i <= NNodes(); i++)
		{
			b.Add(GetNodePosD(i));
		}
		return b;
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES   FACES
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	virtual int NFaces() const = 0;
	virtual const FEFace & GetLocalFace(int i) const = 0;
	virtual FEFace GetFace(int i) const 
	{
		FEFace face = GetLocalFace(i);
		for (int j=1; j<=face.NFaceNodes(); j++)
		{
			face.Node(j) = NodeNum(face.Node(j));
		}
		return face;
	}
	//return if this element has the face 'face' in any cyclic manner
	virtual int HasFace(const FEFace& face) const
	{
		for (int i=1; i <= NFaces(); i++)
		{
			if (face.IsCyclicEqual(GetFace(i))) return 1;
		}
		return 0;
	}
	//compute local face coordinates for drawing or for accessing face surface
	//the vectors v1 and v2 are local vectors which point from refpos to the other nodes
	virtual void GetLocalFaceCoordinates(int face, Vector3D& refpos, Vector3D& v1, Vector3D& v2, Vector3D& v3) const = 0;
	virtual void SetOuterFaces();
	virtual void SetOuterFacesCuttingPlane();
	virtual void ResetOuterFaceFlags() {outer_face = (char)0;}
	char& Outer_Face() {return outer_face;}

	virtual int GetOuterFaceFlag(int face) const
	{
		return (outer_face&GetCharBit(face)) != 0;
	}
	virtual int HasOuterFaces() const {return (int)outer_face != 0;}

	virtual void SetOuterFaceFlag(int face, int set = 1);
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	/// Integral over the element of the displacement with respect to the element coordinates.
	/// Integral over the element of the displacement with respect to the element coordinates \f$\bf{q}\f$
	/// and Jacobian \f$\bf{J}\f$.\n
	/// For finite elements wish shape matrix \f$\bf{S}\f$ and displacements \f$\bf{u}=\bf{Sq}\f$
	/// this reads \f$\bf{H}=\int_{V_{elem}}\bf{S}\,dV_{elem}.\f$
	virtual void GetH(Matrix& H);

	/// Evaluate Mass matrix
	virtual void EvalM(Matrix& m, double t) { EvalMff(m, t); }			// will be overwritten in FFRF case
	// Evaluate Mass matrix for flexible dofs, Attention: Size of mass matrix is not changed, is not initialized with zeros!
	// Size of mass matrix is SOS(), also for Floating Frame formulation!
	virtual void EvalMff(Matrix& m, double t);

	virtual void EvalF2(Vector& f, double t); 
	virtual void EvalF2GeomLin(Vector& fadd, double t); 
	virtual void EvalF2GeomNonlin(Vector& fadd, double t);

	// for volumeloads (gravity ...)
	//in fact it is DuDq Transposed
	virtual void GetIntDuDq(Matrix& dudq) { GetH(dudq); }		// will be overwritten in FFRF case

	// for centrifugal force as MBSLoad - parameters omega, r0 is point on rotation axis (AD)
	virtual void GetIntDuDqFCentrifugal(Matrix& dudq, const Vector3D& omega, const Vector3D& r0); 

	virtual void GetdPosdqT(const Vector3D& ploc, Matrix& d);

	virtual void GetNodedPosdqT(int node, Matrix& dpdqi);

	virtual void AddNodedPosdqTLambda(int node, const Vector3D& lambda, Vector& f) 
	{
		f((node-1)*3+1) += lambda(1);
		f((node-1)*3+2) += lambda(2);
		f((node-1)*3+3) += lambda(3);
	};   // f += dpdq*lambda

	virtual Vector3D GetNodeRefPos(int i) const; //return nodal position of node i in reference configuration
	virtual Vector3D GetNodePos(int i) const;
	virtual Vector3D GetNodePosD(int i) const;
	virtual Vector3D GetNodeVel(int i) const;

	// variables, available for post-processing and sensing
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
	// computation of the variables
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD);

	//for interpolated stress plots!
	virtual void DrawElementPreProc();

	virtual void DrawElement();


	// NONLINEAR MATERIAL LAWS
	virtual double PostNewtonStep(double t);
	virtual void PostprocessingStep();

	// the initial strains may be interpolated over the nodes in a derived class
	virtual void GetInitialStrain(Matrix3D & initial_strain, const Vector3D & ploc)
	{
		GetMaterial().GetInitialStrain(initial_strain);
	}

	virtual void PreAssemble();
	// $EK 2013-10-25 set outer faces
	virtual void PostAssemble();

	virtual double GetVolume() const;

	// $EK 2013-03-21 compute center of mass via integration
	virtual void GetRefCenterOfMass(Vector3D& center) const;
	virtual double GetRefVolume() const;
	
	// $EK 20130408 added in order to be consistent to constraints
	virtual TKinematicsAccessFunctions GetKinematicsAccessFunctions(int mode = 1) const
	{
		// $EK 20130416 not correct to use inheritage here -> addition twice results in the next higher bit to be set!!!!
		//TKinematicsAccessFunctions tkf = FiniteElementGeneric<Body3D>::GetKinematicsAccessFunctions(mode);
		TKinematicsAccessFunctions tkf = TKinematicsAccessFunctions(TKAF_position +TKAF_ref_conf_position+ TKAF_displacement + TKAF_node_position+TKAF_node_ref_conf_position);
		tkf = TKinematicsAccessFunctions(TKAF_velocity + TKAF_node_velocity + TKAF_angular_velocity + tkf);
		tkf = TKinematicsAccessFunctions(TKAF_D_pos_D_q + TKAF_D_node_pos_D_q + TKAF_int_D_u_D_q + tkf);
		//TKAF_position+TKAF_displacement+TKAEF_velocity bereits definiert in Element (Konflikt ????)
		return tkf;
	}

public: //$AD 2017-05-19: functions to determine local position of a given global position on the element
	virtual bool GetLocPosOfGlobalPos(Vector3D& p_loc, const Vector3D& p_glob, int niter) const;
	virtual bool GetLocPosOfGlobalPosPreAssemble(Vector3D& p_loc, const Vector3D& p_glob, int niter) const;

public: //$AD 2016-01-05: functions moved from thermal finite elements to base finite element - must be implemented for all {HexGen,TetGen,PriGen,PyrGen}
	virtual void AddAreaLoad(Vector& f, double pressure, int facet);  
	virtual void AddAreaTraction(Vector& f, Vector3D fvec, int local, int facet);
	virtual void AddElementSurfaceLoad(Vector& f, Vector3D fvec_perunit, int pressure, int facet);
	virtual Vector3D GetSurfaceNormal(int facet);	
	virtual Vector3D GetSurfaceNormalD(int facet);

// functions for ***Generic<T> - geometry/shape dependent
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const = 0;							// for the time being asserts in {TetGen,PriGen,PyrGen} (not implemented yet)
/*additonal by GetFaceJacobi*/ virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const = 0;		// for the time being asserts in {TetGen,PriGen,PyrGen} (not implemented yet)	
	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const = 0;											// for the time being asserts in {TetGen,PriGen,PyrGen} (not implemented yet)
	virtual Vector3D GetFaceCenter(int face) const = 0;																		// done for {HexGen,TetGen,PriGen,PyrGen}
	virtual Vector2D GetFaceCoordOfVolumeCoord(int face, Vector3D p_loc_vol) const = 0;
	virtual Vector3D GetVolumeCoordOfFaceCoord(int face, Vector2D p_loc_face) const = 0;     

// ??? can these functions be removed from derived classes ??? (identical code)
// ??? does this require a separate implementation for all FE geometries ??? (loop over nodes)
	virtual void GetFaceJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const;						// get jacobian of mapping between unit element and element in current (deformed) configuration;    for the time being implemented in base class FiniteElment3D
	virtual void GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const;						// get jacobian of mapping between unit element and element in reference configuration;    for the time being implemented in base class FiniteElment3D - TODO: remove from derived
	virtual void GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const;    						// for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
	virtual Vector3D GetPosFace(int nFace, const Vector2D& p_loc) const;     								// for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
	virtual Vector3D GetPosFaceD(int nFace, const Vector2D& p_loc) const;    								// for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
	virtual double GetFaceArea(int face_nr) const;															// for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
	virtual void GradScal(const Vector& s, const Matrix& jacinvDS, Vector3D& gradScal, int offset=0) const; // for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
	virtual int NSFace(int nFace) const;																	// for the time being implemented in base class FiniteElment3D - TODO: remove from derived	
public:
	//$ SW 2015-9-16: these functions are necessary for the thermal finite elements but may be useful for other fe's too
/*moved*/	//virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const { assert(false && "Function not implemented for this class"); } 
/*moved*/	//virtual void GetFaceJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const { assert(false && "Function not implemented for this class"); }
/*moved*/	//virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const { assert(false && "Function not implemented for this class"); return 0.;}
/*moved*/	//virtual void GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const { assert(false && "Function not implemented for this class"); }
/*moved*/	//virtual Vector3D GetFaceCenter(int face) const { assert(false && "Function not implemented for this class"); return Vector3D();}
/*moved*/	//virtual Vector3D GetPosFace(int nFace, const Vector2D& p_loc) const { assert(false && "Function not implemented for this class"); return Vector3D();}
/*moved*/	//virtual Vector3D GetPosFaceD(int nFace, const Vector2D& p_loc) const { assert(false && "Function not implemented for this class"); return Vector3D();}
/*moved*/	//virtual double GetFaceArea(int face_nr) const { assert(false && "Function not implemented for this class"); return 0.;}
/*moved*/	//virtual void GradScal(const Vector& s, const Matrix& jacinvDS, Vector3D& gradScal, int offset = 0) const { assert(false && "Function not implemented for this class");}
/*moved*/	//virtual int NSFace(int nFace) const { assert(false && "Function not implemented for this class"); return 0;}
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);

	virtual void UpdateDependentVariablesRecursively()
	{
		FiniteElementGeneric::UpdateDependentVariablesRecursively();
		FiniteElement3D::UpdateDependentVariables();
	}
protected:
	// protected constructor for derived classes
	FiniteElement3D(MBS * mbs, TArray<int> nodelist)
	: FiniteElementGeneric<Body3D>(mbs, nodelist)
	{ }
	virtual void UpdateDependentVariables(); // sets x-init correctly

	//drawing:
	char outer_face; //bitwise flags if side 1 - side 6 is an outer (visible) face	



	//rotatoric motion with constant speed rotor_angular_velocity around axis through point rotor_axis_position
protected:
	Vector3D rotor_angular_velocity; //$EDC$[varaccess,EDCvarname="angular_velocity",EDCfolder="RotorDynamics",tooltiptext="Element is rotating with angular_velocity around axis_position."] 
	Vector3D rotor_axis_position; //$EDC$[varaccess,EDCvarname="axis_position",EDCfolder="RotorDynamics",tooltiptext="Element is rotating with angular_velocity around axis_position."] 
public:
	virtual void SetRotorAngularVelocity(const Vector3D& angular_velocity, const Vector3D& axis_position = Vector3D(0) );
	virtual void GyroscopicMatrix(Matrix& gy);
	virtual void StiffnessMatrixCentrifugal(Matrix& krot) const;


	//declaration of EDC auto functionality
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);		//retrieve value of element variable/function named RWdata.variable_name with specified components (vector/matrix); return 1, if variable found and read; 0 if not found, -1 if no readaccess, -2 if range-fault
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //write value of element variable/function named RWdata.variable_name with specified components (vector/matrix); return 1, if variable found and written; 0 if not found, -1 if no writeaccess, -2 if range-fault
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); 

};  //$EDC$[endclass,FiniteElement3D]
