//#*****************************************************************************************
//#
//# filename:             FiniteElement3DThermo.cpp
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "FiniteElement3DThermo.h"
#include "Material.h"
#include "node.h"
#include "solversettings_auto.h"

// this helper class will simplify browsing through all integration points for stiffness
// with the direct access to the associated data units
class IntPointsStiffnessIterator : public IntegrationPointsIterator
{
	FiniteElement3DThermo * fe;
public:
	IntPointsStiffnessIterator(FiniteElement3DThermo * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	FiniteElement3DThermo::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};
// once again for accessing const data
class IntPointsStiffnessIteratorConst : public IntegrationPointsIterator
{
	const FiniteElement3D * fe;
public:
	IntPointsStiffnessIteratorConst(const FiniteElement3D * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	const FiniteElement3D::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};

void FiniteElement3DThermo::SetFiniteElement3DThermo(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	FiniteElement3D::SetFiniteElement3D(bodyindi,nodelist,material_num,coli,CMSelementi);
	//???EK ????
	x_init = Vector(2*SOS());
}

// writes the elements initial conditions (defined in the nodes) in the global initial condition vector x_glob
void FiniteElement3DThermo::SetGlobalInitConditions(Vector& x_glob)
{
	UpdateDependentVariablesRecursively();
	for (int i=1; i<=SS(); i++)
	{
		x_glob(ltg.Get(i)) = x_init(i);
	}
}


int FiniteElement3DThermo::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElement3D::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	SetOuterFaces();
	return rv;
}

void FiniteElement3DThermo::CopyFrom(const Element& e)
{
	FiniteElement3D::CopyFrom(e);
}

void FiniteElement3DThermo::Initialize() 
{
	FiniteElement3D::Initialize();
	UpdateDependentVariables();
}

void FiniteElement3DThermo::UpdateDependentVariables()
{
	x_init = Vector(2*SOS()+ES()+IS());
	int sos = SOS();
	for (int i=1; i<=NNodes(); i++)
	{
		const Node3DThermo& node = static_cast<const Node3DThermo&>(GetNode(i));

		x_init(i) = node.X_Init()(node.GetTemperatureLocalIndex()); // temperature
		x_init(sos+i) = node.X_Init()(node.GetTemperaturePLocalIndex()); // derivative of temperature
	}
}

int FiniteElement3DThermo::NSFace(int nFace) const
{
	return GetFace(nFace).NFaceNodes();
}

int FiniteElement3DThermo::CheckConsistency(mystr &errorstr)
{
	int rv = FiniteElement3D::CheckConsistency(errorstr);
	if (materialnum <= 0 || materialnum > mbs->NMaterials())
	{
		errorstr += "Error in FiniteElement3DThermoMechanic: material number invalid\n";
		return max(rv, 1);
	}
	if (dynamic_cast<MaterialThermoMechanic*>(GetMBS()->GetMaterialPtr(materialnum)) == NULL)
	{
		errorstr += "Material of wrong type, MaterialThermoMechanic is required.\n";
		rv = max(1, rv);
	}
	return rv;
}	

void FiniteElement3DThermo::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	// add all the field variables of the parent class
	//FiniteElement3D::GetAvailableFieldVariables(variables);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_position, FieldVariableDescriptor::FVCI_magnitude);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_temperature);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_density); // = divergence
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_temperature_gradient, FieldVariableDescriptor::FVCI_magnitude);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_volumetric_heat_source);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity, FieldVariableDescriptor::FVCI_magnitude);
}


double FiniteElement3DThermo::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD)
{
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_position:
		return fvd.GetComponent(flagD ? GetPosD(local_position,0.) : GetPos(local_position));
	case FieldVariableDescriptor::FVT_temperature:
		if(flagD)
			return GetTemperatureD(local_position);
		else
			return GetTemperature(local_position);
	case FieldVariableDescriptor::FVT_temperature_gradient:
	{
		return fvd.GetComponent(GetGradient(local_position,flagD));
	}
	// the heat equation is 
	//            d
	// rho * c * --- T(x,t) - lambda * /\T(x,t) = f(x,t)
	//           d t                
	//
	case FieldVariableDescriptor::FVT_volumetric_heat_source:
	{
		Matrix3D jacinv;
		GetJacobi(jacinv, local_position);
		jacinv.Invert();

		Vector3D grad = GetGradient(local_position, flagD);
		Vector3D vel;
		if (flagD)
		{
			vel = GetVelD(local_position, 0);
		}
		else
		{
			vel = GetVel(local_position);
		}

		double condTerm = GetMaterial().GetSpecificHeatCapacity() * GetMaterial().Density() * (grad * vel);

		Matrix3D conductivity = GetMaterial().GetThermalConductivity();

		static const double EPS = 1e-2;
		Vector3D local_position_var(local_position);
		local_position_var.X() -= EPS;
		Vector3D TXM = GetGradient(local_position_var, flagD);
		local_position_var.X() += 2*EPS;
		Vector3D TXP = GetGradient(local_position_var, flagD);
		Vector3D TXx = (TXP - TXM) * (1. / (2.*EPS));
		local_position_var.X() -= EPS;

		local_position_var.Y() -= EPS;
		Vector3D TYM = GetGradient(local_position_var, flagD);
		local_position_var.Y() += 2*EPS;
		Vector3D TYP = GetGradient(local_position_var, flagD);
		Vector3D TYy = (TYP - TYM) * (1. / (2.*EPS));
		local_position_var.Y() -= EPS;

		local_position_var.Z() -= EPS;
		Vector3D TZM = GetGradient(local_position_var, flagD);
		local_position_var.Z() += 2*EPS;
		Vector3D TZP = GetGradient(local_position_var, flagD);
		Vector3D TZz = (TZP - TZM) * (1. / (2.*EPS));
		local_position_var.Z() -= EPS;

		Matrix3D mat(TXx,TYy,TZz);

		return condTerm - (conductivity*mat*jacinv).Trace();
	}
	case FieldVariableDescriptor::FVT_velocity:
		return fvd.GetComponent(flagD ? GetVelD(local_position,0.) : GetVel(local_position));
	case FieldVariableDescriptor::FVT_density:
	{
		Matrix3D jacinv;
		GetJacobi(jacinv, local_position);
		jacinv.Invert();

		Vector3D grad = flagD ? GetVelD(local_position,0) : GetVel(local_position);

		static const double EPS = 1e-2;
		Vector3D local_position_var(local_position);
		local_position_var.X() -= EPS;
		Vector3D TXM = GetGradient(local_position_var, flagD);
		local_position_var.X() += 2*EPS;
		Vector3D TXP = GetGradient(local_position_var, flagD);
		Vector3D TXx = (TXP - TXM) * (1. / (2.*EPS));
		local_position_var.X() -= EPS;

		local_position_var.Y() -= EPS;
		Vector3D TYM = flagD ? GetVelD(local_position_var,0) : GetVel(local_position_var);
		local_position_var.Y() += 2*EPS;
		Vector3D TYP = flagD ? GetVelD(local_position_var,0) : GetVel(local_position_var);
		Vector3D TYy = (TYP - TYM) * (1. / (2.*EPS));
		local_position_var.Y() -= EPS;

		local_position_var.Z() -= EPS;
		Vector3D TZM = flagD ? GetVelD(local_position_var,0) : GetVel(local_position_var);
		local_position_var.Z() += 2*EPS;
		Vector3D TZP = flagD ? GetVelD(local_position_var,0) : GetVel(local_position_var);
		Vector3D TZz = (TZP - TZM) * (1. / (2.*EPS));
		local_position_var.Z() -= EPS;

		Matrix3D mat(TXx,TYy,TZz);

		return (mat*jacinv).Trace();
	}


	}
	

	return 0.;
}

double FiniteElement3DThermo::GetTemperature(const Vector3D &p_loc) const
{
	double temperature = 0.;
	for (int i = 1; i <= NS(); i++)
	{
		temperature += XGtemperature(i) * GetS0(p_loc, i);
	}
	return temperature;
}

double FiniteElement3DThermo::GetRefTemperature(const Vector3D &p_loc) const
{
	double temperature = 0.;
	for (int i = 1; i <= NS(); i++)
	{
		temperature += XInitTemperature(i) * GetS0(p_loc, i);
	}
	return temperature;
}

double FiniteElement3DThermo::GetTemperatureD(const Vector3D &p_loc) const
{
	double temperature = 0.;
	for (int i = 1; i <= NS(); i++)
	{
		temperature += XGtemperatureD(i) * GetS0(p_loc, i);
	}
	return temperature;
}

void FiniteElement3DThermo::StiffnessMatrix(Matrix& m)
{
	static const int MAX_DOFS = 20;
	m.FillWithZeros();
	// was the stiffness matrix computed before?
	if (mbs->GetSolSet().store_FE_matrices 
		&& stiffnessmatrix.Getcols() == NNodes()
		&& stiffnessmatrix.Getrows() == NNodes())
	{
		m.SetSubmatrix(stiffnessmatrix,FlexDOF()+1,FlexDOF()+1);
	}
	else
	{
		
		int ns = NS();
		int dim = Dim();

		Matrix dSi_dxj;

		ConstMatrix<3*MAX_DOFS> temp(ns, dim);
		ConstMatrix<MAX_DOFS*MAX_DOFS> K(ns, ns);

		ConstVector<MAX_DOFS> S0(NNodes());

		const MaterialThermoMechanic& mat = GetMaterial();
		const Matrix3D& lambda = mat.GetThermalConductivity();
		double rhocp = mat.Density()*mat.GetSpecificHeatCapacity();
		double lambda_scalar = lambda(1,1);

		// complete version SUPG 3D
		Vector3D vc = rhocp * GetVel(Vector3D());		// velocity at element center

		//determine element dimensions (could be computed outside of EvalF2)
		Vector3D e_xi =  GetPos(Vector3D( 1., 0., 0.)) - GetPos(Vector3D(-1., 0., 0.));
		Vector3D e_eta = GetPos(Vector3D( 0., 1., 0.)) - GetPos(Vector3D( 0.,-1., 0.));

		double h_xi  = e_xi.Norm();
		double h_eta = e_eta.Norm();

		e_xi  /= h_xi;
		e_eta /= h_eta;

		double v_xi  = vc*e_xi;
		double v_eta = vc*e_eta;

		double alpha_xi  = (h_xi  * v_xi ) / (2. * lambda(1,1));
		double alpha_eta = (h_eta * v_eta) / (2. * lambda(1,1));

		double xi = 0.;
		if (alpha_xi != 0.)
			xi = 1./tanh(alpha_xi) - 1./alpha_xi;

		double eta = 0.;
		if (alpha_eta != 0.)
			eta = 1./tanh(alpha_eta) - 1./alpha_eta;

		double k = 0.5 * (v_xi*h_xi*xi + v_eta*h_eta*eta);

		for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
		{
			GetGrad(ip, dSi_dxj);

			Vector3D v;
			for (int i = 1; i <= ns; i++)
			{
				double s0 = GetS0(ip.Point(),i);
				v += s0*GetNode(i).GetVel();
				S0(i) = s0;
			}
			v *= rhocp;

			for (int i = 1; i <= ns; i++)
			{
				for (int j = 1; j <= dim; j++)
				{
					double diffusion = dSi_dxj(j, i)*lambda(1,1)
						/*+ dSi_dxj(2, i)*lambda(1,1)
						+ dSi_dxj(3, i)*lambda(1,1)*/;

					double convection = S0(i)*v(j);

					double v_norm2 = v.Norm2();
					double v_norm2_inv = v_norm2 != 0. ? 1./v_norm2 : 0.;

					double convection_supg = (v(1)*dSi_dxj(1,i) + v(2)*dSi_dxj(2,i) + v(3)*dSi_dxj(3,i)) * v(j) * k * v_norm2_inv;
					//double convection_supg = (gradT(1)*v(1) + gradT(2)*v(2) + gradT(3)*v(3)) * k * v_norm2_inv * (v(1)*agrad(1,i) + v(2)*agrad(2,i) + v(3)*agrad(3,i));

					temp(i,j) = (convection + convection_supg + diffusion);
				}
			}

			double fact = -fabs(integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet) * ip.Weight();

			Mult(temp, dSi_dxj, K);

			m.AddSubmatrix(K, 1, 1, 1, 1, ns, ns, fact);
		}

		if (mbs->GetSolSet().store_FE_matrices 
			&& stiffnessmatrix.Getcols() != NNodes()
			&& stiffnessmatrix.Getrows() != NNodes())
		{
			stiffnessmatrix.Resize(ns, ns);
			for (int i = 1; i <= ns; i++)
			{
				for (int j = 1; j <= ns; j++)
				{
					stiffnessmatrix(i,j) = m(i,j);
				}
			}
		}

		// SUPG:
		/*double vx_pg = rhocp * GetVel(Vector3D()).X();
		double hx = fabs(GetNodeRefPos(1).X() - GetNodeRefPos(2).X());
		double alpha = (hx * vx_pg)/(2. * lambda_scalar);
		double xi = 0.;
		if (alpha != 0.)
			xi = cosh(alpha)/sinh(alpha) - 1./alpha;
		double k = 0.5 * vx_pg*hx * xi;

		double vx_pg_inv = vx_pg != 0. ? 1./vx_pg : 0.;

		for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
		{
			GetGrad(ip, dSi_dxj);

			Vector3D vel;
			for (int i = 1; i <= ns; i++)
			{
				double s0 = GetS0(ip.Point(),i);
				vel += s0*GetNode(i).GetVel();
				S0(i) = s0;
			}

			for (int i = 1; i <= ns; i++)
			{
				for (int j = 1; j <= dim; j++)
				{
					double convection_supg = vel(j)*rhocp * k * vx_pg_inv;

					temp(i,j) = (convection_supg + lambda_scalar) * dSi_dxj(j,i) + rhocp * S0(i)*vel(j);
				}
			}

			double fact = -fabs(integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet) * ip.Weight();

			Mult(temp, dSi_dxj, K);

			m.AddSubmatrix(K, 1, 1, 1, 1, ns, ns, fact);
		}

		if (mbs->GetSolSet().store_FE_matrices 
			&& stiffnessmatrix.Getcols() == NNodes()
			&& stiffnessmatrix.Getrows() == NNodes())
		{
			stiffnessmatrix.Resize(ns, ns);
			for (int i = 1; i <= ns; i++)
			{
				for (int j = 1; j <= ns; j++)
				{
					stiffnessmatrix(i,j) = m(i,j);
				}
			}
		}*/
	}
}

void FiniteElement3DThermo::GradScal(const Vector& s, const Matrix& jacinvDS, 
																		 Vector3D& gradScal, int ) const
{
	gradScal.Set(0., 0., 0.);
	int dim = Dim();
	for (int i = 1; i <= NS(); i++)
	{
		for (int k = 1; k <= dim; k++)
		{
			gradScal(k) += jacinvDS(k,i)*s(i);
		}
	}
}

void FiniteElement3DThermo::EvalF2(Vector& fadd, double t) 
{

	Body3D::EvalF2(fadd,t);

	int ns = NS();

	// write the XG-vector in xgCache
	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);

	Vector3D gradT;
	
	//!!!!!!!!!!!!!!!!
	//EK if time components are fully integrated and stiffness is integrated reduced -> be aware
	//!!!!!!!!!!!!!!!!

	const MaterialThermoMechanic& tempmat = GetMaterial();
	const Matrix3D& lambda = tempmat.GetThermalConductivity();
	//Matrix3D diffusivity = cond * (1./(tempmat.Density()*tempmat.GetSpecificHeatCapacity()));
	double rhocp = tempmat.Density()*tempmat.GetSpecificHeatCapacity();
	
	Matrix agrad;

	// SUPG:
	/*double vx_pg = rhocp * GetVel(Vector3D()).X();
	double hx = fabs(GetNodeRefPos(1).X() - GetNodeRefPos(2).X());
	double alpha = (hx * vx_pg)/(2. * lambda(1,1));
	double xi = 0.;
	if (alpha != 0.)
		xi = cosh(alpha)/sinh(alpha) - 1./alpha;
	double k = 0.5 * vx_pg*hx * xi;

	double vx_pg_inv = vx_pg != 0. ? 1./vx_pg : 0.;

	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		GetGrad(ip, agrad);
		
		//compute temperature related contributions
		//int_D (grad T)^T k agrad dx \elem \reals^ns
		GradScal(xgCache, agrad, gradT);

		Vector3D lambda_gradT; // = lambda*gradT;
		Mult(lambda, gradT, lambda_gradT);

		Vector3D vel = GetVel(ip.Point());

		for (int i = 1; i <= ns; i++)
		{
			double diffusion = agrad(1, i)*lambda_gradT(1)
				+ agrad(2, i)*lambda_gradT(2)
				+ agrad(3, i)*lambda_gradT(3);

			double convection = (gradT(1)*vel(1) + gradT(2)*vel(2) + gradT(3)*vel(3)) * rhocp * GetS0(ip.Point(), i);
 
			double convection_supg = gradT(1)*vel(1)*rhocp * k * vx_pg_inv * agrad(1, i);

			fadd(FlexDOF()+i) -= (diffusion + convection + convection_supg) * fabs(ip.Data().jacdet) * ip.Weight();
		}
	}*/

	// complete version SUPG 3D
	Vector3D vc = rhocp * GetVel(Vector3D());		// velocity at element center

	//determine element dimensions (could be computed outside of EvalF2)
	Vector3D e_xi =  GetPos(Vector3D( 1., 0., 0.)) - GetPos(Vector3D(-1., 0., 0.));
	Vector3D e_eta = GetPos(Vector3D( 0., 1., 0.)) - GetPos(Vector3D( 0.,-1., 0.));

	double h_xi  = e_xi.Norm();
	double h_eta = e_eta.Norm();

	e_xi  /= h_xi;
	e_eta /= h_eta;

	double v_xi  = vc*e_xi;
	double v_eta = vc*e_eta;

	double alpha_xi  = (h_xi  * v_xi ) / (2. * lambda(1,1));
	double alpha_eta = (h_eta * v_eta) / (2. * lambda(1,1));

	double xi = 0.;
	if (alpha_xi != 0.)
		xi = 1./tanh(alpha_xi) - 1./alpha_xi;

	double eta = 0.;
	if (alpha_eta != 0.)
		eta = 1./tanh(alpha_eta) - 1./alpha_eta;

	double k = 0.5 * (v_xi*h_xi*xi + v_eta*h_eta*eta);

	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		GetGrad(ip, agrad);
		
		//compute temperature related contributions
		//int_D (grad T)^T k agrad dx \elem \reals^ns
		GradScal(xgCache, agrad, gradT);

		Vector3D lambda_gradT; // = lambda*gradT;
		Mult(lambda, gradT, lambda_gradT);

		Vector3D v = GetVel(ip.Point());
		v *= rhocp;

		for (int i = 1; i <= ns; i++)
		{
			double diffusion = agrad(1, i)*lambda_gradT(1)
				+ agrad(2, i)*lambda_gradT(2)
				+ agrad(3, i)*lambda_gradT(3);

			double convection = (gradT(1)*v(1) + gradT(2)*v(2) + gradT(3)*v(3)) * GetS0(ip.Point(), i);
			
			double v_norm2 = v.Norm2();
			double v_norm2_inv = v_norm2 != 0. ? 1./v_norm2 : 0.;

			double convection_supg = (gradT(1)*v(1) + gradT(2)*v(2) + gradT(3)*v(3)) * k * v_norm2_inv * (v(1)*agrad(1,i) + v(2)*agrad(2,i) + v(3)*agrad(3,i));

			fadd(FlexDOF()+i) -= (diffusion + convection + convection_supg) * fabs(ip.Data().jacdet) * ip.Weight();
		}
	}
}

Vector3D FiniteElement3DThermo::GetPos(const Vector3D &ploc) const
{
	Vector3D p; //initialized with zeros
	for (int j = 1; j <= NS(); j++)
	{
		p += GetS0(ploc,j)*GetNodeRefPos(j);
	}
	return p;
}

Vector3D FiniteElement3DThermo::GetPosD(const Vector3D &ploc, int ) const
{
	return GetPos(ploc);
}

Vector3D FiniteElement3DThermo::GetVel(const Vector3D &ploc) const
{
	Vector3D p; //initialized with zeros
	for (int j = 1; j <= NS(); j++)
	{
		p += GetS0(ploc,j)*GetNode(j).GetVel();
	}
	return p;
}


Vector3D FiniteElement3DThermo::GetNodeVel(int i) const
{
	return GetNode(i).GetVel();
}


int FiniteElement3DThermo::FastStiffnessMatrix() const
{
	return 2;
}

void FiniteElement3DThermo::SetComputeTemperature(Vector & xgCache) const
{
	xgCache.SetLen(NS());
	for (int i = 1; i <= xgCache.Length(); i++)
	{
		xgCache(i) = XGtemperature(i);
	}
}
void FiniteElement3DThermo::SetComputeTemperatureOfFace(Vector & xgCache, int face) const
{
	const FEFace& feface = GetLocalFace(face);
	int nnodesFace = feface.NFaceNodes();
	xgCache.SetLen(nnodesFace);
	for (int i = 1; i <= xgCache.Length(); i++)
	{
		xgCache(i) = XGtemperature(feface.Node(i));
	}
}


void FiniteElement3DThermo::GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const
{
	shape.SetLen(NSFace(nFace));
	for(int i = 1; i <= NSFace(nFace); ++i)
	{
		shape(i) = GetS0Face(nFace, ploc, i);
	}
}


//absolute deformed position of face ip
Vector3D FiniteElement3DThermo::GetPosFace(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPos(locpos3D);
}

//absolute deformed position of face ip
Vector3D FiniteElement3DThermo::GetPosFaceD(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPosD(locpos3D,1);
}

//compute element transformation matrix (jacobian)
void FiniteElement3DThermo::GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const
{
	jac.SetSize(3,2);
	int ns = NSFace(nFace);
	int dim = Dim();
	jac.SetAll(0.);
	ConstMatrix<2*8> dsMatrix(2,ns); // size of matrix is at most 2 * 8 since there are no more than 8 nodes at a face
	GetDSMatrix0Face(nFace, ploc, dsMatrix);
	const FEFace& face = GetLocalFace(nFace);
	for (int j = 1; j <= 2; j++)
	{
		for (int k = 1; k <= ns; k++)
		{ 
			Vector3D p0 = GetNodeRefPos(face.Node(k));
			for (int i = 1; i <= dim; i++)
			{ 
				jac(i,j) += dsMatrix(j,k)*p0(i);
			}
		}
	}
}

double FiniteElement3DThermo::GetFaceArea(int face_nr) const
{
	IntegrationRule faceRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
	DefineIntegrationRuleFace(face_nr, faceRule);
	double area = 0.;
	Matrix3D jac;
	for (IntegrationPointsIterator ip(&faceRule); !ip.IsEnd(); ++ip)
	{
		GetFaceJacobi(face_nr, jac, ip.Point());
		Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
		Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
		area += ip.Weight()*dS1.Cross(dS2).Norm();
	}
	return area;
}

Vector3D FiniteElement3DThermo::GetGradient(Vector3D local_position, int flagD)
{
	Matrix jacinvds;
	GetJacInvDS(local_position, jacinvds);
	Vector xg_data(NS());
	for (int i=1; i<=NS(); i++)
	{
		if (flagD)
		{
			xg_data(i) = XGD(i);
		}
		else
		{
			xg_data(i) = XG(i);
		}
	}
	Vector3D gradient;
	GradScal(xg_data, jacinvds,gradient);
	return gradient;
}

void FiniteElement3DThermo::AddSurfaceHeatSource(Vector& f, double heat, int faceNr) const
{
	if (!mbs->IsJacobianComputation())
	{
		IntegrationRule intRule(*GetIntegrationRule(IntegrationRule::IVT_Load));

		DefineIntegrationRuleFace(faceNr, intRule);

		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			Matrix3D jac;
			GetFaceJacobi(faceNr, jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			Vector3D local3DPos = GetVolumeCoordOfFaceCoord(faceNr, ip.Point2D());
			for (int i = 1; i <= NS(); i++)
			{
				f(i) += heat * dS1.Cross(dS2).Norm() * ip.Weight() * GetS0(local3DPos,i);
			}
		}
	}
}

void FiniteElement3DThermo::AddVolumeHeatSource(Vector& f, double heat) const
{
	if (!mbs->IsJacobianComputation())
	{
		// SUPG:
		Matrix agrad;

		const MaterialThermoMechanic& tempmat = (MaterialThermoMechanic&)(GetMaterial());
		const Matrix3D& lambda = tempmat.GetThermalConductivity();

		double rhocp = tempmat.Density()*tempmat.GetSpecificHeatCapacity();

		double vx_pg = rhocp * GetVel(Vector3D()).X();
		double hx = fabs(GetNodeRefPos(1).X() - GetNodeRefPos(2).X());
		double alpha = (hx * vx_pg)/(2. * lambda(1,1));
		double xi = 0.;
		if (alpha != 0.)
			xi = 1./tanh(alpha) - 1./alpha;
		double k = 0.5 * vx_pg*hx * xi;

		double vx_pg_inv = vx_pg != 0. ? 1./vx_pg : 0.;

		for (IntPointsStiffnessIteratorConst ip(this); !ip.IsEnd(); ++ip)
		{
			for (int i = 1; i <= NS(); i++)
			{
				GetGrad(ip, agrad);

				f(i) += heat * fabs(ip.Data().jacdet) * ip.Weight() * (GetS0(ip.Point(),i) + k * vx_pg_inv * agrad(1, i));
			}
		}
	}
}

