//#*****************************************************************************************
//#
//# filename:             FiniteElement3DThermo.h
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#pragma once
#include "FiniteElementScalar3D.h"
#include "ThermoElementInterface.h"
#include "material.h"

class FiniteElement3DThermo : public FiniteElement3D, public ThermoElementInterface
{
public:
	FiniteElement3DThermo(MBS* mbs)
		: FiniteElement3D(mbs, TArray<int>().SetXN(8,1,2,3,4,5,6,7,8))
	{ };

	virtual int CheckConsistency(mystr &errorstr);

	void SetFiniteElement3DThermo(
		int bodyindi,					///< [in] identifier of the body
		const TArray<int>& nodelist,	///< [in] list of the indices of the nodes of the element in the global array
		int material_num,				///< [in] index of the material of the element in the global array
		const Vector3D& coli,			///< [in] color of the graphical representation
		int CMSelnumi=0					///< [in] [optional] number of the corresponding CMS element
		);

	/// Filling in the data from an existing element
	virtual void CopyFrom(const Element& e);

	//$ PG 2016-2-11: [ added
	virtual const MaterialThermoMechanic& GetMaterial() const {return static_cast<const MaterialThermoMechanic&>(GetMBS()->GetMaterial(materialnum));}
	//$ PG 2016-2-11: ]

	virtual int GetLocalNodeDOF(int localNodeNr, int nodeDOF) const
	{
		assert(nodeDOF == 1 || nodeDOF == 2);
		if (nodeDOF == 1)
		{
			return localNodeNr;
		}
		return NNodes() + localNodeNr - 1;
	}

	virtual Vector3D GetPos(const Vector3D& ploc) const;
	virtual Vector3D GetPosD(const Vector3D& ploc, int use_magnification) const;
	virtual Vector3D GetPosD(const Vector3D& ploc) const { return GetPosD(ploc, 1); }
	virtual Vector3D GetVel(const Vector3D& ploc) const;
	virtual Vector3D GetVelD(const Vector3D& ploc, int use_magnification) const { return GetVel(ploc); }

	virtual Vector3D GetNodePos(int i) const
	{
		return GetNodeRefPos(i);
	}

	virtual Vector3D GetNodePosD(int i) const 
	{
		return GetNodeRefPos(i);
	}

	//assign a position to each DOF (for drawing of DOF-constraints)
	virtual Vector3D GetDOFPosD(int idof) const //returns postion of i-th DOF
	{
		return GetPosD(GetNodeLocPos(idof));
	}

	//assign a direction of action to each DOF (for drawing of DOF-constraints)
	virtual Vector3D GetDOFDirD(int idof) const //returns direction of action of i-th DOF
	{
		return Vector3D(0.,0.,0.);
	}

	virtual Vector3D GetNodeVel(int i) const;

	virtual void GradScal(const Vector& s, const Matrix& jacinvDS, Vector3D& gradScal, int offset=0) const;

	virtual TKinematicsAccessFunctions GetKinematicsAccessFunctions(int mode = 1) const
	{
		return  TKinematicsAccessFunctions(TKAF_temperature);
	}
	virtual void Initialize(); ///< After assembly and after initial conditions have been set or loaded
	virtual void UpdateDependentVariables();
	virtual int IS() const { return 0; }
	virtual int ES() const { return 0; }
	virtual int SOS() const { return NNodes(); }
	virtual int XGLength() const { return NNodes(); }
	
	virtual int NSFace(int nFace) const;

	virtual double XGtemperature(int iloc) const
	{
		return XG(iloc);
	}

	virtual double& XGtemperature(int iloc)
	{
		return XG(iloc);
	}

	virtual double XGtemperatureD(int iloc) const
	{
		return XGD(iloc);
	}
	
	virtual double XInitTemperature(int iloc) const
	{
		return x_init(iloc);
	}

	virtual void SetGlobalInitConditions(Vector& x_glob);

	virtual int SetElementData(ElementDataContainer& edc);
	virtual void EvalF2(Vector& f, double t);
	virtual void StiffnessMatrix(Matrix& m);
	virtual int FastStiffnessMatrix() const;
	virtual int FlexDOF() const { return 0; }
	virtual double GetTemperature(const Vector3D& p_loc) const;
	virtual double GetRefTemperature(const Vector3D& p_loc) const;
	virtual double GetTemperatureD(const Vector3D& p_loc) const;

	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD);
	/// Copies global element temperatures of the specific face into some local "compute" cache
	void SetComputeTemperature(Vector & xgCache) const;
	void SetComputeTemperatureOfFace(Vector & xgCache, int face) const;
	
	//$EK 2015-06-30: get integration rule for faces
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const = 0;

	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const = 0;
	virtual void GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const;
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const = 0;
	//absolute deformed position of face ip
	virtual Vector3D GetPosFace(int nFace, const Vector2D& p_loc) const;
	virtual Vector3D GetPosFaceD(int nFace, const Vector2D& p_loc) const;
	virtual Vector3D GetFaceCenter(int face) const = 0;

	virtual void GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const;

	virtual double GetFaceArea(int face_nr) const;
	virtual void AddSurfaceHeatSource(Vector& f, double heat, int face) const;
	virtual void AddVolumeHeatSource(Vector& f, double heat) const;

protected:

private:
	//Matrix temperature_stiffnessmatrix;
	Vector3D GetGradient(Vector3D local_position, int flagD);
};
