//#*****************************************************************************************
//#
//# filename:             FiniteElement3DThermoMechanic.cpp
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#include "FiniteElement3DThermoMechanic.h"
#include "Material.h"
#include "node.h"
#include "solversettings_auto.h"


// this helper class will simplify browsing through all integration points for stiffness
// with the direct access to the associated data units
class IntPointsStiffnessIterator : public IntegrationPointsIterator
{
	FiniteElement3DThermoMechanic * fe;
public:
	IntPointsStiffnessIterator(FiniteElement3DThermoMechanic * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	FiniteElement3DThermoMechanic::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};
// once again for accessing const data
class IntPointsStiffnessIteratorConst : public IntegrationPointsIterator
{
	const FiniteElement3D * fe;
public:
	IntPointsStiffnessIteratorConst(const FiniteElement3D * fe_) :
			IntegrationPointsIterator(fe_->integrationRuleStiffness),
			fe(fe_)
			{
			}
	const FiniteElement3D::IntegrationPointStiffnessMatrixLocalData & Data()
	{
		return *fe->integrationPointStiffnessMatrixLocalData(GetIndex());
	}
};

void FiniteElement3DThermoMechanic::SetFiniteElement3DThermoMechanic(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	FiniteElementScalar3D::SetFiniteElement3D(bodyindi,nodelist,material_num,coli,CMSelementi);
	//???EK ????
	x_init = Vector(2*SOS());
}

// writes the elements initial conditions (defined in the nodes) in the global initial condition vector x_glob
void FiniteElement3DThermoMechanic::SetGlobalInitConditions(Vector& x_glob)
{
	UpdateDependentVariables();
	for (int i=1; i<=SS(); i++)
	{
		x_glob(ltg.Get(i)) = x_init(i);
	}
}

int FiniteElement3DThermoMechanic::SetElementData(ElementDataContainer& edc)
{
	int rv = FiniteElementScalar3D::SetElementData(edc);
	UpdateDependentVariablesRecursively();
	SetOuterFaces();
	return rv;
}

void FiniteElement3DThermoMechanic::CopyFrom(const Element& e)
{
	FiniteElementScalar3D::CopyFrom(e);
}

void FiniteElement3DThermoMechanic::Initialize() 
{
	FiniteElementScalar3D::Initialize();
	UpdateDependentVariables();
}

void FiniteElement3DThermoMechanic::LinkToElements()
{
	// first there are SOS() dofs at position level :
	//   they are all coming from the nodes since this element has no
	//   own variables
	// second there are SOS() dofs at velocity level
	// third there are ES() first order variables:
	// these are stored in the nodes too

	LTGreset();

	//SOS dofs at position level
	for (int i=1; i<=NNodes(); i++)
	{
		const Node& node = GetNode(i);
		for (int j=1; j <= node.Dim(); j++)
		{
			AddLTG(node.Get(j)); // get returns the ltg index of the ith dof of the node
		}
	}

	// degree parameter
	for (int i=1; i<=NNodes(); i++)
	{
		const Node& node = GetNode(i);
		AddLTG(node.Get(4));
	}

	//sos dofs at velocity level
	for (int i=1; i<=NNodes(); i++)
	{
		const Node& node = GetNode(i);
		for (int j=1; j <= node.Dim(); j++)
		{
			AddLTG(node.Get(j + node.SOS()));
		}
	}
	for (int i=1; i<=NNodes(); i++)
	{
		const Node& node = GetNode(i);
		AddLTG(node.Get(8));
	}

}


int FiniteElement3DThermoMechanic::NSFace(int nFace) const
{
	return GetFace(nFace).NFaceNodes();
}


int FiniteElement3DThermoMechanic::CheckConsistency(mystr &errorstr)
{
	int rv = FiniteElementScalar3D::CheckConsistency(errorstr);
	if (materialnum <= 0 || materialnum > mbs->NMaterials())
	{
		errorstr += "Error in FiniteElement3DThermoMechanic: material number invalid\n";
		return max(rv, 1);
	}
	if (dynamic_cast<MaterialThermoMechanic*>(GetMBS()->GetMaterialPtr(materialnum)) == NULL)
	{
		errorstr += "Material of wrong type, MaterialThermoMechanic is required.\n";
		rv = max(1, rv);
	}
	return rv;
}	


void FiniteElement3DThermoMechanic::UpdateDependentVariables()
{
	x_init = Vector(2*SOS()+ES()+IS());
	for (int i=1; i<=NNodes(); i++)
	{
		const Node& node = GetNode(i);
		x_init.Copy(node.X_Init(), 1, (i-1)*node.Dim()+1, node.Dim()); // positional degrees of freedom
		x_init(NNodes()*node.Dim()+i) = node.X_Init()(node.Dim()+1); // temperature
		x_init.Copy(node.X_Init(), 1+node.SOS(), SOS()+(i-1)*node.Dim()+1, node.Dim()); // velocity degrees of freedom
		x_init(SOS()+NNodes()*node.Dim()+i) = node.X_Init()(node.SOS()+node.Dim()+1); // derivative of temperature
	}
}


void FiniteElement3DThermoMechanic::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	// add all the field variables of the parent class
	FiniteElementScalar3D::GetAvailableFieldVariables(variables);

	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_temperature);
}


double FiniteElement3DThermoMechanic::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD)
{
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_temperature:
		{
			if(flagD)
				return GetTemperatureD(local_position);
			else
				return GetTemperature(local_position);
		}
	}
	return FiniteElementScalar3D::GetFieldVariableValue(fvd, local_position, flagD);
}


void FiniteElement3DThermoMechanic::UpdateDependentVariablesRecursively()
{
	FiniteElementScalar3D::UpdateDependentVariablesRecursively();
	UpdateDependentVariables();
}


void FiniteElement3DThermoMechanic::StiffnessMatrix(Matrix& m)
{
	// add mechanical stiffness entries
	FiniteElement3D::StiffnessMatrix(m);

	// was the stiffness matrix computed before?
	if (mbs->GetSolSet().store_FE_matrices 
		&& temperature_stiffnessmatrix.Getcols() == NNodes()
		&& temperature_stiffnessmatrix.Getrows() == NNodes())
	{
		m.SetSubmatrix(temperature_stiffnessmatrix,FlexDOF()+1,FlexDOF()+1);
	}
	else
	{
		
		int sos = SOS();
		int ns = NS();
		// calculate the stiffness matrix
		// 
		//  /      T 
		// -| a(VS) . Vw
		//  /
		// Omega

		Matrix jacinvDS;
		Matrix jacinvDS2;
		Matrix jacinvDSLambda;
		Vector vxGradSt;
		Vector S0(NNodes());
		Matrix conductiveTerm(NNodes(), NNodes());
		temperature_stiffnessmatrix = Matrix(NNodes(), NNodes(), 0, 0.);
		const MaterialThermoMechanic& mat = GetMaterial();
		const Matrix3D& cond = mat.GetThermalConductivity();
		//Matrix3D diffusivity = cond * (1./(mat.Density()*mat.GetSpecificHeatCapacity()));
		double volumetric_heat_capacity = mat.Density()*mat.GetSpecificHeatCapacity();

		for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
		{
			// term a * Laplace(T), where a is the thermal diffusivity and T is the temperature
			GetGrad(ip, jacinvDS);			// get inverse jacobian
			Mult(cond,jacinvDS,jacinvDSLambda);
			MultTp(jacinvDS, jacinvDSLambda, jacinvDS2);
			double det = integrationPointStiffnessMatrixLocalData(ip.GetIndex())->jacdet;

			jacinvDS2.Mult(-abs(det)*ip.Weight());
			temperature_stiffnessmatrix.AddMatrix(1,1,jacinvDS2);


			// in this eulerian specification heat is transported by the movement of the element
			// the thermal flux is:  vel*grad(T)

			double vel_x = 0;
			double vel_y = 0;
			double vel_z = 0;
			for (int i = 1; i <= ns; i++)
			{
				vel_x += GetS0(ip.Point(),i)*x_init(sos+1+3*(i-1));
				vel_y += GetS0(ip.Point(),i)*x_init(sos+2+3*(i-1));
				vel_z += GetS0(ip.Point(),i)*x_init(sos+3+3*(i-1));
				S0(i)=GetS0(ip.Point(), i);
			}
			Vector vel(vel_x, vel_y, vel_z);

			Mult(vel,jacinvDS,vxGradSt);

			vxGradSt *= -abs(det)*ip.Weight()*volumetric_heat_capacity;
			Mult(S0,vxGradSt, conductiveTerm);
			temperature_stiffnessmatrix.AddMatrix(1,1,conductiveTerm);
		}

		m.SetSubmatrix(temperature_stiffnessmatrix,FlexDOF()+1,FlexDOF()+1);
	}
}


void FiniteElement3DThermoMechanic::EvalF2(Vector& fadd, double t) 
{
	//compute deformation gradient, compute strain, compute stress and compute elastic forces:
	FiniteElementScalar3D::EvalF2(fadd,t);

	int ns = NS();
	int dim = Dim();
	int sos = SOS();

	// write the XG-vector in xgCache
	ConstVector<FEmaxDOF> xgCache;
	SetComputeCoordinates(xgCache);

	Vector3D gradT;

	double temp = 0.;
	
	//!!!!!!!!!!!!!!!!
	//EK if time components are fully integrated and stiffness is integrated reduced -> be aware
	//!!!!!!!!!!!!!!!!

	const MaterialThermoMechanic& tempmat = GetMaterial();
	const Matrix3D& cond = tempmat.GetThermalConductivity();
	//Matrix3D diffusivity = cond * (1./(tempmat.Density()*tempmat.GetSpecificHeatCapacity()));
	double volumetric_heat_capacity = tempmat.Density()*tempmat.GetSpecificHeatCapacity();
	Matrix agrad;
	for (IntPointsStiffnessIterator ip(this); !ip.IsEnd(); ++ip)
	{
		GetGrad(ip, agrad);
		
		//compute temperature related contributions
		//int_D (grad T)^T k agrad dx \elem \reals^ns
		GradScal(xgCache, agrad, gradT, dim*ns);

		Vector3D condXGradT = cond*gradT;

		double vel_x = 0;
		double vel_y = 0;
		double vel_z = 0;

		for (int i = 1; i <= ns; i++)
		{
			vel_x += GetS0(ip.Point(),i)*x_init(sos+1+3*(i-1)); // velocity of node i in x-direction
			vel_y += GetS0(ip.Point(),i)*x_init(sos+2+3*(i-1)); // velocity of node i in y-direction
			vel_z += GetS0(ip.Point(),i)*x_init(sos+3+3*(i-1)); // velocity of node i in z-direction
		}
		Vector3D vel = Vector3D(vel_x, vel_y, vel_z);

		for (int i = 1; i <= ns; i++)
		{
			temp = agrad(1, i)*condXGradT(1)
				+ agrad(2, i)*condXGradT(2)
				+ agrad(3, i)*condXGradT(3);
			fadd(FlexDOF()+i) -= fabs(ip.Data().jacdet) * ip.Weight() * temp;

			// in this eulerian specification heat is transported by the movement of the element
			// the thermal flux is:  vel*grad(T)
			fadd(FlexDOF()+i) -= (volumetric_heat_capacity * gradT * vel) * GetS0(ip.Point(),i) * fabs(ip.Data().jacdet) * ip.Weight();
		}
	}
}


int FiniteElement3DThermoMechanic::FastStiffnessMatrix() const
{
	return 2;
}

double FiniteElement3DThermoMechanic::GetRefTemperature(const Vector3D &p_loc) const
	{
	double temperature = 0.;
	for (int i = 1; i <= NS(); i++)
	{
		temperature += XInitTemperature(i) * GetS0(p_loc, i);
	}
	return temperature;
}

void FiniteElement3DThermoMechanic::SetComputeTemperature(Vector & xgCache) const
{
	xgCache.SetLen(NS());
	for (int i = 1; i <= xgCache.Length(); i++)
	{
		xgCache(i) = XGtemperature(i);
	}
}
void FiniteElement3DThermoMechanic::SetComputeTemperatureOfFace(Vector & xgCache, int face) const
{
	const FEFace& feface = GetLocalFace(face);
	int nnodesFace = feface.NFaceNodes();
	xgCache.SetLen(nnodesFace);
	for (int i = 1; i <= xgCache.Length(); i++)
	{
		xgCache(i) = XGtemperature(feface.Node(i));
	}
}


void FiniteElement3DThermoMechanic::GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const
{
	shape.SetLen(NSFace(nFace));
	for(int i = 1; i <= NSFace(nFace); ++i)
	{
		shape(i) = GetS0Face(nFace, ploc, i);
	}
}


//absolute deformed position of face ip
Vector3D FiniteElement3DThermoMechanic::GetPosFace(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPos(locpos3D);
}

//absolute deformed position of face ip
Vector3D FiniteElement3DThermoMechanic::GetPosFaceD(int nFace, const Vector2D& p_loc) const
{
	Vector3D locpos3D = GetVolumeCoordOfFaceCoord(nFace, p_loc);
	return GetPosD(locpos3D);
}

//compute element transformation matrix (jacobian)
void FiniteElement3DThermoMechanic::GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const
{
	jac.SetSize(3,2);
	int ns = NSFace(nFace);
	int dim = Dim();
	jac.SetAll(0.);
	ConstMatrix<2*8> dsMatrix(2,ns); // size of matrix is at most 2 * 8 since there are no more than 8 nodes at a face
	GetDSMatrix0Face(nFace, ploc, dsMatrix);
	const FEFace& face = GetLocalFace(nFace);
	for (int j = 1; j <= 2; j++)
	{
		for (int k = 1; k <= ns; k++)
		{ 
			Vector3D p0 = GetNodeRefPos(face.Node(k));
			for (int i = 1; i <= dim; i++)
			{ 
				jac(i,j) += dsMatrix(j,k)*p0(i);
			}
		}
	}
}

double FiniteElement3DThermoMechanic::GetFaceArea(int face_nr) const
{
	IntegrationRule faceRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
	DefineIntegrationRuleFace(face_nr, faceRule);
	double area = 0.;
	Matrix3D jac;
	for (IntegrationPointsIterator ip(&faceRule); !ip.IsEnd(); ++ip)
	{
		GetFaceJacobi(face_nr, jac, ip.Point());
		Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
		Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
		area += ip.Weight()*dS1.Cross(dS2).Norm();
	}
	return area;
}


void FiniteElement3DThermoMechanic::AddSurfaceHeatSource(Vector& f, double heat, int faceNr) const
{
	if (!mbs->IsJacobianComputation())
	{
		IntegrationRule intRule(*GetIntegrationRule(IntegrationRule::IVT_Load));
		DefineIntegrationRuleFace(faceNr, intRule);
		int offset_thermo_DOF = FlexDOF();
		Matrix3D jac;

		for (IntegrationPointsIterator ip(&intRule); !ip.IsEnd(); ++ip)
		{
			GetFaceJacobi(faceNr, jac, ip.Point());
			Vector3D dS1(jac(1,1), jac(2,1), jac(3,1));
			Vector3D dS2(jac(1,2), jac(2,2), jac(3,2));
			Vector3D local3DPos = GetVolumeCoordOfFaceCoord(faceNr, ip.Point2D());
			for (int i = 1; i <= NS(); i++)
			{
				f(i+offset_thermo_DOF) += heat * dS1.Cross(dS2).Norm() * ip.Weight() * GetS0(local3DPos,i);
			}
		}
	}
}

void FiniteElement3DThermoMechanic::AddVolumeHeatSource(Vector& f, double heat) const
{
	if (!mbs->IsJacobianComputation())
	{
		int offset_thermo_DOF = FlexDOF();
		for (IntPointsStiffnessIteratorConst ip(this); !ip.IsEnd(); ++ip)
		{
			for (int i = 1; i <= NS(); i++)
			{
				f(i + offset_thermo_DOF) += heat * fabs(ip.Data().jacdet) * ip.Weight() * GetS0(ip.Point(),i);
			}
		}
	}
}