//#*****************************************************************************************
//#
//# filename:             FiniteElement3DThermoMechanic.h
//#
//# author:               Humer Alexander, Karer Erwin, Weitzhofer Simon
//#
//# generated:            July 2015
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#pragma once
#include "FiniteElementScalar3D.h"
#include "ThermoElementInterface.h"
#include "material.h"

// This class contains the 

//$ E D C $[beginclass, classname=FiniteElement3DThermoMechanic, parentclassname=FiniteElementScalar]
class FiniteElement3DThermoMechanic : public FiniteElementScalar3D, public ThermoElementInterface
{
public:
	FiniteElement3DThermoMechanic(MBS* mbs)
		: FiniteElementScalar3D(mbs, TArray<int>().SetXN(8,1,2,3,4,5,6,7,8))
	{ };

	virtual int CheckConsistency(mystr &errorstr);

	void SetFiniteElement3DThermoMechanic(
		int bodyindi,					///< [in] identifier of the body
		const TArray<int>& nodelist,	///< [in] list of the indices of the nodes of the element in the global array
		int material_num,				///< [in] index of the material of the element in the global array
		const Vector3D& coli,			///< [in] color of the graphical representation
		int CMSelnumi=0					///< [in] [optional] number of the corresponding CMS element
		);

	/// Filling in the data from an existing element
	virtual void CopyFrom(const Element& e);

	virtual TKinematicsAccessFunctions GetKinematicsAccessFunctions(int mode = 1) const
	{
		return  TKinematicsAccessFunctions(FiniteElementScalar3D::GetKinematicsAccessFunctions(mode) + TKAF_temperature);
	}

	//$ PG 2016-2-11: [ added
	virtual const MaterialThermoMechanic& GetMaterial() const {return static_cast<const MaterialThermoMechanic&>(GetMBS()->GetMaterial(materialnum));}
	//$ PG 2016-2-11: ]

	virtual void Initialize(); ///< After assembly and after initial conditions have been set or loaded
	virtual void LinkToElements();
	virtual int GetLocalNodeDOF(int localNodeNr, int nodeDOF) const
	{
		assert(nodeDOF <= 8);
		if (nodeDOF <= 3)
		{
			return (localNodeNr-1)*3 + nodeDOF;
		}
		if (nodeDOF == 4)
		{
			return FlexDOF() + localNodeNr + nodeDOF - 4;
		}
		if (nodeDOF <= 7)
		{
			return FlexDOF() + NNodes() + (localNodeNr-1)*3 + nodeDOF - 4;
		}
		if (nodeDOF == 8)
		{
			return 2*FlexDOF() + NNodes() + localNodeNr + nodeDOF - 7;
		}
		return -1;
	}

	virtual int IS() const { return 0; }
	virtual int ES() const { return 0; }
	virtual int SOS() const { return (Dim()+1)*NNodes(); }
	virtual int XGLength() const { return FlexDOF() + NNodes(); }
	
	virtual int NSFace(int nFace) const;

	virtual double XGtemperature(int iloc) const
	{
		return XGScalar(iloc);
	}

	virtual double& XGtemperature(int iloc)
	{
		return XGScalar(iloc);
	}
	virtual void SetXGtemperature(Vector& xgCache)
	{
		SetXGScalar(xgCache);
	}
	virtual void SetGlobalInitConditions(Vector& x_glob);

	virtual double XInitTemperature(int iloc) const
	{
		return x_init(FlexDOF()+iloc);
	}


	virtual int SetElementData(ElementDataContainer& edc);
	virtual void EvalF2(Vector& f, double t);

	//virtual void EvalF2GeomNonlin(Vector& f, double t);
	//virtual void EvalF2GeomNonlin2(Vector& f, double t);
	//virtual void EvalF2GeomLin(Vector& f, double t);
	virtual void StiffnessMatrix(Matrix& m);
	virtual int FastStiffnessMatrix() const;
	virtual int FlexDOF() const { return 3 * NNodes(); }
	virtual double GetTemperature(const Vector3D& p_loc) const
	{ return GetScalar(p_loc); }
	virtual double GetTemperatureD(const Vector3D& p_loc) const
	{ return GetScalarD(p_loc); }
	virtual double GetRefTemperature(const Vector3D& p_loc) const;

	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables);
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD);
	/// Copies global element temperatures of the specific face into some local "compute" cache
	virtual void SetComputeTemperature(Vector & xgCache) const;
	virtual void SetComputeTemperatureOfFace(Vector & xgCache, int face) const;
	
	//$EK 2015-06-30: get integration rule for faces
	virtual void DefineIntegrationRuleFace(int nFace, IntegrationRule & integrationRule) const = 0;

	virtual double GetS0Face(int nFace, const Vector3D& ploc, int shape) const = 0;
	virtual void GetS0Face(int nFace, const Vector3D& ploc, Vector& shape) const;
	virtual void GetDSMatrix0Face(int nFace, const Vector3D& ploc, Matrix& sm) const = 0;
	//absolute deformed position of face ip
	virtual Vector3D GetPosFace(int nFace, const Vector2D& p_loc) const;
	virtual Vector3D GetPosFaceD(int nFace, const Vector2D& p_loc) const;
	virtual Vector3D GetFaceCenter(int face) const = 0;

	virtual void GetFaceRefJacobi(int nFace, Matrix3D& jac, const Vector3D& ploc) const;

	virtual double GetFaceArea(int face_nr) const;
	virtual void AddSurfaceHeatSource(Vector& f, double heat, int face) const;
	virtual void AddVolumeHeatSource(Vector& f, double heat) const;

protected:
	virtual void UpdateDependentVariables();
	virtual void UpdateDependentVariablesRecursively();

private:
	// storage for stiffness matrix of the temperature dofs,
	// the stiffness matrix for the mechanical dofs is stored
	// in FiniteElement3D
	Matrix temperature_stiffnessmatrix;
}; //$E D C$[endclass,FiniteElement3DThermoMechanic]
