//#**************************************************************
//#
//# filename:             FiniteElementGeneric.cpp
//#
//# author:               Gerstmayr Johannes, Aigner Larissa, YV
//#
//# generated:						October 2010
//# description:          common finite element base class - for 2D & 3D finite elements
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************



#include "element.h"
#include "body3d.h"
#include "body2d.h"
#include "material.h"
#include "node.h"
#include "finiteelementgeneric.h"
#include "elementdataaccess.h"
#include "solversettings_auto.h"

#pragma warning(disable:4661) // 2015-03: AD surpress Warnings caused by the specialization 
                              //          right below but also at the bottom of the file 
                              //         ( autogenerated function implementation in ElementEDCauto.cpp NOT available at this time ) 

// explicit instantatiation of the template 
template class FiniteElementGeneric<Body3D>;
template class FiniteElementGeneric<Body2D>;


template<class BodyXD>
void FiniteElementGeneric<BodyXD>::CopyFrom(const Element& e)
{
	BodyXD::CopyFrom(e);
	const FiniteElementGeneric<BodyXD>& ce = (const FiniteElementGeneric<BodyXD>&)e;
	massmatrix = ce.massmatrix;
	Hmatrix = ce.Hmatrix;
	stiffnessmatrix = ce.stiffnessmatrix;

	// integration rules
	integrationRuleStiffness = ce.integrationRuleStiffness;
	integrationRuleMass = ce.integrationRuleMass;
	integrationRuleLoad = ce.integrationRuleLoad;

	SetGeometricNonlinearityStatus(ce.GetGeometricNonlinearityStatus());

	perform_initial_value_compuation = ce.perform_initial_value_compuation;

	nodes.SetLen(ce.NNodes());
	for (int i = 1; i <= ce.NNodes(); i++)
	{
		NodeNum(i) = ce.NodeNum(i);
		//concentratedmass[i-1] = ce.concentratedmass[i-1];
	}

	bodyind = ce.bodyind;
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::GetElementData(ElementDataContainer& edc)
{
	BodyXD::GetElementData(edc);
	//GetElementDataAuto(edc); // this calls all the automatic branch of the template class - non-automatic part is missed 
	ElementData ed;
	ed.SetInt((int) geometricNonlinearityStatus,"GeometricNonlinearityStatus"); ed.SetToolTipText("0..GNS_Linear, 1..GNS_NonlinearSmallStrain, 2..GNS_NonlinearLargeStrain"); edc.TreeAdd("FiniteElement",ed);
}

template<class BodyXD>
int FiniteElementGeneric<BodyXD>::SetElementData(ElementDataContainer& edc)
{
	int rv = BodyXD::SetElementData(edc);
	//SetElementDataAuto(edc); // enter FiniteElementGenericChildClass::SetElementDataAuto
	
	int gnls = edc.TreeGetInt("FiniteElement.GeometricNonlinearityStatus");
	geometricNonlinearityStatus = (GeometricNonlinearityStatus) gnls;

	UpdateDependentVariablesRecursively();
	return rv;
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::LinkToElements()
{
	if (SOSowned() != SOS())
	{
		int Nnonode = SOSowned();
		int Nnode = SOS()-Nnonode;
		TArray<int> storeltg(Nnonode);
		for (int i=1; i <= Nnonode*2; i++)
		{
			storeltg.Add(LTG(i));
		}
		LTGreset();

		// Node positions
		for (int j = 1; j <= NNodes(); j++)
		{
			const Node& node = GetNode(j);
			//Position:
			for (int i=1; i <= node.SOS(); i++)
			{
				AddLTG(node.Get(i));
			}
		}
		// remaining, if exist
		for (int i=1; i <= Nnonode; i++)
		{
			AddLTG(storeltg(i));
		}

		// Node velocities
		for (int j = 1; j <= NNodes(); j++)
		{
			const Node& node = GetNode(j);
			//velocity:
			for (int i=1; i <= node.SOS(); i++)
			{
				AddLTG(node.Get(i+node.SOS()));
			}
		}
		// remaining, if exist
		for (int i=1; i <= Nnonode; i++)
		{
			AddLTG(storeltg(i+Nnonode));
		}
	}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::SetMaterialNum(int mnum)
{
	materialnum = mnum;
}

template<class BodyXD>
const IntegrationRule * FiniteElementGeneric<BodyXD>::GetIntegrationRule(IntegrationRule::IntegratedValueType ivt) const
{
	switch(ivt)
	{
	case IntegrationRule::IVT_Stiffness: return integrationRuleStiffness;
	case IntegrationRule::IVT_Mass: return integrationRuleMass;
	case IntegrationRule::IVT_Load: return integrationRuleLoad;
	}
	assert(0 && "unknown entity to integrate");
	return NULL;
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::BuildDSMatrices()
{
	// precompute DShape Matrices and multiply with element transformation
	// in every integration point of the stiffness matrix
	for (IntegrationPointsIterator ip(integrationRuleStiffness); !ip.IsEnd(); ++ip)
	{
		IntegrationPointStiffnessMatrixLocalData * newData = new IntegrationPointStiffnessMatrixLocalData;
		integrationPointStiffnessMatrixLocalData(ip.GetIndex()) = newData;
		if(mbs->GetSolSet().store_FE_matrices)
		{
			newData->grad = new Matrix;
			newData->jacdet = GetJacInvDS(ip.Point(), *(newData->grad));
		}
		else
		{
			static Matrix gradTemp;
			newData->jacdet = GetJacInvDS(ip.Point(), gradTemp);
		}
	}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::SetComputeCoordinates(Vector & xgCache) const
{
	xgCache.SetLen(XGLength());
	for (int i = 1; i <= xgCache.Length(); i++)
		xgCache(i) = XG(i);

	//int nnodes = NNodes();
	//int dof_per_node = FlexDOF()/nnodes;
	//Vector3D mean;
	//for (int i=0; i<nnodes; i++)
	//{
	//	for (int j=1; j<=Dim(); j++)
	//	{
	//		mean(j) += xgCache(i*dof_per_node + j);
	//	}
	//}
	//mean *= 1./nnodes;
	//for (int i=0; i<nnodes; i++)
	//{
	//	for (int j=1; j<=Dim(); j++)
	//	{
	//		xgCache(i*dof_per_node + j) -= mean(j);
	//	}
	//}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::GetStrain(Matrix2D& strain, const Matrix2D& grad_u)
{
	if (GetGeometricNonlinearityStatus() == GNS_Linear)
	{
		// linearized strain tensor
		strain = 0.5*(grad_u.GetTp() + grad_u);
	}
	else
	{
		// Green-St.Venant strain tensor
		Matrix2D Fh = grad_u + Matrix2D(1);       
		strain = 0.5*(Fh.GetTp()*Fh-Matrix2D(1));		
	}
}
template<class BodyXD>
void FiniteElementGeneric<BodyXD>::GetStrain(Matrix3D& strain, const Matrix3D& grad_u)
{
	if (GetGeometricNonlinearityStatus() == GNS_Linear)
	{
		// linearized strain tensor
		strain = 0.5*(grad_u.GetTp() + grad_u);
	}
	else
	{
		// Green-St.Venant strain tensor
		Matrix3D Fh = grad_u + Matrix3D(1);
		strain = 0.5*(Fh.GetTp()*Fh - Matrix3D(1));
	}
}


template<class BodyXD>
void FiniteElementGeneric<BodyXD>::ComputeStressFromStrain(MatrixXD& stress, const MatrixXD& strain, const IntegrationPointsIterator& ip, bool compute_trial_stress) const
{	
	// get old inelastic strain (from last nonlin step, or from last time step, depending on the specific solution method)
	ConstVector<MAXInelasticVariablesCount> inelastic_variables(0);
	if (IsInelasticMaterial() && !perform_initial_value_compuation) //if initial strains are set and we have plastic material, it is necessary to force a first Newton step for elastic material (inelastic_variables keeps length 0) in order to obtain a good initial XG for all the next Newton steps.
	{
		MaterialGeneric::InelasticitySolutionMethod solmeth = GetMaterial().GetInelasticitySolutionMethod();
		switch (solmeth)
		{
		case MaterialGeneric::ISM_ReturnMapping:
			DataLastStepToInelasticVariables(inelastic_variables, ip.GetIndex());  // read old inelastic variables from last computed time/load step
			break;
		case MaterialGeneric::ISM_FixedPoint:
			DataToInelasticVariables(inelastic_variables, ip.GetIndex());  // read old inelastic variables from last nonlinear (discontinuous) step
			break;
		default:
			mbs->UO(UO_LVL_warn) << "WARNING: FiniteElementGeneric<BodyXD>::ComputeStressFromStrain not implemented for inelasticity-solution-method " << solmeth << "\n";
		}
	}
	if (GetMaterial().IsPlanarMaterial())
	{
		int dim = 2;
		assert(stress.Getcols() == dim);
		assert(stress.Getrows() == dim);
		assert(strain.Getcols() == dim);
		assert(strain.Getrows() == dim);
		GetMaterial().ComputeStressFromStrain2D((Matrix2D&) strain, (Matrix2D&) stress, inelastic_variables, compute_trial_stress);
	}
	else
	{
		int dim = 3;
		assert(stress.Getcols() == dim);
		assert(stress.Getrows() == dim);
		assert(strain.Getcols() == dim);
		assert(strain.Getrows() == dim);
		GetMaterial().ComputeStressFromStrain((Matrix3D&) strain, (Matrix3D&) stress, inelastic_variables, compute_trial_stress);
	}
}
#include "timeintlog.h"

template<class BodyXD>
double FiniteElementGeneric<BodyXD>::ComputeInelasticVariablesUpdateFromStrain(const MatrixXD& strain, const IntegrationPointsIterator& ip)
{
	MaterialGeneric::InelasticitySolutionMethod solmeth = GetMaterial().GetInelasticitySolutionMethod();
	if (perform_initial_value_compuation)
		return 0.;

	int n = GetMaterial().GetInelasticVariablesCount();
	static ConstVector<MAXInelasticVariablesCount> inelastic_variables(n);
	static ConstVector<MAXInelasticVariablesCount> inelastic_variables_last_step(n);
	static ConstVector<MAXInelasticVariablesCount> inelastic_variables_update(n);

	// get old inelastic strain (from last time step)
	DataLastStepToInelasticVariables(inelastic_variables_last_step, ip.GetIndex());  // read old inelastic strain from last computed time/load step

	// and last computed update of inelastic strain (either from last nonlin step (fixed_point) or also from last time step (return_mapping))
	switch (solmeth)
	{
	case MaterialGeneric::ISM_ReturnMapping:
		inelastic_variables = inelastic_variables_last_step; // use old inelastic strain from last computed time/load step
		break;
	case MaterialGeneric::ISM_FixedPoint:
		DataToInelasticVariables(inelastic_variables, ip.GetIndex());  // read old inelastic strain from last nonlinear (discontinuous) step
		break;
	default:
		mbs->UO(UO_LVL_warn) << "WARNING: FiniteElementGeneric<BodyXD>::ComputeInelasticVariablesUpdateFromStrain not implemented for inelasticity-solution-method " << solmeth << "\n";
		return 0;
	}

	// compute actual inelastic strain as function from old inelastic strain and elastic strain (minus initial strain)
	double inelastic_variable_update_norm = 0.;
	
	if (!GetMaterial().IsPlanarMaterial())
	{
		int dim = 3;
		assert(strain.Getcols() == dim);
		assert(strain.Getrows() == dim);
		inelastic_variable_update_norm = GetMaterial().ComputeInelasticVariablesUpdateFromStrain(inelastic_variables_update, inelastic_variables, (const Matrix3D&) strain);
	}
	else
	{
		int dim = 2;
		assert(strain.Getcols() == dim);
		assert(strain.Getrows() == dim);
		inelastic_variable_update_norm = GetMaterial().ComputeInelasticVariablesUpdateFromStrain2D(inelastic_variables_update, inelastic_variables, (const Matrix2D&) strain);
	} 

	//$ PG 2014-5-8:[ test
	//double damping = 2.8;
	//inelastic_variable_update_norm *= damping;
	//inelastic_variables_update *= damping;
	//$ PG 2014-5-8:]

	//inelastic_variables = inelastic_variables_last_step + inelastic_variables_update;
	inelastic_variables += inelastic_variables_update;
	InelasticVariablesToData(inelastic_variables, ip.GetIndex());

	return inelastic_variable_update_norm;
}


template<class BodyXD>
void FiniteElementGeneric<BodyXD>::GetNeighbourElements(TArray<int>& neighbours)
{
	neighbours.SetLen(0);

	for (int i=1; i<=NNodes(); i++)
	{
		Node& n = GetNode(i);
		for (int j=1; j<=n.NElements(); j++)
		{
			int el = n.ElemNum(j);
			if (!neighbours.Find(el) && el != GetOwnNum()) neighbours.Add(el);
		}
	}
	//UO() << "elem " << GetOwnNum() << " has " << neighbours.Length() << "neighbours\n";
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::SetFiniteElementGeneric(int bodyindi, const TArray<int>& nodelist, int material_num, const Vector3D& coli, int CMSelementi)
{
	// AH: 2013-11-27: don't set the GNS here, since it needs to be set by EDC
	// moreover, it is set by the InitConstructor anyway
	// SetGeometricNonlinearityStatus(GNS_NonlinearSmallStrain);		// YV

	bodyind = bodyindi;
	nodes = nodelist;
	elementname = GetElementSpec();

	// materials
	SetMaterialNum(material_num);
	col = coli;

	elementname = GetElementSpec();

	//$ SW 2016-6-16: put the following lines from UpdateDependentVariables to here
	//$EK 2015-06-18 - corrected - every ODE (first order) base function needs an init value
	x_init = Vector(2*SOS() + ES());
	if (CMSelementi)
	{
		SetFFRFElement(CMSelementi);
	}
	FiniteElementGeneric<BodyXD>::UpdateDependentVariables();
}

// this function is called 
template<class BodyXD>
void FiniteElementGeneric<BodyXD>::UpdateDependentVariables()
{
	if (x_init.GetLen() != 2*SOS() + ES())
	{
		x_init = Vector(2*SOS() + ES());
	}
	//else //JG 2011-03-21: initialization not needed in case of CMS element
	//{
		// (AD)
		// get initial displacements & velocities from nodes (nodelist) if x_init of (all) nodes are initialized (Length != 0) 
		for (int i=1; i<=NNodes(); i++)
		{
			// AP: Changed mbs->GetNode(nodes(i))  to  FiniteElement3D::GetNode(i),
			//         otherwise, does not work for CMS elements
			// Node node(mbs->GetNode(nodes(i)));
			Node& node(GetNode(i));


			// {x1,y1,z1,x2,y2,z2,... vx1,...} -> twice the size
			if (node.X_Init().GetLen())
			{
				//x_init.Copy(node.X_Init(), 1,								(i-1)*DOFPerNode()+1,          DOFPerNode());
				//x_init.Copy(node.X_Init(), 1+DOFPerNode(),	(i-1+NNodes())*DOFPerNode()+1, DOFPerNode());
				x_init.Copy(node.X_Init(), 1,								(i-1)*node.SOS()+1,          node.SOS());
				x_init.Copy(node.X_Init(), 1+node.SOS(),	(i-1+NNodes())*node.SOS()+1, node.SOS());
				//copy remaining Degrees of Freedom if they exist, e.g., ES() DOF (added afterwards) -> order
				//in this case would be 
				//positions - velocities - ES()-DOF in order of nodes
				int lenRemainingDOF = node.X_Init().GetLen()- 2*node.SOS();
				if (lenRemainingDOF > 0)
				{
					//$EK 2015-06-18 we assume that all nodes do have the same number of remaining dof
					x_init.Copy(node.X_Init(), 1 + 2*SOS(),	2*SOS() + (i-1)*lenRemainingDOF + 1,  lenRemainingDOF);
				}
			}
		}
	//}
	// (AD) end
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::Initialize() 
{
	BodyXD::Initialize();
};

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::PreAssemble()
{
	// we set up the integration rules
	IntegrationRule::IntegrationRuleSettings settings(
		GetElementType(), GetActualInterpolationOrder(), IntegrationRule::IVT_Stiffness, GetGeometricNonlinearityStatus(), 0
		);
	integrationRuleStiffness = GetIntegrationRulesLibrary()->GetIntegrationRule(settings, this);
	settings.integratedValueType = IntegrationRule::IVT_Mass;
	integrationRuleMass = GetIntegrationRulesLibrary()->GetIntegrationRule(settings, this);
	settings.integratedValueType = IntegrationRule::IVT_Load;
	integrationRuleLoad = GetIntegrationRulesLibrary()->GetIntegrationRule(settings, this);
	assert(
		integrationRuleStiffness != NULL &&
		integrationRuleMass != NULL &&
		integrationRuleLoad != NULL
		);

	// Build matrices
	BuildDSMatrices();

	int datas = DataS();   //$ PG 2016-6-20: GetMaterial().GetInelasticVariablesCount()
	if (datas != 0)
	{
		//Vector datainit(datas); //automatically initialized with zeros
		//SetDataInit(datainit); //initialize data variables with zero = initial inelastic strain == zero

		// initializing in case of non-zero inelastic strain, etc.
		data_init.SetLen(datas);
		for (IntegrationPointsIterator ip(this->integrationRuleStiffness); !ip.IsEnd(); ++ip)
		{
			Vector inelastic_variables_IP(GetMaterial().GetInelasticVariablesCount());
			GetMaterial().InitializeInelasticVariables(inelastic_variables_IP);
			InelasticVariablesToDataInit(inelastic_variables_IP, ip.GetIndex());
		}
	}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::GetGrad(IntegrationPointsIterator & it, Matrix & grad) const
{
	if(
		integrationPointStiffnessMatrixLocalData.Length() < it.GetIndex() ||
		integrationPointStiffnessMatrixLocalData(it.GetIndex())->grad == NULL
		)
		GetJacInvDS(it.Point(), grad);		// if nothing is pre-computed, let's do it now
	else
	{
		Matrix * preComputedGrad = integrationPointStiffnessMatrixLocalData(it.GetIndex())->grad;
		grad.LinkWith(preComputedGrad->Getrows(), preComputedGrad->Getcols(), preComputedGrad->GetMatPtr());
	}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::AddMSparse(SparseMatrix& ms, double t) //add sparse mass matrix into full system matrix
{
	if (mbs->GetSolSet().store_FE_matrices || massmatrix.Getcols() != FlexDOF())
	{
		ConstMatrix<FEmaxDOF*FEmaxDOF> m(FlexDOF(),FlexDOF());
		EvalMff(m, t);
		ms.AddMatrix(ltg,ltg,FlexDOF(),FlexDOF(),m); //ltg may be longer than FlexDOF()
	}
	else
	{
		//add dense matrix
		ms.AddMatrix(ltg,ltg,FlexDOF(),FlexDOF(),massmatrix); //ltg may be longer than FlexDOF()
	}
}

template<class BodyXD>
void FiniteElementGeneric<BodyXD>::EraseInelasticVariables()
{
	// here we assume that inelastic variables are stored in each integration point for the stiffness matrix
	int n = GetMaterial().GetInelasticVariablesCount();
	for(int nip = 1; nip <= integrationRuleStiffness->GetNumberOfIntegrationPoints(); nip++)
		for (int i = 1; i <= n; i++)
			XData((nip-1)*n+i) = 0;
}

template<>
Vector3D FiniteElementGeneric<Body3D>::GetAngularMomentum(const Vector3D& p_ref) const
{
	Vector3D D(0.,0.,0.);
	double rho = GetMaterial().Density();

	/*Vector3D v = GetVel(Vector3D(1.,1.,1.));
	Vector3D p = GetPos(Vector3D(1.,1.,1.));

	Vector xgp(NNodes()*Dim()), xg(NNodes()*Dim());*/

	/*for (int i = 1; i <= NNodes()*Dim(); i++)
	{
		xg(i) = XG(i);
		xgp(i) = XGP(i);
	}*/

	/*mbs->UO() << xg << "\n";
	mbs->UO() << xgp << "\n";
	mbs->UO() << v << "\n";
	mbs->UO() << p << "\n\n";*/

	for (IntegrationPointsIterator ip(integrationRuleMass); !ip.IsEnd(); ++ip)
	{
		Vector3D p = ip.Point();
		double det = this->GetPrecomputedJacDet(ip.GetIndex());

		Vector3D r = GetPos(p) - p_ref;
		Vector3D v = GetVel(p);
		D += r.Cross(v) * rho * ip.Weight() * det;
	}
	return D;
}

template<>
Vector3D FiniteElementGeneric<Body2D>::GetAngularMomentum(const Vector3D& p_ref) const
{
	Vector3D D(0.,0.,0.);
	// AH: to be implemented
	assert(0);
	return D;
}

template<class BodyXD>
FiniteElementGeneric<BodyXD>::FiniteElementGeneric(MBS * mbs, TArray<int> nodelist):
	BodyXD(mbs),
	geometricNonlinearityStatus(GNS_NonlinearSmallStrain),
	nodes(nodelist),
	bodyind(1),
	integrationRuleStiffness(0),
	integrationRuleMass(0),
	integrationRuleLoad(0),
	perform_initial_value_compuation(false)
{
	//$ SW 2015-6-17: everything is done in the initialization
	// list. This constructor does everything that is done in
	// the elementdefaultconstructor except that the nodelist
	// is a parameter. 
	// massmatrix, stiffnessmatrix and Hmatrix are implicitely
	// initialized as empty matrices
}