//#*****************************************************************************************
//#
//# filename:             HotintConsoleVersion.cpp
//#
//# author:               Weitzhofer Simon
//#
//# generated:            January 2016
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************


#include "HotintConsoleVersion.h"
#include <stdio.h>
#include <string.h>
#include <sstream>
#include <windows.h>

namespace
{
	class DummyDataStorage : public WCDInterface::DataLoader, public WCDInterface::DataSaver
	{
		virtual void SetTime(double time) {}
		virtual DataSaver & operator << (int) {return *this;}
		virtual DataSaver & operator << (double) {return *this;}
		// this function accepts a pointer to a C-style zero terminated string
		virtual DataSaver & operator << (const char *) {return *this;}

		virtual double GetTime() {return 0;}		// should be called before other
		// data reading operations
		virtual DataLoader & operator >> (int &) {return *this;}
		virtual DataLoader & operator >> (double &) {return *this;}
		// here the working module obtains a new pointer  to
		// the the string is C-style zero terminated string char * p,
		// so that delete[] p should be executed afterwards
		virtual DataLoader & operator >> (char * & p) {return *this;}
	};
}


ConsoleApplication* ca;

BOOL HotintConsoleControlHandler (DWORD dwControlEvent) 
{ 
  switch (dwControlEvent) 
  { 
    case CTRL_C_EVENT:
	case CTRL_CLOSE_EVENT:
	case CTRL_BREAK_EVENT:
		if (ca != 0)
		{
			ca->AddText("received signal to stop execution!");
			ca->StopComputation();
			return TRUE;
		}
		return FALSE;
    default:
      return FALSE; 
  } 
} 

int main(int argc, char **argv)
{
	__argv = argv;
	__argc = argc;
	
	ca = 0;
	SetConsoleCtrlHandler ((PHANDLER_ROUTINE) HotintConsoleControlHandler, TRUE); 

	bool silent_mode = false;
	bool flag_display_version = 0;
	for (int i=1; i < argc; i++)
	{
		if (strcmp(*(argv+i),"-silent")==0 ||
			strcmp(*(argv+i),"-no-console-output")==0)
		{
			silent_mode = true;
			//$ SW 2016-4-4: HACK, should suppress all output, even in dll's
			std::cerr.setstate(std::ios_base::failbit);
			std::cout.setstate(std::ios_base::failbit);
			std::clog.setstate(std::ios_base::failbit);
		}
		if (strcmp(*(argv+i),"-v")==0 ||
			strcmp(*(argv+i),"-ver")==0 ||
			strcmp(*(argv+i),"-version")==0)
		{
			flag_display_version = true;
		}
	}

	if(flag_display_version)
	{
		ca = new ConsoleApplication(1);
		ca->SetSilentMode(0);
		ca->InstantMessageText("\nconsole-HOTINT Version:");
		ca->InstantMessageText(ca->GetHotintVersionText());
		ca->InstantMessageText("\n");
		return 0; // only return version number
	}
	else
	{
		ca = new ConsoleApplication(silent_mode);

		ca->StartComputation();
		int exitcode;
		exitcode = ca->GetExitCode();

		delete ca;	
		return exitcode;
	}
}


ConsoleApplication::ConsoleApplication(bool silent_mode)
: silent_mode(silent_mode),
  exitcode(0)
{
	size_t size = strlen(WORKING_DLL_NAME)+1;
	wchar_t* wString=new wchar_t[size];
	MultiByteToWideChar(CP_ACP, 0, WORKING_DLL_NAME, -1, wString, static_cast<int>(size));
	hWorkingDll = LoadLibrary(wString);
	delete[] wString;

	if(!hWorkingDll)
	{
		AddText("HOTINT: Cannot load dynamic-link library: ");
		AddText(WORKING_DLL_NAME);
		AddText("!\n");
		exitcode = -102;
		ExitProcess(-102);
	}
	pExpFnCreateWCDObject = (PExpFnCreateWCDObject)(GetProcAddress(hWorkingDll, WORKING_OBJECT_CREATION_FUNCTION_NAME));
	if(!pExpFnCreateWCDObject)
	{
		AddText("Wrong dynamic-link library ");
		AddText(WORKING_DLL_NAME);
		AddText("!\n");
		exitcode = -103;
		ExitProcess(-103);
	}
	GetCurrentDirectory(MAX_PATH, ProgramDirectory);
	pExpFnDeleteWCDObject = (PExpFnDeleteWCDObject)(GetProcAddress(hWorkingDll, WORKING_OBJECT_DELETION_FUNCTION_NAME));

	pWCDI = (*pExpFnCreateWCDObject)();
	pWCDI->SetUserInterface(this);

	ReadConfigFile();
	pWCDI->InitializeAfterConfigLoaded();

	// in case log file path was changed, we need to adapt
	pWCDI->OpenLogFile(0);

	//set close when finished
	pWCDI->GetOptions()->GeneralOptions()->ApplicationCloseApplicationWhenFinished() = 1;
}

ConsoleApplication::~ConsoleApplication()
{
	watcher.Stop();

	if(pWCDI != NULL)
	{		
		(*pExpFnDeleteWCDObject)();
		pWCDI = NULL;		
	}

	FreeLibrary(hWorkingDll);
}

void ConsoleApplication::InstantMessageText(const char * text)
{
	if (!silent_mode)
	{
		std::cerr << text << std::endl;
		std::cerr.flush();
	}
}

void ConsoleApplication::AddText(int value)
{
	stringstream ss;
	ss << value;
	string s = ss.str();
	AddText(s.c_str());
}

void ConsoleApplication::AddText(const char *text)
{
	if (!silent_mode)
	{
		std::cout << text;
	}
}

void ConsoleApplication::StatusText(const char * )
{
	//NOTHING TO DO?
}

// if the text-running window has been switch off, this command will switch it on
void ConsoleApplication::AssureTextIsVisible()
{
	//NOTHING TO DO?
}
	

// create directory (only, if it does not exist).
// returns 0 in case of error (wrong path), 1 in case it already existed, and 2 if created
int ConsoleApplication::CreateMissingDirectory(const char * str)
{
	int rv = 2;

	size_t size = strlen(str)+1;
	wchar_t* wString=new wchar_t[size];
	MultiByteToWideChar(CP_ACP, 0, str, -1, wString, static_cast<int>(size));
	
	if (!CreateDirectory(wString, NULL))
	{
		DWORD err = GetLastError();
		if (err == ERROR_ALREADY_EXISTS)
		{
			rv = 1;
		}
		else   // err == ERROR_PATH_NOT_FOUND
		{
			rv = 0;
		}
	}
	delete[] wString;
	return rv;
}

//call a function in WCDriver:
int ConsoleApplication::CallWCDriverFunction(int action, int option, int value, ElementDataContainer* edc)
{
	//action=1: redraw
	//action=2: Edit loads of element 'value', load number 'option'

	int rv = 1;
	//insert actions performed by WCDriver on call by MBS-System

	switch(action)
	{
	
	case 4: //Set Program directory as current
		{
			SetCurrentDirectory(ProgramDirectory);
			break;
		}
	case 5: // Quit Program from MBS
		{
			//ExitProcess(0);
			break;
		}
	case 6: //reset model
		{
			ModelChanged();
			break;
		}

	case 21:
		{
			//SleepX(100);	//$ DR 2014-10-13 some time is necessary to avoid wrong screenshots
			//DialogDataManager.LoadEntry(value);
			InstantMessageText("DialogDataManager not available in console version");
			break;
		}
	case 22:	// return number of entries in DataManager
		{
			InstantMessageText("DialogDataManager not available in console version");
			break;
		}
	case 9: // close all PlotTool instances
		{
			//nothing to do
			break;
		}
	case 23:	// set properties of windows according to hotint options
		{
			//nothing to do: we do not have any window options for console version
			break;
		}
	case 24:	// update hotint options according to properties of windows
		{
			//nothing to do: we do not have any window options for console version
			break;
		}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//Start and Monitor Computation - called from Parser
//action = 30, option:= 1 .. start computation, any.. check if computation is in progress
	case 30:
		{
			if(option==1)
			{
				if(!pWCDI->IsComputationInProgress())
				{
					_beginthread(ThreadControlFunc,0,this);
					//SetThreadPriority(pWinThread->m_hThread,THREAD_PRIORITY_ABOVE_NORMAL);
				}
				else
				{
					pWCDI->StopWhenPossible();
				}
			}
			rv = pWCDI->IsComputationInProgress();
			break;
		}

	default:
		AddText("unrecognized action code ");
		AddText(action);
		AddText("\n");
	}
	return rv;
}

char* ConsoleApplication::GetHotintVersionText()
{
	char* rv = pWCDI->GetHotintVersion().GetString();
	return rv;
}

void ConsoleApplication::ModelChanged()
{
	pWCDI->OpenLogFile(0);
	SetModelModified();
	pWCDI->CallCompFunction(105); //Clear stored initial vector

	//pWCDI->SetDOption2, 0.); 
	//pWCDI->MBS_EDC_TreeSetDouble(0.,"SolverOptions.start_time"); //reset start time, otherwise nothing is computed the second time
	pWCDI->CallCompFunction(105); //$ DR 2014-10-27 no more access to MBS_EDC... reset start time
	//DialogDataManager.RemoveAll(); //remove stored output
}
	//go to sleep for x milliseconds
void ConsoleApplication::SleepX(int xMilliseconds)
{
	Sleep(xMilliseconds);
}


void ConsoleApplication::StartComputation()
{
	if(!pWCDI->IsComputationInProgress())
	{
		//start file watcher (allows user to steer computation)
		mystr command_file_path = pWCDI->GetOptions()->GeneralOptions()->PathsCommandFilePath();
		if (command_file_path.Length()>0)
		{
			mystr homedir = pWCDI->GetAbsolutePath(command_file_path);
			std::string path(homedir.c_str());
			watcher.SetDir(CFileSystemWatcher::s2ws(path));
			watcher.AddFileChangeListener(ca);
			watcher.Start();
		}

		AddText("Starting computation ...\n");

		try
		{
			SetExitCode(0);
			pWCDI->Go(this);			
		}
		catch(int c)
		{ 
			int a=c;
			SetExitCode(a);
		}
	}
}

//$ PG 2021-03-17: crashes on SleepX [
//void ConsoleApplication::RestartComputation()
//{
//	pWCDI->StopWhenPossible();
//
//	while (pWCDI->IsComputationInProgress())
//	{
//		SleepX(1000);
//	}
//	pWCDI->ResetComputation();
//	StartComputation();
//}
//$ PG 2021-03-17: ]

void ConsoleApplication::StopComputation()
{
	if(pWCDI->IsComputationInProgress())
	{
		AddText("Stopping..\n");
		watcher.Stop();
		pWCDI->StopWhenPossible();
	}
}

void ConsoleApplication::ResultsUpdated(int flag)
{
	if((flag==2 || flag==0)&& GetWCDInterface()->StoreResultsIsOn())
	{
		DummyDataStorage ds; double m_TimePoint;
		GetWCDInterface()->StoreResults(ds, m_TimePoint);
	}
}


// Inherited via IFileWatcherListener
void ConsoleApplication::OnFileChange(const std::wstring & path)
{
	//AddText("OnFileChange");
	if (path.compare(L"command.txt") == 0) //compare returns 0 if identical
	{
		mystr homedir = pWCDI->GetOptions()->GeneralOptions()->PathsRelativePathsRelativeToApplication() ? pWCDI->GetOptions()->GeneralOptions()->PathsApplicationPath() : pWCDI->GetOptions()->GeneralOptions()->PathsHotintInputDataPath();
		mystr filepath = homedir + mystr("command.txt");
		
		// read command.txt
		FILE* fCommand = fopen(filepath.c_str(), "r");
		char tCommand[100];

		SleepX(1000);
		fscanf(fCommand, "%s", tCommand);

		if (string(tCommand).compare("stop") == 0)
		{
			AddText("received signal to stop execution!");
			StopComputation();
		}/*
		else if (string(tCommand).compare("restart") == 0)
		{
			ca->AddText("received signal to restart computation!");
			RestartComputation();
		}*/
		fclose(fCommand);
	}
}

// Inherited via IFileWatcherListener
void ConsoleApplication::OnFileAdded(const std::wstring & path)
{
	//AddText("OnFileAdded");
}

// Inherited via IFileWatcherListener
void ConsoleApplication::OnFileRemoved(const std::wstring & path)
{
	//AddText("OnFileRemoved\n");
}

// Inherited via IFileWatcherListener
void ConsoleApplication::OnFileRenamed(const std::wstring & path)
{
	//AddText("OnFileRenamed\n");
}