//#**************************************************************
//# filename:             MBSGraphics3D.h
//#
//# author:               Simon Weitzhofer
//#
//# generated:            February 2014
//# description:          Contains 3D drawing functions
//# remarks:              These functions have been in the TimeInt class
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#pragma once

//see developers documentation

class RenderContext;
#include "..\MBS_Interface\mbs_interface.h"

//class for 3D drawing routines
class MBSGraphics3D : public MBSGraphics3DInterface
{
public:
	MBSGraphics3D(MBSOptionsInterface* mbs_options);
	MBSGraphics3D() {}
	~MBSGraphics3D()
	{
		if (texImage != NULL)
			delete[] texImage;
	}

	void Init();
	void InitFirst();

	//return the current render context
	RenderContext* GetRC() { return pCurrentRC; }
	//sets the render context
	void SetRC(RenderContext * rc) { pCurrentRC = rc; }
	//sets the MBSOptionsInterface
	void SetMBSOptions(MBSOptionsInterface * mbs) { mbs_options = mbs; }

	void DrawLegend(double yoff=0) const;

	void DrawColorQuads(const TArray<Vector3D>& p, const TArray<double>& v, int n1, int n2, 
		int colormode, int drawlines=0, int vres=1); //draw n1*n2 quads with FEcolors, interpolate values v with higher resultion vres

	void DrawQuad(const Vector3D& p1,const Vector3D& p2,const Vector3D& p3,const Vector3D& p4) const;
	void DrawQuad(const Vector3D& p1,const Vector3D& p2,const Vector3D& p3,const Vector3D& p4, const Vector3D& n1,const Vector3D& n2,const Vector3D& n3,const Vector3D& n4) const;

	//Draws a triangle with the vertices at p1, p2 and p3
	void DrawColorTrigs(const TArray<Vector3D>& points, const TArray<double>& values, const TArray<Vector3D>& normals, int colormode, int drawlines, int vres);
	void DrawTrig(const Vector3D& p1,const Vector3D& p2,const Vector3D& p3) const;

	//Draws a hexahedron with vertices p1, ..., p8
	void DrawHex(const Vector3D& p1,const Vector3D& p2,const Vector3D& p3,const Vector3D& p4,
		const Vector3D& p5,const Vector3D& p6,const Vector3D& p7,const Vector3D& p8, int drawouterfaces=1) const;
	
	void DrawCube(const Vector3D& p0, const Vector3D& v1, const Vector3D& v2, const Vector3D& v3) const; //draws cube with reference point and 3 axes
	
	void MyDrawRectangle(const Vector3D& p1, const Vector3D& p2, const Vector3D& p3, const Vector3D& p4, double thickness, const Vector3D* colline, const Vector3D* colfill=0) const;

	void MyDrawLine(const Vector3D& p1, const Vector3D& p2, double thickness) const;
	void MyDrawLine(const Vector3D& p1, const Vector3D& p2, double thickness, const Vector3D& col) const; 
	void MyDrawLine(const Vector3D& p1, const Vector3D& p2, double t, double h) const
	{
		MyDrawLineH(p1,p2,p2-p1,t,h);
	}
	void MyDrawLineH(const Vector3D& p1, const Vector3D&p2, const Vector3D& vy2, 
		double t, double h, int drawouterface=1) const;

	void DrawBox(const Vector3D& pmin, const Vector3D& pmax, const Vector3D& color = Vector3D());
	void DrawBox(const Matrix3D& rotation, const Vector3D& pmin, const Vector3D& pmax, const Vector3D& color = Vector3D());

	void MyDrawCircleXY(const Vector3D p, double r, const Vector3D& col, int res=12, double thickness=1) const;
	void MyDrawArrow(const Vector3D& p1, const Vector3D&p2, const Vector3D& col, 
		double linethickness = -1, double headsize = -1, int resolution=8) const;
	void DrawArrow(const Vector3D& p1, const Vector3D& p2, double linethickness = -1, double headsize = -1, int resolution = 8) const;
	void DrawColorArrow(double v, const Vector3D& p1, const Vector3D& p2, double linethickness = -1, double headsize = -1, int resolution = 8);

	
	void DrawColorZyl(double v, const Vector3D& pz1,const Vector3D& pz2, double rz, int tile);     //draw a zylinder in 3D, with color specified by value v, endpoints p1 and p2, Radius r, number of surfaces (discretized): tile
	void DrawZyl(const Vector3D& pz1, const Vector3D& pz2, double rz, int tile=8, double ri = 0., const Vector3D& vp1 = Vector3D(1.), int twocolors = 0) const;
	void DrawCone(const Vector3D& pz1, const Vector3D& pz2, double rz, int tile=8) const;
	void DrawZyl(const Vector3D& pz1, const Vector3D& pz2, const Vector3D& pz1dir,const Vector3D& pz2dir, double rz, int leftend = 1, int rightend = 1, int tile=8) const;

	//tile ... number of tiles per circumference, fill:1 ...draw whole sphere, fill=0.5 ... draw half sphere
	void DrawSphere(const Vector3D& p1, double r, int tile=8, double fill=1) const;
	void DrawColorSphere(double value, const Vector3D& p, double r, int tile=8, double fill=1);
	void DrawPolygon(const TArray<Vector3D>& p) const;
	void DrawPolygonOutline(const TArray<Vector3D>& p) const;

	// get HSL color specification from RGB
	Vector3D RGB2HSL(const Vector3D& rgb) const;
	// get RGB color specification from HSL
	Vector3D HSL2RGB(const Vector3D& hsl) const;

	Vector3D Lighten(const Vector3D& color, double factor) const;
	void SetColor(const Vector3D& col); //set new color (set actcolor)
	void SetColor(const Vector3D& col, float transparency); //set new color (set actcolor and acttransparency)

	void SetLineColor(const Vector3D& col) {colline = col;}; //set new color
	void SetLineThickness(double thickness) {linethickness = thickness;}
	const Vector3D& GetLineColor() {return colline;}; //set new color
	void SetTransparency(float transp) {transparency = transp; if (transparency>=0) acttransparency = transp;}; // <0 ... use global transparency, 0 ... invisible, 1 ... solid
	float GetTransparency() const { return transparency;}; // <0 ... use global transparency, 0 ... invisible, 1 ... solid
	void ChooseColor(float R, float G, float B) const; //choose color for painting
	void ChooseColor(float R, float G, float B, float transparency) const;
	//void SetDrawlines(int i) {drawlines = i;} //$ SW 2014-4-11: renamed to SetDrawOutline
	//void SetDrawlinesH(int i) {drawlinesh = i;} //$ SW 2014-4-11: removed since it was not used
	void SetDrawFaces(int i) {drawfaces = i;}
	void SetDrawOutline(int i) {drawoutline = i;}

	//int GetDrawlines() {return drawlines;} //$ SW 2014-4-15: replaced by DrawOutline
	//int GetDrawlinesH() {return drawlinesh;} //$ SW 2014-4-15: removed: was not used
	int GetDrawFaces() {return drawfaces;}
	int GetDrawOutline() {return drawoutline;}

	Vector3D FEColor(double val) const;
	void SetFEColor(double val) const;
	void DrawIsoQuad(Vector3D* p, const Vector3D& n, double* v); //draw quad with iso-lines
	void DrawIsoTrig(const TArray<Vector3D>& p, const TArray<Vector3D>& n, Vector3D vals); //draw quad with iso-lines
	void DrawIsoQuadTex(Vector3D* p, const Vector3D& n, double* v, int res);

	const double& GetTImincol() const {return TImincol;}
	double& GetFEmincol() {return TImincol;}
	const double& GetTImaxcol() const {return TImaxcol;}
	double& GetFEmaxcol() {return TImaxcol;}
	int IsFlagComputeMinMaxFECol() const {return flag_compute_minmax_FEcol;}
	void SetFlagComputeMinMaxFECol(int flag) {flag_compute_minmax_FEcol = flag;}

	double TImincol;
	double TImaxcol;
	int flag_compute_minmax_FEcol;

	void UpdateFEMinMaxCol(double val);

	//double tidraw_offx; //$ SW 2014-4-11: Removed since it was not used
	//double tidraw_offy;
	//double tidraw_offz;

	Vector3D actcolor, colline;
	float acttransparency;
	Vector3D & ColLine() { return colline; }	// for interface access
	//int drawlines;
	//int drawlinesh;
	int drawfaces;
	int drawoutline;
	float transparency;
	int linethickness;
	void SetTexStoreResolution(int n);
	int GetTexStoreResolution() const {return texstorenn;}
	int GetTexStoreMaxResolution();
	void SetTexImage(const char* texImage);
	const char* GetTexImage() const;

	void ResetFlags();
	void ResetColor() const;

	void PrintText3D(float x, float y, float z, const char* text);
	void PrintTextStruct(int nLineNo, int nXPos, const char * text);

private:
	int TIdrawtexres;
	int texstorenn;  //size of texture, minimum requirement: 64x64 for opengl
	int reduceimage; //reduce size of image, such that only a part of it is used
	char* texImage;  //pointer to texture image
	unsigned int texName;
	RenderContext* pCurrentRC;
	MBSOptionsInterface* mbs_options;
	
};
