//#**************************************************************
//# filename:             computeeigenmodes.cpp
//#
//# author:								Astrid Pechstein and Karin Nachbagauer            
//#
//# generated:						
//# description:          
//# comments:	
//#							Computes generalized eigenvalues and eigenmodes of the Eigenvalue problem K*v = lambda * M * v
//#							Eigenmodes are loaded into Hotint Visualization
//#							three different options:
//#							 * direct computation: Use Lapack to compute all eigenvalues, full matrices are used
//#							 * sparse computation in Matlab:
//#							   - K and M are written in Matlab-(Sparse-)File
//#							   - gen. EVproblem is solved in Matlab (Arnoldi iteration)
//#							 * sparse computation by lobpcg iteration in Hotint
//#							   - gen. EVproblem is solved by Hotint implementation of lobpcg (Knyazev)
//#							 eigenvalues are loaded for Hotint visualization and can be stored using data manager
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************


#include "mbs.h"


#include "windows.h"					//for shell execute
#include <direct.h>						//for getcwd
#include  <io.h>						//file operation _access
#include  <stdio.h>
#include  <stdlib.h>
#include "lapack_routines.h"

//$ YV 2012-12-14
//by now we explicitly reference particular elements:
// we need here to know, which degrees of freedom are constrained
#include "element.h"
#include "spectra_sparse_ev_solver.h"

//#define __TIMING_COMPUTEEIGENMODES__
#define __DISALLOW_IS_ES__  //$ PG 2020-05-05: We need a concept for integration of algebraic or first order equations. Meanwhile computations of this kind are disallowed.

int MultiBodySystem::ComputeEigenmodes()
{
	UO(UO_LVL_sim) << "\nEV solver -> starting modal analysis\n";

#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStartTimer(28); // assembling
#endif

	//$ SW 2018-9-26:  copy solver-settings
	numsol.NumDiffepsi() = NLS_NumDiffepsi();
	numsol.SymmetricJacobian() = NLS_SymmetricJacobian();
	numsol.UseSparseSolver() = NLS_UseSparseSolver();
	numsol.SolveUndeterminedSystem() = NLS_SolveUndeterminedSystem();
	numsol.EstimatedConditionNumber() = NLS_EstimatedConditionNumber();
	
	double lt = GetClockTime();
	double start_clock_time = lt;

	bComputeEigenmodes = 1; 
	bComputationIsInProgress = 1;
	bStopWhenPossible = 0;
	bPause = 0;

	SetProhibitRedraw(1);			//$ DR 2013-10-01 to avoid crash during computation (if animate scaling factor is active)

	//SetComputationSolverOptions();	//$ DR 2014-10-24 not necessary anymore

	int use_gyroscopic_terms = GetSolSet().eigsolv_gyro;

	//UO() << "old solution: =" << TIx0 << "\n";
	Vector sol = TIx0;

	//$ DR 2013-02-22 check if compute eigenmodes is called from a cpp-model or a hid-model
	mystr model_name = GetOptions()->GeneralOptions()->ModelFileInternalModelFunctionName();
	int nm = GetModelFunctionsIndex0();
	
	//$ SW 2013-11-5 [: in order to calculate the eigenmodes an initialization is necessary.
	// Doing an initialization at the original model would dismiss all the changes done since
	// the last calculation. Hence, if the model is from a HID file we can save the model
	// in a temporary hmc file and do the calculations there.
	// If the model is a cpp model we have to dismiss all changes anyway since we cannot
	// save cpp files.
	mystr hotint_input_data_file;
	if (nm == -1) //not an internal model
	{
		//hotint_input_data_file = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.ModelFile.hotint_input_data_filename", "");
		hotint_input_data_file = GetOptions()->GeneralOptions()->ModelFileHotintInputDataFilename(); //$ DR 2014-10-23 improve speed

		//save current configuration
		ElementDataContainer edc;
		ElementData ed;

		//mystr path = MBS_EDC_TreeGetString("GeneralOptions.Paths.application_path");
		mystr path = GetOptions()->GeneralOptions()->PathsApplicationPath(); //$ DR 2014-10-21 improve speed
		mystr asv_model_name = "model_asv_tmp.hmc";
		ed.SetText(asv_model_name, "File_name"); 
		edc.Add(ed);
		ed.SetText(path, "Directory_name");
		edc.Add(ed);
		CallCompFunction(101, 1, 0, &edc); //save current configuration

		//set the model to be the temporary file and initialize this file
		//GetMBS_EDC_Options()->TreeSetString("GeneralOptions.ModelFile.hotint_input_data_filename", path + asv_model_name);
		GetOptions()->GeneralOptions()->ModelFileHotintInputDataFilename() = mystr(path + asv_model_name); //$ DR 2014-10-24 improve speed
		CallCompFunction(20,3); //Initialize(), each calculation of the eigensystem should start with XG=0 (otherwise Kmat wrong)
	}
	else
	{ // the current model is an internal model
		CallCompFunction(20,3); //Initialize(), each calculation of the eigensystem should start with XG=0 (otherwise Kmat wrong) 
	}
	//$ SW 2013-11-5:]

	UO(UO_LVL_sim) << "EV solver -> checking SolverOptions.Linalg.use_sparse_solver: " << GetSolSet().nls_usesparsesolver << "\n";
	int option = GetSolSet().nls_usesparsesolver ? 3 /*Spectra*/ : 0 /*Lapack*/;	
	if (GetSolSet().eigsolv_solvertype != -1 && GetSolSet().eigsolv_solvertype != option)
		option = GetSolSet().eigsolv_solvertype;
	
	UO(UO_LVL_sim) << "EV solver -> using ";
	switch (option)
	{
	case 0: UO(UO_LVL_sim) << "LAPACK (direct method, full matrices)"; break;
	case 1: UO(UO_LVL_sim) << "MATLAB (iterative method, full matrices)"; break;
	case 2: UO(UO_LVL_sim) << "LOB-PCG (iterative method, sparse matrices)"; break;
	case 3: UO(UO_LVL_sim) << "SPECTRA (iterative method, sparse matrices)"; break;
	}
	UO(UO_LVL_sim) << " - choose solver_type to override\n";

	// GSxs --> starting vector
	//TIx0 contains initial values
	mystatic Vector GSxs;
	int sos= GetSecondOrderSize();//Number of 2nd order ODEs
	int es = GetFirstOrderSize(); //Number of 1st order ODEs
	int is = GetImplicitSize();   //Algebraic Equations (from boundary conditions) 

	//use actual solution for linearization (e.g. for pre-stressed problems):
	int linearize_about_actual_solution = GetSolSet().eigsolv_linearize_about_act_sol;
	UO(UO_LVL_sim) << "EV solver -> read linearize_about_actual_solution: " << linearize_about_actual_solution  << "\n";
	if (linearize_about_actual_solution && sol.Length()!=TIx0.Length())
	{
		UO(UO_LVL_sim) << "EV solver -> no valid actual solution vector found, linearizing about initial state\n";
		linearize_about_actual_solution = 0;
	}
	
	if (!linearize_about_actual_solution)
	{
		sol = TIx0; //$ PG 2020-04-10: TIx0 has been initialized before by CallCompFunction(20,3);
	}

	// store actual state separately for later visualization, modes are transferred from EigenvectorMatrix to Datamanager
	TIx0start.SetLen(sos);
	for (int i = 1; i <= sos; i++)
		TIx0start(i) = sol(i);
	
	GSxs.SetLen(sos+es+is);
	//GSxs is filled with initial values
	for (int i=1; i <= sos; i++) GSxs(i) = sol(i);
	for (int i=1; i <= es; i++) GSxs(i+sos) = sol(i+2*sos);
	for (int i=1; i <= is; i++) GSxs(i+sos+es) = sol(i+2*sos+es);


#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStartTimer(27); // assembling: compute mass matrix (and stiffness matrix in case !use_gyroscopic_terms)
#endif

	// Stiffness Matrix
	SparseMatrix Kmat; //sparse matrix build
	if (! use_gyroscopic_terms) // $ MSax 2013-06-25 : added because otherwise velocities whould be set to zero
	{
		StaticJacobian(Kmat,GSxs);//from starting vector GSxs the stiffness matrix Kmat is generated
		Kmat *= -1;
	}
	// Mass Matrix
	SparseMatrix Mmat;
	Mmat.SetSize((sos+es+is),(sos+es+is), MaxSparseBandwidth());
	if (! use_gyroscopic_terms) // $ MSax 2013-06-25 : added because otherwise velocities whould be set to zero
	{
		EvalM(GSxs, Mmat, 0);
	}
	else
	{
		EvalM(sol, Mmat, 0);
	}

#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStopTimer(27); // assembling: compute mass matrix (and stiffness matrix in case !use_gyroscopic_terms)
#endif

#ifdef __DISALLOW_IS_ES__
	if (use_gyroscopic_terms  && (is != 0 || es != 0)) 
	{
		UO(UO_LVL_err) << "EV solver -> ERROR: Compute Eigenmodes: No constraints or first order equations allowed when including damping or gyroscopic effects -> setting include_damping_and_gyroscopy = 0\n";
		use_gyroscopic_terms = 0;
	}
#else
	if (is > 0 || es > 0)
	{
		UO(UO_LVL_err) << "EV solver -> ERROR: Compute Eigenmodes:\n";
		if (es > 0)
			UO(UO_LVL_err) << "ES()=" << es << ": No first order equations allowed!\n";
		if (is > 0)
			UO(UO_LVL_err) << "IS()=" << is << ": No algebraic equations allowed! Use penalty formulation instead!\n";
		return 1;
	}
#endif

	Matrix A;
	SparseMatrix ASparse; // [0, M; K, G+D]
	SparseMatrix DM;      // [M, 0; 0, M]

	//ofstream testfile(GetSensorOutputPath()+ mystr("test_") + GetSolSet().sol_filename);

	if (use_gyroscopic_terms) // MSax: use gyroscopic terms
	{		
		SparseMatrix Dmat;

		Kmat.SetSize((sos+es+is),(sos+es+is)); // init column size is set in ComputeStiffnessAndDampingMatrix
		Dmat.SetSize((sos+es+is),(sos+es+is)); // init column size is set in ComputeStiffnessAndDampingMatrix

#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStartTimer(26); // assembling: compute Kmat and Dmat (in case use_gyroscopic_terms)
#endif
		ComputeStiffnessAndDampingMatrix(Kmat, Dmat, sol);
#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStopTimer(26); // assembling: compute Kmat and Dmat (in case use_gyroscopic_terms)
#endif

		SparseMatrix GYmat;
#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStartTimer(25); // assembling: compute gyroscopic matrix (in case use_gyroscopic_terms)
#endif
		ComputeGyroscopicMatrix(GYmat); // GY..Gyroscopic matrix
#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStopTimer(25); // assembling: compute gyroscopic matrix (in case use_gyroscopic_terms)
#endif
		
#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStartTimer(24); // assembling: add all matrices in system matrices A and MM, with A = (0 M; K G), MM = (M 0; 0 M)
#endif
		if ((option==2 || option==3)/* && 2*GetSolSet().eigsolv_n_eigvals+1 <= 2*sos*/)
		{
			ASparse = SparseMatrix(2*sos, 2*sos, 2*MaxSparseBandwidth());
			ASparse.FillWithZeros();
			DM = SparseMatrix(2*sos, 2*sos, MaxSparseBandwidth());
			
			ASparse.AddSubmatrix(Mmat,  1, 1, 1      , 1 + sos, sos, sos, 1.);
			ASparse.AddSubmatrix(Kmat,  1, 1, 1 + sos, 1      , sos, sos, 1.);
			ASparse.AddSubmatrix(GYmat, 1, 1, 1 + sos, 1 + sos, sos, sos, 1.);
			ASparse.AddSubmatrix(Dmat,  1, 1, 1 + sos, 1 + sos, sos, sos, 1.);

			DM.AddSubmatrix(Mmat, 1, 1, 1      , 1      , sos, sos, 1.);
			DM.AddSubmatrix(Mmat, 1, 1, 1 + sos, 1 + sos, sos, sos, 1.);
		
			option = 3;
			if (option == 2)
				UO(UO_LVL_warn) << "EV solver -> WARNING: type 2 selected, but switching to type 3 was directed.\n";
		}
		else
		{
			Matrix Mfull; 
			Mmat.GetMatrix(Mfull);

			Matrix Kfull; 
			Kmat.GetMatrix(Kfull);

			Matrix GYfull; 
			GYmat.GetMatrix(GYfull);

			Matrix Dfull;
			Dmat.GetMatrix(Dfull);

			Matrix Id1(sos,sos); 
			Id1.SetDiagMatrix(1.);

			// build  A = [0,E;-M^(-1)*K,-M^(-1)*G] matrix, see HID_USER docu
			A.SetSize(sos*2,sos*2);
			A.FillWithZeros(1,1,sos,sos); //$ SW 2014-8-13: fill left upper area with zeros
			A.SetSubmatrix(Id1,1,1+sos);
			
			int rv=Mfull.Invert2();
			if (!rv) UO(UO_LVL_err) << "ERROR: ComputeEigenmodes: mass matrix not invertable!\n";

			Kfull= Mfull * Kfull;
			GYfull = Mfull * (GYfull + Dfull);

			A.SetSubmatrix(GYfull,1+sos,1+sos);
			A.SetSubmatrix(Kfull,1+sos,1);
		}

#ifdef __TIMING_COMPUTEEIGENMODES__
		TMStopTimer(24); // assembling: add all matrices in system matrices A and MM, with A = (0 M; K G), MM = (M 0; 0 M)
#endif
	}

	int neig;
	int size;
	int numberzeromodes = 0;
	IVector unconstraineddofs(0);
	if (!use_gyroscopic_terms) // $ MSax 2013-07-04 : added
	{
		//$ DR 2013-02-22 error handling added
		if(Kmat.Getcols() * Mmat.Getcols()==0)
		{
			UO(UO_LVL_err) << "\n ERROR: The K- and/or M-matrix has size == 0! Does the system contain elements at all? \n Stopped computation of eigenmodes.\n";
			bComputationIsInProgress = 0;
			SetProhibitRedraw(0); //$ DR 2013-10-01 to avoid crash during computation (if animate scaling factor is active)

			return 0; 
		}

		// List of constrained dofs, which are eliminated from matrix
		TArray<int> CClist;
		CClist.SetLen(0);

		for(int i=1; i <= GetNElements(); i++)//pass all elements
		{
			Element& elem = GetElement(i);//set pointer elem to each element
			//$ YV 2013-01-03: constrained dofs are provided by the basis Element class, so that there is
			// no need to explicitly differentiate Element types here
			if(elem.IS() != 0)		// we consider only the constraints with Lagrange multipliers
				elem.GetGlobalConstrainedDOFs(CClist);
	#if 0
			//we look for CoordinateConstraints:
			CoordConstraint* ccelem = dynamic_cast<CoordConstraint*>(&elem);//we ask, if elem is of type CoordConstraint
			//->0, if elem is not a CoordConstraint, else ->Pointer to elem, if type is CoordConstraint
			if(ccelem!=0)//if pointer is part of ccelem, pointer is added to the list ConstrDOF
			{
				if (ccelem->IS() == 0) continue; // only coord-constraints with Lag.Mult. (IS==1) are of interest, cc with IS==0 are penalty, matrix row is kept
				int GlobConstrNr;
				GlobConstrNr=ccelem->GetGlobalConstrainedDOF();//from local number ccelem a global DOFnumber is generated
				//matout << GlobConstrNr << " ";//add global DOF-Number to the list
				CClist.Add(GlobConstrNr);
			}
			// $LA 2011-08-25:[
			//we look for NodalConstraints:
			NodalConstraint* ncelem = dynamic_cast<NodalConstraint*>(&elem);//we ask, if elem is of type NodalConstraint
			//->0, if elem is not a NodalConstraint, else ->Pointer to elem, if type is NodalConstraint
			if(ncelem!=0)//if pointer is part of ncelem, pointer is added to the list ConstrDOF
			{
				if (ncelem->IS() == 0) continue; // only nodal-constraints with Lag.Mult. (IS==1) are of interest, cc with IS==0 are penalty, matrix row is kept
				TArray<int> GlobConstrNrs;

				// LGA global constrainedDOF ausfindig machen!!!
				ncelem->GetGlobalConstrainedDOFs(GlobConstrNrs);//from local number ccelem a global DOFnumber is generated

				//matout << GlobConstrNr << " ";//add global DOF-Number to the list
				for(int j=1; j<= GlobConstrNrs.Length(); j++)
				{
					CClist.Add(GlobConstrNrs(j));
				}
			}
			// $LA 2011-08-25:]
	#endif
		}
		Quicksort(CClist);
		// Vector containing the degrees of freedom which are not constrained due to coord constraints
		for (int i=1; i<=sos; i++)
		{
			if (!CClist.Find(i))
				unconstraineddofs.Add(i);
		}
		//number of calculated eigenvectors
		if (option==0)
		{
			neig = unconstraineddofs.Length();
			// if (use_gyroscopic_terms) neig = 2*sos; $ MSax 2013-07-04 : removed
		}
		else
		{
			neig = GetSolSet().eigsolv_n_eigvals;
		}
		size = unconstraineddofs.Length();
	}
	else
	{
		if (option == 3)
		{
			neig = GetSolSet().eigsolv_n_eigvals;
			if (neig > sos - 1)
			{
				UO(UO_LVL_warn) << "EV solver -> WARNING: n_eigval = " << neig << ", but upper limit is " << sos-2 << " (tip: solver type 0 computes all EVs, but not sparse)\n";
				neig = sos - 2;
			}
		}
		else
		{
			neig = sos; // $ MSax 2013-07-04 : added
		}
	}

#ifdef __TIMING_COMPUTEEIGENMODES__	
	TMStopTimer(28);  // assembling
#endif


#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStartTimer(29); // computation
#endif

	UO(UO_LVL_sim) << "EV solver -> read n_eigvals: " << GetSolSet().eigsolv_n_eigvals;
	if (GetSolSet().eigsolv_n_eigvals - neig > 0)
		UO(UO_LVL_sim) << ", but only " << neig << " eigen values can be computed";
	UO(UO_LVL_sim) << "\n";
	


	// ---------- start computation ---------
	int computedcorrectly = 0;
	if (use_gyroscopic_terms) // $ MSax: added
	{
		if (option == 3)
		{
			// solving dual system ASparse [q lq] = l DM [q lq] of size sosdual = 2*sos
			// with ASparse=[0 M; -K -D] and DM=[M 0; 0 M]
			// solution: complex valued eigenvectors qk, and complex valued eigenvalues lk (for k in 1..sos_dual)
			// if imag(lk) does not vanish, then also the conjugate of (lk, qk) is eigen pair
			// since user wants neig eigenpairs to be returned we in general have to look for neig_dual=2*neig solutions in spectra
			// however, sectra iteratively solves a sos_dual sized system, and the number of requested solutions is bounded by neig_dual <= sos_dual - 2 (thus neig <= sos - 1)
			// size of working space (ncv_dual) should to be (optimally) at around neigdual*2 + 1, but bounded by 2*sos

			int sos_dual = 2*sos;
			int neig_dual = 2*neig;

			Vector wr(neig_dual); // vector with real part of eigenvalues
			Vector wi(neig_dual); // vector with imaginary parts of eigenvalues
			Matrix evr(neig_dual, sos_dual); // matrix with real parts of eigenvectors
			Matrix evi(neig_dual, sos_dual); // matrix with imag parts of eigenvectors

			double sigma = GetSolSet().eigsolv_use_preconditioning ? -GetSolSet().eigsolv_precond_lambda : 0.;

			int ncv_dual = (sos_dual + neig_dual*2 + 2)/2;
			ncv_dual = min(ncv_dual, sos_dual);
			int spectrainfo = spectra_solve_ev(&ASparse, &DM, neig_dual, ncv_dual, sigma, wr, wi, evr, evi, this);
			if (spectrainfo) 
				computedcorrectly=0;
			else 
				computedcorrectly=1;

			//$ MaSch 2015-06-29: gyroscopic terms lead to imaginary part only (no dissipation); a 2nd order system with n DOF yields n pairs of complex conjugate eigenvalues, i.e., n pairs of the form (+ib, -ib);
			//the n (real) solutions for the absolute values imaginary parts are given by Eq. (9.18) in Gasch, Rotordynamics, Chapter 9 
			TArray<double> evalues;
			TArray<int> index;
			for(int j=1; j<=neig_dual; j++) 
			{
				evalues.Add(wi(j));
				index.Add(j);
			}
			QuicksortDouble(evalues, index); //sort evalues in ascending order; note: index set is not referring to evalues, but to evr, evi, wr, and wi.
			
			EigenvektorMatrix.SetSize(neig, sos_dual); // col (1, .. ,sos): real part    col (sos+1, .. ,sos_dual): imag part
			eigval.SetLen(neig);
			eigdamp.SetLen(neig);
			
			// jth and j+1st egenvalues and eigenvectors are conjugate complex, thus we either have to store + or -imaginary part, and we decide for +. However this has to be done consistently for both eigenvalues and -vectors.
			for(int j=1; j<=neig; j++)
			{
				int j_sorted = j + neig;
				int j_unsorted = index(j_sorted);

				double val = evalues(j_sorted);
				eigval(j) = val;  
				eigdamp(j) = wr(j_unsorted);
				for(int k=1; k<=sos; k++)
				{
					EigenvektorMatrix(j, k) = evr(j_unsorted, k);
					EigenvektorMatrix(j, k + sos) = (val == 0) ? 0. : evi(j_unsorted, k); // imaginary part of eigenvectors is contained in evi, and stored to (sos+1, ..., sos+sos) columns of EigenvectorMatrix
				}
			}
		}
		else
		{
			//$ MaSch 2015-06-29: gyroscopic terms lead to imaginary part only (no dissipation); a 2nd order system with n DOF yields n pairs of complex conjugate eigenvalues, i.e., n pairs of the form (+ib, -ib);
			//the n (real) solutions for the absolute values imaginary parts are given by Eq. (9.18) in Gasch, Rotordynamics, Chapter 9 

			Vector work(8*2*sos+16); // work vector for calculation (necessary for LAPACK)
			Vector wr(2*sos); // vector with real part of eigenvalues
			Vector wi(2*sos); // vector with imaginary parts of eigenvalues
			Matrix ev(2*sos,2*sos); // matrix with eigenvectors

			int lapackinfo = LapackGenEVP_CGEEV(2*sos, &A(1,1), &wr(1), &wi(1), &ev(1,1), &work(1), work.Length()); 
			if (lapackinfo) computedcorrectly=0;
			else computedcorrectly=1;

			TArray<double> evalues(0);
			TArray<int> index(0);

			for(int j=1; j<=2*sos; j++) 
			{
				evalues.Add(wi(j));
				index.Add(j);			
			}
			QuicksortDouble(evalues, index);

			// we now know: evalues have 2*sos sorted entries, and we return #(sos+1, .., sos+sos) of those, and set nel=sos for later use
			neig = sos; // evalues.Length()/2;
			EigenvektorMatrix.SetSize(sos, 2*sos);
			eigval.SetLen(sos);
			eigdamp.SetLen(sos);
			// jth and j+1st egenvalues and eigenvectors are conjugate complex, thus we either have to store + or -imaginary part, and we decide for +. However this has to be done consistently for both eigenvalues and -vectors.
			for(int j=1; j<=sos; j++)
			{
				int j_sorted = j + sos;
				int j_unsorted = index(j_sorted);
				double val = evalues(j_sorted);
				eigval(j) = val;     // imaginary part: frequency
				eigdamp(j) = wr(j_unsorted);       // real part: exponential damping ratio
				for(int k=1; k<=sos; k++)
				{
					EigenvektorMatrix(j, k) = ev(j_unsorted, k);
					EigenvektorMatrix(j, k + sos) = (val == 0) ? 0. : ev(j_unsorted + 1, k); // jth column of ev contains real part, j+1st column contains imaginary part if imaginary part of eigenvalue is not 0.
				}
			}
		}
	}
	else
	{
		EigenvektorMatrix.SetSize(neig,sos);
		EigenvektorMatrix.SetAll(0);
		
		//Solver Parameter (optional)
		int maxit = GetSolSet().eigsolv_maxiterations; //GetIOption(224);
		double tol = GetSolSet().eigsolv_accuracy; //GetDOption(225);
		numberzeromodes = GetSolSet().eigsolv_nzeromodes; //GetIOption(226);
		int use_preconditioning = GetSolSet().eigsolv_use_preconditioning; //GetIOption(228);
		double lambda_precond = GetSolSet().eigsolv_precond_lambda; //GetDOption(226);

		int init_random=1;
		SparseEigenvalueSolver evsolver(this);
		
		//evsolver.Set(&Kmat, &Mmat, maxit, tol, init_random);
		evsolver.Set(&Kmat, &Mmat, maxit, tol, init_random);

		// evsolver can be stopped by MBS
		evsolver.CanBeStopped() = 1;
		if (use_preconditioning)
			evsolver.SetPreconditioner(lambda_precond);
		//   MATLAB SPARSE SOLVER
		if (option == 1)
		{
			mystr pathMatlab="../../internal/tools/Matlab/ComputeEigensystem";
			computedcorrectly = evsolver.ComputeEigenModesMatlab(pathMatlab, neig, eigval, EigenvektorMatrix, unconstraineddofs,
															numberzeromodes);	
		}
		else if (option == 0) // Direct solver by LAPACK
		{
			Matrix Kfull, Mfull;
			neig = size; // compute all possible dofs
			eigval.SetLen(neig);
			Kfull.CopyFrom(Kmat,1,1,size,size, unconstraineddofs);
			Mfull.CopyFrom(Mmat,1,1,size,size, unconstraineddofs);
			if(UO().GetGlobalMessageLevel()==UO_LVL_dbg1)	//$ DR 2011-09-15
			{
				UO(UO_LVL_dbg1) << "K=" << Kfull << "\n";
				UO(UO_LVL_dbg1) << "M=" << Mfull << "\n";
			}
			Vector work(4*size);

			int lapackinfo = LapackGenEVPSPD(size, &Kfull(1,1), &Mfull(1,1), &eigval(1), &work(1), work.Length());
			if (lapackinfo) computedcorrectly=0;
			else computedcorrectly=1;

			// eigenvectors are stored rowwise in Matrix Kfull
			for (int j=1; j<=neig; j++)
			{
				for (int i=1; i<=size; i++)
				{
					EigenvektorMatrix(j,unconstraineddofs(i)) = Kfull(j,i);
				}
			}
		}
		else if (option == 2) // lobpcg - hack
		{
			eigval.SetLen(neig);
			computedcorrectly = evsolver.ComputeEigenModes(neig, eigval, EigenvektorMatrix, unconstraineddofs, numberzeromodes); 
		}
		else if (option == 3) // arnoldi
		{
			if (neig > sos - 2)
			{
				neig = sos - 2;
				UO(UO_LVL_sim) << "EV solver -> no more than " << neig << " eigenvalues can be computed\n";
			}

			Vector wr(neig); // vector with real part of eigenvalues
			Vector wi(neig); // vector with imaginary parts of eigenvalues
			Matrix evr(neig, sos); // matrix with eigenvectors
			Matrix evi(neig, sos);

			double sigma = GetSolSet().eigsolv_use_preconditioning ? -GetSolSet().eigsolv_precond_lambda : 0.;
			int ncv = neig*2 + 1;
			ncv = min(ncv, Mmat.Getrows());
			int spectrainfo = spectra_solve_ev(&Kmat, &Mmat, neig, ncv, sigma, wr, wi, evr, evi, this); 
			if (spectrainfo) computedcorrectly=0;
			else computedcorrectly=1;

			TArray<double> evalues(0);
			evalues.SetLen(neig);
			evalues.SetAll(0);
			TArray<int> index(0);
			index.SetLen(neig);
			index.SetAll(0);

			for(int j=1; j<=neig; j++) 
			{
				//$ MaSch 2015-06-29: gyroscopic terms lead to imaginary part only (no dissipation); a 2nd order system with n DOF yields n pairs of complex conjugate eigenvalues, i.e., n pairs of the form (+ib, -ib);
				//the n (real) solutions for the absolute values imaginary parts are given by Eq. (9.18) in Gasch, Rotordynamics, Chapter 9 
				evalues(j) = wr(j); //sqrt(sqr(wr(j))+sqr(wi(j))); // eigenvalues = (alphar + i alphai) / beta 
				index(j) = j;
				//UO(UO_LVL_ext) << "Index " << j << " : " << wr(j) << " + i* " << wi(j) << "\n";
			}
			QuicksortDouble(evalues, index);	

			eigval.SetLen(neig);
			for(int j=1; j<=neig; j++)
			{
				eigval(j) = evalues(j);
				for(int k=1; k<=sos; k++)
				{
					EigenvektorMatrix(j,k) = evr(index(j),k);
				}
			}
		}
	}

#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStopTimer(29); // computation
#endif

#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStartTimer(30); // sorting and storing
#endif
	
	UO(UO_LVL_sim) << "EV solver -> apply eigenmodes_normalization_mode: " << GetSolSet().eigsolv_normalization_mode << "\n";
	if(GetSolSet().eigsolv_normalization_mode==0)		//$ DR 2011-05-27: normalize such that max(abs(v)) = 1
	{
		for(int j=1; j <= neig; j++)//number of eigenvectors
		{
			double maxval = 0;
			for(int i=1; i <= sos; i++)//length of eigenvectors
			{
				double ejisq = Sqr(EigenvektorMatrix(j, i));
				if (use_gyroscopic_terms)
					ejisq += Sqr(EigenvektorMatrix(j, sos + i));
				maxval = Maximum(maxval, sqrt(ejisq));
			}

			if (maxval!=0)
			{
				if (use_gyroscopic_terms) // $ MSax 2013-07-04 : added
				{
					for(int i=1; i <= sos*2; i++) // real + imaginary part
					{
						EigenvektorMatrix(j, i) /= maxval;
					}
				}
				else
				{
					for(int i=1; i <= sos; i++) //length of eigenvectors
					{
						EigenvektorMatrix(j, i) /= maxval;
					}
				}
			}
		}
	}
	else if(GetSolSet().eigsolv_normalization_mode==1)	// normalize such that v'*v = 1
	{
		for(int j=1; j <= neig; j++)//number of eigenvectors
		{
			double norm = 0;
			for(int i=1; i <= sos; i++)//length of eigenvectors
			{
				norm += Sqr(EigenvektorMatrix(j, i));
				if (use_gyroscopic_terms)
					norm += Sqr(EigenvektorMatrix(j, sos + i));  // imaginary part
			}
			if (norm != 0)
			{
				norm = 1./sqrt(norm);
				if (use_gyroscopic_terms) // $ MSax 2013-07-04 : added
				{
					for(int i=1; i <= sos*2; i++) //real + imaginary part
					{
						EigenvektorMatrix(j, i) *= norm;
					}
				}
				else
				{
					for(int i=1; i <= sos; i++)//length of eigenvectors
					{
						EigenvektorMatrix(j, i) *= norm;
					}
				}
			}
		}
	}

	if (eigenvaluePostProcessingFunction)
	{
		(*eigenvaluePostProcessingFunction) (&EigenvektorMatrix, &eigval);
	}

	double fact = GetSolSet().eigsolv_eigenmodes_scaling_factor;
	if (fact != 1)
	{
		UO(UO_LVL_err) << "\n" << "EV solver -> apply eigenmodes_scaling_factor: " << fact  << "\n";
		EigenvektorMatrix *= GetSolSet().eigsolv_eigenmodes_scaling_factor; //scale eigenvectormatrix
	}	

	if (!use_gyroscopic_terms) // $ MSax 2013-07-04: otherwise already in rad/seconds
	{
		// eigenfrequency in rad/s
		for(int j=1; j <= neig; j++)//number of eigenvectors
		{
			eigval(j) = sqrt(fabs(eigval(j)));
		}
	}

	//eigenvectors are sent to the DataManager (for automized animation without file-loading)
	int orig_storedata = GetOptions()->ViewingOptions()->DataManagerStoreDataEvery();	//$ DR 2014-10-28 renamed option solset.storedata to ViewingOptions()->DataManagerStoreDataEvery()
	draw_eigenvectors = 1;
	pCFB->ResultsUpdated(3); // reset DataManager

	for(int i=1; i<=neig; i++)
	{
		SetTime(i);
		Vector& TIx0draw2 = GetSolVector();
		//Vector& TIx0draw2 = GetDrawVector();
		TIx0draw2.SetLen(EigenvektorMatrix.Getcols());
		TIx0draw2.SetAll(0.);
		if (use_gyroscopic_terms)
		{
			for (int j = 1; j<=sos; j++)
			{
				double c_real = EigenvektorMatrix(i, j);
				double c_imag = EigenvektorMatrix(i, sos + j);
				TIx0draw2(j) = sqrt(Sqr(c_real) + Sqr(c_imag));  // radius of complex number -> dof amplitude
				double val = 0, eps = 1e-8;
				if (fabs(c_real) < eps*fabs(c_imag))
					val = (c_imag < 0) ? -MY_PI/2 : MY_PI/2;
				else
				{
					val = atan(c_imag/c_real);
					if (c_real < 0)
						val += MY_PI;
				}
				TIx0draw2(sos + j) = val;
			}
		}
		else
		{
			for (int j = 1; j<=sos; j++)
			{
				TIx0draw2(j) = EigenvektorMatrix(i, j);
			}
		}

		drawnow = 0;
		GetOptions()->ViewingOptions()->DataManagerStoreDataEvery()=-2;
		TIDrawAndStore();
	}

	//GetSolSet().storedata = orig_storedata;
	GetOptions()->ViewingOptions()->DataManagerStoreDataEvery() = orig_storedata;	//$ DR 2014-10-28 renamed option solset.storedata to ViewingOptions()->DataManagerStoreDataEvery()


	//Output: Eigenfrequencies
	if (!computedcorrectly)
		UO(UO_LVL_err) << "EV solver -> computation not successful!\n";
	else
	{
		UO(UO_LVL_sim) << "EV solver -> computation successful\n";		
		// ----- output to solution data in output dir ----
		WriteSolDataInfoAppendEigenValues();
		// ----- print to output window ----
		WriteEigenValuesToOutputWindow(use_gyroscopic_terms);
	}
		
	//Message for user, if claimed number of zeromodes is not reached:
	if(numberzeromodes!=0)
	{
		if(eigval(numberzeromodes)>0.1)
		{
			UO(UO_LVL_err) << "\n" << "EV solver -> number of zero-modes NOT obtained!" << "\n";
		}
	}

	// --------- output to special file)---------  //$ PG 2020-04-17:  not so clear why... maby for SymSpace(?)... file may become really large with increasing sos
	// DR 2012-04-19: flag to change format of output added	
	int outp_format_flag = GetSolSet().eigval_outp_format_flag; // 1 .. eigenfreq., 2 .. eigenvec., 4 .. eigenfreq. in Hz (otherwise in rad/s)
	mystr outputfile = GetSensorOutputPath()+ mystr("eigenmodes_") + GetSolSet().sol_filename; //$ DR 2015-02-12
	UO(UO_LVL_sim) << "EV solver -> write eigenvalues and eigenvectors to output file " << outputfile << "\n";
	ofstream output(outputfile);
	if(outp_format_flag & 1)
	{
		for(int i=1; i<=neig; i++)
		{
			if(outp_format_flag & 4) 
				output << eigval(i)/2./MY_PI;
			else
				output << eigval(i);

			if (use_gyroscopic_terms)
				output << "\t" << -eigdamp(i);
			
			output <<"\n";
		}
	}

	if(outp_format_flag & 2)
	{
		for(int j=1; j<=EigenvektorMatrix.Getcols(); j++)
		{
			for(int i=1; i<=neig; i++)
			{
				output << EigenvektorMatrix(i,j) <<"\t";
			}
			output << "\n";
		}
	}
	output.close();

	//+++++++++++++++++++++++++++++
	lt = GetClockTime();
	TIcomptime = lt - start_clock_time;

	bComputeEigenmodes = 0;
	bComputationIsInProgress = 0;

	SetProhibitRedraw(0); //$ DR 2013-10-01 to avoid crash during computation (if animate scaling factor is active)

	//reset the filename to the original file name if changed
	if (nm == -1)
	{
		//GetMBS_EDC_Options()->TreeSetString("GeneralOptions.ModelFile.hotint_input_data_filename", hotint_input_data_file);
		GetOptions()->GeneralOptions()->ModelFileHotintInputDataFilename() = hotint_input_data_file; //$ DR 2014-10-24 improve speed
	}

#ifdef __TIMING_COMPUTEEIGENMODES__
	TMStopTimer(30); // sorting and storing
#endif

	return 0;
}

void MultiBodySystem::WriteEigenValuesToOutputWindow(int use_gyroscopic_terms)
{
	int neig = eigval.Length();
	if (UO().GetGlobalMessageLevel() > UO_LVL_multsim)
	{
		mystr table = "";

		table += mystr("\nEigenfrequencies");
		if (use_gyroscopic_terms)
			table += mystr(" & Exponential decay rate (exp(-d*t))");
		table += mystr(":\n");

		for (int i = 1; i<=neig; i++)
		{
			table += mystr(i) + mystr(", ") + mystr(eigval(i)/(2.*MY_PI)) + mystr(" [Hz] = ") + mystr(eigval(i)) + mystr(" [rad/s]");
			if (use_gyroscopic_terms)
				table += mystr(", d = ") + mystr(-eigdamp(i));
			table += mystr("\n");
		}
		UO(UO_LVL_sim) << table;
	}
	else
	{
		UO(UO_LVL_multsim) << "\n" << "Eigenfrequencies [Hz]:";
		for (int i = 1; i<=neig; i++)
		{
			UO(UO_LVL_multsim) << eigval(i)/(2.*MY_PI);
			if (i < neig) UO(UO_LVL_multsim) << ", ";
		}
		UO(UO_LVL_multsim) << "\n" << "Eigenfrequencies [rad/s]:";
		for (int i = 1; i<=neig; i++)
		{
			UO(UO_LVL_multsim) << eigval(i);
			if (i < neig) UO(UO_LVL_multsim) << ", ";
		}
		if (use_gyroscopic_terms)
		{
			UO(UO_LVL_multsim) << "\n" << "Exponential decay rate:";
			for (int i = 1; i<=neig; i++)
			{
				UO(UO_LVL_multsim) << -eigdamp(i);
				if (i < neig) UO(UO_LVL_multsim) << ", ";
			}
		}
	}
	UO(UO_LVL_sim) << "\n";
}

EigenvaluePostProcessingFunctionPtr MultiBodySystem::SetEigenvaluePostProcessingFunction( EigenvaluePostProcessingFunctionPtr func ){
	EigenvaluePostProcessingFunctionPtr old_func = eigenvaluePostProcessingFunction;
	eigenvaluePostProcessingFunction = func;
	return old_func;
}