//#**************************************************************
//#
//# filename:             mbs_communication.cpp
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						March 2010
//# description:          Load/Save mbs data, WCDriver and Windows communication 
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#include "windows.h" //for shell execute
#include <direct.h>  //for getcwd
#include  <io.h>     //file operation _access
#include  <stdio.h>
#include  <stdlib.h>

#include "mbs.h"
#include "parser.h"
#include "script_parser.h"
#include "myfile.h"
#include "element.h"
#include "sensors.h"
#include "node.h"
#include "material.h"
#include "set.h"
#include "mesh.h"
#include "elementdataaccess.h"


int MultiBodySystem::SaveSolutionVector(const mystr& filename, int omit_output_message)  //rv=1 --> OK
{
#if 0 //$ PG 2017-5-16: old redundant code

	int opt = ios::out;


	ofstream* storeout = new ofstream;
#ifdef my_new_stdiostream
	storeout->open(filename.c_str(), opt); //with read sharing allowed!
#else
	storeout->open(filename.c_str(), opt, filebuf::sh_read); //with read sharing allowed!
#endif

	if ((int)storeout->good() == 0 || storeout->fail())  
	{
		UO() << "ERROR: could not write solution!!!\n";
		UO() << "good=" << storeout->good() << ", fail=" << storeout->fail() << "\n";
		UO() << "file=" << filename.c_str() << "\n";

		//storeout->close();
		//storeout->delete;
		//return 0;
	}
	storeout->precision(19);
	int len = GetSolVector().Length();

	char str[256];
	sprintf(str, "HOTINTDataVersion%.3f", GetHotintVersion().GetDoubleValue());
	(*storeout) << str << "\n";
	(*storeout) << len << " 0" << "\n"; //checksums!
	(*storeout) << "1\n"; //only one step stored

	char strtime[64];
	char strtime2[64];
	sprintf(strtime, "%.14f", GetTime());
	sprintf(strtime2, "%.17f", GetTime());

	if (strlen(strtime) < strlen(strtime2)-4)
		(*storeout) << strtime << "\n";
	else
		(*storeout) << strtime2 << "\n";

	(*storeout) << len << "\n";
	for (int i = 1; i <= len; i++)
	{
		(*storeout) << GetSol(i) << "\n";
	}

	storeout->close();
	delete storeout;

	return 1;

#else //$ PG 2017-5-16: using common functionality TimeInt::StoreSolutionStateToFile(CMFile& file) const
	
	int binary_mode = 0;
	CMFile file(filename, TFileMode::TFMwrite, binary_mode);

	if (!file.IsGood())  
	{
		UO() << "ERROR: could not write solution to file " << filename.c_str() << "\n";
		return 0;
	}
	
	file.SetPrecision(19);
	int len = GetSolVector().Length();

	char str[256];
	sprintf(str, "HOTINTDataVersion%.3f", GetHotintVersion().GetDoubleValue());
	mystr mstr(str);
	file.RWSoleStr(mstr);
	
	StoreSolutionDataToFile(file);
	if (!omit_output_message)
		UO(UO_LVL_sim) << "Solution data stored to " << filename << "\n";
	
	return 1;

#endif
}

int MultiBodySystem::LoadInitialVector(const mystr& filename)  //rv=1 --> OK
{
	CMFile storein(filename, TFMread);

	if (!storein.IsGood())
	{
		UO(UO_LVL_err) << "ERROR: load initial vector: could not open file " << filename << "\n";
		return 0;
	}

	//read either version or length:
	mystr header;
	storein.RWSoleStr(header);

	if (header.Length() > 17 && header.Left(17) == mystr("HOTINTDataVersion"))
	{
		mystr version_str = header.Right(header.Length()-17);
		HotintVersionInfo data_version(atof(version_str));
		if (data_version > GetHotintVersion())
		{
			UO(UO_LVL_warn) << "Warning: Your current version of HOTINT: " << GetHotintVersion().GetString() << "\n         Solution data file was stored with version " << data_version.GetString() << ", data may be corrupt\n";
		}

		if (data_version < HotintVersionInfo(1,3,0)) //$ PG 2017-5-16: old redundant code, xdata is not loaded here!
		{
			Vector x0(GetSystemSize());
			stored_initialconditions = GetSolVector();

			int ndof, checksum2, nsteps;
			storein.RWInt(ndof);
			storein.RWInt(checksum2);
			storein.RWInt(nsteps);

			if (ndof != x0.Length()) {UO(UO_LVL_warn) << "Warning: Stored solution does have different size!!!\n";}
			if (nsteps != 1) {UO(UO_LVL_warn) << "Warning: Stored solution contains " << nsteps << " data units, only the first data unit is loaded!\n";}

			double time;
			int len;

			storein.RWDouble(time);
			storein.RWInt(len);

			if (len > x0.Length()) {UO(UO_LVL_err) << "ERROR: Stored solution is incompatible!\n";}
			else
			{
				for (int i=1; i <= len; i++) //additionally check for storein->good() ....???
				{
					storein.RWDouble(x0(i));
				}
				stored_initialconditions = x0;
			}
		} 
		else //$ PG 2017-5-16: using common functionality for reading solution state from files
		{	
			try
			{
				LoadSolutionDataFromFile(storein);
			}
			catch(mystr & err)
			{
				UO(UO_LVL_err) << "ERROR: loading initial solution data: " << err << "\n";
				assert(false);
			}
			stored_initialconditions = TIx0;
			UO(UO_LVL_sim) << "Solution data loaded from " << filename << "\n";
		}
		
		if (GetSolSet().continue_simulation_at_loaded_state)
		{			 
			if (!LoadSensorData())
				return 0;
		}
		else
			SetTime(0.);  // reset simulation time
	}
	else
	{
		UO(UO_LVL_err) << "ERROR: loading initial solution data: " << filename << " is not a valid solution data file\n";
		return 0;
	}


	return 1;
}

int MultiBodySystem::LoadSensorData()  //rv=1 --> OK
{
	int rv = 0;

	CloseFiles();

	//read sensor data from sol file
	mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12
	mystr soln = GetSolSet().sol_filename; //GetTOption81);	// solution file
	mystr fullpath = dir + soln;

	UO(UO_LVL_sim) << "Loading sensor data from " << fullpath << ": ";

	ifstream solin;
	solin.open((dir + soln).c_str(), ios::in); //with read sharing allowed!

	if (solin.good())
	{
		// skip file header
		while (solin.peek() == '%')
			solin.ignore(std::numeric_limits<std::streamsize>::max(), '\n');

		// read time and sensor values		
		double time, val;
		TArray<double> timearray;
		TArrayDynamic<TArray<double> > valuearray(sensors.Length());
		while (!(solin>>time).eof())
		{
			timearray.Add(time);
			for (int i = 1; i<=sensors.Length(); i++)
			{
				solin >> val;
				valuearray(i).Add(val);
			}
		}
		solin.close();

#if 0
		// debug output
		for (int i = 1; i<=timearray.Length(); i++)
		{
			UO() << timearray(i) << " ";
			for (int j = 1; j<=sensors.Length(); j++)
				UO() << valuearray(j)(i) << " ";
			UO() << "\n";
		}
#endif

		for (int j = 1; j<=sensors.Length(); j++)
			for (int i = 1; i<=timearray.Length(); i++)
			{
				(*sensors(j)->GetSignalHistoryTimesPtr()).CopyFrom(timearray);
				(*sensors(j)->GetSignalHistoryValuesPtr()).CopyFrom(valuearray(j));
			}
		rv = 1;
		UO(UO_LVL_sim) << "done.\n";
	}
	else
		UO(UO_LVL_err) << "ERROR at loading " << fullpath << "\nFile does not exist, please check path or do not choose SolverOptions.Solution.continue_simulation_at_loaded_state = 1\n";

	OpenFiles(1);

	return rv;
}

bool MultiBodySystem::DoLoadAnimationDataAtStart() const 
{
	return do_load_animation_data_at_start; 
}

//type=1: element
//type=2: force	i				//$ DR 2012-10 old:force value of element i
//type=3: sensor i
//type=4: geomelement i
//type=5: node i
//type=6: material i
//type=50: KB options (old)
void MultiBodySystem::GetElementData(int i, int type, int value, ElementDataContainer& edc) 
{
	switch (type)
	{
	case 1: //element data
		{
			GetElement(i).GetElementData(edc);
			break;
		}
	case 2: //force value of element i, force value
		{
			//GetElement(i).GetLoad(value).GetElementData(edc); //$ DR 2012-10 old code
			GetLoad(i).GetElementData(edc);
			break;
		}
	case 3: //sensor i data
		{
			GetSensor(i).GetElementData(edc);
			break;
		}
	case 4: //geomelement i data
		{
			GetDrawElement(i)->GetElementData(edc);
			break;
		}
	case 5: //Node i data
		{
			GetNode(i).GetElementData(edc);
			break;
		}
	case 6: //Material i data
		{
			GetMaterial(i).GetElementData(edc);
			break;
		}
	case 7: //Set i data
		{
			GetSetPtr(i)->GetElementData(edc);
			//GetSet(i).GetElementData(edc);
			break;
		}
	case 8: // Mesh i data
		{
			GetMeshPtr(i)->GetElementData(edc);
			break;
		}
	default: ;
	}
}

int MultiBodySystem::SetElementData(int i, int type, int value, ElementDataContainer& edc) 
{
	int rv = 1;
	switch (type)
	{
	case 1: //element data
		{
			rv = GetElement(i).SetElementData(edc);
			break;
		}
	case 2: //force value of element i, force value
		{
			//rv = GetElement(i).GetLoad(value).SetElementData(edc); //$ DR 2012-10 old code
			rv = GetLoad(i).SetElementData(edc); 
			break;
		}
	case 3: //sensor i data
		{
			rv = GetSensor(i).SetElementData(edc);
			break;
		}
	case 4: //geomelement i data
		{
			rv = GetDrawElement(i)->SetElementData(edc);
			break;
		}
	case 5: //Node i data
		{
			rv = GetNode(i).SetElementData(edc);
			break;
		}
	case 6: //Node i data
		{
			rv = GetMaterial(i).SetElementData((ElementDataContainer &)edc);
			break;
		}
	case 7: //Set i data
		{
			rv = GetSetPtr(i)->SetElementData((ElementDataContainer &)edc); // wrtire data from EDC (dialog) to the set variable
			GetSetPtr(i)->Convert();                                        // call the Convert function to generate the entries for the read only fields
			break;
		}
	case 8: //Mesh i data
		{
			rv = GetMeshPtr(i)->SetElementData((ElementDataContainer &)edc); 
			break;
		}
		//#ifdef KB_INCLUDES
		//	case 50: //edit KB menu
		//		{
		//			kbmenu.SetElementData(edc);
		//			break;
		//		}
		//#endif
	default: ;
	}
	return rv;
}

//$AD 2013-07-08: Manipulate Content of arrays in IOElements from 2D Draw Window
void MultiBodySystem::InsertIOElemConNode(int elemnr, int list_idx, int input_nr, Vector2D pos)
{
	GetElement(elemnr).InsertConNode(list_idx, input_nr, pos);
}
void MultiBodySystem::DeleteIOElemConNode(int elemnr, int list_idx)
{
	GetElement(elemnr).DeleteConNode(list_idx);
}
//$AD 2013-07-10: Change Position of a single element (MBS element, conNode, ...) 
void MultiBodySystem::MoveConNode2D(int elemnr, int list_idx, double delta_x, double delta_y) 
{
	GetElement(elemnr).MoveConNode2D(list_idx, delta_x, delta_y);
}
void MultiBodySystem::MoveElement(int elemnr, double delta_x, double delta_y, double delta_z)
{
	GetElement(elemnr).MoveElement(delta_x, delta_y, delta_z);
}

//action=1: option=0: Set initial conditions, option=1: Initialize
//action=2: Add Element: deprecated
//action=3: Assemble system
//action=4: Delete Element value
//action=5: Delete force value
//action=6: Add force: deprecated
//action=7: Add sensor:: deprecated
//action=8: Delete sensor value
//action=9: Add GeomElement: deprecated
//action=10:Delete GeomElement value
//action=11: Add Node: deprecated
//action=12: Delete Node value
//action=13: Swap element: deprecated
//action=14: Add Material: deprecated
//action=15: Delete Material value


//action=20: ModelFunction: option=1: get ModelFunctionList
//
//action=30: Call Element functions in response to action in IOBlockView Window... 
//action=40: specify source file name for an FMU/FMI InputOutput Block and call import Funciton

//action=101: save MBS system to file (file name in edc.Find("File_name")+directory, .GetText() (option=1: with Options, =0: only MBS)
//action=102: load MBS system from file (file name in edc.Find("File_name")+directory, .GetText()
//action=103: save actual solution vector (file name in edc.Find("File_name")+directory, .GetText()
//action=104: load initial solution vector (file name in edc.Find("File_name")+directory, .GetText()
//action=105: clear initial solution vector (setlen = 0 means initial solution is not used)
//action=106: LoadMBSEDCConfiguration
//action=107: SaveMBSEDCConfiguration; ; all:value==0, solver:value==1, hotint:value==2; 
//
//action=110: Load EDC from file (file name in edc.Find("File_name")+directory, .GetText()
//action=111: Save EDC to file (file name in edc.Find("File_name")+directory, .GetText()
//action=112: The input-edc gets written into a mystr, the mystr is returned in the same edc
//
//action=121: plot single sensor value in matlab
//action=122: plot two sensors (x-y-plot) in matlab
//action=123: plot n sensors (vs time) in matlab
//action=150:	set start_time=0
//
//action=201: GetNSensors()
//action=202: Get Element Names in edc-data
//action=203: Get Sensor Names in edc-data
//            option 1: return writeresult flag as int element data
//action=204: Get Sensor value of sensor number 'value' in edc-data
//action=205: Get GeomElement Names in edc-data
//action=206: Get Number of nodes
//action=207: Get Name of nodes
//action=208: Get Number of total DOF
//action=209: Get Material Names in edc-data
//action=210: Get/Set EditAllOptions in edc-data Get:option==0, Set:option==1;    all:value==0, solver:value==1, hotint:value==2; 
//action=211: Get/Set Model ElementDataContainer
//action=212: SetOptions: Options2EDC/EDC2Options, etc.
//action=213: Get Sensors data array(reference), DataArray:option=0, TimesArray:option=1
//action=214: Get Load Names in edc-data	//$ DR 2012-10
//action=215: Get Set Names in edc-data
//            option 1: returns settype as int element data
//action=216: Get Mesh Names in edc-data
//action=217: Get EDC for given Set #val
//            option 1: returns only 
//
//action=301: Show System properties
//action=302: Check System properties

//action=320: Compute Eigenmodes (option=1: sparse, =0: direct)

int MultiBodySystem::CallCompFunction(int action, int option, int value, ElementDataContainer* edc)
{
	int rv = 1;
	switch(action)
	{
	case 1: //Set initial conditions
		{
			if (option == 0)
				SetInitialConditions();
			else if (option == 1)
				Initialize(); //this resets the whole MBS model

			break;
		}
	case 2: //Add Element of type "value"
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;
		}
	case 3: //Assemble system
		{
			Assemble();
			break;
		}
	case 4: //Delete element "value"
		{
			if(value<=NE())
			{
				DeleteElement(value);
			}
			else
			{
				UO() << "Error: element " << value << " does not exist!!!\n";
			}
			break;
		}
	case 5: //Delete force "value" 
		{
			if(value<=NLoads())
			{
				DeleteLoad(value);
			}
			else
			{
				UO() << "Error: load " << value << " does not exist!!!\n";
			}
			break;
		}
	case 6: //Add Force type "value" 
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;
		}
	case 7: //Add Sensor type "value"
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;		}
	case 8: //Delete sensor "value"
		{
			if (value >= 1 && value <= this->NSensors())
			{
				DeleteSensor(value);
			}
			else
			{
				UO() << "Error: sensor " << value << " does not exist!!!\n";
			}
			break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 9: //Add GeomElement SensorTypeNameList above!!!
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;		}
	case 10: //action: Delete GeomElement "value"
		{
			if (value >= 1 && value <= NDrawElements())
			{
				DeleteDrawElement(value);
			}
			else
			{
				UO() << "Error: geom element " << value << " does not exist!!!\n";
			}
			break;
		}
	case 11: //action: Add Node
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;
		}
	case 12: //action: Delete Node "value"
		{
			if (value >= 1 && value <= NNodes())
			{
				DeleteNode(value); //delte this node and reorder node references in elements!
			}
			else
			{
				UO() << "Error: Node " << value << " does not exist!!!\n";
			}
			break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 13: //action: Swap element "value" with element "option"
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;		}

		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 14: //action: Add Material
		{
			UO(UO_LVL_warn)<<"Warning: CallCompFunction: a deprecated function is called!";
			break;		}
	case 15: //action: Delete Material "value"
		{
			if (value >= 1 && value <= NMaterials())
			{
				DeleteMaterial(value); //delete this material and reorder material numbers in elements!
			}
			else
			{
				UO() << "Error: Material " << value << " does not exist!!!\n";
			}
			break;
		}
	case 16: // action : Delete Set "value"
		{
			if(value >= 1 && value <= NSets()) 
			{
				DeleteSet(value); //delete this set
			}
			else
			{
				UO() << "Error: Set " << value << " does not exist!!!\n";
			}
			break;
		}
	case 17: // action : Delete Mesh "value"
		{
			if(value >= 1 && value <= NMeshes()) 
			{
				DeleteMesh(value); //delete this set
			}
			else
			{
				UO() << "Error: Mesh " << value << " does not exist!!!\n";
			}
			break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		//action=20: ModelFunction: option=1: get ModelFunctionList
	case 20: //action: Get ModelFunctionList
		{
			if (option == 1) //get ModelFunctionList
			{
				ElementData ed;
				//$ YV 2013-01-04: model function index is now 1-based
				for (int i=1; i <= GetModelsLibrary()->GetModelsCount(); i++)
				{
					ed.SetText(GetModelsLibrary()->GetModelInterface(i)->GetMBSModelName());
					ed.SetToolTipText(GetModelsLibrary()->GetModelInterface(i)->GetMBSModelDescription());
					edc->Add(ed);
				}
				//GetIOption(12) = GetModelFunctionsIndex0();
			}
			else if (option == 2) //ModelFunctionChanged
			{
				//$ AD: do NOT reset ModelDataContainer when e.g. Model is loaded via Script
				if (GetModelDataContainer()&&value==0) {GetModelDataContainer()->Reset();}		// DR hack
				ModelFunctionChanged();
			}
			else if (option == 3) //Initialize
			{
				Initialize();
			}
			else if (option == 4) //get modelfunctionindex
			{
				ElementData ed;
				ed.SetInt(GetModelFunctionsIndex0(),"model_function_index");
				edc->Add(ed);
			}
			break;
		}
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 30: //action: Call Element functions in response to action in IOBlockView Window... 
		{
			int elnr = option;
			Element& elem = GetElement(elnr);

			if (value == 0) 
			{ 
				if (GetElement(elnr).GetElementSpecification().Compare(mystr("IOMouseResponseElement")) )
					rv = 1; 
				else 
					rv = 0; 
			}

			// allow modifications 1 -> add 1, and -1 -> subtract 1
			if (value == 1)	 
			{ 
				elem.Increase(); 
			}
			if (value == -1) 
			{ 
				elem.Decrease(); 
			}
			break;
		}
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 31: // action: call MBS-wide response to a Keyboard input
		{
			int key = option;
			int nr_of_changes = RespondToKey(key);
			rv = nr_of_changes;
			break;
		}
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 40: // action: specify source file name for an FMU/FMI InputOutput Block (#value) and call import Funciton
		{
			mystr sourcefilename;
			GetElemDataText(this, *edc, "SourceFile",sourcefilename);

			GetElement(value).ImportFile(sourcefilename);

			rv = 1;
			break;
		}
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		//load/save etc.
	case 101: //save to file
		{
			mystr filename, dirname;
			GetElemDataText(this, *edc, "File_name", filename);
			GetElemDataText(this, *edc, "Directory_name", dirname); //include backslash --> dirname+filename==filepath

			isloadsavemode = 1;
			SaveToFile(dirname+filename,option);
			isloadsavemode = 0;
			break;
		}
	case 102: //load from file
		{
			mystr filename, dirname;
			GetElemDataText(this, *edc, "File_name", filename);
			GetElemDataText(this, *edc, "Directory_name", dirname); //include backslash --> dirname+filename==filepath

			isloadsavemode = 1;
			rv = LoadFromFile(dirname+filename);
			isloadsavemode = 0;

			break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 103: //save actual solution vector
		{
			mystr filename, dirname;
			GetElemDataText(this, *edc, "File_name", filename);
			GetElemDataText(this, *edc, "Directory_name", dirname); //include backslash --> dirname+filename==filepath

			SaveSolutionVector(dirname+filename);
			break;
		}
	case 104: //load initial solution vector
		{
			mystr filename, dirname;
			GetElemDataText(this, *edc, "File_name", filename);
			GetElemDataText(this, *edc, "Directory_name", dirname); //include backslash --> dirname+filename==filepath

			if (LoadInitialVector(dirname+filename))
				if (GetSolSet().continue_simulation_at_loaded_state)
					do_load_animation_data_at_start = true;
			break;
		}
	case 105: //clear initial solution vector (setlen = 0 means initial solution is not used)
		{
			//UO() << "Stored initial vector cleared!\n";
			stored_initialconditions.SetLen(0);
			break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 106: //Load_MBS_EDC_config_file
		{
//			mystr filename = "hotint_cfg.txt";
//			CMFile file(filename, TFMread);
//
//			//set default options, if file can not be loaded!
//			SetSolverDialogOptions();
//
//			SetOptions2EDCOptions();//copy numbered i/d/t options to EDC Tree to 
//
//			OpenLogFile(0);
//
//			ElementDataContainer fileedc;
//
//			if (file.IsGood()) 
//			{
//				UO() << "load configuration file 'hotint_cfg.txt'\n";
//
//				mystr str;
//				file.RWF(str); //read file
//
//				// AD 2013-10-22 use new routine to parse
//				EDCParser().ParseAndExecuteString(str,fileedc);
////				EDCParser().String2TreeEDC(/*this, */str, fileedc);
//
//				//$ DR 2013-03-18 for release version, do not use stored model name
//				if(!fileedc.TreeGetBool("GeneralOptions.Application.reload_last_model",0))
//				{
//					fileedc.TreeDelete("GeneralOptions.ModelFile.hotint_input_data_filename");	
//					fileedc.TreeDelete("GeneralOptions.ModelFile.internal_model_function_name");	
//					mystr tmp("");
//					fileedc.TreeSetString("GeneralOptions.ModelFile.hotint_input_data_filename", tmp.c_str());
//					tmp = mystr("new model");
//					fileedc.TreeSetString("GeneralOptions.ModelFile.internal_model_function_name", tmp.c_str());
//				}
//
//				//replace default option parameters with loaded parameters
//				GetMBS_EDC_Options()->TreeReplaceEDCDataWith(&fileedc);
//			}
//			else
//			{
//				UO() << "Warning: Could not read hotint configuration file 'hotint_cfg.txt'. Default values are used instead. Press 'Save Hotint Options' to create this file.\n";
//				rv = 0; 
//			}
//
//			//at this point, use arguments of program start to add some functionality:
//			mystr modeldata_argstr = "";
//			mystr mbsedc_argstr = "";
//			int narg = __argc;
//
//			for (int i=0; i < narg; i++)
//			{
//				mystr argstr = __argv[i];
//				UO(UO_LVL_ext) << "hotint.exe argument " << i << "='" << argstr << "'";
//
//				//$ RL 2012-7-25:[ 
//				if(i == 0) // i==0 --> store application path
//				{
//					// get application path, (path of hotint.exe)
//					mystr app_path("");
//					{
//						int until;
//						for(until=argstr.Length()-1;until>=0;until--)
//						{
//							if(argstr.PosPeek(until) == '\\')
//								break; // until <=> last backslash
//						}
//						for(int i=0;i<=until;i++)
//						{
//							app_path += argstr.PosPeek(i); // path until last backslash
//						}
//						argstr = "GeneralOptions.Paths.application_path=\"" + app_path+ "\" ";
//					}
//				}
//				else if(i == 1) // i==1 --> from drag&drop or command line
//				{
//					mystr hid_file("");
//					int pos = argstr.Find(mystr(".hid"));
//					if(pos == -1)
//					{
//						pos = argstr.Find(mystr(".hmc"));		//$ DR 2012-12-10 this ending is also supported
//					}
//					//if(pos == -1 && GetMBS_EDC_Options()->TreeGetBool("GeneralOptions.ModelFile.accept_txt_file_as_model_file",0))
//					if(pos == -1 && GetOptions()->GeneralOptions()->ModelFileAcceptTxtFileAsModelFile())	//$ DR 2014-10-23 improve speed
//					{
//						pos = argstr.Find(mystr(".txt"));
//					}
//					if (pos != -1)
//					{
//						if(DoesFileExist(argstr) && argstr.Find(mystr("=")) == -1)
//						{
//							hid_file = argstr;
//							argstr = "GeneralOptions.ModelFile.hotint_input_data_filename=\"" + argstr+ "\" ";
//							// get path of Hotint Input Data file
//							mystr path("");
//							{
//								int until;
//								for(until=hid_file.Length()-1;until>=0;until--)
//								{
//									if(hid_file.PosPeek(until) == '\\')
//										break; // until <=> last backslash
//								}
//								for(int i=0;i<=until;i++)
//								{
//									path += hid_file.PosPeek(i); // path until last backslash
//								}
//								//GetMBS_EDC_Options()->TreeSetString("GeneralOptions.Paths.hotint_input_data_path", path.c_str());
//								GetOptions()->GeneralOptions()->PathsHotintInputDataPath() = path; //$ DR 2014-10-24 improve speed
//							}
//						}
//					}
//				}	
//				//$ RL 2012-7-25:] 
//
//				int pos = argstr.Find('=');
//				if (pos != -1)
//				{	
//					//$ PG & AD 2014-7-10:[ check, if '=' occurrs a second time before linebreak (char(10))
//					//$ PG & AD 2014-7-10:  if so, print warning
//					mystr remainder = argstr.Right(argstr.Length()-pos-1);
//					
//					int pos_lbr = remainder.Find(10);
//					int pos_equ = remainder.Find('=');
//					if (pos_equ > -1 && (pos_equ < pos_lbr  || pos_lbr == -1) )
//					{
//						UO(UO_LVL_warn) << "\n WARNING: some command line arguments are not separated by space\n";
//					}
//					//$ PG & AD 2014-7-10:]
//
//					mystr name = argstr.Left(pos);
//					//UO(UO_LVL_ext) << "name = '" << name << "'\n";
//					name.EraseSpaces(); //name anyway should not contain spaces ...
//
//					if (GetMBS_EDC_Options()->TreeFind(name))
//					{
//						UO(UO_LVL_ext) << " ... parsed to MBS_EDC_Option";
//						mbsedc_argstr += argstr + mystr("\n");
//					}
//					else
//					{
//						UO(UO_LVL_ext) << " ... parsed to ModelData";
//						modeldata_argstr += argstr + mystr("\n");
//					}
//				}
//				UO() << "\n";
//			}	// end of for loop narg
//
//			if (mbsedc_argstr.Length() != 0)
//			{
//				UO() << "mbs-edc-argstring=" << mbsedc_argstr << "\n";
//				// AD 2013-10-22 use new routine to parse
//				EDCParser().ParseAndExecuteString(mbsedc_argstr,mbs_edc_args);
////				EDCParser().String2TreeEDC(/*this, */mbsedc_argstr, mbs_edc_args);
//				GetMBS_EDC_Options()->TreeReplaceEDCDataWith(&mbs_edc_args);
//			}
//			if (modeldata_argstr.Length() != 0) //write information in to modeldata container ==> will not be overwritten later
//			{
//				// AD 2013-10-22 use new routine to parse
//				EDCParser().ParseAndExecuteString(modeldata_argstr,modeldata_edc_args);
////				EDCParser().String2TreeEDC(/*this, */modeldata_argstr, modeldata_edc_args);
//
//				if (GetModelDataContainer())
//				{
//					GetModelDataContainer()->TreeReplaceEDCDataWith(&modeldata_edc_args);
//				}
//				else
//				{
//					SetModelDataContainer(modeldata_edc_args);
//				}
//			}
//
//			//edc->CopyFrom(*GetMBS_EDC_Options()); //some options are written in configuration //$ DR 2014-10-24 no configuration anymore --> not necessary
//
//			SetEDCOptions2Options(); //copy EDC Tree to numbered i/d/t options for fast access
//			SetComputationSolverOptions(); //copy EDC Tree to MBSSolSet() 


			//$ AD 2016-03-31 process special command line options ( "-nolog", ... ) at this early stage
			//$ "-nolog" ... no log file is created, no output to existing logfile
            //$ "-silent" ... no log file, no console output
			int narg = __argc;
			int possible_hid_file_pos = 1;
			for (int i=0; i<narg; i++)
			{
				mystr argstr = __argv[i];	
				if(argstr[0] == '-')
				{
					if(argstr.Compare("-nolog") || argstr.Compare("-silent")) // do not create a log file
					{
						flag_nolog = 1;
						if (i==possible_hid_file_pos)
						{
							possible_hid_file_pos++;
						}
					}
				}
			}


			//$ DR 2014-10-24 new implementation due to simplification of options and solver options handling
			mystr filename = "hotint_cfg.txt";
			CMFile file(filename, TFMread);

			//set default options, if file can not be loaded!
			ElementDataContainer edc_options, edc_tmp;
			Options2EDC(&edc_options);
			SolverOptions2EDC(&edc_tmp);
			edc_tmp.TreeReplaceEDCDataWith(&edc_options);		// in edc_tmp now options and solveroptions are stored

			OpenLogFile(0);                                     // NOTE AD: first open / creation of the log file - any "-nolog" option must be checked before this lin

			ElementDataContainer fileedc;

			if (file.IsGood()) 
			{
				UO() << "load configuration file 'hotint_cfg.txt'\n";

				mystr str;
				file.RWF(str); //read file

				EDCParser().ParseAndExecuteString(str,fileedc);

				//$ DR 2013-03-18 for release version, do not use stored model name
				if(!fileedc.TreeGetBool("GeneralOptions.Application.reload_last_model",0))
				{
					fileedc.TreeDelete("GeneralOptions.ModelFile.hotint_input_data_filename");	
					fileedc.TreeDelete("GeneralOptions.ModelFile.internal_model_function_name");	
					mystr tmp("");
					fileedc.TreeSetString("GeneralOptions.ModelFile.hotint_input_data_filename", tmp.c_str());
					tmp = mystr("new model");
					fileedc.TreeSetString("GeneralOptions.ModelFile.internal_model_function_name", tmp.c_str());
				}

				//replace default option parameters with loaded parameters
				//edc_tmp.TreeReplaceEDCDataWith(&fileedc);	// replace options and solveroptions with fileedc
				edc_tmp.TreeReplaceEDCDataWithReplacingOldNames(&fileedc, "HotintOptions", &CEDCParser::GetNewEDCName, 0);
				EDC2Options(&edc_tmp);	// replace options 
				EDC2SolverOptions(&edc_tmp);	// replace solveroptions
				// now all options of config file are transfered to options and solveroptions
			}
			else
			{
				UO() << "Warning: Could not read hotint configuration file 'hotint_cfg.txt'. Default values are used instead. Press 'Save Hotint Options' to create this file.\n";
				rv = 0; 
			}

			//at this point, use arguments of program start to add some functionality:
			mystr modeldata_argstr = "";
			mystr mbsedc_argstr = "";
			/*int*/ narg = __argc;

			for (int i=0; i < narg; i++)
			{
				mystr argstr = __argv[i];
				UO(UO_LVL_ext) << "hotint.exe argument " << i << "='" << argstr << "'";

				//$ RL 2012-7-25:[ 
				if(i == 0) // i==0 --> store application path
				{
					// get application path, (path of hotint.exe)
					mystr app_path("");
					{
						int until;
						for(until=argstr.Length()-1;until>=0;until--)
						{
							if(argstr.PosPeek(until) == '\\')
								break; // until <=> last backslash
						}
						for(int i=0;i<=until;i++)
						{
							app_path += argstr.PosPeek(i); // path until last backslash
						}
						argstr = "GeneralOptions.Paths.application_path=\"" + app_path+ "\" ";
					}
				}
				else if(i == possible_hid_file_pos) // i==1 --> from drag&drop or command line
				{
					mystr hid_file("");
					int pos = argstr.Find(mystr(".hid"));
					if(pos == -1)
					{
						pos = argstr.Find(mystr(".hmc"));		//$ DR 2012-12-10 this ending is also supported
					}
					if(pos == -1 && GetOptions()->GeneralOptions()->ModelFileAcceptTxtFileAsModelFile())	//$ DR 2014-10-23 improve speed
					{
						pos = argstr.Find(mystr(".txt"));
					}
					if (pos != -1)
					{
						if(DoesFileExist(argstr) && argstr.Find(mystr("=")) == -1)
						{
							hid_file = argstr;
							argstr = "GeneralOptions.ModelFile.hotint_input_data_filename=\"" + argstr+ "\" ";
							// get path of Hotint Input Data file
							mystr path("");
							{
								int until;
								for(until=hid_file.Length()-1;until>=0;until--)
								{
									if(hid_file.PosPeek(until) == '\\')
										break; // until <=> last backslash
								}
								for(int i=0;i<=until;i++)
								{
									path += hid_file.PosPeek(i); // path until last backslash
								}
								GetOptions()->GeneralOptions()->PathsHotintInputDataPath() = path; //$ DR 2014-10-24 improve speed todo
							}
						}
					}
				}	
				//$ RL 2012-7-25:] 

				int pos = argstr.Find('=');
				if (pos != -1)
				{	
					//$ PG & AD 2014-7-10:[ check, if '=' occurrs a second time before linebreak (char(10))
					//$ PG & AD 2014-7-10:  if so, print warning
					mystr remainder = argstr.Right(argstr.Length()-pos-1);
					
					int pos_lbr = remainder.Find(10);
					int pos_equ = remainder.Find('=');
					if (pos_equ > -1 && (pos_equ < pos_lbr  || pos_lbr == -1) )
					{
						UO(UO_LVL_warn) << "\n WARNING: some command line arguments are not separated by space\n";
					}
					//$ PG & AD 2014-7-10:]

					mystr name = argstr.Left(pos);
					//UO(UO_LVL_ext) << "name = '" << name << "'\n";
					name.EraseSpaces(); //name anyway should not contain spaces ...

					if (edc_tmp.TreeFind(name))
					{
						UO(UO_LVL_ext) << " ... parsed to MBS_EDC_Option";
						mbsedc_argstr += argstr + mystr("\n");
					}
					else
					{
						UO(UO_LVL_ext) << " ... parsed to ModelData";
						modeldata_argstr += argstr + mystr("\n");
					}
				}
				UO() << "\n";
			}	// end of for loop narg

			// before start replacing with the command line args, reset the edc
			edc_tmp.Reset();
			edc_options.Reset();
			Options2EDC(&edc_options);
			SolverOptions2EDC(&edc_tmp);
			edc_tmp.TreeReplaceEDCDataWith(&edc_options);		// in edc_tmp now options and solveroptions are stored

			if (mbsedc_argstr.Length() != 0)
			{
				UO() << "mbs-edc-argstring=" << mbsedc_argstr << "\n";
				EDCParser().ParseAndExecuteString(mbsedc_argstr,mbs_edc_args);
				//edc_tmp.TreeReplaceEDCDataWith(&mbs_edc_args);	// replace options and solveroptions with mbs_edc_args
				edc_tmp.TreeReplaceEDCDataWithReplacingOldNames(&mbs_edc_args, "HotintOptions", &CEDCParser::GetNewEDCName, 0);
			}
			if (modeldata_argstr.Length() != 0) //write information in to modeldata container ==> will not be overwritten later
			{
				EDCParser().ParseAndExecuteString(modeldata_argstr,modeldata_edc_args);

				if (GetModelDataContainer())
				{
					GetModelDataContainer()->TreeReplaceEDCDataWith(&modeldata_edc_args);
				}
				else
				{
					SetModelDataContainer(modeldata_edc_args);
				}
			}

			EDC2Options(&edc_tmp);	// replace options 
			EDC2SolverOptions(&edc_tmp);	// replace solveroptions

			break;
		}
	case 107: //Save_MBS_EDC_config_file
		{
//!AD 2012-06-26: revised
//!AD 2012-04-17: solver options and hotint options can be stored separately - use autogenerated functions, changed names to better distinguish various EDCs
			mystr filename = "hotint_cfg.txt";
			CMFile file_read(filename, TFMread);
	
// read from current configuration file;
			ElementDataContainer edc_configfile;
			if (file_read.IsGood()) 
			{
			  mystr str;
				file_read.RWF(str);
				// AD 2013-10-22 use new routine to parse
				EDCParser().ParseAndExecuteString(str, edc_configfile);
//				EDCParser().String2TreeEDC(/*this, */str, edc_configfile);
			}
			else
			{
				UO(UO_LVL_warn) << "WARNING: could not read hotint configuration file 'hotint_cfg.txt'\n";
			}
		//	delete &file_read;


////// actualize solver option part of the options edc //AD: revision
////			SetComputationSolverOptions(); // <-- wrong
////			SetOptions2EDCOptions();
			//SetSolverDialogOptions(); //$ DR 2014-10-24 not necessary anymore
			//SetOptions2EDCOptions();//$ DR 2014-10-24 not necessary anymore

// pick the data that will be written to file
			ElementDataContainer edc_tosave;
			if(value == 1) // save solver options only
			{
				SolverOptions2EDC(&edc_tosave);
				UO() << "save configuration file 'hotint_cfg.txt': solver options only\n";
			}
			else if(value == 2) // save hotint options only
			{
			  Options2EDC(&edc_tosave);
			  edc_tosave.TreeReplaceEDCDataWith(edc);	//$ DR 2015-01-23 bugfix. otherwise info of edc (e.g configuration) would be overwritten

				UO() << "save configuration file 'hotint_cfg.txt': hotint options only\n";
			}
			else // rest of old code...
			{
				//edc_tosave = *GetMBS_EDC_Options();
				//$ DR 2014-10-24 new implementation
				ElementDataContainer edc_tmp;
				Options2EDC(&edc_tosave);	// get hotint options

				SolverOptions2EDC(&edc_tmp);
				edc_tosave.TreeReplaceEDCDataWith(&edc_tmp); // add solver options
				edc_tosave.TreeReplaceEDCDataWith(edc);	//$ DR 2015-01-23 bugfix. otherwise info of edc (e.g configuration) would be overwritten

				UO() << "save configuration file 'hotint_cfg.txt': all options\n";
			}
// replace in the content of the config file
			edc_configfile.TreeReplaceEDCDataWith(&edc_tosave);

			//$ AD 2011-09: do not write ComputationSteps to the config file
			RemoveVariableFromEDCTree(edc_configfile, mystr("SolverOptions.ComputationSteps"), 0);
			
// edc contains window placement etc...
			if( value != 1 )
			{
				edc_configfile.TreeReplaceEDCDataWith(edc);
			}

// write to configfile
			CMFile file_write(filename, TFMwrite);

			//UO() << "load arrow=" << edcnew.TreeGetDouble("PostProcOptions.Loads.arrow_size") << "\n";

			//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
			//set some default values, which should not be loaded:
			edc_configfile.TreeSetDouble("SolverOptions.start_time", 0.); //do not load any starttime except 0.
			edc_configfile.TreeSetBool("ViewingOptions.Animation.animate_deformation", 0); //animate at startup
			edc_configfile.TreeSetBool("ViewingOptions.Animation.RecordSingleFrames.record", 0); //do not record at beginning
			//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


			mystr str;
			EDCParser().EDC2String(edc_configfile, str);

			if(UO().GetGlobalMessageLevel()==UO_LVL_dbg2)	//$ DR 2011-09-15
			{
				UO(UO_LVL_dbg2) << str << "\n";
			}
			file_write.RW(mystr("%HOTINT configuration file 'hotint_cfg.txt'\n"));
			file_write.RW(mystr("%handle this file with care, it can crash the startup of HOTINT\n"));
			file_write.RW(mystr("%in case that nothing works anymore ==> delete this file ==> defaults will be created ==> save configuration ==> new 'hotint_cfg.txt' will be created\n"));
			file_write.RW(mystr("%\n"));
			file_write.RW(str);

			break;
		}

		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		//Load and Save ANY EDC
//action=110: Load EDC from file (file name in edc.Find("File_name")+directory, .GetText()
	case 110:
		{
			mystr filename = edc->TreeGetString("File_name");
			mystr the_string;
			edc->Reset();

			File2Str(filename, the_string);
			EDCParser().ParseAndExecuteString(the_string, *edc);
			break;

			//the_file.RWF(the_string);
			//ElementDataContainer the_edc,testedc;
			////EDCParser().String2EDC(the_string, the_edc, testedc);

			//EDCParser().String2TreeEDC(the_string, the_edc);
			//edc->CopyFrom(the_edc);
			//break;
		}
//action=111: Save EDC to file (file name in edc.Find("File_name")+directory, .GetText()
	case 111:
		{
			mystr filename = edc->TreeGetString("File_name");
			edc->Delete(edc->Find("File_name"));
			CMFile the_file(filename, TFMwrite);
			mystr the_string;

			EDCParser().EDC2String(*edc, the_string);
			the_file.RW(the_string);
	    break;
		}
//action=112: The input-edc gets written into a mystr, the mystr is returned in the same edc
	case 112:
		{
			mystr clipboardtext;
			EDCParser().EDC2String(*edc, clipboardtext, mystr("  "));

			//edc = new ElementDataContainer;
			ElementData ed;
			ed.SetText(clipboardtext, "ClipBoardText");
			edc->Add(ed);
			break;
		}
	case 113: //$ DR 2013-06-12
		{
			mystr str = edc->TreeGetString("File_name");
			int fromfile = edc->TreeGetInt("is_file");
			AddModelData(str/*,fromfile*/);
			break;
		}

	case 114: //$ AD 2019-06-12
	{
		mystr str = edc->TreeGetString("File_name");
		WriteSpecial(value/*stepNr*/, str/*File_name*/);
		break;
	}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		//$ DR 2015-02-12:[ not in use anymore
	//	//plot in matlab
	//case 121: //plot sensor 'value' in matlab
	//	{
	//		if (GetSensor(value).GetSignalStorageMode() & SSM_OwnFile == 0) return 0;

	//		UO().CallWCDriverFunction(4); //Set Programm directory as current

	//		//mystr dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path"); //GetTOption80);
	//		//mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath(); //GetTOption80);	//$ DR 2014-10-23 improve speed
	//		mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12
	//		//mystr filename = mystr("S")+mystr(GetSensor(value).GetSensorNumber())+mystr("-")+GetSensor(value).GetTypeName()+mystr(".txt");
	//		mystr filename = mystr("S")+mystr(value)+mystr("-")+GetSensor(value).GetSensorName()+mystr(".txt");
	//		mystr dir2 = dir.Left(dir.Length()-1); //without the '/' sign

	//		ofstream plotfunc((dir+mystr("MBSplotfunc.m")).c_str());

	//		char buffer[_MAX_PATH*4];

	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr oldpath = buffer;

	//		_chdir(dir.c_str());
	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr path = buffer;

	//		char str[1024]; 
	//		//if (path[path.Length()-1] != '\\') path += '\\';

	//		sprintf(str, "f1=load('%s');\n",(filename).c_str());
	//		plotfunc << str;
	//		plotfunc << "h=plot(f1(:,1),f1(:,2),'k');\n";

	//		plotfunc << "title(' Y: " << GetSensor(value).GetSensorName() << "','Fontsize',16);\n"; 
	//		plotfunc << "xlabel(' Time [s] ','Fontsize',16);\n";
	//		//$ YV 2012-06: there is no type name for the sensors
	//		//plotfunc << "ylabel('" << GetSensor(value).GetTypeName() << "','Fontsize',16);\n";
	//		plotfunc << "set(h,'LineWidth',1.0);\n";
	//		//plotfunc << "legend('" << GetSensor(value).GetTypeName() << "');\n";
	//		plotfunc << "waitfor(h);\n"; //wait till window is closed
	//		plotfunc << "exit;\n";       //exit instance of matlab; too many instances will need enormous memory and may crash system 
	//		plotfunc.close();
	//		ShellExecute(NULL, "open", "matlab",	
	//			" -nosplash /r MBSplotfunc",
	//			path.c_str(), SW_HIDE);
	//		//path.c_str(), SW_SHOW);

	//		_chdir(oldpath.c_str());

	//		break;
	//	}
	//case 122: //plot two sensors 'edc.xsensor' vs 'edc.ysensor'  in matlab
	//	{
	//		int xsensor = edc->TreeGetInt("xsensor");
	//		int ysensor = edc->TreeGetInt("ysensor");

	//		if (GetSensor(xsensor).GetSignalStorageMode() & SSM_OwnFile == 0) return 0;
	//		if (GetSensor(ysensor).GetSignalStorageMode() & SSM_OwnFile == 0) return 0;

	//		UO().CallWCDriverFunction(4); //Set Programm directory as current

	//		//mystr dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path"); //GetTOption80);
	//		//mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath(); //GetTOption80);	//$ DR 2014-10-23 improve speed
	//		mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12
	//		mystr dir2 = dir.Left(dir.Length()-1); //without the '/' sign
	//		mystr filename = mystr("S")+mystr(xsensor)+mystr("-")+GetSensor(xsensor).GetSensorName()+mystr(".txt");
	//		mystr filename2 = mystr("S")+mystr(ysensor)+mystr("-")+GetSensor(ysensor).GetSensorName()+mystr(".txt");

	//		ofstream plotfunc((dir+mystr("MBSplotfunc.m")).c_str());

	//		char buffer[_MAX_PATH*4];

	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr oldpath = buffer;

	//		_chdir(dir.c_str());
	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr path = buffer;

	//		char str[1024]; 

	//		sprintf(str, "f1=load('%s');\n",(filename).c_str());
	//		plotfunc << str;
	//		sprintf(str, "f2=load('%s');\n",(filename2).c_str());
	//		plotfunc << str;
	//		plotfunc << "h=plot(f1(:,2),f2(:,2),'k');\n";
	//		plotfunc << "title({[' X: " << GetSensor(xsensor).GetSensorName() << " '];[' Y: " << GetSensor(ysensor).GetSensorName() << " ']},'Fontsize',16);\n"; 
	//		//$ YV 2012-06: there is no type name for the sensors
	//		//plotfunc << "xlabel('" << GetSensor(xsensor).GetTypeName() << "','Fontsize',16);\n";
	//		//plotfunc << "ylabel('" << GetSensor(ysensor).GetTypeName() << "','Fontsize',16);\n";
	//		plotfunc << "xlabel('" << GetSensor(xsensor).GetSensorName() << "','Fontsize',16);\n";
	//		plotfunc << "ylabel('" << GetSensor(ysensor).GetSensorName() << "','Fontsize',16);\n";
	//		plotfunc << "set(h,'LineWidth',1.0);\n";
	//		plotfunc << "waitfor(h);\n"; //wait till window is closed
	//		plotfunc << "exit;\n";       //exit instance of matlab; too many instances will need enormous memory and may crash system 
	//		plotfunc.close();
	//		ShellExecute(NULL, "open", "matlab",	
	//			" -nosplash /r MBSplotfunc",
	//			path.c_str(), SW_HIDE);
	//		//path.c_str(), SW_SHOW);

	//		_chdir(oldpath.c_str());

	//		break;

	//	}
	//case 123: //time plot of N sensors in matlab
	//	{
	//		if(option <=0) 
	//		{
	//			return 0; //no sensors selected
	//		}
	//		TArray<int> ysensor;
	//		ysensor.SetLen(option);

	//		for(int sensNr=1;sensNr<=option;sensNr++)
	//		{
	//			ysensor(sensNr) = edc->TreeGetInt("y"+mystr(sensNr)+"sensor");
	//			if (GetSensor(ysensor(sensNr)).GetSignalStorageMode() & SSM_OwnFile == 0) return 0;
	//		}
	//		UO().CallWCDriverFunction(4); //Set Programm directory as current

	//		//mystr dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path"); //GetTOption80);
	//		//mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath(); //GetTOption80);	//$ DR 2014-10-23 improve speed
	//		mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12
	//		mystr dir2 = dir.Left(dir.Length()-1); //without the '/' sign


	//		ofstream plotfunc((dir+mystr("MBSplotfunc.m")).c_str());

	//		char buffer[_MAX_PATH*4];

	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr oldpath = buffer;

	//		_chdir(dir.c_str());
	//		_getcwd( buffer, _MAX_PATH*4); //about 1000 characters ...
	//		mystr path = buffer;

	//		char str[1024]; 
	//		plotfunc << "scrsz = get(0,'ScreenSize');\n";
	//		plotfunc << "fig = figure;\n";
	//		plotfunc << "set(fig,'Position',[1 (scrsz(4)/2-60) scrsz(3)*2/3 scrsz(4)/2]);\n";
	//		for(int sensNr=1;sensNr<=option;sensNr++)
	//		{
	//			mystr filename = mystr("S")+mystr(ysensor(sensNr))+mystr("-")+GetSensor(ysensor(sensNr)).GetSensorName()+mystr(".txt");
	//			sprintf(str, mystr("f") + mystr(sensNr) + +mystr("=load('%s');\n"),(filename).c_str());				
	//			plotfunc << str;
	//			plotfunc << "h(" << sensNr << ")=plot(f" << sensNr << "(:,1),f" << sensNr << "(:,2),";
	//			if(sensNr==1){plotfunc << "'k'";}
	//			else if(sensNr==1){plotfunc << "'k'";}
	//			else if(sensNr==2){plotfunc << "'b'";}			
	//			else if(sensNr==3){plotfunc << "'g'";}		
	//			else if(sensNr==4){plotfunc << "'r'";}
	//			else if(sensNr==5){plotfunc << "'c'";}
	//			else if(sensNr==6){plotfunc << "'m'";}
	//			else if(sensNr==7){plotfunc << "'k--'";}
	//			else if(sensNr==8){plotfunc << "'b--'";}
	//			else {plotfunc << "'k'";}
	//			plotfunc << ",'linewidth',1);\n";
	//			plotfunc << "hold on;\n";
	//		}
	//		//plotfunc << "title({[' Y1: " << GetSensor(ysensor(1)).GetSensorName() << " ']";
	//		//for(int sensNr=2;sensNr<=option;sensNr++)
	//		//{
	//		//	plotfunc << ";[' Y" << sensNr << ": " << GetSensor(ysensor(sensNr)).GetSensorName() << " ']";
	//		//}
	//		//plotfunc << "},'Fontsize',16);\n";
	//		int sameSensorType = 1;
	//		for(int sensNr=2;sensNr<=option;sensNr++)
	//		{
	//			if(!GetSensor(ysensor(sensNr-1)).GetTypeName().Compare(GetSensor(ysensor(sensNr)).GetTypeName()))
	//			{				
	//				sameSensorType = 0;
	//				break;
	//			}
	//		}
	//		if(sameSensorType)
	//		{
	//			plotfunc << "ylabel('" << GetSensor(ysensor(1)).GetTypeName() << "','Fontsize',16);\n";// same type
	//		}
	//		else
	//		{				
	//			plotfunc << "ylabel({['" << GetSensor(ysensor(1)).GetTypeName() << "']";
	//			for(int sensNr=2;sensNr<=option;sensNr++)
	//			{
	//				plotfunc << ";['" << GetSensor(ysensor(sensNr)).GetTypeName() << "']";	// different types
	//			}
	//			plotfunc << "},'Fontsize',16);\n"; //$ MS 2011-4-5
	//		}

	//		plotfunc << "xlabel('time','Fontsize',16);\n";
	//		plotfunc << "legend('" << GetSensor(ysensor(1)).GetSensorName();
	//		for(int sensNr=2;sensNr<=option;sensNr++)
	//		{
	//			plotfunc	<< "','" << GetSensor(ysensor(sensNr)).GetSensorName();
	//		}
	//		plotfunc << "');\n";
	//		//plotfunc << "set(h,'LineWidth',1.0);\n";
	//		plotfunc << "waitfor(h(1));\n"; //wait till window is closed
	//		plotfunc << "exit;\n";       //exit instance of matlab; too many instances will need enormous memory and may crash system 
	//		plotfunc.close();
	//		ShellExecute(NULL, "open", "matlab",	
	//			" -nosplash /r MBSplotfunc",
	//			dir.c_str(), SW_HIDE);
	//		//path.c_str(), SW_HIDE);
	//		//path.c_str(), SW_SHOW);

	//		_chdir(oldpath.c_str());

	//		break;

	//	}
	//$ DR 2015-02-12:]
	
	//action=150: set start_time = 0
	case 150:
		{
			GetSolSet().starttime = 0;
	    break;
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		//information exchange, etc.
	case 201: //NSensors
		{
			return NSensors();
			break;
		}
	case 202: //Element names
		{
			ElementData ed;
			for (int i=1; i <= NE(); i++)
			{
				int flag = 0; //add information on constraint, rigid, flexible, actuator, spring?
				ed.SetInt(flag, GetElement(i).GetElementName().c_str()); edc->Add(ed);
			}
			break;
		}
	case 203: //Sensor names
		{
			for (int i=1; i <= NSensors(); i++)
			{
				int storagemode = 0; //add information on type?
				if(value==1)
				{
	
					storagemode = GetSensor(i).GetSignalStorageMode();
				}
				ElementData ed;
				ed.SetInt(storagemode, GetSensor(i).GetSensorName().c_str()); edc->Add(ed);
			}
			break;
		}		
	case 204: //Sensor value 
		{
			ElementData ed;
			if(option==1) // return time & all system values
			{
				Vector v(1+NSensors());
				v[0] = GetTime();
				for(int i=1; i <= NSensors(); i++)
				{
					v[i] = GetSensor(i).GetLastValue(); 
				}
				ed.SetVector(v.GetVecPtr(),v.GetLen(),"values");
				edc->Add(ed);
			}
			else //if(option==0) // return single value from sensor #value
			{
				double v = GetSensor(value).GetLastValue();
				ed.SetDouble(v, "value"); edc->Add(ed);
			}
			break;
		}
	case 205: //GeomElement names //MBS: DrawElement == GeomElement
		{
			for (int i=1; i <= NDrawElements(); i++)
			{
				int flag = 0; //add information on type?
				ElementData ed;
				ed.SetInt(flag, GetDrawElement(i)->GetName().c_str()); edc->Add(ed);
			}
			break;
		}
	case 206: //NNodes
		{
			return NNodes();
			break;
		}
	case 207: //Nodes names
		{
			for (int i=1; i <= NNodes(); i++)
			{
				int flag = 0; //add information on type?
				ElementData ed;
				ed.SetInt(flag, (GetNode(i).GetObjectName()+mystr("-")+mystr(GetNode(i).SOS())+("DOF")).c_str()); edc->Add(ed);
			}
			break;
		}
	case 208: //N initial values
		{
			int ss = GetSystemSize();
			return ss;
			break;
		}
	case 209: //Material names
		{
			for (int i=1; i <= NMaterials(); i++)
			{
				int flag = 0; //add information on type?
				ElementData ed;
				//ed.SetInt(flag, (GetMaterial(i).GetMaterialName()+mystr("-")+mystr(GetMaterial(i).GetTypeSpec())).c_str()); edc->Add(ed);
				ed.SetInt(flag, (GetMaterial(i).GetMaterialName()).c_str()); edc->Add(ed);	//$ DR 2013-01-21 change for script language
			}
			break;
		}
	case 210: //EditAllOptions
		//!AD: revised 2012-07-26
		{
			if (option==0) //Get all mbs_edc_options and mbssolset into edc
			{
				edc->Reset();

				if(value == 1) // save solver options only
				{
					SolverOptions2EDC(edc);
				}
				else if(value == 2) // save hotint options only
				{
					Options2EDC(edc);
				}
				else if(value == 0) // entire edc - currently no dialog active for all options
				{
					//SetOptions2EDCOptions();//copy numbered i/d/t options to EDC Tree options ==> this is because old dialogs modify the i/d/toptions //JG
					//edc->CopyFrom(*GetMBS_EDC_Options());
					//$ DR 2014-10-24 new implementation
					Options2EDC(edc);	// hotint options in edc
					ElementDataContainer edc_tmp;
					SolverOptions2EDC(&edc_tmp);	// solver options in edc_tmp
					edc->TreeReplaceEDCDataWith(&edc_tmp);	// options and solver options in edc
				}

				////--------------------------------------------
				////RL:b
				//ElementDataContainer edcnew;
				////copy mbssolset to edc:
				//Options2EDC(&edcnew);
				////replace edcnew data in edc:
				//edc->TreeReplaceEDCDataWith(&edcnew);
				////RL:e
				////--------------------------------------------
			}
			else if (option==1) //Write modified edc-data to mbs_edc_options
			{
				//GetMBS_EDC_Options()->TreeReplaceEDCDataWith(edc); // overwrite MBS-EDC first

				//if(value == 1) // update solver options only
				//{
				//	SetComputationSolverOptions();
				//}
				//else if(value == 2) // update hotint options only
				//{
				//	SetEDCOptions2Options();
				//}
				//else if(value == 0) // update entire edc - currently no dialog active for all options
				//{
				//	SetComputationSolverOptions();
				//	SetEDCOptions2Options();
				//}
				
				//$ DR 2014-10-24 new implementation
				if(value == 1) // update solver options only
				{
					EDC2SolverOptions(edc);
				}
				else if(value == 2) // update hotint options only
				{
					EDC2Options(edc);
				}
				else if(value == 0) // update entire edc - currently no dialog active for all options
				{
					EDC2SolverOptions(edc);
					EDC2Options(edc);
				}

				OpenLogFile(1);
			}

			break;
		}
		//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 211: //Model Data
		{
			if (option==0 && GetModelDataContainer() != 0) //Get Data in EDC
			{
				edc->Reset();
				edc->CopyFrom(*GetModelDataContainer());
			}
			else if (option==1) //Set Data in EDC
			{
				if (!GetModelDataContainer()) //generate empty modeldatacontainer
				{
					ElementDataContainer edcnew;
					SetModelDataContainer(edcnew);
				}
				else
				{
					GetModelDataContainer()->Reset();
				}
				GetModelDataContainer()->CopyFrom(*edc);
			}
			else if (option==2) //Load Model Data
			{
			}
			else if (option==3) //Save Model Data
			{
				ElementDataContainer* medc = GetModelDataContainer();
				mystr filestr;
				EDCParser().EDC2String(*medc, filestr);

				{
					mystr filename = GetTOption( 88); //directory+filename
					int opt = ios::out;


					ofstream* storeout = new ofstream;
#ifdef my_new_stdiostream
					storeout->open(filename.c_str(), opt); //with read sharing allowed!
#else
					storeout->open(filename.c_str(), opt, filebuf::sh_read); //with read sharing allowed!
#endif

					if ((int)storeout->good() == 0 || storeout->fail())  
					{
						UO() << "ERROR: could not write model data!!!\n";
						UO() << "good=" << storeout->good() << ", fail=" << storeout->fail() << "\n";
						UO() << "file=" << filename.c_str() << "\n";
					}
					else
					{
						(*storeout) << filestr.c_str() << "\n";
					}

					storeout->close();
					delete storeout;

				}
			}

			break;
		}
		//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//case 212: //SetOptions //$ DR 2014-10-24 not necessary anymore
	//	{
	//		if (option==1) 
	//		{
	//			SetOptions2EDCOptions();//copy numbered i/d/t options to EDC Tree options
	//		}
	//		else if (option==2) 
	//		{
	//			SetEDCOptions2Options();//copy EDC Tree options to numbered i/d/t options 
	//		}
	//		break;
	//	}
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 214: //Load names
		{
			ElementData ed;
			for (int i=1; i <= NLoads(); i++)
			{
				int val = 0; //add information on type?
				ed.SetInt(val, GetLoad(i).LoadName().c_str()); edc->Add(ed);
			}
			break;
		}
	case 215: // set names - option
		{
			ElementData ed;
			for(int i=1; i <= NSets(); i++)
			{
				int settype = 0;
				if(option==1) 
				{
					settype = GetSetPtr(i)->GetType();
				}			
				ed.SetInt(settype, GetSetPtr(i)->GetName().c_str()); edc->Add(ed);
			}
			break;
		}
	case 216: // mesh names
		{
			ElementData ed;
			for(int i=1; i <= NMeshes(); i++)
			{
				int val = 0;
				ed.SetInt(val, GetMeshPtr(i)->GetMeshName().c_str()); edc->Add(ed);
			}
			break;		
		}
	case 217: // edc-content of a set
		{
			ElementData ed;
			if(value==0)
			{
				for(int i=1; i <= NSets(); i++)
				{
					ElementDataContainer edcSet;
					GetSetPtr(i)->GetElementData(edcSet);
					ed.SetEDC(&edcSet, GetSetPtr(i)->GetName().c_str());  edc->Add(ed);						
				}
			}
			else if(value <= NSets())
			{
				ElementDataContainer edcSet;
				GetSetPtr(value)->GetElementData(edcSet);
				ed.SetEDC(&edcSet, GetSetPtr(value)->GetName().c_str());  edc->Add(ed);						
			}		
		}

		//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	case 301: //Compute system information
		{
			int rs = GetResortSize();
			int sos = GetSecondOrderSize();
			//int sos_rs = GetSecondOrderSize_RS();
			int es = GetFirstOrderSize();
			int is = GetImplicitSize();
			int ss = GetSystemSize();

			int nbodies=0;
			int njoints=0;
			for (int i=1; i <= NE(); i++)
			{
				if (GetElement(i).IsType(TBody)) nbodies++;
				else njoints++;
			}

			UO() << "++++++++++++++++++++++++++++++++++\n";
			UO() << "System properties:\n";
			UO() << "2nd order ODE=" << sos << " (sos)\n";
			UO() << "Algebraic Equ=" << is << " (is)\n";
			UO() << "1st order ODE=" << es << " (es)\n";
			UO() << "number of initial values=" << ss << " (=2*sos+es+is)\n";
			UO() << "Resorted equ =" << rs << "\n";
			UO() << "N. of bodies=" << nbodies << "\n";
			UO() << "N. of joints=" << njoints << "\n";

			UO() << "++++++++++++++++++++++++++++++++++\n";
			break;
		}
	case 302: //check system; rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
		{
			rv = CheckSystemConsistency();
			break;
		}

	case 320: 
		{
			rv = ComputeEigenmodes();
			break;
		}
	default: 
		{
			UO(UO_LVL_warn) << "Called MultiBodySystem::CallCompFunction("<< action <<",... which does not exist!";		//$ DR 2014-10-24 added warning
		}
	}

	return rv;
}

//$ DR 2012-11-29 new code according to skript-language
void MultiBodySystem::SaveToFile(mystr filename,int save_options)
{
	mystr el = "\n"; //end of line
	mystr ob = "{";  //open bracket
	mystr cb = "}";  //closing bracket
	mystr ds = ": "; //double stop
	mystr eq = "= "; //equal sign

	ofstream of(filename.c_str());

	//++++++++++++++++
	//header:
//	char str[32];
	/*sprintf(str, "%0.3f\n\n", HOTINT_version_info);*/
	
	_SYSTEMTIME time;
	GetSystemTime(&time);

	// do not use HOTINT_data_file_version in header
	mystr header = "";

	header += mystr("%======================================================================\n");
	header += mystr("%                      HOTINT model \n");
	header += mystr("%\n");
	//header += mystr("% model name: ") + mystr(MBS_EDC_TreeGetString("GeneralOptions.ModelFile.internal_model_function_name")) + mystr("\n");
	header += mystr("% model name: ") + mystr(GetOptions()->GeneralOptions()->ModelFileInternalModelFunctionName()) + mystr("\n");	//$ DR 2014-10-21 improve speed
	header += mystr("%\n");
	header += mystr("% saved on: ")+mystr(time.wDay)+mystr(".")+mystr(time.wMonth)+mystr(".")+mystr(time.wYear)+mystr(" at ")+mystr(time.wHour)+mystr(":")+mystr(time.wMinute)+mystr("\n");
	header += mystr("%======================================================================\n");
	header +=mystr("HOTINT_data_file_version = \"")+mystr(GetHotintVersion().GetString())+mystr("\" % version of HOTINT, that has been used to generate this file \n\n");

	of << header;

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//		Check if the mbs can be saved at all
	bool error = 0;

	if(error)
	{
		UO() << "Writing of output file aborted due to the error(s) above! \n";
		of << "% Writing of output file aborted, because nodes are not available for skript language.";
		of.close();
		return;
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//materials:
	for (int i=1; i<=NMaterials(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		MaterialGeneric& m = GetMaterial(i);
		m.GetElementData(edc);

		otext += "Material" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket

		if(dynamic_cast<Beam2DProperties*>(&m) || dynamic_cast<Beam3DProperties*>(&m))
		{
			otext += "AddBeamProperties(Material" +mystr(i)+ ")" + el +el;	
		}
		else
		{
			otext += "AddMaterial(Material" +mystr(i)+ ")" + el +el;	
		}

		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// nodes:
	for (int i=1; i<=NNodes(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		Node& m = GetNode(i);
		m.GetElementData(edc);

		otext += "Node" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket
		otext += "AddNode(Node" +mystr(i)+ ")" + el +el;	
		
		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//elements:
	for (int i=1; i<=NE(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		Element& e = GetElement(i);
		e.GetElementData(edc);

		otext += "Element" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket
		
		if (e.IsType(TConstraint) && !e.IsType(TController))
		{
			otext += "AddConnector(Element" +mystr(i)+ ")" + el +el;	// add the element
		}
		else
		{
			otext += "AddElement(Element" +mystr(i)+ ")" + el +el;		// add the element
		}
		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//loads:
	for (int i=1; i<=NLoads(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		MBSLoad& l = GetLoad(i);
		l.GetElementData(edc);

		otext += "Load" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket
		otext += "AddLoad(Load" +mystr(i)+ ")" + el +el;	// add the element
		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//sensors:
	for (int i=1; i<=NSensors(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		Sensor& s = GetSensor(i);
		s.GetElementData(edc);

		otext += "Sensor" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket
		otext += "AddSensor(Sensor" +mystr(i)+ ")" + el +el;	// add the element
		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// GeomElements:
	for (int i=1; i<=NDrawElements(); i++) 
	{
		mystr otext = "";
		ElementDataContainer edc;
		GeomElement *e = GetDrawElement(i);
		e->GetElementData(edc);

		otext += "GeomElement" + mystr(i);
		otext += el + ob + el;			// open bracket

		mystr str;
		EDCParser().EDC2String(edc, str, "  ");
		otext += str;								// element data
		otext += cb + el;						// close bracket
		
		otext += "AddGeomElement(GeomElement" +mystr(i)+ ")" + el +el;		// add the element

		of << otext;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//options:
	if(save_options)
	{
		ElementDataContainer edc;

		//SolverOptions2EDC(edc); // save solver options only
		//Options2EDC(edc); // save hotint options only

		// entire edc - currently no dialog active for all options
		//$ DR 2014-10-24 old code:
		//SetOptions2EDCOptions();//copy numbered i/d/t options to EDC Tree options ==> this is because old dialogs modify the i/d/toptions //JG //$ DR 2014-10-24 not necessary anymore
		//edc.CopyFrom(*GetMBS_EDC_Options());
		//$ DR 2014-10-24 new code:
		ElementDataContainer edc_tmp;
		Options2EDC(&edc);
		SolverOptions2EDC(&edc_tmp);	// get solver options
		edc.TreeReplaceEDCDataWith(&edc_tmp);	// options + solver options in edc now

		////+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		////set some default values, which should not be loaded:
		//edc_configfile.TreeSetDouble("SolverOptions.start_time", 0.); //do not load any starttime except 0.
		//edc_configfile.TreeSetBool("ViewingOptions.Animation.animate_deformation", 0); //animate at startup
		//edc_configfile.TreeSetBool("ViewingOptions.Animation.RecordSingleFrames.record", 0); //do not record at beginning
		////+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

		edc.TreeDelete("GeneralOptions.ModelFile.hotint_input_data_filename");	
		edc.TreeDelete("GeneralOptions.ModelFile.internal_model_function_name");
		edc.TreeDelete("GeneralOptions.ModelFile.recent_file1");
		edc.TreeDelete("GeneralOptions.ModelFile.recent_file2");
		edc.TreeDelete("GeneralOptions.ModelFile.recent_file3");
		edc.TreeDelete("GeneralOptions.ModelFile.recent_file4");
		edc.TreeDelete("GeneralOptions.ModelFile.recent_file5");

		mystr opts;
		EDCParser().EDC2String(edc, opts, "");
		of << "%======================================================================\n";
		of << "%                      Solver Options \n";
		of << opts;	// write the string to file
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//the end 
	of.close();

	UO(UO_LVL_all) << "\nMBS system saved as '" << filename << "'\n";
}

int MultiBodySystem::LoadFromFile(mystr filename)
{
	Destroy();
	int rv = ReadModelData(filename);
	if(rv)
	{
		SetModelData_Initialized(1);
		//Assemble(); // this is done by ModelChanged() wich is called later anyway
	}
	else
	{
		UO() << "ERROR in LoadFromFile during reading '" << filename << "'\n";
	}
	return rv;
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//these functions are called from WCDriver via WinCompInterface:
MBSObjectFactoryInterface * MultiBodySystem::GetObjectFactory()
{
	return EDCParser().GetObjectFactory();
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++               EDC_MBS_OPTIONS             ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//ElementDataContainer* MultiBodySystem::GetMBS_EDC_Options() 
//{
//	UO(UO_LVL_err) << "ERROR: You are using 'GetMBS_EDC_Options' to access (solver) options. The function GetMBS_EDC_Options() has been replaced by GetOptions() which does not use a copy of the options anymore and is more save!";
//	return mbs_edc_options;
//}
//const ElementDataContainer* MultiBodySystem::GetMBS_EDC_Options() const
//{
//	UO(UO_LVL_err) << "ERROR: You are using 'GetMBS_EDC_Options' to access (solver) options. The function GetMBS_EDC_Options() has been replaced by GetOptions() which does not use a copy of the options anymore and is more save!";
//	return mbs_edc_options;
//}
//void MultiBodySystem::SetMBS_EDC_Options(const ElementDataContainer& edc)	
//{
//	if (mbs_edc_options != 0) delete mbs_edc_options;
//
//	mbs_edc_options = new ElementDataContainer;
//
//	mbs_edc_options->CopyFrom(edc);
//}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
ElementDataContainer* MultiBodySystem::GetMBS_EDC_Variables() 
{
	return mbs_edc_global_variables;
}
const ElementDataContainer* MultiBodySystem::GetMBS_EDC_Variables() const
{
	return mbs_edc_global_variables;
}
void MultiBodySystem::SetMBS_EDC_Variables(const ElementDataContainer& edc)
{
	if (mbs_edc_global_variables != 0) delete mbs_edc_global_variables;

	mbs_edc_global_variables = new ElementDataContainer;

	mbs_edc_global_variables->CopyFrom(edc);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//functions to deal with loading and saving standard configurations accessable via the button-bar
void MultiBodySystem::SaveStandardView(int n, ElementDataContainer* current_hotint_configuration)
{
	ElementDataContainer edc_options;
	Options2EDC(&edc_options);										// get HOTINT options
	edc_options.TreeReplaceEDCDataWith(current_hotint_configuration); // add current config

	for(int i=edc_options.Length(); i>0; i--)	// remove all entries but ViewingOptions
	{
		mystr optName = edc_options.Get(i).GetDataName();
		if(! optName.Compare("ViewingOptions"))
		{
			edc_options.Delete(i);
		}
	}

	// write to file
	mystr filename;
	if(n==1) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename1();}
	if(n==2) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename2();}
	if(n==3) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename3();}

	CMFile file_write(filename, TFMwrite);

	mystr str;
	EDCParser().EDC2String(edc_options, str);

	file_write.RW(mystr("%HOTINT ViewingOptions\n"));
	file_write.RW(mystr("%This file was created automatically by clicking on Ctrl + button '1', '2' or '3' in HOTINT.\n"));
	file_write.RW(mystr("%This file holds a copy of the ViewingOptions which can be loaded in HOTINT by clicking on one of these buttons.\n"));
	file_write.RW(mystr("%\n"));
	file_write.RW(str);

	UO() << "Saved 'ViewingOptions" << n << "' to file '"<< filename <<"'\n";

}

void MultiBodySystem::LoadStandardView(int n, ElementDataContainer* current_hotint_configuration)
{
	// read from file
	mystr filename;
	if(n==1) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename1();}
	if(n==2) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename2();}
	if(n==3) { filename = GetOptions()->GeneralOptions()->SavedViewingOptionsFilename3();}

	CMFile file(filename, TFMread);

	ElementDataContainer edc_options; //get default options, if file can not be loaded!
	Options2EDC(&edc_options);

	ElementDataContainer fileedc;
	if (file.IsGood()) 
	{
		UO() << "load 'SavedViewingOptions" << n << "' from file '"<< filename <<"'\n";

		mystr str;
		file.RWF(str); //read file
		EDCParser().ParseAndExecuteString(str,fileedc);

		//replace default option parameters with loaded parameters
		edc_options.TreeReplaceEDCDataWithReplacingOldNames(&fileedc, "HotintOptions", &CEDCParser::GetNewEDCName, 0);
		EDC2Options(&edc_options);	// replace options 
	}
	else
	{
		UO() << "Warning: Could not read '" << filename << "'. This file is specified to be loaded when clicking button '" << n << "', see GeneralOptions.SavedViewingOptions for more details.\n";
	}

	current_hotint_configuration->CopyFrom(edc_options);

	// set the currently displayed field variable properly... is it necessary?
	mystr fieldvariablename =  GetOptions()->ViewingOptions()->FiniteElementsContourPostProcessingVariableName();	
	int idx = GetIndexOfPostProcessingFieldVariable(fieldvariablename);
	SetIndexOfActualPostProcessingFieldVariable(idx);
}


//get values in MBS_EDC:
void MultiBodySystem::MBS_EDC_TreeGetDouble(double& data, const char* name) const
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeGetDouble' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	data = -1;
	//data = GetMBS_EDC_Options()->TreeGetDouble(name);
}
void MultiBodySystem::MBS_EDC_TreeGetInt(int& data, const char* name) const
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeGetInt' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	data = -1;
	//data = GetMBS_EDC_Options()->TreeGetInt(name); //(int == bool internally)
}
const char* MultiBodySystem::MBS_EDC_TreeGetString(const char* name) const
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeGetString' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	//return GetMBS_EDC_Options()->TreeGetString(name);
	return name;
}

//set values in MBS_EDC:
void MultiBodySystem::MBS_EDC_TreeSetDouble(double data, const char* name)
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeSetDouble' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	//GetMBS_EDC_Options()->TreeSetDouble(name, data);
}
void MultiBodySystem::MBS_EDC_TreeSetInt(int data, const char* name)
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeSetInt' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	//GetMBS_EDC_Options()->TreeSetInt(name, data); //(int == bool internally)
}
void MultiBodySystem::MBS_EDC_TreeSetString(const char* data, const char* name)
{
	UO(UO_LVL_err) << "ERROR: You are using 'MBS_EDC_TreeSetString' to access option '"<< name <<"'. The functions MBS_EDC_Tree.. have been replaced by GetOptions()...!";
	//GetMBS_EDC_Options()->TreeSetString(name, data);
}

const CMBSParser& MultiBodySystem::MBSParser() const {return *mbs_parser;}
CMBSParser& MultiBodySystem::MBSParser() {return *mbs_parser;}

const CEDCParser& MultiBodySystem::EDCParser() const {return *edc_parser;}
CEDCParser& MultiBodySystem::EDCParser() {return *edc_parser;}

int MultiBodySystem::File2Str(const char* filename,	mystr& str)
{
	mystr filen = filename;
	CMFile file(filename, TFMread);
	UO()	<< "--------------------------------\n";

	//read parameters from file
	if (file.IsGood()) 
	{
		UO(UO_LVL_all) << "read " << filename << "\n";	//$ DR 2015-05-05 added UO_LVL_all

		//read file
		file.RWF(str);	
	}
	else
	{
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "ERROR: failed to read model data from " << filename << "\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";

		return 0;  // error
	}
	return 1;
}

//this function reads a file and stores it in an edc
int MultiBodySystem::File2EDC(const char* filename, ElementDataContainer* edc_file)
{
	//$ DR 2013-01-14:[ added call for stl-files
	mystr filen = filename;
	if(filen.Find(".stl")>0)
	{
		mystr abspath = GetAbsolutePath(filen);
		return STLFile2EDC(abspath, 0, edc_file);
	}
	//$ DR 2013-01-14:]

	CMFile file(GetAbsolutePath(filen).c_str(), TFMread);
	UO()	<< "--------------------------------\n";

	//read parameters from file
	if (file.IsGood()) 
	{
		UO(UO_LVL_all) << "read " << filename << "\n"; //$ DR 2015-05-05 added UO_LVL_all

		//read file
		mystr str;
		file.RWF(str);		

		// overwrite edc with contents from file
//		EDCParser().String2TreeEDC(/*this, */str, *edc_file);

		EDCParser().ParseAndExecuteString(str, *edc_file);
	}
	else
	{
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "ERROR: failed to read model data from " << filename << "\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";

		return 0;  // error
	}
	return 1;
}

//DR 2013-01-14 this function reads a stl-file and writes the data in the edc
int MultiBodySystem::STLFile2EDC(char* file, int binary, ElementDataContainer* return_value)
{
	int initsize = 10000;
	TArray<Vector3D> newnormals(initsize);
	TArray<Vector3D> newpoints(3*initsize);

	if (!binary)
	{
		ifstream ist(file);

		char buf[200];
		Vector3D pts[3];
		Vector3D normal;

		int cntface = 0;
		int vertex = 0;
		int badnormals = 0;

		while (ist.good())
		{
			ist >> buf;
			if (strcmp (buf, "facet") == 0)
			{
				cntface++;
			}

			normal = 0;

			if (strcmp (buf, "normal") == 0)
			{
				ist >> normal.X() 
					>> normal.Y()
					>> normal.Z();
				if (normal.Norm() > 0)
					normal /= normal.Norm();
			}


			if (strcmp (buf, "vertex") == 0)
			{
				ist >> pts[vertex].X()
					>> pts[vertex].Y()
					>> pts[vertex].Z();


				vertex++;
				if (vertex == 3)
				{
					if (normal.Norm() <= 1e-5)
					{
						Vector3D v1 = pts[1]-pts[0];
						Vector3D v2 = pts[2]-pts[0];
						normal = v1.Cross(v2);
						if (normal.Norm() != 0) normal *= 1./normal.Norm();		  
					}

					vertex = 0;

					if ( (Dist (pts[0], pts[1]) < 1e-16) ||
						(Dist (pts[0], pts[2]) < 1e-16) ||
						(Dist (pts[1], pts[2]) < 1e-16) )
					{
						;
					}
					else
					{
						newnormals.Add(normal);
						newpoints.Add(pts[0]);
						newpoints.Add(pts[1]);
						newpoints.Add(pts[2]);
					}
				}
			}
		}

	}	//end read STL ASCII format
	else
	{ 
		//read STL BINARY format

		int binary_mode = 1;
		CMFile infile(file, TFMread, binary_mode);

		Vector3D pts[3];
		Vector3D normal;

		int nofacets = 0;
		int vertex = 0;

		if (sizeof(int) != 4 || sizeof(float) != 4)
			UO() << "for stl-binary compatibility only use 32 bit compilation!!!";

		//specific settings for stl-binary format
		const int namelen = 80; //length of name of header in file
		const int nospaces = 2; //number of spaces after a triangle

		//read header: name
		//char buf[namelen+1];

		mystr str;
		//read leading 80 characters
		infile.RWchars(namelen, str);
		//global_uo << "STL-header=" << str << "\n";

		//read 4-byte int number of facets
		infile.RWbinaryInt(nofacets);
		//global_uo << "No. facets=" << nofacets << "\n";

		float f;
		char spaces[nospaces];

		//read triangles:
		int cntface, j;
		for (cntface = 1; cntface <= nofacets; cntface++)
		{
			//read 3 floats for triangle normal
			infile.RWbinaryFloat(f); normal.X() = (double)f;
			infile.RWbinaryFloat(f); normal.Y() = (double)f;
			infile.RWbinaryFloat(f); normal.Z() = (double)f;

			//read 3 x 3 floats for (x,y,z) coordinates of triangle points
			for (j = 0; j < 3; j++)
			{
				infile.RWbinaryFloat(f); pts[j].X() = (double)f;
				infile.RWbinaryFloat(f); pts[j].Y() = (double)f;
				infile.RWbinaryFloat(f); pts[j].Z() = (double)f;	  
			} 

			//read two bytes for attributes (usually unused)
			infile.RWchar(spaces[0]); 
			infile.RWchar(spaces[1]);
			//global_uo << "sp1=" << (int)spaces[0] << ", sp1=" << (int)spaces[1] << "\n";

			newnormals.Add(normal);
			newpoints.Add(pts[0]);
			newpoints.Add(pts[1]);
			newpoints.Add(pts[2]);
		}	    
	}

	Matrix m(newnormals.Length(), 3);
	Matrix n(newnormals.Length(), 3);
	int idx = 1;
	for (int i=1; i <= newnormals.Length(); i++)
	{
		m(i,1) = idx++;
		m(i,2) = idx++;
		m(i,3) = idx++;
		n(i,1) = newnormals(i).X();
		n(i,2) = newnormals(i).Y();
		n(i,3) = newnormals(i).Z();
	}

	//ElementDataContainer edc;

	ElementData ed;
	ed.SetMatrix(m.GetMatPtr(),m.Getrows(),m.Getcols(),"triangles");
	return_value->Add(ed);

	ed.SetMatrix(n.GetMatPtr(),n.Getrows(),n.Getcols(),"normals");
	return_value->Add(ed);

	Matrix p(newpoints.Length(), 3);
	for (int i=1; i <= newpoints.Length(); i++)
	{
		p(i,1) = newpoints(i).X();
		p(i,2) = newpoints(i).Y();
		p(i,3) = newpoints(i).Z();
	}

	ed.SetMatrix(p.GetMatPtr(),p.Getrows(),p.Getcols(),"points");
	return_value->Add(ed);

	//return_value.SetEDC(edc.GetCopy(),"MeshData");
	return 1;
}

//this function computes mass, moment of inertia, volume and center of gravity based on the data about the geometry and the material
int MultiBodySystem::ComputeInertia(ElementDataContainer* data, ElementDataContainer* return_value)
{
	int material_number = 0;
	double rho = 0;

	// set rho according to material_number or directly entered density, if there is no material_number
	if(!data->Find("material_number"))	// if there is no material_number search for entry "density"
	{
		rho = data->TreeGetDouble("density",0);
	}
	else
	{
		material_number = data->TreeGetInt("material_number",0);		
		if(material_number <= NMaterials())
		{
			rho = GetMaterial(material_number).Density();
		}
		else
		{
			return 0;		// not a valid material_number
		}
	}

	if(!rho) 	{	return 0; }// density == 0
	// rho is now set

	// start the computation of the inertia values
	Matrix3D Iphi(0.);
	double mass, volume;
	Vector3D cog;

	if(data->Find("MeshData"))
	{
		//see Maple file "maple/integration_triangle_volume.mws" and paper 'computing moments of piecewise polynomial surfaces'
		double im11=0, im12=0, im13=0, im22=0, im23=0, im33=0;
		double vol=0;
		Vector3D center(0.,0.,0.);

		int dummy = 0;
		int Npoints, Ntrigs;

		double * p;
		double * t;
		data->TreeGetMatrix("MeshData.points",&p,Npoints,dummy);
		if(dummy!=3) {return 0;}	// error
		data->TreeGetMatrix("MeshData.triangles",&t,Ntrigs,dummy);
		if(dummy!=3) {return 0;}	// error

		Matrix pts(Npoints,3,p);
		Matrix trigs(Ntrigs,3,t);

		for (int i=1; i <= Ntrigs; i++)
		{
			//const Vector3D& p1 = locpoints(trigs(i).Get(1));
			//const Vector3D& p2 = locpoints(trigs(i).Get(2));
			//const Vector3D& p3 = locpoints(trigs(i).Get(3));

			const Vector3D& p1 = Vector3D(pts(trigs(i,1),1),pts(trigs(i,1),2),pts(trigs(i,1),3));
			const Vector3D& p2 = Vector3D(pts(trigs(i,2),1),pts(trigs(i,2),2),pts(trigs(i,2),3));
			const Vector3D& p3 = Vector3D(pts(trigs(i,3),1),pts(trigs(i,3),2),pts(trigs(i,3),3));

			double v1x = p2.X()-p1.X();
			double v1y = p2.Y()-p1.Y();
			double v1z = p2.Z()-p1.Z();

			double v2x = p3.X()-p1.X();
			double v2y = p3.Y()-p1.Y();
			double v2z = p3.Z()-p1.Z();
			double px = p1.X();
			double py = p1.Y();
			double pz = p1.Z();

			im11 += v1x*v1x*v1z*(v1x*v2y-v2x*v1y)/20.0+px*px*pz*(v1x*v2y-v2x*v1y)/2.0+(
				2.0*px*v1x*v1z+v1x*v1x*pz)*(v1x*v2y-v2x*v1y)/12.0+(2.0*v2x*v1x*v1z+v1x*v1x*v2z)
				*(v1x*v2y-v2x*v1y)/60.0+(2.0*px*v2x*v1z+2.0*px*v1x*v2z+2.0*v2x*v1x*pz)*(v1x*v2y
				-v2x*v1y)/24.0+(v2x*v2x*v1z+2.0*v2x*v1x*v2z)*(v1x*v2y-v2x*v1y)/60.0+(px*px*v1z+
				2.0*px*v1x*pz)*(v1x*v2y-v2x*v1y)/6.0+(px*px*v2z+2.0*px*v2x*pz)*(v1x*v2y-v2x*v1y
				)/6.0+(2.0*px*v2x*v2z+v2x*v2x*pz)*(v1x*v2y-v2x*v1y)/12.0+v2x*v2x*v2z*(v1x*v2y-
				v2x*v1y)/20.0;
			double MapleGenVar1 = ((px*v1y+v1x*py)*v1z+v1x*v1y*pz)*(v1x*v2y-v2x*v1y)/12.0+((
				v2x*v1y+v1x*v2y)*v1z+v1x*v1y*v2z)*(v1x*v2y-v2x*v1y)/60.0+((px*v2y+v2x*py)*v1z+(
				px*v1y+v1x*py)*v2z+(v2x*v1y+v1x*v2y)*pz)*(v1x*v2y-v2x*v1y)/24.0+(v2x*v2y*v1z+(
				v2x*v1y+v1x*v2y)*v2z)*(v1x*v2y-v2x*v1y)/60.0+(px*py*v2z+(px*v2y+v2x*py)*pz)*(
				v1x*v2y-v2x*v1y)/6.0;
			im12 += MapleGenVar1+((px*v2y+v2x*py)*v2z+v2x*v2y*pz)*(v1x*v2y-v2x*v1y)/12.0
				+px*py*pz*(v1x*v2y-v2x*v1y)/2.0+(px*py*v1z+(px*v1y+v1x*py)*pz)*(v1x*v2y-v2x*v1y
				)/6.0+v1x*v1y*v1z*(v1x*v2y-v2x*v1y)/20.0+v2x*v2y*v2z*(v1x*v2y-v2x*v1y)/20.0;
			im13 += px*pz*pz*(v1x*v2y-v2x*v1y)/4.0+v1x*v1z*v1z*(v1x*v2y-v2x*v1y)/40.0+
				v2x*v2z*v2z*(v1x*v2y-v2x*v1y)/40.0+(px*v1z*v1z+2.0*v1x*pz*v1z)*(v1x*v2y-v2x*v1y
				)/24.0+(v2x*v1z*v1z+2.0*v1x*v2z*v1z)*(v1x*v2y-v2x*v1y)/120.0+(2.0*(px*v2z+v2x*
				pz)*v1z+2.0*v1x*pz*v2z)*(v1x*v2y-v2x*v1y)/48.0+(2.0*v2x*v2z*v1z+v1x*v2z*v2z)*(
				v1x*v2y-v2x*v1y)/120.0+(2.0*px*pz*v2z+v2x*pz*pz)*(v1x*v2y-v2x*v1y)/12.0+(px*v2z
				*v2z+2.0*v2x*pz*v2z)*(v1x*v2y-v2x*v1y)/24.0+(2.0*px*pz*v1z+v1x*pz*pz)*(v1x*v2y-
				v2x*v1y)/12.0;
			im22 += py*py*pz*(v1x*v2y-v2x*v1y)/2.0+v2y*v2y*v2z*(v1x*v2y-v2x*v1y)/20.0+
				v1y*v1y*v1z*(v1x*v2y-v2x*v1y)/20.0+(2.0*py*v1y*v1z+v1y*v1y*pz)*(v1x*v2y-v2x*v1y
				)/12.0+(2.0*v2y*v1y*v1z+v1y*v1y*v2z)*(v1x*v2y-v2x*v1y)/60.0+(2.0*py*v2y*v1z+2.0
				*py*v1y*v2z+2.0*v2y*v1y*pz)*(v1x*v2y-v2x*v1y)/24.0+(v2y*v2y*v1z+2.0*v2y*v1y*v2z
				)*(v1x*v2y-v2x*v1y)/60.0+(py*py*v1z+2.0*py*v1y*pz)*(v1x*v2y-v2x*v1y)/6.0+(py*py
				*v2z+2.0*py*v2y*pz)*(v1x*v2y-v2x*v1y)/6.0+(2.0*py*v2y*v2z+v2y*v2y*pz)*(v1x*v2y-
				v2x*v1y)/12.0;
			im23 += v2y*v2z*v2z*(v1x*v2y-v2x*v1y)/40.0+py*pz*pz*(v1x*v2y-v2x*v1y)/4.0+
				v1y*v1z*v1z*(v1x*v2y-v2x*v1y)/40.0+(py*v1z*v1z+2.0*v1y*pz*v1z)*(v1x*v2y-v2x*v1y
				)/24.0+(v2y*v1z*v1z+2.0*v1y*v2z*v1z)*(v1x*v2y-v2x*v1y)/120.0+(2.0*(py*v2z+v2y*
				pz)*v1z+2.0*v1y*pz*v2z)*(v1x*v2y-v2x*v1y)/48.0+(2.0*v2y*v2z*v1z+v1y*v2z*v2z)*(
				v1x*v2y-v2x*v1y)/120.0+(2.0*py*pz*v2z+v2y*pz*pz)*(v1x*v2y-v2x*v1y)/12.0+(py*v2z
				*v2z+2.0*v2y*pz*v2z)*(v1x*v2y-v2x*v1y)/24.0+(2.0*py*pz*v1z+v1y*pz*pz)*(v1x*v2y-
				v2x*v1y)/12.0;
			im33 += v2z*v1z*v1z*(v1x*v2y-v2x*v1y)/60.0+v1z*v1z*v1z*(v1x*v2y-v2x*v1y)/
				60.0+pz*v2z*v1z*(v1x*v2y-v2x*v1y)/12.0+pz*pz*v1z*(v1x*v2y-v2x*v1y)/6.0+v2z*v2z*
				v1z*(v1x*v2y-v2x*v1y)/60.0+pz*pz*v2z*(v1x*v2y-v2x*v1y)/6.0+pz*v2z*v2z*(v1x*v2y-
				v2x*v1y)/12.0+pz*pz*pz*(v1x*v2y-v2x*v1y)/6.0+v2z*v2z*v2z*(v1x*v2y-v2x*v1y)/60.0
				+pz*v1z*v1z*(v1x*v2y-v2x*v1y)/12.0;

			vol += 1./6.*v1z*(v1x*v2y-v2x*v1y)+1./6.*v2z*(v1x*v2y-v2x*v1y)+1./2.*pz*(v1x*v2y-v2x*v1y);

			center.X() += v1x*v1z*(v1x*v2y-v2x*v1y)/12.0+(v2x*v1z+v1x*v2z)*(v1x*v2y-v2x*v1y)/24.0+v2x*v2z*(v1x*v2y-v2x*v1y)/12.0+(px*v1z+v1x*pz)*(v1x*v2y-v2x*v1y)/6.0+(px*v2z+v2x*pz)*(v1x*v2y-v2x*v1y)/6.0+px*pz*(v1x*v2y-v2x*v1y)/2.0;
			center.Y() += v1y*v1z*(v1x*v2y-v2x*v1y)/12.0+(v2y*v1z+v1y*v2z)*(v1x*v2y-v2x*v1y)/24.0+v2y*v2z*(v1x*v2y-v2x*v1y)/12.0+(py*v1z+v1y*pz)*(v1x*v2y-v2x*v1y)/6.0+(py*v2z+v2y*pz)*(v1x*v2y-v2x*v1y)/6.0+py*pz*(v1x*v2y-v2x*v1y)/2.0;
			center.Z() += v1z*v1z*(v1x*v2y-v2x*v1y)/24.0+v2z*v1z*(v1x*v2y-v2x*v1y)/24.0+v2z*v2z*(v1x*v2y-v2x*v1y)/24.0+pz*v1z*(v1x*v2y-v2x*v1y)/6.0+pz*v2z*(v1x*v2y-v2x*v1y)/6.0+pz*pz*(v1x*v2y-v2x*v1y)/4.0;
		}
		volume = vol;
		mass = volume*rho;
		Iphi = rho * Matrix3D(im22+im33,-im12,-im13, -im12, im11+im33, -im23, -im13, -im23, im11+im22);
		if (vol != 0) 
		{
			cog = center * (1./vol);
		}
	}
	else if(data->Find("Cube"))
	{
		double x,y,z;
		data->TreeGetVector3D("Cube.body_dimensions",x,y,z);
		
		volume = x*y*z;
		mass = volume*rho;
		Iphi(1,1)= 1./12.*mass*(Sqr(y)+Sqr(z));
		Iphi(2,2)= 1./12.*mass*(Sqr(x)+Sqr(z));
		Iphi(3,3)= 1./12.*mass*(Sqr(x)+Sqr(y));
		cog = Vector3D(0.);
	}
	else if(data->Find("Cylinder"))
	{

	}

	// set the new edc
	ElementData ed;
	ed.SetDouble(volume,"volume"); return_value->Add(ed);
	ed.SetDouble(mass,"mass"); return_value->Add(ed);
	ed.SetVector3D(cog.X(),cog.Y(),cog.Z(),"center_of_mass"); return_value->Add(ed);
	ed.SetMatrix(Iphi.GetMatPtr(),3,3,"moment_of_inertia"); return_value->Add(ed);

	return 1;

}

//this function reads a column of a file and returns it as vector, rv = 1.. success, 0..file is not good
int MultiBodySystem::LoadVectorFromFile(const char* filename, int col, ElementDataContainer* return_value, int readRowInsteadOfCol)
{
	CMatrixFile file( filename, TFMread );
	if(file.IsGood())
	{
		Vector test;
		if(readRowInsteadOfCol)	// read 1 row (slow but it works) //$ DR 2014-09-25
		{
			file.ReadAllColumns();
			test.SetLen(file.NColumns());
			for(int i=1; i<=file.NColumns(); i++)
			{
				test(i) = file.GetData(col,i);
			}
		}
		else	// read 1 column
		{
			file.ReadSingleColumn(col);
			TArray<double>& coldata = file.Column(col);
			test.SetVector(coldata);
		}
		// set the new edc
		ElementData ed;
		ed.SetVector(test.GetVecPtr(), test.Length(),"vector"); return_value->Add(ed);
		return 1;
	}
	else
	{
		return 0;
	}
}

//void MultiBodySystem::AddReplaceModelDataEDC(/*const */ElementDataContainer& edc)
//{
//	//$ DR 2013-06-12:[ code that has been in AddModelData
//	// verbosity level is needed for further opertation
//	if (edc.TreeFind("LoggingOptions.output_level"))    // temporary solution - best would be to have an entry like "GeneralOptions.show_general_warnings" in edc_file     //$ PG 2012-2-3: substituted SolverOptions.Log.output_level by LoggingOptions.output_level
//	{
//		GetOptions()->LoggingOptions()->OutputLevel() = edc.TreeGetInt("LoggingOptions.output_level");       //$ PG 2012-2-3: substituted solset.output_level by GetOptions()->LoggingOptions()->OutputLevel() and SolverOptions.Log.output_level by LoggingOptions.output_level
//	}
//	//$ DR 2013-06-12:] code that has been in AddModelData
//
//	bool show_warnings = ( GetOptions()->LoggingOptions()->OutputLevel() >= UO_LVL_dbg1 );    //$ PG 2012-2-3: substituted solset.output_level by GetOptions()->LoggingOptions()->OutputLevel()
//
////	ElementData* edmodel;
//
//	// loop over all default main branches
//	for (int i=1; i <= edc.Length(); i++) 
//	{
//		// get i-th main branch
//		if(!GetMBS_EDC_Options()->TreeFind(edc.GetPtr(i)->GetDataName()))
//		{
//			ElementDataContainer edci;edci.TreeAdd("",edc.Get(i));
//			edc_modeldata->TreeReplaceEDCDataWith(&edci, show_warnings, mystr(edc.GetPtr(i)->GetDataName())+mystr("."));
//		}
//	}
//
//	//$ DR 2013-06-12:[ code that has been in AddModelData
//	// replace solveroptions & other settings
//	ElementData* eddef;
//	ElementData* edfile;
//
//	// loop over all default main branches
//	for (int i=1; i <= GetMBS_EDC_Options()->Length(); i++) 
//	{
//		// get i-th main branch
//		//eddef = edc_default.GetPtr(i);		
//		eddef = GetMBS_EDC_Options()->GetPtr(i);
//		edfile = edc.TreeFind(eddef->GetDataName());
//		// replace the default settings by the user's file settings
//		if (edfile != 0 && eddef->IsEDC() && edfile->IsEDC())
//		{
//			if(GetMBS_EDC_Options()->TreeFind(edfile->GetDataName()))
//			{				
//				UO(UO_LVL_ext) << "replacing hotint option: " << edfile->GetDataName() << "\n";
//				// use file parameters (when existing) instead of default variables
//				eddef->GetEDC()->TreeReplaceEDCDataWith(edfile->GetEDC(), show_warnings, mystr(edfile->GetDataName())+mystr("."));
//			}
//		}
//	}
//	// overwrite default options by user file data 
//	EDC2SolverOptions(GetMBS_EDC_Options());
//	EDC2Options(GetMBS_EDC_Options());       // copy EDC to i/d/toptions
//
//	OpenLogFile(0);
//
//	UO() << "done.\n";
//	UO() << "--------------------------------\n";
//	//$ DR 2013-06-12:] code that has been in AddModelData
//}


//int MultiBodySystem::ReadModelData(mystr filename)
//{
//	ElementDataContainer edc_file;
//
//	// read parameters from file and store it in ElementDataContainer of mbs (details: see forum)
//	CMFile file(filename, TFMread);
//	UO()	<< "--------------------------------\n";
//	int rv = 1;
//
//	//read parameters from file
//	if (file.IsGood()) 
//	{
//		UO(UO_LVL_0) << "read " << filename << "\n";
//
//		//read file
//		mystr str;
//		file.RWF(str);		
//		//$ DR 2013-03-09:[
//		// Get the variables of the parameter variation
//		ElementDataContainer edc_tmp;
//		ElementData ed_tmp;
//		//$ EK 2014-01-12:[ if edc_modeldata initally does not exist 
//		//if(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.activate") && edc_modeldata)
//		if(GetSolSet().parvar_activate && edc_modeldata)	//$ DR 2014-10-23 improve speed
//		//$ EK  2014-01-12:]
//		{
//			//mystr varname = GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.MBS_EDC_variable_name");
//			mystr varname = GetSolSet().parvar_EDCvarname; //$ DR 2014-10-23 improve speed
//			double val = edc_modeldata->TreeGetDouble(varname);
//			ed_tmp.SetDouble(val,varname);
//			edc_tmp.Add(ed_tmp);
//			//if(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.Var2.activate"))
//			if(GetSolSet().parvar2_activate)	//$ DR 2014-10-23 improve speed
//			{
//				//varname = GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.Var2.MBS_EDC_variable_name");
//				varname = GetSolSet().parvar2_EDCvarname; //$ DR 2014-10-23 improve speed
//				val = edc_modeldata->TreeGetDouble(varname);
//				ed_tmp.SetDouble(val,varname);
//				edc_tmp.Add(ed_tmp);
//			}
//		}		
//		//$ DR 2013-03-09:]
//		//$ RE & MSax 2013-06-27:[
//		//else if(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Optimization.activate") && edc_modeldata)
//		else if(GetSolSet().paropt_activate && edc_modeldata) //$ DR 2014-10-23 improve speed
//		{
//			//int number_of_params = GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Optimization.Parameters.number_of_params");
//			int number_of_params = GetSolSet().par_opt_N_pars; //$ DR 2014-10-23 improve speed
//			for (int col=1; col <= number_of_params; col++)
//			{
//				mystr varname =	GetMBS_EDC_Options()->TreeGetString(mystr("SolverOptions.Optimization.Parameters.param_name")+mystr(col));
//				double val = edc_modeldata->TreeGetDouble(varname);
//				ed_tmp.SetDouble(val,varname);
//				edc_tmp.Add(ed_tmp);
//			}
//		}
//		//$ RE & MSax 2013-06-27:]
//		
////$!AD 26-09-2013: parse line by line directly into modeldata EDC [
//// NEW CODE:
//		edc_modeldata = 0;
//		if (edc_modeldata != 0)
//		{ 
//			delete edc_modeldata;
//			edc_modeldata = 0;
//		}
//		edc_modeldata = new ElementDataContainer;
//
//		// overwrite edc with contents from file
//		//edc_modeldata = &edc_file; //JG, this is just temporary in order to get variables, change to MBSParser->readedc_global/local!!!
//		ElementData ed;
//		ed.SetText(filename, "last_file_name");edc_modeldata->Add(ed); // path is stored e.g. for getting relative paths in include command
//
//		//$ DR 2013-03-09:[
//		// set the values for the parameter variation and save them with flagging them as readonly
//		for(int i=1; i<=edc_tmp.Length();i++)
//		{
//			edc_modeldata->Add(edc_tmp.Get(i));
//			edc_modeldata->Last().SetLocked(1);
//		}
//		edc_tmp.Reset();
//		//$ DR 2013-03-09:]
// 
//		this->EDCParser().GetMBSParser()->SetLocalEDC2(edc_modeldata);
//		rv = this->EDCParser().ParseAndExecuteString(str, *edc_modeldata);
//		edc_file = *(edc_modeldata->GetCopy());	//$ DR+AD 2013-10-01 bugfix, copy is necessary for solver options
//
//// OLD CODE:
//		////////
//		//////////UO() << "WARNING:set read EDC==>modeldata!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
//		////////rv = EDCParser().String2TreeEDC(/*this, */str, edc_file);  
//		//////////UO() << "delete read EDC!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
//		//////////delete read EDC
//		////////edc_modeldata = 0;
//
//		////////SetModelDataContainer(edc_file); //store empty edc to avoid crashes (due to null-pointer exeptions) at start of Generate - function
//		////////
////$!AD ]
//	}
//	else // FileNotGood
//	{
//		UO()	<< "*****************\n";
//		UO()	<< "*****************\n";
//		UO()	<< "*****************\n";
//		UO()	<< "ERROR: failed to read model data from " << filename << "\n";
//		UO()	<< "*****************\n";
//		UO()	<< "*****************\n";
//		UO()	<< "*****************\n";
//		SetModelDataContainer(edc_file); //store empty edc to avoid crashes (due to null-pointer exeptions) at start of Generate - function
//
//		return 0;  //$JG2012-02-21 changed from 2 to 0 in case of error
//	}
//
//	// verbosity level is needed for further opertation
//	if (edc_file.TreeFind("LoggingOptions.output_level"))    // temporary solution - best would be to have an entry like "GeneralOptions.show_general_warnings" in edc_file         //$ PG 2012-2-3: substituted SolverOptions.Log.output_level by LoggingOptions.output_level
//	{
//		GetOptions()->LoggingOptions()->OutputLevel() = edc_file.TreeGetInt("LoggingOptions.output_level");    //$ PG 2012-2-3: substituted solset.output_level by GetOptions()->LoggingOptions()->OutputLevel(), and SolverOptions.Log.output_level by LoggingOptions.output_level
//	}
//	bool show_warnings = ( GetOptions()->LoggingOptions()->OutputLevel() >= UO_LVL_warn );     //$ PG 2012-2-3: substituted solset.output_level by GetOptions()->LoggingOptions()->OutputLevel()
//	// load default parameter settings
//	ElementDataContainer* edc_default = GetMBS_EDC_Options();
//
//	RemoveVariableFromEDCTree(*edc_default, mystr("SolverOptions.ComputationSteps"), 1);
//
//	ElementData* eddef;
//	ElementData* edfile;
//
//	//$ AD: 2011-11-08: special treatment for Computation Step entries in EDC 
//	// as these entries would produce a popupwindow every time when using TreeReplace
//	// copy computation steps manually, assume that entry does NOT exist in default EDC
//	ElementData* edcsfile = edc_file.TreeFind("SolverOptions.ComputationSteps");
//	if(edcsfile != 0 && edcsfile->IsEDC())
//	{
//		ElementDataContainer* edc_solopt_default = edc_default->TreeFind("SolverOptions")->GetEDC();
//		edc_solopt_default->Add(*edcsfile);
//		UO(UO_LVL_ext) << "Added Computation Steps: (" << edcsfile->GetEDC()->Length() << ")\n";
//	// one could eliminate entries from FileEDC now...
//		RemoveVariableFromEDCTree(edc_file, mystr("SolverOptions.ComputationSteps"), 0);
//	}
//
//	//TreeGetString was not used here to avoid an additional warning in case HOTINT_data_file_version is not defined
//	ElementData* HOTINT_data_file_version = edc_file.TreeFind("HOTINT_data_file_version");
//	mystr hidVersion;
//	if (HOTINT_data_file_version != 0)
//	{
//		hidVersion = HOTINT_data_file_version->GetText();
//	}
//	else
//	{
//		hidVersion = "0.0.0";
//	}
//
//	// loop over all default main branches
//	for (int i=1; i <= edc_default->Length(); i++) 
//	{
//
//		// get i-th main branch
//		eddef = edc_default->GetPtr(i);		
//		edfile = edc_file.TreeFind(eddef->GetDataName());
//
//		// replace the default settings by the user's file settings
//		if (edfile != 0 && eddef->IsEDC() && edfile->IsEDC())
//		{
//			mystr edcName(edfile->GetDataName());
//			if(edcName.Compare("PostProcOptions")||edcName.Compare("GraphicsOptions"))	//$ DR 2014-10-16
//			{
//				UO(UO_LVL_warn) << "WARNING: You are using PostProcOptions or GraphicsOptions which are now both included in ViewingOptions. Correct the entries in your model file from 'PostProcOptions...' or 'GraphicsOptions...' to 'ViewingOptions...' otherwise these entries are ignored!\n";
//			}
//			else
//			{
//				UO(UO_LVL_ext) << "replacing defaults: " << edfile->GetDataName() << "\n";
//			}
//			// use file parameters (when existing) instead of default variables
//			//eddef->GetEDC()->TreeReplaceEDCDataWith(edfile->GetEDC(), show_warnings, mystr(edfile->GetDataName())+mystr("."));
//			// use global EDC
//			ElementDataContainer rootedEdFile;
//			rootedEdFile.Add(*edfile);
//
//			edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", hidVersion, &CEDCParser::GetNewEDCName, show_warnings);
//		}
//	}
//
//	// extra treatment for PostProcOptions and GraphicsOptions:
//	// check for PostProcOptions (they have been deleted)
//	if (edfile = edc_file.TreeFind("PostProcOptions"))
//	{
//		ElementDataContainer rootedEdFile;
//		rootedEdFile.Add(*edfile);
//		edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", hidVersion, &CEDCParser::GetNewEDCName, show_warnings);
//	}
//
//	// check for GraphicsOptions (they have been deleted)
//	if (edfile = edc_file.TreeFind("GraphicsOptions"))
//	{
//		ElementDataContainer rootedEdFile;
//		rootedEdFile.Add(*edfile);
//		edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", hidVersion, &CEDCParser::GetNewEDCName, show_warnings);
//	}
//
//	//edc_default->Delete(iPostProcOptions);
//	//edc_default->Delete(iGraphicsOptions);
//
//	// overwrite default options by user file data 
//	EDC2SolverOptions(edc_default);
//	EDC2Options(edc_default);       // copy EDC to i/d/toptions
//
//	UO() << "done.\n";
//	UO() << "--------------------------------\n";
//
//	return rv; //1==success, 0==error
//}

//$ AD 2017-10-10: OBSOLETE, COPYOptimized to be used
//$ DR 2014-10-28 not very beautiful. it is the easiest way to get rid of GetMBS_EDC_Options
// modus: 0.. sensitivity 1..optimization 
//void MultiBodySystem::AddOptimizedParameters2EDC(int nrOfParam, int modus, ElementDataContainer* edc)
//{
//	mystr varname;
//	if(modus==1)
//	{
//		switch(nrOfParam)
//		{
//		case 1:
//			varname = GetSolSet().par_opt_name1;
//			break;
//		case 2:
//			varname = GetSolSet().par_opt_name2;
//			break;
//		case 3:
//			varname = GetSolSet().par_opt_name3;
//			break;
//		case 4:
//			varname = GetSolSet().par_opt_name4;
//			break;
//		case 5:
//			varname = GetSolSet().par_opt_name5;
//			break;
//		case 6:
//			varname = GetSolSet().par_opt_name6;
//			break;
//		case 7:
//			varname = GetSolSet().par_opt_name7;
//			break;
//		case 8:
//			varname = GetSolSet().par_opt_name8;
//			break;
//		case 9:
//			varname = GetSolSet().par_opt_name9;
//			break;
//		case 10:
//			varname = GetSolSet().par_opt_name10;
//			break;
//		}
//	}
//	double val = 0;
//	if(edc_modeldata->Find(varname))	//$ DR 2014-11-24 if not found: varname is part of the (hotint or solver) options
//	{
//		val = edc_modeldata->TreeGetDouble(varname);
//	}
//	edc->TreeSetDouble(varname,val);
//}

//$ AD 2017-10-10: allows to copy entries with fixed (solset) names between EDCS
// modus: 0.. sensitivity 1..optimization
void MultiBodySystem::CopyOptimizedParametersEDC2EDC(int nrOfParam, int modus, ElementDataContainer* edc_source, ElementDataContainer* edc_target)
{
	mystr varname;
	if(modus==1)
	{
		switch(nrOfParam)
		{
		case 1: varname = GetSolSet().par_opt_name1; break;
		case 2:	varname = GetSolSet().par_opt_name2; break;
		case 3: varname = GetSolSet().par_opt_name3; break;
		case 4: varname = GetSolSet().par_opt_name4; break;
		case 5: varname = GetSolSet().par_opt_name5; break;
		case 6: varname = GetSolSet().par_opt_name6; break;
		case 7: varname = GetSolSet().par_opt_name7; break;
		case 8: varname = GetSolSet().par_opt_name8; break;
		case 9: varname = GetSolSet().par_opt_name9; break;
		case 10:varname = GetSolSet().par_opt_name10;break;
		}
	}
	double val = 0;
	if(edc_source->TreeFind(varname))
	{
		val = edc_source->TreeGetDouble(varname);
	}
	edc_target->TreeSetDouble(varname,val);
	edc_target->TreeFind(varname)->SetLocked(1);
}


//$ DR 2014-10-24 new implementation
int MultiBodySystem::ReadModelData(mystr filename)
{
	ElementDataContainer edc_file;

	// read parameters from file and store it in ElementDataContainer of mbs (details: see forum)
	CMFile file(filename, TFMread);
	UO()	<< "--------------------------------\n";
	int rv = 1;

	//read parameters from file
	if (file.IsGood()) 
	{
		UO(UO_LVL_multsim) << "read " << filename << "\n";	//$ DR 2015-05-05 removed "UO_LVL_0" 

		//read file
		mystr str;
		file.RWF(str);		

		// Get the variables of the parameter variation
		ElementDataContainer edc_tmp_parvar;
		ElementDataContainer edc_tmp_paropt;
		if(GetSolSet().parvar_activate && edc_modeldata)	// transfer value of variated parameter(s) to edc_tmp
		{
			mystr varname = GetSolSet().parvar_EDCvarname;
			if(edc_modeldata->TreeFind(varname))	//$ DR 2014-11-24 if not found: varname is part of the (hotint or solver) options
			{
				double val = edc_modeldata->TreeGetDouble(varname);
				edc_tmp_parvar.TreeSetDouble(varname,val);
				if(GetSolSet().parvar2_activate)
				{
					varname = GetSolSet().parvar2_EDCvarname; 
					if(edc_modeldata->TreeFind(varname)) //$ DR 2014-11-24 if not found: varname is part of the (hotint or solver) options
					{
						val = edc_modeldata->TreeGetDouble(varname);
						edc_tmp_parvar.TreeSetDouble(varname,val);
					}
				}
			}
		}		
		else if(GetSolSet().paropt_activate && edc_modeldata) // transfer value of optimized parameter(s) to edc_tmp
		{
			int number_of_params = GetSolSet().par_opt_N_pars; //$ DR 2014-10-23 improve speed
			for (int col=1; col <= number_of_params; col++)
			{
				//AddOptimizedParameters2EDC(col,1,&edc_tmp_paropt);
				CopyOptimizedParametersEDC2EDC(col,1,edc_modeldata,&edc_tmp_paropt);
			}
		}

//$ AD 2017-04-06: keep entries in case a script file runs a compiled model ( Command Open/RunCompiledModel )
		int flag_nested_compiled = 0;
		mystr callingHIDname;
		if (edc_modeldata != 0 && edc_modeldata->Find("HidCalledCpp"))
		{
			flag_nested_compiled = 1;
			callingHIDname = edc_modeldata->TreeGetString("HidCalledCpp");
			edc_modeldata->TreeDelete("last_file_name");
		}
		else // branch "old code": resets the modeldata
		{
			// initialize edc_modeldata
			edc_modeldata = 0;
			if (edc_modeldata != 0)
			{ 
				delete edc_modeldata;
				edc_modeldata = 0;
			}
			edc_modeldata = new ElementDataContainer;
		}

		// some default entries
		ElementData ed;
		ed.SetText(filename, "last_file_name");
		edc_modeldata->	Add(ed); // path is stored e.g. for getting relative paths in include command


		if (modeldata_edc_args.Length())
		{
			modeldata_edc_args.TreeDelete("last_file_name");
			edc_modeldata->TreeReplaceEDCDataWith(&modeldata_edc_args);	//$ DR 2015-12-17 add the parameters of ModelDataContainer (necessary for command line args in hid-models)
		}

		// set the values for the parameter variation and save them with flagging them as readonly
		if(GetSolSet().parvar_activate)
		{
			mystr varname = GetSolSet().parvar_EDCvarname;
			if(edc_tmp_parvar.TreeFind(varname))	
			{
				double val = edc_tmp_parvar.TreeGetDouble(varname);
				edc_modeldata->TreeSetDouble(varname,val);
				edc_modeldata->TreeFind(varname)->SetLocked(1);
				if(GetSolSet().parvar2_activate)
				{
					mystr varname = GetSolSet().parvar2_EDCvarname;
					if(edc_tmp_parvar.TreeFind(varname))	
					{
						double val = edc_tmp_parvar.TreeGetDouble(varname);
						edc_modeldata->TreeSetDouble(varname,val);
						edc_modeldata->TreeFind(varname)->SetLocked(1);
					}
				}
			}
		}
		else if(GetSolSet().paropt_activate && edc_modeldata) // transfer value of optimized parameter(s) back to edc_modeldata
		{
			int number_of_params = GetSolSet().par_opt_N_pars;
			for (int col=1; col <= number_of_params; col++)
			{
				CopyOptimizedParametersEDC2EDC(col,1,&edc_tmp_paropt,edc_modeldata);
			}		
		}

		// set the values fot the optimization 


	
		//for(int i=1; i<=edc_tmp_paropt.Length();i++)
		//{
		//	edc_modeldata->Add(edc_tmp_paropt.Get(i));
		//	edc_modeldata->Last().SetLocked(1);
		//}
		edc_tmp_parvar.Reset();
		edc_tmp_paropt.Reset();
 
		// parse line by line directly into modeldata EDC
		this->EDCParser().GetMBSParser()->SetLocalEDC2(edc_modeldata);
		rv = this->EDCParser().ParseAndExecuteString(str, *edc_modeldata);

// case of nested: set name of last file back again
		if(flag_nested_compiled)
		{
			edc_modeldata->TreeDelete("last_file_name");
			ed.SetText(callingHIDname,"last_file_name");
			edc_modeldata->Add(ed);
			edc_modeldata->TreeDelete("HidCalledCpp");
		}

		//$ PG 2014-11-4: [ the following line caused memory leaks, since GetCopy() creates an instance with the new-operator, which is not being deleted automatically
		//edc_file = *(edc_modeldata->GetCopy());	//$ DR+AD 2013-10-01 bugfix, copy is necessary for solver options
		//$ PG 2014-11-4:   instead, just copy from modeldata
		edc_file.CopyFrom(*edc_modeldata);
		//$ PG 2014-11-4: ]

	}
	else // FileNotGood
	{
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "ERROR: failed to read model data from " << filename << "\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		UO()	<< "*****************\n";
		SetModelDataContainer(edc_file); //store empty edc to avoid crashes (due to null-pointer exeptions) at start of Generate - function

		return 0; 
	}

	// verbosity level is needed for further opertation
	if (edc_file.TreeFind("LoggingOptions.output_level"))    // temporary solution - best would be to have an entry like "GeneralOptions.show_general_warnings" in edc_file   
	{
		GetOptions()->LoggingOptions()->OutputLevel() = edc_file.TreeGetInt("LoggingOptions.output_level");   
	}
	bool show_warnings = ( GetOptions()->LoggingOptions()->OutputLevel() >= UO_LVL_warn ); 

	// load default parameter settings
	ElementDataContainer edc_options, edc_tmp;
	Options2EDC(&edc_options);
	SolverOptions2EDC(&edc_tmp);
	edc_options.TreeReplaceEDCDataWith(&edc_tmp); // options+solver options are now in edc_options

	ElementDataContainer* edc_default = &edc_options;
	RemoveVariableFromEDCTree(*edc_default, mystr("SolverOptions.ComputationSteps"), 1);

	ElementData* eddef;
	ElementData* edfile;

	//$ AD: 2011-11-08: special treatment for Computation Step entries in EDC 
	// as these entries would produce a popupwindow every time when using TreeReplace
	// copy computation steps manually, assume that entry does NOT exist in default EDC
	ElementData* edcsfile = edc_file.TreeFind("SolverOptions.ComputationSteps");
	if(edcsfile != 0 && edcsfile->IsEDC())
	{
		ElementDataContainer* edc_solopt_default = edc_default->TreeFind("SolverOptions")->GetEDC();
		edc_solopt_default->Add(*edcsfile);
		UO(UO_LVL_ext) << "Added Computation Steps: (" << edcsfile->GetEDC()->Length() << ")\n";
	// one could eliminate entries from FileEDC now...
		RemoveVariableFromEDCTree(edc_file, mystr("SolverOptions.ComputationSteps"), 0);
	}

	// loop over all default main branches
	for (int i=1; i <= edc_default->Length(); i++) 
	{

		// get i-th main branch
		eddef = edc_default->GetPtr(i);		
		edfile = edc_file.TreeFind(eddef->GetDataName());

		// replace the default settings by the user's file settings
		if (edfile != 0 && eddef->IsEDC() && edfile->IsEDC())
		{
			mystr edcName(edfile->GetDataName());
			if(edcName.Compare("PostProcOptions")||edcName.Compare("GraphicsOptions"))	//$ DR 2014-10-16
			{
				UO(UO_LVL_warn) << "WARNING: You are using PostProcOptions or GraphicsOptions which are now both included in ViewingOptions. Correct the entries in your model file from 'PostProcOptions...' or 'GraphicsOptions...' to 'ViewingOptions...' otherwise these entries are ignored!\n";
			}
			else
			{
				UO(UO_LVL_ext) << "replacing defaults: " << edfile->GetDataName() << "\n";
			}
			// use file parameters (when existing) instead of default variables
			//eddef->GetEDC()->TreeReplaceEDCDataWith(edfile->GetEDC(), show_warnings, mystr(edfile->GetDataName())+mystr("."));
			// use global EDC
			ElementDataContainer rootedEdFile;
			rootedEdFile.Add(*edfile);

			edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", &CEDCParser::GetNewEDCName, show_warnings);
		}
	}

	// extra treatment for PostProcOptions and GraphicsOptions:
	// check for PostProcOptions (they have been deleted)
	if (edfile = edc_file.TreeFind("PostProcOptions"))
	{
		ElementDataContainer rootedEdFile;
		rootedEdFile.Add(*edfile);
		edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", &CEDCParser::GetNewEDCName, show_warnings);
	}

	// check for GraphicsOptions (they have been deleted)
	if (edfile = edc_file.TreeFind("GraphicsOptions"))
	{
		ElementDataContainer rootedEdFile;
		rootedEdFile.Add(*edfile);
		edc_default->TreeReplaceEDCDataWithReplacingOldNames(&rootedEdFile, "HotintOptions", &CEDCParser::GetNewEDCName, show_warnings);
	}

	// overwrite default options by user file data 
	EDC2SolverOptions(edc_default);
	EDC2Options(edc_default);      

	UO() << "done.\n";
	UO() << "--------------------------------\n";

	return rv; //1==success, 0==error
}

// use add model data after command read model data; assumption: model data container is not empty!!!
//$ DR 2013-07-25: assumption is not necessary anymore.
//$ AD 2013-10-22: uses ParseAndExecuteString to instantly write to edc_modeldata
int MultiBodySystem::AddModelData(mystr filename)
{
	UO(UO_LVL_multsim) << "read " << filename << "\n"; //$ DR 2015-05-05 removed "UO_LVL_0"

	ElementDataContainer edc_file;
	
	// store file name for include command with relative paths
	ElementData ed;
	ed.SetText(filename, "last_file_name");
	if(edc_modeldata != 0)
	{
		edc_modeldata->TreeSet("",ed);
	}
	else
	{
		SetModelDataContainer(edc_file); //store empty edc to avoid crashes (due to null-pointer exeptions) 
	}

//$ AD 2013-10-22: read the string and parse it directly into the modeldatacontainer
	mystr str;
	int rv = File2Str(filename, str);
	if(!rv)return rv; 
	
	EDCParser().ParseAndExecuteString(str, *edc_modeldata);

	//$ DR 2014-10-24 new implementation
	ElementDataContainer edc_options, edc_tmp;
	Options2EDC(&edc_options);
	SolverOptions2EDC(&edc_tmp);
	edc_options.TreeReplaceEDCDataWith(&edc_tmp); // now all old options in edc_options
	edc_options.TreeReplaceEDCDataWithReplacingOldNames(edc_modeldata,"HotintOptions", CEDCParser::GetNewEDCName, 0);	// replace the options with edc_modeldata
	EDC2Options(&edc_options);
	EDC2SolverOptions(&edc_options);

	OpenLogFile(0);

	UO() << "done.\n";
	UO() << "--------------------------------\n";

	return 1; //success
}

//// writes all solversettings from the modeldata-edc into the solversettings-edc //$ DR 2014-10-24 not necessary anymore
//int MultiBodySystem::UpdateSettingsWithModelData(ElementDataContainer& edc_modeldata)
//{
//	// replace solveroptions & other settings
//	bool show_warnings = GetOptions()->LoggingOptions()->OutputLevel() >= UO_LVL_dbg1;
//	ElementData* eddef;
//	ElementData* edfile;
//
//	// loop over all default main branches
//	for (int i=1; i <= GetMBS_EDC_Options()->Length(); i++) 
//	{
//		// get i-th main branch
//		//eddef = edc_default.GetPtr(i);		
//		eddef = GetMBS_EDC_Options()->GetPtr(i);
//		edfile = edc_modeldata.TreeFind(eddef->GetDataName());
//		// replace the default settings by the user's file settings
//		if (edfile != 0 && eddef->IsEDC() && edfile->IsEDC())
//		{
//			if(GetMBS_EDC_Options()->TreeFind(edfile->GetDataName()))
//			{				
//				UO(UO_LVL_ext) << "replacing hotint option: " << edfile->GetDataName() << "\n";
//				// use file parameters (when existing) instead of default variables
//				eddef->GetEDC()->TreeReplaceEDCDataWith(edfile->GetEDC(), show_warnings, mystr(edfile->GetDataName())+mystr("."));
//			}
//		}
//	}
//	// overwrite default options by user file data 
//	EDC2SolverOptions(GetMBS_EDC_Options());
//	EDC2Options(GetMBS_EDC_Options());       // copy EDC to i/d/toptions	
//	return 1;
//}

// remove a variable or branch from the edc by name - this is used to prevent ComputationSteps to be written to file and to be remembered from a previous model		
int MultiBodySystem::RemoveVariableFromEDCTree(ElementDataContainer& edc, mystr& varname, int warning)
{
	mystr treename, elementname;
	edc.SplitIntoTreeAndElementName(varname, treename, elementname);
	ElementData* ed_branch = edc.TreeFind(varname);
	
	if (ed_branch != 0) // variable exists
	{
		ElementDataContainer* edc_nesting;
		if(treename.Length())
		{
			edc_nesting = edc.TreeFind(treename)->GetEDC();
		}
		else
		{
			edc_nesting = &edc;
		}

		if (warning)
		{
			if( ed_branch -> IsEDC() ) UO(UO_LVL_ext) << "removing sub-edc " + varname + " from edc\n";
			else UO(UO_LVL_ext) << "removing entry " + varname + "from edc \n";
		}

		edc_nesting->Delete(edc_nesting->Find(elementname));
		return 1;
	}
	return 0;
}

int MultiBodySystem::AddFileSensor(const char* filename, const mystr sensor_name, const int ncolumns, const int colTime, const int colSignal, const int interpolation, const int nrOfHeaderLines)
{
	/*
	MathFunction mf;
	if(mf.SetPiecewiseFromFile(filename, ncolumns, colTime, colSignal, interpolation, nrOfHeaderLines)) // interpolation: 0...steps, 1...piecewise linear
	{
		InstantMessageText(mystr("Error1: File " + mystr(filename) + " not found!"));
		return 1; // error
	}
	Vector2D posDraw(0.,0.); // drawing position
	Vector3D drawDim = Vector3D(0.2,0.2,0.); // draw dimension of control element
	IOTime lt_time(this);
	lt_time.SetElementName(sensor_name+"-IOTime");
	lt_time.SetRefPos2D(Vector2D(0.,3.0)+posDraw);//define reference position for drawing
	lt_time.SetDrawDim(drawDim);     //define size of rectangle for drawing (B, H, dummy)
	int nLt_time = AddElement(&lt_time);

	IOMathFunction iom_ty(this);
	iom_ty.SetIOMathFunction(mf);
	iom_ty.SetRefPos2D(Vector2D(0.5,3.0)+posDraw);
	iom_ty.SetDrawDim(drawDim);
	iom_ty.SetElementName(sensor_name+"-IOMathFcn");						//Change name of element from default name				
	iom_ty.AddInput(nLt_time, IOInputTypeElement, 1);  // connection(nLt_time,  typ=io, output 1)
	int nIo_ty = AddElement(&iom_ty);

	MBSSensor sens_mes(this, TMBSSensor(TSOutputSensor), nIo_ty, 1); //measure output 1
	sens_mes.SetSensorName(sensor_name);
	AddSensor(&sens_mes);
	*/
	return 0; //success
}
int MultiBodySystem::AddFileSensor2(const char* filename, const mystr sensor_name, const int colTime, const int colSignal, const int interpolation, mystr comment)
{
	/*
	MBSSensor sens_mes(this, TMBSSensor(TSFile), colTime, colSignal); //measure output 1
	if(sens_mes.SetTSFile(filename, colTime, colSignal, interpolation, comment)) // set file name
	{
		return 1; // error, don't add sensor
	}
	sens_mes.SetSensorName(sensor_name);
	AddSensor(&sens_mes);	
	*/
	return 0; //success
}

void MultiBodySystem::SetView3DCenterOffset(int type, int idx)
{
	switch (type)
	{
	case DTElement:        SetView3DCenterOffset(GetElementPtr(idx)->GetRefPosD()); break;
	case DTConstraint:     SetView3DCenterOffset(GetElementPtr(idx)->GetRefPosD()); break;
	case DTNode:           SetView3DCenterOffset(GetNodePtr(idx)->GetPosD()); break;
	default:               SetView3DCenterOffset(Vector3D(0));  break; 
	}
}

void MultiBodySystem::SetView3DCenterOffset(const Vector3D& pos)
{
	GetOptions()->ViewingOptions()->View3DCenterOffsetXpos() = pos.X();
	GetOptions()->ViewingOptions()->View3DCenterOffsetYpos() = pos.Y();
	GetOptions()->ViewingOptions()->View3DCenterOffsetZpos() = pos.Z();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//void MultiBodySystem::SetOptions2EDCOptions() //$ DR 2014-10-24 not necessary anymore
//{
//	ElementDataContainer* edc = GetMBS_EDC_Options();
//	ElementDataContainer edcnew;
//	Options2EDC(&edcnew);                //copy options to edc:
//	edc->TreeReplaceEDCDataWith(&edcnew);//replace edcnew data in edc:
//}

//void MultiBodySystem::SetEDCOptions2Options() //$ DR 2014-10-24 not necessary anymore
//{
//	ElementDataContainer* edc = GetMBS_EDC_Options();
//	EDC2Options(edc);
//}

//void MultiBodySystem::SetComputationSolverOptions() //$ DR 2014-11-21 removed completely
//{
//	ElementDataContainer* edc = GetMBS_EDC_Options();
//	EDC2SolverOptions(edc);
//}

//void MultiBodySystem::SetSolverDialogOptions() //$ DR 2014-10-24 not necessary anymore
//{
//	//update computation options with solver settings:
//
//	ElementDataContainer* edc = GetMBS_EDC_Options();
//	ElementDataContainer edcnew;
//
//	//copy mbssolset to edc:
//	SolverOptions2EDC(&edcnew);
//	//replace edcnew data in edc:
//	edc->TreeReplaceEDCDataWith(&edcnew);
//}


