//#**************************************************************
//#
//# filename:             performcomp.cpp
//# 
//# author:               Gerstmayr Johannes
//#
//# generated:						July 2004
//# description:          Driver and model for timeintegration
//#                       Model of a rigid arm and a hydraulic zylinder
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#include "MBS.h"
#include "myfile.h"
#include "sensitivity.h"
#include "optimization.h"
#include "element.h"
#include "sensors.h"

////#include "..\SuperLU\slu_ddefs.h"
//#include "..\SuperLU\SuperLUmain.h"
//
////ofstream sol("..\\..\\output\\sol.txt"); //
//ofstream solpar2("..\\output\\par_rest.txt");

void MultiBodySystem::PerformComputation()
{
	simulationStatus.SetStatusFlag(TSimulationRunning); //$ MaSch 2015-07-02

	//check if parameter variation or compute eigenmodes:
	if(ptr2PerformComputation_Function != 0)
	{
		UO(UO_LVL_ext) << "\n****************************\nDo user-defined computation!\n****************************\n\n";
		CallPtr2PerformComputation_Function(this);
	}
	//else if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.activate"))
	else if (GetSolSet().parvar_activate)	//$ DR 2014-10-23 improve speed
	{
		UO(UO_LVL_ext) << "\n****************************\nDo parameter variation\n****************************\n\n";
		PerformParameterVariation();
	}
	//else if(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Optimization.activate"))
	else if(GetSolSet().paropt_activate)	//$ DR 2014-10-23 improve speed
	{
		UO(UO_LVL_ext) << "\n*****************************************\nDo parameter optimization\n*****************************************\n\n";
		Optimization(this).PerformOptimization();		//$ YV 2012-11-28
	}
	//else if(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Sensitivity.activate"))
	else if(GetSolSet().par_sens_activate) //$ DR 2014-10-23 improve speed
	{
		UO(UO_LVL_ext) << "\n*****************************************\nDo sensitivity analysis\n*****************************************\n\n";		
		Sensitivity(this).PerformSensitivityAnalysis();
	}
	else 
	{
		//normal mode: do a single computation
		PerformSingleStaticDynamicComputation();
	}

	simulationStatus.AddSetStatusFlag(TSimulationProcessFinished); //$ MaSch 2013-08-19
	//ComputationFinished(); //$ DR 2014-10-13 removed because already called in the functions above

	//if option is set==>close application after computation!
	//if (GetMBS_EDC_Options()->TreeGetInt("GeneralOptions.Application.close_application_when_finished")) // GetIOption(220))
	if (GetOptions()->GeneralOptions()->ApplicationCloseApplicationWhenFinished()) // GetIOption(220)) //$ DR 2014-10-24 improve speed
	{
		// AD 2018-02-15: explicitly set an exit code in some cases ( finished computation but with some error )
		int exitcode = 0;
		if(simulationStatus.GetStatusFlag(TSimulationProcessFinished) && simulationStatus.GetStatusFlag(TSimulationStoppedDueError)) { exitcode = 64+8; } // case: did not converge	
		if(simulationStatus.GetStatusFlag(TSimulationProcessFinished) && simulationStatus.GetStatusFlag(TSimulationStoppedByUser)) { exitcode = 64+16; } // case: stopped by user	

		UO().CallWCDriverFunction(5,0,exitcode); //quit
	}
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void MultiBodySystem::PerformSingleStaticDynamicComputation()
{
	int rv = PreComputationOperations();
	if (!rv) return;

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	rv = ComputeSystem(); //Main solution procedure !!! (time integration/static solver)
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	if (!rv) simulationStatus.SetStatusFlag(TSimulationStoppedDueError); //$ MS + RL 2012-2-29: //$ MaSch 2013-08-19 
	
	//OpenFiles() is done in following procedure:
	PostComputationOperations(rv);

	//WriteSolutionVector(); //$ PG 2017-5-9: using following method instead: virtual int SaveSolutionVector(const mystr& filename)
	if (GetSolSet().storesolution_state && !StopCalculation())
	{
		SaveSolutionVector(GetSensorOutputPath()+GetSolSet().storesolution_statename);
	}

	DoFinalSensorEvaluations();

	CloseFiles();

	if (GetOptions()->GeneralOptions()->ApplicationCaptureFinalFrame())
	{
		mystr path = GetSensorOutputPath() + mystr("final_frame.bmp");
		uo.pUI->CreateScreenshot(path.c_str());
	}

	pCFB->FinishedComputation();
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int MultiBodySystem::ComputeSystem()
{
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Eigensolver.do_eigenmode_computation"))
	if (GetSolSet().do_eigenmode_comp) //$ DR 2014-10-23 improve speed
	{
		UO(UO_LVL_ext) << "\n****************************\nDo eigenvalue/eigenmode computation\n****************************\n\n";
		ComputeEigenmodes();
		return 1;
	}
	else
	{
		return Integrate();
	}
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//$!PG 2011-2-22:[  new local function in order to handle more cases of parameter variation (also varstep<0 and varstepfact<1) 	
bool MultiBodySystem::DoParameterVariationComputation(bool const isgeometric, const double par, const double varstep, const double varstepfact, const double varend) const
{
	double some_space = 1e-8;
	if (isgeometric)
	{
		if ((varstepfact >= 1) && (par <= varend*(1.+some_space)))
			return true;
		else if ((varstepfact < 1) && (par >= varend*(1.-some_space)))
			return true;
	}
	else
	{
		if ((varstep >= 0) && (par <= varend*(1.+some_space)))
			return true;
		else if ((varstep < 0) && (par >= varend*(1.-some_space)))
			return true;
	}

	return false;
}
//$!PG 2011-2-22:]

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void MultiBodySystem::PerformParameterVariation()
{
	//###################################################################################################################
	//
	//     do parametric variation
	//
	//###################################################################################################################
	
	clearSolpar = 1;

	//double varstart = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.start_value"); //
	//double varend = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.end_value"); //=11;
	double varstart = GetSolSet().parvar_start_value;  //$ DR 2014-10-23 improve speed
	double varend = GetSolSet().parvar_end_value; //$ DR 2014-10-23 improve speed
	
	double varstep = 0; //!!!!!!!!!!!!!!!!!!
	double varstepfact = 1; //1 = incremental

	//$!PG 2011-2-22:[  new passage in order to handle more cases of parameter variation (also varstep<0 and varstepfact<1) 
	bool isgeometric = false;
	//$!PG 2011-2-22:]

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.geometric") == 0) 
	if (GetSolSet().parvar_geometric == 0) 	//$ DR 2014-10-23 improve speed
	{ //arithmetic
		//varstep = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.arithmetic_step");
		varstep = GetSolSet().parvar_step_lin; //$ DR 2014-10-23 improve speed
	}
	else //geometric 
	{
		//$!PG 2011-2-22:[  new line in order to handle more cases of parameter variation (also varstep<0 and varstepfact<1) 
		isgeometric = true;
		//$!PG 2011-2-22:]

		varstep = 0;
		//varstepfact = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.geometric_step");
		varstepfact = GetSolSet().parvar_step_geom; //$ DR 2014-10-23 improve speed
	}

	//###################################################################################################################
	//second variation

	//double varstart2 = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.Var2.start_value"); //
	//double varend2 = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.Var2.end_value"); //=11;
	double varstart2 = GetSolSet().parvar2_start_value; //$ DR 2014-10-23 improve speed
	double varend2 = GetSolSet().parvar2_end_value; //$ DR 2014-10-23 improve speed
	

	double varstep2 = 0; //!!!!!!!!!!!!!!!!!!
	double varstepfact2 = 1; //1 = incremental

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.Var2.geometric") == 0) 
	if (GetSolSet().parvar2_geometric == 0) 	 //$ DR 2014-10-23 improve speed
	{ //arithmetic
		//varstep2 = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.Var2.arithmetic_step");
		varstep2 = GetSolSet().parvar2_step_lin;  //$ DR 2014-10-23 improve speed
	}
	else //geometric 
	{
		varstep2 = 0;
		//varstepfact2 = GetMBS_EDC_Options()->TreeGetDouble("SolverOptions.ParameterVariation.Var2.geometric_step");
		varstepfact2 = GetSolSet().parvar2_step_geom; //$ DR 2014-10-23 improve speed
	}

	varparameter2 = varstart2;
	//int var2_activate = GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.Var2.activate");
	int var2_activate = GetSolSet().parvar2_activate; //$ DR 2014-10-23 improve speed

	if (!var2_activate)
	{
		varstart2 = 0;
		varparameter2 = 0;
		varend2 = 0;
		varstep2 = 1;
		varstepfact2 = 1;
	}

	while (DoParameterVariationComputation(isgeometric, varparameter2, varstep2, varstepfact2, varend2) && !StopCalculation())
	{
		if (var2_activate)
		{
			//mystr varname2 = GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.Var2.MBS_EDC_variable_name");
			mystr varname2 = GetSolSet().parvar2_EDCvarname; //$ DR 2014-10-23 improve speed
			// RL 2011-02:[ TODO: function of class PerformComputation should be used --> no duplicate code!
			if (varname2 != mystr(""))
			{				
				//if (GetMBS_EDC_Options()->TreeFind(varname2))	// set new value in options
				//{
				//	GetMBS_EDC_Options()->TreeFind(varname2)->SetLocked(0);	//$ DR 2013-04-09: parameter variation for script language
				//	GetMBS_EDC_Options()->TreeSetDouble(varname2, varparameter2);
				//	GetMBS_EDC_Options()->TreeFind(varname2)->SetLocked(1); //$ DR 2013-04-09: parameter variation for script language
				if (GetModelDataContainer() != 0 && GetModelDataContainer()->TreeFind(varname2)) // set new value in model data
				{
					GetModelDataContainer()->TreeFind(varname2)->SetLocked(0);	//$ DR 2013-04-09: parameter variation for script language
					GetModelDataContainer()->TreeSetDouble(varname2, varparameter2);		
					GetModelDataContainer()->TreeFind(varname2)->SetLocked(1); //$ DR 2013-04-09: parameter variation for script language
				} 
				else	//$ DR 2014-10-29 new implementation: variation of options
				{
					ElementDataContainer edc_tmp, edc_opt;
					Options2EDC(&edc_opt);
					SolverOptions2EDC(&edc_tmp);
					edc_opt.TreeReplaceEDCDataWith(&edc_tmp);	// options and solver options in edc_opt
					if(edc_opt.TreeFind(varname2)) // replace options
					{
						edc_opt.TreeSetDouble(varname2, varparameter2);
						EDC2Options(&edc_opt);
						EDC2SolverOptions(&edc_opt);
					}
					else
					{
						UO() << "*********************\nWarning: can not access variable '" << varname2 << "' for parameter variation!\n*********************\n";
					}
				}
			}
			// RL 2011-02:] TODO: function of class PerformComputation should be used --> no duplicate code!

			UO(UO_LVL_all) << "\n\n\n\n********************\n********************\n********************\n********************\n";
			UO(UO_LVL_all) << "\n********************\nStart simulation with varparameter 2\n" << varname2 << " = " << varparameter2 
				<< "\n********************\n********************\n********************\n********************\n";
		}

		double par = varstart;


		//while (par <= varend*(1.+1e-8) && !StopCalculation())
		//$!PG 2011-2-22: why only for (par <= varend)? parameter variation should also handle the case (varstepfact < 1)
		//$!PG 2011-2-22: Instead I implemented MultiBodySystem::DoParameterVariationComputation(...)
		while (DoParameterVariationComputation(isgeometric, par, varstep, varstepfact, varend) && !StopCalculation())
		{
			varparameter = par;	//par and varparameter is effectively the same

			//GetMBS_EDC_Options()->TreeSetDouble("SolverOptions.start_time", 0.);
			GetSolSet().starttime = 0.; //$ DR 2014-10-23 improve speed

			//mystr varname = GetMBS_EDC_Options()->TreeGetString("SolverOptions.ParameterVariation.MBS_EDC_variable_name");
			mystr varname = GetSolSet().parvar_EDCvarname; //$ DR 2014-10-23 improve speed
			if (varname != mystr(""))
			{
				//if (GetMBS_EDC_Options()->TreeFind(varname)) // set new value in options
				//{
				//	GetMBS_EDC_Options()->TreeFind(varname)->SetLocked(0); //$ DR 2013-04-09: parameter variation for script language
				//	GetMBS_EDC_Options()->TreeSetDouble(varname, varparameter);
				//	GetMBS_EDC_Options()->TreeFind(varname)->SetLocked(1); //$ DR 2013-04-09: parameter variation for script language
				//}
				if (GetModelDataContainer() != 0 && GetModelDataContainer()->TreeFind(varname)) // set new value in model data
				{
					GetModelDataContainer()->TreeFind(varname)->SetLocked(0); //$ DR 2013-04-09: parameter variation for script language
					GetModelDataContainer()->TreeSetDouble(varname, varparameter);
					GetModelDataContainer()->TreeFind(varname)->SetLocked(1); //$ DR 2013-04-09: parameter variation for script language
				} 
				else
				{
					ElementDataContainer edc_tmp, edc_opt;
					Options2EDC(&edc_opt);
					SolverOptions2EDC(&edc_tmp);
					edc_opt.TreeReplaceEDCDataWith(&edc_tmp);	// options and solver options in edc_opt
					if(edc_opt.TreeFind(varname)) // replace options
					{
						edc_opt.TreeSetDouble(varname, varparameter);
						EDC2Options(&edc_opt);
						EDC2SolverOptions(&edc_opt);
					}
					else
					{
						UO() << "*********************\nWarning: can not access variable '" << varname << "' for parameter variation!\n*********************\n";
					}
				}
			}


			UO(UO_LVL_all) << "\n********************\nStart simulation with varparameter\n" << varname << " = " << varparameter << "\n********************\n";

			//********************************************************************
			//********************************************************************
			//repeatedly call computation:
			RepeatedlyPerformComputation();

			//********************************************************************
			//next iteration:
			par += varstep;
			par *= varstepfact;
		}
		//********************************************************************
		//next iteration, varparameter2:
		varparameter2 += varstep2;
		varparameter2 *= varstepfact2;
	}

	pCFB->FinishedComputation();

	//if (GetMBS_EDC_Options()->TreeGetInt("GeneralOptions.Application.close_application_when_finished")) // GetIOption(220))
	if (GetOptions()->GeneralOptions()->ApplicationCloseApplicationWhenFinished()) // GetIOption(220))	//$ DR 2014-10-23 improve speed
	{
		UO().CallWCDriverFunction(5); //quit
	}

}

double MultiBodySystem::EvalSensorCostFunctionVal() //$ RL 2011-02: evaluate cost function (e.g. for parameter optimization)
{
	double cost_function_val = 0.;
	if (GetTIFinished() == -1)//$ MS 2011-9-8: if time-integration failed  //$ YV -> MS: 2011?
	{
		cost_function_val += 1.e10;
	}
	else
	{
		for (int i=1; i <= NSensors(); i++)
		{
			if(GetSensor(i).HasSensorProcessingEvaluationData())
			{				
				TArray<double> signalProcessingEvaluationData = GetSensor(i).GetSignalProcessingEvaluationData();
				for(int i = 1; i <= signalProcessingEvaluationData.Length(); i++)
					cost_function_val += signalProcessingEvaluationData(i);

				if(GetSolSet().paropt_sensors != 0)
				{
					UO(UO_LVL_warn) << "Depriciated data from sensor-processing is used as cost function.\n Change option to \"SolverOptions.Optimization.sensors=0\".\n";
				}
			}
			// for script language, the end value of the sensor is used
			else if(GetSolSet().paropt_sensors == i)
			{
				GetSensor(i).Evaluate(GetTime()); //$ SW 2015-4-8: evaluate since the sensor might not have been evaluated (in case of a eigenvalue optimization e.g.)
				cost_function_val += GetSensor(i).GetLastValue();
			}
		}
	}
	return cost_function_val;
}

bool MultiBodySystem::IsEmptyFile(std::ofstream& ofs) const
{
	std::streampos pos = ofs.tellp();  // store current location
	ofs.seekp(0, std::ios_base::end);  // go to end
	bool empty = (ofs.tellp() == 0);   // check size == 0 ?
	ofs.seekp(pos);                    // restore location
	
	return empty;
}

int MultiBodySystem::RepeatedlyPerformComputation(int flags)
{

	//Model must be always initialized! ==> always isfirst
	int isfirst = isfirstcomputation;
	Initialize();
	isfirstcomputation = isfirst; //because is overwritten in PreComputationOperations / Initialize

	//OpenFiles() is done in following procedure:
	int rv = PreComputationOperations(1+2); //1..always set initial conditions; 2..always append to parameter solution file
	if(!rv)
	{
		UO(UO_LVL_err).InstantMessageText("Error in pre-computation operations, simulation stopped!\n");
		return rv; //$ RL 2011-8-5: if system is inconsistent, computation must not be started,
	}

	isfirstcomputation = 0; //append to files after this point!
	rv = ComputeSystem();



	if (!rv) simulationStatus.SetStatusFlag(TSimulationStoppedDueError); //$ MS + RL 2012-2-29: //$ MaSch 2013-08-19 
	PostComputationOperations(rv);
	if (!(flags&1)) //decide wheter to write solpar file directly
	{
		DoFinalSensorEvaluations();
	}
	CloseFiles();

	return rv;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MultiBodySystem::PreComputationOperations(int flags)
{
	if (isfirstcomputation)
	{
		//isfirstcomputation = 0;
		
		UO(UO_LVL_ext) << "solution output file=" << GetSolSet().sol_filename << "\n";
		//UO(UO_LVL_ext) << "output directory = " << GetOptions()->GeneralOptions()->PathsSensorOutputPath() << "\n";		//$ DR 2014-10-23 improve speed
		UO(UO_LVL_ext) << "output directory = " << GetSensorOutputPath() << "\n";	//$ DR 2015-02-12 absolute path
		
		//$ PG 2012-4-19:[ for saving solution data to files
		//if (solset.sol_data_to_files)
		if (GetOptions()->ViewingOptions()->DataManagerStoreDataToFiles())	//$ DR 2014-10-28 renamed option
		{
			//mystr sol_data_dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath()+mystr("solution_data\\"); //$ DR 2014-10-23 improve speed
			mystr sol_data_dir = GetSensorOutputPath()+mystr("solution_data\\"); //$ DR 2014-10-23 improve speed
			mystr sol_data_dir_name("solution data directory"); 
			int rv = UOfull().pUI->CreateMissingDirectory(sol_data_dir.c_str());
			switch (rv)
			{
			case 0:				// path to directory is wrong
				UO(UO_LVL_err) << mystr("ERROR: ") + sol_data_dir_name + mystr(" cannot be created, since path ") + sol_data_dir + mystr(" is wrong!\n"); 
				return 0;
			case 1:				// directory existed
				UO(UO_LVL_ext) << sol_data_dir_name + mystr(" = ") + sol_data_dir + mystr("\n");
				break;
			case 2:				// directory created
				UO(UO_LVL_ext) << sol_data_dir_name + mystr(" = ") + sol_data_dir + mystr(" (created) \n");
				break;
			default:
				assert(0);
				break;
			}
		}
		//$ PG 2012-4-19:] for saving solution data to files 


		//UO(UO_LVL_ext) << "log file = " << GetSolSet().log_filename << "\n";  //$ PG 2012-2-3: modified (see next line)
		UO(UO_LVL_ext) << "log file = " << GetOptions()->LoggingOptions()->DefaultLogFilename() << "\n";

		int appendflag = 0;
		if ((flags&2)) appendflag += 2; //always append to parameter file (e.g. in parameter variation)
		if (clearSolpar) //$ SW 2015-1-21: added flag to clear the parameter file at the beginning of a parameter variation
		{
			appendflag |= 4;
			clearSolpar = 0;
		}
		OpenFiles(appendflag);

		SetInitialConditions();

	}
	else
	{
		//the following option is used for continuation of a computation (PAUSE): ==> NEEDED???
		//if (GetMBS_EDC_Options()->TreeGetBool("SolverOptions.Timeint.reset_after_simulation") || (flags&1) != 0)
		if (GetSolSet().reset_after_simulation || (flags&1) != 0) //$ DR 2014-10-23 improve speed
		{
			SetInitialConditions(); //this is set in Initialize(), for the case that initial con
		}

		//decide whether to append solution to existing file, or to replace it:
		int appendflag = 1;

		if (GetSolSet().sol_replace_files) appendflag = 0;
		if ((flags&2)) appendflag += 2; //always append to parameter file (e.g. in parameter variation)
		if (clearSolpar)//$ SW 2015-1-21: added flag to clear the parameter file at the beginning of a parameter variation
		{
			appendflag |= 4;
			clearSolpar = 0;
		}

		OpenFiles(appendflag);
		//PG: OpenLogFile(1);
	}

	//$ DR 2012-11-08 moved from "if (isfirstcomputation)-branch" to here. Consistency should be checked always when there are changes!
	int modified = UOfull().pUI->GetModelModified();
	if(isfirstcomputation || modified)
	{
		isinconsistent = CheckSystemConsistency();
		if (isinconsistent) 
		{
			UO(UO_LVL_err).InstantMessageText("Error: Due to an inconsistent system, all output files are closed!\n"); //$ RL 2012-5-22: user should now, what happens. 
			CloseFiles();
			return 0;
		}
	}
	isfirstcomputation = 0;

	//SetMaxIndex(GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Timeint.max_index"));
	SetMaxIndex(GetSolSet().maxindex); //$ DR 2014-10-23 improve speed
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.do_static_computation")) SetMaxIndex(3); //dostaticcomputation
	if (GetSolSet().dostaticcomputation) SetMaxIndex(3); //dostaticcomputation //$ DR 2014-10-23 improve speed

	//simulationStatus.SetStatusFlag(TSimulationRunning); //$ MaSch 2013-11-20  //$ MaSch 2015-07-02: moved to PerformComputation()  

	TMResetTimer(); //reset timers for evaluation of function computational time
	TMStartTimer(1); //timer for total computational time, only solution/integration!!!
	return 1;
}



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MultiBodySystem::PostComputationOperations(int integrate_rv)
{
	solfile << flush;
	solparfile << flush;

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //$ RL 2011-7-14: FFT code
	//$ RL 2011-7-14:[ FFT-sensor files are created
	for(int i=1; i<=sensors.Length();i++)
	{
		if(GetSensor(i).GetOwnOutputFile())
		{
			(*GetSensor(i).GetOwnOutputFile()) << flush;
		}

		// now we perform post-computation evaluation of the sensor signal

		if(GetSensor(i).NeedsFileForPostComputationProcessing())
		{
			// we need to treat the file with the results here
			ofstream post_computation_output_file;
			//mystr dir = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path");
			//mystr dir = GetOptions()->GeneralOptions()->PathsSensorOutputPath();	//$ DR 2014-10-24 improve speed
			mystr dir = GetSensorOutputPath(); //$ DR 2015-02-12 absolute path
			post_computation_output_file.open((dir+mystr("S")+mystr(i)+mystr("-")+GetSensor(i).GetSensorName()+mystr("-post_comp.txt")).c_str(), ios::out);
			GetSensor(i).ApplyPostComputationSensorProcessing(&post_computation_output_file);
			post_computation_output_file.close();
		}
		else
			GetSensor(i).ApplyPostComputationSensorProcessing(NULL);		// postcomputation processing is performed without an output file stream
	}
	//$ RL 2011-7-14:] FFT-sensor files are created
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	

	TMStopTimer(1);


	//$ PG 2019-6-24: moved call of WriteSolDataInfo() to void TimeInt::TIDrawAndStore()
	//WriteSolDataInfo();


	//$ AD AD 2017-02-13: write special output
	WriteSpecialAfterFinished();


	if (!integrate_rv)
	{
		UO(UO_LVL_err) << "*** time integration not finished! ***" << "\n";
		simulationStatus.SetStatusFlag(TSimulationStoppedDueError); //$ MS + RL 2012-2-29: //$ MaSch 2013-08-19 
	}
	else
	{
		if (!StopCalculation())
		{
			if (simulationStatus.GetStatusFlag(TSimulationStoppedByElement))
			{
				UO(UO_LVL_err) << "computation stopped by element\n";
			}
			else
			{
				UO(UO_LVL_err) << "computation finished\n";
				simulationStatus.SetStatusFlag(TSimulationEndedRegularly); //$ MS + RL 2012-2-29: //$ MaSch 2013-08-19
			}
		}
		else
		{
			UO(UO_LVL_err) << "computation stopped by user\n";
			simulationStatus.SetStatusFlag(TSimulationStoppedByUser); //$ MS + RL 2012-2-29: //$ MaSch 2013-08-19
		}
	}


#if 1 //$ PG 2020-04-23: faster output
	mystr text("---------------------------------------------------------------\nCOMPUTATION SUMMARY:\n");
	text += mystr("computational time = ") + mystr(TIcomptime) + mystr(" seconds\n")
		+ mystr("rejected steps = ") + mystr(TIrejectedsteps) + mystr("\n");
	UO(UO_LVL_err) << text;
	if (TIit != 0 && TIcomptime!=0)
	{
		text = mystr("timesteps per second = ") + mystr((double)TIit/TIcomptime)  + mystr(" steps/s\n");
		UO(UO_LVL_all) << text;
	}
	text = mystr("number of timesteps = ") + mystr(GetTimeStep()) + mystr("\n")
		+ mystr("number of jacobians = ") + mystr(GetJacCount()) + mystr("\n");
	// + mystr(FullNewtonCnt()) + mystr(" Full Jacobians computed\n");
	UO(UO_LVL_err) << text;
	UO(UO_LVL_dbg1) << mystr("changes in the timestep = ") + mystr(GetStepChanges()) + mystr("\n");
	text = mystr("total newton iterations = ") + mystr(GetNewtonItSum()) + mystr("\n")
		+ mystr("full evaluations of EvalF2 = ") + mystr(log.evalfcnt) + mystr("\n")
		+ mystr("partial Jacobi evaluations of EvalF2 = ") + mystr(log.evalf_jaccnt) + mystr("\n")
#ifdef gencnt
		+ mystr("allocated matrices = ") + mystr(GenMatCnt()) + mystr("\n")
		+ mystr("allocated vectors = ") + mystr(GenVecCnt()) + mystr("\n")
#endif
		+ mystr("---------------------------------------------------------------\n");
	UO(UO_LVL_err) << text;

#else

	UO(UO_LVL_err) << "---------------------------------------------------------------\nCOMPUTATION SUMMARY:\n";
	UO(UO_LVL_err) << "computational time = " << TIcomptime << " seconds" << ENDL;
	UO(UO_LVL_err) << "rejected steps = " << TIrejectedsteps << ENDL;
	if (TIit != 0 && TIcomptime!=0) UO(UO_LVL_all) << "timesteps per second = " << (double)TIit/TIcomptime << " steps/s" << ENDL;

	UO(UO_LVL_err) << "number of timesteps = "  << GetTIit()<< "\n";
	UO(UO_LVL_err) << "number of jacobians = " << GetJacCount() << "\n";
	//UO(UO_LVL_ext) << FullNewtonCnt() << " Full Jacobians computed." << "\n";
	UO(UO_LVL_dbg1) << "changes in the timestep = " << GetStepChanges() << "\n";
	UO(UO_LVL_err) << "total newton iterations = " << GetNewtonItSum() << "\n";
	UO(UO_LVL_err) << "full evaluations of EvalF2 = " << log.evalfcnt << "\n";
	UO(UO_LVL_err) << "partial Jacobi evaluations of EvalF2 = " << log.evalf_jaccnt << "\n";
#ifdef gencnt
	UO(UO_LVL_err) << "allocated matrices = " << GenMatCnt() << "\n";
	UO(UO_LVL_err) << "allocated vectors = " << GenVecCnt() << "\n";
#endif
	UO(UO_LVL_err) << "---------------------------------------------------------------\n";

#endif

	//if (StopCalculation()) return; //this can not be earlier - otherwise the timer is not accurate and the output is not done on manual stop

	ComputationFinished();
}

void MultiBodySystem::DoFinalSensorEvaluations() //do final sensor evaluations (min/max/diff, etc.) for sensors after computation
{
	if (StopCalculation()) return;
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//output sensor data at final step:

	if(DoFinalEigenValueComputation())
	{
		CloseFiles();
		double simulation_time = GetTime(); //$ PG 2013-11-27: memo needed, since time is set to length of eigensystem (-> for vizualizing Eigenmodes with DataManager)
		// sensor of type TSEigenValue exist
		ComputeEigenmodes();

		//$ PG 2013-11-26: perform recalculation of all system sensors after this final eigenvalue computation (at least for all ev-systemsensors this is required)
		for (int i=1; i <= NSensors(); i++)
		{
			Sensor& s = GetSensor(i);
			if(s.IsSensorOnEigensystem())
			{
				s.Evaluate(simulation_time);
			}
		}

		int append = 1;
		OpenFiles(append);
		//PG: OpenLogFile(1);
	}

	int is_sensor_computation = 0;
	for (int i=1; i <= NSensors(); i++)
	{
		// we need to know if the sensor has post evaluation data;
		if(GetSensor(i).HasSensorProcessingEvaluationData())
			is_sensor_computation = 1;
	}

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// write headers
	if (WriteSolParFileHeader() && IsEmptyFile(solparfile))
		solparfile << GetSolParFileHeaderString();
	
	solparfile.precision(16);

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.activate")) {solpar << varparameter << " ";}
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.ParameterVariation.Var2.activate")) {solpar << varparameter2 << " ";}
	if (GetSolSet().parvar_activate) {solparfile << varparameter << " ";} //$ DR 2014-10-23 improve speed
	if (GetSolSet().parvar2_activate) {solparfile << varparameter2 << " ";} //$ DR 2014-10-23 improve speed

	//write size of system (only second order equations):
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_second_order_size")) //$ RL 2011-02: OPTIONAL default: no
	if (GetSolSet().sol_parfile_write_second_order_size) //$ RL 2011-02: OPTIONAL default: no //$ DR 2014-10-23 improve speed
	{
		solparfile << GetSecondOrderSize() << " "; 
	}

                                            
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_CPU_time")) //$ RL 2011-02: OPTIONAL default: no
	if (GetSolSet().sol_parfile_write_CPU_time) //$ RL 2011-02: OPTIONAL default: no	//$ DR 2014-10-23 improve speed
	{
		//write CPU-time:
		solparfile << TMGetTimer(1) << " "; //computational time (CPU-costs)
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Eigensolver.do_eigenmode_computation")) // MSax 2013-07-04 : added eigenvalues, for campbell diagram
	if (GetSolSet().do_eigenmode_comp) // MSax 2013-07-04 : added eigenvalues, for campbell diagram	//$ DR 2014-10-23 improve speed
	{
		for(int i=1; i <= eigval.Length(); i++)
		{
			solparfile << eigval(i)/(2.*MY_PI) << " ";
		}
	}

	//write final sensor values:
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_final_sensor_values")) //$ RL 2011-02: OPTIONAL default: yes
	if (GetSolSet().sol_parfile_write_final_sensor_values) //$ RL 2011-02: OPTIONAL default: yes	//$ DR 2014-10-23 improve speed
	{
		for (int i=1; i <= NSensors(); i++)
		{
			if(GetSensor(i).GetSignalStorageMode() != SSM_None)
			{
				//$ YV 2012-06: the sensors may produce just one scalar value
				solparfile << GetSensor(i).GetLastValue() << " ";
			}
		}
	}

	//write sensor evaluation values:
	if (is_sensor_computation)
	{
		UO(UO_LVL_ext) << "***************************\nSensor computations:\n";
		for (int i=1; i <= NSensors(); i++)
		{
			if(GetSensor(i).HasSensorProcessingEvaluationData())
			{
				UO(UO_LVL_ext) << "sensor " << i << " : [ ";
				TArray<double> & signalProcessingEvaluationData = GetSensor(i).GetSignalProcessingEvaluationData();
				for (int j = 1; j <= signalProcessingEvaluationData.Length(); j++)
				{
					double v = signalProcessingEvaluationData(i);
					solparfile << v << " ";
					UO(UO_LVL_ext) << v << " ";
				}
				UO(UO_LVL_ext) << "]\n";
			}
		}
		UO(UO_LVL_ext) << "***************************\n";
	}

	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.ParameterFile.write_cost_function")) //$ RL 2011-02: OPTIONAL default: yes
	if (GetSolSet().sol_parfile_write_cost_function) //$ RL 2011-02: OPTIONAL default: yes	//$ DR 2014-10-23 improve speed
	{
		// write cost function:
		double costfunc = EvalSensorCostFunctionVal();		
		solparfile << costfunc << " ";
	}

	solparfile << "\n" << flush;

}


//$ PG 2017-5-9:[deleted method due to redundancy with method virtual int SaveSolutionVector(const mystr& filename)
//$ PG 2017-5-9: single call of this method was in performcomp.cpp:100 in method void MultiBodySystem::PerformSingleStaticDynamicComputation()
/*
void MultiBodySystem::WriteSolutionVector() //write solution vector into file if option is set
{
	if (StopCalculation()) return;
	
	//store solution
	//if (GetMBS_EDC_Options()->TreeGetInt("SolverOptions.Solution.store_solution_state"))
	if (GetSolSet().storesolution_state)	//$ DR 2014-10-23 improve speed
	{
		//mystr path = GetMBS_EDC_Options()->TreeGetString("GeneralOptions.Paths.sensor_output_path");
		//mystr path = GetOptions()->GeneralOptions()->PathsSensorOutputPath(); //$ DR 2014-10-23 improve speed
		mystr path = GetSensorOutputPath(); //$ DR 2015-02-12 absolute path

		int strlen = path.Length();
		if (strlen != 0)
		{

			if (path[strlen-1] != '\\' && path[strlen-1] != '/') path += mystr('\\');
			//mystr file = path+GetMBS_EDC_Options()->TreeGetString("SolverOptions.Solution.store_solution_state_name");
			mystr file = path+GetSolSet().storesolution_statename;	//$ DR 2014-10-23 improve speed

			//OLD: mystr file = path+mystr("store_out.txt");
			ofstream storeout(file.c_str()); 

			UO(UO_LVL_ext) << "final solution vector stored in file '" << file << "'\n";

			Vector x = GetSolVector();

			if ((int)storeout.good() == 0 || storeout.fail())
			{
				UO(UO_LVL_err) << "ERROR: could not save solution, check filename!!!\n";
			}
			else
			{

				storeout.precision(19);
				(storeout) << x.Length() << "\n";
				for (int i=1; i <= x.Length(); i++)
				{
					(storeout) << x(i) << " ";
				}
				(storeout) << endl;
				(storeout) << GetTime() << endl;
			}
			storeout.close();
		}
	}

}
*/
//$ PG 2017-5-9:]
