//#**************************************************************
//#
//# filename:             material.h           
//#
//# author:               Peter Gruber, Johannes Gerstmayr, Larissa Aigner
//#
//# generated:            January 2004
//# updated:			  September 2016
//# description:          class contains material laws called during computation by functions of element classes
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#ifndef MATERIAL__H
#define MATERIAL__H



/////////////////////////////////////////////////////////////////////////
// class:   MaterialGeneric 
// purpose: manages functionality which is common to all Material classes
/////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=MaterialGeneric,addelementtypename=MaterialGeneric,pythonclassdescription="Base class for all Materials",pythonAddAs=Material,
//texdescription="MaterialGeneric is base class for all Material classes.",
//texdescriptionComments="$It manages \mathtt{name}$, $\mathtt{color}$, access to MBS and other basic facilities."]
class MaterialGeneric
{
public:

	// enums
	enum InelasticitySolutionMethod {ISM_FixedPoint = 1, ISM_ReturnMapping = 2, ISM_NotSpecified = 29582};
	
	// static dummy variables, introduced by YV to avoid frequent initialization of Matrices etc.
	static const Matrix3D dummyMatrix3D;
	static const Matrix2D dummyMatrix2D;
	static const Vector dummyVector;
	
	// constructors and copying
	MaterialGeneric()	{mbs = 0; DefaultConstructor();}
	MaterialGeneric(MBS* mbsi) {mbs = mbsi; DefaultConstructor();}
	MaterialGeneric(const MaterialGeneric& m) {CopyFrom(m);}
	virtual int CheckConsistency(mystr& errorstr);   	//$ PG 2012-12-19: CheckConsistency added:   return value must be rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute

	// set functions
	void SetMaterialGeneric(double densityI, int planeI, int planestressI, const Vector3D& color);
	virtual void SetMaterialFromEDCData();  //$Set material type and parameters according to user specified edc data
	
	// element data access (manually)
	virtual void GetElementData(ElementDataContainer& edc); 		//fill in all material data
	virtual int SetElementData(ElementDataContainer& edc); //set material data
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) {return 0;}
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) {return 0;}

	// element data access (automatically: function definitions generated by EDC auto converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter
	
	// color, name, number, mbs
	virtual const mystr& GetMaterialName() const {return materialname;}
	virtual mystr& GetMaterialName() {return materialname;}
	virtual Vector3D GetMaterialColor() const {return materialcolor;}
	virtual void SetMaterialColor(Vector3D col) {materialcolor=col;}
	virtual int GetMaterialNumber() const {return materialnumber;}
	virtual void SetMaterialNumber(int mnum) {materialnumber = mnum;}
	virtual const MBS* GetMBS() const {return mbs;}
	virtual MBS* GetMBS() {return mbs;}
	
	// dimensionality
	virtual int IsPlanarMaterial() const {return plane == 1;}
	virtual int IsPlaneStrain() const {return plane_stress == 0;}

	//density:
	virtual const double& Density() const {return density;}
	virtual double& Density() {return density;}

	// Rayleigh damping ( C = dampingM * M + dampingK * K )
	virtual const double& DampingK() const {return dampingK;}
	virtual double& DampingK() {return dampingK;}
	virtual const double& DampingM() const {return dampingM;}
	virtual double& DampingM() {return dampingM;}

	// following functions are pure virtual and must be defined in derived classes (mandatory)
	virtual MaterialGeneric* GetCopy() const = 0;
	virtual mystr GetTypeName() const = 0;
	virtual int IsInelasticMaterial() const = 0;
	virtual int GetInelasticVariablesCount() const = 0;	
	virtual void InitializeInelasticVariables(Vector& inel_variables) const = 0;

	// set inelasticity solution method
	virtual void SetInelasticitySolutionMethod(const InelasticitySolutionMethod& ism);

	// field variables managed by material (internal state variables)
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const;
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const;

	// access inelasticity solution method (and string)
	virtual InelasticitySolutionMethod GetInelasticitySolutionMethod() const { return inelasticity_solution_method; }
	virtual mystr GetInelasticitySolutionMethodString() const { return inelasticity_solution_method_str; }
	
	// convert inelasticity solution method into string and vice versa
	static mystr GetInelasticitySolutionMethodString(InelasticitySolutionMethod ism);
	static InelasticitySolutionMethod GetInelasticitySolutionMethod(const mystr& ism_str, const MBS* mbs);	

	virtual bool HasInitialStrains() const { return false; }
	virtual void GetInitialStrain2D(Matrix2D & initial_strain) const
	{
		mbs->UO(UO_LVL_warn) << "WARNING: GetInitialStrain2D(..) not implemented!\n";
	}
	virtual double GetInitialStrain33() const
	{
		mbs->UO(UO_LVL_warn) << "WARNING: GetInitialStrain33(..) not implemented!\n";
		return 0.;
	}
	virtual void GetInitialStrain(Matrix3D & initial_strain) const
	{
		mbs->UO(UO_LVL_warn) << "WARNING: GetInitialStrain(..) not implemented!\n";
	}

protected:

	virtual void DefaultConstructor();
	virtual void CopyFrom(const MaterialGeneric& s);

	mystr material_type;          //$EDC$[varaccess,EDCvarname="material_type",tooltiptext="type of the material"]
	mystr materialname;						//$EDC$[varaccess,EDCvarname="name",tooltiptext="name of the material",python]
	Vector3D materialcolor;				//$EDC$[varaccess,EDCvarname="color",EDCfolder="Graphics",tooltiptext="material color (as yet used with FEMesh, only)"]
	int plane;										//$EDC$[varaccess,EDCvarname="plane",EDCfolder="Solid",int_bool,tooltiptext="true: 2D, false: 3D",python]
	int plane_stress;							//$EDC$[varaccess,EDCvarname="plane_stress",EDCfolder="Solid",int_bool,tooltiptext="for 2D-Elements only; 1: plane stress, 0: plane strain",python]
	double density;								//$EDC$[varaccess,EDCvarname="density",EDCfolder="Solid",tooltiptext="material density, must not be set for static problems",python]
	double dampingM;              //$EDC$[varaccess,EDCvarname="c_M",EDCfolder="Damping",tooltiptext="used for Rayleigh damping ( C = c_M * M + c_K * K )"]
	double dampingK;              //$EDC$[varaccess,EDCvarname="c_K",EDCfolder="Damping",tooltiptext="used for Rayleigh damping ( C = c_M * M + c_K * K )"]
	mystr inelasticity_solution_method_str; //$EDC$[varaccess,EDCvarname="inelasticity_solution_method",EDCfolder="Inelasticity",tooltiptext="fixed_point, return_mapping (see Simo and Hughes, Computational Inelasticity 1998)"]
	InelasticitySolutionMethod inelasticity_solution_method; // flag (int) computes faster than inelasticity_solution_method_str (string)
	int materialnumber;						//material number in MBS system
	MBS* mbs;
};
//$EDC$[endclass,MaterialGeneric]


////////////////////////////////////////////////////////////
// class:   MaterialContinuumMechanics 
// purpose: base class for all continuum mechanics materials
////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=MaterialContinuumMechanics,parentclassname=MaterialGeneric,addelementtypename=MaterialContinuumMechanics,
//texdescription="MaterialContinuumMechanics is base class for all continuum mechanics materials.",
//texdescriptionComments="$It provides computation of stress tensors and update strategies for interior variables."]
class MaterialContinuumMechanics : public MaterialGeneric
{
public:
	// constructors and copying
	MaterialContinuumMechanics() : MaterialGeneric()	{}
	MaterialContinuumMechanics(MBS* mbsi) : MaterialGeneric(mbsi) {}
	MaterialContinuumMechanics(const MaterialContinuumMechanics& m) {}

	// set functions
	virtual void SetMaterialFromEDCData(){ MaterialGeneric::SetMaterialFromEDCData(); }  //$Set material type and parameters according to user specified edc data
	
	// element data access (automatically: function definitions generated by EDC_converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter



	//////////////////////////////////////////////////////////////////////
	// required functionality from derived classes (pure virtual functions)
	//
	//compute stress (piola2) from strain in 3D
	virtual void ComputeStressFromStrain(const Matrix3D& strain, Matrix3D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const = 0;
	virtual void ComputeStressFromStrain2D(const Matrix2D& strain, Matrix2D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const = 0;
	virtual void ComputeStressFromStrain2D(const Vector3D& strain, Vector3D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const {};
	
	// 'C' returns Elasticity Matrix in engineering notation (Matrix of size 6x6 in 3D, or 3x3 in 2D) //$ PG 2016-2-16: get rid of this method in this base class (i.e. perform different call in Elements FiniteElement2D::StiffnessMatrix, FiniteElement3D::StiffnessMatrix)
	virtual void ComputeElasticityMatrix(Matrix& C) const = 0;
	//
	//////////////////////////////////////////////////////////////////////
	// optional functionality from derived classes 
	//
	//compute update of inelastic variables as function from total strain and old state of inelastic variables
	virtual double ComputeInelasticVariablesUpdateFromStrain(Vector& inel_variables_update, const Vector& inel_variables, const Matrix3D& strain) const { mbs->UO(UO_LVL_err) << "WARNING: ComputeInelasticVariablesUpdateFromStrain not implemented\n"; return 1e99; }
	virtual double ComputeInelasticVariablesUpdateFromStrain2D(Vector& inel_variables_update, const Vector& inel_variables, const Matrix2D& strain) const { mbs->UO(UO_LVL_err) << "WARNING: ComputeInelasticVariablesUpdateFromStrain2D not implemented\n"; return 1e99; }
	//
	// 'C' [IN] may already contain Elasticity Matrix in engineering notation,
	// 'C' [OUT] returns Consistent Tangent Stiffness Matrix in engineering notation,
	// 'strain' contains total strain in engineering notation: (e_xx, e_yy, 2 e_xy) in 2D, and (e_xx, e_yy, e_zz, 2 e_yz, 2 e_zx, 2 e_xy) in 3D
	// 'inelastict_variables' contains plastic strain in engineering notation, as well as other variables such as a hardening parameter (depending on the respective InelasticityType)
	virtual void ComputeConsistentTangentStiffnessMatrix(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const { mbs->UO(UO_LVL_err) << "WARNING: ComputeConsistentTangentStiffnessMatrix not implemented\n"; }
	virtual void ComputeConsistentTangentStiffnessMatrix2D(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const { mbs->UO(UO_LVL_err) << "WARNING: ComputeConsistentTangentStiffnessMatrix2D not implemented\n"; }
	//
	//////////////////////////////////////////////////////////////////////


	virtual double ComputeYieldFunction(const Matrix3D& strain, const Vector& inelastic_variables) const { return 0.; }
	virtual double ComputeYieldFunction2D(const Matrix2D& strain, const Vector& inelastic_variables) const { return 0.; }

	//[STRESS] at input stress, at output deviator of stress
	//returns effective von-Mises stress
	virtual double ComputeDeviator(Matrix3D& stress) const;
	
	//[STRESS2D] at input components (11, 12, 21, 22) of stress,
	//           at output components (11, 12, 21, 22) of deviator of stress
	//returns effective von-Mises stress (invoking also component 33 either of stress or strain!)
	virtual double ComputeDeviator2D(Matrix2D& stress2d, double strain_offset_zz) const;

	// 2D plain strain: computes the 33 stress component from the 11, 12 and 22 components and the plastic variables.
	// Note that the 13 and 23 components always vanish in 2D.
	virtual double ComputeStress33_2D(const Matrix2D& stress2d, double strain_offset_zz) const;

	virtual double ComputeThickness2D(const Matrix2D& strain, const Vector& inelastic_variables, double thickness) const { return thickness; }
};
//$EDC$[endclass,MaterialContinuumMechanics]




////////////////////////////////////////////////////////////////////////////////////////////////////////
// enum:    TMBSMaterial
// purpose: quick access to material type; not used/set when material is instantiated by script language
////////////////////////////////////////////////////////////////////////////////////////////////////////
typedef enum
{
	TMatEmpty = 0, 
	TMatRigidBody = 1, 
	TMatFlexGeneral = 2, 
	TMat2D = 4, 
	TMatBeam = 8, 
	TMatPlate = 16, 
	TMatShell = 32, 
	TMatBeam2D = 64, 
	/*TMatOrthotropic = 128,*/ //$ PG 2016-2-10: deleted
	TMatAnisotropic = 256, 
	TMatPlaneStrain = 512, 
	TMatTemperature = 1024
} TMBSMaterial;


//const int MAXInelasticVariablesCount = 20;   //$ PG 2016-3-14: increased due to implementation of viscoelastic material
const int MAXInelasticVariablesCount = 100;

//class IntegrationPointVariables
//{
//	double* data;
//};


//$EDC$[beginclass,classname=ElasticParametersLinear]
class ElasticParametersLinear
{
public:	
	struct OrthotropicConstants
	{
		OrthotropicConstants() { e1=0.; e2=0.; e3=0.; nu12=0.; nu13=0.; nu23=0.; g12=0.; g13=0.; g23=0.; }
	public:
		double& E1() {return e1;} double& E2() {return e2;} double& E3() {return e3;}
		double& NU12() {return nu12;}	double& NU13() {return nu13;}	double& NU23() {return nu23;}
		double& G12() {return g12;}	double& G13() {return g13;}	double& G23() {return g23;}
		
		const double E1() const {return e1;} const double& E2() const {return e2;} const double& E3() const {return e3;}
		const double NU12() const {return nu12;}	const double& NU13() const {return nu13;}	const double& NU23() const {return nu23;}
		const double G12() const {return g12;}	const double& G13() const {return g13;}	const double& G23() const {return g23;}

	private:
		double e1, e2, e3;
		double nu12, nu13, nu23;
		double g12, g13, g23;
	};

	//constructing and copying
	ElasticParametersLinear() {DefaultConstructor();}
private: 
	MBS* mbs2;
//public:
	virtual const MBS* GetMBS() const { return mbs2; }
	virtual MBS* GetMBS() { return mbs2; }
public:
	ElasticParametersLinear(MBS* mbsi) { mbs2 = mbsi; DefaultConstructor();}
	ElasticParametersLinear(const ElasticParametersLinear& m) {CopyFrom(m);}

	//setting
	void SetElasticParametersLinear(double youngs_modulusI, double poisson_ratioI);
	void SetElasticParametersLinear(const OrthotropicConstants& oc);

	//checking
	int CheckConsistency(mystr& errorstr);   //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute

	//get and set orthotropic constants:
	virtual int IsOrthotropicMaterial() const {return is_orthotropic;}
	virtual void SetOrthotropicConstants(const OrthotropicConstants oc) {orthotropic_constants = oc;}
	virtual const OrthotropicConstants& GetOrthotropicConstants() const {return orthotropic_constants;}
	virtual OrthotropicConstants& GetOrthotropicConstants() {return orthotropic_constants;}

	//get and set isotropic constants:
	virtual const double& YoungsModulus() const {return youngs_modulus;}
	virtual double& YoungsModulus() {return youngs_modulus;}
	virtual const double& PoissonRatio() const {return poisson_ratio;}
	virtual double& PoissonRatio() {return poisson_ratio;}
	virtual double GetMu() const { return youngs_modulus / (2. + 2.*poisson_ratio); }
	virtual double GetLambda() const { return poisson_ratio*youngs_modulus / (1. + poisson_ratio) / (1. - 2*poisson_ratio); }
	virtual double GetLambdaPlaneStress() const { return poisson_ratio*youngs_modulus / (1. - poisson_ratio*poisson_ratio);} 	// 0 = sigma_33 = 2 mu eps_33 + la tr(eps) I  ==>  eps_33 = -(la/(2 mu + la)) * (eps_11 + eps_22)  ==>  sigma_2d = 2 mu eps_2d + la_2d tr(eps_2d) I_2d, with  la2d = la * (1 - la/(la + 2 mu)) = la * (2 mu /(la + 2 mu)) = nu E / ((1+nu)*(1-nu))

	//'C' returns Elasticity Matrix in engineering notation (Matrix of size 6x6 in 3D, or 3x3 in 2D)
	virtual void ComputeElasticityMatrix(Matrix& C, int is_planar_material, int is_plane_strain) const;


	// required for edc auto converter (manual element data access)
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) {return 0;}
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) {return 0;}

	// required for edc auto converter (function definitions generated by EDC_converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

protected:
	virtual void DefaultConstructor();
	virtual void CopyFrom(const ElasticParametersLinear& s);

	//isotropic parameters:
	double youngs_modulus; //$EDC$[varaccess,EDCvarname="youngs_modulus",EDCfolder="Solid",tooltiptext="Youngs modulus"]
	double poisson_ratio;  //$EDC$[varaccess,EDCvarname="poisson_ratio",EDCfolder="Solid",tooltiptext="Poisson ratio"]

	//orthotropic parameters:
	OrthotropicConstants orthotropic_constants;

	//flag
	int is_orthotropic;    //$EDC$[varaccess,EDCvarname="is_orthotropic_material",EDCfolder="Solid",int_bool,tooltiptext="check to enable orthotropic parameters"]

	//EDC double GetOrthotropicConstants().E1();	//$EDC$[varaccess,EDCvarname="E1",EDCfolder="Solid.Orthotropic",tooltiptext="elasticity along axis 1"]
	//EDC double GetOrthotropicConstants().E2();	//$EDC$[varaccess,EDCvarname="E2",EDCfolder="Solid.Orthotropic",tooltiptext="elasticity along axis 2"]
	//EDC double GetOrthotropicConstants().E3();	//$EDC$[varaccess,EDCvarname="E3",EDCfolder="Solid.Orthotropic",tooltiptext="elasticity along axis 3"]
	//EDC double GetOrthotropicConstants().NU12();	//$EDC$[varaccess,EDCvarname="NU12",EDCfolder="Solid.Orthotropic",tooltiptext="poisson ration 12"]
	//EDC double GetOrthotropicConstants().NU13();	//$EDC$[varaccess,EDCvarname="NU13",EDCfolder="Solid.Orthotropic",tooltiptext="poisson ration 13"]
	//EDC double GetOrthotropicConstants().NU23();	//$EDC$[varaccess,EDCvarname="NU23",EDCfolder="Solid.Orthotropic",tooltiptext="poisson ration 23"]
	//EDC double GetOrthotropicConstants().G12();	//$EDC$[varaccess,EDCvarname="G12",EDCfolder="Solid.Orthotropic",tooltiptext="shear modulus 12"]
	//EDC double GetOrthotropicConstants().G13();	//$EDC$[varaccess,EDCvarname="G13",EDCfolder="Solid.Orthotropic",tooltiptext="shear modulus 13"]
	//$EK 2018-09-05 : [bug fix in orthotropic EDC access:
	//EDC double GetOrthotropicConstants().G23();	//$EDC$[varaccess,EDCvarname="G23",EDCfolder="Solid.Orthotropic",tooltiptext="shear modulus 23"]
	//$EK 2018-09-05 : ]
};
//$EDC$[endclass,ElasticParametersLinear]

//////////////////////////////////////////////////////////////////////////////////////////////
// class:   Material
// purpose: linear elastic (isotropic (Hooke) or orthotropic) material for continuum mechanics
//          serves as base class for MaterialElastoplastic and MaterialNonLinSimoHughes
//////////////////////////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=Material,parentclassname=MaterialContinuumMechanics,parentclassname=ElasticParametersLinear,addelementtypename=Material,
//texdescription="Material is the basic Object for defining material properties for standard finite elements (in contrast to structural finite elements such as beams and plates).",
//texdescriptionComments="For static problems define the elastic properties $\mathtt{Solid.youngs\_modulus}$ and $\mathtt{Solid.poisson\_ratio}$, whereas for dynamic problems also $\mathtt{Solid.density}$ is required. If the problem is planar ($\mathtt{Solid.plane}$ is set to $1$), then the plane strain case is assumed unless $\mathtt{Solid.plane\_stress}$ is set to $1$. If the material is inelastic, then also the properties in the subtree $\mathtt{Inelasticity}$ have to be set.",
//example="Material.txt"]
class Material : public MaterialContinuumMechanics, public ElasticParametersLinear
{
public:
	// constructors, copying
	Material() : MaterialContinuumMechanics(), ElasticParametersLinear() { DefaultConstructor();	} //empty constructor
	Material(MBS* mbsi) : MaterialContinuumMechanics(mbsi), ElasticParametersLinear(mbsi)	{	DefaultConstructor();	} //recommended constructor
	Material(const Material& s)	{	CopyFrom(s); }; // copy constructor	
	virtual MaterialGeneric* GetCopy() const; // get copy of instance outside from class

	//$ PG 2016-1-18: [ all depricated:
	/*
	//deprecated - use Material(MBS* mbsi) instead
	Material(MBS* mbsi, double densityI)
	{
		mbs = mbsi;
		MaterialDefaultConstructor();

		type = TMatRigidBody;
		density = densityI;

		mbs->UO() << "WARNING: constructor deprecated - use __thiscall Material::Material(class MBS *) instead of " << __FUNCSIG__  << ".\n";
	}

	//deprecated - use Material(MBS* mbsi) instead
	Material(MBS* mbsi, double densityI, double youngs_modulusI, double poisson_ratioI, int plane = 0, int planestress = 1)
	{
		mbs = mbsi;
		MaterialDefaultConstructor();

		if (!plane)
		{
			type = TMatFlexGeneral;
		}
		else
		{
			type = TMBSMaterial(TMatFlexGeneral + TMat2D);
		}

		if (!planestress)
		{
			//type = TMatPlaneStrain;
			type = TMBSMaterial(type | TMatPlaneStrain);
		}

		density = densityI;
		youngs_modulus = youngs_modulusI;
		poisson_ratio = poisson_ratioI;
		
		mbs->UO() << "WARNING: constructor deprecated - use __thiscall Material::Material(class MBS *) instead of " << __FUNCSIG__  << ".\n";
	}

	//$ SW 2014-8-18: Removed since deprecated and wrong implemented (In C++03, constructors of a class are not allowed to call other constructors of that class, see http://en.wikipedia.org/wiki/C%2B%2B11#Object_construction_improvement)
	
	//deprecated - use Material(MBS* mbsi) instead
	Material(MBS* mbsi, double densityI, double youngs_modulusI, double poisson_ratioI, Vector3D color, int plane = 0, int planestress = 1)
	{
		Material(mbsi, densityI, youngs_modulusI, poisson_ratioI, plane, planestress);		
		materialcolor = color;
		
		mbs->UO() <<  WARNING: constructor deprecated - use __thiscall Material::Material(class MBS *) instead of " << __FUNCSIG__  << ".\n";
	} 

	//depricated - old Set-Function
	virtual void SetMaterial(int nonlinmattypeI, double densityI, double youngs_modulusI, double poisson_ratioI);
	*/
	//$ PG 2016-1-26: ]


	// name of material type
	virtual mystr GetTypeName() const;

	// setting
	virtual void SetMaterial(double densityI, double youngs_modulusI, double poisson_ratioI, Vector3D color, int plane = 0, int planestress = 1);
	virtual void SetMaterial(double densityI, Vector3D color, const Material::OrthotropicConstants& oc);
	virtual void SetMaterialFromEDCData(); //$Set material type and parameters according to user specified edc data
	
	//$ PG 2013-1-21: setting material for rigid bodies is deprecated, since rigids should not use class Material!
	virtual Material& SetMaterialRigid(double densityI);

	// checking
	virtual int CheckConsistency(mystr& errorstr);

	// initializing inelastic variables
	virtual void InitializeInelasticVariables(Vector& inel_variables) const {};

	// get status information
	virtual int IsInelasticMaterial() const;
	virtual int GetInelasticVariablesCount() const;	


	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) { return 0; }
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) { return 0; }
	virtual const MBS* GetMBS() const { return MaterialContinuumMechanics::GetMBS(); }
	virtual MBS* GetMBS() { return MaterialContinuumMechanics::GetMBS(); }


	// required for edc auto converter (function definitions generated by EDC_converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter


	//$ PG 2016-1-25: do we really need it?
	//Material& operator=(const Material& s) 
	//{
	//	if (this == &s) {return *this;}
	//	CopyFrom(s);
	//	return *this;
	//}

	//$ PG 2016-1-26: [ to be erased in the near future
	virtual TMBSMaterial GetType() const;
	virtual void SetType(TMBSMaterial t);
	virtual void AddType(TMBSMaterial t);
	//$ PG 2016-1-26: ]


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// stress computation etc.

	//compute stress (piola2) from strain in 3D
	virtual void ComputeStressFromStrain(
		const Matrix3D& total_strain, Matrix3D& stress,
		const Vector& inel_variables = dummyVector,
		bool compute_trial_stress = false) const;

	//compute stress (piola2) from strain in 2D
	virtual void ComputeStressFromStrain2D(
		const Matrix2D& total_strain, Matrix2D& stress,
		const Vector& inel_variables = dummyVector,
		bool compute_trial_stress = false
		) const;

	// 2D plain strain: computes the 33 stress component from the 11, 12 and 22 components and the plastic variables.
	// Note that the 13 and 23 components always vanish in 2D.
	virtual double ComputeStress33_2D(const Matrix2D& stress2d, double strain_offset_zz) const;

	// 'C' returns Elasticity Matrix in engineering notation (Matrix of size 6x6 in 3D, or 3x3 in 2D)
	virtual void ComputeElasticityMatrix(Matrix& C) const;


	////////////////////////////////  OLD AND DEPRICATED FUNCTIONALITY (BEGIN):

	//virtual double ComputeAveragedContactStress(FiniteElement3D* const fe);	

	//$ PG 2013-8-1: flag TMatNonlinElast and member nonlinmattype have been deleted. nonlinear elasticity is now specified via setting inelasticity_type = IT_NonlinElasticSimoHughes.
	//$ PG 2013-8-1: therefore, the following function may be deleted
	//virtual int IsNonlinElasticMaterial() const {if (type&TMatNonlinElast) return 1; else return 0;}



	//$ PG 2013-8-7: replaced by ComputeInelasticVariablesFromStrain(...)
	//virtual double DoNonlinStep(FiniteElement3D* const fe, const Matrix3D& F, Vector& inel_strain, Vector& inel_strain_old) const 
	//{
	//	//------------------------------
	//	// equations used:             |
	//	//                             |
	//	// int_A (P : d F / d q) dA    |
	//	// P = F * S                   |
	//	// S = D : E                   |
	//	// E = 1/2*(F^T F-I)           |
	//	//------------------------------

	//	return 0;
	//}

	//virtual double DoNonlinStep(FiniteElement2D* const fe, const Matrix2D& F, Vector& inel_strain, Vector& inel_strain_old) const 
	//{
	//	//------------------------------
	//	// equations used:             |
	//	//                             |
	//	// int_A (P : d F / d q) dA    |
	//	// P = F * S                   |
	//	// S = D : E                   |
	//	// E = 1/2*(F^T F-I)           |
	//	//------------------------------

	//	return 0;
	//}

	//$ PG 2013-8-1: deleted nonlinmattype, nonlinear elastic material is now specified via setting inelasticity_type = IT_NonlinElasticSimoHughes
	//$ PG 2013-8-1: the use of the following two functions has been replaced in all models and elements, they are not needed anymore and shall be erased from this file
	//virtual const int& NonlinMatType() const {return nonlinmattype;}
	//virtual int& NonlinMatType() {return nonlinmattype;}

	virtual void GetStress2D(const Vector3D& strain, Vector3D& stress, const Vector3D& inel_strain, const double contactstress = -1) const;
	virtual void GetLinearizedStress2D(const Vector3D& strain, Vector3D& stress, const Vector3D& inel_strain, const Vector3D& stress_lin) const;

	//PG: This routine addresses geometrical properties only.
	//*   Hence, it should be provided by the element - not by the material.
	virtual void ComputeStrain(Matrix3D& strain, const Matrix3D& grad_u, int isgeomnonlin);
	virtual void ComputeStrain2D(Matrix2D& strain, const Matrix2D& grad_u, int isgeomnonlin);

	//General access functions:
	//$ PG 2016-1-26: [ methods seem not to be used:
	//virtual int NParameters() const {return 21;} //this should be the number of possible parameters 
	//virtual void GetParam(int i, double& value, mystr& name, int& group, int& flag);
	//$ PG 2016-1-26: ]

	////////////////////////////////  OLD AND DEPRICATED FUNCTIONALITY (END):

protected:
	virtual void CopyFrom(const Material& s);
	virtual void DefaultConstructor();


	
	//$ PG 2016-1-19: flag type is for fast internal computation, may become redundant with new class architecture (?)
	TMBSMaterial type;
};
//$EDC$[endclass,Material]


//////////////////////////////////////////////////////////////////////////////////////////////
// class:   MaterialThermalExpansion
// purpose: linear elastic (isotropic (Hooke) or orthotropic) material for continuum mechanics
//          serves as base class for MaterialElastoplastic and MaterialNonLinSimoHughes
//////////////////////////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=MaterialThermalExpansion,parentclassname=Material,addelementtypename=MaterialThermalExpansion,
//texdescription="Material which considers thermal expansion.",
//texdescriptionComments=""]
class MaterialThermalExpansion : public Material
{
public:
	// constructors, copying
	MaterialThermalExpansion() : Material() { DefaultConstructor();	} //empty constructor
	MaterialThermalExpansion(MBS* mbsi) : Material(mbsi)	{	DefaultConstructor();	} //recommended constructor
	MaterialThermalExpansion(const MaterialThermalExpansion& s)	{	CopyFrom(s); }; // copy constructor	
	virtual MaterialGeneric* GetCopy() const; // get copy of instance outside from class

	// name of material type
	virtual mystr GetTypeName() const;

	virtual void SetMaterialFromEDCData(); //$Set material type and parameters according to user specified edc data

	// required for edc auto converter (function definitions generated by EDC_converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	virtual void CopyFrom(const MaterialThermalExpansion& s);
	virtual void DefaultConstructor();

	virtual bool HasInitialStrains() const { return !(delta_T == 0. || thermal_expansion_coefficient == 0.); }
	virtual void GetInitialStrain2D(Matrix2D & initial_strain) const;
	virtual double GetInitialStrain33() const;
	virtual void GetInitialStrain(Matrix3D & initial_strain) const; 

	void SetThermalExpansion(double delta_T, double thermal_expansion_coefficient);
	

protected:
	double delta_T;                       //$EDC$[varaccess,EDCvarname="delta_T",EDCfolder="ThermalExpansion",tooltiptext="temperature difference to reference material"]
	double thermal_expansion_coefficient; //$EDC$[varaccess,EDCvarname="coefficient",EDCfolder="ThermalExpansion",tooltiptext="thermal expansion coefficient"]
};
//$EDC$[endclass,MaterialThermalExpansion]




class ElastoplasticParametersPrandtlReuss
{
public:
	//constructors
	ElastoplasticParametersPrandtlReuss() {DefaultConstructor();}
	ElastoplasticParametersPrandtlReuss(const ElastoplasticParametersPrandtlReuss& s) {CopyFrom(s);}

	//setting
	virtual void SetElastoplasticParametersPrandtlReuss(double yield_stressI, double tangent_moduleI);

	//checking
	virtual int CheckConsistency(mystr& errorstr);

	// general information
	virtual int GetInelasticVariablesCount(int planar) const;
	
	//accessing
	virtual double YieldStress() const {return yieldstress;}
	virtual double HardeningModule() const {return sqrt(tangentmodule)/yieldstress;}

protected:
	virtual void CopyFrom(const ElastoplasticParametersPrandtlReuss& s);
	virtual void DefaultConstructor();

// elastoplastic parameters
	double yieldstress;
	double tangentmodule;
};




//$EDC$[beginclass,classname=MaterialElastoplastic,parentclassname=Material,addelementtypename=MaterialElastoplastic,
//texdescription="MaterialElastoplastic is a material which obeys linear elasticity and Prandtl Reuss flow rule with Mises yield condition. Also, material hardening is considered.",
//texdescriptionComments="Please makr sure, that all items in subtree $\mathtt{Inelasticity}$ are defined."]
class MaterialElastoplastic : public Material, public ElastoplasticParametersPrandtlReuss
{
public:
	//constructing and copying
	MaterialElastoplastic() : Material (), ElastoplasticParametersPrandtlReuss () { DefaultConstructor(); }
	MaterialElastoplastic(MBS* mbs) : Material (mbs), ElastoplasticParametersPrandtlReuss () { DefaultConstructor(); }
	MaterialElastoplastic(const MaterialElastoplastic& me) { CopyFrom(me); }
	virtual MaterialGeneric* GetCopy() const;
	
	//checking
	virtual int CheckConsistency(mystr& errorstr);
	
	//setting
	virtual void SetMaterialElastoplastic(double densityI, double youngs_modulusI, double poisson_ratioI, 
		InelasticitySolutionMethod inelasticity_solution_methodI, double yield_stressI, double tangent_moduleI, Vector3D color, int plane = 0, int planestress = 1);
	
	//initializing
	virtual int GetInelasticVariablesCount() const;
	virtual void InitializeInelasticVariables(Vector& inel_variables) const { /*all inelastic variables initialized with 0*/ };

	//accessing
	virtual mystr GetTypeName() const;
	virtual int IsInelasticMaterial() const;

	//edc auto
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	// computing stress, inelastic variable update, and consistent tangent stiffness matrix
	//
protected:
	virtual void Get3DStrainsTensors2D(Matrix3D& total_strain, Matrix3D& old_plastic_strain, double& hardening_parameter, const Matrix2D& strain, const Vector& inel_variables) const;
	virtual void ComputeStressFromStrain(Matrix3D& stress, const Matrix3D& total_strain, const Matrix3D& plastic_strain, double hardening_param, bool compute_trial_stress) const;
public:
	virtual void ComputeStressFromStrain(const Matrix3D& strain, Matrix3D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const;
	virtual void ComputeStressFromStrain2D(const Matrix2D& strain, Matrix2D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const;

protected:
	virtual double ComputeInelasticVariablesUpdateFromStrain(Matrix3D& inel_strain_update, double& hardening_param_update, const Matrix3D& strain, const Matrix3D& inel_strain, double hardening_param) const;
public:
	virtual double ComputeInelasticVariablesUpdateFromStrain(Vector& inel_variables_update, const Vector& inel_variables, const Matrix3D& strain) const;
	virtual double ComputeInelasticVariablesUpdateFromStrain2D(Vector& inel_variables_update, const Vector& inel_variables, const Matrix2D& strain) const;


	// 'C' [IN] may contain Elasticity Matrix in engineering notation,
	// 'C' [OUT] returns Consistent Tangent Stiffness Matrix in engineering notation,
	// 'strain' contains total strain in engineering notation: (e_xx, e_yy, 2 e_xy) in 2D, and (e_xx, e_yy, e_zz, 2 e_yz, 2 e_zx, 2 e_xy) in 3D
	// 'inelastict_variables' contains plastic strain in engineering notation, as well as other variables such as a hardening parameter (depending on the respective InelasticityType)
	virtual void ComputeConsistentTangentStiffnessMatrix(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const;
	virtual void ComputeConsistentTangentStiffnessMatrix2D(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const;

	// computes yield function (which is required for determining plastic strain update, etc.); definition is model-specific 
	virtual double ComputeYieldFunction(const Matrix3D& strain, const Vector& inelastic_variables) const;
	virtual double ComputeYieldFunction2D(const Matrix2D& strain, const Vector& inelastic_variables) const;

	// field variables managed by material (internal state variables)
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const;
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const;


protected:
	virtual void DefaultConstructor();
	void CopyFrom(const MaterialElastoplastic& m);

	//to be added to edc from base class ElastoplasticParametersPrandtlReuss
	//EDC	double yieldstress;						//$EDC$[varaccess,EDCvarname="yield_stress",EDCfolder="Inelasticity",tooltiptext="Yield Stress S_y, used for yield criterion: Mises(S) := sqrt(2./3.)*|dev S| <= S_y, with S denoting 2PK stress tensor"]
	//EDC	double tangentmodule;					//$EDC$[varaccess,EDCvarname="tangent_module",EDCfolder="Inelasticity",tooltiptext="Tangent module K used in plastic zones, i.e. K = d(Mises(S))/d|Ep|, corresponds to formerly used module of hardening H = sqrt(K)/S_y"]
};
//$EDC$[endclass,MaterialElastoplastic]








class ElastoplasticParametersEterovicBathe
{
public:
	//constructors
	ElastoplasticParametersEterovicBathe() {DefaultConstructor();}
	ElastoplasticParametersEterovicBathe(const ElastoplasticParametersEterovicBathe& s) {CopyFrom(s);}

	//setting
	virtual void SetElastoplasticParametersEterovicBathe(double yield_stressI, double tangent_moduleI);

	//checking
	virtual int CheckConsistency(mystr& errorstr);

	// general information
	virtual int GetInelasticVariablesCount(int planar) const;
	
	//accessing
	virtual double YieldStress() const {return yieldstress;}
	virtual double HardeningModule() const {return sqrt(tangentmodule)/yieldstress;}

protected:
	virtual void CopyFrom(const ElastoplasticParametersEterovicBathe& s);
	virtual void DefaultConstructor();

// elastoplastic parameters
	double yieldstress;
	double tangentmodule;
};

//$EDC$[beginclass,classname=MaterialElastoplasticLargeStrain,parentclassname=Material,addelementtypename=MaterialElastoplasticLargeStrain,
//texdescription="MaterialElastoplasticLargeStrain is a obeys linear elasticity and Prandtl Reuss flow rule with Mises yield condition. Also, material hardening is considered.",
//texdescriptionComments="Please makr sure, that all items in subtree $\mathtt{Inelasticity}$ are defined."]
class MaterialElastoplasticLargeStrain : public Material, public ElastoplasticParametersEterovicBathe
{
public:
	//constructing and copying
	MaterialElastoplasticLargeStrain() : Material (), ElastoplasticParametersEterovicBathe () { DefaultConstructor(); }
	MaterialElastoplasticLargeStrain(MBS* mbs) : Material (mbs), ElastoplasticParametersEterovicBathe () { DefaultConstructor(); }
	MaterialElastoplasticLargeStrain(const MaterialElastoplasticLargeStrain& me) { CopyFrom(me); }
	virtual MaterialGeneric* GetCopy() const;
	
	//checking
	virtual int CheckConsistency(mystr& errorstr);
	
	//setting
	virtual void SetMaterialElastoplasticLargeStrain(double densityI, double youngs_modulusI, double poisson_ratioI, 
		InelasticitySolutionMethod inelasticity_solution_methodI, double yield_stressI, double tangent_moduleI, Vector3D color, int plane = 0, int planestress = 1);
	
	//initializing
	virtual int GetInelasticVariablesCount() const;
	virtual void InitializeInelasticVariables(Vector& inel_variables) const { /*all inelastic variables initialized with 0*/ };

	//accessing
	virtual mystr GetTypeName() const;
	virtual int IsInelasticMaterial() const;
	
	//edc auto
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	// computing stress, inelastic variable update, and consistent tangent stiffness matrix
	//
protected:
	virtual void Get3DStrainsTensors2D(Matrix3D& total_strain, Matrix3D& old_plastic_strain, double& hardening_parameter, const Matrix2D& strain, const Vector& inel_variables) const;
	virtual double GetEquivalentTensileStress(const Vector& inelastic_variables) const;
	virtual double GetYieldStress(const Vector& inelastic_variables) const;
	virtual Matrix3D GetPlasticStrainTensor(const Vector& inelastic_variables) const;

public:
	virtual void ComputeStressFromStrain(const Matrix3D& strain, Matrix3D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const;
	virtual void ComputeStressFromStrain2D(const Matrix2D& strain, Matrix2D& stress, const Vector& inel_variables = dummyVector, bool compute_trial_stress = false) const;

protected:
	virtual double ComputeInelasticVariablesUpdateFromStrain(Matrix3D& inel_strain_update, double& hardening_param_update, const Matrix3D& strain, const Matrix3D& inel_strain, double hardening_param) const;
public:
	virtual double ComputeInelasticVariablesUpdateFromStrain(Vector& inel_variables_update, const Vector& inel_variables, const Matrix3D& strain) const;
	virtual double ComputeInelasticVariablesUpdateFromStrain2D(Vector& inel_variables_update, const Vector& inel_variables, const Matrix2D& strain) const;


	// 'C' [IN] may contain Elasticity Matrix in engineering notation,
	// 'C' [OUT] returns Consistent Tangent Stiffness Matrix in engineering notation,
	// 'strain' contains total strain in engineering notation: (e_xx, e_yy, 2 e_xy) in 2D, and (e_xx, e_yy, e_zz, 2 e_yz, 2 e_zx, 2 e_xy) in 3D
	// 'inelastict_variables' contains plastic strain in engineering notation, as well as other variables such as a hardening parameter (depending on the respective InelasticityType)
	virtual void ComputeConsistentTangentStiffnessMatrix(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const;
	virtual void ComputeConsistentTangentStiffnessMatrix2D(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const;

	// computes yield function (which is required for determining plastic strain update, etc.); definition is model-specific 
	virtual double ComputeYieldFunction(const Matrix3D& strain, const Vector& inelastic_variables) const;
	virtual double ComputeYieldFunction2D(const Matrix2D& strain, const Vector& inelastic_variables) const;

	// field variables managed by material (internal state variables)
	virtual void GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const;
	virtual double GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const;


protected:
	virtual void DefaultConstructor();
	void CopyFrom(const MaterialElastoplasticLargeStrain& m);

	//to be added to edc from base class ElastoplasticParametersEterovicBathe
	//EDC	double yieldstress;						//$EDC$[varaccess,EDCvarname="yield_stress",EDCfolder="Inelasticity",tooltiptext="Yield Stress S_y, used for yield criterion: Mises(S) := sqrt(2./3.)*|dev S| <= S_y, with S denoting 2PK stress tensor"]
	//EDC	double tangentmodule;					//$EDC$[varaccess,EDCvarname="tangent_module",EDCfolder="Inelasticity",tooltiptext="Tangent module K used in plastic zones, i.e. K = d(Mises(S))/d|Ep|, corresponds to formerly used module of hardening H = sqrt(K)/S_y"]
};
//$EDC$[endclass,MaterialElastoplasticLargeStrain]





//////////////////////////////////////////////////////////////////////////////////////////////
// class:   MaterialThermalExpansion
// purpose: linear elastic (isotropic (Hooke) or orthotropic) material for continuum mechanics
//          serves as base class for MaterialElastoplastic and MaterialNonLinSimoHughes
//////////////////////////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=MaterialElastoplasticThermalExpansion,parentclassname=MaterialElastoplastic,addelementtypename=MaterialElastoplasticThermalExpansion,
//texdescription="MaterialElastoplasticThermalExpansion is an elastoplatic material which considers thermal expansion"]
class MaterialElastoplasticThermalExpansion : public MaterialElastoplastic
{
public:
	// constructors, copying
	MaterialElastoplasticThermalExpansion() : MaterialElastoplastic() { DefaultConstructor();	} //empty constructor
	MaterialElastoplasticThermalExpansion(MBS* mbsi) : MaterialElastoplastic(mbsi)	{	DefaultConstructor();	} //recommended constructor
	MaterialElastoplasticThermalExpansion(const MaterialElastoplasticThermalExpansion& s)	{	CopyFrom(s); }; // copy constructor	
	virtual MaterialGeneric* GetCopy() const; // get copy of instance outside from class

	// name of material type
	virtual mystr GetTypeName() const;

	virtual void SetMaterialFromEDCData(); //$Set material type and parameters according to user specified edc data

	// required for edc auto converter (function definitions generated by EDC_converter in ElementEDCauto.cpp)
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	virtual void CopyFrom(const MaterialElastoplasticThermalExpansion& s);
	virtual void DefaultConstructor();

	virtual bool HasInitialStrains() const { return !(delta_T == 0. || thermal_expansion_coefficient == 0.); }
	virtual void GetInitialStrain2D(Matrix2D & initial_strain) const;
	virtual double GetInitialStrain33() const;
	virtual void GetInitialStrain(Matrix3D & initial_strain) const; 

	void SetThermalExpansion(double delta_T, double thermal_expansion_coefficient);
	

protected:
	double delta_T;                       //$EDC$[varaccess,EDCvarname="delta_T",EDCfolder="ThermalExpansion",tooltiptext="temperature difference to reference material"]
	double thermal_expansion_coefficient; //$EDC$[varaccess,EDCvarname="coefficient",EDCfolder="ThermalExpansion",tooltiptext="thermal expansion coefficient"]
};
//$EDC$[endclass,MaterialElastoplasticThermalExpansion]


class MaterialNonlinElasticSimoHughes : public Material
{
public:
	MaterialNonlinElasticSimoHughes() : Material() {}
	MaterialNonlinElasticSimoHughes(MBS* mbs) : Material(mbs) {}
	MaterialNonlinElasticSimoHughes(const MaterialNonlinElasticSimoHughes& me) { CopyFrom(me); }

	virtual MaterialGeneric* GetCopy() const;
	virtual mystr GetTypeName() const;

	virtual void ComputeStressFromStrain(
		const Matrix3D& strain, Matrix3D& stress,
		const Vector& inel_variables = dummyVector,
		bool compute_trial_stress = false) const;

	virtual void ComputeStressFromStrain2D(
		const Matrix2D& strain, Matrix2D& stress,
		const Vector& inelastic_variables = dummyVector,
		bool compute_trial_stress = false) const;

protected:
	virtual void DefaultConstructor();
};






/////////////////////////////
// class BeamCrossSectionInfo
/////////////////////////////
class BeamCrossSectionInfo
{
public:
	typedef enum
	{
		TCSS_NONE = 0,
		TCSS_rectangular = 1,
		TCSS_circular = 2,
		TCSS_tubular = 3,
		TCSS_polygonal = 4,
		TCSS_MAX
	} TCrossSectionShape;

	BeamCrossSectionInfo();
	BeamCrossSectionInfo(const BeamCrossSectionInfo& b);
	
	virtual void SetBeamCrossSectionInfo(int _beamCrossSectionShape, const Vector& _beamCrossSectionSize);
	virtual int CheckConsistency(mystr& errorstr);

	virtual double GetBeamCrossSectionSize(int i) const { return beamCrossSectionSize(i); }
	virtual TCrossSectionShape GetCrossSectionShape() const { return beamCrossSectionShape; }
	virtual double GetCrossSectionArea() const;
	virtual int IsRectangularCrossSection() const;
	virtual int IsCircularCrossSection() const;
	virtual int IsTubularCrossSection() const;
	virtual double GetBeamThicknessY() const { return IsRectangularCrossSection() ? beamCrossSectionSize(1) : 0; }
	virtual double GetBeamThicknessZ() const {return IsRectangularCrossSection() ? beamCrossSectionSize(2) : 0; }
	//$EK 2015-10-14 bug fix - beamCrossSectionSize for circular and tubular cross sections contains the diameter;
	//therefore the radius is 1/2 of the crossSectionSize
	virtual double GetRadius() const { return (beamCrossSectionShape == TCSS_tubular || beamCrossSectionShape == TCSS_circular) ? beamCrossSectionSize(1)/2. : 0.; }
	virtual double GetInnerRadius() const { return (beamCrossSectionShape == TCSS_tubular) ? beamCrossSectionSize(2)/2. : 0.; }

protected:
	virtual void CopyFrom(const BeamCrossSectionInfo& b);
	virtual int ComputeCrossSectionArea();   // returns 0 if error, 1 if ok

	TCrossSectionShape beamCrossSectionShape;
	Vector beamCrossSectionSize;

	double beamCrossSectionArea;
	double beamIx; // polar moment of inertia
	double beamIy; // area moment of inertia in direction e2
	double beamIz; // area moment of inertia in direction e3
};




///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class Beam3DProperties : public MaterialGeneric, public BeamCrossSectionInfo, public ElasticParametersLinear, public ElastoplasticParametersPrandtlReuss
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=Beam3DProperties,parentclassname=MaterialGeneric,addelementtypename=Beam3DProperties,pythonclassdesctiption="Material for 3D Beams",pythonAddAs=BeamProperties,
//texdescription="Beam3DProperties defines material and geometric properties for beam structural finite elements.",
//texdescriptionComments="First, specify the $\mathtt{cross\_section\_type}$ of the beam, which may be either rectangular (if set to $1$), circular (if set to $2$), or tubular (if set to $3$) or polygonal (if set to $4$). In either case the $\mathtt{cross\_section\_size}$ is a vector of $2$, $1$, $2$, or $2\,n$ entries, where $n$ confers to the number of vertices of a closed polygon. Then specify the stiffnesses and moments of inertias, as they are neede by your beam and problem.",
//example="Beam3DProperties.txt"]
class Beam3DProperties: public MaterialGeneric, public BeamCrossSectionInfo, public ElasticParametersLinear, public ElastoplasticParametersPrandtlReuss
{
public:
	// constructing and copying
	Beam3DProperties() : MaterialGeneric(), BeamCrossSectionInfo(), ElasticParametersLinear(), ElastoplasticParametersPrandtlReuss() { DefaultConstructor(); }
	Beam3DProperties(MBS* mbs) : MaterialGeneric(mbs), BeamCrossSectionInfo(), ElasticParametersLinear(mbs), ElastoplasticParametersPrandtlReuss() { DefaultConstructor(); }
	Beam3DProperties(const Beam3DProperties& bp) { CopyFrom(bp); }
	virtual MaterialGeneric* GetCopy() const;
	
	// material name
	virtual mystr GetTypeName() const;
	
	// setting
	virtual void SetBeam3DProperties(int _beamCrossSectionType /*1: rectangular, 2: circular, 3: tubular, 4: polygonal*/, const Vector& _beamCrossSectionSize, double _rho, double _EA, double _EIy, double _EIz=0, double _GJkx=0, double _GAky=0, double _GAkz=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0, double _rhoIz=0);  //$ PG 2016-2-11: recommended set function
	virtual void SetBeam3DPropertiesCM(int _beamCrossSectionType /*1: rectangular, 2: circular, 3: tubular, 4: polygonal*/, const Vector& _beamCrossSectionSize, double _rho, double E, double nu, double _EA=0, double _EIy=0, double _EIz=0, double _GJkx=0, double _GAky=0, double _GAkz=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0, double _rhoIz=0); //$ PG 2016-2-11: recommended set function for continuum mechanics formulation of beams
	virtual void SetBeam3DProperties(double _rho, double _EA, double _EIy, double _EIz=0, double _GJkx=0, double _GAky=0, double GAkz=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0, double _rhoIz=0); //$ PG 2016-2-11: deprecated, old
	virtual void SetMaterialFromEDCData(); 

	// checking
	virtual int CheckConsistency(mystr& errorstr);
	
	// status information 
	virtual int IsInelasticMaterial() const { return YieldStress() > 0; }
	virtual int IsHardeningMaterial() const { return HardeningModule() > 0; }
	virtual int GetInelasticVariablesCount() const { return IsInelasticMaterial()*ElastoplasticParametersPrandtlReuss::GetInelasticVariablesCount(plane); }
	
	// initializing inelastic variables (called from each integration point of elements)
	virtual void InitializeInelasticVariables(Vector& inel_variables) const {}
	
	virtual const MBS* GetMBS() const { return MaterialGeneric::GetMBS(); }
	virtual MBS* GetMBS() { return MaterialGeneric::GetMBS(); }
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) { return 0; }
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) { return 0; }
	
	// required for edc auto-converter
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	// stiffnesses and moments of inertia
	virtual const double& BeamRhoA() const {return beamRhoA;   }
	virtual double& BeamRhoA() {return beamRhoA;   }
	virtual const double& BeamRhoIx() const {return beamRhoIx;   }
	virtual double& BeamRhoIx() {return beamRhoIx;   }
	virtual const double& BeamRhoIy() const {return beamRhoIy;   }
	virtual double& BeamRhoIy() {return beamRhoIy;   }
	virtual const double& BeamRhoIz() const {return beamRhoIz;   }
	virtual double& BeamRhoIz() {return beamRhoIz;   }
	virtual const double& BeamEA() const {return beamEA;   }
	virtual double& BeamEA() {return beamEA;   }
	virtual const double& BeamEIy() const {return beamEIy; }
	virtual double& BeamEIy() {return beamEIy; }
	virtual const double& BeamEIz() const {return beamEIz; }
	virtual double& BeamEIz() {return beamEIz; }
	virtual const double& BeamGAky() const {return beamGAky;}
	virtual double& BeamGAky() {return beamGAky;}
	virtual const double& BeamGAkz() const {return beamGAkz;}
	virtual double& BeamGAkz() {return beamGAkz;}
	virtual const double& BeamGJkx() const {return beamGJkx;}
	virtual double& BeamGJkx() {return beamGJkx;}

protected:
	virtual void DefaultConstructor();	
	virtual void CopyFrom(const Beam3DProperties& s); 

	//to be added to edc from base class BeamCrossSectionInfo
	//EDC int beamCrossSectionShape;	//$EDC$[varaccess,EDCvarname="cross_section_type",EDCfolder="",tooltiptext="1: rectangular, 2: circular, 3: tubular, 4: polygonal",python]
	//EDC Vector beamCrossSectionSize;	//$EDC$[varaccess,EDCvarname="cross_section_size",EDCfolder="",variable_length_vector,tooltiptext="vector length of cross_section_size depends on cross_section_type: length 2 for rectangular cross section (y and z extension), length 1 for circular cross section (diameter), length 2 for tubular cross section (outer and inner diameter), and length 2*n for polygonal cross section (p1y,p1z,p2y,p2z,...,pny,pnz)",python];
	//EDC double beamCrossSectionArea;  //$EDC$[varaccess,readonly,EDCvarname="cross_section_area",EDCfolder="",tooltiptext="area of the beam's cross section"]
	//EDC double beamIx;  //$EDC$[varaccess,readonly,EDCvarname="Ix",EDCfolder="",tooltiptext="polar moment of inertia of the beams cross section"]
	//EDC double beamIy;  //$EDC$[varaccess,readonly,EDCvarname="Iy",EDCfolder="",tooltiptext="area moment of inertia of the beams cross section w.r.t. y-axis (2D-beam)"]
	//EDC double beamIz;  //$EDC$[varaccess,readonly,EDCvarname="Iz",EDCfolder="",tooltiptext="area moment of inertia of the beams cross section w.r.t. z-axis"]

	// own members
  double beamEA;								//$EDC$[varaccess,EDCvarname="EA",EDCfolder="",tooltiptext="youngs modulus * area",python]
	double beamEIy;								//$EDC$[varaccess,EDCvarname="EIy",EDCfolder="",tooltiptext="bending stiffness w.r.t. y-axis (2D-beam)",python]
	double beamEIz;								//$EDC$[varaccess,EDCvarname="EIz",EDCfolder="",tooltiptext="bending stiffness w.r.t. z-axis",python]
	double beamGAky;							//$EDC$[varaccess,EDCvarname="GAky",EDCfolder="",tooltiptext="shear stiffness including shear correction factor ky (2D-beam)",python]
	double beamGAkz;							//$EDC$[varaccess,EDCvarname="GAkz",EDCfolder="",tooltiptext="shear stiffness including shear correction factor kz",python]
	double beamGJkx;							//$EDC$[varaccess,EDCvarname="GJkx",EDCfolder="",tooltiptext="torsional stiffness including shear correction factor kx",python]
	double beamRhoA;							//$EDC$[varaccess,EDCvarname="RhoA",EDCfolder="",tooltiptext="density * area",python]
	double beamRhoIx;							//$EDC$[varaccess,EDCvarname="RhoIx",EDCfolder="",tooltiptext="density * second area of moment w.r.t. x-axis",python]
	double beamRhoIy;							//$EDC$[varaccess,EDCvarname="RhoIy",EDCfolder="",tooltiptext="density * second area of moment w.r.t. y-axis (2D-beam)",python]
	double beamRhoIz;							//$EDC$[varaccess,EDCvarname="RhoIz",EDCfolder="",tooltiptext="density * second area of moment w.r.t. z-axis",python]

	// to be removed from edc of base class MaterialGeneric
	//EDC int plane_stress;							//$EDC$[varaccess,EDCvarname="plane_stress",remove,EDCfolder="Solid",int_bool,tooltiptext="for 2D-Elements only; 1: plane stress, 0: plane strain",python]
	//EDC int plane;										//$EDC$[varaccess,EDCvarname="plane",EDCfolder="Solid",remove,python]

	//to be added to edc from base classes ElasticParametersLinear and ElastoplasticParametersPrandtlReuss
	//EDC double youngs_modulus;				//$EDC$[varaccess,EDCvarname="youngs_modulus",EDCfolder="Solid",tooltiptext="Youngs modulus"]
	//EDC double poisson_ratio;					//$EDC$[varaccess,EDCvarname="poisson_ratio",EDCfolder="Solid",tooltiptext="Poisson ratio"]
	//EDC	double yieldstress;						//$EDC$[varaccess,EDCvarname="yield_stress",EDCfolder="Inelasticity",tooltiptext="Yield Stress s_y, e.g., |dev s| <= s_y"]
	//EDC	double tangentmodule;					//$EDC$[varaccess,EDCvarname="tangent_module",EDCfolder="Inelasticity",tooltiptext="Modulus of hardening H"]
};
//$EDC$[endclass,Beam3DProperties]


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class Beam2DProperties : public MaterialGeneric, public BeamCrossSectionInfo, public ElasticParametersLinear, public ElastoplasticParametersPrandtlReuss
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=Beam2DProperties,parentclassname=MaterialGeneric,addelementtypename=Beam2DProperties,
//texdescription="Beam2DProperties defines material and geometric properties for beam structural finite elements in (x,y)-plane.",
//texdescriptionComments="First, specify the $\mathtt{cross\_section\_type}$ of the beam, which may be either rectangular (if set to $1$), circular (if set to $2$), or tubular (if set to $3$) or polygonal (if set to $4$). In either case the $\mathtt{cross\_section\_size}$ is a vector of $2$, $1$, $2$, or $2\,n$ entries, where $n$ confers to the number of vertices of a closed polygon. Then specify the stiffnesses and moments of inertias, as they are required by your beam and problem.",
//example="Beam2DProperties.txt"]
class Beam2DProperties: public MaterialGeneric, public BeamCrossSectionInfo, public ElasticParametersLinear, public ElastoplasticParametersPrandtlReuss
{
public:
	// constructing and copying
	Beam2DProperties() : MaterialGeneric(), BeamCrossSectionInfo(), ElasticParametersLinear(), ElastoplasticParametersPrandtlReuss() { DefaultConstructor(); }
	Beam2DProperties(MBS* mbs) : MaterialGeneric(mbs), BeamCrossSectionInfo(), ElasticParametersLinear(mbs), ElastoplasticParametersPrandtlReuss() { DefaultConstructor(); }
	Beam2DProperties(const Beam2DProperties& bp) { CopyFrom(bp); }
	virtual MaterialGeneric* GetCopy() const;
	
	// material name
	virtual mystr GetTypeName() const;
	
	// setting
	virtual void SetBeam2DProperties(int _beamCrossSectionType, Vector _beamCrossSectionSize, double _rho, double _EA, double _EIy, double _GAky=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0);  //$ PG 2016-2-11: recommended set function
	virtual void SetBeam2DPropertiesCM(int _beamCrossSectionType, Vector _beamCrossSectionSize, double _rho, double E, double nu, double _EA=0, double _EIy=0, double _GAky=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0);  //$ PG 2016-2-11: recommended set function for continuum mechanics formulation of beams
	virtual void SetBeam2DProperties(double _rho, double _EA, double _EIy, double _GAky=0, double _rhoA=0, double _rhoIx=0, double _rhoIy=0); //$ PG 2016-2-11: deprecated, old
	virtual void SetMaterialFromEDCData(); 
	
	// checking
	virtual int CheckConsistency(mystr& errorstr);
	
	// status information 
	virtual int IsInelasticMaterial() const { return YieldStress() > 0; }
	virtual int IsHardeningMaterial() const { return HardeningModule() > 0; }
	virtual int GetInelasticVariablesCount() const { return IsInelasticMaterial()*ElastoplasticParametersPrandtlReuss::GetInelasticVariablesCount(plane); }
	
	// initializing inelastic variables (called from each integration point of elements)
	virtual void InitializeInelasticVariables(Vector& inel_variables) const {}
	
	virtual const MBS* GetMBS() const { return MaterialGeneric::GetMBS(); }
	virtual MBS* GetMBS() { return MaterialGeneric::GetMBS(); }
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) { return 0; }
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) { return 0; }

	// required for edc auto-converter
	virtual void GetElementDataAuto(ElementDataContainer& edc); //fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc);  //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		    //automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	// stiffnesses and moments of inertia
	virtual const double& BeamRhoA() const {return beamRhoA;   }
	virtual double& BeamRhoA() {return beamRhoA;   }
	virtual const double& BeamRhoIx() const {return beamRhoIx;   }
	virtual double& BeamRhoIx() {return beamRhoIx;   }
	virtual const double& BeamRhoIy() const {return beamRhoIy;   }
	virtual double& BeamRhoIy() {return beamRhoIy;   }
	virtual const double& BeamEA() const {return beamEA;   }
	virtual double& BeamEA() {return beamEA;   }
	virtual const double& BeamEIy() const {return beamEIy; }
	virtual double& BeamEIy() {return beamEIy; }
	virtual const double& BeamGAky() const {return beamGAky;}
	virtual double& BeamGAky() {return beamGAky;}
	virtual const double& BeamGJkx() const {return beamGJkx;}
	virtual double& BeamGJkx() {return beamGJkx;}

	// some methods concerning plasticity:

	// AP Dez. 2012
	// Material law due to return mapping algorithm by Simo and Hughes, plane stress and strain case
	// plane strain case differs from ComputeInelasticVariablesFromStrain2D by definition of TangentModule
	// plane stress case differs from ComputeInelasticVariablesFromStrain2D substancially
	void ComputeInelasticVariablesFromStrainPlaneStress2D(Vector& inel_variables, const Matrix2D& strain) const;
	void ComputeInelasticVariablesFromStrainPlaneStrain2D(Vector& inel_variables, const Matrix2D& strain) const;
	void ComputeInelasticVariablesFromStrain1D(Vector& inel_variables, double strain) const;

	//$ PG 2016-2-11: [ seems to be old code (at least used in ANCFCable2D):
	//virtual double DoNonlinStepBeam2D(double& stress, double eps, double& epsplast, double epsplast_old) const;
	virtual double DoNonlinStepBeam2D(double& stress, double eps, double& epsplast, double& internalvar, 
		double epsplast_old, double internalvar_old, int use_tangent_stiffness=0) const;
	//$ PG 2016-2-11: ]

protected:
	virtual void DefaultConstructor();
	virtual void CopyFrom(const Beam2DProperties& s);
	
	//to be added to edc from base class BeamCrossSectionInfo
	//EDC int beamCrossSectionShape;		//$EDC$[varaccess,EDCvarname="cross_section_type",EDCfolder="",tooltiptext="1: rectangular, 2: circular, 3: tubular, 4: polygonal"]
	//EDC Vector beamCrossSectionSize;	//$EDC$[varaccess,EDCvarname="cross_section_size",EDCfolder="",variable_length_vector,tooltiptext="vector length of cross_section_size depends on cross_section_type: length 1 for circular cross section, length 2 for rectangular cross section (y and z extension) or tubular cross section (outer and inner diameter) , and length 2*n for polygonal cross section (p1y,p1z,p2y,p2z,...,pny,pnz)"];
	//EDC double beamCrossSectionArea;  //$EDC$[varaccess,readonly,EDCvarname="cross_section_area",EDCfolder="",tooltiptext="area of the beam's cross section"]
	//EDC double beamIx;  //$EDC$[varaccess,readonly,EDCvarname="Ix",EDCfolder="",tooltiptext="polar moment of inertia of the beams cross section"]
	//EDC double beamIy;  //$EDC$[varaccess,readonly,EDCvarname="Iy",EDCfolder="",tooltiptext="area moment of inertia of the beams cross section w.r.t. y-axis (2D-beam)"]

	// own members
  double beamEA;								//$EDC$[varaccess,EDCvarname="EA",EDCfolder="",tooltiptext="youngs modulus * area"]
	double beamEIy;								//$EDC$[varaccess,EDCvarname="EIy",EDCfolder="",tooltiptext="bending stiffness w.r.t. y-axis (2D-beam)"]
	double beamGAky;							//$EDC$[varaccess,EDCvarname="GAky",EDCfolder="",tooltiptext="shear stiffness including shear correction factor ky (2D-beam)"]
	double beamGJkx;							//$EDC$[varaccess,EDCvarname="GJkx",EDCfolder="",tooltiptext="torsional stiffness including shear correction factor kx"]
	double beamRhoA;							//$EDC$[varaccess,EDCvarname="RhoA",EDCfolder="",tooltiptext="density * area"]
	double beamRhoIx;							//$EDC$[varaccess,EDCvarname="RhoIx",EDCfolder="",tooltiptext="density * second area of moment w.r.t. x-axis"]
	double beamRhoIy;							//$EDC$[varaccess,EDCvarname="RhoIy",EDCfolder="",tooltiptext="density * second area of moment w.r.t. y-axis (2D-beam)"]

	// to be removed from edc of base class MaterialGeneric
	//EDC int plane_stress;							//$EDC$[varaccess,EDCvarname="plane_stress",remove,EDCfolder="Solid",int_bool,tooltiptext="for 2D-Elements only; 1: plane stress, 0: plane strain"]
	//EDC int plane;										//$EDC$[varaccess,EDCvarname="plane",EDCfolder="Solid",remove]

	//to be added to edc from base classes ElasticParametersLinear and ElastoplasticParametersPrandtlReuss
	//EDC double youngs_modulus;				//$EDC$[varaccess,EDCvarname="youngs_modulus",EDCfolder="Solid",tooltiptext="Youngs modulus"]
	//EDC double poisson_ratio;					//$EDC$[varaccess,EDCvarname="poisson_ratio",EDCfolder="Solid",tooltiptext="Poisson ratio"]
	//EDC	double yieldstress;						//$EDC$[varaccess,EDCvarname="yield_stress",EDCfolder="Inelasticity",tooltiptext="Yield Stress s_y, e.g., |dev s| <= s_y"]
	//EDC	double tangentmodule;					//$EDC$[varaccess,EDCvarname="tangent_module",EDCfolder="Inelasticity",tooltiptext="Modulus of hardening H"]
};
//$EDC$[endclass,Beam2DProperties]




























////////////////////////////////////////////////////////
// class MaterialThermoMechanic : public MaterialGeneric
////////////////////////////////////////////////////////

//$EDC$[beginclass,classname=MaterialThermoMechanic,addelementtypename=MaterialThermoMechanic,parentclassname=Material,
//texdescription="MaterialThermoMechanic is a basic material class extended with thermal properties such as thermal conductivity or heat capacity."]
class MaterialThermoMechanic : public Material
{
public:
	struct ThermalProperties
	{
	public:
		ThermalProperties(double specificHeatCapacity, const Matrix3D& thermalConductivity);
		double GetSpecificHeatCapacity() const;
		const Matrix3D& GetThermalConductivity() const;
		void SetSpecificHeatCapacity(double specificHeatCapacity);
		void SetThermalConductivity(const Matrix3D& thermalConductivity);
	private:
		double specific_heat_capacity;
		Matrix3D thermal_conductivity;
	};

	// constructing
	MaterialThermoMechanic(MBS* mbs);
	MaterialThermoMechanic(const Material& material, const ThermalProperties& temperatureProperties);
	
	// creates a material with temperature properties
	// the specific heat capacity is the heat capacity per mass
	MaterialThermoMechanic( const Material& material, double specificHeatCapacity, const Matrix3D& thermalConductivity);

	// required by MaterialGeneric
	virtual MaterialGeneric* GetCopy() const;
	virtual mystr GetTypeName() const {	return "MaterialThermoMechanic"; }
	virtual int IsInelasticMaterial() const { return 0; }
	virtual int GetInelasticVariablesCount() const { return 0; }
	virtual void InitializeInelasticVariables(Vector& inel_variables) const {}

	// returns the heat capacity per mass
	double GetSpecificHeatCapacity() const;
	
	// returns the thermal conductivity
	const Matrix3D& GetThermalConductivity() const;

	const ThermalProperties& GetThermalProperties() const;
	void SetThermalProperties(const ThermalProperties& temperatureProperties);
	void SetThermalProperties(double specificHeatCapacity, const Matrix3D& thermalConductivity);

	virtual void GetElementData(ElementDataContainer& edc);
	virtual void GetElementDataAuto(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc);
	virtual int SetElementDataAuto(ElementDataContainer& edc);
	virtual int CheckConsistency(mystr &errorstr);
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); // Read access to a single element variable 
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); // Write access to a single element variable

protected:
	ThermalProperties temperature_properties;
};
//$EDC$[endclass,MaterialThermoMechanic]


#endif