//#***************************************************************************************
//#
//# filename:             mesh.h
//#
//# author:               Dorninger Alexander
//#
//# generated:						September 2014
//# description:          introduces meshes to the MBS
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#ifndef MESH__H
#define MESH__H

#include "mbs_interface.h"
#include "ElementsAndModelsLibraryInterface.h"
#include "..\UtilityLib\elementdataaccess.h"
#include "element.h"
#include "node.h"

#include "mesh_aux.h"
#include "set.h"



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                                BASEMESH
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

typedef enum {
	TMBSMeshUnknown = 0,

	TMBSMeshStructural = 1,
	
	TMBSMeshSolid = 11

} TMBSMeshType;

//class MeshElement;
//class MeshNodes;
//class MeshComponent;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                                 BASEMESH
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// base class for all meshes, containting information to identify by name and type
// contains data shared by all types of meshes
class BaseMesh //$EDC$[beginclass,classname=BaseMesh]
{
public: // lifecycle I
	BaseMesh(): mbs(NULL),name(""),type(TMBSMeshUnknown) { Reset(); }
	BaseMesh(MBS* mbsi): mbs(mbsi),name(""),type(TMBSMeshUnknown) { Reset(); }
	BaseMesh(const BaseMesh& other) { CopyFrom(other); }
	~BaseMesh() { }

public: // lifecycle II
	BaseMesh(MBS* mbsi,mystr namei,TMBSMeshType typei): mbs(mbsi),name(namei),type(typei) { Reset(); }
	virtual BaseMesh* GetCopy()
	{
		BaseMesh* mc = new BaseMesh();
		mc->CopyFrom(*this);
		return mc;
	}
	virtual void CopyFrom(const BaseMesh& other)
	{
		mbs = other.mbs;
		name = other.name;
		type = other.type;
		flag_autocomputesurface           = other.flag_autocomputesurface;
		symmetric_nodes_mbs_numbers       = other.symmetric_nodes_mbs_numbers;
		symmetric_constraints_mbs_numbers = other.symmetric_constraints_mbs_numbers;
		periodic_nodes_mbs_numbers_1      = other.periodic_nodes_mbs_numbers_1;
		periodic_nodes_mbs_numbers_2      = other.periodic_nodes_mbs_numbers_2;
		periodic_constraints_mbs_numbers  = other.periodic_constraints_mbs_numbers;
	}
	virtual void Reset() {
		flag_autocomputesurface = 0;
	}

public: // access
	virtual const MBS* GetMBS() const { return mbs; }
	virtual MBS* GetMBS() { return mbs; }
	virtual void SetMBS(MBS* mbsi) { mbs = mbsi; }

	virtual const TMBSMeshType GetType() const { return type; }
	virtual void SetType(TMBSMeshType typei) { type = typei; }
	virtual TMBSMeshType& Type() { return type; }

	virtual const mystr GetMeshName() const { return name; }
	virtual void SetMeshName(mystr namei) { name = namei; }
	virtual mystr& MeshName() { return name; }

	virtual MeshComponent* GetSubordinateComponent(int i) { return GetMeshComponent(i); }
	virtual MeshComponent* GetMeshComponent(int i) { return meshcomponents(i); }
	virtual void SetMeshComponent(int i, MeshComponent* comp) { meshcomponents(i) = comp; }
	
	virtual BaseSet* GetSet(int i) { return meshsets(i); }

public: // operations and operators 
	int NSubs() { return NComp(); }
	int NComp() { return meshcomponents.Length(); }
	MeshComponent* GetComponent(int i) { return meshcomponents(i); }
	void DeleteComponent(int i) { meshcomponents.Erase(i); }
	int NSets() { return meshsets.Length(); }
	int AddSet(BaseSet* set) { return meshsets.Add(set->GetCopy()); }
	void DeleteSet(int i) { meshsets.Erase(i); }

public: // MESH COMPONENT GENERATION - PRIMITIVE COMPONENT
	virtual int AddComponent(MeshComponent* comp);
public: // MESH COMPONENT GENERATION - EXTERNAL SOURCE
	virtual int LoadMeshExternal(mystr& filetype, mystr& filename);
public: // MESH COMPONENT GENERATION - EXTERNAL TARGET
	virtual int WriteMeshExternal(int compnr, mystr& filetype, mystr& filename);
public: // MESH MODIFICATION - LINEAR TO QUADRATIC
	virtual int LinearToQuadratic(int compnr);
public: // MESH MODIFICATION - SPLIT HEXES
	virtual int SplitHexes(int compnr, int split);
public: // MESH MODIFICATION - REFINE
	virtual int Refine(int compnr, const TArray<int>& sdlvl, int method=0);
public: // MESH MODIFICATION - EXTRUDE
	virtual int Extrude(int compnr, int axis=3, int discr=1, double dist=1., double thick=0.1);
public: // MESH MODIFICATION - ROTATIONAL
	virtual int Rotational(int compnr, int axis=1, int angleseg=2, double angle=90.);
public: // MESH MODIFICATION - MIRROR
	virtual int Mirror(int compnr, int plane=1, double dist=0.);

public: // MESH MODIFICATION - TRANSFORM
	virtual int TransformXYZ(int compnr, Vector3D translate, Vector3D rotate=Vector3D(0.,0.,0.), Vector3D scale=Vector3D(1.,1.,1.));
public: // MESH MODIFICATION - DISTORTION
	virtual int DistortXYZ(int compnr, mystr fx="x", mystr fy="y", mystr fz="z", int subordinate=0);
public: // HESH MODIFICATION - MODIFY
	virtual int Modify(int compnr, MeshModifier::WorkingOnEntity entity, int newIntVal, double newDoubleVal, mystr& newStringVal, 
		                                                                 int oldIntVal, double oldDoubleVal, mystr& oldStringVal, BaseSet* filter);
	
public: // MESH MODIFICATION - GLUE
	virtual int Glue(TArray<int> compnr,int use_penalty=0);


public: // MESH INSTANTIATION - ADD TO MBB
	int MeshToMBS(MBS* mbs);
private:
	int NodesToMBS(MBS* mbs);                         // all MeshNodes are instantiated in the MBS
	int ElemsToMBS(MBS* mbs);                         // all MeshElements are instantiated in the MBS 
	int SetsToMBS(MBS* mbs);                          // all sets added to the MBS ( with index numbers adjusted to MBS order )
	int UpdateStatistics();                           // compute current Domain Statistics for the Mesh after adding to the MBS
   	int AppyAutomaticConstraints(MBS* mbs);           // automatically create constraints for coincident pairs, component intersections, 

public: // MESH INSTANTIATION - COMPUTE MESH 
	//NOTE: the list of components could contain NULL entries ( remnants of Glue operations )
	int MeshAllComponents();
	int MeshSingleComponent(int compnr);
	int NNodes();
	int NElems();
	int NCoincident();
public: // meshing helper functions - indirect access - resolve a string to ObjectFactoryTypeIndex
	int GetComponentNodeTypeIdx(MeshComponent* comp); // use object factory only in Mesh, not in components
	int GetComponentElemTypeIdx(MeshComponent* comp); // use object factory only in Mesh, not in components
	int GetComponentNodeTypeIdx(mystr& typestr); // use object factory only in Mesh, not in components
	int GetComponentElemTypeIdx(mystr& typestr); // use object factory only in Mesh, not in components

public: // MESH - MBS INTERFACE - NODES TO ELEMENT LIST
	TArray<TArray<int>*>& N2E()
	{
		if(node2element.Length() != NNodes())
		{
			if(node2element.Length()!=0)	
				ClearNodesToElementsList();
			AllocateNodesToElementsList();
			AssembleNodesToElementsList();
		}
		return node2element;
	}
	int ClearNodesToElementsList();
	int AllocateNodesToElementsList();
	int AssembleNodesToElementsList();

public: // MESH PICKING
	MeshNode* GetNodePtr(int nodenumber);
	MeshElement* GetElementPtr(int elemenetnumber);
	TArray<int> GetNodesInBox(Box3D box);
	TArray<int> GetNodesInCylinderShell(Vector3D center_bottom, Vector3D center_top, double rad_outer, double rad_inner=0.);
	TArray<int> GetNodesInSphericalShell(Vector3D center, double rad_outer, double rad_inner=0.);
	TArray<int>GetNodesInMathFunction(MathFunction& mf);
	TArray<int> GetElementsInBox(Box3D box); 
	int GetElementAtPos(Vector3D pos);
	Vector3D GetLocalPosOnElem(int elnr, Vector3D globpos);
	Vector3D GetGlobalPosOfLocalPosOnElem(int elnr, Vector3D ploc);
public:
	int GetFacesFromNodes(const TArray<int>& nodenumbers, const TArray<int>& elementnumbers);
	int ComputeSurface(FaceSetA& rv_surface, const TArray<int>& filter_elementnumbers); // attention Domain borders are always surfaces, no nodes shared, just coincident nodes
	void SetFlagAutoComputeSurface(int flag) { flag_autocomputesurface = flag; }
private:
	int flag_autocomputesurface;

private: 
	int WarnSkipOperationIfNoComponents(mystr& caller);
	virtual int ComputeBoundingBox() { return 0; }

	// these functions create the nodes and elements for the MBS and thus require a valid mesh
public: // instantiation

private: // 

public:
	int MapNodeListToMBSNodes(TArray<int>& nodelist);
	int MapElemListToMBSElems(TArray<int>& elemlist);
	
public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(name.c_str(),"mesh_name"); ed.SetToolTipText("the name of the mesh"); edc.Add(ed); // name of the mesh, may be changed
		ed.SetText(GetMeshSpecification().c_str(),"mesh_type"); ed.SetToolTipText("type of the mesh");  /*ed.SetLocked(1);*/ edc.Add(ed); // type of the mesh 

		for(int i=1; i<=NComp(); i++)
		{
			ElementDataContainer edc_comp;
			MeshComponent* comp = GetMeshComponent(i);
			if(comp!=NULL)
			{
				comp->GetElementData(edc_comp);
				ed.SetEDC(&edc_comp,comp->GetName());
				ed.SetLocked(1);
				edc.Add(ed);
			}
		}

		ed.SetInt(NNodes(),"number_of_nodes"); ed.SetToolTipText("number of nodes in the mesh"); ed.SetLocked(1); edc.Add(ed); // number of nodes
		Vector nodenums(NNodes()); for(int i=1; i<=NNodes(); i++) { nodenums(i) = mbsnodenrs(i); }
		ed.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "list_of_nodes", 1 , NNodes()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS node numbers of the entire mesh"); edc.Add(ed);

		ed.SetInt(NElems(),"number_of_elements"); ed.SetToolTipText("number of elements in the mesh"); ed.SetLocked(1); edc.Add(ed); // number of elements
		Vector elemnums(NElems()); for(int i=1; i<=NElems(); i++) { elemnums(i) = mbselemnrs(i); }
		ed.SetVector(elemnums.GetVecPtr(), elemnums.Length(), "list_of_elements", 1 , NElems()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the entire mesh"); edc.Add(ed);

		ed.SetBool(flag_autocomputesurface,"compute_surface"); ed.SetToolTipText("flag to automatically identify domain surfaces"); edc.Add(ed); // auto compute surface flag
		for(int i=1; i<=domainstatistics.Length(); i++)
		{
			ElementDataContainer edc_sub;
			domainstatistics(i)->GetElementData(edc_sub);
			ed.SetEDC(&edc_sub,mystr("Domain")+mystr(i));
			edc.TreeAdd("Stats",ed);		
		}

		for(int i=1, n=NSets(); i<=n; ++i)
		{
			//face Set added to Subcontainer BoundaryConditions
			if(GetSet(i)->GetType()==TMBSSetType::TMBSSetFaces)
			{
				ElementDataContainer edc_bc;
				GetSet(i)->GetElementData(edc_bc);
				ed.SetEDC(&edc_bc,GetSet(i)->GetName());
				edc.TreeAdd("BoundaryConditions",ed);		
			}
		}

		ed.SetVector(symmetric_nodes_mbs_numbers.GetVecPtr(), symmetric_nodes_mbs_numbers.Length(), "node_numbers"); ; ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("global node numbers of the nodes lying on a symmetry boundary"); ed.SetLocked(true); edc.TreeAdd("BoundaryConditions.MeshApplied.Symmetry",ed);
		ed.SetVector(symmetric_constraints_mbs_numbers.GetVecPtr(), symmetric_constraints_mbs_numbers.Length(), "element_numbers"); ; ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("global element numbers of the symmetry boundary constraints"); ed.SetLocked(true); edc.TreeAdd("BoundaryConditions.MeshApplied.Symmetry",ed);

		ed.SetVector(periodic_nodes_mbs_numbers_1.GetVecPtr(), periodic_nodes_mbs_numbers_1.Length(), "node_numbers_1"); ; ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("global node numbers of the nodes lying on a periodic boundary (side 1)"); ed.SetLocked(true); edc.TreeAdd("BoundaryConditions.MeshApplied.Periodic", ed);
		ed.SetVector(periodic_nodes_mbs_numbers_2.GetVecPtr(), periodic_nodes_mbs_numbers_2.Length(), "node_numbers_2"); ; ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("global node numbers of the nodes lying on a periodic boundary (side 2)"); ed.SetLocked(true); edc.TreeAdd("BoundaryConditions.MeshApplied.Periodic", ed);
		ed.SetVector(periodic_constraints_mbs_numbers.GetVecPtr(), periodic_constraints_mbs_numbers.Length(), "element_numbers"); ; ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("global element numbers of the periodic boundary constraints"); ed.SetLocked(true); edc.TreeAdd("BoundaryConditions.MeshApplied.Periodic", ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		name = edc.TreeGetString("mesh_name");
		mystr type_old = GetMeshSpecification();
		mystr type_new = edc.TreeGetString("mesh_type");
		if(!type_new.Compare(type_old))
		{
			GetMBS()->UO().InstantMessageText("ERROR: You MUST NOT change the type of the set!");
			return 0;
		}
		if(edc.Find("compute_surface"))
		{
			flag_autocomputesurface = edc.TreeGetBool("compute_surface");
		}
		return rv;
	}
// class specific
	virtual mystr GetMeshSpecification() { return mystr("abstract"); } //$EDC$[funcaccess,EDCvarname"mesh_type",tooltiptext="specification of mesh type. Once the mesh is added to the MBS, you MUST NOT change this type anymore"]
	virtual void GetAvailableScriptFunctions(ElementDataContainer& edc_available_functions)
	{
		ElementData ed;
//		ed.SetText("returns the number of nodes in the Mesh","NNodes"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
//		ed.SetText("returns the number of elementss in the Mesh","NElements"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
	
		ed.SetText("loads a mesh from an external file into a new Mesh component","LoadMesh"); ed.SetLocked(1); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("writes a mesh to an external file","WriteMesh"); ed.SetLocked(1); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);

		ed.SetText("allows to change properties for mesh elements ","Modify"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
//		ed.SetText("applies a processing operation on an entire Mesh or a single Mesh component","Process"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);

		ed.SetText("applies a transformation on an entire Mesh or a single Mesh component","Transform"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("applies a distortion on an entire Mesh or a single Mesh component","Distort"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("converts elements of an entire Mesh or a single Mesh component","Linear2Quadratic"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("converts elements of an entire Mesh or a single Mesh component","SplitHexes"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("applies a refinement operation on a single subordinate component","Refine"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("rotates a 2d mesh around a given axis","Rotate"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("mirrors a mesh at a given plane","Mirror"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("extrudes a 1d or 2d mesh along a given axis","Extrude"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);

		ed.SetText("creates an instance of the mesh in the MBS","AddMeshToMBS"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);

		ed.SetText("remove redundant nodes or create constraints","GlueMesh"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns nodes in the defined box","GetNodesInBox"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns elements fully contained in the defined box","GetElementsInBox"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns nodes in the cylinder shell","GetNodesInCylinder"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns nodes in the shperical shell","GetNodesInSphere"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns nodes in the region defined by the math function","GetNodesInFunction"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns global position of the node","GetNodePos"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);

		ed.SetText("returns element and local position","GetElementAtPosition"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("returns lists of element numbers and element face numbers that can be constructed from the given nodeset","GetFacesFromNodes"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
	}

protected:
	MBS* mbs;
	mystr name;           //$EDC$[varaccess,EDCvarname="name",EDCfolder="",tooltiptext="name of the mesh"]
	TMBSMeshType type;
protected:
	TArray<MeshComponent*> meshcomponents;
	TArray<BaseSet*> meshsets;
protected:
	TArray<int> mbsnodenrs;
	TArray<int> mbselemnrs;
protected: // domain statistics
	struct DomainStatistics 
	{
		DomainStatistics(MBS* mbsi) { surface = new FaceSetA(mbsi); }
		~DomainStatistics() { delete surface; }

		virtual void GetElementData(ElementDataContainer& edc)
		{
			ElementData ed;
			
			ed.SetInt(nodes.Length(),"number_of_nodes"); ed.SetToolTipText("number of nodes in the domain"); ed.SetLocked(1); edc.Add(ed); // number of nodes
			Vector nodenums(nodes.Length()); for(int i=1; i<=nodes.Length(); i++) { nodenums(i) = nodes(i); }
			ed.SetVector(nodenums.GetVecPtr(), nodenums.Length(), mystr("list_of_nodes"), 1, nodes.Length());
			ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS node numbers of the domain"); edc.Add(ed); // list of nodes	

			ed.SetInt(elements.Length(),"number_of_elements"); ed.SetToolTipText("number of elements in the domain"); ed.SetLocked(1); edc.Add(ed); // number of elements
			Vector elemnums(elements.Length()); for(int i=1; i<=elements.Length(); i++) { elemnums(i) = elements(i); }
			ed.SetVector(elemnums.GetVecPtr(), elemnums.Length(), mystr("list_of_elements"), 1, elements.Length());
			ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the domain"); edc.Add(ed); // list of elements	

			if(surface->GetNrOfEntries()>0)
			{
				ElementDataContainer edc_surface;
				surface->GetElementData(edc_surface);
				ed.SetEDC(&edc_surface,"Surface");
				edc.Add(ed);

				ed.SetInt(surfacenodes.Length(),"number_of_surface_nodes"); ed.SetToolTipText("number of nodes in the surface"); ed.SetLocked(1); edc.TreeAdd("Surface",ed); // number of nodes
				Vector surfacenodenums(surfacenodes.Length()); for(int i=1; i<=surfacenodes.Length(); i++) { surfacenodenums(i) = surfacenodes(i); }
				ed.SetVector(surfacenodenums.GetVecPtr(), surfacenodenums.Length(), mystr("list_of_surface_nodes"), 1, surfacenodes.Length());
				ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS node numbers of the surface"); edc.TreeAdd("Surface",ed); // list of nodes	

				ed.SetInt(surfaceelements.Length(),"number_of_surface_elements"); ed.SetToolTipText("number of elements in the surface"); ed.SetLocked(1); edc.TreeAdd("Surface",ed); // number of elements
				Vector surfaceelementnums(surfaceelements.Length()); for(int i=1; i<=surfaceelements.Length(); i++) { surfaceelementnums(i) = surfaceelements(i); }
				ed.SetVector(surfaceelementnums.GetVecPtr(), surfaceelementnums.Length(), mystr("list_of_surface_elements"), 1, surfaceelements.Length());
				ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the surface"); edc.TreeAdd("Surface",ed); // list of elements
			}
		}

		TArray<int> nodes;                              // contains node MBS numbers 
		TArray<int> elements;                           // contains element MBS number
		FaceSetA* surface;								
		TArray<int> surfaceelements;
		TArray<int> surfacenodes;
		
	};
	TArray<DomainStatistics*> domainstatistics;

protected: // buffer variables
	TArray<TArray<int>*> node2element;

	virtual int ApplyPeriodicBCs(MBS* mbs);
	virtual int ApplySymmetryBCs(MBS* mbs);

private:
	Vector symmetric_nodes_mbs_numbers;
	Vector symmetric_constraints_mbs_numbers;
	Vector periodic_nodes_mbs_numbers_1;
	Vector periodic_nodes_mbs_numbers_2;
	Vector periodic_constraints_mbs_numbers;



}; //$EDC$[endclass,BaseMesh]

//$EDC$[beginclass,classname=StructuralMesh,parentclassname=BaseMesh,
//texdescription="A mesh for structural elements",example="Mesh.txt"]
class StructuralMesh : public BaseMesh 
{
public: // lifecycle I
	StructuralMesh(MBS* mbsi): BaseMesh(mbsi,"StructuralMesh",TMBSMeshStructural) {;}
	StructuralMesh(const StructuralMesh& other) { CopyFrom(other); }
	~StructuralMesh() {}
public: // lifecycle II
	StructuralMesh(MBS* mbsi, mystr namei): BaseMesh(mbsi,namei,TMBSMeshStructural) {;}
	BaseMesh* GetCopy()
	{
		BaseMesh* mc = new StructuralMesh(*this);
		return mc;
	}
	void CopyFrom(const StructuralMesh& other)
	{
		// copy base class content
		BaseMesh::CopyFrom(other);	
		// specific for each derived class
	}
public: // access
public: // operations and operators
public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseMesh::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		BaseMesh::SetElementData(edc);
		return rv;
	}
	virtual mystr GetMeshSpecification() { return mystr("StructuralMesh"); }
	virtual void GetAvailableScriptFunctions(ElementDataContainer& edc_available_functions)
	{
		BaseMesh::GetAvailableScriptFunctions(edc_available_functions);	

		ElementData ed;
		ed.SetText("generates discretized Beams","GenerateBeam"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("generates discretized Plates","GeneratePlate"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
	}
protected:
}; //$EDC$[endclass,StructuralMesh]

//$EDC$[beginclass,classname=SolidMesh,parentclassname=BaseMesh,
//texdescription="A mesh for solid elements",example="Mesh.txt"]
class SolidMesh : public BaseMesh 
{
public: // lifecycle I
	SolidMesh(MBS* mbsi): BaseMesh(mbsi,"SolidMesh",TMBSMeshSolid) {;}
	SolidMesh(const SolidMesh& other) { CopyFrom(other); }
	~SolidMesh() {}
public: // lifecycle II
	SolidMesh(MBS* mbsi, mystr namei): BaseMesh(mbsi,namei,TMBSMeshSolid) {;}
	BaseMesh* GetCopy()
	{
		BaseMesh* mc = new SolidMesh(*this);
		return mc;
	}
	void CopyFrom(const SolidMesh& other)
	{
		// copy base class content
		BaseMesh::CopyFrom(other);	
		// specific for each derived class
	}
public: // access
public: // operations and operators
//public: // generation
//	virtual int Mesh(int compnr); 

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseMesh::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		BaseMesh::SetElementData(edc);
		return rv;
	}
	virtual mystr GetMeshSpecification() { return mystr("SolidMesh"); }
	virtual void GetAvailableScriptFunctions(ElementDataContainer& edc_available_functions)
	{
		BaseMesh::GetAvailableScriptFunctions(edc_available_functions);	

		ElementData ed;
		ed.SetText("generates discretized Blocks","GenerateBlock"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
		ed.SetText("generates discretized Cylinder","GenerateCylinder"); ed.SetLocked(1); ed.SetToolTipText("This Command is registerd for a " + GetMeshSpecification()); edc_available_functions.Add(ed);
	}
protected:
}; //$EDC$[endclass,SolidMesh]

#endif