//#***************************************************************************************
//#
//# filename:             mesh_aux.h
//#
//# author:               Dorninger Alexander
//#
//# generated:						September 2014
//# description:          introduces meshes to the MBS
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#ifndef MESHAUX__H
#define MESHAUX__H

#include "mbs_interface.h"
#include "..\UtilityLib\elementdataaccess.h"
#include "elementdata.h"
#include "finite_element_definitions.h" // GeometricNonlinearityStatus
#include <map>
#include <vector>

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// BASEMESHELEMENT
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
enum MeshElementType 
{ 
	MELine=102,  METrig=203,  MEQuad=204,  METet=304,  MEPyr=305,  MEPri=306,  MEHex=308,
	MEQLine=103, MEQTrig=206, MEQQuad=208, MEQTet=310, MEQPyr=313, MEQPri=315, MEQHex=320
};

class MeshElement //$EDC$[beginclass,classname=MeshElement]
{
public: // lifecycle I
	MeshElement() {;}
	MeshElement(const MeshElement& other) { CopyFrom(other); }
	~MeshElement() {}

public: // lifecycle II
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshElement();
		mec->CopyFrom(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshElement& other)
	{
		nodes.CopyFrom(other.nodes);
		typenr = other.typenr;
		matnr = other.matnr;
	}
	virtual void Reset()
	{
		nodes.Flush();
		typenr = 1;
		matnr = 1;
	}
public: // access
	virtual int GetNode(int localnode) { return nodes(localnode); }
	virtual void SetNode(int localnode, int globalnode) { nodes(localnode) = globalnode; }
	virtual int NNodes() { return nodes.Length(); }
	virtual TArray<int> GetNodeList() { return nodes; }
	virtual void SetNodeList(TArray<int>& n) { nodes = n; }
	virtual TArray<int>& NodeList() { return nodes; }
	virtual TArray<int> GetFaceNodeList(int f) { return GetFace(f).NodeList(); }
	virtual int ContainsNode(int globalnode); // in cpp

	virtual int GetTypeNr() { return typenr; }
	virtual void SetTypeNr(int t) { typenr = t; }
	virtual int& TypeNr() { return typenr; }
	virtual int GetMatNr() { return matnr; }
	virtual void SetMatNr(int m) { matnr = m; }
	virtual int& MatNr() { return matnr; }

public: // operations and operators
	virtual int IsMeshHex()  { return 0; }
	virtual int IsMeshTet()  { return 0; }
	virtual int IsMeshPri()  { return 0; }
	virtual int IsMeshPyr()  { return 0; }
	virtual int IsMeshQuad()  { return 0; }
	virtual int IsMeshTrig() { return 0; }
	virtual int IsMeshLine()  { return 0; }
	virtual int GetMEType() { return 0; }
	virtual int Dim() { return 0; }
	virtual int Invert() { return 0; }

public: // identity, mirrorsymmetry
	virtual int IsSameBasicType(MeshElement& other); // in cpp                       // ensures only elements of same basic type but different order are compared
	virtual int IsCyclicEqual(MeshElement& other); // in cpp                         // compares nodelist, returns 0 for no match, >0 for matching elements, 1 for same orientaiton 
	virtual int NOrientations() { return 0; }                                                                                            // returns number of unique orientations for the given element type
	virtual int GetNodeListOriented(int orientation,TArray<int>& nodelist, TArray<int>& inverse) { return 0; }                           // returns nodelist - permutation of the node numbers such that the element in gien orientation is depicted AND also the inverse permutaion

public: // faces, edges, ...
	virtual int NFaces() { return 0; }
	virtual MeshElement GetLocalFace(int i) { return MeshElement(); }                       // returns element face with local node numbers (element specific)
	virtual MeshElement GetFace(int i)							                            // returns element face with global node numbers inserted
	{
		MeshElement localface = GetLocalFace(i);
		for(int i=1; i<=localface.NNodes(); i++)
		{
			localface.SetNode(i, GetNode(localface.GetNode(i)));
		}
		return localface;
	}
public: // higher order 
	virtual int NInter() { return 0; }
	virtual int2 GetInterPair(int i) { return int2(0,0); }               // returns a pair of nodenumers to interpolate position of an intermediate node (lin2quad)

public: // interpolation - redundant to MBS-Element
	virtual Vector3D InterpolateVector3D(Vector3D ploc, const TArray<Vector3D>& nodalvalues);     // interpolates the vector values at the nodes for a given local position depending on the element order
	virtual double InterpolateScalar(Vector3D ploc, TArray<double>& nodalvalues);           // interploates the scalar calues at the nodes for a given local position depending on the element order

	virtual Vector3D GetLocPosOfGlobalPos(Vector3D pglob, TArray<Vector3D>& nodalposition); // determine the local position of a global position
	virtual Matrix3D GetJacobi(Vector3D ploc, TArray<Vector3D>& nodalposition);                                
	
	virtual Vector3D GetLocPosOfCenter() { return Vector3D(); }       // override for all derived element types
	virtual Vector3D GetLocPosOfIthNode(int i) { return Vector3D(); } // override for all derived element types

	virtual double GetS0(Vector3D ploc, int shape) { return 0.; }     // override for all derived element types                                 
	virtual Matrix GetDSMatrix0(Vector3D ploc) { return Matrix(); }   // override for all derived element types 

public: // refinement - template
	virtual int3 IdentifyRequiredTemplate(TArray<int>& lvl, int method=0) { return int3(0,0,0); }                                        // computes pattern & orientation of the required template
	virtual int IsTemplateImplemented(TArray<int>& lvl, int method=0) { return 1; }                                                      // checks if template is implemented, provides node subdividsion level for alternative template
public: // refinement - orientation
	virtual Vector3D ApplyOrientation(Vector3D vec0, int orientation) { return Vector3D(); }                                             // applies orientation to a Vector
	virtual Matrix3D GetOrientationMatrix(int orientation) { return Matrix3D(); }                                                        // provides a Rotation-Matrix  for each orientaion
public: // refinement - node generation
	virtual int GetNumberOfAdditionalNodesForRefine(int3 pattern, int method=0) { return 0; }                                            // counts the additional nodes required for the template replacement (NOT conting the nodes keps from unrefined)
	virtual int GetGridNumberOfIthAdditionalNode(int i, int3 pattern, int method=0) { return 0; }                                        // gives the grid index number of the additional node
	virtual Vector3D GetLocPosOfIthAdditionalNode(int i, int3 pattern, int method=0) { return Vector3D(); }                              // computes local positions for required additional nodes for refinement
	virtual int GetSubdivisionLevelOfIthAdditionalNode(int i, int3 pattern, TArray<int>& cornersdlvls, int method=0) { return 0; }       // computes node subdivision level for newly generated node
public: // refinement - elements
	virtual int GetNumberOfElementsForRefinement(int3 pattern, int method=0) { return 0; }                                               // counts the elements required for the template replacement ( parent element is not changed - although it gets replaced when its a result of a previous iteration of this refinement )
	virtual int GetNodeListOfIthRefinedElement(int i, int3 pattern, TArray<int>& nodelist, int method=0) { return 0; }                   // computes the nodelist for the refined element
	virtual Vector3D GetLocPosCenterOfIthRefinedElement(int i, int3 pattern, int method=0) { return Vector3D(); }                        // position of center in local coordinates of parent element

public: // local position to global position mapping
	//virtual Vector3D ComputeLocPos(Vector3D pglob, TArray<Vector3D>& noderefposlist); // returns the local coordinates of a input global position for an element with input global node positions (type dependent) 
	//virtual Vector3D ComputeGlobPos(Vector3D ploc, TArray<Vector3D>& noderefposlist); // returns the global coordinates of a input local position for an element with input global node positions (type dependent) 
	//virtual Vector3D GetGlobPosOfLocPos000(TArray<Vector3D>& noderefposlist) { return Vector3D(); }         // override for all derived element typess
	//virtual Matrix3D GetElementAxisVectorMatrix(TArray<Vector3D>& noderefposlist) { return Matrix3D(); }    // override for all derived element types
	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.) { return 0; }              // returns 1 if local position is within bounds of the element type    

public: // helper function
	virtual int SwapNodeNumberEntries(int i, int j)
	{
		if(i==j) { return 1; }
		else if(i<1 || i>NNodes()) { return -1; }
		else if(j<1 || j>NNodes()) { return -1; }
		else
		{
			int tmp = nodes(i); 
			nodes(i) = nodes(j);
			nodes(j) = tmp;
			return 1;
		}
	}

public: // EDC-interaction
protected: //
	TArray<int> nodes;                // list of nodes
	int typenr;						  // type of the element as in ObjectFactory
	int matnr;                        // material number in MBS
}; //$EDC$[endclass,MeshElement]

class MeshHex : public MeshElement
{
public: // lifecycle I
	MeshHex() { Reset(); }
	MeshHex(const MeshHex& other) { CopyFrom(other); }
	~MeshHex() {}

public: // lifecycle II
	MeshHex(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshHex(int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8)
	{
		nodes.SetXN(8, n1,n2,n3,n4,n5,n6,n7,n8); 
	}
	MeshHex(int n1, int n2, int n3, int n4, int n5, int n6, int n7, int n8,
		int n9, int n10, int n11, int n12, int n13, int n14, int n15, int n16, int n17, int n18, int n19, int n20)
	{
		nodes.SetXN(20, n1,n2,n3,n4,n5,n6,n7,n8, n9,n10,n11,n12,n13,n14,n15,n16,n17,n18,n19,n20); 
	}
		    
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshHex(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshHex& other)
	{
		MeshElement::CopyFrom(other);

		const MeshHex& o = (const MeshHex&) other;
// specific for derived
	}
	virtual int IsMeshHex()  { return 1; }
	virtual int GetMEType() { if(NNodes()==8) return MEHex; else if(NNodes()==20) return MEQHex; else return 0; }
	virtual int Dim() { return 3; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 24; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 6; }
	virtual MeshElement GetLocalFace(int i); //in cpp

	virtual int NInter() { return 12; }
	virtual int2 GetInterPair(int i); // in cpp

	virtual Vector3D GetLocPosOfCenter() { return Vector3D(0.,0.,0.); }
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); // in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp

	virtual int3 IdentifyRequiredTemplate(TArray<int>& lvl, int method=0); //in cpp
	virtual int IsTemplateImplemented(TArray<int>& lvl, int method=0); // in cpp

	virtual Vector3D ApplyOrientation(Vector3D vec0, int orientation); // in cpp
	virtual Matrix3D GetOrientationMatrix(int orientation); // in cpp

	virtual int GetNumberOfAdditionalNodesForRefine(int3 pattern, int method=0); // in cpp
	virtual int GetGridNumberOfIthAdditionalNode(int i, int3 pattern, int method=0); // in cpp
	virtual Vector3D GetLocPosOfIthAdditionalNode(int i, int3 pattern, int method=0); // in cpp
	virtual int GetSubdivisionLevelOfIthAdditionalNode(int i, int3 pattern, TArray<int>& cornersdlvls, int method=0); // in cpp
	
	virtual int GetNumberOfElementsForRefinement(int3 pattern, int method=0); // in cpp
	virtual int GetNodeListOfIthRefinedElement(int i, int3 pattern, TArray<int>& nodelist, int method=0); // in cpp
	virtual Vector3D GetLocPosCenterOfIthRefinedElement(int i, int3 pattern, int method=0); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps=0.); // in cpp
};

class MeshTet : public MeshElement
{
public: // lifecycle I
	MeshTet() { Reset(); }
	MeshTet(const MeshTet& other) { CopyFrom(other); }
	~MeshTet() {}

public: // lifecycle II
	MeshTet(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshTet(int n1, int n2, int n3, int n4)
	{
		nodes.SetXN(4, n1,n2,n3,n4); 
	}
	MeshTet(int n1, int n2, int n3, int n4, 
		int n5, int n6, int n7, int n8,int n9, int n10)
	{
		nodes.SetXN(10, n1,n2,n3,n4, n5,n6,n7,n8,n9,n10); 
	}
		
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshTet(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshTet& other)
	{
		MeshElement::CopyFrom(other);

		const MeshTet& o = (const MeshTet&) other;
// specific for derived
	}

	virtual int IsMeshTet()  { return 1; }
	virtual int GetMEType() { if(NNodes()==4) return METet; else if(NNodes()==10) return MEQTet; else return 0; }
	virtual int Dim() { return 3; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 12; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 4; }
	virtual MeshElement GetLocalFace(int i); //in cpp
	
	virtual int NInter() { return 6; }
	virtual int2 GetInterPair(int i); // in cpp	
	
	virtual Vector3D GetLocPosOfCenter() { return Vector3D(1./4.,1./4.,1./4.); } // TetrahedralGeneric::GetRefPos
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); //in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

class MeshPyr : public MeshElement
{
public: // lifecycle I
	MeshPyr() { Reset(); }
	MeshPyr(const MeshPyr& other) { CopyFrom(other); }
	~MeshPyr() {}

public: // lifecycle II
	MeshPyr(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshPyr(int n1, int n2, int n3, int n4, int n5)
	{
		nodes.SetXN(5, n1,n2,n3,n4,n5); 
	}
	MeshPyr(int n1, int n2, int n3, int n4, int n5,
		int n6, int n7, int n8, int n9, int n10, int n11, int n12, int n13)
	{
		nodes.SetXN(13, n1,n2,n3,n4,n5, n6,n7,n8,n9,n10,n11,n12,n13); 
	}
		
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshPyr(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshPyr& other)
	{
		MeshElement::CopyFrom(other);

		const MeshPyr& o = (const MeshPyr&) other;
// specific for derived
	}
	virtual int IsMeshPyr() { return 1; }
	virtual int GetMEType() { if(NNodes()==5) return MEPyr; else if(NNodes()==13) return MEQPyr; else return 0; }
	virtual int Dim() { return 3; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 16; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 5; }
	virtual MeshElement GetLocalFace(int i); //in cpp

	virtual int NInter() { return 8; }
	virtual int2 GetInterPair(int i); // in cpp	
	
	virtual Vector3D GetLocPosOfCenter() { return Vector3D(3./8.,3./8.,1./4.); } // PyramidGeneric::GetRefPos
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D plox, int shape); // in cpp 
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

class MeshPri : public MeshElement
{
public: // lifecycle I
	MeshPri() { Reset(); }
	MeshPri(const MeshPri& other) { CopyFrom(other); }
	~MeshPri() {}

public: // lifecycle II
	MeshPri(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshPri(int n1, int n2, int n3, int n4, int n5, int n6)
	{
		nodes.SetXN(6, n1,n2,n3,n4,n5,n6); 
	}
	MeshPri(int n1, int n2, int n3, int n4, int n5, int n6, 
		int n7, int n8, int n9, int n10, int n11, int n12, int n13, int n14, int n15)
	{
		nodes.SetXN(15, n1,n2,n3,n4,n5,n6, n7,n8,n9,n10,n11,n12,n13,n14,n15); 
	}		
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshPri(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshPri& other)
	{
		MeshElement::CopyFrom(other);

		const MeshPri& o = (const MeshPri&) other;
// specific for derived
	}
	virtual int IsMeshPri() { return 1; }
	virtual int GetMEType() { if(NNodes()==6) return MEPri; else if(NNodes()==15) return MEQPri; else return 0; }
	virtual int Dim() { return 3; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 18; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 5; }
	virtual MeshElement GetLocalFace(int i); //in cpp

	virtual int NInter() { return 9; }
	virtual int2 GetInterPair(int i); // in cpp

	virtual Vector3D GetLocPosOfCenter() { return Vector3D(1./3.,1./3.,1./2.); } // PrismGeneric::GetRefPos
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); // in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

class MeshQuad : public MeshElement
{
public: // lifecycle I
	MeshQuad() { Reset(); }
	MeshQuad(const MeshQuad& other) { CopyFrom(other); }
	~MeshQuad() {}

public: // lifecycle II
	MeshQuad(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshQuad(int n1, int n2, int n3, int n4)
	{
		nodes.SetXN(4, n1,n2,n3,n4); 
	}
	MeshQuad(int n1, int n2, int n3, int n4, 
		int n5, int n6, int n7, int n8)
	{
		nodes.SetXN(8, n1,n2,n3,n4, n5,n6,n7,n8);
	}
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshQuad(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshQuad& other)
	{
		MeshElement::CopyFrom(other);

		const MeshQuad& o = (const MeshQuad&) other;
// specific for derived
	}
	virtual int IsMeshQuad()  { return 1; }
	virtual int GetMEType() { if(NNodes()==4) return MEQuad; else if(NNodes()==8||NNodes()==9) return MEQQuad; else return 0; }
	virtual int Dim() { return 2; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 4; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 4; }
	virtual MeshElement GetLocalFace(int i); //

	//virtual int NInter() { return 5; } // ANCF Plate ?
	virtual int NInter() { return 4; } // FE2DQuadrilateral
	virtual int2 GetInterPair(int i); // in cpp

	virtual Vector3D GetLocPosOfCenter() { return Vector3D(0.0,0.0,0.0); }
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); // in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp
	
	virtual int3 IdentifyRequiredTemplate(TArray<int>& lvl, int method=0); //in cpp
	virtual int IsTemplateImplemented(TArray<int>& lvl, int method=0) { return 1; } // all templates for 3-refinement are implemented
	virtual Vector3D ApplyOrientation(Vector3D vec0, int orientation); // in cpp
	virtual Matrix3D GetOrientationMatrix(int orientation); // in cpp

	virtual int GetNumberOfAdditionalNodesForRefine(int3 pattern, int method=0); // in cpp
	virtual int GetGridNumberOfIthAdditionalNode(int i, int3 pattern, int method=0); // in cpp
	virtual Vector3D GetLocPosOfIthAdditionalNode(int i, int3 pattern, int method=0); // in cpp
	virtual int GetSubdivisionLevelOfIthAdditionalNode(int i, int3 pattern, TArray<int>& cornersdlvls, int method=0); // in cpp

	virtual int GetNumberOfElementsForRefinement(int3 pattern, int method=0); // in cpp
	virtual int GetNodeListOfIthRefinedElement(int i, int3 pattern, TArray<int>& nodelist, int method=0); // in cpp
	virtual Vector3D GetLocPosCenterOfIthRefinedElement(int i, int3 pattern, int method=0); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

class MeshTrig : public MeshElement
{
public: // lifecycle I
	MeshTrig() { Reset(); }
	MeshTrig(const MeshTrig& other) { CopyFrom(other); }
	~MeshTrig() {}

public: // lifecycle II
	MeshTrig(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshTrig(int n1, int n2, int n3)
	{
		nodes.SetXN(3, n1,n2,n3); 
	}
	MeshTrig(int n1, int n2, int n3, 
		int n4, int n5, int n6)
	{
		nodes.SetXN(6, n1,n2,n3, n4,n5,n6);
	}
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshTrig(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshTrig& other)
	{
		MeshElement::CopyFrom(other);

		const MeshTrig& o = (const MeshTrig&) other;
	}
	virtual int IsMeshTrig() { return 1; }
	virtual int GetMEType() { if(NNodes()==3) return METrig; else if(NNodes()==6) return MEQTrig; else return 0; }
	virtual int Dim() { return 2; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 3; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 3; }
	virtual MeshElement GetLocalFace(int i); //in cpp

	virtual int NInter() { return 3; }
	virtual int2 GetInterPair(int i); // in cpp

	virtual Vector3D GetLocPosOfCenter() { return Vector3D(1./3.,1./3.,0.); }
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); // in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp

	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

class MeshLine : public MeshElement
{
public: // lifecycle I
	MeshLine() { Reset(); }
	MeshLine(const MeshLine& other) { CopyFrom(other); }
	~MeshLine() {}
public: // lifecycle II
	MeshLine(TArray<int>& nodelist,int matnri)
	{
		nodes.CopyFrom(nodelist);
		matnr = matnri;
	}
	MeshLine(int n1, int n2)
	{
		nodes.SetXN(2, n1,n2); 
	}
	MeshLine(int n1, int n2, 
		int n3)
	{
		nodes.SetXN(3, n1,n2, n3);
	}
	virtual MeshElement* GetCopy()
	{
		MeshElement* mec = new MeshLine(*this);
		return mec;
	}
	virtual void CopyFrom(const MeshLine& other)
	{
		MeshElement::CopyFrom(other);

		const MeshLine& o = (const MeshLine&) other;
// specific for derived
	}
	virtual int IsMeshLine()  { return 1; }
	virtual int GetMEType() { if(NNodes()==2) return MELine; else if(NNodes()==3) return MEQLine; else return 0; }
	virtual int Dim() { return 1; }
	virtual int Invert(); // in cpp

	virtual int NOrientations() { return 2; }
	virtual int GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse); // in cpp
	virtual int NFaces() { return 2; }
	virtual MeshElement GetLocalFace(int i); //in cpp

	virtual int NInter() { return 1; }
	virtual int2 GetInterPair(int i); // in cpp

	virtual Vector3D GetLocPosOfCenter() { return Vector3D(0.0,0.0,0.0); }
	virtual Vector3D GetLocPosOfIthNode(int i); // in cpp
	virtual double GetS0(Vector3D ploc, int shape); // in cpp
	virtual Matrix GetDSMatrix0(Vector3D ploc); // in cpp
	
	virtual int TestPLocOnElem(Vector3D ploc, double eps = 0.); // in cpp
protected:
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// BASEMESHNODE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// NOTES 
// required member variables:
// * coordinates - computed by the component when creating the MeshNode
// * typenumber - represents the typenumber for the ObjectFactory, string comparison done once for each component
// * bodyidx - BodyIndex / Domain number, determined by element at time of generation or parameter when loading
// NOT really required
// * flag for additional orientation values associated with the node - not the value themselves.
//   --> problem is that the flag would be dependeing on the MBSNodeType and no MBSNode is instantiated before this value would come in handy... 
// * the number of the node in the MBS

//typedef enum { THasRefSlope1 = 1, THasRefSlope2 = 2, THasRefSlope3 = 4,  THasRefAngles = 8} TMeshNodeOrientation; 

class MeshNode //$EDC$[beginclass,classname=MeshNode]
{
public: // lifecycle I
	MeshNode():position(0.,0.,0.),typenr(0),bodyidx(0){;}
	MeshNode(const MeshNode& other) { CopyFrom(other); }
	~MeshNode() {}

public: // lifecycle II
	virtual MeshNode* GetCopy()
	{
		MeshNode* mnc = new MeshNode(*this);
		return mnc;
	}
	virtual void CopyFrom(const MeshNode& other)
	{
		position = other.position;
		typenr = other.typenr;
		bodyidx = other.bodyidx;
	}
public: // access
	virtual Vector3D GetNodePos() { return position; }
	virtual void SetNodePos(Vector3D positioni) { position = positioni; }
	virtual Vector3D& NodePos() { return position; }
	
	virtual int GetTypeNr() { return typenr; }
	virtual void SetTypeNr(int t) { typenr = t; }
	virtual int& TypeNr() { return typenr; }

	virtual int GetBodyIdx() { return bodyidx; }
	virtual void SetBodyIdx(int b) { bodyidx = b; }
	virtual int& BodyIdx() { return bodyidx; }

	virtual int HasOwnOrientationStored() { return 0; }

public: // EDC-interaction
protected:
	Vector3D position;				// untransformed coordinates
	int typenr;						// node type number from ObjectFactory
	int bodyidx;                    // body index / domain number

}; //$EDC$[endclass,MeshNode]

/*
class MeshNodeExtended : public MeshNode //$EDC$[beginclass,classname=MeshNodeExtended]
{
public: // lifecycle I
	MeshNodeExtended():MeshNode(),n(Vector3D(0,0,0)) {;}
	MeshNodeExtended(const MeshNodeExtended& other) { CopyFrom(other); }
	~MeshNodeExtended() {}

public: // lifecycle II
	virtual MeshNode* GetCopy()
	{
		MeshNode* mnc = new MeshNodeExtended(*this);
		return mnc;
	}
	virtual void CopyFrom(const MeshNodeExtended& other)
	{
		MeshNode::CopyFrom(other);

		const MeshNodeExtended& o = (const MeshNodeExtended&) other;
		n = other.n;
	}
	
	virtual int HasOrientationStored() { return 1; }
protected:
	Vector3D n;
};
*/


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH COMPONENT - BASE CLASS
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// provides identifiers for the mesh component
// provides basic properties
class MeshComponent //$EDC$[beginclass,classname=MeshComponent]
{
public: // lifecycle I
	MeshComponent() { Reset();}
	//MeshComponent(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshComponent(const MeshComponent& other) { CopyFrom(other); }
	virtual ~MeshComponent() {	}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshComponent(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshComponent& other)
	{
		mesh = other.mesh;
		name = other.name;
		type = other.type;

		bodynr = other.bodynr;
		matnr = other.matnr;
		color = other.color;
		component_gnls = other.component_gnls;

		boundingbox = other.boundingbox;
		nodeIdx = other.nodeIdx;
		elemIdx = other.elemIdx;
	}
	virtual void Reset()
	{
		//mesh = NULL;
		name = GetComponentSpecification();
		type = GetComponentSpecification();
		
		bodynr = 1;
		matnr = 1;
		color = Vector3D(.7,.7,.7);
		component_gnls = -1; // default: no change

		boundingbox.Empty();
		nodeIdx.Flush();
		elemIdx.Flush();
	}
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("abstract"); } //$EDC$[funcaccess,EDCvarname"component_type",tooltiptext="specification of meshcomponent type. Once the mesh is added to the MBS, you MUST NOT change this type anymore"]
	virtual int IsSource() { return 0; }
	virtual int IsLoader() { return 0; }
	virtual int IsTransformation() { return 0; }

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc); // moved to .cpp - uses mesh funciton now
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		
		name = edc.TreeGetString("component_name");
		type = edc.TreeGetString("component_type");
		SetBodyIndex(edc.TreeGetInt("Generation.body_index"));
		SetMaterialNumber(edc.TreeGetInt("Generation.material_number"));
		SetGeometricNonlinearityStatus(edc.TreeGetInt("Generation.GeometricNonlinearityStatus"));
		Vector3D col;
		edc.TreeGetVector3D("Graphics.RGB_color",col.X(),col.Y(),col.Z());
		SetColor(col);
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // owner
	virtual const BaseMesh* GetMesh() const { return mesh; }
	virtual BaseMesh* GetMesh() { return mesh; }
	virtual void SetMesh(BaseMesh* meshi) { mesh = meshi; }
public: // identification
	virtual const mystr GetName() const { return name; }
	virtual void SetName(mystr namei) { name = namei; }
public: // basic component properties
	virtual void SetBodyIndex(int i) { bodynr = i; }
	virtual int GetBodyIndex(int i=0) { return bodynr; }
	virtual void SetMaterialNumber(int i) { matnr = i; }
	virtual int GetMaterialNumber(int i=0) { return matnr; }
	virtual void SetColor(Vector3D col) { color = col; }
	virtual Vector3D GetColor(int i=0)
	{
		if (NSubs() > 0)
		{
			return GetSubordinateComponent(1)->GetColor(i);
		}
		return color; 
	}
	virtual void SetGeometricNonlinearityStatus(int i) { component_gnls = i; }
	virtual int GetGeometricNonlinearityStatus(int i=0) { return component_gnls; }
public: // 
	virtual int InitNodeIdx() { nodeIdx.SetLen(NNodes()); nodeIdx.SetAll(0); return nodeIdx.Length(); }
	virtual int InitElemIdx() { elemIdx.SetLen(NElems()); elemIdx.SetAll(0); return elemIdx.Length(); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// to be implemented in derived classes
public: // overrideables - access
	virtual int NNodes() { return 0; }
	virtual int NElems() { return 0; }
	virtual MeshNode* GetNodePtr(int i) { return NULL; }
	virtual MeshElement* GetElemPtr(int i) { return NULL; }

public: // overrideables - meshing
	virtual int MeshYourself() { return 0; } 
	virtual int GenerateOwnNodes() { return 0; }
	virtual int GenerateOwnElements() { return 0; }

public: // overrideables - node generation
	virtual int GetNodeObjectFactoryType(int i) { return 0; }
	virtual void SetNodeMeshNumber(int i, int idx) { return; }
	virtual int GetNodeMeshNumber(int i) { return 0; }
	
	virtual int ComputeOrientation(int i) { return 0; }
	virtual int ComputeOrientation(Vector3D& locpos) { return 0; }
	virtual Vector3D GetNodePosition(int i) { return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope1(int i) { return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope2(int i) { return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope3(int i) { return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefAngles(int i) { return Vector3D(0.,0.,0.); }
	virtual int GetNodeBodyIndex(int i) { return 0; }

public: // overrideables - element generation
	virtual int GetElementObjectFactoryType(int i) { return 0; }
	virtual int GetElementMaterialNumber(int i) { return 0; } 
	virtual int GetElementGeometricNonlinearityStatus(int i) { return -1; } // default -1 --> use from subordinate or element itself
	virtual void SetElementMeshNumber(int i, int idx) { return; }
	virtual int GetElementMeshNumber(int i) { return 0; }
	virtual TArray<int> GetElementNodeList(int i) { return TArray<int>(0); } // TArray<int>& should be used when possible
	virtual TArray<int> GetElementFaceNodeList(int i, int f) { return TArray<int>(0); }
	virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);

	virtual int ComputeElementSize(int i) { return 0; }
	virtual Vector3D GetElementSizeVector(int i) { return Vector3D(0.,0.,0.); }
	virtual int GetElemAlternateFlag(int i) { return 0; }
	virtual mystr GetElemType(int i=0) { return 0; } // to be removed after finite elemen base class is extended to return the required additional parameters OR ALL the finite can compute all sizes themselves
	virtual mystr GetNodeType(int i=0) { return 0; }

public: // overrideables - subordinate mesh components
	virtual int NSubs() { return 0; }
	virtual MeshComponent* GetSubordinateComponent(int i=1) { return NULL; }
	virtual void SetSubordinate(MeshComponent* , int) { }
	virtual int NCoincident() { return 0; }
	virtual int2 GetCoincidentNodeNumbers(int i) { return int2(0,0); }
	virtual int GetCoincidentPenaltyFlag(int i) { return 0; }
	virtual void SetAutoConstraintNumber(int i, int cnr) { return; }
public: // overrideables - picking
	virtual int ComputeBoundingBox() { return 0; }
	virtual Box3D GetBoundingBox() { return boundingbox; }

	virtual TArray<int2> GetPeriodicBCNodePairs() const {return TArray<int2>();}
	virtual TArray<int> GetSymmetricNodes() const { return TArray<int>(); }
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
protected: // own variables 
	BaseMesh* mesh;
	mystr name;
	mystr type;
	int bodynr;
	int matnr;
	Vector3D color;
	int component_gnls;

	Box3D boundingbox;
	TArray<int> nodeIdx; // node number mapping to index in mesh class
	TArray<int> elemIdx; // elem number mapping to index in mesh class
}; //$EDC$[endclass,MeshComponent]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SOURCE MESHES
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// source meshes have their OWN Nodes and Elements STORED
// * can generate Nodes and Elements themselves OR
// * OR can load Nodes and Elements form an external source ( e.g. textfile )
// * MUST NOT have subordinate Meshes
class MeshComponentSource : public MeshComponent
{
public: // lifecycle I
	MeshComponentSource() { Reset();}
//	MeshComponentSource(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshComponentSource(const MeshComponentSource& other) { CopyFrom(other); }
	virtual ~MeshComponentSource()
	{
	}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshComponentSource(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshComponentSource& other)
	{
		MeshComponent::CopyFrom(other);
	
		nodes.CopyFrom(other.nodes);
		elems.CopyFrom(other.elems);
		nodetype = other.nodetype;
		elemtype = other.elemtype;
		thickness = other.thickness;
		component_gnls = other.component_gnls;
	}
	virtual void Reset()
	{
		MeshComponent::Reset();

		nodes.Flush();
		elems.Flush();
		//nodetype
		//elemtype
		thickness = 0.0;
		component_gnls = -1; // use_element_default
	}
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Source"); } 
	virtual int IsSource() { return 1; }

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponent::GetElementData(edc);

		ElementData ed;
		ed.SetText(nodetype.c_str(),"used_node_type"); ed.SetToolTipText("node type used to instantiate the component"); edc.TreeAdd("Generation",ed);;
		ed.SetText(elemtype.c_str(),"used_elem_type"); ed.SetToolTipText("element type used to instantiate the component"); edc.TreeAdd("Generation",ed);
		// thickness: by derived class
		// gnls in Primitive and DomainContainer
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);

		NodeType() = edc.TreeGetString("Generation.used_node_type");
		ElemType() = edc.TreeGetString("Generation.used_elem_type");

		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // node and element types 
	virtual void SetNodeType(mystr type) { nodetype = type; }
	virtual mystr& NodeType() { return nodetype; }
	virtual void SetElemType(mystr type) { elemtype = type; }
	virtual mystr& ElemType() { return elemtype; }
public: // functionality - lists
	virtual int AddNode(MeshNode& node) { return nodes.Add(node.GetCopy()); }
	virtual int AddElem(MeshElement& elem) { return elems.Add(elem.GetCopy()); }
public: // additional properties
	virtual int& GNLS() { return component_gnls; }
	virtual double& Thickness() { return thickness; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { return nodes.Length(); }
	virtual int NElems() { return elems.Length(); }
	virtual MeshNode* GetNodePtr(int i) { return nodes(i); }
	virtual MeshElement* GetElemPtr(int i) { return elems(i); }

public: // overrideables - meshing
//NOT	virtual int MeshYourself() { return 0; }
//NOT virtual int GenerateOwnNodes() { return 0; }
//NOT virtual int GenerateOwnElements() { return 0; }

public: // overrideables - node generation
	virtual int GetNodeObjectFactoryType(int i) { return GetNodePtr(i)->GetTypeNr(); }
	virtual void SetNodeMeshNumber(int i, int idx) { nodeIdx(i) = idx; }
	virtual int GetNodeMeshNumber(int i) { return nodeIdx(i); }
//NOT virtual int ComputeOrientation(int i) { return 0; }
//NOT virtual int ComputeOrientation(Vector3D& locpos) { return 0; }
  virtual Vector3D GetNodePosition(int i) { return GetNodePtr(i)->GetNodePos(); }
//NOT virtual Vector3D GetNodeRefSlope1(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefSlope2(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefSlope3(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefAngles(int i) { return Vector3D(0.,0.,0.); }
  	virtual int GetNodeBodyIndex(int i) { return GetNodePtr(i)->GetBodyIdx(); }
	virtual mystr GetNodeType(int i=0) { return nodetype; }

public: // overrideables - element generation
	virtual int GetElementObjectFactoryType(int i) { return GetElemPtr(i)->GetTypeNr(); }
	virtual int GetElementMaterialNumber(int i) { return GetElemPtr(i)->GetMatNr(); }
	virtual int GetElementGeometricNonlinearityStatus(int i=0) { return component_gnls; }
	virtual void SetElementMeshNumber(int i, int idx) { elemIdx(i) = idx; }
	virtual int GetElementMeshNumber(int i) { return elemIdx(i); }
	virtual TArray<int> GetElementNodeList(int i); // { return GetElemPtr(i)->GetNodeList(); }
	virtual TArray<int> GetElementFaceNodeList(int i, int f); // { return GetElemPtr(i)->GetFaceNodeList(f); }
//NOT virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);
//NOT virtual int ComputeElementSize(int i) { return 0; }
//NOT virtual int GetElemAlternateFlag(int i) { return 0; }
//NOT	virtual Vector3D GetElementSizeVector(int i) { return Vector3D(0.,0.,0.); }
	virtual mystr GetElemType(int i=0) { return elemtype; }

public: // overrideables - subordinate mesh components
//NOT virtual int NSubs() { return 0; }
//NOT virtual MeshComponent* GetSubordinateComponent(int i=1) { return NULL; }
//NOT virtual int NCoincident() { return 0; }
//NOT virtual int2 GetCoincidentNodeNumbers(int i) { return int2(0,0); }
//NOT virtual int GetCoincidentPenaltyFlag(int i) { return 0; }
//NOT virtual void SetAutoConstraintNumber(int i, int cnr) { return; }
public: // overrideables - picking
//NOT virtual int ComputeBoundingBox() { return 0; }
//NOT virtual Box3D GetBoundingBox() { return boundingbox; }

protected: // own variables
	TArray<MeshNode*> nodes;
	TArray<MeshElement*> elems;

	mystr nodetype;
	mystr elemtype;
	double thickness;    
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH PRIMITIVE  
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// common base class for all classes that generate own nodes and elements
// specific geometries in derived classes
class MeshComponentPrimitive : public MeshComponentSource
{
public: // lifecycle I
	MeshComponentPrimitive() { Reset();}
//	MeshComponentPrimitive(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshComponentPrimitive(const MeshComponentPrimitive& other) { CopyFrom(other); }
	~MeshComponentPrimitive() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshComponentPrimitive(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshComponentPrimitive& other)
	{
		MeshComponentSource::CopyFrom(other);
		
		points.CopyFrom(other.points);	
		discretization = other.discretization;
		
		posdependent = other.posdependent;
		orientationcomputed = other.orientationcomputed;
		refslope1 = other.refslope1;
		refslope2 = other.refslope2;
		refslope3 = other.refslope3;
		refangles = other.refangles;

		constantelementsize = other.constantelementsize;
		sizecomputed = other.sizecomputed;
		size1 = other.size1;
		size2 = other.size2;
	}
	virtual void Reset()
	{
		MeshComponentSource::Reset();

		points.Flush();
		discretization = int3(1,1,1);

		posdependent = 0;
		orientationcomputed = 0;
		refslope1 = Vector3D(0.,0.,0.);
		refslope2 = Vector3D(0.,0.,0.);
		refslope3 = Vector3D(0.,0.,0.);
		refangles = Vector3D(0.,0.,0.);

		constantelementsize = 1;
		sizecomputed = 0;
		size1 = 0.;
		size2 = 0.;
	}
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Primitive"); } 

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentSource::GetElementData(edc);

		ElementData ed;
		char label[16];
		for(int i=1; i<=NPoints(); i++)
		{
			sprintf(label,"P%d",i);
			ed.SetVector3D(Point(i).X(),Point(i).Y(),Point(i).Z(),label); ed.SetToolTipText("defining points of the component in basic configuration"); edc.TreeAdd("Generation",ed);
		} 
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentSource::SetElementData(edc);

		char label[32];
		points.SetLen(NPoints());
		for(int i=1; i<=NPoints(); i++)
		{
			sprintf(label,"Generation.P%d",i);
			edc.TreeGetVector3D(label, Point(i).X(), Point(i).Y(), Point(i).Z());
		}
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // defining points - list 
	virtual void SetPointList(TArray<Vector3D>& pointlist) { points = pointlist; }
	virtual int NPoints() { return points.Length(); }	
	virtual Vector3D& Point(int i) { return points(i); }
public: // discretization of generated mesh
	virtual void SetDiscretization(int dx, int dy=0, int dz=0) { discretization = int3(dx,dy,dz); }
	virtual int& Discretization(int i) { return discretization(i); }
public: // orientation
	virtual Vector3D GetRefSlope1() { return refslope1; }
	virtual Vector3D GetRefSlope2() { return refslope2; }
	virtual Vector3D GetRefSlope3() { return refslope3; }
	virtual Vector3D GetRefAngles() { return refangles; }
	virtual double GetSize1() { return size1; }
	virtual double GetSize2() { return size2; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
//NOT	virtual int NNodes() { return nodes.Length(); }
//NOT	virtual int NElems() { return elems.Length(); }
//NOT	virtual MeshNode* GetNodePtr(int i) { return nodes(i); }
//NOT	virtual MeshElement* GetElemPtr(int i) { return elems(i); }

public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp - primitives generate nodes and elements
//NOT virtual int GenerateOwnNodes() { return 0; }
//NOT virtual int GenerateOwnElements() { return 0; }

public: // overrideables - node generation
//NOT virtual int GetNodeObjectFactoryType(int i) { return GetNodePtr(i)->GetTypeNr(); }
//NOT virtual void SetNodeMeshNumber(int i, int idx) { nodeIdx(i) = idx; }
//NOT virtual int GetNodeMeshNumber(int i) { return nodeIdx(i); }
//NOT virtual int ComputeOrientation(int i) { return 0; }
//NOT virtual int ComputeOrientation(Vector3D& locpos) { return 0; }
//NOT virtual Vector3D GetNodePosition(int i) { return GetNodePtr(i)->GetNodePos(); }
	virtual Vector3D GetNodeRefSlope1(int i) { return GetRefSlope1(); } 
	virtual Vector3D GetNodeRefSlope2(int i) { return GetRefSlope2(); }
	virtual Vector3D GetNodeRefSlope3(int i) { return GetRefSlope3(); }
	virtual Vector3D GetNodeRefAngles(int i) { return GetRefAngles(); }
//NOT virtual int GetNodeBodyIndex(int i) { return GetNodePtr(i)->GetNodeBodyIndex(i); }

public: // overrideables - element generation
//NOT virtual int GetElementObjectFactoryType(int i) { return GetElemPtr(i)->GetTypeNr(); }
//NOT virtual int GetElementMaterialNumber(int i) { return GetElemPtr(i)->GetMatNr(); }
//NOT virtual int GetElementGeometricNonlinearityStatus(int i=0) { return component_gnls; }
//NOT virtual void SetElementMeshNumber(int i, int idx) { elemIdx(i) = idx; }
//NOT virtual int GetElementMeshNumber(int i) { return elemIdx(i); }
//NOT virtual TArray<int> GetElementNodeList(int i); // { return GetElemPtr(i)->GetNodeList(); }
//NOT virtual TArray<int> GetElementFaceNodeList(int i,int f); // { return GetElemPtr(i)->GetFaceNodeList(f); }
//NOT virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);
//NOT virtual int ComputeElementSize(int i) { return 0; }
	virtual Vector3D GetElementSizeVector(int i) { ComputeElementSize(i); return Vector3D(thickness,size1,size2); }
//NOT virtual mystr GetElemType(int i=0) { return elemtype; }

public: // overrideables - subordinate mesh components
//NOT virtual int NSubs() { return 0; }
//NOT virtual MeshComponent* GetSubordinateComponent(int i=1) { return NULL; }
//NOT virtual int NCoincident() { return 0; }
//NOT virtual int2 GetCoincidentNodeNumbers(int i) { return int2(0,0); }
//NOT virtual int GetCoincidentPenaltyFlag(int i) { return 0; }
//NOT virtual void SetAutoConstraintNumber(int i, int cnr) { return; }
public: // overrideables - picking
  virtual int ComputeBoundingBox();
//NOT virtual Box3D GetBoundingBox() { return boundingbox; }

protected: // own variables
// defining points and discretization
	TArray<Vector3D> points;
	int3 discretization;
// mesh generation - orientation (nodes)
	int posdependent;
	int orientationcomputed;
	Vector3D refslope1;
	Vector3D refslope2;
	Vector3D refslope3;
	Vector3D refangles;
// mesh generation - size (elements)
	int constantelementsize;
	int sizecomputed;
	double size1;
	double size2;
};

class MeshBlock : public MeshComponentPrimitive 
{
public: // lifecycle I
	MeshBlock() {	Reset(); }
	//MeshBlock(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshBlock(const MeshBlock& other) { CopyFrom(other); }
	~MeshBlock() {}	
	
public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshBlock(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshBlock& other)
	{
		MeshComponentPrimitive::CopyFrom(other);

		const MeshBlock& o = (const MeshBlock&)other;
// specific for derived
	}
	virtual void Reset()
	{
		MeshComponentPrimitive::Reset();

		points.SetLen(NPoints());
		Point(1) = Vector3D(0,0,0); Point(5) = Vector3D(0,0,1);
		Point(2) = Vector3D(1,0,0); Point(6) = Vector3D(1,0,1);
		Point(3) = Vector3D(0,1,0); Point(7) = Vector3D(0,1,1);
		Point(4) = Vector3D(1,1,0); Point(8) = Vector3D(1,1,1);

		nodetype = mystr("Node3D");
		elemtype = mystr("Hexahedral");
		component_gnls = -1; // use_element_default
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Block"); } 

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentPrimitive::GetElementData(edc);
		
		ElementData ed;
		Vector3D tmp(Discretization(1),Discretization(2),Discretization(3));
		ed.SetVector(tmp.GetVecPtr(),3,"discretization"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("discretization for block (x,y,z) "); edc.TreeAdd("Generation",ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentPrimitive::SetElementData(edc);
		
		double* v;
		int l;
		edc.TreeGetVector("Generation.discretization", &v, l);  
		Discretization(1) = (int) (v[0]+.5);
		Discretization(2) = (int) (v[1]+.5);
		Discretization(3) = (int) (v[2]+.5);
		return rv;
	}	

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: 	
	virtual int NPoints() { return 8; }
private: // additional functions meshing
	virtual int GetNodesForHex(int node1idx, TArray<int>& nodelist);	

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
private: // overrideables - meshing
	virtual int GenerateOwnNodes();
	virtual int GenerateOwnElements();

	virtual int GetElemAlternateFlag(int i);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
};

class MeshCylinder : public MeshComponentPrimitive 
{
public: // lifecycle I
	MeshCylinder() { Reset(); }
	//MeshCylinder(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshCylinder(const MeshCylinder& other) { CopyFrom(other); }
	~MeshCylinder() {}	
	
public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshCylinder(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshCylinder& other)
	{
		MeshComponentPrimitive::CopyFrom(other);

		const MeshCylinder& o = (const MeshCylinder&)other;
// specific for derived
	}
	virtual void Reset()
	{
		MeshComponentPrimitive::Reset();

		points.SetLen(NPoints());             //        z-axis (rotational)		
		Point(1) = Vector3D(0,0,0);           //        ( ^ )			
		Point(2) = Vector3D(1,0,0);           //          |			    
		Point(3) = Vector3D(0,0,1);           //          |(3)+---------+(4)		
		Point(4) = Vector3D(1,0,1);           //          |   |#########|
                                              //          |   |#########|
		nodetype = mystr("Node3D");           //          |(1)+---------+(2)	
		elemtype = mystr("Hexahedral");       //   y-axis X--------> x-axis
        component_gnls = -1; // elem_ default // NOTE: when P1.x==0&&P3.x==0 -> full cylinder
                                              //       when P1.x!=0&&P3.x!=0 -> hollow cylinder
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Cylinder"); } 


public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentPrimitive::GetElementData(edc);
		
		ElementData ed;
		Vector3D tmp(Discretization(1),Discretization(2),Discretization(3));
		ed.SetVector(tmp.GetVecPtr(),3,"discretization"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("discretization for cylinder (rad,ang,axi) "); edc.TreeAdd("Generation",ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentPrimitive::SetElementData(edc);
		
		double* v;
		int l;
		edc.TreeGetVector("Generation.discretization", &v, l);  
		Discretization(1) = (int) (v[0]+.5);
		Discretization(2) = (int) (v[1]+.5);
		Discretization(3) = (int) (v[2]+.5);
		return rv;
	}	

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public:
	virtual int NPoints() { return 4; }
private: // additional functions - meshing
	virtual int IsFullCylinder() { return ((Point(1).X()==0.) && (Point(3).X()==0.)); }
	virtual int GenerateNodesDisc(Vector3D pi, Vector3D pa, double inner_fraction=0.5);
	virtual int GenerateNodesShim(Vector3D pi, Vector3D pa);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
private: // overrideables - meshing
	virtual int GenerateOwnNodes();
	virtual int GenerateOwnElements();

	virtual int GetElemAlternateFlag(int i);
public: 
	virtual int ComputeBoundingBox(); // required special implementation for cylinder
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
};

class MeshQuadrilateral : public MeshComponentPrimitive
{
public: // lifecycle I
	MeshQuadrilateral() { Reset(); }
	//MeshQuadrilateral(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshQuadrilateral(const MeshQuadrilateral& other) { CopyFrom(other); }
	~MeshQuadrilateral() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshQuadrilateral(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshQuadrilateral& other)
	{
		MeshComponentPrimitive::CopyFrom(other);

		const MeshQuadrilateral& o = (const MeshQuadrilateral&)other;
// specific for derives
	}
	virtual void Reset()
	{
		MeshComponentPrimitive::Reset();

		points.SetLen(NPoints());
		Point(1) = Vector3D(0,0,0);
		Point(2) = Vector3D(1,0,0);
		Point(3) = Vector3D(0,1,0);
		Point(4) = Vector3D(1,1,0);
		
		nodetype = mystr("Node3DS1S2");
		elemtype = mystr("ANCFThinPlate3D");
		component_gnls = -1; // use_element_default
		thickness = 1e-3;
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Quadrilateral"); } 

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentPrimitive::GetElementData(edc);
		
		ElementData ed;
		Vector3D tmp (Discretization(1),Discretization(2),0);
		ed.SetVector(tmp.GetVecPtr(),2, "discretization"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("discretization for quadrilateral (x,y,*)"); edc.TreeAdd("Generation",ed);
		ed.SetDouble(Thickness(),"thickness"); ed.SetToolTipText("thickness for all elements"); edc.TreeAdd("Generation",ed);
	}

	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentPrimitive::SetElementData(edc);

		double* v;
		int l;
		edc.TreeGetVector("Generation.discretization", &v, l);  
		Discretization(1) = (int) (v[0]+.5);
		Discretization(2) = (int) (v[1]+.5);
		Thickness() = edc.TreeGetDouble("Generation.thickness");
		return rv;
	}	

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public:
	virtual int NPoints() { return 4; }
	virtual int GetNodesForQuad(int node1idx, TArray<int>& nodelist);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
private: // overrideables - meshing
	virtual int GenerateOwnNodes();
	virtual int GenerateOwnElements();
public: // overrideables - node generation
	virtual int ComputeOrientation(Vector3D& pos);
	virtual int ComputeOrientation(int i) { return ComputeOrientation(GetNodePosition(i)); }
public: // overrideables - element generation
	virtual int ComputeElementSize(int i);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
};

class MeshCurve : public MeshComponentPrimitive
{
public: // lifecycle I
	MeshCurve() {	Reset(); }
	//MeshCurve(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshCurve(const MeshCurve& other) { CopyFrom(other); }
	~MeshCurve() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshCurve(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshCurve& other)
	{
		MeshComponentPrimitive::CopyFrom(other);

		const MeshCurve& o = (const MeshCurve&)other;
// specific for derived
	}
	virtual void Reset()
	{
		MeshComponentPrimitive::Reset();

		points.SetLen(2);
		Point(1) = Vector3D(0,0,0);
		Point(2) = Vector3D(1,0,0);

		nodetype = mystr("Node3DRxyz");
		elemtype = mystr("LinearBeam3D");
		component_gnls = -1; // use_element_default
		// beam cross section is defined in material
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Linear"); } 

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentPrimitive::GetElementData(edc);
		
		ElementData ed;
		Vector3D tmp (Discretization(1),0,0);
		ed.SetVector(tmp.GetVecPtr(),1 , "discretization"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("discretization for linear (x,*,*)"); edc.TreeAdd("Generation",ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentPrimitive::SetElementData(edc);

		double* v;
		int l;
		edc.TreeGetVector("Generation.discretization", &v, l);  
		Discretization(1) = (int) (v[0]+.5);
		return rv;
	}	

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public:
	virtual int NPoints() { return 2; }
	virtual int GetNodesForLinear(int node1idx, TArray<int>& nodelist);	
	virtual int GetNodesForQuadratic(int node1idx, TArray<int>& nodelist);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
private: // overrideables - meshing
	virtual int GenerateOwnNodes();
	virtual int GenerateOwnElements();
	
	//function returns if element is of quadratic type (3-noded element)
	//is slow! use only outside of nested loops!
	virtual bool IsQuadraticElementType()
	{
		return (elemtype.Find(mystr("Quadratic")) != -1);    // Find(..) returns -1 if not found, otherwise starting position
	}

public: // overrideables - node generation
	virtual int ComputeOrientation(Vector3D& pos);
	virtual int ComputeOrientation(int i) { return ComputeOrientation(GetNodePosition(i)); }
public: // overrideables - element generation
	virtual int ComputeElementSize(int i);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH SOURCE EXTENDED and derived
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// base class for all source extensions - those classes introduce additional nodes/elelemts, some completely replace the entries of the subordinate components
class MeshSourceExtended : public MeshComponentSource
{
public: // lifecycle I
	MeshSourceExtended() {	Reset(); }
	//MeshSourceExtended(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshSourceExtended(const MeshSourceExtended& other) { CopyFrom(other); }
	~MeshSourceExtended() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshSourceExtended(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshSourceExtended& other)
	{
		MeshComponentSource::CopyFrom(other);

		const MeshSourceExtended& o = (const MeshSourceExtended&)other;
// specific for derived
		if(o.NSubs()>0) 
		{ 
			subordinate = o.subordinate->GetCopy();
		}
	}
	virtual void Reset()
	{
		MeshComponentSource::Reset();
		subordinate = NULL;		
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("SourceExtended"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentSource::GetElementData(edc);

		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.body_index"); if(ped!=NULL) { ped->SetText("dependent","same_as_seed_node"); ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.material_number"); if(ped!=NULL) { ped->SetText("dependent","same_as_seed_element"); ped->SetLocked(1); }

		ElementData ed;
		for(int i=1; i<=NSubs(); i++) 
		{
			ElementDataContainer edc_sub;
			GetSubordinateComponent(i)->GetElementData(edc_sub);
			ed.SetEDC(&edc_sub,GetSubordinateComponent(i)->GetName()); edc.Add(ed);
		}
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentSource::SetElementData(edc);
		
		for(int i=1; i<=NSubs(); i++)
		{
			ElementDataContainer* p_edc_sub;
			ElementData* p_ed_sub = edc.TreeFind(GetSubordinateComponent(i)->GetName());
			if(p_ed_sub!=NULL && p_ed_sub->IsEDC())
			{
				p_edc_sub = p_ed_sub->GetEDC();
				GetSubordinateComponent(i)->SetElementData(*p_edc_sub);
			}
		}
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual void SetSubordinate(MeshComponent* sub) { subordinate = sub; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
//NOT	virtual int NNodes() { if(subordinate!=NULL) return GetSubordinateComponent()->NNodes(); else return 0; }
//NOT	virtual int NElems() { if(subordinate!=NULL) return GetSubordinateComponent()->NElems(); else return 0; }
//NOT	virtual MeshNode* GetNodePtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePtr(i); else return NULL; }
//NOT	virtual MeshElement* GetElemPtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElemPtr(i); else return NULL; }

public: // overrideables - meshing
//NOT	virtual int MeshYourself() { return 0; }
//NOT virtual int GenerateOwnNodes() { return 0; }
//NOT virtual int GenerateOwnElements() { return 0; 

public: // overrideables - node generation
//NOT virtual int GetNodeObjectFactoryType(int i) { return GetNodePtr(i)->GetTypeNr(); }
//NOT virtual void SetNodeMeshNumber(int i, int idx) { nodeIdx(i) = idx; }
//NOT virtual int GetNodeMeshNumber(int i) { return nodeIdx(i); }
//NOT virtual int ComputeOrientation(int i) { return 0; }
//NOT virtual int ComputeOrientation(Vector3D& locpos) { return 0; }
//NOT virtual Vector3D GetNodePosition(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefSlope1(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefSlope2(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefSlope3(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual Vector3D GetNodeRefAngles(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual int GetNodeBodyIndex(int i) { return GetNodePtr(i)->GetNodeBodyIndex(i); }

public: // overrideables - element generation
//NOT virtual int GetElementObjectFactoryType(int i) { return GetElemPtr(i)->GetTypeNr(); }
//NOT virtual int GetElementMaterialNumber(int i) { return GetElemPtr(i)->GetMatNr(); }
virtual int GetElementGeometricNonlinearityStatus(int i=0);
//NOT virtual void SetElementMeshNumber(int i, int idx) { elemIdx(i) = idx; }
//NOT virtual int GetElementMeshNumber(int i) { return elemIdx(i); }
//NOT virtual TArray<int> GetElementNodeList(int i); // { return GetElemPtr(i)->GetNodeList(); }
//NOT virtual TArray<int> GetElementFaceNodeList(int i, int f); // { return GetElemPtr(i)->GetFaceNodeList(f); }
//NOT virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);
//NOT virtual int ComputeElementSize(int i) { return 0; }
//NOT	virtual Vector3D GetElementSizeVector(int i) { return Vector3D(0.,0.,0.); }
//NOT virtual mystr GetElemType(int i=0) { return elemtype; }

public: // overrideables - subordinate mesh components
	virtual int NSubs() { if(subordinate!= NULL) return 1; else return 0; } 
	virtual const int NSubs() const { if(subordinate!= NULL) return 1; else return 0; }
	virtual MeshComponent* GetSubordinateComponent(int i=1) { return subordinate; }
//NOT virtual int NCoincident() { return 0; }
//NOT virtual int2 GetCoincidentNodeNumbers(int i) { return int2(0,0); }
//NOT virtual int GetCoincidentPenaltyFlag(int i) { return 0; }
//NOT virtual void SetAutoConstraintNumber(int i, int cnr) { return; }
public: // overrideables - picking
//NOT virtual int ComputeBoundingBox() { return 0; }
//NOT virtual Box3D GetBoundingBox() { return boundingbox; }

protected: // own variables
	MeshComponent* subordinate;
	int flag;
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH LINEAR TO QUADRATIC 
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// changes elements from linear to quadratic
// introduces additional nodes and fully replaces the elements
class MeshLinearToQuadratic : public MeshSourceExtended
{
public: // lifecycle I
	MeshLinearToQuadratic() { Reset(); }
	//MeshLinearToQuadratic(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLinearToQuadratic(const MeshLinearToQuadratic& other) { CopyFrom(other); }
	~MeshLinearToQuadratic() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLinearToQuadratic(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLinearToQuadratic& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshLinearToQuadratic& o = (const MeshLinearToQuadratic&) other;
	// specific for derived
		parentnodes.CopyFrom(o.parentnodes);
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();
		
		parentnodes.Flush();
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("LinearToQuadratic"); } 

	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetText("dependent","used_node_type"); ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetText("dependent","used_elem_type"); ped->SetLocked(1); }
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
private: // additional functions - select node and element types
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int lineartype=0);
	virtual int GetElemTypeLookup(int lineartype=0);
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { if(NSubs()!=0) { return GetSubordinateComponent()->NNodes()+nodes.Length(); } else return 0; }	// first nodes from subordinate, then additional nodes 
  virtual int NElems() { return elems.Length(); }																																				// only new elements, ignore subordinate
	virtual MeshNode* GetNodePtr(int i);																																									// first nodes from subordinate, then additional nodes 
  virtual MeshElement* GetElemPtr(int i) { return elems(i); }																														// only new elements, ignore subordinate

public: // overrideables - meshing
	virtual int MeshYourself(); 

public: // overrideables - node generation 
	virtual int GetNodeObjectFactoryType(int i); 
	virtual void SetNodeMeshNumber(int i, int idx); 
	virtual int GetNodeMeshNumber(int i); 
	virtual int ComputeOrientation(int i); 
	virtual Vector3D GetNodePosition(int i); 

	virtual Vector3D GetNodeRefAngles(int i);
	virtual Vector3D GetNodeRefSlope1(int i);
	virtual Vector3D GetNodeRefSlope2(int i);
	virtual Vector3D GetNodeRefSlope3(int i);

public: // overrideables - element generation
//	NOT virtual void MapNodeListComponentToMesh(TArray<int>& nodelist); 

	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
protected: // own variables
	TArray<int2> parentnodes; // original nodes for all intermediate nodes
	
	TArray<int> nodetypelist;               // lookup table: list of object factory nodetypes - must be initialized with ::InitializeLookupTable() 
	TArray<int> elementtypelist;            // lookup table: list of object factory elementtypes - must be initialized with ::InitializeLookupTable() 
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH HEXAEDER TO TETRAEDER
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// splits existing elements 
// introduces NO new nodes and completely replaces elements
class MeshSplitHexes : public MeshSourceExtended
{
public: // lifecycle I
	MeshSplitHexes() { Reset(); }
	MeshSplitHexes(const MeshSplitHexes& other) { CopyFrom(other); }
	~MeshSplitHexes() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshSplitHexes(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshSourceExtended& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshSplitHexes& o = (const MeshSplitHexes&)other;
// specific for derived
		alternate = o.alternate;
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();
		
		//split = 5;
		alternate = 1;
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("SplitHexes"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData ed;
		//ed.SetInt(split,"split_into"); ed.SetToolTipText("split single Hex into: 2..Prisms, 3..Pyramids, 5..Tets"); edc.Add(ed);
		ed.SetBool(alternate,"alternate"); ed.SetToolTipText("use alternate orientation for split mesh"); edc.Add(ed);
		
		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetLocked(1); }

	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentSource::SetElementData(edc);

		//split = edc.TreeGetInt("split_into");
		alternate = edc.TreeGetBool("alternate");

		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
private: // additional
	int GetNodesFor5Tets(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate=0);
	int GetNodesFor3Pyramids(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate=0);
	int GetNodesFor2Prisms(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate=0);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { if(NSubs()!=0) { return GetSubordinateComponent()->NNodes(); } else return 0; }	// only nodes from subordinate
  virtual int NElems() { return elems.Length(); }																													// only new elements, ignore subordinate 
	virtual MeshNode* GetNodePtr(int i) { if(NSubs()!=0) { return GetSubordinateComponent()->GetNodePtr(i); } else return NULL; }	// only nodes from subordinate
  virtual MeshElement* GetElemPtr(int i) { return elems(i); }																							// only new elements, ignore subordinate

public: // overrideables - meshing
	virtual int MeshYourself();

public: // overrideables
	virtual int GetNodeObjectFactoryType(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeObjectFactoryType(i); else return 0; }
	virtual void SetNodeMeshNumber(int i, int idx) { if(subordinate!=0) GetSubordinateComponent()->SetNodeMeshNumber(i,idx); }
	virtual int GetNodeMeshNumber(int i) { if(subordinate!=0) return GetSubordinateComponent()->GetNodeMeshNumber(i); else return 0;}
	virtual int ComputeOrientation(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->ComputeOrientation(i); else return 0; }
	virtual Vector3D GetNodePosition(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePosition(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope1(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope1(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope2(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope2(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope3(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope3(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefAngles(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefAngles(i); else return Vector3D(0.,0.,0.); }
//TODO: virtual int GetNodeBodyIndex(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeBodyIndex( (i-1)/5 +1); }

protected:
	int alternate;
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH REFINE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// defines and applies refinement on subordinate component
// introduces additional nodes 
// introduces additional elements and replaces some
class MeshRefine : public MeshSourceExtended
{
	//TODO: implement other initial elements - so far refine is possible for {8-node hex, 4-node quad} only
public: // lifecycle I
	MeshRefine() { Reset(); }
	//MeshRefine(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshRefine(const MeshRefine& other) { CopyFrom(other); }
	~MeshRefine() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshRefine(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshSourceExtended& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshRefine& o = (const MeshRefine&) other;
	// specific for derived
		method=o.method;
		nodesdlvl_seed.CopyFrom(o.nodesdlvl_seed);
		nodesdlvl.CopyFrom(o.nodesdlvl);
		elementmapping_R2PC.CopyFrom(o.elementmapping_R2PC);
		unrefinedelementofnode.CopyFrom(o.unrefinedelementofnode);
		locposnodeinunrefinedelement.CopyFrom(o.locposnodeinunrefinedelement);
		locposelemcenterunrefinedelement.CopyFrom(o.locposelemcenterunrefinedelement);
		elementshrinkingfactor.CopyFrom(o.elementshrinkingfactor);
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();

		method=0;
		nodesdlvl_seed.Flush();
		nodesdlvl.Flush();
		elementmapping_R2PC.Flush();
		unrefinedelementofnode.Flush();
		locposnodeinunrefinedelement.Flush();
		locposelemcenterunrefinedelement.Flush();
		elementshrinkingfactor.Flush();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("Refine"); }

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData ed;
		if(subordinate!=NULL)
		{
			int len = subordinate->NNodes();
			Vector parentnodenrs(len); for(int i=1; i<=len; i++) { parentnodenrs(i) = subordinate->GetNodeMeshNumber(i); }
			ed.SetVector(parentnodenrs.GetVecPtr(), len, "sourcenodes"); ed.SetValuesInt(); ed.SetLocked(1); ed.SetVariableLength(); ed.SetToolTipText("nodes of original component"); edc.TreeAdd("Generation",ed);
			Vector nodesubdivisionlevel(len); for(int i=1; i<=len; i++) { nodesubdivisionlevel(i) = nodesdlvl_seed(i); }
			ed.SetVector(nodesubdivisionlevel.GetVecPtr(), len, "level"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("node's subdivision level"); edc.TreeAdd("Generation",ed);
			ed.SetInt(method,"method"); ed.SetToolTipText("used refinement method: 0 for \'all directions\', 1 for \'no refinement in local x direction\' 2 and 3 for no y and z direction"); edc.TreeAdd("Generation",ed);
		}
		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetLocked(1); }

	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshSourceExtended::SetElementData(edc);

		double* v;
		int l;
		edc.TreeGetVector("Generation.level", &v, l);
		for(int i=1; i<=l; i++) nodesdlvl_seed(i) = (int) (v[i-1]+0.5);
		method = edc.TreeGetInt("Generation.method");
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	void SetSDlvl(const TArray<int>& sdlvl) { nodesdlvl_seed.CopyFrom(sdlvl); }
	void SetMethod(int m) { method = m; }
public: // functionality
	int ParentOfIthElement(int i) { return elementmapping_R2PC(i)(1); }               // returns the array index in the unrefined component
	int ChildNrOfIthElement(int i) { return elementmapping_R2PC(i)(2); }              // returns the child number {1..n} for the given parent element, "0" for unchanged elements
	
	int UnrefinedElementOfIthNode(int i) { return unrefinedelementofnode(i); }                             // returns the element number of the unrefined element the node is associated (lowest number for multiple choices)
	Vector3D LocalPositionOfIthNodeOnUnrefinedElement(int i) { return locposnodeinunrefinedelement(i); }   // returns the local position of the node on the original element
	

public: // overrides - basic component properties
//	virtual int GetBodyIndex(int i=0) { if(subordinate!=NULL) { if(i==0) return GetSubordinateComponent()->GetBodyIndex(); else return GetSubordinateComponent()->GetBodyIndex(ParentOfIthElement(i)); } else return 0; }
	virtual int GetBodyIndex(int i=0) 
	{ 
		if(subordinate!=NULL) 
		{ 
			if(i==0) 
				return GetSubordinateComponent()->GetBodyIndex(); 
			else 
			{
				int nparentelement = UnrefinedElementOfIthNode(i);
				int nparentnode1 = GetSubordinateComponent()->GetElemPtr(nparentelement)->GetNode(1);
				return GetSubordinateComponent()->GetBodyIndex(nparentnode1); 
			}
		} 
		else
			return 0; 
	}
	virtual int GetMaterialNumber(int i=0) { if(subordinate!=NULL) { if(i==0) return GetSubordinateComponent()->GetMaterialNumber(); else return GetSubordinateComponent()->GetMaterialNumber(ParentOfIthElement(i)); } else return 0; }
	virtual Vector3D GetColor(int i=0) { if(subordinate!=NULL) { if(i==0) return GetSubordinateComponent()->GetColor(); else return GetSubordinateComponent()->GetColor(ParentOfIthElement(i)); } else return Vector3D(0.,0.,0.); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { if(NSubs()!=0) { return GetSubordinateComponent()->NNodes()+nodes.Length(); } else return 0; }	// first nodes from subordinate, then additional nodes 
	virtual int NElems() { return elementmapping_R2PC.Length(); } 
	virtual MeshNode* GetNodePtr(int i);
	virtual MeshElement* GetElemPtr(int i);


public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp
private: 
	virtual int Refine(int method=0); // in cpp - one refinement (recursive)
	virtual int FixSDLVLList(); // in cpp - increases  the subdividion level for individual nodes
	virtual int MapTemplateNodeListToActualNodes(TArray<int>& templatenodelist, TArray<int>& unrefinedelementnodes, TArray<int>& additionalnodes); // on cpp - replace entries from template with the actual (component-)node numbers

public: // overrideables - element generation
	virtual int ComputeElementSize(int i) { if(NSubs()!=0) { return GetSubordinateComponent()->ComputeElementSize(i); } else return 0; }
	virtual Vector3D GetElementSizeVector(int i) { if(NSubs()!=0) { Vector3D rv=GetSubordinateComponent()->GetElementSizeVector(i); rv.Y()/=3.; rv.Z()/=3.; return rv; } else return Vector3D(0.,0.,0.); }
	virtual TArray<int> GetElementNodeList(int i);


protected: // own variables
	int method;                              // refining method - 0.. all directions 1,2,3 to skip local 1,2,3 direction 
	TArray<int> nodesdlvl_seed;              // node subdivision level for each node of the original configuration
	TArray<int> nodesdlvl;                   // node subdivision level for each node of the current configuration ( refining operation extends the list and decreses the level continuously )
	TArray<int2> elementmapping_R2PC;        // mapping of all elements in the refined configuration to their parent elements in the original configuration: #refined -> (#parent,#child) #child=0 for untouched
// mapping of nodes back to unrefined	
	TArray<int> unrefinedelementofnode;                  // mapping of all nodes in the refined configuration to their parent elements in the unrefined configuration
	TArray<Vector3D> locposnodeinunrefinedelement;       // mapping of all nodes to the local position in the unrefined parent element
	TArray<Vector3D> locposelemcenterunrefinedelement;   // mapping of the center of childelement to unrefined parent element
	TArray<Vector3D> elementshrinkingfactor;               // size factor of the child elelemts in comparison to the unrefined - expected to be powers of (1/3) for elements that can be split further (same shape as unrefined)
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH EXTRUDE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// extrudes a 1d-curve to quadrilaterals
// or a 2d-planar mesh hexahedrals and prisms
// completely replaces all nodes and elements
class MeshExtrude : public MeshSourceExtended
{
	//TODO: extrudes 2D->3D, 1D->2D
public: // lifecycle I
	MeshExtrude() { Reset(); }
	//MeshExtrude(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshExtrude(const MeshExtrude& other) { CopyFrom(other); }
	~MeshExtrude() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshExtrude(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshSourceExtended& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshExtrude& o = (const MeshExtrude&) other;
	// specific for derived
		extrusion_axis_number = o.extrusion_axis_number;
		extrusion_discretization = o.extrusion_discretization;
		extrusion_distance = o.extrusion_distance;
		extrusion_platethickness = o.extrusion_platethickness;
		nodemapping.CopyFrom(o.nodemapping);

		nodetype2D = o.nodetype2D;
		elemtype2D = o.elemtype2D;

		nodetype3D = o.nodetype3D;
		elemtype3DHex = o.elemtype3DHex;
		elemtype3DPri = o.elemtype3DPri;
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();

		extrusion_axis_number = 3;
		extrusion_discretization = 1;
		extrusion_distance = 1.;
		extrusion_platethickness = 0.1;

		nodemapping.Flush();

		nodetype2D = mystr("Node3DS1S2");
		elemtype2D = mystr("ANCFThinPlate3D");

		nodetype3D = mystr("Node3D");
		elemtype3DHex = mystr("Hexahedral");
		elemtype3DPri = mystr("Prism");
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("Extrude"); }

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetText("dependent","used_node_type"); ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetText("dependent","used_elem_type"); ped->SetLocked(1); }
		ElementData ed;
		ed.SetInt(extrusion_axis_number,"axis_number"); ed.SetToolTipText("direction for extrusion along main axis"); edc.TreeAdd("Generation",ed);
		ed.SetInt(extrusion_discretization,"discretization"); ed.SetToolTipText("discretization along extrusion axis"); edc.TreeAdd("Generation",ed);
		ed.SetDouble(extrusion_distance,"total_extrusion"); ed.SetToolTipText("distance along extrusion axis"); edc.TreeAdd("Generation",ed);
		ed.SetDouble(extrusion_platethickness,"thickness"); ed.SetToolTipText("plate thickness after extrusion"); edc.TreeAdd("Generation",ed);

		int layers = extrusion_discretization+1; // always +1 for extrusion
		if(GetSubordinateComponent()!=NULL)
		{
			int original = GetSubordinateComponent()->NNodes();
			Vector curlayer(original);
			Matrix fullmapping(layers,original);
			ed.SetMatrix(fullmapping.GetMatPtr(),1,1,"FullNodeMapping"); edc.TreeAdd("NodeMapping",ed); // add empty such that entry is first in folder
			for(int i=1; i<=layers; i++)
			{
				for(int j=1; j<=original; j++)
				{
					curlayer(j) = nodemapping( j+(i-1)*original );
					fullmapping(i,j) = curlayer(j);
				}
				mystr layername = mystr("Layer") + mystr(i);
				ed.SetVector(curlayer.GetVecPtr(), curlayer.Length(), layername, 1, curlayer.Length());
				ed.SetValuesInt(); ed.SetLocked(1); ed.SetToolTipText(""); edc.TreeAdd("NodeMapping",ed);
			}
			ElementData* pEd = edc.TreeFind("NodeMapping.FullNodeMapping");
			pEd->SetMatrix(fullmapping.GetMatPtr(),fullmapping.Getrows(),fullmapping.Getcols(),"FullNodeMapping");
			pEd->SetValuesInt(); pEd->SetLocked(1); pEd->SetToolTipText(""); 
		}
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshSourceExtended::SetElementData(edc);
		extrusion_axis_number = edc.TreeGetInt("Generation.axis_number");
		extrusion_discretization = edc.TreeGetInt("Generation.discretization");
		extrusion_distance = edc.TreeGetDouble("Generation.total_extrusion");
		extrusion_platethickness = edc.TreeGetDouble("Generation.thickness");
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	int& ExtrusionAxis() { return extrusion_axis_number; }
	int& ExtrusionDiscretization() { return extrusion_discretization; }
	double& ExtrusionDistance() { return extrusion_distance; }
	double& ExtrusionPlateThickness() { return extrusion_platethickness; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { return nodes.Length(); }									// only new nodes, ignore subordinate
	virtual int NElems() { return elems.Length(); }									// only new elements, ignore subordinate
	virtual MeshNode* GetNodePtr(int i) { return nodes(i); }				// only new nodes, ignore subordinate
	virtual MeshElement* GetElemPtr(int i) { return elems(i); }			// only new elements, ignore subordinate

// ++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp
private: 
	virtual int GenerateOwnNodes(); // in cpp
	virtual int GenerateOwnElements(); // in cpp

public:
	virtual Vector3D GetNodeRefSlope1(int i); // in cpp
	virtual Vector3D GetNodeRefSlope2(int i); // in cpp
	virtual Vector3D GetElementSizeVector(int i); // in cpp
	virtual mystr GetElemType(int i); // in cpp

public:
	virtual int NCoincident();
	virtual int2 GetCoincidentNodeNumbers(int i);
	virtual int GetCoincidentPenaltyFlag(int i);
	virtual void SetAutoConstraintNumber(int i, int cnr);

protected: // own variables
	mystr nodetype2D;
	mystr elemtype2D;
	mystr nodetype3D;
	mystr elemtype3DHex;
	mystr elemtype3DPri;
	int extrusion_axis_number;
	int extrusion_discretization;
	double extrusion_distance;
	double extrusion_platethickness;
	TArray<int> nodemapping;
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH ROTATIONAL
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// rotates a triangular mesh around the selected axis
// completely replaces all nodes and elements
class MeshRotational : public MeshSourceExtended
{
public: // lifecycle I
	MeshRotational() { Reset(); }
	//MeshRotational(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshRotational(const MeshRotational& other) { CopyFrom(other); }
	~MeshRotational() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshRotational(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshRotational& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshRotational& o = (const MeshRotational&) other;
	// specific for derived
		rotation_axis_number = o.rotation_axis_number;
		angular_segments = o.angular_segments;
		final_angle_deg = o.final_angle_deg;
		nodemapping.CopyFrom(o.nodemapping);
		typepri = o.typepri;
		typepyr = o.typepyr;
		typetet = o.typetet;
		typehex = o.typehex;
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();
		nodetype = mystr("Node3D");
		elemtype = mystr("Hexahedral");

		rotation_axis_number = 3;
		angular_segments = 2;
		//final_angle_deg = 90;
		nodemapping.Flush();
		typepri = mystr("Prism");
		typepyr = mystr("Pyramid");
		typetet = mystr("Tetrahedral");
		typehex = mystr("Hexahedral");
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("Rotate2D"); }

	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData ed;
		ed.SetInt(NNodes(),"nnodes"); ed.SetToolTipText("number of nodes read from external source file"); ed.SetLocked(1); edc.TreeAdd("Generation",ed);
		ed.SetInt(NElems(),"nelems"); ed.SetToolTipText("number of elements read from external source file"); ed.SetLocked(1); edc.TreeAdd("Generation",ed);

		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetText("dependent","used_node_type"); ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetText("dependent","used_elem_type"); ped->SetLocked(1); }

		int layers = angular_segments;
		if(final_angle_deg<360.0) { layers++; } //  +1 for incomplete only 

		if(GetSubordinateComponent()!=NULL)
		{
			int original = GetSubordinateComponent()->NNodes();
			Vector curlayer(original);
			Matrix fullmapping(layers,original);
			ed.SetMatrix(fullmapping.GetMatPtr(),1,1,"FullNodeMapping"); edc.TreeAdd("NodeMapping",ed); // add empty such that entry is first in folder
			for(int i=1; i<=layers; i++)
			{
				for(int j=1; j<=original; j++)
				{
					curlayer(j) = nodemapping( j+(i-1)*original );
					fullmapping(i,j) = curlayer(j);
				}
				mystr layername = mystr("Layer") + mystr(i);
				ed.SetVector(curlayer.GetVecPtr(), curlayer.Length(), layername, 1, curlayer.Length());
				ed.SetValuesInt(); ed.SetLocked(1); ed.SetToolTipText(""); edc.TreeAdd("NodeMapping",ed);
			}
			ElementData* pEd = edc.TreeFind("NodeMapping.FullNodeMapping");
			pEd->SetMatrix(fullmapping.GetMatPtr(),fullmapping.Getrows(),fullmapping.Getcols(),"FullNodeMapping");
			pEd->SetValuesInt(); pEd->SetLocked(1); pEd->SetToolTipText(""); 
		}
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	int& RotationAxis() { return rotation_axis_number; }
	int& AngularSegments() { return angular_segments; }
	double& FinalAngleDeg() { return final_angle_deg; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { return nodes.Length(); }									// only new nodes, ignore subordinate
	virtual int NElems() { return elems.Length(); }									// only new elements, ignore subordinate
	virtual MeshNode* GetNodePtr(int i) { return nodes(i); }				// only new nodes, ignore subordinate
	virtual MeshElement* GetElemPtr(int i) { return elems(i); }			// only new elements, ignore subordinate

// ++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp
private: 
	virtual int GenerateOwnNodes(); // in cpp
	virtual int GenerateOwnElements(); // in cpp

protected: // own variables
	mystr typepri;
	mystr typepyr;
	mystr typetet;
	mystr typehex;
	int rotation_axis_number;
	int angular_segments;
	double final_angle_deg;
	TArray<int> nodemapping;
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH MIRROR
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// generates the mirror image of a given mesh (in addition to the original)
// completely replaces all nodes and elements
class MeshMirror : public MeshSourceExtended
{
public: // lifecycle I
	MeshMirror() { Reset(); }
	//MeshMirror(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshMirror(const MeshMirror& other) { CopyFrom(other); }
	~MeshMirror() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshMirror(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshMirror& other)
	{
		MeshSourceExtended::CopyFrom(other);

		const MeshMirror& o = (const MeshMirror&) other;
	// specific for derived	
		mirror_plane_number = o.mirror_plane_number;
		mirror_distance = o.mirror_distance;
		nodemapping.CopyFrom(o.nodemapping);
	}
	virtual void Reset()
	{
		MeshSourceExtended::Reset();
		
		mirror_plane_number = 1;
		mirror_distance = 0.0;
		nodemapping.Flush();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("Mirror"); }

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshSourceExtended::GetElementData(edc);

		ElementData ed;
		ed.SetInt(MirrorPlane(), "plane"); ed.SetToolTipText("1 for YZ plane, 2 for XZ plane, 3 for XY plane"); edc.Add(ed);
		ed.SetDouble(MirrorDistance(), "distance"); ed.SetToolTipText("mirror planes' distance from origin"); edc.Add(ed);

		ElementData* ped; // lock type entries
		ped = edc.TreeFind("Generation.used_node_type"); if(ped!=NULL) { ped->SetText("dependent","used_node_type"); ped->SetLocked(1); }
		ped = edc.TreeFind("Generation.used_elem_type"); if(ped!=NULL) { ped->SetText("dependent","used_elem_type"); ped->SetLocked(1); }

		if(GetSubordinateComponent()!=NULL)
		{
			int original = GetSubordinateComponent()->NNodes();
			Vector originalnodes(original);
			Vector copiednodes(original);
			Matrix fullmapping(2,original);
			for(int j=1; j<=original; j++)
			{
				originalnodes(j) = nodemapping(j);
				fullmapping(1,j) = originalnodes(j);
				copiednodes(j) =   nodemapping(j+original);
				fullmapping(2,j) = copiednodes(j);
			}
			ed.SetMatrix(fullmapping.GetMatPtr(),fullmapping.Getrows(),fullmapping.Getcols(),"FullNodeMapping");
			ed.SetValuesInt(); ed.SetLocked(1); edc.TreeAdd("NodeMapping",ed);
			ed.SetVector(originalnodes.GetVecPtr(), originalnodes.Length(), "original", 1, originalnodes.Length());
			ed.SetValuesInt(); ed.SetLocked(1); ed.SetToolTipText(""); edc.TreeAdd("NodeMapping",ed);
			ed.SetVector(copiednodes.GetVecPtr(), copiednodes.Length(), "copies", 1, copiednodes.Length());
			ed.SetValuesInt(); ed.SetLocked(1); ed.SetToolTipText(""); edc.TreeAdd("NodeMapping",ed);
		}
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshSourceExtended::SetElementData(edc);
		MirrorPlane()=edc.TreeGetInt("plane");
		MirrorDistance()=edc.TreeGetDouble("distance");
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	int& MirrorPlane() { return mirror_plane_number; }
	double& MirrorDistance() { return mirror_distance; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access
	virtual int NNodes() { if(NSubs()!=0) { return GetSubordinateComponent()->NNodes()+nodes.Length(); } else return 0; } // first nodes from subordinate, then additional nodes 
	virtual int NElems() { if(NSubs()!=0) { return GetSubordinateComponent()->NElems()+elems.Length(); } else return 0; } // first elems from subordinate, then additional elems
	virtual MeshNode* GetNodePtr(int i);																																									// first nodes from subordinate, then additional nodes 
	virtual MeshElement* GetElemPtr(int i);																																								// first elems from subordinate, then additional elems

public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp
private: 
	virtual int GenerateOwnNodes(); // in cpp
	virtual int GenerateOwnElements(); // in cpp

public: // overrideables - node generation 
	virtual int GetNodeObjectFactoryType(int i); 
	virtual void SetNodeMeshNumber(int i, int idx); 
	virtual int GetNodeMeshNumber(int i); 
	virtual int ComputeOrientation(int i); 
	virtual Vector3D GetNodePosition(int i); 
	virtual Vector3D GetNodeRefSlope1(int i);
	virtual Vector3D GetNodeRefSlope2(int i);
	virtual Vector3D GetNodeRefSlope3(int i);
	virtual Vector3D GetNodeRefAngles(int i);
//NOT virtual int GetNodeBodyIndex(int i);

public: // overrideables - element generation
	virtual int GetElementObjectFactoryType(int i);
	virtual int GetElementMaterialNumber(int i);
	//	virtual int GetElementGeometricNonlinearityStatus(int i=0) { if(component_gnls!=-1) return component_gnls; /*else*/ if(subordinate!=NULL) return GetSubordinateComponent()->GetElementGeometricNonlinearityStatus(i); else return -1; }
	virtual void SetElementMeshNumber(int i, int idx);
	virtual int GetElementMeshNumber(int i);
	virtual TArray<int> GetElementNodeList(int i);
	virtual TArray<int> GetElementFaceNodeList(int i, int f);
//NOT	virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);

	virtual int ComputeElementSize(int i);
	virtual Vector3D GetElementSizeVector(int i);
	virtual int GetElemAlternateFlag(int i);
	virtual mystr GetElemType(int i=0);

public:
	virtual int NCoincident();
	virtual int2 GetCoincidentNodeNumbers(int i);
	virtual int GetCoincidentPenaltyFlag(int i);
	virtual void SetAutoConstraintNumber(int i, int cnr);

protected: // own variables
	int mirror_plane_number;
	double mirror_distance;
	TArray<int> nodemapping;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH PROCESSOR
// MESH MULTI PROCESSOR
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MeshProcessors have subordinate Meshes and NO own nodes and element lists
// * simple processors only manipulate existing nodes or elements
// **: Transform, Distort
// * MULTI processors (glue) have more than one subordinate mesh components 
// **: Glue, Loader
class MeshProcessor : public MeshComponent
{
public: // lifecycle I
	MeshProcessor():filter(NULL) {	Reset(); }
	//MeshProcessor(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshProcessor(const MeshProcessor& other) { CopyFrom(other); }
	~MeshProcessor() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshProcessor(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshProcessor& other)
	{
		MeshComponent::CopyFrom(other);

		const MeshProcessor& o = (const MeshProcessor&)other;
// specific for derived
		if(o.NSubs()>0)	
		{ 
			subordinate = o.subordinate->GetCopy();
		}
		filter = o.filter;
	}
	virtual void Reset()
	{
		MeshComponent::Reset();
		subordinate = NULL;
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Processor"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponent::GetElementData(edc);

		ElementData ed;
		for(int i=1; i<=NSubs(); i++) 
		{
			ElementDataContainer edc_sub;
			GetSubordinateComponent(i)->GetElementData(edc_sub);
			ed.SetEDC(&edc_sub,GetSubordinateComponent(i)->GetName()); edc.Add(ed);
		}
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
		
		for(int i=1; i<=NSubs(); i++)
		{
			ElementDataContainer* p_edc_sub;
			ElementData* p_ed_sub = edc.TreeFind(GetSubordinateComponent(i)->GetName());
			if(p_ed_sub!=NULL && p_ed_sub->IsEDC())
			{
				p_edc_sub = p_ed_sub->GetEDC();
				GetSubordinateComponent(i)->SetElementData(*p_edc_sub);
			}
		}
		return rv;
	}
	virtual void SetFilter(BaseSet* filter)
	{
		this->filter = filter;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual void SetSubordinate(MeshComponent* sub, int = 0) { subordinate = sub; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access - always go to subordinate
	virtual int NNodes() { if(subordinate!=NULL) return GetSubordinateComponent()->NNodes(); else return 0; }
	virtual int NElems() { if(subordinate!=NULL) return GetSubordinateComponent()->NElems(); else return 0; }
	virtual MeshNode* GetNodePtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePtr(i); else return NULL; }
	virtual MeshElement* GetElemPtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElemPtr(i); else return NULL; }

public: // overrideables - meshing
//NOT	virtual int MeshYourself() { return 0; }
//NOT virtual int GenerateOwnNodes() { return 0; }
//NOT virtual int GenerateOwnElements() { return 0; 

public: // overrideables - node generation - always go to subordinate
	virtual int GetNodeObjectFactoryType(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeObjectFactoryType(i); else return 0; }
	virtual void SetNodeMeshNumber(int i, int idx) { if(subordinate!=NULL) GetSubordinateComponent()->SetNodeMeshNumber(i,idx); }
	virtual int GetNodeMeshNumber(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeMeshNumber(i); else return 0;}
	virtual int ComputeOrientation(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->ComputeOrientation(i); else return 0; }
	virtual Vector3D GetNodePosition(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePosition(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope1(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope1(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope2(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope2(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope3(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefSlope3(i); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefAngles(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeRefAngles(i); else return Vector3D(0.,0.,0.); }
	virtual int GetNodeBodyIndex(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeBodyIndex(i); else return 0; }

public: // overrideables - element generation - always go to subordinate
	virtual int GetElementObjectFactoryType(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementObjectFactoryType(i); else return 0; }
	virtual int GetElementMaterialNumber(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementMaterialNumber(i); else return 0; }
	virtual int GetElementGeometricNonlinearityStatus(int i=0) { if(component_gnls!=-1) return component_gnls; /*else*/ if(subordinate!=NULL) return GetSubordinateComponent()->GetElementGeometricNonlinearityStatus(i); else return -1; }
	virtual void SetElementMeshNumber(int i, int idx) { if(subordinate!=NULL) GetSubordinateComponent()->SetElementMeshNumber(i,idx); }
	virtual int GetElementMeshNumber(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementMeshNumber(i); else return 0; }
	virtual TArray<int> GetElementNodeList(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementNodeList(i); return TArray<int>(0); } // TArray<int>& should be used when possible
	virtual TArray<int> GetElementFaceNodeList(int i, int f) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementFaceNodeList(i,f); return TArray<int>(0); }
	virtual void MapNodeListComponentToMesh(TArray<int>& nodelist) { if(subordinate!=NULL) GetSubordinateComponent()->MapNodeListComponentToMesh(nodelist); }
	virtual int ComputeElementSize(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->ComputeElementSize(i); else return 0; }
	virtual Vector3D GetElementSizeVector(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElementSizeVector(i); return Vector3D(0.,0.,0.); }
	virtual mystr GetElemType(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElemType(i); else return mystr("");; }

public: // overrideables - subordinate mesh components
	virtual int NSubs() { if(subordinate!= NULL) return 1; else return 0; } 
	virtual int const NSubs() const { if(subordinate!= NULL) return 1; else return 0; } 
	virtual MeshComponent* GetSubordinateComponent(int i=1) { return subordinate; }
	virtual int NCoincident() { if(subordinate!=NULL) return GetSubordinateComponent()->NCoincident(); else return 0; }
	virtual int2 GetCoincidentNodeNumbers(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetCoincidentNodeNumbers(i); else return int2(0,0); }
	virtual int GetCoincidentPenaltyFlag(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetCoincidentPenaltyFlag(i); else return 0; }
	virtual void SetAutoConstraintNumber(int i, int cnr) { if(subordinate!=NULL) return GetSubordinateComponent()->SetAutoConstraintNumber(i,cnr); }
public: // overrideables - picking
//NOT virtual int ComputeBoundingBox() { return 0; }
//NOT virtual Box3D GetBoundingBox() { return boundingbox; }

protected: // own variables
	MeshComponent* subordinate;
	BaseSet * filter;
};

class MeshMultiProcessor : public MeshProcessor
{
public: // lifecycle I
	MeshMultiProcessor() { Reset(); }
	//MeshMultiProcessor(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshMultiProcessor(const MeshMultiProcessor& other) { CopyFrom(other); }
	~MeshMultiProcessor() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshMultiProcessor(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshMultiProcessor& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshMultiProcessor& o = (const MeshMultiProcessor&)other;
// specific for derived
		subordinates.CopyFrom(other.subordinates);
		nodemapping_I2CN.CopyFrom(other.nodemapping_I2CN);
		nodemapping_CN2I.CopyFrom(other.nodemapping_CN2I);
		elementmapping_I2CE.CopyFrom(other.elementmapping_I2CE);
		elementmapping_CE2I.CopyFrom(other.elementmapping_CE2I);
		coincident_pairs.CopyFrom(other.coincident_pairs);
	}
	virtual void Reset()
	{
		MeshProcessor::Reset();
		subordinates.Flush();
		nodemapping_I2CN.Flush();
		nodemapping_CN2I.Flush();
		elementmapping_I2CE.Flush();
		elementmapping_CE2I.Flush();
		coincident_pairs.Flush();
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("MultiProcessor"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshProcessor::GetElementData(edc);

// remove entries
		ElementData* gen = edc.TreeFind("Generation");
		ElementData* gra = edc.TreeFind("Graphics");
		for(int i=1; i<=edc.Length(); i++)
		{
			ElementData* cur = edc.GetPtr(i);
			if(cur==gen || cur==gra) 
			{ 
				edc.Delete(i--); 
			}
		}
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual void SetSubordinate(MeshComponent* sub, int i=1) 
	{
		if(i>NSubs())
		{
			int nsubsbefore = NSubs();
			subordinates.ReSize(i);
			for(int j=nsubsbefore+1; j<=i-1; j++)
			{
				subordinates(j) = NULL;
			}
		}
		subordinates(i) = sub; 
	}
	int AddSubordinate(MeshComponent* sub) { return subordinates.Add(sub); }

public: // access node mapping (array based)
	int IthNode(int comp, int node) { return nodemapping_CN2I(comp)(node); }
	int ComponentNumberOfIthNode(int i) { return nodemapping_I2CN(i)(1); }
	int NCofI(int nodenr) { return ComponentNumberOfIthNode(nodenr); }
	int ComponentNodeNumberOfIthNode(int i) { return nodemapping_I2CN(i)(2); }
	int NofI(int nodenr) { return ComponentNodeNumberOfIthNode(nodenr); }
public: // access element mapping (array ased)
	int IthElement(int comp, int elem) { return elementmapping_CE2I(comp)(elem); }
	int ComponentNumberOfIthElement(int i) { return elementmapping_I2CE(i)(1); }
	int ECofI(int elemnr) { return ComponentNumberOfIthElement(elemnr); }
	int ComponentElementNumberOfIthElement(int i) { return elementmapping_I2CE(i)(2); }
	int EofI(int elemnr) { return ComponentElementNumberOfIthElement(elemnr); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overridables - access
	virtual int NNodes() { return nodemapping_I2CN.Length(); } 
	virtual int NElems() { return elementmapping_I2CE.Length(); } 
	virtual MeshNode* GetNodePtr(int i); // in cpp { return NULL; } 
	virtual MeshElement* GetElemPtr(int i); // in cpp { return NULL; } 
	//$EK 2017-01-11: pass alternate-flag of subordinates, e.g. in case of MeshGlue
	virtual int GetElemAlternateFlag(int i); // in cpp { return NULL; } 
public: // overrideables - meshing
	virtual int MeshYourself() { return 0; } // different for each derived class

public: // overrideables - node generation 
	virtual int GetNodeObjectFactoryType(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeObjectFactoryType(NofI(i)); else return 0; }
	virtual void SetNodeMeshNumber(int i, int idx); // in cpp - mapping in Glue AND forward to source
	virtual int GetNodeMeshNumber(int i) { return nodeIdx(i); } // mapping is in Glue component

	virtual int ComputeOrientation(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> ComputeOrientation(NofI(i)); else return 0; }
	virtual Vector3D GetNodePosition(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodePosition(NofI(i)); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefSlope1(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeRefSlope1(NofI(i)); else return Vector3D(0.,0.,0.); } 
	virtual Vector3D GetNodeRefSlope2(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeRefSlope2(NofI(i)); else return Vector3D(0.,0.,0.); } 
	virtual Vector3D GetNodeRefSlope3(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeRefSlope3(NofI(i)); else return Vector3D(0.,0.,0.); }
	virtual Vector3D GetNodeRefAngles(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeRefAngles(NofI(i)); else return Vector3D(0.,0.,0.); }
	virtual int GetNodeBodyIndex(int i) { if(NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetNodeBodyIndex(NofI(i)); else return 0; }

	virtual int GetElementObjectFactoryType(int i) { if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetElementObjectFactoryType(EofI(i)); else return 0; }
	virtual int GetElementMaterialNumber(int i) { if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetElementMaterialNumber(EofI(i)); else return 0; }
	virtual int GetElementGeometricNonlinearityStatus(int i=0) { if(component_gnls!=-1) return component_gnls; /*else*/ if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetElementGeometricNonlinearityStatus(EofI(i)); else return -1; }
	virtual void SetElementMeshNumber(int i, int idx); // in cpp 
	virtual int GetElementMeshNumber(int i) { return elemIdx(i); }  //{ if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetElementMeshNumber(EofI(i)); else return 0; }
	virtual TArray<int> GetElementNodeList(int i); // TArray<int>& should be used when possible
	virtual TArray<int> GetElementFaceNodeList(int i,int f);
	virtual void MapNodeListComponentToMesh(TArray<int>& nodelist);
	virtual int ComputeElementSize(int i) { if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> ComputeElementSize(EofI(i)); else return 0; }
	virtual Vector3D GetElementSizeVector(int i) { if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetElementSizeVector(EofI(i)); else return Vector3D(0.,0.,0.);; }
	virtual mystr GetElemType(int i=0)
	{
		mystr type;
		if (i==0)
		{
			if (NSubs()>0)
			{
				type = GetSubordinateComponent(1) -> GetElemType(1);
			}
			else
			{
				type = mystr(""); 
			}
		}
		else if (NSubs()>0)
		{
			type = GetSubordinateComponent(ECofI(i)) -> GetElemType(EofI(i));
		}
		else
		{
			type = mystr(""); 
		}
		return type;
	}

public: // overrides - basic component properties
	virtual int GetBodyIndex(int i=0) { if(i==0) return bodynr; /*else*/ if (NSubs()>0) return GetSubordinateComponent(NCofI(i)) -> GetBodyIndex(NofI(i)); else return 0; } // BodyIndex is a property of the NODE
	virtual int GetMaterialNumber(int i=0) { if(i==0) return matnr; /*else*/ if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetMaterialNumber(EofI(i)); else return 0; } // MaterialIndex is a property of the ELEMENT
	virtual Vector3D GetColor(int i=0) { if(i==0) return color; /*else*/ if(NSubs()>0) return GetSubordinateComponent(ECofI(i)) -> GetColor(EofI(i)); else return Vector3D(0.,0.,0.); } // Color is a property of the ELEMENT

public: // overrideables - subordinate mesh components
	virtual int NSubs() { return subordinates.Length(); }
	virtual MeshComponent* GetSubordinateComponent(int i=1) { return subordinates(i); }
	virtual int NCoincident(); // in cpp
	virtual int2 GetCoincidentNodeNumbers(int i); // in cpp - returns the MAPPED nodenumbers
protected: // overrideables - meshing
	virtual int ComputeBoundingBox(); // in cpp

protected: // own variables
	TArray<MeshComponent*> subordinates;
// node mapping, the multiprocessor does not have a nodelist on its own...
	TArray<int2> nodemapping_I2CN;         // maps node "i" in the MultiProcessor object to the first occurrence (in all subordinates) as node "n" in the subordinate "c"
	TArray<TArray<int>> nodemapping_CN2I;  // maps every node "n" in subordinate"c" to node "i" in multiprocessor
// element mapping, the multiprocessor does not have a elementmapping on its own...
	TArray<int2> elementmapping_I2CE;
	TArray<TArray<int>> elementmapping_CE2I;
// list of special pairs
	TArray<int4> coincident_pairs;         // pairs two entries (componentnumber,componentnodenumber) that are coincident but not redundant ( == same position, different domain or orientation )

// AD: NOTE on node/element mappings:
// indexed I is always the nodes / elements visible for the 'outside', the nodes/elements actually forwarded to the MBS
// indexed CN,CE are always the full set of nodes/elements available. During LoadFromFile operations additional nodes/elements may be generated (domain splitting)
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH TRANSFORMATION 
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// applies the transformation matrix to compute position and orientation from subordinate (untransformed) mesh component
// constant transformation matrix
class MeshTransformation : public MeshProcessor
{
public: // lifecycle I
	MeshTransformation() {	Reset(); }
	//MeshTransformation(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshTransformation(const MeshTransformation& other) { CopyFrom(other); }
	~MeshTransformation() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshTransformation(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshTransformation& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshTransformation& o = (const MeshTransformation&)other;

// specific for derived
		transformation4D = o.transformation4D;
	}
	virtual void Reset()
	{
		MeshProcessor::Reset();
	
		transformation4D.Set44(1., 0., 0., 0.,
                               0., 1., 0., 0.,
							   0., 0., 1., 0.,
							   0., 0., 0., 1.);
	}
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Transformation"); } 
	virtual int IsTransformation() { return 1; }

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshProcessor::GetElementData(edc);

		ElementData ed;
		ed.SetMatrix(TransformationMatrix().GetMatPtr(),TransformationMatrix().Getrows(),TransformationMatrix().Getcols(),"transformation_matrix"); 
		ed.SetToolTipText("4 dimensional transformation matrix - translation+scale+rotation"); 
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
			
		double* v;
		int r,c;
		edc.TreeGetMatrix("Geometry.transformation_matrix",&v,r,c);
		TransformationMatrix().Set44( v[ 0], v[ 1], v[ 2], v[ 3],
		                              v[ 4], v[ 5], v[ 6], v[ 7],
		                              v[ 8], v[ 9], v[10], v[11],
		                              v[12], v[13], v[14], v[15]);
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	Matrix3D& TransformationMatrix() { return transformation4D; }
public: // functionality
// push transformation matrix
	virtual int ScaleXYZ(Vector3D& scale); 
	virtual int RotateXYZ(Vector3D& rotate);
	virtual int TranslateXYZ(Vector3D& translate);
// apply transformation 
	virtual void TransformPositionVector(Vector3D& position);
	virtual void TransformOrientationVector(Vector3D& orientation); 

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overrides - basic component properties
	virtual int GetBodyIndex(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetBodyIndex(i); else return 0; }
	virtual int GetMaterialNumber(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetMaterialNumber(i); else return 0; }
	virtual Vector3D GetColor(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetColor(i); else return Vector3D(0.,0.,0.); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access - always go to subordinate
	virtual int NNodes() { if(subordinate!=NULL) return GetSubordinateComponent()->NNodes(); else return 0; }
	virtual int NElems() { if(subordinate!=NULL) return GetSubordinateComponent()->NElems(); else return 0; }
	virtual MeshNode* GetNodePtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePtr(i); else return NULL; }
	virtual MeshElement* GetElemPtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElemPtr(i); else return NULL; }

public: // overrideables - meshing
	virtual int MeshYourself(); 

public: // overrideables - node generation
	virtual Vector3D GetNodePosition(int i); 
	virtual Vector3D GetNodeRefSlope1(int i); 
	virtual Vector3D GetNodeRefSlope2(int i); 
	virtual Vector3D GetNodeRefSlope3(int i);
	virtual Vector3D GetNodeRefAngles(int i);
	virtual int GetNodeBodyIndex(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodeBodyIndex(i); else return 0; }

public: // overrideables - picking
	virtual int ComputeBoundingBox(); // in cpp - apply transformation on the Bounding box of the subordinate component
	                                  // more accurate would be to loop over transformed nodes ...

protected: // own variables
	Matrix3D transformation4D; // accumulated transformation matrix
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH DISTORTION
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// applies a mathfunction to compute position and orientation from a subordinate (undistorted) mesh
// transformation matrix function of position
class MeshDistortion : public MeshProcessor
{
	//MeshDistortion() {	Reset(); }
	public: // lifecycle I
	MeshDistortion(MBS* mbs): pParser(&(mbs->MBSParser())) {	Reset(); }
	//MeshDistortion(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshDistortion(const MeshDistortion& other) { CopyFrom(other); }
	~MeshDistortion() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshDistortion(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshDistortion& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshDistortion& o = (const MeshDistortion&)other;
// specific for derived
		PXofXYZ.CopyFrom(o.PXofXYZ);
		PYofXYZ.CopyFrom(o.PYofXYZ);
		PZofXYZ.CopyFrom(o.PZofXYZ);
		OXofXYZ.CopyFrom(o.OXofXYZ);
		OYofXYZ.CopyFrom(o.OYofXYZ);
		OZofXYZ.CopyFrom(o.OZofXYZ);
	}
	virtual void Reset(); // in .cpp
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Distortion"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshProcessor::GetElementData(edc);

		ElementData ed;
		ed.SetText(PXofXYZ.GetParsedFunction(), "fX"); ed.SetToolTipText("function defining the new x coordinate dependent on original (x,y,z) coordinates"); edc.Add(ed);
		ed.SetText(PYofXYZ.GetParsedFunction(), "fY"); ed.SetToolTipText("function defining the new y coordinate dependent on original (x,y,z) coordinates"); edc.Add(ed);
		ed.SetText(PZofXYZ.GetParsedFunction(), "fZ"); ed.SetToolTipText("function defining the new z coordinate dependent on original (x,y,z) coordinates"); edc.Add(ed);
		ed.SetText(OXofXYZ.GetParsedFunction(), "nX"); ed.SetToolTipText("function defining the new x direction of the orientation vector dependent on original (x,y,z) coordinates"); edc.Add(ed);
		ed.SetText(OYofXYZ.GetParsedFunction(), "nY"); ed.SetToolTipText("function defining the new y direction of the orientation vector dependent on original (x,y,z) coordinates"); edc.Add(ed);
		ed.SetText(OZofXYZ.GetParsedFunction(), "nZ"); ed.SetToolTipText("function defining the new z direction of the orientation vector dependent on original (x,y,z) coordinates"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
		PXofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("fX"), "x,y,z");
		PYofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("fY"), "x,y,z");
		PYofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("fZ"), "x,y,z");
		OXofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("nX"), "x,y,z");
		OYofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("nY"), "x,y,z");
		OYofXYZ.SetParsedFunction3D(pParser, edc.TreeGetString("nZ"), "x,y,z");
		return rv;
	}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual void SetFunctionString(int i, mystr& str)
	{
		if(i==1) { PXofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
		if(i==2) { PYofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
		if(i==3) { PZofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
		if(i==4) { OXofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
		if(i==5) { OYofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
		if(i==6) { OZofXYZ.SetParsedFunction3D(pParser, str, "x,y,z"); }
	}
	virtual mystr GetFunctionString(int i)
	{
		if(i==1) { return PXofXYZ.GetParsedFunction(); }
		if(i==2) { return PYofXYZ.GetParsedFunction(); }
		if(i==3) { return PZofXYZ.GetParsedFunction(); }
		if(i==4) { return OXofXYZ.GetParsedFunction(); }
		if(i==5) { return OYofXYZ.GetParsedFunction(); }
		if(i==6) { return OZofXYZ.GetParsedFunction(); }
		return "";
	}
public: // functionality
// apply transformation 
	virtual void TransformPositionVector(Vector3D& position);
	virtual void TransformOrientationVector(Vector3D& orientation); 

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overrides - basic component properties
	virtual int GetBodyIndex(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetBodyIndex(i); else return 0; }
	virtual int GetMaterialNumber(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetMaterialNumber(i); else return 0; }
	virtual Vector3D GetColor(int i=0) { if(subordinate!=NULL) return GetSubordinateComponent()->GetColor(i); else return Vector3D(0.,0.,0.); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public: // overridables - access - always go to subordinate
	virtual int NNodes() { if(subordinate!=NULL) return GetSubordinateComponent()->NNodes(); else return 0; }
	virtual int NElems() { if(subordinate!=NULL) return GetSubordinateComponent()->NElems(); else return 0; }
	virtual MeshNode* GetNodePtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetNodePtr(i); else return NULL; }
	virtual MeshElement* GetElemPtr(int i) { if(subordinate!=NULL) return GetSubordinateComponent()->GetElemPtr(i); else return NULL; }

public: // overrideables - meshing
	virtual int MeshYourself(); 

public: // overrideables - node generation
	virtual Vector3D GetNodePosition(int i); 
	//virtual Vector3D GetNodeRefSlope1(int i); 
	//virtual Vector3D GetNodeRefSlope2(int i); 
	//virtual Vector3D GetNodeRefSlope3(int i);
	//virtual Vector3D GetNodeRefAngles(int i);

public: // overrideables - picking
	virtual int ComputeBoundingBox(); // in cpp

protected: // own variables
	CParsedFunction PXofXYZ;
	CParsedFunction PYofXYZ;
	CParsedFunction PZofXYZ;
	CParsedFunction OXofXYZ;
	CParsedFunction OYofXYZ;
	CParsedFunction OZofXYZ;
	CMBSParser* pParser;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH MODIFIER
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// allows to change properties of the mesh element
// change is defined by : { entity to change, new value, (filterset) }
class MeshModifier : public MeshProcessor
{
public: // lifecycle I
	MeshModifier() {	Reset(); }
	//MeshModifier(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshModifier(const MeshModifier& other) { CopyFrom(other); }
	~MeshModifier() {}	

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshModifier(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshModifier& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshModifier& o = (const MeshModifier&)other;
// specific for derived
		entity = o.entity;
		newInt = o.newInt;
		newDouble = o.newDouble;
		newString = o.newString;
		oldInt = o.oldInt;
		oldDouble = o.oldDouble;
		oldString = o.oldString;
	}
	virtual void Reset()
	{
		MeshProcessor::Reset();
		entity = WOE_none;
		newInt = 0;
		newDouble = 0.0;
		newString = mystr("");
		oldInt = 0;
		oldDouble = 0.0;
		oldString = mystr("");
	}
public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Modifier"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshProcessor::GetElementData(edc);

		ElementData ed;
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
			
		return rv;
	}


public: // specific overrides
	virtual int MeshYourself();
	virtual int GetNodeObjectFactoryType(int i); 
	virtual int GetElementObjectFactoryType(int i); 
	virtual int GetElementMaterialNumber(int i);
	virtual int GetElementGeometricNonlinearityStatus(int i);

	enum WorkingOnEntity /*binary!*/
	{
		WOE_none = 0,
		WOE_elemtype = 1,
		WOE_nodetype = 2,
		WOE_materialnumber = 4,
		WOE_nonlinearity = 8
	};
	virtual void SetEntity(WorkingOnEntity entityi) { entity = entityi; }
	virtual void SetNewInt(int newVal) { newInt = newVal; }
	virtual void SetNewDouble(double newVal) { newDouble = newVal; }
	virtual void SetNewString(mystr newVal) { newString = newVal; }
	virtual void SetOldInt(int oldVal) { oldInt = oldVal; }
	virtual void SetOldDouble(double oldVal) { oldDouble = oldVal; }
	virtual void SetOldString(mystr oldVal) { oldString = oldVal; }

protected: // own variables
	WorkingOnEntity entity;
	TArray<int> applymask;

	int newInt,oldInt;
	double newDouble,oldDouble;
	mystr newString,oldString;
};



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH WRITER & derived
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MeshComponentWriter : public MeshProcessor
{
public: // lifecycle I 
	MeshComponentWriter() { Reset(); }
	//MeshComponentWriter(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshComponentWriter(const MeshComponentWriter& other) { CopyFrom(other); }
	~MeshComponentWriter() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshComponentWriter(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshComponentWriter& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshComponentWriter& o = (const MeshComponentWriter&) other;
	// specific for derived
		filename = other.filename;
	}
	virtual void Reset()
	{
		MeshProcessor::Reset();
		filename = mystr("");
	}

public: // class specific
	mystr GetComponentSpecification() { return mystr("Loader"); }

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshProcessor::GetElementData(edc);

		ElementData ed;
		ed.SetText(filename.c_str(),"target_file"); ed.SetToolTipText("target file for the mesh"); edc.TreeAdd("Generation",ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshProcessor::SetElementData(edc);

		FileName() = edc.TreeGetString("Generation.target_file");
		return rv;
	}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual const mystr GetFileName() const { return filename; }
	virtual void SetFileName(mystr filenamei) { filename = filenamei; }
	virtual mystr& FileName() { return filename; }
private: // additional functions - load
	virtual int WriteToFile() { return 0; } 
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
	virtual int MeshYourself(); // in cpp - Writer classes write to file
//NOT virtual int GenerateOwnNodes() { return 0; }
//NOT virtual int GenerateOwnElements() { return 0; }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
protected: // own variables
	mystr filename;                         // file to load data from
};

class MeshWriterNeutral3D : public MeshComponentWriter
{
public: // lifecycle I
	MeshWriterNeutral3D() { Reset(); }
	//MeshWriterNeutral3D(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshWriterNeutral3D(const MeshWriterNeutral3D& other) { CopyFrom(other); }
	~MeshWriterNeutral3D() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshWriterNeutral3D(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshWriterNeutral3D& other)
	{
		MeshComponentWriter::CopyFrom(other);

		const MeshWriterNeutral3D& o = (const MeshWriterNeutral3D&) other;
	// specific for derived	
	}
	virtual void Reset() 
	{
		MeshComponentWriter::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("MeshWriterNeutral3D"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int WriteToFile();
	virtual int PermuteNodeList(TArray<int>& nodelist);
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ADVANCED MESH MANIPULATIONS 
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH GLUE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MeshGlue : public MeshMultiProcessor
{
public: // lifecycle I
	MeshGlue() {	Reset(); }
	//MeshGlue(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshGlue(const MeshGlue& other) { CopyFrom(other); }
	~MeshGlue() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{	
		MeshComponent* mcc = new MeshGlue(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshGlue& other)
	{
		MeshProcessor::CopyFrom(other);

		const MeshGlue& o = (const MeshGlue&)other;
// specific for derived
		usepenalty = o.usepenalty;
		redundant_pairs.CopyFrom(other.redundant_pairs);
		coincident_pairs.CopyFrom(other.coincident_pairs);
		autoconstraintnrs.CopyFrom(other.autoconstraintnrs);
	}
	virtual void Reset()
	{
		MeshProcessor::Reset();

		usepenalty=0;
		redundant_pairs.Flush();
		coincident_pairs.Flush();
		autoconstraintnrs.Flush();
	}

public: // class specific
	virtual mystr GetComponentSpecification() { return mystr("Glue"); } 

public: // EDC-Interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshMultiProcessor::GetElementData(edc);

		ElementData ed;
		ed.SetBool(usepenalty,"penalty"); ed.SetToolTipText("use penalty formulation for the automatically generated constraints"); edc.Add(ed);
		
		Vector cnrs(autoconstraintnrs.Length()); for(int i=1; i<=autoconstraintnrs.Length(); i++) { cnrs(i) = autoconstraintnrs(i); } // nodes for this component in MBS numbering
		ed.SetVector(cnrs.GetVecPtr(), cnrs.Length(), "constraints"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the automatically generated constraints"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponent::SetElementData(edc);
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	void SetUsePenaltyFlag(int use_penalty) { usepenalty = use_penalty; }
	int GetUsePenaltyFlag() { return usepenalty; }

public: // functionality
	int IdentifyNodePairs();	       // component-by-component, computes two list of node pairs, redundant_pairs and coincident_pairs
private:
	int GenerateNodeMapping();         //
	int GenerateElementMapping();      //
	virtual int GetCoincidentPenaltyFlag(int i) { return GetUsePenaltyFlag(); }
	virtual void SetAutoConstraintNumber(int i, int cnr); // in cpp { autoconstraintnrs(i)=cnr; }
private:
//	virtual int NSubs();
//	virtual MeshComponent* GetSubordinateComponent(int i);
	virtual int NSubsWLoaders();       // class specific implementation to deal with Loaders ( loader as a single subordinate of Glue contributes all of its domains to the component count )
	virtual MeshComponent* GetSubordinateComponentWLoaders(int i=1);

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // overrideables - meshing
	virtual int MeshYourself(); // in cpp  - generate meshes for subordinates first 

protected: // own variables
	int usepenalty;
	TArray<int4> redundant_pairs;
	TArray<int> autoconstraintnrs;
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MESH LOADER & derived
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// container class that holds the nodes and elements of one domain
// loader may generate more than one of these containers
class MeshDomainContainer : public MeshComponentSource
{
public: // lifecycle I
	MeshDomainContainer() { Reset(); }
	//MeshDomainContainer(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshDomainContainer(const MeshDomainContainer& other) { CopyFrom(other); }
	~MeshDomainContainer() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshDomainContainer(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshDomainContainer& other)
	{
		MeshComponentSource::CopyFrom(other);

		const MeshDomainContainer& o = (const MeshDomainContainer&) other;
	}
	virtual void Reset()
	{
		MeshComponentSource::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("DomainContainer"); }
	virtual int IsSource() { return 1; }

public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc)
	{
		MeshComponentSource::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshComponentSource::SetElementData(edc);
		
		ElementData ed;
//		ed.SetInt(GetGeometricNonlinearityStatus(1),"GeometricNonlinearityStatus"); ed.SetToolTipText("-1..use element defaults, 0..GNS_Linear, 1..GNS_NonlinearSmallStrain, 2..GNS_NonlinearLargeStrain"); edc.TreeAdd("Generation",ed);

		return rv;
	}
	virtual int ComputeBoundingBox()
	{
		boundingbox.Clear();
		for(int i=1; i<=NNodes(); i++)
		{
			Vector3D pos = GetNodePosition(i);
			boundingbox.Add(pos);
		}
		return NNodes();
	}
	virtual Box3D GetBoundingBox() { return boundingbox; }

protected: // own variables
};
// helper class for loading surfaces, boundaries, FaceSets
// feasable for 2D and 1D boundaries of individual elmenets
class FaceInfo
{
public:
//	int entryNr;                 // index of the Facet/Edge -- as loaded from file
	int bcTypeNr;                // Type(number) of the boundary condition this Facet/Edge has -- as loaded from file
	int domin,domout;            // inner and outer domain of the #facet/Edge -- as loaded from file
	TArray<int> globalNodeNrs;   // global node numbers of the Facet/Edge -- as loaded from file
	
	int elNr;                    // element number of the nesting element -- identified from meshelements
	int faceNr;                  // face number on the element -- identified from meshelements 
	TArray<int> localNodeNrs;    // local node numbers on the element (redundant with faceNr) -- identified from meshelements
	double area;                 // facet area// edge length -- computed from meshelements
};
// helper class for mapping boundary conditions
class BoundaryMapping
{
public: 
	int bcTypeNr;
	int domin,domout;
	int setNrin,setNrout;
};



// common base class for all classes that load nodes and elements from an external file
// format specific load routines in derived classes
class MeshComponentLoader : public MeshMultiProcessor
{
public: // lifecycle I
	MeshComponentLoader() {	Reset(); }
	//MeshComponentLoader(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshComponentLoader(const MeshComponentLoader& other) { CopyFrom(other); }
	~MeshComponentLoader() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshComponentLoader(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshComponentLoader& other);

	virtual void Reset();

public: // class specific
	mystr GetComponentSpecification() { return mystr("Loader"); }
	virtual int IsLoader() { return 1; }
public: // EDC-interaction
	virtual void GetElementData(ElementDataContainer& edc);

	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = MeshMultiProcessor::SetElementData(edc);

		FileName() = edc.TreeGetString("Generation.source_file");
		return rv;
	}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // access
	virtual const mystr GetFileName() const { return filename; }
	virtual void SetFileName(mystr filenamei) { filename = filenamei; }
	virtual mystr& FileName() { return filename; }
private: // additional functions - load
	virtual int LoadFromFile() { return 0; } 
private: // additional functions - format specific - NOTE: move to derived classes ?
	virtual int InitializeLookupTables() { return 0; }
	virtual int GetNodeTypeLookup(int i=0) { return 0; }
	virtual int GetElemTypeLookup(int nnodes, int i=0) { return 0; }
	virtual int LoadFromFileToBuffers() { return 0; }
	virtual int MatchSurfaceElements() { return 0; }           // TODO: this should used for all formats, get rid of bc_s array...
protected: // additional functions - common
	virtual int RemoveUnusedNodes();
	virtual int GenerateDomainContainers();
	virtual int SplitIntoDomains();
	virtual int SplitBoundaries();
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// overrides
public:  // overrideables - meshing
	virtual int MeshYourself(); // in cpp - Loader classes load from file
	virtual int GenerateOwnNodes();
	virtual int GenerateOwnElements();
	virtual int GenerateOwnSets();

	virtual int AddBuffer(MeshElement& elem) { return elementbuffer.Add(elem.GetCopy()); }
public: // overrideables - picking
//	virtual int ComputeBoundingBox(); // in cpp - loop over all loaded nodes

public: // functionality - let a Glue component access the coincidentbuffer
	virtual int NCoincidentBuffer() { return coincidentbuffer.Length(); }
	virtual int2 GetNodeNumbersFromCoincidentBuffer(int i)
	{
		int4 entry = coincidentbuffer(i);
		int nnr1 = IthNode(entry(1),entry(2));
		int nnr2 = IthNode(entry(3),entry(4));

		return int2(nnr1,nnr2); // returns the i-indexed nodes of the coincident pair
	}

protected: // own variables
	mystr filename;                         // file to load data from
	int usedtypelist;                       // define the lookup table to use ( chose among several possible default sets )
	
	TArray<int> nodetypelist;				// lookup table: list of object factory nodetypes - must be initialized with ::InitializeLookupTable() in derived class
	TArray<int> elementtypelist;            // lookup table: list of object factory elementtypes - must be initialized with ::InitializeLookupTable() in derived class
	
// import buffers for nodes and elements
	int2 pnesource;                         // points and elements actually provided by source file
	int npoints,nelements;                  // points and elements actually used 
	TArray<Vector3D> pointbuffer;
	TArray<MeshElement*> elementbuffer;     // MeshElements must keep their Type (MEHex,...)
	TArray<int> useddomainsbuffer;

// import buffers for boundary conditions
	int nborder;
// currently 3D, later all
	TArray<int2> identifications;         // node pairs from periodic boundary conditions
	TArray<FaceInfo> surfacesbuffer;      // holds all surface elements (Facet/Edge) for boundary conditions

// currently 2D, later remove
	TArrayDynamic<TArray<int>> bc_s; // 2D boundary_conditions
	TArray<int2> idents; // identic node numbers (periodic boudary condition)


// mapping buffers
// buffers for domain splitting
	TArray<int> nodedomainmultiplicity;
	TArrayDynamic<TArray<int>> nodenumberindomain;
	TArray<int> nodedomain;
	TArray<int4> coincidentbuffer;

// used to split and store the individual face/edge lists for the BCs
	std::map< int, std::map< int, std::vector<FaceInfo> > > collectedBCs;
	std::map < int, BoundaryMapping > bcmapping;

};


class MeshLoaderSTL : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderSTL() { Reset(); }
	//MeshLoaderSTL(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderSTL(const MeshLoaderSTL& other) { CopyFrom(other); }
	~MeshLoaderSTL() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderSTL(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderSTL& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderSTL& o = (const MeshLoaderSTL&) other;
	// specific for derived
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("LoaderSTL"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadFromFile();
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int i=0);
	virtual int GetElemTypeLookup(int nnodes, int i=0);
	virtual int LoadFromFileToBuffers();
};


class MeshLoaderNetgen2D : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderNetgen2D() { Reset(); }
	//MeshLoaderNetgen2D(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderNetgen2D(const MeshLoaderNetgen2D& other) { CopyFrom(other); }
	~MeshLoaderNetgen2D() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderNetgen2D(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderNetgen2D& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderNetgen2D& o = (const MeshLoaderNetgen2D&) other;
	// specific for derived	
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("LoaderNetgen2D"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadFromFile();
private: // additional functions - format specific
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int i=0);
	virtual int GetElemTypeLookup(int nnodes, int i=0);
	virtual int LoadFromFileToBuffers();
	virtual int LoadFromFileToBuffers_old();
};

class MeshLoaderNetgen3D : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderNetgen3D() { Reset(); }
	//MeshLoaderNetgen3D(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderNetgen3D(const MeshLoaderNetgen3D& other) { CopyFrom(other); }
	~MeshLoaderNetgen3D() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderNetgen3D(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderNetgen3D& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderNetgen3D& o = (const MeshLoaderNetgen3D&) other;
	// specific for derived	
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("LoaderNetgen3D"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadFromFile();
private: // additional functions - format specific
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int i=0);
	virtual int GetElemTypeLookup(int nnodes, int i=0);
	virtual int LoadFromFileToBuffers();
};

class MeshLoaderNeutral3D : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderNeutral3D() { Reset(); }
	//MeshLoaderNeutral3D(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderNeutral3D(const MeshLoaderNeutral3D& other) { CopyFrom(other); }
	~MeshLoaderNeutral3D() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderNeutral3D(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderNeutral3D& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderNeutral3D& o = (const MeshLoaderNeutral3D&) other;
	// specific for derived	
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("MeshLoaderNeutral3D"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadFromFile();
private: // additional functions - format specific
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int i=0);
	virtual int GetElemTypeLookup(int nnodes,int i=0);
	virtual int LoadFromFileToBuffers();
	virtual int PermuteNodeList(TArray<int>& nodelist);
};

class MeshLoaderVOL : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderVOL() { Reset(); }
	//MeshLoaderNetgen3D(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderVOL(const MeshLoaderVOL& other) { CopyFrom(other); }
	~MeshLoaderVOL() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderVOL(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderVOL& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderVOL& o = (const MeshLoaderVOL&) other;
		symmetricNodes.CopyFrom(other.symmetricNodes);
		dim = o.dim;
	// specific for derived	
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("LoaderVOL"); }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadFromFile();
	void UpdateSymmetryNodes();
	virtual Vector3D GetColor(int i=0);
	virtual void GetElementData(ElementDataContainer& edc);

protected:
	virtual TArray<int2> GetPeriodicBCNodePairs() const;
	virtual TArray<int> GetSymmetricNodes() const;
private: // additional functions - format specific
	virtual int InitializeLookupTables();
	virtual int GetNodeTypeLookup(int i=0);
	virtual int GetElemTypeLookup(int nnodes, int i=0);
	virtual int LoadFromFileToBuffers();
	virtual int MatchSurfaceElements();
	TArray<int> symmetricNodes;
	int dim;
};
 
class MeshLoaderDataArrays : public MeshComponentLoader
{
public: // lifecycle I
	MeshLoaderDataArrays() { Reset(); }
	//MeshLoaderDataArrays(BaseMesh* meshi) { mesh = meshi; Reset(); }
	MeshLoaderDataArrays(const MeshLoaderDataArrays& other) { CopyFrom(other); }
	~MeshLoaderDataArrays() {}

public: // lifecycle II
	virtual MeshComponent* GetCopy()
	{
		MeshComponent* mcc = new MeshLoaderDataArrays(*this);
		return mcc;
	}
	virtual void CopyFrom(const MeshLoaderDataArrays& other)
	{
		MeshComponentLoader::CopyFrom(other);

		const MeshLoaderDataArrays& o = (const MeshLoaderDataArrays&) other;
	// specific for derived
	}
	virtual void Reset()
	{
		MeshComponentLoader::Reset();
		
		useddomainsbuffer.Add(1);
		GenerateDomainContainers();
	}
public: // class specific
	mystr GetComponentSpecification() { return mystr("LoaderDataArrays"); }



	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public: // additional functions - load
	virtual int LoadNodesFromArrays(int npoints, int dofsperpoint, double* pointdofs);
	virtual int LoadElementsFromArrays(int nelems, int nodesperelement, int* elementnodelists);

	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// previously inherited from Source
public: // node and element types - To keep compatible with existing models
	virtual void SetNodeType(mystr type) { nodetype = type; }
	virtual mystr& NodeType() { return nodetype; }
	virtual void SetElemType(mystr type) { elemtype = type; }
	virtual mystr& ElemType() { return elemtype; }

	virtual mystr GetNodeType(int = 0)
	{
		return nodetype;
	}
	virtual mystr GetElementType(int = 0)
	{
		return elemtype;
	}
protected:
	mystr nodetype;
	mystr elemtype;

};


#endif

