//#***************************************************************************************
//#
//# filename:             set.h
//#
//# author:               Dorninger Alexander
//#
//# generated:						July 2014
//# description:          introduces sets to the MBS
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************

#pragma once

#include "mbs_interface.h"
#include "node.h"
#include "..\UtilityLib\elementdataaccess.h"
#include "node.h"

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                                BASESET
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

typedef enum {
	TMBSSetUnknown = 0,
	
	TMBSSetElements = 1,
	
	TMBSSetPoints = 11,
	TMBSSetGlobalNodes = 12,
	TMBSSetLocalNodesA = 13,
	TMBSSetLocalNodesB = 14,
	TMBSSetGlobalCoords = 15,
	TMBSSetLocalCoordsA = 16, // this is the format required by e.g. constraints
	TMBSSetLocalCoordsB = 17,

	TMBSSetFaces = 21,

	TMBSSetSensors = 51

} TMBSSetType;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                                 BASESET
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// base class for all sets, containting information to identify by name and type
//$EDC$[beginclass,classname=BaseSet]
class BaseSet 
{
public: // lifecycle I
	BaseSet(): mbs(NULL),name(""),type(TMBSSetUnknown) {;}
	BaseSet(MBS* mbsi): mbs(mbsi),name(""),type(TMBSSetUnknown) {;}
	BaseSet(const BaseSet& other) { CopyFrom(other); }
	~BaseSet() {}

public: // lifecycle II
	BaseSet(MBS* mbsi,mystr namei,TMBSSetType typei): mbs(mbsi),name(namei),type(typei) {;}
	virtual BaseSet* GetCopy()
	{
		BaseSet* sc = new BaseSet();
		sc->CopyFrom(*this);
		return sc;
	}
	virtual void CopyFrom(const BaseSet& other)
	{
		mbs = other.mbs;
		name = other.name;
		type = other.type;
	}

public: // access
	virtual const MBS* GetMBS() const { return mbs; }
	virtual MBS* GetMBS() { return mbs; }
	virtual void SetMBS(MBS* mbsi) { mbs = mbsi; }

	virtual const TMBSSetType GetType() const { return type; }
	virtual void SetType(TMBSSetType typei) { type = typei; }
	virtual TMBSSetType& Type() { return type; }

	virtual const mystr GetName() const { return name; }
	virtual void SetName(mystr namei) { name = namei; }
	virtual mystr& Name() { return name; }


public: // operations and operators
	// all AddEntry(...)
	virtual void AddEntry(int nr) {;}				// ElementSet::AddEntry(elementNumber), SetOfNodes::AddEntry(nodeNumber)
	virtual void AddEntry(Vector3D pos) {;}	// SetOfPoints::AddEntry(globalPosition)
	
	// force implementation for the following funcitons by abstract function declaration ? 
	// conflicts with all the access functions...
	virtual void DeleteEntry(int i) {;} //= 0;
	virtual void Reset() {;} //= 0;
	virtual int GetNrOfEntries() {return 0;} //= 0;
	virtual int Check() {return 0;} //= 0;
	virtual int Convert () {return 0;} //= 0;
	virtual mystr ToString() {return mystr("");} //= 0;

	bool operator== (const BaseSet& other) // equality operator for TArray<T>::Find(const T& t)
	{
		return name.Compare(other.name) == 1;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(name.c_str(),"set_name"); ed.SetToolTipText("the name of the set"); edc.Add(ed); // name of the set, may be changed
		ed.SetText(GetSetSpecification().c_str(),"set_type"); ed.SetToolTipText("type of the set");  /*ed.SetLocked(1);*/ edc.Add(ed); // type of the set 
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		name = edc.TreeGetString("set_name");

		mystr type_old = GetSetSpecification();
		mystr type_new = edc.TreeGetString("set_type");
		if(!type_new.Compare(type_old))
		{
			GetMBS()->UO().InstantMessageText("ERROR: You MUST NOT change the type of the set!");
			return 0;
		}

		return rv;
	}
	virtual int ReadSingleElementData(ReadWriteElementDataVariableType& RWdata) {return 0;}
	virtual int WriteSingleElementData(const ReadWriteElementDataVariableType& RWdata) {return 0;}
	virtual mystr GetSetSpecification() { return mystr("abstract"); } //*EDC*[funcaccess,EDCvarname"set_type",tooltiptext="specification of set type. Once the set is added to the MBS, you MUST NOT change this type anymore"]

protected:
	MBS* mbs;
	mystr name;           //*EDC*[varaccess,EDCvarname="name",EDCfolder="",tooltiptext="name of the set"]
	TMBSSetType type;
}; //$EDC$[endclass,BaseSet]


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                              ELEMENTSET
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// set containting elements
//$EDC$[beginclass,classname=ElementSet,parentclassname=BaseSet,
//texdescription="A set containing only elements (as element numbers).",example="sets.txt"]
class ElementSet : public BaseSet 
{
public: // lifecycle I
	ElementSet(MBS* mbsi): BaseSet(mbsi,"ElementSet",TMBSSetElements), elementnumbers(0) {;}
	ElementSet(const ElementSet& other) { CopyFrom(other); }
	~ElementSet() {}

public: // lifecycle II
	ElementSet(MBS* mbsi, mystr namei): BaseSet(mbsi,namei,TMBSSetElements), elementnumbers(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new ElementSet(*this);
		return sc;
	}
	void CopyFrom(const ElementSet& other) 
	{
		// copy base class content
		BaseSet::CopyFrom(other);
		// specific for each derived class
		const ElementSet& o = (const ElementSet&)other;
		elementnumbers.CopyFrom(o.elementnumbers);
	}

public: // access
	virtual int GetElementNumber(int i) { return elementnumbers(i); }
	virtual void SetElementNumber(int i, int elnr) { elementnumbers(i) = elnr; }
	virtual int& ElementNumber(int i) { return elementnumbers(i); }
	virtual TArray<int>& ElementNumbers() { return elementnumbers; }

public: // operations and operators
	virtual void AddEntry(int elnr) { elementnumbers.Add(elnr); }
	virtual void DeleteEntry(int i) { elementnumbers.Erase(i); }
	virtual void Reset() { elementnumbers.Flush(); }
	virtual int GetNrOfEntries() { return elementnumbers.Length(); }

	virtual int Check() // all entries of elementnumbers must exist in mbs
	{
		int nerrors = 0;
		int nelements = GetMBS()->NE();
		for(int i=1; i<=elementnumbers.Length(); i++)
		{
			if(elementnumbers(i)>nelements || elementnumbers(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in ElementSet " << GetName() << " - for " << nerrors << " entries the element does not exist\n";
		}
		return nerrors;
	}

	virtual mystr ToString()
	{
		mystr rv("The ElementSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += mystr(GetElementNumber(i)); 
			rv += ", ";
		}
		rv += mystr(GetElementNumber(GetNrOfEntries()));
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseSet::GetElementData(edc);

		ElementData ed;
		Vector vv(elementnumbers.Length()); 
		for(int i=1; i<=elementnumbers.Length(); i++) 
		{ 
			vv(i) = elementnumbers(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "element_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Elements in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		BaseSet::SetElementData(edc);

		GetElemDataIVector(GetMBS(), edc, "element_numbers", elementnumbers, 1);
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("ElementSet"); }
protected:
	TArray<int> elementnumbers; //*EDC*[varaccess,EDCvarname="elements",EDCfolder="",variable_length_vector,tooltiptext="members of the element set"]
}; //$EDC$[endclass,ElementSet]


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                               POINTSET (BaseClass)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// base class for all sets on points - provides arrays for element number and local position that are required for further operations on point sets
//$EDC$[beginclass,classname=PointSetBC,parentclass=BaseSet]
class PointSetBC : public BaseSet 
{
public:
	PointSetBC():BaseSet() {;}
	PointSetBC(MBS* mbsi): BaseSet(mbsi,"PointSet",TMBSSetPoints), good2use(false), elementnumbers(0), positions_local(0) {;}
	PointSetBC(const PointSetBC& other) { CopyFrom(other); }
	virtual ~PointSetBC()
	{
		for (int i = 1; i <= local_dofs.Length(); i++)
		{
			delete local_dofs.Get(i);
		}
		local_dofs.Flush();
	}

public: // lifecycle II
	PointSetBC(MBS* mbsi, mystr namei,TMBSSetType typei): BaseSet(mbsi,namei,typei), good2use(false), elementnumbers(0), positions_local(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new PointSetBC(*this);
		return sc;
	}
	void CopyFrom(const PointSetBC& other) 
	{
		// copy base class content
		BaseSet::CopyFrom(other);
		// specific for each derived class
		const PointSetBC& o = (const PointSetBC&)other;
		good2use = o.good2use;
		elementnumbers.CopyFrom(o.elementnumbers);
		positions_local.CopyFrom(o.positions_local);
	}

public: // access
	virtual int GetElementNumber(int i) { return elementnumbers(i); }
	virtual void SetElementNumber(int i, int elnr) { elementnumbers(i) = elnr; }
	virtual int& ElementNumber(int i) { return elementnumbers(i); }
	virtual TArray<int>& ElementNumbers() { return elementnumbers; }
	virtual Vector3D GetLocalPosition(int i) { return positions_local(i); }
	virtual void SetLocalPosition(int i, Vector3D locpos) { positions_local(i) = locpos; }
	virtual const IVector & GetLocalDOFs(int i) { return *local_dofs(i); }
	virtual void SetLocalDOFs(int i, const IVector& locdofs) { local_dofs(i) = new IVector(locdofs); }
	virtual Vector3D& LocalPosition(int i) { return positions_local(i); }
	virtual int GetLocalNodeNumber(int i) { return 0; }
	virtual void SetLocalNodeNumber(int i, int nnr) { return; }
	virtual bool IsGood2Use() { return good2use; }

public: // operations and operators
	virtual void AddEntry(int elnr, Vector3D locpos) { elementnumbers.Add(elnr); positions_local.Add(locpos); }
	virtual void DeleteEntry(int i) { elementnumbers.Erase(i); positions_local.Erase(i); }
	virtual void Reset() { elementnumbers.Flush(); positions_local.Flush(); }
	virtual int GetNrOfEntries() { return elementnumbers.Length(); }

	virtual int Check() { return 0;} // no consistency checks in base class
	virtual int Convert() { return 0;} // no conversion operations in base class 
	virtual mystr ToString() { return "";} // no string for base class 

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseSet::GetElementData(edc);

		ElementData ed;
		Vector vv(elementnumbers.Length()); 
		for(int i=1; i<=elementnumbers.Length(); i++) 
		{ 
			vv(i) = elementnumbers(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "element_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Elements in this set");
		edc.Add(ed);
		
		Matrix mm(positions_local.Length(),3);
		for(int i=1; i<=positions_local.Length(); i++) 
		{
			mm(i,1) = positions_local(i).X();
			mm(i,2) = positions_local(i).Y();
			mm(i,3) = positions_local(i).Z();
		}
		ed.SetMatrix(mm.GetMatPtr(), mm.Getrows(), mm.Getcols(), "local_positions");
		ed.SetVariableLength(); ed.SetToolTipText("Local positions on the elements in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		BaseSet::SetElementData(edc);

		GetElemDataIVector(GetMBS(), edc, "element_numbers", elementnumbers, 1);
		Matrix lps;
		GetElemDataMatrix(GetMBS(), edc, "local_positions", lps, 1);
		for(int i=1; i<=lps.Getrows(); i++)
		{
			Vector3D lp(lps(i,1),lps(i,2),lps(i,3));
			SetLocalPosition(i,lp);
		}
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("abstract"); } 

protected:
	bool good2use; // flag to determine if all the current entries are converted for further use
	TArray<int> elementnumbers; //*EDC*[varaccess,EDCvarname="elements",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
	TArray<Vector3D> positions_local; //*EDC*[varaccess,EDCvarname="ploc",EDCfolder="",variable_length_vector,tooltiptext="local positions on the elementss of the set"]
	TArray<IVector*> local_dofs; //*EDC*[varaccess,EDCvarname="locdofs",EDCfolder="",variable_length_vector,tooltiptext="map from local node dofs to local element dofs"]
}; //$EDC$[endclass,PointSetBC]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                          GLOBALNODESET (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (node-based) point set that is defined by global node numbers
//$EDC$[beginclass,classname=GlobalNodeSet,parentclass=PointSetBC, 
//texdescription="A (node-based) point set that is defined by global node numbers."]
class GlobalNodeSet : public PointSetBC 
{ 
public: // lifecycle I
	GlobalNodeSet(MBS* mbsi): PointSetBC(mbsi,"GlobalNodesSet",TMBSSetGlobalNodes), nodenumbers_global(0), nodenumbers_local(0) {;}
	GlobalNodeSet(const GlobalNodeSet& other) { CopyFrom(other); }
	~GlobalNodeSet() {}

public: // lifecycle II
	GlobalNodeSet(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetGlobalNodes), nodenumbers_global(0), nodenumbers_local(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new GlobalNodeSet(*this);
		return sc;
	}
	void CopyFrom(const GlobalNodeSet& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const GlobalNodeSet& o = (const GlobalNodeSet&)other;
		nodenumbers_global.CopyFrom(o.nodenumbers_global);
		nodenumbers_local.CopyFrom(o.nodenumbers_local);
	}

public: // access
	virtual int GetGlobalNodeNumber(int i) { return nodenumbers_global(i); }
	virtual void SetGlobalNodeNumber(int i, int nnr) { nodenumbers_global(i) = nnr; }
	virtual int& GlobalNodeNumber(int i) { return nodenumbers_global(i); }
	virtual TArray<int>& GlobalNodeNumbers() { return nodenumbers_global; }
	virtual int GetLocalNodeNumber(int i) { return nodenumbers_local(i); }
	virtual void SetLocalNodeNumber(int i, int nnr) { nodenumbers_local(i) = nnr; }
	
public: // operations and operators
	virtual void AddEntry(int nnr) 
	{ 
		nodenumbers_global.Add(nnr);
		good2use = false;
	}
	virtual void DeleteEntry(int i) 
	{ 
		if(IsGood2Use())
		{
			elementnumbers.Erase(i);
			positions_local.Erase(i);
			nodenumbers_local.Erase(i);
		}
		nodenumbers_global.Erase(i);
	}
	virtual void Reset() { elementnumbers.Flush(); positions_local.Flush(); nodenumbers_local.Flush(); nodenumbers_global.Flush(); }
	virtual int GetNrOfEntries() { return nodenumbers_global.Length(); }

	virtual int Check() // all entries of nodenumbers must exist in mbs
	{
		int nerrors = 0;
		int nnodes = GetMBS()->NNodes();
		for(int i=1; i<=nodenumbers_global.Length(); i++)
		{
			if(nodenumbers_global(i)>nnodes || nodenumbers_global(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in GlobalNodeSet " << GetName() << " - for " << nerrors << " entries the node does not exist\n";
		}
		return nerrors;
	}
	virtual int Convert() // compute the according element numbers and local positions from global node numbers
	{
		//ASSUME
		// this is only done in the MBS environment !
		mbs->ComputeNodesToElementsList(/*MBS*/);

		for(int i=1; i<=nodenumbers_global.Length(); i++)
		{
		// local node number in 1st element
			int gnnr = nodenumbers_global(i);
			int elnr = GetMBS()->GetElementNumberOfGlobalNode(gnnr);
			int lnnr = GetMBS()->GetLocalNodeNumberOfGlobalNode(elnr,gnnr);
			const Node * node = GetMBS()->GetNodePtr(gnnr);
			const Element *elem = GetMBS()->GetElementPtr(elnr);
			IVector DOFnrs(node->LTGLength());
			DOFnrs.SetAll(0);

			for (int j=1; j<=node->LTGLength(); j++)
			{
				for (int k=1; k<=elem->LTGlength(); k++)
				{
					if (node->LTG(j) == elem->LTG(k))
					{
						DOFnrs(j) = k;
					}
				}
			}

		// local pos from local node number
			Vector3D lpos = GetMBS()->GetLocalPositionOfLocalNode(elnr,lnnr);
		// write converted data into the according fields
			SetLocalDOFs(i, DOFnrs);
			SetElementNumber(i,elnr);
			SetLocalPosition(i,lpos);
			SetLocalNodeNumber(i,lnnr);
		}

		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The GlobalNodeSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += mystr(GetGlobalNodeNumber(i)); 
			rv += ", ";
		}
		rv += mystr(GetGlobalNodeNumber(GetNrOfEntries()));
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		PointSetBC::GetElementData(edc);
		ElementData* edp;
		edp = edc.GetPtr(edc.Find("element_numbers")); edp->SetLocked(1);
		edp = edc.GetPtr(edc.Find("local_positions")); edp->SetLocked(1);
		
		ElementData ed;
		Vector vv(nodenumbers_global.Length()); 
		for(int i=1; i<=nodenumbers_global.Length(); i++) 
		{ 
			vv(i) = nodenumbers_global(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "global_node_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Global Nodes in this set");
		edc.Add(ed);

		vv.SetLen(nodenumbers_local.Length());
		for(int i=1; i<=nodenumbers_local.Length(); i++)
		{
			vv(i) = nodenumbers_local(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "local_node_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Local node numbers on the elements in this set");
		ed.SetLocked(1);
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		PointSetBC::SetElementData(edc);
		GetElemDataIVector(GetMBS(), edc, "global_node_numbers", nodenumbers_global, 1);
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("GlobalNodeSet"); } 
	
protected:
	TArray<int> nodenumbers_global; //*EDC*[varaccess,EDCvarname="nodes_global",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
	TArray<int> nodenumbers_local;  
}; //$EDC$[endclass,GlobalNodeSet]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                           LOCALNODESETA (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (node-based) point set that is defined by element number and local node numbers - holds two lists of equal length forming pairs of (elnr,locnnr)
//$EDC$[beginclass,classname=LocalNodeSetA,parentclass=PointSetBC,
//texdescription="A (node-based) point set that is defined by element number and local node numbers - holds two lists of equal length forming pairs of (elnr,locnnr)."]
class LocalNodeSetA : public PointSetBC 
{ 
public: // lifecycle I
	LocalNodeSetA(MBS* mbsi): PointSetBC(mbsi,"LocalNodeSetA",TMBSSetLocalNodesA), nodenumbers_local(0) {;}
	LocalNodeSetA(const LocalNodeSetA& other) { CopyFrom(other); }
	~LocalNodeSetA() {}

public: // lifecycle II
	LocalNodeSetA(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetLocalNodesA), nodenumbers_local(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new LocalNodeSetA(*this);
		return sc;
	}
	void CopyFrom(const LocalNodeSetA& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const LocalNodeSetA& o = (const LocalNodeSetA&)other;
		nodenumbers_local.CopyFrom(o.nodenumbers_local);
	}

public: // access
	virtual int GetLocalNodeNumber(int i) { return nodenumbers_local(i); }
	virtual void SetLocalNodeNumber(int i, int nnr) { nodenumbers_local(i) = nnr; }
	virtual int& LocalNodeNumber(int i) { return nodenumbers_local(i); }

public: // operations and operators
	virtual void AddEntry(int elnr, int nnr) 
	{ 
		elementnumbers.Add(elnr);
		nodenumbers_local.Add(nnr);
		good2use = false;
	}
	virtual void DeleteEntry(int i) 
	{ 
		if(IsGood2Use())
		{
			positions_local.Erase(i);
		}
		elementnumbers.Erase(i);
		nodenumbers_local.Erase(i);
	}
	virtual void Reset() { elementnumbers.Flush(); positions_local.Flush(); nodenumbers_local.Flush(); }
	
	virtual int Check() // all entries of localnodenumbers must exist in the elements mbs
	{
		int nerrors = 0;
		if(nodenumbers_local.Length() != elementnumbers.Length())
		{
			GetMBS()->UO() << "Found inconsistencies in LocalNodeSet " << GetName() << " - list of local node numbers and list of element numbers are not the same length\n";			
		}

		for(int i=1; i<=nodenumbers_local.Length(); i++)
		{
			Element* el = GetMBS()->GetElementPtr(elementnumbers(i));
			if(nodenumbers_local(i)>el->NNodes() || nodenumbers_local(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in LocalNodeSet " << GetName() << " - for " << nerrors << " entries - the node does not exist in the element\n";
		}
		return nerrors;
	}
	virtual int Convert() // compute the according element numbers and local positions from global node numbers
	{
		//ASSUME
		// this is only done in the MBS environment !

		for(int i=1; i<=nodenumbers_local.Length(); i++)
		{
			int elnr = elementnumbers(i);
			int lnnr = nodenumbers_local(i);
		// local pos from local node number
			Vector3D lpos = GetMBS()->GetLocalPositionOfLocalNode(elnr,lnnr);
			SetLocalPosition(i,lpos);
		}
		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The LocalNodeSet ");
		rv += name;
		rv += " contains pairs: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += mystr(GetElementNumber(i));
			rv += ".";
			rv += mystr(GetLocalNodeNumber(i)); 
			rv += ", ";
		}
		rv += mystr(GetElementNumber(GetNrOfEntries()));
		rv += ".";
		rv += mystr(GetLocalNodeNumber(GetNrOfEntries()));
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		PointSetBC::GetElementData(edc);
		ElementData* edp;
		edp = edc.GetPtr(edc.Find("local_positions")); edp->SetLocked(1);
		
		ElementData ed;
		Vector vv(nodenumbers_local.Length()); 
		for(int i=1; i<=nodenumbers_local.Length(); i++) 
		{ 
			vv(i) = nodenumbers_local(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "local_node_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Local Nodes in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		PointSetBC::SetElementData(edc);
		GetElemDataIVector(GetMBS(), edc, "local_node_numbers", nodenumbers_local, 1);
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("LocalNodeSetA"); } 

protected:
	TArray<int> nodenumbers_local; //*EDC*[varaccess,EDCvarname="nodes_local",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
}; //$EDC$[endclass,LocalNodeSetA]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                           LOCALNODESETB (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (node-based) point set that is defined by global node numbers - holds two lists of different length for elements and local node numbers - the entire list of (elnr/locnode) pairs contains every combination
//$EDC$[beginclass,classname=LocalNodeSetB,parentclass=PointSetBC,
//texdescription="A (node-based) point set that is defined by global node numbers - holds two lists of different length for elements and local node numbers - the entire list of (elnr,locnode) pairs contains every combination."]
class LocalNodeSetB : public PointSetBC 
{ 
public: // lifecycle I
	LocalNodeSetB(MBS* mbsi): PointSetBC(mbsi,"LocalNodeSetB",TMBSSetLocalNodesB), elementnumbers_shortlist(0), nodenumbers_local_shortlist(0) {;}
	LocalNodeSetB(const LocalNodeSetB& other) { CopyFrom(other); }
	~LocalNodeSetB() {}

public: // lifecycle II
	LocalNodeSetB(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetLocalNodesB), elementnumbers_shortlist(0), nodenumbers_local_shortlist(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new LocalNodeSetB(*this);
		return sc;
	}
	void CopyFrom(const LocalNodeSetB& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const LocalNodeSetB& o = (const LocalNodeSetB&)other;
		elementnumbers_shortlist.CopyFrom(o.elementnumbers_shortlist);
		nodenumbers_local_shortlist.CopyFrom(o.nodenumbers_local_shortlist);
	}

public: // access
	virtual int GetElementNumberSL(int i) { return elementnumbers_shortlist(i); }              // ElementNumber SHORTLIST - GetElementNumber MUST always give back the element from the full list (after conversion)
	virtual void SetElementNumberSL(int i, int elnr) { elementnumbers_shortlist(i) = elnr; }
	virtual int& ElementNumberSL(int i) { return elementnumbers_shortlist(i); }
	virtual int GetLocalNodeNumberSL(int i) { return nodenumbers_local_shortlist(i); }
	virtual void SetLocalNodeNumberSL(int i, int nnr) { nodenumbers_local_shortlist(i) = nnr; }
	virtual int& LocalNodeNumberSL(int i) { return nodenumbers_local_shortlist(i); }

public: // operations and operators
	virtual void AddEntryElem(int elnr)
	{ 
		elementnumbers_shortlist.Add(elnr);
		good2use = false;
	}
	virtual void AddEntryNode(int nnr)
	{ 
		nodenumbers_local_shortlist.Add(nnr);
		good2use = false;
	}
	virtual void DeleteEntryElem(int i) 
	{ 
		if(IsGood2Use())
		{
		// ASSUME - fixed order of entries in the (element,localposition) list
		// erase entries (#nodes*i)+1 through (#nodes*i)+#nodes		
			int nnodes = nodenumbers_local_shortlist.Length();
			for(int j=nnodes; j>1; j--)
			{
				int idx = nnodes*i+j;
				elementnumbers.Erase(idx);
				positions_local.Erase(idx);
			}
		}
		elementnumbers_shortlist.Erase(i);
	}
	virtual void DeleteEntryNode(int i) 
	{ 
		if(IsGood2Use())
		{
		// ASSUME - fixed order of entries in the (element,localposition) list
		// erase entries i, i+k*#nodes 
			int nnodes = nodenumbers_local_shortlist.Length();
			int nelems = elementnumbers_shortlist.Length();
			for(int j=nelems; j>1; j--)
			{
				int idx = nnodes*(j-1)+i;
				elementnumbers.Erase(idx);
				positions_local.Erase(idx);
			}
		}
		nodenumbers_local_shortlist.Erase(i);
	}
	virtual void Reset() { elementnumbers.Flush(); elementnumbers_shortlist.Flush(); positions_local.Flush(); nodenumbers_local_shortlist.Flush(); }

	virtual int Check() // all entries of localnodenumbers must exist in the elements mbs
	{
		int nerrors = 0;
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++)
		{
			//$EK 2020-02-25: bug fix use correct element_number list (elementnumbers_shortlist)
			Element* el = GetMBS()->GetElementPtr(elementnumbers_shortlist(i));
			for(int j=1; j<=nodenumbers_local_shortlist.Length(); j++)
			{
				if(nodenumbers_local_shortlist(j)>el->NNodes() || nodenumbers_local_shortlist(j)<=0)
				{
					nerrors++;
				}			
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in LocalNodeSet " << GetName() << " - for " << nerrors << " entries the node does not exist in the element\n";
		}
		return nerrors;
	}
	virtual int Convert() // compute the according element numbers and local positions from global node numbers
	{
		//ASSUME
		// this is only done in the MBS environment !
		
		// mind the loops to construct the set product in the following way:
		// { (el1,n1), (el1,n2), (el1,n3), (el2,n1), (el2,n2), ... }
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++)
		{
			for(int j=1; j<=nodenumbers_local_shortlist.Length(); j++)
			{
				int elnr = elementnumbers_shortlist(i);
				int lnnr = nodenumbers_local_shortlist(j);
				int idx = (i-1)*nodenumbers_local_shortlist.Length()+j;
		// local pos from local node number
				Vector3D lpos = GetMBS()->GetLocalPositionOfLocalNode(elnr,lnnr);
				SetElementNumber(idx,elnr);
				SetLocalPosition(idx,lpos);		
			}
		}
		// local pos from local node number
		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The LocalNodeSet ");
		rv += name;
		rv += " contains all combinations of the following ELEMENTS: {";
		for(int i=1; i<elementnumbers_shortlist.Length(); i++)
		{
			rv += mystr(GetElementNumberSL(i));
			rv += ", ";
		}
		rv += mystr(GetElementNumberSL(elementnumbers_shortlist.Length()));
		rv += "}\n";

		rv += " and LOCALNODENUMBERS: {";
		for(int i=1; i<nodenumbers_local_shortlist.Length(); i++)
		{
			rv += mystr(GetLocalNodeNumberSL(i));
			rv += ", ";
		}
		rv += mystr(GetLocalNodeNumberSL(nodenumbers_local_shortlist.Length()));
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		PointSetBC::GetElementData(edc);
		ElementData* edp;
		edp = edc.GetPtr(edc.Find("element_numbers")); edp->SetLocked(1);
		edp = edc.GetPtr(edc.Find("local_positions")); edp->SetLocked(1);
		
		ElementData ed;
		Vector vv(elementnumbers_shortlist.Length()); 
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++) 
		{ 
			vv(i) = elementnumbers_shortlist(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "element_numbers_shortlist");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Shortlist of Elements in this set");
		edc.Add(ed);

		vv.SetLen(nodenumbers_local_shortlist.Length()); 
		for(int i=1; i<=nodenumbers_local_shortlist.Length(); i++) 
		{ 
			vv(i) = nodenumbers_local_shortlist(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "local_node_numbers_shortlist");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Shortlist of Local Nodes in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		PointSetBC::SetElementData(edc);
		GetElemDataIVector(GetMBS(), edc, "element_numbers_shortlist", elementnumbers_shortlist, 1);
		GetElemDataIVector(GetMBS(), edc, "local_node_numbers_shortlist", nodenumbers_local_shortlist, 1);
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("LocalNodeSetB"); } 

protected:
	TArray<int> elementnumbers_shortlist; //*EDC*[varaccess,EDCvarname="elements_shortlist",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
	TArray<int> nodenumbers_local_shortlist; //*EDC*[varaccess,EDCvarname="nodes_local_shortlist",EDCfolder="",variable_length_vector,tooltiptext="local node numbers of the set"]
}; //$EDC$[endclass,LocalNodeSetB]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                         GLOBALCOORDSET (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (coordinate-based) point set that is defined by global coordinates
//$EDC$[beginclass,classname=GlobalCoordSet,parentclass=PointSetBC,
//texdescription="A (coordinate-based) point set that is defined by global coordinates."]
class GlobalCoordSet : public PointSetBC 
{ 
public: // lifecycle I
	GlobalCoordSet(MBS* mbsi): PointSetBC(mbsi,"GlobalCoordSet",TMBSSetGlobalCoords), positions_global(0) {;}
	GlobalCoordSet(const GlobalCoordSet& other) { CopyFrom(other); }
	~GlobalCoordSet() {}

public: // lifecycle II
	GlobalCoordSet(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetGlobalCoords), positions_global(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new GlobalCoordSet(*this);
		return sc;
	}
	void CopyFrom(const GlobalCoordSet& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const GlobalCoordSet& o = (const GlobalCoordSet&)other;
		positions_global.CopyFrom(o.positions_global);
	}

public: // access
	virtual Vector3D GetGlobalPosition(int i) { return positions_global(i); }
	virtual void SetGlobalPosition(int i, Vector3D coord) { positions_global(i) = coord; }
	virtual Vector3D& GlobalPosition(int i) { return positions_global(i); }

public: // operations and operators
	virtual void AddEntry(Vector3D coord) 
	{ 
		positions_global.Add(coord);
		good2use = false;
	}
	virtual void DeleteEntry(int i) 
	{ 
		if(IsGood2Use())
		{
			elementnumbers.Erase(i);
			positions_local.Erase(i);
		}
		positions_global.Erase(i);
	}
	virtual void Reset() { elementnumbers.Flush(); positions_local.Flush(); positions_global.Flush(); }
	virtual int GetNrOfEntries() { return positions_global.Length(); }

	virtual int Check() 
	{
		int nerrors = 0;
		// TODO: actually check anything...
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in GlobalCoordSet " << GetName() << " - for " << nerrors << " entries ...\n";
		}
		return nerrors;
	}
	virtual int Convert() // compute the according element numbers and local positions from global node numbers
	{
		//ASSUME
		// this is only done in the MBS environment !
	
		for(int i=1; i<=positions_global.Length(); i++)
		{
			Vector3D gpos = positions_global(i);
			int elnr = GetMBS()->GetElementNumberOfGlobalPosition(gpos);
			Vector3D lpos = GetMBS()->GetLocalPositionOfGlobalPosition(elnr,gpos);
		// write converted data into the according fields
			SetElementNumber(i,elnr);
			SetLocalPosition(i,lpos);
		}

		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The GlobalCoordSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += "(";
			rv += mystr(GetGlobalPosition(i).X()); 
			rv += ", ";
			rv += mystr(GetGlobalPosition(i).Y()); 
			rv += ", ";
			rv += mystr(GetGlobalPosition(i).Z()); 
			rv += "), ";
		}
		rv += "(";
		rv += mystr(GetGlobalPosition(GetNrOfEntries()).X()); 
		rv += ", ";
		rv += mystr(GetGlobalPosition(GetNrOfEntries()).Y()); 
		rv += ", ";
		rv += mystr(GetGlobalPosition(GetNrOfEntries()).Z()); 
		rv += ")}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		PointSetBC::GetElementData(edc);
		ElementData* edp;
		edp = edc.GetPtr(edc.Find("element_numbers")); edp->SetLocked(1);
		edp = edc.GetPtr(edc.Find("local_positions")); edp->SetLocked(1);
		
		ElementData ed;
		Matrix mm(positions_global.Length(),3); 
		for(int i=1; i<=positions_global.Length(); i++) 
		{ 
			mm(i,1) = positions_global(i).X();
			mm(i,2) = positions_global(i).Y();
			mm(i,3) = positions_global(i).Z();
		}
		ed.SetMatrix(mm.GetMatPtr(), mm.Getrows(), mm.Getcols(), "global_positions");
		ed.SetVariableLength(); ed.SetToolTipText("Global positions registered in this set");
		edc.Add(ed);
	}
	
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		PointSetBC::SetElementData(edc);

		Matrix gps;
		GetElemDataMatrix(GetMBS(), edc, "global_positions", gps, 1);
		for(int i=1; i<=gps.Getrows(); i++)
		{
			Vector3D gp(gps(i,1),gps(i,2),gps(i,3));
			SetGlobalPosition(i,gp);
		}
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("GlobalCoordSet"); } 

protected:
	TArray<Vector3D> positions_global; //*EDC*[varaccess,EDCvarname="positions_global",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
}; //$EDC$[endclass,GlobalCoordSet]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                         LOCALCOORDSETA (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (coordinate-based) point set that is defined by element number and local coordinates - holds two lists of equal length forming pairs of (elnr,locpos)
//$EDC$[beginclass,classname=LocalCoordSetA,parentclass=PointSetBC,
//texdescription="A (coordinate-based) point set that is defined by element number and local coordinates - holds two lists of equal length forming pairs of (elnr,locpos)."]
class LocalCoordSetA : public PointSetBC
{ 
public: // lifecycle I
	LocalCoordSetA(MBS* mbsi): PointSetBC(mbsi,"LocalCoordSetA",TMBSSetLocalCoordsA) {;}
	LocalCoordSetA(const LocalCoordSetA& other) { CopyFrom(other); }
	~LocalCoordSetA() {}

public: // lifecycle II
	LocalCoordSetA(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetLocalCoordsA) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new LocalCoordSetA(*this);
		return sc;
	}
	void CopyFrom(const LocalCoordSetA& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const LocalCoordSetA& o = (const LocalCoordSetA&)other;
	}

public: // access
	// no own members
public: // operations and operators
	// AddEntry(int,Vector3D) ... in BaseClass
	// DeleteEntry(int)       ... in BaseClass
	// Reset()                ... in BaseClass
	// GetNrOfEntries()       ... in BaseClass

	virtual int Check() 
	{
		int nerrors = 0;
		int nelem = GetMBS()->NE();
		for(int i=1; i<=elementnumbers.Length(); i++)
		{
			if(elementnumbers(i)>nelem || elementnumbers(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in LocalCoordSet " << GetName() << " - for " << nerrors << " entries the element does not exist\n";
		}
		return nerrors;
	}
	virtual int Convert()
	{
		//ASSUME
		// this is only done in the MBS environment !
		
		// nothing to do

		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The LocalCoordSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += mystr(GetElementNumber(i));
			rv += ".(";
			rv += mystr(GetLocalPosition(i).X()); 
			rv += ", ";
			rv += mystr(GetLocalPosition(i).Y()); 
			rv += ", ";
			rv += mystr(GetLocalPosition(i).Z()); 
			rv += "), ";
		}
		rv += mystr(GetElementNumber(GetNrOfEntries()));
		rv += ".(";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).X()); 
		rv += ", ";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).Y()); 
		rv += ", ";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).Z()); 
		rv += ")}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	// GetElementData         ... in BaseClass
	// SetElementData         ... in BaseClass
	virtual mystr GetSetSpecification() { return mystr("LocalCoordSetA"); } 

protected:
	// no own members
}; //$EDC$[endclass,LocalCoordSetA]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                         LOCALCOORDSSETB (PointSet)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// (coordinate-based) point set that is defined by element number and local coordinates - holds two lists of equal length holds two lists of different length for elements and local coordinates  - the entire list of (elnr/locpos) pairs contains every combination
//$EDC$[beginclass,classname=LocalCoordSetB,parentclass=PointSetBC,
//texdescription="A (coordinate-based) point set that is defined by element number and local coordinates - holds two lists of different length for elements and local coordinates  - the entire list of (elnr/locpos) pairs contains every combination."]
class LocalCoordSetB : public PointSetBC 
{ 
public: // lifecycle I
	LocalCoordSetB(MBS* mbsi): PointSetBC(mbsi,"LocalCoordSetB",TMBSSetLocalCoordsB), elementnumbers_shortlist(0), positions_local_shortlist(0) {;}
	LocalCoordSetB(const LocalCoordSetB& other) { CopyFrom(other); }
	~LocalCoordSetB() {}

public: // lifecycle II
	LocalCoordSetB(MBS* mbsi, mystr namei): PointSetBC(mbsi,namei,TMBSSetLocalCoordsB), elementnumbers_shortlist(0), positions_local_shortlist(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new LocalCoordSetB(*this);
		return sc;
	}
	void CopyFrom(const LocalCoordSetB& other) 
	{
		// copy base class content
		PointSetBC::CopyFrom(other);
		// specific for each derived class
		const LocalCoordSetB& o = (const LocalCoordSetB&)other;
		elementnumbers_shortlist.CopyFrom(o.elementnumbers_shortlist);
		positions_local_shortlist.CopyFrom(o.positions_local_shortlist);
	}

public: // access
	virtual int GetElementNumberSL(int i) { return elementnumbers_shortlist(i); }              // ElementNumber SHORTLIST - GetElementNumber MUST always give back the element from the full list (after conversion)
	virtual void SetElementNumberSL(int i, int elnr) { elementnumbers_shortlist(i) = elnr; }
	virtual int& ElementNumberSL(int i) { return elementnumbers_shortlist(i); }
	virtual Vector3D GetLocalPositionSL(int i) { return positions_local_shortlist(i); }
	virtual void SetLocalPositionSL(int i, Vector3D pos) { positions_local_shortlist(i) = pos; }
	virtual Vector3D& LocalPositionsSL(int i) { return positions_local_shortlist(i); }

public: // operations and operators
	virtual void AddEntryElem(int elnr)
	{ 
		elementnumbers_shortlist.Add(elnr);
		good2use = false;
	}
	virtual void AddEntryPosition(Vector3D pos)
	{ 
		positions_local_shortlist.Add(pos);
		good2use = false;
	}
	virtual void DeleteEntryElem(int i) 
	{ 
		if(IsGood2Use())
		{
		// ASSUME - fixed order of entries in the (element,localposition) list
		// erase entries (#nodes*i)+1 through (#nodes*i)+#nodes		
			int nposs = positions_local_shortlist.Length();
			for(int j=nposs; j>1; j--)
			{
				int idx = nposs*i+j;
				elementnumbers.Erase(idx);
				positions_local.Erase(idx);
			}
		}
		elementnumbers_shortlist.Erase(i);
	}
	virtual void DeleteEntryPosition(int i) 
	{ 
		if(IsGood2Use())
		{
		// ASSUME - fixed order of entries in the (element,localposition) list
		// erase entries i, i+k*#nodes 
			int nposs = positions_local_shortlist.Length();
			int nelems = elementnumbers_shortlist.Length();
			for(int j=nelems; j>1; j--)
			{
				int idx = nposs*(j-1)+i;
				elementnumbers.Erase(idx);
				positions_local.Erase(idx);
			}
		}
		positions_local_shortlist.Erase(i);
	}
	virtual void Reset() { elementnumbers.Flush(); elementnumbers_shortlist.Flush(); positions_local.Flush(); positions_local_shortlist.Flush(); }

	virtual int Check() 
	{
		int nerrors = 0;
		int nelem = GetMBS()->NE();
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++)
		{
			if(elementnumbers_shortlist(i)>nelem || elementnumbers_shortlist(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in LocalCoordSet " << GetName() << " - for " << nerrors << " entries the element does not exist\n";
		}
		return nerrors;
	}
	virtual int Convert() // compute the according element numbers and local positions from global node numbers
	{
		//ASSUME
		// this is only done in the MBS environment !

		// mind the loops to construct the set product in the following way:
		// { (el1,lp1), (el1,lp2), (el1,lp3), (el2,lp1), (el2,lp2), ... }
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++)
		{
			for(int j=1; j<=positions_local_shortlist.Length(); j++)
			{
				int elnr = elementnumbers_shortlist(i);
				Vector3D lpos = positions_local_shortlist(j);
				int idx = (i-1)*positions_local_shortlist.Length()+j;
		// local pos from local node number
				SetElementNumber(idx,elnr);
				SetLocalPosition(idx,lpos);		
			}
		}

		good2use = true;
		return 0;
	}
	virtual mystr ToString()
	{
		mystr rv("The LocalNodeSet ");
		rv += name;
		rv += " contains all combinations of the following ELEMENTS: {";
		for(int i=1; i<elementnumbers_shortlist.Length(); i++)
		{
			rv += mystr(GetElementNumberSL(i));
			rv += ", ";
		}
		rv += mystr(GetElementNumberSL(elementnumbers_shortlist.Length()));
		rv += "}\n";

		rv += " and LOCALPOSITIONS: {";
		for(int i=1; i<positions_local_shortlist.Length(); i++)
		{
			rv += "(";
			rv += mystr(GetLocalPosition(GetNrOfEntries()).X()); 
			rv += ", ";
			rv += mystr(GetLocalPosition(GetNrOfEntries()).Y()); 
			rv += ", ";
			rv += mystr(GetLocalPosition(GetNrOfEntries()).Z()); 
			rv += "), ";
		}
		rv += "(";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).X()); 
		rv += ", ";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).Y()); 
		rv += ", ";
		rv += mystr(GetLocalPosition(GetNrOfEntries()).Z()); 
		rv += ")}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		PointSetBC::GetElementData(edc);
		ElementData* edp;
		edp = edc.GetPtr(edc.Find("element_numbers")); edp->SetLocked(1);
		edp = edc.GetPtr(edc.Find("local_positions")); edp->SetLocked(1);
		
		ElementData ed;
		Vector vv(elementnumbers_shortlist.Length()); 
		for(int i=1; i<=elementnumbers_shortlist.Length(); i++) 
		{ 
			vv(i) = elementnumbers_shortlist(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "element_numbers_shortlist");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Shortlist of Elements in this set");
		edc.Add(ed);

		Matrix mm(positions_local_shortlist.Length(),3);
		for(int i=1; i<=positions_local_shortlist.Length(); i++) 
		{ 
			mm(i,1) = positions_local_shortlist(i).X();
			mm(i,2) = positions_local_shortlist(i).Y();
			mm(i,3) = positions_local_shortlist(i).Z();
		}
		ed.SetMatrix(mm.GetMatPtr(), mm.Getrows(), mm.Getcols(), "local_positions_shortlist");
		ed.SetVariableLength(); ed.SetToolTipText("Shortlist of Local Positions in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		PointSetBC::SetElementData(edc);
		GetElemDataIVector(GetMBS(), edc, "element_numbers_shortlist", elementnumbers_shortlist, 1);
		Matrix lps;
		GetElemDataMatrix(GetMBS(), edc, "local_positions_shortlist", lps, 1);
		for(int i=1; i<=lps.Getrows(); i++)
		{
			Vector3D lp(lps(i,1),lps(i,2),lps(i,3));
			SetLocalPositionSL(i,lp);
		}
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("LocalCoordSetB"); }

protected:
	TArray<int> elementnumbers_shortlist; //*EDC*[varaccess,EDCvarname="elements_shortlist",EDCfolder="",variable_length_vector,tooltiptext="elements of the set"]
	TArray<Vector3D> positions_local_shortlist; //*EDC*[varaccess,EDCvarname="positions_local_shortlist",EDCfolder="",variable_length_vector,tooltiptext="local node numbers of the set"]
}; //$EDC$[endclass,LocalCoordSetB]



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                            FACESETA
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// set containting Faces
//$EDC$[beginclass,classname=FaceSet,parentclassname=BaseSet,
//texdescription="A face set that is defined by element number and local (element) face number - holds two lists of equal length.",example="sets.txt"]
class FaceSetA : public BaseSet 
{
public: // lifecycle I
	FaceSetA(MBS* mbsi): BaseSet(mbsi,"FaceSetA",TMBSSetFaces), elementnumbers(0), facenumbers(0), facenodelengths(0), facenodenumbers(0), facenodelocalnumbers(0), faceareas(0) {;}
	FaceSetA(const FaceSetA& other) { CopyFrom(other); }
	~FaceSetA() {}

public: // lifecycle II
	FaceSetA(MBS* mbsi, mystr namei): BaseSet(mbsi,namei,TMBSSetFaces), elementnumbers(0), facenumbers(0), facenodelengths(0), facenodenumbers(0), facenodelocalnumbers(0), faceareas(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new FaceSetA(*this);
		return sc;
	}
	void CopyFrom(const FaceSetA& other)
	{
		// copy base class content
		BaseSet::CopyFrom(other);
		// specific for each derived class
		const FaceSetA& o = (const FaceSetA&)other;
		elementnumbers.CopyFrom(o.elementnumbers);
		facenumbers.CopyFrom(o.facenumbers);
		facenodelengths.CopyFrom(o.facenodelengths);
		facenodenumbers.CopyFrom(o.facenodenumbers);
		facenodelocalnumbers.CopyFrom(o.facenodelocalnumbers);
		faceareas.CopyFrom(o.faceareas);
		usednodes.CopyFrom(o.usednodes);
	}

public: // access
	virtual int GetElementNumber(int i) { return elementnumbers(i); }
	virtual void SetElementNumber(int i, int elnr) { elementnumbers(i) = elnr; }
	virtual int& ElementNumber(int i) { return elementnumbers(i); }
	virtual TArray<int>& ElementNumbers() { return elementnumbers; }

	virtual int GetFaceNumber(int i) { return facenumbers(i); }
	virtual void SetFaceNumber(int i, int fanr) { facenumbers(i) = fanr; }
	virtual int& FaceNumber(int i) { return facenumbers(i); }
	virtual TArray<int>& FaceNumbers() { return facenumbers; }

	virtual int GetFaceNodeLength(int i) { return facenodelengths(i); }
	virtual void SetFaceNNodeLength(int i, int flen) { facenodelengths(i) = flen; }
	virtual int& FaceNodeLength(int i) { return facenodelengths(i); }
	virtual TArray<int>& FaceNodeLengths() { return facenodelengths; }

	virtual TArray<int> GetFaceNodeNumber(int i) { return facenodenumbers(i); }
	virtual void SetFaceNodeNumber(int i, TArray<int>& fnnr) { facenodenumbers(i).CopyFrom(fnnr); }
	virtual TArray<int>& FaceNodeNumber(int i) { return facenodenumbers(i); }
	virtual TArrayDynamic<TArray<int>>& FaceNodeNumbers() { return facenodenumbers; }

	virtual TArray<int> GetFaceNodeLocalNumber(int i) { return facenodelocalnumbers(i); }
	virtual void SetFaceNodeLocalNumber(int i, TArray<int>& fnnr) { facenodelocalnumbers(i).CopyFrom(fnnr); }
	virtual TArray<int>& FaceNodeLocalNumber(int i) { return facenodelocalnumbers(i); }
	virtual TArrayDynamic<TArray<int>>& FaceNodeLocalNumbers() { return facenodelocalnumbers; }

	virtual double GetFaceArea(int i) { return faceareas(i); }
	virtual void SetFaceArea(int i, double farea) { faceareas(i) = farea; }
	virtual double& FaceArea(int i) { return faceareas(i); }
	virtual TArray<double>& FaceAreas() { return faceareas; }

public: // operations and operators
	virtual void AddEntry(int elnr, int facenr, int facelen, const TArray<int>& facenodes, const TArray<int>& facenodeslocal, double farea) 
	{ 
		elementnumbers.Add(elnr); 
		facenumbers.Add(facenr);
		facenodelengths.Add(facelen);
		facenodenumbers.Add(facenodes);
		facenodelocalnumbers.Add(facenodeslocal);
		faceareas.Add(farea);
		for(int i=1; i<=facenodes.Length(); i++) { usednodes.AddIfNotExists(facenodes(i)); }
	}
	virtual void DeleteEntry(int i) 
	{
		elementnumbers.Erase(i); 
		facenumbers.Erase(i);
		facenodelengths.Erase(i);
		facenodenumbers.Erase(i);
		facenodelocalnumbers.Erase(i);
		faceareas.Erase(i);

	}
	virtual void Reset() 
	{
		elementnumbers.Flush(); 
		facenumbers.Flush();
		facenodelengths.Flush();
		facenodenumbers.Flush();
		facenodelocalnumbers.Flush();
		faceareas.Flush();
		usednodes.Flush();
	}
	virtual int GetNrOfEntries() { return elementnumbers.Length(); }

	virtual int Check()
	{
		int nerrors = 0;
		int nelements = GetMBS()->NE();
		int nnodes = GetMBS()->NNodes();
// all arrays must have same length
		if(elementnumbers.Length()!=facenumbers.Length()) nerrors++;
		if(elementnumbers.Length()!=facenodelengths.Length()) nerrors++;
		if(elementnumbers.Length()!=facenodenumbers.Length()) nerrors++;
		if(elementnumbers.Length()!=facenodelocalnumbers.Length()) nerrors++;
		if(elementnumbers.Length()!=faceareas.Length()) nerrors++;
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in FaceSet " << GetName() << " - lenghts of lists do not match\n";
		}

		nerrors=0;
		for(int i=1; i<=elementnumbers.Length(); i++)
		{
			if(elementnumbers(i)>nelements || elementnumbers(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in FaceSet " << GetName() << " - for " << nerrors << " entries the element does not exist\n";
		}
		
		nerrors=0;
		for(int i=1; i<=facenodenumbers.Length(); i++)
		{
			for(int j=1; j<=facenodelengths(i); j++)
			{
				if(facenodenumbers(i)(j)>nnodes || facenodenumbers(i)(j)<=0)
				{
					nerrors++;
				}
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in FaceSet " << GetName() << " - for " << nerrors << " entries the node does not exist\n";
		}
		return nerrors;
	}
	virtual mystr ToString()
	{
		mystr rv("The FaceSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += "E";
			rv += mystr(GetElementNumber(i)); 
			rv += "F";
			rv += mystr(GetFaceNumber(i)); 
			rv += ", ";
		}
		rv += "E";
		rv += mystr(GetElementNumber(GetNrOfEntries()));
		rv += "F";
		rv += mystr(GetFaceNumber(GetNrOfEntries())); 
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseSet::GetElementData(edc);

		ElementData ed;
		Vector vv(elementnumbers.Length()); 
		for(int i=1; i<=elementnumbers.Length(); i++) 
		{ 
			vv(i) = elementnumbers(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "element_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Elements in this set");
		edc.Add(ed);

		// local face index in element
		for(int i=1; i<=facenumbers.Length(); i++) { vv(i) = facenumbers(i); } 
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "face_numbers"); 		
		ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("ElementFaces in this set");
		edc.Add(ed);

		// number of nodes per face
		for(int i=1; i<=facenodelengths.Length(); i++) { vv(i) = facenodelengths(i); } 
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "number_of_nodes"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("number of nodes for each face in this set"); 
		edc.Add(ed);
		
		// global node numbers
		Matrix mm(facenodenumbers.Length(),8);
		mm.SetAll(0);
		for(int i=1; i<=facenodenumbers.Length(); i++)
		{
			for(int j=1; j<=facenodelengths(i); j++) { mm(i,j) = facenodenumbers(i)(j);	}
		}
		ed.SetMatrix(mm.GetMatPtr(),facenodenumbers.Length(),8,"node_lists"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("node list for each face in this set");
		edc.Add(ed);

		// local node index in element
		for(int i=1; i<=facenodelocalnumbers.Length(); i++)
		{
			for(int j=1; j<=facenodelengths(i); j++) { mm(i,j) = facenodelocalnumbers(i)(j);	}
		}
		ed.SetMatrix(mm.GetMatPtr(),facenodelocalnumbers.Length(),8,"node_lists_local"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("local node number in element");
		edc.Add(ed);

		for(int i=1; i<=faceareas.Length(); i++) { vv(i) = faceareas(i); }
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "face_areas"); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("ElementFaceAreas in this set");
		edc.Add(ed);		

		Vector vv2(usednodes.Length()); // not the same length
		for(int i=1; i<=usednodes.Length(); i++) { vv2(i) = usednodes(i); }
		ed.SetVector(vv2.GetVecPtr(), vv2.Length(), "used_nodes"); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("all global nodes used in this set");
		edc.Add(ed);		
	}
	virtual int SetElementData(ElementDataContainer &edc)
	{
		int rv = 1;
		BaseSet::SetElementData(edc);

		GetElemDataIVector(GetMBS(), edc, "element_numbers", elementnumbers, 1);
		GetElemDataIVector(GetMBS(), edc, "face_numbers", facenumbers, 1);

		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("FaceSetA"); }

protected:
	TArray<int> elementnumbers; //*EDC*[varaccess,EDCvarname="elements",EDCfolder="",variable_length_vector,tooltiptext="members of the element set"]
	TArray<int> facenumbers;
	TArray<int> facenodelengths;
	//TArray<int4> facenodenumbers;
	TArrayDynamic<TArray<int>> facenodenumbers;
	TArrayDynamic<TArray<int>> facenodelocalnumbers;
	TArray<double> faceareas;
	TArray<int> usednodes;
}; //$EDC$[endclass,FaceSet]

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                              SENSORSET
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// set containting sensors
//$EDC$[beginclass,classname=SensorSet,parentclassname=BaseSet,
//texdescription="A set containing only sensors (as sensor numbers).",example="sets.txt"]
class SensorSet : public BaseSet 
{
public: // lifecycle I
	SensorSet(MBS* mbsi): BaseSet(mbsi,"SensorSet",TMBSSetSensors), sensornumbers(0) {;}
	SensorSet(const SensorSet& other) { CopyFrom(other); }
	~SensorSet() {}

public: // lifecycle II
	SensorSet(MBS* mbsi, mystr namei): BaseSet(mbsi,namei,TMBSSetSensors), sensornumbers(0) {;}
	BaseSet* GetCopy()
	{
		BaseSet* sc = new SensorSet(*this);
		return sc;
	}
	void CopyFrom(const SensorSet& other) 
	{
		// copy base class content
		BaseSet::CopyFrom(other);
		// specific for each derived class
		const SensorSet& o = (const SensorSet&)other;
		sensornumbers.CopyFrom(o.sensornumbers);
	}

public: // access
	virtual int GetSensorNumber(int i) { return sensornumbers(i); }
	virtual void SetSensorNumber(int i, int sensnr) { sensornumbers(i) = sensnr; }
	virtual int& SensorNumber(int i) { return sensornumbers(i); }
	virtual TArray<int>& SensorNumbers() { return sensornumbers; }

public: // operations and operators
	virtual void AddEntry(int sensnr) { sensornumbers.Add(sensnr); }
	virtual void DeleteEntry(int i) { sensornumbers.Erase(i); }
	virtual void Reset() { sensornumbers.Flush(); }
	virtual int GetNrOfEntries() { return sensornumbers.Length(); }

	virtual int Check() // all entries of sensornumbers must exist in mbs
	{
		int nerrors = 0;
		int nsensors = GetMBS()->NSensors();
		for(int i=1; i<=sensornumbers.Length(); i++)
		{
			if(sensornumbers(i)>nsensors || sensornumbers(i)<=0)
			{
				nerrors++;
			}
		}
		if(nerrors)
		{
			GetMBS()->UO() << "Found inconsistencies in SensorSet " << GetName() << " - for " << nerrors << " entries the sensor does not exist\n";
		}
		return nerrors;
	}

	virtual mystr ToString()
	{
		mystr rv("The SensorSet ");
		rv += name;
		rv += " contains: {";
		for(int i=1; i<GetNrOfEntries(); i++)
		{
			rv += mystr(GetSensorNumber(i)); 
			rv += ", ";
		}
		rv += mystr(GetSensorNumber(GetNrOfEntries()));
		rv += "}\n";
		return rv;
	}

public: // EDC-interaction
	//virtual void GetElementDataAuto(ElementDataContainer& edc);
	//virtual int SetElementDataAuto(ElementDataContainer& edc);
	//virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata);
	//virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata);
	//virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables);
	virtual void GetElementData(ElementDataContainer& edc) 
	{
		BaseSet::GetElementData(edc);

		ElementData ed;
		Vector vv(sensornumbers.Length()); 
		for(int i=1; i<=sensornumbers.Length(); i++) 
		{ 
			vv(i) = sensornumbers(i);
		}
		ed.SetVector(vv.GetVecPtr(), vv.Length(), "sensor_numbers");
		ed.SetValuesInt();
		ed.SetVariableLength(); ed.SetToolTipText("Sensors in this set");
		edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = 1;
		BaseSet::SetElementData(edc);

		GetElemDataIVector(GetMBS(), edc, "sensor_numbers", sensornumbers, 1);
		return rv;
	}
	virtual mystr GetSetSpecification() { return mystr("SensorSet"); }

protected:
	TArray<int> sensornumbers; //*EDC*[varaccess,EDCvarname="sensors",EDCfolder="",variable_length_vector,tooltiptext="members of the sensor set"]
}; //$EDC$[endclass,SensorSet]