//#**************************************************************
//#
//# filename:             script_parser.h
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						July 2004
//# description:          parser for creating objects in script language
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


#ifndef SCRIPT_PARSER__H
#define SCRIPT_PARSER__H

#include "options_class_auto.h"

class CEDCParser;
class CMBSParser;
struct MBSObjectFactoryInterface;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//Parsing input files:

//typedef enum {TAEflexible=1, TAEconstraint=2, TAE2D = 4, TAEspecial_connector=8} TAddElementType; //$ DR 2012-07: old code

////+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
////parser for ElementDataContainer:
////$JG2012-01-27: created
////string-->EDC
////EDC-->string

//these are default values for the creation of bodies, sensors, etc.
class Default_ParserObjectValues
{
public:
	Default_ParserObjectValues()
	{
		defaultbodycol = Vector3D(0.2,0.2,0.8);
		defaultconstraintcol = Vector3D(0.2,0.8,0.2);
		defaultbodypos = Vector3D(0.,0.,0.);
		defaultbodyvel = Vector3D(0.,0.,0.);
		defaultbodyangle = Vector3D(0.,0.,0.);
		defaultbodyangularvel = Vector3D(0.,0.,0.);
		defaultbodysize = Vector3D(1.,1.,1.);
		defaultbodyrho = 1000;
		defaultconstraintdim = Vector3D(0.1,0.1,16.);

		defaultsensorvisible = 1;
		defaultsensorwriteresults = 3;
		defaultsensorprecision = 17;
		defaultsensordrawdim = Vector3D(0.001,6,0);

		defaultgeomtile = 16;
		defaultgeomcol = Vector3D(0.5,0.5,0.5);
    defaultgeomdim = Vector3D(1,1,1); 
    defaulttransparency=-1;

		//flexible bodies:
		defaultbodyEm = 1e8;
		defaultbodynu = 0.;
	}
public:
	Vector3D defaultbodycol;
	Vector3D defaultconstraintcol;
	Vector3D defaultbodypos;
	Vector3D defaultbodyvel;
	Vector3D defaultbodyangle;
	Vector3D defaultbodyangularvel;
	Vector3D defaultbodysize;
	double defaultbodyrho;
	Vector3D defaultconstraintdim;

	int defaultsensorvisible;
	int defaultsensorwriteresults;
	int defaultsensorprecision;
	Vector3D defaultsensordrawdim;

	int defaultgeomtile;
	Vector3D defaultgeomcol;
	Vector3D defaultgeomdim; 
  double defaulttransparency;

	//flexible bodies:
	double defaultbodyEm;
	double defaultbodynu;
};

//typedef enum {TCT_AddElement=1, TCT_AddConstraint=2, TCT_AddSensor=3, TCT_AddLoad=4, TCT_AddGeomElement=5, } TEDCParser_CommandType;

typedef enum { TCO_None = 0, TCO_FollowedByContainer = 1, TCO_LeftHandSideOnly = 2, TCO_RequiresHandle = 4} TEDCParser_CommandOption;

//class for EDC-based commands, which can be parsed
class CEDC_Command
{
public:
	//initialization
	CEDC_Command(int id) 
	{
		command_id = id;
	}

	//unit command id, which is automatically assigned
	virtual int GetCommandID() const {return command_id;}

	//name of command in parser
	virtual const char* CommandName() const {assert(0); return "";}

	//Tex description for the docu
	virtual mystr GetCommandTexDescription() const {return "";};
	virtual mystr GetCommandTexParameterDescription() const {return "";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "";};
	virtual mystr GetCommandExampleFileName() const {return "";};

	//number of fixed parameters, which are necessary for the execution of the command
	virtual int NParameters() const {return 0;}

	virtual int HasReturnValue() const {return 0;} //0: no return value, 1: has a return value (ElementData)
	virtual int GetOptions() const {return 0;} //some options, which are not yet specified

	//execute command, 1==success, 0==failed
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option) {assert(0); return 0;}

	//convert textual (inline, not previously defined as variables) parameter to numerical value
	//NOTE: inline definition of Vector is NOT working: 
	//in function GetParameterEDC spaces get contracted, "," are interpreted as seperators for arguments
	virtual int EvaluateTextualParameterEntry(MBS* mbs, CEDCParser * edcParser, ElementDataContainer* param_edc); 

	virtual int ResolveParameterEntryToSetNumber(MBS* mbs, CEDCParser * edcParser, ElementData* param_ed); 
	virtual int ResolveParameterEntryToMaterialNumber(MBS* mbs, CEDCParser * edcParser, ElementData* param_ed); 
	virtual int ResolveParameterEntryToLoadNumber(MBS* mbs, CEDCParser * edcParser, ElementData* param_ed); 
	virtual int ResolveParameterEntryToMeshNumber(MBS* mbs, CEDCParser * edcParser, ElementData* param_ed); 
	virtual int ResolveParameterEntryToComponentNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed, int meshIdx);

private:
	int command_id;
};

class CEDC_ComAddElement: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddElement(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddElement";}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "Adds an element to the system. See the description of the elements above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the element. ATTENTION: the entry element\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the element in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddElement.txt";};

	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return 0;}

	virtual int IsAddElement() const {return 1;} //this only distinguishes the AddElement from the AddConstraint command

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComAddConstraint: public CEDC_ComAddElement
{
public:
	//initialization
	CEDC_ComAddConstraint(int id): CEDC_ComAddElement(id) {}

	virtual const char* CommandName() const {return "AddConnector";}
	virtual int IsAddElement() const {return 0;} //this only distinguishes the AddElement from the AddConstraint command
	virtual mystr GetCommandTexDescription() const {return "Adds a connector to the system. See the description of the connectors above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the connector. ATTENTION: the entry element\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the connector in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddConnector.txt";};
};
//$ RL 2012-7-19:[  
class CEDC_ComAddGeomElement: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddGeomElement(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddGeomElement";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command adds an geometric element.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the geometric element. ATTENTION: the entry geom\\_element\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the geometric element in the MBS.";}
	virtual mystr GetCommandExampleFileName() const {return "AddGeomElement.txt";};


	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComInclude: public CEDC_Command
{
public:
	//initialization
	CEDC_ComInclude(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Include";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 0;}


	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command includes a file.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is the absolut or relative filename. If a relative filename is used, then the path is relative to the last file! Be carefull, if you use this command more than one time in a file.";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "There is no return value defined yet.";}
	virtual mystr GetCommandExampleFileName() const {return "Include.txt";};


	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ RL 2012-7-19:] 

class CEDC_ComPrint: public CEDC_Command
{
public:
	//initialization
	CEDC_ComPrint(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Print";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 0;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "Prints a text to the output window";};
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const {return "There is no return value for this command";};
	virtual mystr GetCommandExampleFileName() const {return "Print.txt";};

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComPrintIf: public CEDC_ComPrint
{
public:
	//initialization
	CEDC_ComPrintIf(int id): CEDC_ComPrint(id) {}

	virtual const char* CommandName() const {return "PrintIf";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "additional boolean, Prints a text to the output window";};
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const {return "There is no return value for this command";};
	virtual mystr GetCommandExampleFileName() const {return "Print.txt";};

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};


class CEDC_ComAddLoad: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddLoad(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddLoad";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a load to the system. See the description of the loads above in order to get the available options. You have to adjust the value 'loads' in the element to assign the load to the element.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the load. ATTENTION: the entry load\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the load in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddLoad.txt";};
private:
};

//$ DR 2012-10
class CEDC_ComAddSensor: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddSensor(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddSensor";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a sensor to the system. See the description of the sensors above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the sensor. ATTENTION: the entry sensor\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the sensor in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddSensor.txt";};
private:
};

class CEDC_ComAddSensorProcessor: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddSensorProcessor(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddSensorProcessor";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a sensorProcessor to an existing. See the description of the sensorProcessors above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The 1st parameter of this command is an ElementDataContainer with the data of the sensor followed by the index number of the sensor to add to as 2nd parameter. ATTENTION: the entry processor\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the sensor in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddSensorProcessor.txt";};
private:
};

//$ DR 2012-12-11:[
class CEDC_ComLoadSTL: public CEDC_Command
{
public:
	//initialization
	CEDC_ComLoadSTL(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "ReadSTLFile";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}


	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command reads a stl-mesh from a file and stores the data in an ElementDataContainer.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is the absolut or relative filename.";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is an ElementDataContainer with 2 entries: triangles and points.";}
	virtual mystr GetCommandExampleFileName() const {return "ReadSTLFile.txt";};


	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ DR 2012-12-11:] 

class CEDC_ComAddMaterial: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddMaterial(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddMaterial";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a material to the system. See the description of the materials above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the material. ATTENTION: the entry material\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the material in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddMaterial.txt";};
private:
};

//$ DR 2013-01
class CEDC_ComAddBeam3DProperties: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddBeam3DProperties(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddBeamProperties";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a BeamProperty to the system. See the description of the BeamProperties above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the BeamProperties. ATTENTION: the entry material\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the node in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddBeamProperties.txt";};
private:
};

//$ DR 2013-01
class CEDC_ComAddNode: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddNode(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddNode";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a node to the system. See the description of the nodes above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the node. ATTENTION: the entry node\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the node in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddNode.txt";};
private:
};

//$ AD 2014-08
class CEDC_ComAddSet: public CEDC_Command
{
public:
	//initialization
	CEDC_ComAddSet(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddSet";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
	virtual mystr GetCommandTexDescription() const {return "Adds a set to the system. See the description of the set above in order to get the available options.";};
	virtual mystr GetCommandTexParameterDescription() const {return "The parameter of this command is an ElementDataContainer with the data of the set. ATTENTION: the entry set\\_type must exist!";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value of this command is the number of the set in the MBS.";};
	virtual mystr GetCommandExampleFileName() const {return "AddSet.txt";};
private:
};


//$ DR 2013-01-30:[
class CEDC_ComComputeInertia: public CEDC_Command
{
public:
	//initialization
	CEDC_ComComputeInertia(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "ComputeInertia";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command computes the mass, moment of inertia, volume and center of mass based on the information about the geometry and the material of a body";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is an ElementDataContainer with 4 entries: volume, mass, moment\\_of\\_inertia and center\\_of\\_mass";}
	virtual mystr GetCommandExampleFileName() const {return "ComputeInertia.txt";};


	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ DR 2013-01-30:] 

//$ DR 2013-02-19:[
class CEDC_ComTransformPoints: public CEDC_Command
{
public:
	//initialization
	CEDC_ComTransformPoints(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "TransformPoints";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "With this command, the geometry described by the points can be transformed. It is possible to apply rotation and/or translation and/or scaling. The new point pN is computed according to the formula pN = trans + rot*p.";};
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a Matrix containing the transformed points pN.";}
	virtual mystr GetCommandExampleFileName() const {return "TransformPoints.txt";};

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ DR 2013-02-19:] 

//$ AD: 2017-08-01
class CEDC_ComRotMatToAngles: public CEDC_Command
{
public:
	//initialization
	CEDC_ComRotMatToAngles(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "RotMat2Angles";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "For a given rotation matrix the Euler Angles Z-X-Z and Kardan Angles are computed";};
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a set Vector3D containing the angles [rad].";}
	virtual mystr GetCommandExampleFileName() const {return "RotMat2Angles.txt";};

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ PG 2020-05-04: 
class CEDC_ComRotMatFromAngles : public CEDC_Command
{
public:
	//initialization
	CEDC_ComRotMatFromAngles(int id) : CEDC_Command(id) {}

	virtual const char* CommandName() const { return "Angles2RotMat"; }
	virtual int NParameters() const { return 2; }
	virtual int HasReturnValue() const { return 1; }

	//description for the docu
	virtual mystr GetCommandTexDescription() const { return "For a given sequence of Euler Angles Z-X-Z or Kardan Angles X-Y-Z, both in [rad], the rotation matrix is computed. Euler Angles Z-X-Z are assumed if second parameter is 0; Kardan Angles X-Y-Z are assumed if second parameter is 1."; };
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const { return "The return value is an orthogonal rotation matrix (3D)."; }
	virtual mystr GetCommandExampleFileName() const { return "Angles2RotMat.txt"; };

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ DR 2013-07-04:[
class CEDC_ComLoadVecFromFile: public CEDC_Command
{
public:
	//initialization
	CEDC_ComLoadVecFromFile(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "LoadVectorFromFile";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command reads a vector from a file and returns this vector.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the vector.";}
	virtual mystr GetCommandExampleFileName() const {return "LoadVectorFromFile.txt";};

	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ DR 2013-07-04:] 

//$ AD 2013-07-11:[
class CEDC_ComSum: public CEDC_Command
{
public:
	//initialization
	CEDC_ComSum(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Sum";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command adds two components of the same type ( scalar, vector or matrix ).";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the sum of the two inputs.";}
	virtual mystr GetCommandExampleFileName() const {return "Sum.txt";};
	
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComProduct: public CEDC_Command
{
public:
	//initialization
	CEDC_ComProduct(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Product";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command multiplies two components of the type ( scalar, vector or matrix ) when the operation is defined.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the product of the two inputs.";}
	virtual mystr GetCommandExampleFileName() const {return "Product2.txt";};
	
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComTranspose: public CEDC_Command
{
public:
	//initialization
	CEDC_ComTranspose(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Transpose";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command transposes a matrix or vector.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a matrix or a vector.";}
	virtual mystr GetCommandExampleFileName() const {return "Transpose.txt";};
	
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$! AD 2013-07-15 commented out Class Scalar product for the time being, functionality is in ComProduct...
////class CEDC_ComScalarProduct: public CEDC_Command
////{
////public:
////	//initialization
////	CEDC_ComScalarProduct(int id): CEDC_Command(id) {}
////
////	virtual const char* CommandName() const {return "ScalarProduct";}
////	virtual int NParameters() const {return 2;}
////	virtual int HasReturnValue() const {return 1;}
////
////	//description for the docu
////	virtual mystr GetCommandTexDescription() const {return "This command computes the scalar product of two vectors.";};
////	virtual mystr GetCommandTexParameterDescription() const ;
////	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the scalar product.";}
////	virtual mystr GetCommandExampleFileName() const {return "AddMult.txt";};
////	
////	//execute command, 1==success, 0==failed
////	//it is already checked, that the number of parameters is equal to the desired number of parameters
////	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
////private:
////};		

class CEDC_ComCrossProduct: public CEDC_Command
{
public:
	//initialization
	CEDC_ComCrossProduct(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "CrossProduct";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command computes the cross product of two vectors.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the scalar cross product.";}
	virtual mystr GetCommandExampleFileName() const {return "CrossProduct.txt";};
	
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ AD 2013-07-11:] 

//$ AD 2013-09-13:[ loop and condition
class CEDC_ComFor: public CEDC_Command
{
public:
	//initialization
	CEDC_ComFor(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "for";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_FollowedByContainer + TCO_LeftHandSideOnly;} // the command is followed by a code block in a { } container

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command starts a FOR loop for the subsequent block.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the number of iterations.";}
	virtual mystr GetCommandExampleFileName() const {return "LoopCond.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComIf: public CEDC_Command
{
public:
	//initialization
	CEDC_ComIf(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "if";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_FollowedByContainer + TCO_LeftHandSideOnly;} // the command is followed by a code block in a { } container

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command evaluates an IF condition for the subsequent block.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the 1 for true and 0 for false.";}
	virtual mystr GetCommandExampleFileName() const {return "LoopCond.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ AD 2013-09-13:] 

//$ AD 2013-10-09:[ functions to generate specific handles
class CEDC_ComMesh_GenerateNewMesh: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GenerateNewMesh(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GenerateNewMesh";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Handle to a Mesh Object for further operations.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a special EDC (Handle) that must be assigned to a new variable.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGenerateMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ AD 2013-10-09:]

//$ AD 2013-10-09:[ Mesh commands
class CEDC_ComMesh_GenerateBeam: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GenerateBeam(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GenerateBeam";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;} 

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Beam within a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated beam within the mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGenerateBeam.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GeneratePlate: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GeneratePlate(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GeneratePlate";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Plate within a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated plate within the mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGeneratePlate.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GenerateBlock : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GenerateBlock(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "GenerateBlock";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Block within a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated block within the mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGenerateBlock.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GenerateCylinder : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GenerateCylinder(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "GenerateCylinder";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Cylinder within a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated cylinder within the mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGenerateCylinder.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_LoadMesh : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_LoadMesh(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "LoadMesh";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command loads a mesh from an external file into a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component containing the external mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshLoadMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_WriteMesh : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_WriteMesh(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "WriteMesh";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command loads a mesh from an external file into a Mesh Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component containing the external mesh.";}
	virtual mystr GetCommandExampleFileName() const {return "MeshLoadMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

/*
class CEDC_ComMesh_Process : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Process(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Process";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command applies a processing operation on an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return ".";}
	virtual mystr GetCommandExampleFileName() const {return "MeshProcessMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
*/

class CEDC_ComMesh_Transform : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Transform(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Transform";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command applies a transformation on an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshTransformMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Distort : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Distort(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Distort";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command applies a distortion on an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshDistort.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Modify : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Modify(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Modify";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command applies a modification on an entire Mesh or a single Mesh component. A filter list can be used";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshModify.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};


class CEDC_ComMesh_Linear2Quadratic : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Linear2Quadratic(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Linear2Quadratic";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command converts elements for an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshLinear2Quadratic.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_SplitHexes : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_SplitHexes(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "SplitHexes";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command converts hexahedrals to (5)tetrahedrals, (3)pyramids or (2)prisms for an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshSplitHexes.txt";}
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Refine : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Refine(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Refine";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command applies a refinement on an entire Mesh or a single Mesh component.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshRefine.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Rotate : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Rotate(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Rotate";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command rotates a subordinate mesh around a given axis ( 2D -> 3D )";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshRotate.txt";}
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Mirror : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Mirror(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Mirror";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command mirrors a subordinate mesh at a given plane ";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshMirror.txt";}
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_Extrude : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_Extrude(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Extrude";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command extrudes a subordinate mesh along a given axis ( 1D->2D, 2D->3D )";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the newly generated component (usually the same number as the input component number)";}
	virtual mystr GetCommandExampleFileName() const {return "MeshExtrude.txt";}
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_AddMeshToMBS : public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_AddMeshToMBS(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "AddMeshToMBS";}
	virtual int NParameters() const {return 1;}     // "()" gets resolved to 1 parameter...
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates an instance of the mesh in the MBS.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the Element Data conatiner of the added mesh (as shown in the Edit Mesh Menu).";}
	virtual mystr GetCommandExampleFileName() const {return "MeshAddMeshToMBS.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetNodesInBox: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetNodesInBox(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetNodesInBox";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of nodes (registered to the mesh) in a given box.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of node numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetNodesInBox.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetNodesInCylinderShell: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetNodesInCylinderShell(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetNodesInCylinder";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of nodes (registered to the mesh) in a given cylinder or cylinder shell.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of node numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetNodesInCylinder.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetNodesInSphericalShell: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetNodesInSphericalShell(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetNodesInSphere";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of nodes (registered to the mesh) in a given sphere or spherical shell.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of node numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetNodesInSphere.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetNodesInFunction: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetNodesInFunction(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetNodesInFunction";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of nodes (registered to the mesh) in a region where a given function f(x,y,z)>0 .";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of node numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetNodesInFunction.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetNodePos: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetNodePos(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetNodePos";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returnsthe global position of a given node.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a 3D position vector.";}
	virtual mystr GetCommandExampleFileName() const {return "GetNodePosition.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetFacesFromNodes: public CEDC_Command
{
public:
	CEDC_ComMesh_GetFacesFromNodes(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "GetFacesFromNodes";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of elements and face numbers (registered to the mesh) that can be built from nodes in the input.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a container with a list of element numbers and a list of face numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

class CEDC_ComMesh_GlueMesh: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComMesh_GlueMesh(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GlueMesh";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command glues mesh together.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the component number of the MeshGlue-Component (usually the lowest of the input component numbers";}
	virtual mystr GetCommandExampleFileName() const {return "MeshGlueMesh.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ AD 2013-10-09:]

class CEDC_GetLocalPosOfGlobalPos: public CEDC_Command
{
public:
	//initialization
	CEDC_GetLocalPosOfGlobalPos(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetLocalPosOfGlobalPos";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns the local position of a global position on a specified element.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a 3D vector.";}
	virtual mystr GetCommandExampleFileName() const {return "GetLocalPositionOfGlobalPosition.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetElementsInBox: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetElementsInBox(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetElementsInBox";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
/**/	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of elements (registered to the mesh) in a given box.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of element numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetElementsInBox.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComMesh_GetElementAtPos: public CEDC_Command
{
public:
	//initialization
	CEDC_ComMesh_GetElementAtPos(int id): CEDC_Command(id) {}
	
	virtual const char* CommandName() const {return "GetElementAtPosition";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
/**/	virtual int GetOptions() const {return TCO_RequiresHandle;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns a list of elements and local positions (registered to the mesh) at a given position.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a list of element numbers and a list of local positions numbers.";}
	virtual mystr GetCommandExampleFileName() const {return "GetElementAtPosition.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};




//$ AD 2017-03-21: functions to generate specific handle
class CEDC_ComPlot_GenerateNewPlot: public CEDC_Command
{
public:
	//initialization
	CEDC_ComPlot_GenerateNewPlot(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "GenerateNewPlot";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates a Handle to a PlotTool Object for further operations.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is a special EDC (Handle) that must be assigned to a new variable.";}
	virtual mystr GetCommandExampleFileName() const {return "PlotGenreatePlot.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};
//$ AD 2017-03-21:]
class CEDC_ComPlot_ExportPlot: public CEDC_Command
{
public:
	//initialization
	CEDC_ComPlot_ExportPlot(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "ExportToFile";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;} 

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command saves the content PlotTool Object for further operations.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the number of the generated files.";}
	virtual mystr GetCommandExampleFileName() const {return "ExportPlot.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

class CEDC_ComPlot_ClosePlot: public CEDC_Command
{
public:
	//initialization
	CEDC_ComPlot_ClosePlot(int id): CEDC_Command(id) {}

	virtual const char* CommandName() const {return "Close";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_RequiresHandle;} 

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command closes the content PlotTool Object.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "The return value is the filenabe of the generated file.";}
	virtual mystr GetCommandExampleFileName() const {return "ExportPlot.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};


//$ AD 2013-11-27: command to determine if a specific container exists
class CEDC_Check_Entry_Exists: public CEDC_Command
{
	public:
	//initialization
	CEDC_Check_Entry_Exists(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "DoesEntryExist";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command checks if the specified entry exists.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 if the entry does not exist, returns 1 if the entry exists, returns 2 if the entry is an EDC";}
	virtual mystr GetCommandExampleFileName() const {return "Container.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2015-03-19: command to get a container/value by its tree-name ( and assign it to a variable )
class CEDC_GetByName: public CEDC_Command
{
	public:
	//initialization
	CEDC_GetByName(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "GetByName";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command lets you get any (existing) EDC entry by name.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns the entry associated with the string.";}
	virtual mystr GetCommandExampleFileName() const {return "Container.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2015-03-19: command to set a container/value by its tree-name
class CEDC_SetByName: public CEDC_Command
{
	public:
	//initialization
	CEDC_SetByName(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "SetByName";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command lets you set any EDC entry by name, the name contains the absolute treename.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "has no return value.";}
	virtual mystr GetCommandExampleFileName() const {return "Container.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2013-12-13: command compare strings - could be used for other comparisons as well
class CEDC_Compare: public CEDC_Command
{
	public:
	//initialization
	CEDC_Compare(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "Compare";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command compares two strings.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 if both strings are identical, returns >0 or <0 otherwise indicating which string has higher value .";}
	virtual mystr GetCommandExampleFileName() const {return "strings.txt";}
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2013-12-13: command concatinate strings 
class CEDC_StrCat: public CEDC_Command
{
	public:
	//initialization
	CEDC_StrCat(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "StrCat";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command joins two strings together";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns a single string - strA+strB.";}
	virtual mystr GetCommandExampleFileName() const {return "strings.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2014-01-08: command zeros - creates a vector, matrix of a given dimension
class CEDC_Zeros: public CEDC_Command
{
	public:
	//initialization
	CEDC_Zeros(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "Zeros";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command sets a vector or matrix variable to agiven dimension and sets all entries to 0";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns a vector or matrix variable";}
	virtual mystr GetCommandExampleFileName() const {return "lists.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2017-03-07: command IntArrayOp - collection of operations with integer arrays
class CEDC_IntArrayOp: public CEDC_Command
{
	public:
	//initialization
	CEDC_IntArrayOp(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "IntArrayOp";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const { return "This command allows operations on up two integer arrays, length of an array may be one, some operations only require one array to operate on"; };
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const { return "returns a vector variable"; }
	virtual mystr GetCommandExampleFileName() const { return "ArrayOps.txt"; };
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2021-04-07: command StringOp - collection of operation with strings
class CEDC_StringOp : public CEDC_Command
{
public:
	//initialization
	CEDC_StringOp(int id) : CEDC_Command(id) {}
	virtual const char* CommandName() const { return "StringOp"; }
	virtual int NParameters() const { return 3; }
	virtual int HasReturnValue() const { return 1; }
	virtual int GetOptions() const { return TCO_None; }

	//description for the docu
	virtual mystr GetCommandTexDescription() const { return "This command allows operations on up two Strings, some operations only require one string to operate on"; };
	virtual mystr GetCommandTexParameterDescription() const;
	virtual mystr GetCommandTexReturnValueDescription() const { return "returns various variable types"; }
	virtual mystr GetCommandExampleFileName() const { return "StringOps.txt"; };
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};


//$ AD 2017-12-07: command Timer - system time in milliseconds
class CEDC_Timer: public CEDC_Command
{
	public:
	//initialization
	CEDC_Timer(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "Timer";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command returns the current system time in milliseconds";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns an integer variable";}
	virtual mystr GetCommandExampleFileName() const {return "Timer.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};

//$ AD 2017-03-07: command ArrayAdd - creates a vector, matrix of a given dimension
class CEDC_ComSet_AccessSet: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComSet_AccessSet(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "AccessSet";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}

	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command makes a set accessable for script by copying it to Model variables";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns the EDC of the set";}
	virtual mystr GetCommandExampleFileName() const {return "AccessSet.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD 2015-01-29: command GenerateConstraints(nodeset,constraintparameter) - generates constraints for the given set
class CEDC_ComSet_GenerateConstraints: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComSet_GenerateConstraints(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "GenerateConstraints";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates constraints for the given set";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns a list of element numbers for the generated constraints";}
	virtual mystr GetCommandExampleFileName() const {return "GenerateConstraints.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD 2016-02-03: command GenerateSensors(nodeset,sensorparameter) - genreates sensors for the given set
class CEDC_ComSet_GenerateSensors: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComSet_GenerateSensors(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "GenerateSensors";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command generates sensors for the given set";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns a list of sensor numbers for the generated sensors";}
	virtual mystr GetCommandExampleFileName() const {return "GenerateSensors.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

class CEDC_ComSetInitialCondition : public CEDC_Command
{
public:
	CEDC_ComSetInitialCondition(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "SetInitialCondition";}
	virtual int NParameters() const {return 3;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}

	virtual mystr GetCommandTexDescription() const {return "This command sets the initial condition of all members of the element or node set.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	//virtual mystr GetCommandTexReturnValueDescription() const {return "";}
	virtual mystr GetCommandExampleFileName() const {return "";};

	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};


//$ AD 2014-08-01: command AssignMaterial(elementset,material) - sets the material number of all the elements of hte set to the given number
class CEDC_ComAssignMaterial: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComAssignMaterial(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "AssignMaterial";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command sets the material number of all elemenets of the element-set to a given number";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "AssignMaterial.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD 2014-08-11: command AssignLoad(elementset,loadnumber) - assigns a(n additional) load to  all the elements of the set
class CEDC_ComAssignLoad: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComAssignLoad(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "AssignLoad";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command adds a load to all elemenets of the element-set";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "AssignLoad.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};


//$ DR 2015-02-17: command AssignGeomElementToElement(elemen_number,geomelment_number) - assigns a(n additional) geomelement to an elements
class CEDC_ComAssignGeomElementToElement: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComAssignGeomElementToElement(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "AssignGeomElementToElement";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command assigns a geom element to an element. The reference point and rotation of the element are evaluated and the settings of the geom element are modified automatically, such that the current relative orientation of element and geom element keeps the same. You can therefore add and align the geom element independently from the element first and afterwards decide to connect the geom element to the element without the need of changing the settings of the geom element again.";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "AssignGeomElementToElement.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD 2015-01-28: command ComChangeProperties(elementset||nodeset,substitute) - change any properties for all elements of the set
class CEDC_ComChangeProperties: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComChangeProperties(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "ChangeProperties";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command changes properties of the elements of the set";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "ChangeProperties.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD + SW 2016-1-12: sorts set "set" such that its elements match (position) the elements in set "order"
class CEDC_ComSet_MatchSets: public CEDC_Command
{
	public:
	//initialization
	CEDC_ComSet_MatchSets(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "MatchSets";}
	virtual int NParameters() const {return 2;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "this command sorts the second set such that ith element of both sets are at the same position";};
	virtual mystr GetCommandTexParameterDescription() const {return "";};
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns nothing";}
	virtual mystr GetCommandExampleFileName() const {return "MatchSets.txt";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

class CEDC_DeleteElement: public CEDC_Command
{
public:
	public:
	//initialization
	CEDC_DeleteElement(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "DeleteElement";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command deletes an element from the MBS";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

class CEDC_DeleteNode: public CEDC_Command
{
public:
	public:
	//initialization
	CEDC_DeleteNode(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "DeleteNode";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 0;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command deletes a node from the MBS";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 or an error code";}
	virtual mystr GetCommandExampleFileName() const {return "";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
};

//$ AD 2017-04-03: command OpenModel - opens a compiled model
class CEDC_OpenCompiledModel: public CEDC_Command
{
	public:
	//initialization
	CEDC_OpenCompiledModel(int id): CEDC_Command(id) {}
	virtual const char* CommandName() const {return "OpenCompiledModel";}
	virtual int NParameters() const {return 1;}
	virtual int HasReturnValue() const {return 1;}
	virtual int GetOptions() const {return TCO_None;}
	
	//description for the docu
	virtual mystr GetCommandTexDescription() const {return "This command loads a model from the compiled models";};
	virtual mystr GetCommandTexParameterDescription() const ;
	virtual mystr GetCommandTexReturnValueDescription() const {return "returns 0 for fail, 1 for successful load";}
	virtual mystr GetCommandExampleFileName() const {return "";};
	//execute command, 1==success, 0==failed
	//it is already checked, that the number of parameters is equal to the desired number of parameters
	virtual int ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);
private:
};


class CEDCParser 
{
public:
	//initialize parser
	//CEDCParser(MBS * mbs);
	CEDCParser(): commands(), class_names(), class_descriptions(), hid_version_number(0,0,0), hotint_version_number(0,0,0)
	{
 		mbs = NULL;
		Initialize();
	}

	virtual ~CEDCParser()
	{
		for(int i=1; i<=commands.Length(); i++)
		{
			delete commands(i);
		}
	}

	//$ YV 2012-12-30: expression parser is also stored as a member here
	virtual void SetMBS(MBS * mbs, CMBSParser * mbsParser)
	{
		this->mbs = mbs;
		this->mbsParser = mbsParser;
	}
	virtual const MBS * GetMBS() const {return mbs;}
	virtual MBS* GetMBS() {return mbs;}

	virtual void Initialize();

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//auxiliary functions:
	virtual int ReadComment(mystr& str, int& pos, ElementData& ed, const mystr& dataname); //read comment at end of statement into ElementData
	virtual int ReadBoolIntDoubleExpression(mystr& text, int& pos, ElementData& ed, const mystr& elemname); //read expression containing bool (yes/no), int/double or a mathematical expression
	virtual int ReadMatrixVector(const mystr& text, ElementData& ed, const mystr& elemname, const mystr& dataname); //read structure in [] containing matrix or vector
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	virtual void EDCError(const mystr& str) {mbs->UO(UO_LVL_err) << "EDC Parser error: '" << str << "' at line " << line_cnt << "\n";}
	virtual void EDCWarning(const mystr& str) {mbs->UO(UO_LVL_err) << "EDC Parser warning: '" << str << "' at line " << line_cnt << "\n";}
	virtual void EDCDbgMsg(const mystr& str) {mbs->UO(UO_LVL_dbg1) << "EDC Parser message: '" << str << "' at line " << line_cnt << "\n";}

	virtual int MakeFilenameAbsolute(mystr& filename);	// checks if filename has relative path, and if yes, converts it into an absolute one. returns 1 if success
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//commands, set up all possible commands:
	virtual void InitializeCommands() 
	{
		//add all EDC-commands manually here!
		int comcnt = 0;

		for(int i=1; i<=commands.Length(); i++) {delete commands(i);}	//$ DR 2012-10: clear array, because this function is called several times
		commands.Flush();

		CEDC_ComAddElement* cae = new CEDC_ComAddElement(++comcnt);
		commands.Add(cae);

		CEDC_ComAddGeomElement* cage = new CEDC_ComAddGeomElement(++comcnt);
		commands.Add(cage);

		CEDC_ComAssignGeomElementToElement* asge = new CEDC_ComAssignGeomElementToElement(++comcnt);	//$ DR 2015-02-17
		commands.Add(asge);

		CEDC_ComAddConstraint* cac = new CEDC_ComAddConstraint(++comcnt);		//$ DR 2012-08
		commands.Add(cac);

		CEDC_ComAddLoad* cal = new CEDC_ComAddLoad(++comcnt);		//$ DR 2012-10
		commands.Add(cal);

		CEDC_ComAddSensor* cas = new CEDC_ComAddSensor(++comcnt);		//$ DR 2012-10
		commands.Add(cas);

		CEDC_ComAddSensorProcessor* casp = new CEDC_ComAddSensorProcessor(++comcnt);		//$ AD 2017-06
		commands.Add(casp);

		CEDC_ComAddMaterial* cam = new CEDC_ComAddMaterial(++comcnt);		//$ DR+PG 2012-12
		commands.Add(cam);

		CEDC_ComAddBeam3DProperties* cabp = new CEDC_ComAddBeam3DProperties(++comcnt);		//$ DR+PG 2012-12
		commands.Add(cabp);

		CEDC_ComAddNode* can = new CEDC_ComAddNode(++comcnt);			//$ DR 2013-01
		commands.Add(can);

		//CEDC_ComAddSet* caset = new CEDC_ComAddSet(++comcnt);			//$ AD 2014-08 //$ SW 2015-4-24: removed since defined twice (see ~100 lines below)
		//commands.Add(caset);

		CEDC_ComInclude* caincl = new CEDC_ComInclude(++comcnt);
		commands.Add(caincl);

		CEDC_ComPrint* cpr = new CEDC_ComPrint(++comcnt);		//$ DR 2012-11
		commands.Add(cpr);
		
		CEDC_ComPrintIf* cprif = new CEDC_ComPrintIf(++comcnt);		//$ AD 2017-09
		commands.Add(cprif);

		CEDC_ComLoadSTL* clstl = new CEDC_ComLoadSTL(++comcnt);		//$ DR 2012-12
		commands.Add(clstl);

		CEDC_ComRotMatToAngles* crm2a = new CEDC_ComRotMatToAngles(++comcnt);       //$ AD-2017-08
		commands.Add(crm2a);

		CEDC_ComRotMatFromAngles* cra2m = new CEDC_ComRotMatFromAngles(++comcnt);       //$ PG-2020-05
		commands.Add(cra2m);

		CEDC_ComLoadVecFromFile* clvff = new CEDC_ComLoadVecFromFile(++comcnt);			//$ DR 2013-07
		commands.Add(clvff);

		CEDC_ComTransformPoints* ctp = new CEDC_ComTransformPoints(++comcnt);		//$ DR 2013-02-19
		commands.Add(ctp);
		
		CEDC_ComComputeInertia* cci = new CEDC_ComComputeInertia(++comcnt);			//$ DR 2013-01
		commands.Add(cci);

		CEDC_ComSum* cs = new CEDC_ComSum(++comcnt);			//$ AD 2013-07
		commands.Add(cs);

		CEDC_ComProduct* cp = new CEDC_ComProduct(++comcnt);    //$ AD 2013-07
		commands.Add(cp);

		CEDC_ComTranspose* ct = new CEDC_ComTranspose(++comcnt);    //$ AD 2013-07
		commands.Add(ct);

		//CEDC_ComScalarProduct* csp = new CEDC_ComScalarProduct(++comcnt);    //$ AD 2013-07
		//commands.Add(csp);
		
		CEDC_ComCrossProduct* ccp = new CEDC_ComCrossProduct(++comcnt);    //$ AD 2013-07
		commands.Add(ccp);

		CEDC_ComFor* cfor = new CEDC_ComFor(++comcnt);    //$ AD 2013-09
		commands.Add(cfor);
		
		CEDC_ComIf* cif = new CEDC_ComIf(++comcnt);    //$ AD 2013-09
		commands.Add(cif);

//++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH SCRIPT COMMANDS
//#ifndef __RELEASE_VERSION__
		CEDC_ComMesh_GenerateNewMesh* cmgmesh = new CEDC_ComMesh_GenerateNewMesh(++comcnt);    //$ AD 2013-09
		commands.Add(cmgmesh);

		CEDC_ComMesh_GenerateBeam* cmgbeam = new CEDC_ComMesh_GenerateBeam(++comcnt);    //$ AD 2013-09
		commands.Add(cmgbeam);

		CEDC_ComMesh_GeneratePlate* cmgplate = new CEDC_ComMesh_GeneratePlate(++comcnt);    //$ AD 2013-10
		commands.Add(cmgplate);

		CEDC_ComMesh_GenerateBlock* cmgblock = new CEDC_ComMesh_GenerateBlock(++comcnt);		//$ AD 2014-09
		commands.Add(cmgblock);

		CEDC_ComMesh_GenerateCylinder* cmgcyl = new CEDC_ComMesh_GenerateCylinder(++comcnt);		//$ AD 2014-11
		commands.Add(cmgcyl);

		CEDC_ComMesh_LoadMesh* cmload = new CEDC_ComMesh_LoadMesh(++comcnt);	//$ AD 2015-01
		commands.Add(cmload);

		CEDC_ComMesh_WriteMesh* cmwrite = new CEDC_ComMesh_WriteMesh(++comcnt);	//$ AD 2016-07
		commands.Add(cmwrite);
/*
		CEDC_ComMesh_Process* cmprocess = new CEDC_ComMesh_Process(++comcnt); //$ AD 2015-01
		commands.Add(cmprocess);
*/
		CEDC_ComMesh_Transform* cmtrans = new CEDC_ComMesh_Transform(++comcnt); //$ AD 2014-10
		commands.Add(cmtrans);

		CEDC_ComMesh_Distort* cmdist = new CEDC_ComMesh_Distort(++comcnt); //$ AD 2015-06
		commands.Add(cmdist);

		CEDC_ComMesh_Modify* cmmod = new CEDC_ComMesh_Modify(++comcnt); //$ AD 2017-01
		commands.Add(cmmod);

		CEDC_ComMesh_Linear2Quadratic* cml2q = new CEDC_ComMesh_Linear2Quadratic(++comcnt); //$ AD 2015-01
		commands.Add(cml2q);

		CEDC_ComMesh_SplitHexes* cmsplit = new CEDC_ComMesh_SplitHexes(++comcnt); //$ AD 2015-01
		commands.Add(cmsplit);

		CEDC_ComMesh_Refine* cmref = new CEDC_ComMesh_Refine(++comcnt); //$ AD 2015-07
		commands.Add(cmref);
		
		CEDC_ComMesh_Rotate* cmrotate = new CEDC_ComMesh_Rotate(++comcnt); //$ AD 2015-01
		commands.Add(cmrotate);

		CEDC_ComMesh_Mirror* cmmirror = new CEDC_ComMesh_Mirror(++comcnt); //$ AD 2015-06
		commands.Add(cmmirror);

		CEDC_ComMesh_Extrude* cmextrude = new CEDC_ComMesh_Extrude(++comcnt); //$ AD 2015-01
		commands.Add(cmextrude);

		CEDC_ComMesh_AddMeshToMBS* cmam2m = new CEDC_ComMesh_AddMeshToMBS(++comcnt);		//$ AD 2014-09
		commands.Add(cmam2m);

		CEDC_ComMesh_GetNodesInBox* cmgetnodesbox = new CEDC_ComMesh_GetNodesInBox(++comcnt);    //$ AD 2013-10
		commands.Add(cmgetnodesbox);

		CEDC_ComMesh_GetNodesInCylinderShell* cmgetnodescyl = new CEDC_ComMesh_GetNodesInCylinderShell(++comcnt);    //$ AD 2017-04
		commands.Add(cmgetnodescyl);

		CEDC_ComMesh_GetNodesInSphericalShell* cmgetnodessph = new CEDC_ComMesh_GetNodesInSphericalShell(++comcnt);    //$ AD 2017-04
		commands.Add(cmgetnodessph);

		CEDC_ComMesh_GetNodesInFunction* cmgetnodesfctn = new CEDC_ComMesh_GetNodesInFunction(++comcnt);    //$ AD 2017-04
		commands.Add(cmgetnodesfctn);

		CEDC_ComMesh_GetNodePos* cmgetnodepos = new CEDC_ComMesh_GetNodePos(++comcnt);    //$ AD 2017-05
		commands.Add(cmgetnodepos);

		CEDC_ComMesh_GetFacesFromNodes* cmgetfacesfromnodes = new CEDC_ComMesh_GetFacesFromNodes(++comcnt);     //$ AD 2015-12
		commands.Add(cmgetfacesfromnodes);

		CEDC_ComMesh_GlueMesh* cmgluemesh = new CEDC_ComMesh_GlueMesh(++comcnt);    //$ AD 2013-10
		commands.Add(cmgluemesh);
//#endif
//++++++++++++++++++++++++++++++++++++++++++++++++++++
		
		CEDC_GetLocalPosOfGlobalPos* clocofglob = new CEDC_GetLocalPosOfGlobalPos(++comcnt);    //$ AD 2017-05
		commands.Add(clocofglob);

		CEDC_ComMesh_GetElementsInBox* cmgetelements = new CEDC_ComMesh_GetElementsInBox(++comcnt);    //$ AD 2013-10
		commands.Add(cmgetelements);

		CEDC_ComMesh_GetElementAtPos* cmgetelatpos = new CEDC_ComMesh_GetElementAtPos(++comcnt);    //$ AD 2014-05
		commands.Add(cmgetelatpos);



//++++++++++++++++++++++++++++++++++++++++++++++++++++
// PLOT TOOL COMMANDS
// commands to work with PlotTool instances via Script
		CEDC_ComPlot_GenerateNewPlot* cmgplot = new CEDC_ComPlot_GenerateNewPlot(++comcnt);
		commands.Add(cmgplot);

		CEDC_ComPlot_ExportPlot* cmxplot = new CEDC_ComPlot_ExportPlot(++comcnt);
		commands.Add(cmxplot);

		CEDC_ComPlot_ClosePlot* cm0plot = new CEDC_ComPlot_ClosePlot(++comcnt);
		commands.Add(cm0plot);



		CEDC_Check_Entry_Exists* cmentryexists = new CEDC_Check_Entry_Exists(++comcnt);    //$ AD 2013-11
		commands.Add(cmentryexists);

		CEDC_GetByName* cmgetbyname = new CEDC_GetByName(++comcnt); //$ AD 2015-03
		commands.Add(cmgetbyname);

		CEDC_SetByName* cmsetbyname = new CEDC_SetByName(++comcnt); //$ AD 2015-03
		commands.Add(cmsetbyname);

		CEDC_Compare* cmcompare = new CEDC_Compare(++comcnt);    //$ AD 2013-12
		commands.Add(cmcompare);

		CEDC_StrCat* cmstrcat = new CEDC_StrCat(++comcnt);    //$ AD 2013-12
		commands.Add(cmstrcat);

		CEDC_Zeros* cmzeros = new CEDC_Zeros(++comcnt);			//$ AD 2014-01
		commands.Add(cmzeros);

		CEDC_IntArrayOp* cmarrop = new CEDC_IntArrayOp(++comcnt);    //$ AD 2017-03
		commands.Add(cmarrop);

		CEDC_StringOp* cmstrop = new CEDC_StringOp(++comcnt);    //$ AD 2021-04
		commands.Add(cmstrop);

		CEDC_Timer* cmtimer = new CEDC_Timer(++comcnt);  //$ AD 2017-12
		commands.Add(cmtimer);

//++++++++++++++++++++++++++++++++++++++++++++++++++++
// SET SCRIPT COMMANDS
// commands to create/manipulate SETS //$ AD 2014-08 ++
		CEDC_ComAddSet* cmaddset = new CEDC_ComAddSet(++comcnt);
		commands.Add(cmaddset);

		CEDC_ComSet_AccessSet* cmaccset = new CEDC_ComSet_AccessSet(++comcnt);
		commands.Add(cmaccset);

		CEDC_ComSet_GenerateConstraints* cmgencon = new CEDC_ComSet_GenerateConstraints(++comcnt);
		commands.Add(cmgencon);

		CEDC_ComSet_GenerateSensors* cmgensens = new CEDC_ComSet_GenerateSensors(++comcnt);
		commands.Add(cmgensens);

// commands with SETS as parameter //$ AD 2014-08 ++
		CEDC_ComAssignMaterial* cmassmat = new CEDC_ComAssignMaterial(++comcnt);
		commands.Add(cmassmat);

		CEDC_ComAssignLoad* cmassloa = new CEDC_ComAssignLoad(++comcnt);
		commands.Add(cmassloa);

		CEDC_ComChangeProperties* cmchprop = new CEDC_ComChangeProperties(++comcnt);
		commands.Add(cmchprop);

		CEDC_ComSetInitialCondition* csetinit = new CEDC_ComSetInitialCondition(++comcnt);
		commands.Add(csetinit);

#ifndef __RELEASE_VERSION__
		CEDC_ComSet_MatchSets* cmmatchsets = new CEDC_ComSet_MatchSets(++comcnt);
		commands.Add(cmmatchsets);

// command to remove an element from the MBS via Script
		CEDC_DeleteElement* cmdel = new CEDC_DeleteElement(++comcnt);
		commands.Add(cmdel);

		CEDC_DeleteNode* cmdno = new CEDC_DeleteNode(++comcnt);
		commands.Add(cmdno);
#endif
// command to open a ompiled model via script
		CEDC_OpenCompiledModel* cmopen = new CEDC_OpenCompiledModel(++comcnt);
		commands.Add(cmopen);

//$ AD 03-04-2017

	};

	//get the command type_id from the command_name; also provide the number of parameters and optional flags for the command
	//return 0, if failed
	virtual int GetCommandType(const mystr& command_name, int& n_parameters, int& has_return_value, int& option)
	{
		n_parameters = 0;
		has_return_value = 0;
		option = 0;
		for (int i=1; i <= commands.Length(); i++)
		{
			if (mystr(commands(i)->CommandName()) == command_name) //==>STringCMP would be more efficient
			{
				n_parameters = commands(i)->NParameters();
				has_return_value = commands(i)->HasReturnValue();
				option = commands(i)->GetOptions();
				return i;
			}
		}
		return 0;
	}

	virtual mystr GetCommandTypeName(int i) const {return mystr(commands(i)->CommandName());}
	virtual int NCommands() const {return commands.Length();}
	virtual CEDC_Command* GetCommand(int i) const {return commands(i);}

// AD 2013-10-16: single call for commands - collect code from String2TreeEDC and ReadBoolIntDoubleExpression - returns "com_id" (including 0 for no command) or "-1" on error 
	virtual int IdentifyAndExecuteCommand(mystr& dataname, mystr& str, int& pos, ElementData& return_value, int flag_RHS = 0);

// AD 2013-10-16: identifies command type and checks restrictions - returns "com_id" (including 0 for no command) or "-1" on error
	virtual int IdentifyCommand(mystr& dataname, int& n_params, int& has_return_value, int& option, int flag_RHS );

// AD 2013-10-16: parses parameters into EDCs - returns "1" on success and "0" on error
	virtual int ParseCommandParameter(TArray<ElementDataContainer*>& parameter_EDCs, mystr& dataname, int n_params, mystr& str, int& pos, int option);

//execute a command with type_id, a list of EDCs for command parameters and an option 
	virtual int ExecuteCommand(int command_type_id, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);

// AD 2013-10-16: perform a check for the datafile version (once)	
	virtual int HotintDataFileVersionCheck(int& version_check_performed, ElementDataContainer& edc_version);




	//get TArray<ElementDataContainer*> of comma-separated parameter sets for a command
	virtual int GetParameterEDC(mystr text, TArray<ElementDataContainer*>& parameter_EDCs);
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	//main EDC parsing functions:
	//transform edc into string, works already for TreeEDCs!
	virtual void EDC2String(const ElementDataContainer& edc, mystr& str, const mystr& indent = "");
	//virtual void EDC2TexTable(const ElementDataContainer& edc, mystr& str, const mystr& indent = "", const mystr& treename = ""); //$ DR 2013-01-11 moved to models_auto_documentation

	//transforms string into edc until a keyword is read, returns remaining string:
	virtual int String2EDC(/*MBS * mbs, */mystr& str, ElementDataContainer& edc, ElementDataContainer& testedc); 

	////////transforms a string automatically into a tree edc; only accepts: double, string (""), vector, matrix
	//////virtual int String2TreeEDC(/*MBS * mbs, */mystr str, ElementDataContainer& edc); 
	////////recursive version of string to TreeEDC conversion
	//////virtual int String2TreeEDC_recursive(mystr str, ElementDataContainer& edc);

  // AD 2013-09-24: these functions parses a string ans applies it ENTRY BY ENTRY to the target element data container
	virtual int ParseAndExecuteString(mystr str, ElementDataContainer& edc); 
	virtual int ParseAndExecuteString_recursive(mystr str, ElementDataContainer& edc);
	//virtual int AddReplaceSingleEntry(ElementData& entryED, ElementDataContainer& targetEDC);

	virtual Default_ParserObjectValues& DPOV() {return dpov;}
	virtual const Default_ParserObjectValues& DPOV() const {return dpov;}
	
	// DR 2013-01-14: this function evaluates a command of the parser, using the parameter_EDC as additional input data 
	// AD 2013-10-16: remark: can be called from .cpp model for a) testing new script function, b) use existing script routines
	virtual int ExecuteParserCommand(mystr & command, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option);

	// +++++++++++++ the classs descriptions are not used at the moment +++++++++++++++++++++++	DR 2013-01-11
	//virtual void AddClassDescriptions_Auto();
	//add a class name with a description to the parser
	virtual void AddClassDescription(const mystr& name, const mystr& description)
	{
		class_names.Add(name);
		class_descriptions.Add(description);
	}
	//get a description for a certain class name
	virtual mystr GetClassDescription(const mystr& name)
	{
		for (int i=1; i <= class_names.Length(); i++)
		{
			if (name == mystr(class_names(i))) return mystr(class_descriptions(i));
		}
		return "";
	}
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	virtual int PrintOutput() {
		if (mbs)
		{
			return mbs->GetOptions()->LoggingOptions()->EDCParserGeneralInformation(); //$ DR 2015-07-01 simplification of code
		}
		return 0;

 		//int val = oo;
 		//if (mbs)
 		//{
 		//	if (mbs->GetOptions()->LoggingOptions()->EDCParserGeneralInformation() == 1)
 		//	{
 		//		val = 1;
 		//	}
 		//	else
 		//	{
 		//		val = 0;
 		//	}
 		//}

 		//return val;
 	}

public:
	MBSObjectFactoryInterface * GetObjectFactory() { return pObjectFactory; }
	void SetObjectFactory(MBSObjectFactoryInterface * pObjectFactory) { this->pObjectFactory = pObjectFactory; }

	// Writes the new name and path of the EDC variable in old_name_and_path into new_name and new_path. If the variable name has not changed,
	// new_path = "" and new_name = ""
	static void GetNewEDCName(const mystr& element_specification,  const mystr& old_name_and_path, mystr& new_name, mystr& new_path);

public: 
	CMBSParser* GetMBSParser() { return mbsParser; }

private:
	MBS * mbs;
	CMBSParser * mbsParser;
	MBSObjectFactoryInterface * pObjectFactory;		// a pointer to the instance in the models and elements module

	Default_ParserObjectValues dpov; //default parser object values

  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//commands:
	TArray<CEDC_Command*> commands;
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	//$ DR 2013-01-11 not used at the moment:
	MyStrList class_names, class_descriptions; //this are two lists, which provide a detailed .tex based description for a single class

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//definitions of characters for parsing:
	mystr el; //end of line
	char elc;
	char obc;  //open brace, for sub structure
	char cbc;  //closing brace
	mystr ob;
	mystr cb;
	char ofc;  //open function (round '(') bracket
	char cfc;  //close function (round ')') bracket
	mystr of;
	mystr cf;
	char osbc;  //open square bracket, for vector or matrix
	char csbc;  //closing square bracket
	mystr osb;  //open square bracket, for vector or matrix
	mystr csb;  //closing square bracket
	mystr ds;	//double stop - not used any more
	mystr eq;		//equal sign
	mystr semicolon; //double stop
	char semicolonc; //double stop
	char commac; //comma
	mystr comma; //comma
	char doublequote; //for strings or text
	char dotc; //dot
	mystr dot; //dot
	mystr space; //space
	char spacec;
	char tabc;
	mystr tab;
	mystr bool_yes;
	mystr bool_no;

	char commentc; //comment
	mystr comment; //comment
	mystr matrixsep; //may be changed to semicolon ...

  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//other variables and switches:
	int line_cnt; //line counter in EDC
	int total_lines; //total number of lines in string
	//int oo; //turn on / off detailed output //$ DR 2015-07-01 removed for simplification of code

	int version_check_performed;    // flag to perform a version check when first command is encountered
	HotintVersionInfo hid_version_number; // hotint version of hid file //$ SW 2014-8-11: added
	HotintVersionInfo hotint_version_number; // hotint version //$ SW 2014-8-11: added
	void LoadError(const mystr& str) {mbs->UO() << "Error in load file: " << str.c_str() << "\n";}
};
#endif
