//#**************************************************************
//#
//# filename:             MBSload.cpp
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						17.October 2004
//# description:          Driver and model for timeintegration
//#                       Model of a rigid arm and a hydraulic zylinder
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#include "element.h"
#include "mbsload.h"
#include "elementdataaccess.h"
#include "constraint.h"
#include "sensors.h"
#include "control.h"
#include "body3d.h"
#include "body2d.h"
#include "mass1d.h"
#include "femathhelperfunctions.h" //Quicksort
#include "FiniteElement3D.h"
#include "FiniteElement2D.h"


MBSLoad::MBSLoad(): /*data(),*/ steps()
{
	Reset();
};

void MBSLoad::Reset()
{
	val=0; gc=0; loadtype=TGCload; specpos = 0; /*el=NULL; */loadfunc = TLoadValue;
	forcepos = Vector3D(0,0,0); vecval = Vector3D(0,0,0);
	IOelement = 0; IOelement_outputnum = 0; //loadsteps initialized by 0 by default
	mathfunc.Init();
	node_1 = 0; node_2 = 0; node_f = 0; // DR  initialize by 0
	dim = 0;	//$ DR 2012-10
	SetLoadName("Load");//$ DR 2012-10
	localBodyAxis = 1; // SM 2012-01-11
};

mystr MBSLoad::LoadName() const 
{
	return loadname;
}

mystr MBSLoad::GetLoadTypeString() const
{
	switch(LoadType()) 
	{
		//the following names MUST coincide with the type names defined in MBS::MBS() !!!!
	case TGCload:						return "GCLoad";				break;
	case Tbodyload:					return "BodyLoad";			break;
	case Tsurfaceload:			return "SurfaceLoad";		break;
	case Tgravity:					return "Gravity";				break;
	case Tforcevector3D:		return "ForceVector3D"; break; //$ DR 2012-10
	case Tforcevector2D:		return "ForceVector2D"; break; //$ DR 2012-10
		//case Tpointload: 
		//	{
		//		if (el->Dim() == 3)	return "ForceVector3D"; 
		//		else	return "ForceVector2D"; 
		//		break;
		//	}
	case Tmomentvector3D:		return "MomentVector3D"; break; //$ DR 2012-10
	case Tmomentvector2D:		return "Moment2D";			break; //$ DR 2012-10
		//case Tpointmoment: 
		//	{
		//		if (el->Dim() == 3)	return "MomentVector3D"; 
		//		else	return "Moment2D"; 
		//		break;
		//	}
	case Tsurfacepressure:	return "SurfacePressure"; break;
	case Tcentrifugal:			return "CentrifugalLoad";		break;
	case Tbodyloadspatial:	return "BodyLoadSpatial";		break;
	case Tsurfaceheatsource: return "SurfaceHeatSource";
	case Tvolumeheatsource: return "VolumeHeatSource";
	case Tareaload: return "AreaLoad"; break;
	case Tareatraction: return "AreaTraction"; break;
	case Tsufaceload: return "SurfaceLoad"; break;
	default:								return "Load";
	}
}

//$ DR 2012-10
// documentation of the LoadTypes, used for autogenerated docu
mystr MBSLoad::GetLoadTypeTexDescription() const
{
	mystr descr="";
	switch(LoadType()) 
	{
		//the following names MUST coincide with the type names defined in MBS::MBS() !!!!
	case TGCload:						
		descr = "A load acting on a generalized coordinate (gc) of the element.";
		break;
	case Tbodyload:			
		descr = "The load value is integrated over the volume of the body and applied to the body in the specified direction. For the case of a rigid body, a force of size load\\_value = density*gravity\\_constant applies a force according to the gravitational force.";
		break;
	case Tsurfaceload:				
		descr = "";	// description is still missing
		break;
	case Tgravity:							
		descr = "The load is integrated over the volume of the body and applied to the body in the specified direction. The density of the body is used to compute the force.";
		break;
	case Tforcevector3D:				
		descr = "A load acting on an element at a specified (local) position.";	
		break;
	case Tforcevector2D:					
		descr = "";	// description is still missing
		break;
	case Tmomentvector3D:					
		descr = "A torque acting on an element at a specified (local) position. ";
		break;
	case Tmomentvector2D:				
		descr = "";	// description is still missing
		break;
	case Tsurfacepressure:				
		descr = "";	// description is still missing
		break;
	case Tcentrifugal:				
		descr = "";	// description is still missing
		break;
	}

	return descr;
}

void MBSLoad::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Vector3D force;
	Vector3D pos;
	Vector2D force2;
	Vector2D pos2;
	
	ElementData ed;

	mystr eleminfo = LoadName();
	
	ed.SetText(LoadName(), "name"); ed.SetToolTipText("name of the load"); edc.Add(ed);
	ed.SetText(GetLoadTypeString(), "load_type"); ed.SetToolTipText("specification of load type. Once the load is added to the mbs, you MUST NOT change this type anymore!"); edc.Add(ed);
	ed.SetInt(GetOwnNum(), "load_number"); ed.SetToolTipText("number of the load in the mbs"); ed.SetLocked(1); edc.Add(ed);

	switch(LoadType())
	{
	case TGCload: 
		{
			double val;
			int gc;
			GetGCLoad(val, gc);
			ed.SetInt(gc, "generalized_coordinate");	ed.SetToolTipText("(local) number of the generalized coordinate"); edc.Add(ed);
			ed.SetDouble(val, "load_value");					ed.SetToolTipText("value of the load acting in the direction of generalized_coordinate"); edc.Add(ed);
			break;
		}
	case Tbodyload: 
		{
			double val;
			int dir;
			GetBodyLoad(val, dir);
			ed.SetInt(dir, "direction", 1,Dim(),1,0);			ed.SetToolTipText("direction of the load"); edc.Add(ed);//$ DR 2011-07-21 removed upper limit Dim() because it is not set yet, when it is checked
			ed.SetDouble(val, "load_value");					ed.SetToolTipText("value of the load acting"); edc.Add(ed);
			break;
		}
	case Tforcevector2D: 
		{
			GetForceVector2D(force2, pos2);
			SetElemDataVector2D(edc, force2, "force_vector","defines the magnitude and direction of the force");
			SetElemDataVector2D(edc, pos2, "position","(local) position where the force is applied to the element");
			ed.SetInt(localBodyAxis, "local_force", 0,1,1,1);			ed.SetToolTipText("flag which describes, if local or global coordinate system is used: 1 = force in local body coordinate system, 0 = global force"); edc.Add(ed);//$ DR 2011-07-21 adjusted the flags for the limits
			break;
		}
	case Tforcevector3D: 
		{
			GetForceVector3D(force, pos);
			SetElemDataVector3D(edc, force, "force_vector","defines the magnitude and direction of the force");
			SetElemDataVector3D(edc, pos, "position","(local) position where the force is applied to the element");
			ed.SetInt(localBodyAxis, "local_force", 0,1,1,1);			ed.SetToolTipText("flag which describes, if local or global coordinate system is used: 1 = force in local body coordinate system, 0 = global force"); edc.Add(ed);//$ DR 2011-07-21 adjusted the flags for the limits
			break;
		}
	case Tmomentvector2D: 
		{
			double moment;
			GetMomentVector2D(moment, pos2);
			ed.SetDouble(val, "moment");							ed.SetToolTipText("magnitude of the moment");edc.Add(ed);
			SetElemDataVector2D(edc, pos2, "position","(local) position where the moment is applied to the element");
			break;
		}
	case Tmomentvector3D: 
		{
			GetMomentVector3D(force, pos);
			SetElemDataVector3D(edc, force, "moment_vector","defines the magnitude and direction of the moment");
			SetElemDataVector3D(edc, pos, "position","(local) position where the moment is applied to the element");
			ed.SetInt(localBodyAxis, "local_moment", 0,1,1,1);			ed.SetToolTipText("flag which describes, if local or global coordinate system is used: 1 = moment in local body coordinate system, 0 = global moment"); edc.Add(ed);//$ DR 2011-07-21 adjusted the flags for the limits
			break;
		}
		//case Tpointload: 
		//	{
		//		if (el->Dim() == 2)
		//		{
		//			GetForceVector2D(force2, pos2);
		//			SetElemDataVector2D(edc, force2, "Force_vector");
		//			SetElemDataVector2D(edc, pos2, "Position");
		//		}
		//		else
		//		{
		//			GetForceVector3D(force, pos);
		//			SetElemDataVector3D(edc, force, "Force_vector");
		//			SetElemDataVector3D(edc, pos, "Position");
		//		}
		//		break;
		//	}
		//case Tpointmoment: 
		//	{
		//		if (el->Dim() == 2)
		//		{
		//			double moment;
		//			GetMomentVector2D(moment, pos2);
		//			ed.SetDouble(val, "Moment"); edc.Add(ed);
		//			SetElemDataVector2D(edc, pos2, "Position");
		//		}
		//		else
		//		{
		//			GetMomentVector3D(force, pos);
		//			SetElemDataVector3D(edc, force, "Moment_vector");
		//			SetElemDataVector3D(edc, pos, "Position");
		//		}
		//		break;
		//	}
	case Tgravity: 
		{
			double val;
			int dir;
			GetGravity(val, dir);
			ed.SetInt(dir, "direction", 1, Dim(),1,0);		ed.SetToolTipText("global direction of the gravity");edc.Add(ed);//$ DR 2011-07-21 removed upper limit Dim() because it is not set yet, when it is checked
			ed.SetDouble(val, "gravity_constant");		ed.SetToolTipText("use negative sign if necessary");edc.Add(ed);
			break;
		}
	case Tsurfaceload: 		
		{
			break;
		}
	case Tsurfacepressure: 
		{
			double val;
			int dir;
			GetSurfacePressure(val, dir);
			ed.SetInt(dir, "direction", 1, 2*Dim(),1,0);	ed.SetToolTipText("local surface (inner/outer)");		edc.Add(ed);//$ DR 2011-07-21 removed upper limit Dim() because it is not set yet, when it is checked
			ed.SetDouble(val, "surface_pressure");	ed.SetToolTipText("use negative sign if necessary");	edc.Add(ed);
			break;
		}
	case Tsurfaceheatsource: 
		{
			double val;
			int faceNr;
			GetSurfaceHeatSource(val, faceNr);
			ed.SetInt(faceNr, "face", 1, 6,1,1);	ed.SetToolTipText("face number");		edc.Add(ed);
			ed.SetDouble(val, "heat");	ed.SetToolTipText("power density per area");	edc.Add(ed);
			break;
		}
	case Tvolumeheatsource: 
		{
			double val;
			GetVolumeHeatSource(val);
			ed.SetDouble(val, "heat");	ed.SetToolTipText("power density per volume");	edc.Add(ed);
			break;
		}
	case Tcentrifugal:
		{
			GetCentrifugal(force,pos);
			SetElemDataVector3D(edc, force, "angular_velocity_vector","direction is the rotation axis and magnitude the magnitude of the angular velocity");
			SetElemDataVector3D(edc, pos, "point_on_axis", "one point on the rotation axis");
			break;
		}
	default: ;
	}

	//$ DR 2012-10:[ changed all to MathFunction, old code
	//double omega = 0;
	//double phase = 0;
	//double ts = 0, te = 0;
	//Vector v(5);
	//mystr loadname = "constant";

	//if (loadfunc == 2)
	//{
	//	GetHarmonic(omega, phase);
	//	loadname = "harmonic";
	//}
	//else if (loadfunc == 3)
	//{
	//	GetTimeSpan(ts, te);
	//	loadname = "time_span";
	//}
	//else if (loadfunc == 4)
	//{
	//	GetTimeRamp(v(1), v(2), v(3), v(4), v(5));
	//	loadname = "time_ramp";
	//}
	//else if (loadfunc == 6)
	//{
	//  loadname = "math_func";

	//ed.SetText(loadname.c_str(), "Time_dependency"); ed.SetToolTipText("{constant | harmonic | time_span | time_ramp}"); edc.Add(ed);
	//SetElemDataVector2D(edc, Vector2D(omega, phase), "Harmonic"); edc.Get(edc.Length()).SetToolTipText("[omega, phase(rad)]");
	//SetElemDataVector2D(edc, Vector2D(ts, te), "Time_span"); edc.Get(edc.Length()).SetToolTipText("[t_start, t_end]");
	//ed.SetVector(v.GetVecPtr(), 5, "Time_ramp"); ed.SetToolTipText("[magnitude, ramp_start, ramp_end, ramp2_start=-1, ramp2_end=-1]"); edc.Add(ed);

	//$ DR 2012-10:] changed all to MathFunction, old code

	ed.SetInt(loadfunc, "load_function_type", 0,2,1,1);			ed.SetToolTipText("time dependency of the load: 0..constant, 1..MathFunction, 2..IOElement"); edc.Add(ed);//$ DR 2011-07-21 adjusted the flags for the limits
	
	// MathFunction
	ElementDataContainer edc_mf;
	mathfunc.GetElementData(edc_mf);
	ed.SetEDC(&edc_mf,"MathFunction"); ed.SetToolTipText("MathFunction is used, if load_function_type = 1"); edc.Add(ed);	

	// IOelement
	ElementDataContainer edc_io;
	ed.SetInt(IOelement, "input_element_number");							ed.SetToolTipText("number of IOElement in the mbs");  edc_io.Add(ed);
	ed.SetInt(IOelement_outputnum, "input_local_number");			ed.SetToolTipText("number of output of IOElement connected to this element");  edc_io.Add(ed);
	ed.SetEDC(&edc_io,"IOElement"); ed.SetToolTipText("IOElement is used, if load_function_type = 2"); edc.Add(ed);	

	//GetElementDataAuto(edc);
}

int MBSLoad::SetElementData(ElementDataContainer& edc) 		//update element data
{
	int rv = 1;

	Vector2D force2;
	Vector2D pos2;

	GetElemDataText(mbs,edc,"name",loadname,1);

	mystr type_old = GetLoadTypeString();
	mystr type_new = edc.TreeGetString("load_type");
	if(!type_new.Compare(type_old))
	{
		GetMBS()->UO().InstantMessageText("ERROR: You MUST NOT change the type of the load!");
		return 0;
	}

	switch(LoadType())
	{
	case TGCload: 
		{
			GetElemDataInt(mbs, edc, "generalized_coordinate", gc, 1);
			GetElemDataDouble(mbs, edc, "load_value", val, 1);
			break;
		}
	case Tbodyload: 
		{
			GetElemDataInt(mbs, edc, "direction", specpos, 1);
			GetElemDataDouble(mbs, edc, "load_value", val, 1);
			break;
		}
	case Tforcevector2D: 
		{
			GetElemDataVector2D(mbs, edc, "force_vector", force2, 1);
			GetElemDataVector2D(mbs, edc, "position", pos2, 1);
			GetElemDataInt(mbs, edc, "local_force", localBodyAxis, 1);
			SetForceVector2D(force2, pos2,localBodyAxis);
			break;
		}
	case Tforcevector3D: 
		{
			GetElemDataVector3D(mbs, edc, "force_vector", vecval, 1);
			GetElemDataVector3D(mbs, edc, "position", forcepos, 1);
			GetElemDataInt(mbs, edc, "local_force", localBodyAxis, 1);
			break;
		}
	case Tmomentvector2D: 
		{
			GetElemDataDouble(mbs, edc, "moment", val, 1);
			GetElemDataVector2D(mbs, edc, "position", pos2, 1);
			SetMomentVector2D(val, pos2);
			break;
		}
	case Tmomentvector3D: 
		{
			GetElemDataVector3D(mbs, edc, "moment_vector", vecval, 1);
			GetElemDataVector3D(mbs, edc, "position", forcepos, 1);
			GetElemDataInt(mbs, edc, "local_moment", localBodyAxis, 1);
			break;
		}
	case Tgravity: 
		{
			GetElemDataInt(mbs, edc, "direction", specpos, 1);
			GetElemDataDouble(mbs, edc, "gravity_constant", val, 1);
			break;
		}
	case Tsurfaceload: 		
		{
			break;
		}
	case Tsurfacepressure: 
		{
			GetElemDataInt(mbs, edc, "direction", specpos, 1);
			GetElemDataDouble(mbs, edc, "surface_pressure", val, 1);
			break;
		}
	case Tsurfaceheatsource: 
		{
			GetElemDataInt(mbs, edc, "face", specpos, 1);
			GetElemDataDouble(mbs, edc, "heat", val, 1);
			break;
		}
	case Tvolumeheatsource:
		{
			GetElemDataDouble(mbs, edc, "heat", val, 1);
			break;
		}
	case Tcentrifugal:
		{
			GetElemDataVector3D(mbs, edc, "angular_velocity_vector", vecval, 1);
			GetElemDataVector3D(mbs, edc, "point_on_axis", forcepos, 1);
			break;
		}
	default: ;
	}

	//$ DR 2012-10:[ changed all to MathFunction, old code
	//mystr str;
	//Vector2D v2;
	//Vector v5(5);

	//if (GetElemDataText(mbs, edc, "Time_dependency", str, 0))
	//{
	//	if (str == mystr("harmonic"))
	//	{
	//		if (GetElemDataVector2D(mbs, edc, "Harmonic", v2, 1))
	//		{
	//			SetHarmonic(v2(1), v2(2));
	//		}
	//	}
	//	else if (str == mystr("time_span"))
	//	{
	//		if (GetElemDataVector2D(mbs, edc, "Time_span", v2, 1))
	//		{
	//			SetTimeSpan(v2(1), v2(2));
	//		}
	//	}
	//	else if (str == mystr("time_ramp"))
	//	{
	//		if (GetElemDataVector(mbs, edc, "Time_ramp", v5, 1) && v5.Length() == 5)
	//		{
	//			SetTimeRamp(v5(1), v5(2), v5(3), v5(4), v5(5));
	//		}
	//	}
	//}
	//$ DR 2012-10:] changed all to MathFunction, old code

	int lf;
	GetElemDataInt(mbs,edc,"load_function_type",lf,1);
	loadfunc = (TMBSLoadFunc) lf;

	if(loadfunc == TLoadMathFunc)
	{
		int idx = edc.Find(mystr("MathFunction"));
		if(idx && (edc.Get(idx)).IsEDC())
		{
			//const ElementDataContainer* edcp_mf = (edc.Get(idx)).GetEDC();
			//const ElementDataContainer& edc_mf = (const ElementDataContainer&) *edcp_mf;
			//mathfunc.SetElementData(mbs,edc_mf);
			ElementDataContainer* edcp_mf = (edc.Get(idx)).GetEDC();
			ElementDataContainer edc_mf = *edcp_mf;
			mathfunc.SetElementData(mbs,edc_mf);	//$ DR 2012-12-12
			val = 1; //$ DR 2014-08-21 when MathFunction is combined with scalar value, only the MathFunction should define the value (no additional multiplication!)
		}
	}

	if(loadfunc == TLoadIOElement)
	{
			GetElemDataInt(mbs,edc,"IOElement.input_element_number",IOelement,1);
			GetElemDataInt(mbs,edc,"IOElement.input_local_number",IOelement_outputnum,1);
			val = 1; //$ DR 2014-08-21 when IOElement is combined with scalar value, only the IOElement should define the value (no additional multiplication!)
	}

	//SetElementDataAuto(edc);
	return rv;
}

int MBSLoad::CheckConsistency(Element& el, mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = 0;

	if(el.Dim()!=Dim())
	{
		errorstr += mystr("    The dimension of element ") + mystr(el.GetOwnNum()) + mystr(" and the applied load ") +mystr(GetOwnNum())+ mystr(" are different!\n");
		rv = Maximum(rv,1); 
	}

	switch(LoadType())
	{
	case TGCload: 
		{
			if(el.SS() < gc || 0 > gc)
			{
				errorstr += mystr("    The generalized coordinate of load ") +mystr(GetOwnNum()) + mystr(" is not fitting to the number of equations of element ") + mystr(el.GetOwnNum()) + mystr("!\n");
				rv = Maximum(rv,1); 
			}
			break;
		}
	case Tbodyload: 
		{
			break;
		}
	case Tforcevector2D: 
		{
			break;
		}
	case Tforcevector3D: 
		{
			break;
		}
	case Tmomentvector2D: 
		{
			break;
		}
	case Tmomentvector3D: 
		{
			break;
		}
	case Tgravity: 
		{
			break;
		}
	case Tsurfaceload: 		
		{
			break;
		}
	case Tsurfacepressure: 
		{
			break;
		}
	case Tcentrifugal:
		{
			break;
		}
	default: ;
	}

	return rv;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++        Get Force type and values       +++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


void MBSLoad::GetGCLoad(double& vali, int& gci)
{
	vali=val;
	gci =gc;
}

void MBSLoad::GetBodyLoad(double& vali, int& dir)
{
	vali=val; 
	dir = specpos;
}

void MBSLoad::GetGravity(double& gravityconstant, int& dir)
{
	gravityconstant=val; 
	dir = specpos;
}

void MBSLoad::GetForceVector3D(Vector3D & force, Vector3D& pos)
{
	//if(!localBodyAxis)
	//{
		force = vecval;
		pos = forcepos;
	//}
	//else // SM 2013-01-11: TODO --> not implemented for this case
	//{
	//	force = Vector3D(0.);
	//	pos = Vector3D(0.);
	//}
}

void MBSLoad::GetMomentVector3D(Vector3D & moment, Vector3D& pos)
{
	//if(!localBodyAxis)
	//{
		moment = vecval;
		pos = forcepos;
	//}
	//else // SM 2013-01-11: TODO --> not implemented for this case
	//{
	//	moment = Vector3D(0.);
	//	pos = Vector3D(0.);
	//}

}

void MBSLoad::GetForceVector2D(Vector2D & force, Vector2D& pos)
{
	//if(!localBodyAxis)
	//{
		force.X() = vecval.X();
		force.Y() = vecval.Y();
		pos.X() = forcepos.X();
		pos.Y() = forcepos.Y();
	//}
	//else // SM 2013-01-11: TODO --> not implemented for this case
	//{
	//	force.X() = 0;
	//	force.Y() = 0;
	//	pos.X() = 0;
	//	pos.Y() = 0;
	//}
}

void MBSLoad::GetMomentVector2D(double& moment, Vector2D& pos)
{
	//if(!localBodyAxis)
	//{
		moment = vecval.Z();
		pos.X() = forcepos.X();
		pos.Y() = forcepos.Y();
	//}
	//else // SM 2013-01-11: TODO --> not implemented for this case
	//{
	//	moment = 0;
	//	pos.X() = 0;
	//	pos.Y() = 0;
	//}
}

void MBSLoad::GetCentrifugal(Vector3D & omega, Vector3D& pos)
{
	omega = vecval;
	pos = forcepos;
}

void MBSLoad::GetContactZ(double& zpos)
{
	zpos = val ;
}

void MBSLoad::GetSurfacePressure(double& pressure, int& surfacenum)
{
	pressure = val; 
	surfacenum = specpos;
}

void MBSLoad::GetRotatingForce(double& vali, int& node_force, int& node_ax1, int& node_ax2)
{
	vali = val;
	node_force = node_f;
	node_ax1 = node_1;
	node_ax2 = node_2;

	//$ DR 2011-04-14: old code:
	// specpos is used to store node number of node where force is applied
	//node_force = specpos;
	//// IOelement and IOelement_outputnum are used to store node number of axis nodes
	//node_ax1 = IOelement;
	//node_ax2 = IOelement_outputnum;
}

void MBSLoad::GetSurfaceHeatSource(double& heat, int& surfacenum)
{
	heat = val;
	surfacenum = specpos;
}

void MBSLoad::GetVolumeHeatSource(double& heat)
{
	heat = val;
}

void MBSLoad::GetParameterInt(int* & p_inti)
{
	p_inti = p_int;
}

void MBSLoad::GetParameterDouble(double* & p_doublei)
{
	p_doublei = p_double;
}

void MBSLoad::GetParameterVector3D(double* & p_vector3di)
{
	p_vector3di = p_vector3di;
}

//$ DR 2012-10:[ changed all to MathFunction, old code
//void MBSLoad::GetHarmonic(double& omega_i, double& phase_i) 
//{
//	omega_i = data(1);
//	phase_i = data(2);
//}
//
//void MBSLoad::GetTimeSpan(double& ts, double& te)
//{
//	ts = data(1);
//	te = data(2);
//}
//
//void MBSLoad::GetTimeRamp(double& mag, double& t1on, double& t2on, double& t1off, double& t2off)
//{
//	mag  = data(1);
//	t1on = data(2);
//	t2on = data(3);
//	t1off= data(4);
//	t2off= data(5);
//}
//$ DR 2012-10:] changed all to MathFunction, old code

void MBSLoad::GetMathFunction(MathFunction& userfunction)
{
	userfunction = mathfunc;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++        Set Force type and values       +++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void MBSLoad::SetGCLoad(double vali, int gci)
{
	val=vali; gc=gci;
	loadtype = TGCload;
}

void MBSLoad::SetBodyLoad(double vali, int dir)
{
	val=vali; specpos = dir;
	if (specpos > 3) {GetMBS()->UO() << "ERROR: SetBodyLoad only accepts dir = 1 ... 3\n"; specpos = 3;}
	loadtype = Tbodyload;
}

void MBSLoad::SetGravity(double gravityconstant, int dir)
{
	val=gravityconstant; specpos = dir;
	if (specpos > 3) {GetMBS()->UO() << "ERROR: SetGravity only accepts dir = 1 ... 3\n"; specpos = 3;}
	loadtype = Tgravity;
}

void MBSLoad::SetForceVector3D(const Vector3D & force, const Vector3D& pos, const int useLocalBodyAxis)
{
	val=1.;
	loadtype = Tforcevector3D;
	vecval = force;
	forcepos = pos;
	node_f = 0;	// DR 2011-05-18
	localBodyAxis = useLocalBodyAxis; // SM 2013-01-11
}

void MBSLoad::SetForceVector3D(const Vector3D & force, const int& nodenr, const int useLocalBodyAxis)
{
	val=1.;
	loadtype = Tforcevector3D;
	vecval = force;
	forcepos = Vector3D(0.0);
	node_f = nodenr;
	localBodyAxis = useLocalBodyAxis; // SM 2013-01-11
}

void MBSLoad::SetContactZ(double zpos)
{
	val = zpos;
	loadtype = TcontactZ;
	specpos = 3; //z-dir
}

void MBSLoad::SetMomentVector3D(const Vector3D & moment, const Vector3D& pos, const int useLocalBodyAxis)
{
	val=1.;
	loadtype = Tmomentvector3D;
	vecval = moment;
	forcepos = pos;
	localBodyAxis = useLocalBodyAxis; // SM 2013-01-11
}

void MBSLoad::SetForceVector2D(const Vector2D & force, const Vector2D& pos, const int useLocalBodyAxis)
{
	val=1.;
	loadtype = Tforcevector2D;
	vecval.X() = force.X();
	vecval.Y() = force.Y();
	forcepos.X() = pos.X();
	forcepos.Y() = pos.Y();
	localBodyAxis = useLocalBodyAxis; // SM 2013-01-11
}

void MBSLoad::SetMomentVector2D(double moment, const Vector2D& pos)
{
	val=1.;
	loadtype = Tmomentvector2D;
	vecval.Z() = moment;
	forcepos.X() = pos.X();
	forcepos.Y() = pos.Y();
}

void MBSLoad::SetSurfacePressure(double pressure, int surfacenum)
{
	val = pressure; 
	specpos = surfacenum;
	loadtype = Tsurfacepressure;
}

void MBSLoad::SetDistributedMoment2D(double moment)
{
	val = moment;
	loadtype = Tdistributedmoment;
}

void MBSLoad::SetCentrifugalLoad(const Vector3D& omega, const Vector3D& pos)
{
	val = 1.;
  vecval = omega;
	forcepos = pos;
	loadtype = Tcentrifugal;
}

// ATTENTION: it is assumed, that the nodes are local node_numbers of the SAME Element (therefore only for CMS-element and equivalents)
void MBSLoad::SetRotatingForce(double vali, int node_force, int node_ax1, int node_ax2)
{
	val = vali;
	node_f = node_force;
	node_1 = node_ax1;
	node_2 = node_ax2;
	loadtype = Trotatingforce;
	
	//$ DR 2011-04-14: old code:
	// specpos is used to store node number where nodal force is applied
	//specpos = node_force;
	//// IOelement and IOelement_outputnum are used to store node numbers for axis
	//IOelement = node_ax1;
	//IOelement_outputnum = node_ax2;
}

void MBSLoad::SetSurfaceHeatSource(double heat, int surfacenum)
{
	val = heat; 
	specpos = surfacenum;
	loadtype = Tsurfaceheatsource;
}

void MBSLoad::SetVolumeHeatSource(double heat)
{
	val = heat;
	loadtype = Tvolumeheatsource;
}

void MBSLoad::SetParameterInt(int vali, int* p_inti)
{
	loadtype = Tparameterint;
	val = vali;
	p_int = p_inti;
}

void MBSLoad::SetParameterDouble(double vali, double* p_doublei)
{
	loadtype = Tparameterdouble;
	val = vali;
	p_double = p_doublei;
}

void MBSLoad::SetParameterVector3D(Vector3D& vecvali, Vector3D* p_vector3di)
{
	val = 1.;
	loadtype = Tparametervector3d;
	vecval = vecvali;
	p_vector3d = p_vector3di;
}

void MBSLoad::SetHarmonic(double omega_i, double phase_i) 
{
// alternative (depreciated feature of mathfunction
//	mathfunc.SetHarmonic(Vector(omega_i),Vector(phase_i),Vector(1.));
//
	mystr expr = mystr("sin(")+ mystr(omega_i) + mystr("*t)");
	//CMBSParser parser;

	mathfunc.SetExpression(expr,"t",mbs);

	loadfunc = TLoadMathFunc;
	
	//$ DR 2012-10: changed all to MathFunction, old code
	//data.Flush();
	//data.Add(omega_i);
	//data.Add(phase_i);
	//loadfunc = 2; 
}

void MBSLoad::SetTimeSpan(double ts, double te)
{
	mathfunc.SetPiecewise(Vector(0.,ts,te,0.),Vector(0.,1.,1.,0.),0);
	loadfunc = TLoadMathFunc;

	//$ DR 2012-10: changed all to MathFunction, old code
	//data.Flush();
	//data.Add(ts);
	//data.Add(te);
	//loadfunc = 3; 
}

void MBSLoad::SetTimeRamp(double mag, double t1on, double t2on, double t1off, double t2off)
{
	val = 1;	//$ DR 2013-01-29: change log 372, if MathFunction or IOElement are used val=1
	if(t1off == -1)
	{
		mathfunc.SetPiecewise(Vector(0.,t1on,t2on,t2on+1),Vector(0.,0.,mag,mag),1);
	}
	else
	{
		if(t2off == -1)
		{
			mathfunc.SetPiecewise(Vector(0.,t1on,t2on,t2on+1),Vector(0.,0.,mag,mag),1);
		}
		else
		{
			mathfunc.SetPiecewise(Vector(0.,t1on,t2on,t1off,t2off,t2off+1),Vector(0.,0.,mag,mag,0.,0.),1);
		}
	}
	loadfunc = TLoadMathFunc;
	//$ DR 2012-10: changed all to MathFunction, old code
	//data.Flush();
	//data.Add(mag);
	//data.Add(t1on);
	//data.Add(t2on);
	//data.Add(t1off);
	//data.Add(t2off);
	//loadfunc = 4; 
}

void MBSLoad::SetIOElement(int IOelementI, int IOelement_outputnumI)
{
	loadfunc = TLoadIOElement; //$ DR 2012-10: changed loadfunc, IOElement changed from 5 to 2
	val = 1;	//$ DR 2013-01-29: change log 372, if MathFunction or IOElement are used val=1
	IOelement = IOelementI;
	IOelement_outputnum = IOelement_outputnumI;
}

void MBSLoad::SetMathFunction(MathFunction userfunction)
{
	//data.Flush();
	mathfunc = userfunction;
	loadfunc = TLoadMathFunc; //$ DR 2012-10: changed loadfunc, MathFunction changed from 6 to 1
	val = 1;	//$ DR 2013-01-29: change log 372, if MathFunction or IOElement are used val=1
}

void MBSLoad::SetLoadSteps(const TArray<StepSettings>& settings)
{
	steps = settings;
}

int MBSLoad::GetCStepNumber(double t) const
{ 
	return mbs->GetCSNumber(t); // do indeed calculate
//	return el->GetMBS()->ComputationStepNumber(); // access-function: gives wrong values for t=0 during assemble
}

double MBSLoad::GetCStepTime(double t) const
{ 
	return mbs->GetCSTime(t); // do indeed calculate
//	return el->GetMBS()->ComputationStepTime(); // access-function: gives wrong values for t=0 during assemble
}

 int MBSLoad::GetRampMode(int i) const
{ 
	return steps(i).GetRampMode(); 
}

double MBSLoad::GetFinalValue(int i) const
{ 
	return steps(i).GetLoadFactor(); 
}

int const MBSLoad::NLoadSteps() const 
{ 
	return steps.Length(); 
}

// Get load factor from loadsteps array
double MBSLoad::GetCStepsFact(double t) const
{
	if(t <= 0) 
		return 0.;
	
  double steptime = GetCStepTime(t);    // step number (computation step number from TimeInt)
	int stepnumber = GetCStepNumber(t);   // step time   (computation step time from TimeInt)

// return value when time is after last computation step endtime...
	if(stepnumber < 0 && t > 0.) 
		return GetFinalValue(steps.Length());

// return value when too few load steps are defined
	if(stepnumber > NLoadSteps()) 
		return GetFinalValue(steps.Length());


	double finalvalue = GetFinalValue(stepnumber); // value at the end of this interval
// return value for STEP ( finalvalue is valid for entire interval )
	if(GetRampMode(stepnumber) == TCSRStep)
		return finalvalue; 

	double startvalue = 0.;  // value at beginning of this interval, =0. means that all constraints start OFF !!!
	if (stepnumber != 1) 
    startvalue = GetFinalValue(stepnumber-1);
// return value for LINEAR ( linear interpolation between beginning and end of the interval )
  if(GetRampMode(stepnumber) == TCSRLinear)
	{
		double stepfactor = startvalue + (finalvalue-startvalue)*steptime;
		return stepfactor;
	}

	const double eps_t = 1e-12;
// return value for EXPONENTIAL ( eponential interpolation between beginning and end of the interval )
	if(steptime < eps_t) 
		return startvalue;
	else if(steptime == 1.)
		return finalvalue;
	else
	{
		double factor_growth_decay = 1.;
		if(finalvalue < startvalue) 
			factor_growth_decay=-1;
// exponential interpolation:  y(x) = y(0) * (y(1) / y(0))^x ... with y(0) and y(1) may not be 0.
// with y(x) = a*e^+x + d , x[0..1] --> y(0) = a+d, y(1) = ae+d    --> a = (y(1)-y(0)) / (e-1),    d = y(0) - a  (growth)
// with y(x) = a*e^-x + d , x[0..1] --> y(0) = a+d, y(1) = ae^-1+d --> a = (y(1)-y(0)) / (e^-1-1), d = y(0) - a  (decay)
		double a = (finalvalue-startvalue) / (exp(factor_growth_decay)-1.);
		double d = startvalue - a;
		double stepfactor = a * exp(steptime*factor_growth_decay) + d;
		return stepfactor;
	}
  return 0.;
}

double MBSLoad::Evaluate(double t) const
{
	// special cases: -------------------------
	//$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
	if(! mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted) && t == 0.)   //$ MaSch 2013-08-19
		return val;

	//$ AD 2011-10-28: all parameter loads are not touched at t=0;
	if (t==0. && ( loadtype == Tparameterdouble || loadtype == Tparametervector3d) )
		return val;

	// general cases: -------------------------
	// get loadstep (static computation)
	double loadstep = 1.0;		// range 0..1, used for static computations

	if (NLoadSteps()==0)
	{
		if(GetCStepNumber(t) == 1)
		{
			loadstep = mbs->LoadFact();        // linear turn on in first CS Interval
		}
	}
	else
	{ 
	 loadstep = GetCStepsFact(t);             
	}

	// get loadfunc (static+dynamic computation)
	if (!loadfunc)				// no Mathfunction and no IOElement used
	{
		return val*loadstep;
	}
	else
	{
		switch (loadfunc)
		{
		case TLoadMathFunc: 				
			{
				return val * mathfunc.Evaluate(t); //$AD 2015-11-06: removed dependency on load-step for MathFunction Loads
				//OLD: return val * loadstep * mathfunc.Evaluate(t); 
				break;
			}
		case TLoadIOElement:		// loadsteps are not used if IOElements are used
			{
				const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
				return val * ioe.GetOutput(t, IOelement_outputnum);
				break;
			}
		}
	}
	return 0;
}

//double MBSLoad::Evaluate(double t) const
//{
//	if(! mbs->GetSimulationStatus() == TSimulationNotStarted && t == 0.)   //$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
//		return val;
//	//    ------ XXXXXXXX --------
//	double vv;
////	if (loadsteps.Length()==0)
//	if (NLoadSteps()==0)
//	{
//		if(GetCStepNumber(t) == 1)
//		{
//			vv = val*mbs->LoadFact();        // linear turn on in first CS Interval
//		}
//		else 
//		{
//			vv = val;                                 // full load after first CS Interval
//		}
//	}
//	else
//	{ 
//	 vv = val*GetCStepsFact(t);                   // 
//	}
//	//double vv = val*el->GetMBS()->LoadFact();
//
////$ AD 2011-10-28: all parameter loads are not touched ar t=0;
//	if (t==0. && ( loadtype == Tparameterdouble || loadtype == Tparametervector3d) )
//		return val;
//
//	if (!loadfunc) return vv;
//	else
//	{
//		switch (loadfunc)
//		{
//			//$ DR 2012-10:[ changed all to MathFunction, old code
//			//case 1: if (t < 1.) return 0.5*vv*(1.-cos(MY_PI*t)); else return vv; break;
//			//case 2: return vv*sin(data(1)*t+data(2)); break;
//			//case 3: if (t >= data(1) && t < data(2)) return vv; else return 0; break;
//			//case 4: //ramp with ton and tof
//			//	{
//			//		double mag = data(1)*vv;
//			//		if (t < data(2)) return 0; 
//			//		if (t < data(3)) return (t-data(2))/(data(3)-data(2))*mag; 
//			//		if (t < data(4) || data(4) == -1) return mag; 
//			//		if (t < data(5)) return (data(5)-t)/(data(5)-data(4))*mag;
//			//		return 0;
//			//		break;
//			//	}
//			//case 5:
//			//	{
//			//		const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
//			//		//const InputOutputElement* ioe = (const InputOutputElement*)el->GetMBS()->GetElementPtr(IOelement);
//			//		double val = ioe.GetOutput(t, IOelement_outputnum);
//			//		return val;
//
//			//		break;
//			//	}
//			//case 6:
//			//	{
//			//		double val = mathfunc.Evaluate(t);
//			//		return val;
//			//	}
//			//default: ;
//			//$ DR 2012-10:] changed all to MathFunction, old code
//		case 1: 				
//			{
//				double val = mathfunc.Evaluate(t);
//				return val;
//			}
//		case 2:
//			{
//				const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
//				double val = ioe.GetOutput(t, IOelement_outputnum);
//				return val;
//				break;
//			}
//		default: ;
//		}
//	}
//	return 0;
//}


void MBSLoad::AddElementLoad(Vector& f, double t, Element* el)
//void MBSLoad::AddElementLoad(Vector& f, double t) //$ DR 2012-10: loads moved from element to mbs, old code
{
	if (loadtype == TGCload)
	{
		f(gc) += Evaluate(t);
	} 
	else if (loadtype == Tbodyload || loadtype == Tgravity)
	{
		//$ PG 2019-11-05: [ MaSch 2015-07-01: temporary hack to enable Tbodyload and Tgravity for NodalDiskMass3D
		//if (el->IsRigid())
		if (el->IsRigid() || strcmp(el->GetElementSpec(), "NodalDiskMass3D")==0)
		//$ PG 2019-11-05: ]
		{
			dudq.SetLen(f.GetLen());
			if (specpos == 1){el->GetDuxDq(dudq);}
			else if (specpos == 2){el->GetDuyDq(dudq);}
			else if (specpos == 3){el->GetDuzDq(dudq);}

			if (loadtype==Tbodyload)
			{
				dudq *= Evaluate(t)*el->GetVolume();
			}
			else // Tgravity
			{
				dudq *= Evaluate(t)*el->GetMass();
			}
			f += dudq;
		}
		else
		{
			H.SetSize(f.Length(),el->Dim());
			if(loadtype == Tbodyload)
			{
				el->GetIntDuDq(H); //in fact it is DuDq Transposed
			}
			else // Tgravity
			{
				el->GetIntRhoDuDq(H); //in fact it is int_V rho DuDq Transposed
			}
			//GetMBS()->UO() << "H=" << H << "\n";
			if (el->Dim() == 3)
			{
				Vector3D l(0.,0.,0.);
				l(specpos) = Evaluate(t);
				Mult(H,l,dudq);
				f += dudq;
			}
			else
			{
				Vector2D l(0.,0.);
				if (specpos > 2) {GetMBS()->UO() << "ERROR: bodyload/gravity accepts only dir = 1..2 in 2D bodies!!!\n"; specpos = 2;}
				l(specpos) = Evaluate(t);
				Mult(H,l,dudq);
				f += dudq;
			}
		}
	}
	else if ((loadtype == Tforcevector2D) || (loadtype == Tforcevector3D)) 
	{
		//IF BODY IS RIGID
		//add force load in mass center
		dudq.SetLen(f.GetLen());

		//if (!el->IsRigid())	//$ DR 2014-07-17: differentiation between rigid or not is not necessary anymore
		//{
		if(el->Dim()==3)
		{
			Body3D* bo = (Body3D*)el;
			if (!bo->HasConstdPosdqT() || H.Getrows() == 0)
			{
				H.SetSize(f.Length(),el->Dim());
				
				if(node_f)	
				{
					forcepos = el->GetNodePos(node_f);
					bo->GetNodedPosdqT(node_f, H);
				}
				else
				{
					bo->GetdPosdqT(forcepos, H);
				}
			}
			Vector3D force = Evaluate(t)*vecval;
			if(localBodyAxis) // MSax: transform local force to global force
			{
				Matrix3D A = bo->GetRotMatrix(forcepos);
				force = A*force;
			}
			Mult(H, force, dudq);
			f += dudq;
		}
		else
		{
			Body2D* bo = (Body2D*)el;
			Vector2D force(Evaluate(t)*vecval.X(),Evaluate(t)*vecval.Y());

			if(localBodyAxis) // MSax: transform local force to global force
			{
				force = bo->GetRotMatrix2D()*force;
			}

			H.SetSize(f.Length(),el->Dim());

			bo->GetdPosdqT(Vector2D(forcepos.X(),forcepos.Y()), H);
			Mult(H, force, dudq);
			f += dudq;
		}
			//$ DR 2014-07-17: differentiation between rigid or not is not necessary anymore
		//}
		//else
		//{
		//	if(el->Dim()==3)
		//	{
		//		Body3D* bo = (Body3D*)el;
		//		Vector3D force = Evaluate(t)*vecval;

		//		if(localBodyAxis) // MSax: transform local force to global force
		//		{
		//			Matrix3D A = bo->GetRotMatrix(); // rigid body --> no local position needed
		//			force = A*force;
		//		}

		//		if(bo->ApplyLoadAtPos(dudq, force, forcepos))	//$!DR 2014-07-15 hack for Rigid3DMinCoord
		//		{
		//			f += dudq;
		//		}
		//		else
		//		{
		//			bo->ApplyDprefdq(dudq, force);
		//			f += dudq;
		//			//add load due to moments:
		//			if (forcepos.X() != 0 || forcepos.Y() != 0 || forcepos.Z() != 0)
		//			{
		//				//assert(forcepos.X() == 0 || forcepos.Y() == 0 || forcepos.Z() == 0);  //$ MaSch 2012-03: Whoever needed this, please tell JG why.

		//				Vector3D pglob = bo->GetRotMatrix()*forcepos;
		//				Vector3D mF = pglob.Cross(force);
		//				bo->ApplyDrotrefdq(dudq, mF);
		//				//el->UO() << "f_rotref=" << dudq << "\n";
		//				f += dudq;
		//			}
		//		}
		//	}
		//	else if (el->Dim()==2)
		//	{
		//		Body2D* bo = (Body2D*)el;
		//		Vector2D force(vecval.X(),vecval.Y());
		//		force *= Evaluate(t);

		//		if(localBodyAxis) // MSax: transform local force to global force
		//		{
		//			force = bo->GetRotMatrix2D()*force;
		//		}

		//		bo->ApplyDprefdq(dudq, force);
		//		f += dudq;
		//		//add load due to moments:
		//		if (forcepos.X() != 0 || forcepos.Y() != 0 || forcepos.Z() != 0)
		//		{
		//			//assert(forcepos.X() == 0 || forcepos.Y() == 0 || forcepos.Z() == 0);  //$ MaSch 2012-03: Whoever needed this, please tell JG why.

		//			Vector2D pglob = bo->GetRotMatrix2D()*Vector2D(forcepos.X(),forcepos.Y());
		//			double mF = pglob.Cross(force); //Applied moment
		//			bo->ApplyDrotrefdq(dudq, mF);
		//			f += dudq;
		//		}
		//	}
		//	else if (el->Dim()==1)	//$ DR 2013-06-19
		//	{
		//		//$ DR 2013-06-24 removed again, according to JG
		//		//Mass1D* bo = (Mass1D*)el;
		//		//double force = vecval.X()*Evaluate(t);
		//		//bo->ApplyDprefdq(dudq, force);
		//		//f += dudq;
		//	}
		//}
	}
	else if ((loadtype == Tmomentvector2D) || (loadtype == Tmomentvector3D)) 
	{
		//IF BODY IS RIGID
		//add force load in mass center
		dudq.SetLen(f.GetLen());
		if (el->IsRigid())
		{
			if (el->Dim() == 3)
			{
				Body3D* bo = (Body3D*)el;
				//el->UO() << "vecval=" << Evaluate(t)*vecval << "\n";

				Vector3D force = Evaluate(t)*vecval;
				if(localBodyAxis) // MSax: transform local force to global force
				{
					Matrix3D A = bo->GetRotMatrix();
					force = A*force;
				}

				bo->ApplyDrotrefdq(dudq, force);
				//el->UO() << "dudq=" << dudq << "\n";
				f += dudq;
			}
			else if (el->Dim() == 2)
			{
				Body2D* bo = (Body2D*)el;
				bo->ApplyDrotrefdq(dudq, Evaluate(t)*vecval.Z());
				f += dudq;
			}
		}
		else //flexible:
		{
			if(el->Dim()==3)
			{
				Body3D* bo = (Body3D*)el;

				Vector3D force = Evaluate(t)*vecval;
				if(localBodyAxis)
				{
					Matrix3D A = bo->GetRotMatrix(forcepos);
					force = A*force;
				}
			
				H.SetSize(f.Length(),el->Dim());
				bo->GetdRotdqT(forcepos, H);
				Mult(H, force, dudq);
				f += dudq;
			}
			else if(el->Dim()==2)
			{
				Body2D* bo = (Body2D*)el;
				double force = Evaluate(t) * vecval.Z();
				Vector2D fpos(forcepos.X(), forcepos.Y());
				H.SetSize(f.Length(),1);
				bo->GetdAngle2DdqT(fpos, H);
				for (int i=1; i <= f.Length(); i++)
				{
					f(i) += force*H(i,1);
				}
			}
		}
	}
	else if (loadtype == TcontactZ)
	{
		//for ANCF-plate elements only!!!
		//add force load in mass center
		dudq.SetLen(f.GetLen());

		Body3D* bo = (Body3D*)el;
		Vector3D force(0,0,1);

		double cstiff = 1e5;

		int np = 4;
		Vector3D lp[4];
		Vector3D size = bo->GetSize();
		double r = 0.3;//0.5/sqrt(3.);
		lp[0] = Vector3D(-r*size.X(),-r*size.Y(),0);
		lp[1] = Vector3D(-r*size.X(), r*size.Y(),0);
		lp[2] = Vector3D( r*size.X(),-r*size.Y(),0);
		lp[3] = Vector3D( r*size.X(), r*size.Y(),0);

		for (int i=0; i < np; i++)
		{
			double fdamp = -cstiff*0.005;
			Vector3D p = bo->GetPos(lp[i]);
			double gap = (p.Z()-val);
			if (gap <= 0)
			{

				Vector3D v = bo->GetVel(lp[i]);
				//double sgn = Sgn(fdamp);
				//if (fabs(fdamp)  > fabs(gap*cstiff)) fdamp = sgn*fabs(gap*cstiff);
				double fact = -(gap*cstiff-fdamp*v.Z()); //XG(1) = gap * cstiff  -> pressure is negative, XG(1) always negative!

				H.SetSize(f.Length(),el->Dim());

				bo->GetdPosdqT(lp[i], H);
				Mult(H, fact*force, dudq);
				f += dudq;
			}
		}
	}
	else if (loadtype == Tsurfacepressure)
	{
		//Add surface pressure
		el->AddSurfacePressure(f, Evaluate(t), specpos);
	}
	else if (loadtype == Tsurfaceheatsource)
	{
		el->AddSurfaceHeatSource(f, Evaluate(t), specpos);
	}
	else if (loadtype == Tvolumeheatsource)
	{
		el->AddVolumeHeatSource(f, Evaluate(t));
	}
	else if (loadtype == Tdistributedmoment)
	{
		if (el->Dim() == 2)
		{
			if (el->IsRigid())
			{
				GetMBS()->UO() << "Error: distributed moment only for 2D flexible beams!\n";
			}
			else
			{
				dudq.SetLen(f.Length());
				el->GetIntDkappaDq2D(dudq); //in fact it is DkappaDq Transposed

				dudq *= Evaluate(t);

				f += dudq;
			}
		}
		else
		{
			GetMBS()->UO() << "Error: distributed moment only for 2D flexible beams!\n";
		}
	}
	else if(loadtype == Tcentrifugal)
	{
		if (el->IsRigid())
		{
			if (el->Dim() == 3)
			{
				dudq.SetLen(f.Length());
				
				Vector3D p = el->GetPos(Vector3D(0.0,0.0,0.0)); // cm
				Vector3D r_rel = forcepos-p;
				Vector3D cforce = vecval.Cross(vecval.Cross(r_rel));

				el->GetDuxDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.X());
				f += dudq;

				el->GetDuyDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.Y());
				f += dudq;

				el->GetDuzDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.Z());
				f += dudq;
			}
			else if (el->Dim() == 2)
			{
				dudq.SetLen(f.Length());

				Vector3D p = el->GetPos(Vector3D(0.0,0.0,0.0)); // cm
				Vector3D r_rel = forcepos-p;
				Vector3D cforce = vecval.Cross(vecval.Cross(r_rel));

				el->GetDuxDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.X());
				f += dudq;

				el->GetDuxDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.X());
				f += dudq;

				el->GetDuyDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.Y());
				f += dudq;

				el->GetDuzDq(dudq);
				dudq *= (Evaluate(t)*el->GetMass()*cforce.Z());
				f += dudq;
			}
		}
		else 
		{
			if (el->Dim() == 3)
			{
				H.SetSize(f.Length(),el->Dim());
				el->GetIntDuDqFCentrifugal(H,vecval,forcepos);
				H *= Evaluate(t);
				
				Mult(H,Vector3D(1.0,1.0,1.0),dudq);
				f += dudq;
			}	
			else if (el->Dim() == 2)
			{
				H.SetSize(f.Length(),el->Dim());
				el->GetIntDuDqFCentrifugal(H,vecval,forcepos);
				H *= Evaluate(t);
				
				Mult(H,Vector2D(1.0,1.0),dudq);
				f += dudq;
			}				
		}
	}
	else if (loadtype == Trotatingforce)
	{
		//Trotatingforce ... force applied to a point in direction normal to axis
		//                   two such forces can be used to simulate moment (Kraeftepaar)
		//                   2D: only one axis point needed, axis is in direction Z, node_ax2 can be dummy
		dudq.SetLen(f.GetLen());
		// node numbers of node where force attacs and two nodes on axis
		int node_force, node_ax1, node_ax2;
		double moment;
		GetRotatingForce(moment, node_force, node_ax1, node_ax2);

		if(el->Dim()==3)
		{
			Body3D* bo = (Body3D*)el;
			// v_ax is direction vector of axis (normed)
			Vector3D v_ax;
			v_ax = el->GetNodePos(node_ax2) - el->GetNodePos(node_ax1);
			v_ax.Normalize();
			// r is vector from point on axis to point where force attacs
			Vector3D r;
			r = el->GetNodePos(node_force) - el->GetNodePos(node_ax1);
			// force = moment /  |v x r|^2 (v x r)
			Vector3D force;
			force = v_ax.Cross(r);
			force /= force.Norm2();
			force *= Evaluate(t);

			H.SetSize(f.Length(),el->Dim());

			bo->GetNodedPosdqT(node_force, H); //$ DR 2011-04-14: spec_pos replaced by node_force
			Mult(H, force, dudq);
			f += dudq;
		}
		else
		{
			Body2D* bo = (Body2D*)el;
			// p_ax is point on axis, axis is in Z-direction
			Vector2D p_ax = el->GetNodePos2D(node_ax1);
			// r is vector from point on axis to point where force attacs
			Vector2D r = p_ax - el->GetNodePos2D(node_force);
       
			// force = moment /  |r|^2 (r_hat)
			Vector2D force(-r.Y(), r.X());
			force /= r.Norm2();
			force *= Evaluate(t);

			H.SetSize(f.Length(),el->Dim());

			bo->GetNodedPosdqT(node_force, H);
			Mult(H, force, dudq);
			f += dudq;
		}
	}
	else if (loadtype == Tparameterint)
	{
		*p_int = (int)(Evaluate(t));
	}
	else if (loadtype == Tparameterdouble)
	{
		*p_double = Evaluate(t);
	}
	else if (loadtype == Tparametervector3d)
	{
	  *p_vector3d = vecval*Evaluate(t);
	}
}

void MBSLoad::DrawLoad(MBS* mbs, Element* el)
//virtual void MBSLoad::DrawLoad(MBS* mbs) //$ DR 2012-10: loads moved from element to mbs, old code
{
	double size = mbs->GetDOption(120);
	double linethickness = size * 0.04;
	double headsize = size * 0.2;
	Vector3D p(0.,0.,0.); //position of force
	Vector3D dir(0.,0.,0.);

//$!AD 2011-10: loads from IO-Elements are visible as the IO element itself
	if (this->IOelement != 0)
	{
		if(mbs->GetElement(IOelement).DrawElementFlag() == 0)
			return; // dont draw controlled load if control element is not visible
	}

	if (loadtype == TGCload)
	{
		dir = el->GetDOFDirD(gc);
		double val = Evaluate(mbs->GetDrawTime());
		//mbs->UO() << "load=" << val << "\n";
		if (val < 0) dir = -1*dir;
		p = el->GetDOFPosD(gc);
	} 
	else if (loadtype == Tbodyload || loadtype==Tgravity)
	{
		// AH: Check if this works for Rigids...
		// p = el->GetRefPosD();
		p = el->GetPosD(Vector3D());

		if (specpos == 1) {dir = Vector3D(1.,0.,0.);}
		else if (specpos == 2){dir = Vector3D(0.,1.,0.);}
		else if (specpos == 3){dir = Vector3D(0.,0.,1.);}

		double val = Evaluate(mbs->GetDrawTime());
		if (val < 0) dir = -1.*dir;
	}
	else if (loadtype == Tsurfacepressure)
	{
		p = el->GetPosD(Vector3D());

		dir = el->GetSurfaceNormalD(specpos);

		double val = Evaluate(mbs->GetDrawTime());
		if (val < 0) dir = -1.*dir;
	}
	else if ((loadtype == Tforcevector2D) || (loadtype == Tforcevector3D))
	{
		if(el->Dim()==3 || el->Dim()==2)
		{
			Matrix3D A;
			if(node_f > 0)	// DR 2011-05-20: the constructor with node-nr is used
			{
				//forcepos = el->GetNodeLocPos(node_f);
				//forcepos = el->GetNodePosD(node_f);
				p=el->GetNodePosD(node_f);
				if(localBodyAxis) // transform local force to global force
				{
					if (el->Dim()==3) 
					{
						Body3D* bo = (Body3D*)el;
						A = bo->GetRotMatrixD(); // rigid body --> no local position needed
						dir = A*vecval;
					}
					else
					{
						Body2D* bo = (Body2D*)el;
						A = bo->GetRotMatrix2DD();
						Vector2D dir2D = Vector2D(vecval(1),vecval(2));
						dir2D = A*dir2D;
						dir = Vector3D(dir2D(1),dir2D(2),0.);
					}
				}
				else
				{
					dir = vecval;
				}
			}
			else
			{
				p = el->GetPosD(forcepos);
				dir = vecval;
				if(localBodyAxis)
				{
					if (el->Dim()==3) 
					{
						Body3D* bo = (Body3D*)el;
						if (el->IsRigid())
						{
							A = bo->GetRotMatrixD(); // rigid body --> no local position needed
						}
						else
						{
							A = bo->GetRotMatrixD(forcepos); // rigid body --> no local position needed
						}
						dir = A*dir;
					}
					else
					{
						Body2D* bo = (Body2D*)el;
						A = bo->GetRotMatrix2DD();
						Vector2D	dir2D = Vector2D(vecval(1),vecval(2));
						dir2D = A*dir2D;
						dir = Vector3D(dir2D(1),dir2D(2),0.);
					}
				}
			}

			dir.Normalize();
			double val = Evaluate(mbs->GetDrawTime()*vecval.Norm());
			if (val < 0) dir = -1.*dir;
		}
	}
	else if ((loadtype == Tmomentvector2D) || (loadtype == Tmomentvector3D))
	{
		if(el->Dim()==3 || el->Dim()==2)
		{
			p = el->GetPosD(forcepos);
			dir = vecval;
			if(localBodyAxis && el->Dim()==3) // transform local force to global force
			{
				Matrix3D A;
				Body3D* bo = (Body3D*)el;
				if (el->IsRigid())
				{
					A = bo->GetRotMatrixD(); // rigid body --> no local position needed
				}
				else
				{
					A = bo->GetRotMatrixD(forcepos); // rigid body --> no local position needed
				}
				//Matrix3D A = el->GetRotMatrixD(); // rigid body --> no local position needed
				dir = A*dir;
			}
			dir.Normalize();
			dir *= 2; //showing that this is a moment

			double val = Evaluate(mbs->GetDrawTime()*vecval.Norm());
			if (val < 0) dir = -1.*dir;
		}
	}
	else if (loadtype == Tcentrifugal)
	{
		p = el->GetRefPosD(); // @c.m.
		dir = p - forcepos;
		vecval.GramSchmidtNormalized(dir);
		dir.Normalize();
	}
	else if(loadtype == Trotatingforce) //$ DR 2011-04-14
	{
		Vector3D r, v_ax;
		p = el->GetNodePosD(node_f);
		v_ax = el->GetNodePosD(node_2) - el->GetNodePosD(node_1);
		r = el->GetNodePosD(node_f) - el->GetNodePosD(node_1);
		dir = v_ax.Cross(r);
		dir.Normalize();
	}

	else if (loadtype == Tparameterint)
	{
		*p_int = (int)(Evaluate(mbs->GetDrawTime()));
	}
	else if (loadtype == Tparameterdouble)
	{
		*p_double = Evaluate(mbs->GetDrawTime());
	}
	else if (loadtype == Tparametervector3d)
	{
	  *p_vector3d = vecval*Evaluate(mbs->GetDrawTime()); // val == 1.
	}

	if (loadfunc == TLoadIOElement)
	{
		if(mbs->GetIOption(144))	// do not draw the arrow from control element to load, if control elements are not shown in 3D window
		{
			InputOutputElement* ioe = (InputOutputElement*)mbs->GetElementPtr(IOelement);	
			Vector3D po = ioe->ToP3D(ioe->GetOutputPosD(IOelement_outputnum));
			mbs->Graphics3D()->MyDrawArrow(po, p, Vector3D(0.3,0.3,0.3));
		}
	}

	if ( loadtype == Tvolumeheatsource )
	{
		mbs->Graphics3D()->DrawSphere(el->GetPosD(Vector3D()),
			size * 0.5, 
			6);
	}
	if (loadtype == Tsurfaceheatsource )
	{
		double val;
		int faceNr;
		GetSurfaceHeatSource(val, faceNr);

		FiniteElement3D* elFE = dynamic_cast<FiniteElement3D*>(el); 
		Vector3D locpos(0., 0., 0.);
		if (elFE != 0)
		{
			locpos = elFE->GetFaceCenter(faceNr);
			dir = locpos;
			dir.Normalize();
			p = el->GetPos(locpos);
			mbs->Graphics3D()->MyDrawArrow(p + (-1.*size)*dir, p,
			Vector3D(mbs->GetDOption(121),mbs->GetDOption(122),mbs->GetDOption(123)), linethickness, headsize, 6);
		}
		else
		{
			p = el->GetPos(locpos);
			mbs->Graphics3D()->DrawSphere(p, 
				size * 0.5, 
				6);
		}

	}
	else if (RelApproxi(dir.Norm(),1.)) //force
	{
		mbs->Graphics3D()->MyDrawArrow(p + (-1.*size)*dir, p,
			Vector3D(mbs->GetDOption(121),mbs->GetDOption(122),mbs->GetDOption(123)), linethickness, headsize, 6);
	}
	else if (RelApproxi(dir.Norm(),2.)) //moment (torque)
	{
		dir.Normalize();
		mbs->Graphics3D()->MyDrawArrow(p + (-1.*size)*dir, p,
			Vector3D(mbs->GetDOption(121),mbs->GetDOption(122),mbs->GetDOption(123)), linethickness, headsize, 6);
		mbs->Graphics3D()->MyDrawArrow(p + (-1.2*size)*dir, p + (-0.2*size)*dir,
			Vector3D(mbs->GetDOption(121),mbs->GetDOption(122),mbs->GetDOption(123)), linethickness, headsize, 6);
	}
}


//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
//MBSBodyLoadSpatial
//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------

MBSBodyLoadSpatial::MBSBodyLoadSpatial(): MBSLoad()
{
	loadtype=Tbodyloadspatial; 
	loadfunc = TLoadMathFunc;
	SetLoadName("BodyLoadSpatial");//$ DR 2012-10
	//mbs->UO(UO_LVL_warn) << "WARNING in MBSBodyLoadSpatial: Not implemented/tested for all types of elements (done for ANCFBeamShear3DGeneric).\n";
};

void MBSBodyLoadSpatial::SetBodyLoadSpatial(double vali, int dir)
{
	val=vali; specpos = dir;
	if (specpos > 3) {GetMBS()->UO() << "ERROR: SetBodyLoadSpatial only accepts dir = 1 ... 3\n"; specpos = 3;}
	loadtype = Tbodyloadspatial;
}

void MBSBodyLoadSpatial::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	
	Vector3D pos;
	Vector2D pos2;
	
	ElementData ed;

	mystr eleminfo = LoadName();
	
	ed.SetText(LoadName(), "name"); ed.SetToolTipText("name of the load"); edc.Add(ed);
	ed.SetText(GetLoadTypeString(), "load_type"); ed.SetToolTipText("specification of load type. Once the load is added to the mbs, you MUST NOT change this type anymore!"); 
	ed.SetLocked(1);	edc.Add(ed);

	ed.SetInt(GetOwnNum(), "load_number"); ed.SetToolTipText("number of the load in the mbs"); ed.SetLocked(1); edc.Add(ed);

	
	double val;
	int dir;
	GetBodyLoad(val, dir);
	ed.SetInt(dir, "direction", 1,Dim(),1,0);			ed.SetToolTipText("direction of the load"); edc.Add(ed);//$ DR 2011-07-21 removed upper limit Dim() because it is not set yet, when it is checked
	ed.SetDouble(val, "load_value");					ed.SetToolTipText("value of the load acting"); edc.Add(ed);
	
	ed.SetInt(loadfunc, "load_function_type", 0,2,1,1);			ed.SetToolTipText("time dependency of the load: 0..constant, 1..MathFunction, 2..IOElement"); 
	ed.SetLocked(1); edc.Add(ed); 
	

	// MathFunction
	ElementDataContainer edc_mf;
	mathfunc.GetElementData(edc_mf);
	edc_mf.TreeDelete("piecewise_points");
	edc_mf.TreeDelete("piecewise_values");
	edc_mf.TreeDelete("piecewise_diff_values");
	
	ElementData* tmped = edc_mf.TreeFind("parsed_function_parameter");
	tmped->SetText("t,x,y,z");
	tmped->SetLocked(1);
	
	tmped = edc_mf.TreeFind("piecewise_mode");
	tmped->SetInt(-1);
	tmped->SetLocked(1);

	ed.SetEDC(&edc_mf,"MathFunction"); 
	ed.SetToolTipText("MathFunction is used"); 
	edc.Add(ed);	
}

int MBSBodyLoadSpatial::SetElementData(ElementDataContainer& edc) 		//update element data
{
	int rv = 1;

	GetElemDataText(mbs,edc,"name",loadname,1);
	GetElemDataInt(mbs, edc, "direction", specpos, 1);
	GetElemDataDouble(mbs, edc, "load_value", val, 1);
	
	int idx = edc.Find(mystr("MathFunction"));
	if(idx && (edc.Get(idx)).IsEDC())
	{

		ElementDataContainer* edcp_mf = (edc.Get(idx)).GetEDC();
		ElementDataContainer edc_mf = *edcp_mf;
		ElementData ed;
		ed.SetText("","piecewise_points");
		edc_mf.Add(ed);
		ed.SetText("","piecewise_values");
		edc_mf.Add(ed);
		ed.SetText("","piecewise_diff_values");
		edc_mf.Add(ed);
		mathfunc.SetElementData(mbs,edc_mf);	//$ DR 2012-12-12
		val = 1; //$ DR 2014-08-21 when MathFunction is combined with scalar value, only the MathFunction should define the value (no additional multiplication!)
	}
	//CParser parser;
	
	
	//parsedfunc.SetParsedFunctionMD(&parser,mathfunc.parsedFunctionExpression,mathfunc.parsedFunctionVariables,4);
	//}
	return rv;
}

int MBSBodyLoadSpatial::CheckConsistency(Element& el, mystr& errorstr) //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = 0;
	rv = MBSLoad::CheckConsistency(el,errorstr);
	return rv;
}

double MBSBodyLoadSpatial::Evaluate(double t) const
{
	// special cases: -------------------------
	//$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
	if(! mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted) && t == 0.)   //$ MaSch 2013-08-19
		return val;

	//$ AD 2011-10-28: all parameter loads are not touched at t=0;
	if (t==0. && ( loadtype == Tparameterdouble || loadtype == Tparametervector3d) )
		return val;

	// general cases: -------------------------
	// get loadstep (static computation)
	double loadstep = 1.0;		// range 0..1, used for static computations

	if (NLoadSteps()==0)
	{
		if(GetCStepNumber(t) == 1)
		{
			loadstep = mbs->LoadFact();        // linear turn on in first CS Interval
		}
		else
		{ 
			loadstep = GetCStepsFact(t);             
		}
		return val * loadstep * mathfunc.Evaluate(t);
	}
	return 0;
}


void MBSBodyLoadSpatial::AddElementLoad(Vector& f, double t, Element* el)
{
	if (el->IsRigid())
	{
		//does this make sense ???
		//yes if preprocessing is not done carried out

		dudq.SetLen(f.GetLen());
		if (specpos == 1){el->GetDuxDq(dudq);}
		else if (specpos == 2){el->GetDuyDq(dudq);}
		else if (specpos == 3){el->GetDuzDq(dudq);}

		//evaluate at zero (constant value over rigid body)
		Vector3D pos = el->GetPos(Vector3D(0,0,0));	
		Vector mathfuncIn(t, pos.X(), pos.Y(), pos.Z());
		mathfunc.SetInputVector(mathfuncIn);
		double mathfuncVal = mathfunc.Evaluate(t);
		Vector& mathfuncVec = mathfunc.GetOutputVector();
		
		assert(0); //to be further implemented
		dudq *= Evaluate(t)*el->GetVolume();

		//Body3D::GetAltShape
		//	el->GetVolume()
		f += dudq;
	}
	else
	{
		//H.SetSize(f.Length(),el->Dim());
		dudq.SetLen(f.Length());
		el->GetIntfDuDq(mathfunc, dudq, t);
		f += dudq;
	}
}

void MBSBodyLoadSpatial::DrawLoad(MBS* mbs, Element* el)
{
	double size = mbs->GetDOption(120);
	double linethickness = size * 0.04;
	double headsize = size * 0.2;
	Vector3D p(0.,0.,0.); //position of force
	double t = mbs->GetDrawTime();
	
	p = el->GetPosD(Vector3D());

	Vector3D pos = el->GetPos(Vector3D());	
	Vector mathfuncIn(t, pos.X(), pos.Y(), pos.Z());
	mathfunc.SetInputVector(mathfuncIn);
  double mathfuncVal = mathfunc.Evaluate(t);
	
	Vector& mathfuncVec = mathfunc.GetOutputVector();
	Vector3D dir(mathfuncVec(1),mathfuncVec(2),mathfuncVec(3));
	double val = dir.Norm();
	dir = (1./val)*dir;
	
	//draw arrow
	mbs->Graphics3D()->MyDrawArrow(p + (-1.*size)*dir, p,
			Vector3D(mbs->GetDOption(121),mbs->GetDOption(122),mbs->GetDOption(123)), linethickness, headsize, 6);
}

//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
//MBSSurfaceLoad
//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
void MBSSurfaceLoad::SetSurfaceLoad(double force_abs, Vector3D force_dir, int pressure, int perunit, TArray<int>& elements, TArray<int>& facets)
{
	loadtype = Tsurfaceload;
	val = force_abs;
	vecval = force_dir;
	flag_pressure = pressure;
	localBodyAxis = pressure; // for the time being the pressure variant is always local, the forvevector variant is always global
	flag_forceperunitarea = perunit;
	elementlist.CopyFrom(elements);
	facetlist.CopyFrom(facets);	

	int flag_error=0;
// NOTE:
	// is mbs a valid pointer at this time ? --> this is NOT sure
// must have same length
	if(elementlist.Length()!=facetlist.Length())
	{
		GetMBS()->UO() << "ERROR: SetAreaLoad requires lists Elementnumbers and Facetnumbers to have equal length \n";
	}
// elements must exist and element facet must exist
	int nelem = GetMBS()->NE();
	for(int i=1; i<=elementlist.Length(); i++)
	{
		if(elementlist(i)>nelem)
		{
			GetMBS()->UO() << "ERROR: SetAreaLoad element list contains a number that exceeds number of elements in MBS \n";
		}
		else
		{
			Element* e = GetMBS()->GetElementPtr(i);
			// nfacets
		}
	}		
	total_area = ComputeTotalArea();
	SortEntries();  
}

void MBSSurfaceLoad::GetElementData(ElementDataContainer& edc)
{
	MBSLoad::GetElementData(edc);

// specific for derived class
	ElementData ed;
	ed.SetBool(flag_pressure, "pressure"); ed.SetToolTipText("use only z-component of force as pressure (both flags 'local_force' and 'per_unit_area' are automatically set to true)"); edc.Add(ed);
	ed.SetDouble(val,"total_force"); ed.SetToolTipText("value of pressure or absolute value (length) of the force vector"); edc.Add(ed);

	if(flag_pressure)
	{
		ed.SetVector3D(vecval.X(),vecval.Y(),vecval.Z(),"force_direction"); ed.SetLocked(1); ed.SetToolTipText("direction of the applied force, this vector is scaled to length 1"); edc.Add(ed);
		ed.SetBool(1,"force_per_unit_area"); ed.SetLocked(1); ed.SetToolTipText("setting pressure requires force_per_unit_area"); edc.Add(ed);
	}
	else
	{
		ed.SetVector3D(vecval.X(),vecval.Y(),vecval.Z(),"force_direction"); ed.SetToolTipText("direction of the applied force, this vector is scaled to length 1"); edc.Add(ed);
		ed.SetBool(flag_forceperunitarea, "force_per_unit_area"); ed.SetToolTipText("force is defined per unit area rather than total force on entire area"); edc.Add(ed);
	}

	Vector elnrs(elementlist.Length()); for(int i=1; i<=elementlist.Length(); i++) { elnrs(i) = elementlist(i); }
	ed.SetVector(elnrs.GetVecPtr(), elnrs.Length(), "element_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("MBS Element numbers to apply the load to");
	edc.Add(ed);
	Vector fanrs(facetlist.Length()); for(int i=1; i<=facetlist.Length(); i++) { fanrs(i) = facetlist(i); }
	ed.SetVector(fanrs.GetVecPtr(), fanrs.Length(), "element_facet_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("Facet numbers to apply the load to");
	edc.Add(ed);
	Vector faars(facetarealist.Length()); for(int i=1; i<=facetarealist.Length(); i++) { faars(i) = facetarealist(i); }
	ed.SetVector(faars.GetVecPtr(), faars.Length(), "element_facet_area"); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("Facet areas to apply the load to");
	edc.Add(ed);
	ed.SetDouble(total_area,"total_area");  ed.SetLocked(1); ed.SetToolTipText("Total areas to apply the load to");
}

int MBSSurfaceLoad::SetElementData(ElementDataContainer& edc)
{
	int rv = 1;
	rv = MBSLoad::SetElementData(edc);

	// specific for derived class
	GetElemDataBool(mbs, edc, "pressure", flag_pressure, 1);
	GetElemDataDouble(mbs, edc, "total_force", val, 1);
	GetElemDataVector3D(mbs, edc, "force_direction", vecval, 1);
	if(flag_pressure)
	{
		localBodyAxis = 1;
		flag_forceperunitarea = 1;
	}
	else
	{
		localBodyAxis = 0;
		GetElemDataBool(mbs, edc, "force_per_unit_area", flag_forceperunitarea, 1);
	}
	GetElemDataIVector(mbs, edc, "element_numbers", elementlist, 1);
	GetElemDataIVector(mbs, edc, "element_facet_numbers", facetlist, 1);

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   	total_area = ComputeTotalArea();
	return rv;
}

int MBSSurfaceLoad::CheckConsistency(Element& el, mystr& errorstr)
{
	int rv = 0;
	rv = MBSLoad::CheckConsistency(el,errorstr);
// specific for derived class
	return  rv;
}

void MBSSurfaceLoad::SortEntries()
{
//void Quicksort(TArray<int>& a, TArray<int>& b, TArray<double>& c)
//Sorts an array a(1..a.Length()) into ascending numerical order by Shells method (diminishing increment
//sort). a is replaced on output by its sorted rearrangement. b and c are sorted according to the order of a.
	TArray<double> dummy(elementlist.Length());
	Quicksort(elementlist,facetlist,facetarealist);
	lastidx = 0;
}

double MBSSurfaceLoad::ComputeTotalArea()
{
	total_area = 0.;
	facetarealist.SetLen(elementlist.Length());
	facetarealist.SetAll(0.);

	for(int i=1; i<=elementlist.Length(); i++)
	{
		double area = 0;
		int elnr = elementlist(i);
// TODO: ??? define function in base class Element ???
		Element *el = GetMBS()->GetElementPtr(elnr);
		FiniteElement3D* fel3 = dynamic_cast<FiniteElement3D*>(el);
		if (fel3)
		{
			area = fel3->GetFaceArea(facetlist(i));
		}
		else
		{
			FiniteElement2D* fel2 = dynamic_cast<FiniteElement2D*>(el);
			if (fel2)
			{
				fel2->PreAssemble();
				area = fel2->GetEdgeLength(facetlist(i));
			}
			else
			{
				assert(false && "Elements must be of type FiniteElement2D or FiniteElement3D");
			}
		}

		facetarealist(i) = area;
		total_area += area;
	}
	return total_area;
}

double MBSSurfaceLoad::Evaluate(double t) const
{
	//$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
	if(! mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted) && t == 0.)   
		return val;

	if(!loadfunc)
	{
		return val*mbs->LoadFact(); // time intependent scaling factor for forcevector
	}
	else
	{
		switch (loadfunc)
		{
		case TLoadMathFunc:
			return val * mathfunc.Evaluate(t); // time dependent (mathfunction) scaling factor for forcevector
			break;
		case TLoadIOElement:		
			const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
			return val * ioe.GetOutput(t, IOelement_outputnum); // time dependent (controlelement) scaling factor for forcevector
			break;
		}
	}
	return 0;
}

void MBSSurfaceLoad::AddElementLoad(Vector& f, double t, Element* el)
{
	int mbselementnumber = el->GetOwnNum();

// reset search mark if necessary
	if(lastidx<=0) { lastidx=1; }
	if(mbselementnumber < elementlist(lastidx)) { lastidx=1; } // start from beginning to search for a number smaller than the current position (likely new timestep)

// find and process all entries for the given element
	for(int i=lastidx; i<=elementlist.Length(); i++)
	{
		if(mbselementnumber == elementlist(i))
		{
			int facetnumber = facetlist(i);
			double facearea = facetarealist(i);
			double tfact = Evaluate(t);
			Vector3D dir0 = vecval;
			dir0.Normalize();
				
			if(flag_pressure) 
			{ 
				//double z = tfact*vecval.Z();
				double z = -tfact; // AD: NOTE that val is already included in tfact !!
				el->AddElementSurfaceLoad(f, Vector3D(0.,0.,z), 1, facetnumber); // always local
			}
			else
			{
				double x,y,z;
				if(flag_forceperunitarea)
				{
					x = tfact*dir0.X();
					y = tfact*dir0.Y();
					z = tfact*dir0.Z();
				}
				else
				{
					//double afact = facearea/total_area;
					//x = tfact*dir0.X()/afact;
					//y = tfact*dir0.Y()/afact;
					//z = tfact*dir0.Z()/afact;
					//$EK 2018-05-16 bug fixing: correct scaling of total load 
					//final integration is
					//total_force = force_per_area * area ===> force_per_area = total_force/ total_area
					x = tfact*dir0.X()/total_area;
					y = tfact*dir0.Y()/total_area;
					z = tfact*dir0.Z()/total_area;
				}
				el->AddElementSurfaceLoad(f, Vector3D(x,y,z), 0, facetnumber); // always global
			}
			lastidx = i;
		}
		else if(mbselementnumber < elementlist(i))
		{
			break; // 
		}
		else /*(mbselementnumber > elementlist(i)*/
		{ 
			/*continue;*/ // to next in sequence
		}
	}
}

void MBSSurfaceLoad::DrawLoad(MBS* mbs, Element* el)
{
	double size = mbs->GetDOption(120);
	double linethickness = size * 0.04;
	double headsize = size * 0.2;
	double t = mbs->GetDrawTime();
	FiniteElement2D* FE2D = dynamic_cast<FiniteElement2D*>(el);

	int mbselementnumber = el->GetOwnNum();

	for(int i=1; i<=elementlist.Length(); i++)
	{
		if(mbselementnumber == elementlist(i))
		{
			int facetnumber = facetlist(i);
			Vector3D p1;
			Vector3D n0;
			if (FE2D != nullptr)
			{
				p1 = FE2D->GetPosD(FE2D->GetAreaCoordOfEdgeCoord(facetnumber, 0.).MakeV3D());
				n0 = FE2D->GetEdgeNormalD(facetnumber).MakeV3D();
			}
			else
			{
				p1 = el->GetPosD(el->GetFaceCenter(facetnumber));
				n0 = el->GetSurfaceNormalD(facetnumber);
			}
			double sign_val = 1 - 2 * (val < 0);

			if (flag_pressure == 0)
			{
				// draw absolute direction
				Vector3D dir = vecval;
				dir.Normalize();

				mbs->Graphics3D()->MyDrawArrow(p1 + (-1.*2.*size)*dir*sign_val, p1, Vector3D(0., 1., 0.), linethickness, headsize, 6);
			}
			else
			{
				Vector3D dir = vecval.Z()*n0;
				dir.Normalize();

				mbs->Graphics3D()->MyDrawArrow(p1 + (2.*size)*dir*sign_val, p1, Vector3D(0., 0., 1.), linethickness, headsize, 6);
			}
		}
	}
}




//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
//MBSAreaLoad
//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
// AD: obsolete code - to be removed when SurfaceLoad is fully tested
//////void MBSAreaLoad::SetAreaLoad(double forceperunitarea, TArray<int>& elements, TArray<int>& facets)
//////{
//////	loadtype = Tareaload;
//////	val = forceperunitarea;
//////	elementlist.CopyFrom(elements);
//////	facetlist.CopyFrom(facets);
//////
//////	int flag_error=0;
//////// NOTE:
//////	// is mbs a valid pointer at this time ? --> this is NOT sure
//////// must have same length
//////	if(elementlist.Length()!=facetlist.Length())
//////	{
//////		GetMBS()->UO() << "ERROR: SetAreaLoad requires lists Elementnumbers and Facetnumbers to have equal length \n";
//////	}
//////// elements must exist and element facet must exist
//////	int nelem = GetMBS()->NE();
//////	for(int i=1; i<elementlist.Length(); i++)
//////	{
//////		if(elementlist(i)>nelem)
//////		{
//////			GetMBS()->UO() << "ERROR: SetAreaLoad element list contains a number that exceeds number of elements in MBS \n";
//////		}
//////		else
//////		{
//////			Element* e = GetMBS()->GetElementPtr(i);
//////			// nfacets
//////		}
//////	}	
//////	
//////	SortEntries();
//////}
//////
//////void MBSAreaLoad::GetElementData(ElementDataContainer& edc)
//////{
//////	MBSLoad::GetElementData(edc);
//////
//////// specific for derived class
//////	ElementData ed;
//////	ed.SetDouble(val,"pressure"); ed.SetToolTipText("pressure - normal force per unit area");
//////	edc.Add(ed);
//////	Vector elnrs(elementlist.Length()); for(int i=1; i<=elementlist.Length(); i++) { elnrs(i) = elementlist(i); }
//////	ed.SetVector(elnrs.GetVecPtr(), elnrs.Length(), "element_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("MBS Element numbers to apply the load to");
//////	edc.Add(ed);
//////	Vector fanrs(facetlist.Length()); for(int i=1; i<=facetlist.Length(); i++) { fanrs(i) = facetlist(i); }
//////	ed.SetVector(fanrs.GetVecPtr(), fanrs.Length(), "element_facet_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("Facet numbers to apply the load to");
//////	edc.Add(ed);
//////}
//////
//////int MBSAreaLoad::SetElementData(ElementDataContainer& edc)
//////{
//////	int rv = 1;
//////	rv = MBSLoad::SetElementData(edc);
//////
//////// specific for derived class
//////	GetElemDataDouble(mbs, edc, "pressure", val, 1);
//////	GetElemDataIVector(mbs, edc, "element_numbers", elementlist, 1);
//////	GetElemDataIVector(mbs, edc, "element_facet_numbers", facetlist, 1);
//////	
//////	return rv;
//////}
//////
//////int MBSAreaLoad::CheckConsistency(Element& el, mystr& errorstr)
//////{
//////	int rv = 0;
//////	rv = MBSLoad::CheckConsistency(el,errorstr);
//////// specific for derived class
//////	return  rv;
//////}
//////
//////void MBSAreaLoad::SortEntries()
//////{
////////void Quicksort(TArray<int>& a, TArray<int>& b, TArray<double>& c)
////////Sorts an array a(1..a.Length()) into ascending numerical order by Shells method (diminishing increment
////////sort). a is replaced on output by its sorted rearrangement. b and c are sorted according to the order of a.
//////	TArray<double> dummy(elementlist.Length());
//////	Quicksort(elementlist,facetlist,dummy);
//////	lastidx = 0;
//////}
//////
//////double MBSAreaLoad::Evaluate(double t) const
//////{
//////	//$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
//////	if(! mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted) && t == 0.)   
//////		return val;
//////
//////	if(!loadfunc)
//////	{
//////		return val*mbs->LoadFact(); // time intependent pressure
//////	}
//////	else
//////	{
//////		switch (loadfunc)
//////		{
//////		case TLoadMathFunc:
//////			return val * mathfunc.Evaluate(t); // time dependent (mathfunction) pressure
//////			break;
//////		case TLoadIOElement:		
//////			const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
//////			return val * ioe.GetOutput(t, IOelement_outputnum); // time dependent (controlelement) pressure
//////			break;
//////		}
//////	}
//////	return 0;
//////}
//////
//////void MBSAreaLoad::AddElementLoad(Vector& f, double t, Element* el)
//////{
//////	int mbselementnumber = el->GetOwnNum();
//////
//////// reset search mark if necessary
//////	if(lastidx<=0) { lastidx=1; }
//////	if(mbselementnumber < elementlist(lastidx)) { lastidx=1; } // start from beginning to search for a number smaller than the current position (likely new timestep)
//////
//////// find and process all entries for the given element
//////	for(int i=lastidx; i<=elementlist.Length(); i++)
//////	{
//////		if(mbselementnumber == elementlist(i))
//////		{
//////			int facetnumber = facetlist(i);
//////			el->AddAreaLoad(f, Evaluate(t), facetnumber);
//////			lastidx = i;
//////		}
//////		else if(mbselementnumber < elementlist(i))
//////		{
//////			break; // 
//////		}
//////		else /*(mbselementnumber > elementlist(i)*/
//////		{ 
//////			/*continue;*/ // to next in sequence
//////		}
//////	}
//////}
//////
//////
//////void MBSAreaLoad::DrawLoad(MBS* mbs, Element* el)
//////{
//////	double size = mbs->GetDOption(120);
//////	double linethickness = size * 0.04;
//////	double headsize = size * 0.2;
//////	double t = mbs->GetDrawTime();
//////	
//////	int mbselementnumber = el->GetOwnNum();
//////
//////	for(int i=1; i<=elementlist.Length(); i++)
//////	{
//////		if(mbselementnumber == elementlist(i))
//////		{
//////			int facetnumber = facetlist(i);	
//////			Vector3D p1 = el->GetPosD(el->GetFaceCenter(facetnumber));
//////			Vector3D n0 = el->GetSurfaceNormalD(facetnumber);
//////			double sign_val = 1-2*(val<0);
//////
//////			mbs->Graphics3D()->MyDrawArrow(p1 + (-1.*size)*n0*sign_val, p1, Vector3D(1.,0.,0.), linethickness, headsize, 6);
//////		}
//////	}
////////	
//////}

//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
//MBSAreaTraction
//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
// AD: obsolete code - to be removed when surfaceLoad is fully tested
//////void MBSAreaTraction::SetAreaTraction(Vector3D forceperunitarea, int localforce, TArray<int>& elements, TArray<int>& facets)
//////{
//////	loadtype = Tareatraction;
//////	val = 1.0;
//////	vecval = forceperunitarea;
//////	localBodyAxis = localforce;
//////	elementlist.CopyFrom(elements);
//////	facetlist.CopyFrom(facets);	
//////
//////	int flag_error=0;
//////// NOTE:
//////	// is mbs a valid pointer at this time ? --> this is NOT sure
//////// must have same length
//////	if(elementlist.Length()!=facetlist.Length())
//////	{
//////		GetMBS()->UO() << "ERROR: SetAreaLoad requires lists Elementnumbers and Facetnumbers to have equal length \n";
//////	}
//////// elements must exist and element facet must exist
//////	int nelem = GetMBS()->NE();
//////	for(int i=1; i<elementlist.Length(); i++)
//////	{
//////		if(elementlist(i)>nelem)
//////		{
//////			GetMBS()->UO() << "ERROR: SetAreaLoad element list contains a number that exceeds number of elements in MBS \n";
//////		}
//////		else
//////		{
//////			Element* e = GetMBS()->GetElementPtr(i);
//////			// nfacets
//////		}
//////	}		
//////	SortEntries();
//////}
//////
//////void MBSAreaTraction::GetElementData(ElementDataContainer& edc)
//////{
//////	MBSLoad::GetElementData(edc);
//////
//////// specific for derived class
//////	ElementData ed;
//////	ed.SetVector3D(vecval.X(),vecval.Y(),vecval.Z(),"force_vector"); ed.SetToolTipText("force per unit area");
//////	edc.Add(ed);	
//////	ed.SetInt(localBodyAxis, "local_force", 0,1,1,1); ed.SetToolTipText("force is defined in the coordinate system of the facet, with 3rd coord being the outward surface normal");
//////	edc.Add(ed);
//////	Vector elnrs(elementlist.Length()); for(int i=1; i<=elementlist.Length(); i++) { elnrs(i) = elementlist(i); }
//////	ed.SetVector(elnrs.GetVecPtr(), elnrs.Length(), "element_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("MBS Element numbers to apply the load to");
//////	edc.Add(ed);
//////	Vector fanrs(facetlist.Length()); for(int i=1; i<=facetlist.Length(); i++) { fanrs(i) = facetlist(i); }
//////	ed.SetVector(fanrs.GetVecPtr(), fanrs.Length(), "element_facet_numbers"); ed.SetValuesInt(); ed.SetVariableLength(); /*ed.SetLocked(1);*/ ed.SetToolTipText("Facet numbers to apply the load to");
//////	edc.Add(ed);
//////}
//////
//////int MBSAreaTraction::SetElementData(ElementDataContainer& edc)
//////{
//////	int rv = 1;
//////	rv = MBSLoad::SetElementData(edc);
//////
//////// specific for derived class
//////	GetElemDataVector3D(mbs, edc, "force_vector", vecval, 1);
//////	GetElemDataInt(mbs, edc, "local_force", localBodyAxis, 1);
//////
//////	GetElemDataIVector(mbs, edc, "element_numbers", elementlist, 1);
//////	GetElemDataIVector(mbs, edc, "element_facet_numbers", facetlist, 1);
//////	
//////	return rv;
//////}
//////
//////int MBSAreaTraction::CheckConsistency(Element& el, mystr& errorstr)	
//////{
//////	int rv = 0;
//////	rv = MBSLoad::CheckConsistency(el,errorstr);
//////// specific for derived class
//////	return  rv;
//////}
//////
//////void MBSAreaTraction::SortEntries()                                     
//////{
////////void Quicksort(TArray<int>& a, TArray<int>& b, TArray<double>& c)
////////Sorts an array a(1..a.Length()) into ascending numerical order by Shells method (diminishing increment
////////sort). a is replaced on output by its sorted rearrangement. b and c are sorted according to the order of a.
//////	TArray<double> dummy(elementlist.Length());
//////	Quicksort(elementlist,facetlist,dummy);
//////	lastidx = 0;
//////}
//////
//////double MBSAreaTraction::Evaluate(double t) const
//////{
//////	//$ AD 2011-09-15: draw all loads before computation starts (even if they are off at t=0)
//////	if(! mbs->GetSimulationStatus().GetStatusFlag(TSimulationNotStarted) && t == 0.)   
//////		return val;
//////
//////	if(!loadfunc)
//////	{
//////		return val*mbs->LoadFact(); // time intependent pressure
//////	}
//////	else
//////	{
//////		switch (loadfunc)
//////		{
//////		case TLoadMathFunc:
//////			return val * mathfunc.Evaluate(t); // time dependent (mathfunction) pressure
//////			break;
//////		case TLoadIOElement:		
//////			const InputOutputElement& ioe = (const InputOutputElement&)(mbs->GetElement(IOelement));
//////			return val * ioe.GetOutput(t, IOelement_outputnum); // time dependent (controlelement) pressure
//////			break;
//////		}
//////	}
//////	return 0;
//////}
//////
//////void MBSAreaTraction::AddElementLoad(Vector& f, double t, Element* el)
//////{
//////	int mbselementnumber = el->GetOwnNum();
//////
//////// reset search mark if necessary
//////	if(lastidx<=0) { lastidx=1; }
//////	if(mbselementnumber < elementlist(lastidx)) { lastidx=1; } // start from beginning to search for a number smaller than the current position (likely new timestep)
//////
//////// find and process all entries for the given element
//////	for(int i=lastidx; i<=elementlist.Length(); i++)
//////	{
//////		if(mbselementnumber == elementlist(i))
//////		{
//////			int facetnumber = facetlist(i);
//////			double v = Evaluate(t);
//////			double x = v*vecval.X();
//////			double y = v*vecval.Y();
//////			double z = v*vecval.Z();
//////			el->AddAreaTraction(f, Vector3D(x,y,z), localBodyAxis, facetnumber);
//////			lastidx = i;
//////		}
//////		else if(mbselementnumber < elementlist(i))
//////		{
//////			break; // 
//////		}
//////		else /*(mbselementnumber > elementlist(i)*/
//////		{ 
//////			/*continue;*/ // to next in sequence
//////		}
//////	}
//////}
//////
//////void MBSAreaTraction::DrawLoad(MBS* mbs, Element* el)
//////{
//////	double size = mbs->GetDOption(120);
//////	double linethickness = size * 0.04;
//////	double headsize = size * 0.2;
//////	double t = mbs->GetDrawTime();
//////	
//////	int mbselementnumber = el->GetOwnNum();
//////
//////	for(int i=1; i<=elementlist.Length(); i++)
//////	{
//////		if(mbselementnumber == elementlist(i))
//////		{
//////			int facetnumber = facetlist(i);	
//////			Vector3D p1 = el->GetPosD(el->GetFaceCenter(facetnumber));
//////			Vector3D n0 = el->GetSurfaceNormalD(facetnumber);
//////			double sign_val = 1-2*(val<0);
//////
//////			if(localBodyAxis==0)
//////			{
//////				// draw absolute direction
//////				Vector3D dir = vecval;
//////				dir.Normalize();
//////
//////				
//////				mbs->Graphics3D()->MyDrawArrow(p1 + (-1.*2.*size)*dir*sign_val, p1, Vector3D(0.,1.,0.), linethickness, headsize, 6);
//////			}
//////			else
//////			{	
//////				// draw relative to surface
//////				int facetnumber = facetlist(i);	
//////			
//////				FiniteElement3D* fe3d = (FiniteElement3D*) el;
//////				FEFace locface = fe3d->GetLocalFace(facetnumber);
//////
//////				Vector3D p1D = fe3d->GetNodePosD(locface.Node(1)); 
//////				Vector3D p2D = fe3d->GetNodePosD(locface.Node(2)); 
//////				Vector3D pnD = fe3d->GetNodePosD(locface.Node(locface.NFaceNodes()));
//////
//////				Vector3D v12 = p2D-p1D;
//////				Vector3D v1n = pnD-p1D;
//////				Vector3D n0 = v12.Cross(v1n);
//////
//////				v12.Normalize();
//////				v1n.Normalize();
//////				n0.Normalize();
//////
//////				Vector3D dir = vecval.X()*v12+vecval.Y()*v1n+vecval.Z()*n0;
//////				dir.Normalize();
//////
//////				mbs->Graphics3D()->MyDrawArrow(p1 + (-1.*2.*size)*dir*sign_val, p1, Vector3D(0.,0.,1.), linethickness, headsize, 6);
//////			}
//////		}
//////	}
//////}