//#**************************************************************
//#
//# filename:             material.cpp
//#
//# author:               Gerstmayr Johannes, Aigner Larissa
//#
//# generated:						July 2004
//# updated:							September 2010
//# description:          
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#include "mbs_interface.h"
#include "element.h"
#include "material.h"
#include "solversettings_auto.h"
#include "elementdataaccess.h"
#include "interpolators1D.h"

const Matrix3D MaterialGeneric::dummyMatrix3D = Matrix3D(0);
const Matrix2D MaterialGeneric::dummyMatrix2D = Matrix2D(0);
const Vector MaterialGeneric::dummyVector = Vector(0);



////////////////////////////////////////////////
// function definitions of class MaterialGeneric
////////////////////////////////////////////////

void MaterialGeneric::DefaultConstructor()
{
	material_type = mystr("MaterialGeneric");
	materialcolor.Set(0, 0, 1);
	plane = 0;
	plane_stress = 0;
	density = 0.;
	dampingM = 0.;
	dampingK = 0.;
	inelasticity_solution_method = ISM_FixedPoint;
	inelasticity_solution_method_str = GetInelasticitySolutionMethodString(inelasticity_solution_method);
}

void MaterialGeneric::CopyFrom(const MaterialGeneric& s)
{
	mbs = s.mbs;
	plane = s.plane;  
	plane_stress = s.plane_stress;
	density = s.density;
	dampingM = s.dampingM;
	dampingK = s.dampingK;
	materialname = s.materialname;
	materialnumber = s.materialnumber;
	materialcolor = s.materialcolor;
	material_type = s.material_type;

	inelasticity_solution_method = s.inelasticity_solution_method;
	inelasticity_solution_method_str = s.inelasticity_solution_method_str;
}

void MaterialGeneric::SetMaterialGeneric(double densityI, int planeI, int planestressI, const Vector3D& color)
{
	density = densityI;
	plane = planeI;
	plane_stress = planestressI;
	materialcolor = color;
	material_type = GetTypeName();
}

void MaterialGeneric::SetMaterialFromEDCData()
{
	// align internal flag with user defined string for inelasticity solution method
	inelasticity_solution_method = GetInelasticitySolutionMethod(inelasticity_solution_method_str, mbs);
}

int MaterialGeneric::CheckConsistency(mystr& errorstr)
{
	int rv = 0;
	if (mbs == NULL)
	{
		errorstr += mystr("ERROR: pointer to mbs not assigned!\n");
		rv = max(rv, 2);
	}
	if (materialnumber == 0)
	{
		errorstr += mystr("ERROR: id not assigned!\n");
		rv = max(rv, 2);
	}
	if (dampingM < 0)
	{
		errorstr += mystr("WARNING: dampingM is negative!\n");
		rv = max(rv, 0);
	}
	if (dampingK < 0)
	{
		errorstr += mystr("WARNING: dampingK is negative!\n");
		rv = max(rv, 0);
	}

	return rv;
}

void MaterialGeneric::GetElementData(ElementDataContainer& edc) 		//fill in all material data
{
	ElementData ed;
	ed.SetInt(GetMaterialNumber(), "Material_number"); ed.SetLocked(1); edc.Add(ed);
	GetElementDataAuto(edc);
	ed.SetInt(GetInelasticVariablesCount(), "nr_inelastic_variables"); ed.SetToolTipText("number of inelastic variables used by material class"); ed.SetLocked(1); edc.TreeAdd("Inelasticity", ed);
}

int MaterialGeneric::SetElementData(ElementDataContainer& edc) //set material data according to ElementDataContainer
{
	int rv = SetElementDataAuto(edc);
	SetMaterialFromEDCData();
	return rv;
}

mystr MaterialGeneric::GetInelasticitySolutionMethodString(MaterialGeneric::InelasticitySolutionMethod ism)
{
	switch (ism)
	{
	case ISM_FixedPoint:
		return mystr("fixed_point");
	case ISM_ReturnMapping:
		return mystr("return_mapping");
	default:
		return mystr("");
	}
}

MaterialGeneric::InelasticitySolutionMethod MaterialGeneric::GetInelasticitySolutionMethod(const mystr& ism_str, const MBS* mbs) 
{
	if (ism_str.Compare(mystr("fixed_point")))
		return ISM_FixedPoint;
	else if (ism_str.Compare(mystr("return_mapping")))
		return ISM_ReturnMapping;

	mbs->UO(UO_LVL_warn) << "WARNING in MaterialGeneric:GetInelasticitySolutionMethod(): inelasticity_solution_method = \"" << ism_str << "\" is not known, switching to \"fixed_point\"!\n";
	return ISM_FixedPoint;
	//return ISM_NotSpecified;
}

void MaterialGeneric::SetInelasticitySolutionMethod(const MaterialGeneric::InelasticitySolutionMethod& ism)
{
	inelasticity_solution_method = ism;
	inelasticity_solution_method_str = GetInelasticitySolutionMethodString(ism);
}	

void MaterialGeneric::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const
{
	if (HasInitialStrains())
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_initial_strain, FieldVariableDescriptor::FVCI_z, true);
}

double MaterialGeneric::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const
{
	switch (fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_initial_strain:
	{
		Matrix3D initial_strain;
		GetInitialStrain(initial_strain);
		return fvd.GetComponent(initial_strain);
	}
	}
	return FIELD_VARIABLE_NO_VALUE;
}



////////////////////////////////////////////////////////
// function definitions of class ElasticParametersLinear
////////////////////////////////////////////////////////

void ElasticParametersLinear::DefaultConstructor()
{
	youngs_modulus = 0.;
	poisson_ratio = 0.;
	is_orthotropic = 0;
}
	
void ElasticParametersLinear::CopyFrom(const ElasticParametersLinear& m)
{
	youngs_modulus = m.youngs_modulus;
	poisson_ratio = m.poisson_ratio;
	orthotropic_constants = m.orthotropic_constants;
	is_orthotropic = m.is_orthotropic;
	mbs2 = m.mbs2;
}

void ElasticParametersLinear::SetElasticParametersLinear(double youngs_modulusI, double poisson_ratioI)
{
	youngs_modulus = youngs_modulusI;
	poisson_ratio = poisson_ratioI;
	is_orthotropic = 0;
}

void ElasticParametersLinear::SetElasticParametersLinear(const OrthotropicConstants& oc)
{
	orthotropic_constants = oc;
	is_orthotropic = 1;

	//GetMBS()->UO(UO_LVL_all) << "ortho.E1=" << orthotropic_constants.E1 << "\n";
	//GetMBS()->UO(UO_LVL_all) << "ortho.E2=" << orthotropic_constants.E2 << "\n";
	//GetMBS()->UO(UO_LVL_all) << "ortho.E3=" << orthotropic_constants.E3 << "\n";
	//GetMBS()->UO(UO_LVL_all) << "ortho.G12=" << orthotropic_constants.G12 << "\n";

	//GetMBS()->UO(UO_LVL_all) << "ortho.nu12=" << orthotropic_constants.nu12 << "\n";
	//GetMBS()->UO(UO_LVL_all) << "ortho.nu13=" << orthotropic_constants.nu13 << "\n";
	//GetMBS()->UO(UO_LVL_all) << "ortho.nu23=" << orthotropic_constants.nu23 << "\n";
}

int ElasticParametersLinear::CheckConsistency(mystr& errorstr)   //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = 0;

	if (IsOrthotropicMaterial())
	{
		// still to be specified
	}
	else  // isotropic
	{
		if (youngs_modulus < 0)
		{
			errorstr += mystr("ERROR: Young's module must not be negative\n");
			rv = max(rv, 1);
		}
		if (poisson_ratio < 0 || poisson_ratio > 0.5)
		{
			errorstr += mystr("ERROR: Poisson's ratio must be in [0, 1/2]\n");
			rv = max(rv, 1);
		}
	}

	return rv;
}

void ElasticParametersLinear::ComputeElasticityMatrix(Matrix& C, int is_planar_material, int is_plane_strain) const 
{
	//assert ( (C.Getcols() == 6 && !IsPanarMaterial())||(C.Getcols() == 3 && IsPanarMaterial()) );

	double Em = YoungsModulus();
	double nu = PoissonRatio();

	if (!is_planar_material)
	{	//3D
		if (!IsOrthotropicMaterial())
		{
			C.SetSize(6,6);
			C.SetAll(0.);
			C(1,1) = 1.-nu;	C(1,2) = nu;	  C(1,3) = nu;
			C(2,1) = nu;	  C(2,2) = 1.-nu;	C(2,3) = nu;
			C(3,1) = nu;	  C(3,2) = nu;	  C(3,3) = 1.-nu;
			C(4,4) = (1.-2.*nu)/2.;
			C(5,5) = (1.-2.*nu)/2.;
			C(6,6) =(1.-2.*nu)/2.;
			C *= Em/((1.+nu)*(1.-2.*nu));
		}
		else
		{
			C.SetSize(6,6);
			C.SetAll(0.);
			const OrthotropicConstants& oc = orthotropic_constants;


			C(1,1) = (-oc.E2()+Sqr(oc.NU23())*oc.E3())*Sqr(oc.E1());    C(1,2) = -(oc.NU12()*oc.E2()+oc.NU13()*oc.NU23()*oc.E3())*oc.E1()*oc.E2();	  C(1,3) = -(oc.NU12()*oc.NU23()+oc.NU13())*oc.E1()*oc.E2()*oc.E3();
			C(2,1) = -(oc.NU12()*oc.E2()+oc.NU13()*oc.NU23()*oc.E3())*oc.E1()*oc.E2();	  C(2,2) = -(oc.E1()-Sqr(oc.NU13())*oc.E3())*Sqr(oc.E2());    C(2,3) = -(oc.NU23()*oc.E1()+oc.NU12()*oc.NU13()*oc.E2())*oc.E2()*oc.E3();
			C(3,1) = -(oc.NU12()*oc.NU23()+oc.NU13())*oc.E1()*oc.E2()*oc.E3();	  C(3,2) = -(oc.NU23()*oc.E1()+oc.NU12()*oc.NU13()*oc.E2())*oc.E2()*oc.E3();	  C(3,3) = -(oc.E1()-Sqr(oc.NU12())*oc.E2())*oc.E2()*oc.E3();
			C *= 1./(-oc.E1()*oc.E2()+oc.E1()*Sqr(oc.NU23())*oc.E3()+Sqr(oc.NU12())*Sqr(oc.E2())+2*oc.NU12()*oc.E2()*oc.NU13()*oc.NU23()*oc.E3()+Sqr(oc.NU13())*oc.E2()*oc.E3());


			C(4,4) = oc.G23();
			C(5,5) = oc.G13();
			C(6,6) = oc.G12();

		}
	}
	else
	{
		C.SetSize(3,3);
		if (!IsOrthotropicMaterial())
		{	
			if (!is_plane_strain)
			{ //plane stress
				double k = Em/(1.-nu*nu);
				C(1,1) = k;    C(1,2) = k*nu; C(1,3) = 0;
				C(2,1) = k*nu; C(2,2) = k; 		C(2,3) = 0;
				C(3,1) = 0; 	 C(3,2) = 0;  	C(3,3) = k * (1.-nu)/2.;
			}
			else
			{ //plane strain
				double k = Em/((1.+nu)*(1.-2*nu));
				C(1,1) = k*(1-nu); C(1,2) = k*nu;	    C(1,3) = 0;
				C(2,1) = k*nu;		 C(2,2) = k*(1-nu);	C(2,3) = 0;
				C(3,1) = 0; 			 C(3,2) = 0;				C(3,3) = k * (1.-2*nu)/2.;
			}
		}
		else
		{ 
			if (is_plane_strain)
			{
				//EK 2012-01-24 : plane strain case of orthotropic material 
				const OrthotropicConstants& oc = orthotropic_constants;
				C.SetAll(0.);
				double oce2 = oc.E2();
				double nu12 = oc.NU12();

				C(1,1) = oc.E1() * oce2 * (oce2 * Sqr(oc.NU23())-oc.E3());    C(1,2) = -oc.E1() * oce2 * (oc.E3() * nu12 + oce2 * oc.NU13() * oc.NU23());	 
				C(2,1) = -oc.E1() * oce2 *(nu12 * oc.E3() + oce2 * oc.NU13() * oc.NU23());	  C(2,2) = Sqr(oce2) * ( oc.E1() * Sqr(oc.NU13())-oc.E3());    
				C *= 1./(-oce2 * oc.E3()+Sqr(oc.NU23()) * Sqr(oce2)+oce2*Sqr(oc.NU13())*oc.E1()+oc.E1()*Sqr(nu12)*oc.E3()+2*oc.E1()*nu12*oc.NU13()*oc.NU23()*oce2);
				
				C(3,3) = oc.G12();
			}
		}
	}
}






////////////////////////////////////////////////////////////////////////////////////
// function definitions of class MaterialContinuumMechanics : public MaterialGeneric
////////////////////////////////////////////////////////////////////////////////////

//[STRESS] at input 2nd piola kirchoff stress tensor, at output deviatoric part of stress tensor
//returns von-Mises stress
double MaterialContinuumMechanics::ComputeDeviator(Matrix3D& stress) const
{
	double pressure = stress.Trace()/3.;
	for (int i=1; i<=3; i++)
	{
		stress(i,i) -= pressure;
	}
	double vonMises = sqrt(3./2.*stress.NormSqr());

	return vonMises;
}

// 2D plain strain: computes the 33 stress component from the 11, 12 and 22 components and the plastic variables.
// Note that the 13 and 23 components always vanish in 2D.
double MaterialContinuumMechanics::ComputeStress33_2D(const Matrix2D& stress2d, double strain_offset_zz) const
{
	assert (IsPlanarMaterial());
	
	if (IsPlaneStrain())
	{
		mbs->UO()<< "MaterialContinuumMechanics::ComputeStress33_2D requires plane stress case, only derived classes may have necessary information for plane strain case.";
		return 0.;
	}

	return 0.;
}

//[STRESS2D] at input components (11, 12, 21, 22) of stress,
//           at output components (11, 12, 21, 22) of deviator of stress
//returns effective von-Mises stress (invoking also component 33 of deviator of stress!)
double MaterialContinuumMechanics::ComputeDeviator2D(Matrix2D& stress2d, double strain_offset_zz) const
{
	assert (IsPlanarMaterial());
	
	Matrix3D stress3D(stress2d);
	stress3D(3,3) = ComputeStress33_2D(stress2d, strain_offset_zz);

	double pressure = stress3D.Trace()/3.;
	stress2d(1,1) -= pressure;
	stress2d(2,2) -= pressure;

	return ComputeDeviator(stress3D);
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// function definitions of class Material : public MaterialContinuumMechanics, public ElasticParametersLinear
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Material::DefaultConstructor()
{	
	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
	type = TMatEmpty;
}

MaterialGeneric* Material::GetCopy() const
{
	MaterialGeneric* ec = new Material(*this);
	return ec;
}

void Material::CopyFrom(const Material& m)
{
	//MaterialGeneric::DefaultConstructor(); // (AD) explicit call of class Init() function
	//Material::DefaultConstructor(); // (AD) explicit call of class Init() function
	MaterialContinuumMechanics::CopyFrom(m);
	ElasticParametersLinear::CopyFrom(m);

	type = m.type;
}

void Material::SetMaterialFromEDCData()    //set material type and material parameters by user specified EDC variables
{
	MaterialContinuumMechanics::SetMaterialFromEDCData();

	if (plane)
	{
		type = TMBSMaterial(TMat2D+TMatFlexGeneral);
		if (!plane_stress)
		{
			type = (TMBSMaterial)(type + TMatPlaneStrain);
		}
	}
	else
	{
		type = TMatFlexGeneral;
	}
}


////$ PG 2013-1-21: Function Deprecated, Rigid should not use class Material!
Material& Material::SetMaterialRigid(double densityI)
{
	//$ PG 2012-12-19: Call of Init() should be removed: 
	//Init();

	type = TMatRigidBody;
	density = densityI;
	return *this;
}

//$ PG 2016-1-19: default set function
void Material::SetMaterial(double densityI, double youngs_modulusI, double poisson_ratioI, Vector3D color, int planeI, int planestressI)
{
	density = densityI;
	plane = planeI;
	plane_stress = planestressI;
	materialcolor = color;

	SetElasticParametersLinear(youngs_modulusI, poisson_ratioI);
		
	if (planeI)
	{
		type = TMBSMaterial(TMat2D+TMatFlexGeneral);
		if (!planestressI)
		{
			type = (TMBSMaterial)(type + TMatPlaneStrain);
		}
	}
	else
	{
		type = TMatFlexGeneral;
	}
}

void Material::SetMaterial(double densityI, Vector3D color, const OrthotropicConstants& oc)
{
	type = (TMBSMaterial)(TMatFlexGeneral);
	density = densityI;
	materialcolor = color;
	SetElasticParametersLinear(oc);
}

mystr Material::GetTypeName() const
{
	return "Material";
}

//$ PG 2016-2-11: [ old, and no more used
//mystr Material::GetTypeSpec() const
//{
//	mystr addon = "";
//
//	if (type&TMatRigidBody) return mystr("rigid")+addon;
//	if (type&TMatFlexGeneral) return mystr("general3D")+addon;
//	if (type&TMat2D) return mystr("2D")+addon;
//	if (type&TMatBeam) return mystr("beam3D")+addon;
//	if (type&TMatPlate) return mystr("plate")+addon;
//	if (type&TMatTemperature) return mystr("temperature")+addon;
//
//	return addon;
//}
//$ PG 2016-2-11: ]

TMBSMaterial Material::GetType() const 
{
	return type; 
}

void Material::SetType(TMBSMaterial t)
{
	type = t;
}

void Material::AddType(TMBSMaterial t)
{
	type = TMBSMaterial(type|t);
}

int Material::CheckConsistency(mystr& errorstr)   //rv==0 --> OK, rv==1 --> can not compute, rv==2 --> can not draw and not compute
{
	int rv = max(MaterialContinuumMechanics::CheckConsistency(errorstr), ElasticParametersLinear::CheckConsistency(errorstr));

	if (density <= 0 && GetMBS()->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("ERROR: density must be positive in case of a dynamic computation\n");
		rv = max(rv, 1);
	}

	return rv;
}

int Material::IsInelasticMaterial() const
{
	return 0;
}

int Material::GetInelasticVariablesCount() const
{
	return 0;
}

void Material::ComputeStressFromStrain(const Matrix3D& total_strain, Matrix3D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
	double la = GetLambda();
	double mu = GetMu();

	Matrix3D strain = total_strain;
	if (HasInitialStrains())
	{
		Matrix3D initial_strain;
		GetInitialStrain(initial_strain);
		strain -= initial_strain;
	}

	if (!IsOrthotropicMaterial())
	{
		stress = ((2*mu) * strain + Matrix3D(la * strain.Trace()));
	}
	else
	{
		ConstMatrix<36> C;
		this->ComputeElasticityMatrix(C);

		ConstVector<6> elastic_strain(strain(1,1),strain(2,2),strain(3,3), 2*strain(2,3), 2*strain(1,3), 2*strain(1,2));
		ConstVector<6> stress_vec(6);
		stress_vec = C*elastic_strain;

		stress(1,1) = stress_vec(1); stress(1,2) = stress_vec(6); stress(1,3) = stress_vec(5); 
		stress(2,1) = stress_vec(6); stress(2,2) = stress_vec(2); stress(2,3) = stress_vec(4); 
		stress(3,1) = stress_vec(5); stress(3,2) = stress_vec(4); stress(3,3) = stress_vec(3); 
	}
}

//$ YV 2014.04.22
// if this flag is on, then the total number of integration points
// with active plastic flow will be counted per Newton iteration
//#define PRINT_PLASTIC_FLOW_STATS

void Material::ComputeStressFromStrain2D(
		const Matrix2D& total_strain, Matrix2D& stress,
		const Vector& inelastic_variables, bool compute_trial_stress) const
{
	assert(IsPlanarMaterial());
	double mu = GetMu();
	double la = GetLambda();
	if (!IsPlaneStrain()) { la = GetLambdaPlaneStress(); }

	Matrix2D strain = total_strain;
	if (HasInitialStrains())
	{
		Matrix2D initial_strain;
		GetInitialStrain2D(initial_strain);
		strain -= initial_strain;
	}		

	if (!IsOrthotropicMaterial())
	{
		stress = ((2*mu)*(strain) + Matrix2D(la*strain.Trace()));
	}
	else
	{
		ConstMatrix<9> C;
		this->ComputeElasticityMatrix(C);

		Vector3D elastic_strain_vec(strain(1,1), strain(2,2), 2*strain(1,2));
		Vector3D stress_vec = C*elastic_strain_vec;
		stress(1,1) = stress_vec(1); stress(1,2) = stress_vec(3); 
		stress(2,1) = stress_vec(3); stress(2,2) = stress_vec(2); 
	}
}


// 2D plain strain: computes the 33 stress component from the 11, 12 and 22 components and the plastic variables.
// Note that the 13 and 23 components always vanish in 2D.
double Material::ComputeStress33_2D(const Matrix2D& stress2d, double strain_offset_zz) const
{
	assert (IsPlanarMaterial());
	
	double stress_33;  // return value
	if (IsPlaneStrain())
	{
		double nu = PoissonRatio();
		double tr2d = stress2d.Trace();
		stress_33 = nu*tr2d - YoungsModulus()*strain_offset_zz;
	}
	else // plane stress
	{
		stress_33 = 0.;
	}

	return stress_33;
}

void Material::ComputeElasticityMatrix(Matrix& C) const 
{
	ElasticParametersLinear::ComputeElasticityMatrix(C, IsPlanarMaterial(), IsPlaneStrain());
}

//PG: This routines are deprecated, since they addresses geometrical properties only.
//*   It should be provided by the element - not by the material.
void Material::ComputeStrain(Matrix3D& strain, const Matrix3D& grad_u, int isgeomnonlin)
{
	if (isgeomnonlin)
	{
		Matrix3D Fh = grad_u + Matrix3D(1);
		strain = 0.5*(Fh.GetTp()*Fh-Matrix3D(1));
	}
	else
	{
		strain = 0.5*(grad_u.GetTp() + grad_u);
	}
}
void Material::ComputeStrain2D(Matrix2D& strain, const Matrix2D& grad_u, int isgeomnonlin)
{
	if (isgeomnonlin)
	{
		Matrix2D Fh = grad_u + Matrix2D(1);
		strain = 0.5*(Fh.GetTp()*Fh-Matrix2D(1));
	}
	else
	{
		strain = 0.5*(grad_u.GetTp() + grad_u);
	}
}

//$ PG 2013-8-2: replaced by DoAveragingStep/DoAveragingStep2D
//// Computes value of stress component zz averaged over element
//double Material::ComputeAveragedContactStress(FiniteElement3D* const fe)
//{
//	return 0;
//}

//$ PG 2016-1-26: method seems not to be used anymore
/*
void Material::GetParam(int i, double& value, mystr& name, int& group, int& flag)
{
	//flag&1  ==integer
	//flag&2  ==positive with zero
	//flag&4  ==check (0 or 1)
	//flag&8  ==for plane problems only
	//flag&16 ==for beam problems only
	//flag&32 ==for special materials only
	//flag&64 ==for 3D problems only

	switch(i)
	{
	case 1: {value = density; name = "density"; group = 1; flag = 2; break;}
	case 2: {value = dampingM; name = "mass_prop_damping"; group = 1; flag = 2; break;}
	case 3: {value = dampingK; name = "stiffness_prop_damping"; group = 1; flag = 2; break;}

	case 4: {value = youngs_modulus; name = "youngs_modulus"; group = 2; flag = 2; break;}
	case 5: {value = poisson_ratio; name = "poisson_ratio"; group = 2; flag = 0; break;}
	case 6: {value = !IsPlaneStrain(); name = "planestress"; group = 3; flag = 8+4; break;}

	case 7: {value = beamA   ; name = "beamA";    group = 4; flag = 2+16; break;}
	case 8: {value = beamEIy ; name = "beamEIy";  group = 4; flag = 2+16; break;}
	case 9: {value = beamEIz ; name = "beamEIz";  group = 4; flag = 2+16+64; break;}
	case 10:{value = beamGAky; name = "beamGAky"; group = 4; flag = 2+16; break;}
	case 11:{value = beamGAkz; name = "beamGAkz"; group = 4; flag = 2+16+64; break;}
	case 12:{value = beamGJkx; name = "beamGJkx"; group = 4; flag = 2+16+64; break;}
	case 13:{value = beamEA  ; name = "beamEA";   group = 4; flag = 2+16; break;}

	case 14:{value = beamRhoA; name = "beamRhoA";   group = 4; flag = 2+16; break;}
	case 15:{value = beamRhoIx; name = "beamRhoIx"; group = 4; flag = 2+16+64; break;}
	case 16:{value = beamRhoIy; name = "beamRhoIy"; group = 4; flag = 2+16; break;}
	case 17:{value = beamRhoIz; name = "beamRhoIz"; group = 4; flag = 2+16+64; break;}



	case 21: {value = (int)inelasticity_type;     name = "inelasticity_type";     group = 3; flag = 1; break;}
	default:
		value=0; name=""; group=0; flag=0;
		GetMBS()->UO() << "Error in Material, parameter not found!\n";
	}
}
*/

void Material::GetLinearizedStress2D(const Vector3D& strain, Vector3D& stress, const Vector3D& inel_strain, const Vector3D& stress_lin) const
{
	ConstMatrix<36> C(6,6);
	ComputeElasticityMatrix(C);

	stress = C*(strain);
}

void Material::GetStress2D(const Vector3D& strain, Vector3D& stress, const Vector3D& inel_strain, const double contactstress) const
{
	ConstMatrix<36> C(6,6);
	ComputeElasticityMatrix(C);

	stress = C*(strain);
}

/////////////////////////////////////////////////////////
// function definitions of class MaterialThermalExpansion
/////////////////////////////////////////////////////////


mystr MaterialThermalExpansion::GetTypeName() const
{
	return "MaterialThermalExpansion";
}
MaterialGeneric* MaterialThermalExpansion::GetCopy() const
{
	return new MaterialThermalExpansion(*this);
}

void MaterialThermalExpansion::DefaultConstructor()
{	
	Material::DefaultConstructor();
	delta_T = 0;
	thermal_expansion_coefficient = 0.;
}

void MaterialThermalExpansion::SetMaterialFromEDCData()
{
	Material::SetMaterialFromEDCData();
}

void MaterialThermalExpansion::CopyFrom(const MaterialThermalExpansion& mat)
{
	Material::CopyFrom(mat);
	delta_T = mat.delta_T;
	thermal_expansion_coefficient = mat.thermal_expansion_coefficient;
}

void MaterialThermalExpansion::GetInitialStrain2D(Matrix2D & initial_strain) const
{
	Matrix3D initial_strain_3D;
	GetInitialStrain(initial_strain_3D);
	if (plane_stress)
	{
		initial_strain.SetDiag((1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient);
	}
	if (!plane_stress) // plane strain
	{
		initial_strain.SetDiag(1*(1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient
			+ PoissonRatio()*GetInitialStrain33());
	}
}

double MaterialThermalExpansion::GetInitialStrain33() const
{
	return (1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient;
}

void MaterialThermalExpansion::GetInitialStrain(Matrix3D & initial_strain) const
{
	initial_strain.SetDiag((1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient);
} 

void MaterialThermalExpansion::SetThermalExpansion(double delta_T, double thermal_expansion_coefficient)
{
	this->delta_T = delta_T;
	this->thermal_expansion_coefficient = thermal_expansion_coefficient;
}


//////////////////////////////////////////////////////////////////////
// function definitions of class MaterialElastoplasticThermalExpansion
//////////////////////////////////////////////////////////////////////


mystr MaterialElastoplasticThermalExpansion::GetTypeName() const
{
	return "MaterialElastoplasticThermalExpansion";
}
MaterialGeneric* MaterialElastoplasticThermalExpansion::GetCopy() const
{
	return new MaterialElastoplasticThermalExpansion(*this);
}

void MaterialElastoplasticThermalExpansion::DefaultConstructor()
{	
	MaterialElastoplastic::DefaultConstructor();
	delta_T = 0;
	thermal_expansion_coefficient = 0.;
}

void MaterialElastoplasticThermalExpansion::SetMaterialFromEDCData()
{
	MaterialElastoplastic::SetMaterialFromEDCData();
}

void MaterialElastoplasticThermalExpansion::CopyFrom(const MaterialElastoplasticThermalExpansion& mat)
{
	MaterialElastoplastic::CopyFrom(mat);
	delta_T = mat.delta_T;
	thermal_expansion_coefficient = mat.thermal_expansion_coefficient;
}

void MaterialElastoplasticThermalExpansion::GetInitialStrain2D(Matrix2D & initial_strain) const
{
	if (plane_stress)
	{
		initial_strain.SetDiag((1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient);
	}
	else // plane strain
	{
		initial_strain.SetDiag(1*(1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient
			+ PoissonRatio()*GetInitialStrain33());
	}
}
double MaterialElastoplasticThermalExpansion::GetInitialStrain33() const
{
	return (1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient;
}
void MaterialElastoplasticThermalExpansion::GetInitialStrain(Matrix3D & initial_strain) const
{
	initial_strain.SetDiag((1 + 0.5*delta_T*thermal_expansion_coefficient)*delta_T*thermal_expansion_coefficient);
} 

void MaterialElastoplasticThermalExpansion::SetThermalExpansion(double delta_T, double thermal_expansion_coefficient)
{
	this->delta_T = delta_T;
	this->thermal_expansion_coefficient = thermal_expansion_coefficient;
}



////////////////////////////////////////////////////////////////////
// function definitions of class ElastoplasticParametersPrandtlReuss
////////////////////////////////////////////////////////////////////

void ElastoplasticParametersPrandtlReuss::DefaultConstructor()
{
	yieldstress = 0;
	tangentmodule = 0;
}
	
void ElastoplasticParametersPrandtlReuss::CopyFrom(const ElastoplasticParametersPrandtlReuss& m)
{
	yieldstress = m.yieldstress;
	tangentmodule = m.tangentmodule;
}

void ElastoplasticParametersPrandtlReuss::SetElastoplasticParametersPrandtlReuss(double yield_stressI, double tangent_moduleI)
{
	yieldstress = yield_stressI;
	tangentmodule = tangent_moduleI;
}

int ElastoplasticParametersPrandtlReuss::CheckConsistency(mystr& errorstr)
{
	int rv = 0;

	if(yieldstress < 0)
	{
		errorstr += mystr("ERROR: yield stress must not be negative!\n");
		rv = max(1,rv);
	}
	if(tangentmodule < 0)
	{
		errorstr += mystr("ERROR: tangent module must not be negative!\n");
		rv = max(1,rv);
	}

	return rv;
}

int ElastoplasticParametersPrandtlReuss::GetInelasticVariablesCount(int planar) const
{
	if (planar)
		return 4;     // 3*strain+1*hardening_parameter
		
	return 7;     // 6*strain+1*hardening_parameter
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// function definitions of class MaterialElastoplastic : public Material, public ElastoplasticParametersPrandtlReuss
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void MaterialElastoplastic::DefaultConstructor()
{
	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
}

MaterialGeneric* MaterialElastoplastic::GetCopy() const
{
	MaterialGeneric* ec = new MaterialElastoplastic(*this);
	return ec;
}

void MaterialElastoplastic::CopyFrom(const MaterialElastoplastic& m)
{
	Material::CopyFrom(m);
	ElastoplasticParametersPrandtlReuss::CopyFrom(m);
}

void MaterialElastoplastic::SetMaterialElastoplastic(double densityI, double youngs_modulusI, double poisson_ratioI,
																			Material::InelasticitySolutionMethod inelasticity_solution_methodI, double yield_stressI, double tangent_moduleI,
																			Vector3D color, int planeI, int planestressI)
//$ PG 2016-1-19: default set function
{
	Material::SetMaterial(densityI, youngs_modulusI, poisson_ratioI, color, planeI, planestressI);
	ElastoplasticParametersPrandtlReuss::SetElastoplasticParametersPrandtlReuss(yield_stressI, tangent_moduleI);
	SetInelasticitySolutionMethod(inelasticity_solution_methodI);
}
int MaterialElastoplastic::CheckConsistency(mystr& errorstr)
{
	int rv = max(Material::CheckConsistency(errorstr), ElastoplasticParametersPrandtlReuss::CheckConsistency(errorstr));
	return rv;
}

int MaterialElastoplastic::IsInelasticMaterial() const
{
	return 1;
}

int MaterialElastoplastic::GetInelasticVariablesCount() const
{
	return ElastoplasticParametersPrandtlReuss::GetInelasticVariablesCount(IsPlanarMaterial());
}

mystr MaterialElastoplastic::GetTypeName() const
{
	return ("MaterialElastoplastic");
}


void MaterialElastoplastic::ComputeStressFromStrain(
									Matrix3D& stress,	
									const Matrix3D& total_strain, const Matrix3D& plastic_strain, double hardening_parameter,
									bool compute_trial_stress) const
{
	//material is inelastic (or plastic), but elastic law is linear (Hookes law)
	double mu = GetMu();

	Matrix3D s(total_strain);
	s -= plastic_strain;

	if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
	{
		Matrix3D trial_stress;
		Material::ComputeStressFromStrain(s, trial_stress);
		double von_mises = ComputeDeviator(trial_stress);
		double yield_stress = YieldStress();

		double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_parameter);
		if ( beta > 0 )   // new plasticity occurs
		{
			double xi = 2.*mu + pow(yield_stress*HardeningModule(), 2)*2./3.;
			s -= (beta/xi/von_mises)*trial_stress;
		}
	}

	Material::ComputeStressFromStrain(s, stress);
}

void MaterialElastoplastic::ComputeStressFromStrain(
									const Matrix3D& strain, Matrix3D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
	//material is inelastic (or plastic), but elastic law is linear (Hookes law)

	assert (inel_variables.Length() >= 7);
	Matrix3D inel_strain;
	StrainVectorToMatrix3D(inel_strain, inel_variables);	
	double hardening_parameter = inel_variables(7);

	ComputeStressFromStrain(stress,	strain, inel_strain, hardening_parameter, compute_trial_stress);
}

//$ PG 2017-3-7: save version: transfrom total and plastic strain to 3D tensors, compute stress using 3D functionality, and return upper 2x2 block of stress tensor
void MaterialElastoplastic::ComputeStressFromStrain2D(
									const Matrix2D& strain, Matrix2D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
	double hardening_parameter = 0.;
	Matrix3D total_strain, old_plastic_strain;	
	Get3DStrainsTensors2D(total_strain, old_plastic_strain, hardening_parameter, strain, inel_variables);

	Matrix3D stress3D;
	ComputeStressFromStrain(stress3D, total_strain, old_plastic_strain, hardening_parameter, compute_trial_stress);	
	stress.Set(stress3D(1,1), stress3D(2,1), stress3D(1,2), stress3D(2,2));
}

void MaterialElastoplastic::Get3DStrainsTensors2D(Matrix3D& total_strain, Matrix3D& old_plastic_strain, double& hardening_parameter, const Matrix2D& strain, const Vector& inel_variables) const
{
  total_strain.Set(
    strain(1,1),strain(1,2),0,
    strain(2,1),strain(2,2),0,
    0,0,0);
  if (!IsPlaneStrain())
  {
    if (!IsOrthotropicMaterial())
      total_strain(3,3) = - (strain(1,1) + strain(2,2));
    else
      mbs->UO(UO_LVL_err) << "ERROR in MaterialElastoplastic: Case (!IsPlaneStrain() && IsOrthotropicMaterial()) not implemented!\n";
  }

  if (inel_variables.Length() >= 4)
  {
    hardening_parameter = inel_variables(4);
    double diag1 = inel_variables(1);
    double diag2 = inel_variables(2);
    double diag1plus2 = diag1 + diag2;
    double offdiag = 0.5*inel_variables(3);
    old_plastic_strain.Set(
      diag1, offdiag, 0,
      offdiag, diag2, 0,
      0, 0, -diag1plus2);  //$ PG 2017-3-9: old plastic strain must be tracefree

    if (!IsPlaneStrain())
    {
      if (!IsOrthotropicMaterial())
      {
        total_strain(3,3) += 2.*GetMu()/GetLambda()*diag1plus2;
      }
    }
  }
}

//$ PG 2017-3-7: tested, copy from 3D and modified (for plane strain only)
//void MaterialElastoplastic::ComputeStressFromStrain2D(
//									const Matrix2D& strain, Matrix2D& stress,
//									const Vector& inel_variables,
//									bool compute_trial_stress) const
//{
//	//material is inelastic (or plastic), but elastic law is linear (Hookes law)
//	assert (inel_variables.Length() >= 4);
//
//	Material::ComputeStressFromStrain2D(strain, stress);
//
//	Matrix2D old_plastic_strain;
//	StrainVectorToMatrix2D(old_plastic_strain, inel_variables);
//	double mu = GetMu();
//	stress -= 2*mu*old_plastic_strain;
//
//	if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
//	{
//		double hardening_param = inel_variables(4);
//		Matrix2D dev_stress(stress);
//		double von_mises = ComputeDeviator2D(dev_stress, inel_variables); //transforms stress into its deviator (dev_stress) and returns it in argument list (apart from retrun value, which is von Mises stress)!
//		double yield_stress = YieldStress();
//
//		double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_param);
//		if ( beta > 0 )   // new plasticity occurs
//		{
//			double xi = 2.*mu + pow(yield_stress*HardeningModule(), 2)*2./3.;
//			Matrix2D plastic_strain_update = (beta/xi/von_mises)*dev_stress;
//			stress -= 2*mu*plastic_strain_update;
//		}
//	}
//}

////$ PG 2017-3-7: original
//void MaterialElastoplastic::ComputeStressFromStrain2D(
//	const Matrix2D& strain, Matrix2D& stress,
//	const Vector& inelastic_variables, bool compute_trial_stress) const
//{
//	assert(IsPlanarMaterial());
//
//#ifdef PRINT_PLASTIC_FLOW_STATS
//	static int old_newton_its = 0;
//	static int active_plastic_flows = 0;
//	static int calls = 0;
//	if(!GetMBS()->IsJacobianComputation())
//		calls++;
//
//	if(old_newton_its != GetMBS()->NumSolver().GetNewtonIts())
//	{
//		old_newton_its = GetMBS()->NumSolver().GetNewtonIts();
//		GetMBS()->UO(UO_LVL_0) << "its = " << old_newton_its << ", active flows = " << active_plastic_flows << ", calls = " << calls << "\n";
//		active_plastic_flows = 0;
//		calls = 0;
//	}
//#endif
//
//
//	double mu = GetMu();
//	double la = GetLambda();
//	if (!IsPlaneStrain()) { la = GetLambdaPlaneStress(); }
//
//	if (!IsOrthotropicMaterial())
//	{
//		stress = ((2*mu)*(strain) + Matrix2D(la*strain.Trace()));
//	}
//	else
//	{
//		ConstMatrix<9> C;
//		this->ComputeElasticityMatrix(C);
//
//		Vector3D elastic_strain_vec(strain(1,1), strain(2,2), 2*strain(1,2));
//		Vector3D stress_vec = C*elastic_strain_vec;
//		stress(1,1) = stress_vec(1); stress(1,2) = stress_vec(3); 
//		stress(2,1) = stress_vec(3); stress(2,2) = stress_vec(2); 
//	}
//
//	if (inelastic_variables.Length() != 0)
//	{ 
//		Matrix2D inelastic_strain;
//		StrainVectorToMatrix2D(inelastic_strain, inelastic_variables);
//
//		stress -= (2*mu)*inelastic_strain;
//		if (!IsPlaneStrain())
//		{
//			double inelastic_strain_33 = -(inelastic_strain(1,1) + inelastic_strain(2,2));
//			double val = la * inelastic_strain_33;
//			stress(1,1) += val;
//			stress(2,2) += val;
//		}
//
//		if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
//		{
//			Matrix2D trial_stress(stress);
//			double von_mises = ComputeDeviator2D(trial_stress, inelastic_variables);
//			double yield_stress = YieldStress();
//			double hardening_param = inelastic_variables(4);
//			double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_param);
//			if ( beta > 0. )   // new plasticity occurs
//			{
//#ifdef PRINT_PLASTIC_FLOW_STATS
//				if(!GetMBS()->IsJacobianComputation())
//					active_plastic_flows++;
//#endif
//				double xi = 2.*mu + Sqr(yield_stress*HardeningModule())*2./3.;
//				inelastic_strain = (beta/xi/von_mises)*trial_stress;  // inelastic strain update saved in matrix of inealstic strain
//			}
//
//			stress -= (2*mu)*inelastic_strain; 
//			if (!IsPlaneStrain())
//			{
//				double inelastic_strain_33 = -(inelastic_strain(1,1) + inelastic_strain(2,2));
//				double val = la * inelastic_strain_33;
//				stress(1,1) += val;
//				stress(2,2) += val;
//			}
//		}
//	}
//}


double MaterialElastoplastic::ComputeInelasticVariablesUpdateFromStrain(Matrix3D& inel_strain_update, double& hardening_param_update, const Matrix3D& strain, const Matrix3D& inel_strain, double hardening_param) const
{
	double inel_strain_update_norm = 0.;   // return value in case of ISM_FixedPoint
	double mu = GetMu();

	Matrix3D trial_stress;
	Material::ComputeStressFromStrain(strain-inel_strain, trial_stress);
	double von_mises = ComputeDeviator(trial_stress);  // returns vonMises via return value (double), and deviator of trial stress via Matrix3D& trial_stress
	double yield_stress = YieldStress();

	double beta = von_mises - yield_stress*(1 + HardeningModule()*hardening_param);
	if ( beta > 0. )   // new plasticity occurs
	{
		double xi = 2.*mu + pow(yield_stress*HardeningModule(),2)*2./3.;
		inel_strain_update = (beta/xi/von_mises)*trial_stress;
		

		inel_strain_update_norm = sqrt(inel_strain_update.InnerProduct(inel_strain_update));
		hardening_param_update = sqrt(2./3.)*yield_stress*HardeningModule()*inel_strain_update_norm;
	}
	else
	{
		inel_strain_update.SetAll(0.);
		hardening_param_update = 0.;
	}

	switch (GetInelasticitySolutionMethod())
	{
	case ISM_FixedPoint:
		//return norm of inelastic strain update
		return inel_strain_update_norm;
	case ISM_ReturnMapping:
		//in this case return 0, since the update formula is substituted in the equations of motion (i.e., in EvalF2()), PostNewtonStep is just used for the final update of the old inelastic variables
		return 0.;
	}

	return 1e99;
}

// in 3D: compute the update of the inelastic variables as a function from the total strain and the old state of inelastic variables
double MaterialElastoplastic::ComputeInelasticVariablesUpdateFromStrain(Vector& inel_variables_update, const Vector& inel_variables, const Matrix3D& strain) const
{
	assert (inel_variables.Length() >= 7);
	Matrix3D inel_strain;
	StrainVectorToMatrix3D(inel_strain, inel_variables);
	double hardening_param = inel_variables(7);
	
	Matrix3D inel_strain_update;
	double hardening_param_update;
	double err = ComputeInelasticVariablesUpdateFromStrain(inel_strain_update, hardening_param_update, strain, inel_strain, hardening_param);
	Matrix3DToStrainVector(inel_variables_update, inel_strain_update);
	inel_variables_update(7) = hardening_param_update;
	return err;
}

// in 2D: save version: compute the update of the inelastic variables as a function from the total strain and the old state of inelastic variables
double MaterialElastoplastic::ComputeInelasticVariablesUpdateFromStrain2D(Vector& inelastic_variables_update, const Vector& inelastic_variables, const Matrix2D& strain) const
{
	double hardening_parameter;
	Matrix3D total_strain, old_plastic_strain;	
	Get3DStrainsTensors2D(total_strain, old_plastic_strain, hardening_parameter, strain, inelastic_variables);
	
	Matrix3D inel_strain_update;
	double hardening_param_update;
	double err = ComputeInelasticVariablesUpdateFromStrain(inel_strain_update, hardening_param_update, total_strain, old_plastic_strain, hardening_parameter);
	inelastic_variables_update(1) = inel_strain_update(1,1);
	inelastic_variables_update(2) = inel_strain_update(2,2);
	inelastic_variables_update(3) = 2.*inel_strain_update(1,2);
	inelastic_variables_update(4) = hardening_param_update;
	return err;
}

////$ PG 2017-3-10: original
//double MaterialElastoplastic::ComputeInelasticVariablesUpdateFromStrain2D(Vector& inelastic_variables_update, const Vector& inelastic_variables, const Matrix2D& strain) const
//{ // calculate the actual state of inelastic variables as a function of the total strain in the planar case
//	double mu = GetMu();
//	double inelastic_strain_update_norm = 0.;   // return value in case of ISM_FixedPoint
//
//	Matrix2D trial_stress;
//	ComputeStressFromStrain2D(strain, trial_stress, inelastic_variables, true);  // compute trial stress: sigma = C (eps - eps^p_old)
//	double von_mises = ComputeDeviator2D(trial_stress, inelastic_variables);
//	double yield_stress = YieldStress();
//	double xi = 2.*mu + Sqr(yield_stress*HardeningModule())*2./3.;
//	double hardening_param = inelastic_variables(4);
//
//	//bool converged = false;
//	//while (!converged)  // fixed point iteration  (in case of plane strain: one iteration leads to exact update)
//	//{
//	double beta = von_mises - yield_stress*(1 + HardeningModule()*hardening_param);
//	Matrix2D inelastic_strain_update;
//	if ( beta > 0. )   // new plasticity occurs
//	{
//		inelastic_strain_update = (beta/xi/von_mises)*trial_stress;  //only up to 2nd component need to be stored, 33 component of plastic strain satisfies tr(inel_strain_update) = 0
//#if 1 //PG 2015-05-12: accelerated but less stable convergence
//		if (InelasticitySolutionMethod() == ISM_FixedPoint)
//			inelastic_strain_update *= 1.8;
//#endif
//		Matrix2DToStrainVector(inelastic_variables_update, inelastic_strain_update);
//
//		inelastic_strain_update_norm = sqrt(inelastic_strain_update.NormSqr() + Sqr(inelastic_strain_update.Trace()));
//		inelastic_variables_update(4) = sqrt(2./3.)*yield_stress*HardeningModule()*inelastic_strain_update_norm;
//	}
//	else
//	{
//		inelastic_variables_update.SetAll(0.);
//	}
//
//	//	if(IsPlaneStrain())
//	//	{
//	//		converged = true;
//	//	}
//	//	else// plane stress
//	//	{
//	//		double plane_stress_correction = inelastic_strain_update.Trace()*(-GetLambdaPlaneStress()+2*mu)/3.;
//	//		trial_stress(1,1) += plane_stress_correction; 
//	//		trial_stress(2,2) += plane_stress_correction;
//	//		trial_stress -= 2*mu*inelastic_strain_update;
//	//		norm_trial_stress = sqrt(trial_stress.NormSqr() + Sqr(trial_stress.Trace()));
//	//		converged = (fabs(plane_stress_correction) < 1e-5*scaled_yield_stress);
//	//	}
//	//}
//
//	switch (GetInelasticitySolutionMethod())
//	{
//	case ISM_FixedPoint:
//		//return norm of inelastic strain update
//		return inelastic_strain_update_norm;
//	case ISM_ReturnMapping:
//		//in this case return 0, since the update formula is substituted in the equations of motion (i.e., in EvalF2()), PostNewtonStep is just used for the final update of the old inelastic variables
//		return 0.;
//	default:
//		return 1e99;
//	}
//
//	return 0.;
//}



// computes the consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// C [output]: consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// strain: (eps_xx, eps_yy, eps_zz, 2 eps_yz, 2 eps_zx, 2 eps_xy)
// inelastic_variables: (eps^p_xx, eps^p_yy, eps^p_zz, 2 eps^p_yz, 2 eps^p_zx, 2 eps^p_xy, hardening_parameter)
void MaterialElastoplastic::ComputeConsistentTangentStiffnessMatrix(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const
{
	// the consisten tangent stiffness matrix is returned in C, depending on strain and inelastic varibles (in integration point)
	assert(!IsPlanarMaterial() && "use ComputeTangentStiffnessMatrix2D for 2D case!");

	if (GetInelasticitySolutionMethod() == ISM_ReturnMapping)
	{
		ConstMatrix<6*6> dPdE(6,6,1), K(6,6), N(6,6);  // K & N initialized with 0, dPdE filled later
    
		N(1,1) = 1;
		N(2,2) = 1;
		N(3,3) = 1;
		N(4,4) = 2;
		N(5,5) = 2;
		N(6,6) = 2;

		K(1,1) = 1;
		K(2,2) = 1;
		K(3,3) = 1;
		K(4,4) = 1;
		K(5,5) = 1;
		K(6,6) = 1;
		double val = -1./3.;
		K(1,1) += val;
		K(1,2) += val;
		K(1,3) += val;
		K(2,1) += val;
		K(2,2) += val;
		K(2,3) += val;
		K(3,1) += val;
		K(3,2) += val;
		K(3,3) += val;

		ComputeElasticityMatrix(C);  //$ PG 2016-1-26:  might be redundant, since function shall already be called with correct C, check that!
				
		assert (inelastic_variables.Length() >= 7);
		ConstVector<6> inelastic_strain(6,1);
		inelastic_strain(1) = inelastic_variables(1);
		inelastic_strain(2) = inelastic_variables(2);
		inelastic_strain(3) = inelastic_variables(3);
		inelastic_strain(4) = 0.5*inelastic_variables(4);
		inelastic_strain(5) = 0.5*inelastic_variables(5);
		inelastic_strain(6) = 0.5*inelastic_variables(6);
		double hardening_parameter = inelastic_variables(7);

		double mu2 = 2*GetMu();
		ConstVector<6> dev_trial_stress = (ConstVector<6>&) (K*(C*strain) - mu2*inelastic_strain);

		double norm_dev_trial_stress_sq = dev_trial_stress*(N*dev_trial_stress);
		double norm_dev_trial_stress = sqrt(norm_dev_trial_stress_sq);
		double von_mises = norm_dev_trial_stress/sqrt(2./3.);
		double yield_stress = YieldStress();
		double beta = von_mises - yield_stress*(1 + hardening_parameter*HardeningModule());

		if (beta > 0)
		{
			dev_trial_stress *= 1./norm_dev_trial_stress;
			double xi =  mu2/(mu2 + Sqr(yield_stress*HardeningModule())*2./3.);
			beta /= von_mises;

			ConstVector<6> N_times_dev_trial_stress((ConstVector<6>&)(N*dev_trial_stress));
			dev_trial_stress *= xi*(1-beta);

			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				for (int j=1; j<=dPdE.Getcols(); j++)
				{
					dPdE(i,j) = dev_trial_stress(i) * N_times_dev_trial_stress(j);
				}
			}
			double beta_xi = beta*xi;
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				dPdE(i,i) += beta_xi;
			}
			C -= dPdE*K*C; //C = Celast - 2*mu*dPdE   (2*mu was already muptiplied in xi)
		}
	}
}

// computes the consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// C [output]: consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// strain: (eps_xx, eps_yy, 2 eps_xy)
// inelastic_variables: (eps^p_xx, eps^p_yy, 2 eps^p_xy, hardening_parameter)
void MaterialElastoplastic::ComputeConsistentTangentStiffnessMatrix2D(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const
{
	// the consistent tangent stiffness matrix is returned in C, depending on strain and inelastic varibles (in integration point)
	assert(IsPlanarMaterial() && "use ComputeTangentStiffnessMatrix for 3D case!");

	if (GetInelasticitySolutionMethod() == ISM_ReturnMapping)
	{
		ConstMatrix<3*3> dPdE(3,3,1), K(3,3), N(3,3);   // K & N initialized with 0, dPdE filled later
		N(1,1) = 2;
		N(2,2) = 2;
		N(3,3) = 2;
		N(1,2) = 1;
		N(2,1) = 1;

		K(1,1) = 1;
		K(2,2) = 1;
		K(3,3) = 1;
		double val = -(PoissonRatio() + 1.)/3.;
		K(1,1) += val;
		K(1,2) += val;
		K(2,2) += val;
		K(2,1) += val;

		ComputeElasticityMatrix(C);
				
		assert (inelastic_variables.Length() >= 4);
		ConstVector<3> inelastic_strain(3,1);
		inelastic_strain(1) = inelastic_variables(1);
		inelastic_strain(2) = inelastic_variables(2);
		inelastic_strain(3) = 0.5*inelastic_variables(3);
		double hardening_parameter = inelastic_variables(4);

		double mu2 = 2*GetMu();
		ConstVector<3> dev_trial_stress = (ConstVector<3>&) (K*(C*strain) - mu2*inelastic_strain);

		double norm_dev_trial_stress_sq = dev_trial_stress*(N*dev_trial_stress);
		double norm_dev_trial_stress = sqrt(norm_dev_trial_stress_sq);
		double von_mises = norm_dev_trial_stress/sqrt(2./3.);
		double yield_stress = YieldStress();
		double beta = von_mises - yield_stress*(1 + hardening_parameter*HardeningModule());

		if (beta > 0)
		{
			dev_trial_stress *= 1./norm_dev_trial_stress;
			double xi =  mu2/(mu2 + Sqr(yield_stress*HardeningModule())*2./3.);
			beta /= von_mises;

			ConstVector<3> N_times_dev_trial_stress((ConstVector<3>&)(N*dev_trial_stress));
			dev_trial_stress *= xi*(1-beta);
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				for (int j=1; j<=dPdE.Getcols(); j++)
				{
					dPdE(i,j) = dev_trial_stress(i) * N_times_dev_trial_stress(j);
				}
			}
			double beta_xi = beta*xi;
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				dPdE(i,i) += beta_xi;
			}
			C -= dPdE*K*C; //C = Celast - 2*mu*dPdE   (2*mu was already muptiplied in xi)
		}
	}
}

double MaterialElastoplastic::ComputeYieldFunction(const Matrix3D& strain, const Vector& inelastic_variables) const
{	
	Matrix3D inel_strain;
	StrainVectorToMatrix3D(inel_strain, inelastic_variables);
	Matrix3D trial_stress;
	Material::ComputeStressFromStrain(strain - inel_strain, trial_stress);
	double vonMises = ComputeDeviator(trial_stress);
	double hardening_param = inelastic_variables(7);
	double yield_function = vonMises - YieldStress()*(1 + HardeningModule()*hardening_param);
	
	return yield_function;
}

double MaterialElastoplastic::ComputeYieldFunction2D(const Matrix2D& strain, const Vector& inelastic_variables) const
{
	double mu = GetMu();

	Matrix2D trial_stress;
	ComputeStressFromStrain2D(strain, trial_stress, inelastic_variables, true); // compute trial stress: sigma = C (eps - eps^p_old)
	double strain_offset_33 = 0.;
	if (HasInitialStrains())
		strain_offset_33 += GetInitialStrain33();
	if (inelastic_variables.Length() > 1)
		strain_offset_33 -= inelastic_variables(1) + inelastic_variables(2);
	double vonMises = ComputeDeviator2D(trial_stress, strain_offset_33);
	double hardening_param = inelastic_variables(4);
	double yield_function = vonMises - YieldStress()*(1 + HardeningModule()*hardening_param);

	return yield_function;
}


void MaterialElastoplastic::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const
{
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_inelastic_strain, FieldVariableDescriptor::FVCI_y, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_hardening_parameter);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_equivalent_plastic_strain);
}


double MaterialElastoplastic::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const
{
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_inelastic_strain:
	case FieldVariableDescriptor::FVT_equivalent_plastic_strain:
		{
			Matrix3D inelastic_strain = 0;
			if (IsInelasticMaterial())
			{
				if (IsPlanarMaterial())
				{
					inelastic_strain(1,1) = inelastic_variables(1);
					inelastic_strain(2,2) = inelastic_variables(2);
					inelastic_strain(1,2) = inelastic_strain(1,2) = .5*inelastic_variables(3);
				}
				else
				{
					StrainVectorToMatrix3D(inelastic_strain, inelastic_variables);
				}
			}
			if (fvd.VariableType() == FieldVariableDescriptor::FVT_equivalent_plastic_strain)
			{
				return sqrt(2. / 3.*inelastic_strain.InnerProduct(inelastic_strain));
			}
			return fvd.GetComponent(inelastic_strain);
		}
	case FieldVariableDescriptor::FVT_hardening_parameter:
		{
			return inelastic_variables(GetInelasticVariablesCount());
		}
	}
	return FIELD_VARIABLE_NO_VALUE;
}




////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// function definitions of class ElastoplasticParametersEterovicBathe
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void ElastoplasticParametersEterovicBathe::DefaultConstructor()
{
	yieldstress = 0;
	tangentmodule = 0;
}
	
void ElastoplasticParametersEterovicBathe::CopyFrom(const ElastoplasticParametersEterovicBathe& m)
{
	yieldstress = m.yieldstress;
	tangentmodule = m.tangentmodule;
}

void ElastoplasticParametersEterovicBathe::SetElastoplasticParametersEterovicBathe(double yield_stressI, double tangent_moduleI)
{
	yieldstress = yield_stressI;
	tangentmodule = tangent_moduleI;
}

int ElastoplasticParametersEterovicBathe::CheckConsistency(mystr& errorstr)
{
	int rv = 0;

	if(yieldstress < 0)
	{
		errorstr += mystr("ERROR: yield stress must not be negative!\n");
		rv = max(1,rv);
	}
	if(tangentmodule < 0)
	{
		errorstr += mystr("ERROR: tangent module must not be negative!\n");
		rv = max(1,rv);
	}

	return rv;
}

int ElastoplasticParametersEterovicBathe::GetInelasticVariablesCount(int planar) const
{
	if (planar)
		return 4;     // 3*strain+1*hardening_parameter
		
	return 7;     // 6*strain+1*hardening_parameter
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// function definitions of class MaterialElastoplasticLargeStrain : public Material, public ElastoplasticParametersEterovicBathe
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void MaterialElastoplasticLargeStrain::DefaultConstructor()
{
	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
}

MaterialGeneric* MaterialElastoplasticLargeStrain::GetCopy() const
{
	MaterialGeneric* ec = new MaterialElastoplasticLargeStrain(*this);
	return ec;
}

void MaterialElastoplasticLargeStrain::CopyFrom(const MaterialElastoplasticLargeStrain& m)
{
	Material::CopyFrom(m);
	ElastoplasticParametersEterovicBathe::CopyFrom(m);
}

void MaterialElastoplasticLargeStrain::SetMaterialElastoplasticLargeStrain(double densityI, double youngs_modulusI, double poisson_ratioI,
																			Material::InelasticitySolutionMethod inelasticity_solution_methodI, double yield_stressI, double tangent_moduleI,
																			Vector3D color, int planeI, int planestressI)
//$ PG 2016-1-19: default set function
{
	Material::SetMaterial(densityI, youngs_modulusI, poisson_ratioI, color, planeI, planestressI);
	ElastoplasticParametersEterovicBathe::SetElastoplasticParametersEterovicBathe(yield_stressI, tangent_moduleI);
	SetInelasticitySolutionMethod(inelasticity_solution_methodI);
}
int MaterialElastoplasticLargeStrain::CheckConsistency(mystr& errorstr)
{
	int rv = max(Material::CheckConsistency(errorstr), ElastoplasticParametersEterovicBathe::CheckConsistency(errorstr));
	return rv;
}

int MaterialElastoplasticLargeStrain::IsInelasticMaterial() const
{
	return 1;
}

int MaterialElastoplasticLargeStrain::GetInelasticVariablesCount() const
{
	return ElastoplasticParametersEterovicBathe::GetInelasticVariablesCount(IsPlanarMaterial());
}

mystr MaterialElastoplasticLargeStrain::GetTypeName() const
{
	return ("MaterialElastoplasticLargeStrain");
}

double MaterialElastoplasticLargeStrain::GetEquivalentTensileStress(const Vector& inelastic_variables) const
{
	return inelastic_variables(1);
}
double MaterialElastoplasticLargeStrain::GetYieldStress(const Vector& inelastic_variables) const
{
	return inelastic_variables(2);
}

Matrix3D MaterialElastoplasticLargeStrain::GetPlasticStrainTensor(const Vector& inelastic_variables) const
{
	Matrix3D m;
	double m11 = inelastic_variables(3);
	double m22 = inelastic_variables(4);
	double m33 = inelastic_variables(5);
	double m12 = inelastic_variables(6);
	double m23 = inelastic_variables(7);
	double m31 = inelastic_variables(8);
	
	m.Set(m11,m12,m31,
		m12,m22,m23,
		m31,m23,m33);
	
	return m;
}

void MaterialElastoplasticLargeStrain::ComputeStressFromStrain(
									const Matrix3D& strain, Matrix3D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
	assert (inel_variables.Length() >= 7);
	//total and plastic strain, are indeed total deformation gradient and inverse plastic deformation gradient, respectively, stress is cauchy stress tensor!
	//material is inelastic (or plastic), but elastic law is linear (Hookes law)
	double mu = GetMu();                   // shear modulus
	double kappa = GetLambda() + 2.*mu/3.; // bluk modulus

	// a) obtain the trial elastic deformation gradient (strain... total strain, plastic_strain... already inverted)
	Matrix3D pstrain_inv = GetPlasticStrainTensor(inel_variables).Invert();
	Matrix3D Xe_trial = strain * pstrain_inv;
	
	// b) perform the polar decomposition
	// to obtain a polar decomposition X = R U (with X ... square matrix, R ... orthogonal matrix, U ... positive semidefinite symmetric matrix)
	// we first apply a singular value decomposition X = M Lambda N^T (M,N ... orthogonal, Lambda ... positive semidefinite diagonal matrix) via Lapack,
	// and finaly compute R = M N^T, U = N Lambda N^T
	Matrix3D M, N;
	Vector3D lambda;
	Xe_trial.SVD(M, lambda, N);
	Matrix3D NT = N.GetTp();
	Matrix3D NL(N);
	for(int j=1; j<=3; j++)
	{
		double val = lambda(j);
		for(int i=1; i<=3; i++)
		{
			NL(i,j) *= val;
		}
	}	
	Matrix3D Re_trial(M*NT);
	Matrix3D Ue_trial(NL*NT);
	
	// c) obtain the trial elastic strain tensor
	Matrix3D Ee_trial(Ue_trial.GetLog());

	// d) obtain the trial elastic stress tensor and the trial elastic effective stress
	double trial_rotated_stress_trace = 3*kappa*Ee_trial.Trace();
	Matrix3D trial_rotated_stress_deviator(Ee_trial);
	trial_rotated_stress_deviator *= 2.*mu;
	//trial_rotated_stress_deviator -= back_stress; //if kinematic hardening was considered

	double equivalent_tensile_stress = sqrt(trial_rotated_stress_deviator.NormSqr()*3./2.);

	
	if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
	{
		//$ PG 2017-11-14: large strain plasticity to be continued here


		//$ PG 2017-11-14: small strain plasticity:
		//Matrix3D trial_stress;
		//Material::ComputeStressFromStrain(s, trial_stress);
		//double von_mises = ComputeDeviator(trial_stress);
		//double yield_stress = YieldStress();

		//double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_parameter);
		//if ( beta > 0 )   // new plasticity occurs
		//{
		//	double xi = 2.*mu + pow(yield_stress*HardeningModule(), 2)*2./3.;
		//	s -= (beta/xi/von_mises)*trial_stress;
		//}
	}

	//Material::ComputeStressFromStrain(s, stress);
}

//$ PG 2017-3-7: save version: transfrom total and plastic strain to 3D tensors, compute stress using 3D functionality, and return upper 2x2 block of stress tensor
void MaterialElastoplasticLargeStrain::ComputeStressFromStrain2D(
									const Matrix2D& strain, Matrix2D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
//...to be implemented
}

void MaterialElastoplasticLargeStrain::Get3DStrainsTensors2D(Matrix3D& total_strain, Matrix3D& old_plastic_strain, double& hardening_parameter, const Matrix2D& strain, const Vector& inel_variables) const
{
  total_strain.Set(
    strain(1,1),strain(1,2),0,
    strain(2,1),strain(2,2),0,
    0,0,0);
  if (!IsPlaneStrain())
  {
    if (!IsOrthotropicMaterial())
      total_strain(3,3) = - (strain(1,1) + strain(2,2));
    else
      mbs->UO(UO_LVL_err) << "ERROR in MaterialElastoplasticLargeStrain: Case (!IsPlaneStrain() && IsOrthotropicMaterial()) not implemented!\n";
  }

  if (inel_variables.Length() >= 4)
  {
    hardening_parameter = inel_variables(4);
    double diag1 = inel_variables(1);
    double diag2 = inel_variables(2);
    double diag1plus2 = diag1 + diag2;
    double offdiag = 0.5*inel_variables(3);
    old_plastic_strain.Set(
      diag1, offdiag, 0,
      offdiag, diag2, 0,
      0, 0, -diag1plus2);  //$ PG 2017-3-9: old plastic strain must be tracefree

    if (!IsPlaneStrain())
    {
      if (!IsOrthotropicMaterial())
      {
        total_strain(3,3) += 2.*GetMu()/GetLambda()*diag1plus2;
      }
    }
  }
}

//$ PG 2017-3-7: tested, copy from 3D and modified (for plane strain only)
//void MaterialElastoplasticLargeStrain::ComputeStressFromStrain2D(
//									const Matrix2D& strain, Matrix2D& stress,
//									const Vector& inel_variables,
//									bool compute_trial_stress) const
//{
//	//material is inelastic (or plastic), but elastic law is linear (Hookes law)
//	assert (inel_variables.Length() >= 4);
//
//	Material::ComputeStressFromStrain2D(strain, stress);
//
//	Matrix2D old_plastic_strain;
//	StrainVectorToMatrix2D(old_plastic_strain, inel_variables);
//	double mu = GetMu();
//	stress -= 2*mu*old_plastic_strain;
//
//	if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
//	{
//		double hardening_param = inel_variables(4);
//		Matrix2D dev_stress(stress);
//		double von_mises = ComputeDeviator2D(dev_stress, inel_variables); //transforms stress into its deviator (dev_stress) and returns it in argument list (apart from retrun value, which is von Mises stress)!
//		double yield_stress = YieldStress();
//
//		double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_param);
//		if ( beta > 0 )   // new plasticity occurs
//		{
//			double xi = 2.*mu + pow(yield_stress*HardeningModule(), 2)*2./3.;
//			Matrix2D plastic_strain_update = (beta/xi/von_mises)*dev_stress;
//			stress -= 2*mu*plastic_strain_update;
//		}
//	}
//}

////$ PG 2017-3-7: original
//void MaterialElastoplasticLargeStrain::ComputeStressFromStrain2D(
//	const Matrix2D& strain, Matrix2D& stress,
//	const Vector& inelastic_variables, bool compute_trial_stress) const
//{
//	assert(IsPlanarMaterial());
//
//#ifdef PRINT_PLASTIC_FLOW_STATS
//	static int old_newton_its = 0;
//	static int active_plastic_flows = 0;
//	static int calls = 0;
//	if(!GetMBS()->IsJacobianComputation())
//		calls++;
//
//	if(old_newton_its != GetMBS()->NumSolver().GetNewtonIts())
//	{
//		old_newton_its = GetMBS()->NumSolver().GetNewtonIts();
//		GetMBS()->UO(UO_LVL_0) << "its = " << old_newton_its << ", active flows = " << active_plastic_flows << ", calls = " << calls << "\n";
//		active_plastic_flows = 0;
//		calls = 0;
//	}
//#endif
//
//
//	double mu = GetMu();
//	double la = GetLambda();
//	if (!IsPlaneStrain()) { la = GetLambdaPlaneStress(); }
//
//	if (!IsOrthotropicMaterial())
//	{
//		stress = ((2*mu)*(strain) + Matrix2D(la*strain.Trace()));
//	}
//	else
//	{
//		ConstMatrix<9> C;
//		this->ComputeElasticityMatrix(C);
//
//		Vector3D elastic_strain_vec(strain(1,1), strain(2,2), 2*strain(1,2));
//		Vector3D stress_vec = C*elastic_strain_vec;
//		stress(1,1) = stress_vec(1); stress(1,2) = stress_vec(3); 
//		stress(2,1) = stress_vec(3); stress(2,2) = stress_vec(2); 
//	}
//
//	if (inelastic_variables.Length() != 0)
//	{ 
//		Matrix2D inelastic_strain;
//		StrainVectorToMatrix2D(inelastic_strain, inelastic_variables);
//
//		stress -= (2*mu)*inelastic_strain;
//		if (!IsPlaneStrain())
//		{
//			double inelastic_strain_33 = -(inelastic_strain(1,1) + inelastic_strain(2,2));
//			double val = la * inelastic_strain_33;
//			stress(1,1) += val;
//			stress(2,2) += val;
//		}
//
//		if (!compute_trial_stress && GetInelasticitySolutionMethod() == ISM_ReturnMapping)   //use the updated instead of the old inelastic strain
//		{
//			Matrix2D trial_stress(stress);
//			double von_mises = ComputeDeviator2D(trial_stress, inelastic_variables);
//			double yield_stress = YieldStress();
//			double hardening_param = inelastic_variables(4);
//			double beta = von_mises - yield_stress*(1. + HardeningModule()*hardening_param);
//			if ( beta > 0. )   // new plasticity occurs
//			{
//#ifdef PRINT_PLASTIC_FLOW_STATS
//				if(!GetMBS()->IsJacobianComputation())
//					active_plastic_flows++;
//#endif
//				double xi = 2.*mu + Sqr(yield_stress*HardeningModule())*2./3.;
//				inelastic_strain = (beta/xi/von_mises)*trial_stress;  // inelastic strain update saved in matrix of inealstic strain
//			}
//
//			stress -= (2*mu)*inelastic_strain; 
//			if (!IsPlaneStrain())
//			{
//				double inelastic_strain_33 = -(inelastic_strain(1,1) + inelastic_strain(2,2));
//				double val = la * inelastic_strain_33;
//				stress(1,1) += val;
//				stress(2,2) += val;
//			}
//		}
//	}
//}


double MaterialElastoplasticLargeStrain::ComputeInelasticVariablesUpdateFromStrain(Matrix3D& inel_strain_update, double& hardening_param_update, const Matrix3D& strain, const Matrix3D& inel_strain, double hardening_param) const
{
	double inel_strain_update_norm = 0.;   // return value in case of ISM_FixedPoint
	double mu = GetMu();

	Matrix3D trial_stress;
	Material::ComputeStressFromStrain(strain-inel_strain, trial_stress);
	double von_mises = ComputeDeviator(trial_stress);  // returns vonMises via return value (double), and deviator of trial stress via Matrix3D& trial_stress
	double yield_stress = YieldStress();

	double beta = von_mises - yield_stress*(1 + HardeningModule()*hardening_param);
	if ( beta > 0. )   // new plasticity occurs
	{
		double xi = 2.*mu + pow(yield_stress*HardeningModule(),2)*2./3.;
		inel_strain_update = (beta/xi/von_mises)*trial_stress;
		

		inel_strain_update_norm = sqrt(inel_strain_update.InnerProduct(inel_strain_update));
		hardening_param_update = sqrt(2./3.)*yield_stress*HardeningModule()*inel_strain_update_norm;
	}
	else
	{
		inel_strain_update.SetAll(0.);
		hardening_param_update = 0.;
	}

	switch (GetInelasticitySolutionMethod())
	{
	case ISM_FixedPoint:
		//return norm of inelastic strain update
		return inel_strain_update_norm;
	case ISM_ReturnMapping:
		//in this case return 0, since the update formula is substituted in the equations of motion (i.e., in EvalF2()), PostNewtonStep is just used for the final update of the old inelastic variables
		return 0.;
	}

	return 1e99;
}

// in 3D: compute the update of the inelastic variables as a function from the total strain and the old state of inelastic variables
double MaterialElastoplasticLargeStrain::ComputeInelasticVariablesUpdateFromStrain(Vector& inel_variables_update, const Vector& inel_variables, const Matrix3D& strain) const
{
	assert (inel_variables.Length() >= 7);
	Matrix3D inel_strain;
	StrainVectorToMatrix3D(inel_strain, inel_variables);
	double hardening_param = inel_variables(7);
	
	Matrix3D inel_strain_update;
	double hardening_param_update;
	double err = ComputeInelasticVariablesUpdateFromStrain(inel_strain_update, hardening_param_update, strain, inel_strain, hardening_param);
	Matrix3DToStrainVector(inel_variables_update, inel_strain_update);
	inel_variables_update(7) = hardening_param_update;
	return err;
}

// in 2D: save version: compute the update of the inelastic variables as a function from the total strain and the old state of inelastic variables
double MaterialElastoplasticLargeStrain::ComputeInelasticVariablesUpdateFromStrain2D(Vector& inelastic_variables_update, const Vector& inelastic_variables, const Matrix2D& strain) const
{
	double hardening_parameter;
	Matrix3D total_strain, old_plastic_strain;	
	Get3DStrainsTensors2D(total_strain, old_plastic_strain, hardening_parameter, strain, inelastic_variables);
	
	Matrix3D inel_strain_update;
	double hardening_param_update;
	double err = ComputeInelasticVariablesUpdateFromStrain(inel_strain_update, hardening_param_update, total_strain, old_plastic_strain, hardening_parameter);
	inelastic_variables_update(1) = inel_strain_update(1,1);
	inelastic_variables_update(2) = inel_strain_update(2,2);
	inelastic_variables_update(3) = 2.*inel_strain_update(1,2);
	inelastic_variables_update(4) = hardening_param_update;
	return err;
}

////$ PG 2017-3-10: original
//double MaterialElastoplasticLargeStrain::ComputeInelasticVariablesUpdateFromStrain2D(Vector& inelastic_variables_update, const Vector& inelastic_variables, const Matrix2D& strain) const
//{ // calculate the actual state of inelastic variables as a function of the total strain in the planar case
//	double mu = GetMu();
//	double inelastic_strain_update_norm = 0.;   // return value in case of ISM_FixedPoint
//
//	Matrix2D trial_stress;
//	ComputeStressFromStrain2D(strain, trial_stress, inelastic_variables, true);  // compute trial stress: sigma = C (eps - eps^p_old)
//	double von_mises = ComputeDeviator2D(trial_stress, inelastic_variables);
//	double yield_stress = YieldStress();
//	double xi = 2.*mu + Sqr(yield_stress*HardeningModule())*2./3.;
//	double hardening_param = inelastic_variables(4);
//
//	//bool converged = false;
//	//while (!converged)  // fixed point iteration  (in case of plane strain: one iteration leads to exact update)
//	//{
//	double beta = von_mises - yield_stress*(1 + HardeningModule()*hardening_param);
//	Matrix2D inelastic_strain_update;
//	if ( beta > 0. )   // new plasticity occurs
//	{
//		inelastic_strain_update = (beta/xi/von_mises)*trial_stress;  //only up to 2nd component need to be stored, 33 component of plastic strain satisfies tr(inel_strain_update) = 0
//#if 1 //PG 2015-05-12: accelerated but less stable convergence
//		if (InelasticitySolutionMethod() == ISM_FixedPoint)
//			inelastic_strain_update *= 1.8;
//#endif
//		Matrix2DToStrainVector(inelastic_variables_update, inelastic_strain_update);
//
//		inelastic_strain_update_norm = sqrt(inelastic_strain_update.NormSqr() + Sqr(inelastic_strain_update.Trace()));
//		inelastic_variables_update(4) = sqrt(2./3.)*yield_stress*HardeningModule()*inelastic_strain_update_norm;
//	}
//	else
//	{
//		inelastic_variables_update.SetAll(0.);
//	}
//
//	//	if(IsPlaneStrain())
//	//	{
//	//		converged = true;
//	//	}
//	//	else// plane stress
//	//	{
//	//		double plane_stress_correction = inelastic_strain_update.Trace()*(-GetLambdaPlaneStress()+2*mu)/3.;
//	//		trial_stress(1,1) += plane_stress_correction; 
//	//		trial_stress(2,2) += plane_stress_correction;
//	//		trial_stress -= 2*mu*inelastic_strain_update;
//	//		norm_trial_stress = sqrt(trial_stress.NormSqr() + Sqr(trial_stress.Trace()));
//	//		converged = (fabs(plane_stress_correction) < 1e-5*scaled_yield_stress);
//	//	}
//	//}
//
//	switch (GetInelasticitySolutionMethod())
//	{
//	case ISM_FixedPoint:
//		//return norm of inelastic strain update
//		return inelastic_strain_update_norm;
//	case ISM_ReturnMapping:
//		//in this case return 0, since the update formula is substituted in the equations of motion (i.e., in EvalF2()), PostNewtonStep is just used for the final update of the old inelastic variables
//		return 0.;
//	default:
//		return 1e99;
//	}
//
//	return 0.;
//}



// computes the consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// C [output]: consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// strain: (eps_xx, eps_yy, eps_zz, 2 eps_yz, 2 eps_zx, 2 eps_xy)
// inelastic_variables: (eps^p_xx, eps^p_yy, eps^p_zz, 2 eps^p_yz, 2 eps^p_zx, 2 eps^p_xy, hardening_parameter)
void MaterialElastoplasticLargeStrain::ComputeConsistentTangentStiffnessMatrix(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const
{
	// the consisten tangent stiffness matrix is returned in C, depending on strain and inelastic varibles (in integration point)
	assert(!IsPlanarMaterial() && "use ComputeTangentStiffnessMatrix2D for 2D case!");

	if (GetInelasticitySolutionMethod() == ISM_ReturnMapping)
	{
		ConstMatrix<6*6> dPdE(6,6,1), K(6,6), N(6,6);  // K & N initialized with 0, dPdE filled later
    
		N(1,1) = 1;
		N(2,2) = 1;
		N(3,3) = 1;
		N(4,4) = 2;
		N(5,5) = 2;
		N(6,6) = 2;

		K(1,1) = 1;
		K(2,2) = 1;
		K(3,3) = 1;
		K(4,4) = 1;
		K(5,5) = 1;
		K(6,6) = 1;
		double val = -1./3.;
		K(1,1) += val;
		K(1,2) += val;
		K(1,3) += val;
		K(2,1) += val;
		K(2,2) += val;
		K(2,3) += val;
		K(3,1) += val;
		K(3,2) += val;
		K(3,3) += val;

		ComputeElasticityMatrix(C);  //$ PG 2016-1-26:  might be redundant, since function shall already be called with correct C, check that!
				
		assert (inelastic_variables.Length() >= 7);
		ConstVector<6> inelastic_strain(6,1);
		inelastic_strain(1) = inelastic_variables(1);
		inelastic_strain(2) = inelastic_variables(2);
		inelastic_strain(3) = inelastic_variables(3);
		inelastic_strain(4) = 0.5*inelastic_variables(4);
		inelastic_strain(5) = 0.5*inelastic_variables(5);
		inelastic_strain(6) = 0.5*inelastic_variables(6);
		double hardening_parameter = inelastic_variables(7);

		double mu2 = 2*GetMu();
		ConstVector<6> dev_trial_stress = (ConstVector<6>&) (K*(C*strain) - mu2*inelastic_strain);

		double norm_dev_trial_stress_sq = dev_trial_stress*(N*dev_trial_stress);
		double norm_dev_trial_stress = sqrt(norm_dev_trial_stress_sq);
		double von_mises = norm_dev_trial_stress/sqrt(2./3.);
		double yield_stress = YieldStress();
		double beta = von_mises - yield_stress*(1 + hardening_parameter*HardeningModule());

		if (beta > 0)
		{
			dev_trial_stress *= 1./norm_dev_trial_stress;
			double xi =  mu2/(mu2 + Sqr(yield_stress*HardeningModule())*2./3.);
			beta /= von_mises;

			ConstVector<6> N_times_dev_trial_stress((ConstVector<6>&)(N*dev_trial_stress));
			dev_trial_stress *= xi*(1-beta);

			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				for (int j=1; j<=dPdE.Getcols(); j++)
				{
					dPdE(i,j) = dev_trial_stress(i) * N_times_dev_trial_stress(j);
				}
			}
			double beta_xi = beta*xi;
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				dPdE(i,i) += beta_xi;
			}
			C -= dPdE*K*C; //C = Celast - 2*mu*dPdE   (2*mu was already muptiplied in xi)
		}
	}
}

// computes the consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// C [output]: consistent tangent stiffness matrix C (I - d actual_plastic_strain/d strain)
// strain: (eps_xx, eps_yy, 2 eps_xy)
// inelastic_variables: (eps^p_xx, eps^p_yy, 2 eps^p_xy, hardening_parameter)
void MaterialElastoplasticLargeStrain::ComputeConsistentTangentStiffnessMatrix2D(Matrix& C, const Vector& strain, const Vector& inelastic_variables) const
{
	// the consistent tangent stiffness matrix is returned in C, depending on strain and inelastic varibles (in integration point)
	assert(IsPlanarMaterial() && "use ComputeTangentStiffnessMatrix for 3D case!");

	if (GetInelasticitySolutionMethod() == ISM_ReturnMapping)
	{
		ConstMatrix<3*3> dPdE(3,3,1), K(3,3), N(3,3);   // K & N initialized with 0, dPdE filled later
		N(1,1) = 2;
		N(2,2) = 2;
		N(3,3) = 2;
		N(1,2) = 1;
		N(2,1) = 1;

		K(1,1) = 1;
		K(2,2) = 1;
		K(3,3) = 1;
		double val = -(PoissonRatio() + 1.)/3.;
		K(1,1) += val;
		K(1,2) += val;
		K(2,2) += val;
		K(2,1) += val;

		ComputeElasticityMatrix(C);
				
		assert (inelastic_variables.Length() >= 4);
		ConstVector<3> inelastic_strain(3,1);
		inelastic_strain(1) = inelastic_variables(1);
		inelastic_strain(2) = inelastic_variables(2);
		inelastic_strain(3) = 0.5*inelastic_variables(3);
		double hardening_parameter = inelastic_variables(4);

		double mu2 = 2*GetMu();
		ConstVector<3> dev_trial_stress = (ConstVector<3>&) (K*(C*strain) - mu2*inelastic_strain);

		double norm_dev_trial_stress_sq = dev_trial_stress*(N*dev_trial_stress);
		double norm_dev_trial_stress = sqrt(norm_dev_trial_stress_sq);
		double von_mises = norm_dev_trial_stress/sqrt(2./3.);
		double yield_stress = YieldStress();
		double beta = von_mises - yield_stress*(1 + hardening_parameter*HardeningModule());

		if (beta > 0)
		{
			dev_trial_stress *= 1./norm_dev_trial_stress;
			double xi =  mu2/(mu2 + Sqr(yield_stress*HardeningModule())*2./3.);
			beta /= von_mises;

			ConstVector<3> N_times_dev_trial_stress((ConstVector<3>&)(N*dev_trial_stress));
			dev_trial_stress *= xi*(1-beta);
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				for (int j=1; j<=dPdE.Getcols(); j++)
				{
					dPdE(i,j) = dev_trial_stress(i) * N_times_dev_trial_stress(j);
				}
			}
			double beta_xi = beta*xi;
			for (int i=1; i<=dPdE.Getrows(); i++)
			{
				dPdE(i,i) += beta_xi;
			}
			C -= dPdE*K*C; //C = Celast - 2*mu*dPdE   (2*mu was already muptiplied in xi)
		}
	}
}

double MaterialElastoplasticLargeStrain::ComputeYieldFunction(const Matrix3D& strain, const Vector& inelastic_variables) const
{	
	Matrix3D inel_strain;
	StrainVectorToMatrix3D(inel_strain, inelastic_variables);
	Matrix3D trial_stress;
	Material::ComputeStressFromStrain(strain - inel_strain, trial_stress);
	double vonMises = ComputeDeviator(trial_stress);
	double hardening_param = inelastic_variables(7);
	double yield_function = vonMises - YieldStress()*(1 + HardeningModule()*hardening_param);
	
	return yield_function;
}

double MaterialElastoplasticLargeStrain::ComputeYieldFunction2D(const Matrix2D& strain, const Vector& inelastic_variables) const
{
	double mu = GetMu();

	Matrix2D trial_stress;
	ComputeStressFromStrain2D(strain, trial_stress, inelastic_variables, true); // compute trial stress: sigma = C (eps - eps^p_old)
	double strain_offset_33 = GetInitialStrain33();
	if (inelastic_variables.Length() > 1)
		strain_offset_33 -= inelastic_variables(1) + inelastic_variables(2);
	double vonMises = ComputeDeviator2D(trial_stress, strain_offset_33);
	double hardening_param = inelastic_variables(4);
	double yield_function = vonMises - YieldStress()*(1 + HardeningModule()*hardening_param);

	return yield_function;
}


void MaterialElastoplasticLargeStrain::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables) const
{
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_inelastic_strain, FieldVariableDescriptor::FVCI_y, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_hardening_parameter);
}


double MaterialElastoplasticLargeStrain::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector& inelastic_variables) const
{
	switch(fvd.VariableType())
	{
	case FieldVariableDescriptor::FVT_inelastic_strain:
		{
			Matrix3D inelastic_strain = 0;
			if (IsInelasticMaterial())
			{
				if (IsPlanarMaterial())
				{
					inelastic_strain(1,1) = inelastic_variables(1);
					inelastic_strain(2,2) = inelastic_variables(2);
					inelastic_strain(1,2) = inelastic_strain(1,2) = .5*inelastic_variables(3);
				}
				else
				{
					StrainVectorToMatrix3D(inelastic_strain, inelastic_variables);
				}
			}
			return fvd.GetComponent(inelastic_strain);
		}
	case FieldVariableDescriptor::FVT_hardening_parameter:
		{
			return inelastic_variables(GetInelasticVariablesCount());
		}
	}
	return FIELD_VARIABLE_NO_VALUE;
}





////////////////////////////////////////////////////////////////////////
// function definitions of class MaterialNonlinElasticSimoHughes : public Material
////////////////////////////////////////////////////////////////////////

void MaterialNonlinElasticSimoHughes::DefaultConstructor()
{
	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
}

mystr MaterialNonlinElasticSimoHughes::GetTypeName() const
{
	return mystr("MaterialNonlinElasticSimoHughes");
}

MaterialGeneric* MaterialNonlinElasticSimoHughes::GetCopy() const
{
	MaterialGeneric* ec = new MaterialNonlinElasticSimoHughes(*this);
	return ec;
}

void MaterialNonlinElasticSimoHughes::ComputeStressFromStrain(
									const Matrix3D& strain, Matrix3D& stress,
									const Vector& inel_variables,
									bool compute_trial_stress) const
{
	//material is nonlinear elastic (something different from Hookes law), but not inelastic (eg., no strain splitting)

	Matrix3D C = 2.*strain; //C = 2E+I = F^T F
	C(1,1)+=1.; C(2,2)+=1.; C(3,3)+=1.;
	if (!C.Invert()) {GetMBS()->UO() << "ERROR: cannot invert strain tensor in Material\n";}

	//PG: F not needed in the interface
	// since J^2 = det(F)^2 = det(F^T F) = det(C)
	double J_square = C.Det();
	//double J = gradient_F.Det();

	double la = GetLambda();
	double mu = GetMu();
	stress = 0.5*la*(J_square - 1.)*C + mu*(Matrix3D(1.)-C); //C is inverse of C !!!!
}

void MaterialNonlinElasticSimoHughes::ComputeStressFromStrain2D(
	const Matrix2D& strain, Matrix2D& stress,
	const Vector& inelastic_variables, bool compute_trial_stress) const
{
	assert(IsPlanarMaterial());

	double mu = GetMu();
	double la = IsPlaneStrain() ? GetLambda() : GetLambdaPlaneStress();

	Matrix2D C = 2.*strain; //C = 2E+I = F^T F
	C(1,1)+=1.; C(2,2)+=1.;
	if (!C.Invert()) {GetMBS()->UO() << "ERROR: cannot invert strain tensor in Material\n";}

	//PG: F not needed in the interfacem, since J^2 = det(F)^2 = det(F^T F) = det(C)
	double J_square = C.Det();
	//double J = gradient_F.Det();

	stress = 0.5*la*(J_square - 1.)*C + mu*(Matrix2D(1.)-C); //C is inverse of C !!!!
}




















/////////////////////////////////////////////////////
// function definitions of class BeamCrossSectionInfo
/////////////////////////////////////////////////////

BeamCrossSectionInfo::BeamCrossSectionInfo()
{
	beamCrossSectionShape = TCSS_rectangular;
	beamCrossSectionSize.SetLen(2);
	beamCrossSectionSize.SetAll(0);
	beamCrossSectionArea = 0.;
	beamIx = 0.;
	beamIy = 0.;
	beamIz = 0.;
}

BeamCrossSectionInfo::BeamCrossSectionInfo(const BeamCrossSectionInfo& b)
{
	CopyFrom(b);
}

void BeamCrossSectionInfo::CopyFrom(const BeamCrossSectionInfo& b)
{
	beamCrossSectionShape = b.beamCrossSectionShape;
	beamCrossSectionSize = b.beamCrossSectionSize;
	beamCrossSectionArea = b.beamCrossSectionArea;
	beamIx = b.beamIx;
	beamIy = b.beamIy;
	beamIz = b.beamIz;
}

void BeamCrossSectionInfo::SetBeamCrossSectionInfo(int _beamCrossSectionShape, const Vector& _beamCrossSectionSize)
{
	if (_beamCrossSectionShape > TCSS_NONE && _beamCrossSectionShape < TCSS_MAX)
	{
		beamCrossSectionShape = (TCrossSectionShape) _beamCrossSectionShape; /*1: rectangular, 2: circular, 3: tubular, 4: polygonal*/ //$ AD 2015-05-07 removed variable 'beamCrossSectionType'
		beamCrossSectionSize = _beamCrossSectionSize; //Vector of size 1 if circular, 2 if rectangular, and 2*n if polyognal (n... number of nodes: [p1y,p1z,p2y,p2z,...,pny,pnz])
		ComputeCrossSectionArea();
	}
}

int BeamCrossSectionInfo::CheckConsistency(mystr& errorstr)
{
	int rv = 0;
	if (!(beamCrossSectionShape > TCSS_NONE && beamCrossSectionShape < TCSS_MAX))
	{
		errorstr += mystr("ERROR in BeamCrossSectionInfo: beamCrossSectionShape has to be specified appropriately!\n");
		rv = max(rv, 2);
	}
	if (beamCrossSectionShape == TCSS_rectangular && beamCrossSectionSize.Length() != 2)
	{
		errorstr += mystr("ERROR in BeamCrossSectionInfo: rectangular cross section requires 2 cross section size components!\n");
		rv = max(rv, 2);
	}
	if (beamCrossSectionShape == TCSS_circular && beamCrossSectionSize.Length() != 1)
	{
		errorstr += mystr("ERROR in BeamCrossSectionInfo: circular cross section requires 1 cross section size component!\n");
		rv = max(rv, 2);
	}
	if (beamCrossSectionShape == TCSS_tubular && beamCrossSectionSize.Length() != 2)
	{
		errorstr += mystr("ERROR in BeamCrossSectionInfo: tubular cross section requires 2 cross section size components!\n");
		rv = max(rv, 2);
	}
	if (beamCrossSectionShape == TCSS_polygonal && beamCrossSectionSize.Length()%2 != 0)
	{
		errorstr += mystr("ERROR in BeamCrossSectionInfo: polygonal cross section requires 2*n cross section size components!\n");
		rv = max(rv, 2);
	}

	return rv;
}

int BeamCrossSectionInfo::IsRectangularCrossSection() const 
{
	return beamCrossSectionShape == TCSS_rectangular;
}

int BeamCrossSectionInfo::IsCircularCrossSection() const 
{
	return beamCrossSectionShape == TCSS_circular;
}

int BeamCrossSectionInfo::IsTubularCrossSection() const 
{
	return beamCrossSectionShape == TCSS_tubular;
}

double BeamCrossSectionInfo::GetCrossSectionArea() const
{
	return beamCrossSectionArea;
}

int BeamCrossSectionInfo::ComputeCrossSectionArea()  //$ PG 2016-1-25: returns 1 on success, 0 on error
{
	switch(beamCrossSectionShape)
	{
	case TCSS_rectangular: // rectangular    A = h w
		{
			double h = beamCrossSectionSize(1); // cross section length in direction of e_2
			double b = beamCrossSectionSize(2); // cross section length in direction of e_3
			beamCrossSectionArea = fabs(h*b);
			
			// computation of area/polar moment of inertia
			// see literature, e.g., http://materiales.azc.uam.mx/gjl/Clases/MA10_I/Roark's%20formulas%20for%20stress%20and%20strain.pdf
			beamIy = Cub(h)*b/12.;
			beamIz = Cub(b)*h/12.;
			TArray<double> data_points(10), data_values(10);
			data_points(1)  = 1.0;    data_values(1)  = 0.141;
			data_points(2)  = 1.5;    data_values(2)  = 0.196;
			data_points(3)  = 2.0;    data_values(3)  = 0.229;
			data_points(4)  = 2.5;    data_values(4)  = 0.249;
			data_points(5)  = 3.0;    data_values(5)  = 0.263;
			data_points(6)  = 4.0;    data_values(6)  = 0.281;
			data_points(7)  = 5.0;    data_values(7)  = 0.291;
			data_points(8)  = 6.0;    data_values(8)  = 0.299;
			data_points(9)  = 1.0e1;  data_values(9)  = 0.312;
			data_points(10) = 1.0e10; data_values(10) = 0.333;
			ListInterpolator kx(data_points, data_values);
			double longside = max(h,b);
			double shortside = min(h,b);
			beamIx = kx.GetValueAtPoint(longside/shortside)*longside*Cub(shortside); // see 1) Advanced Strength and Applied Elasticity, Ugural & Fenster, Elsevier, ISBN 0-444-00160-3, or 2) Roark's Formulas for stress & Strain, 7th Edition, Warren C. Young & Richard G. Budynas
			return 1; //OK
		}
	case TCSS_circular: // circular       A = r^2 pi
		{
			//$EK 2015-11 cross section size contains the diameter
			beamCrossSectionArea = Sqr(beamCrossSectionSize(1)/2.)*MY_PI;
			beamIx = MY_PI*Sqr(Sqr(GetRadius()))/2.;
			beamIy = beamIx/2.;
			beamIz = beamIy;
			return 1; //OK
		}
	case TCSS_tubular: // tubular        A = (ra^2-ri^2) pi
		{
			//$EK 2015-11 cross section size contains the diameter
			beamCrossSectionArea = (Sqr(beamCrossSectionSize(1)/2.)-Sqr(beamCrossSectionSize(2)/2.))*MY_PI;
			beamIx = MY_PI*(Sqr(Sqr(GetRadius())) - Sqr(Sqr(GetInnerRadius())))/2.;
			beamIy = beamIx/2.;
			beamIz = beamIy;
			return 1; //OK
		}
	case TCSS_polygonal: // polygonal     2A = |\sum_i={1}^n (x_i - x_{i+1})(y_i + y_{i+1})|    with x_{n+1}:=x_{1} and y_{n+1}:=y_{1} 
		{
			int n = beamCrossSectionSize.Length()/2;
			double result = (beamCrossSectionSize(2*n-1) - beamCrossSectionSize(1))*(beamCrossSectionSize(2*n) + beamCrossSectionSize(2));
			for (int i=1; i<=n-1; i++)
			{				
				int iy = 2*i; 
				int ix = iy-1;
				result += (beamCrossSectionSize(ix) - beamCrossSectionSize(ix+2))*(beamCrossSectionSize(iy) + beamCrossSectionSize(iy+2));
			}
			beamCrossSectionArea = fabs(result)/2;
			beamIx = 0; // to be implemented!
			beamIy = 0; // to be implemented!
			beamIz = 0; // to be implemented!
			return 1; //OK
		}
	}

	return 0; //ERROR
}



//////////////////////////////////////////////////////////////////////////
// function definitions of class Beam3DProperties : public MaterialGeneric
//////////////////////////////////////////////////////////////////////////

mystr Beam3DProperties::GetTypeName() const 
{
	return "Beam3DProperties";
}

void Beam3DProperties::DefaultConstructor()
{
	density = 0;   //$ PG 2016-1-25: depricated, but used in some old models

	//beamA = 0;    //$ PG 2016-1-25: depricated
	beamRhoA = -1;
	beamRhoIx = -1;
	beamRhoIy = -1;
	beamRhoIz = -1;
	beamEA = -1;
	beamEIy = -1;
	beamEIz = -1;
	beamGAky = -1;
	beamGAkz = -1;
	beamGJkx = -1;

	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
}

MaterialGeneric* Beam3DProperties::GetCopy() const
{
	MaterialGeneric* ec = new Beam3DProperties(*this);
	return ec;
}

void Beam3DProperties::CopyFrom(const Beam3DProperties& s)
{
	MaterialGeneric::CopyFrom(s);
	BeamCrossSectionInfo::CopyFrom(s);
	ElasticParametersLinear::CopyFrom(s);
	ElastoplasticParametersPrandtlReuss::CopyFrom(s);

	//beamA = s.beamA;   //$ PG 2016-1-25: depricated
	beamRhoA = s.beamRhoA;
	beamRhoIx = s.beamRhoIx;
	beamRhoIy = s.beamRhoIy;
	beamRhoIz = s.beamRhoIz;
	beamEA = s.beamEA;
	beamEIy = s.beamEIy;
	beamEIz = s.beamEIz;
	beamGAky = s.beamGAky;
	beamGAkz = s.beamGAkz;
	beamGJkx = s.beamGJkx;
}

void Beam3DProperties::SetMaterialFromEDCData()    //set material type and material parameters by user specified EDC variables
{
	//$ PG 2018-3-25: [ old way of initialization
	//// do not change anything
	//SetBeam3DProperties(beamCrossSectionShape, beamCrossSectionSize, density, beamEA, beamEIy, beamEIz, beamGJkx, beamGAky, beamGAkz, beamRhoA, beamRhoIx, beamRhoIy, beamRhoIz);
	//$ PG 2018-3-25: ]

	//$ PG 2018-3-25: [ if youngs modulus is set, we would like to overwrite shear/bending/torsional stiffnesses and moments of inertia if not defined
	ComputeCrossSectionArea();
	if (YoungsModulus()>0)
	{
		double E = YoungsModulus();  // [N/m^2]
		double nu = PoissonRatio();  // []
		double G = E/(2.+2.*nu);     // [N/m^2]
		double ky = 1;               // []
		if (IsRectangularCrossSection())
			ky = 10*(1+nu)/(12+11*nu); // see, e.g. http://iopscience.iop.org/article/10.1088/0022-3727/10/11/009/meta
		else if (IsCircularCrossSection() || IsTubularCrossSection())
			ky = (6+12*nu+6*Sqr(nu))/(7+12*nu+4*Sqr(nu)); // see, e.g. http://iopscience.iop.org/article/10.1088/0022-3727/10/11/009/meta  
		else
			mbs->UO(UO_LVL_warn) << "WARNING: no default shear correction factors available in case beamCrossSectionShape = " << beamCrossSectionShape << "\n";

		if (beamEA == -1) beamEA = E*beamCrossSectionArea; // [N]
		if (beamEIy == -1) beamEIy = E*beamIy; // [N m^2]
		if (beamEIz == -1) beamEIz = E*beamIz; // [N m^2]
		if (beamGJkx == -1) beamGJkx = G*beamIx;   // [N m^2]
		if (beamGAky == -1) beamGAky = G*beamCrossSectionArea*ky;   // [N m^2]
		if (beamGAkz == -1) beamGAkz = beamGAky; // [N m^2]
	}
	if (Density()>0)
	{
		double rho = Density();
		if (beamRhoA == -1) beamRhoA = rho*beamCrossSectionArea; //[kg/m]
		if (beamRhoIx == -1) beamRhoIx = rho*beamIx;
		if (beamRhoIy == -1) beamRhoIy = rho*beamIy;
		if (beamRhoIz == -1) beamRhoIz = rho*beamIz;
	}
	//$ PG 2018-3-25: ]

	//$ PG 2016-1-25: [ commented out, since Beam3DProperties now inherits from Material, and not MaterialElastoplastic
	//inelasticity_type = GetInelasticityType(inelasticity_type_str);
	//inelasticity_solution_method = GetInelasticitySolutionMethod(inelasticity_solution_method_str);
	//$ PG 2016-1-25: ]
}

//$ PG 2016-2-11: recommended set function
void Beam3DProperties::SetBeam3DProperties(int _beamCrossSectionShape /*1: rectangular, 2: circular, 3: tubular, 4: polygonal*/, const Vector& _beamCrossSectionSize, double _rho, double _EA, double _EIy, double _EIz, double _GJkx, double _GAky, double _GAkz, double _rhoA, double _rhoIx, double _rhoIy, double _rhoIz)
{
	SetBeamCrossSectionInfo(_beamCrossSectionShape, _beamCrossSectionSize);

	Density() = _rho;
	BeamEA() = _EA;
	BeamEIy() = _EIy;
	BeamEIz() = _EIz;
	BeamGJkx()= _GJkx;
	BeamGAky()= _GAky;
	BeamGAkz()= _GAkz;
	BeamRhoA() = _rhoA;
	BeamRhoIx() = _rhoIx;
	BeamRhoIy() = _rhoIy;
	BeamRhoIz() = _rhoIz;
	
	if (Density() == 0)
	{
		double A = GetCrossSectionArea();
		if (A > 0)
		{
			Density() = BeamRhoA()/A;
		}
	}
}

//$ PG 2016-2-11: recommended set function for continuum mechanics formulation
void Beam3DProperties::SetBeam3DPropertiesCM(int _beamCrossSectionShape /*1: rectangular, 2: circular, 3: tubular, 4: polygonal*/, const Vector& _beamCrossSectionSize, double _rho, double E, double nu, double _EA, double _EIy, double _EIz, double _GJkx, double _GAky, double _GAkz, double _rhoA, double _rhoIx, double _rhoIy, double _rhoIz)
{
	SetElasticParametersLinear(E,nu);
	SetBeam3DProperties(_beamCrossSectionShape, _beamCrossSectionSize, _rho, _EA, _EIy, _EIz, _GJkx, _GAky, _GAkz, _rhoA, _rhoIx, _rhoIy, _rhoIz);
}

//JG, depricated, because it does not initialize cross section size and shape, and it does not specify density from rhoA and cross section size
void Beam3DProperties::SetBeam3DProperties(double _rho, double _EA, double _EIy, double _EIz, double _GJkx, double _GAky, double _GAkz, double _rhoA, double _rhoIx, double _rhoIy, double _rhoIz)
{
	Density() = _rho;
	BeamEA() = _EA;
	BeamEIy() = _EIy;
	BeamEIz() = _EIz;
	BeamGJkx()= _GJkx;
	BeamGAky()= _GAky;
	BeamGAkz()= _GAkz;
	BeamRhoA() = _rhoA;
	BeamRhoIx() = _rhoIx;
	BeamRhoIy() = _rhoIy;
	BeamRhoIz() = _rhoIz;
}

int Beam3DProperties::CheckConsistency(mystr& errorstr)
{
	int rv = MaterialGeneric::CheckConsistency(errorstr);
	rv = max(rv, BeamCrossSectionInfo::CheckConsistency(errorstr));
	rv = max(rv, ElasticParametersLinear::CheckConsistency(errorstr));
	rv = max(rv, ElastoplasticParametersPrandtlReuss::CheckConsistency(errorstr));

	if (beamRhoA <= 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: rhoA is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	if (beamRhoIx <= 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: beamRhoIx is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	if (beamRhoIy <= 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: beamRhoIy is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	if (beamRhoIz <= 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: beamRhoIz is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	return rv;
}




//////////////////////////////////////////////////////////////////////////
// function definitions of class Beam2DProperties : public MaterialGeneric
//////////////////////////////////////////////////////////////////////////

mystr Beam2DProperties::GetTypeName() const 
{
	return "Beam2DProperties";
}

void Beam2DProperties::DefaultConstructor()
{
	//beamA = 0;   //$ PG 2016-1-25: depricated
	beamRhoA = -1;
	beamRhoIx = -1;
	beamRhoIy = -1;
	beamEA = -1;
	beamEIy = -1;
	beamGAky = -1;
	beamGJkx = -1;

	material_type = GetTypeName(); 
	materialname = GetTypeName();  // just for initialization, to be overwritten by the user
}

MaterialGeneric* Beam2DProperties::GetCopy() const
{
	MaterialGeneric* ec = new Beam2DProperties(*this);
	return ec;
}

void Beam2DProperties::CopyFrom(const Beam2DProperties& s)
{
	MaterialGeneric::CopyFrom(s);
	BeamCrossSectionInfo::CopyFrom(s);
	ElasticParametersLinear::CopyFrom(s);
	ElastoplasticParametersPrandtlReuss::CopyFrom(s);

	//beamA = s.beamA;   //$ PG 2016-1-25: depricated
	beamRhoA = s.beamRhoA;
	beamRhoIx = s.beamRhoIx;
	beamRhoIy = s.beamRhoIy;
	beamEA = s.beamEA;
	beamEIy = s.beamEIy;
	beamGAky = s.beamGAky;
	beamGJkx = s.beamGJkx;
}

int Beam2DProperties::CheckConsistency(mystr& errorstr)
{
	int rv = MaterialGeneric::CheckConsistency(errorstr);
	rv = max(rv, BeamCrossSectionInfo::CheckConsistency(errorstr));
	rv = max(rv, ElasticParametersLinear::CheckConsistency(errorstr));
	rv = max(rv, ElastoplasticParametersPrandtlReuss::CheckConsistency(errorstr));

	if (beamRhoA < 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: rhoA is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	if (beamRhoIx < 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: beamRhoIx is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	if (beamRhoIy < 0 && mbs->GetSolSet().dostaticcomputation == 0)
	{
		errorstr += mystr("WARNING: beamRhoIy is non-positive in dynamic computation\n");
		rv = max(rv, 0);
	}
	return rv;
}

void Beam2DProperties::SetMaterialFromEDCData()    //set material type and material parameters by user specified EDC variables
{
	//$ PG 2018-3-25: [ old way of initialization
	//// do not change anything
	//SetBeam2DProperties(beamCrossSectionShape, beamCrossSectionSize, density, beamEA, beamEIy, beamGAky, beamRhoA, beamRhoIx, beamRhoIy);
	//$ PG 2018-3-25: ]

	//$ PG 2018-3-25: [ if youngs modulus is set, we would like to overwrite shear/bending/torsional stiffnesses and moments of inertia if not defined
	ComputeCrossSectionArea();
	if (YoungsModulus()>0)
	{
		double E = YoungsModulus();  // [N/m^2]
		double nu = PoissonRatio();  // []
		double G = E/(2.+2.*nu);     // [N/m^2]
		double ky = 1;               // []
		if (IsRectangularCrossSection())
			ky = 10*(1+nu)/(12+11*nu); // see, e.g. http://iopscience.iop.org/article/10.1088/0022-3727/10/11/009/meta
		else if (IsCircularCrossSection() || IsTubularCrossSection())
			ky = (6+12*nu+6*Sqr(nu))/(7+12*nu+4*Sqr(nu)); // see, e.g. http://iopscience.iop.org/article/10.1088/0022-3727/10/11/009/meta  
		else
			mbs->UO(UO_LVL_warn) << "WARNING: no default shear correction factors available in case beamCrossSectionShape = " << beamCrossSectionShape << "\n";

		if (beamEA == -1) beamEA = E*beamCrossSectionArea; // [N]
		if (beamEIy == -1) beamEIy = E*beamIy; // [N m^2]
		if (beamGJkx == -1) beamGJkx = G*beamIx;   // [N m^2]
		if (beamGAky == -1) beamGAky = G*beamCrossSectionArea*ky;   // [N m^2]
	}
	if (Density()>0)
	{
		double rho = Density();
		if (beamRhoA == -1) beamRhoA = rho*beamCrossSectionArea; //[kg/m]
		if (beamRhoIx == -1) beamRhoIx = rho*beamIx;
		if (beamRhoIy == -1) beamRhoIy = rho*beamIy;
	}
	//$ PG 2018-3-25: ]

	//$ PG 2016-1-25: [ commented out, since Beam3DProperties now inherits from Material, and not MaterialElastoplastic
	//inelasticity_type = GetInelasticityType(inelasticity_type_str);
	//inelasticity_solution_method = GetInelasticitySolutionMethod(inelasticity_solution_method_str);
	//$ PG 2016-1-25: ]
}

//$ PG 2016-2-11: recommended set function for continuum mechanics formulation
void Beam2DProperties::SetBeam2DProperties(int _beamCrossSectionShape, Vector _beamCrossSectionSize, double _rho, double _EA, double _EIy, double _GAky, double _rhoA, double _rhoIx, double _rhoIy)
{
	SetBeamCrossSectionInfo(_beamCrossSectionShape, _beamCrossSectionSize);
	plane = 1;

	Density() = _rho;
	BeamEA() = _EA;
	BeamEIy() = _EIy;
	BeamGAky()= _GAky;
	BeamRhoA() = _rhoA;
	BeamRhoIx() = _rhoIx;
	BeamRhoIy() = _rhoIy;

	double A = GetCrossSectionArea();
	if (A > 0 && Density() < 1e-15)
	{
		Density() = BeamRhoA()/A;
	}
}

//$ PG 2016-2-11: recommended set function for continuum mechanics formulation
void Beam2DProperties::SetBeam2DPropertiesCM(int _beamCrossSectionShape, Vector _beamCrossSectionSize, double _rho, double E, double nu, double _EA, double _EIy, double _GAky, double _rhoA, double _rhoIx, double _rhoIy)
{
	ElasticParametersLinear::SetElasticParametersLinear(E, nu);
	SetBeam2DProperties(_beamCrossSectionShape, _beamCrossSectionSize, _rho, _EA, _EIy, _GAky, _rhoA, _rhoIx, _rhoIy);
}

//JG, depricated, because it does not initialize cross section size and shape, and it does not specify density from rhoA and cross section size
void Beam2DProperties::SetBeam2DProperties(double _rho, double _EA, double _EIy, double _GAky, double _rhoA, double _rhoIx, double _rhoIy)
{
	plane = 1;

	Density() = _rho;
	BeamEA() = _EA;
	BeamEIy() = _EIy;
	BeamGAky()= _GAky;
	BeamRhoA() = _rhoA;
	BeamRhoIx() = _rhoIx;
	BeamRhoIy() = _rhoIy;
}



//////////////////////////////////////////////////////////////////////////
//$ PG 2013-12-5:[ the following functions are deprecated for general use!



// new plastic strain, internal variable are computed
// new stress is computed
// input: from actual iteration step: eps, epsplast, internalvar, 
// input: from last time step: epsplast_old, internalvar_old
double Beam2DProperties::DoNonlinStepBeam2D(double& stress, double eps, double& epsplast, double& internalvar, double epsplast_old, double internalvar_old, int use_tangent_stiffness) const
{
	// discontinuous accuracy
	double da = GetMBS()->GetSolSet().discontinuousaccuracy; //JG: old DOption8

	if (yieldstress < 0 ) {GetMBS()->UO() << "Material:: Warning -- yieldstress " << yieldstress << " < 0 for inelastic material!\n"; return 0;}
	if (HardeningModule() < 0) {GetMBS()->UO() << "Material:: Warning -- Tangent module " << HardeningModule() << " < 0 for inelastic material, softening?!\n"; return 0;}

	stress = YoungsModulus()*(eps - epsplast);
	double depsplast = 0;//epsplast - epsplast_old;
	double dinternalvar = 0;//internalvar - internalvar_old;
	double depsplastold = depsplast;
	double smin, smax, help, yieldfun;
	double error = 0;

	if (!IsHardeningMaterial()) // perfect plasticity
	{
		smax = yieldstress;
		smin = -yieldstress;
		help = 0;
		yieldfun = fabs(stress) - yieldstress;
		if (stress > smax ) 
		{
			depsplast = yieldfun / YoungsModulus();
			error = yieldfun / YoungsModulus();
		}
		else if (stress < smin ) 
		{
			depsplast = -yieldfun / YoungsModulus();
			error = yieldfun / YoungsModulus();
		}
		else if (yieldfun < -da/**YoungsModulus()*/ && fabs(depsplast) > da /YoungsModulus() && !use_tangent_stiffness) 
		{
			error = fabs(depsplast);
			depsplast = epsplast_old - epsplast;
		}
		dinternalvar = 0;
	}
	else // isotropic hardening
	{
		help = YoungsModulus()*HardeningModule()/(YoungsModulus()-HardeningModule());
		smax = yieldstress + help*internalvar;
		smin = -yieldstress - help*internalvar;
		yieldfun = fabs(stress) - smax;
		if (stress > smax ) 
		{
			dinternalvar = yieldfun/(YoungsModulus()+help);
			depsplast = dinternalvar;
			error = yieldfun / YoungsModulus();
		}
		else if (stress < smin ) 
		{
			dinternalvar = yieldfun/(YoungsModulus()+help);
			depsplast = -dinternalvar;
			error = yieldfun / YoungsModulus();
		}
		else if (yieldfun < -da/*YoungsModulus()*/ && fabs(depsplast) > da/YoungsModulus() && !use_tangent_stiffness ) 
		{
			error = fabs(depsplast);
			depsplast = 1*(epsplast_old - epsplast);
			dinternalvar = 1*(internalvar_old - internalvar);
		}
		internalvar = internalvar + dinternalvar;
	}
	epsplast = epsplast + depsplast;

	return error;
}



// AP Dez. 2012
// Material law due to return mapping algorithm by Simo and Hughes, 
// plane stress case differs from ComputeInelasticVariablesFromStrain2D substancially
void Beam2DProperties::ComputeInelasticVariablesFromStrainPlaneStress2D(Vector& inel_variables, const Matrix2D& strain) const
{ // calculate the actual state of inelastic variables as a function of the total strain in the planar case
	
	assert(IsPlanarMaterial());

	if (inel_variables.Length()>=5)
	{ 
		double Em = YoungsModulus();
		double nu = PoissonRatio();
		// standard Lame parameters
		double mu = Em / 2. / (1.+nu);
		double la = Em * nu / ((1.+nu)*(1.-2.*nu));
		// modified parameters for plane stress case
		la *= mu/(mu + .5*la);
		nu = .5*la/(la+mu);

		Vector3D inel_strain;    //  ( comp_11, comp_22, comp_12 )
		inel_strain(1) = inel_variables(1);
		inel_strain(2) = inel_variables(2);
		inel_strain(3) = inel_variables(3);					
		double hardening_param = inel_variables(4);
		double yield_function = inel_variables(5);

		Vector3D total_strain;      //  ( comp_11, comp_22, comp_12 )
		total_strain(1) = strain(1,1);
		total_strain(2) = strain(2,2);
		//AP Nov 2012: do not use engineering strains, but strain vector with eps_xy in third component
		//             now total_strain and inel_strain have the same form, and can be added, subtracted...
		total_strain(3) = strain(1,2);

		Matrix3D elasticity_tensor(2.*mu);
		elasticity_tensor(1,1) += la;
		elasticity_tensor(1,2) += la;
		elasticity_tensor(2,1) += la;
		elasticity_tensor(2,2) += la;
		//AP Nov 2012: do not use engineering strains, but strain vector with eps_xy in third component
		//             elasticity tensor has 2*mu on (3,3) position instead of standard 1*mu

		// Matrix P according to SimoHughes page 93, generate deviator, 
		// in case of plane stress, sigma_zz = 0
		Matrix3D P(2.,-1.,0.,
			-1.,2.,0.,
			0.,0.,6.);
		P *= 1./3.;

		// trial stress
		Vector3D trial_stress = elasticity_tensor*(total_strain - inel_strain);
		// deviator of trial stress
		Vector3D dev_trial_stress = P*trial_stress;

		// elementary computation:
		// ||dev sigma_0||^2 = (trial_stress^T*P*trial_stress)
		double sqrnorm_dev_trial_stress = dev_trial_stress*trial_stress;
		double norm_dev_trial_stress = sqrt(sqrnorm_dev_trial_stress);

		// value of (sqared) yield function:   ||dev sigma_0||^2 - 2/3(sigma_y + K hardening_param)^2
		// K .. plastic modulus
		double K = (YoungsModulus()*HardeningModule())/(YoungsModulus()-HardeningModule());
		double beta = sqrnorm_dev_trial_stress - 2./3.*sqr(YieldStress()+K*hardening_param);
		if ( beta > 0 )   // new plasticity occurs
		{
			// solve quadratic equation for gamma
			// update = gamma * dev sigma_0 / ||dev sigma_0||
			Vector3D C_dev_trial_stress = elasticity_tensor*dev_trial_stress;

			double A = (C_dev_trial_stress*P*C_dev_trial_stress)/sqrnorm_dev_trial_stress - 4./9.*K*K;
			double B = -2/norm_dev_trial_stress*(dev_trial_stress*C_dev_trial_stress) - 4./3.*(YieldStress()+K*hardening_param)*sqrt(2./3.)*K;
			double C = beta;
			double gamma = (-B - sqrt(B*B-4*A*C))/(2*A);
			Vector3D inel_strain_update = (gamma/norm_dev_trial_stress)*dev_trial_stress;
			inel_strain += inel_strain_update;


			inel_variables(1) = inel_strain(1);
			inel_variables(2) = inel_strain(2);
			inel_variables(3) = inel_strain(3);

			hardening_param += sqrt(2./3.)*gamma;
			inel_variables(4) = hardening_param;
			// update the trial stress
			trial_stress = elasticity_tensor*(total_strain - inel_strain);
			dev_trial_stress = P*trial_stress;
		}
		//yield_function (as additional 5th inelastic variable)
		norm_dev_trial_stress = sqrt(dev_trial_stress*trial_stress);
		inel_variables(5) = norm_dev_trial_stress - sqrt(2./3.)*(YieldStress()+K*hardening_param);
	}
	else
	{
		assert (0 && "This type of inelasticity has not yet been implemented!\n");
	}
}

// AP Dez. 2012
// Material law due to return mapping algorithm by Simo and Hughes, plane stress and strain case
// plane strain case differs from ComputeInelasticVariablesFromStrain2D by definition of TangentModule
// here, TangentModule is elastoplastic tangent module
void Beam2DProperties::ComputeInelasticVariablesFromStrainPlaneStrain2D(Vector& inel_variables, const Matrix2D& strain) const
{ // calculate the actual state of inelastic variables as a function of the total strain in the planar case
	
	assert(IsPlanarMaterial());
	
	if (inel_variables.Length() >= 5)
	{ 
		double Em = YoungsModulus();
		double nu = PoissonRatio();
		// standard Lame parameters
		double mu = Em / 2. / (1.+nu);
		double la = Em * nu / ((1.+nu)*(1.-2.*nu));

		Vector3D inel_strain;    //  ( comp_11, comp_22, comp_12 )
		inel_strain(1) = inel_variables(1);
		inel_strain(2) = inel_variables(2);
		inel_strain(3) = inel_variables(3);					
		double hardening_param = inel_variables(4);
		double yield_function = inel_variables(5);

		Vector3D total_strain;      //  ( comp_11, comp_22, comp_12 )
		total_strain(1) = strain(1,1);
		total_strain(2) = strain(2,2);
		//AP Nov 2012: do not use engineering strains, but strain vector with eps_xy in third component
		//             now total_strain and inel_strain have the same form, and can be added, subtracted...
		total_strain(3) = strain(1,2);

		Matrix3D elasticity_tensor(2.*mu);
		elasticity_tensor(1,1) += la;
		elasticity_tensor(1,2) += la;
		elasticity_tensor(2,1) += la;
		elasticity_tensor(2,2) += la;
		//AP Nov 2012: do not use engineering strains, but strain vector with eps_xy in third component
		//             modify elasticity tensor accordingly, 2*mu on (3,3) position instead of 1*mu
		//elasticity_tensor(3,3) -= mu;

		// Matrix P which generates deviator of stress sigma = C epsilon in case of plane strain, 
		// in case of plane stress, sigma_zz = 0
		Matrix3D P(1.-(1.+nu)/3.,-(1.+nu)/3.,0.,
			-(1.+nu)/3.,1.-(1.+nu)/3.,0.,
			0.,0.,1);

		// deviator of trial stress
		Vector3D dev_trial_stress = P*elasticity_tensor*total_strain - 2*mu*inel_strain;

		// matrix which represents norm for deviatoric stress vector
		Matrix3D norm(2.,1.,0.,
			1.,2.,0.,
			0.,0.,2);
		double sqrnorm_dev_trial_stress = dev_trial_stress*norm*dev_trial_stress;
		double norm_dev_trial_stress = sqrt(sqrnorm_dev_trial_stress);

		// value of yield function:   ||dev sigma_0|| - sqrt(2/3)(sigma_y + K hardening_param)
		// K .. plastic modulus
		double K = (YoungsModulus()*HardeningModule())/(YoungsModulus()-HardeningModule());
		double beta = norm_dev_trial_stress - sqrt(2./3)*(YieldStress()+K*hardening_param);
		if ( beta > 0 )   // new plasticity occurs
		{
			// solve linear update equation
			double gamma = beta/(2*mu+2./3.*K);
			Vector3D inel_strain_update = (gamma/norm_dev_trial_stress)*dev_trial_stress;
			inel_strain += inel_strain_update;

			inel_variables(1) = inel_strain(1);
			inel_variables(2) = inel_strain(2);
			inel_variables(3) = inel_strain(3);

			hardening_param += sqrt(2./3.)*gamma;
			inel_variables(4) = hardening_param;
			// different stress updates for plane strain/plane stress
			dev_trial_stress -= 2*mu*inel_strain_update;
		}
		//yield_function (as additional 5th inelastic variable)
		norm_dev_trial_stress = sqrt(dev_trial_stress*norm*dev_trial_stress);
		inel_variables(5) = norm_dev_trial_stress - sqrt(2./3.)*(YieldStress()+K*hardening_param);
	}
}

void Beam2DProperties::ComputeInelasticVariablesFromStrain1D(Vector& inel_variables, double strain) const
{ // calculate the actual state of inelastic variables as a function of the total strain in the planar case

	assert(IsPlanarMaterial());

	if (inel_variables.Length() >= 3)
	{ 
		double Em = YoungsModulus();
		double inel_strain;   
		inel_strain = inel_variables(1);
		double hardening_param = inel_variables(2);
		double yield_function = inel_variables(3);

		double el_strain;    
		el_strain = strain;

		double trial_stress = Em*(el_strain - inel_strain);

		double norm_trial_stress = fabs(trial_stress);

		double scaled_yield_stress = YieldStress();
		double beta = norm_trial_stress - scaled_yield_stress*(1 + HardeningModule()*hardening_param);
		if ( beta > 0 )   // new plasticity occurs
		{
			double xi = Em + pow(scaled_yield_stress*HardeningModule(), 2);
			double inel_strain_update = (beta/xi/norm_trial_stress)*trial_stress;
			inel_strain += inel_strain_update;

			inel_variables(1) = inel_strain;

			double inel_strain_update_norm = fabs(inel_strain_update);
			hardening_param += scaled_yield_stress*HardeningModule()*inel_strain_update_norm;
			inel_variables(2) = hardening_param;
			trial_stress += - Em * inel_strain_update;
		}
		//yield_function (as additional 5th inelastic variable)
		norm_trial_stress = fabs(trial_stress);
		inel_variables(3) = norm_trial_stress - scaled_yield_stress*(1 + HardeningModule()*hardening_param);
	}
}
//$ PG 2013-12-5:]
//////////////////////////////////////////////////////////////////////////











/////////////////////////////////////////////////////////////////////////
// function definitions of class MaterialThermoMechanic : public Material
/////////////////////////////////////////////////////////////////////////

// -------------------------
//   ThermalProperties 
// -------------------------
MaterialThermoMechanic::ThermalProperties::ThermalProperties(double specificHeatCapacity, const Matrix3D& thermalConductivity):
	specific_heat_capacity(specificHeatCapacity),
	thermal_conductivity(thermalConductivity)
{}
double MaterialThermoMechanic::ThermalProperties::GetSpecificHeatCapacity() const
{
	return specific_heat_capacity;
}
const Matrix3D& MaterialThermoMechanic::ThermalProperties::GetThermalConductivity() const
{
	return thermal_conductivity;
}
void MaterialThermoMechanic::ThermalProperties::SetSpecificHeatCapacity(double specificHeatCapacity)
{
	specific_heat_capacity = specificHeatCapacity;
}
void MaterialThermoMechanic::ThermalProperties::SetThermalConductivity(const Matrix3D& thermalConductivity)
{
	thermal_conductivity = thermalConductivity;
}

MaterialThermoMechanic::MaterialThermoMechanic(MBS* mbs):
	Material(mbs),
	temperature_properties(0., Matrix3D())
{ }

MaterialThermoMechanic::MaterialThermoMechanic( const Material& material, const ThermalProperties& temperatureProperties):
	Material(material),
	temperature_properties(temperatureProperties)
{
	//type = TMatTemperature; //$ PG 2016-2-17: type is now left to class Material, it is no more part of the base class MaterialGeneric!
}

MaterialThermoMechanic::MaterialThermoMechanic( const Material& material, double specificHeatCapacity, const Matrix3D& thermalConductivity):
	Material(material),
	temperature_properties(specificHeatCapacity, thermalConductivity)
{
	//type = TMatTemperature; //$ PG 2016-2-17: type is now left to class Material, it is no more part of the base class MaterialGeneric!
}

int MaterialThermoMechanic::SetElementData(ElementDataContainer &edc)
{
	int rv = MaterialThermoMechanic::SetElementDataAuto(edc);

	double specific_heat_capacity;
	Matrix thermal_conductivity;
	rv |= GetElemDataDouble(GetMBS(), edc, "Temperature.specific_heat_capacity", specific_heat_capacity, 1);
	rv |= GetElemDataMatrix(GetMBS(), edc, "Temperature.thermal_conductivity", thermal_conductivity, 1);
	SetThermalProperties(ThermalProperties(specific_heat_capacity, thermal_conductivity));


	return rv;
}

void MaterialThermoMechanic::GetElementData(ElementDataContainer &edc)
{
	MaterialThermoMechanic::GetElementDataAuto(edc);

	ElementData ed;

	ed.SetDouble(GetSpecificHeatCapacity(), "specific_heat_capacity", 0., -1, 1, 0); // lower limit ist zero, no upper limit is used
	ed.SetToolTipText("specific heat capacity (per mass)");
	edc.TreeAdd("Temperature",ed);

	Matrix thermal_conductivity(GetThermalConductivity());

	ed.SetMatrix(thermal_conductivity.GetMatPtr(),3,3,"thermal_conductivity");
	ed.SetToolTipText("");
	edc.TreeAdd("Temperature",ed);

}

MaterialGeneric* MaterialThermoMechanic::GetCopy() const
{
		MaterialGeneric* ec = new MaterialThermoMechanic(*this);
		return ec;
}

int MaterialThermoMechanic::CheckConsistency(mystr& errorstr)
{
	int rv = Material::CheckConsistency(errorstr);

	Matrix3D tc = GetThermalConductivity();

	// check if the thermal conductivity is positive semidefinite
	if ((tc(1,1)<0) ||
		(tc(1,1)*tc(2,2) - tc(2,1)*tc(1,2) < 0) ||
		(tc.Det() < 0))
	{
		errorstr += "Error in MatrialTemperature: thermal_conductivity must be a positive semidefinite matrix";
		rv = max(rv,1);
	}
	if (GetSpecificHeatCapacity() <= 0.)
	{
		errorstr += "Error in MatrialTemperature: the specific heat capacity must be positive";
		rv = max(rv,1);
	}
	return rv;
}

double MaterialThermoMechanic::GetSpecificHeatCapacity() const
{
	return temperature_properties.GetSpecificHeatCapacity();
}

const Matrix3D& MaterialThermoMechanic::GetThermalConductivity() const
{
	return temperature_properties.GetThermalConductivity();
}

const MaterialThermoMechanic::ThermalProperties& MaterialThermoMechanic::GetThermalProperties() const
{
	return  temperature_properties;
}

void MaterialThermoMechanic::SetThermalProperties(const MaterialThermoMechanic::ThermalProperties& temperatureProperties)
{
	temperature_properties = temperatureProperties;
}

void MaterialThermoMechanic::SetThermalProperties(double specificHeatCapacity, const Matrix3D& thermalConductivity)
{
	temperature_properties.SetSpecificHeatCapacity(specificHeatCapacity);
	temperature_properties.SetThermalConductivity(thermalConductivity);
}

