//#***************************************************************************************
//#
//# filename:             mesh.cpp
//#
//# author:               Dorninger Alexander
//#
//# generated:						September 2014
//# description:          implementation file for mesh
//#                       contains access and management functions
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
#include "mesh.h"
#include <femathHelperFunctions.h> // QSort

int BaseMesh::WarnSkipOperationIfNoComponents(mystr& caller)
{
	if(NComp()==0)
	{
		mbs->UO(UO_LVL_warn) << mystr("WARNING: ") + caller + mystr(" - the Mesh '") + GetMeshName() + ("' contains no Components. \n");
		return 1; // warning
	}
	return 0; // no warning
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH COMPONENT GENERATION - PRIMITIVE COMPONENT
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::AddComponent(MeshComponent* comp)
{
	comp->SetMesh(this);
	return meshcomponents.Add(comp->GetCopy());
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH COMPONENT GENERATION - EXTERNAL SOURCE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::LoadMeshExternal(mystr& filetype, mystr& filename)
{
	MeshComponentLoader* loader;
	if(filetype.Compare("Netgen2D"))
	{
		loader = new MeshLoaderNetgen2D();
	}
	else if(filetype.Compare("Netgen3D"))
	{
		loader = new MeshLoaderNetgen3D();
	}
	else if(filetype.Compare("STL"))
	{
		loader = new MeshLoaderSTL();
	}
	else if(filetype.Compare("Neutral3D"))
	{
		loader = new MeshLoaderNeutral3D();
	}
	else if(filetype.Compare("VOL"))
	{
		loader = new MeshLoaderVOL();
	}
	else
	{
		mbs->UO(UO_LVL_err) << "Error: Mesh file type " << filetype << " unknown!\n";
	}
	loader->SetMesh(this);
	loader->SetFileName(filename);

	return meshcomponents.Add(loader);
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH COMPONENT GENERATION - EXTERNAL TARGET
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::WriteMeshExternal(int compnr, mystr& filetype, mystr& filename)
{
	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			mystr filenameidx;
			filenameidx += filename;
			filenameidx += mystr(i);
			WriteMeshExternal(i,filetype,filenameidx);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL)
		{
			MeshComponentWriter* writer;
			if(filetype.Compare("Neutral3D"))
			{
				writer = new MeshWriterNeutral3D();
				writer->SetName("MeshWriterNeutral3D");
			}
			writer->SetFileName(filename);
			writer->SetMesh(this);
			writer->SetSubordinate(sub);
			SetMeshComponent(compnr,writer);			
		}
	}
	return 1; 
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - LINEAR TO QUADRATIC
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::LinearToQuadratic(int compnr)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::LinearToQuadratic"))) return 0;

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			LinearToQuadratic(i);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshLinearToQuadratic* linear2quadratic = new MeshLinearToQuadratic();
			//$EK 2017: test added component number: linear2quadratic->SetName(mystr("Linear2Quadratic")+mystr(compnr));
			linear2quadratic->SetName("Linear2Quadratic");
			linear2quadratic->SetMesh(this);
			linear2quadratic->SetSubordinate(sub);
			
			linear2quadratic->SetElemType("dependent");  // this entry is not actually used in the generation but the type of the element in the subordinate component may be changed
			linear2quadratic->SetNodeType("dependent");  // this entry is not actually used in the generation but the type of the element in the subordinate component may be changed

			SetMeshComponent(compnr,linear2quadratic);
		}
	}
	return compnr;
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - SPLIT HEXES
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::SplitHexes(int compnr, int split)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::SplitHexes"))) return 0;
	
	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			SplitHexes(i,split);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshSplitHexes* splithexes = new MeshSplitHexes();
			//splithexes->SetNodeType("Node3D");
			splithexes->SetNodeType(sub->GetNodeType()); // use same node type as hex

			//$EK 2017-01-11: in case of Stacked hexahedrals, stacked finite elements
			//(tets, prism, pyramids) should be used. Therefore a valid (source) subordinate 
			//has to be found which possesses the element type method
			//$AD 2018-08-16: also allow to split Thermo and ThermoMechanic Hexes to Tets only
			MeshComponent* tmpsub = sub;
			while (!tmpsub->IsSource() && tmpsub != NULL)
			{
				if (tmpsub->GetSubordinateComponent() != 0)
				{
					tmpsub = tmpsub->GetSubordinateComponent();
				}
			}
			mystr type = "Hexahedral";
			if (tmpsub != NULL)
			{
				type = ((MeshComponentSource*)tmpsub)->ElemType();
			}

			if(split==2)
			{
				splithexes->SetName("SplitToPrisms");

				if (type.Compare("HexahedralStacked"))
					splithexes->SetElemType("PrismStacked");
				else
					splithexes->SetElemType("Prism");
			}
			else if(split==3)
			{
				splithexes->SetName("SplitToPyramids");
				
				if (type.Compare("HexahedralStacked"))
					splithexes->SetElemType("PyramidStacked");
				else
					splithexes->SetElemType("Pyramid");
			}
			else if(split==5)
			{
				splithexes->SetName("SplitToTetraherals");

				if (type.Compare("HexahedralStacked"))
					splithexes->SetElemType("TetrahedralStacked");
				else if (type.Compare("HexahedralThermo"))
					splithexes->SetElemType("TetrahedralThermo");
				else if (type.Compare("HexahedralThermoMechanic"))
					splithexes->SetElemType("TetrahedralThermoMechanil");
				else
					splithexes->SetElemType("Tetrahedral");
			}
			splithexes->SetMesh(this);
			splithexes->SetSubordinate(sub);
			SetMeshComponent(compnr,splithexes);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - REFINE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Refine(int compnr, const TArray<int>& sdlvl, int method)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Refine"))) return 0;
	
	if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL)
		{
			MeshRefine* refine = new MeshRefine(); 
			refine->SetName("Refinement");
			refine->SetElemType(sub->GetElemType());
			refine->SetNodeType(sub->GetNodeType());
			refine->SetSDlvl(sdlvl);
			refine->SetMethod(method);
			refine->SetMesh(this);
			refine->SetSubordinate(sub);
			SetMeshComponent(compnr,refine);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - EXTRUDE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Extrude(int compnr, int axis, int discr, double dist, double thick)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Extrude"))) return 0;

	if(axis<1 || axis>3) { mbs->UO(UO_LVL_warn) << mystr("value of 'axis' must be 1, 2 or 3. \n"); return 0; }
	if(discr<1) { mbs->UO(UO_LVL_warn) << mystr("value of 'discr' must be at least 1. \n"); return 0; }
	if(abs(dist)<1.e-10) { mbs->UO(UO_LVL_warn) << mystr("value of 'dist' must not be 0. \n"); return 0; }

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			Extrude(i,axis,discr,dist,thick);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshExtrude* extrude = new MeshExtrude();

			extrude->SetName("extrude");
			extrude->ExtrusionAxis() = axis;
			extrude->SetMesh(this);
			extrude->SetSubordinate(sub);
			extrude->ExtrusionDiscretization() = discr;
			extrude->ExtrusionDistance() = dist;
			extrude->ExtrusionPlateThickness() = thick;
			SetMeshComponent(compnr,extrude);
		}
	}
	return compnr;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - ROTATIONAL
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Rotational(int compnr, int axis, int angleseg, double finalangle)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Rotational"))) return 0;
	
	if(axis<1 || axis>3) { mbs->UO(UO_LVL_warn) << mystr("value of 'axis' must be 1, 2 or 3. \n"); return 0; }
	if(angleseg<1) { mbs->UO(UO_LVL_warn) << mystr("value of 'angleseg' must be at least 1. \n"); return 0; }
	if(finalangle<0. || finalangle>360.) { mbs->UO(UO_LVL_warn) << mystr("value of 'finalangle' must be between 0 and 360. \n"); return 0; }

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			Rotational(i,axis,angleseg,finalangle);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshRotational* rotational = new MeshRotational();

			rotational->SetName("Rotational");
			rotational->RotationAxis() = axis;
			rotational->SetMesh(this);
			rotational->SetSubordinate(sub);
			rotational->AngularSegments() = angleseg;
			rotational->FinalAngleDeg() = finalangle;
			SetMeshComponent(compnr,rotational);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - MIRROR
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Mirror(int compnr, int plane, double dist)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Extrude"))) return 0;

	if(plane<1 || plane>3) { mbs->UO(UO_LVL_warn) << mystr("value of 'plane' must be 1, 2 or 3. \n"); return 0; }

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			Mirror(i,plane,dist);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshMirror* mirror = new MeshMirror();

			mirror->SetName("mirror");
			mirror->MirrorPlane() = plane;
			mirror->SetMesh(this);
			mirror->SetSubordinate(sub);
			mirror->MirrorDistance() = dist;
			SetMeshComponent(compnr,mirror);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - TRANSFORM
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::TransformXYZ(int compnr, Vector3D translate, Vector3D rotate, Vector3D scale)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::TransformXYZ"))) return 0;

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			TransformXYZ(i, translate, rotate, scale);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshTransformation* transformation;
			if(sub->IsTransformation())
			{ // change an existing MeshTransformation
				transformation = (MeshTransformation*) sub;
			}
			else
			{ // Insert a MeshTransformation 
				transformation = new MeshTransformation();
				transformation->SetName("Transformation");
				transformation->SetMesh(this);
				transformation->SetSubordinate(sub);
				SetMeshComponent(compnr,transformation);
			}
			transformation->TranslateXYZ(translate);
			transformation->RotateXYZ(rotate);
			transformation->ScaleXYZ(scale);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - Distort
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::DistortXYZ(int compnr, mystr fx, mystr fy, mystr fz, int subordinate)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::DistortXYZ"))) return 0;

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			DistortXYZ(i, fx, fy, fz, subordinate);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			if (subordinate>0)
			{					
				MeshDistortion* distortion = new MeshDistortion(GetMBS()); // Parser is required for MeshDistortion::Reset
				distortion->SetName("Distortion");
				distortion->SetMesh(this);
				distortion->SetFunctionString(1,fx);
				distortion->SetFunctionString(2,fy);
				distortion->SetFunctionString(3,fz);

				if (sub->NSubs() > 0)
				{				
					distortion->SetSubordinate(sub->GetSubordinateComponent(subordinate));
				}
				sub->SetSubordinate(distortion,subordinate);
			}
			else
			{
				MeshDistortion* distortion = new MeshDistortion(GetMBS()); // Parser is required for MeshDistortion::Reset
				distortion->SetName("Distortion");
				distortion->SetMesh(this);
				distortion->SetSubordinate(sub);
				distortion->SetFunctionString(1,fx);
				distortion->SetFunctionString(2,fy);
				distortion->SetFunctionString(3,fz);
				SetMeshComponent(compnr,distortion);
			}
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - Modify
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Modify(int compnr, MeshModifier::WorkingOnEntity entity, int newIntVal, double newDoubleVal, mystr& newStringVal, 
					                                                   int oldIntVal, double oldDoubleVal, mystr& oldStringVal, BaseSet* filter)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Modify"))) return 0;

	if(compnr==0)
	{
		for(int i=1; i<=NComp(); i++)
		{
			Modify(i, entity, newIntVal, newDoubleVal, newStringVal, 
				              oldIntVal, oldDoubleVal, oldStringVal, filter);
		}
	}
	else if(0<compnr && compnr<=NComp())
	{
		MeshComponent* sub = GetMeshComponent(compnr);
		if(sub!=NULL) 
		{
			MeshModifier* modifier = new MeshModifier(); 
			modifier->SetName("Modifier");
			modifier->SetMesh(this);
			modifier->SetSubordinate(sub);
			
			modifier->SetEntity(entity);
			if(entity == MeshModifier::WorkingOnEntity::WOE_elemtype ||
			   entity == MeshModifier::WorkingOnEntity::WOE_nodetype ||
			   entity == MeshModifier::WorkingOnEntity::WOE_materialnumber )
			{
				modifier->SetNewInt(newIntVal);
				modifier->SetOldInt(oldIntVal);
			}
			modifier->SetFilter(filter);

			SetMeshComponent(compnr,modifier);
		}
	}
	return compnr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH MODIFICATION - GLUE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::Glue(TArray<int> compnr, int use_penalty)
{
	if(WarnSkipOperationIfNoComponents(mystr("BaseMesh::Glue"))) return 0;	

	MeshGlue* glue = new MeshGlue();
	glue->SetName("Glue");
	glue->SetMesh(this);
	glue->SetUsePenaltyFlag(use_penalty);
	for(int i=1; i<=compnr.Length(); i++)
	{
		MeshComponent* sub = GetMeshComponent(compnr(i));
		glue->AddSubordinate(sub);
		SetMeshComponent(compnr(i),NULL);
	}
	SetMeshComponent(compnr(1),glue);
	return compnr(1);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH INSTANTIATION - ADD TO MBS
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::MeshToMBS(MBS* mbs)
{	
	MeshAllComponents();
	NodesToMBS(mbs);
	ElemsToMBS(mbs);
	SetsToMBS(mbs); 
	UpdateStatistics();
	AppyAutomaticConstraints(mbs);
	ApplyPeriodicBCs(mbs);
	ApplySymmetryBCs(mbs);
	return 0;
}
int BaseMesh::NodesToMBS(MBS* mbs)
{
	mbsnodenrs.Flush();
	mbsnodenrs.SetLen(NNodes());
	mbsnodenrs.SetAll(0);

// add the nodes to the mbs (root-)component by (root-)component
	int prevcomp = 0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)
		{
			for(int i=1; i<=comp->NNodes(); i++)
			{
	// create a node of required type and register in MBS
				int nodetype = comp->GetNodeObjectFactoryType(i);
				int nnr = mbs->GetObjectFactory()->AddObject(OFCNode,nodetype);
				comp->SetNodeMeshNumber(i,i+prevcomp);
				mbsnodenrs(i+prevcomp) = nnr;
		
	// alter the nodes properties accordingly - basic properties
				GenericNode* pMBSNode = (GenericNode*) mbs->GetNodePtr(nnr);	
				pMBSNode->Pos() = comp->GetNodePosition(i); 
				pMBSNode->SetBodyInd( comp->GetNodeBodyIndex(i) );

	// alter the nodes properties accordingly - optional properties
				if(pMBSNode->HasRefSlope1() || pMBSNode->HasRefSlope2() || pMBSNode->HasRefSlope3() || pMBSNode->HasRefAngles()) 
				{
					comp->ComputeOrientation(i); // recompute the orientation if node type requires 
				}
				if(pMBSNode->HasRefSlope1()) pMBSNode->SetRefSlope1( comp->GetNodeRefSlope1(i) );
				if(pMBSNode->HasRefSlope2()) pMBSNode->SetRefSlope2( comp->GetNodeRefSlope2(i) );
				if(pMBSNode->HasRefSlope3()) pMBSNode->SetRefSlope3( comp->GetNodeRefSlope3(i) );
				if(pMBSNode->HasRefAngles()) pMBSNode->SetRefAngles( comp->GetNodeRefAngles(i) );
			}
			prevcomp += comp->NNodes();
		}
	}
	return mbsnodenrs.Length();
}
int BaseMesh::ElemsToMBS(MBS* mbs)
{
	mbselemnrs.Flush();
	mbselemnrs.SetLen(NElems());
	mbselemnrs.SetAll(0);

// add the elements to the mbs (root-)component by (root-)component
	int prevcomp = 0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)
		{
			for(int i=1; i<=comp->NElems(); i++)
			{
	// create an element of required type and register in MBS and Mesh
				int elemtype = comp->GetElementObjectFactoryType(i);
				int enr = mbs->GetObjectFactory()->AddObject(OFCElement,elemtype);
				comp->SetElementMeshNumber(i,i+prevcomp);
				mbselemnrs(i+prevcomp) = enr;

	// alter the elements properties accordingly - basic properties - common to all FiniteElements, Beams, ...
				TArray<int> nodelist = comp->GetElementNodeList(i); // nodelist containing component node numbering
				int bodyind = comp->GetBodyIndex(nodelist(1));
				comp->MapNodeListComponentToMesh(nodelist);         // nodelist containing mesh node numbering 
				MapNodeListToMBSNodes(nodelist);                    // nodelist containing mbs node numbering
				int matnr = comp->GetElementMaterialNumber(i);
				Vector3D& col = comp->GetColor(i);
				mbs->GetElementPtr(enr)->SetFiniteElementProperties(bodyind,nodelist,matnr,col);

				int gnls = comp->GetElementGeometricNonlinearityStatus(i);
				if(gnls>=0) { mbs->GetElementPtr(enr)->SetGeometricNonlinearityStatus(gnls); }

	// alter the elements properties accordingly - optional properties
	// NOTE: parts of code below are to be removed - e.g. plate elements should compute their own size...
				mystr elemtypestring = comp->GetElemType(i);	

				if(elemtypestring.Compare("ANCFThinPlate3D"))
				{
					comp->ComputeElementSize(i);
					Vector3D thicknessSize1Size2 = comp->GetElementSizeVector(i);
					double thickness = thicknessSize1Size2.X();
					double size1 = thicknessSize1Size2.Y();
					double size2 = thicknessSize1Size2.Z();
					mbs->GetElementPtr(enr)->SetFiniteElementAdditionalProperties(thickness,size1,size2);
				}
			}
			prevcomp += comp->NElems();
		}
	}
	return mbselemnrs.Length();
}

#include "KinematicPairs.h" // BaseBodyJoint
#include "RigidBodyJoints.h" // PointJoint2D
int BaseMesh::AppyAutomaticConstraints(MBS* mbs)
{
	if(NCoincident()==0) { return 0; }

	// in MBS generate a Set containing the coincident nodes
	GlobalNodeSet coincident_nodes(mbs,mystr("coincident_nodes"));
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = (MeshComponent*) meshcomponents(j);
		if(comp!=NULL) 
		{
			for(int i=1; i<=comp->NCoincident(); i++)
			{
				int2 coin_pair = comp->GetCoincidentNodeNumbers(i); // already mapped to numbering from root mesh

				coin_pair(1) = comp->GetMesh()->mbsnodenrs(coin_pair(1));
				coin_pair(2) = comp->GetMesh()->mbsnodenrs(coin_pair(2));

				coincident_nodes.AddEntry(coin_pair(1));
				coincident_nodes.AddEntry(coin_pair(2));
			}
		}
	}
	coincident_nodes.Convert(); // converts to default -> (Element numbers & local coordinates)

	// helper array for additional settings - FTTB contains only a single flag per constraint
	TArray<int> flagspenalty;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = (MeshComponent*) meshcomponents(j);
		if(comp!=NULL) 
		{
			for(int i=1; i<=comp->NCoincident(); i++)
			{
				flagspenalty(i) = comp->GetCoincidentPenaltyFlag(i);
			}
		}
	}

// NOTES:
// can not typecast to BasePointJoint* ( that is where the TArray<Vector3D> loccoords is introduced )
// no accessfunctions to these TArray in Constraint
// functions to set:  SetPos1ToLocalCoord, ...
// simply #include "KinematicPairs.h" gives linker error - SetPosToLocalCoord implemented in .cpp - project dependency :(
// BUT: Hack exists
// this functionality could be moved to MBS ?!?


	Element* p1st = GetMBS()->GetElementPtr(coincident_nodes.GetElementNumber(1));
	int dim = p1st->Dim();

	TArray<int> constraintnumbers;

	if (dim==3) 
	{
// parameters for the automatically generated constraints - some of those should be changeable - possibly defined in the GlueMesh
// fttb only the use_penalty flag is implemented as additional parameter
		mystr elemtype = mystr("GenericBodyJoint");
		int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);

		for(int i=1; i<=coincident_nodes.GetNrOfEntries()/2; i++)
		{
			int j1 = i*2-1;
			int j2 = i*2;

// request joint from ObjectFactory
			int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr); // new constraint generated here
			
			BasePointJoint* pbpj = (BasePointJoint*) mbs->GetElementPtr(cnr);
// change the joint parameters according to Glue Component parameters
			int use_penalty = flagspenalty(i);
			if(use_penalty) 
			{
				pbpj->SetPenaltyFormulation(use_penalty);
				pbpj->SetStiffnessMatrix(Matrix3D(1e10,1e10,1e10));
			}

			int enr1 = coincident_nodes.GetElementNumber(j1);
			Vector3D& lpos1 = coincident_nodes.GetLocalPosition(j1);
		//	pbpj->SetPos1ToLocalCoord(enr1, lpos1); // dependency

			int enr2 = coincident_nodes.GetElementNumber(j2);
			Vector3D& lpos2 = coincident_nodes.GetLocalPosition(j2);
		//	pbpj->SetPos2ToLocalCoord(enr2, lpos2); // dependency

			TArray<int>& el = pbpj->GetElementsArray();
			TArray<Vector3D>& lp = pbpj->GetLoccoordsArray();
			el(1) = enr1;
			el(2) = enr2;
			lp(1) = lpos1;
			lp(2) = lpos2;

			constraintnumbers.Add(cnr);
		}
	}
	else 
	{
		mystr elemtype = mystr("PointJoint2D");
		int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);

		for(int i=1; i<=coincident_nodes.GetNrOfEntries()/2; i++)
		{
			int j1 = i*2-1;
			int j2 = i*2;

// request joint from ObjectFactory
			int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr); // new constraint generated here

			PointJoint2D* ppj2 = (PointJoint2D*) mbs->GetElementPtr(cnr);
// change the joint parameters according to Glue Component parameters
			int use_penalty = flagspenalty(i);
			if(use_penalty) 
			{
				ppj2->SetPenaltyFormulation(use_penalty);
				ppj2->SetPenaltyStiffness2(Vector2D(1e10,1e10));
			}

			int enr1 = coincident_nodes.GetElementNumber(j1);
			Vector3D& lpos1 = coincident_nodes.GetLocalPosition(j1);
			//ppj2->SetPos1ToLocalCoord(enr1, lpos1); // dependency

			int enr2 = coincident_nodes.GetElementNumber(j2);
			Vector3D& lpos2 = coincident_nodes.GetLocalPosition(j2);
		//	ppj2->SetPos2ToLocalCoord(enr2, lpos2); // dependency

			TArray<int>& el = ppj2->GetElementsArray();
			TArray<Vector2D>& lp = ppj2->GetLoccoordsArray();
			el(1) = enr1;
			el(2) = enr2;
			lp(1) = Vector2D(lpos1.X(),lpos1.Y());
			lp(2) = Vector2D(lpos2.X(),lpos2.Y());

			constraintnumbers.Add(cnr);
		}
	}


// register the generated constraints in the origin glue mesh components
	int componentidx=1;
	int prevcompconstraints=0;
	for(int i=1; i<=constraintnumbers.Length(); i++)
	{
		if(componentidx <= NComp()) // iterate over all top level components
		{
			MeshComponent* comp = (MeshComponent*) meshcomponents(componentidx);
			if(comp!=NULL)
			{
				int j=i-prevcompconstraints;
				if(j<=comp->NCoincident())
				{
					comp->SetAutoConstraintNumber(j, constraintnumbers(i));
				}
				else
				{
					prevcompconstraints += comp->NCoincident();
					componentidx++; // try next component
				}
			}
			else
			{
				componentidx++; // skip invalid components
			}
		}
	}

	return 1;
}

int BaseMesh::ApplyPeriodicBCs(MBS* mbs)
{
	GlobalNodeSet periodic_nodes_1(mbs,mystr("periodic_nodes_1"));
	GlobalNodeSet periodic_nodes_2(mbs,mystr("periodic_nodes_2"));

	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = (MeshComponent*) meshcomponents(j);
		if(comp!=NULL) 
		{
			TArray<int2> node_pairs = comp->GetPeriodicBCNodePairs();
			for(int i=1, n=node_pairs.Length(); i<=n; i++)
			{
				int2& node_pair = node_pairs(i); // component index

				if (node_pair(1) == 0)
				{
					mbs->UO() << "ERROR: ";
					mbs->UO() << "(" << node_pair(1) << ", " << node_pair(2) << ")\n";
					node_pair(1) = node_pair(2);
				}
				if (node_pair(2) == 0)
				{
					mbs->UO() << "ERROR: ";
					mbs->UO() << "(" << node_pair(1) << ", " << node_pair(2) << ")\n";
					node_pair(2) = node_pair(1);
				}
				// component index => mesh index
				node_pair(1) = comp->GetNodeMeshNumber(node_pair(1));
				node_pair(2) = comp->GetNodeMeshNumber(node_pair(2));

				// mesh index => mbs index
				node_pair(1) = comp->GetMesh()->mbsnodenrs(node_pair(1));
				node_pair(2) = comp->GetMesh()->mbsnodenrs(node_pair(2));

				periodic_nodes_1.GlobalNodeNumbers().Add(node_pair(1));
				periodic_nodes_2.GlobalNodeNumbers().Add(node_pair(2));
			}
		}
	}

	periodic_nodes_1.Convert();
	periodic_nodes_2.Convert();

	periodic_constraints_mbs_numbers.SetLen(periodic_nodes_1.GetNrOfEntries());
	periodic_nodes_mbs_numbers_1.SetLen(periodic_nodes_1.GetNrOfEntries());
	periodic_nodes_mbs_numbers_2.SetLen(periodic_nodes_1.GetNrOfEntries());

	for (int i = 1, n=periodic_nodes_1.GetNrOfEntries(); i <= n; ++i)
	{
		int e1 = periodic_nodes_1.ElementNumber(i);
		int e2 = periodic_nodes_2.ElementNumber(i);
		int en1 = periodic_nodes_1.GetLocalNodeNumber(i);
		int en2 = periodic_nodes_2.GetLocalNodeNumber(i);



		int dim = mbs->GetElement(e1).Dim();

		if (e1 != e2 || en1 != en2)
		{
			int typenr = 0;
			if (dim == 2)
			{
				mystr elemtype("NodalSymmetryConstraint2D");
				typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
			}
			else if (dim == 3)
			{
				mystr elemtype("NodalSymmetryConstraint");
				typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
			}

			int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr);
			Element* c = mbs->GetElementPtr(cnr);
			ElementDataContainer edc;
			c->GetElementData(edc);

			edc.TreeSetInt("Position1.element_number",e1);
			edc.TreeSetInt("Position2.element_number",e2);
			edc.TreeSetInt("Position1.node_number", en1);
			edc.TreeSetInt("Position2.node_number", en2);
			ElementData ed;
		
			int n1 = periodic_nodes_1.GetGlobalNodeNumber(i);
			int n2 = periodic_nodes_2.GetGlobalNodeNumber(i);

			//mbs->UO(UO_LVL_all) << mystr("< ") << n1 << ", " << n2 << " >  ==>  < ( " << e1 << ", " << en1 << " ), (" << e2 << ", " << en2 << mystr(" ) > \n");

			periodic_constraints_mbs_numbers(i) = cnr;
			periodic_nodes_mbs_numbers_1(i) = periodic_nodes_1.GlobalNodeNumber(i);
			periodic_nodes_mbs_numbers_2(i) = periodic_nodes_2.GlobalNodeNumber(i);


			if (dim == 2)
			{
				Vector2D v1 = mbs->GetNode(n1).RefConfPos2D();
				Vector2D v2 = mbs->GetNode(n2).RefConfPos2D();
	
				v1.Normalize();
				v2.Normalize();
	
				double x1 = v1.X();
				double x2 = v2.X();
	
				double y1 = v1.Y();
				double y2 = v2.Y();
	
	
				Matrix2D rel_rot(x1*x2+y1*y2,  -(x1*y2 - x2*y1),
								 x1*y2 - x2*y1, x1*x2+y1*y2);
	
				Matrix rel_rot_1(rel_rot);
				// rotatiton matrix from coordinate system of first node/area into second node/area
				ed.SetMatrix(rel_rot_1.GetMatPtr(),rel_rot_1.Getrows(),rel_rot_1.Getcols(),"relative_rotation");
			}
			else if (dim == 3)
			{
				Vector3D v1 = mbs->GetNode(n1).RefConfPos();
				Vector3D v2 = mbs->GetNode(n2).RefConfPos();
	
				v1.Normalize();
				v2.Normalize();
	
				Vector3D vn = v1.Cross(v2);
				Matrix3D vnSkew;
				vnSkew.SetSkew(vn);

				double s = vn.Norm();
				double c = v1*v2;

				Matrix3D rel_rot;
				if (abs(s) > 1e-8)
				{
					 rel_rot = Matrix3D(1.)+vnSkew+(1-c)/(s*s)*vnSkew*vnSkew;
				}
				else if (abs(c-1) < 1.) // cos == 1
				{
					rel_rot = Matrix3D(1.);
				}
				else // cos == -1
				{
					rel_rot = Matrix3D(-1.);
				}

	
				Matrix rel_rot_1(rel_rot);
				// rotatiton matrix from coordinate system of first node/area into second node/area
				ed.SetMatrix(rel_rot_1.GetMatPtr(),rel_rot_1.Getrows(),rel_rot_1.Getcols(),"relative_rotation");

			}
			edc.TreeSet("Geometry",ed, 1, 0);
	
			c->SetElementData(edc);
		}
		else
		{			
			int typenr = 0;
			if (dim == 2)
			{
				mystr elemtype("PointJoint2D");
				typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
			}
			else if (dim == 3)
			{
				mystr elemtype("PointJoint");
				typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
			}

			int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr);

			Element* c = mbs->GetElementPtr(cnr);
			ElementDataContainer edc;
			c->GetElementData(edc);

			periodic_constraints_mbs_numbers(i)=cnr;
			periodic_nodes_mbs_numbers_1(i) = periodic_nodes_1.GlobalNodeNumber(i);
			periodic_nodes_mbs_numbers_2(i) = periodic_nodes_2.GlobalNodeNumber(i);

			edc.TreeSetInt("Position1.element_number",e1);

			if (dim == 2)
			{
				Vector2D locpos2D = periodic_nodes_1.GetLocalPosition(i).MakeV2D();
				double locpos[2];
				locpos[0] = locpos2D.X();
				locpos[1] = locpos2D.Y();
				edc.TreeSetVectorC("Position1.position", locpos,2,"");
			}
			else
			{
				edc.TreeSetInt("Position1.node_number", en1);
			}
			c->SetElementData(edc);
		}
	}
	return 0;
}

int BaseMesh::ApplySymmetryBCs(MBS* mbs)
{
	GlobalNodeSet symmetric_nodes(mbs, mystr("symmetry_nodes"));
	GlobalNodeSet origin_nodes(mbs, mystr("symmetric_origin_nodes"));//nodes at origin

	for (int j = 1; j <= NComp(); j++)
	{
		MeshComponent* comp = (MeshComponent*)meshcomponents(j);
		if (comp != NULL)
		{
			TArray<int> symmetricNodes = comp->GetSymmetricNodes();
			for (int i = 1, n = symmetricNodes.Length(); i <= n; i++)
			{
				int globalNode = symmetricNodes(i); // component index

				// component index => mesh index
				globalNode = comp->GetNodeMeshNumber(globalNode);

				// mesh index => mbs index
				globalNode = comp->GetMesh()->mbsnodenrs(globalNode);

				//Node at origin
				if (mbs->GetNodePtr(globalNode)->RefConfPos().Norm2() == 0)
				{
					origin_nodes.GlobalNodeNumbers().Add(globalNode);
				}
				else
				{
					symmetric_nodes.GlobalNodeNumbers().Add(globalNode);
				}
			}
		}
	}

	symmetric_nodes.Convert();
	origin_nodes.Convert();

	symmetric_constraints_mbs_numbers.SetLen(symmetric_nodes.GetNrOfEntries());
	symmetric_nodes_mbs_numbers.SetLen(symmetric_nodes.GetNrOfEntries());

	for (int i = 1, n = symmetric_nodes.GetNrOfEntries(); i <= n; ++i)
	{
		int elNr = symmetric_nodes.ElementNumber(i);
		int localNode = symmetric_nodes.GetLocalNodeNumber(i);
		int dim = mbs->GetElement(elNr).Dim();
		int typenr = -1;
		if (dim == 2)
		{
			mystr elemtype("PointOnRadialLine2D");
			typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement, elemtype);
		}
		else if (dim == 3)
		{
			release_assert("BaseMesh::ApplySymmetryBCs: Not implemented for 3D-Elements" && false);
			return -1;
		}

		int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr);

		symmetric_nodes_mbs_numbers(i) = symmetric_nodes.GlobalNodeNumber(i);
		symmetric_constraints_mbs_numbers(i) = cnr;


		Element* c = mbs->GetElementPtr(cnr);
		ElementDataContainer edc;
		c->GetElementData(edc);

		edc.TreeSetInt("Position1.element_number", elNr);
		edc.TreeSetInt("Position1.node_number", localNode);
		c->SetElementData(edc);
	}
	for (int i = 1, n = origin_nodes.GetNrOfEntries(); i <= n; ++i)
	{
		int elNr = origin_nodes.ElementNumber(i);
		int localNode = origin_nodes.GetLocalNodeNumber(i);
		int dim = mbs->GetElement(elNr).Dim();
		int typenr = -1;
		if (dim == 2)
		{
			mystr elemtype("PointJoint2D");
			typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement, elemtype);
		}
		else if (dim == 3)
		{
			mystr elemtype("PointJoint");
			typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement, elemtype);
		}

		int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr);
		Element* c = mbs->GetElementPtr(cnr);
		ElementDataContainer edc;
		c->GetElementData(edc);

		edc.TreeSetInt("Position1.element_number", elNr);
		edc.TreeSetInt("Position1.node_number", localNode);
		c->SetElementData(edc);

		symmetric_nodes_mbs_numbers.Append(symmetric_nodes.GlobalNodeNumber(i));
		symmetric_constraints_mbs_numbers.Append(cnr);
	}
	return 0;
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH INSTANTIATION - COMPUTE MESH
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::MeshAllComponents()
{
	if(NComp()==0)
	{
		mbs->UO(UO_LVL_warn) << "WARNING: BaseMesh::MeshAllComponents - the Mesh contains no Components \n";
	}
	for(int i=1; i<=NComp(); i++)
	{
		MeshSingleComponent(i);
	}
	return 1;
}
int BaseMesh::MeshSingleComponent(int compnr)
{
	MeshComponent* comp = GetMeshComponent(compnr);
	if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
	{	
		comp->MeshYourself();
	}
	return 1;
}
int BaseMesh::NNodes()
{
	int nn=0;
	for(int i=1; i<=NComp(); i++)
	{
		MeshComponent* comp = GetMeshComponent(i);
		if(comp!=NULL)
		{
			nn += comp->NNodes();
		}
	}
	return nn;
}
int BaseMesh::NElems()
{
	int ne=0;
	for(int i=1; i<=NComp(); i++)
	{
		MeshComponent* comp = GetMeshComponent(i);
		if(comp!=NULL)
		{
			ne += comp->NElems();
		}
	}
	return ne;
}
int BaseMesh::NCoincident()
{
	int nc=0;
	for(int i=1; i<=NComp(); i++)
	{
		MeshComponent* comp = GetMeshComponent(i);
		if(comp!=NULL)
		{
			nc += comp->NCoincident();
		}
	}
	return nc;
}

int BaseMesh::GetComponentNodeTypeIdx(MeshComponent* comp)
{
	if(comp->IsSource())
	{
		mystr nodetype = ((MeshComponentSource*)comp)->GetNodeType();
		return GetComponentNodeTypeIdx(nodetype);
	}
	else
	{
		mbs->UO() << mystr("BaseMesh::GetComponentNodeTypeIdx requests a NodeType for a SpecialPurpose component. This operation is only allowed for Primitive components\n");	
		return 0;
	}
}
int BaseMesh::GetComponentNodeTypeIdx(mystr& nodetype)
{
	int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCNode,nodetype);
	if(typenr<=0)
	{
		mbs->UO() << mystr("BaseMesh::GetComponentNodeTypeIdx - NodeType " + nodetype + " not registered in ObjectFactory\n");
	}
	return typenr;
}
int BaseMesh::GetComponentElemTypeIdx(MeshComponent* comp)
{
	if(comp->IsSource())
	{
		mystr elemtype = ((MeshComponentSource*)comp)->GetElemType();
		return GetComponentElemTypeIdx(elemtype);
	}
	else
	{
		mbs->UO() << mystr("BaseMesh::GetComponentElemTypeIdx requests an ElementType for a SpecialPurpose component. This operation is only allowed for Primitive components\n");	
		return 0;
	}
}
int BaseMesh::GetComponentElemTypeIdx(mystr& elemtype)
{
	int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
	if(typenr<=0)
	{
		mbs->UO() << mystr("BaseMesh::GetComponentElemTypeIdx - ElementType " + elemtype + " not registered in ObjectFactory\n");
	}
	return typenr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH - MBS INTERFACE - NODES TO ELEMENT LIST
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int BaseMesh::ClearNodesToElementsList()
{
	for(int i=1; i<=N2E().Length(); i++)
	{
		TArray<int>* a = N2E().Get(i);
		if(a!=NULL)
		{
			delete a;
			a=NULL;
		}
	}
	N2E().Flush();
	return 0;
}
int BaseMesh::AllocateNodesToElementsList()
{
	int nnodes = NNodes();
	node2element.SetLen(nnodes);
	for(int i=1; i<=N2E().Length(); i++)
	{
		N2E()(i) = new TArray<int>(10); 
		N2E()(i)->SetLen(0);
	}
	return 0;
}
int BaseMesh::AssembleNodesToElementsList()
{
	int prevcomp = 0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Components MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NElems(); i++)
			{
				TArray<int> nodelist = comp->GetElementNodeList(i); // nodelist containing component node numbering
				comp->MapNodeListComponentToMesh(nodelist);         // nodelist containing mesh node numbering
				for(int k=1; k<=nodelist.Length(); k++)
				{
					int nn = nodelist(k);
					if(N2E()(nn)->Last() != (i+prevcomp))
					{
						N2E()(nn)->Add(i+prevcomp);			
					}
				}	
			}
			prevcomp += comp->NElems();
		}
	}
	return 0;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MESH - PICKING
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MeshNode* BaseMesh::GetNodePtr(int nodenumber)
{
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			if(nodenumber <= comp->NNodes())
			{ 
				return comp->GetNodePtr(nodenumber); 
			}
			else
			{
				nodenumber -= comp->NNodes();
			}
		}
	}
	return NULL;
}
MeshElement* BaseMesh::GetElementPtr(int elementnumber)
{
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			if(elementnumber <= comp->NElems())
			{ 
				return comp->GetElemPtr(elementnumber); 
			}
			else
			{
				elementnumber -= comp->NElems();
			}
		}
	}
	return NULL;
}
TArray<int> BaseMesh::GetNodesInBox(Box3D box)
{
	TArray<int> nodes;
	int nn=0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NNodes(); i++)
			{
				Vector3D pos = comp->GetNodePosition(i);
				if(box.IsIn(pos))
				{
					nodes.Add(i+nn);                // Adds the MeshNumber to the Array
				}
			}
			nn += comp->NNodes();
		}
	}
	return nodes;
}
TArray<int> BaseMesh::GetNodesInCylinderShell(Vector3D center_bottom, Vector3D center_top, double rad_outer, double rad_inner)
{
	TArray<int> nodes;
	int nn=0;

	Vector3D axis = center_top - center_bottom;
	double cylinder_height = axis.Norm();
	if(!axis.Norm()) {assert(0);}									// axis must not be (0.0,0.0,0.0)
	
	Vector3D normalized_axis = axis; normalized_axis.Normalize();

	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NNodes(); i++)
			{
				Vector3D pos = comp->GetNodePosition(i);   
				Vector3D vec = pos-center_bottom;          
				Vector3D tmp = vec.Cross(normalized_axis); 
				double radial_distance = tmp.Norm();

				if(rad_inner<=radial_distance && radial_distance<=rad_outer) // within an infinitly long shell
				{
					double axial_projection = (vec*normalized_axis)/cylinder_height;  // 0@bot, 1@top
					if(0<=axial_projection && axial_projection<=1.)
					{
						nodes.Add(i+nn);                // Adds the MeshNumber to the Array
					}
				}	
			}
			nn += comp->NNodes();
		}
	}
	return nodes;
}
TArray<int> BaseMesh::GetNodesInSphericalShell(Vector3D center, double rad_outer, double rad_inner)
{
	TArray<int> nodes;
	int nn=0;

	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NNodes(); i++)
			{
				Vector3D pos = comp->GetNodePosition(i);
				Vector3D vec = pos-center;
				double distance2center = vec.Norm();
				
				if(rad_inner<=distance2center && distance2center<=rad_outer)
				{
					nodes.Add(i+nn);                // Adds the MeshNumber to the Array
				}
			}
			nn += comp->NNodes();
		}
	}
	return nodes;
}
TArray<int> BaseMesh::GetNodesInMathFunction(MathFunction& mf)
{
	TArray<int> nodes;
	int nn=0;

	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NNodes(); i++)
			{
				Vector3D pos = comp->GetNodePosition(i);	

				Vector vin(pos.X(),pos.Y(),pos.Z());
				mf.SetInputVector(vin);

				double retval = mf.Evaluate(0.);
				if(0<retval)
				{
					nodes.Add(i+nn);
				}
			}
			nn += comp->NNodes();
		}
	}
	return nodes;
}


TArray<int> BaseMesh::GetElementsInBox(Box3D box)
{
	TArray<int> elements;
	int nn=0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NElems(); i++)
			{
				const TArray<int>& nodelist = comp->GetElementNodeList(i);
				bool is_fully_contained = true;
				for (int k=1; is_fully_contained && k<=nodelist.Length(); k++)
				{
					Vector3D pos = comp->GetNodePosition(nodelist(k));
					if (!box.IsIn(pos))
					{
						is_fully_contained = false;
					}
				}
				if (is_fully_contained)
				{
					elements.Add(nn+i);
				}
			}
			nn += comp->NElems();
		}
	}
	return elements;
}
int BaseMesh::GetElementAtPos(Vector3D pos)
{
	int visited = 0;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)                              // the array Mesh::Compnenets MAY contain NULL pointers ( as remnant of Glue operations )
		{
			for(int i=1; i<=comp->NElems(); i++)
			{
				MeshElement* el = comp->GetElemPtr(i);
				if(el!=NULL)
				{
					TArray<int> elementnodes = comp->GetElementNodeList(i);
					TArray<Vector3D> nodeposlist(elementnodes.Length());
					for(int j=1; j<=elementnodes.Length(); j++)
					{
						int nnr = elementnodes(j);
						Vector3D nodepos = comp->GetNodePosition(nnr);
						nodeposlist(j) = nodepos;
					}
					Vector3D ploc = el->GetLocPosOfGlobalPos(pos,nodeposlist); 
					int ison = el->TestPLocOnElem(ploc);                  
					if(ison==1)
					{
						return visited+i;
					}
				}
			}
			visited += comp->NElems();
		}
	}
	return 0;
}
Vector3D BaseMesh::GetGlobalPosOfLocalPosOnElem(int elnr, Vector3D ploc)
{
	int remaining=elnr;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)
		{
			if(remaining<=comp->NElems())
			{
				MeshElement* el = comp->GetElemPtr(remaining);
				if(el!=NULL)
				{
					TArray<int> elementnodes = comp->GetElementNodeList(remaining);
					TArray<Vector3D> nodeposlist(el->NNodes());
					for(int j=1; j<=el->NNodes(); j++)
					{
						int nnr = elementnodes(j);
	
						Vector3D nodepos = comp->GetNodePosition(nnr);
						nodeposlist(j) = nodepos;
					}
					Vector3D pglob = el->InterpolateVector3D(ploc,nodeposlist);
					return pglob;
				}
			}
			else
			{
				remaining -= comp->NElems();
			}
		}
	}
	return Vector3D(0.,0.,0.);
}

Vector3D BaseMesh::GetLocalPosOnElem(int elnr, Vector3D pos)
{
	//NOTE AD: Revisit ?!?
	int remaining=elnr;
	for(int j=1; j<=NComp(); j++)
	{
		MeshComponent* comp = GetMeshComponent(j);
		if(comp!=NULL)
		{
			if(remaining<=comp->NElems())
			{
				MeshElement* el = comp->GetElemPtr(remaining);
				if(el!=NULL)
				{
					TArray<int> elementnodes = comp->GetElementNodeList(remaining);
					TArray<Vector3D> nodeposlist(el->NNodes());
					for(int j=1; j<=el->NNodes(); j++)
					{
						int nnr = elementnodes(j);
	
						Vector3D nodepos = comp->GetNodePosition(nnr);
						nodeposlist(j) = nodepos;
					}
					Vector3D ploc = el->GetLocPosOfGlobalPos(pos,nodeposlist); 
					return ploc;
				}
			}
			else
			{
				remaining -= comp->NElems();
			}
		}
	}
	return Vector3D(0.,0.,0.);
}
int BaseMesh::GetFacesFromNodes(const TArray<int>& nodenumbers, const TArray<int>& elementnumbers)
{
	this->ClearNodesToElementsList();
	this->AllocateNodesToElementsList();
	this->AssembleNodesToElementsList();
	
// prepare a tick-off list
	const int MAX_FACES_PER_ELEMENT = 6;
	enum TickOffStatus { TProcessing = 'p', TNotVisited = 'o', TFailed = 'n', TPassed = 'y'	};

	TArray<int> tickoffelemnr;
	if(elementnumbers.Length()==0) { tickoffelemnr = NaturalNumbers(NElems()); }
	else { tickoffelemnr.CopyFrom(elementnumbers); }

	TArray<char> tickoffmatrix;
	tickoffmatrix.SetLen(tickoffelemnr.Length()*MAX_FACES_PER_ELEMENT); 
	tickoffmatrix.SetAll(TNotVisited);

// nodelist is with MBS numbering, the elements are still in their original components 
// - mirror the access and nodelist conversion as in the ::ElemsToMBS() function
// - alternatively introduce a buffer in BaseMesh that stores mapped nodelists when generated...

// process the input nodes: identify all complete faces containig the active node that can be built from the set.
	for(int i=1; i<=nodenumbers.Length(); i++) // loop over all nodes to find faces for
	{
		int curnodenr = nodenumbers(i);
		
		for(int j=1; j<=node2element(curnodenr)->Length(); j++) // loop over all elements that contain this node
		{
// determine base mesh relative indices for the elemenet
			int curelemnr = node2element(curnodenr)->Get(j);     // element number in MBS numbering
			int tickoffelnr = tickoffelemnr.Find(curelemnr);     // index number in TickOff List

// determine component relative indices for the element
			MeshComponent* comp = NULL;
			//int prevcomp = 0;
			int curelnrincomp = curelemnr;
			for(int k=1; k<=NComp(); k++)
			{
				comp = GetMeshComponent(k);
				if(comp != NULL)
				{
					int nelincomp = comp->NElems();
					if(curelnrincomp <= nelincomp)
					{
						break;
					}
					else 
					{ 
						curelnrincomp -= nelincomp; 
					}
				}
			}

			if(tickoffelnr>0)
			{
				MeshElement* pcurelem = GetElementPtr(curelemnr); 
				for(int k=1; k<= pcurelem->NFaces(); k++) // loop over all faces of an element
				{
					TArray<int> facenodelist = comp->GetElementFaceNodeList(curelnrincomp,k);

// face node list contains nodnumbers in component numbering, convert as for ::ElemsToMBS
					comp->MapNodeListComponentToMesh(facenodelist);         // nodelist containing mesh node numbering 
					MapNodeListToMBSNodes(facenodelist);                    // nodelist containing mbs node numbering
					
					if(facenodelist.Find(curnodenr) > 0) // process only faces that contain the current node 
					{
						int tickoffindex = k + (tickoffelnr-1)*MAX_FACES_PER_ELEMENT;
						if(tickoffmatrix(tickoffindex) == TNotVisited)
						{
							tickoffmatrix(tickoffindex) = TProcessing;

							int iscomplete = 1; 
							for(int l=1; l<=facenodelist.Length(); l++) // loop over all nodes of the face
							{
								int found = nodenumbers.Find(facenodelist(l));
								if(found==0)
								{
									iscomplete = 0;
									break; // breaks the loop over nodes
								}
							}

							if(iscomplete==0) { tickoffmatrix(tickoffindex) = TFailed; }
							else { tickoffmatrix(tickoffindex) = TPassed; }
						}
					}
				}
			}
		} // end loop elements of current node
	} // end loop input nodes

// 
	FaceSetA facesfromnodes(mbs,"Faces from Nodes");
	for(int i=1; i<= tickoffelemnr.Length(); i++)
	{
		for(int j=1; j <=MAX_FACES_PER_ELEMENT; j++)
		{
			int curelemnr = tickoffelemnr(i);
			
			MeshComponent* comp = NULL;
			int curelnrincomp = curelemnr;
			for(int k=1; k<=NComp(); k++)
			{
				comp = GetMeshComponent(k);
				if(comp != NULL)
				{
					int nelincomp = comp->NElems();
					if(curelnrincomp <= nelincomp) {	break; }
					else { curelnrincomp -= nelincomp; }
				}
			}
			int tickoffindex = j + (i-1)*MAX_FACES_PER_ELEMENT;			
			if(tickoffmatrix(tickoffindex) == TPassed)
			{
				TArray<int> facenodelist = comp->GetElementFaceNodeList(curelnrincomp,j);	
// face node list contains nodnumbers in component numbering, convert as for ::ElemsToMBS
				comp->MapNodeListComponentToMesh(facenodelist);         // nodelist containing mesh node numbering 
				MapNodeListToMBSNodes(facenodelist);                    // nodelist containing mbs node numbering
				
				TArray<int> facenodelocallist(facenodelist.Length());
				for (int m=1; m<=facenodelist.Length(); m++)
				{
					if (facenodelist(m) > 0)
					{
						int global_node_nr = facenodelist(m);
						//Element * el= mbs->GetElementPtr(i);
						Element * el= mbs->GetElementPtr(curelemnr);
						for (int ln=1; ln<=el->NNodes(); ln++)
						{
							if (el->GetNode(ln).NodeNum() == global_node_nr)
							{
								facenodelocallist(m) = ln;
								continue;
							}
						}
					}
				}
				// compute area of facet - TODO: make an element funciton for that - also used by AreaLoad / AreaTraction 
				Vector3D posN1 = ( (GenericNode*) mbs->GetNodePtr(facenodelist(1)) ) ->GetRefPos();
				Vector3D posN2 = ( (GenericNode*) mbs->GetNodePtr(facenodelist(2)) ) ->GetRefPos();
				Vector3D vecN1N2 = posN2-posN1;
				double area = 0.;

				if(facenodelist.Length()==3 || facenodelist.Length()==6) 
				{ 
					Vector3D posN3 = ( (GenericNode*) mbs->GetNodePtr(facenodelist(3)) ) ->GetRefPos(); 
					Vector3D vecN1N3 = posN3-posN1;
					Vector3D normal = vecN1N2.Cross(vecN1N3);
					area = normal.Norm()*0.5;
				}
				else if(facenodelist.Length()==4 || facenodelist.Length()==8) 
				{ 
					Vector3D posN3 = ( (GenericNode*) mbs->GetNodePtr(facenodelist(3)) ) ->GetRefPos(); 
					Vector3D vecN1N3 = posN3-posN1;
					Vector3D normal = vecN1N2.Cross(vecN1N3);
					area = normal.Norm()*0.5;

					Vector3D posN4 = ( (GenericNode*) mbs->GetNodePtr(facenodelist(4)) ) ->GetRefPos(); 
					Vector3D vecN1N4 = posN4-posN1;
					normal = vecN1N4.Cross(vecN1N3);
					area += normal.Norm()*0.5;	
				}
				else { ; }
				//facesfromnodes.AddEntry(i, j, facenodelist.Length(), facenodelist, facenodelocallist, 0.0);
				facesfromnodes.AddEntry(curelemnr, j, facenodelist.Length(), facenodelist, facenodelocallist, area);
			}
		}
	}
	int setnr = mbs->AddSet(&facesfromnodes);

	return setnr;
}

int BaseMesh::ComputeSurface(FaceSetA& rv_surface, const TArray<int>& filter_elementnumbers)
{
	this->ClearNodesToElementsList();
	this->AllocateNodesToElementsList();
	this->AssembleNodesToElementsList();

	TArray<int> loopelems;
	if(filter_elementnumbers.Length()==0) { loopelems = NaturalNumbers(NElems()); }
	else { loopelems.CopyFrom(filter_elementnumbers); }

// process all the elements, for each face try to find a matching face in a neighboring element
	for(int i=1; i<=loopelems.Length(); i++)
	{
		int curelemnr = loopelems(i);     // element number in mesh numbering

// determine component relative indices for the element
		MeshComponent* comp = NULL;
		int curelnrincomp = curelemnr;
		for(int k=1; k<=NComp(); k++)
		{
			comp = GetMeshComponent(k);
			if(comp != NULL)
			{
				int nelincomp = comp->NElems();
				if(curelnrincomp <= nelincomp)
				{
					break;
				}
				else 
				{ 
					curelnrincomp -= nelincomp; 
				}
			}
		}

// loop over all faces of the current element
		MeshElement* pcurelem = GetElementPtr(curelemnr); 
		int nfaces = pcurelem->NFaces();
		for(int k=1; k<=nfaces; k++)
		{
			TArray<int> facenodelist = comp->GetElementFaceNodeList(curelnrincomp,k);			
// face node list contains nodnumbers in component numbering, convert as for ::ElemsToMBS
			comp->MapNodeListComponentToMesh(facenodelist);         // nodelist containing mesh node numbering 

// find an element that contains all of the face nodes (other than the current element) as intersection of the node2element list for all face nodes
			TArray<int> intersection = *(node2element(facenodelist(1)));
			for(int fnnr=2; fnnr<=facenodelist.Length(); fnnr++)
			{
				TArray<int> nodeXelems = *(node2element(facenodelist(fnnr)));
				intersection = intersection.Intersection(nodeXelems);
			}

			if(intersection.Length()==0) { assert("epic fail"); }	// at least one entry for current element expected
			if(intersection.Length()>2)  { assert("epic fail"); }	// at most two entries expected, a face can only be shared by two elements	
			if(intersection.Length()==2) 
			{ 
// TODO: check orientation ?
			}
			else //(intersection.Length()==1)
			{
				MapNodeListToMBSNodes(facenodelist);                    // nodelist containing mbs node numbering
				int mbscurelnr = this->mbselemnrs(curelemnr);
				rv_surface.AddEntry(mbscurelnr, k, facenodelist.Length(), facenodelist,  pcurelem->GetLocalFace(k).GetNodeList(), 0.0);
			}
		}
	}
	return 0;
}

int BaseMesh::SetsToMBS(MBS* mbs)
{
	for(int i=1; i<=meshsets.Length(); i++)
	{
		if(meshsets(i)->GetType() == TMBSSetElements)
		{
			ElementSet* tmpMBSSet = (ElementSet*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		} 
		else if(meshsets(i)->GetType() == TMBSSetGlobalNodes) // change the global node numbers accordingly
		{
			GlobalNodeSet* tmpMBSSet = (GlobalNodeSet*) meshsets(i)->GetCopy();
			MapNodeListToMBSNodes(tmpMBSSet->GlobalNodeNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		}
		else if(meshsets(i)->GetType() == TMBSSetLocalNodesA) // change the element number 
		{
			LocalNodeSetA* tmpMBSSet = (LocalNodeSetA*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		} 
		else if(meshsets(i)->GetType() == TMBSSetLocalNodesB) // change the element number 
		{
			LocalNodeSetB* tmpMBSSet = (LocalNodeSetB*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		}
		else if(meshsets(i)->GetType() == TMBSSetGlobalCoords) // do nothing
		{
			;
		}
		else if(meshsets(i)->GetType() == TMBSSetLocalCoordsA) // change the element number
		{
			LocalCoordSetA* tmpMBSSet = (LocalCoordSetA*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		}
		else if(meshsets(i)->GetType() == TMBSSetLocalCoordsB) // change the element number
		{
			LocalCoordSetB* tmpMBSSet = (LocalCoordSetB*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		}
		else if (meshsets(i)->GetType() == TMBSSetFaces) // change global node numbers and element numbers 
		{
			FaceSetA* tmpMBSSet = (FaceSetA*) meshsets(i)->GetCopy();
			MapElemListToMBSElems(tmpMBSSet->ElementNumbers());
			for(int faceNr=1; faceNr<=tmpMBSSet->GetNrOfEntries(); faceNr++)
			{
				TArray<int>& curFaceGlobalNodeNrs = tmpMBSSet->FaceNodeNumber(faceNr);
				MapNodeListToMBSNodes(curFaceGlobalNodeNrs);
			}
			mbs->AddSet(tmpMBSSet);
			delete tmpMBSSet;
		}
		else 
		{
			mbs->UO() << mystr("BaseMesh::SetsToMBS - Mesh Set #") + mystr(i) + mystr(" cannot be converted into MBS Set.\n");
		}
	}
	return meshsets.Length();
}

int BaseMesh::UpdateStatistics()
{
	TArray<int> elementdomainnumber(NElems());
	elementdomainnumber.SetAll(0);
	TArray<int> nodedomainnumber(NNodes());
	nodedomainnumber.SetAll(0);
	int ndomains=0;
	for(int i=1; i<=NElems(); i++)
	{
		Element* el = GetMBS()->GetElementPtr(mbselemnrs(i));
		Node* nd = GetMBS()->GetNodePtr(el->NodeNum(1));
		int domainnumber = nd->GetBodyInd();
		
		elementdomainnumber(i) = domainnumber;
		if(ndomains<domainnumber) { ndomains = domainnumber; }
	}
	for(int i=1; i<=NNodes(); i++)
	{
		Node* nd = GetMBS()->GetNodePtr(mbsnodenrs(i));
		int domainnumber = nd->GetBodyInd();

		nodedomainnumber(i) = domainnumber;
		if(ndomains<domainnumber) { ndomains = domainnumber; }
	}

	domainstatistics.Flush();
	for(int j=1; j<=ndomains; j++)
	{
		DomainStatistics* pCurDomain = new DomainStatistics(GetMBS());
		TArray<int> nnrs_mesh,enrs_mesh; // keep a temporary list with the mesh index numbers, required for all cases with more than one Mesh 
		for(int i=1; i<=NNodes(); i++)
		{
			if(nodedomainnumber(i)==j) { pCurDomain->nodes.Add(mbsnodenrs(i)); nnrs_mesh.Add(i); }
		}
		for(int i=1; i<=NElems(); i++)
		{
			if(elementdomainnumber(i)==j) { pCurDomain->elements.Add(mbselemnrs(i)); enrs_mesh.Add(i); }
		}		

// optional identification of Domain surface
		if(flag_autocomputesurface)
		{
			ComputeSurface(*(pCurDomain->surface),enrs_mesh); 
// the resulting elementnumbers / facenodes are in MBS numbering

// identify all nodes used in surfaceelements
			int maxnodenr = mbsnodenrs.Last(); int nsurfnodes = 0;
			TArray<int> issurfacenode; issurfacenode.SetLen(maxnodenr);	issurfacenode.SetAll(0);	
			int maxelemnr = mbselemnrs.Last(); int nsurfelems = 0;
			TArray<int> issurfaceelem; issurfaceelem.SetLen(maxelemnr);	issurfaceelem.SetAll(0);

			for(int i=1; i<pCurDomain->surface->GetNrOfEntries(); i++)
			{
				TArray<int>& gnnr = pCurDomain->surface->FaceNodeNumber(i);
				for(int ii=1; ii<=gnnr.Length(); ii++)
				{
					if(!issurfacenode(gnnr(ii)))
					{
						issurfacenode(gnnr(ii)) = 1;
						nsurfnodes++;
					}
				}
				int elnr = pCurDomain->surface->GetElementNumber(i);
				if(!issurfaceelem(elnr))
				{
					issurfaceelem(elnr) = 1;
					nsurfelems++;
				}
			}
			pCurDomain->surfacenodes.Flush();
			for(int i=1; i<=issurfacenode.Length(); i++)
			{
				if(issurfacenode(i)) 
				{ 
					pCurDomain->surfacenodes.Add(i); 
				}
			}
			pCurDomain->surfaceelements.Flush();
			for(int i=1; i<=issurfaceelem.Length(); i++)
			{
				if(issurfaceelem(i)) 
				{ 
					pCurDomain->surfaceelements.Add(i); 
				}
			}
		}
		domainstatistics.Add(pCurDomain);
	}
	return ndomains;
}

/*
#include "KinematicPairs.h"
int BaseMesh::AppyAutomaticConstraints(MBS* mbs)
{
	if(coincident_pairs.Length()==0)
	{
		return 0;
	}

	GlobalNodeSet coincident_nodes(mbs,mystr("coincident_nodes"));
	for(int i=1; i<=coincident_pairs.Length(); i++)
	{
		coincident_nodes.AddEntry(coincident_pairs(i).Get(1));
		coincident_nodes.AddEntry(coincident_pairs(i).Get(2));
	}

	//(1,9,3,11) -> (1,(-1,-1, 0)), (4,(-1,-1, 0)), (1,(-1,+1, 0)), (4,(-1,+1, 0)) 
	//              (1, 1st),       (4, 1st),       (1, 3rd),       (4, 3rd)    
	coincident_nodes.Convert();

// NOTES:
// can not typecast to BasePointJoint* ( that is where the TArray<Vector3D> loccoords is introduced )
// no accessfunctions to these TArray in Constraint
// functions to set:  SetPos1ToLocalCoord, ...
// simply #include "KinematicPairs.h" gives linker error - SetPosToLocalCoord implemented in .cpp - project dependency :(
// BUT: Hack exists

// parameters for the automatically generated constraints - some of those should be changeable
	mystr elemtype = mystr("GenericBodyJoint");
	int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);

	for(int i=1; i<=coincident_nodes.GetNrOfEntries()-1; i+=2) // -1 to remove unpaired entries 
	{
		int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr);
		BasePointJoint* pbpj = (BasePointJoint*) mbs->GetElementPtr(cnr);
		
		int enr1 = coincident_nodes.GetElementNumber(i);
		Vector3D& lpos1 = coincident_nodes.GetLocalPosition(i);
	//	pbpj->SetPos1ToLocalCoord(enr1, lpos1); // dependency

		int enr2 = coincident_nodes.GetElementNumber(i+1);
		Vector3D& lpos2 = coincident_nodes.GetLocalPosition(i+1);
	//	pbpj->SetPos2ToLocalCoord(enr1, lpos2); // dependency

		TArray<int>& el = pbpj->GetElementsArray();
		TArray<Vector3D>& lp = pbpj->GetLoccoordsArray();

		el(1) = enr1;
		el(2) = enr2;
		lp(1) = lpos1;
		lp(2) = lpos2;
	}

	return 1;
}
*/

int BaseMesh::MapNodeListToMBSNodes(TArray<int>& nodelist)
{
	for(int i=1; i<=nodelist.Length(); i++)
	{
		int meshnr = nodelist(i);
		if(meshnr<0) { nodelist(i) = -1; } //	mesh number never set properly - complete code in class ??? (most likely derived from SourceExtended) - node was replaced...
		else if(meshnr>NNodes()) { nodelist(i) = -1; } //	mesh number never set properly - complete code in class ??? (most likely derived from SourceExtended) - node was replaced...
		else if(meshnr==0) { nodelist(i) = 0; } // node is not used by mesh
		else // meshnr>0 
		{
			int mbsnr = mbsnodenrs(meshnr);
			nodelist(i) = mbsnr;
		}
	}
	return nodelist.Length();
}

int BaseMesh::MapElemListToMBSElems(TArray<int>& elemlist)
{
	for(int i=1; i<=elemlist.Length(); i++)
	{
		int meshnr = elemlist(i);
		if(meshnr<0) { elemlist(i) = -1; } //	mesh number never set properly - complete code in class ??? (most likely derived from SourceExtended) - element was replaced...
		else if(meshnr>NElems()) { elemlist(i) = -1; } //	mesh number never set properly - complete code in class ??? (most likely derived from SourceExtended) - element was replaced...
		else if(meshnr==0) { elemlist(i) = 0; } // element is not used by mesh
		else // meshnr>0
		{
			int mbsnr = mbselemnrs(meshnr);
			elemlist(i) = mbsnr;
		}
	}
	return elemlist.Length();
}

/*
int BaseMesh::RemoveRedundantNodes()
{
// * compute mapping current node numbers to revised node numbers
// * replace nodennumbers in all elements, sets, ...
// * replace also in coincident pairs, remove any now identival entries
// * erase the nodes from nodelist 

//        +                          G
//       /|                         /|
//	+---y---+      7---8    3---4  F |
//	|   | + |      |   |    |   |  | E
//	|   |/  |      |   |    |   |  |/
//	+---X---z      5---6    1---2  D
//      |   |               B---C 
//      |   |               |   |
//      +---+               |   |
//                          9---A
// example for :  redundant_pairs =  {(1,6),(1,11),(6,11), +2 more @ y,z (3,8), (2,12)}
// at pos "X"     coincident_pairs = {(1,13),(6,13),(11,13), +2 more @ y (3,15),(8,15)}

// loop over all entries of redundant_pairs and assemble a mapping for nodes
//   1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16
//   |  |  |  |  |  |  |  |  |   |   |   |   |   |   |   |
//   v  v  v  v  v  v  v  v  v   v   v   v   v   v   v   v
//   1  2  3  4  6  1  7  3  9  10   1   2  13  14  15  16		// redundant nodes have same number now
	TArray<int> mapping = NaturalNumbers(meshnodes.Length());
 	for(int i=1; i<=redundant_pairs.Length(); i++) // loop over all identified redundant pairs
	{
		int nnr1 = redundant_pairs(i)(1); // this node will be kept
		int nnr2 = redundant_pairs(i)(2);	// this node will be removed
		
		if(mapping(nnr2)>nnr1)
		{
			mapping(nnr2) = nnr1;
		}
	}
// store the numbers of the nodes to delete in an array ( mapping(i) differs from i )
// adjust the other nodenumbers down by the amount of redundate nodes that are to be removed
//   1  2  3  4  5  1  7  3  9  10   1   2  13  14  15  16		// redundant nodes have same number now
//   |  |  |  |  |  |  |  |  |   |   |   |   |   |   |   |
//   v  v  v  v  v  v  v  v  v   v   v   v   v   v   v   v
//   1  2  3  4  5  1  6	3  7   8	 1	 2	 9  10  11  12		// list of new node numbers
//                  6     8         11  12										// indices in nodelist for the entries to be removed
	TArray<int> todelete;
	int decrement = 0;
	for(int i=1; i<=mapping.Length(); i++)
	{
		if(i != mapping(i))
		{
			todelete.Add(i);
			decrement++;
		}
		else
		{
			mapping(i) -= decrement;
		}
	}

// replace in all (Mesh)Components
// (1,2,3,4)->(1,2,3,4), (5,6,7,8)->(5,1,6,3), (9,10,11,12)->(7,8,1,2), (13,14,15,16)->(9,10,11,12)
	for(int i=1; i<=meshcomponents.Length(); i++)
	{
		MeshComponent* comp = meshcomponents(i);
		TArray<int>& compnodes = comp->NodeNrsList();
		for(int j=1; j<=compnodes.Length(); j++)
		{
			compnodes(j) = mapping(compnodes(j));
		}
	}

// replace in all (Mesh)Elements
// (1,2,4,3)->(1,2,4,3) - (5,6,8,7)->(5,1,3,6) - (9,10,12,11)->(7,8,2,1) - (13,14,15,16)->(9,10,12,11)
	for(int i=1; i<=meshelements.Length(); i++)
	{
		MeshElement* el = meshelements(i);
		for(int j=1; j<=el->NNodes(); j++)
		{
			el->SetNode(j, mapping(el->GetNode(j)));
		}
	}

// replace in all (GlobalNode)Sets
	for(int i=1; i<=meshsets.Length(); i++) // loop over all (GlobalNode)Sets and replace the redundant node
	{
		if(meshsets(i)->GetType() == TMBSSetGlobalNodes)
		{
			//GlobalNodeSet* gns = (GlobalNodeSet*) sets(i);
			TArray<int>& nnrs = ((GlobalNodeSet*)meshsets(i))->GlobalNodeNumbers();
			for(int j=1; j<=nnrs.Length(); j++)
			{
				nnrs(j) = mapping(nnrs(j));
			}
		}
	}
// update list of coincident pairs and remove identical entries
// {(1,13),(6,13),(11,13), +2 more @ y (3,15),(8,15)}
// {(1,9), (1,9), (1,9),               (3,11),(3,11)} --> (1,9), (3,11)
	for(int i=1; i<=coincident_pairs.Length(); i++)
	{
		coincident_pairs(i)(1) = mapping(coincident_pairs(i)(1));
		coincident_pairs(i)(2) = mapping(coincident_pairs(i)(2));
	}
	Quicksort(coincident_pairs);
	for(int i=1; i<=(coincident_pairs.Length()-1); i++)
	{
		if(coincident_pairs(i) == coincident_pairs(i+1))
		{
			coincident_pairs.Erase(i+1);
			i--;
		}
	}
// delete (Mesh)Nodes
	while(todelete.Length()>0)
	{
		meshnodes.Erase(todelete.Last());
		todelete.Erase(todelete.Length());
	}

	redundant_pairs.Flush();
	return 1;
}

*/