//#**************************************************************
//#
//# filename:             model_tools.cpp
//#
//# author:               Peter Gruber
//#
//# generated:						December 2012
//# description:          Modeling-functionality for automated node and element generation
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#include "mbs_interface.h"
#include "element.h"
#include "material.h"
#include "body2d.h"
#include "body3d.h"
#include "ANCFBeamBE2D.h"
#include "ANCFBeamShear2D.h"
//$ PG 2014-6-5:[ for testing ANCFBeamShearFE2DLinear, ANCFBeamShearFE2DQuadratic
#include "ANCFBeamShearFE2D.h"
//$ PG 2014-6-5:]
#include "ANCFBeam3DTorsion.h"
#include "ANCFBeamShear3D.h"
#include "node.h"
#include "graphicsConstants.h"

#include "model_tools.h"


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++    Automated Node Generation     +++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//Funktion mit StartPunkt, Orientierung und Lnge, templatisiert fr alle Knoten!!!!==>PG!!!!!! JG
//IVector GenerateNodesOnStraightLine(MBS* mbs, const ANCFNodeS1rot1_3D& first_node, const ANCFNodeS1rot1_3D& last_node, int number_of_nodes)
//{
//	int number_of_segments = number_of_nodes - 1;
//	assert(number_of_segments>0 && "GenerateNodesOnStraightLine: number of nodes > 1 required");
//
//	Vector3D first_node_pos = first_node.Pos();
//	Vector3D last_node_pos = last_node.Pos();
//	Vector3D vector_to_next_node(last_node_pos - first_node_pos); 
//	double beam_length = vector_to_next_node.Norm();
//	assert(beam_length > 0 && "first and last node must not have same position");
//	vector_to_next_node *= 1./number_of_segments;
//
//	// create nodes
//	IVector node_list(number_of_nodes);
//	Vector3D pos(first_node_pos);
//
//	Vector3D ref_kardan_angles(first_node.GetRefAngles());
//	assert( (ref_kardan_angles - last_node.GetRefAngles()).Norm() < 1e-14 && "first and last node must have same orientation!");
//
//	for (int i=1; i<=number_of_nodes; i++)
//	{
//		ANCFNodeS1rot1_3D node(mbs);
//		node.SetANCFNodeS1rot1_3D(pos,ref_kardan_angles);
//		node_list(i) = mbs->AddNode(&node);
//		pos += vector_to_next_node;
//	}
//
//	return node_list;
//}
//
//IVector GenerateNodesOnStraightLine(MBS* mbs, const ANCFNodeS2S3_3D& first_node, const ANCFNodeS2S3_3D& last_node, int number_of_nodes)
//{
//	int number_of_segments = number_of_nodes - 1;
//	assert(number_of_segments>0 && "GenerateNodesOnStraightLine: number of nodes > 1 required");
//
//	Vector3D first_node_pos = first_node.Pos();
//	Vector3D last_node_pos = last_node.Pos();
//	Vector3D vector_to_next_node(last_node_pos - first_node_pos); 
//	double beam_length = vector_to_next_node.Norm();
//	assert(beam_length > 0 && "first and last node must not have same position");
//	vector_to_next_node *= 1./number_of_segments;
//
//	// create nodes
//	IVector node_list(number_of_nodes);
//	Vector3D pos(first_node_pos);
//
//	Vector3D refslope2 = first_node.GetRefSlope2();
//	Vector3D refslope3 = first_node.GetRefSlope3();
//	assert((refslope2 - last_node.GetRefSlope2()).Norm() + (refslope3 - last_node.GetRefSlope3()).Norm() < 2e-14 && "first and last node must have same orientation!");
//
//	for (int i=1; i<=number_of_nodes; i++)
//	{
//		ANCFNodeS2S3_3D node(mbs);
//		node.SetANCFNodeS2S3_3D(pos,refslope2,refslope3);
//		node_list(i) = mbs->AddNode(&node);
//		pos += vector_to_next_node;
//	}
//	return node_list;
//}

template<class TNode>
IVector GenerateNodesOnStraightLine(MBS* mbs, const TNode& first_node, const TNode& last_node, int number_of_nodes)
{
	int dim = first_node.Dim();
	assert((dim == 2 || dim == 3) && "GenerateNodesOnStraightLine: wrong dimension!");
	assert(number_of_nodes>1 && "GenerateNodesOnStraightLine: number of nodes > 1 required!");
	IVector node_list(number_of_nodes);

	Vector3D first_node_pos = first_node.Pos();
	Vector3D last_node_pos = last_node.Pos();
	Vector3D vector_to_next_node(last_node_pos - first_node_pos); 
	double beam_length = vector_to_next_node.Norm();
	assert(beam_length > 0 && "first and last node must not have same position");
	vector_to_next_node /= number_of_nodes - 1;

	// create nodes
	Vector3D pos(first_node_pos);

	for (int i=1; i<=number_of_nodes; i++)
	{
		TNode node(mbs);
		node.SetBodyInd(first_node.GetBodyInd());
		node.Pos() = pos;

		if (first_node.HasRefSlope1()) node.SetRefSlope1(first_node.GetRefSlope1());
		if (first_node.HasRefSlope2()) node.SetRefSlope2(first_node.GetRefSlope2());
		if (dim == 3)
		{
			if (first_node.HasRefSlope3()) node.SetRefSlope3(first_node.GetRefSlope3());
			if (first_node.HasRefAngles()) node.SetRefAngles(first_node.GetRefAngles());
		}
		node_list(i) = mbs->AddNode(&node);
		pos += vector_to_next_node;
	}

	return node_list;
}

//$ PG 2014-6-5: function hard coded for class ANCFNodeS1rot1_3D, has to be templated yet
IVector GenerateNodesOnArc(MBS* mbs, const ANCFNodeS1rot1_3D& first_node, const ANCFNodeS1rot1_3D& last_node, int number_of_nodes)
{
		int number_of_segments = number_of_nodes - 1;
	assert(number_of_segments>0 && "GenerateNodesOnArc: number of nodes > 1 required");

	Vector3D first_node_e1, last_node_e1;
	Vector3D first_node_e2, last_node_e2;
	Vector3D first_node_e3, last_node_e3;
	for (int i=1; i<=3; i++)
	{
		first_node_e1(i) = first_node.GetLocalFrame(i,1);
		first_node_e2(i) = first_node.GetLocalFrame(i,2);
		first_node_e3(i) = first_node.GetLocalFrame(i,3);
		last_node_e1(i) = last_node.GetLocalFrame(i,1);
		last_node_e2(i) = last_node.GetLocalFrame(i,2);
		last_node_e3(i) = last_node.GetLocalFrame(i,3);
	}

	double delta_e3 = (first_node_e3 - last_node_e3).Norm();
	assert(delta_e3 >= 1e-14 && "GenerateNodesOnArc: use GenerateNodesOnStraightLine(...) for generation of nodes on a straight line");

	double radius = (first_node.Pos() - last_node.Pos()).Norm() / delta_e3;
	Vector3D center = first_node.Pos() - radius*first_node_e3;

	// test for consistency: 
	assert((last_node_e2-first_node_e2).Norm() < 1e-14);  // by convention: e2 must be normal to the plane in which the arc lies

	// some renaming, in order to make the generation algorithm independent from interface (to be substituted for acceleration of computation)
	Vector3D tangential_vector_at_first_node(first_node_e1);
	Vector3D normal(first_node_e2);
	Vector3D vector_to_first_node(first_node_e3);
	Vector3D vector_to_last_node(last_node_e3);
			
	double delta_phi = acos(vector_to_first_node*vector_to_last_node)/number_of_segments;

	//create nodes
	IVector node_list(number_of_nodes);
	for (int i=0; i<=number_of_segments; i++)
	{
		Vector3D unit_vector_to_node_i(cos(i*delta_phi)*vector_to_first_node + sin(i*delta_phi)*tangential_vector_at_first_node);
		Vector3D tangential_vector_at_node_i(normal.Cross(unit_vector_to_node_i));

		Vector3D pos(center+radius*unit_vector_to_node_i);
		
		Matrix3D local_frame;
		for (int j=1; j<=3; j++) 
		{
			local_frame(j,1)=tangential_vector_at_node_i(j);
			local_frame(j,2)=normal(j);
			local_frame(j,3)=unit_vector_to_node_i(j);
		}
		
		Vector3D ref_angles;
		RotMatToKardanAngles(local_frame, ref_angles);
		ANCFNodeS1rot1_3D node(mbs); node.SetANCFNodeS1rot1_3D(pos, ref_angles);
		node_list(i+1) = mbs->AddNode(&node);
	}

	return node_list;
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++    Automated Beam Generation with given Nodes    ++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//function generates n=(node_list.Length()-1) elements of type ANCFBeam3DTorsion (in serial order). The elements are located between nodes with MBS-index as in node_list, and adds those new elements to mbs.
//  material_idx:                index (as listed in MBS) of material to be used for each of the elements.
//  update_directors:            if set to 0, then orientation of the directors is fixed during computation steps, else the directors are updated.
//  kinematic_computation_mode:  0 .. exact kinematic terms + 5th order gaussian integration (slow), 1 .. exact terms + 1st order lobatto integration (fast), 2 .. constant mass matrix approximation (fastest).
IVector GenerateANCFBeam3DTorsion(MBS* mbs, const IVector& node_list, int material_idx, int update_directors, int kinematic_computation_mode)
{
	int number_of_elements = node_list.Length() - 1;
	assert(number_of_elements>0);

	//create elements
	IVector	element_list(number_of_elements);

	for (int i=1; i<=number_of_elements; i++)
	{
		ANCFBeam3DTorsion beam_element(mbs);
		beam_element.SetANCFBeam3DTorsion(node_list(i),node_list(i+1), material_idx, colblue, update_directors, kinematic_computation_mode);
		element_list(i) = mbs->AddElement(&beam_element);
	}

	return element_list;
}

IVector GenerateANCFBeamShear3D(MBS* mbs, const IVector& node_list, int material_idx, int is_quadratic, int beam_formulation)
{
	assert(node_list.Length() > 1 && "number of nodes has to be > 1");
	if (is_quadratic)
	{
		assert(node_list.Length() % 2 == 1 && "odd number of nodes is required here!");
	}

	int number_of_elements = node_list.Length() - 1;
	if (is_quadratic)
	{
		number_of_elements /= 2;
	}
	
	//create elements
	IVector element_list;
	int iterator = 1;
	for (int i=1; i<=number_of_elements; i++)
	{
		if (!is_quadratic)
		{
			//get the global node index (the local index of nodes at each element is 1->2)
			int node1_global_idx = node_list(iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShear3DLinear linear_beam_element(mbs);
			linear_beam_element.SetANCFBeamShear3DLinear(node1_global_idx, node2_global_idx, material_idx, colblue, beam_formulation); //2-noded beam element
			//linear_beam_element.SetANCFBeamShear3DLinear(node1_global_idx, node2_global_idx, material_idx, colblue); //2-noded beam element
			//linear_beam_element.SetBeamFormulation(beam_formulation);
			element_list.Add(mbs->AddElement(&linear_beam_element));
		}
		else
		{
			//get the global node index (the local index of nodes at each element is 1->3->2)
			int node1_global_idx = node_list(iterator);
			int node3_global_idx = node_list(++iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShear3DQuadratic quadratic_beam_element(mbs);
			quadratic_beam_element.SetANCFBeamShear3DQuadratic(node1_global_idx, node2_global_idx, node3_global_idx, material_idx, colblue, beam_formulation); //3-noded beam element
			//quadratic_beam_element.SetANCFBeamShear3DQuadratic(node1_global_idx, node2_global_idx, node3_global_idx, material_idx, colblue); //3-noded beam element
			//quadratic_beam_element.SetBeamFormulation(beam_formulation);
			element_list.Add(mbs->AddElement(&quadratic_beam_element));
		}
	}

	return element_list;
}

IVector GenerateANCFBeamBE2D(MBS* mbs, const IVector& node_list, int material_idx)
{
	assert(node_list.Length() > 1 && "number of nodes has to be > 1");
	int number_of_elements = node_list.Length() - 1;
	
	//create elements
	IVector element_list;
	int iterator = 1;
	for (int i=1; i<=number_of_elements; i++)
	{
		//get the global node index (the local index of nodes at each element is 1->2)
		int node1_global_idx = node_list(iterator);
		int node2_global_idx = node_list(++iterator);
		ANCFBeamBE2D beam_element(mbs);
		beam_element.SetANCFBeamBE2D(node1_global_idx, node2_global_idx, material_idx, colblue);
		element_list.Add(mbs->AddElement(&beam_element));
	}
	return element_list;
}

IVector GenerateANCFBeamShear2D(MBS* mbs, const IVector& node_list, int material_idx, int is_quadratic)
{
	assert(node_list.Length() > 1 && "number of nodes has to be > 1");
	if (is_quadratic)
	{
		assert(node_list.Length() % 2 == 1 && "odd number of nodes is required here!");
	}

	int number_of_elements = node_list.Length() - 1;
	if (is_quadratic)
	{
		number_of_elements /= 2;
	}
	
	//create elements
	IVector element_list;
	int iterator = 1;
	for (int i=1; i<=number_of_elements; i++)
	{
		if (!is_quadratic)
		{
			//get the global node index (the local index of nodes at each element is 1->2)
			int node1_global_idx = node_list(iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShear2D linear_beam_element(mbs);
			linear_beam_element.SetANCFBeamShear2D(node1_global_idx, node2_global_idx, material_idx, colblue);
			element_list.Add(mbs->AddElement(&linear_beam_element));
		}
		else
		{
			//get the global node index (the local index of nodes at each element is 1->3->2)
			int node1_global_idx = node_list(iterator);
			int node3_global_idx = node_list(++iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShear2D quadratic_beam_element(mbs);
			quadratic_beam_element.SetANCFBeamShear2D(node1_global_idx, node2_global_idx, node3_global_idx, material_idx, colblue); //3-noded beam element
//$ PG 2014-6-5:[ test
			//quadratic_beam_element.SetFormulation(2);
			//quadratic_beam_element.SetRI(1);
//$ PG 2014-6-5:]
			element_list.Add(mbs->AddElement(&quadratic_beam_element));
		}
	}
	return element_list;
}

IVector GenerateANCFBeamShearFE2D(MBS* mbs, const IVector& node_list, int material_idx, int is_quadratic, int use_contmech_formulation)
{
	assert(node_list.Length() > 1 && "number of nodes has to be > 1");
	if (is_quadratic)
	{
		assert(node_list.Length() % 2 == 1 && "odd number of nodes is required here!");
	}

	int number_of_elements = node_list.Length() - 1;
	if (is_quadratic)
	{
		number_of_elements /= 2;
	}
	
	//create elements
	IVector element_list;
	int iterator = 1;
	for (int i=1; i<=number_of_elements; i++)
	{
		if (!is_quadratic)
		{
			//get the global node index (the local index of nodes at each element is 1->2)
			int node1_global_idx = node_list(iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShearFE2DLinear linear_beam_element(mbs);
			linear_beam_element.SetANCFBeamShearFE2DLinear(node1_global_idx, node2_global_idx, material_idx, colblue); //2-noded beam element
			linear_beam_element.SetUseContinuumMechanicsFormulation(use_contmech_formulation);
			linear_beam_element.SetReducedIntegration(1);
			linear_beam_element.SetReducedIntegrationPoisson(1);
			element_list.Add(mbs->AddElement(&linear_beam_element));
		}
		else
		{
			//get the global node index (the local index of nodes at each element is 1->3->2)
			int node1_global_idx = node_list(iterator);
			int node3_global_idx = node_list(++iterator);
			int node2_global_idx = node_list(++iterator);
			ANCFBeamShearFE2DQuadratic quadratic_beam_element(mbs);
			quadratic_beam_element.SetANCFBeamShearFE2DQuadratic(node1_global_idx, node2_global_idx, node3_global_idx, material_idx, colblue); //3-noded beam element
			quadratic_beam_element.SetUseContinuumMechanicsFormulation(use_contmech_formulation);
//$ PG 2014-6-5:[ or
//			Vector xc1(4); xc1.SetAll(0);
//			xc1(4)=0.1; 
//			Vector xc2(xc1), xc3(xc1);
//			xc1(1)=mbs->GetNode(node1_global_idx).Pos().X();
//			xc2(1)=mbs->GetNode(node2_global_idx).Pos().X();
//			xc3(1)=mbs->GetNode(node3_global_idx).Pos().X();
//			quadratic_beam_element.SetANCFBeamShearFE2DQuadratic(1, xc1, xc2, xc3, node1_global_idx, node2_global_idx, node3_global_idx, material_idx, Vector3D(2,0.1,0.1), colblue);
//			quadratic_beam_element.SetUseContinuumMechanicsFormulation(2);
//			quadratic_beam_element.UseReducedIntegrationPoisson();
//$ PG 2014-6-5:]
			quadratic_beam_element.SetReducedIntegration(1);
			quadratic_beam_element.SetReducedIntegrationPoisson(1);
			element_list.Add(mbs->AddElement(&quadratic_beam_element));
		}
	}

	return element_list;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++    Automated Beam Generation w/o given Nodes    +++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//template<class TBeam>
//Node3D* CreateNode3D(MBS* mbs)
//{
//	ANCFNodeS1rot1_3D n;
//	return &n;
//}
//
//template<class TBeam>
//IVector GenerateBeam3DStraight<TBeam>(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx)
//{
//	int number_of_nodes = number_of_elements + 1;
//	
//	Node3D *first_node = CreateNode3D(mbs);
//	Node3D *last_node = CreateNode3D(mbs);
//	first_node->SetBodyInd(body_idx);
//	last_node->SetBodyInd(body_idx);
//
//	beam_axis_e1.Normalize();
//	beam_cross_section_e2.Normalize();
//	Vector3D beam_cross_section_e3 = beam_axis_e1.Cross(beam_cross_section_e2);
//	
//	if (first_node.HasRefSlope1()) first_node.SetRefSlope1(beam_axis_e1);
//	if (first_node.HasRefSlope2()) first_node.SetRefSlope2(beam_axis_e2);
//	if (first_node.HasRefSlope3()) first_node.SetRefSlope3(beam_axis_e3);
//	if (first_node.HasRefAngles())
//	{
//		Matrix3D A(
//			beam_axis_e1.X(),beam_cross_section_e2.X(),beam_cross_section_e3.X(),
//			beam_axis_e1.Y(),beam_cross_section_e2.Y(),beam_cross_section_e3.Y(),
//			beam_axis_e1.Z(),beam_cross_section_e2.Z(),beam_cross_section_e3.Z());
//		Vector3D phi;
//		RotMatToKardanAngles(A, phi);
//		first_node->SetRefAngles(phi);
//		last_node->SetRefAngles(phi);
//	}
//
//
//	first_node->Pos() = start_point;
//	last_node->Pos() = start_point + length*beam_axis_e1;
//	
//	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS1rot1_3D(??)>(mbs, first_node, last_node, number_of_nodes);
//	return GenerateBeam<TBeam>(mbs, node_list, material_idx, colblue);
//}

IVector GenerateANCFBeam3DTorsionAtLine(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int update_directors, int kinematic_computation_mode)
{
	int number_of_nodes = number_of_elements + 1;
	
	ANCFNodeS1rot1_3D first_node(mbs);

	beam_axis_e1.Normalize();
	beam_cross_section_e2.Normalize();
	Vector3D beam_cross_section_e3 = beam_axis_e1.Cross(beam_cross_section_e2);
	
	Matrix3D A(
		beam_axis_e1.X(),beam_cross_section_e2.X(),beam_cross_section_e3.X(),
		beam_axis_e1.Y(),beam_cross_section_e2.Y(),beam_cross_section_e3.Y(),
		beam_axis_e1.Z(),beam_cross_section_e2.Z(),beam_cross_section_e3.Z());
	Vector3D phi;
	RotMatToKardanAngles(A, phi);
	first_node.SetRefAngles(phi);

	ANCFNodeS1rot1_3D last_node(first_node);
	first_node.Pos() = start_point;
	last_node.Pos() = start_point + length*beam_axis_e1;
	
	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS1rot1_3D>(mbs, first_node, last_node, number_of_nodes);
	return GenerateANCFBeam3DTorsion(mbs, node_list, material_idx, update_directors, kinematic_computation_mode);
}

IVector GenerateANCFBeamShear3DAtLine(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic, int beam_formulation)
{
	int number_of_nodes = number_of_elements*(1+is_quadratic) + 1;
	
	ANCFNodeS2S3_3D first_node(mbs), last_node(mbs);
	beam_axis_e1.Normalize();
	beam_cross_section_e2.Normalize();
	Vector3D beam_cross_section_e3 = beam_axis_e1.Cross(beam_cross_section_e2);
	first_node.SetANCFNodeS2S3_3D(start_point, beam_cross_section_e2, beam_cross_section_e3);
	last_node.SetANCFNodeS2S3_3D(start_point + length*beam_axis_e1, beam_cross_section_e2, beam_cross_section_e3);

	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS2S3_3D>(mbs, first_node, last_node, number_of_nodes);
	return GenerateANCFBeamShear3D(mbs, node_list, material_idx, is_quadratic, beam_formulation);
}

IVector GenerateANCFBeamBE2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, double length, int number_of_elements, int material_idx)
{
	int number_of_nodes = number_of_elements + 1;
	
	ANCFNodeS1_2D first_node(mbs), last_node(mbs);
	beam_axis_e1.Normalize();
	first_node.SetANCFNodeS1_2D(start_point, beam_axis_e1);
	last_node.SetANCFNodeS1_2D(start_point + length*beam_axis_e1, beam_axis_e1);
	
	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS1_2D>(mbs, first_node, last_node, number_of_nodes);
	return GenerateANCFBeamBE2D(mbs, node_list, material_idx);
}

IVector GenerateANCFBeamShear2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, Vector2D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic)
{
	int number_of_nodes = number_of_elements*(1+is_quadratic) + 1;
	
	ANCFNodeS2_2D first_node(mbs), last_node(mbs);
	beam_axis_e1.Normalize();
	beam_cross_section_e2.Normalize();
	first_node.SetANCFNodeS2_2D(start_point, beam_cross_section_e2);
	last_node.SetANCFNodeS2_2D(start_point + length*beam_axis_e1, beam_cross_section_e2);
	
	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS2_2D>(mbs, first_node, last_node, number_of_nodes);
	return GenerateANCFBeamShear2D(mbs, node_list, material_idx, is_quadratic);
}

IVector GenerateANCFBeamShearFE2DAtLine(MBS* mbs, Vector2D start_point, Vector2D beam_axis_e1, Vector2D beam_cross_section_e2, double length, int number_of_elements, int material_idx, int is_quadratic, int use_contmech_formulation)
{
	int number_of_nodes = number_of_elements*(1+is_quadratic) + 1;
	
	ANCFNodeS2_2D first_node(mbs), last_node(mbs);
	beam_axis_e1.Normalize();
	beam_cross_section_e2.Normalize();
	first_node.SetANCFNodeS2_2D(start_point, beam_cross_section_e2);
	last_node.SetANCFNodeS2_2D(start_point + length*beam_axis_e1, beam_cross_section_e2);
	
	IVector node_list = GenerateNodesOnStraightLine<ANCFNodeS2_2D>(mbs, first_node, last_node, number_of_nodes);
	return GenerateANCFBeamShearFE2D(mbs, node_list, material_idx, is_quadratic, use_contmech_formulation);
}



//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++    Apply Loads or Constraints to Beam    ++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void ApplyLoadToElements(MBS* mbs, const IVector& element_list, const MBSLoad& load)
{
	for (int i=1; i<=element_list.Length(); i++)
	{
		int idx = element_list(i);
		assert(idx > 0);
		mbs->GetElementPtr(idx)->AddLoad(load);
	}
}








/////////////////////////////////////////
//// Generate BeamShearQuadraticAtLine //
/////////////////////////////////////////
//
////this function is used to generate beam elements at a line defined by the point "start_point", the length "length" and 
////the directions "beam_axis_e1", which is a vector in direction of the beam axis and "beam_cross_section_e2" which is the cross section principal axis e2 (e3 is computed automatically from the cross product)
////the axis "beam_axis_e1" and "beam_cross_section_e2" do not need to be unit vectors 
////the number of beam elements generated corresponds to "number_of_elements"
////"quadratic" indicates, if the linear (quadratic=0) or the quadratic (quadratic=1) version of the element is used
////"material_idx" is the index to the material/beam_properties which is used for the beam definitions
////the list of generated elements is returned withing the "element_list"
////returns element_list
////author: Johannes Gerstmayr
//IVector GenerateANCFBeamShear3DAtLine(MBS* mbs, Vector3D start_point, Vector3D beam_axis_e1, Vector3D beam_cross_section_e2, double length, int number_of_elements,
//																 int quadratic, int material_idx)
//{
//	//create nodes and elements
//	//Vector3D last_node_position;
//	beam_axis_e1.Normalize();
//	beam_cross_section_e2.Normalize();
//	
//	Vector3D end_point = length*beam_axis_e1+start_point;
//	int nel = number_of_elements;
//
//	int element_segments_loc = 1+quadratic; //1..linear, 2..quadratic shape functions
//	IVector element_list;
//	int dofs_per_node = 9; //for ANCF degenerated shear deformable beam
//
//	Vector3D vector_to_next_node = (1./(element_segments_loc*nel))*(end_point - start_point);
//
//	Vector3D node_position = start_point;
//
//	//generate Vector in direction of beam axis
//	Vector3D e_1 = beam_axis_e1;
//	Vector3D e_2 = beam_cross_section_e2;
//	Vector3D e_3 = beam_axis_e1.Cross(beam_cross_section_e2);
//  
//	//create nodes
//	ANCFNodeS2S3_3D node(mbs);
//	//node.SetANCFNodeS2S3_3D(node_position, e_2, e_3);
//	//int node_idx = mbs->AddNode(&node);
//	int node_idx;
//	int node_offset = 0;
//
//	for (int m=1; m<=element_segments_loc*nel+1; m++)
//	{
//		node.SetANCFNodeS2S3_3D(node_position, e_2, e_3);
//		node_idx = mbs->AddNode(&node);
//		if (m==1) {node_offset = node_idx-1;} 
//
//		node_position += vector_to_next_node;
//	}
//
//	//create elements
//	for (int i=1; i<=nel; i++)
//	{
//		//get the global node index (the local index of nodes at each element is 1->3->2)
//		int node1_global_idx, node2_global_idx, node3_global_idx;
//		node1_global_idx = i+node_offset;
//		node2_global_idx = node1_global_idx+1;
//
//		if(quadratic)
//		{
//			node3_global_idx = 2*i + node_offset;
//			node1_global_idx = node3_global_idx-1;
//			node2_global_idx = node3_global_idx+1;
//		}
//
//		//get the positions of the nodes
//		Vector3D node1_global_position = mbs->GetNode(node1_global_idx).Pos();
//		Vector3D node2_global_position = mbs->GetNode(node2_global_idx).Pos();
//		Vector3D node3_global_position(0.);
//		if(quadratic)
//		{
//			node3_global_position = mbs->GetNode(node3_global_idx).Pos();
//		}
//
//		ANCFBeamShear3DGeneric* element;
//		int element_nr;
//		if (!quadratic)
//		{
//			ANCFBeamShear3DLinear linear_beam_element(mbs);
//			linear_beam_element.SetANCFBeamShear3DLinear(node1_global_idx, node2_global_idx, material_idx, colblue); //2-noded beam element
//			element_nr = mbs->AddElement(&linear_beam_element);
//		}
//		else
//		{   
//			ANCFBeamShear3DQuadratic quadratic_beam_element(mbs);
//			quadratic_beam_element.SetANCFBeamShear3DQuadratic(node1_global_idx, node2_global_idx, node3_global_idx, material_idx, colblue);
//			element_nr = mbs->AddElement(&quadratic_beam_element);
//		}
//		element = dynamic_cast<ANCFBeamShear3DGeneric*>(&mbs->GetElement(element_nr));
//		element_list.Add(element_nr);
//
//		////set penalty for structural mechanics based formulation
//		//for (int k=1; k<=penalty.Length(); k++)
//		//{
//		//	element->SetPenalty(k, penalty(k));
//		//}
//		////set cont. mech. based (CMF) or struct. mech. based formulation (SMF)
//		//element->SetBeamFormulation(BeamFormulation);
//
//		////integration order for SMF is fixed as follows:
//		////int order_x_bend = 5;
//		////int order_x_thickness_eps = 5;
//		////int order_x_axial_eps = 3;
//
//		////set integration order for CMF:
//		//element->SetOrderAxial(edc->TreeGetInt("ANCFBeamShear3D.order_axial"));
//		//element->SetOrderCrossSectional(edc->TreeGetInt("ANCFBeamShear3D.order_crosssectional"));
//
//		//int flexible_body_nr = mbs->AddElement(&element);
//		//apply loads
//		//if (distributed_load_idx != 0) {mbs->GetElement(element_nr).AddLoad(distributed_load_idx);}
//	}
//
//	return element_list;
//}

