//#**************************************************************
//#
//# filename:             sensorProcessorsSpecific.h
//#
//# author:               Gerstmayr Johannes
//#												Vetyukov Yury
//#
//# generated:						February-June 2012
//# description:          
//#                       
//# remarks:						  HotInt sensors - specific sensor processors
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#pragma once

#include "sensors.h"

class OffsetScaleSensorProcessor : public SensorProcessor //$EDC$
{
	double scale;
	double offset;

public:
	OffsetScaleSensorProcessor(double scale, double offset)
	{
		this->scale = scale;
		this->offset = offset;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		return value * scale + offset;
	}
	virtual SensorProcessor * GetCopy()
	{
		OffsetScaleSensorProcessor * sp = new OffsetScaleSensorProcessor(scale, offset);
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("OffsetScaleSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(scale,"scaling_factor"); ed.SetToolTipText("scaling factor of signal"); edc.Add(ed);
		ed.SetDouble(offset,"offset"); ed.SetToolTipText("(scaled) signal offset"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		scale = edc.TreeGetDouble("scaling_factor",1.0);
		offset = edc.TreeGetDouble("offset",0.0);
		return 2;
	}
};

class MaxAbsValueSensorProcessor : public SensorProcessor
{
	double currentValue;

public:
	MaxAbsValueSensorProcessor()
	{
		currentValue = 0;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		currentValue = max(currentValue, fabs(value));
		return currentValue;
	}
	virtual SensorProcessor * GetCopy()
	{
		MaxAbsValueSensorProcessor * sp = new MaxAbsValueSensorProcessor();
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("MaxAbsValueSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(currentValue,"current_value"); ed.SetLocked(1); ed.SetToolTipText("current maximum absolute value"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};

class ReferenceComparerSensorProcessor : public SensorProcessor
{
	MathFunction referenceSignal;
	MBS * mbs;

public:
	ReferenceComparerSensorProcessor(MBS * mbs, MathFunction & referenceSignal)	// the provided object should be deleted by the client code
	{
		this->referenceSignal = referenceSignal;
		this->mbs = mbs;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		return value - referenceSignal.Evaluate(time);
	}
	virtual SensorProcessor * GetCopy()
	{
		ReferenceComparerSensorProcessor * sp = new ReferenceComparerSensorProcessor(mbs, referenceSignal);
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("ReferenceComparerSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
				
		ElementDataContainer edcMF;
		referenceSignal.GetElementData(edcMF);
		ed.SetEDC(&edcMF,"referenceSignal"); ed.SetLocked(1); ed.SetToolTipText("reference Signal for the Sensor Processor as MathFunction"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		//TODO: read mathfunction ?

		return 0;
	}
};

//$ DR 2012-08-24 ReferenceSensorComparerSensorProcessor added
// computes the difference of the actual value of the sensor and the actual value of the reference sensor
// output = sensor - reference
class ReferenceSensorComparerSensorProcessor : public SensorProcessor
{
	Sensor *  sensP;

public:
	ReferenceSensorComparerSensorProcessor(Sensor * referenceSensor)	
	{
		this->sensP = referenceSensor;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		return value - this->sensP->GetCurrentValueWithSensorProcessing(time);
	}
	virtual SensorProcessor * GetCopy()
	{
		ReferenceSensorComparerSensorProcessor * sp = new ReferenceSensorComparerSensorProcessor(sensP);
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("ReferenceSensorComparerSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
				
		int ownNum = sensP->GetOwnNum();
		ed.SetInt(ownNum,"reference_sensor"); ed.SetLocked(1); ed.SetToolTipText("index number of the reference sensor"); edc.Add(ed);
		
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		//TODO: read sensornumber and reset Sensor*  ?

		return 0;
	}
};


class FFTSensorProcessor : public SensorProcessor
{
public:
	virtual bool NeedsSignalTimeHistoryForPostComputationProcessing() { return true; }
	virtual bool NeedsFileForPostComputationProcessing() { return true; }
	virtual TArray<double> DoPostComputationProcessing(TArray<double> & sensorHistoryTimes, TArray<double> & sensorHistoryValues, ofstream & outputFile)
	{
		// TODO
		assert(0);
		return TArray<double>();
	}
	virtual FFTSensorProcessor * GetCopy()
	{
		FFTSensorProcessor * sp = new FFTSensorProcessor();
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("FFTSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};

//$ YV 2014-02-28: added
// this sensor processor filters higher frequencies out of the signal (and thus makes it smoother)
// by solving a simple differential equation
// T * output'[t] + output[t] = input[t]
// the time constant T defines the frequency range (the higher T is, the lower is the frequency range).
// T shall not be smaller than the time step of the simulation.
// the filter is designed such that repeated invokations for the same time instant
// (e.g. owing to post-Newton (discontinuous) iterations)
// shall be consistently recognized and processed.
class LowPassFilterSensorProcessor : public SensorProcessor
{
	double T;
	double previousOutput;
	double currentOutput;
	double previousTime;
	double prePreviousTime;

public:
	LowPassFilterSensorProcessor(double T)
	{
		this->T = T;
		previousTime = -1e10;
	}
	virtual double ProcessCurrentValue(double time, double value);
	virtual SensorProcessor * GetCopy()
	{
		LowPassFilterSensorProcessor * sp = new LowPassFilterSensorProcessor(T);
		return sp;
	}
// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("LowPassFilterSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(),"processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(T,"time_constant"); ed.SetToolTipText("T * output'[t] + output[t] = input[t]"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		T = edc.TreeGetDouble("time_constant",1.0);
		return 1;
	}
};

//$ PG 2021-06-14: processor computes backwards derivative of sensor signal
class DerivativeSensorProcessor : public SensorProcessor
{
	double valueOld;     // old sensor value
	double timeOld;      // old sensor time
	double currentValue;

public:
	DerivativeSensorProcessor()
	{
		valueOld = 0;
		timeOld = -1e10;
		currentValue = 0.;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		if (time > timeOld)
			currentValue = (value - valueOld)/(time - timeOld);

		valueOld = value;
		timeOld = time;
		
		return currentValue;
	}
	virtual SensorProcessor * GetCopy()
	{
		DerivativeSensorProcessor * sp = new DerivativeSensorProcessor();
		return sp;
	}
	// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("DerivativeSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(), "processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(currentValue, "current_value"); ed.SetLocked(1); ed.SetToolTipText("current derivative value"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};

//$ PG 2021-06-14: processor computes backwards derivative of sensor signal
class IntegralSensorProcessor : public SensorProcessor
{
	double valueOld;     // old sensor value
	double timeOld;      // old sensor time
	double currentValue;

public:
	IntegralSensorProcessor()
	{
		timeOld = -1e10;
		currentValue = 0.;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		if (time > timeOld)
		{
			currentValue += value*(time - timeOld);
			timeOld = time;
		}			
		
		return currentValue;
	}
	virtual SensorProcessor * GetCopy()
	{
		IntegralSensorProcessor * sp = new IntegralSensorProcessor();
		return sp;
	}
	// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("IntegralSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(), "processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(currentValue, "current_value"); ed.SetLocked(1); ed.SetToolTipText("current integral value"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};

//$ PG 2021-06-14: processor computes backwards derivative of sensor signal
class ResettingIntegralSensorProcessor : public SensorProcessor
{
	double valueOld;     // old sensor value
	double timeOld;      // old sensor time
	double currentValue;

public:
	ResettingIntegralSensorProcessor()
	{
		timeOld = -1e10;
		currentValue = 0.;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		if (time > timeOld)
		{
			if (value <= 0)
				currentValue = 0;
			else
				currentValue += value*(time - timeOld);
			
			timeOld = time;
		}

		return currentValue;
	}
	virtual SensorProcessor * GetCopy()
	{
		ResettingIntegralSensorProcessor * sp = new ResettingIntegralSensorProcessor();
		return sp;
	}
	// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("ResettingIntegralSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(), "processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(currentValue, "current_value"); ed.SetLocked(1); ed.SetToolTipText("current resetting integral value"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};

class BandwidthSensorProcessor : public SensorProcessor
{
	double minValue;
	double maxValue;
	double currentValue;

public:
	BandwidthSensorProcessor(double minValue, double maxValue)
	{
		this->minValue = minValue;
		this->maxValue = maxValue;
		currentValue = 0.;
	}
	virtual double ProcessCurrentValue(double time, double value)
	{
		currentValue = max(minValue, min(maxValue, value));
		return currentValue;
	}
	virtual SensorProcessor * GetCopy()
	{
		BandwidthSensorProcessor * sp = new BandwidthSensorProcessor(minValue, maxValue);
		return sp;
	}
	// script functionality - gui representation
	virtual mystr GetSensorProcessorSpec() { return mystr("BandwidthSensorProcessor"); }
	virtual void GetElementData(ElementDataContainer& edc)
	{
		ElementData ed;
		ed.SetText(GetSensorProcessorSpec().c_str(), "processor_type"); ed.SetLocked(1); ed.SetToolTipText("type of the sensor processor"); edc.Add(ed);
		ed.SetDouble(currentValue, "current_value"); ed.SetLocked(1); ed.SetToolTipText("current bandwidth value"); edc.Add(ed);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		return 0;
	}
};
