//#**************************************************************
//#
//# filename:             sensorsSpecific.cpp
//#
//# author:               Gerstmayr Johannes
//#												Vetyukov Yury
//#
//# generated:						February-June 2012
//# description:          
//#                       
//# remarks:						  HotInt sensors - specific sensors
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


#include "mbs_interface.h"
#include "element.h"
#include "sensors.h"
#include "sensorsSpecific.h"
#include "elementdataaccess.h"
#include "timeintlog.h"
#include "FiniteElement3D.h"
#include "FiniteElement2D.h"
#include "IntegrationRule.h"
#include "set.h"
#include <limits>
#include <cmath>


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FieldVariableElementSensor
// 2012-11 added by YV
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

double FieldVariableElementSensor::GetCurrentValue(double time)
{
	if (average_mode)
	{
		if (dimension == 3)
		{
			const FiniteElement3D &fe3D = static_cast<const FiniteElement3D&>(GetElement());
			const IntegrationRule *intRule = fe3D.GetIntegrationRule(IntegrationRule::IVT_Load);

			//ConstVector<FEmaxDOF> xgCache;
			//fe3D.SetComputeCoordinates(xgCache);
			//ConstMatrix<FEmaxDOF> agrad(3, fe3D.NS());
			//Matrix3D F;
			Matrix3D jac;
			double volume = 0.;
			double value = 0.;

			// Integration routine
			for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
			{
				//IntegrationRule::IntegrationPoint ip = intRule->integrationPoints.Get(i);
				//fe3D.GetGrad(ip, agrad);
				//fe3D.Gradu(xgCache, agrad, F);
				//F(1, 1) += 1; F(2, 2) += 1; F(3, 3) += 1;
				fe3D.GetJacobi(jac, ip.Point());
				double fact = fabs(/*F.Det() * */jac.Det()) * ip.Weight();
				volume += fact;
				value += fact*GetElement().GetFieldVariableValue(fieldVariableDescriptor, ip.Point(), false);
			}
			return value / volume;
		}
		if (dimension == 2)
		{
			const FiniteElement2D &fe2D = static_cast<const FiniteElement2D&>(GetElement());
			const IntegrationRule *intRule = fe2D.GetIntegrationRule(IntegrationRule::IVT_Load);
			//ConstVector<FE2DmaxDOF> xgCache;
			//fe2D.SetComputeCoordinates(xgCache);
			//ConstMatrix<2*FE2DmaxDOF> agrad(2, fe2D.NS());
			//Matrix2D F;
			Matrix2D jac;
			double area = 0.;
			double value = 0.;

			// Integration routine
			for (IntegrationPointsIterator ip(intRule); !ip.IsEnd(); ++ip)
			{
				//IntegrationRule::IntegrationPoint ip = intRule->integrationPoints.Get(i);
				//fe2D.GetGrad(ip, agrad);
				//fe2D.Gradu(xgCache, agrad, F);
				//F(1, 1) += 1; F(2, 2) += 1; F(3, 3) += 1;
				fe2D.GetJacobi(jac, ip.Point2D());
				double fact = fabs(/*F.Det() * */jac.Det()) * ip.Weight();
				area += fact;
				value += fact*GetElement().GetFieldVariableValue(fieldVariableDescriptor, ip.Point2D(), false);
			}
			return value / area;
		}
	}
	if(localNodeNumber == 0)
	{
		//if(!localPosition2D) 
		//	return GetElement().GetFieldVariableValue(fieldVariableDescriptor, localPosition, false);
		//else
		//	return GetElement().GetFieldVariableValue(fieldVariableDescriptor, (const Vector2D &)localPosition, false);
		if(dimension==3)
			return GetElement().GetFieldVariableValue(fieldVariableDescriptor, localPosition, false);
		else if(dimension==2)
			return GetElement().GetFieldVariableValue(fieldVariableDescriptor, (const Vector2D &)localPosition, false);
		else
			return GetElement().GetFieldVariableValue(fieldVariableDescriptor, (const double &)localPosition.X(), false);
	}
	return GetElement().GetFieldVariableValue(fieldVariableDescriptor, localNodeNumber, false);
}

void FieldVariableElementSensor::SetFVESPos3D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector3D localPosition)
{
	SetElementNumber(elementNumber);
	fieldVariableDescriptor = fieldVariableDescriptor_;
	localNodeNumber = 0;
	this->localPosition = localPosition;
	//localPosition2D = false;
	dimension = 3;
	AfterSetFunction();
}

void FieldVariableElementSensor::SetFVESPos2D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector2D localPosition)
{
	SetElementNumber(elementNumber);
	fieldVariableDescriptor = fieldVariableDescriptor_;
	localNodeNumber = 0;
	this->localPosition.X() = localPosition.X();
	this->localPosition.Y() = localPosition.Y();
	this->localPosition.Z() = 0;
	//localPosition2D = true;
	dimension = 2;
	AfterSetFunction();
}

void FieldVariableElementSensor::SetFVESNode(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, int localNodeNumber)
{
	//$ YV 2012-10-22 with a node number we do not need to differ between 2D and 3D, as we use the same function from the element to retrieve the value
	SetElementNumber(elementNumber);
	fieldVariableDescriptor = fieldVariableDescriptor_;
	this->localNodeNumber = localNodeNumber;
	// AD: include dimension of node here ?
	AfterSetFunction();
}

void FieldVariableElementSensor::CopyFrom(const Sensor& s)
{
	ElementSensor::CopyFrom(s);
	FieldVariableElementSensor & S = (FieldVariableElementSensor&)s;
	fieldVariableDescriptor = S.fieldVariableDescriptor;
	localPosition = S.localPosition;
	//localPosition2D = S.localPosition2D;
	localNodeNumber = S.localNodeNumber;
	dimension = S.dimension;
	average_mode = S.average_mode;
}

mystr FieldVariableElementSensor::GetTypeName()
{
	return "FVElementSensor"; //$ DR 2012-10 changed in order to be consistent with CEDCParser
	//if(localNodeNumber > 0)
	//	return "Element_" + mystr(GetElementNumber()) + "-" + "Node_" + mystr(localNodeNumber) + "-" + fieldVariableDescriptor.GetTextualIdentifier();
	//return "Element_" + mystr(GetElementNumber()) + "-" + fieldVariableDescriptor.GetTextualIdentifier();
}

mystr FieldVariableElementSensor::GetSensorTypeTexDescription()	// for the auto generated documentation //$ DR 2012-10 added
{
	mystr descr = mystr("The FieldVariableElementSensor evaluates the value of a field variable at a specified position. There are two possibilities to define this position: \n")
		+ mystr("\\begin{itemize} \n")
		+ mystr("   \\item element number + local position \n")
		+ mystr("   \\item element number + local node number \n")
		+ mystr("\\end{itemize} \n")
		+ mystr("The descriptions of the elements above include a list of available field variables for each element. ")
		+ mystr("Possible field variables are e.g.")
		+ mystr("\\begin{itemize} \n")
		+ mystr("   \\item position, velocity and displacement \n")
		+ mystr("   \\item bryant\\_angle and angular\\_velocity \n")
		+ mystr("   \\item beam\\_axial\\_extension, beam\\_torsion, beam\\_curvature \n")		
		+ mystr("   \\item many more \n")		
		+ mystr("\\end{itemize} \n");
	return descr;
}

bool FieldVariableElementSensor::IsConsistent(mystr & errorStr)
{
	if(!ElementSensor::IsConsistent(errorStr))
		return false;

	// AD: 2013-02-19: included checks for SetFVESNode
	if (localNodeNumber == 0)
	{ // Element & posiiton 
		//if( !( (GetElement().Dim() == 2 && localPosition2D) || (GetElement().Dim() == 3 && !localPosition2D)) )
		//{
		//	errorStr = "Body and sensor dimension are different (2D <==> 3D)\n";
		//	return false;
		//}
	}
	else
	{ // element & node
		return true; // dimension of node is not checked against dimension of element...
	}

	TArray<FieldVariableDescriptor> availableFieldVariables;
	GetElement().GetAvailableFieldVariables(availableFieldVariables);
	if(! (availableFieldVariables.Find(fieldVariableDescriptor)))
	{
		errorStr = "The FieldVariable '" + mystr(fieldVariableDescriptor.GetTextualIdentifier()) + "' is not available in Element " + mystr(GetElementNumber()) +"!\n";
		return false;
	}

	return true;
}

void FieldVariableElementSensor::GetElementData(ElementDataContainer& edc)
{
	ElementSensor::GetElementData(edc);

	ElementData ed;
	ed.SetText(fieldVariableDescriptor.GetTextualIdentifierWithoutComponents(), "field_variable"); ed.SetToolTipText("name of the field variable, e.g. 'position', see the documentation of the elements for the available field variables"); edc.Add(ed);
	ed.SetText(fieldVariableDescriptor.GetTextualIdentifierComponentsOnly(), "component");  ed.SetToolTipText("component of the field variable, e.g. 'x'"); edc.Add(ed);
}

int FieldVariableElementSensor::SetElementData(ElementDataContainer& edc) //$ DR 2012-10 return value changed from bool to int
{
	int rv = 1;
	rv = ElementSensor::SetElementData(edc);

	mystr FV_temp;
	GetElemDataText(GetMBS(),edc,"field_variable",FV_temp,1);

	mystr FVcomp_temp = mystr("");
	GetElemDataText(GetMBS(),edc,"component",FVcomp_temp,1);

	fieldVariableDescriptor.InitializeFromTextualIdentifier(FV_temp + mystr(FieldVariableDescriptor::GetComponentsDelimiter()) + FVcomp_temp);
	
	AfterSetFunction();	// DR 2012-06-25 added
	return rv;
}



// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FieldVariableGlobalPositionSensor
// 2012-11 added by YV
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

double FieldVariableGlobalPositionSensor::GetCurrentValue(double time)
{
	assert(dimension==3);

	Vector3D guess;
	int index=0;
	while (index++ < GetNumberOfRelatedElements())
	{	
		Body3D* b = (Body3D*) &GetRelatedElement(index);
		if (b->FindLocalPosition(guess, globalPosition, globalDirection))
		{
			elementNumber = index;
			localPosition = guess;			
			return b->GetFieldVariableValue(fieldVariableDescriptor, localPosition, false);
		}
	}

	// if not found, reset state 
	elementNumber = 0;
	localPosition.Set(0,0,0);

	return 0;
}

void FieldVariableGlobalPositionSensor::SetFVGPSPos3D(const IVector& relatedElements, FieldVariableDescriptor fieldVariableDescriptor_, const Vector3D& globalPosition, const Vector3D& globalDirection)
{
	SetRelatedElements(relatedElements);
	fieldVariableDescriptor = fieldVariableDescriptor_;
	this->globalPosition = globalPosition;
	this->globalDirection = globalDirection;
	AfterSetFunction();
}

//void FieldVariableElementSensor::SetFVESPos2D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector2D localPosition)
//{
//	SetElementNumber(elementNumber);
//	fieldVariableDescriptor = fieldVariableDescriptor_;
//	localNodeNumber = 0;
//	this->localPosition.X() = localPosition.X();
//	this->localPosition.Y() = localPosition.Y();
//	this->localPosition.Z() = 0;
//	//localPosition2D = true;
//	dimension = 2;
//	AfterSetFunction();
//}

void FieldVariableGlobalPositionSensor::CopyFrom(const Sensor& s)
{
	MultipleElementsSensor::CopyFrom(s);
	FieldVariableGlobalPositionSensor & S = (FieldVariableGlobalPositionSensor&)s;

	fieldVariableDescriptor = S.fieldVariableDescriptor;
	globalPosition = S.globalPosition;
	globalDirection = S.globalDirection;
	dimension = S.dimension;

	elementNumber = S.elementNumber;
	localPosition = S.localPosition;
}

mystr FieldVariableGlobalPositionSensor::GetTypeName()
{
	return "FVGlobalPositionSensor";
}

mystr FieldVariableGlobalPositionSensor::GetSensorTypeTexDescription()	// for the auto generated documentation //$ DR 2012-10 added
{
	mystr descr = mystr("The FieldVariableGlobalPositionSensor measures a particular field variable of a set of elements at the first intersection found \n")
		+ mystr("\\begin{itemize} \n")
		+ mystr("   \\item with a given plane (in case of beam elements), or \n")
		+ mystr("   \\item with a given straight line (in case of plate/shell elements), or \n")
		+ mystr("   \\item with a given global position (in case of solid elements). \n")
		+ mystr("\\end{itemize} \n")
		+ mystr("The user specifies a list of related elements (related\\_elements), whith which this sensor may operate. ")
		+ mystr("In sequential order it is analysed if an element intersects with the given global hyperplane, global straight line, or global position. ")
		+ mystr("As soon as an intersection is found (in terms of element number and a local position) the fieldvariable value at this position is returned. ")
		+ mystr("If no intersection is found, then the sensor returns 0. ")
		+ mystr("Two member variables are required in total to define all geometric properties of the three modes (a)-(c): one position vector (global\\_position), and one direction vector (global\\_direction).\n");

	return descr;
}

bool FieldVariableGlobalPositionSensor::IsConsistent(mystr & errorStr)
{
	if(!MultipleElementsSensor::IsConsistent(errorStr))
		return false;

	TArray<FieldVariableDescriptor> availableFieldVariables;

	for (int index=1; index<=GetNumberOfRelatedElements(); index++)
	{
		GetRelatedElement(index).GetAvailableFieldVariables(availableFieldVariables);
		if(! (availableFieldVariables.Find(fieldVariableDescriptor)))
		{
			errorStr = "The FieldVariable '" + mystr(fieldVariableDescriptor.GetTextualIdentifier()) + "' is not available in Element " + mystr(GetRelatedElementNumber(index)) +"!\n";
			return false;
		}
	}

	return true;
}

void FieldVariableGlobalPositionSensor::GetElementData(ElementDataContainer& edc)
{
	MultipleElementsSensor::GetElementData(edc);

	ElementData ed;
	ed.SetText(fieldVariableDescriptor.GetTextualIdentifierWithoutComponents(), "field_variable"); ed.SetToolTipText("name of the field variable, e.g. 'position', see the documentation of the elements for the available field variables"); edc.Add(ed);
	ed.SetText(fieldVariableDescriptor.GetTextualIdentifierComponentsOnly(), "component");  ed.SetToolTipText("component of the field variable, e.g. 'x'"); edc.Add(ed);
}

int FieldVariableGlobalPositionSensor::SetElementData(ElementDataContainer& edc) //$ DR 2012-10 return value changed from bool to int
{
	int rv = 1;
	rv = MultipleElementsSensor::SetElementData(edc);

	mystr FV_temp;
	GetElemDataText(GetMBS(),edc,"field_variable",FV_temp,1);

	mystr FVcomp_temp = mystr("");
	GetElemDataText(GetMBS(),edc,"component",FVcomp_temp,1);

	fieldVariableDescriptor.InitializeFromTextualIdentifier(FV_temp + mystr(FieldVariableDescriptor::GetComponentsDelimiter()) + FVcomp_temp);
	
	AfterSetFunction();	// DR 2012-06-25 added
	return rv;
}



// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SingleElementDataSensor
// 2012-11 added by YV
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


double SingleElementDataSensor::GetCurrentValue(double time)
{
	RWdata.time = time;
	RWdata.value = std::numeric_limits<double>::quiet_NaN();
	if (successfully_converted)
	{
		GetElement().ReadSingleElementDataAuto(RWdata);
	}
	else
	{
		RWdata.value = 0;
	}
	return RWdata.value;
}

void SingleElementDataSensor::SetSingleElementDataSensor(int elementNumber, const char * variable_name)
{
	SetElementNumber(elementNumber);

	successfully_converted = RWdata.GetVariableNameAndComponents(variable_name); //convert variable name into name and components in RWdata structure
	RWdata.RWaccess = TRWElementDataRead; //JG 2013-01-11: we only want to read values in sensors

	//if (!successfully_converted) 
	//{
	//	UO() << "ERROR: SingleElementDataSensor: variable name='" << variable_name << "' is not legal. Sensor returns 0.\n";
	//}

	AfterSetFunction();   //$ PG 2013-1-16: call added, in order to be consistent with FieldVariableElementSensor (see , e.g., FieldVariableElementSensor::SetFVESPos3D)
}

bool SingleElementDataSensor::IsConsistent(mystr & errorStr)
{
	if(!ElementSensor::IsConsistent(errorStr))
	{
		return false;
	}
	if(!successfully_converted)
	{
		errorStr = mystr("Variable of Sensor ") +mystr(GetOwnNum()) +mystr(" could not be converted to a ReadWriteElementDataVariableType data structure");
		return false;
	}

	// a single call to this function sets the variable name to the new value if necessary
	GetElement().ReadSingleElementDataAuto(RWdata);

	TArrayDynamic<ReadWriteElementDataVariableType> available_variables;
	GetElement().GetAvailableSpecialValues(available_variables);

	for(int i = 1; i <= available_variables.Length(); i++)
	{		
		if(available_variables(i).IsAvailable(RWdata))
		{
			return true;
		}
	}
	errorStr = "Variable " + RWdata.variable_name + " is not supported by the specified element\n";
	return false;
}

void SingleElementDataSensor::CopyFrom(const Sensor& s)
{
	ElementSensor::CopyFrom(s);
	SingleElementDataSensor & S = (SingleElementDataSensor&)s;
	successfully_converted = S.successfully_converted;
	RWdata = S.RWdata;
}

//$ DR 2012-11-06 added
void SingleElementDataSensor::GetElementData(ElementDataContainer& edc)	//fill in all element data 
{
	ElementSensor::GetElementData(edc);

	mystr text = RWdata.variable_name;
	if(RWdata.comp1)	
	{
		text+=mystr("[")+mystr(RWdata.comp1);
		if(RWdata.comp2)
		{
			text+= mystr(",")+mystr(RWdata.comp2);
		}
		text+=mystr("]");
	}
	ElementData ed;
	ed.SetText(text,"value"); 
	ed.SetToolTipText("special value of the element, use ''[]'' to access vector or matrix values, e.g. force[1] or stress[2,3]"); 
	edc.TreeAdd("",ed);
}

//$ DR 2012-11-06 added
int SingleElementDataSensor::SetElementData(ElementDataContainer& edc) //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
{
	int rv = 1;
	rv = ElementSensor::SetElementData(edc);

	mystr full_string="";
	rv = GetElemDataText(GetMBS(), edc, "value",full_string, 1);
	if(rv)
	{
		rv=RWdata.GetVariableNameAndComponents(full_string);
		successfully_converted = rv;
	}

	AfterSetFunction();	// DR 2012-06-25 added
	return rv;
}

//$ DR 2012-11-06 added
mystr SingleElementDataSensor::GetSensorTypeTexDescription()	// for the auto generated documentation //$ DR 2012-10 added
{
	mystr descr = mystr("The ElementSensor returns special values evaluated in the element. It can be used e.g. for measuring a specific degree of freedom of an element. \n")
		+ mystr("The descriptions of the elements above include a list of available special values for each element.");
	return descr;
}


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// MultipleSensor
// 2012-11 added by DR
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// sensor that consists of a list of sensors
// this replaces the old MBSMultipleSensor
void MultipleSensor::SetMultipleSensor(TArray<int> sensorNumbers, mystr operationMinMaxAverage, Vector weights_i)
{
	sensorNrs = sensorNumbers;
	operation = GetOperationFlagFromTextualIdentifier(operationMinMaxAverage);
	weights = weights_i;
	AfterSetFunction(); //$ PG 2013-1-16: call added, in order to be consistent with FieldVariableElementSensor (see , e.g., FieldVariableElementSensor::SetFVESPos3D)
}

int MultipleSensor::GetOperationFlagFromTextualIdentifier(mystr operation_str)
{
	// this list must coincide with the function GetCurrentValue
	if(operation_str.Compare("average")) { return 1;}
	if(operation_str.Compare("minimum")) { return 2;}
	if(operation_str.Compare("maximum")) { return 3;}
	if(operation_str.Compare("sum"))		 { return 4;}
	if(operation_str.Compare("norm"))		 { return 5;}  //$ PG 2014-3-10: added operation "norm" to MultipleSensor
	if(operation_str.Compare("norm2"))   { return 6;}  //$ PG 2014-3-10: added operation "norm2" to MultipleSensor
	if(operation_str.Compare("maxabs"))   { return 7;}
	else return 0;
}

mystr MultipleSensor::TextualIdentifierOfOperation()
{
	// this list must coincide with the function GetOperationFlagFromTextualIdentifier
	switch(operation)
	{
	case 1:	// average
		return "average";
		break;
	case 2:	// minimum
		return "minimum";
		break;
	case 3:	// maximum
		return "maximum";
		break;
	case 4:	// sum
		return "sum";
		break;
	case 5:	// norm
		return "norm";
		break;
	case 6:	// norm2
		return "norm2";
		break;
	case 7:	// maxabs
		return "maxabs";
		break;
	}
	return "";
}

mystr MultipleSensor::GetSensorTypeTexDescription()	// for the auto generated documentation 
{
	mystr descr = mystr("The MultipleSensor applies mathematical operations to a list of sensors. \n");
	descr+= mystr("The sensor can be used, e.g. to get the maximum or average value of a list of sensors. \n");
	descr+= mystr("The following mathematical operations are possible (use these words for 'operation'): \\\\");
	descr+= mystr("\n \\begin{itemize} \n");

	mystr op = "start";
	int i=1;
	while(!op.Compare(""))
	{
		operation = i;
		op = TextualIdentifierOfOperation();
		i++;
		if(!op.Compare(""))
		{
			descr+= mystr("  \\item ") +op + mystr("\n");
		}
	}
	descr+= mystr("\\end{itemize} \n");
	descr+= mystr("If weights are used, then the value of each sensor is multiplied with the weight before the mathematical operation is performed.\n");
	descr+= mystr("To compute a weighted sum of the first 4 sensors, the entries would be e.g. sensor\\_numbers = [1,2,3,4] and weights = [0.125,0.125,0.25,0.5].\n");
	return descr;
}

double MultipleSensor::GetCurrentValue(double time)
{
	bool use_weights = (weights.Length() != 0);
	double w=1.;
	double value=0.;

	if (operation > 0 && operation < 5) // average, minimum, maximum, sum
	{
		switch(operation)
		{
		case 1:	// average
			for(int i=1; i<=sensorNrs.Length(); i++)
			{
				if(use_weights){w=weights(i);}
				value += w*this->GetSensor(i).GetCurrentValueWithSensorProcessing(time);
			}
			value /= (sensorNrs.Length());	
			break;
		case 2:	// minimum
			if(0<sensorNrs.Length()) 
			{
				if(use_weights){w=weights(1);}
				value = w*this->GetSensor(1).GetCurrentValueWithSensorProcessing(time);
			}
			for(int i=1; i<=sensorNrs.Length(); i++)
			{
				if(use_weights){w=weights(i);}
				value = min(value,w*this->GetSensor(i).GetCurrentValueWithSensorProcessing(time));
			}
			break;
		case 3:	// maximum
			if(0<sensorNrs.Length()) 
			{
				if(use_weights){w=weights(1);}
				value = w*this->GetSensor(1).GetCurrentValueWithSensorProcessing(time);
			}			
			for(int i=1; i<=sensorNrs.Length(); i++)
			{
				if(use_weights){w=weights(i);}
				value = max(value,w*this->GetSensor(i).GetCurrentValueWithSensorProcessing(time));
			}
			break;
		case 4:	// sum
			for(int i=1; i<=sensorNrs.Length(); i++)
			{
				if(use_weights){w=weights(i);}
				value += w*this->GetSensor(i).GetCurrentValueWithSensorProcessing(time);
			}
			break;
		}
	}
	else if (operation == 5) // norm
	{
		for(int i=1; i<=sensorNrs.Length(); i++)
		{
			double sv = this->GetSensor(i).GetCurrentValueWithSensorProcessing(time);
			if(use_weights){w=weights(i);}
			value += w*sv*sv;
		}
		value = sqrt(value);
	}
	else if (operation == 6) // norm2
	{
		for(int i=1; i<=sensorNrs.Length(); i++)
		{
			double sv = this->GetSensor(i).GetCurrentValueWithSensorProcessing(time);
			if(use_weights){w=weights(i);}
			value += w*sv*sv;
		}
	}
	else if (operation == 7) // max abs
	{
		if(0 < sensorNrs.Length()) 
		{
			if(use_weights)
			{
				w = weights(1);
			}
			value = w*fabs(this->GetSensor(1).GetCurrentValueWithSensorProcessing(time));
		}			
		for(int i=1; i<=sensorNrs.Length(); i++)
		{
			if(use_weights)
			{
				w = weights(i);
			}
			value = max(value,w*fabs(this->GetSensor(i).GetCurrentValueWithSensorProcessing(time)));
		}
	}
	else
	{
		GetMBS()->UO(UO_LVL_err) << "ERROR: Operation " << operation << " not defined in SensorProcessor!\n";
	}

	return value;
}

bool MultipleSensor::IsConsistent(mystr & errorStr)
{
	if(sensorNrs.Length()<1)
	{
		errorStr = "MultipleSensor: There has to be at least one sensor in the list!\n";
		return false;
	}

	for(int i=1; i<=sensorNrs.Length(); i++)
	{
		if((sensorNrs(i)<1) || (sensorNrs(i)> GetMBS()->NSensors()) || (sensorNrs(i)==GetOwnNum()))
		{
			errorStr = mystr("MultipleSensor: It is not possible to use sensor ") +mystr(sensorNrs(i)) +mystr(" for the MultipleSensor!\n");
			return false;
		}
	}

	if(!((weights.Length()==0) || (weights.Length() == sensorNrs.Length())))
	{
		errorStr = mystr("MultipleSensor: The length of weights must be equal to the length of the sensor_numbers or zero!\n");
		return false;	
	}


	return true;
}

void MultipleSensor::GetElementData(ElementDataContainer & edc)
{
	Sensor::GetElementData(edc);

	ElementData ed;
	ed.SetText(TextualIdentifierOfOperation() ,"operation"); ed.SetToolTipText("mathematical operation that is applied to the sensor values, e.g. 'maximum','average',..."); edc.Add(ed);

}
int MultipleSensor::SetElementData(ElementDataContainer& edc) 
{
	int rv = 1;
	rv = Sensor::SetElementData(edc);

	mystr temp;
	GetElemDataText(GetMBS(),edc,"operation",temp,1);
	operation = GetOperationFlagFromTextualIdentifier(temp);
	
	AfterSetFunction();	// DR 2012-06-25 added
	return rv;

}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// LoadSensor
// 2012-11 added by DR
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

double LoadSensor::GetCurrentValue(double time)
{
	if(GetLoadNumber())
	{
		return GetLoad().Evaluate(time);
	}
	return -1;
}

mystr LoadSensor::GetSensorTypeTexDescription()	// for the auto generated documentation 
{
	mystr descr = mystr("The LoadSensor can be applied to loads in order to measure their time dependency. \n");
	descr+= mystr("The value $F(t)$ of a load at time $t$ is computed (see the description of the loads for more details) as: \n");
	descr+= mystr("	\\begin{equation} \n");
	descr+= mystr("	   F(t) = f(t) \\vec{F} \n");
	descr+= mystr("	\\end{equation} \n");
	descr+= mystr("The LoadSensor returns the value of the factor $f(t)$ and not the value $F(t)$. ");
	descr+= mystr("If the LoadSensor is used for a scalar load (e.g. GCLoad), then $f(t)$ and $F(t)$ are equal. If the LoadSensor is used for a load vector (e.g. ForceVector3D) then $f(t)$ and $F(t)$ may not be equal.  \\\\");
	descr+= mystr("The LoadSensor can not be shown in the graphical output, because the load does not have a position by itself and may be applied to several elements or nodes.");

	return descr;
}



// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// SystemSensor
// 2013-11 added by PG
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void SystemSensor::SetSystemSensor(Object object, int global_index)
{
	SetObject(object);
	SetGlobalIndex(global_index);
	AfterSetFunction();

	mystr name("Systemsensor");
	if (object != None && object != Unknown)
	{
		name += mystr("-")+object_str;
		if (IsGlobalIndexObject())
		{
			name += mystr(global_index);
		}
	}
	SetSensorName(name);
}

void SystemSensor::SetObject(Object object)
{
	this->object = object;
	if (object != Unknown)
	{
		this->object_str = GetObjectString(object);
	}
}

void SystemSensor::SetGlobalIndex(int global_index)
{
	this->global_index = global_index;
}

mystr SystemSensor::GetSensorTypeTexDescription()	// for the auto generated documentation 
{
	mystr descr = mystr("The SystemSensor can be applied to global degrees of freedom, eigenvalues, several iteration numbers or performance indicators. \n");
	descr+= mystr("It returns the value of the specified quantity at time $t$, ");
	descr+= mystr("and can not be shown in the graphical output, because a system quantity does in general not have a position by itself.");

	return descr;
}

void SystemSensor::PostAssemble()
{
	if (object != MinMaxFieldVariableValue)
	{
		return;
	}

	MBS* mbs = GetMBS();

	if (set_number > 0)
	{
		if (mbs->NSets() < set_number || mbs->GetSetPtr(set_number)->GetType() != TMBSSetElements)
		{
			// the error message will be thrown by the function IsConsistent
			return;
		}
		ElementSet *elSet = static_cast<ElementSet*>(mbs->GetSetPtr(set_number));
		active_elements = elSet->ElementNumbers();
	}
	else
	{
		active_elements = NaturalNumbers(mbs->NE());
	}


	mystr object_str = this->object_str;
	// check if field variables are available
	mystr evaluateVariable_textualIdentifier;
	evaluateVariable_active = false;

	int posAt = object_str.Find(0, '@');
	if (posAt != -1)
	{
		evaluateVariable_textualIdentifier = object_str.Left(posAt);
		object_str = object_str.SubString(posAt + 1, object_str.Length() - 1);
		evaluateVariable_active = true;
	}

	// find last dot
	int p1 = 0;
	int p2 = 0;
	while ((p1 = object_str.Find(p1, '.')) != -1)
	{
		p2 = p1;
		p1++;
	}

	textualIdentifier = object_str.Left(p2);
	mystr typestr = object_str.SubString(p2+1, object_str.Length()-1);

	if (typestr.Compare("min"))
	{
		operation = Min;
	}
	else if (typestr.Compare("max"))
	{
		operation = Max;
	}
	else if (typestr.Compare("avg"))
	{
		operation = Avg;
	}
	else if (typestr.Compare("sum"))
	{
		operation = Sum;
	}

	//check if at least one element knows these FieldVariables or SpecialValues
	
	fvDescriptor.InitializeFromTextualIdentifier(textualIdentifier);
	svDescriptor = ReadWriteElementDataVariableType(textualIdentifier);


	if (evaluateVariable_active)
	{
		evaluateVariable_fvDescriptor.InitializeFromTextualIdentifier(evaluateVariable_textualIdentifier);
		evaluateVariable_svDescriptor = ReadWriteElementDataVariableType(evaluateVariable_textualIdentifier);
	}

	
	TArray<int> filtered_element_list;
	filtered_element_list.ReSize(active_elements.Size());

	for (int i = 1, ne= active_elements.Length(); i <= ne; i++)
	{
		bool has_fv = false;
		Element* el = mbs->GetElementPtr(active_elements(i));
		TArray<FieldVariableDescriptor> available_fv;
		el->GetAvailableFieldVariables(available_fv);

		for (int j = 1; j <= available_fv.Length(); j++)
		{
			if (available_fv(j).VariableType() == fvDescriptor.VariableType())
			{
				has_fv = true;
				isFieldVariableValue = true;
			}
		}

		TArrayDynamic<ReadWriteElementDataVariableType> available_special_values;
		el->GetAvailableSpecialValues(available_special_values);

		for (int j = 1; j <= available_special_values.Length(); j++)
		{
			if (available_special_values(j) == svDescriptor)
			{
				has_fv = true;
				isFieldVariableValue = false;
			}
		}

		if (evaluateVariable_active)
		{
			bool evaluateVariable_has_fv = false;
			for (int j = 1; j <= available_fv.Length(); j++)
			{
				if (available_fv(j).VariableType() == evaluateVariable_fvDescriptor.VariableType())
				{
					evaluateVariable_has_fv = true;
					evaluateVariable_isFieldVariableValue = true;
				}
			}

			TArrayDynamic<ReadWriteElementDataVariableType> available_special_values;
			el->GetAvailableSpecialValues(available_special_values);

			for (int j = 1; j <= available_special_values.Length(); j++)
			{
				if (available_special_values(j) == evaluateVariable_svDescriptor)
				{
					evaluateVariable_has_fv = true;
					evaluateVariable_isFieldVariableValue = false;
				}
			}
			// we search an element which has both
			has_fv = has_fv && evaluateVariable_has_fv;
		}
		if (has_fv)
		{
			filtered_element_list.Add(active_elements(i));
		}
	}
	active_elements = filtered_element_list;
}

bool SystemSensor::IsConsistent(mystr & errorStr)
{
	mystr object_str = this->object_str;

	if(object == Unknown)
	{
		errorStr = mystr("Sensor ") +mystr(GetOwnNum())  + mystr(" has an unknown object '") + object_str + "'\n";
		return false;
	}

	if(IsGlobalIndexObject())
	{
		if ((global_index <= 0)
			//|| (object==EV && ??? size of eigenvalue vector not known at the time this function is called)
			|| (object==DOF && global_index > GetMBS()->GetXact().Length()))
		{
			errorStr = mystr("Sensor ") +mystr(GetOwnNum()) + mystr(" has an invalid global index number: ") + mystr(global_index) + "\n";
			return false;
		}
	}
	else
	{
		if (global_index != 0)
		{
			GetMBS()->UO(UO_LVL_warn) << "WARNING: SystemSensor: global index was set to " << global_index << ", but object '" << object_str << "' does not need a global index!\n";
		}
	}
	if (object == MinMaxFieldVariableValue)
	{
		MBS* mbs = GetMBS();

		if (set_number > 0)
		{
			if (mbs->NSets() < set_number || mbs->GetSetPtr(set_number)->GetType() != TMBSSetElements)
			{
				errorStr = mystr("Sensor ") +mystr(GetOwnNum()) + mystr(" has an invalid set number: ") + mystr(set_number) + "\n";
				return false;
			}
		}

		// check if field variables are available

		int posAt = object_str.Find(0, '@');
		if (posAt != -1)
		{
			object_str = object_str.SubString(posAt + 1, object_str.Length() - 1);
		}

		// find last dot
		int p1 = 0;
		int p2 = 0;
		while ((p1 = object_str.Find(p1, '.')) != -1)
		{
			p2 = p1;
			p1++;
		}
		if (p2 > 0)
		{
			mystr typestr = object_str.SubString(p2 + 1, object_str.Length() - 1);
			if (!typestr.Compare("min") && !typestr.Compare("max") && !typestr.Compare("avg") && !typestr.Compare("sum"))
			{
				errorStr = mystr("Sensor ") + mystr(GetOwnNum()) + mystr(" has an unknown operation '") + this->object_str + "'\n";
				return false;
			}
		}
		else
		{
			errorStr = mystr("Sensor ") + mystr(GetOwnNum()) + mystr(" has an unknown operation '") + this->object_str + "'\n";
			return false;
		}

		if (active_elements.Length()==0)
		{
			errorStr = mystr("Sensor ") + mystr(GetOwnNum()) + mystr(": No element found which has the variable '") + object_str.Left(p2);
			if (posAt != -1)
			{
				errorStr += mystr("and the variable ") + this->object_str.Left(posAt) + "'\n";
			}
			return false;
		}
	}
	return true;
}

mystr SystemSensor::GetObjectString(Object object)
{
	switch(object)
	{
	case None:
		return mystr("none");
	case NewtonIterations:
		return mystr("newton_iterations");
	case NewtonIterationsTotal:
		return mystr("newton_iterations_total");
	case DiscontinuousIterations:
		return mystr("discontinuous_iterations");
	case Jacobians:
		return mystr("jacobians");
	case RHSEvaluationsJacobian:
		return mystr("rhs_evaluations_jacobian");
	case RHSEvaluations:
		return mystr("rhs_evaluations");
	case Volume:
		return mystr("volume");
	case PotentialEnergy:
		return mystr("potential_energy");
	case KineticEnergy:
		return mystr("kinetic_energy");
	case ExternalWork:
		return mystr("external_work");
	case TimeStep:
		return mystr("time_steps");
	case TimeStepSize:
		return mystr("time_step_size");
	case DOF:
		return mystr("DOF");
	case EV:
		return mystr("EV");		
	case FEmincol:
		return mystr("FE_color_minimum");	
	case FEmaxcol:
		return mystr("FE_color_maximum");	
	case NNodes:
		return mystr("NNodes");
	case NElements:
		return mystr("NElements");
	case ElapsedTime:
		return mystr("ElapsedTime");
	case MinMaxFieldVariableValue:
		return object_str;
	}

	return mystr("unknown");
}
SystemSensor::Object SystemSensor::GetObject(const mystr& object_str)
{
	if (object_str.Compare("none", 1))
	{
		return None;
	}
	else if (object_str.Compare("newton_iterations", 1))
	{
		return NewtonIterations;
	}
	else if (object_str.Compare("newton_iterations_total", 1))
	{
		return NewtonIterationsTotal;
	}
	else if (object_str.Compare("discontinuous_iterations", 1))
	{
		return DiscontinuousIterations;
	}
	else if (object_str.Compare("jacobians", 1))
	{
		return Jacobians;
	}
	else if (object_str.Compare("rhs_evaluations_jacobian", 1))
	{
		return RHSEvaluationsJacobian;
	}
	else if (object_str.Compare("rhs_evaluations", 1))
	{
		return RHSEvaluations;
	}
	else if (object_str.Compare("volume", 1))
	{
		return Volume;
	}
	else if (object_str.Compare("potential_energy", 1))
	{
		return PotentialEnergy;
	}
	else if (object_str.Compare("kinetic_energy", 1))
	{
		return KineticEnergy;
	}
	else if (object_str.Compare("external_work", 1))
	{
		return ExternalWork;
	}
	else if (object_str.Compare("time_steps", 1))
	{
		return TimeStep;
	}
	else if (object_str.Compare("time_step_size", 1))
	{
		return TimeStepSize;
	}
	else if (object_str.Compare("DOF", 1))
	{
		return DOF;
	}
	else if (object_str.Compare("EV", 1))
	{
		return EV;
	}
	else if (object_str.Compare("FE_color_minimum", 1))
	{
		return FEmincol;
	}
	else if (object_str.Compare("FE_color_maximum", 1))
	{
		return FEmaxcol;
	}
	else if (object_str.Compare("NNodes", 1))
	{
		return NNodes;
	}
	else if (object_str.Compare("NElements", 1))
	{
		return NElements;
	}
	else if (object_str.Compare("ElapsedTime",1))
	{
		return ElapsedTime;
	}
	else
	{
		return MinMaxFieldVariableValue;
	}
	GetMBS()->UO(UO_LVL_warn) << "WARNING: SystemSensor: object '" << object_str << "' unknown!\n";
	return Unknown;
}

double SystemSensor::GetCurrentValue(double time)   
{
	switch(object)
	{
	case NewtonIterations:
		return GetMBS()->GetTimeIntLog().TInewtonit;
	case NewtonIterationsTotal:
		return GetMBS()->GetTimeIntLog().TInewtonitsum;
	case DiscontinuousIterations:
		return GetMBS()->GetTimeIntLog().TInonlinit;
	case RHSEvaluations:
		return GetMBS()->GetTimeIntLog().evalfcnt;
	case RHSEvaluationsJacobian:
		return GetMBS()->GetTimeIntLog().evalf_jaccnt;
	case Jacobians:
		return GetMBS()->GetTimeIntLog().jaccount;
	case Volume:
		return GetMBS()->GetVolume();
	case PotentialEnergy:
		return GetMBS()->GetPotentialEnergy();
	case KineticEnergy:
		return GetMBS()->GetKineticEnergy();
	case ExternalWork:
		return GetMBS()->GetExternalWork();
	case TimeStepSize:
		return GetMBS()->GetStepSize();
	case TimeStep:
		return GetMBS()->GetTimeStep();
	case DOF:
		return GetMBS()->GetXact(global_index);
	case EV:
		return GetMBS()->GetEigenValue(global_index);
	case FEmincol:
		return GetMBS()->Graphics3D()->GetFEmincol();
	case FEmaxcol:
		return GetMBS()->Graphics3D()->GetFEmaxcol();
	case NNodes:
		return GetMBS()->NNodes();
	case NElements:
		return GetMBS()->NE();
	case ElapsedTime:
		return GetMBS()->GetElapsedTime();
	case MinMaxFieldVariableValue:
		return getMinMaxFieldVariableValue(time);
	}

	return 0.;
}

double SystemSensor::getMinMaxFieldVariableValue(double time)
{
	MBS* mbs = GetMBS();

	double ret;
	if (operation == Min)
	{
		ret = std::numeric_limits<double>::max();
	}
	else if (operation == Max)
	{
		ret = -std::numeric_limits<double>::max();
	}
	else // avg or sum
	{
		ret = 0.;
	}
	
	double val = 0;

	double vol = 0.; //2D: area, 3D: volume
	bool oneElementHasFieldVariable = false;

	int element_number=-1;
	Vector3D local_position;
	
	for (int i = 1, ne = active_elements.Length(); i<=ne; i++)
	{
		Element* el = mbs->GetElementPtr(active_elements(i));
			
		if (isFieldVariableValue)
		{
			TArray<Vector3D> pos;
			TArray<double> weight;
			const IntegrationRule* ir = nullptr;
			const FiniteElement3D* fel3D = dynamic_cast<const FiniteElement3D*> (el);
			if (fel3D)
			{
				ir = fel3D->GetIntegrationRule(IntegrationRule::IVT_Stiffness);
			}
			else
			{
				const FiniteElement2D* fel2D = dynamic_cast<const FiniteElement2D*> (el);
				if (fel2D)
				{
					ir = fel2D->GetIntegrationRule(IntegrationRule::IVT_Stiffness);
				}
			}
			if (ir)
			{
				for (int j = 1; j<=ir->GetNumberOfIntegrationPoints(); j++)
				{
					IntegrationRule::IntegrationPoint ip = ir->integrationPoints(j);
					pos.Add(Vector3D(ip.x, ip.y, ip.z));
					weight.Add(ip.weight);
				}
			}
			else
			{
				pos.Add(el->GetRefPos());
				weight.Add(1.);
			}

			int n_pos = pos.Length();
			for (int j = 1; j<=n_pos; j++)
			{
				double val;
				if (el->Dim()==2)
				{
					val = el->GetFieldVariableValue(fvDescriptor,pos(j).MakeV2D(), false);
				}
				else
				{
					val = el->GetFieldVariableValue(fvDescriptor,pos(j), false);
				}
				switch (operation)
				{
				case Min:
					if (val < ret || std::isnan(val))
					{
						element_number = active_elements(i);
						local_position = pos(j);
						ret = val;
					}
					break;
				case Max:
					if (val > ret || std::isnan(val))
					{
						element_number = active_elements(i);
						local_position = pos(j);
						ret = val;
					}
					break;
				case Avg:
					vol += weight(j);
				case Sum:
					ret += val * weight(j);
					break;
				}
			}
		}
		else //isFieldVariableValue==false:
		{
			TArrayDynamic<ReadWriteElementDataVariableType> available_variables;
			el->GetAvailableSpecialValues(available_variables);
			for (int j = 1; j <= available_variables.Length(); j++)
			{
				const ReadWriteElementDataVariableType& available_variable = available_variables(j);
				if (available_variable == svDescriptor && available_variable.RWaccess & TRWElementDataRead)
				{
					svDescriptor.value = std::numeric_limits<double>::quiet_NaN();
					svDescriptor.time = time;
					el->ReadSingleElementDataAuto(svDescriptor);
					double val = svDescriptor.value;
					switch (operation)
					{
					case Min:
						if (val < ret || std::isnan(val))
						{
							element_number = active_elements(i);
							local_position = el->GetRefPos();
							ret = val;
						}
						break;
					case Max:
						if (val > ret || std::isnan(val))
						{
							element_number = active_elements(i);
							local_position = el->GetRefPos();
							ret = val;
						}
						break;
					case Avg:
						vol += 1;
					case Sum:
						ret += val;
						break;
					}
				}
			}
		}
	}

	if (operation == Avg)
	{
		if (vol == 0.)
		{
			ret = std::numeric_limits<double>::quiet_NaN();
		}
		else
		{
			ret /= vol;
		}
	}

	// sensortype 2: we do not want the maximum value but another field variable at this position
	if (evaluateVariable_active)
	{
		Element* el = mbs->GetElementPtr(element_number);
		if (evaluateVariable_isFieldVariableValue)
		{
			return el->GetFieldVariableValue(evaluateVariable_fvDescriptor, local_position, false);
		}
		// not a field variable
		TArrayDynamic<ReadWriteElementDataVariableType> available_variables;
		el->GetAvailableSpecialValues(available_variables);
		for (int i = 1; i <= available_variables.Length(); i++)
		{
			const ReadWriteElementDataVariableType& available_variable = available_variables(i);
			if (available_variable == evaluateVariable_svDescriptor && available_variable.RWaccess & TRWElementDataRead)
			{
				evaluateVariable_svDescriptor.time = time;
				evaluateVariable_svDescriptor.value = std::numeric_limits<double>::quiet_NaN();
				el->ReadSingleElementDataAuto(evaluateVariable_svDescriptor);
				return evaluateVariable_svDescriptor.value;
			}
		}
		return std::numeric_limits<double>::quiet_NaN();
	}
	return ret;
	
}