//#**************************************************************
//#
//# filename:             sensorsSpecific.h
//#
//# author:               Gerstmayr Johannes
//#												Vetyukov Yury
//#
//# generated:						February-June 2012
//# description:          
//#                       
//# remarks:						  HotInt sensors - particular sensor types
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#pragma once

#include "sensors.h"

// added by YV
// this is an abstract base class for sensors, which are attached to a particular element in MBS
class ElementSensor : public Sensor //$EDC$[beginclass,classname=ElementSensor,parentclassname=Sensor,pythonclassdescription="Sensor on an element",pythonAddAs=Sensor]
{
	int elementNumber; //$EDC$[varaccess,EDCvarname="element_number",EDCfolder="",tooltiptext="number of the element, to which the sensor is applied",python]

protected:
	int& GetElementNumber() { return elementNumber; }
	Element & GetElement() { return GetMBS()->GetElement(GetElementNumber()); }		//$ DR 2012-09-12 changed GetElement(elementNumber) to GetElement(GetElementNumber())

	// default constructor is just for copy making
	ElementSensor() {}
	
public:
	ElementSensor(MBS * mbs) : Sensor(mbs)
	{
	}
	void SetElementNumber(int elementNumber)
	{
		this->elementNumber = elementNumber;
	}
	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		ElementSensor & S = (ElementSensor&)s;
		elementNumber = S.elementNumber;
	}
	virtual bool IsConsistent(mystr & errorStr)
	{
		if(elementNumber <= 0 || elementNumber > GetMBS()->NE())
		{
			errorStr = mystr("Sensor ") +mystr(GetOwnNum()) + mystr(" has an invalid element number: ") + mystr(elementNumber) + "\n";
			return false;
		}
		return true;
	}
	virtual int GetNumberOfRelatedElements() { return 1; }
	virtual int& GetRelatedElementNumber(int nElement) { return elementNumber; }
	virtual void GetElementData(ElementDataContainer & edc)
	{
		Sensor::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc) //$ DR 2012-10 return value changed from bool to int
	{
		int rv = Sensor::SetElementData(edc);
		AfterSetFunction();	// DR 2012-06-25 added
		return rv;
	}
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
  virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

};//$EDC$[endclass,ElementSensor]

// added by PG
// this is an abstract base class for sensors, which are attached to a set of element in MBS
class MultipleElementsSensor : public Sensor //$EDC$[beginclass,classname=MultipleElementsSensor,parentclassname=Sensor]
{
	IVector relatedElements; //$EDC$[varaccess,EDCvarname="related_elements",EDCfolder="",variable_length_vector,tooltiptext="list of elements, on which the sensor acts"]

protected:
	Element& GetRelatedElement(int index) { return GetMBS()->GetElement(GetRelatedElementNumber(index)); }
	MultipleElementsSensor() {}
	
public:
	MultipleElementsSensor(MBS * mbs) : Sensor(mbs)
	{
	}
	void SetRelatedElements(const IVector& relatedElements)
	{
		// we store element numbers and avoid redundant entries
		for(int i=1; i<=relatedElements.Length(); i++)
			this->relatedElements.AddIfNotExists(relatedElements(i));
	}
	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		MultipleElementsSensor & S = (MultipleElementsSensor&)s;
		relatedElements = S.relatedElements;
	}
	virtual bool IsConsistent(mystr & errorStr)
	{
		if (GetNumberOfRelatedElements() == 0)
		{
			errorStr = mystr("Sensor ") +mystr(GetOwnNum()) + mystr(" has no element numbers specified\n");
			return false;
		}

		int nElementsInSystem = GetMBS()->NE();
		for (int index=1; index<=GetNumberOfRelatedElements(); index++)
		{			
			if(relatedElements(index) <= 0 || relatedElements(index) > nElementsInSystem)
			{
				errorStr = mystr("Sensor ") +mystr(GetOwnNum()) + mystr(" has an invalid element number: ") + mystr(relatedElements(index)) + "\n";
				return false;
			}
		}
		return true;
	}
	virtual int GetNumberOfRelatedElements() { return relatedElements.Length(); }
	virtual int& GetRelatedElementNumber(int index) { return relatedElements(index); }
	virtual void GetElementData(ElementDataContainer & edc)
	{
		Sensor::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc)
	{
		int rv = Sensor::SetElementData(edc);
		AfterSetFunction();
		return rv;
	}
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
  virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

};//$EDC$[endclass,MultipleElementsSensor]

// added by PG
// these sensors measure a particular field variable of a set of elements at the first intersection found
//   a) with a given plane (in case of beam elements), or 
//   b) with a given straight line (in case of plate/shell elements), or
//   c) with a given global position (in case of solid elements).
// the user specifies a number of elements, whith which this sensor may operate
// in sequential order we look if an element intersects with the given global hyperplane, global straight line, or global position.
// as soon as an intersection is found, this yields an element number and a local position, at which the fieldvariable value is returned.
// if no intersection is found, then the sensor returns 0
// two member variables are required to cover all of the three modes (a)-(c): one position vector (global_position), and one direction vector (global_direction)
//$ PG 2017-5-23: meanwhile the implementation is restricted to the three dimensional case, and mode (a).
class FieldVariableGlobalPositionSensor : public MultipleElementsSensor //$EDC$[beginclass,classname=FieldVariableGlobalPositionSensor,parentclassname=MultipleElementsSensor]
{
protected:
	FieldVariableDescriptor fieldVariableDescriptor;
	Vector3D globalPosition;		//$EDC$[varaccess,EDCvarname="global_position",EDCfolder="",tooltiptext="global position."]
	Vector3D globalDirection;		//$EDC$[varaccess,EDCvarname="global_direction",EDCfolder="",tooltiptext="either pane normal (case a) or direction of a straight line (case b) or meaningless (case c)."]
	int dimension;	//$ PG 2017-5-23: for the moment: required to equal 3
	
	// members storing the solution of the inverse search
	int elementNumber;          
	Vector3D localPosition;

	virtual double GetCurrentValue(double time);

	virtual int GetNumberOfDrawingPositions() { return 1; }
	virtual Vector3D GetDrawPosition(int i)
	{
		if (!elementNumber)
			return Vector3D(0.);

		return GetRelatedElement(elementNumber).GetPosD(localPosition);
	}

	virtual void AfterSetFunction()
	{
		dimension = 3;
		assert(dimension == GetRelatedElement(1).Dim());
	}

	// default constructor is just for copy making
	FieldVariableGlobalPositionSensor() {}

public:
	FieldVariableGlobalPositionSensor(MBS * mbs) : MultipleElementsSensor(mbs)
	{
		elementNumber = 0;
		localPosition.Set(0,0,0);
	}
	// three set-functions for three variants of specification of the local position
	void SetFVGPSPos3D(const IVector& relatedElements, FieldVariableDescriptor fieldVariableDescriptor_, const Vector3D& globalPosition, const Vector3D& globalDirection = Vector3D(0.));
	//void SetFVGPSPos2D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector2D localPosition);
	virtual void CopyFrom(const Sensor& s);
	virtual Sensor * GetCopy()
	{
		FieldVariableGlobalPositionSensor * S = new FieldVariableGlobalPositionSensor();
		S->CopyFrom(*this);
		return S;
	}
	virtual mystr GetTypeName();
	virtual mystr GetSensorTypeTexDescription(); 	// for the auto generated documentation //$ DR 2012-10 added
	virtual bool IsConsistent(mystr & errorStr);
	virtual void GetElementData(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc); //$ DR 2012-10 return value changed from bool to int

	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
  virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter
};//$EDC$[endclass,FieldVariableGlobalPositionSensor]

// added by YV
// these sensors measure a particular field variable for a given element
// either a local node number or a local position may be specified
// the local position may be either 2D or 3D
class FieldVariableElementSensor : public ElementSensor //$EDC$[beginclass,classname=FieldVariableElementSensor,parentclassname=ElementSensor,pythonclassdescription="Sensor to a field variable on an element at a local positoin",pythonAddAs=Sensor]
{
protected:
	FieldVariableDescriptor fieldVariableDescriptor;
	int localNodeNumber;     //$EDC$[varaccess,EDCvarname="node_number",EDCfolder="",tooltiptext="local node number. If > 0, then the position of this node is used."]
	Vector3D localPosition;  //$EDC$[varaccess,EDCvarname="local_position",EDCfolder="",tooltiptext="local position at which the field variable is evaluated.",python]
	//bool localPosition2D;	 //[varaccess,EDCvarname="local_position2D",EDCfolder="",tooltiptext="Only the two first components in local_position are used!"]
	int average_mode;        //$EDC$[varaccess,EDCvarname="average",EDCfolder="",int_bool,tooltiptext="return the average value on the element (only for finite elements) is returned, if activated node_number and local_position are ignored"]
	int dimension;	//$ DR 2013-06-19 replaces bool localPosition2D

	virtual double GetCurrentValue(double time);

	virtual int GetNumberOfDrawingPositions() { return 1; }
	virtual Vector3D GetDrawPosition(int i)
	{
		return GetElement().GetPosD(localPosition);
	}

	virtual void AfterSetFunction()
	{
		//SetSensorName(GetTypeName());	//$ DR 2012-06-25 commented out
		if (localNodeNumber == 0 && (GetElementNumber() <= GetMBS()->NE()))
		{ 
			dimension = GetElement().Dim();
		}
	}

	// default constructor is just for copy making
	FieldVariableElementSensor() {}

public:
	FieldVariableElementSensor(MBS * mbs) : ElementSensor(mbs)
	{
		average_mode = 0;
	}
	// three set-functions for three variants of specification of the local position
	void SetFVESPos3D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector3D localPosition);
	void SetFVESPos2D(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, Vector2D localPosition);
	void SetFVESNode(int elementNumber, FieldVariableDescriptor fieldVariableDescriptor_, int localNodeNumber);
	virtual void CopyFrom(const Sensor& s);
	virtual Sensor * GetCopy()
	{
		FieldVariableElementSensor * S = new FieldVariableElementSensor();
		S->CopyFrom(*this);
		return S;
	}
	virtual mystr GetTypeName();
	virtual mystr GetSensorTypeTexDescription(); 	// for the auto generated documentation //$ DR 2012-10 added
	virtual bool IsConsistent(mystr & errorStr);
	virtual void GetElementData(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc); //$ DR 2012-10 return value changed from bool to int
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	Vector3D GetLocalPosition()	{ return localPosition; }
};//$EDC$[endclass,FieldVariableElementSensor]

// added by YV
// this sensor reads values from a given element using Element::ReadSingleElementData access function.
class SingleElementDataSensor : public ElementSensor 
{
protected:
	ReadWriteElementDataVariableType RWdata;		// corresponding identifier in the internal presentation
	int successfully_converted;
	//virtual void AfterSetFunction();
	virtual double GetCurrentValue(double time);

	// default constructor is just for copy making
	SingleElementDataSensor() {}

public:
	SingleElementDataSensor(MBS * mbs) : ElementSensor(mbs)
	{
	}
	void SetSingleElementDataSensor(int elementNumber, const char * variable_name);
	virtual bool IsConsistent(mystr & errorStr);
	virtual void GetElementData(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc);
	virtual mystr GetTypeName(){return "ElementSensor"; }
	virtual mystr GetSensorTypeTexDescription(); 	// for the auto generated documentation //$ DR 2012-11 added

	virtual void CopyFrom(const Sensor& s);
	virtual Sensor * GetCopy()
	{
		SingleElementDataSensor * S = new SingleElementDataSensor();
		S->CopyFrom(*this);
		return S;
	}
};

struct SensorFixedValue : public Sensor
{
	double actvalue;
	SensorFixedValue(MBS * mbs) : Sensor(mbs)
	{
		actvalue = 0;
		ModifySignalStorageMode(SSM_OwnFile);
	}
	virtual Sensor* GetCopy()
	{
		Sensor * s = new SensorFixedValue(GetMBS());
		s->CopyFrom(*this);
		return s;
	}
	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		actvalue = ((SensorFixedValue&)s).actvalue;
	}
	virtual double GetCurrentValue(double time)
	{
		return actvalue;
	}
	void SetCurrentValue(double value)
	{
		actvalue = value;
	}
	void SetCurrentValueIfLarger(double value)
	{
		if (value > actvalue)
			actvalue = value;
	}
	void SetCurrentValueIfSmaller(double value)
	{
		if (value < actvalue)
			actvalue = value;
	}
	virtual mystr GetTypeName() { return ""; }
};

// added by YV; to be implemented
// prospective description:
// abstract base class for sensors, which are attached to a particular node in MBS
class NodeSensor// : public Sensor
{
};

// added by YV; to be implemented
// prospective description:
// computes an average for all elements, which meet at this node;
// the field variable can be computed either directly in the node or with some offset into the elements
class FieldVariableNodeSensor : public NodeSensor
{
};

// added by YV; to be implemented
// prospective description:
// nodal DOF, etc.
class SpecialValueNodeSensor : public NodeSensor
{
};




// added by DR
// sensor that consists of a list of sensors
// this replaces the old MBSMultipleSensor
class MultipleSensor : public Sensor //$EDC$[beginclass,classname=MultipleSensor,parentclassname=Sensor]
{
	TArray<int> sensorNrs; //$EDC$[varaccess,variable_length_vector,EDCvarname="sensor_numbers",EDCfolder="",tooltiptext="number of the sensors, that are used for computation"]
	int operation;	
	Vector weights; //$EDC$[varaccess,variable_length_vector,EDCvarname="weights",EDCfolder="",tooltiptext="weights for e.g. a weighted sum. This vector must have the same length as sensor_numbers or must be empty!"]

protected:
	int& GetSensorNumber(int i) { return sensorNrs(i); }
	Sensor& GetSensor(int i) { return GetMBS()->GetSensor(GetSensorNumber(i)); }		

	// default constructor is just for copy making
	MultipleSensor() {}
	
public:
	MultipleSensor(MBS * mbs) : Sensor(mbs)
	{
	}

	void SetMultipleSensor(TArray<int> sensorNumbers, mystr operationMinMaxAverage, Vector weights = Vector(0,1,0.));
	virtual mystr GetTypeName() {return "MultipleSensor";};	
	virtual mystr GetSensorTypeTexDescription(); 	// for the auto generated documentation 

	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		MultipleSensor & S = (MultipleSensor&)s;
		sensorNrs = S.sensorNrs;
		operation = S.operation;
		weights = S.weights;		//$ YV 2014-05-13
	}
	
	virtual Sensor * GetCopy()
	{
		MultipleSensor * S = new MultipleSensor();
		S->CopyFrom(*this);
		return S;
	}

	virtual bool IsConsistent(mystr & errorStr);
	
	virtual void GetElementData(ElementDataContainer& edc);
	virtual int SetElementData(ElementDataContainer& edc); 
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
  virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter


	virtual int GetOperationFlagFromTextualIdentifier(mystr operationMinMaxAverage);
	virtual mystr TextualIdentifierOfOperation();

	virtual double GetCurrentValue(double time);

	virtual int GetNumberOfDrawingPositions() { return 1; }
	virtual Vector3D GetDrawPosition(int i)
	{
		return GetSensor(1).GetDrawPosition(i);
	}
	virtual int GetNumberOfRelatedSensors() { return sensorNrs.Length(); }			//$ DR 2013-05-21
	virtual int& GetRelatedSensorNumber(int nSensor) { return sensorNrs(nSensor); } //$ DR 2013-05-21


};//$EDC$[endclass,MultipleSensor]


// added by DR
class LoadSensor : public Sensor //$EDC$[beginclass,classname=LoadSensor,parentclassname=Sensor]
{
	int loadNumber; //$EDC$[varaccess,EDCvarname="load_number",EDCfolder="",tooltiptext="number of the load, to which the sensor is applied"]

protected:
	int& GetLoadNumber() { return loadNumber; }
	MBSLoad& GetLoad() { return GetMBS()->GetLoad(GetLoadNumber()); }		//$ DR 2012-09-12 changed GetElement(elementNumber) to GetElement(GetElementNumber())

	// default constructor is just for copy making
	LoadSensor() {}
	
public:
	LoadSensor(MBS * mbs) : Sensor(mbs)
	{
	}

	virtual void SetLoadSensor(int load_nr)
	{
		loadNumber = load_nr;
		AfterSetFunction(); //$ PG 2013-1-16: call added, in order to be consistent with FieldVariableElementSensor (see , e.g., FieldVariableElementSensor::SetFVESPos3D)
	}

	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		LoadSensor & S = (LoadSensor&)s;
		loadNumber = S.loadNumber;
	}

	virtual Sensor * GetCopy()
	{
		LoadSensor * S = new LoadSensor();
		S->CopyFrom(*this);
		return S;
	}

	virtual bool IsConsistent(mystr & errorStr)
	{
		if(loadNumber <= 0 || loadNumber > GetMBS()->NLoads())
		{
			errorStr = mystr("Sensor ") +mystr(GetOwnNum()) +mystr(" has an invalid load number: ") + mystr(loadNumber) + "\n";
			return false;
		}
		return true;
	}

	virtual double GetCurrentValue(double time);

	//virtual int GetNumberOfRelatedElements() { return 1; }
	//virtual int& GetRelatedElementNumber(int nElement) { return elementNumber; }
	virtual void GetElementData(ElementDataContainer & edc)
	{
		Sensor::GetElementData(edc);
	}
	virtual int SetElementData(ElementDataContainer& edc) 
	{
		int rv = Sensor::SetElementData(edc);
		AfterSetFunction();	// DR 2012-06-25 added
		return rv;
	}
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
  virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	virtual int GetNumberOfDrawingPositions() { return 0; } // if just the load is known, but no element, it is not possible to draw the load
	virtual Vector3D GetDrawPosition(int i)		{	return Vector3D(0);	}

	virtual mystr GetTypeName(){return "LoadSensor"; }
	virtual mystr GetSensorTypeTexDescription(); 	// for the auto generated documentation 

	virtual int GetNumberOfRelatedLoads() { return 1; }			//$ DR 2013-05-21
	virtual int& GetRelatedLoadNumber(int nLoad) {return loadNumber;} //$ DR 2013-05-21

}; //$EDC$[endclass,LoadSensor]


// added by PG
// this class provides SYSTEM sensors, which are attached to a particular multibody SYSTEM quantity such as a global DOF, global Eigenvalue, iteration numbers, or scalar performance values.
// it should be constantly extended by all hotint developers by modifying
// 1) typedef enum {...} Object,
// 2) mystr SystemSensor::GetObjectString(Object object), and
// 3) SystemSensor::Object SystemSensor::GetObject(const mystr& object_str)
// 4) double SystemSensor::GetCurrentValue(double time)
// 5) EDCAuto-comment regarding the class member 'mystr object_str'
class SystemSensor : public Sensor //$EDC$[beginclass,classname=SystemSensor,parentclassname=Sensor]
{
public:

// the following typedef defines all possible objects being tracked by the system sensor
// it should be constantly extended by all hotint users in particular accordance with
// 1) mystr SystemSensor::GetObjectString(Object object), and
// 2) SystemSensor::Object SystemSensor::GetObject(const mystr& object_str)
// 3) double SystemSensor::GetCurrentValue(double time)
// 4) EDCAuto-comment regarding the class member 'mystr object_str'
	typedef enum
	{
		None,
		RHSEvaluations,
		RHSEvaluationsJacobian,
		Jacobians,
		NewtonIterations,
		NewtonIterationsTotal,
		DiscontinuousIterations,
		Volume,
		PotentialEnergy,
		KineticEnergy,
		ExternalWork,
		DOF,
		EV,
		FEmincol,
		FEmaxcol,
		NNodes,
		NElements,
		ElapsedTime,
		TimeStep,
		TimeStepSize,
		MinMaxFieldVariableValue,
		Unknown
	} Object;

	SystemSensor(MBS * mbs) : Sensor(mbs)
	{
		SetObject(None);
		global_index = 0;
		set_number = 0;
		SetSensorName(mystr("Systemsensor"));
	}
	virtual void CopyFrom(const Sensor& s)
	{
		Sensor::CopyFrom(s);
		SystemSensor & S = (SystemSensor&)s;
		global_index = S.global_index;
		set_number = S.set_number;
		object = S.object;
		object_str = S.object_str;
	}
	virtual Sensor * GetCopy()
	{
		SystemSensor * S = new SystemSensor();
		S->CopyFrom(*this);
		return S;
	}
	virtual bool IsConsistent(mystr & errorStr);
	virtual mystr GetTypeName(){return "SystemSensor"; }
	virtual mystr GetSensorTypeTexDescription();
	virtual void GetElementData(ElementDataContainer & edc)
	{
		Sensor::GetElementData(edc);
	}
	virtual void PostAssemble();
	virtual int SetElementData(ElementDataContainer& edc) //$ DR 2012-10 return value changed from bool to int
	{
		int rv = Sensor::SetElementData(edc);
		SetObject(GetObject(object_str));
		AfterSetFunction();
		return rv;
	}
	virtual void GetElementDataAuto(ElementDataContainer& edc); 		//fill in all element data
	virtual int SetElementDataAuto(ElementDataContainer& edc); //set element data acc. to ElementDataContainer, return 0 if failed or values invalid!
	virtual int ReadSingleElementDataAuto(ReadWriteElementDataVariableType& RWdata); 		//automatically generated function from EDC_converter
	virtual int WriteSingleElementDataAuto(const ReadWriteElementDataVariableType& RWdata); //automatically generated function from EDC_converter
	virtual int GetAvailableSpecialValuesAuto(TArrayDynamic<ReadWriteElementDataVariableType>& available_variables); //automatically generated function from EDC_converter

	virtual bool IsSensorOnEigensystem() { return (object == EV); }

	// main operation functions: setting and evaluating
	virtual void SetSystemSensor(Object object, int global_index=0);
	virtual double GetCurrentValue(double time);

protected:
	// internal access functions
	int GetGlobalIndex() { return global_index; }
	void SetGlobalIndex(int global_index);
	Object GetObject() { return object; }
	void SetObject(Object object); // sets object and object_str at once
	bool IsGlobalIndexObject() { return (object==EV || object==DOF); }

	// conversion between object and object_str
	mystr GetObjectString(Object object);
	Object GetObject(const mystr& object_str);
	
	SystemSensor() {}  // just for copy making
	
private:
	enum MinMaxOperation {
		Min,
		Max,
		Avg,
		Sum
	} operation;
	double getMinMaxFieldVariableValue(double time);
	mystr object_str; //$EDC$[varaccess,EDCvarname="object",EDCfolder="",tooltiptext="Object tracked by systemsensor. Is either 'DOF' (global degree of freedom), 'EV' (global eigenvalue), 'time_step_size', 'jacobians', 'newton_iterations', 'newton_iterations_total', 'discontinuous_iterations', 'rhs_evaluations', 'rhs_evaluations_jacobian', 'volume', 'potential_energy', 'kinetic_energy', 'FE_color_minimum', 'FE_color_maximum', 'NNodes', 'NElements', or 'ElapsedTime'"]
	Object object;
	int global_index; //$EDC$[varaccess,EDCvarname="global_index",EDCfolder="",tooltiptext="Number of the global index. Has to be set if (and only if) object is 'DOF' or 'EV'."]
	int set_number;   //$EDC$[varaccess,EDCvarname="set_number",EDCfolder="",tootiptext="Set number of elements that should be considered (if 0, all elements are considered). This is only used if a maximum/minimum of a field/special value is requested."]

	//used in case of minmaxsensor
	//these values will be set in PostAssemble

	bool isFieldVariableValue; // true ... fieldVariableValue, false ... SpecialValue
	mystr textualIdentifier;
	FieldVariableDescriptor fvDescriptor;
	ReadWriteElementDataVariableType svDescriptor;

	//used in case of a second field variable.
	// position.x@displacement.x.max
	// position.x would be the 
	bool evaluateVariable_active;
	bool evaluateVariable_isFieldVariableValue;
	FieldVariableDescriptor evaluateVariable_fvDescriptor;
	ReadWriteElementDataVariableType evaluateVariable_svDescriptor;
	
	TArray<int> active_elements;
	// end minmaxsensorvalues

};//$EDC$[endclass,SystemSensor]