//#**************************************************************
//#
//# filename:             interpolators1D.h           
//#
//# author:               Yury Vetyukov, Peter Gruber
//#
//# generated:						March 2018
//# updated:							March 2018
//# description:          filie contains interpolator classes using 1 dimensional input
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2016 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

// this is the base class, which performs linear interpolation of values
class Interpolator
{
protected:
	virtual int NumberOfNodes() = 0;
	virtual double NodePosition(int i) = 0;		// 1-based
	virtual double NodeValue(int i) = 0;			// 1-based

public:
	// interpolators will be stored in the input provider without explicit typing
	virtual ~Interpolator() {}

	// interpolates values to a point with a given global vertical position
	double GetValueAtPoint(double position)
	{
		// first we search for the segment between two nodes, to which the queried point belongs
		double previous_position;
		double previous_value;
		double subsequent_position;
		double subsequent_value;
		int i;
		for(i = 1; i <= NumberOfNodes(); i++)
		{
			double y = NodePosition(i);
			if(y < position)
			{
				previous_position = y;
				previous_value = NodeValue(i);
			}
			else
			{
				subsequent_position = y;
				subsequent_value = NodeValue(i);
				break;
			}
		}

		double result;
		if(i == 1)
			result = subsequent_value;
		else if(i > NumberOfNodes())
			result = previous_value;
		else
		{
			// we linearly interpolate the values
			double k = (position - previous_position) / (subsequent_position - previous_position);
			result = previous_value * (1 - k) + subsequent_value * k;
		}
		return result;
	}
};

// simplest case: just one value
class ConstantInterpolator : public Interpolator
{
	virtual int NumberOfNodes() { return 1; }
	virtual double NodePosition(int i) { return 0; }
	virtual double NodeValue(int i) { return value; }

public:
	ConstantInterpolator() { value = 0; }
	ConstantInterpolator(double value) { this->value = value; }
	double value;
};


// this interpolator works with a set of values per node,
class ListInterpolator : public Interpolator
{
	TArray<double> points;
	TArray<double> values;
protected:
	virtual int NumberOfNodes() { return points.Length(); }
	virtual double NodePosition(int i) { return points(i); }
	virtual double NodeValue(int i)
	{
		return values(i);
	}
public:
	/*
	ListInterpolator(TArray<double> & points_)
		: points(points_)
	{
	}
	*/
	ListInterpolator(TArray<double> & points_, TArray<double> & values_)
		: points(points_), values(values_)
	{
	}
	double & Value(int i) { return values(i); }
};

