//#**************************************************************
//#
//# filename:             spectra_sparse_ev_solver.h           
//#
//# author:               Peter Gruber, Simon Weitzhofer
//#
//# generated:            June 2016
//# description:          wrapper for the spectra ev solver
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************

#pragma once
#include <Eigen/Core>
#include <Spectra/GenEigsRealShiftSolver.h>
#include <Spectra/GenEigsComplexShiftSolver.h>
#include <string.h>
#include "mbs_interface.h"
//#define __TIMING_SPARSEMATRIXSHIFTINVERTER__

class SparseMatrixShiftInverter
{
public:
	SparseMatrixShiftInverter(const SparseMatrix* A_, const SparseMatrix* M_, MBS* mbsi):
		A(A_),
		M(M_),
		mbs(mbsi)
	{ }

	virtual int rows()
	{
		return A->Getrows();
	}

	virtual int cols()
	{
		return A->Getcols();
	}

	virtual void perform_op(const double* x_in, double* y_out) = 0;

protected:
	const SparseMatrix* A;
	const SparseMatrix* M;
	MBS* mbs;
};

class SparseMatrixShiftInverterReal : public SparseMatrixShiftInverter
{
public:
	SparseMatrixShiftInverterReal(const SparseMatrix* A_, const SparseMatrix* M_, MBS* mbsi) : SparseMatrixShiftInverter(A_, M_, mbsi),
		AMinusSigmaMInv(0)
	{ }

	~SparseMatrixShiftInverterReal()
	{
		delete AMinusSigmaMInv;
	}

	void set_shift(double sigma)
	{
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStartTimer(13);
#endif
		sigma = sigma;
		AMinusSigmaM = SparseMatrix(*A);
		SparseMatrix SigmaM = SparseMatrix(*M);
		SigmaM *= sigma;
		AMinusSigmaM -= SigmaM;
		AMinusSigmaMInv = new SparseInverse(AMinusSigmaM);
		AMinusSigmaMInv->Factorize();
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStopTimer(13);
#endif
	}

	void perform_op(const double* x_in, double* y_out)
	{
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStartTimer(14);
#endif
		// y_out := (A-sig*M)^-1*M*x_in
		Vector x((*A).Getrows(), x_in);
		Vector y((*A).Getrows());
		Mult(*M, x, y);
		AMinusSigmaMInv->Apply(y);
		memcpy(y_out, y.GetVecPtr(), (*A).Getcols() * sizeof(double));
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStopTimer(14);
#endif
	}

protected:
	SparseMatrix AMinusSigmaM;
	SparseInverse *AMinusSigmaMInv;
	double sigma;
};

class SparseMatrixShiftInverterComplex : public SparseMatrixShiftInverter
{
	/*
	Function \emph{perform_op} shall return $\Re{(A - \sigma M)^{-1} M v}$ with $\sigma \in \mathbb{C}$, $v \in \mathbb{C}^n$, and $A, M \in \mathbb{R}{n\times n}$.
	In other words, for any given v we aim to solve
	\begin{equation}
	(A - \sigma M) y = M v \,,
	\end{equation}
	and then return $\Re{y}$. Due to details of the solution procedure we may assume Im{v} = 0 (i.e. what we really get in perform_op is x_in = Re{v}. But lets turn to the general case...
	Writing it in block notation we have to solve
	\begin{equation}
	\begin{pmatrix}
	A-\Re{\sigma}M  &   \Im{\sigma}M  \\
	 -\Im{\sigma}M  & A-\Re{\sigma}M  \\
	\end{pmatrix}
	\begin{pmatrix}
	\Re{y} \\ \Im{y}
	\end{pmatrix}
	=
	\begin{pmatrix}
	M \Re{v} \\ M \Im{v}
	\end{pmatrix}	
	\end{equation}

	Solution:
	\begin{equation}
	\Re{y} = (a + b a^{-1} b)^{-1} M (\Re{v} - b a^{-1} \Im{v})
	\end{equation}
	with $a = A -\Re{\sigma}M$, $b=\Im(\sigma)M$, and $\Im{v}=0$	
	*/
public:
	SparseMatrixShiftInverterComplex(const SparseMatrix* A_, const SparseMatrix* M_, MBS* mbsi) : SparseMatrixShiftInverter(A_, M_, mbsi),
		shiftInv(0)
	{ }

	~SparseMatrixShiftInverterComplex()
	{
		delete shiftInv;
	}

	void set_shift(double sigmaReal, double sigmaImag)
	{
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStartTimer(13);
#endif
		this->sigmaReal = sigmaReal;
		this->sigmaImag = sigmaImag;
				
		// now factorize $(a + b a^{-1} b)^{-1}$ with $a = A -\Re{\sigma}M$, $b=\Im(\sigma)M$,
		shift = SparseMatrix(*A);
		SparseMatrix sigmaM = SparseMatrix(*M);
		if (sigmaReal != 0.)
		{
			sigmaM *= sigmaReal;
			shift -= sigmaM;
			sigmaM *= sigmaImag/sigmaReal;
		}
		else
		{
			sigmaM *= sigmaImag;
		}

		// update a += b a^-1 b, with b = sigmaImag * M, and a = A - sigmaReal M
		if (sigmaImag != 0.)
		{
			SparseMatrix a(shift);
			SparseInverse aInv(a);
			aInv.Factorize();

			Vector bj(rows(), 1), cj(rows(), 1); // Vector not initialized

			for (int j = 1; j<=cols(); j++)
			{
				// c_j = a^-1 b_{.,j}
				for (int k = 1; k<=rows(); k++) //$ PG 2020-04-30: naive approach, better would be to get j'th column of M via access function in SparseMatrix
					cj(k) = sigmaM.Get(k, j);
				
				aInv.Apply(cj);

				// a += b_{i,.}*c_j
				Mult(sigmaM, cj, bj);
#if 0
				SparseVector sbj(rows(), 0);
				for (int i = 1; i<=rows(); i++)
				{
					double val = bj(i);
					if (fabs(val)>sparsetolzero)
						sbj.AddEntry(i, val);
				}				
				shift.AddColVector(sbj, j);
#else
				shift.AddColVector(bj, j);
#endif
			}
		}

		shiftInv = new SparseInverse(shift);
		shiftInv->Factorize();

#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStopTimer(13);
#endif
	}

	void perform_op(const double* x_in, double* y_out)
	{
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStartTimer(14);
#endif
		// y_out := shitInv*M*x_in
		Vector x((*A).Getrows(), x_in);
		Vector y((*A).Getrows());
		Mult(*M, x, y);
		shiftInv->Apply(y);
		memcpy(y_out, y.GetVecPtr(), (*A).Getcols() * sizeof(double));
#ifdef __TIMING_SPARSEMATRIXSHIFTINVERTER__
		TMStopTimer(14);
#endif
	}

protected:
	SparseMatrix shift;
	SparseInverse *shiftInv;
	double sigmaReal, sigmaImag;
};

int spectra_solve_ev(const SparseMatrix* A, const SparseMatrix* M, int nev, int ncv, double sigma, Vector& eval_real, Vector& eval_imag, Matrix& evecs_real, Matrix& evecs_imag, MBS* mbs)
{
	mystr outtext("EV solver -> create shifted inverse operator (preconditioner) using sigma: ");
	outtext += mystr(sigma) + mystr("\n");
	mbs->UO(UO_LVL_sim) << outtext;

	SparseMatrixShiftInverterReal op(A, M, mbs);
	Spectra::GenEigsRealShiftSolver< double, Spectra::LARGEST_MAGN, SparseMatrixShiftInverterReal > eigs(&op, nev, ncv, sigma);
	//SparseMatrixShiftInverterComplex op(A, M, mbs);
	//Spectra::GenEigsComplexShiftSolver< double, Spectra::LARGEST_MAGN, SparseMatrixShiftInverterComplex > eigs(&op, nev, ncv, 0, sigma);
	
	// Initialize and compute
	eigs.init();
	int maxit = mbs->GetSolSet().eigsolv_maxiterations;
	outtext = mystr("EV solver -> read max_iterations: ") + mystr(maxit) + mystr("\n");

	double accuracy = mbs->GetSolSet().eigsolv_accuracy;
	outtext += mystr("EV solver -> read accuracy: ") + mystr(accuracy) + mystr("\n");
	outtext += mystr("EV solver -> computing... ");	
	mbs->UO(UO_LVL_sim) << outtext;

	int nconv = eigs.compute(maxit, accuracy, Spectra::LARGEST_MAGN);

	mbs->UO(UO_LVL_sim) << mystr("iterations: ") + mystr(eigs.num_iterations());
	if (eigs.num_iterations() >= maxit)
		mbs->UO(UO_LVL_sim) << mystr(" (maximum number of iterations reached)");
	mbs->UO(UO_LVL_sim) << mystr("\n");

	if (nconv < nev)
		return 1;

    // Retrieve results
	int return_value = 2;  //not successful: >0, successful: ==0
	if(eigs.info() == Spectra::SUCCESSFUL)
	{
        Eigen::VectorXcd evalues = eigs.eigenvalues();
		Eigen::MatrixXcd evectors = eigs.eigenvectors(nev);
		int n = (*A).Getcols(); // =2*sos
		for(int i=0; i<nev; i++)
		{
			std::complex<double> ev = evalues(i);
			eval_real[i] = ev.real();
			eval_imag[i] = ev.imag();

			for (int j=0; j< n; j++)
			{
				evecs_real(i+1, j+1) = evectors(j, i).real();
				evecs_imag(i+1, j+1) = evectors(j, i).imag();
			}
		}
		return_value = 0;
	}
	return return_value;
}