//#**************************************************************
//#
//# filename:             timer.h
//#
//# author:               Gruber Peter
//#
//# generated:            14.12.2019
//# description:          timer class
//# remarks:              
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************


#pragma once
//#pragma warning(disable: 4995 4996)

// This class allows for hierarchical timing reports at runtime.
// Create an instance of Timer, repeatedly Start() & Stop() increases member "double duration" of instance.
// Additionally a list of children may be added. Duration of children timers (and percentage w.r.t. total duration of parent) is printed when calling this->Report().
// Finally call Report() to print timer report. In addition to durations (in seconds [s]) and percentage of durations, the descriptions of children are printed on Report(), as well as the description of their parent (which is printed in parantheses):
//
// Example of a report:
//28.0344[s] top level:
//	0.905306[s](3.22927 %) 1st child (top level)
//	0.391393[s](1.39612 %) 2nd child (top level)
//	25.0293[s](89.2807 %) 3rd child (top level):
//		0.126877[s](0.506914 %) 1st grandchild (3rd child)
//		24.7324[s](98.8135 %) 2nd grandchild (3rd child):
//			15.2696[s](61.7394 %) 1st grandgrandchild (2nd grandchild)
//			8.92194[s](36.074 %) 2nd grandgrandchild (2nd grandchild)
//	1.14244[s](4.07513 %) 4th child (top level)

class Timer
{
	__int64 startCtr, freq;    // used to compute duration
	const Timer* parent;       // pointer to parent timer
	TArray<Timer*> children;   // array of children timers

protected:
	double duration;           // timer duration so far
	mystr description;         // textual description

public:

	// construct timer instance, with textual description, and link to parent timer (NULL if has no parent)
	Timer(const mystr& description, const Timer* parent = NULL);
	
	// destructor
	~Timer();

	// add another child timer to this timer, with textual description, this is parent of child
	Timer* Add(const mystr& description);
	
	// get i-th child timer
	Timer* Get(int i) const;
	
	// reset timer
	void Reset();

	// start timer
	void Start();

	// stop timer
	void Stop();

	// report to string
	mystr Report() const;

	// return hierarchy level (how many generations above, zero if has no parent)
	int GetLevel() const;
};

// dummy instance: use as optional parameter in function declarations
static Timer dummyTimer("DUMMYTIMER");
