//#**************************************************************
//# filename:             ANCFBeamShear3D.cpp
//#
//# author:               Karin & Astrid
//#
//# generated:						Nov. 2010
//# description:          3D ANCF beam element with shear deformation
//# comments:							formulation is based on the following paper: 
//#												"A 3D Shear Deformable Finite Element Based on the Absolute Nodal Coordinate Formulation"
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
 
#include "body3d.h"
#include "femathhelperfunctions.h"
#include "material.h"
#include "ANCFBeamShear3D.h"
#include "Node.h"
//#include "graphicsconstants.h"
#include "elementdataaccess.h"
#include "solversettings_auto.h"
#include "options_class_auto.h" //for HOTINTOptions




#pragma region ANCFBeamShear3DGeneric


double ANCFBeamShear3DGeneric::CalculateElementLength() const
{
	//to obtain stress free reference configuration: same numerical integration as in axial deformation part of EvalF2()
	int order = (InContinuumMechanicsFormulation()) ? GetIntegrationOrderCMAxial() : GetIntegrationOrderStretchAndShear();   // must be set equally to integration order in axial deformation part of strain energy (EvalF2()) for each BeamFormulation, respectively!!!
	Vector xT, wT;
	GetIntegrationRule(xT, wT, order); 

	double val = 0.;
	for (int i1=1; i1<=xT.GetLen(); i1++)  //i1-loop over all integration points
	{
		double x = xT(i1);		//x in unit configuration [-1 , 1]
		double dxidx = GetLx()*0.5;	//determinant of the jacobian (from element transformation [-1 , 1] -> [-lx/2 , lx/2])
		double xi = x*dxidx;	//xi in scaled configuration [-lx/2 , lx/2]
		Vector3D p0(xi, 0., 0.);

		// integrate r,xi along xi in [-lx/2, +lx/2]
		Vector sfx(NS(), 1); GetShapesx(sfx, p0);
		Matrix Sfx(Dim(), NS()*Dim());
		for (int i=1; i<=NS(); i++)
		{
			for (int j=1; j<=Dim(); j++)
			{
				Sfx(j, (i-1)*Dim() + j) = sfx(i);
			}
		}
		Vector posx = Sfx*q0;
		double norm_posx = posx.GetNorm();
		val += norm_posx*wT(i1)*dxidx;
	}

	return val;
}


// default set function
void ANCFBeamShear3DGeneric::SetANCFBeamShear3DGeneric(int materialnumi, const Vector3D& coli, int beam_formulation)
{
	//set reference configuration
	int idx = 1;
	for(int i=1; i <= nnodes; i++)   
	{
		ANCFNodeS2S3_3D& n((ANCFNodeS2S3_3D&)GetNode(i));
		q0(idx++) = n.Pos().X();
		q0(idx++) = n.Pos().Y();
		q0(idx++) = n.Pos().Z();
		q0(idx++) = n.GetRefSlope2().X();
		q0(idx++) = n.GetRefSlope2().Y();
		q0(idx++) = n.GetRefSlope2().Z();
		q0(idx++) = n.GetRefSlope3().X();
		q0(idx++) = n.GetRefSlope3().Y();
		q0(idx++) = n.GetRefSlope3().Z();
	}

	materialnum = materialnumi;

	double lx = CalculateElementLength();
	double ly, lz;
	const Beam3DProperties& bp = GetBeam3DProperties();
	if (bp.IsRectangularCrossSection())
	{
		ly = bp.GetBeamThicknessY();
		lz = bp.GetBeamThicknessZ();
		size.Set(lx,ly,lz);
	}
	else
	{
		size.Set(lx,0,0);
	// $EK 2015-12-03 TO BE DONE	mbs->UO(UO_LVL_err) << "WARNING: Only rectangular cross section implemented for ANCFBeamShear3D at the moment!\n";
	}


	xg.SetLen(SOS());
	xg.SetAll(0.);
	
	col = coli;
	BuildDSMatrices();

	x_init.SetLen(2*SOS());
	x_init.SetAll(0.);
	
	SetBeamFormulation(beam_formulation);

	SetElementName(GetElementSpec());
}

// deprecated set function!!!
void ANCFBeamShear3DGeneric::SetANCFBeamShear3DGeneric(int materialnumi, const Vector3D& si, const Vector3D& coli)
{
	size = si;

	xg.SetLen(SOS());
	xg.SetAll(0.);
	
	//lx = size.X(); ly = size.Y(); lz = size.Z(); //lx is parameter to identify reference configuration
	materialnum = materialnumi;
	col = coli;
	BuildDSMatrices();

	x_init.SetLen(2*SOS());
	x_init.SetAll(0.);

	SetElementName(GetElementSpec());
}

void ANCFBeamShear3DGeneric::GetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	Body3D::GetElementData(edc);
}

//user has changed data ==> write new data and parameters to element
int ANCFBeamShear3DGeneric::SetElementData(ElementDataContainer& edc) 		//fill in all element data
{
	int rv = Body3D::SetElementData(edc);
	return rv;
}

void ANCFBeamShear3DGeneric::BuildDSMatrices()
{
};

//$ EK 2015-10-21: added in order to set the outer faces correctly 
void ANCFBeamShear3DGeneric::PostAssemble() 
{
	SetOuterFaces();
}

//if interpolated values should be used for drawing the average (integral) along specific points
//on the cross section are computed



void ANCFBeamShear3DGeneric::Gradu(const Vector3D& ploc, const Vector& u, Matrix3D& gradu) const
{
	gradu.SetSize(3,3);
	gradu.SetAll(0);

	int dim = Dim();
	int l;
	for (int j = 1; j <= dim; j++) 
	{
		for (int i = 1; i <= NS(); i++)
		{
			l = (i-1)*dim+j;
			gradu(j,1) += GetSFx(i,ploc)*u(l);
			gradu(j,2) += GetSFy(i,ploc)*u(l);
			gradu(j,3) += GetSFz(i,ploc)*u(l);
		}
	}
}

//----------------
//Velocity
//----------------

//flagD = 0 fr Berechnungen, flagD = 1 for Visualization

Vector3D ANCFBeamShear3DGeneric::GetVel3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	for (int i = 1; i <= Dim(); i++) //i=1,2,3
	{
		for (int j = 1; j <= NS(); j++)
		{	
			if(flagD==0)
				p(i) += GetSF(j,p_loc)*XGP((j-1)*Dim()+i);  //XGP=actual solution vector

			else
				p(i) += GetSF(j,p_loc)*XGPD((j-1)*Dim()+i);
		}
	}
	return p;
};

//----------------
//Displacement
//----------------
//only for visualization
Vector3D ANCFBeamShear3DGeneric::GetDisplacementD(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	for (int i = 1; i <= Dim(); i++)
	{
		for (int j = 1; j <= NS(); j++)
		{
			p(i) += GetSF(j,p_loc)*(XGD((j-1)*Dim()+i));
		}
	}
	return p;
};

Vector3D ANCFBeamShear3DGeneric::GetDisplacement(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	for (int i = 1; i <= Dim(); i++)
	{
		for (int j = 1; j <= NS(); j++)
		{
			p(i) += GetSF(j,p_loc)*(XG((j-1)*Dim()+i));
		}
	}
	return p;
};

//Vector3D ANCFBeamShear3DGeneric::GetDisplacement3D(const Vector3D& p_loc, const Vector& xg) const
//{
//	Vector3D p(0.,0.,0.);
//	for (int i = 1; i <= Dim(); i++)
//	{
//		for (int j = 1; j <= NS(); j++)
//		{
//			p(i) += GetSF(j,p_loc)*(xg((j-1)*Dim()+i));
//		}
//	}
//	return p;
//};
//

void ANCFBeamShear3DGeneric::GetIntDuDq(Matrix& dudq)
{
	//integrate du/dq over beam volume   => (crossection area) * (integration over axis)
	//u(q) = r(q) - r0  ==>  du/dq = dr/dq

	dudq.SetSize(NS()*Dim(), Dim());
	dudq.SetAll(0);

	ConstVector<4> w1, x1;//max. 4 integration points -> order 7
	int order = InContinuumMechanicsFormulation() ? GetIntegrationOrderCMAxial() : GetIntegrationOrderStretchAndShear();
	GetIntegrationRule(x1, w1, order);
	//$EK 2015-11	- use cross section type of beam material
	//double fact = GetLx()*0.5*GetLy()*GetLz();									//crosssection area times determinant for line integral
	double fact = GetLx()*0.5*GetBeam3DProperties().GetCrossSectionArea();							//correct crosssection area times determinant for line integral

	for (int i1=1; i1<=x1.GetLen(); i1++)
	{
		double scaled_weight = w1(i1) * fact;			//scale weight
		double xi = 0.5*GetLx()*x1(i1);
		Vector3D ploc(xi, 0., 0.);

		for (int j=1; j<=NS(); j++)
		{
			double sf_value = GetSF(j, ploc);			//value returned by i'th (positional) shape function at point xi
			double block_j = Dim()*(j-1);						//addresses (j-1)st block in dudq
			for (int k=1; k<=Dim(); k++)
			{
				dudq(k+block_j, k) += scaled_weight*sf_value;
			}
		}
	}
}

//$EK 2015-11-18 spatially and time varying body loads for flexible bodies - therefore integration
void ANCFBeamShear3DGeneric::GetIntfDuDq(MathFunction& mathfunc, Vector& dudq, double t) //in fact it is f*DuDq Transposed
{
	//integrate f du/dq over beam volume   => (crossection area) * (integration over axis)
	//u(q) = r(q) - r0  ==>  du/dq = dr/dq

	//dudq.SetSize(NS()*Dim(), Dim());
	dudq.SetLen(NS()*Dim());
	dudq.SetAll(0);

	ConstVector<4> w1, x1;//max. 4 integration points -> order 7
	ConstVector<3> w2, x2, w3, x3;//max. 3 integration points -> order 5

	int order = InContinuumMechanicsFormulation() ? GetIntegrationOrderCMAxial() : GetIntegrationOrderStretchAndShear();
	GetIntegrationRule(x1, w1, order);  //axial x-direction
	GetIntegrationRule(x2, w2, order);  //transversal y-direction
	GetIntegrationRule(x3, w3, order);  //vertical z-direction
	


	//$EK 2015-11	- use cross section type of beam material
	//double fact = GetLx()*0.5*GetLy()*GetLz();									//crosssection area times determinant for line integral
	//double fact = GetLx()*0.5*GetBeam3DProperties().BeamA();							//correct crosssection area times determinant for line integral
	//fact = volume of beam/ volume of reference(=8)
	double fact = GetLx()*GetBeam3DProperties().GetCrossSectionArea()/8.;							//correct crosssection area times determinant for line integral

	for (int i1=1; i1<=x1.GetLen(); i1++)
	{
		double x1i = 0.5*GetLx()*x1(i1);
		for (int i2=1; i2<=x2.GetLen(); i2++)
		{
			double x2i = 0.5*GetLy()*x2(i2);
			for (int i3=1; i3<=x3.GetLen(); i3++)
			{
				double x3i = 0.5*GetLz()*x3(i3);
				double scaled_weight = w1(i1) * w2(i2) * w3(i3) * fact;			//scale weight
				Vector3D ploc(x1i, x2i, x3i);
				Vector3D pos = GetPos3D(ploc, 0 );	
				Vector mathfuncIn(t, pos.X(), pos.Y(), pos.Z());
				mathfunc.SetInputVector(mathfuncIn);

				double mathfuncVal = mathfunc.Evaluate(t);
				Vector& mathfuncVec = mathfunc.GetOutputVector();

				for (int j=1; j<=NS(); j++)
				{
			
					double sf_value = GetSF(j, ploc);			//value returned by i'th (positional) shape function at point xi
					double block_j = Dim()*(j-1);						//addresses (j-1)st block in dudq

					for (int k=1; k<=Dim(); k++)
					{
						//dudq(k+block_j, k) += scaled_weight*mathfuncVal*sf_value;
						//dudq(k+block_j, k) += scaled_weight*sf_value*mathfuncVec(k);
						dudq(k+block_j) += scaled_weight*sf_value*mathfuncVec(k);
					}
				}
			}
		}
	}
}




//void ANCFBeamShear3DGeneric::GetIntDuDqxMathFunction(Matrix& dudq, const MathFunction& mf)
//{
//	//integrate du/dq over beam volume   => (crossection area) * (integration over axis)
//	//u(q) = r(q) - r0  ==>  du/dq = dr/dq
//
//	dudq.SetSize(NS()*Dim(), Dim());
//	dudq.SetAll(0);
//
//	ConstVector<4> w1, x1;//max. 4 integration points -> order 7
//	GetIntegrationRule(x1, w1, 3);
//	//$EK 2015-11	- use cross section type of beam material
//	//double fact = GetLx()*0.5*GetLy()*GetLz();									//crosssection area times determinant for line integral
//	double fact = GetLx()*0.5*GetBeam3DProperties().BeamA();							//correct crosssection area times determinant for line integral
//
//	for (int i1=1; i1<=x1.GetLen(); i1++)
//	{
//		double scaled_weight = w1(i1) * fact;			//scale weight
//		double xi = 0.5*GetLx()*x1(i1);
//		Vector3D ploc(xi, 0., 0.);
//		//compute average over cross section
//		//depending on crosssecion type
//		doulbe mfval = CalcMathFunctionAverageOverCrossSection(xi, mf);
//
//		for (int j=1; j<=NS(); j++)
//		{
//			double sf_value = GetSF(j, ploc);			//value returned by i'th (positional) shape function at point xi
//			double block_j = Dim()*(j-1);						//addresses (j-1)st block in dudq
//			for (int k=1; k<=Dim(); k++)
//			{
//				dudq(k+block_j, k) += scaled_weight*sf_value*mfval;
//			}
//		}
//	}
//}

double ANCFBeamShear3DGeneric::	CalcIntegratefdA(double xi, const CParsedFunction& mf) const
{
//	mf.Evaluate
	return 0.;
}
//----------------
//GetPos
//p_loc in [-Lx/2,Lx/2] x [-Ly/2,Ly/2] x [-Lz/2,Lz/2]
//----------------

Vector3D ANCFBeamShear3DGeneric::GetPosD(const Vector3D& p_loc) const 
{ 
	return GetPos3D(p_loc, 1); 
}

//Vector3D ANCFBeamShear3DGeneric::GetPosD(const Vector3D& p_loc) const
//// p_loc in [-Lx/2,Lx/2] x [-Ly/2,Ly/2] x [-Lz/2,Lz/2]
//{
//	//mbs->UO() << "posy=" << GetPosy3D(p_loc, 1) << "\n";
//	//mbs->UO() << "posz=" << GetPosz3D(p_loc, 1) << "\n";
//	//mbs->UO() << "ploc=" << p_loc << "\n";
//	Vector3D p0(p_loc.X(),0.,0.);
//
//	return GetPos3D(p_loc, 1) + p_loc.Y()*GetPosy3D(p_loc, 1) + p_loc.Z()*GetPosz3D(p_loc, 1);
//}

//GetPos2D berechnet Vektor Positionsvektor r im deformierten Element
//GetPos2D computes r in deformed element
//xg = generalized coordinates (= our unknowns q)
Vector3D ANCFBeamShear3DGeneric::GetPos3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0.);
	//TMStartTimer(24);
	Vector sf(NS(), 1);
	GetShapes(sf, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	//TMStopTimer(24);
	return p;
};

//GetPos2D von oben, aber falls im Aufruf als zweite Eingabe kein Vektor, sondern ein Integer steht,
//wird diese Fkt fr Grafik aufgerufen; globale Variable XG wird fr Rechnung verwendet anstatt Eingabe xg
//xg=XG wird einmal gesetzt, weshalb GetPos3D mit xg schneller ist, wohingegen GetPos3D mit XG jedes Mal aufs globale XG zugreift und das dauert lange

//same GetPos2D function as above, but with the second parameter flag D the function for the visualization is activated
Vector3D ANCFBeamShear3DGeneric::GetPos3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapes(sf, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sf(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{ 
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sf(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
};

//Ableitungen von GetPos entsprechen unseren Richtungsvektoren
//GetPosx2D (Ableitung in Richtung der Balkenachse: dr/dx=r,x)
//GetPosx2D is the derivative with respect to the direction of the beam centerline
Vector3D ANCFBeamShear3DGeneric::GetPosx3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0.);
	Vector sfx(NS(), 1);
	GetShapesx(sfx, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sfx(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosx3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfx(NS(), 1);
	GetShapesx(sfx, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfx(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{ 
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfx(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
}


void ANCFBeamShear3DGeneric::GetdPosdx(const Vector3D& ploc, Vector3D& dpdx) 
{
	dpdx = GetPosx3D(ploc);
};


//GetPosy2D (Ableitung in Richtung des Querschnitts: dr/dy=r,y)
//GetPosy2D is the derivative with respect to the direction of the cross section
Vector3D ANCFBeamShear3DGeneric::GetPosy3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0.);
	Vector sfy(NS(), 1);
	GetShapesy(sfy, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sfy(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosy3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfy(NS(), 1);
	GetShapesy(sfy, p_loc);
	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfy(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfy(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosyP3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfy(NS(), 1);
	GetShapesy(sfy, p_loc);
	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfy(j)*XGP(offset + i);
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfy(j)*XGPD(offset + i);
			}
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosz3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0);
	Vector sfz(NS(), 1);
	GetShapesz(sfz, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sfz(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosz3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfz(NS(), 1);
	GetShapesz(sfz, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfz(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfz(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPoszP3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfz(NS(), 1);
	GetShapesz(sfz, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfz(j)*XGP(offset + i);
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfz(j)*XGPD(offset + i);
			}
		}
	}
	return p;
}


//GetPosxy2D: d2r/dydx=r,xy
//is used in DeltaThetax
Vector3D ANCFBeamShear3DGeneric::GetPosxy3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0.);
	Vector sfxy(NS(), 1);
	GetShapesxy(sfxy, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sfxy(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosxy3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfxy(NS(), 1);
	GetShapesxy(sfxy, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfxy(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfxy(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosxz3D(const Vector3D& p_loc, const Vector& xg) const
{
	Vector3D p(0.,0.,0.);
	Vector sfxz(NS(), 1);
	GetShapesxz(sfxz, p_loc);

	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sfxz(j)*(xg(offset + i) + q0(offset + i)); //r,y=S,y*u+S,y*q_0		
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetPosxz3D(const Vector3D& p_loc, int flagD) const
{
	Vector3D p(0.,0.,0.);
	Vector sfxz(NS(), 1);
	GetShapesxz(sfxz, p_loc);

	if (!flagD)
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfxz(j)*(XG(offset + i) + q0(offset + i));
			}
		}
	}
	else
	{
		for (int j = 1; j <= NS(); j++)
		{
			int offset = (j-1)*Dim();
			for (int i = 1; i <= Dim(); i++)
			{
				p(i) += sfxz(j)*(XGD(offset + i) + q0(offset + i));
			}
		}
	}
	return p;
}

//computes position vector in the reference element: r_0 
//p_loc in [-Lx/2,Lx/2] x [-Ly/2,Ly/2] x [-Lz/2,Lz/2]
Vector3D ANCFBeamShear3DGeneric::GetInitPos3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapes(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
};

Vector3D ANCFBeamShear3DGeneric::GetInitPosx3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapesx(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
};

Vector3D ANCFBeamShear3DGeneric::GetInitPosy3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapesy(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetInitPosz3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapesz(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetInitPosxy3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapesxy(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
}

Vector3D ANCFBeamShear3DGeneric::GetInitPosxz3D(const Vector3D& p_loc) const
{
	Vector3D p(0.,0.,0.);
	Vector sf(NS(), 1);
	GetShapesxz(sf, p_loc);
	for (int j = 1; j <= NS(); j++)
	{
		int offset = (j-1)*Dim();
		for (int i = 1; i <= Dim(); i++)
		{
			p(i) += sf(j)*q0(offset + i);
		}
	}
	return p;
}

//for visualization
Vector3D ANCFBeamShear3DGeneric::GetPos3D0D(const Vector3D& p_loc) const 
{
	Vector3D plocscaled;
	plocscaled(1)=p_loc(1)*GetLx()/2.;
	plocscaled(2)=p_loc(2)*GetLy()/2.;
	plocscaled(3)=p_loc(3)*GetLz()/2.;
	return GetPos3D(plocscaled,1); 
};

Vector3D ANCFBeamShear3DGeneric::GetPos3D0D(const Vector3D& p_loc, double defscale) const 
{
	Vector3D plocscaled;
	plocscaled(1)=p_loc(1)*GetLx()/2.;
	plocscaled(2)=p_loc(2)*GetLy()/2.;
	plocscaled(3)=p_loc(3)*GetLz()/2.;
	return GetInitPos3D(plocscaled) + defscale*GetDisplacementD(plocscaled); 

	GetNodePos(1);
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Vector3D ANCFBeamShear3DGeneric::GetNodeLocPos(int i) const //returns position of i-th node
{
	switch(i)
	{
	case 1:	return Vector3D(-GetLx()*.5, 0, 0);
	case 2:	return Vector3D(+GetLx()*.5, 0, 0);
	case 3:	assert(nnodes >= 3); return Vector3D(0.);
	}
	assert(0);
	return Vector3D(0.);
}


Vector3D ANCFBeamShear3DGeneric::GetNodePos(int i) const //returns position of i-th node
{
	switch(i)
	{
	case 1:	return GetPos(Vector3D(-GetLx()*.5, 0, 0));
	case 2:	return GetPos(Vector3D(+GetLx()*.5, 0, 0));
	case 3:	assert(nnodes >= 3); return GetPos(Vector3D(0.));
	}
	assert(0);
	return Vector3D(0.);
}

Vector3D ANCFBeamShear3DGeneric::GetNodePosD(int i) const //returns position of i-th node (draw mode)
{
	switch(i)
	{
	case 1:	return GetPosD(Vector3D(-GetLx()*.5, 0, 0));
	case 2:	return GetPosD(Vector3D(+GetLx()*.5, 0, 0));
	case 3:	assert(nnodes >= 3); return GetPosD(Vector3D(0.));
	}
	assert(0);
	return Vector3D(0.);
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Vector3D ANCFBeamShear3DGeneric::GetDOFDirD(int idof) const //returns direction of action of i-th DOF
{
	if ((idof >= 1 && idof <= 3) ||
		(idof >= 10 && idof <= 12) ||
		(idof >= 19 && idof <= 21))
	{
		Vector3D v(0.);
		int component = (idof-1)%3+1;
		v(component) = 1;
		return v;
	}
	//for other components, the direction can not be easily identified and is therefore set to zero (default)

	return Vector3D(0.,0.,0.);
}

Vector3D ANCFBeamShear3DGeneric::GetDOFPosD(int idof) const //returns position of i-th DOF
{
	if (idof <= 9) //left node: 1,...,9
	{
		return GetPos3D(Vector3D(-0.5*GetLx(),0.,0.),1);//1 fr flagD
	}
	else if (idof <= 18) //right node: 10,...,18
	{
		return GetPos3D(Vector3D(0.5*GetLx(),0.,0.),1);
	}
	else if (idof <= 27) //middle node: 19,...,27
	{
		return GetPos3D(Vector3D(0.,0.,0.),1);//middle node: (0,0)
	}
	else
	{
		return Vector3D(0, 0, 0);
	}
}


Matrix3D ANCFBeamShear3DGeneric::GetRotMatrix(const Vector3D& p0, int on_reference_element, int flagD) const //p0(1) in [-lx/2,lx/2]
{
	//ROTATION TENSOR: 
	//A = [ e1 | e2 | e3 ]

	//e1 = e1^{\bar}/|e1^{\bar}| with e1^{\bar} = drdy x drdz
	//e3 = e3^{\bar}/|e3^{\bar}| with e3^{\bar} = drdz
	//e2 = e2^{\bar}/|e2^{\bar}| with e2^{\bar} = drdz x (drdy x drdz) = e3^{\bar} x e1^{\bar}
	
	//Attention: Getei(i) returns ei^{\bar} 

  Vector3D drdy, drdz;

	if (on_reference_element)
	{
		drdy = GetInitPosy3D(p0);
		drdz = GetInitPosz3D(p0);
	}
	else
	{
		drdy = GetPosy3D(p0, flagD);
		drdz = GetPosz3D(p0, flagD);
	}
	
	TArray<Vector3D> e(Dim());
	for (int i=1; i<=Dim(); i++)
	{
		e(i) = Getei(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		e(i).Normalize();
	}

	Matrix3D A(e(1).X(),e(2).X(),e(3).X(),
					   e(1).Y(),e(2).Y(),e(3).Y(),
					   e(1).Z(),e(2).Z(),e(3).Z());
	return A;
}

Matrix3D ANCFBeamShear3DGeneric::GetRotMatrixP(const Vector3D& p0, int flagD) const
{
	//Time derivative of rotation tensor: 
	//AP = [ e1P | e2P | e3P ]

	//e1 = e1^{\bar}/|e1^{\bar}| with eP1^{\bar} = drdyP x drdz + drdy x drdzP
	//e3 = e3^{\bar}/|e3^{\bar}| with eP3^{\bar} = drdzP
	//e2 = e2^{\bar}/|e2^{\bar}| with eP2^{\bar} = drdzP x (drdy x drdz) + drdz x (drdyP x drdz + drdy x drdzP) = eP3^{\bar} x e1^{\bar} + e3^{\bar} x eP1^{\bar}
	
	//Attention: Getei(i) returns ei^{\bar}, and GetePi(i) returns ePi^{\bar} 

  Vector3D drdy, drdz, drPdy, drPdz;

	drdy = GetPosy3D(p0, flagD);
	drdz = GetPosz3D(p0, flagD);
	drPdy = GetPosyP3D(p0, flagD);
	drPdz = GetPoszP3D(p0, flagD);
	
	Matrix3D AP;
	for (int i=1; i<=3; i++)
	{
		Vector3D ei = Getei(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		double ei_norm = ei.Norm();
		Vector3D eiP = GeteiP(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z(), drPdy.X(), drPdy.Y(), drPdy.Z(), drPdz.X(), drPdz.Y(), drPdz.Z());
		double ei_eiP = ei*eiP;
		eiP *= (1./ei_norm);
		eiP += (-ei_eiP/pow(ei_norm,3))*ei;	
		for (int j=1; j<=3; j++)
		{
			AP(j,i) = eiP(j);
		}
	}

	return AP;
}

Vector3D ANCFBeamShear3DGeneric::GetTwistAndCurvature(const Vector3D& ploc, int on_reference_element, int flagD) const 
// ploc.X() in [ -Lx/2 , Lx/2 ]
// calculate vector of twist and curvature k = 1/2 sum_i e_i x e_i'
// or, for the reference element (see flag) k_0 = 1/2 sum_i e0_i x e0_i'
// NOTE: both vectors are returned with their components w.r.t. the local frame basis (e_i)!
{
	if(IsStraightBeamInReferenceConfiguration() && on_reference_element)  // for speedup
	{
		return Vector3D();   //std constructor initializes with zero
	}

	Vector3D drdy, drdz, ddrdxdy, ddrdxdz;
	if (on_reference_element)
	{
		drdy = GetInitPosy3D(ploc);
		drdz = GetInitPosz3D(ploc);
		ddrdxdy = GetInitPosxy3D(ploc);
		ddrdxdz = GetInitPosxz3D(ploc);
	}
	else
	{
		drdy = GetPosy3D(ploc, flagD);
		drdz = GetPosz3D(ploc, flagD);
		ddrdxdy = GetPosxy3D(ploc, flagD);
		ddrdxdz = GetPosxz3D(ploc, flagD);
	}
	
	Vector3D ei;     // returns ei^{\bar}, ei = ei^{\bar} / |ei^{\bar}|
                                 //e1 = e1^{\bar}/|e1^{\bar}| with e1^{\bar} = drdy x drdz
	                               //e3 = e3^{\bar}/|e3^{\bar}| with e3^{\bar} = drdz
	                               //e2 = e2^{\bar}/|e2^{\bar}| with e2^{\bar} = drdz x (drdy x drdz) = e3^{\bar} x e1^{\bar}
	Vector3D deidxi; // returns d ei^{\bar} / d xi

	Vector3D k;
	for (int i=1; i<=Dim(); i++)
	{
		ei = Getei(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		deidxi = Getdeidxi(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z(), ddrdxdy.X(), ddrdxdy.Y(), ddrdxdy.Z(), ddrdxdz.X(), ddrdxdz.Y(), ddrdxdz.Z());

		k += ( 0.5 / ei.Norm2() ) * ei.Cross(deidxi);  //k = 1/2 sum_i e_i x e_i' -> simplified written using e_i^{\bar}-Vectors (see the paper)
	}

	Matrix3D AT = GetRotMatrix(ploc, on_reference_element, flagD).GetTp();   //in order to represent the vector k in the local frame basis

	return AT * k;
}

Vector3D ANCFBeamShear3DGeneric::GetAngularVel(const Vector3D& p_loc) const
{
	return GetAngularVel(p_loc, 0);

  //use one single function for draw/compute JG2013-09
	//Matrix retmat;
	//Vector xp(SOS());
	//for (int j=1; j<=SOS(); j++)
	//{
	//	xp(j)=XGP(j);
	//}
	//Matrix3D A = GetRotMatrix(p_loc);
	//Matrix3D omega_skew;
	//for (int i=1; i<=Dim(); i++)
	//{
	//	Vector3D ei(A(1,i),A(2,i),A(3,i));
	//	GetdRotvdqT(ei, p_loc, retmat, 1);
	//	Vector wi(3);
	//	Mult(xp, retmat, wi);
	//	for (int j=1; j<=Dim(); j++)
	//	{
	//		omega_skew(i,j) = wi(j);
	//	}
	//}
	////Matrix3D omega_skew = GetRotMatrixP(p_loc)*GetRotMatrix(p_loc).GetTp();
	////omega_skew * A = RotMatrixP

	//return Vector3D(omega_skew(2,3),omega_skew(1,3),-omega_skew(1,2));
}

//use one single function for draw/compute JG2013-09
Vector3D ANCFBeamShear3DGeneric::GetAngularVel(const Vector3D& p_loc, int flagD) const
{
	Matrix retmat;
	Vector xp(SOS());
	for (int j=1; j<=SOS(); j++)
	{
		if (flagD) {xp(j)=XGPD(j);}
		else       {xp(j)=XGP(j);}
	}
	Matrix3D A = GetRotMatrix(p_loc, 0, flagD);
	Matrix3D omega_skew;
	for (int i=1; i<=Dim(); i++)
	{
		Vector3D ei(A(1,i),A(2,i),A(3,i));
		GetdRotvdqT(ei, p_loc, retmat, 1, flagD);
		Vector wi(3);
		Mult(xp, retmat, wi);
		for (int j=1; j<=Dim(); j++)
		{
			omega_skew(i,j) = wi(j);
		}
	}
	//Matrix3D omega_skew = GetRotMatrixP(p_loc)*GetRotMatrix(p_loc).GetTp();
	//omega_skew * A = RotMatrixP

	return Vector3D(omega_skew(2,3),omega_skew(1,3),-omega_skew(1,2));
}

void ANCFBeamShear3DGeneric::GetdRotvdqT(const Vector3D& vloc, const Vector3D& ploc, Matrix& drotvdqT, int use_transposed_rotmatrix, int flagD) const
{
	// rot = (rot11 rot12 rot13
	//        rot21 rot22 rot23
	//        rot31 rot32 rot33)
	// Basis vectors are columns of rotation tensor
	// rotT.v = (rot11*v1 + rot21*v2 + rot31*v3, rot12*v1 + rot22*v2 + rot32*v3, rot13*v1 + rot23*v2 + rot33*v3)
	// p = (r,y , r,z) = ( r1dy , r2dy , r3dy , r1dz , r2dz , r3dz )
	// drotTv/dq = drotTv/dp . dp/dq

	double xi = ploc.X();		// ploc.X() in [ -Lx/2 , Lx/2 ]

	//rotation tensor only depends on r,y and r,z
	Vector3D drdy = GetPosy3D(xi, flagD);
	Vector3D drdz = GetPosz3D(xi, flagD);

	Vector3D e1(Getei(1, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
	Vector3D e2(Getei(2, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
	Vector3D e3(Getei(3, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
	TArray<Vector3D> de1dp(GetdeidDiffVar(1, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
	TArray<Vector3D> de2dp(GetdeidDiffVar(2, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
	TArray<Vector3D> de3dp(GetdeidDiffVar(3, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));

// calculate derivatives of normalized ei
	double e1norm = e1.Norm();
	double e2norm = e2.Norm();
	double e3norm = e3.Norm();
	for (int j=1; j<=de1dp.Length(); j++)
	{
		de1dp(j) = de1dp(j)*(1./e1norm) - e1*((e1*de1dp(j))/Cub(e1norm));
		de2dp(j) = de2dp(j)*(1./e2norm) - e2*((e2*de2dp(j))/Cub(e2norm));
		de3dp(j) = de3dp(j)*(1./e3norm) - e3*((e3*de3dp(j))/Cub(e3norm));
	}

  Matrix drotvdp(Dim(), 6);//3x6
	if (use_transposed_rotmatrix)    //use transposed A: A^T*v
	{
		for (int h=1; h<=6; h++)//h...DiffVar
		{
			drotvdp(1,h) = de1dp(h)*vloc;
			drotvdp(2,h) = de2dp(h)*vloc;
			drotvdp(3,h) = de3dp(h)*vloc;
		}
	}
	else //dont use transposed A: A*v
	{
		for (int h=1; h<=6; h++) //h...DiffVar
		{
			drotvdp(1,h) = de1dp(h).X()*vloc.X() + de2dp(h).X()*vloc.Y() + de3dp(h).X()*vloc.Z();
			drotvdp(2,h) = de1dp(h).Y()*vloc.X() + de2dp(h).Y()*vloc.Y() + de3dp(h).Y()*vloc.Z();
			drotvdp(3,h) = de1dp(h).Z()*vloc.X() + de2dp(h).Z()*vloc.Y() + de3dp(h).Z()*vloc.Z();
		}
	}
	//h = 1,2,3 -> SFy
	//h = 4,5,6 -> SFz
	drotvdqT.SetSize(SOS(),Dim());
	drotvdqT.SetAll(0.);
	Vector sfy(NS(), 1); GetShapesy(sfy, ploc);
	Vector sfz(NS(), 1); GetShapesz(sfz, ploc);

	for (int k=1; k<=Dim(); k++)
	{
		for (int i=1; i<=3; i++)
			for (int j=1; j<=NS(); j++)
				drotvdqT((j-1)*Dim()+i, k) = drotvdp(k,i)*sfy(j) + drotvdp(k,Dim()+i)*sfz(j);
	}
}

void ANCFBeamShear3DGeneric::GetdRotdqT(const Vector3D& ploc, Matrix& d)
{
	//ploc in [-lx/2,lx/2] x [-ly/2,ly/2] x [-lz/2,lz/2]
	////////////////////////////////////////////////////////////
	//KN&PG's approach   \cite{hodges06} 
	//let e_i denote the i-th base vector of the actual system (i-th column of the rotation matrix)
	//\delta \theta = \sum_{i=1}^3  e_i (\delta e_j . e_k)    where j = i%3+1, k = j%3+1
	//(\grad_q \theta^T)_{l,m} = (e_i)_m (e_k)_n (partial (e_j)_n/ \partial q_l\)  
	//note: for this method the call of routine GetdRotvdqT has to sum up (e_j)_n v_n instead of (e_j)_n v_j, i.e. use_transposed_rotmatrix = 1
	////////////////////////////////////////////////////////////

	d.SetSize(SOS(),Dim());
	d.FillWithZeros();

	double xi = ploc.X();		// ploc.X() in [ -Lx/2 , Lx/2 ]

	//rotation tensor only depends on r,y and r,z
	Vector3D drdy = GetPosy3D(xi);
	Vector3D drdz = GetPosz3D(xi);

	//-----------------------------------------
	// new version
	//-----------------------------------------
	Vector3D ei;
	Vector3D ej;
	Vector3D ek;
	TArray<Vector3D> dejdp(6);
	TArray<Vector3D> dekdp(6);
	TArray<Vector3D> deidp(6);
	ConstVector<27> helpy;
	helpy.SetAll(0.);


	bool define_theta_by_a_transposed = false;
	bool derivatives_wrt_rotated_vars = false;

	for (int i=1; i<=Dim(); i++)	//i=1,2,3
	{
		int j = i%3 + 1;		//i=1->j=2->k=3, i=2->j=3->k=1, i=3->j=1->k=2
		int k = j%3 + 1;

		// e_i = ebar_i	/ |ebar_i|
		ei = Getei(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		ej = Getei(j, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		ek = Getei(k, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());

		if (define_theta_by_a_transposed)
		{
			double memo;
			memo = ei(j);
			ei(j) = ej(i);
			ej(i) = memo;
			memo = ej(k);
			ej(k) = ek(j);
			ek(j) = memo;
			memo = ek(i);
			ek(i) = ei(k);
			ei(k) = memo;
		}

		ei.Normalize();
		ek.Normalize();
		double ej_norm = ej.Norm();
		double ej_norm3 = Cub(ej_norm);
		ej.Normalize();

		// GetdeidDiffVar(j,...) returns debar_j_dp
		deidp = GetdeidDiffVar(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		dejdp = GetdeidDiffVar(j, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		dekdp = GetdeidDiffVar(k, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());

		// calculate all n=2*3=6 partial derivatives of ej (=normalized ebar_j) by using the derivatives of ebar_j
		for (int n=1; n<=dejdp.Length(); n++)
		{
			if (define_theta_by_a_transposed)
			{
				dejdp(n)(i) = deidp(n)(j);
				dejdp(n)(k) = dekdp(n)(j);
			}

			//dejdp(n) = dejdp(n)*(1./ej_norm) - ej*((ej*dejdp(n))/ej_norm3);
			dejdp(n) *= 1./ej_norm;
			dejdp(n) -= ej*(ej*dejdp(n));
		}

		if (!derivatives_wrt_rotated_vars)
		{
			//assembling of a help-vector saving at the (m*Dim()+n)th position the value
			//(ek^T * dejdpy(n))*SFy(m) + (ek^T * dejdpz(n))*SFz(m)
			for (int n=1; n<=Dim(); n++)
			{
				double fac_y = ek*dejdp(n);
				double fac_z = ek*dejdp(n+Dim());
				for (int m=1; m<=NS(); m++)
				{
					helpy((m-1)*Dim() + n) = fac_y*GetSFy(m, ploc) + fac_z*GetSFz(m, ploc);
				}
			}
		}
		else
		{
			//assembling of a help-vector saving at the (m*Dim()+n)th position the value
			//(ek^T * dejdpy(n))*A^T*SFy(m) + (ek^T * dejdpz(n))*A^T*SFz(m)

			Vector3D vec_y, vec_z;
			for (int n=1; n<=Dim(); n++)
			{
				vec_y(n) = ek*dejdp(n);
				vec_z(n) = ek*dejdp(n+Dim());
			}

			Matrix3D A = GetRotMatrix(ploc);

			vec_y = vec_y*A.GetTp();
			vec_z = vec_z*A.GetTp();


			for (int n=1; n<=Dim(); n++)
			{
				for (int m=1; m<=NS(); m++)
				{
					helpy((m-1)*Dim() + n) = vec_y(n)*GetSFy(m, ploc) + vec_z(n)*GetSFz(m, ploc);
				}
			}
		}

		//dimension of output matrix d is SOS() times Dim()   (transpose of dAdq)
		for (int m=1; m<=SOS(); m++)
		{
			for (int n=1; n<=Dim(); n++)
			{
				d(m,n) += helpy(m)*ei(n);
			}
		}
	}


	////-----------------------------------------	
	////old version:
	////-----------------------------------------
	////
	//Matrix3D A = GetRotMatrix(ploc);
	////Matrix3D I = A.GetTp()*A;
	////mbs->UO() << I.DoubleContract(I)-3. << "\n";

	//Matrix drotvdqT(SOS(),Dim());
	//for (int i=1; i<=Dim(); i++)//i=1,2,3
	//{
	//	int j = i%3 + 1;//i=1->j=2->k=3, i=2->j=3->k=1, i=3->j=1->k=2
	//	int k = j%3 + 1;
	//	Vector3D ei(A(1,i), A(2,i), A(3,i)); //i-th column of A
	//	GetdRotvdqT(Vector3D(A(1,k), A(2,k), A(3,k)), ploc, drotvdqT, 1);  //k-th column of A (=v) * dRotdq -> dRotvdq
	//	for (int l=1; l<=SOS(); l++)//SOS=Dim*NS=3*9=27
	//		for (int m=1; m<=Dim(); m++)//3
	//			d(l,m) += drotvdqT(l,j)*ei(m);
	//}
}


//----------------
//Massmatrix M
//----------------
void ANCFBeamShear3DGeneric::EvalM(Matrix& m, double t)
{
	if (massmatrix.Getcols() == SOS())
	{
		m = massmatrix;
		return;
	}
	else
	{
		int dim = Dim(); 
		int ns = NS();
		Vector SV;
		SV.SetLen(ns);
		Matrix HL(ns*dim,dim);
		m.SetSize(SOS(), SOS());
		m.SetAll(0.);

		if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_rectangular)
		{
			Vector x1,x2,w1,w2; //integration points and according weights

			//GetIntegrationRule => Gau points on interval [-1,+1]
			GetIntegrationRule(x1,w1,7); //optimal: 6x3x3, 6x1x1 geht auch; (GetIntegrationRule-Order == 6 complies with == 7)
			GetIntegrationRule(x2,w2,3);

			for (int i1=1; i1<=x1.GetLen(); i1++)
			{
				for (int i2=1; i2<=x2.GetLen(); i2++)
				{
					for (int i3=1; i3<=x2.GetLen(); i3++)
					{
						double x = x1(i1)*0.5*GetLx();  //x1 in [-1,+1], scaled rect. reference element: \xi
						double y = x2(i2)*0.5*GetLy();
						double z = x2(i3)*0.5*GetLz();
						Vector3D ploc(x,y,z);
						Vector3D rx0 = GetInitPosx3D(ploc);
						double rxn = rx0.Norm(); //curvature in reference element
						double jacdet = 0.5*GetLx()*0.5*GetLy()*0.5*GetLz() * rxn; //element transformation

						GetShapes(SV,ploc);
						for (int i=0; i<ns; i++)
						{
							for (int j=1; j<=dim; j++)
							{
								HL(i*dim+j,j)=SV(i+1);
							}
						}
						m += fabs (jacdet) * Rho() * w1(i1)*w2(i2)*w2(i3) * (HL*HL.GetTp());
					}
				}
			}
		}
		else if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_circular || GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_tubular)
		{
			TArray<Vector3D> ip;
			Vector w;
			GetIntegrationRuleHollowCylinder(ip, w, 7, 2, GetIPNumberPerDiscQuadrant(), GetLx(), GetBeam3DProperties().GetRadius(), GetBeam3DProperties().GetInnerRadius());

			for (int ip_idx=1; ip_idx<=ip.Size(); ip_idx++)
			{
				Vector3D ploc(ip(ip_idx));
				Vector3D rx0 = GetInitPosx3D(ploc);
				double rxn = rx0.Norm(); //curvature in reference element
				double jacdet = rxn; //element transformation

				GetShapes(SV, ploc);
				for (int i=0; i<ns; i++)
				{
					for (int j=1; j<=dim; j++)
					{
						HL(i*dim + j, j) = SV(i+1);
					}
				}
				m += fabs(jacdet) * Rho() * w(ip_idx) * (HL*HL.GetTp());
			}
		}
		else
		{
			mbs->UO(UO_LVL_err) << "ERROR in ANCFBeamShear3DGeneric: assembling of mass matrix not implemented in case of GetBeam3DProperties().GetCrossSectionShape() == " << GetBeam3DProperties().GetCrossSectionShape() << "!\n";
			return;
		}

		massmatrix = m;
	}  
};

double ANCFBeamShear3DGeneric::GetPotentialEnergy()
{
	double energy = 0.;

	if (InContinuumMechanicsFormulation())
	{
		//TODO
		mbs->UO(UO_LVL_warn) << "WARNING: ANCFBeamShear3DGeneric::GetPotentialEnergy() not yet implemented!\n";
	}
	else
	{
		static Vector wT,xT;      // container for integration points and weights
		Vector3D gk(0.);          // container for Gamma (1,2,3), Kappa (1,2,3), and Strain E=(E_22,E_33,E_23) respectively
		double det = 0.5*GetLx(); // determinant of the jacobian (from element transformation [-1 , 1] -> [-lx/2 , lx/2])
		Vector3D d;               // container for stiffneses
		
		// material parameters
		double beamGJ = GetBeam3DProperties().BeamGJkx();
		double beamEIy = GetBeam3DProperties().BeamEIy();
		double beamEIz = GetBeam3DProperties().BeamEIz();
		double beamEA = GetBeam3DProperties().BeamEA();
		double beamGAky = GetBeam3DProperties().BeamGAky();
		double beamGAkz = GetBeam3DProperties().BeamGAkz();
		double beamGAkyz = BeamGAkyz(beamGAky, beamGAkz);

		// integration of bending terms
		GetIntegrationRule(xT, wT, GetIntegrationOrderBending());
		for (int i1=1; i1<=xT.GetLen(); i1++)  // loop over all integration points
		{
			double x = xT(i1);          // x in unit configuration [-1 , 1]
			double xi = x*det;  // xi in scaled configuration [-lx/2 , lx/2]

			d(1) = penalty(1)*beamGJ*wT(i1)*det;   // beamGJ  = torsional stiffness (including torsional correction factor, depending on the shape of the cross section)
			d(2) = penalty(2)*beamEIy*wT(i1)*det;  // beamEIy = bending stiffness (in e2=y direction))
			d(3) = penalty(3)*beamEIz*wT(i1)*det;  // beamEIz = bending stiffness (in e3=z direction)

			GetKappa(xi, gk);
			for (int j=1; j<=3; j++)
			{
				energy += d(j)*gk(j)*gk(j);
			}
		}

		// integration of terms of axial strain and shear deformation
		GetIntegrationRule(xT, wT, GetIntegrationOrderStretchAndShear());
		for (int i1=1; i1<=xT.GetLen(); i1++)
		{
			double x = xT(i1);
			double xi = x*GetLx()*0.5;
			double det = 0.5*GetLx();

			d(1) = penalty(4)*beamEA*wT(i1)*det;
			d(2) = penalty(5)*beamGAky*wT(i1)*det;
			d(3) = penalty(6)*beamGAkz*wT(i1)*det;

			GetGamma(xi, gk);
			for (int j=1; j<=3; j++)
			{
				energy += d(j)*gk(j)*gk(j);
			}
		}

		// integration of cross section deformation
		GetIntegrationRuleLobatto(xT, wT, GetIntegrationOrderThickness());
		for (int i1=1; i1<=xT.GetLen(); i1++)
		{
			double x = xT(i1);
			double xi = x*GetLx()*0.5;
			double det = 0.5*GetLx();

			d(1) = penalty(7)*beamEA*wT(i1)*det;
			d(2) = penalty(8)*beamEA*wT(i1)*det;
			d(3) = penalty(9)*beamGAkyz*wT(i1)*det;

			GetThicknessStrain(xi, gk);
			for (int j=1; j<=3; j++)
			{
				energy += d(j)*gk(j)*gk(j);
			}
		}
	}

	return energy*.5; 
}

double ANCFBeamShear3DGeneric::GetKineticEnergy()
{
	ConstVector<27> qdot(SOS(), 1); //nofill

	for (int k=1; k<=SOS(); k++)
	{
		qdot(k)=XGP(k);
	}

	ConstMatrix<27*27> m(SOS(),SOS(),1);
	EvalM(m, mbs->GetTime());
	double energy = ((m*qdot)*qdot)/2;
	
	return energy;
}

void ANCFBeamShear3DGeneric::GetJacobi(Matrix3D& jac, const Vector3D& ploc, const Vector& xg0) const
{
	jac.SetSize(3,3);
	jac.SetAll(0.);
	for (int i = 1; i <= Dim(); i++)
	{ 
		for (int k=1; k <= NS(); k++)
		{ 
			jac(i,1) += GetSFx(k, ploc)*xg0((k-1)*Dim()+i);
			jac(i,2) += GetSFy(k, ploc)*xg0((k-1)*Dim()+i);
			jac(i,3) += GetSFz(k, ploc)*xg0((k-1)*Dim()+i);
		}
	}
}

void ANCFBeamShear3DGeneric::EvalF2(Vector& f, double t)
{
	Body3D::EvalF2(f,t);
	//TMStartTimer(22);  //CPU timing

	bool add_damping_terms = !mbs->GetSolSet().dostaticcomputation && !GetMBS()->IsJacobianComputation() && (GetBeam3DProperties().DampingK() > 0 || GetBeam3DProperties().DampingM() > 0);
	static ConstVector<27> velocities(SOS(), 1);
	if (add_damping_terms)
	{
		for (int i = 1; i <= SOS(); i++)
			velocities(i) = XGP(i);
	}


	ConstVector<27> f_elastic_forces(SOS());	 //element residual - components are set to 0
	if (InContinuumMechanicsFormulation())
	{
		EvalF2CM(f_elastic_forces, t, add_damping_terms, velocities);
	}
	else
	{
		EvalF2SM(f_elastic_forces, t, add_damping_terms, velocities);
	}

	f -= f_elastic_forces;  //f=f-fadd, (Ku=f -> Residual=f-Ku, Ku=fadd)
	//TMStopTimer(22);


	// Reighley damping (mass proportional part)
	if (add_damping_terms && GetMaterial().DampingM() > 0)
	{
		if (mbs->GetSolSet().store_FE_matrices) // mass matrix is stored
		{
			if (massmatrix.Getcols() != SOS()) {EvalM(massmatrix, t); return;}
			Mult(massmatrix, velocities, f_elastic_forces);
		}
		else
		{
			ConstMatrix<27*27> m(SOS(), SOS());
			EvalM(m, t);
			Mult(m, velocities, f_elastic_forces);
		}

		f_elastic_forces *= GetMaterial().DampingM();
		f -= f_elastic_forces;
	}
}

//residual of variational formulation in structural mechanics formulation   (Reissner-Simo-VuQuoc Formulation)
void ANCFBeamShear3DGeneric::EvalF2SM(Vector& fadd, double t, int add_damping_terms, const Vector& velocities)
{
	if (!InContinuumMechanicsFormulation())
	{
		double beamGJ = GetBeam3DProperties().BeamGJkx();
		double beamEIy = GetBeam3DProperties().BeamEIy();
		double beamEIz = GetBeam3DProperties().BeamEIz();
		double beamEA = GetBeam3DProperties().BeamEA();
		double beamGAky = GetBeam3DProperties().BeamGAky();
		double beamGAkz = GetBeam3DProperties().BeamGAkz();
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		double beamGAkyz = BeamGAkyz(beamGAky, beamGAkz);   // ACHTUNG: Literaturstudie notwendig - dieser Faktor wird fr den letzten Term bzgl. Dickendeformation verwendet (GAkyz*2E_yz*2deltaE_yz).
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

		double damping_K = GetBeam3DProperties().DampingK();

		ConstVector<27> faddadd(SOS(), 1);
		ConstVector<3> gkp(3, 1);
		Vector3D gk(0.);								//container for Gamma, Kappa, and Strain E=(E_yy,E_zz,E_yz) respectively
		ConstMatrix<27*3> delta_gk(SOS(),Dim());		//container for DeltaGamma, and DeltaKappa, and DeltaStrain respectively
		static Vector wT,xT;										//integration points and weights
		GetIntegrationRule(xT,wT, GetIntegrationOrderBending());

		// add kappa terms (torsion + bending)
		//TMStartTimer(23);  //CPU timing %52%
		for (int i1=1; i1<=xT.GetLen(); i1++)  //i1-loop over all integration points
		{
			double x = xT(i1);		//x in unit configuration [-1 , 1]
			double xi = x*GetLx()*0.5;	//xi in scaled configuration [-lx/2 , lx/2]
			double det = 0.5*GetLx();	//determinant of the jacobian (from element transformation [-1 , 1] -> [-lx/2 , lx/2])


			GetDeltaKappa(xi, delta_gk, gk);

			// Reighley damping (stiffness proportional part)
			if (add_damping_terms && damping_K > 0)
			{
				Mult(velocities, delta_gk, gkp);
				gk(1) += damping_K*gkp(1);
				gk(2) += damping_K*gkp(2);
				gk(3) += damping_K*gkp(3);
			}

			gk(1) *= penalty(1)*beamGJ*wT(i1)*det;	// beamGJ  = torsional stiffness (including torsional correction factor, depending on the shape of the cross section)
			gk(2) *= penalty(2)*beamEIy*wT(i1)*det;	// beamEIy = bending stiffness (in e2=y direction))
			gk(3) *= penalty(3)*beamEIz*wT(i1)*det;	// beamEIz = bending stiffness (in e3=z direction)			
			
			Mult(delta_gk, gk, faddadd);
			fadd += faddadd;
		}

		// add gamma terms (axial and shear strains)
		GetIntegrationRule(xT, wT, GetIntegrationOrderStretchAndShear());
		for (int i1=1; i1<=xT.GetLen(); i1++)  //i1-loop over all integration points
		{
			double x = xT(i1);		//x in unit configuration [-1 , 1]
			double xi = x*GetLx()*0.5;	//xi in scaled configuration [-lx/2 , lx/2]
			double det = 0.5*GetLx();	//determinant of the jacobian (from element transformation [-1 , 1] -> [-lx/2 , lx/2])

			//GetGamma(xi, gk);
			GetDeltaGamma(xi, delta_gk, gk);

			// Reighley damping (stiffness proportional part)
			if (add_damping_terms && damping_K > 0)
			{
				Mult(velocities, delta_gk, gkp);
				gk(1) += damping_K*gkp(1);
				gk(2) += damping_K*gkp(2);
				gk(3) += damping_K*gkp(3);
			}

			gk(1) *= penalty(4)*beamEA*wT(i1)*det;		  // beamEA  = axial stiffness 
			gk(2) *= penalty(5)*beamGAky*wT(i1)*det;		
			gk(3) *= penalty(6)*beamGAkz*wT(i1)*det;		

			Mult(delta_gk, gk, faddadd);
			fadd += faddadd;
		}

		// add thickness terms (E_yy E_zz 2*E_yz)
		//TMStartTimer(25);  //CPU timing %7.5%
		GetIntegrationRuleLobatto(xT, wT, GetIntegrationOrderThickness());
		for (int i1=1; i1<=xT.GetLen(); i1++)  //i1-loop over all integration points
		{
			double x = xT(i1);		//x in unit configuration [-1 , 1]
			double xi = x*GetLx()*0.5;	//xi in scaled configuration [-lx/2 , lx/2]
			double det = 0.5*GetLx();	//determinant of the jacobian (from element transformation [-1 , 1] -> [-lx/2 , lx/2])
			
			GetDeltaThicknessStrain(xi, delta_gk, gk);

			// Reighley damping (stiffness proportional part)
			if (add_damping_terms && damping_K > 0)
			{
				Mult(velocities, delta_gk, gkp);
				gk(1) += damping_K*gkp(1);
				gk(2) += damping_K*gkp(2);
				gk(3) += damping_K*gkp(3);
			}

			gk(1) *= penalty(7)*beamEA*wT(i1)*det;
			gk(2) *= penalty(8)*beamEA*wT(i1)*det;
			gk(3) *= penalty(9)*beamGAkyz*wT(i1)*det;		// ACHTUNG: beamGAkyz: Literaturstudie noch notwendig!!!!!!!
			
			Mult(delta_gk, gk, faddadd);
			fadd += faddadd;
		}
		//TMStopTimer(25);  //CPU timing
	}
}

void ANCFBeamShear3DGeneric::GetIntegrationRuleCM(Vector& x1, Vector& x2, Vector& w1, Vector& w2, TArray<Vector3D>& ip, Vector& ip_weight, int intOrderAxial, int intOrderCrossSection, int ipNumberPerDiscQuadrant) const
{
	if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_rectangular)
	{
		//x_i ... integration points, w_i ... integration weights
		GetIntegrationRule(x1, w1, intOrderAxial);//for dx
		GetIntegrationRule(x2, w2, intOrderCrossSection);//for dy
	}
	else if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_circular || GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_tubular)
	{
		GetIntegrationRuleHollowCylinder(ip, ip_weight, intOrderAxial, intOrderCrossSection, ipNumberPerDiscQuadrant, GetLx(), GetBeam3DProperties().GetRadius(), GetBeam3DProperties().GetInnerRadius());
	}
}

//residual of variational formulation in continuum mechanics formulation
//$ PG 2016-10-7: stiffness proportional part of Reighley damping not yet implemented
void ANCFBeamShear3DGeneric::EvalF2CM(Vector& fadd, double t, int add_damping_terms, const Vector& velocities)
{
	double EA = GetBeam3DProperties().BeamEA();				//Em*area
	double A = EA/Em();					//cross section area
	double GAky = GetBeam3DProperties().BeamGAky();		
	double GAkz = GetBeam3DProperties().BeamGAkz();		
	double G = Em()/(2.+2.*Nu());

	SetComputeCoordinates();//xg(i) = XG(i);

	int poissoncorrection = 0; //1==reduced integration of poisson part
	if (BeamFormulation == 2) poissoncorrection = 1;

	static Vector x1,x2,w1,w2;//KN-todo: ersetzen durch ConstVector

	//Vector3D ks(1.,GAky/(G*A),GAkz/(G*A)); //vector of shear correction factors
	//Matrix3D R = GetRotMatrix(Vector3D(0.,0.,0.), 1).GetTp(); //WARNING:for curved beams, this must be evaluated at every point of integration!
	//Vector3D ksR = R*ks;
	//ksR.X() = fabs(ksR.X());
	//ksR.Y() = fabs(ksR.Y());
	//ksR.Z() = fabs(ksR.Z());
	//UO() << "rot=" << R << "\n";
	//UO() << "ks=" << ks << "\n";
	//UO() << "ksR=" << ksR << "\n";

	TArray<Vector3D> ip(0);
	Vector ip_weight(0);

	for (int kk=1; kk <= 1+poissoncorrection; kk++)
	{
		ConstMatrix<36> Dm;
		Dm.SetSize(6,6);
		Dm.SetAll(0.);
		if (poissoncorrection)
		{
			if (BeamFormulation == 2)//locking compensated
			{
				if (kk == 1)//D^0 (equ. (25))
				{
					//poisson ratio zero:
					Dm(1,1) = Em();
					Dm(2,2) = Em();//reduced thickness stiffness 0.5*
					Dm(3,3) = Em();//reduced thickness stiffness

					//Dm(4,4) = G*5./6.;//1.;//*ksR.X();//statt GAks/A; //gamma_yz
					//Dm(5,5) = G*5./6.;//*ksR.Y();           //gamma_xz
					//Dm(6,6) = G*5./6.;//*ksR.Z();//event. weglassen -> konstant ber Querschnitt //gamma_xy
					//UO() << "CMF (locking-free)" << "\n";
					Dm(4,4) = G;//statt GAks/A; //gamma_yz
					Dm(5,5) = GAky/A;           //gamma_xz
					Dm(6,6) = GAkz/A;//event. weglassen -> konstant ber Querschnitt //gamma_xy

					//x_i, w_i ... integration points and weights in case of rectangular domain
					//ip, ip_weight ... integration points and weights in case of cylindrical domain
					GetIntegrationRuleCM(x1, x2, w1, w2, ip, ip_weight, GetIntegrationOrderCMAxial(), GetIntegrationOrderCMCrossSection(), GetIPNumberPerDiscQuadrant());
				}
				else if (kk == 2)//D^v (equ. (26))
				{
					//$ KN 2011-10-28: help1, help2 mit Paper verglichen -> passt
					double help1 = (Em()* (1-Nu()) / ((1+Nu())*(1-2.*Nu()))) - Em();
					//help1 = (2.*Em()*Nu()*Nu())/((1+Nu())*(1-2.*Nu()))) ... equ.(26)
					Dm(1,1) = help1;
					Dm(2,2) = help1;
					Dm(3,3) = help1;
					
					double help2 = Em()*Nu()/((1+Nu())*(1-2.*Nu()));
					Dm(1,2) = help2;
					Dm(2,1) = help2;
					Dm(1,3) = help2;
					Dm(3,1) = help2;
					Dm(2,3) = help2;
					Dm(3,2) = help2;

					//x_i, w_i ... integration points and weights in case of rectangular domain
					//ip, ip_weight ... integration points and weights in case of cylindrical domain
					GetIntegrationRuleCM(x1, x2, w1, w2, ip, ip_weight, GetIntegrationOrderCMAxial(), 1, 1); //only 1 int. point along cross section
				}
			}
		}
		else if(BeamFormulation == 1)//cont. mech. approach which shows locking phenomenon
		{
			double help3 = Em()* (1-Nu()) / ((1+Nu())*(1-2*Nu()));
			double help4 = Em()*Nu() / ((1+Nu())*(1-2*Nu()));

			Dm(1,1)=help3;
			Dm(1,2)=help4;
			Dm(1,3)=help4;

			Dm(2,1)=help4;
			Dm(2,2)=help3;
			Dm(2,3)=help4;

			Dm(3,1)=help4;
			Dm(3,2)=help4;
			Dm(3,3)=help3;

			Dm(4,4) = G;//statt GAks/A; //gamma_yz
			Dm(5,5) = GAky/A;           //gamma_xz
			Dm(6,6) = GAkz/A;//event. weglassen -> konstant ber Querschnitt //gamma_xy
			/*UO() << "false!\n";*/
			//Dm(4,4)=G*5./6.;
			//Dm(5,5)=G*5./6.;
			//Dm(6,6)=G*5./6.;
			//Dm(4,4)=10*Em();//G*5./6.;
			//Dm(5,5)=10*Em();//G*5./6.;
			//Dm(6,6)=10*Em();//G*5./6.;
			//UO() << "D=" << Dm << "\n";

			//x_i, w_i ... integration points and weights in case of rectangular domain
			//ip, ip_weight ... integration points and weights in case of cylindrical domain
			GetIntegrationRuleCM(x1, x2, w1, w2, ip, ip_weight, GetIntegrationOrderCMAxial(), GetIntegrationOrderCMCrossSection(), GetIPNumberPerDiscQuadrant());
		}
		else if(BeamFormulation == 3)// only D^0 (equ. (25))
		{
			Dm(1,1) = Em();
			Dm(2,2) = Em();
			Dm(3,3) = Em();
			//Dm(4,4) = G*ksR.X();//statt GAks/A; //gamma_yz
			//Dm(5,5) = G*ksR.Y();           //gamma_xz
			//Dm(6,6) = G*ksR.Z();//event. weglassen -> konstant ber Querschnitt //gamma_xy
			/*UO() << "false!\n";*/
			Dm(4,4) = G*5./6.;//statt GAks/A;
			Dm(5,5) = G*5./6.;
			Dm(6,6) = G*5./6.;

			//x_i, w_i ... integration points and weights in case of rectangular domain
			//ip, ip_weight ... integration points and weights in case of cylindrical domain
			GetIntegrationRuleCM(x1, x2, w1, w2, ip, ip_weight, GetIntegrationOrderCMAxial(), GetIntegrationOrderCMCrossSection(), GetIPNumberPerDiscQuadrant());
		}
		//mbs->UO()<<"Dm="<<Dm<<"\n";

		if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_rectangular)
		{
			for (int i1=1; i1<=x1.GetLen(); i1++)
			{
				for (int i2=1; i2<=x2.GetLen(); i2++)
				{
					for ( int i3=1; i3<=x2.GetLen(); i3++)
					{
						Vector3D p(x1(i1)*0.5*GetLx(), x2(i2)*0.5*GetLy(), x2(i3)*0.5*GetLz());
						double p_weight = 0.5*GetLx()*0.5*GetLy()*0.5*GetLz() * w1(i1)*w2(i2)*w2(i3);
						EvalF2CMAtIntegrationPoint(fadd, Dm, p, p_weight);
					}
				}
			}
		}
		else if (GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_circular || GetBeam3DProperties().GetCrossSectionShape() == Beam3DProperties::TCSS_tubular)
		{
			for (int ip_idx=1; ip_idx<=ip.Size(); ip_idx++)
			{
				EvalF2CMAtIntegrationPoint(fadd, Dm, ip(ip_idx), ip_weight(ip_idx));
			}
		}
		else
		{
			mbs->UO(UO_LVL_err) << "ERROR in ANCFBeamShear3DGeneric: assembling of strain energy not implemented in case of GetBeam3DProperties().GetCrossSectionShape() == " << GetBeam3DProperties().GetCrossSectionShape() << "!\n";
			return;
		}
	}
}

void ANCFBeamShear3DGeneric::EvalF2CMAtIntegrationPoint(Vector& fadd, const Matrix& Dm, const Vector3D& p, double p_weight) const
{
	int dim = Dim(); //=3
	int ns = NS();   //=9
	int sos = SOS(); //=9*3

	Matrix3D strain, piola1, F;
	strain.SetAll(0);
	piola1.SetAll(0);

	Matrix3D jac;
	GetJacobi(jac,p,q0);//p(1) in [-lx/2,lx/2]
	double jacdet = jac.Det();
	//mbs->UO()<<"jacdet = "<<jacdet<<"\n";
	//mbs->UO()<<"0.5*lx*0.5*ly*0.5*lz = "<<0.5*lx*0.5*ly*0.5*lz<<"\n";

	Matrix3D jacinv;
	jac.GetInverse(jacinv);
	jacinv = jacinv.GetTp();
	//mbs->UO() << "jacinv = " << jacinv << "\n";

	static Matrix grad0, grad;//am Einheitselement
	grad0.SetSize(dim, ns);
	for(int i=1; i<=ns; i++)
	{
		grad0(1,i)=GetSFx(i,p);//3x9-matrix
		grad0(2,i)=GetSFy(i,p);
		grad0(3,i)=GetSFz(i,p);
	}

	Mult(jacinv, grad0, grad);//grad0*jacinv
	F.SetAll(0);//muss in jedem Integrationsschritt zurck auf 0 gesetzt werden!!!!
	for (int j = 1; j <= dim; j++) 
	{
		for (int i = 1; i <= ns; i++)
		{ 
			int l = (i-1)*dim+j;
			F(j,1) += grad(1,i)*xg(l);//3x9-matrix*9-vector
			F(j,2) += grad(2,i)*xg(l);
			F(j,3) += grad(3,i)*xg(l);
		}
	}
	//F = F.GetTp();//FTranspose?
	if(calclinear == 0)
	{
		F(1,1) += 1;//Positionsgrad (nicht Grad(u), sondern Grad(r))
		F(2,2) += 1;
		F(3,3) += 1;
		//E = 1/2*(F^T F-I)
		strain = 0.5*(F.GetTp()*F);//eps=0.5*(F.GetTp()*F-Id)
		strain(1,1) -= 0.5;
		strain(2,2) -= 0.5;
		strain(3,3) -= 0.5;
	}
	else if(calclinear == 1)
	{
		strain = 0.5*(F + F.GetTp());
	}

	ConstVector<6> s3(strain(1,1), strain(2,2), strain(3,3), 2.*strain(2,3), 2.*strain(1,3), 2.*strain(1,2));
	ConstVector<6> stress;
	stress = Dm * s3;//Dm=6x6-Matrix, //S = D : E

	//generate 2. Piola-Kirchhoff-tensor, 2.PK = symm. tensor 
	piola1(1,1)=stress(1);
	piola1(2,2)=stress(2);
	piola1(3,3)=stress(3);

	piola1(2,3)=stress(4);
	piola1(3,2)=stress(4);
	piola1(1,3)=stress(5);
	piola1(3,1)=stress(5);
	piola1(1,2)=stress(6);
	piola1(2,1)=stress(6);

	if(calclinear == 0)
	{
		piola1 = F * piola1;//P = F * S, 1.PK = F * 2.PK
	}

	ConstVector<27> temp;//KN-todo: 27 ersetzen durch ConstInt MaxDOF ...
	temp.SetLen(sos);
	temp.SetAll(0);

	//grad = grad.GetTp();//gradTranspose?
	for (int j=1; j <= dim; j++)
	{
		for (int i = 0; i < ns; i++)
		{
			temp(dim*i+j) = grad(1,i+1)*piola1(j,1) + grad(2,i+1)*piola1(j,2) + grad(3,i+1)*piola1(j,3);
		}
	}
	//integrate(P : dF/dq) 
	fadd.MultAdd(fabs(jacdet)*p_weight, temp);		
}

//----------------------------------------------------------------------------
//Gamma & Kappa & ThicknessStrain  +  Derivatives of those with respect to dof
//----------------------------------------------------------------------------
void ANCFBeamShear3DGeneric::GetGamma(const double& xi, Vector3D& gamma, int flagD)		// -Lx/2 <= xi <= Lx/2
{
	//gamma1 = t1*drdx/|e1| - 1
	//gamma2 = t2*drdxi
	//gamma3 = t3*drdxi
	//
	//gamma = (gamma1, gamma2, gamma3)
	Vector3D p0(xi, 0., 0.);     //xi in [-Lx/2, Lx/2]

	Vector3D drdx, drdy, drdz, e1, e2, e3;
	drdx = GetPosx3D(p0, flagD);
	drdy = GetPosy3D(p0, flagD);
	drdz = GetPosz3D(p0, flagD);
	e1 = Getei(1, 
		drdy.X(), drdy.Y(), drdy.Z(), 
		drdz.X(), drdz.Y(), drdz.Z());
	e2 = Getei(2,
		drdy.X(), drdy.Y(), drdy.Z(),
		drdz.X(), drdz.Y(), drdz.Z());
	e3 = Getei(3,
		drdy.X(), drdy.Y(), drdy.Z(),
		drdz.X(), drdz.Y(), drdz.Z());
	e1.Normalize();
	e2.Normalize();
	e3.Normalize();

	gamma.Set(
		e1*drdx - 1.,
		e2*drdx,
		e3*drdx);
}

void ANCFBeamShear3DGeneric::GetDeltaGamma(const double& xi, Matrix& dgammadq, Vector3D& gamma)		// -Lx/2 <= xi <= Lx/2
{
	//dgamma1dq = ( ... )   see paper
	//dgamma2dq = ( r,y.S,x + r,x.S,y - (r,x.r,y)/(r,y.r,y).r,y.S,y ) / (norm(r,y))
	//dgamma3dq = ( r,z.S,x + r,x.S,z - (r,x.r,z)/(r,z.r,z).r,z.S,z ) / (norm(r,z))
	//
	//dgammadq = (dgamma1dq, dgamma2dq, dgamma3dq)

	Vector3D p0(xi, 0., 0.);     //xi in [-Lx/2, Lx/2]
	Vector sfx(NS(), 1); GetShapesx(sfx, p0);
	Vector sfy(NS(), 1); GetShapesy(sfy, p0);
	Vector sfz(NS(), 1); GetShapesz(sfz, p0);

	dgammadq.SetSize(SOS(), Dim());
	dgammadq.SetAll(0.);

	Vector3D drdx, drdy, drdz;
	drdx = GetPosx3D(p0);
	drdy = GetPosy3D(p0);
	drdz = GetPosz3D(p0);

	// help variables
	double norm_drdy_squ = drdy.Norm2();
	double norm_drdy_lin = sqrt(norm_drdy_squ);
	double norm_drdy_cub = norm_drdy_lin * norm_drdy_squ;
	double norm_drdz_squ = drdz.Norm2();
	double norm_drdz_lin = sqrt(norm_drdz_squ);
	double norm_drdz_cub = norm_drdz_lin * norm_drdz_squ;
	
	for (int i=1; i<=Dim(); i++)
	{
		Vector3D ei = Getei(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z());
		TArray<Vector3D> deidp(GetdeidDiffVar(i, drdy.X(), drdy.Y(), drdy.Z(), drdz.X(), drdz.Y(), drdz.Z()));
		double norm_ei_squ = ei.Norm2();
		double norm_ei_lin = sqrt(norm_ei_squ);
		double norm_ei_cub = norm_ei_lin * norm_ei_squ;

		gamma(i) = (ei*drdx)/norm_ei_lin;
		Vector3D v = (1./norm_ei_lin)*drdx - (gamma(i)/norm_ei_squ)*ei;
		//double facy = v * (de1dp(1) + de1dp(2) + de1dp(3));
		//double facz = v * (de1dp(4) + de1dp(5) + de1dp(6));

		//assembling of dgamma1dq
		for (int j=1; j<=NS(); j++)
		{
			for (int k=1; k<=Dim(); k++)
			{
				double rowidx = (j-1)*Dim() + k;
				dgammadq(rowidx, i) = ((1./norm_ei_lin)*ei(k))*sfx(j);
				dgammadq(rowidx, i) += (v*deidp(k))*sfy(j);
				dgammadq(rowidx, i) += (v*deidp(k+Dim()))*sfz(j);			
			}
		}
	}
	gamma(1) -= 1;
}


void ANCFBeamShear3DGeneric::GetKappa(const double& xi, Vector3D& kappa, int flagD)		// -Lx/2 <= xi <= Lx/2
{
	Vector3D p0 = Vector3D(xi, 0., 0.);
	kappa = GetTwistAndCurvature(p0, 0, flagD) - GetTwistAndCurvature(p0, 1, flagD);
}


void ANCFBeamShear3DGeneric::GetDeltaKappa(const double& xi, Matrix& dkappadq, Vector3D& kappa)   //derivative of vector Kappa wrt vector q at xi (which is between -Lx/2 and Lx/2)
{
	//TMStartTimer(28);
	Vector3D p0 = Vector3D(xi, 0, 0);

	Vector sfy(NS(), 1);  GetShapesy(sfy, p0);
	Vector sfz(NS(), 1);  GetShapesz(sfz, p0);
	Vector sfxy(NS(), 1); GetShapesxy(sfxy, p0);
	Vector sfxz(NS(), 1); GetShapesxz(sfxz, p0);


	Vector3D drdy = GetPosy3D(p0);
	Vector3D drdz = GetPosz3D(p0);
	Vector3D ddrdxdy = GetPosxy3D(p0);
	Vector3D ddrdxdz = GetPosxz3D(p0);

	// e1 = drdy x drdz, e3 = drdz, e2 = e3 x e1 = -e1 x e3
	Vector3D ei;
	Vector3D deidxi;
	TArray<Vector3D> deidp(6);
	TArray<Vector3D> ddeidxidp(12);

	// ei     depends on p = (dr1dy dr2dy dr3dy dr1dz dr2dz dr3dz),
	// deidxi depends on p = (dr1dy dr2dy dr3dy dr1dz dr2dz dr3dz ddr1dxdy ddr2dxdy ddr3dxdy ddr1dxdz ddr2dxdz ddr3dxdz)
	// delta kappa = delta k - delta(Rot*k_0)

	// calculation of delta k:  in the way  dkdq = dkdp * dpdq

	TArray<Vector3D> dkdp(12);    // Vector3D() sets components to 0!
	for (int i=1; i<=Dim(); i++) 
	{
		ei = Getei(i, 
			drdy.X(), drdy.Y(), drdy.Z(), 
			drdz.X(), drdz.Y(), drdz.Z());
		deidxi = Getdeidxi(i, 
			drdy.X(), drdy.Y(), drdy.Z(), 
			drdz.X(), drdz.Y(), drdz.Z(), 
			ddrdxdy.X(), ddrdxdy.Y(), ddrdxdy.Z(), 
			ddrdxdz.X(), ddrdxdz.Y(), ddrdxdz.Z());
		deidp = GetdeidDiffVar(i, 
			drdy.X(), drdy.Y(), drdy.Z(), 
			drdz.X(), drdz.Y(), drdz.Z());
		ddeidxidp = GetddeidxidDiffVar(i, 
			drdy.X(), drdy.Y(), drdy.Z(), 
			drdz.X(), drdz.Y(), drdz.Z(), 
			ddrdxdy.X(), ddrdxdy.Y(), ddrdxdy.Z(), 
			ddrdxdz.X(), ddrdxdz.Y(), ddrdxdz.Z());

		double einormsquare = ei*ei;

		for (int j=1; j<=12; j++)
		{
			if (j<=6) 
			{
				dkdp(j) -= ((ei*deidp(j)) / Sqr(einormsquare)) * ei.Cross(deidxi);
				dkdp(j) += (0.5/einormsquare) * deidp(j).Cross(deidxi);
			}

			dkdp(j) += (0.5/einormsquare) * ei.Cross(ddeidxidp(j));
		}
	}

	// delta k = [ d k / d p ] . [ d p / dq ]
	dkappadq.SetAll(0.);
	Vector3D rowvec;

	for(int k=1; k<=NS(); k++)
	{
		for(int j=1; j<=Dim(); j++) 
		{
			rowvec =
				dkdp(j  )*sfy(k) + 
				dkdp(j+3)*sfz(k) + 
				dkdp(j+6)*sfxy(k) + 
				dkdp(j+9)*sfxz(k); 
			int rowidx = j + (k-1)*Dim();

			dkappadq(rowidx, 1) = rowvec.X();
			dkappadq(rowidx, 2) = rowvec.Y();
			dkappadq(rowidx, 3) = rowvec.Z();
		}
	}

	// [kappa]_e = A^T (k - k_0)              [..] be the representation in the local frame basis
	// d [kappa]_e / d q = d A^T /dq (k - k_0) + A^T dk/dq

	Matrix helper = dkappadq;
	Matrix3D A = GetRotMatrix(p0);
	Mult(helper, A, dkappadq); //dkappadq = helper * A;

	//Matrix3D AT = A;//.GetTp();
	//for(int i=1; i<=NS()*Dim(); i++)
	//{
	//	Vector3D v;
	//	v(1)=helper(i,1);
	//	v(2)=helper(i,2);
	//	v(3)=helper(i,3);
	//	v = AT*v;
	//	dkappadq(i,1) = v(1);
	//	dkappadq(i,2) = v(2);
	//	dkappadq(i,3) = v(3);
	//}

//TMStopTimer(29);

//TMStartTimer(30);
	kappa = GetTwistAndCurvature(p0, 0) - GetTwistAndCurvature(p0, 1);

//TMStopTimer(30);
//TMStartTimer(20);
	GetdRotvdqT(A*kappa, p0, helper, 1);
//TMStopTimer(20);
	dkappadq += helper;

	//kappa = A.GetTp()*kappa;
}


void ANCFBeamShear3DGeneric::GetThicknessStrain(const double& xi, Vector3D& thicknessstrain, int flagD)		// -Lx/2 <= xi <= Lx/2
{
	//thicknessstrain_yy = 1/2 * (drdy*drdy - 1)
	//thicknessstrain_zz = 1/2 * (drdz*drdz - 1)
	//thicknessstrain_yz_twice = 2 * thicknessstrain_yz = drdy*drdz
	//
	//thicknessstrain = (thicknessstrain_yy, thicknessstrain_zz, thicknessstrain_yz_twice)

	Vector3D p0(xi, 0., 0.);     //xi in [-Lx/2, Lx/2]

	Vector3D drdy = GetPosy3D(p0, flagD);
	Vector3D drdz = GetPosz3D(p0, flagD);

	thicknessstrain.Set(
		(drdy*drdy - 1.)/2.,
		(drdz*drdz - 1.)/2.,
		drdy*drdz );
}

void ANCFBeamShear3DGeneric::GetDeltaThicknessStrain(const double& xi, Matrix& dthicknessstraindq, Vector3D& thicknessstrain)		// -Lx/2 <= xi <= Lx/2
{
	//dthicknessstrainyydq = r,y.S,y
	//dthicknessstrainzzdq = r,z.S,z
	//dthicknessstrainyztwicedq = r,y.S,z + r,z.S,y
	//
	//dthicknessstraindq = (dthicknessstrainyydq, dthicknessstrainzzdq, dthicknessstrainyztwicedq)

	Vector3D p0(xi, 0., 0.);     //xi in [-Lx/2, Lx/2]
	Vector sfy(NS(), 1); GetShapesy(sfy, p0);
	Vector sfz(NS(), 1); GetShapesz(sfz, p0);

	dthicknessstraindq.SetSize(SOS(), Dim());
	dthicknessstraindq.SetAll(0.);

	Vector3D drdy = GetPosy3D(p0);
	Vector3D drdz = GetPosz3D(p0);

	//set thickness strain
	thicknessstrain.Set(
		(drdy*drdy - 1.)/2.,
		(drdz*drdz - 1.)/2.,
		drdy*drdz );

	//assembling of dthicknessstraindq
	for (int i=1; i<=NS(); i++)
	{
		for (int j=1; j<=Dim(); j++)
		{
			double rowidx = (i-1)*Dim() + j;
			dthicknessstraindq(rowidx, 1) += drdy(j)*sfy(i);
			dthicknessstraindq(rowidx, 2) += drdz(j)*sfz(i);
			dthicknessstraindq(rowidx, 3) += drdy(j)*sfz(i) + drdz(j)*sfy(i);
		}
	}
}


//for plasticity:
double ANCFBeamShear3DGeneric::PostNewtonStep(double t)		// changes plastic strains, returns yieldfunction(sigma)/Em
{
	return 0;
}


//---------------------------------------------------------------
//for Displacement/Stress/Strain/Stress Resultants-Visualization:
//---------------------------------------------------------------
double ANCFBeamShear3DGeneric::GetFieldVariableValue(const FieldVariableDescriptor & fvd, const Vector3D & local_position, bool flagD)
{		

	//ploc_scaled = local position: -l/2 <= ploc_scaled <= +l/2
	Vector3D p0 = local_position;//-l/2 <= p0 <= +l/2

	xgd.SetLen(SOS());
	GetDrawCoordinates(xgd);  //xgd=XGD

	//Position:
	if (fvd.VariableType() == FieldVariableDescriptor::FVT_position)
	{
		Vector3D u;
		if(flagD) 
			u = GetPosD(p0);
		else
			u = GetPos(p0);

		return fvd.GetComponent(u);
	}

	//Displacement:
	if (fvd.VariableType() == FieldVariableDescriptor::FVT_displacement)
	{
		Vector3D u;
		if(flagD) 
			u = GetDisplacementD(p0);
		else
			u = GetDisplacement(p0);

		return fvd.GetComponent(u);
	}

	//Velocity:
	if (fvd.VariableType() == FieldVariableDescriptor::FVT_velocity)
	{
		Vector3D u;
		u = GetVel3D(p0, flagD);

		return fvd.GetComponent(u);
	}

	//angular velocity:
	if (fvd.VariableType() == FieldVariableDescriptor::FVT_angular_velocity)
	{
		Vector3D u;
		u = GetAngularVel(p0, flagD);

		return fvd.GetComponent(u);
	}


	//JG2014-03-07
	//rotations:
	if (fvd.VariableType() == FieldVariableDescriptor::FVT_bryant_angle)
	{
		//same as in rigid3d.cpp
		Vector3D phi;
		RotMatToKardanAngles(GetRotMatrix(p0, 0, flagD), phi);
		return fvd.GetComponent(phi);
	}

	if (fvd.VariableType() == FieldVariableDescriptor::FVT_rotations_312)
	{
		Vector3D phi;
		RotMatTo312Angles(GetRotMatrix(p0, 0, flagD), phi);
		return fvd.GetComponent(phi);
	}

	//Stress and Strain:
	if (InContinuumMechanicsFormulation())
	{
		Matrix3D strain(0), F, piola1;
		ConstMatrix<36> Dm;
		Dm.SetSize(6,6);
		Dm.SetAll(0);
		Vector3D rx, ry, rz;
		rx=GetPosx3D(p0,flagD);
		ry=GetPosy3D(p0,flagD);
		rz=GetPosz3D(p0,flagD);

		//strain:
		F.SetAll(0); //$JG: WRONG: this is the deformation gradient in local coordinates
		for (int j = 1; j <= Dim(); j++) 
		{
			F(j,1) = rx(j);//grad*xg
			F(j,2) = ry(j);
			F(j,3) = rz(j);
		}
		if(calclinear == 0)
		{
			//F(1,1) += 1;//Positionsgrad (nicht Grad(u), sondern Grad(r))//ux->rx: +1, hier haben wir schon rx, kein +1
			//F(2,2) += 1;
			//F(3,3) += 1;
			//E = 1/2*(F^T F-I)
			strain = 0.5*(F.GetTp()*F);//eps=0.5*(F.GetTp()*F-Id)
			strain(1,1) -= 0.5;
			strain(2,2) -= 0.5;
			strain(3,3) -= 0.5;
		}
		else if(calclinear == 1)
		{
			F(1,1) -= 1;//rx->ux: -1
			F(2,2) -= 1;
			F(3,3) -= 1;
			strain = 0.5*(F + F.GetTp());
		}
		//stress:
		double help3 = Em()* (1-Nu()) / ((1+Nu())*(1-2*Nu()));
		double help4 = Em()*Nu() / ((1+Nu())*(1-2*Nu()));
		double GAky = GetBeam3DProperties().BeamGAky();		
		double GAkz = GetBeam3DProperties().BeamGAkz();		
		double G = Em()/(2.+2.*Nu());
		double A = GetBeam3DProperties().GetCrossSectionArea();

		Dm(1,1)=help3;
		Dm(1,2)=help4;
		Dm(1,3)=help4;

		Dm(2,1)=help4;
		Dm(2,2)=help3;
		Dm(2,3)=help4;

		Dm(3,1)=help4;
		Dm(3,2)=help4;
		Dm(3,3)=help3;

		Dm(4,4)=G;
		Dm(5,5)=G;
		Dm(6,6)=G;

		ConstVector<6> s3(strain(1,1), strain(2,2), strain(3,3), 2.*strain(2,3), 2.*strain(1,3), 2.*strain(1,2));
		ConstVector<6> stress;
		stress = Dm * s3;//Dm=6x6-Matrix, //S = D : E

		//generate 2. Piola-Kirchhoff-tensor, 2.PK = symm. tensor 
		piola1(1,1)=stress(1);
		piola1(2,2)=stress(2);
		piola1(3,3)=stress(3);

		piola1(2,3)=stress(4);
		piola1(3,2)=stress(4);
		piola1(1,3)=stress(5);
		piola1(3,1)=stress(5);
		piola1(1,2)=stress(6);
		piola1(2,1)=stress(6);

		switch(fvd.VariableType())
		{
		case FieldVariableDescriptor::FVT_stress_mises:							return piola1.Mises();
		case FieldVariableDescriptor::FVT_stress:										return fvd.GetComponent(piola1);
		case FieldVariableDescriptor::FVT_total_strain:							return fvd.GetComponent(strain);
		}
	}
	else		// SM-Formulation
	{
		//$ EK 2015-12-03
		if (fvd.VariableType() == FieldVariableDescriptor::FVT_stress)
		{
			return GetStressSMF(local_position, fvd, flagD);
		}
		if (fvd.VariableType() == FieldVariableDescriptor::FVT_stress_mises)
		{
			FieldVariableDescriptor fvd_stress(FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_magnitude);
			return GetStressSMF(local_position, fvd_stress, flagD);
		}
		if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_force_axial || fvd.VariableType() == FieldVariableDescriptor::FVT_beam_force_transversal)
		{
			Vector3D gamma;
			GetGamma(local_position.X(), gamma, flagD);
			gamma.X() *= GetBeam3DProperties().BeamEA();
			gamma.Y() *= GetBeam3DProperties().BeamGAky();
			gamma.Z() *= GetBeam3DProperties().BeamGAkz();
			if (fvd.VariableType() == FieldVariableDescriptor::FVT_beam_force_axial)
				return gamma.X();
			return fvd.GetComponent(gamma);
		}
		//$ PG 2014-5-14:]

		double xi = p0.X();
		Vector3D gamma; GetGamma(xi, gamma, flagD);
		Vector3D kappa; GetKappa(xi, kappa, flagD);
		Vector3D thicknessstrain; GetThicknessStrain(xi, thicknessstrain, flagD);


		int use_interpolated_values = 0;
		//compute interpolated values:
		if (use_interpolated_values)
		{
			ConstVector<10> wT;										//integration points and weights
			ConstVector<10> xT;										//integration points and weights
			GetIntegrationRule(xT, wT, 3);

			double x1 = GetLx()*0.5*xT(1);
			double x2 = GetLx()*0.5*xT(2);
			Vector3D gamma_1; GetGamma(x1, gamma_1, flagD);
			//		Vector3D kappa_1; GetKappa(x1, kappa_1, flagD);
			Vector3D gamma_2; GetGamma(x2, gamma_2, flagD);
			//		Vector3D kappa_2; GetKappa(x2, kappa_2, flagD);

			gamma = 1./(x2-x1)*((x2-xi)*gamma_1+(xi-x1)*gamma_2); //linear interpolation of all gamma values
		}
		
		double beamEA = GetBeam3DProperties().BeamEA();
		double beamGAky = GetBeam3DProperties().BeamGAky();
		double beamGAkz = GetBeam3DProperties().BeamGAkz();
		double beamGAkyz = BeamGAkyz(beamGAky, beamGAkz);      
		double beamGJkx = GetBeam3DProperties().BeamGJkx();
		double beamEIy = GetBeam3DProperties().BeamEIy();
		double beamEIz = GetBeam3DProperties().BeamEIz();

		Vector3D moment(kappa);
		moment.X() *= beamGJkx;
		moment.Y() *= beamEIy;
		moment.Z() *= beamEIz;

		Vector3D beam_force(gamma);
		beam_force.X() *= beamEA;
		beam_force.Y() *= beamGAky;
		beam_force.Z() *= beamGAkz;

		/*
		//strains and stresses are computed in WRONG WAY JG2013-09 !!!
		//the part of bending strain has been forgotten!!!

		// for sm-formulation the total strain contains:
		//  Gamma_1 Gamma_2 Gamma_3
		//  Gamma_2  E_yy    E_yz
		//  Gamma_3  E_yz    E_zz
		// ... same for the stress

		Matrix3D strains(
			gamma.X(), gamma.Y(), gamma.Z(),
			gamma.Y(), thicknessstrain.X(), 0.5*thicknessstrain.Z(),
			gamma.Z(), 0.5*thicknessstrain.Z(), thicknessstrain.Y()
			);

		Matrix3D stresses(
			beamEA*gamma.X(), beamGAky*gamma.Y(), beamGAkz*gamma.Z(),
			beamGAky*gamma.Y(), beamEA*thicknessstrain.X(), beamGAkyz*0.5*thicknessstrain.Z(),
			beamGAkz*gamma.Z(), beamGAkyz*0.5*thicknessstrain.Z(), beamEA*thicknessstrain.Y()
			);
		*/
		
		switch(fvd.VariableType())
		{
			//beam_curvature is defined as 2-component vector! JG2013-09 
			//case FieldVariableDescriptor::FVT_beam_curvature: return fvd.GetComponent(Vector3D(kappa.X(), kappa.Y(), kappa.Z()));

			//strains and stresses are computed in WRONG WAY JG2013-09 !!!
			//case FieldVariableDescriptor::FVT_total_strain:	  return fvd.GetComponent(strains);
			//case FieldVariableDescriptor::FVT_stress:         return fvd.GetComponent(stresses);

			case FieldVariableDescriptor::FVT_beam_torsion: return kappa.X();
			case FieldVariableDescriptor::FVT_beam_curvature: return fvd.GetComponent(kappa);

			case FieldVariableDescriptor::FVT_beam_moment_torsional: return moment.X();
			case FieldVariableDescriptor::FVT_beam_moment_bending: return fvd.GetComponent(moment);

			case FieldVariableDescriptor::FVT_beam_axial_extension: return gamma.X();
			case FieldVariableDescriptor::FVT_beam_shear: return fvd.GetComponent(gamma);

			case FieldVariableDescriptor::FVT_beam_force_axial: return beam_force.X();
			case FieldVariableDescriptor::FVT_beam_force_transversal: return fvd.GetComponent(beam_force);
		}
	}

	return FIELD_VARIABLE_NO_VALUE;
}


//$ EK 2015-10: compute stress components and mises stress, respectively
//this procedure is for teh structural mechanics formulation (SMF) - the continuum mechanics based formulation is
//considered in the GetFieldVariable-routine directly
double ANCFBeamShear3DGeneric::GetStressSMF(const Vector3D& local_position, const FieldVariableDescriptor& fvd, bool flagD)
{
	const Beam3DProperties& mat = GetBeam3DProperties();

	double EA = mat.BeamEA();
	double EIz = mat.BeamEIz();
	double EIy = mat.BeamEIy();
	double GJkx = mat.BeamGJkx();
	double b = mat.GetBeamThicknessY();
	double h = mat.GetBeamThicknessZ();
	double r = mat.GetRadius();
	double ri = mat.GetInnerRadius();

	double x0 = local_position.X();
	double y = local_position.Y();
	double z = local_position.Z();

	Vector3D gamma(0.), kappa(0.);
	GetGamma(x0, gamma, flagD);
	GetKappa(x0, kappa, flagD);

	double Qz = mat.BeamGAkz() * gamma.Z();
	double Qy = mat.BeamGAky() * gamma.Y();
	
	double Mx = mat.BeamGJkx() * kappa.X();
	double My = mat.BeamEIy() * kappa.Y();
	double Mz = mat.BeamEIz() * kappa.Z();

	double A, Iy = 0., Iz = 0., J = 0.;
	A = mat.GetCrossSectionArea();
	
	//calculate the second moments of area of the cross section
	switch (mat.GetCrossSectionShape())
	{
	case BeamCrossSectionInfo::TCSS_rectangular:
		{
			Iy = A * Sqr(h) / 12.;
			Iz = A * Sqr(b) / 12.;
			// the old-german Tg in Szabo's book is the tanh (which could be verified by integrating and tanh = sinh/cosh)
			J = (1. - 192./pow(MY_PI, 5.) * b/h * tanh(MY_PI/2. * h/b)) / 3. * Cub(b) * h;	// cf. approximate relation (14.55) in I. Szabo - Hoehere Technische Mechanik
			break;
		}
	case BeamCrossSectionInfo::TCSS_circular:
		{
			Iy = Iz = A / 4. * Sqr(r);
			J = Iy + Iz;	// torsional area moment equals polar moment of inertia
			break;
		}
	case BeamCrossSectionInfo::TCSS_tubular:
		{
			Iy = Iz = A / 4. * (Sqr(r) + Sqr(ri));
			J = Iy + Iz;	// torsional area moment equals polar moment of inertia
			break;
		}
	default:
		{
			assert(0);
			return 0.;
		}
	}
	
	// sigma_xx - identical for all cross sections
	if (fvd.ComponentIndex1() == FieldVariableDescriptor::FVCI_x && fvd.ComponentIndex2() == FieldVariableDescriptor::FVCI_x)
	{
		double N =  mat.BeamEA() * gamma.X();

		return N / A + My / Iy * local_position.Z() + Mz / Iz * local_position.Y();
	}
	// sigma_xy - different on the cross sections - see 
	else if (fvd.ComponentIndex1() == FieldVariableDescriptor::FVCI_x && fvd.ComponentIndex2() == FieldVariableDescriptor::FVCI_y)
	{
		switch (mat.GetCrossSectionShape())
		{
		case BeamCrossSectionInfo::TCSS_rectangular:
			{
				return GetStressSMFxyRectangle(local_position, Iz, J, Qy, Mx);
				//double Sy = -h / 2. * (Sqr(b / 2.) - Sqr(y));
				//double tau_xy_1 = -Qy * Sy / (Iz * h);

				//// approximate (Ritz) solution according to (14.47) I. Szabo - Hoehere Technische Mechanik
				////for torsional influences
				//double h1 = h/2.;
				//double b1 = b/2.;
				//
				////compared to Szabo - den is scaled by h^6
				////in total "Szabo's c1" = 1/h^2 c1, "Szabo's c2" = 1/h^4 c2, "Szabo's c3" = 1/h^4 c3
				////in total the values are correct since the y^2 and z^2 in the caculations of tau_xy_2
				////are also scaled by h^2 and therefore the factors cancel 
				//double den = (45.*(1.+pow(b1/h1,6.)) + 509.*Sqr(b1/h1)*(1.+Sqr(b1/h1)));
				//double c1 = -35./16. *  (9.*(1+pow(b1/h1,4.)) + 130.*Sqr(b1/h1)) / den;
				//double c2 = -105./16. * (1. + 9.*Sqr(b1/h1)) / den;
				//double c3 = -105./16. * (9. + Sqr(b1/h1)) / den;

				//double tau_xy_2 = -4. * Mx / J * z * (Sqr(y/h1)-Sqr(b1/h1)) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c3*(Sqr(z/h1) - 1.));
				//
				//return tau_xy_1 + tau_xy_2;
			}
		case BeamCrossSectionInfo::TCSS_circular:
			{
				return GetStressSMFxyCircle(local_position, Iy, Iz, J, Qy, Qz, Mx);
				//double sin_beta_1 = sqrt(1.-Sqr(z/r));
				////P_z is the intersection point of the shear stress vectors (- z)
				//double P_z = (abs(z) > 1e-7) ? Sqr(r) / z - z : 1e20; 
				////double tan_gamma = y / P_z;
				//double tan_gamma = (abs(P_z) < 1e-7) ? 0 : y / P_z;
				////tau_xz contributes to tau_xy with tan_gamma *tau_xz
				//double tau_xz = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r);
				//double tau_xy_1 = tau_xz * tan_gamma;	

				////shear stress due to transversal force - see Hagedorn (Schubspannungen infolge Querkraft bei Biegung)
				//double sin_beta_2 = sqrt(1.-Sqr(y/r));
				//double tau_xy_2 = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r);

				////alt von AH double tau_r = Mx / J * sqrt(y*y + z*z);
				////alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r;
				////Szabo - Hoehere Technische Mechanik (14.29)
				//double tau_xy_3 = -Mx/J * z;

				//return tau_xy_1 + tau_xy_2 + tau_xy_3;
			}
		case BeamCrossSectionInfo::TCSS_tubular:
			{
				//return GetStressSMFxyCircle(local_position, Iy, Iz, J, Qy, Qz, Mx);
				return GetStressSMFxyTube(local_position, Iy, Iz, J, Qy, Qz, Mx);

				//double y2 = Sqr(y);
				//double z2 = Sqr(z);
				//double tmp = Sqr(r) - y2;
				//double tmpi = Sqr(ri) - min(y2,Sqr(ri)) 
				//double Sy = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
				//double height = 2*(Sqrt(tmp) - Sqrt(tmpi));
				//tmp = Sqr(r) - z2;
				//tmpi = Sqr(ri) - min(z2,Sqr(ri)) 
				//double Sz = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
				//double width = 2*(Sqrt(tmp) - Sqrt(tmpi));

				//double tau_xy_1 = - Qy*Sy/Iz / height;
				//double tau_xz = - Qz*Sz/Iy / width;

				////P_z is the intersection point of the shear stress vectors (- z)
				//double P_z = (abs(z) > 1e-7) ? Sqr(r) / z - z : 1e20; 
				//double tan_gamma = (abs(P_z) < 1e-7) ? 0 : y / P_z;
				////tau_xz contributes to tau_xy with tan_gamma *tau_xz
				//double tau_xy_2 = tau_xz * tan_gamma;	


				////alt von AH double tau_r_3 = Mx / J * sqrt(y*y + z*z);
				////alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r_3 * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r_3;
				////Szabo - Hoehere Technische Mechanik (14.29)
				//double tau_xy_3 = -Mx/J * z;

				//return tau_xy_1 + tau_xy_2 + tau_xy_3;
			}
		}
	}
	//sigma_xz
	else if (fvd.ComponentIndex1() == FieldVariableDescriptor::FVCI_x && fvd.ComponentIndex2() == FieldVariableDescriptor::FVCI_z)
	{
		switch (mat.GetCrossSectionShape())
		{
		case BeamCrossSectionInfo::TCSS_rectangular:
			{

				return GetStressSMFxzRectangle(local_position, Iy, J, Qz, Mx);

				//double Sz = -b / 2. * (Sqr(h / 2.) - Sqr(z));
				//double tau_xz_1 = -Qz * Sz / (Iy * b);

				//// approximate (Ritz) solution according to (14.47) I. Szabo - Hoehere Technische Mechanik
				////for torsional influences
				//double h1 = h/2.;
				//double b1 = b/2.;
				//
				////compared to Szabo - den is scaled by h^6
				////in total "Szabo's c1" = 1/h^2 c1, "Szabo's c2" = 1/h^4 c2, "Szabo's c3" = 1/h^4 c3
				////in total the values are correct since the y^2 and z^2 in the caculations of tau_xy_2
				////are also scaled by h^2 and therefore the factors cancel 
				//double den = (45.*(1.+pow(b1/h1,6.)) + 509.*Sqr(b1/h1)*(1.+Sqr(b1/h1)));
				//double c1 = -35./16. *  (9.*(1+pow(b1/h1,4.)) + 130.*Sqr(b1/h1)) / den;
				//double c2 = -105./16. * (1. + 9.*Sqr(b1/h1)) / den;
				//double c3 = -105./16. * (9. + Sqr(b1/h1)) / den;

				//double tau_xz_2 = -4. * Mx / J * y * (Sqr(z/h1)-1.) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c3*(Sqr(y/h1) - Sqr(b1/h1)));
				//
				//return tau_xz_1 + tau_xz_2;
			}
		case BeamCrossSectionInfo::TCSS_circular:
			{
				return GetStressSMFxzCircle(local_position, Iy, Iz, J, Qy, Qz, Mx);

				//double sin_beta_1 = sqrt(1.-Sqr(z/r));
				//double sin_beta_2 = sqrt(1.-Sqr(y/r));
				////shear stress due to transversal force - see Hagedorn (Schubspannungen infolge Querkraft bei Biegung)			
				//double tau_xz_1 = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r);

				////P_y is the intersection point of the shear stress vectors (- y)
				//double P_y = (abs(y) > 1e-7) ? Sqr(r) / y - y : 1e20; 
				////double tan_gamma = y / P_z;
				//double tan_gamma = (abs(P_y) < 1e-7) ? 0 : z / P_y;
				////tau_xz contributes to tau_xy with tan_gamma *tau_xz
				//double tau_xy = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r);
				//double tau_xz_2 = tau_xy * tan_gamma;	

				////alt von AH double tau_r = Mx / J * sqrt(y*y + z*z);
				////alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r;
				////Szabo - Hoehere Technische Mechanik (14.29)
				//double tau_xz_3 = -Mx/J * y;

				//return tau_xz_1 + tau_xz_2 + tau_xz_3;
			}
		case BeamCrossSectionInfo::TCSS_tubular:
			{
				return GetStressSMFxzTube(local_position, Iy, Iz, J, Qy, Qz, Mx);

				//double y2 = Sqr(y);
				//double z2 = Sqr(z);
				//double tmp = Sqr(r) - y2;
				//double tmpi = Sqr(ri) - min(y2,Sqr(ri)) 
				//double Sy = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
				//double height = 2*(Sqrt(tmp) - Sqrt(tmpi));
				//tmp = Sqr(r) - z2;
				//tmpi = Sqr(ri) - min(z2,Sqr(ri)) 
				//double Sz = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
				//double width = 2*(Sqrt(tmp) - Sqrt(tmpi));

				//double tau_xy = - Qy*Sy/Iz / height;
				//double tau_xz^_1 = - Qz*Sz/Iy / width;

				////P_y is the intersection point of the shear stress vectors (- y)
				//double P_y = (abs(y) > 1e-7) ? Sqr(r) / y - y : 1e20; 
				//double tan_gamma = (abs(P_y) < 1e-7) ? 0 : z / P_y;
				////tau_xy contributes to tau_xz with tan_gamma *tau_xy
				//double tau_xz_2 = tau_xy * tan_gamma;	

				////alt von AH double tau_r_3 = Mx / J * sqrt(y*y + z*z);
				////alt von AH double tau_xz_3 = (abs(y) > 1e-7) ? Sgn(y) * tau_r_3 * 1./sqrt(1. + Sqr(z/y)) : 0;
				////Szabo - Hoehere Technische Mechanik (14.29)
				//double tau_xz_3 = -Mx/J * y;

				//return tau_xz_1 + tau_xz_2 + tau_xz_3;

				//double rmid = 0.5 * (r + ri);
				//double sin_phi = y / sqrt(Sqr(y)+Sqr(z));
				//double cos_phi = z / sqrt(Sqr(y)+Sqr(z));

				//double tau_r_1 = Qz * Sqr(rmid) * sin_phi / Iy;
				//double tau_xz_1 = tau_r_1 * sin_phi;

				//double tau_r_2 = -Qy * Sqr(rmid) * cos_phi / Iz;
				//double tau_xz_2 = tau_r_2 * sin_phi;

				//double tau_r_3 = Mx / J * sqrt(y*y + z*z);
				//double tau_xz_3 = (abs(y) > 1e-7) ? Sgn(y) * tau_r_3 * 1./sqrt(1. + Sqr(z/y)) : 0;
			
			}
		default:
			{
				assert(0);
				return 0.;
			}
		}
	}
	else if (fvd.ComponentIndex1() == FieldVariableDescriptor::FVCI_magnitude)
	{
		//double N = EA * ux;
		//double My = -EIy * wxx;
		//double Mz = -EIz * vxx;

		double N =  mat.BeamEA() * gamma.X();
		double sigma_xx = N / A + My / Iy * local_position.Z() + Mz / Iz * local_position.Y();

		double tau_xy = 0.;
		double tau_xz = 0.;

		switch (mat.GetCrossSectionShape())
		{
		case BeamCrossSectionInfo::TCSS_rectangular:
			{
				// tau_xy
				//double Sy = -h / 2. * (Sqr(b / 2.) - Sqr(y));
				//double tau_xy_1 = -Qy * Sy / (Iz * h);

				//// approximate (Ritz) solution according to (14.47) I. Szabo - Hoehere Technische Mechanik
				//double h1 = h/2.;
				//double b1 = b/2.;
				//double den = (45.*(1.+pow(b1/h1,6.)) + 509.*Sqr(b1/h1)*(1.+Sqr(b1/h1)));

				//double c1 = -35./16. *  (9.*(1+pow(b1/h1,4.)) + 130.*Sqr(b1/h1)) / den;
				//double c2 = -105./16. * (1. + 9.*Sqr(b1/h1)) / den;
				//double c3 = -105./16. * (9. + Sqr(b1/h1)) / den;

				//double tau_xy_2 = -4. * Mx / J * z * (Sqr(y/h1)-Sqr(b1/h1)) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c3*(Sqr(z/h1) - 1.));
				
				//tau_xy = tau_xy_1 + tau_xy_2;
				tau_xy = GetStressSMFxyRectangle(local_position, Iz, J, Qy, Mx);

				//// tau_xz
				//double Sz = -b / 2. * (Sqr(h / 2.) - Sqr(z)); 
				//double tau_xz_1 = -Qz * Sz / (Iy * b);

				//double tau_xz_2 = 4. * Mx / J * y * (Sqr(z/h1)-1.) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c2*(Sqr(y/h1) - Sqr(b1/h1)));

				//tau_xz = tau_xz_1 + tau_xz_2;
				tau_xz = GetStressSMFxzRectangle(local_position, Iy, J, Qz, Mx);
				break;
			}
		case BeamCrossSectionInfo::TCSS_circular:
			{
				// tau_xy
				//double sin_beta_1 = sqrt(1.-Sqr(z/r));
				//double P_z = (abs(z) > 1e-7) ? -Sqr(r) / z - z : 1e20; 
				//double tan_gamma = y / P_z;
				//double tau_xy_1 = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r) * tan_gamma;

				//double sin_beta_2 = sqrt(1.-Sqr(y/r));
				//double tau_xy_2 = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r);

				//double tau_r = Mx / J * sqrt(y*y + z*z);
				//double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r;
				//
				//tau_xy = tau_xy_1 + tau_xy_2 + tau_xy_3;
				tau_xy = GetStressSMFxyCircle(local_position,Iy,Iz,J,Qy,Qz,Mx);

				// tau_xz
				//sin_beta_1 = sqrt(1.-Sqr(z/r));
				//double tau_xz_1 = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r);
				//
				//sin_beta_2 = sqrt(1.-Sqr(y/r));
				//double P_y = (abs(y) > 1e-7) ? -Sqr(r) / y - y : 1e20; 
				//tan_gamma = z / P_y;
				//double tau_xz_2 = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r) * tan_gamma;

				//tau_r = Mx / J * sqrt(y*y + z*z);
				//double tau_xz_3 = (abs(y) > 1e-7) ? Sgn(y) * tau_r * 1./sqrt(1. + Sqr(z/y)) : 0;
				//
				//tau_xz = tau_xz_1 + tau_xz_2 + tau_xz_3;
				tau_xz = GetStressSMFxzCircle(local_position,Iy,Iz,J,Qy,Qz,Mx);
				break;
			}
		case BeamCrossSectionInfo::TCSS_tubular:
			{
				// tau_xy
				//double rmid = 0.5 * (r + ri);
				//double sin_phi = y / sqrt(Sqr(y)+Sqr(z));
				//double cos_phi = z / sqrt(Sqr(y)+Sqr(z));

				//double tau_r_1 = -Qz * Sqr(rmid) * sin_phi / Iy;
				//double tau_xy_1 = tau_r_1 * cos_phi;

				//double tau_r_2 = Qy * Sqr(rmid) * cos_phi / Iz;
				//double tau_xy_2 = tau_r_2 * cos_phi;

				//double tau_r_3 = Mx / J * sqrt(y*y + z*z);
				//double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r_3 * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r_3;

				//tau_xy = tau_xy_1 + tau_xy_2 + tau_xy_3;
				tau_xy = GetStressSMFxyTube(local_position,Iy,Iz,J,Qy,Qz,Mx);

				// tau_xz
				//tau_r_1 = Qz * Sqr(rmid) * sin_phi / Iy;
				//double tau_xz_1 = tau_r_1 * sin_phi;

				//tau_r_2 = -Qy * Sqr(rmid) * cos_phi / Iz;
				//double tau_xz_2 = tau_r_2 * sin_phi;

				//tau_r_3 = Mx / J * sqrt(y*y + z*z);
				//double tau_xz_3 = (abs(y) > 1e-7) ? Sgn(y) * tau_r_3 * 1./sqrt(1. + Sqr(z/y)) : 0;

				//tau_xz = tau_xz_1 + tau_xz_2 + tau_xz_3;
				tau_xz = GetStressSMFxzTube(local_position,Iy,Iz,J,Qy,Qz,Mx);
				break;
			}
		default:
			{
				assert(0);
				return 0.;
			}
		}
		double stress_mises = sqrt(sigma_xx*sigma_xx + 3.*(tau_xy*tau_xy + tau_xz*tau_xz));
		return stress_mises;
	}
	return FIELD_VARIABLE_NO_VALUE;
}




double ANCFBeamShear3DGeneric::GetStressSMFxyRectangle(const Vector3D& local_position, double Iz, double J, double Qy, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double b = mat.GetBeamThicknessY();
	double h = mat.GetBeamThicknessZ();
	double b1 = b/2.;
	double h1 = h/2.;	

	double Sy = -h / 2. * (Sqr(b / 2.) - Sqr(y));
	double tau_xy_1 = -Qy * Sy / (Iz * h);

	// approximate (Ritz) solution according to (14.47) I. Szabo - Hoehere Technische Mechanik
	//for torsional influences
	
	//compared to Szabo - den is scaled by h^6
	//in total "Szabo's c1" = 1/h^2 c1, "Szabo's c2" = 1/h^4 c2, "Szabo's c3" = 1/h^4 c3
	//in total the values are correct since the y^2 and z^2 in the caculations of tau_xy_2
	//are also scaled by h^2 and therefore the factors cancel 
	double den = (45.*(1.+pow(b1/h1,6.)) + 509.*Sqr(b1/h1)*(1.+Sqr(b1/h1)));
	double c1 = -35./16. *  (9.*(1+pow(b1/h1,4.)) + 130.*Sqr(b1/h1)) / den;
	double c2 = -105./16. * (1. + 9.*Sqr(b1/h1)) / den;
	double c3 = -105./16. * (9. + Sqr(b1/h1)) / den;

	double tau_xy_2 = -4. * Mx / J * z * (Sqr(y/h1)-Sqr(b1/h1)) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c3*(Sqr(z/h1) - 1.));
	return tau_xy_1 + tau_xy_2;
}

double ANCFBeamShear3DGeneric::GetStressSMFxyCircle(const Vector3D& local_position, double Iy, double Iz, double J, double Qy, double Qz, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double r = mat.GetRadius();

	double sin_beta_1 = sqrt(1.-Sqr(z/r));
	//P_z is the intersection point of the shear stress vectors (- z)
	double P_z = (abs(z) > 1e-7) ? Sqr(r) / z - z : 1e20; 
	//double tan_gamma = y / P_z;
	double tan_gamma = (abs(P_z) < 1e-7) ? 0 : y / P_z;
	//tau_xz contributes to tau_xy with tan_gamma *tau_xz
	double tau_xz = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r);
	double tau_xy_1 = tau_xz * tan_gamma;	

	//shear stress due to transversal force - see Hagedorn (Schubspannungen infolge Querkraft bei Biegung)
	double sin_beta_2 = sqrt(1.-Sqr(y/r));
	double tau_xy_2 = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r);

	//alt von AH double tau_r = Mx / J * sqrt(y*y + z*z);
	//alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r;
	//Szabo - Hoehere Technische Mechanik (14.29)
	double tau_xy_3 = -Mx/J * z;

	return tau_xy_1 + tau_xy_2 + tau_xy_3;
}

double ANCFBeamShear3DGeneric::GetStressSMFxyTube(const Vector3D& local_position, double Iy, double Iz, double J, double Qy, double Qz, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double r = mat.GetRadius();
	double ri = mat.GetInnerRadius();

	double y2 = Sqr(y);
	double z2 = Sqr(z);
	double tmp = Sqr(r) - y2;
	double tmpi = Sqr(ri) - min(y2,Sqr(ri));
	double Sy = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
	double height = 2*(Sqrt(tmp) - Sqrt(tmpi));
	tmp = Sqr(r) - z2;
	tmpi = Sqr(ri) - min(z2,Sqr(ri));
	double Sz = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
	double width = 2*(Sqrt(tmp) - Sqrt(tmpi));

	//problem if Sy = 0 since then height = 0 which would imply tau_xy_1 = 0 (instead of numerical problems)
	double tau_xy_1 = (abs(Sy)> 1e-10) ? - Qy*Sy/Iz / height : 0.;
	//problem if Sz = 0 since then width = 0 which would imply tau_xy_1 = 0 (instead of numerical problems)
	double tau_xz = (abs(Sz)> 1e-10) ? - Qz*Sz/Iy / width : 0.;

	//P_z is the intersection point of the shear stress vectors (- z)
	double P_z = (abs(z) > 1e-7) ? Sqr(r) / z - z : 1e20; 
	double tan_gamma = (abs(P_z) < 1e-7) ? 0 : y / P_z;
	//tau_xz contributes to tau_xy with tan_gamma *tau_xz
	double tau_xy_2 = tau_xz * tan_gamma;	

	//alt von AH double tau_r_3 = Mx / J * sqrt(y*y + z*z);
	//alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r_3 * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r_3;
	//Szabo - Hoehere Technische Mechanik (14.29)
	double tau_xy_3 = -Mx/J * z;

	return tau_xy_1 + tau_xy_2 + tau_xy_3;
}

double ANCFBeamShear3DGeneric::GetStressSMFxzRectangle(const Vector3D& local_position, double Iy, double J, double Qz, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double b = mat.GetBeamThicknessY();
	double h = mat.GetBeamThicknessZ();
	double b1 = b/2.;
	double h1 = h/2.;	
	
	double Sz = -b / 2. * (Sqr(h / 2.) - Sqr(z));
	double tau_xz_1 = -Qz * Sz / (Iy * b);
	// approximate (Ritz) solution according to (14.47) I. Szabo - Hoehere Technische Mechanik
	//for torsional influences
	
	//compared to Szabo - den is scaled by h^6
	//in total "Szabo's c1" = 1/h^2 c1, "Szabo's c2" = 1/h^4 c2, "Szabo's c3" = 1/h^4 c3
	//in total the values are correct since the y^2 and z^2 in the caculations of tau_xy_2
	//are also scaled by h^2 and therefore the factors cancel 
	double den = (45.*(1.+pow(b1/h1,6.)) + 509.*Sqr(b1/h1)*(1.+Sqr(b1/h1)));
	double c1 = -35./16. *  (9.*(1+pow(b1/h1,4.)) + 130.*Sqr(b1/h1)) / den;
	double c2 = -105./16. * (1. + 9.*Sqr(b1/h1)) / den;
	double c3 = -105./16. * (9. + Sqr(b1/h1)) / den;

	double tau_xz_2 = -4. * Mx / J * y * (Sqr(z/h1)-1.) * (c1 + c2*Sqr(y/h1) + c3*Sqr(z/h1) + c3*(Sqr(y/h1) - Sqr(b1/h1)));
	
	return tau_xz_1 + tau_xz_2;
}

double ANCFBeamShear3DGeneric::GetStressSMFxzCircle(const Vector3D& local_position, double Iy, double Iz, double J, double Qy, double Qz, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double r = mat.GetRadius();

	double sin_beta_1 = sqrt(1.-Sqr(z/r));
	double sin_beta_2 = sqrt(1.-Sqr(y/r));
	//shear stress due to transversal force - see Hagedorn (Schubspannungen infolge Querkraft bei Biegung)			
	double tau_xz_1 = 2./3. * Qz * Cub(r) * Sqr(sin_beta_1) / (Iy * 2*r);

	//P_y is the intersection point of the shear stress vectors (- y)
	double P_y = (abs(y) > 1e-7) ? Sqr(r) / y - y : 1e20; 
	//double tan_gamma = y / P_z;
	double tan_gamma = (abs(P_y) < 1e-7) ? 0 : z / P_y;
	//tau_xz contributes to tau_xy with tan_gamma *tau_xz
	double tau_xy = 2./3. * Qy * Cub(r) * Sqr(sin_beta_2) / (Iz * 2*r);
	double tau_xz_2 = tau_xy * tan_gamma;	

	//alt von AH double tau_r = Mx / J * sqrt(y*y + z*z);
	//alt von AH double tau_xy_3 = (abs(y) > 1e-7) ?  -Sgn(y) * tau_r * z/y/sqrt(1. + Sqr(z/y)) : -Sgn(y)*tau_r;
	//Szabo - Hoehere Technische Mechanik (14.29)
	double tau_xz_3 = -Mx/J * y;

	return tau_xz_1 + tau_xz_2 + tau_xz_3;

}

double ANCFBeamShear3DGeneric::GetStressSMFxzTube(const Vector3D& local_position, double Iy, double Iz, double J, double Qy, double Qz, double Mx)
{
	double y = local_position.Y();
	double z = local_position.Z();

	const Beam3DProperties& mat = GetBeam3DProperties();
	double r = mat.GetRadius();
	double ri = mat.GetInnerRadius();

	double y2 = Sqr(y);
	double z2 = Sqr(z);
	double tmp = Sqr(r) - y2;
	double tmpi = Sqr(ri) - min(y2,Sqr(ri));
	double Sy = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
	double height = 2*(Sqrt(tmp) - Sqrt(tmpi));
	tmp = Sqr(r) - z2;
	tmpi = Sqr(ri) - min(z2,Sqr(ri));
	double Sz = -2.* (sqrt(Cub(tmp)) - sqrt(Cub(tmpi)))/3.;
	double width = 2*(Sqrt(tmp) - Sqrt(tmpi));

	//problems if either Sy=0 or Sz =0 -> division by 0
	double tau_xy =(abs(Sy)> 1e-10) ? - Qy*Sy/Iz / height : 0.;
	double tau_xz_1 = (abs(Sz)> 1e-10) ? - Qz*Sz/Iy / width : 0.;

	//P_y is the intersection point of the shear stress vectors (- y)
	double P_y = (abs(y) > 1e-7) ? Sqr(r) / y - y : 1e20; 
	double tan_gamma = (abs(P_y) < 1e-7) ? 0 : z / P_y;
	//tau_xy contributes to tau_xz with tan_gamma *tau_xy
	double tau_xz_2 = tau_xy * tan_gamma;	

	//alt von AH double tau_r_3 = Mx / J * sqrt(y*y + z*z);
	//alt von AH double tau_xz_3 = (abs(y) > 1e-7) ? Sgn(y) * tau_r_3 * 1./sqrt(1. + Sqr(z/y)) : 0;
	//Szabo - Hoehere Technische Mechanik (14.29)
	double tau_xz_3 = -Mx/J * y;

	return tau_xz_1 + tau_xz_2 + tau_xz_3;

}

void ANCFBeamShear3DGeneric::GyroscopicMatrix(Matrix& m)
{
	m.SetSize(SOS(), SOS());
	m.FillWithZeros();

	if (fabs(rotor_angular_velocity) > 1e-14)
	{
		int dim = Dim(); 
		int ns = NS();
		Vector SV(ns, 1), SVy(ns, 1), SVz(ns, 1);
		Matrix HL(ns*dim,dim), HLCS(ns*dim,dim);
		Vector x1,x2,x3,w1,w2,w3; //integration points and according weights

		//GetIntegrationRule => Gau points on interval [-1,+1] (see also EvalM)
		GetIntegrationRule(x1,w1,7); 
		GetIntegrationRule(x2,w2,3);
		GetIntegrationRule(x3,w3,3);

		for (int i1=1; i1<=x1.GetLen(); i1++)
		{
			for (int i2=1; i2<=x2.GetLen(); i2++)
			{
				for (int i3=1; i3<=x3.GetLen(); i3++)
				{
					double x = x1(i1)*0.5*GetLx();  //x1 in [-1,+1], scaled rect. reference element: \xi
					double y = x2(i2)*0.5*GetLy();
					double z = x3(i3)*0.5*GetLz();
					Vector3D ploc(x,y,z);
					Vector3D rx0 = GetInitPosx3D(ploc);
					double rxn = rx0.Norm(); //curvature in reference element
					double jacdet = 0.5*GetLx()*0.5*GetLy()*0.5*GetLz() * rxn; //element transformation

					GetShapes(SV,ploc);
					GetShapesy(SVy, ploc);
					GetShapesz(SVz, ploc);

					for (int i=0; i<ns; i++)
					{
						for (int j=1; j<=dim; j++)
						{
							HL(i*dim+j,j) = SV(i+1);
							HLCS(i*dim+j,j) = y*SVz(i+1) - z*SVy(i+1);
						}
					}
					m += 2. * rotor_angular_velocity * fabs (jacdet) * Rho() * w1(i1)*w2(i2)*w3(i3) * (HL*HLCS.GetTp());
				}
			}
		}
	}
}

void ANCFBeamShear3DGeneric::GetAvailableFieldVariables(TArray<FieldVariableDescriptor> & variables)
{
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_position, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_displacement, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_bryant_angle, FieldVariableDescriptor::FVCI_z); //$JG 2014-03-07 added
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_rotations_312, FieldVariableDescriptor::FVCI_z); //$JG 2012-03-07 added

	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_velocity, FieldVariableDescriptor::FVCI_z);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_angular_velocity, FieldVariableDescriptor::FVCI_z);
	
	//$EK 2015-10 added stresses to the SMF 
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress,
			FieldVariableDescriptor::FVCI_z, true);
	FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress_mises);

	if (InContinuumMechanicsFormulation())
	{
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_total_strain,
			FieldVariableDescriptor::FVCI_z, true);	}
	else
	{
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_curvature, FieldVariableDescriptor::FVCI_z);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_torsion);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_moment_bending, FieldVariableDescriptor::FVCI_z);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_moment_torsional);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_shear, FieldVariableDescriptor::FVCI_z);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_axial_extension);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_force_transversal, FieldVariableDescriptor::FVCI_z);
		FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_force_axial);

		//this is the global beam moment! JG2013-09:
		//FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_beam_moment, FieldVariableDescriptor::FVCI_z);
		// for sm-formulation the total strain contains:
		//  Gamma_1 Gamma_2 Gamma_3
		//  Gamma_2  E_yy    E_yz
		//  Gamma_3  E_yz    E_zz
		// ... same for the stress
		
		//strains and stresses are computed in WRONG WAY JG2013-09 !!!
		//FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_total_strain, FieldVariableDescriptor::FVCI_z, true);   
		//FieldVariableDescriptor::AddTypeIntoArray(variables, FieldVariableDescriptor::FVT_stress, FieldVariableDescriptor::FVCI_z, true);

	}
}

static bool warning3362_tobeprinted = true;
void ANCFBeamShear3DGeneric::DrawElementPreProc()
{
	//if plot interpolated is not checked not precomputation necessary
	//old: if (!GetMBS()->GetIOption(118)) return;
	if (!GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourPlotInterpolated()) return;
	//if nothing is chosen return
	if (GetMBS()->GetActualPostProcessingFieldVariable() == NULL) return;

	double lx1 = 1, ly1 = 1, lz1 = 1;
	if(GetMBS()->GetMagnifyYZ() != 1)
	{
		ly1 *= GetMBS()->GetMagnifyYZ();
		lz1 *= GetMBS()->GetMagnifyYZ();
		if (warning3362_tobeprinted)
		{
			mbs->UO(UO_LVL_warn) << "WARNING in ANCFBeamShear3DGeneric: GetMagnifyYZ() leads to wrong stress evaluation\n";
			warning3362_tobeprinted = false;
		}
	}

	double def_scale = GetMBS()->GetDOption(105); //deformation scaling
	double thickness = GetMBS()->GetDOption(102); //line thickness
	double vLeft=0;
	double vRight=0;

	//depending on the field variable and the cross section type to be plotted the cross section is subdivided or not
	//the following FVs need a subdivision: position?, displacement, velocity, angular_velocity(?), stresses
	bool needsDivision = FieldVarialbeNeedsSubdivision();

	if (needsDivision)
	{
		const Beam3DProperties& bp = GetBeam3DProperties();
		if(bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_rectangular)
		{
			int tileyn = (int)GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(); //formerly GetIOption(138)
			averagedFVvaluesForDrawing.SetLen(Sqr(tileyn+1));

			//case 5: // left
			Vector3D p0(-lx1, ly1,-lz1);
			Vector3D vx(0,-2.*ly1/tileyn,0);
			Vector3D vy(0,0,2.*lz1/tileyn); 
	
			for (double ix = 0; ix <= tileyn+1e-10; ix++)
			{
				for (double iy = 0; iy <= tileyn+1e-10; iy++)
				{
					Vector3D ploc = (p0+ix*vx+iy*vy);//-1<=p0<=+1
					averagedFVvaluesForDrawing((int)(ix*(tileyn+1) + iy)+1) = GetAveragedFVvalue(ploc);
				}
			}

		}
		//in case of circular and tubular shapes the FV values have to be computed only once for a specific x if 
		//NeedsDivision is false
		else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_circular)
		{
			
			size.Y() = 2*bp.GetRadius();
			size.Z() = 2*bp.GetRadius();


			int tilern = Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2));//formerly GetIOption(138)
			int tilexn = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
			averagedFVvaluesForDrawing.SetLen((tilexn+1)*(tilern+1));

			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{

					double radius = 1;
					Vector3D ploc = Vector3D(0.,-radius/tilexn*ix*cos(del_phi*ir),-radius/tilexn*ix*sin(del_phi*ir));
										
					//Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);
					Vector3D plocScaledLeft(-GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2
					Vector3D plocScaledRight(GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2

					vLeft = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledLeft, true);	//else v = 0.;		
					vRight = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledRight, true);	//else v = 0.;		
	
					averagedFVvaluesForDrawing((int)(ix*(tilern+1) + ir)+1) = (vLeft + vRight)/2.;
				}
			}
		}
		else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_tubular)
		{
		
			size.Y() = bp.GetBeamCrossSectionSize(1);
			size.Z() = bp.GetBeamCrossSectionSize(1);

			//double radius = 0.5*bp.GetBeamCrossSectionSize(1);
			//double inner_radius = 0.5*bp.GetBeamCrossSectionSize(2);

			int tilern = Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2));  //formerly GetIOption(138)
			int tilexn = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
			averagedFVvaluesForDrawing.SetLen((tilexn+1)*(tilern+1));

			double inner_radius = bp.GetBeamCrossSectionSize(2)/bp.GetBeamCrossSectionSize(1); // inner radius in local coordinates
			double radius = 1;
			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{

//				ploc = px+Vector3D(0.,-radius/tilexn*ix*cos(del_phi*ir),-radius/tilexn*ix*sin(del_phi*ir));
//				Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);

					Vector3D ploc = Vector3D(0.,-(inner_radius+(radius-inner_radius)/tilexn*ix)*cos(del_phi*ir),-(inner_radius+(radius-inner_radius)/tilexn*ix)*sin(del_phi*ir));					

					//Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);
					//Vector3D plocScaledLeft(-GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2
					//Vector3D plocScaledRight(GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2

					//vLeft = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledLeft, true);	//else v = 0.;		
					//vRight = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledRight, true);	//else v = 0.;		
	
					//averagedFVvaluesForDrawing((int)(ix*(tilern+1) + ir)+1) = (vLeft + vRight)/2.;
					averagedFVvaluesForDrawing((int)(ix*(tilern+1) + ir)+1) = GetAveragedFVvalue(ploc);
				}
			}	
		}
	}
	//only the center has to be evaluated
	else
	{
		averagedFVvaluesForDrawing.SetLen(1);
		//double vLeft = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), Vector3D(-GetLx()/Sqrt(3)/2,0.,0.), true);	//else v = 0.;		
		//double vRight = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), Vector3D(GetLx()/Sqrt(3)/2,0.,0.), true);	//else v = 0.;		
		//averagedFVvaluesForDrawing(1) = (vLeft+vRight)/2.;
		averagedFVvaluesForDrawing(1) = GetAveragedFVvalue(Vector3D(0.,0.,0.));
	}
}

//$ EK 2015-12 average the actual field variable along the line determined through the points at the cross section given by (ploc(2), ploc(3))
double ANCFBeamShear3DGeneric::GetAveragedFVvalue(Vector3D& ploc)
{
	//Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);
	Vector3D plocScaledLeft(-GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2
	Vector3D plocScaledRight(GetLx()*0.5/Sqrt(3), ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2

	double vLeft = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledLeft, true);	//else v = 0.;		
	double vRight = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), plocScaledRight, true);	//else v = 0.;		
	return (vLeft + vRight)/2;
}

//$ EK 2015-12-02 splitted into general drawing and interpolated drawing
void ANCFBeamShear3DGeneric::DrawElement()
{
	//if plot interpolated is not checked no precomputation necessary
	//old if (!GetMBS()->GetIOption(118)) 
	if (GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourPlotInterpolated())
	{
		DrawElementInterpolated();
	}
	else
	{
		DrawElementGeneral();
	}
}


//$ EK 2015-12-02 determines if the FV is non-constant along the cross section
bool ANCFBeamShear3DGeneric::FieldVarialbeNeedsSubdivision()
{
	//in case of body color not field variable is chosen
	if (GetMBS()->GetActualPostProcessingFieldVariable() == NULL)
	{ return false; }
	switch (GetMBS()->GetActualPostProcessingFieldVariable()->VariableType())
	{
		case FieldVariableDescriptor::FVT_position: return true;
		case FieldVariableDescriptor::FVT_displacement: return true;
		case FieldVariableDescriptor::FVT_velocity: return true;
		case FieldVariableDescriptor::FVT_acceleration: return true;
		case FieldVariableDescriptor::FVT_angular_velocity: return true;
		case FieldVariableDescriptor::FVT_stress: return true;
		case FieldVariableDescriptor::FVT_stress_mises: return true;
		default: return false;
	}
	return false;
}

//$ EK 2015-12-02 used in order to accelerate drawing (do not draw inner faces)
//set all outer faces active, which have no opposite face from an other element
void ANCFBeamShear3DGeneric::SetOuterFaces()
{
	SetAllOuterFaceFlags();
	Node& n1 = GetNode(1);
	Node& n2 = GetNode(1);
	TArray<TArray<int>*>& n2e = GetMBS()->N2E();
	const Beam3DProperties& bp = GetBeam3DProperties();

	if(bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_rectangular)
	{
		if (n2e(NodeNum(1))->Length() > 1)
		{
			//left face
			SetOuterFaceFlag(5,0);
		}
		if (n2e(NodeNum(2))->Length() > 1)
		{
			//right face
			SetOuterFaceFlag(6,0);
		}
	}
	else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_circular || bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_tubular)
	{
		if (n2e(NodeNum(1))->Length() > 1)
		{
			//left face
			SetOuterFaceFlag(2,0);
		}
		if (n2e(NodeNum(2))->Length() > 1)
		{
			//right face
			SetOuterFaceFlag(3,0);
		}
	}
}
//$ EK 2015-12-02 used in order to accelerate drawing (do not draw inner faces)
void ANCFBeamShear3DGeneric::SetOuterFaceFlag(int face, int set /* = 1 */)
{
	if (set == 1)
	{
		outer_face = outer_face|GetCharBit(face);
	}
	else
	{
		if (GetOuterFaceFlag(face))
		{
			outer_face = (char)((int)outer_face - GetIntBit(face));
		}
	}
}


void ANCFBeamShear3DGeneric::DrawElementGeneral()
{
	mbs->Graphics3D()->SetColor(col);
	mbs->Graphics3D()->SetDrawOutline(1);

	//double lx1 = 1; double ly1 = 1/**GetMBS()->GetMagnifyYZ()*/; double lz1 = 1/**GetMBS()->GetMagnifyYZ()*/; //GetMagnifyYZ() leads to wrong stress evaluation
	double lx1 = 1, ly1 = 1, lz1 = 1;
	if(GetMBS()->GetMagnifyYZ() != 1)
	{
		ly1 *= GetMBS()->GetMagnifyYZ();
		lz1 *= GetMBS()->GetMagnifyYZ();
		if (warning3362_tobeprinted)
		{
			mbs->UO(UO_LVL_warn) << "WARNING in ANCFBeamShear3DGeneric: GetMagnifyYZ() leads to wrong stress evaluation\n";
			warning3362_tobeprinted = false;
		}
	}

	double def_scale = GetMBS()->GetDOption(105); //deformation scaling

	//GetMBS()->GetOptions()->ViewingOptions()->BodiesDeformationScaleFactor()
	double thickness = GetMBS()->GetDOption(102); //line thickness

	int linemode = GetLineMode();
	//int linemode = 1; //0=no lines, 1=outline+color, 2=outline, 3=elementline+color, 4=elementline


	int colormode = 0;
	const FieldVariableDescriptor * fv = GetMBS()->GetActualPostProcessingFieldVariable();
	if (fv != NULL && HasFieldVariable(*fv))
	{
		colormode = 1;
	}

	//depending on the field variable and the cross section type to be plotted the cross section is subdivided or not
	//the following FVs need a subdivision: position?, displacement, velocity, angular_velocity(?), stresses
	bool needsDivision = FieldVarialbeNeedsSubdivision();

	const Beam3DProperties& bp = GetBeam3DProperties();
	if(bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_rectangular)
	{
		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
		double tiley = 1.;
		if (needsDivision)		
		{
		 tiley = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection();  //formerly GetIOption(138)GetMBS()->GetIOption(138);
		}
		TArray<Vector3D> points((int)(tilex+1)*(int)(tiley+1));
		TArray<double> vals((int)(tilex+1)*(int)(tiley+1));
		double v=0;

		for (int side=1; side <= 6; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); vals.SetLen(0);
			Vector3D p0, vx, vy;
			int tileyn = (int)tiley;
			int tilexn = (int)tilex;
			//in case of left or right side, the cross section division has to be used
			if (side >= 5)
			{
				tilexn = (int)tiley;
			}

			switch(side)
			{
			case 1: // bottom
				p0 = Vector3D(-lx1,-ly1,-lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			case 2: // top
				p0 = Vector3D(-lx1, ly1, lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,0,-2.*lz1/tileyn); break;
			case 3: // front
				p0 = Vector3D(-lx1,-ly1, lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,2.*ly1/tileyn,0); break;
			case 4: // back
				p0 = Vector3D(-lx1, ly1,-lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,-2.*ly1/tileyn,0); break;
			case 5: // left
				p0 = Vector3D(-lx1, ly1,-lz1);
				vx = Vector3D(0,-2.*ly1/tilexn,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			case 6: // right
				p0 = Vector3D( lx1,-ly1,-lz1);
				vx = Vector3D(0,2.*ly1/tileyn,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			default: ;
			}
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double iy = 0; iy <= tileyn+1e-10; iy++)
				{

					Vector3D ploc = (p0+ix*vx+iy*vy);//-1<=p0<=+1
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5,ploc(2)*GetLy()*0.5,ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2
					Vector3D pg = GetPos3D0D(ploc, def_scale);
		
				//	mbs->UO()<<"point("<< side << "," << iy << "," << ix << "): " << ploc << " --> " << ploc_scaled << "-->" << pg << "\n";
			
					points.Add(pg);
					//if (colormode) ComputeStressD(ploc,type,comp1,comp2,v);//old_KN
					//if no division is needed in y-direction (constant on cross-section area) the value has to be computed only
					//for the first time
					//additionally if side is left or right only the first value has to be computed (ix=0 & iy = 0)
					//in case of !needsDivision the point-information in y and z direction is NOT used  for the FV calculation
					if (colormode && (needsDivision || ((iy == 0) && ((side >= 5 && ix == 0) || (side <= 4)))))
					{
						v = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(), ploc_scaled, true);	//else v = 0.;
					}
					vals.Add(v);
				}
			}
			mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tileyn+1,(int)tilexn+1,colormode,linemode);
		}
	}
	//in case of circular and tubular shapes the FV values have to be computed only once for a specific x if 
	//NeedsDivision is false
	else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_circular)
	{
		//$EK 2015-10-14 bug fix - size has to contain the diameter -> before GetRadius returned erroneously the diameter
		size.Y() = 2*bp.GetRadius();
		size.Z() = 2*bp.GetRadius();
		
		//in case of circular or tubular cross sections a tiling has to be performed
		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
		double tiler = Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2)); //formerly GetIOption(138)

		TArray<Vector3D> points((int)(tilex+1)*(int)(tiler+1));
		TArray<double> vals((int)(tilex+1)*(int)(tiler+1));
		double v=0;

		for (int side=1; side <= 3; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); vals.SetLen(0);
			Vector3D px, vx, vr;
			int tilern = (int)tiler;
			int tilexn = (int)tilex;

			switch(side)
			{
			case 1: //cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			case 2: //left side
				px = Vector3D(-lx1,0., 0.); break;
			case 3: //right side
				px = Vector3D(lx1,0., 0.); break;
			default: ;
			}

			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{
					Vector3D ploc;
					double radius = 1;
					if (side==1) 
					{
						ploc = px+ix*vx+Vector3D(0.,-radius*cos(del_phi*ir),-radius*sin(del_phi*ir));
					} 
					else 
					{
						ploc = px+Vector3D(0.,-radius/tilexn*ix*cos(del_phi*ir),-radius/tilexn*ix*sin(del_phi*ir));
					}
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);
					Vector3D pg = GetPos3D0D(ploc, def_scale);

			//		mbs->UO()<<"point(" << side << "," << ir << "," << ix << "): " << ploc << " --> " << ploc_scaled << "-->" << pg << "\n";
					
					points.Add(pg);
					if (colormode && (needsDivision || ( ir == 0 && ((side == 1) || (side >= 2 && ix == 0)))))
					{
						v = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(),ploc_scaled, true); //else v = 0.;
					}
					vals.Add(v);
				}
			}
			// note: using linemode variable here gives bad result (outline of the quadstrip: top and bottom circle but also single edge on the surface)
			//in case of linemode == 2 no quads are drawn anyway
			if(linemode == 2)
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,linemode);
			}
			else
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,0/*linemode*/);
				if (linemode != 0)
				{
					for (double ir = 0; ir <= tilern+1e-10; ir++) // draw lines last for better visibility
					{
						Vector3D p1loc = px+Vector3D(0.,-cos(del_phi* ir   ),-sin(del_phi* ir   ));
						Vector3D p2loc = px+Vector3D(0.,-cos(del_phi*(ir+1)),-sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness);

					}
				}
			}
		}
	}
	else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_tubular)
	{
		size.Y() = bp.GetBeamCrossSectionSize(1);
		size.Z() = bp.GetBeamCrossSectionSize(1);

		double radius = 0.5*bp.GetBeamCrossSectionSize(1)*GetMBS()->GetMagnifyYZ();
		double inner_radius = 0.5*bp.GetBeamCrossSectionSize(2)*GetMBS()->GetMagnifyYZ();

		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
		double tiler =  Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2));  //formerly GetIOption(138), 2));  
		TArray<Vector3D> points((int)(tilex+1)*(int)(tiler+1));
		TArray<double> vals((int)(tilex+1)*(int)(tiler+1));
		double v=0;

		for (int side=1; side <= 4; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); vals.SetLen(0);
			Vector3D px, vx, vr;
			int tilern = (int)tiler;
			int tilexn = (int)tilex;
			
			switch(side)
			{
			case 1: //outer cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			case 2: //left side
				px = Vector3D(-lx1,0., 0.); break;
			case 3: //right side
				px = Vector3D(lx1,0., 0.); break;
			case 4: //inner cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			default: ;
			}

			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{
					Vector3D ploc;
					double inner_radius = bp.GetBeamCrossSectionSize(2)/bp.GetBeamCrossSectionSize(1); // inner radius in local coordinates
					double radius = 1;                       // outer radius in local coordinates is always one
					if (side==1) 
					{
						ploc = px+ix*vx+Vector3D(0.,-radius*cos(del_phi*ir),-radius*sin(del_phi*ir));
					} 
					else if (side==4)
					{
						ploc = px+ix*vx+Vector3D(0.,-inner_radius*cos(del_phi*ir),-inner_radius*sin(del_phi*ir));						
					}
					else 
					{
						ploc = px+Vector3D(0.,-(inner_radius+(radius-inner_radius)/tilexn*ix)*cos(del_phi*ir),-(inner_radius+(radius-inner_radius)/tilexn*ix)*sin(del_phi*ir));
					}
					//Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);		
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLy()*0.5);		
					Vector3D pg = GetPos3D0D(ploc, def_scale);		

					points.Add(pg);

					if (colormode && (needsDivision || ( ir == 0 && ((side == 1 || side == 4) || ((side == 2 || side == 3) && ix == 0)))))
					{
						v = GetFieldVariableValue(*GetMBS()->GetActualPostProcessingFieldVariable(),ploc_scaled, true); //else v = 0.;
					}
					vals.Add(v);
				}
			}
			// note: using linemode variable here gives bad result (outline of the quadstrip: top and bottom circle but also single edge on the surface)
			//in case of linemode == 2 no quads are drawn anyway
			if(linemode == 2)
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,linemode);		
			}
			else
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,0/*linemode*/);		
				if (linemode != 0)
				{
					double ri = bp.GetBeamCrossSectionSize(2)/bp.GetBeamCrossSectionSize(1);
					for (double ir = 0; ir <= tilern+1e-10; ir++) // draw lines last for better visibility
					{
						Vector3D p1loc = px+Vector3D(0.,-cos(del_phi* ir   ),-sin(del_phi* ir   ));
						Vector3D p2loc = px+Vector3D(0.,-cos(del_phi*(ir+1)),-sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness); // outer surface of the tube

						p1loc = px+Vector3D(0.,-ri*cos(del_phi* ir   ),-ri*sin(del_phi* ir   ));
						p2loc = px+Vector3D(0.,-ri*cos(del_phi*(ir+1)),-ri*sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness); // inner surface of the tube
					}
				}
			}
		}
	}

	else
	{
		assert(0);
	}
};


//ACHTUNG: nur kopiert von ANCFBeam3D und ComputeStress-Aufrufe auskommentiert
void ANCFBeamShear3DGeneric::DrawElementInterpolated()
{
	mbs->Graphics3D()->SetColor(col);
	mbs->Graphics3D()->SetDrawOutline(1);

	//double lx1 = 1; double ly1 = 1/**GetMBS()->GetMagnifyYZ()*/; double lz1 = 1/**GetMBS()->GetMagnifyYZ()*/; //GetMagnifyYZ() leads to wrong stress evaluation
	double lx1 = 1, ly1 = 1, lz1 = 1;
	if(GetMBS()->GetMagnifyYZ() != 1)
	{
		ly1 *= GetMBS()->GetMagnifyYZ();
		lz1 *= GetMBS()->GetMagnifyYZ();
		if (warning3362_tobeprinted)
		{
			mbs->UO(UO_LVL_warn) << "WARNING in ANCFBeamShear3DGeneric: GetMagnifyYZ() leads to wrong stress evaluation\n";
			warning3362_tobeprinted = false;
		}
	}

	double def_scale = GetMBS()->GetDOption(105); //deformation scaling
	double thickness = GetMBS()->GetDOption(102); //line thickness

	int linemode = GetLineMode();

	int colormode = 0;
	const FieldVariableDescriptor * fv = GetMBS()->GetActualPostProcessingFieldVariable();
	if (fv != NULL && HasFieldVariable(*fv))
	{
		colormode = 1;
	}

	//depending on the field variable and the cross section type to be plotted the cross section is subdivided or not
	//the following FVs need a subdivision: position?, displacement, velocity, angular_velocity(?), stresses
	bool needsDivision = FieldVarialbeNeedsSubdivision();
	//factor which guarantees the correct index access
	int needsDivisionFactor = (needsDivision) ? 1 : 0;
	//get left and right averaged field variable values
	TArray<double> leftVals;
	TArray<double> rightVals;
	//determine averaged vales at neighboring elements if necessary
	TArray<TArray<int>*>& n2e = GetMBS()->N2E();
	//first the left neighbor (if exists)
	TArray<int>* node2elemLeft = n2e(GetNode(1).NodeNum());			
	////then the right neighbor (if exists)
	TArray<int>* node2elemRight = n2e(GetNode(2).NodeNum());			
	////determine corresponding element numbers
	leftVals = DetermineAveragedFVValues( node2elemLeft);
	rightVals = DetermineAveragedFVValues( node2elemRight);

	const Beam3DProperties& bp = GetBeam3DProperties();
	if(bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_rectangular)
	{
// scaling is done later
		//////double lx1 = 0.5*size.X()*GetMBS()->GetMagnifyYZ();
		//////double ly1 = 0.5*size.Y()*GetMBS()->GetMagnifyYZ();
		//////double lz1 = 0.5*size.Z()*GetMBS()->GetMagnifyYZ();
	
		//tilex for side 1-4 is 1 for side 5/6 it is tiley
		
		//formerly double tilex = GetMBS()->GetIOption(137);
		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis();
		double tiley = 1.;
		if (needsDivision)		
		{
		 tiley = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(); //formerly GetMBS()->GetIOption(138)
		}

		int nCrossSectionValues = 1;
		if (needsDivision)
		{
			nCrossSectionValues = (int)(Sqr((tiley+1))+1e-10);
		}
		TArray<Vector3D> points((int)Sqr(tiley+1));
		TArray<double> vals((int)Sqr(tiley+1));
		
		double v=0;
		for (int side=1; side <= 6; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); 
			vals.SetLen(0);
			Vector3D p0, vx, vy;
			int tileyn = (int)tiley;
			int tilexn = (int)tilex;
			//in case of left or right side, the cross section division has to be used
			if (side >= 5)
			{
				tilexn = (int)tiley;
			}

			switch(side)
			{
			case 1: // bottom  ----->>>> back
				p0 = Vector3D(-lx1,-ly1,-lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			case 2: // top ----->>>> front
				p0 = Vector3D(-lx1, ly1, lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,0,-2.*lz1/tileyn); break;
			case 3: // front ----->>>> top
				p0 = Vector3D(-lx1,-ly1, lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,2.*ly1/tileyn,0); break;
			case 4: // back ----->>>> bottom
				p0 = Vector3D(-lx1, ly1,-lz1);
				vx = Vector3D(2.*lx1/tilexn,0,0);
				vy = Vector3D(0,-2.*ly1/tileyn,0); break;
			case 5: // left
				p0 = Vector3D(-lx1, ly1,-lz1);
				vx = Vector3D(0,-2.*ly1/tilexn,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			case 6: // right
				p0 = Vector3D( lx1,-ly1,-lz1);
				vx = Vector3D(0,2.*ly1/tileyn,0);
				vy = Vector3D(0,0,2.*lz1/tileyn); break;
			default: ;
			}
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double iy = 0; iy <= tileyn+1e-10; iy++)
				{

					Vector3D ploc = (p0+ix*vx+iy*vy);//-1<=p0<=+1
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5,ploc(2)*GetLy()*0.5,ploc(3)*GetLz()*0.5); //-l/2<=ploc_scaled<=+l/2
					Vector3D pg = GetPos3D0D(ploc, def_scale);
		

				//	mbs->UO()<<"point("<< side << "," << iy << "," << ix << "): " << ploc << " --> " << ploc_scaled << "-->" << pg << "\n";
			
					points.Add(pg);
					int index;
					if (side <= 4)
					{
						//note that -Ly/2 corresponds to ix==0 at side 5 (which is the starting point for the averaging)
						double vLeft, vRight;
		
						switch (side)
						{
						case 4: 
							//bottom: along line with iy=0 (iy (on bottom side) --> ix (on left side))					
							index = needsDivisionFactor*iy*(tileyn+1) + 1;
							break;
						case 3: 						
							//top: along line with iy=tileyn (iy (on top side) --> ix (on left side))
							//on top y-direction runs positive while on reference side (side 5) y runs in negative direction -> revert index
							index = needsDivisionFactor*((tileyn-iy)*(tileyn+1) + tileyn) + 1;
							//vLeft=((*leftVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							//vRight=((*rightVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							//v = ((tilexn - ix) * vLeft + ix*vRight)/tilexn;
							break;
						case 2:
							//back: along line with ix=0 (iy (on front side) --> iy (on left side))
							//vertical coordinate runs in negative direction
							index = needsDivisionFactor*(tileyn-iy) + 1;
							//vLeft = ((*leftVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							//vRight = ((*rightVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							//v = ((tilexn - ix) * vLeft + ix*vRight)/tilexn;
							break;
						case 1:
							//front: along line with ix=tileyn (iy (on front side) --> iy (on left side))
							index = needsDivisionFactor*(tileyn*(tileyn+1) + iy) + 1;
							/*vLeft = ((*leftVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							vRight = ((*rightVals)(index) + averagedFVvaluesForDrawing(index))/2.;  
							v = ((tilexn - ix) * vLeft + ix*vRight)/tilexn;*/
							break;
						}
						//v = ((tilexn - ix) * (*leftVals)(index) + ix*(*rightVals)(index))/tilexn;
						v = ((tilexn - ix) * leftVals(index) + ix*rightVals(index))/tilexn;
						//v = ((tilexn - ix) * vLeft + ix*vRight)/tilexn;
					}
					else if (side == 5)
					{
						//left side
						index = needsDivisionFactor*(ix*(tileyn+1) + iy) + 1;
						//v = (*leftVals)(index);  
						v = leftVals(index);  
					}
					else if (side == 6)
					{
						//right side
						index = needsDivisionFactor*((tilexn-ix)*(tileyn+1) + iy) + 1;
						//v = (*rightVals)(index);  
						v = rightVals(index);  
					}					
					vals.Add(v);
				}
			}
			mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tileyn+1,(int)tilexn+1,colormode,linemode);
		}
	}
	//in case of circular and tubular shapes the FV values have to be computed only once for a specific x if 
	//NeedsDivision is false
	else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_circular)
	{
		// BUT: radius is required for ploc_scaled and pg later
		//$EK 2015-10-14 bug fix - size has to contain the diameter -> before GetRadius returned erroneously the diameter
		size.Y() = 2*bp.GetRadius();
		size.Z() = 2*bp.GetRadius();
		
		//in case of circular or tubular cross sections a tiling has to be performed
		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis();
		double tiler = Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2)); //formerly GetIOption(138), 2));  

		TArray<Vector3D> points((int)(tilex+1)*(int)(tiler+1));
		TArray<double> vals((int)(tilex+1)*(int)(tiler+1));
		double v=0;

		for (int side=1; side <= 3; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); vals.SetLen(0);
			Vector3D px, vx, vr;
			int tilern = (int)tiler;
			int tilexn = (int)tilex;

			switch(side)
			{
			case 1: //cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			case 2: //left side
				px = Vector3D(-lx1,0., 0.); break;
			case 3: //right side
				px = Vector3D(lx1,0., 0.); break;
			default: ;
			}

			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{
					Vector3D ploc;
					double radius = 1;
					if (side==1) 
					{
						ploc = px+ix*vx+Vector3D(0.,-radius*cos(del_phi*ir),-radius*sin(del_phi*ir));
					} 
					else 
					{
						ploc = px+Vector3D(0.,-radius/tilexn*ix*cos(del_phi*ir),-radius/tilexn*ix*sin(del_phi*ir));
					}
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLz()*0.5);
					Vector3D pg = GetPos3D0D(ploc, def_scale);

			//		mbs->UO()<<"point(" << side << "," << ir << "," << ix << "): " << ploc << " --> " << ploc_scaled << "-->" << pg << "\n";
					
					points.Add(pg);


					int index;
					
					//note that -Ly/2 corresponds to ix==0 at side 5 (which is the starting point for the averaging)
					double vLeft, vRight;
	
					switch (side)
					{
					//cylinder
					case 1: 
						//bottom: along line with iy=0 (iy (on bottom side) --> ix (on left side))					
						index = needsDivisionFactor*((tilern+1)*tilexn + ir) + 1;
						v = ((tilexn - ix) * leftVals(index) + ix*rightVals(index))/tilexn;
						break;
					//left
					case 2: 						
						//top: along line with iy=tileyn (iy (on top side) --> ix (on left side))
						//on top y-direction runs positive while on reference side (side 5) y runs in negative direction -> revert index
						index = needsDivisionFactor*(ix*(tilern+1) + ir) + 1;
						v = leftVals(index);
						break;
					//right
					case 3:
						//back: along line with ix=0 (iy (on front side) --> iy (on left side))
						//vertical coordinate runs in negative direction
						index = needsDivisionFactor*(ir + (tilern+1)*ix) + 1;
						v = rightVals(index);
						break;
					}
					vals.Add(v);
				}
			}
			// note: using linemode variable here gives bad result (outline of the quadstrip: top and bottom circle but also single edge on the surface)
			//in case of linemode == 2 no quads are drawn anyway
			if(linemode == 2)
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,linemode);
			}
			else
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,0/*linemode*/);
				if (linemode != 0)
				{
					for (double ir = 0; ir <= tilern+1e-10; ir++) // draw lines last for better visibility
					{
						Vector3D p1loc = px+Vector3D(0.,-cos(del_phi* ir   ),-sin(del_phi* ir   ));
						Vector3D p2loc = px+Vector3D(0.,-cos(del_phi*(ir+1)),-sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness);
					}
				}
			}
		}
	}
	else if (bp.GetCrossSectionShape() == BeamCrossSectionInfo::TCSS_tubular)
	{
		size.Y() = bp.GetBeamCrossSectionSize(1);
		size.Z() = bp.GetBeamCrossSectionSize(1);

		double radius = 0.5*bp.GetBeamCrossSectionSize(1)*GetMBS()->GetMagnifyYZ();
		double inner_radius = 0.5*bp.GetBeamCrossSectionSize(2)*GetMBS()->GetMagnifyYZ();

		double tilex = GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionAxis(); //formerly GetIOption(137)
		double tiler =  Sqr(max(GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourResolutionCrossSection(), 2));  //formerly GetIOption(138)
		TArray<Vector3D> points((int)(tilex+1)*(int)(tiler+1));
		TArray<double> vals((int)(tilex+1)*(int)(tiler+1));
		double v=0;

		for (int side=1; side <= 4; side++)
		{
			if (GetOuterFaceFlag(side) == 0) continue;
			points.SetLen(0); vals.SetLen(0);
			Vector3D px, vx, vr;
			int tilern = (int)tiler;
			int tilexn = (int)tilex;
			
			switch(side)
			{
			case 1: //outer cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			case 2: //left side
				px = Vector3D(-lx1,0., 0.); break;
			case 3: //right side
				px = Vector3D(lx1,0., 0.); break;
			case 4: //inner cylinder
				px = Vector3D(-lx1,0.,0.);
				vx = Vector3D(2.*lx1/tilexn,0,0); break;
			default: ;
			}

			double del_phi = 2.*MY_PI/tilern;
			for (double ix = 0; ix <= tilexn+1e-10; ix++)
			{
				for (double ir = 0; ir <= tilern+1e-10; ir++)
				{
					Vector3D ploc;
					double inner_radius = bp.GetBeamCrossSectionSize(2)/bp.GetBeamCrossSectionSize(1); // inner radius in local coordinates
					double radius = 1;                       // outer radius in local coordinates is always one
					if (side==1) 
					{
						ploc = px+ix*vx+Vector3D(0.,-radius*cos(del_phi*ir),-radius*sin(del_phi*ir));
					} 
					else if (side==4)
					{
						ploc = px+ix*vx+Vector3D(0.,-inner_radius*cos(del_phi*ir),-inner_radius*sin(del_phi*ir));						
					}
					else 
					{
						ploc = px+Vector3D(0.,-(inner_radius+(radius-inner_radius)/tilexn*ix)*cos(del_phi*ir),-(inner_radius+(radius-inner_radius)/tilexn*ix)*sin(del_phi*ir));
					}
					Vector3D ploc_scaled(ploc(1)*GetLx()*0.5, ploc(2)*GetLy()*0.5, ploc(3)*GetLy()*0.5);		
					Vector3D pg = GetPos3D0D(ploc, def_scale);		

					points.Add(pg);

					int index;
					
					//note that -Ly/2 corresponds to ix==0 at side 5 (which is the starting point for the averaging)
					double vLeft, vRight;
	
					switch (side)
					{
					//outer cylinder
					case 1: 
						//bottom: along line with iy=0 (iy (on bottom side) --> ix (on left side))					
						index = needsDivisionFactor*((tilern+1)*tilexn + ir) + 1;
						v = ((tilexn - ix) * leftVals(index) + ix*rightVals(index))/tilexn;
						break;
					//left
					case 2: 						
						//top: along line with iy=tileyn (iy (on top side) --> ix (on left side))
						//on top y-direction runs positive while on reference side (side 5) y runs in negative direction -> revert index
						index = needsDivisionFactor*(ix*(tilern+1) + ir) + 1;
						v = leftVals(index);
						break;
					//right
					case 3:
						//back: along line with ix=0 (iy (on front side) --> iy (on left side))
						//vertical coordinate runs in negative direction
						index = needsDivisionFactor*(ir + (tilern+1)*ix) + 1;
						v = rightVals(index);
						break;
					//inner cylinder
					case 4: 
						//bottom: along line with iy=0 (iy (on bottom side) --> ix (on left side))					
						index = needsDivisionFactor*ir + 1;
						v = ((tilexn - ix) * leftVals(index) + ix*rightVals(index))/tilexn;
						break;
					}
					vals.Add(v);
				}
			}
			// note: using linemode variable here gives bad result (outline of the quadstrip: top and bottom circle but also single edge on the surface)
			//in case of linemode == 2 no quads are drawn anyway
			if(linemode == 2)
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,linemode);		
			}
			else
			{
				mbs->Graphics3D()->DrawColorQuads(points,vals,(int)tilern+1,(int)tilexn+1,colormode,0/*linemode*/);		
				if (linemode != 0)
				{
					double ri = bp.GetBeamCrossSectionSize(2)/bp.GetBeamCrossSectionSize(1);
					for (double ir = 0; ir <= tilern+1e-10; ir++) // draw lines last for better visibility
					{
						Vector3D p1loc = px+Vector3D(0.,-cos(del_phi* ir   ),-sin(del_phi* ir   ));
						Vector3D p2loc = px+Vector3D(0.,-cos(del_phi*(ir+1)),-sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness); // outer surface of the tube

						p1loc = px+Vector3D(0.,-ri*cos(del_phi* ir   ),-ri*sin(del_phi* ir   ));
						p2loc = px+Vector3D(0.,-ri*cos(del_phi*(ir+1)),-ri*sin(del_phi*(ir+1)));
						mbs->Graphics3D()->MyDrawLine(GetPos3D0D(p1loc, def_scale), GetPos3D0D(p2loc, def_scale), thickness); // inner surface of the tube
					}
				}
			}
		}
	}

	else
	{
		assert(0);
	}
};

int ANCFBeamShear3DGeneric::GetLineMode()
{
	//0=no lines, 1=outline+color, 2=outline, 3=elementline+color, 4=elementline
	//old GetMBS()->GetIOption(110) ---> GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsMeshShow()
	//old GetMBS()->GetIOption(111) ---> GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourActivate()
	
	if (GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsMeshShow() && 
		!GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourActivate())
	{
		return 2;
	}
	if (!GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsMeshShow() && 
		GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourActivate())
	{
		return 0;
	}
	if (!GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsMeshShow() && 
		!GetMBS()->GetOptions()->ViewingOptions()->FiniteElementsContourActivate())
	{
		return 4;
	}
	return 1;
}

TArray<double> ANCFBeamShear3DGeneric::DetermineAveragedFVValues(TArray<int>* node2elem)
{
	if (node2elem->Length() > 1)
	{
		ANCFBeamShear3DGeneric* neighbor;
		if ((*node2elem)(1) == GetOwnNum()) 
		{
			neighbor = (ANCFBeamShear3DGeneric*)(& GetMBS()->GetElement((*node2elem)(2)));
		}
		else
		{
			neighbor = (ANCFBeamShear3DGeneric*)(& GetMBS()->GetElement((*node2elem)(1)));
		}
		//get attached element
		if (neighbor == 0)
		{
			mbs->UO(UO_LVL_warn) << "WARNING in ANCFBeamShear3DGeneric: DrawElementInterpolated - connected element is not a ANCFBeamShear3D.\n";
			return TArray<double>();
		}		
		TArray<double> vals(neighbor->GetAveragedFVvaluesForDrawing());
		//for (int i = 1; i <= rightVals->Length(); ++i)
		for (int i = 1; i <= vals.Length(); ++i)
		{
			vals(i) += averagedFVvaluesForDrawing(i);
			vals(i) /= 2.;
		}
		return vals;
	}
	else
	{
		//else use the actual values
		return GetAveragedFVvaluesForDrawing();
	}
}



#pragma endregion

#pragma region ANCFBeamShear3DLinear

void ANCFBeamShear3DLinear::ElementDefaultConstructor() 
{
	// set nodes
	nnodes=2;
	n1=1; n2=2;
	
	//set reference configuration
	q0.SetLen(SOS());
	q0.SetAll(0.);
	size = Vector3D(1,0.1,0.1);
	penalty.SetLen(9);
	penalty.SetAll(1);
	x_init.SetLen(2*SOS());
	x_init.SetAll(0.);

	xg.SetLen(SOS());
	xg.SetAll(0.);
	
	is_straight_beam_in_reference_configuration = 0;
	actFieldVariableValue = 0.;
}

// default set function!!!
void ANCFBeamShear3DLinear::SetANCFBeamShear3DLinear(int n1i, int n2i, int materialnumi, const Vector3D& coli, int beam_formulation)
{
	//ElementDefaultConstructor(); //$ PG 2018-7-5: should not be called in set phase

	// set nodes
	n1=n1i; n2=n2i;

	// set common stuff for ancfbeamshear3d elements
	SetANCFBeamShear3DGeneric(materialnumi, coli, beam_formulation);

	//used for computation speedup: test if beam element is straight in reference configuration
	is_straight_beam_in_reference_configuration = 1; // linear (2-noded) element is always straight, only cross section may vary
};

// deprecated set function!!!
void ANCFBeamShear3DLinear::SetANCFBeamShear3DLinear(const Vector& xc1, const Vector& xc2, int n1i, int n2i, int materialnumi,
																				 const Vector3D& si, const Vector3D& coli)
{
	//ElementDefaultConstructor(); //$ PG 2018-7-5: should not be called in set phase
	// set nodes
	//nnodes=2;		// is now in ElementDefaultConstructorInitialization
	n1=n1i; n2=n2i;
	
	//set reference configuration
	//q0.SetLen(SOS()); // is now in ElementDefaultConstructorInitialization
	//q0.SetAll(0.); // is now in ElementDefaultConstructorInitialization
	int ndofs_per_node = SOS()/nnodes;    //each of the nodes owns ndofs_per_node degrees of freedom
	for(int i=1;i<=ndofs_per_node;i++)   
	{
		q0(i) = xc1(i);
		q0(i+ndofs_per_node) = xc2(i);
	}

	// set common stuff for ancfbeamshear3d elements
	SetANCFBeamShear3DGeneric(materialnumi, si, coli);

	//used for computation speedup: test if beam element is straight in reference configuration
	is_straight_beam_in_reference_configuration = 1; // linear (2-noded) element is always straight, only cross section may vary
};

void ANCFBeamShear3DLinear::LinkToElements()
{
	if (SOSowned() == 0)
	{
		//UO() << "Link nodes to elements in ANCFBeamShear3DLinear\n";
		const Node& node1 = GetMBS()->GetNode(n1);
		const Node& node2 = GetMBS()->GetNode(n2);
		for (int i=1; i <= node1.SOS(); i++)
		{
			AddLTG(node1.Get(i));
		}
		for (int i=1; i <= node2.SOS(); i++)
		{
			AddLTG(node2.Get(i));
		}
		for (int i=1; i <= node1.SOS(); i++)
		{
			AddLTG(node1.Get(i+node1.SOS()));
		}
		for (int i=1; i <= node2.SOS(); i++)
		{
			AddLTG(node2.Get(i+node2.SOS()));
		}
	}
}

//-------------------------------------------------------
//Shapefunctions
//
//defined on scaled rectangular element: ploc=(\xi,\eta,\zeta)
//-L/2<\xi<+L/2, -H/2<\eta<+H/2, -B/2<\zeta<+B/2
//-------------------------------------------------------


void ANCFBeamShear3DLinear::GetShapes(Vector& sf, const Vector3D& ploc) const 
//compute all Shapefunctions at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = ploc(1)/GetLx();
	double sf1 = .5 - fact;
	double sf4 = .5 + fact;
	sf(1) = sf1;
	sf(2) = sf1*ploc(2);
	sf(3) = sf1*ploc(3);
	sf(4) = sf4;
	sf(5) = sf4*ploc(2);
	sf(6) = sf4*ploc(3);
}

double ANCFBeamShear3DLinear::GetSF(int i, const Vector3D& ploc) const
//computes i-th Shapefunction at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return (.5 - ploc(1)/GetLx());
	case 2: return (.5 - ploc(1)/GetLx())*ploc(2);
	case 3: return (.5 - ploc(1)/GetLx())*ploc(3);
	case 4: return (.5 + ploc(1)/GetLx());
	case 5: return (.5 + ploc(1)/GetLx())*ploc(2);
	case 6: return (.5 + ploc(1)/GetLx())*ploc(3);
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DLinear::GetShapesx(Vector& sf, const Vector3D& ploc) const 
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = 1./GetLx();
	sf(1) = -fact;
	sf(2) = -fact*ploc(2);
	sf(3) = -fact*ploc(3);
	sf(4) = fact;
	sf(5) = fact*ploc(2);
	sf(6) = fact*ploc(3);
}

double ANCFBeamShear3DLinear::GetSFx(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(1)=xi at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return -1./GetLx();
	case 2: return -ploc(2)/GetLx();
	case 3: return -ploc(3)/GetLx();
	case 4: return 1./GetLx();
	case 5: return ploc(2)/GetLx();
	case 6: return ploc(3)/GetLx();
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DLinear::GetShapesy(Vector& sf, const Vector3D& ploc) const 
//compute derivative of all Shapefunctions w.r.t. ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = ploc(1)/GetLx();
	sf(1) = 0.;
	sf(2) = .5 - fact;
	sf(3) = 0.;
	sf(4) = 0.;
	sf(5) = .5 + fact;
	sf(6) = 0.;
}

double ANCFBeamShear3DLinear::GetSFy(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return 0.;
	case 2: return .5 - ploc(1)/GetLx();
	case 3: return 0.;
	case 4: return 0.;
	case 5: return .5 + ploc(1)/GetLx();
	case 6: return 0.;
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DLinear::GetShapesz(Vector& sf, const Vector3D& ploc) const 
//compute derivative of all Shapefunctions w.r.t. ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = ploc(1)/GetLx();
	sf(1) = 0.;
	sf(2) = 0.;
	sf(3) = .5 - fact;
	sf(4) = 0.;
	sf(5) = 0.;
	sf(6) = .5 + fact;
}

double ANCFBeamShear3DLinear::GetSFz(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return 0.;
	case 2: return 0.;
	case 3: return .5 - ploc(1)/GetLx();
	case 4: return 0.;
	case 5: return 0.;
	case 6: return .5 + ploc(1)/GetLx();
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DLinear::GetShapesxy(Vector& sf, const Vector3D& ploc) const 
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi and ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = 1./GetLx();
	sf(1) = 0.;
	sf(2) = -fact;
	sf(3) = 0.;
	sf(4) = 0.;
	sf(5) = fact;
	sf(6) = 0.;
}

double ANCFBeamShear3DLinear::GetSFxy(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. plox(1)=xi and ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return 0.;
	case 2: return -1./GetLx();
	case 3: return 0.;
	case 4: return 0.;
	case 5: return 1./GetLx();
	case 6: return 0.;
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DLinear::GetShapesxz(Vector& sf, const Vector3D& ploc) const 
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi and ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());
	double fact = 1./GetLx();
	sf(1) = 0.;
	sf(2) = 0.;
	sf(3) = -fact;
	sf(4) = 0.;
	sf(5) = 0.;
	sf(6) = fact;
}

double ANCFBeamShear3DLinear::GetSFxz(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. plox(1)=xi and ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return 0.;
	case 2: return 0.;
	case 3: return -1./GetLx();
	case 4: return 0.;
	case 5: return 0.;
	case 6: return 1./GetLx();
	default: mbs->UO()<<"only 6 Shapefunctions\n"; return 0.;
	}
	return 0.;
}
#pragma endregion



#pragma region ANCFBeamShear3DQuadratic

void ANCFBeamShear3DQuadratic::ElementDefaultConstructor() 
{
	// set nodes
	nnodes=3;
	n1=1; n2=2; n3 = 3;
	
	//set reference configuration
	q0.SetLen(SOS());
	q0.SetAll(0.);

	//lx = 1; 
	//ly = 0.1;
	//lz = 0.1;
	size = Vector3D(1,0.1,0.1);

	penalty.SetLen(9);
	penalty.SetAll(1);

	x_init.SetLen(2*SOS());
	x_init.SetAll(0.);

	xg.SetLen(SOS());
	xg.SetAll(0.);
	
	is_straight_beam_in_reference_configuration = 0;
}

// default set function
void ANCFBeamShear3DQuadratic::SetANCFBeamShear3DQuadratic(int n1i, int n2i, int n3i, int materialnumi, const Vector3D& coli, int beam_formulation)
{
	//ElementDefaultConstructor(); //$ PG 2018-7-5: should not be called in set phase
	// set nodes
	//nnodes=3;	// is now in ElementDefaultConstructorInitialization()
	n1=n1i; n2=n2i, n3=n3i;
	
	// set common stuff for ancfbeamshear3d elements
	SetANCFBeamShear3DGeneric(materialnumi, coli, beam_formulation);

	//used for computation speedup: test if beam element is straight in reference configuration
	Vector3D pos_left(q0(1), q0(2), q0(3));
	Vector3D pos_right(q0(10), q0(11), q0(12));
	Vector3D pos_mid(q0(19), q0(20), q0(21));
	Vector3D v1 = pos_right - pos_mid;
	Vector3D v2 = pos_mid - pos_left;
	v1.Normalize();
	v2.Normalize();
	is_straight_beam_in_reference_configuration = (v1*v2 > 1-1e-14);   // if both sections of the beam are colinear, then the scalar product of v1 and v2 numerically yields 1
};

// deprecated set function!!!
void ANCFBeamShear3DQuadratic::SetANCFBeamShear3DQuadratic(const Vector& xc1, const Vector& xc2, const Vector& xc3, int n1i, int n2i, int n3i, int materialnumi,
																				 const Vector3D& si, const Vector3D& coli)
{
	//ElementDefaultConstructor(); //$ PG 2018-7-5: should not be called in set phase
	// set nodes
	//nnodes=3;	// is now in ElementDefaultConstructorInitialization()
	n1=n1i; n2=n2i, n3=n3i;
	
	//set reference configuration
	//q0.SetLen(SOS()); // is now in ElementDefaultConstructorInitialization()
	//q0.SetAll(0.); // is now in ElementDefaultConstructorInitialization()
	int ndofs_per_node = SOS()/nnodes;    //each of the nodes owns ndofs_per_node degrees of freedom
	for(int i=1;i<=ndofs_per_node;i++)   
	{
		q0(i) = xc1(i);
		q0(i+ndofs_per_node) = xc2(i);
		q0(i+2*ndofs_per_node) = xc3(i);
	}

	//used for computation speedup: test if beam element is straight in reference configuration
	Vector3D pos_left(q0(1), q0(2), q0(3));
	Vector3D pos_right(q0(10), q0(11), q0(12));
	Vector3D pos_mid(q0(10), q0(11), q0(12));
	Vector3D v1 = pos_right - pos_mid;
	Vector3D v2 = pos_mid - pos_left;
	v1.Normalize();
	v2.Normalize();
	is_straight_beam_in_reference_configuration = (v1*v2 > 1-1e-14);   // if both sections of the beam are colinear, then this scalar product gives numerically 1

	// set common stuff for ancfbeamshear3d elements
	SetANCFBeamShear3DGeneric(materialnumi, si, coli);
};

void ANCFBeamShear3DQuadratic::LinkToElements()
{
	if (SOSowned() == 0)
	{
		//UO() << "Link nodes to elements in ANCFBeamShear3DQuadratic\n";
		const Node& node1 = GetMBS()->GetNode(n1);
		const Node& node2 = GetMBS()->GetNode(n2);
		const Node& node3 = GetMBS()->GetNode(n3);
		for (int i=1; i <= node1.SOS(); i++)
		{
			AddLTG(node1.Get(i));
		}
		for (int i=1; i <= node2.SOS(); i++)
		{
			AddLTG(node2.Get(i));
		}
		for (int i=1; i <= node3.SOS(); i++)
		{
			AddLTG(node3.Get(i));
		}
		for (int i=1; i <= node1.SOS(); i++)
		{
			AddLTG(node1.Get(i+node1.SOS()));
		}
		for (int i=1; i <= node2.SOS(); i++)
		{
			AddLTG(node2.Get(i+node2.SOS()));
		}
		for (int i=1; i <= node3.SOS(); i++)
		{
			AddLTG(node3.Get(i+node3.SOS()));
		}
	}
}

//-------------------------------------------------------
//Shapefunctions
//
//defined on scaled rectangular element: ploc=(\xi,\eta,\zeta)
//-L/2<\xi<+L/2, -H/2<\eta<+H/2, -B/2<\zeta<+B/2
//-------------------------------------------------------


void ANCFBeamShear3DQuadratic::GetShapes(Vector& sf, const Vector3D& ploc) const
//compute all Shapefunctions at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact=2./(GetLx()*GetLx());
	sf(1)=-fact*(GetLx()/2.-ploc(1))*ploc(1);
	sf(2)=sf(1)*ploc(2);
	sf(3)=sf(1)*ploc(3);
	sf(4)=fact*(GetLx()/2.+ploc(1))*ploc(1);
	sf(5)=sf(4)*ploc(2);
	sf(6)=sf(4)*ploc(3);
	sf(7)=-2.*fact*(GetLx()/2.+ploc(1))*(ploc(1)-GetLx()/2.);
	sf(8)=sf(7)*ploc(2);
	sf(9)=sf(7)*ploc(3);
}

double ANCFBeamShear3DQuadratic::GetSF(int i, const Vector3D& ploc) const
//computes i-th Shapefunction at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{ 
	switch(i)
	{
	case 1: return -2./(GetLx()*GetLx())*ploc(1)*(GetLx()/2.-ploc(1));
	case 2: return -2./(GetLx()*GetLx())*ploc(1)*ploc(2)*(GetLx()/2.-ploc(1));
	case 3: return -2./(GetLx()*GetLx())*ploc(1)*ploc(3)*(GetLx()/2.-ploc(1));
	case 4: return 2./(GetLx()*GetLx())*ploc(1)*(GetLx()/2.+ploc(1));
	case 5: return 2./(GetLx()*GetLx())*ploc(1)*ploc(2)*(GetLx()/2.+ploc(1));
	case 6: return 2./(GetLx()*GetLx())*ploc(1)*ploc(3)*(GetLx()/2.+ploc(1));
	case 7: return -4./(GetLx()*GetLx())*(ploc(1)-GetLx()/2.)*(ploc(1)+GetLx()/2.);
	case 8: return -4./(GetLx()*GetLx())*ploc(2)*(ploc(1)-GetLx()/2.)*(ploc(1)+GetLx()/2.);
	case 9: return -4./(GetLx()*GetLx())*ploc(3)*(ploc(1)-GetLx()/2.)*(ploc(1)+GetLx()/2.);
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DQuadratic::GetShapesx(Vector& sf, const Vector3D& ploc) const
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact = 4.*ploc(1)/(GetLx()*GetLx());
	double sf1 = -1./GetLx()+fact;
	double sf4 = 1./GetLx()+fact;
	double sf7 = -2.*fact;
	sf(1) = sf1;
	sf(2) = sf1*ploc(2);
	sf(3) = sf1*ploc(3);
	sf(4) = sf4;
	sf(5) = sf4*ploc(2);
	sf(6) = sf4*ploc(3);
	sf(7) = sf7;
	sf(8) = sf7*ploc(2);
	sf(9) = sf7*ploc(3);
}

double ANCFBeamShear3DQuadratic::GetSFx(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(1)=xi at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{
	switch(i)
	{
	case 1: return -1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 2: return -ploc(2)/GetLx()+4.*ploc(1)*ploc(2)/(GetLx()*GetLx());
	case 3: return -ploc(3)/GetLx()+4.*ploc(1)*ploc(3)/(GetLx()*GetLx());
	case 4: return 1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 5: return ploc(2)/GetLx()+4.*ploc(1)*ploc(2)/(GetLx()*GetLx());
	case 6: return ploc(3)/GetLx()+4.*ploc(1)*ploc(3)/(GetLx()*GetLx());
	case 7: return -8.*ploc(1)/(GetLx()*GetLx());
	case 8: return -8.*ploc(1)*ploc(2)/(GetLx()*GetLx());
	case 9: return -8.*ploc(1)*ploc(3)/(GetLx()*GetLx());
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DQuadratic::GetShapesy(Vector& sf, const Vector3D& ploc) const
//compute derivative of all Shapefunctions w.r.t. ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact1 = ploc(1)/GetLx();
	double fact2 = 2.*fact1*fact1;
	sf(1) = 0;
	sf(2) = -fact1 + fact2;
	sf(3) = 0;
	sf(4) = 0;
	sf(5) = fact1 + fact2;
	sf(6) = 0;
	sf(7) = 0;
	sf(8) = 1. - 2.*fact2;
	sf(9) = 0;
}

double ANCFBeamShear3DQuadratic::GetSFy(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{
	switch(i)
	{
	case 1: return 0;
	case 2: return -ploc(1)/GetLx()+2.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	case 3: return 0;
	case 4: return 0;
	case 5: return ploc(1)/GetLx()+2.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	case 6: return 0;
	case 7: return 0;
	case 8: return 1.-4.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	case 9: return 0;
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DQuadratic::GetShapesxy(Vector& sf, const Vector3D& ploc) const
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi and ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact1 = 1./GetLx();
	double fact2 = 4.*ploc(1)*fact1*fact1;
	sf(1) = 0;
	sf(2) = -fact1 + fact2;
	sf(3) = 0;
	sf(4) = 0;
	sf(5) = fact1 + fact2;
	sf(6) = 0;
	sf(7) = 0;
	sf(8) = -2.*fact2;
	sf(9) = 0;
}

double ANCFBeamShear3DQuadratic::GetSFxy(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(1)=xi and ploc(2)=eta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{
	switch(i)
	{
	case 1: return 0;
	case 2: return -1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 3: return 0;
	case 4: return 0;
	case 5: return 1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 6: return 0;
	case 7: return 0;
	case 8: return -8.*ploc(1)/(GetLx()*GetLx());
	case 9: return 0;
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DQuadratic::GetShapesz(Vector& sf, const Vector3D& ploc) const
//compute derivative of all Shapefunctions w.r.t. ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact1 = ploc(1)/GetLx();
	double fact2 = 2.*fact1*fact1;
	sf(1) = 0;
	sf(2) = 0;
	sf(3) = -fact1 + fact2;
	sf(4) = 0;
	sf(5) = 0;
	sf(6) = fact1 + fact2;
	sf(7) = 0;
	sf(8) = 0;
	sf(9) = 1. - 2.*fact2;
}

double ANCFBeamShear3DQuadratic::GetSFz(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2] 
{
	switch(i)
	{
	case 1: return 0;
	case 2: return 0;
	case 3: return -ploc(1)/GetLx()+2.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	case 4: return 0;
	case 5: return 0;
	case 6: return ploc(1)/GetLx()+2.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	case 7: return 0;
	case 8: return 0;
	case 9: return 1.-4.*ploc(1)*ploc(1)/(GetLx()*GetLx());
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

void ANCFBeamShear3DQuadratic::GetShapesxz(Vector& sf, const Vector3D& ploc) const
//compute derivative of all Shapefunctions w.r.t. ploc(1)=xi and ploc(3)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{ 
	sf.SetLen(NS());  //NS=3*nnodes, vector sf has 9 entries
	double fact1 = 1./GetLx();
	double fact2 = 4.*ploc(1)*fact1*fact1;
	sf(1) = 0;
	sf(2) = 0;
	sf(3) = -fact1 + fact2;
	sf(4) = 0;
	sf(5) = 0;
	sf(6) = fact1 + fact2;
	sf(7) = 0;
	sf(8) = 0;
	sf(9) = -2.*fact2;
}

double ANCFBeamShear3DQuadratic::GetSFxz(int i, const Vector3D& ploc) const
//computes derivative of i-th Shapefunction w.r.t. ploc(1)=xi and ploc(2)=zeta at ploc=(xi,eta,zeta) in [-L/2,L/2] x [-h/2,h/2] x [-w/2,w/2]
{
	switch(i)
	{
	case 1: return 0;
	case 2: return 0;
	case 3: return -1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 4: return 0;
	case 5: return 0;
	case 6: return 1./GetLx()+4.*ploc(1)/(GetLx()*GetLx());
	case 7: return 0;
	case 8: return 0;
	case 9: return -8.*ploc(1)/(GetLx()*GetLx());
	default: mbs->UO()<<"only 9 Shapefunctions\n"; return 0.;
	}
	return 0.;
}

#pragma endregion
