//#**************************************************************
//#
//# filename:             script_parser.cpp
//#
//# author:               Gerstmayr Johannes
//#
//# generated:						July 2004
//# description:          parser for creating objects in script language
//#                       
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#**************************************************************
#include "mbs_interface.h"
#include "element.h"
#include "options_class_auto.h"
#include "script_parser.h"
//$ YV 2012-12-14: the following include is to be removed; general object factory should be used in the future
//#include "sensorsSpecific.h"	//$ DR 2012-10: necessary to include for CEDCParser
#include "elementdataaccess.h"
#include "graphicsconstants.h"
#include "material.h"
#include "geomelements.h"
#include "femathhelperfunctions.h"
#include "sensors.h"
#include "parser.h"
#include "elementsandmodelslibraryinterface.h"
#include "node.h"
#include "set.h"
#include "mesh.h"
#include "Windows.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++EvaluateTextualParameterEntry++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::EvaluateTextualParameterEntry(MBS* mbs, CEDCParser* edcParser, ElementDataContainer* param_edc)
{
// convert any textual entry to scalar,vector,matrix
// assume that the parameter EDC has only a single entry
	ElementData* edp = param_edc->GetPtr(1);
	if(edp->IsText())
	{
		mystr entry = mystr(edp->GetDataName()) + mystr("= ") + mystr(edp->GetText());
		ElementDataContainer tmp;

		// evaluate the textual entry
		edcParser->ParseAndExecuteString(entry,tmp);

		param_edc->Delete(1);
		param_edc->TreeReplaceEDCDataWith(&tmp);

		return 1;
	}
	return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++ResolveParameterEntryToSetNumber+++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::ResolveParameterEntryToSetNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed)
{
// resolves single-entry EDC to the set number 
	int rv = 0;
// the entry may contain: a) Integer Variable, b) Text Numeral (inline definition), c) Text Name
	ElementData* edp = param_ed;
	if(edp->IsInt())
	{
		rv = edp->GetInt(); // integer variabe
	}
	else if(edp->IsDouble())
	{
		rv = (int) edp->GetDouble();
	}
	else if(edp->IsText())
	{
		mystr str = edp->GetText();
		if(str.IsValidNumber())
		{
			rv = str.MakeInt(); // inline number
		}
		else
		{
		// VERY BAD - VERY SLOW - Find() Method
			int nsets = mbs->NSets();
			for(int i=1; i<=nsets; i++)
			{
				int cmp = str.Compare(mbs->GetSetPtr(i)->GetName());
				if(cmp) // ==1 when steings are identical
				{
					rv = i;
					break;
				}
			}
			if(rv==0)
				edcParser->EDCError(mystr("ERROR: no Set with name \"") + str + mystr("\"was found in the list of sets. Check spelling."));
		}
	}
	else 
	{
		;
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++ResolveParameterEntryToMaterialNumber+++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::ResolveParameterEntryToMaterialNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed)
{
// resolves single-entry EDC to the material number 
	int rv = 0;
// the entry may contain: a) Integer Variable, b) Text Numeral (inline definition), c) Text Name
	ElementData* edp = param_ed;
	if(edp->IsInt())
	{
		rv = edp->GetInt(); // integer variabe
	}
	else if(edp->IsText())
	{
		mystr str = edp->GetText();
		if(str.IsValidNumber())
		{
			rv = str.MakeInt(); // inline number
		}
		else
		{
		// VERY BAD - VERY SLOW - Find() Method
			int nmats = mbs->NMaterials();
			for(int i=1; i<=nmats; i++)
			{
				int cmp = str.Compare(mbs->GetMaterialPtr(i)->GetMaterialName());
				if(cmp) // ==1 when steings are identical
				{
					rv = i;
					break;;
				}
			}
			if(rv==0)
				edcParser->EDCError(mystr("ERROR: no Material with name \"") + str + mystr("\"was found in the list of materials.  Check spelling."));
		}
	}
	else 
	{
		;
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++ResolveParameterEntryToLoadNumber+++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::ResolveParameterEntryToLoadNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed)
{
// resolves single-entry EDC to the load number 
	int rv = 0;
// the entry may contain: a) Integer Variable, b) Text Numeral (inline definition), c) Text Name
	ElementData* edp = param_ed;
	if(edp->IsInt())
	{
		rv = edp->GetInt(); // integer variabe
	}
	else if(edp->IsText())
	{
		mystr str = edp->GetText();
		if(str.IsValidNumber())
		{
			rv = str.MakeInt(); // inline number
		}
		else
		{
		// VERY BAD - VERY SLOW - Find() Method
			int nloads = mbs->NLoads();
			for(int i=1; i<=nloads; i++)
			{
				int cmp = str.Compare(mbs->GetLoadPtr(i)->LoadName());
				if(cmp) // ==1 when steings are identical
				{
					rv = i;
					break;
				}
			}
			if(rv==0)
				edcParser->EDCError(mystr("ERROR: no Load with name \"") + str + mystr("\"was found in the list of loads. Check spelling!"));
		}
	}
	else 
	{
		;
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++ResolveParameterEntryToMeshNumber+++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::ResolveParameterEntryToMeshNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed)
{
// resolves single-entry EDC to the mesh number 
	int rv = 0;

// the entry may contain:  c) Text Name ( name of the handle )
	ElementData* edp = param_ed;
	if(edp->IsText())
	{
		mystr str = edp->GetText();
	// VERY BAD - VERY SLOW - Find() Method
		int nmeshes = mbs->NMeshes();
		for(int i=1; i<=nmeshes; i++)
		{
			int cmp = str.Compare(mbs->GetMeshPtr(i)->GetMeshName());
			if(cmp) // ==1 when strings are identical
			{
				rv = i;
				break;
			}
		}
		if(rv==0)
			edcParser->EDCError(mystr("ERROR: no Mesh with name \"") + str + mystr("\"was found in the list of meshes"));
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++ResolveParameterEntryToComponentNumber++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_Command::ResolveParameterEntryToComponentNumber(MBS* mbs, CEDCParser* edcParser, ElementData* param_ed, int meshIdx)
{
	int rv = 0;
	ElementData* edp = param_ed;

	if(edp->IsInt())
	{
		rv = edp->GetInt(); // integer variabe
	}
	else if(edp->IsText())
	{
		mystr str = edp->GetText();
		if(str.IsValidNumber())
		{
			rv = str.MakeInt(); // inline number
		}
		else
		{
		// VERY BAD - VERY SLOW - Find() Method
			int ncomp = mbs->GetMeshPtr(meshIdx)->NComp();
			for(int i=1; i<=ncomp; i++)
			{
				int cmp = str.Compare(mbs->GetMeshPtr(meshIdx)->GetMeshComponent(i)->GetName());
				if(cmp) // ==1 when strings are identical
				{
					rv = i;
					break;
				}
			}
			if(rv==0)
				edcParser->EDCError(mystr("ERROR: no Component with name \"") + str + mystr("\"was found in the list of mesh components"));
		}
	}
	else 
	{
		;
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++ExecuteCommand ++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_ComAddElement::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option) 
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int element_type_index = edc->Find("element_type"); //find index in element EDC
	if (element_type_index)
	{
		if (!edc->Get(element_type_index).IsText()) {assert(0);}

		mystr element_type_str = edc->Get(element_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCElement, element_type_str);
		if (type_id < 0 ) 
		{
			edcParser->EDCError(mystr("Element type '") +  element_type_str + mystr("' not found in command AddElement(...)! Check spelling of element edc."));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		} //does not exist in element type list

		int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCElement, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + element_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			if (IsAddElement() && (typeflags & TAEconstraint)) //constraints are in the same list, but should not be added with "AddElement"
			{
				edcParser->EDCError(mystr("Connector '")+element_type_str+mystr("' can not be added with command AddElement(...), use AddConnector(...) instead!"));
			}
			else if (!IsAddElement() && !(typeflags & TAEconstraint)) //constraints are in the same list, but should not be added with "AddElement"
			{
				edcParser->EDCError(mystr("Element '")+element_type_str+mystr("' can not be added with command AddConnector(...), use AddElement(...) instead!"));
			}

			//add new element of selected type to MBS
			//int elnum = edcParser->AddElement(type_id, option); //old: AddElement(element_type_index, option);
			int elnum = edcParser->GetObjectFactory()->AddObject(OFCElement, type_id);

			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetElement(elnum).GetElementData(edc);
			
			//$ SW 2014-8-11: added check for EDC entries that are not found
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1),mbs->GetElement(elnum).GetElementSpec(), &CEDCParser::GetNewEDCName, 1);

			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetElement(elnum).SetElementData(edc);

			return_value.SetInt(elnum, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'element_type' in command ") + CommandName() + mystr(" Check spelling of element edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_ComAddGeomElement::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int hackDR = 1*0;	// test of command StartComputation
	if(hackDR)
	{
		mbs->UO().InstantMessageText("start auto comp");
		mbs->Assemble();
		//mbs->PerformSingleStaticDynamicComputation();
		mbs->UO().CallWCDriverFunction(30,1); //start
		mbs->UO().InstantMessageText("finished auto comp");

		return 1;
	}



	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int element_type_index = edc->Find("geom_element_type"); //find index in element EDC
	if (element_type_index)
	{
		if (!edc->Get(element_type_index).IsText()) {assert(0);}

		mystr element_type_str = edc->Get(element_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCGeomElement, element_type_str);
		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("Element type '") +  element_type_str + mystr("' not found in command AddGeomElement(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCGeomElement, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + element_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new element of selected type to MBS
			//int geom_elnum=edcParser->AddGeomElement(type_id,0); //Add GeomElement with default values
			int geom_elnum = edcParser->GetObjectFactory()->AddObject(OFCGeomElement, type_id); //Add GeomElement with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetDrawElement(geom_elnum)->GetElementData(edc);
			//now overwrite the default with the user-defined values
			//$ SW 2014-8-11: added check for EDC entries that are not found
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), element_type_str, &CEDCParser::GetNewEDCName, 1);
			//edc.TreeReplaceEDCDataWith(parameter_EDCs(1),1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetDrawElement(geom_elnum)->SetElementData(edc);

			return_value.SetInt(geom_elnum, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'geom_element_type' in command ") + CommandName() + mystr(" Check spelling of element edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//$ DR 2012-11

mystr CEDC_ComPrint::GetCommandTexParameterDescription() const 
{
	mystr descr;
	descr += "There are three possibilities to use the command. The parameter can either be: \\\\ \n";
	descr += mystr("\\begin{itemize} \n");
	descr+= mystr("  \\item a text, e.g. Print(\"Hello world\")  \n");
	descr+= mystr("  \\item an ElementDataContainer, e.g. Print(my\\_mass)  \n");
	descr+= mystr("  \\item an ElementData, e.g. Print(my\\_mass.density)  \n");
	descr += mystr("\\end{itemize} \n");
	descr += mystr("In the case of a text or an ElementData, only the text itself is printed. In the case of an ElementDataContainer, also the name of the ElementData is printed.");
	return descr;
}


int CEDC_ComPrint::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	mystr str, num;

	//symbols defined in CEDCParser:
	mystr el = "\n"; //end of line
	mystr ob = "{";  //open brace
	mystr cb = "}";  //closing brace
	mystr osb = "[";  //open square bracket
	mystr csb = "]";  //closing square bracket
	mystr ds = ": "; //double stop
	mystr semicolon = ";"; //semicolon
	char commentc = '%'; //comment
	mystr doublequote = mystr('"'); //for strings or text

	//symbols with additional whitespaces:
	mystr eq = "= "; //equal sign
	mystr comma = ", "; //comma
	mystr comment = mystr("  ")+mystr(commentc); //comment
	mystr matrixsep = el; //may be changed to semicolon ...


	ElementData ed = parameter_EDCs(1)->Get(1);
	if ((parameter_EDCs(1)->Length()==1) && (! ed.IsEDC()))		// if it is not an EDC, than do not write "text = my text" but only "my text"
	{
		if (ed.IsText()) 
		{
			str = parameter_EDCs(1)->Get(1).GetText();
			str.Replace("\\n","\n");
		}
		else if (ed.IsBool())
		{
			str += mystr(ed.GetBool());
		}
		else if (ed.IsInt())
		{
			str += mystr(ed.GetInt());
		}
		else if (ed.IsDouble())
		{
			num.SmartDouble2String(ed.GetDouble());
			str += num;
		}
		else if (ed.IsText())
		{
			str += mystr(doublequote) + mystr(ed.GetText()) + mystr(doublequote);
		}
		else if (ed.IsVector())
		{
			str += osb;
			for (int j = 1; j <= ed.GetVectorLen(); j++)
			{
				num.SmartDouble2String(ed.GetVectorVal(j));
				str += num;
				if (j < ed.GetVectorLen()) str += comma;
			}
			str += csb;
		}
		else if (ed.IsMatrix())
		{
			str += osb;
			for (int j1 = 1; j1 <= ed.GetMatrixRows(); j1++)
			{
				for (int j2 = 1; j2 <= ed.GetMatrixCols(); j2++)
				{
					num.SmartDouble2String(ed.GetMatrixVal(j1, j2));
					str += num;
					if (j2 < ed.GetMatrixCols()) str += comma;
				}
				if (j1 < ed.GetMatrixRows()) str += matrixsep + mystr("    ");
			}
			str += csb;
		}
	}
	else
	{
		ElementDataContainer edc;
		edc.CopyFrom(*parameter_EDCs(1));
		edcParser->EDC2String(edc, str,"");
	}
	//mbs->UO(UO_LVL_warn) << str;
	mbs->UO(UO_LVL_sim) << str; //$ DR 2015-05-05 removed "UO_LVL_0"

	return 1;
}

mystr CEDC_ComPrintIf::GetCommandTexParameterDescription() const 
{
	mystr descr;
	descr += "First parameter is a flag whether to print at all. \n";
	descr += "There are three possibilities to use the command. The 2nd parameter can either be: \\\\ \n";
	descr += mystr("\\begin{itemize} \n");
	descr+= mystr("  \\item a text, e.g. Print(\"Hello world\")  \n");
	descr+= mystr("  \\item an ElementDataContainer, e.g. Print(my\\_mass)  \n");
	descr+= mystr("  \\item an ElementData, e.g. Print(my\\_mass.density)  \n");
	descr += mystr("\\end{itemize} \n");
	descr += mystr("In the case of a text or an ElementData, only the text itself is printed. In the case of an ElementDataContainer, also the name of the ElementData is printed.");
	return descr;
}

int CEDC_ComPrintIf::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	// extract first parameter - must be a boolean
	ElementData ed = parameter_EDCs(1)->Get(1);

	int flagvalue = 0;
	if(ed.IsBool())
	{
		flagvalue = ed.GetBool();
	}
	else if(ed.IsInt())
	{
		flagvalue = ed.GetInt();
	}
	else if (ed.IsDouble())
	{
		flagvalue = ed.GetDouble();
	}
	else if(ed.IsText())
	{
		mystr str = ed.GetText();
		if(str.IsValidNumber())
		{
			flagvalue = str.MakeInt(); // inline number
		}		
		else
		{
			if(str.Compare("y")||str.Compare("yes")||str.Compare("True"))
			{
				flagvalue = 1;
			}
		}
	}
	if (flagvalue > 0)
	{
		ElementDataContainer* edc_pass = parameter_EDCs(2);
		TArray<ElementDataContainer*> parameter_EDCs_pass;
		parameter_EDCs_pass.Add(edc_pass);

		return CEDC_ComPrint::ExecuteCommand(mbs, edcParser, parameter_EDCs_pass, return_value, option);
	}
	else return 1;
}


int CEDC_ComInclude::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	if (!parameter_EDCs(1)->Get(1).IsText()) 
	{
		edcParser->EDCError(mystr("Error happened during parsing command ") + CommandName());
		return 0;
	}

	mystr param_text = parameter_EDCs(1)->Get(1).GetText(); // filename is absolute or alternatively relative to hid-file

	
	mystr filename = param_text; // param_text contains filename without double quotes

	int rv = 1; // success
	edcParser->MakeFilenameAbsolute(filename);		//$ DR 2013-07-04

	if (DoesFileExist(filename))
	{
		ElementDataContainer edc_file;
		if(mbs->GetModelDataContainer() == 0)
		{
			rv = mbs->ReadModelData(filename); // empty model data container --> create new one
		}
		else
		{
			rv = mbs->AddModelData(filename);  // model data container has contents --> add and replace
		}
	}
	else
	{
		edcParser->EDCError(mystr("Error in Include - command: Can not find 'filename' '") + filename + mystr("'!"));
		rv = 0;
	}
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2012-10
int CEDC_ComAddLoad::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int load_type_index = edc->Find("load_type"); //find index in element EDC
	if (load_type_index)
	{
		if (!edc->Get(load_type_index).IsText()) {assert(0);}

		mystr load_type_str = edc->Get(load_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCLoad, load_type_str);

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("Load type '") +  load_type_str + mystr("' not found in command AddLoad(...)!"));
			//AD: two obsolete Load Types {AreaLoad,AreaTraction) --> SurfaceLoad
			if(load_type_str.Compare("AreaLoad")) { edcParser->EDCError(mystr("Load type 'AreaLoad' is obsolete, please use 'SurfaceLoad' instead. ")); }
			if(load_type_str.Compare("AreaTraction")) { edcParser->EDCError(mystr("Load type 'AreaTraction' is obsolete, please use 'SurfaceLoad' instead. ")); }

			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCLoad, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + load_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new load of selected type to MBS
			int load_elnum = edcParser->GetObjectFactory()->AddObject(OFCLoad, type_id); //Add Load with default values

			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetLoad(load_elnum).GetElementData(edc);
			//now overwrite the default with the user-defined values
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), load_type_str, &CEDCParser::GetNewEDCName, 1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetLoad(load_elnum).SetElementData(edc);

			return_value.SetInt(load_elnum, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'load_type' in command ") + CommandName() + mystr(" Check spelling of load edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2012-10
int CEDC_ComAddSensor::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int sensor_type_index = edc->Find("sensor_type"); //find index in element EDC
	if (sensor_type_index)
	{
		if (!edc->Get(sensor_type_index).IsText()) {assert(0);}

		mystr sensor_type_str = edc->Get(sensor_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCSensor, sensor_type_str);

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("Sensor type '") +  sensor_type_str + mystr("' not found in command AddSensor(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCSensor, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + sensor_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new sensor of selected type to MBS
			int sensor_elnum = edcParser->GetObjectFactory()->AddObject(OFCSensor, type_id); //Add Load with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edcSensor;
			mbs->GetSensor(sensor_elnum).GetElementData(edcSensor);
//!AD: 2017-06-12: any entries for SensorProcessors woutld cause a warning popup here, prevent this by removing these entries for the treeReplace Operation:
			ElementDataContainer cutoutProcessors;
			int spidx = 1;
			mystr processorContainerName = mystr("Processor")+mystr(spidx);
			while(parameter_EDCs(1)->TreeFind(processorContainerName)!=NULL && parameter_EDCs(1)->TreeFind(processorContainerName)->IsEDC()) 
			{
				cutoutProcessors.Add(*(parameter_EDCs(1)->TreeFind(processorContainerName)));
				parameter_EDCs(1)->TreeDelete(processorContainerName);
				processorContainerName = mystr("Processor")+mystr(++spidx);
			}
			//now overwrite the default with the user-defined values
			edcSensor.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1),sensor_type_str, &CEDCParser::GetNewEDCName, 1);

			// and without warnings add the sensorprocessors
			edcSensor.TreeReplaceEDCDataWithReplacingOldNames(&cutoutProcessors,sensor_type_str, &CEDCParser::GetNewEDCName, 0/*no warning*/);

			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetSensor(sensor_elnum).SetElementData(edcSensor);

			return_value.SetInt(sensor_elnum, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'sensor_type' in command ") + CommandName() + mystr(" Check spelling of sensor edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2017-06
int CEDC_ComAddSensorProcessor::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;

	// second parameter is the number of the 
	ElementDataContainer* edcSensNr  = parameter_EDCs(2);
	// when parameters are defined inline the parameter EDC contains text.
	if(edcSensNr->GetPtr(1)->IsText())
		EvaluateTextualParameterEntry(mbs, edcParser, edcSensNr);
	int sensNr = edcSensNr->GetPtr(1)->GetInt();
	
	if(sensNr>mbs->NSensors())
	{
		edcParser->EDCError(mystr("can not access sensor in command ") + CommandName());
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;	
	}
	//Sensor* masterSensor = mbs->GetSensorPtr(sensNr);
	
	
	// first container is the EDC for the SensorProcessor
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int processor_type_index = edc->Find("processor_type"); //find index in element EDC
	if (processor_type_index)
	{
		if (!edc->Get(processor_type_index).IsText()) {assert(0);}

		mystr processor_type_str = edc->Get(processor_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCSensorProcessor, processor_type_str);		

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("SensorProcessor type '") +  processor_type_str + mystr("' not found in command AddSensorProcessor(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCSensorProcessor, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		{
			edcParser->EDCError(mystr("ERROR: ") + processor_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}
		else
		{
			int dummy = mbs->GetObjectFactory()->AddObject(OFCSensorProcessor,type_id,int2(sensNr,sensNr));
			Sensor* masterSensor = mbs->GetSensorPtr(sensNr);
			masterSensor->GetSensorProcessorPtr(dummy)->SetElementData(*parameter_EDCs(1));
			return dummy;
		}
	}

	return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2012-10
int CEDC_ComLoadSTL::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	if (!parameter_EDCs(1)->Get(1).IsText()) 
	{
		edcParser->EDCError(mystr("Error happened during parsing command ") + CommandName());
		return 0;
	}

	mystr param_text = parameter_EDCs(1)->Get(1).GetText(); // filename is absolute or alternatively relative to hid-file
	
	mystr filename = param_text; // param_text contains filename without double quotes

	int rv = 1; // success
	rv = edcParser->MakeFilenameAbsolute(filename);		//$ DR 2013-07-04

	ElementDataContainer edc;

	if(rv)
	{
		rv = mbs->File2EDC(filename, &edc);
		return_value.SetEDC(&edc,"MeshData");
	}
	return rv;

}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDC_ComAddMaterial::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int continuum_type_index = edc->Find("material_type"); //find index in element EDC
	mystr type = edc->TreeGetString("material_type");
	if (continuum_type_index)
	{
		if (!edc->Get(continuum_type_index).IsText()) {assert(0);}

		mystr mat_type_str = edc->Get(continuum_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCMaterial, mat_type_str);

		if (type_id < 0)
		{
			edcParser->EDCError(mystr("Material type '") +  mat_type_str + mystr("' not found in command AddMaterial(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCMaterial, type_id); //this are the typeflags, e.g. constraint, 2D, etc.

		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + mat_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new material to MBS
			int mat_num = edcParser->GetObjectFactory()->AddObject(OFCMaterial, type_id); //Add Material with default values
			//retrieve the default EDC for the new material
			ElementDataContainer edc;
			mbs->GetMaterialPtr(mat_num)->GetElementData(edc);
			//now overwrite the default with the user-defined values
			// [ $EK - 2018-05-23:
			//in case of SerialMaterial ommit warnings which are due to missing entries in the EDC
			//since some entries are not used/necessary in that case
			if (mat_type_str.Compare("MaterialSerial")==1)
			{
				edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), mat_type_str, &CEDCParser::GetNewEDCName, 0);
			}
			else
			{
				edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), mat_type_str, &CEDCParser::GetNewEDCName, 1);
			}
			//$EK - 2018-05-23 ]
			//now set the user-defined EDC in the newly generated material
			int rv = mbs->GetMaterial(mat_num).SetElementData(edc);

			return_value.SetInt(mat_num, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'material_type' in command ") + CommandName() + mystr(" Check spelling of material edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2013-01
int CEDC_ComAddBeam3DProperties::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int beam_property_type_index = edc->Find("material_type"); //find index in element EDC
	if (beam_property_type_index)
	{
		if (!edc->Get(beam_property_type_index).IsText()) {assert(0);}

		mystr beam_property_type_str = edc->Get(beam_property_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCBeamProperties, beam_property_type_str);

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("BeamProperty type '") +  beam_property_type_str + mystr("' not found in command AddBeamProperties(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCBeamProperties, type_id); //this are the typeflags, e.g. constraint, 2D, etc.
		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + beam_property_type_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new beam property of selected type to MBS
			int beamProp_matnum = edcParser->GetObjectFactory()->AddObject(OFCBeamProperties, type_id); //Add beam property with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetMaterial(beamProp_matnum).GetElementData(edc);
			//now overwrite the default with the user-defined values
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), beam_property_type_str, &CEDCParser::GetNewEDCName, 1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetMaterial(beamProp_matnum).SetElementData(edc);

			return_value.SetInt(beamProp_matnum, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'beam_poperty_type' in command ") + CommandName() + mystr(" Check spelling of beam property edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2013-01
int CEDC_ComAddNode::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	int node_type_index = edc->Find("node_type"); //find index in element EDC
	if (node_type_index)
	{
		if (!edc->Get(node_type_index).IsText()) {assert(0);}

		mystr node_property_str = edc->Get(node_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCNode, node_property_str);

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("Node type '") +  node_property_str + mystr("' not found in command AddNode(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}
		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCNode, type_id); //this are the typeflags, e.g. constraint, 2D, etc.
		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + node_property_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new node of selected type to MBS
			int node_num = edcParser->GetObjectFactory()->AddObject(OFCNode, type_id); //Add Node with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetNode(node_num).GetElementData(edc);
			//now overwrite the default with the user-defined values
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), node_property_str, &CEDCParser::GetNewEDCName, 1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetNode(node_num).SetElementData(edc);

			return_value.SetInt(node_num, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'node_type' in command ") + CommandName() + mystr(" Check spelling of node edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2014-08
int CEDC_ComAddSet::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	int set_type_index = parameter_EDCs(1)->Find("set_type"); //find index in element EDC
	if(set_type_index)
	{
		if (!parameter_EDCs(1)->Get(set_type_index).IsText()) {assert(0);}

		mystr set_property_str = parameter_EDCs(1)->Get(set_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCSet, set_property_str);	

		if (type_id < 0) 
		{
			edcParser->EDCError(mystr("Set type '") +  set_property_str + mystr("' not found in command AddSet(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCSet, type_id); //this are the typeflags, e.g. constraint, 2D, etc.
		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + set_property_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new set of selected type to MBS
			int set_num = edcParser->GetObjectFactory()->AddObject(OFCSet, type_id); //Add Set with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetSetPtr(set_num)->GetElementData(edc);
			//now overwrite the default with the user-defined values
			edc.TreeReplaceEDCDataWithReplacingOldNames(parameter_EDCs(1), set_property_str, &CEDCParser::GetNewEDCName, 1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetSetPtr(set_num)->SetElementData(edc);

			return_value.SetInt(set_num, ""); //return value has no name!

			return 1;
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'set_type' in command ") + CommandName() + mystr(" Check spelling of set edc."));
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2013-01-30
int CEDC_ComComputeInertia::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edc = parameter_EDCs(1); //this contains the element Data
	ElementDataContainer edc_rv;
	int rv = mbs->ComputeInertia(edc,&edc_rv);
	if(rv)
	{
		return_value.SetEDC(&edc_rv,"InertiaValues");
	}
	return rv;
}
mystr CEDC_ComComputeInertia::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameter of this command is an ElementDataContainer, with the following entries: \n");
	descr +=			mystr("\\begin{itemize} \n");
	descr +=			mystr("  \\item density or material\\_number (one of these 2 has to be set!) \n");
	descr +=			mystr("	 \\item One of the following options to define the geometry: \n");
	descr +=			mystr("		  \\begin{itemize} \n");
	descr +=			mystr("				\\item  MeshData.triangles and MeshData.points \\\\ \n");
	descr +=			mystr("							  both entries are Matrices with 3 columns \n");
	descr +=			mystr("				\\item  Cube.body\\_dimensions \n");
	descr +=			mystr("			\\end{itemize} \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2013-02-19

mystr CEDC_ComTransformPoints::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are as follows \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("  \\item points: Matrix of the points: Each line represents a point p. The 3 columns are the x-, y- and z-coordinate \n");
	descr +=			mystr("	 \\item trans: Vector of translation, 3 dimensions! \n");
	descr +=			mystr("	 \\item rot: rotation matrix (3x3), can be used for scaling as well as rotation \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

int CEDC_ComTransformPoints::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edcpoints = parameter_EDCs(1);	//the points that should be transformed
	ElementDataContainer* edcvec = parameter_EDCs(2);			//translation
	ElementDataContainer* edcMat = parameter_EDCs(3);			//rotation matrix
	
	int nRows = 0;
	int nCols = 3;
	double * p;

	ElementData* edp = edcMat->GetPtr(1);
	p = edp->GetMatrix();
	nRows = edp->GetMatrixRows();
	Matrix rotation(nRows,nCols,p);
	
	edp = edcpoints->GetPtr(1);
	p = edp->GetMatrix();
	nRows = edp->GetMatrixRows();
	Matrix points(nRows,nCols,p);

	Matrix new_points;
	new_points.SetSize(nRows,nCols);

	Vector3D translation;
	edp = edcvec->GetPtr(1);
	edp->GetVector(translation.X(),translation.Y(),translation.Z());

	Vector3D help;
	for (int i=1; i<=nRows; i++)
	{
		help = Vector3D(points(i,1),points(i,2),points(i,3));
		help = translation + rotation*help;
		new_points(i,1)=help.X();
		new_points(i,2)=help.Y();
		new_points(i,3)=help.Z();
	}

	return_value.SetMatrix(new_points.GetMatPtr(),new_points.Getrows(),new_points.Getcols(),"points");

	int rv = 1;
	return rv;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
mystr CEDC_ComRotMatToAngles::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are as follows \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("  \\item rotationatrix \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

int CEDC_ComRotMatToAngles::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edc = parameter_EDCs(1); 
	Matrix3D rotMat3D;
	
	Vector3D angles;

	ElementData* edpmatrix = parameter_EDCs(1)->GetPtr(1);
	if( edpmatrix->IsMatrix() && edpmatrix->GetMatrixCols()==3 && edpmatrix->GetMatrixRows()==3 )
	{
		double* m = edpmatrix->GetMatrix();
		Matrix rotationFull(3,3,m);
		rotMat3D = Matrix3D(rotationFull);
	}
	else
	{
		edcParser->EDCError(mystr("first parameter must be a 3x3 matrix "));
		return_value.SetVector3D(0.,0.,0., ""); //return value has no name!
		return 0;
	}

	ElementDataContainer edc_angles;
	ElementData ed;

	Vector3D eulerZXZ;
	RotMatToEulerAngles(rotMat3D, eulerZXZ);
	ed.SetVector3D(eulerZXZ.X(), eulerZXZ.Y(), eulerZXZ.Z(), "EulerZXZ");
	edc_angles.Add(ed);

	Vector3D kardan;
	RotMatToKardanAngles(rotMat3D,kardan);
	ed.SetVector3D(kardan.X(), kardan.Y(), kardan.Z(), "Kardan");
	edc_angles.Add(ed);

	return_value.SetEDC(&edc_angles,"Angles");

	return 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ PG 2020-05-04: 
mystr CEDC_ComRotMatFromAngles::GetCommandTexParameterDescription() const
{
	mystr descr = mystr("The parameters of this command are as follows \n");
	descr += mystr("\\begin{enumerate} \n");
	descr += mystr("  \\item Sequence of rotation angles (3D vector)\n");
	descr += mystr("  \\item Integer value, deciding if Euler-ZXZ (0) or Kardan-XYZ (1) are considered\n");
	descr += mystr("\\end{enumerate} \n");
	descr += mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

int CEDC_ComRotMatFromAngles::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edc = parameter_EDCs(1);
	Vector3D angles;

	ElementData* edpmatrix = parameter_EDCs(1)->GetPtr(1);
	if (edpmatrix->IsVector() && edpmatrix->GetVectorLen()==3)
	{
		double* v = edpmatrix->GetVector();
		angles.Set(v[0], v[1], v[2]);
	}
	else
	{
		edcParser->EDCError(mystr("first parameter must be a 3 vector "));
		double v[] = {
			1.0, 0.0, 0.0,
			0.0, 1.0, 0.0,
			0.0,0.0,1.0 };
		return_value.SetMatrix(v, 3, 3, ""); //return value has no name!
		return 0;
	}


	// read second parameter (bool, deciding whether sequence of angles has to be interpreted as Euler-ZXZ (true) or KardanXYZ (false)
	edc = parameter_EDCs(2);
	mystr tmp;
	int interpreteAnglesAsKardanXYZInsteadOfEulerZXZ = 0;
	if (edc->Get(1).IsDouble())
	{
		interpreteAnglesAsKardanXYZInsteadOfEulerZXZ = (int)(edc->Get(1).GetDouble());
	}
	else if (edc->Get(1).IsInt())
	{
		interpreteAnglesAsKardanXYZInsteadOfEulerZXZ = edc->Get(1).GetInt();
	}
	else
	{
		tmp = edc->TreeGetString("parameter2");
	}

	if (tmp.IsValidNumber())
	{
		interpreteAnglesAsKardanXYZInsteadOfEulerZXZ = tmp.MakeInt();
	}
	else
	{
		if (interpreteAnglesAsKardanXYZInsteadOfEulerZXZ<0 || interpreteAnglesAsKardanXYZInsteadOfEulerZXZ>1)
		{
			edcParser->EDCError("Error while executing command 'Angles2RotMat': 2nd parameter \"interpreteAnglesAsKardanXYZInsteadOfEulerZXZ\" has to be either 0 or 1\n");
			return 0;
		}
	}


	ElementDataContainer edc_rotMat;
	ElementData ed;
	Matrix3D rotMat3D;
	if (interpreteAnglesAsKardanXYZInsteadOfEulerZXZ)
	{
		rotMat3D = ComputeRotMatrixWithKardanAngles(angles(1), angles(2), angles(3));
	}
	else
	{
		rotMat3D = ComputeRotMatrixEuler(angles(1), angles(2), angles(3));
	}
	ed.SetMatrix(rotMat3D.GetMatPtr(), 3, 3, "A");
	edc_rotMat.Add(ed);
	
	return_value.SetEDC(&edc_rotMat, "RotationMatrix");

	return 1;
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ DR 2013-07-04
int CEDC_ComLoadVecFromFile::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edc = parameter_EDCs(1); 
	mystr filename;
	if(edc->Get(1).IsText())
	{
		filename = edc->Get(1).GetText();
	}
	else
	{
		filename = edc->TreeGetString("parameter1");
	}
	edcParser->MakeFilenameAbsolute(filename);

	edc = parameter_EDCs(2);
	mystr tmp;
	int col = 0;
	if(edc->Get(1).IsDouble())
	{
		col = (int)(edc->Get(1).GetDouble());
	}
	else if(edc->Get(1).IsInt())
	{
		col = edc->Get(1).GetInt();
	}
	else
	{
		tmp = edc->TreeGetString("parameter2");
	}

	if(tmp.IsValidNumber())
	{
		col = tmp.MakeInt();
	}
	else
	{
		if(col<=0)	
		{
			edcParser->EDCError("Error while executing command 'LoadVectorFromFile': 2nd parameter has to be a positive integer!\n");
			return 0;	//$ DR 2014-09-25 added this return to stop execution
		}
	}

	int readRowInsteadOfCol = 0;
	if(parameter_EDCs.Length()>2) //$ DR 2014-09-25 added this optional argument
	{
		edc = parameter_EDCs(3);
		if(edc->Get(1).IsDouble()) 	{			readRowInsteadOfCol = (int)(edc->Get(1).GetDouble()); }
		else if(edc->Get(1).IsInt()){			readRowInsteadOfCol = edc->Get(1).GetInt();	}
		else
		{
			tmp = edc->TreeGetString("parameter3");
			if(tmp.IsValidNumber())		{			readRowInsteadOfCol = tmp.MakeInt();	}
			else
			{
				mystr warn = mystr("Error while executing command 'LoadVectorFromFile': 3rd parameter has to be a 0 or 1 but received") + tmp + mystr("! 0 is assumed as default.");
				edcParser->EDCWarning(warn);
			}
		}
	}

	ElementDataContainer edc_rv;
	int rv = mbs->LoadVectorFromFile(filename, col, &edc_rv,readRowInsteadOfCol);
	if(rv)
	{
		double* vptr;
		int len;
		edc_rv.TreeGetVector("Vector",&vptr,len);
		return_value.SetVector(vptr,len,"vector");
	}
	else
	{
		edcParser->EDCError("Error while executing command 'LoadVectorFromFile': could not read file!\n");
	}
	return rv;
}

mystr CEDC_ComLoadVecFromFile::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item The name of the file as string\n");
	descr +=			mystr("  \\item An integer defining in which column (default) or row of the file the vector is stored\n");
	descr +=			mystr("  \\item (optional) 0.. take the column (default), 1.. take the row\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-07-11
int CEDC_ComSum::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* summand1 = parameter_EDCs(1);   // first summand
	ElementDataContainer* summand2 = parameter_EDCs(2);   // second summand
	ElementData *edp1, *edp2;

	edp1 = summand1->GetPtr(1);
	edp2 = summand2->GetPtr(1);

	int rv = 1;
// scalar+scalar
	if (edp1->IsDouble() && edp2->IsDouble())
	{
		double s1 = edp1->GetDouble();
		double s2 = edp2->GetDouble();
		return_value.SetDouble(s1+s2,"sum");
	}
// vector+vector 
	else if (edp1->IsVector() && edp2->IsVector())
	{
		int l1 = edp1->GetVectorLen();
		int l2 = edp2->GetVectorLen();
		if(l1==l2)
		{
			Vector v1,v2;
			v1.LinkWith(edp1->GetVector(),l1);
			v2.LinkWith(edp2->GetVector(),l2);

			//+++++++++++++++++++++++++++++++++++++++++++
			Vector res = v1+v2; //internal representation of command
			//+++++++++++++++++++++++++++++++++++++++++++

			return_value.SetVector(res.GetVecPtr(), res.GetLen(), "sum");
		}
		else
		{
			edcParser->EDCError("Error while executing command 'Add': adding two vectors of different length\n");
		}
	}
// matrix + matrix ( this includes Vector Transposed )
	else if (edp1->IsMatrix() && edp2->IsMatrix())
	{
		int r1 = edp1->GetMatrixRows();
		int r2 = edp2->GetMatrixRows();
		int c1 = edp1->GetMatrixCols();
		int c2 = edp2->GetMatrixCols();

		if(r1==r2 && c1==c2)
		{
			Matrix m1(r1, c1, edp1->GetMatrix());
			Matrix m2(r2, c2, edp2->GetMatrix());

			//+++++++++++++++++++++++++++++++++++++++++++
			Matrix res = m1+m2; //internal representation of command
			//+++++++++++++++++++++++++++++++++++++++++++

			return_value.SetMatrix(res.GetMatPtr(), res.Getrows(), res.Getcols(), "sum");
		}
		else
		{
			edcParser->EDCError("Error while executing command 'Add': adding two matrixes of different length\n");
		}
	}
	else
	{
		edcParser->EDCError("Error while executing command 'Add': cannot add different types of scalar/vector/matrix\n");
	}
	return rv;
}

mystr CEDC_ComSum::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ summand, either scalar, vector or matrix \n");
	descr +=			mystr("  \\item $2^{nd}$ summand, either scalar, vector or matrix \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-07-11
int CEDC_ComProduct::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* summand1 = parameter_EDCs(1);   // first summand
	ElementDataContainer* summand2 = parameter_EDCs(2);   // second summand
	ElementData *edp1, *edp2;

	// when parameters are defined inline the parameter EDC contains text.
	if(summand1->GetPtr(1)->IsText())
		EvaluateTextualParameterEntry(mbs, edcParser, summand1);
	if(summand2->GetPtr(1)->IsText())
		EvaluateTextualParameterEntry(mbs, edcParser, summand2);

	edp1 = summand1->GetPtr(1);
	edp2 = summand2->GetPtr(1);


	int rv = 1;
// involving a scalar: {s*s, s*v, s*m, v*s, m*s}
	if (edp1->IsDouble() || edp2->IsDouble())
	{
		double s;
		ElementData *any;
		if (edp1->IsDouble())
		{
			s = edp1->GetDouble();
			any = edp2;
		}
		else
		{
			s = edp2->GetDouble();
			any = edp1;
		}
		
		if(any->IsDouble()) //s*s
		{
			double r = any->GetDouble();
			return_value.SetDouble(s*r,"product");
		}
		else //s*v, s*m,
		{
			if(any->IsVector())
			{
				Vector V;
				V.LinkWith(any->GetVector(),any->GetVectorLen());

				//+++++++++++++++++++++++++++++++++++++++++++
				Vector res = V*s; //internal representation of command
				//+++++++++++++++++++++++++++++++++++++++++++

				return_value.SetVector(res.GetVecPtr(), res.GetLen(), "product");
			}
			else
			{
				Matrix M(any->GetMatrixRows(), any->GetMatrixCols(), any->GetMatrix());

				//+++++++++++++++++++++++++++++++++++++++++++
				Matrix res = M*s; //internal representation of command
				//+++++++++++++++++++++++++++++++++++++++++++

				return_value.SetMatrix(res.GetMatPtr(), res.Getrows(), res.Getcols(), "product");
			}
		}
	}
// involving only vector and matrix: {v*v, v*m, m*v, m*m}
	else
	{
		Vector V1,V2;
		Matrix M1,M2;

//left factor
		if(edp1->IsVector())
		{
			V1.LinkWith(edp1->GetVector(), edp1->GetVectorLen());
		}
		else if(edp1->IsMatrix())
		{
			M1 = Matrix(edp1->GetMatrixRows(),edp1->GetMatrixCols(),edp1->GetMatrix());
		}
		else
		{
			edcParser->EDCError("Error while executing command 'Product': first factor neither scalar nor vector nor matrix \n");
		}
// right factor
		if(edp2->IsVector())
		{
			V2.LinkWith(edp2->GetVector(), edp2->GetVectorLen());
		}
		else if(edp2->IsMatrix())
		{
			M2 = Matrix(edp2->GetMatrixRows(),edp2->GetMatrixCols(),edp2->GetMatrix());
		}
		else
		{
			edcParser->EDCError("Error while executing command 'Product': second factor neither scalar nor vector nor matrix \n");
		}

// compute the product
// case m*m = m --> matrix product
		if(edp1->IsMatrix()&&edp2->IsMatrix()) 
		{
			if(M1.Getcols() == M2.Getrows())
			{
				//+++++++++++++++++++++++++++++++++++++++++++
				//internal representation of command
				Matrix res;
				Mult(M1,M2,res);
				//+++++++++++++++++++++++++++++++++++++++++++
				return_value.SetMatrix(res.GetMatPtr(), res.Getrows(), res.Getcols(), "product");
			}
			else
			{
				edcParser->EDCError("Error while executing command 'Product': dimension check failed - number of columns in left factor must be equal to number of rows in right factor \n");
			}
		}
// case v*m = v --> matrix product
		else if(edp1->IsVector()&&edp2->IsMatrix())
		{
			if(V1.GetLen() == M2.Getrows())
			{
				//+++++++++++++++++++++++++++++++++++++++++++
				//internal representation of command
				Vector res;
				Mult(V1,M2,res);
				//+++++++++++++++++++++++++++++++++++++++++++
				return_value.SetVector(res.GetVecPtr(), res.GetLen(), "product");
			}
			else
			{
				edcParser->EDCError("Error while executing command 'Product': dimension check failed - number of columns in left factor must be equal to number of rows in right factor \n");
			}
		}

//! AD: !!! MAYBE TAKE THESE CASES OUT !!! more commands eg transpose(), scalarproduct()
// case v*v =s -> scalar product
		else if(edp1->IsVector()&&edp2->IsVector())
		{
			//$ AD: warning removed, now in docu 
			//edcParser->EDCError("WARNING: multiplying two (row-)vectors is autmatically assumed to compute the scalar product of the two vectors \n");
			if(V1.GetLen() == V2.GetLen())
			{
				double scalarproduct = 0;
				for(int i=1; i<=V1.GetLen(); i++)
				{
					scalarproduct += ( V1(i)*V2(i) );
				}
				return_value.SetDouble(scalarproduct,"product");	
			}
			else
			{
				edcParser->EDCError("Error while executing command 'Product': factors (vector) must have same length \n");
			}
		}
// case m*v --> m*vT=vT
		else if(edp1->IsMatrix()&&edp2->IsVector())
		{
			//$ AD: warning removed, now in docu 
			//edcParser->EDCError("WARNING: multiplying a matrix with a (row-)vector is autmatically assumed to compute the product with the transposed vector \n");
			if(M1.Getcols() == V2.GetLen())
			{
				//+++++++++++++++++++++++++++++++++++++++++++
				//internal representation of command
				Vector res;
				Mult(M1,V2,res);
				//+++++++++++++++++++++++++++++++++++++++++++
				return_value.SetMatrix(res.GetVecPtr(), res.GetLen(), 1, "product");
			}
			else
			{
				edcParser->EDCError("Error while executing command 'Product': dimension check failed - number of columns in left factor must be equal to number of rows in right factor \n");
			}
		}
		else
		{
			edcParser->EDCError("Error while executing command 'Product': dimension check failed - number of columns in left factor must be equal to number of rows in right factor \n");
		}
	}
	return rv;
}

mystr CEDC_ComProduct::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ factor, either scalar, vector or matrix \n");
	descr +=			mystr("  \\item $2^{nd}$ factor, either scalar, vector or matrix \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("  product of two vectors is always computed as scalar product \n");
	descr +=			mystr("  for vector times Matrix the vector is automatically transposed if required \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-07-11
int CEDC_ComTranspose::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* edc1 = parameter_EDCs(1);   // sole input
	ElementData *edp1;
	edp1 = edc1->GetPtr(1);

	int rv = 1;
// transpose a vector -> vT is a matrix
	if (edp1->IsVector())
	{
		Vector V;
		V.LinkWith(edp1->GetVector(), edp1->GetVectorLen());;
		//+++++++++++++++++++++++++++++++++++++++++++
		//internal representation of command
		Matrix res(V.GetLen(), 1, V.GetVecPtr());
		//+++++++++++++++++++++++++++++++++++++++++++
		return_value.SetMatrix(res.GetMatPtr(), res.Getrows(), res.Getcols(), "transposed");
	}
	else if (edp1->IsMatrix())
	{
		Matrix M(edp1->GetMatrixRows(),edp1->GetMatrixCols(),edp1->GetMatrix());
		//+++++++++++++++++++++++++++++++++++++++++++
		//internal representation of command
		Matrix res = M.GetTp();
		//+++++++++++++++++++++++++++++++++++++++++++
		if(res.Getrows() == 1)
		{
		// this is actually a vector ...
			return_value.SetVector(res.GetMatPtr(), res.Getcols(), "product");
		}
		else
		{
			return_value.SetMatrix(res.GetMatPtr(), res.Getrows(), res.Getcols(), "product");
		}
	}
	else
	{
		edcParser->EDCError("Error while executing command 'Transpose': only vectors and matrices can be transposed\n");
	}
	return rv;
}

mystr CEDC_ComTranspose::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item vector or matrix to be transposed\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-07-11
//$! AD 2013-07-15 commented out Class Scalar product for the time being, functionality is in ComProduct...

////int CEDC_ComScalarProduct::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
////{
////	ElementDataContainer* summand1 = parameter_EDCs(1);   // first summand
////	ElementDataContainer* summand2 = parameter_EDCs(2);   // second summand
////	ElementData *edp1, *edp2;
////
////	edp1 = summand1->GetPtr(1);
////	edp2 = summand2->GetPtr(1);
////
////	int rv = 1;
////	if(edp1->IsVector()&&edp2->IsVector())
////	{	
////		Vector V1,V2;
////		V1.LinkWith(edp1->GetVector(), edp1->GetVectorLen());
////		V2.LinkWith(edp2->GetVector(), edp2->GetVectorLen());
////
////		if(V1.GetLen() == V2.GetLen())
////		{
////			double scalarproduct = 0;
////			for(int i=1; i<=V1.GetLen(); i++)
////			{
////				scalarproduct += ( V1(i)*V2(i) );
////			}
////			return_value.SetDouble(scalarproduct,"product");	
////		}
////		else
////		{
////			edcParser->EDCError("Error while executing command 'ScalarProduct': vectors must have same length \n");
////		}
////	}
////	else
////	{
////		edcParser->EDCError("Error while executing command 'ScalarProduct': scalar product only defined for two vectors\n");
////	}
////	return rv;
////}
////
////mystr CEDC_ComScalarProduct::GetCommandTexParameterDescription() const 
////{
////	mystr descr = mystr("The parameters of this command are \n");
////	descr +=			mystr("\\begin{enumerate} \n");
////	descr +=			mystr("	 \\item 1^{st} vector \n");
////	descr +=			mystr("  \\item 2^{nd} vector \n");
////	descr +=			mystr("\\end{enumerate} \n");
////	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
////
////	return descr;
////};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-07-11
int CEDC_ComCrossProduct::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementDataContainer* summand1 = parameter_EDCs(1);   // first summand
	ElementDataContainer* summand2 = parameter_EDCs(2);   // second summand
	ElementData *edp1, *edp2;

	edp1 = summand1->GetPtr(1);
	edp2 = summand2->GetPtr(1);

	int rv = 1;
	if(edp1->IsVector()&&edp2->IsVector())
	{
		Vector V1,V2;
		V1.LinkWith(edp1->GetVector(), edp1->GetVectorLen());
		V2.LinkWith(edp2->GetVector(), edp2->GetVectorLen());

		if(V1.GetLen() == V2.GetLen() && (V1.GetLen() == 3 || V1.GetLen() == 2) )   // allow  3Dx3D and 2Dx2D
		{
			Vector3D v1,v2;
			if(V1.GetLen() == 2)
			{
				v1 = Vector3D(V1(1), V1(2), 0.);
				v2 = Vector3D(V2(1), V2(2), 0.);
			}
			else 
			{
				v1 = Vector3D(V1(1), V1(2), V1(3));
				v2 = Vector3D(V2(1), V2(2), V2(3));
			}
						
			//+++++++++++++++++++++++++++++++++++++++++++
			//internal representation of command
			Vector3D res = v1.Cross(v2);
			//+++++++++++++++++++++++++++++++++++++++++++
			if(V1.GetLen() == 3)
			{
				return_value.SetVector(res.GetVecPtr(), 3, "crossproduct");	// return 3D Vector  3Dx3D -> 3D
			}
			else if(V1.GetLen() == 2)
			{
				return_value.SetDouble(res(3), "crossproduct");	// return scalar 2Dx2D -> scalar
			}
		}
		else
		{
			edcParser->EDCError("Error while executing command 'CrossProduct': vectors must have same length (2D or 3D)\n");
		}
	}
	else
	{
		edcParser->EDCError("Error while executing command 'CrossProduct': scalar product only defined for two vectors\n");
	}
	return rv;
}

mystr CEDC_ComCrossProduct::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ vector (2D or 3D)\n");
	descr +=			mystr("  \\item $2^{nd}$ vector (2D or 3D)\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("for two 3D vectors the retuen value is also a 3D vector. For two 2D vectors the return value is a scalar.\n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex

	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-09-13
int CEDC_ComFor::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ED_loop_init = parameter_EDCs(1)->GetPtr(1);
	ElementData* ED_loop_cond = parameter_EDCs(2)->GetPtr(1);
	ElementData* ED_loop_incr = parameter_EDCs(3)->GetPtr(1);
	ElementData* ED_loop_code = parameter_EDCs(4)->GetPtr(1);

// CHECK PARAMETERS:
// parameter 1: loop initialization
	mystr str_loop_init;
	if(ED_loop_init->IsText())
	{
		str_loop_init = ED_loop_init->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing FOR loop: first parameter is not a text (loop initialization)");
		return 0;// rv;
	}
// parameter 2: loop condition
	mystr str_loop_cond;
	if(ED_loop_cond->IsText())
	{
		str_loop_cond = ED_loop_cond->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing FOR loop: second parameter is not a text (loop condition)");
		return 0;// rv;
	}
// parameter 3: loop increment
	mystr str_loop_incr;
	if(ED_loop_incr->IsText())
	{
		str_loop_incr = ED_loop_incr->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing FOR loop: third parameter is not a text (loop increment)");
		return 0;// rv;
	}

// parameter 4: loop code
	mystr str_loop_code;
	if(ED_loop_code->IsText())
	{
		str_loop_code = ED_loop_code->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing FOR loop: loop code is to text (text in brackets required)");
		return 0;// rv;
	}

// the actual FOR Loop

// initialize local loop variable container
	ElementDataContainer& localEDC = *(edcParser->GetMBSParser()->GetLocalEDC());  // existing local EDC for the FOR command ("outside" the FOR command)

	// initialize the loop variable
	int rv_init = edcParser->ParseAndExecuteString(str_loop_init, localEDC);
	if(!rv_init)
	{
		edcParser->EDCError("Error while executing FOR loop: first parameter could not be translated to a loop initialization statement (\"i=1\")");
		return 0;// rv;
	}

	int iter_count = 0;
// loop condition 1st time
	int parserErrorFlag = 0;
	int flag_loop_condition = (int) (edcParser->GetMBSParser()->ExpressionToDouble(str_loop_cond, parserErrorFlag)+0.5);				// loop condition
	if (parserErrorFlag)
	{
		edcParser->EDCError("Error while executing FOR condition: could not evaluate condition " + str_loop_cond);
		return 0;// rv;
	}

	while(flag_loop_condition) 
	{
// execute loop code 
		int rv_codeblock = edcParser->ParseAndExecuteString(str_loop_code, localEDC);
		if(!rv_codeblock)
		{
			edcParser->EDCError("Error while executing FOR loop: failed to execute loop code");
			return 0;// rv;
		}
		iter_count++;

// increment
		int rv_inc = edcParser->ParseAndExecuteString(str_loop_incr, localEDC);
		if (!rv_inc)
		{
			edcParser->EDCError("Error while executing FOR loop: failed to increment "+str_loop_incr);
			return 0;// rv;
		}

// condition again
		flag_loop_condition = (int) (edcParser->GetMBSParser()->ExpressionToDouble(str_loop_cond, parserErrorFlag)+0.5);
		if (parserErrorFlag)
		{
			edcParser->EDCError("Error while executing FOR condition: could not evaluate condition " + str_loop_cond);
			return 0;// rv;
		}
	}

	return 1;
}

mystr CEDC_ComFor::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ define and initialize loop variable (\"i=1\")\n");
	descr +=			mystr("  \\item $2^{nd}$ loop condition (\"i<5\")\n");
	descr +=			mystr("  \\item $3^{rd}$ loop increment (\"i=i+1\")\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("the command must be followed by a container for the loop code \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int CEDC_ComIf::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ED_loop_cond = parameter_EDCs(1)->GetPtr(1);
	ElementData* ED_loop_code = parameter_EDCs(2)->GetPtr(1);

// CHECK PARAMETERS:
// parameter 1: loop condition
	mystr str_loop_cond;
	if(ED_loop_cond->IsText())
	{
		str_loop_cond = ED_loop_cond->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing IF condition: second parameter is not a text (loop condition)");
		return 0;// rv;
	}
// parameter 2: loop code
	mystr str_loop_code;
	if(ED_loop_code->IsText())
	{
		str_loop_code = ED_loop_code->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing IF condition: loop code is to text (text in brackets required)");
		return 0;// rv;
	}

// the actual IF condition

// initialize local loop variable container
	ElementDataContainer& localEDC = *(edcParser->GetMBSParser()->GetLocalEDC());  // existing local EDC for the IF command ("outside" the IF command)

// loop condition 1st time
	int parserErrorFlag = 0;
	int flag_loop_condition = (int) (edcParser->GetMBSParser()->ExpressionToDouble(str_loop_cond, parserErrorFlag)+0.5);				// loop condition

	if (parserErrorFlag)
	{
		edcParser->EDCError("Error while executing IF condition: could not evaluate condition "+str_loop_cond);
		return 0;// rv;
	}

	if(flag_loop_condition)
	{
// loop condition true -->  execute conditional code 
		int rv_codeblock = edcParser->ParseAndExecuteString(str_loop_code, localEDC); 
		return rv_codeblock;
	}

// loop contition false
	return 1;
	
}

mystr CEDC_ComIf::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ condition (\"i<10\")\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("the command must be followed by a container for the conditional code \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2013-10-09
//$ AD 2014-09-16 - changed to true Mesh object
int CEDC_ComMesh_GenerateNewMesh::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	edcParser->EDCDbgMsg(mystr("identified: command GenerateNewMesh."));

// step 1: generate the Mesh within the MBS
	int mesh_num = 0;
	int mesh_type_index = parameter_EDCs(1)->Find("mesh_type"); // find index in element EDC
	int mesh_flag_autocompute = 0;

	// HACK AD: for use with the old version that did not require a flag for automatic surface identification
	if(parameter_EDCs(1)->Find("compute_surface"))
	{
		parameter_EDCs(1)->TreeGetBool("compute_surface",0);
	}

	// HACK AD: for use with the old version that did not require a mesh_type entry...	
	if(mesh_type_index==0)
	{
		ElementData ed;
		ed.SetText("StructuralMesh","mesh_type");
		parameter_EDCs(1)->Add(ed);
		mesh_type_index = parameter_EDCs(1)->Find("mesh_type");
	}
	
	if(mesh_type_index)
	{
		if(!parameter_EDCs(1)->Get(mesh_type_index).IsText()) {assert(0);}

		mystr mesh_property_str = parameter_EDCs(1)->Get(mesh_type_index).GetText();
		int type_id = edcParser->GetObjectFactory()->GetObjectTypeId(OFCMesh, mesh_property_str);

		if(type_id < 0)
		{
			edcParser->EDCError(mystr("Mesh type '") +  mesh_property_str + mystr("' not found in command GenerateNewMesh(...)!"));
			return_value.SetInt(rv_error, ""); //return value has no name!
			return 0;
		}

		//int typeflags = edcParser->GetObjectFactory()->GetTypeFlags(OFCMesh, type_id); //these are the typeflags, e.g. constraint, 2D, etc.
		//if( (typeflags & TAENotInRelease) && (edcParser->GetObjectFactory()->ExcludeExperimentalObjects()) )
		//{
		//	edcParser->EDCError(mystr("ERROR: ") + mesh_property_str + mystr(" is not implemented yet. Check Homepage for available updates."));
		//	return_value.SetInt(rv_error, ""); //return value has no name!
		//	return 0;
		//}
		//else
		{
			//add new mesh of selected type to MBS
			mesh_num = edcParser->GetObjectFactory()->AddObject(OFCMesh, type_id); //Add Mesh with default values
			//retrieve the default EDC for the new element
			ElementDataContainer edc;
			mbs->GetMeshPtr(mesh_num)->GetElementData(edc);
			//now overwrite the default with the user-defined values
			edc.TreeReplaceEDCDataWith(parameter_EDCs(1),1);
			//now set the user-defined EDC in the newly generated element
			int rv = mbs->GetMeshPtr(mesh_num)->SetElementData(edc);

			edcParser->EDCDbgMsg(mystr("Mesh handle created."));
			return_value.SetInt(mesh_num, ""); //return value has no name!
		}
	}
	else
	{
		edcParser->EDCError(mystr("did not find 'mesh_type' in command ") + CommandName() + " (new reqirement as of 2014-09-17)");
		return_value.SetInt(rv_error, ""); //return value has no name!
		return 0;
	}

// step 2: entry in global variables
	ElementDataContainer edc_meshproperties;
	ElementData ed;
// why cant bool be locked ?
	ed.SetBool(1,"isHandle"); ed.SetLocked(1); ed.SetToolTipText("This Container represents a Handle. Commands may be executed for this object"); edc_meshproperties.Add(ed);
// available script commands directly from actual instanced mesh 
	ElementDataContainer edc_registered_functions;
	mbs->GetMeshPtr(mesh_num)->GetAvailableScriptFunctions(edc_registered_functions);
	ed.SetEDC(&edc_registered_functions,"Functions"); // this container name is checked in ::IdentifyCommand
	edc_meshproperties.Add(ed);
	
	return_value.SetEDC(&edc_meshproperties,"MeshProperties");
	edcParser->EDCDbgMsg(mystr("Mesh handle registered."));

	return 1;
}


mystr CEDC_ComMesh_GenerateNewMesh::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC to overwrite the default properties\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("the return vaule of the command MUST be assigned to a new variable(handle) \n\n");
	descr +=			mystr("overwritable enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item mesh\\_name \n");
	descr +=			mystr("	 \\item mesh\\_type: may be \\bf{StructuralMesh} (default) or \\bf{SolidMesh} \n");
	descr +=            mystr("  \\item compute\\_surface: set to 1 to automatically compute surface of the mesh \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#ifdef mode_old
int CEDC_ComMesh_GenerateBeam::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// Default values for a BeamParameters
// * this is to be autogenerated at a later time
	ElementDataContainer edc_beamparameters;
	ElementData ed;
	ed.SetVector3D(0.,0.,0.,"P1");							edc_beamparameters.Add(ed);
	ed.SetVector3D(1.,0.,0.,"P2");							edc_beamparameters.Add(ed);
	ed.SetInt(1,"matnr");												edc_beamparameters.Add(ed);
	ed.SetInt(1,"discretization");							edc_beamparameters.Add(ed);
	ed.SetText("LinearBeam3D","element_type");	edc_beamparameters.Add(ed);
	ed.SetText("Node3DRxyz","node_type");				edc_beamparameters.Add(ed);

// Overwrite defaults with input from container 1
	ElementDataContainer* edc_override = parameter_EDCs(1);
	edc_beamparameters.TreeReplaceEDCDataWith(edc_override);

// container 2 contains the name of the handle ( can not pass the EDC* itself because that is going to be deleted... )
// consistency already checked in ParseCommandParameters...
	mystr handlename = parameter_EDCs(2)->TreeGetString("handle");
	ElementDataContainer* edc_local = edcParser->GetMBSParser()->GetLocalEDC(); 
	int nr_handle = edc_local->Find(handlename);
	ElementDataContainer* edc_handle_mesh = edc_local->Get(nr_handle).GetEDC();


// Actual Generation of BeamElements - to be a function of the MeshElement
// read geometry parameters
	Vector3D p1,p2;
	edc_beamparameters.TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());
	edc_beamparameters.TreeGetVector3D("P2",p2.X(), p2.Y(), p2.Z());
	int matnr = edc_beamparameters.TreeGetInt("matnr");
	int discr = edc_beamparameters.TreeGetInt("discretization");
	TArray<double> loads;
	int i = edc_beamparameters.Find("load"); 
	if (i != 0)
	{
		int load = edc_beamparameters.TreeGetInt("load");
		if (load != 0)
		{
			loads.Add(load);
		}
	}

// when more then one type of element is implemented, perform a type check 
	mystr elem_type = edc_beamparameters.TreeGetString("element_type");
	mystr node_type = edc_beamparameters.TreeGetString("node_type");

// get current lists for mesh object
	double* vdata;
	int vlen;
	edc_handle_mesh->TreeGetVector("list_of_nodes",&vdata,vlen);            
	TArray<double> mesh_nnrs; mesh_nnrs.CopyFrom(vdata,vlen); mesh_nnrs.SetLen(vlen);			// node numbers already associated with the Mesh
	edc_handle_mesh->TreeGetVector("list_of_elements",&vdata,vlen);
	TArray<double> mesh_enrs; mesh_enrs.CopyFrom(vdata,vlen); mesh_enrs.SetLen(vlen);			// element numbers already associated with the Mesh

// *******************
// GENERATION OF NODES

// use existing script command AddNode
	mystr commandname("AddNode");
	int nparam, hasretval, c_option;
	CEDC_Command* command_addnode = edcParser->GetCommand( edcParser->GetCommandType(commandname, nparam, hasretval, c_option) );

// prepare EDC to be able to use existing script commands 
	ElementDataContainer edc_nodeproperties;
	TArray<ElementDataContainer*> paramEDCforNode;
	paramEDCforNode.Add(&edc_nodeproperties);
	ed.SetText(node_type,"node_type");																		edc_nodeproperties.Add(ed);

//  todo: automatically compute orientation for all nodes - depending on node_type ...
	// Node3DRxyz 
	Vector3D p12 = p2-p1;
	double pyth_xy= pow((p12.X()*p12.X()+p12.Y()*p12.Y()),(.5)); 

	double roll = 0;
	double pitch = atan2(p12.Y(), p12.X());
	double yaw = atan2(p12.Z(), pyth_xy);
	ed.SetVector3D(roll,-yaw,pitch,"reference_rot_angles");											edc_nodeproperties.TreeAdd("Geometry",ed);


// override discretization if entry element_size is present
	int j = edc_beamparameters.Find("element_size");											
	if (j != 0 && edc_beamparameters.Get(j).IsDouble())
	{
		double elemsize = edc_beamparameters.TreeGetDouble("element_size");
		double total_length = p12.Norm();
		double required = total_length / elemsize;
		int discr_new = (int) ceil(required);
		discr = discr_new;
	}


	TArray<double> nnrs;                                                  // node numbers for the current call GenerateBeam
	for(int i=0; i<=discr; i++)
	{
		Vector3D pos = ( (discr-i)*p1 + i*p2 ) * (1./double(discr));
		ed.SetVector3D(pos.X(),pos.Y(),pos.Z(),"reference_position");				edc_nodeproperties.TreeAdd("Geometry",ed);

// call AddNode with prepared EDC, returnvalue is number of added node
		command_addnode->ExecuteCommand(mbs,edcParser,paramEDCforNode,ed,0);
		
		nnrs.Add(ed.GetInt());
	}
// return value:nodes from current generation function
	ed.SetVector( nnrs.GetDataPtr(), nnrs.Length(), "list_of_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Nodes associated with the Beam or empty");
	edc_rv.Add(ed);
// update list_of_nodes in Mesh
	mesh_nnrs.Merge(nnrs);
	ed.SetVector( mesh_nnrs.GetDataPtr(), mesh_nnrs.Length(), "list_of_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Nodes associated with the Mesh or empty");
	ElementDataContainer update_nnrs; update_nnrs.Add(ed);
	edc_handle_mesh->TreeReplaceEDCDataWith(&update_nnrs);

// ***************************
// GENERATION OF BEAM ELEMENTS

// use existing script command AddElement
	commandname = mystr("AddElement");
	CEDC_Command* command_addelement = edcParser->GetCommand( edcParser->GetCommandType(commandname, nparam, hasretval, c_option) );
	
	// prepare EDC to be able to use existing script commands 
	ElementDataContainer edc_elementproperties;
	ed.SetText(elem_type,"element_type");																	edc_elementproperties.Add(ed);
	ed.SetInt(matnr,"material_number");																		edc_elementproperties.TreeAdd("Physics",ed);

	if (loads.Length() > 0)
	{
		ed.SetVector(loads.GetDataPtr(), loads.Length(), "loads");	edc_elementproperties.Add(ed);
	}

	TArray<ElementDataContainer*> paramEDCforElement;
	paramEDCforElement.Add(&edc_elementproperties);

	TArray<double> enrs;
	for(int i=1; i<=discr; i++)
	{
		ed.SetInt((int)(nnrs(i)),"node_1");																				edc_elementproperties.TreeAdd("Geometry",ed);
		ed.SetInt((int)(nnrs(i+1)),"node_2");																			edc_elementproperties.TreeAdd("Geometry",ed);

// call AddElement with prepared EDC, returnvalue is number of added element
		command_addelement->ExecuteCommand(mbs,edcParser,paramEDCforElement,ed,0);

		enrs.Add(ed.GetInt());
	}

// return value:nodes from current generation function
	ed.SetVector( enrs.GetDataPtr(), enrs.Length(), "list_of_elements");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Elements associated with the Beam or empty");
	edc_rv.Add(ed);
// update list_of_nodes in Mesh
	mesh_enrs.Merge(enrs);
	ed.SetVector( mesh_enrs.GetDataPtr(), mesh_enrs.Length(), "list_of_elements");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Elements associated with the Mesh or empty");
	ElementDataContainer update_enrs; update_enrs.Add(ed);
	edc_handle_mesh->TreeReplaceEDCDataWith(&update_enrs);


	return_value.SetEDC(edc_rv.GetCopy(), "RV_Beam");

	return 1;
}
#else 
int CEDC_ComMesh_GenerateBeam::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GenerateBeam - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GenerateBeam - Handle could not be resolved into a Mesh"));
		return 0;
	}
	if(mbs->GetMeshPtr(meshIdx)->GetType() != /*TMBSMeshType::*/TMBSMeshStructural) // (double) check type - already checked if 
	{
		edcParser->EDCError(mystr("Error in GenerateBeam - requires mesh_type = structural mesh"));
		return 0;	
	}

// first container contains the beam parameters	
	ElementDataContainer* edc_beam_params = parameter_EDCs(1);
// generate the component
	MeshCurve curve;
	ElementDataContainer edc_curve;
	curve.GetElementData(edc_curve);
	edc_curve.TreeReplaceEDCDataWith(edc_beam_params);
	curve.SetElementData(edc_curve);
// add the component to the mesh 
	int compnr = mbs->GetMeshPtr(meshIdx)->AddComponent(&curve);
	
	return_value.SetInt(compnr,"component_number");
	return compnr;//rv;
}
#endif

mystr CEDC_ComMesh_GenerateBeam::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the beam properties \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item component\\_name \n");
	descr +=			mystr("	 \\item component\\_type: Linear(default) \n");	
	descr +=			mystr("	 \\item Generation.P1 - position of left outer node \n");
	descr +=			mystr("	 \\item Generation.P2 - position of right outer node \n");
	descr +=      mystr("  \\item Generation.material\\_number - number of the material to be used for the beam elements (Beam3DProperties) \n");
	descr +=      mystr("  \\item Generation.discretization - number of the beam elements \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls of the generated elements \n");
//N.I.Y.	descr +=      mystr("  \\item elementsize - (optional) maximum element size ( if set has priority over discretization ) \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

#define mode_old_not
#ifdef mode_old
int CEDC_ComMesh_GeneratePlate::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

	// Default values for a PlateParameters
// * this is to be autogenerated at a later time
	ElementDataContainer edc_plateparameters;
	ElementData ed;
	ed.SetVector3D(0.,0.,0.,"P1");				edc_plateparameters.Add(ed);
	ed.SetVector3D(1.,0.,0.,"P2");				edc_plateparameters.Add(ed);
	ed.SetVector3D(0.,1.,0.,"P3");				edc_plateparameters.Add(ed);
	ed.SetInt(1,"matnr");						edc_plateparameters.Add(ed);
	ed.SetVector2D(1.,1.,"discretization");		ed.SetValuesInt();	edc_plateparameters.Add(ed);
	
	ed.SetText("ANCFThinPlate3D","element_type");	edc_plateparameters.Add(ed);
	ed.SetText("Node3DS1S2","node_type");			edc_plateparameters.Add(ed);
	ed.SetBool(1, "nonlinear");		edc_plateparameters.Add(ed);

// Overwrite defaults with input from container 1
	ElementDataContainer* edc_override = parameter_EDCs(1);
	edc_plateparameters.TreeReplaceEDCDataWith(edc_override);

// container 2 contains the name of the handle ( can not pass the EDC* itself because that is going to be deleted... )
// consistency already checked in ParseCommandParameters...
	mystr handlename = parameter_EDCs(2)->TreeGetString("handle");
	ElementDataContainer* edc_local = edcParser->GetMBSParser()->GetLocalEDC(); 
	int nr_handle = edc_local->Find(handlename);
	ElementDataContainer* edc_handle_mesh = edc_local->Get(nr_handle).GetEDC();


// Actual Generation of PlateElements - to be a function of the MeshElement
// read geometry parameters 
	Vector3D p1,p2,p3;
	edc_plateparameters.TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());
	edc_plateparameters.TreeGetVector3D("P2",p2.X(), p2.Y(), p2.Z());
	edc_plateparameters.TreeGetVector3D("P3",p3.X(), p3.Y(), p3.Z());
	int matnr = edc_plateparameters.TreeGetInt("matnr");
	double thickness = edc_plateparameters.TreeGetDouble("thickness");
	Vector2D discr;
	edc_plateparameters.TreeGetVector2D("discretization", discr.X(), discr.Y());
	int is_nonlinear = edc_plateparameters.TreeGetBool("nonlinear", 1);
	
	TArray<double> loads;
	int i = edc_plateparameters.Find("load"); 
	if (i != 0)
	{
		if(edc_plateparameters.Get(i).IsInt())
		{
			int load = edc_plateparameters.TreeGetInt("load");
			if (load != 0)
			{
				loads.Add(load);
			}
		}
		else if(edc_plateparameters.Get(i).IsVector())
		{
			int len = edc_plateparameters.Get(i).GetVectorLen();
			double* v = edc_plateparameters.Get(i).GetVector();
			for(int i0=0; i0<len; i0++)
			{
				loads.Add(v[i0]);
			}
		}
		else
		{
				edcParser->EDCError("Error while executing GeneratePlate operation: parameter for load must be integer or vector(of integers)");
		}

	}

// when more then one type of element is implemented, perform a type check 
	mystr elem_type = edc_plateparameters.TreeGetString("element_type");
	mystr node_type = edc_plateparameters.TreeGetString("node_type");

// get current lists for mesh object
	double* vdata;
	int vlen;
	edc_handle_mesh->TreeGetVector("list_of_nodes",&vdata,vlen);            
	TArray<double> mesh_nnrs; mesh_nnrs.CopyFrom(vdata,vlen); mesh_nnrs.SetLen(vlen);			// node numbers already associated with the Mesh
	edc_handle_mesh->TreeGetVector("list_of_elements",&vdata,vlen);
	TArray<double> mesh_enrs; mesh_enrs.CopyFrom(vdata,vlen); mesh_enrs.SetLen(vlen);			// element numbers already associated with the Mesh

// *******************

// GENERATION OF NODES

// use existing script command AddNode
	mystr commandname("AddNode");
	int nparam, hasretval, c_option;
	CEDC_Command* command_addnode = edcParser->GetCommand( edcParser->GetCommandType(commandname, nparam, hasretval, c_option) );


// prepare EDC to be able to use existing script commands 
	ElementDataContainer edc_nodeproperties;
	TArray<ElementDataContainer*> paramEDCforNode;
	paramEDCforNode.Add(&edc_nodeproperties);
	ed.SetText(node_type,"node_type");																		edc_nodeproperties.Add(ed);

//  todo: automatically compute orientation for all nodes - depending on node_type ...
	Vector3D p12 = p2-p1;						//  e1 = p12/|p12|
	Vector3D p13 = p3-p1;						//  e2 = p13/|p13|
	Vector3D n(p12.Cross(p13));			// normal to the plane od the plate

	Vector3D e1(p12); e1.Normalize();
	ed.SetVector3D(e1.X(), e1.Y(), e1.Z(), "reference_slope1");	edc_nodeproperties.TreeAdd("Geometry",ed);
	Vector3D e2(p13); e2.Normalize();
	ed.SetVector3D(e2.X(), e2.Y(), e2.Z(), "reference_slope2");	edc_nodeproperties.TreeAdd("Geometry",ed);

// override discretization if entry element_size is present
	int j = edc_plateparameters.Find("element_size");										
	if (j != 0 && edc_plateparameters.Get(j).IsVector())
	{
		Vector2D elemsize;
		edc_plateparameters.TreeGetVector2D("element_size",elemsize.X(), elemsize.Y());
		double total_length, required;
		total_length = p12.Norm();
		required = total_length / elemsize.X();
		discr.X() = ceil(required);
		if(discr.X() <= 0) discr.X() = 1;
		total_length = p13.Norm();
		required = total_length / elemsize.Y();
		discr.Y() = ceil(required);
		if(discr.Y() <= 0) discr.Y() = 1;
	}

// catch case that the plate has no width -> do not create any new elements
	if(p12.Norm() < 1e-8)
	{
		return rv;
	}

	TArray<double> nnrs;                                                  // node numbers for the current call GenerateBeam
	for(int i=0; i<=(int)discr.Y(); i++)
	{
		for(int j=0; j<=(int)discr.X(); j++)
		{
			Vector3D pos = p1 + p12*((double)j/discr.X()) + p13*((double)i/discr.Y());
			ed.SetVector3D(pos.X(),pos.Y(),pos.Z(),"reference_position");				edc_nodeproperties.TreeAdd("Geometry",ed);

// call AddNode with prepared EDC, returnvalue is number of added node
			command_addnode->ExecuteCommand(mbs,edcParser,paramEDCforNode,ed,0);

			nnrs.Add(ed.GetInt());
		}
	}
// return value:nodes from current generation function
	ed.SetVector( nnrs.GetDataPtr(), nnrs.Length(), "list_of_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Nodes associated with the Plate or empty");
	edc_rv.Add(ed);
// update list_of_nodes in Mesh
	mesh_nnrs.Merge(nnrs);
	ed.SetVector( mesh_nnrs.GetDataPtr(), mesh_nnrs.Length(), "list_of_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Nodes associated with the Mesh or empty");
	ElementDataContainer update_nnrs; update_nnrs.Add(ed);
	edc_handle_mesh->TreeReplaceEDCDataWith(&update_nnrs);


// ***************************
// GENERATION OF PLATE ELEMENTS

	double size1 = p12.Norm() / discr.X();
	double size2 = p13.Norm() / discr.Y();

// use existing script command AddElement
	commandname = mystr("AddElement");
	CEDC_Command* command_addelement = edcParser->GetCommand( edcParser->GetCommandType(commandname, nparam, hasretval, c_option) );
	
	// prepare EDC to be able to use existing script commands 
	ElementDataContainer edc_elementproperties;
	ed.SetText(elem_type,"element_type");	edc_elementproperties.Add(ed);
	ed.SetInt(matnr,"material_number");		edc_elementproperties.TreeAdd("Physics",ed);
	ed.SetDouble(thickness,"thickness");	edc_elementproperties.TreeAdd("Geometry",ed);
	ed.SetDouble(size1,"size1");	edc_elementproperties.TreeAdd("Geometry",ed);
	ed.SetDouble(size2,"size2");	edc_elementproperties.TreeAdd("Geometry",ed);
	ed.SetBool(is_nonlinear, "is_geometrically_nonlinear");		edc_elementproperties.TreeAdd("Physics",ed);
	
	if (loads.Length() > 0)
	{
		ed.SetVector(loads.GetDataPtr(), loads.Length(), "loads");	edc_elementproperties.Add(ed);
	}

	TArray<ElementDataContainer*> paramEDCforElement;
	paramEDCforElement.Add(&edc_elementproperties);


	TArray<double> enrs;
	for(int i=1; i<=discr.Y(); i++)
	{
		for(int j=1; j<=discr.X(); j++)
		{
			ed.SetInt((int)nnrs(j   +(i-1)*((int)(discr.X())+1)),"node_1");		edc_elementproperties.TreeAdd("Geometry",ed);
			ed.SetInt((int)nnrs(j+1 +(i-1)*((int)(discr.X())+1)),"node_2");		edc_elementproperties.TreeAdd("Geometry",ed);
			ed.SetInt((int)nnrs(j+1 +(i  )*((int)(discr.X())+1)),"node_3");		edc_elementproperties.TreeAdd("Geometry",ed);
			ed.SetInt((int)nnrs(j   +(i  )*((int)(discr.X())+1)),"node_4");		edc_elementproperties.TreeAdd("Geometry",ed);

	// call AddElement with prepared EDC, returnvalue is number of added element
			command_addelement->ExecuteCommand(mbs,edcParser,paramEDCforElement,ed,0);

			enrs.Add(ed.GetInt());
		}
	}
// return value:elements from current generation function
	ed.SetVector( enrs.GetDataPtr(), enrs.Length(), "list_of_elements");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Elements associated with the Beam or empty");
	edc_rv.Add(ed);
// update list_of_elements in Mesh
	mesh_enrs.Merge(enrs);
	ed.SetVector( mesh_enrs.GetDataPtr(), mesh_enrs.Length(), "list_of_elements");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Elements associated with the Mesh or empty");
	ElementDataContainer update_enrs; update_enrs.Add(ed);
	edc_handle_mesh->TreeReplaceEDCDataWith(&update_enrs);

	// first approach, slow

	return rv;
}
#else
int CEDC_ComMesh_GeneratePlate::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GeneratePlate - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GeneratePlate - Handle could not be resolved into a Mesh"));
		return 0;
	}
	if(mbs->GetMeshPtr(meshIdx)->GetType() != /*TMBSMeshType::*/TMBSMeshStructural) // (double) check type - already checked if 
	{
		edcParser->EDCError(mystr("Error in GeneratePlate - requires mesh_type = structural mesh"));
		return 0;	
	}

// first container contains the plate parameters	
	ElementDataContainer* edc_plate_params = parameter_EDCs(1);
// generate the component
	MeshQuadrilateral plate;
	ElementDataContainer edc_plate;
	plate.GetElementData(edc_plate);
	edc_plate.TreeReplaceEDCDataWith(edc_plate_params);
	plate.SetElementData(edc_plate);
// add the component to the mesh 
	int compnr = mbs->GetMeshPtr(meshIdx)->AddComponent(&plate);

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}
#endif

mystr CEDC_ComMesh_GeneratePlate::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the plate properties \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item component\\_name \n");
	descr +=			mystr("	 \\item component\\_type: Quadrilateral(default) \n");	
	descr +=			mystr("	 \\item Generation.P1 - position of first node \n");
	descr +=			mystr("	 \\item Generation.P2 - position of outer node along first direction\n");
	descr +=			mystr("	 \\item Generation.P3 - position of outer node along second direction\n");
	descr +=			mystr("	 \\item Generation.P4 - position of node opposite to the first node\n");
	descr +=      mystr("  \\item Generation.material\\_number - number of the material to be used for the plate elements\n");
	descr +=      mystr("  \\item Generation.discretization - number of the plate elements both directions - enter as 2-component ector\n");
	descr +=      mystr("  \\item Generation.thickness - thickness for the plate elements\n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls of the generated elements \n");
//N.I.Y.	descr +=      mystr("  \\item elementsize - (optional) maximum element size ( if set has priority over discretization ) \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GenerateBlock::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	
// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GenerateBlock - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GenerateBlock - Handle could not be resolved into a Mesh"));
		return 0;
	}
	if(mbs->GetMeshPtr(meshIdx)->GetType() != /*TMBSMeshType::*/TMBSMeshSolid) // (double) check type - already checked if 
	{
		edcParser->EDCError(mystr("Error in GenerateBlock - requires mesh_type = solid mesh"));
		return 0;	
	}

// first container contains the block parameters	
	ElementDataContainer* edc_block_params = parameter_EDCs(1);
// generate the component
	MeshBlock block;
	ElementDataContainer edc_block;
	block.GetElementData(edc_block);
	edc_block.TreeReplaceEDCDataWith(edc_block_params);
	block.SetElementData(edc_block);	
// add the component to the mesh 
	int compnr = mbs->GetMeshPtr(meshIdx)->AddComponent(&block);
	
	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_GenerateBlock::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the block properties \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item component\\_name \n");
	descr +=			mystr("	 \\item component\\_type: Block(default) \n");	
	descr +=			mystr("	 \\item Generation.P1 .. P8 - position of corner nodes \n");
	descr +=      mystr("  \\item Generation.material\\_number - number of the material \n"); 
	descr +=      mystr("  \\item Generation.discretization - number of the block elements all three directions\n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls of the generated elements \n");
//N.I.Y. ////descr +=      mystr("  \\item elementsize - (optional) maximum element size ( if set has priority over discretization ) \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GenerateCylinder::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	
// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GenerateCylinder - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GenerateCylinder - Handle could not be resolved into a Mesh"));
		return 0;
	}
	if(mbs->GetMeshPtr(meshIdx)->GetType() != /*TMBSMeshType::*/TMBSMeshSolid) // (double) check type - already checked if 
	{
		edcParser->EDCError(mystr("Error in GenerateCylinder - requires mesh_type = solid mesh"));
		return 0;	
	}

// first container contains the cylinder parameters	
	ElementDataContainer* edc_cylinder_params = parameter_EDCs(1);
// generate the component
	MeshCylinder cylinder;
	ElementDataContainer edc_cylinder;
	cylinder.GetElementData(edc_cylinder);
	edc_cylinder.TreeReplaceEDCDataWith(edc_cylinder_params);
	cylinder.SetElementData(edc_cylinder);	
// add the component to the mesh 
	int compnr = mbs->GetMeshPtr(meshIdx)->AddComponent(&cylinder);
	
	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_GenerateCylinder::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the cylinder properties \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=      mystr("it is assumed that the cylinder rotates around the z-axis\n");
	descr +=      mystr("the y component for all points should be 0.0\n");
	descr +=			mystr("entries in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item component\\_name \n");
	descr +=			mystr("	 \\item component\\_type: Cylinder(default) \n");	
	descr +=			mystr("	 \\item Generation.P1 - inner point of the base\n");
	descr +=			mystr("	 \\item Generation.P2 - outer point of the base\n");
	descr +=			mystr("	 \\item Generation.P3 - inner point of the top\n");
	descr +=			mystr("	 \\item Generation.P4 - outer point of the top\n");
	descr +=      mystr("  \\item Generation.material\\_number - number of the material \n"); 
	descr +=      mystr("  \\item Generation.discretization - number of the cylinder elements in radial, tangential and axial direction.\n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls of the generated elements \n");
//N.I.Y.	////descr +=      mystr("  \\item elementsize - (optional) maximum element size ( if set has priority over discretization ) \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_LoadMesh::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	
	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1); 
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in LoadMesh - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in LoadMesh - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the file type
	ElementDataContainer* edc_filetype = parameter_EDCs(1);
	ElementData* ed_filetype = edc_filetype->GetPtr(1);
	mystr filetype;
	if(ed_filetype!=NULL && ed_filetype->IsText()) { filetype = ed_filetype->GetText(); }
	else { edcParser->EDCError(mystr("Error in LoadMesh - first entry (filetype) must be a string")); }

	// second container contains the file name
	ElementDataContainer* edc_filename = parameter_EDCs(2);
	ElementData* ed_filename = edc_filename->GetPtr(1);
	mystr filename;
	if(ed_filename!=NULL && ed_filename->IsText()) { filename = ed_filename->GetText(); }
	else { edcParser->EDCError(mystr("Error in LoadMesh - second entry (filename) must be a string")); }
	
	edcParser->MakeFilenameAbsolute(filename);

	// generate the component
	int compnr = mbs->GetMeshPtr(meshIdx)->LoadMeshExternal(filetype,filename);

	return_value.SetInt(compnr,"");
	return compnr;// rv;
};

mystr CEDC_ComMesh_LoadMesh::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ operation code defining the file format \n");
	descr +=			mystr("	 \\item $2^{nd}$ filename of the file containing the mesh \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("currently implemented file formats and thus allowed values for the $1^{st}$ parameter are\n");
	descr +=			mystr("\\begin{itemize} \n");
	descr +=			mystr("  \\item Neutral3D \n");
	descr +=			mystr("	 \\item Netgen2D \n");
	descr +=			mystr("	 \\item STL \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_WriteMesh::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1); 
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in WriteMesh - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in WriteMesh - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the file type
	ElementDataContainer* edc_filetype = parameter_EDCs(2);
	ElementData* ed_filetype = edc_filetype->GetPtr(1);
	mystr filetype;
	if(ed_filetype!=NULL && ed_filetype->IsText()) { filetype = ed_filetype->GetText(); }
	else { edcParser->EDCError(mystr("Error in WriteMesh - first entry (filetype) must be a string")); }

	// third container contains the file name
	ElementDataContainer* edc_filename = parameter_EDCs(3);
	ElementData* ed_filename = edc_filename->GetPtr(1);
	mystr filename;
	if(ed_filename!=NULL && ed_filename->IsText()) { filename = ed_filename->GetText(); }
	else { edcParser->EDCError(mystr("Error in WriteMesh - second entry (filename) must be a string")); }

	// generate the component
	compnr = mbs->GetMeshPtr(meshIdx)->WriteMeshExternal(compnr,filetype,filename);

	return_value.SetInt(compnr,"");
	return compnr;// rv;
};

mystr CEDC_ComMesh_WriteMesh::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=            mystr("  \\item $1^{st}$ component number or zero for all \n");
	descr +=			mystr("	 \\item $2^{nd}$ operation code defining the file format \n");
	descr +=			mystr("	 \\item $3^{rd}$ filename of the file containing the mesh \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("currently implemented file formats and thus allowed values for the $1^{st}$ parameter are\n");
	descr +=			mystr("\\begin{itemize} \n");
	descr +=			mystr("	 \\item Neutral3D \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Transform::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Translate - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Translate - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the transformation parameters
	ElementDataContainer* edc_transformation_params = parameter_EDCs(2);
	Vector3D scale(1.,1.,1.);
	if(edc_transformation_params->TreeFind("scale"))
	{
		edc_transformation_params->TreeGetVector3D("scale",scale.X(),scale.Y(),scale.Z());
	}
	Vector3D rotate(0.,0.,0.);
	if(edc_transformation_params->TreeFind("rotate"))
	{
		edc_transformation_params->TreeGetVector3D("rotate",rotate.X(),rotate.Y(),rotate.Z());
	}
	Vector3D translate(0.,0.,0.);
	if(edc_transformation_params->TreeFind("translate"))
	{
		edc_transformation_params->TreeGetVector3D("translate",translate.X(),translate.Y(),translate.Z());
	}

	// generate the component from parameters
	compnr = mbs->GetMeshPtr(meshIdx)->TransformXYZ(compnr, translate, rotate, scale);

	// additional properties
	if(edc_transformation_params->TreeFind("name"))
	{
		mystr name = edc_transformation_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_transformation_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_transformation_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Transform::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the transformation parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=      mystr("multiple entries for a single call are allowed and are processed in the same sequence as stated below");
	descr +=			mystr("entries in the transformation parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item scale - scaling vector  \n");
	descr +=      mystr("  \\item rotate - rotation vector - rotation angles around global axis, processed in sequence (x,y,z) \n");
	descr +=			mystr("	 \\item translate - translation vector \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Distort::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Distort - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Distort - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the distortion parameters
	ElementDataContainer* edc_distortion_params = parameter_EDCs(2);
	
	mystr fx = "x";
	if(edc_distortion_params->TreeFind("fx"))
	{
		fx = edc_distortion_params->TreeGetString("fx");
	}
	mystr fy = "y";
	if(edc_distortion_params->TreeFind("fy"))
	{
		fy = edc_distortion_params->TreeGetString("fy");
	}
	mystr fz = "z";
	if(edc_distortion_params->TreeFind("fz"))
	{
		fz = edc_distortion_params->TreeGetString("fz");
	}
	int subordinate = 0;
	if (edc_distortion_params->TreeFind("subordinate"))
	{
		subordinate = edc_distortion_params->TreeGetInt("subordinate");
	}

	// generate the component from parameters
	compnr = mbs->GetMeshPtr(meshIdx)->DistortXYZ(compnr, fx, fy, fz, subordinate);

	// additional properties
	if(edc_distortion_params->TreeFind("name"))
	{
		mystr name = edc_distortion_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_distortion_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_distortion_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Distort::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the distortion parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the distortion parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item fx - function to compute new x-coordinate - string, use (x,y,z) as function parameters \n");
	descr +=      mystr("  \\item fy - function to compute new y-coordinate - string, use (x,y,z) as function parameters \n");
	descr +=			mystr("	 \\item fz - function to compute new z-coordinate - string, use (x,y,z) as function parameters \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=      mystr("when no function defined the coordinate is not changed. \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Modify::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Modify - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Modify - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the modify parameters
	ElementDataContainer* edc_modify_params = parameter_EDCs(2);

	// required entry ** entity **
	MeshModifier::WorkingOnEntity entity = MeshModifier::WorkingOnEntity::WOE_none;
	if(edc_modify_params->TreeFind("entity"))
	{
		ElementData* ed_entity =edc_modify_params->TreeFind("entity");
		if(ed_entity->IsText())
		{
			mystr entitystr = ed_entity->GetText();
// all the allowed entities here
			if(entitystr.Compare("ElementType")==1)
			{
				entity = MeshModifier::WorkingOnEntity::WOE_elemtype;
			}
			else if(entitystr.Compare("NodeType")==1)
			{
				entity = MeshModifier::WorkingOnEntity::WOE_nodetype;
			}
			else if(entitystr.Compare("MaterialNumber")==1 ||
					entitystr.Compare("FiniteElement.material_number")==1)
			{
				entity = MeshModifier::WorkingOnEntity::WOE_materialnumber;
			}
			else if(entitystr.Compare("GeometricNonlinearityStatus")==1 ||
				    entitystr.Compare("Generation.GeometricNonlinearityStatus")==1)
			{
				entity = MeshModifier::WorkingOnEntity::WOE_nonlinearity;
			}
			else 
			edcParser->EDCError(mystr("Error in Mesh.Modify(entity) - Parameter entity is not valid"));
		}
		else
		{
			edcParser->EDCError(mystr("Error in Mesh.Modify - Parameter entity is missing"));
		}
	}
	
	// required entry ** new value ** expected type may depend on entity
	int newinteger = 0;
	double newdouble = 0.;
	mystr newmystr = "";
	if(edc_modify_params->TreeFind("newValue"))
	{
		//ElementData* ed_newval = edc_modify_params->Get(edc_modify_params->TreeFind("newValue"));
		if(entity == MeshModifier::WorkingOnEntity::WOE_elemtype)
		{
			mystr newTypeStr = edc_modify_params->TreeGetString("newValue");
			newinteger = mbs->GetMeshPtr(meshIdx)->GetComponentElemTypeIdx(newTypeStr);
			if(newinteger<=0) 
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(ElementType) - Parameter newValue is not a valid ElementType string"));
				return 0;			
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_nodetype)
		{
			mystr newTypeStr = edc_modify_params->TreeGetString("newValue");
			newinteger = mbs->GetMeshPtr(meshIdx)->GetComponentNodeTypeIdx(newTypeStr);
			if(newinteger<=0) 
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(NodeType) - Parameter newValue is not a valid NodeType string"));
				return 0;			
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_materialnumber)
		{
			newinteger = edc_modify_params->TreeGetInt("newValue");
			if(newinteger < mbs->NMaterials())
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(MaterialNumber) - Parameter newValue outside of range of existing materials"));
				return 0;
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_nonlinearity)
		{
			newinteger = edc_modify_params->TreeGetInt("newValue");
			if(newinteger>2||newinteger<-1)
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(GeometricNonlinearityStatus) - Parameter newValue outside of allowed range [-1..+2] "));
				return 0;				
			}
		}
		//else ;
	}
	else
	{
		edcParser->EDCError(mystr("Error in Mesh.Modify - Parameter newValue is missing"));
		return 0;	
	}

	// optional entry ** old value ** expected type may depend on entity 
	int oldinteger = 0;
	double olddouble = 0.;
	mystr oldmystr = "";
	if(edc_modify_params->TreeFind("oldValue"))
	{
		if(entity == MeshModifier::WorkingOnEntity::WOE_elemtype)
		{
			mystr oldTypeStr = edc_modify_params->TreeGetString("oldValue");
			oldinteger = mbs->GetMeshPtr(meshIdx)->GetComponentElemTypeIdx(oldTypeStr);
			if(oldTypeStr.Length()>0 && oldinteger<=0) 
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(ElementType) - Parameter oldValue is not a valid ElementType string"));
				return 0;			
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_nodetype)
		{
			mystr oldTypeStr = edc_modify_params->TreeGetString("oldValue");
			oldinteger = mbs->GetMeshPtr(meshIdx)->GetComponentNodeTypeIdx(oldTypeStr);
			if(oldTypeStr.Length()>0 && oldinteger<=0) 
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(NodeType) - Parameter oldValue is not a valid NodeType string"));
				return 0;			
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_materialnumber)
		{
			oldinteger = edc_modify_params->TreeGetInt("oldValue");
			if(oldinteger < mbs->NMaterials())
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(MaterialNumber) - Parameter oldValue outside of range of existing materials"));
				return 0;
			}
		}
		else
		if(entity == MeshModifier::WorkingOnEntity::WOE_nonlinearity)
		{
			oldinteger = edc_modify_params->TreeGetInt("oldValue");
			if(oldinteger>2||oldinteger<-1)
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify(GeometricNonlinearityStatus) - Parameter oldValue outside of allowed range [-1..+2] "));
				return 0;				
			}
		}
		//else ;
	}

	// *optional entry * filter, an existing set **
	BaseSet* pFilter = NULL;
	if(edc_modify_params->TreeFind("filter"))
	{
		ElementData* ed_set = edc_modify_params->TreeFind("filter"); 
		int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
		if(setIdx==0) 	// check existence
		{
			edcParser->EDCError(mystr("Error in Mesh.Modify - parameter entry \"filter\" could not be resolved into a Set"));
			return 0;
		}
		pFilter = mbs->GetSetPtr(setIdx);
		if(entity == MeshModifier::WorkingOnEntity::WOE_elemtype ||
		   entity == MeshModifier::WorkingOnEntity::WOE_nodetype ||
		   entity == MeshModifier::WorkingOnEntity::WOE_materialnumber ||
		   entity == MeshModifier::WorkingOnEntity::WOE_nonlinearity)
		{
			if(pFilter->GetType() != /*TMBSSetType::*/TMBSSetElements)
			{
				edcParser->EDCError(mystr("Error in Mesh.Modify - requires set_type = set_of_elements"));
				return 0;
			}
		}
	}

	// generate the component
	compnr = mbs->GetMeshPtr(meshIdx)->Modify(compnr, entity, 
		                                                      newinteger, newdouble, newmystr, 
															  oldinteger, olddouble, oldmystr,                     
															                                   pFilter);

	// additional properties
	if(edc_modify_params->TreeFind("name"))
	{
		mystr name = edc_modify_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Modify::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the modify parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the modify parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("  \\item entity - string - valid entries are: \"NodeType\",\"ElementType\",\"MaterialNumber\",\"GeometricNonlinearityStatus\" \n");
	descr +=			mystr("  \\item newValue - various, matching the entity... - e.g. for entity ElementType any valid ElementTypeString    \n");
	descr +=			mystr("	 \\item filter - integer - index number of a Set added to the system or zero \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Linear2Quadratic::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	edcParser->EDCDbgMsg(mystr("identified command: Linear2Quadratic."));

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in AddMeshToMBS - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in AddMeshToMBS - Handle could not be resolved into a Mesh"));
		return 0;
	}
	edcParser->EDCDbgMsg(mystr("meshIdx resolved."));

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the additional parameters
	ElementDataContainer* edc_lin2quad_params = parameter_EDCs(2);

	// generate the component
	compnr = mbs->GetMeshPtr(meshIdx)->LinearToQuadratic(compnr);

	// additional properties
	if(edc_lin2quad_params->TreeFind("name"))
	{
		mystr name = edc_lin2quad_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_lin2quad_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_lin2quad_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}
	
	return_value.SetInt(compnr,"");

	return compnr;// rv;
}

mystr CEDC_ComMesh_Linear2Quadratic::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the additional parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the additional parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_SplitHexes::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in SplitHexes - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in SplitHexes - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the transformation parameters
	ElementDataContainer* edc_split_params = parameter_EDCs(2);
	int split = 5;
	if(edc_split_params->TreeFind("split"))
	{
		split = edc_split_params->TreeGetInt("split");
	}

	// generate the component from parameters
	if(split==2) { compnr = mbs->GetMeshPtr(meshIdx)->SplitHexes(compnr,2); } // Hex -> Prisms
	else if(split==3) { compnr = mbs->GetMeshPtr(meshIdx)->SplitHexes(compnr,3); } // Hex -> Pyramids
	else if(split==5) { compnr = mbs->GetMeshPtr(meshIdx)->SplitHexes(compnr,5); } // Hex -> Tets
	else { edcParser->EDCError(mystr("Error in SplitHexes - can only split into 2, 3 or 5")); return 0; }

	// additional properties
	if(edc_split_params->TreeFind("name"))
	{
		mystr name = edc_split_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_split_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_split_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_SplitHexes::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the conversion parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item split - number of items to split the hexahedral into - may be (2), (3), (5)  \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");	
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Refine::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Refine - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Refine - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the distortion parameters
	ElementDataContainer* edc_refinement_params = parameter_EDCs(2);
	
	double* v;
	int l;
	edc_refinement_params->TreeGetVector("Generation.level", &v, l);
	TArray<int> nodesdlvl(l);
	for(int i=1; i<=l; i++) nodesdlvl(i) = (int) (v[i-1]+0.5);

	int method = 0;
	if(edc_refinement_params->Find("Generation.method")!=0) { edc_refinement_params->TreeGetInt("Generation.method",0); }

// generate the component from parameters
	compnr = mbs->GetMeshPtr(meshIdx)->Refine(compnr, nodesdlvl, method);

	// additional properties
	if(edc_refinement_params->TreeFind("name"))
	{
		mystr name = edc_refinement_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_refinement_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_refinement_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}
	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Refine::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the refinementparameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the distortion parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("  \\item Generation.level - list of subdivision levels for all nodes of the subordinate component  \n");
	descr +=			mystr("  \\item Generation.method - string defining the method: default 0 for all directions. 1,2 or 3 to skip single local direction  \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Rotate::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Rotate - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Rotate - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the transformation parameters
	ElementDataContainer* edc_rotate_params = parameter_EDCs(2);
	int axisnr=1;
	if(edc_rotate_params->TreeFind("axis_number"))
	{
		axisnr = edc_rotate_params->TreeGetInt("axis_number");
	}
	int angseg=2;
	if(edc_rotate_params->TreeFind("angular_segments"))
	{
		angseg = edc_rotate_params->TreeGetInt("angular_segments");
	}
	double totangle=90.0; 
	if(edc_rotate_params->TreeFind("total_angle"))
	{
		totangle = edc_rotate_params->TreeGetDouble("total_angle");
	}

	// generate the component
	compnr = mbs->GetMeshPtr(meshIdx)->Rotational(compnr,axisnr,angseg,totangle);

	// additional properties 
	if(edc_rotate_params->TreeFind("name"))
	{
		mystr name = edc_rotate_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_rotate_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_rotate_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Rotate::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the conversion parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item axis\\_number - rotates around the axis 1..x 2..y 3..z \n");
	descr +=      mystr("  \\item angular\\_segments - number of elements around the circumflex \n");
	descr +=      mystr("  \\item total\\_angle - 360 for a full rotation \n");
	descr +=			mystr("\\end{itemize} \n");	
	descr +=			mystr("additionally the following parameters may be specified for the extruded component \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item thickness - thickness for the plates when extruding curves \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");	
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Mirror::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Mirror - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Mirror - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the number of the mirror plane
	ElementDataContainer* edc_mirror_params = parameter_EDCs(2);
	int planenr = 3;
// for the old version: 2nd parameter could be a simple integer - passed in as text
	if(edc_mirror_params->TreeFind("parameter2"))
	{
		ElementData* ed_planenr = edc_mirror_params->TreeFind("parameter2");
		planenr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_planenr, meshIdx);
		edcParser->EDCWarning(mystr("The call to MeshMirror has been changed, please use a parameterEDC as second parameter!"));
	}
	else if(edc_mirror_params->TreeFind("plane"))
	{
		planenr = edc_mirror_params->TreeGetInt("plane",0);
	}

	// generate the component from parameters
	compnr = mbs->GetMeshPtr(meshIdx)->Mirror(compnr,planenr);

	// additional properties
	if(edc_mirror_params->TreeFind("name"))
	{
		mystr name = edc_mirror_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_mirror_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_mirror_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}

	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Mirror::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the mirror parameters");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("entries in the mirror parameters EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("  \\item plane - 1..x=0, 2..y=0, 3..z=0 \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_Extrude::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in Extrude - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in Extrude - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number
	ElementData* ed_compnr = parameter_EDCs(1)->GetPtr(1);
	int compnr = ResolveParameterEntryToComponentNumber(mbs, edcParser, ed_compnr, meshIdx);

	// second container contains the transformation parameters
	ElementDataContainer* edc_extrude_params = parameter_EDCs(2);
	int axisnr=3;
	if(edc_extrude_params->TreeFind("axis_number"))
	{
		axisnr = edc_extrude_params->TreeGetInt("axis_number");
	}
	int discr=1;
	if(edc_extrude_params->TreeFind("discretization"))
	{
		discr = edc_extrude_params->TreeGetInt("discretization");
	}
	double dist=1.0; 
	if(edc_extrude_params->TreeFind("total_extrusion"))
	{
		dist = edc_extrude_params->TreeGetDouble("total_extrusion");
	}
	double thickness=0.1;
	if(edc_extrude_params->TreeFind("thickness"))
	{
		thickness = edc_extrude_params->TreeGetDouble("thickness");
	}

	// generate the component from parameters
	compnr = mbs->GetMeshPtr(meshIdx)->Extrude(compnr, axisnr,discr,dist, thickness);

	// additional properties
	if(edc_extrude_params->TreeFind("name"))
	{
		mystr name = edc_extrude_params->TreeGetString("name");
		if(name.Length()>0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
		}
	}
	if(edc_extrude_params->TreeFind("Generation.GeometricNonlinearityStatus"))
	{
		int	gnls = edc_extrude_params->TreeGetInt("Generation.GeometricNonlinearityStatus",-1);
		if(gnls>=0)
		{
			mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetGeometricNonlinearityStatus(gnls);
		}
	}
	return_value.SetInt(compnr,"");
	return compnr;// rv;
}

mystr CEDC_ComMesh_Extrude::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component number \n");	
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing the conversion parameters \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=      mystr("  \\item axis\\_number - extrude along the axis 1..x 2..y 3..z \n");
	descr +=      mystr("  \\item discretizaiton - number of elements along the axis \n");
	descr +=      mystr("  \\item total\\_extrusion - extrusion distance \n");
	descr +=			mystr("	 \\item name - name of the component \n");
	descr +=      mystr("  \\item Generation.GeometricNonlinearityStatus - gnls for all nodes below - -1..dont change 0..Linear 1..NonlinearSmallStrain 2..NonlinearLargeStrain \n");
	descr +=			mystr("\\end{itemize} \n");	
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_AddMeshToMBS::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	edcParser->EDCDbgMsg(mystr("identified: command AddMeshToMBS."));

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in AddMeshToMBS - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in AddMeshToMBS - Handle could not be resolved into a Mesh"));
		return 0;
	}
	edcParser->EDCDbgMsg(mystr("meshIdx resolved."));
	
	// generates the nodes, elements, constraints in MBS 
	int hook = mbs->AddMeshToMBS(meshIdx);
	edcParser->EDCDbgMsg(mystr("mesh added to MBS."));

	ElementDataContainer* edc_mesh = new ElementDataContainer;
	mbs->GetMeshPtr(meshIdx)->GetElementData(*edc_mesh); // NOTE AD: this is the problematic line of code !!! 
	                                                     // one identified cause of crash here is: mesh pointer in any component not set properly 
	                                                     // --> crashes in BaseMesh::MapNodeListToMBSNodes()
	edcParser->EDCDbgMsg(mystr("meshEDC retrieved."));

	return_value.SetEDC(edc_mesh, "rv");
	edcParser->EDCDbgMsg(mystr("meshEDC stored as return value."));
	edcParser->EDCDbgMsg(mystr("================================="));


	mbs->Assemble();
	edcParser->EDCDbgMsg(mystr("MBS assembled."));

	return 1;
}

mystr CEDC_ComMesh_AddMeshToMBS::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("There are no parameters for this command, for compatibility please enter a dummy '0' at this position. \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

#ifdef mode_old
int CEDC_ComMesh_GetNodesInBox::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// container 1 contains the corners of the box
	Vector3D p1, p2;
	ElementDataContainer* edc_boxparameters = parameter_EDCs(1);
	edc_boxparameters->TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());
	edc_boxparameters->TreeGetVector3D("P2",p2.X(), p2.Y(), p2.Z());
	Box3D box(p1,p2);

// container 2 contains the name of the handle ( can not pass the EDC* itself because that is going to be deleted... )
// consistency already checked in ParseCommandParameters...
	mystr handlename = parameter_EDCs(2)->TreeGetString("handle");
	ElementDataContainer* edc_local = edcParser->GetMBSParser()->GetLocalEDC(); 
	int nr_handle = edc_local->Find(handlename);
	ElementDataContainer* edc_handle_mesh = edc_local->Get(nr_handle).GetEDC();


// get current lists for mesh object
	double* vdata;
	int vlen;
	edc_handle_mesh->TreeGetVector("list_of_nodes",&vdata,vlen);            
	TArray<double> mesh_nnrs; mesh_nnrs.CopyFrom(vdata,vlen); mesh_nnrs.SetLen(vlen);			// node numbers already associated with the Mesh

	TArray<double> nnrs;
// loop over all nodes
	for (int i=1; i<= mesh_nnrs.Length(); i++)
	{
		Vector3D nodepos = mbs->GetNode((int)mesh_nnrs(i)).RefConfPos();
		if (box.IsIn(nodepos) /*&& !mesh_redundant_nodes.Find(mesh_nnrs(i))*/)
		{
			nnrs.Add(mesh_nnrs(i));
		}
	}

// return value:nodes from current generation function
	return_value.SetVector( nnrs.GetDataPtr(), nnrs.Length(), "list_of_nodes");		return_value.SetValuesInt();	return_value.SetVariableLength();	return_value.SetToolTipText("List of Nodes returned from GetNodesInBox");

	return rv;
}
#else
int CEDC_ComMesh_GetNodesInBox::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetNodesInBox - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetNodesInBox - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the box parameters
	ElementDataContainer* edc_box_parameters = parameter_EDCs(1);
	Vector3D p1, p2;
	edc_box_parameters->TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());
	edc_box_parameters->TreeGetVector3D("P2",p2.X(), p2.Y(), p2.Z());
	
	TArray<int> nodelist = mbs->GetMeshPtr(meshIdx)->GetNodesInBox(Box3D(p1,p2));
	Vector nodenums(nodelist.Length()); for(int i=1; i<=nodelist.Length(); i++) { nodenums(i) = nodelist(i); }

	return_value.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "nodes_in_box"); return_value.SetValuesInt(); return_value.SetVariableLength(); /* return_value.SetLocked(1); */ return_value.SetToolTipText("retured by a call of 'GetNodesInBox'");
	return 1;// rv;
};
#endif
mystr CEDC_ComMesh_GetNodesInBox::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the box (defined by two corner) \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item P1 - position of corner 1 \n");
	descr +=			mystr("	 \\item P2 - position of corner 2 \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetNodesInCylinderShell::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetNodesInCylinder - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetNodesInCylinder - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the cylinder shell parameters
	ElementDataContainer* edc_cylinder_parameters = parameter_EDCs(1);
	Vector3D bot(0.,0.,0.);
	if(edc_cylinder_parameters->TreeFind("P1"))
	{
		edc_cylinder_parameters->TreeGetVector3D("P1",bot.X(), bot.Y(), bot.Z());
	}
	Vector3D top(1.,0.,0.);
	if(edc_cylinder_parameters->TreeFind("P2"))
	{
		edc_cylinder_parameters->TreeGetVector3D("P2",top.X(), top.Y(), top.Z());
	}
	double rout = 1.;
	if(edc_cylinder_parameters->TreeFind("Rout"))
	{
		rout = edc_cylinder_parameters->TreeGetDouble("Rout");
	}
	double rin = 0.;
	if(edc_cylinder_parameters->TreeFind("Rin"))
	{
		rin = edc_cylinder_parameters->TreeGetDouble("Rin");
	}
	
	TArray<int> nodelist = mbs->GetMeshPtr(meshIdx)->GetNodesInCylinderShell(bot,top,rout,rin);
	Vector nodenums(nodelist.Length()); for(int i=1; i<=nodelist.Length(); i++) { nodenums(i) = nodelist(i); }

	return_value.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "nodes_in_cylinder"); return_value.SetValuesInt(); return_value.SetVariableLength(); /* return_value.SetLocked(1); */ return_value.SetToolTipText("retured by a call of 'GetNodesInBox'");
	return 1;// rv;
};
mystr CEDC_ComMesh_GetNodesInCylinderShell::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the cylinder (defined by axis and two radii) \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item P1 - position of bottom center \n");
	descr +=			mystr("	 \\item P2 - position of top center \n");
	descr +=			mystr("	 \\item Rout - outer shell radius \n");
	descr +=			mystr("	 \\item Rin - inner shell radius, default 0 for full cylinder \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetNodesInSphericalShell::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetNodesInSphere - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetNodesInSphere - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the cylinder shell parameters
	ElementDataContainer* edc_sphere_parameters = parameter_EDCs(1);
	Vector3D center(0.,0.,0.);
	if(edc_sphere_parameters->TreeFind("P1"))
	{
		edc_sphere_parameters->TreeGetVector3D("P1",center.X(), center.Y(), center.Z());
	}
	double rout = 1.;
	if(edc_sphere_parameters->TreeFind("Rout"))
	{
		rout = edc_sphere_parameters->TreeGetDouble("Rout");
	}
	double rin = 0.;
	if(edc_sphere_parameters->TreeFind("Rin"))
	{
		rin = edc_sphere_parameters->TreeGetDouble("Rin");
	}

	TArray<int> nodelist = mbs->GetMeshPtr(meshIdx)->GetNodesInSphericalShell(center,rout,rin);
	Vector nodenums(nodelist.Length()); for(int i=1; i<=nodelist.Length(); i++) { nodenums(i) = nodelist(i); }

	return_value.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "nodes_in_sphere"); return_value.SetValuesInt(); return_value.SetVariableLength(); /* return_value.SetLocked(1); */ return_value.SetToolTipText("retured by a call of 'GetNodesInBox'");
	return 1;// rv;
};
mystr CEDC_ComMesh_GetNodesInSphericalShell::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the sphere (defined by center and two radii) \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item P1 - position of center \n");
	descr +=			mystr("	 \\item Rout - outer shell radius \n");
	descr +=			mystr("	 \\item Rin - inner shell radius, default 0 for full cylinder \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetNodesInFunction::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetNodesInFunction - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetNodesInFunction - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the cylinder shell parameters
	ElementDataContainer* edc_function_parameters = parameter_EDCs(1);
	mystr function_string("0");
	if(edc_function_parameters->TreeFind("function"))
	{
		function_string = edc_function_parameters->TreeGetString("function");
	}

	MathFunction mf;
	mf.SetExpression(function_string,mystr("x,y,z"),mbs);

	TArray<int> nodelist = mbs->GetMeshPtr(meshIdx)->GetNodesInMathFunction(mf);
	Vector nodenums(nodelist.Length()); for(int i=1; i<=nodelist.Length(); i++) { nodenums(i) = nodelist(i); }

	return_value.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "nodes_in_function"); return_value.SetValuesInt(); return_value.SetVariableLength(); /* return_value.SetLocked(1); */ return_value.SetToolTipText("retured by a call of 'GetNodesInBox'");
	return 1;// rv;
};
mystr CEDC_ComMesh_GetNodesInFunction::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the function \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item function - definition of the function as single string, dependencies on (x,y,z) as node coordinates are assumed.  \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetNodePos::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetNodePosition - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetNodePosition - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the global node number
	ElementData* ed_gnnr = parameter_EDCs(1)->GetPtr(1);
	int globalnodenumber = 0;
	if(ed_gnnr->IsInt())
	{
		globalnodenumber = ed_gnnr->GetInt(); // integer variabe
	}
	else if(ed_gnnr->IsDouble())
	{
		globalnodenumber = (int) ed_gnnr->GetDouble();
	}
	else if(ed_gnnr->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr str = ed_gnnr->GetText();
		globalnodenumber = str.MakeInt();
	}

	GenericNode* gn = (GenericNode*)mbs->GetNodePtr(globalnodenumber);
	Vector3D nodepos = gn->GetRefPos();

	return_value.SetVector3D(nodepos.X(),nodepos.Y(),nodepos.Z(),"global_position");
	return 1;
}

mystr CEDC_ComMesh_GetNodePos::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item global node number \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetFacesFromNodes::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in FacesFromNodes - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in FacesFromNodes - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container holds the set of global node numbers
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in FacesFromNodes - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in FacesFromNodes - requires set_type = set_of_global_nodes"));
		return 0;
	}
	GlobalNodeSet* gnset = (GlobalNodeSet*) mbs->GetSetPtr(setIdx);
	TArray<int>& nodenumbers = gnset->GlobalNodeNumbers();

	// second container may hold a set of element numbers
	ed_set = parameter_EDCs(2)->GetPtr(1);
	setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	//if(setIdx==0) 	// ZERO is explicitely allowed here by parameter definitions (no restrictions for elements)
	//{
	//	edcParser->EDCError(mystr("Error in FacesFromNodes - Parameter 2 could not be resolved into a Set"));
	//	return 0;
	//}
	if(setIdx==0) 
	{
		setIdx = mbs->GetMeshPtr(meshIdx)->GetFacesFromNodes( gnset->GlobalNodeNumbers(), TArray<int>(0));
	}
	else
	{
		if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetElements) 	// check type
		{
			edcParser->EDCError(mystr("Error in FacesFromNodes - requires set_type = set_of_elements"));
			return 0;
		}
		ElementSet* elset = (ElementSet*) mbs->GetSetPtr(setIdx);
		setIdx = mbs->GetMeshPtr(meshIdx)->GetFacesFromNodes( gnset->GlobalNodeNumbers(), elset->ElementNumbers() );
	}

	FaceSetA* theFaces = (FaceSetA*)mbs->GetSetPtr(setIdx);

	TArray<int> elementnumbers = theFaces->ElementNumbers();
	TArray<int> elementfacenumbers = theFaces->FaceNumbers();
	//TArray<int4> facenodenumbers;
	//TArray<double> facearea;

	ElementData ed;
	Vector elnrs(elementnumbers.Length()); for(int i=1; i<=elementnumbers.Length(); i++) { elnrs(i) = elementnumbers(i); }
	ed.SetVector(elnrs.GetVecPtr(), elnrs.Length(), "element_numbers"); ed.SetInt(1); ed.SetVariableLength(); ed.SetToolTipText("element numbers for the identified faces"); edc_rv.Add(ed);
	Vector elfanrs(elementfacenumbers.Length()); for(int i=1; i<=elementfacenumbers.Length(); i++) { elfanrs(i) = elementfacenumbers(i); }
	ed.SetVector(elfanrs.GetVecPtr(), elfanrs.Length(), "element_facet_numbers"); ed.SetInt(1); ed.SetVariableLength(); ed.SetToolTipText("element facet numbers for the identified faces"); edc_rv.Add(ed);

	TArray<int> & facenodelengths = theFaces->FaceNodeLengths();
	Vector vv(facenodelengths.Length());
	for(int i=1; i<=facenodelengths.Length(); i++)
	{
		vv(i) = facenodelengths(i);
	} 
	ed.SetVector(vv.GetVecPtr(), vv.Length(), "number_of_nodes"); ed.SetValuesInt(); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("number of nodes for each face in this set"); 
	edc_rv.Add(ed);
		
	TArrayDynamic<TArray<int>> & facenodenumbers = theFaces->FaceNodeNumbers();
	Matrix mm(facenodenumbers.Length(),8);
	mm.SetAll(0);
	for(int i=1; i<=facenodenumbers.Length(); i++)
	{
		for(int j=1; j<=facenodelengths(i); j++)
		{
			mm(i,j) = facenodenumbers(i)(j);
		}
	}
	ed.SetMatrix(mm.GetMatPtr(),facenodenumbers.Length(),8,"node_lists"); ed.SetValuesInt(); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("node list for each face in this set");
	edc_rv.Add(ed);

	mm.SetAll(0);
	for(int i=1; i<=facenodenumbers.Length(); i++)
	{
		for(int j=1; j<=facenodelengths(i); j++)
		{
			Element * el= mbs->GetElementPtr(elementnumbers(i));
			int index = 0;
			for (int k=1; k<=el->NNodes(); k++)
			{
				if (el->GetNode(k).NodeNum() == facenodenumbers(i)(j))
				{
					index = k;
				}
			}
			mm(i,j) = index;
		}
	}
	ed.SetMatrix(mm.GetMatPtr(),facenodenumbers.Length(),8,"node_lists_local"); ed.SetValuesInt(); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("node list for each face in this set");
	edc_rv.Add(ed);

	TArray<double> & faceareas = theFaces->FaceAreas();
	vv.SetAll(0.);
	for(int i=1; i<=faceareas.Length(); i++)
	{
		vv(i) = faceareas(i);
	} 
	ed.SetVector(vv.GetVecPtr(), vv.Length(), "face_areas"); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("ElementFaceAreas in ths set"); 
	edc_rv.Add(ed);


	ElementDataContainer edc_gn_match;
	
	Matrix em(nodenumbers.Length(),8);
	Matrix fm(nodenumbers.Length(),8);
	Matrix lnm(nodenumbers.Length(),8);

	Vector n_entries(nodenumbers.Length());
	const TArray<TArray<int>*> &n2e = mbs->GetMeshPtr(meshIdx)->N2E();


	int lastCompIndex = 0;
	for (int i = 1, n = mbs->GetMeshPtr(meshIdx)->NComp(); i<=n; ++i)
	{
		if (mbs->GetMeshPtr(meshIdx)->GetMeshComponent(i) != NULL)
		{
			lastCompIndex=i;
		}
	}

	if (lastCompIndex != 1 || meshIdx!=1 )
	{
		mbs->UO(UO_LVL_warn) << "GetFacesFromNodes.fullMatch only working for mesh with mbs number 1 if it has only one component";
	}
	else
	{
		MeshComponent* mesh_comp = mbs->GetMeshPtr(meshIdx)->GetMeshComponent(1);
		for(int i=1; i<=nodenumbers.Length(); i++)
		{
			TArray<int> *elements = n2e(nodenumbers(i));
			int index = 1;
			for(int j=1; j<=elements->Length(); j++)
			{
				//Element * el= mbs->GetElementPtr(elements->Get(j));

				MeshElement * m_el = mbs->GetMeshPtr(meshIdx)->GetElementPtr(elements->Get(j));

				int nface = m_el->NFaces();
				TArray<int> elementnodelist = mesh_comp->GetElementNodeList(elements->Get(j));
				for (int k = 1; k<=nface; k++)
				{
					TArray<int> facenodelist = mesh_comp->GetElementFaceNodeList(elements->Get(j),k);
					int n_nodes = facenodelist.Length();
					facenodelist &= nodenumbers;
					if (n_nodes == facenodelist.Length())
					{
						em(i,index) = elements->Get(j);
						fm(i,index) = k;
						lnm(i,index)= elementnodelist.Find(nodenumbers(i));
						index++;
					}
				}
			}
			n_entries(i) = index-1.;
		}

		ed.SetMatrix(em.GetMatPtr(),nodenumbers.Length(),8,"element_list"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("element list of each global node of the set");
		edc_gn_match.Add(ed);

		ed.SetMatrix(fm.GetMatPtr(),nodenumbers.Length(),8,"facet_list"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("face list of each global node of the set");
		edc_gn_match.Add(ed);

		ed.SetMatrix(lnm.GetMatPtr(),nodenumbers.Length(),8,"local_node_list"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("local node number list of each global node of the set");
		edc_gn_match.Add(ed);

		ed.SetVector(n_entries.GetVecPtr(), n_entries.Length(), "number_of_faces"); ed.SetValuesInt(); ed.SetVariableLength(); 
		ed.SetLocked(1); ed.SetToolTipText("number of faces for each node in this set"); 
		edc_gn_match.Add(ed);
		
		ed.SetEDC(&edc_gn_match, "global_nodes_match");
		edc_rv.Add(ed);
	}

	return_value.SetEDC(&edc_rv,"result");
	return 1;// rv;
}

mystr CEDC_ComMesh_GetFacesFromNodes::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of global node numbers \n");
	descr +=            mystr("  \\item $2^{nd}$ parameter: a set of elements to process, use 0 for all elements \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

#ifdef mode_old
int CEDC_ComMesh_GlueMesh::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	// default parameter values, to be overridden by user
	ElementDataContainer edc_glue_parameters;
	ElementData ed;

	Vector nodes;
	ed.SetVector(nodes.GetVecPtr(), 0, "nodes_to_constrain");	ed.SetValuesInt(); ed.SetVariableLength();			edc_glue_parameters.Add(ed);
	ed.SetBool(0, "is_ground");	ed.SetToolTipText("default 0: constrain node pairs - set to 1 for ground constraint");	edc_glue_parameters.Add(ed);
	ed.SetBool(0, "remove_redundant_only");		ed.SetToolTipText("default 0: connect pairs - set to 1 to only remove redundant nodes"); edc_glue_parameters.Add(ed);
	ed.SetVector3D(1., 1., 1., "constrained_directions");	ed.SetValuesInt();										edc_glue_parameters.Add(ed);
	ed.SetVector3D(1., 1., 1., "constrained_rotations");	ed.SetValuesInt();										edc_glue_parameters.Add(ed);

	edc_glue_parameters.TreeReplaceEDCDataWith(parameter_EDCs(1));

	int len; double* p_vec;
	edc_glue_parameters.TreeGetVector("nodes_to_constrain", &p_vec, len);	// ist das so gut????
	nodes.LinkWith(p_vec, len);

	int is_ground = edc_glue_parameters.TreeGetBool("is_ground", 0);
	int redundant_only = edc_glue_parameters.TreeGetBool("remove_redundant_only", 0);


	Vector constrained_dirs, constrained_rots;
	edc_glue_parameters.TreeGetVector("constrained_directions", &p_vec, len);
	constrained_dirs.LinkWith(p_vec, len);
	edc_glue_parameters.TreeGetVector("constrained_rotations", &p_vec, len);
	constrained_rots.LinkWith(p_vec, len);

	// HACK: first of all, build a node2elements list
	mystr handlename = parameter_EDCs(2)->TreeGetString("handle");
	ElementDataContainer* edc_local = edcParser->GetMBSParser()->GetLocalEDC(); 
	int nr_handle = edc_local->Find(handlename);
	ElementDataContainer* edc_handle_mesh = edc_local->Get(nr_handle).GetEDC();

	double* elems_data, * nodes_data, * redundant_nodes_data; 
	int elems_len, nodes_len, redundant_nodes_len;
	
	edc_handle_mesh->TreeGetVector("list_of_nodes",&nodes_data,nodes_len);            
	TArray<double> mesh_nodes; mesh_nodes.CopyFrom(nodes_data,nodes_len); mesh_nodes.SetLen(nodes_len);			// node numbers already associated with the Mesh
	
	edc_handle_mesh->TreeGetVector("list_of_elements",&elems_data,elems_len);
	TArray<double> mesh_elems; mesh_elems.CopyFrom(elems_data,elems_len); mesh_elems.SetLen(elems_len);			// element numbers already associated with the Mesh
	
	edc_handle_mesh->TreeGetVector("list_of_redundant_nodes",&redundant_nodes_data,redundant_nodes_len);            
	TArray<double> mesh_redundant_nodes; mesh_redundant_nodes.CopyFrom(redundant_nodes_data,redundant_nodes_len); mesh_redundant_nodes.SetLen(redundant_nodes_len);			// node numbers already associated with the Mesh

	TArray<TArray<int>> nodes2elems;
	nodes2elems.SetLen(mesh_nodes.Length()); // AD only works for continuous list...
	nodes2elems.SetLen((int)mesh_nodes.Last());   // initialize
	for (int i = 1; i <= mesh_elems.Length(); i++)
	{
		Element& e = mbs->GetElement((int)mesh_elems(i));
		for (int j = 1; j <= e.NNodes(); j++)
		{
			nodes2elems(e.GetNode(j).NodeNum()).Add((int)mesh_elems(i));
		}
	}

	ElementDataContainer edc_elementproperties;		
	ed.SetText("GenericBodyJoint","element_type");		edc_elementproperties.Add(ed);

	int ed_num = edc_glue_parameters.Find("constrained_directions");
	if (ed_num)
	{
		edc_elementproperties.TreeAdd("Physics.Lagrange", edc_glue_parameters.Get(ed_num));
	}
	ed_num = edc_glue_parameters.Find("constrained_rotations");
	if (ed_num)
	{
		edc_elementproperties.TreeAdd("Physics.Lagrange", edc_glue_parameters.Get(ed_num));
	}
	ed_num = edc_glue_parameters.Find("size");
	/*ed.SetText("UniversalJoint", "element_type");	edc_elementproperties.Add(ed);
	ed.SetVector3D(0., 0., 1., "axis");				edc_elementproperties.TreeAdd("Position1", ed);
	ed.SetVector3D(0., 0., 1., "axis");				edc_elementproperties.TreeAdd("Position2", ed);*/

	mystr commandname = mystr("AddConnector");
	int nparam, hasretval, c_option;
	CEDC_Command* command_addrigidjoint = edcParser->GetCommand( edcParser->GetCommandType(commandname, nparam, hasretval, c_option) );

	TArray<double> new_redundant_nodes;		// nodes to remove from mbs
	// pairs
	if (!is_ground)
	{
		TArray<int2> node_pairs;
		
		for (int i = 1; i <= nodes.Length(); i++)
		{
			for (int j = i+1; j <= nodes.Length(); j++)
			{
				if ( nodes(j)!=0 && nodes(i)!=0 ) // skip redundant nodes ( marked in a previous iteration )
				{
					Node& n1 = mbs->GetNode((int)nodes(i));
					Node& n2 = mbs->GetNode((int)nodes(j));

					Vector3D dist = n1.RefConfPos() - n2.RefConfPos();

					if (dist.Norm() < 1e-10)
					{
						// check if nodes are identical; if so, replace n2 by n1
						// TODO: == operator for nodes!
						if (
						// code for ANCFNodeS1S2_3D ( node for generate plates
							(typeid(n1) == typeid(ANCFNodeS1S2_3D) && typeid(n2) == typeid(ANCFNodeS1S2_3D) &&
							(dynamic_cast<ANCFNodeS1S2_3D&>(n1)).GetRefSlope1() == (dynamic_cast<ANCFNodeS1S2_3D&>(n2)).GetRefSlope1() &&
							(dynamic_cast<ANCFNodeS1S2_3D&>(n1)).GetRefSlope2() == (dynamic_cast<ANCFNodeS1S2_3D&>(n2)).GetRefSlope2())
						||
						// code for Node3DRxyz ( node for generate beams
							(typeid(n1) == typeid(Node3DRxyz) && typeid(n2) == typeid(Node3DRxyz) &&
							(dynamic_cast<Node3DRxyz&>(n1)).GetRefAngles() == (dynamic_cast<Node3DRxyz&>(n2)).GetRefAngles()) )

						{
							// replace in elements
							for (int k = 1; k <= nodes2elems((int)nodes(j)).Length(); k++)
							{
								Element& e = mbs->GetElement(nodes2elems((int)nodes(j))(k));
								for (int l = 1; l <= e.NNodes(); l++)
								{
									if (e.NodeNum(l) == n2.NodeNum())
									{
										e.NodeNum(l) = n1.NodeNum();
									}
								}
							}
							// remove redundant entries from pair list
							for (int m = 1; m <= node_pairs.Length(); m++)
							{
								if( node_pairs(m)(1) == nodes(j) || node_pairs(m)(2) == nodes(j))  
								{
									node_pairs.Erase(m);
								}
							}
							new_redundant_nodes.Add(nodes(j));
							nodes(j) = 0;             // ignore this nodelist entry - has beed identified as redundant
						}
						// if not, apply RigidBodyJoint later on
						else
						{
							if (!redundant_only)
							{
								int2 pair((int)nodes(i),(int)nodes(j));
								node_pairs.Add(pair);
							}
						}
					}
				}
			}
		}

		for (int i = 1; i <= node_pairs.Length(); i++)
		{
			Node& n1 = mbs->GetNode(node_pairs(i)(1));
			Node& n2 = mbs->GetNode(node_pairs(i)(2));

			Element& e1 = mbs->GetElement(nodes2elems(n1.NodeNum())(1));
			Element& e2 = mbs->GetElement(nodes2elems(n2.NodeNum())(1));

			//mbs->UO() << n1.NodeNum() << " "  << n2.NodeNum() << "\n";
			//mbs->UO() << nodes2elems(n1.NodeNum())(1) << " "  << nodes2elems(n2.NodeNum())(1) << "\n";
			
			// find the local position of a node within an element
			Vector3D p_loc_n1, p_loc_n2;
			for (int j = 1; j <= e1.NNodes(); j++)
			{
				if (e1.NodeNum(j) == n1.NodeNum())
				{
					p_loc_n1 = e1.GetNodeLocPos(j);
				}
				if (e2.NodeNum(j) == n2.NodeNum())
				{
					p_loc_n2 = e2.GetNodeLocPos(j);
				}
			}
			//mbs->UO() << p_loc_n1 << " "  << p_loc_n2 << "\n";
			
			// prepare EDC to be able to use existing script commands 
			ed.SetInt(nodes2elems(n1.NodeNum())(1), "element_number");	edc_elementproperties.TreeAdd("Position1", ed);
			ed.SetInt(nodes2elems(n2.NodeNum())(1), "element_number");	edc_elementproperties.TreeAdd("Position2", ed);
			ed.SetVector3D(p_loc_n1.X(), p_loc_n1.Y(), p_loc_n1.Z(), "position");	edc_elementproperties.TreeAdd("Position1", ed);
			ed.SetVector3D(p_loc_n2.X(), p_loc_n2.Y(), p_loc_n2.Z(), "position");	edc_elementproperties.TreeAdd("Position2", ed);

			TArray<ElementDataContainer*> paramEDCforElement;
			paramEDCforElement.Add(&edc_elementproperties);

			command_addrigidjoint->ExecuteCommand(mbs,edcParser,paramEDCforElement,ed,0);
		}
	}
	else		// is_ground
	{
		// again, find node pairs in order to avoid redundant constraints
		// assume, no redundant nodes exist eny longer 
		TArray<int2> node_pairs;
		for (int i = 1; i <= nodes.Length(); i++)
		{
			for (int j = i+1; j <= nodes.Length(); j++)
			{
				Node& n1 = mbs->GetNode((int)nodes(i));
				Node& n2 = mbs->GetNode((int)nodes(j));

				Vector3D dist = n1.RefConfPos() - n2.RefConfPos();

				if (dist.Norm() < 1e-10)
				{
					int2 pair((int)nodes(i),(int)nodes(j));
					node_pairs.Add(pair);
				}
			}
		}

		for (int i = 1; i <= nodes.Length(); i++)
		{
			Node& n = mbs->GetNode((int)nodes(i));
			Vector3D refpos = n.RefConfPos();

			Element& e = mbs->GetElement(nodes2elems(n.NodeNum())(1));
			int is_in_pair = 0;
			for (int j = 1; j <= node_pairs.Length(); j++)
			{
				if (node_pairs(j)(2) == nodes(i))
				{
					is_in_pair = 1; break;
				}
			}

			if (!is_in_pair)
			{
				// find the local position of a node within an element
				Vector3D p_loc;
				for (int j = 1; j <= e.NNodes(); j++)
				{
					if (e.NodeNum(j) == n.NodeNum())
					{
						p_loc = e.GetNodeLocPos(j); break;
					}
				}
				
				// prepare EDC to be able to use existing script commands 
				ed.SetInt(nodes2elems(n.NodeNum())(1), "element_number");		edc_elementproperties.TreeAdd("Position1", ed);
				ed.SetInt(0, "element_number");									edc_elementproperties.TreeAdd("Position2", ed);
				ed.SetVector3D(p_loc.X(), p_loc.Y(), p_loc.Z(), "position");	edc_elementproperties.TreeAdd("Position1", ed);
				ed.SetVector3D(refpos.X(), refpos.Y(), refpos.Z(), "position");	edc_elementproperties.TreeAdd("Position2", ed);

				TArray<ElementDataContainer*> paramEDCforElement;
				paramEDCforElement.Add(&edc_elementproperties);

				command_addrigidjoint->ExecuteCommand(mbs,edcParser,paramEDCforElement,ed,0);
			}
		}
	}

	TArray<int> permutation; permutation.SetLen(new_redundant_nodes.Length());
	QuicksortDouble(new_redundant_nodes, permutation);

	for (int i = new_redundant_nodes.Length(); i >= 1; i--)
	{
		int node_to_delete = (int)new_redundant_nodes(i);
		mbs->DeleteNode(node_to_delete);
		mesh_nodes.Erase(mesh_nodes.Find(new_redundant_nodes(i)));
		
		for (int j = 1; j <= mesh_nodes.Length(); j++)
		{
			if (mesh_nodes(j) > node_to_delete)
			{
				mesh_nodes(j)--;
			}
		}
	}

	ElementDataContainer update_nodes;
	mesh_redundant_nodes.Merge(new_redundant_nodes);
	ed.SetVector( mesh_redundant_nodes.GetDataPtr(), mesh_redundant_nodes.Length(), "list_of_redundant_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of redundant nodes in mesh");
	update_nodes.Add(ed);
	ed.SetVector( mesh_nodes.GetDataPtr(), mesh_nodes.Length(), "list_of_nodes");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Nodes associated with the Plate or empty");
	update_nodes.Add(ed);
	edc_handle_mesh->TreeReplaceEDCDataWith(&update_nodes);

	return 1;
}
#else
int CEDC_ComMesh_GlueMesh::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GlueMesh - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GlueMesh - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the component number(s)
	ElementDataContainer* edc_compnrs = parameter_EDCs(1);
	TArray<int> subIdx; subIdx.Add(0);

	ElementData* ed_compnrs = edc_compnrs->GetPtr(1);
	if(ed_compnrs->IsInt()) { subIdx(1) = ed_compnrs->GetInt(); }
	else if(ed_compnrs->IsDouble()) { subIdx(1) = ed_compnrs->GetInt(); }
	else if(ed_compnrs->IsVector()) 
	{ 
		int len = ed_compnrs->GetVectorLen();
		for(int i=1; i<=len; i++)
		{
			subIdx(i) = (int)ed_compnrs->GetVectorVal(i);	
		}
	}
	else /* textual ? */ ;
	if(subIdx(1)==0) 
	{ 
		subIdx = NaturalNumbers(mbs->GetMeshPtr(meshIdx)->NComp());
	// AD: changed to "use all non-NULL pointers" for any cascaded glue operations...
		BaseMesh* meshPtr = mbs->GetMeshPtr(meshIdx);
		subIdx.Flush();
		for(int i=1; i<=meshPtr->NComp(); i++)
		{
			if(meshPtr->GetComponent(i) != NULL)
			{
				subIdx.Add(i);
			}
		}
	}
	
	// second container contains glue prameter
	ElementDataContainer* edc_glueparams = parameter_EDCs(2);
	int usepenalty=0;
	if(edc_glueparams->TreeFind("usepenalty"))
	{
		usepenalty=edc_glueparams->TreeGetInt("usepenalty");
	}
	mystr name;
	if(edc_glueparams->TreeFind("name"))
	{
		name = edc_glueparams->TreeGetString("name");
	}

	// generate the component
	int compnr = mbs->GetMeshPtr(meshIdx)->Glue(subIdx,usepenalty);
	
	if(name.Length()>0)
	{
		mbs->GetMeshPtr(meshIdx)->GetMeshComponent(compnr)->SetName(name);
	}

	return_value.SetInt(compnr,"");
	return 1;// rv;
}
#endif

mystr CEDC_ComMesh_GlueMesh::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the component numbers \n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter EDC containing additional parameters for the glue operation\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item usepenalty - set to one for penalty constraints, 0(default) for lagrangian \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int CEDC_GetLocalPosOfGlobalPos::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

	// first container contains the element number
	ElementData* ed_elnr = parameter_EDCs(1)->GetPtr(1);
	int elnr = 0;
	if(ed_elnr->IsInt())
	{
		elnr = ed_elnr->GetInt(); // integer variabe
	}
	else if(ed_elnr->IsDouble())
	{
		elnr = (int) ed_elnr->GetDouble();
	}
	else if(ed_elnr->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr str = ed_elnr->GetText();
		elnr = str.MakeInt();
	}

	// second container contains the global position as vector3D
	ElementData* ed_pglob = parameter_EDCs(2)->GetPtr(1);
	Vector3D pglob;
	if(ed_pglob->IsVector()&&ed_pglob->GetVectorLen()==3)
	{
		ed_pglob->GetVector(pglob.X(),pglob.Y(),pglob.Z());
	}

	// this does lead to a crash, at this point the initialization of the elements is not 
	//Vector3D ploc = mbs->GetLocalPositionOfGlobalPosition(elnr,gpos);

	Element* el = mbs->GetElementPtr(elnr);
	Vector3D ploc;
	bool found = el->GetLocPosOfGlobalPosPreAssemble(ploc,pglob,4);

	return_value.SetVector3D(ploc.X(),ploc.Y(),ploc.Z(),"local_position");

	return found;// rv;
}


mystr CEDC_GetLocalPosOfGlobalPos::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item (integer) number of the element \n");
	descr +=            mystr("  \\item (vector3D) global position \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_ComMesh_GetElementsInBox::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetElementsInBox - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetElementsInBox - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the box parameters
	ElementDataContainer* edc_box_parameters = parameter_EDCs(1);
	Vector3D p1, p2;
	edc_box_parameters->TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());
	edc_box_parameters->TreeGetVector3D("P2",p2.X(), p2.Y(), p2.Z());
	
	TArray<int> elementslist = mbs->GetMeshPtr(meshIdx)->GetElementsInBox(Box3D(p1,p2));
	Vector elementnums(elementslist.Length()); for(int i=1; i<=elementslist.Length(); i++) { elementnums(i) = elementslist(i); }

	return_value.SetVector(elementnums.GetVecPtr(), elementnums.Length(), "elements_in_box"); return_value.SetValuesInt(); return_value.SetVariableLength(); /* return_value.SetLocked(1); */ return_value.SetToolTipText("retured by a call of 'GetElementsInBox'");
	return 1;// rv;
}

mystr CEDC_ComMesh_GetElementsInBox::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the box (defined by two corners) \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item P1 - position of first corner \n");
	descr +=			mystr("	 \\item P2 - position of second corner\n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};


#ifdef mode_old
int CEDC_ComMesh_GetElementAtPos::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// container 1 contains the corners of the box
	Vector3D p1,p2;
	ElementDataContainer* edc_pointparameters = parameter_EDCs(1);
	edc_pointparameters->TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());

// container 2 contains the name of the handle ( can not pass the EDC* itself because that is going to be deleted... )
// consistency already checked in ParseCommandParameters...
	mystr handlename = parameter_EDCs(2)->TreeGetString("handle");
	ElementDataContainer* edc_local = edcParser->GetMBSParser()->GetLocalEDC(); 
	int nr_handle = edc_local->Find(handlename);
	ElementDataContainer* edc_handle_mesh = edc_local->Get(nr_handle).GetEDC();

// get current lists for mesh object
	double* vdata;
	int vlen;
	edc_handle_mesh->TreeGetVector("list_of_nodes",&vdata,vlen);            
	TArray<double> mesh_nnrs; mesh_nnrs.CopyFrom(vdata,vlen); mesh_nnrs.SetLen(vlen);			// node numbers already associated with the Mesh
	
	edc_handle_mesh->TreeGetVector("list_of_elements",&vdata,vlen);            
	TArray<double> mesh_enrs; mesh_enrs.CopyFrom(vdata,vlen); mesh_enrs.SetLen(vlen);			// element numbers already associated with the Mesh

	TArray<double> enrs;
	TArray<Vector3D> locpos;
// loop over all elements
	for (int i=1; i<= mesh_enrs.Length(); i++)
	{
		Element& curel = mbs->GetElement((int)mesh_enrs(i));
		Node& typeidnode = curel.GetNode(1); // for typetesting via node, don want to include all elements .h
		if(curel.Dim()==1)
		{
		// find locpos on beam element
		}
		if( typeid(typeidnode) == typeid(ANCFNodeS1S2_3D) )   // associated plate element ANCFThinPlate3D
		{
		// find locpos on ANCFThinPlate3D plate element
    // origin in the center of the plate, local coordinates from -1 to +1
			// LN4 --- LN3
			//  |       |
			// -1   O  +1 -->x      
			//  |       |
			// LN1 --- LN2

			Vector3D c1,c2,c3,c4,cc;
			c1 = curel.GetNode(1).RefConfPos();
			c2 = curel.GetNode(2).RefConfPos();
			c3 = curel.GetNode(3).RefConfPos();
			c4 = curel.GetNode(4).RefConfPos();
			cc = (c1+c2+c3+c4)*0.25;

			Vector3D e1,e2,e3;
			e1 = (c2-c1)*0.5;
			e2 = (c4-c1)*0.5;
			e3 = e1.Cross(e2);

			Matrix3D A = Matrix( e1(1), e2(1), e3(1),
				                   e1(2), e2(2), e3(2),
													 e1(3), e2(3), e3(3) );
			Vector3D b = cc-p1;
			
			Matrix3D Ainv;
			A.GetInverse(Ainv);
			Vector3D x = Ainv*b;

			if(abs(x.X()) <= 1. && abs(x.Y()) <= 1. && abs(x.Z()) <= 1e-1)
			{
				enrs.Add(mesh_enrs(i));
				locpos.Add(x);
			}
		}
		if(curel.Dim()==3) 
		{
		// find locpos on volume element
		}
	}

// return value: all elements and local position in first element
	ElementData ed;
	ed.SetVector( enrs.GetDataPtr(), enrs.Length(), "list_of_elements");	ed.SetValuesInt();	ed.SetVariableLength();	ed.SetToolTipText("List of Elements associated with the Beam or empty");
	edc_rv.Add(ed);
	ed.SetVector3D( locpos(1).X(), locpos(1).Y(), locpos(1).Z(),"locpos"); ed.SetToolTipText("local position in first element");
	edc_rv.Add(ed);
	return_value.SetEDC(&edc_rv,"result");
	return rv;
}
#else
int CEDC_ComMesh_GetElementAtPos::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementDataContainer edc_rv;

// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in GetElementAtPos - entry for Handle could not be found in parameters"));
		return 0;	
	}
	int meshIdx = ResolveParameterEntryToMeshNumber(mbs, edcParser, ed_handle);	
	if(meshIdx==0)	// check existance
	{
		edcParser->EDCError(mystr("Error in GetElementAtPos - Handle could not be resolved into a Mesh"));
		return 0;
	}

	// first container contains the global
	ElementDataContainer* edc_pos_parameters = parameter_EDCs(1);
	Vector3D p1;
	edc_pos_parameters->TreeGetVector3D("P1",p1.X(), p1.Y(), p1.Z());

	int elnr = mbs->GetMeshPtr(meshIdx)->GetElementAtPos(p1);
	Vector3D ploc = mbs->GetMeshPtr(meshIdx)->GetLocalPosOnElem(elnr,p1);

	ElementData ed;
	ed.SetInt(elnr, "element_number"); ed.SetToolTipText("element number identified by GetElemArPos"); edc_rv.Add(ed);
	ed.SetVector3D(ploc.X(), ploc.Y(), ploc.Z(), "locpos"); ed.SetToolTipText("local coordinates  of the global position - MIND THE ELEMENTTYPE"); edc_rv.Add(ed);
	return_value.SetEDC(&edc_rv,"result");
	return 1;// rv;
}
#endif

mystr CEDC_ComMesh_GetElementAtPos::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing the position \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item P1 - position to look at \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//$ AD 2017-03-21
int CEDC_ComPlot_GenerateNewPlot::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;
	edcParser->EDCDbgMsg(mystr("identified: command GenerateNewPlot."));

// step 1: generate the PlotTool
// TODO: ?? register as MBS object ??

	// first container contains filename - process the string
	ElementDataContainer* edc_filename = parameter_EDCs(1);
	ElementData* ed_filename = edc_filename->TreeFind("layout_file");
	mystr filename;
	if(ed_filename!=NULL && ed_filename->IsText()) { filename = ed_filename->GetText(); }
	else { edcParser->EDCError(mystr("Error in GenerateNewPlot - entry (plotparams.layout_file) must be a string")); }
	edcParser->MakeFilenameAbsolute(filename);

	ed_filename->SetText(filename); // replace entry with 
	


	int plot_num = mbs->UO().CallWCDriverFunction(11, /* 11 = PlotTool functions */
		                                      1, /* 1 = open new instance */
											  0, /* not used */
											  parameter_EDCs(1)); 

	if(plot_num<=0) 
	{
		edcParser->EDCError(mystr("ERROR: no new PlotTool instance could be generated. "));
		return_value.SetInt(rv_error, ""); //return value has no name!
	}
	else
	{
		edcParser->EDCDbgMsg(mystr("PlotTool handle created."));
		return_value.SetInt(plot_num, ""); //return value has no name!		
	}

// step 2:
	ElementDataContainer edc_plotproperties;
	ElementData ed;
// why cant bool be locked ?
	ed.SetBool(1,"isHandle"); ed.SetLocked(1); ed.SetToolTipText("This Container represents a Handle. Commands may be executed for this object"); edc_plotproperties.Add(ed);
// available script commands ???
	ElementDataContainer edc_registered_functions;	
	mbs->UO().CallWCDriverFunction(11, 2, plot_num, &edc_registered_functions);
	ed.SetEDC(&edc_registered_functions,"Functions"); // this container name is checked in ::IdentifyCommand
	edc_plotproperties.Add(ed);

	return_value.SetEDC(&edc_plotproperties,"PlotProperties");
	edcParser->EDCDbgMsg(mystr("Plot handle registered."));

	return 1;
};

mystr CEDC_ComPlot_GenerateNewPlot::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC to overwrite the default properties\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("the return vaule of the command MUST be assigned to a new variable(handle) \n\n");
	descr +=			mystr("overwritable enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item layout\\_file: may be \\bf{StructuralMesh} (default) or \\bf{SolidMesh} \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

//$ AD 2017-03-27
int CEDC_ComPlot_ExportPlot::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in ExportPlot - entry for Handle could not be found in parameters"));
		return 0;	
	}	

	int generated = mbs->UO().CallWCDriverFunction(11,4,0,parameter_EDCs(1)); // 

	return_value.SetInt(generated,"files_generated");

	return generated;// rv_error;
};

mystr CEDC_ComPlot_ExportPlot::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter EDC containing entries for the possible export formats\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("available entries in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item width: width in pixels, use actual window size if not specified");
	descr +=			mystr("	 \\item height: height in pixels, use actual window size if not specified");
	descr +=			mystr("	 \\item export\\_file\\_jpg: generates a .jpg file \"name.jpg\"");
	descr +=			mystr("	 \\item export\\_file\\_png: generates a .png file \"name.jpg\"");
	descr +=			mystr("	 \\item export\\_file\\_bmp: generates a .bmp file \"name.jpg\"");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

//$ AD 2017-03-27
int CEDC_ComPlot_ClosePlot::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv_error = 0;

	// last container contains the name of the handle
	mystr handlename = parameter_EDCs.Last()->TreeGetString("handle");
	ElementData* ed_handle = parameter_EDCs.Last()->GetPtr(1);
	if(ed_handle==NULL)
	{
		edcParser->EDCError(mystr("Error in ClosePlot - entry for Handle could not be found in parameters"));
		return 0;	
	}

	mbs->UO().CallWCDriverFunction(11,9,0,parameter_EDCs(1)); // 9 for "close plot"

	return 1;// rv_error;
};

mystr CEDC_ComPlot_ClosePlot::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};



mystr CEDC_Check_Entry_Exists::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string with the name and tree of the entry to check \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_Check_Entry_Exists::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementData* ED_treename = parameter_EDCs(1)->GetPtr(1);
  mystr str_treename;
	if(ED_treename->IsText())
	{
		str_treename = ED_treename->GetText();
	}

	ElementDataContainer* localEDC = edcParser->GetMBSParser()->GetLocalEDC2();  // existing local EDC for the CHECK_ENTRY_EXISTS command

	ElementData* ed = localEDC->TreeFind(str_treename);
	if(ed==NULL)
	{
	// Entry not found
		return_value.SetInt(0,"rv");
	}
	else if(ed!=NULL && !ed->IsEDC())
	{
	// Entry found, not an EDC
		return_value.SetInt(1,"rv");
	}
	else if(ed!=NULL && ed->IsEDC())
	{
	// Entry found, also an EDC
		return_value.SetInt(2,"rv");
	}
	return 1;
}

mystr CEDC_GetByName::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string with the name and tree of the entry to get \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_GetByName::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementData* ED_treename = parameter_EDCs(1)->GetPtr(1);
  mystr str_treename;
	if(ED_treename->IsText())
	{
		str_treename = ED_treename->GetText();
	}

	ElementDataContainer* localEDC = edcParser->GetMBSParser()->GetLocalEDC2();  // existing local EDC for the CHECK_ENTRY_EXISTS command

	ElementData* ed = localEDC->TreeFind(str_treename);
	if(ed==NULL)
	{
	// Entry not found
		return_value.SetInt(0,"rv");
	}
	else if(ed!=NULL && !ed->IsEDC())
	{
	// Entry found, not an EDC
		return_value.CopyFrom(*ed);
	}
	else if(ed!=NULL && ed->IsEDC())
	{
	// Entry found, also an EDC
		return_value.CopyFrom(*ed);
	}
	return_value.SetLocked(0); // required override if the variable is to be reused
	return 1;
}

mystr CEDC_SetByName::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string with the name and tree of the entry to set \n");
	descr +=      mystr("  \\item $2^{nd}$ parameter: the variable that should be assigned \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_SetByName::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	ElementData* ED_dataname = parameter_EDCs(1)->GetPtr(1);
  mystr dataname;
	if(ED_dataname->IsText())
	{
		dataname = ED_dataname->GetText();
	}
	ElementData* ED_Assign = parameter_EDCs(2)->GetPtr(1);

	ElementDataContainer* rootEDC = mbs->GetModelDataContainer();
	
	mystr treename,elemname;
	rootEDC->SplitIntoTreeAndElementName(dataname,treename,elemname); 
	ED_Assign->SetDataName(elemname.c_str());

	if(rootEDC->TreeFind(dataname))
	{
		rootEDC->TreeDelete(dataname);
	}
	rootEDC->TreeAdd(treename.c_str(),*ED_Assign);

	return 1;
}

mystr CEDC_Compare::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string A \n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: string B \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_Compare::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementData* ED_strA = parameter_EDCs(1)->GetPtr(1);
  mystr strA;
	if(ED_strA->IsText())
	{
		strA = ED_strA->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing COMPARE operation: first parameter is not a string variable");
		return rv;
	}
	ElementData* ED_strB = parameter_EDCs(2)->GetPtr(1);
  mystr strB;
	if(ED_strB->IsText())
	{
		strB = ED_strB->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing COMPARE operation: second parameter is not a string variable");
		return rv;
	}

	int result = strcmp(strA.c_str(), strB.c_str());
	
	return_value.SetInt(result,"rv");

	return 1;
}

mystr CEDC_StrCat::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string A \n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: string B \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("You can also use integer or double values instead of the strings.\n");
	descr +=			mystr("Inline definition of strings, e.g. StrCat(''this is a '',''test''), do not work properly. The spaces are not taken into account correctly!\n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_StrCat::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	ElementData* ED_strA = parameter_EDCs(1)->GetPtr(1);
  mystr strA;
	if(ED_strA->IsText())
	{
		strA = ED_strA->GetText();
	}
	else
	{
		if(ED_strA->IsInt() || ED_strA->IsDouble())
		{
			double tmp;
			if(ED_strA->IsInt())		{ tmp = (double) ED_strA->GetInt();}
			if(ED_strA->IsDouble())	{ tmp = ED_strA->GetDouble();}
			strA = mystr(tmp);
		}
		else
		{
			edcParser->EDCError("Error while executing STRCAT operation: first parameter has to be string, int or double.");
			return rv;
		}
	}
	ElementData* ED_strB = parameter_EDCs(2)->GetPtr(1);
  mystr strB;
	if(ED_strB->IsText())
	{
		strB = ED_strB->GetText();
	}
	else
	{
		if(ED_strB->IsInt() || ED_strB->IsDouble())
		{
			double tmp;
			if(ED_strB->IsInt())		{ tmp = (double) ED_strB->GetInt();}
			if(ED_strB->IsDouble())	{ tmp = ED_strB->GetDouble();}
			strB = mystr(tmp);
		}
		else
		{
			edcParser->EDCError("Error while executing STRCAT operation: second parameter has to be string, int or double.");
			return rv;
		}
	}

	mystr result = strA + strB;

	return_value.SetText(result,"rv");

	return 1;
}


mystr CEDC_Zeros::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: length of the vector or first dimension of the matrix \n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: second dimenstion of the matrix \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_Zeros::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ED_par1 = parameter_EDCs(1)->GetPtr(1);
  int par1;
	if(ED_par1->IsInt()) // entry is a variable of integer type
	{
		par1 = ED_par1->GetInt();
	}
	else if(ED_par1->IsDouble()) // entry is a variable of double type - result of multiplcation for instance...
	{
		par1 = (int) ED_par1->GetDouble();
	}
	else if(ED_par1->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr p1 = ED_par1->GetText();
		par1 = p1.MakeInt();
	}
	else
	{
		edcParser->EDCError("Error while executing ZEROS operation: first parameter is not an integer");
		return rv;
	}
	if(par1 <= 0)
	{
		edcParser->EDCError("Error while executing ZEROS operation: first parameter is not an integer");
		return rv;		
	}

	ElementData* ED_par2 = parameter_EDCs(2)->GetPtr(1);
  int par2;

	if(ED_par2->IsInt()) // entry is a variable of integer type
	{
		par2 = ED_par2->GetInt();
	}
	else if(ED_par2->IsDouble()) // entry is a variable of double type - result of multiplcation for instance...
	{
		par2 = (int) ED_par2->GetDouble();
	}
	else if(ED_par2->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr p2 = ED_par2->GetText();
		par2 = p2.MakeInt();
	}
	else
	{
		edcParser->EDCError("Error while executing ZEROS operation: second parameter is not an integer");
		return rv;
	}

	if(par2 < 0)
	{
		edcParser->EDCError("Error while executing ZEROS operation: second parameter is not an integer");
		return rv;		
	}

	TArray<double> zeros;
	zeros.SetLen(par1*par2);
	zeros.SetAll(0.);
	if(par2 == 1) // vector
	{
		return_value.SetVector(zeros.GetDataPtr(), par1, "rv");
	}
	else // matrix
	{
		return_value.SetMatrix(zeros.GetDataPtr(),par1,par2,"rv");
	}
	return 1;
}

//$ AD 2017-03-07: command CEDC_IntArrayOp
mystr CEDC_IntArrayOp::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: type of operation as string \n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: first array operand \n");
	descr +=			mystr("	 \\item $3^{nd}$ parameter: second array operand \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("allowed operations are: \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item operation 'Append' or 'Add' - adds the second array to the tail of the first array \n");
	descr +=			mystr("	 \\item operation 'Union', 'Inter' or 'Diff' - computes union, intersection and difference set of the two arrays \n");
	descr +=			mystr("	 \\item operation 'Asc' and 'Desc' - sorts the array, only the first input is processed \n");
	descr +=			mystr("  \\item operation 'Find' - returns a list of index numbers where numbers of the second array occurr in the first \n");
	descr +=			mystr("	 \\item operation 'Unique' removes multiple entries in the first array \n");
	descr +=            mystr("  \\item operation 'AddConst', 'MultConst' \n");
	descr +=            mystr("  \\item operation 'Sequence' \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_IntArrayOp::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	enum TArrayOp { opNoOp, opAppend, opUnion, opIntersection, opDifference, opSortAsc, opSortDesc, opFind, opUnique, opAddConst, opMultConst, opAddArrays, opSequence };
	
	// first entry contains a string identifier
	ElementData* ED_StrOpID = parameter_EDCs(1)->GetPtr(1);
	mystr strOpID;
	if(ED_StrOpID->IsText())
	{
		strOpID = ED_StrOpID->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing INTARRAYOP operation: first parameter has to be string.");
		return rv;
	}
	int operation = 0;
	if(strOpID.Compare("Append")==1) { operation = opAppend; }
	else if(strOpID.Compare("Add")==1) { operation = opAppend; }
	else if(strOpID.Compare("Union")==1) { operation = opUnion; }
	else if(strOpID.Compare("Inter")==1) { operation = opIntersection; }
	else if(strOpID.Compare("Diff")==1) { operation = opDifference; }
	else if(strOpID.Compare("Asc")==1) { operation = opSortAsc; }
	else if(strOpID.Compare("Desc")==1) { operation = opSortDesc; }
	else if(strOpID.Compare("Find")==1) { operation = opFind; }
	else if(strOpID.Compare("Unique")==1) { operation = opUnique; }
	else if(strOpID.Compare("AddConst")==1) { operation = opAddConst; }
	else if(strOpID.Compare("MultConst")==1) { operation = opMultConst; }
	else if(strOpID.Compare("AddArrays")==1) { operation = opAddArrays; }
	else if(strOpID.Compare("Sequence")==1) { operation = opSequence; }
	else
	{
		edcParser->EDCError("Error while executing INTARRAYOP operation: no valid operation identifier found.");
		return rv;	
	}

	// second entry contains a vector or single integer
	ElementData* ED_Arg1 = parameter_EDCs(2)->GetPtr(1);
	TArray<int> arg1;
	if(ED_Arg1->IsVector())
	{
		int len = ED_Arg1->GetVectorLen();
		double* v = ED_Arg1->GetVector();
		for(int i=1; i<=len; i++)
		{
			int entry = (int) v[i-1]+.5;
			arg1.Add(entry);
		}
	}
	else if(ED_Arg1->IsInt())
	{
		arg1.Add(ED_Arg1->GetInt());
	}
	else if(ED_Arg1->IsText())
	{
		mystr str = ED_Arg1->GetText();
		if(str.IsValidNumber())
		{
			arg1.Add(str.MakeInt());
		}
		else
		{
			edcParser->EDCError("Error while executing INTARRAYOP operation: 2nd parameter is neither vector nor integer.");
			return rv;	
		}
	}
	else if (ED_Arg1->IsDouble())
	{
		double entry = ED_Arg1->GetDouble();
		int int_entry = static_cast<int>(entry + 0.5);
		static const double eps = 1e-9;
		if (abs(entry-int_entry) > eps)
		{
			edcParser->EDCError("Error while executing INTARRAYOP operation: 2nd parameter is neither vector nor integer.");
			return rv;	
		}
		arg1.Add(int_entry);
	}
	else
	{
		edcParser->EDCError("Error while executing INTARRAYOP operation: 2nd parameter is neither vector nor integer.");
		return rv;	
	}

	// third entry contains a vector or single integer
	ElementData* ED_Arg2 = parameter_EDCs(3)->GetPtr(1);
	TArray<int> arg2;
	if(ED_Arg2->IsVector())
	{
		int len = ED_Arg2->GetVectorLen();
		double* v = ED_Arg2->GetVector();
		for(int i=1; i<=len; i++)
		{
			int entry = (int) v[i-1]+.5;
			arg2.Add(entry);
		}
	}
	else if(ED_Arg2->IsInt())
	{
		arg2.Add(ED_Arg2->GetInt());
	}
	else if(ED_Arg2->IsText())
	{
		mystr str = ED_Arg2->GetText();
		if(str.IsValidNumber())
		{
			arg2.Add(str.MakeInt());
		}
		else
		{
			edcParser->EDCError("Error while executing INTARRAYOP operation: 3nd parameter is neither vector nor integer.");
			return rv;	
		}
	}
	else if (ED_Arg2->IsDouble())
	{
		double entry = ED_Arg2->GetDouble();

		int int_entry = static_cast<int>(entry + 0.5);
		static const double EPS = 1e-9;
		if (abs(entry-int_entry) > EPS)
		{
			edcParser->EDCError("Error while executing INTARRAYOP operation: 3rd parameter is neither vector nor integer.");
			return rv;	
		}
		arg2.Add(int_entry);
	}
	else
	{
		edcParser->EDCError("Error while executing INTARRAYOP operation: 3nd parameter is neither vector nor integer.");
		return rv;	
	}

// compute result
	TArray<int> result;
	if(operation==opAppend)
	{
		arg1.Append(arg2);
		result = arg1;
	}
	else if(operation==opUnion)
	{
		result = arg1.Union(arg2);
	}
	else if(operation==opIntersection)
	{
		result = arg1.Intersection(arg2);
	}
	else if(operation==opDifference)
	{
		result = arg1.Difference(arg2);
	}
	else if(operation==opSortAsc)
	{
		Quicksort(arg1);
		result = arg1;
	}
	else if(operation==opSortDesc)
	{
		Quicksort(arg1);
		for(int i=arg1.Length(); i>0; i--)
		{
			result.Add(arg1(i));
		}
	}
	else if(operation==opFind)
	{
		for(int i=1; i<=arg1.Length(); i++)
		{
			if(arg2.Find(arg1(i))>0)
			{
				result.Add(i); // the i-ht entry of the list to search contained one of the elements to find
			}
		}
	}
	else if(operation==opUnique)
	{
		RemoveRedundantEntries(arg1);
		result = arg1;
	}
	else if(operation==opAddConst)
	{
		for(int i=1; i<=arg1.Length(); i++)
		{
			result.Add(arg1(i)+arg2(1));
		}
	}
	else if(operation==opMultConst)
	{
		for(int i=1; i<=arg1.Length(); i++)
		{
			result.Add(arg1(i)*arg2(1));
		}
	}
	else if(operation==opAddArrays)
	{
		if(arg1.Length()==arg2.Length())
		{
			for(int i=1; i<=arg1.Length(); i++)
			{
				result.Add(arg1(i)+arg2(i));
			}
		}
		else
		{
			edcParser->EDCError("Error while executing INTARRAYOP operation opAddArrays, the arrays must be of same length when items are added pairwise.");
		}
	}
	else if (operation==opSequence)
	{
		int first = arg1(1);
		int last = arg2(1);
		if(arg1(1) <  arg2(1))
		{
			for(int i=arg1(1); i<=arg2(1); i++)
				result.Add(i);
		}
		else if(arg1(1) > arg2(1))
		{
			for(int i=arg1(1); i>=arg2(1); i--)
				result.Add(i);
		}
		else
			result.Add(arg1(1));
	}


// return the result
	Vector res(result.Length()); for(int i=1; i<=result.Length(); i++) { res(i) = result(i); }
	return_value.SetVector(res.GetVecPtr(), res.Length(), "result"); return_value.SetValuesInt(); return_value.SetVariableLength(); return_value.SetToolTipText("returned by a call of 'IntArrayOp'\n");


	return 1;
}

//$ AD 2021-04-07: command StringOp - collection of operation with strings
mystr CEDC_StringOp::GetCommandTexParameterDescription() const
{
	mystr descr = mystr("The parameters of this command are \n");
	descr += mystr("\\begin{enumerate} \n");
	descr += mystr("	 \\item $1^{st}$ parameter: type of operation as string \n");
	descr += mystr("	 \\item $2^{nd}$ parameter: first string operand \n");
	descr += mystr("	 \\item $3^{nd}$ parameter: second string operand or other (e.g. list of numbers) \n");
	descr += mystr("\\end{enumerate} \n");
	descr += mystr("allowed operations are: \n");
	descr += mystr("\\begin{enumerate} \n");
	descr += mystr("  \\item operation 'Compare' or CompareCase' - returns 0 if first and second string are equal");
	descr += mystr("  \\item operation 'CompareNoCase' - returns 0 if first and second string are equal when both are converted to lower case");
	descr += mystr("  \\item operation 'ToLower' - returns first string with only lower case letters");
	descr += mystr("  \\item operation 'ToUpper' - returns first string with only upper case letters");
	descr += mystr("\\end{enumerate} \n");
	descr += mystr("use variables as arguments, inline definitions are not stable !\n");
	descr += mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_StringOp::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	return_value.SetText("processing...","rv");    // 'meaningful' return value in case of parameter error - on premature return
	enum TStringOp { opNoOp,                       // input --> result
		             opCompare, opCompareNoCase,   // str,str --> int
					 opToLower, opToUpper,         // str     --> str             
					 opFind,                       // str,str --> int
					 
					 opStrCat,                     // str,s/n --> str

		
					 opSplit,                      // str,s/n --> EDC
		             opPrintf };                   // str,vec --> str

// options for 3 arguments
//					 opStrCat,					   // str,s/n,s/n --> str
//					 opReplace,                    // str,str,str --> str
//		             opInsert,                     // str,str,int --> str   
//		             opErase };                    // str,int,int --> str
	               

	// first entry contains a string identifier
	ElementData* ED_StrOpID = parameter_EDCs(1)->GetPtr(1);
	mystr strOpID;
	if (ED_StrOpID->IsText())
	{
		strOpID = ED_StrOpID->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing STRINGOP operation: first parameter has to be string.");
		return rv;
	}
	int operation = 0;
	if(strOpID.Compare("Compare") == 1) { operation = opCompare; }
	else if(strOpID.Compare("CompareCase") == 1) { operation = opCompare; }
	else if(strOpID.Compare("CompareNoCase") == 1) { operation = opCompareNoCase; }
	else if(strOpID.Compare("ToLower") == 1) { operation = opToLower; }
	else if(strOpID.Compare("ToUpper") == 1) { operation = opToUpper; }
	else if(strOpID.Compare("Find") == 1) { operation = opFind; }
	else if (strOpID.Compare("StrCat") == 1) { operation = opStrCat; }
	else if (strOpID.Compare("Split") == 1) { operation = opSplit; }
	else if (strOpID.Compare("Printf") == 1) { operation = opPrintf; }
	else
	{
		edcParser->EDCError("Error while executing STRINGOP operation: no valid operation identifier found.");
		return rv;
	}

	// second entry always contains a string
	ElementData* ED_Arg1 = parameter_EDCs(2)->GetPtr(1);
	mystr strArg1;
	if(ED_Arg1->IsText())
	{
		strArg1 = ED_Arg1->GetText();
	}
	else
	{
		edcParser->EDCError("Error while executing STRINGOP operation: 2nd parameter has to be string.");
		return rv;
	}

	// third entry contains a String or a Number/Vector
	ElementData* ED_Arg2 = parameter_EDCs(3)->GetPtr(1);
	mystr strArg2;
	TArray<double> vecArg2;
	if(ED_Arg2->IsText())
	{
		strArg2 = ED_Arg2->GetText();
		// try to convert a single number argument - this removes any whitespace characters thus copy
		//                                           only required when the argument may be a string 
		mystr strArg2Copy = strArg2;
		if (strArg2Copy.IsValidNumber())
		{
			vecArg2.Add(strArg2Copy.MakeDouble());
		}
	}
	else if(ED_Arg2->IsVector())
	{
		int len = ED_Arg2->GetVectorLen();
		double* v = ED_Arg2->GetVector();
		for (int i = 1; i <= len; i++)
		{
			vecArg2.Add(v[i-1]);
		}
	}
	else if(ED_Arg2->IsInt() || ED_Arg2->IsDouble())
	{
		vecArg2.Add(ED_Arg2->GetDouble());
	}
	else
	{
		edcParser->EDCError("Error while executing STRINGOP operation: 3nd parameter has to be string a vector or a single number");
		return rv;
	}

	// operations with specific type restrictions - assume that arg1 is always a string with len > 0
	//   this is already covered when parsing arg 1
	if(operation == opCompare || operation == opCompareNoCase ||
		operation == opFind )
	{
		if(strArg2.Length() < 1)
		{
			edcParser->EDCError("Error while executing STRINGOP operation: specified operation requires two string arguments");
			return rv;
		}
	}
	if(operation == opStrCat)
	{ 
		if(vecArg2.Length() > 1)
		{
			edcParser->EDCError("Error while executing STRINGOP operation: specified operation requires string, integer or double as second argument");
			return rv;
		}
	}
	if (operation == opSplit)
	{
		if (strArg2.Length() != 1 && vecArg2.Length() != 1)
		{
			edcParser->EDCError("Error while executing STRINGOP operation: specified operation requires a single character or single integer as second argument");
			return rv;
		}
	}
	

	// compute result
	if(operation==opCompare)
	{
		int strcmpVal = strcmp(strArg1.c_str(), strArg2.c_str());
		return_value.SetInt(strcmpVal, "rv");
	}
	if(operation == opCompareNoCase)
	{
		int strcmpVal = _stricmp(strArg1.c_str(), strArg2.c_str());
		return_value.SetInt(strcmpVal, "rv");
	}

	if(operation == opToLower)
	{
		mystr tmp(strArg1);
		int changedChars = tmp.ToLower();
		return_value.SetText(tmp.c_str(), "rv");
	}
	if(operation == opToUpper)
	{
		mystr tmp(strArg1);
		int changedChars = tmp.ToUpper();
		return_value.SetText(tmp.c_str(), "rv");
	}
	if (operation == opFind)
	{
		int pos = strArg1.Find(strArg2);
		return_value.SetInt(pos, "rv");
	}
	if (operation == opStrCat)
	{
		if (vecArg2.Length() == 1 && strArg2.Length() == 0) // when argument is an existing numerical variable
		{
			strArg2 = mystr(vecArg2(1));
		}
		mystr tmp = strArg1 + strArg2;
		return_value.SetText(tmp.c_str(), "rv");
	}
	if (operation == opSplit)
	{
		ElementDataContainer edc_rv;
		int pos = 0, count = 0;
		mystr subStr;
		if (vecArg2.Length() < 1) // split character, arbitary number of resulting substrings
		{
			while (pos > -1 && pos < strArg1.Length())
			{
				int newPos = strArg1.Find(pos, strArg2);          // position of the split char      
				if (newPos >= 0)                              // includes newPos == 0, newpos == length()-1
				{
					subStr = strArg1.SubString(pos, newPos - 1);  // -1 means that the split char is not included in substring
					pos = newPos + 1;                                // continue with letter after split
				}
				else
				{
					subStr = strArg1.SubString(pos, strArg1.Length() - 1);  // no more split found
					pos = strArg1.Length();
				}
				if (subStr.Length() > 0)
				{
					mystr tag = mystr("Split") + mystr(++count);
					ElementData ed; ed.SetText(subStr, tag); edc_rv.Add(ed);
				}
			}
		}
		else  // split at given position
		{
			int splitPos = int(vecArg2(1))-1;
			ElementData ed;

			subStr = strArg1.SubString(0, splitPos - 1);
			ed.SetText(subStr, "Split1"); edc_rv.Add(ed);
			subStr = strArg1.SubString(splitPos + 1, strArg1.Length()-1);
			ed.SetText(subStr, "Split2"); edc_rv.Add(ed);
		}	
		return_value.SetEDC(&edc_rv, "result");
	}
	if (operation == opPrintf)
	{
		char buffer[1024];
		mystr final, subStr;
		int pos = 0, count = 0;
		
	// reference: https://www.cplusplus.com/reference/cstdio/printf/
	// begin with '%',  ends with specifier from list, first letter that follows %
	// 
		// head (before first %)
		pos = strArg1.Find(0, '%');
		if (pos == -1)   // no placeholders in format string
			{ final = strArg1; } 
		else 
			{ final = strArg1.SubString(0, pos - 1); }

		// body - begins with a  %,   look ahead for the next '%'
		while (pos > -1 && pos < strArg1.Length())
		{
			int newPos = strArg1.Find(pos+1, '%');

			if (newPos - pos == 1) // case '%%'
			{
				final += mystr('%');
				pos = newPos + 1;
				// also find next '%' (as in case head)
				newPos = strArg1.Find(pos, '%');
				int lastChar = newPos > 0 ? newPos - 1 : strArg1.Length() - 1; // look ahead until next '%' or end of string
				final += strArg1.SubString(pos, lastChar);
				pos = lastChar + 1;
			}
			else
			{
				int lastChar = newPos > 0 ? newPos-1 : strArg1.Length()-1; // look ahead until next '%' or end of string
				mystr subStr = strArg1.SubString(pos, lastChar);
				pos = lastChar +1 ;

				// TO DO: check length of argument vector
				if (++count > vecArg2.Length())
				{
					edcParser->EDCError("Error while executing STRINGOP operation: argument vector too short for format string: '" + strArg1 + "'");
					return_value.SetText(final, "rv");
					return rv;
				}

				char spec = 0;
				for (int i = 0; i < subStr.Length(); i++)
				{
					if (IsLetter(subStr[i]))
					{
						spec = subStr[i];
						if (IsUpperCaseLetter(spec)) { spec += ('a' - 'A'); } // convert to lower case
						break;
					}
				}
				// allowed, requires an integer argument
				if (spec == 'd' || spec == 'i' || spec == 'u' ||       // decimal
					spec == 'o' || spec == 'x')                        // octal, hexadecimal
				{
					double dblArg = vecArg2(count);
					int intArg = (int)(dblArg + .5);
					int len = sprintf(buffer, subStr, intArg);
				}
				// allowed, requires a floating point argument
				else if (spec == 'f' || spec == 'e' || spec == 'g')   // float, scentific, shortest of both
				{
					double dblArg = vecArg2(count);
					int len = sprintf(buffer, subStr, dblArg);
				}
				// allowed, requires a string argument // not implemented yet
	//				else if (spec == 'c' || spec == 's')                    // single character, string
	//				{
	//
	//				}
				// not a valid specifier
				else
				{
					edcParser->EDCError("Error while executing STRINGOP operation: could not process format string: '" + subStr + "'");
					return_value.SetText(final, "rv");
					return rv;
				}
				final += mystr(buffer);
			} 
		} // end while
		return_value.SetText(final, "rv");
	}

	return 1;
	};

//$ AD 2017-12-08: command TimeMillis()
mystr CEDC_Timer::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item 0..no ouput, 1.. line in  output window and log file \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_Timer::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// first container holds the output flag
	ElementData* ed_flag = parameter_EDCs(1)->GetPtr(1);
	int flag_output = 0;
	if(ed_flag->IsBool())
	{
		flag_output = ed_flag->GetBool(); // integer variabe
	}
	else if(ed_flag->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr str = ed_flag->GetText();
		flag_output = str.MakeInt();
	}

	if(flag_output)
	{
		time_t t; time(&t); // seconds
		struct tm *ts; ts = localtime(&t); mystr str = mystr("=========\nTIMESTAMP: ") + asctime(ts) + mystr("=========\n"); // output string
		mbs->UO() << str;
	}

	int timeMS = GetTickCount();
	return_value.SetInt(timeMS,"ticks");
	return 1;// rv;
}


//$ AD 2017-05-24: command AccessSet(anyset)
mystr CEDC_ComSet_AccessSet::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a setnumber \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComSet_AccessSet::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// first container holds the set
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in AccessSet - Parameter 1 could not be resolved into a Set"));
		return 0;
	}

	BaseSet* theSet = mbs->GetSetPtr(setIdx);
	theSet->Convert();

	ElementDataContainer edc_set;
	theSet->GetElementData(edc_set);


	return_value.SetEDC(&edc_set,"result");
	return 1;//rv;
};


//$ AD 2015-01-29: command GenerateConstraints(nodeset,) - generates constraints for the given set
mystr CEDC_ComSet_GenerateConstraints::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of global node numbers (more types will be added)\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: parameters for the constraints\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item mode - 'ground' or 'pair'(default) \n");
	descr +=            mystr("  \\item type - type string for the constraint to use \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

class BasePointJoint;
int CEDC_ComSet_GenerateConstraints::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// first container holds the set
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in GenerateConstraints - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in GenerateConstraints - requires set_type = set_of_global_nodes"));
		return 0;
	}
	GlobalNodeSet* gnset = (GlobalNodeSet*) mbs->GetSetPtr(setIdx);
//	gnset->Convert(); // elements and local positions required ...

	// second container holds the parameters for the command
	ElementDataContainer* edc_params = parameter_EDCs(2); 
	int mode = 2;  // 'default mode' is pairs 
	if(edc_params->TreeFind("mode"))
	{
		mystr modestr = edc_params->TreeGetString("mode");
		if(modestr.Compare("ground"))
		{
			mode=1;
		}
		else if(modestr.Compare("find"))
		{
			mode=3;
		}
	}
	mystr elemtype = "GenericBodyJoint";
	if(edc_params->TreeFind("type"))
	{
		elemtype = edc_params->TreeGetString("type");
	}
	int loccoord1 = 0, loccoord2 = 0;
	ElementData *ed = edc_params->TreeFind("coordinate1");
	if (ed)
	{
		if (ed->GetType() == 1) //int
		{
			loccoord1 = ed->GetInt();
		}
		else if (ed->GetType() == 2) //double
		{
			loccoord1 = (int) (ed->GetDouble() + 0.5);
		}
		else if (ed->GetType() == 8) //text
		{
			mystr text(ed->GetText());
			if (text.IsValidNumber())
			{
				loccoord1 = text.MakeInt();
			}
			else
			{
				edcParser->EDCError("Error in GenerateConstraints: \"" + text + "\" cannot be parsed into a number");
			}
		}
		else
		{
			edcParser->EDCError("Error in GenerateConstraints: wrong datatype for entry coordinate1");
		}
	}
	ed = edc_params->TreeFind("coordinate2");
	if(ed)
	{
		if (ed->GetType() == 1) //int
		{
			loccoord2 = ed->GetInt();
		}
		else if (ed->GetType() == 2) //double
		{
			loccoord2 = (int) (ed->GetDouble() + 0.5);
		}
		else if (ed->GetType() == 8) //text
		{
			mystr text(ed->GetText());
			if (text.IsValidNumber())
			{
				loccoord2 = text.MakeInt();
			}
			else
			{
				edcParser->EDCError("Error in GenerateConstraints: \"" + text + "\" cannot be parsed into a number");
			}
		}
		else
		{
			edcParser->EDCError("Error in GenerateConstraints: wrong datatype for entry coordinate2");
		}
	}

	if (loccoord1 != 0 && loccoord2 == 0)
	{
		mode = 1;
	}
	ed = edc_params->TreeFind("findfilter");
	if(ed && mode==3) // AD: override for new mode - use parameter loccoord2 to pass a set number for the allowed matching elements...
	{
		int setIdx = ResolveParameterEntryToSetNumber(mbs,edcParser,ed);
		loccoord2 = setIdx;
	}

	ElementDataContainer * constraintTemplate = 0;
	int templateIndex = edc_params->Find("template");
	if (templateIndex != 0)
	{
		ed = edc_params->GetPtr(templateIndex);
		if (!ed->IsEDC())
		{
			edcParser->EDCError("Error in GenerateConstraints: template must be a container");
		}
		else
		{
			constraintTemplate = ed->GetEDC();
			if(constraintTemplate->TreeFind("element_type"))
			{
				elemtype = constraintTemplate->TreeGetString("element_type");
			}
		}
	}

	TArray<int> constraintnumbers = mbs->GenerateConstraints(gnset, elemtype, mode, loccoord1, loccoord2, constraintTemplate);
	Vector cnrs(constraintnumbers.Length()); for(int i=1; i<=constraintnumbers.Length(); i++) { cnrs(i) = constraintnumbers(i); }

	return_value.SetVector(cnrs.GetVecPtr(), cnrs.Length(), "constraint_numbers"); return_value.SetValuesInt(); return_value.SetVariableLength(); return_value.SetLocked(0); return_value.SetToolTipText("returned by a call of 'GenerateConstrains'\n");

	// generate the constraints 
	//////////int typenr = mbs->GetObjectFactory()->GetObjectTypeId(OFCElement,elemtype);
	//////////TArray<int> constraintnumbers;
	
	//////////int inc = mode; // one entry per constraint for a ground constraint, two for pairs
	//////////for(int i=1; i<=gnset->GetNrOfEntries(); i+=inc)
	//////////{
	//////////	int cnr = mbs->GetObjectFactory()->AddObject(OFCElement, typenr); // new constraint generated here
	//////////	BasePointJoint* pbpj = (BasePointJoint*) mbs->GetElementPtr(cnr);
	//////////	constraintnumbers.Add(cnr);

	//////////	TArray<int>& el = pbpj->GetElementsArray();
	//////////	TArray<Vector3D>& lp = pbpj->GetLoccoordsArray();		
	//////////	el(1) = gnset->GetElementNumber(i);
	//////////	lp(1) = gnset->GetLocalPosition(i);

	//////////	if(mode==1)
	//////////	{
	//////////		el(2) = 0; 
	//////////		lp(2) = mbs->GetNodePtr(gnset->GetGlobalNodeNumber(i))->GetPos());
	//////////	}
	//////////	else 
	//////////	{ 
	//////////		el(2) = gnset->GetElementNumber(i+1);
	//////////		lp(2) = gnset->GetLocalPosition(i+1);
	//////////	}
	//////////}
	return 1;// rv;
}



//$ AD 2016-02-03: command GenerateSensors(nodeset,parameter) - generates sensorss for the given set
mystr CEDC_ComSet_GenerateSensors::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of global node numbers (more types will be added)\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: parameters for the sensors\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("enties in the properties EDC are: \n");
	descr +=			mystr("\\begin{itemize} \n");	
	descr +=			mystr("	 \\item template - a template container containing SensorType etc... \n");
	descr +=			mystr("\\end{itemize} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComSet_GenerateSensors::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	// first container holds the set
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in GenerateSensors - Parameter 1 could not be resolved into a Set"));
		return 0;
	}

	//$EK 2020-02-25: GenerateSensors can be called with all node sets instead of only GlobalNodeSet (PointSetBC-> GlobalNodeSet, LocalNodesA, LocalNodesB)
	TMBSSetType stype = mbs->GetSetPtr(setIdx)->GetType();
	if((stype != /*TMBSSetType::*/TMBSSetGlobalNodes) & (stype != TMBSSetPoints) & 
		(stype !=TMBSSetLocalNodesA) & (stype != TMBSSetLocalNodesB))  	// check type
	{
		edcParser->EDCError(mystr("Error in GenerateSensors - requires set_type = set_of_global_nodes/TMBSSetPoints/TMBSSetLocalNodesA/TMBSSetLocalNodesB"));
		return 0;
	}
	//$EK 2020-02-25: GenerateSensors can be called with all node sets instead of only GlobalNodeSet (PointSetBC-> GlobalNodeSet, LocalNodesA, LocalNodesB)
	PointSetBC* nset = (PointSetBC*) mbs->GetSetPtr(setIdx);

	// second container holds the parameters for the command
	ElementDataContainer* edc_params = parameter_EDCs(2); 
	ElementDataContainer * sensorTemplate = 0;
	int templateIndex = edc_params->Find("template");
	if (templateIndex != 0)
	{
		ElementData* ed = edc_params->GetPtr(templateIndex);
		if (!ed->IsEDC())
		{
			edcParser->EDCError("Error in GenerateSensors: template must be a container");
		}
		else
		{
			sensorTemplate = ed->GetEDC();
			if(sensorTemplate->TreeFind("sensor_type"))
			{
				mystr elemtype = sensorTemplate->TreeGetString("sensor_type");
				
				//$EK 2020-02-25: GenerateSensors can be called with all node sets instead of only GlobalNodeSet (PointSetBC-> GlobalNodeSet, LocalNodesA, LocalNodesB)
				TArray<int> sensornumbers = mbs->GenerateSensors(nset, elemtype, sensorTemplate);
				Vector snrs(sensornumbers.Length()); for(int i=1; i<=sensornumbers.Length(); i++) { snrs(i) = sensornumbers(i); }

				return_value.SetVector(snrs.GetVecPtr(), snrs.Length(), "sensor_numbers");  return_value.SetValuesInt(); return_value.SetVariableLength(); /*return_value.SetLocked(1);*/ return_value.SetToolTipText("returned by a call of 'GenerateSensors'\n");
			}
		}
	}
	else
	{
		edcParser->EDCError("Error in GenerateSensors: a Sensor template must be defined");
		return 0;
	}

	return 1;// rv;
}



//$ AD 2014-08-01: command ApplyMaterial(elset,nr) - sets the material number of all the elements of hte set to the given number
mystr CEDC_ComAssignMaterial::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of elements\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: material number to be assigned\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComAssignMaterial::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in AssignMaterial - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetElements) 	// check type
	{
		edcParser->EDCError(mystr("Error in AssignMaterial - requires set_type = set_of_elements"));
		return 0;
	}
	ElementSet* elset = (ElementSet*) mbs->GetSetPtr(setIdx);
	
	ElementData* ed_mat = parameter_EDCs(2)->GetPtr(1); // input parameter "Material"
	int matIdx = ResolveParameterEntryToMaterialNumber(mbs, edcParser, ed_mat);
	if(matIdx==0) // check existence
	{
		edcParser->EDCError(mystr("Error in AssignMaterial - Parameter 2 could not be resolved into a Material"));
		return 0;
	}

	// actually change the material type of the elements
	for(int i=1; i<= elset->GetNrOfEntries(); i++)
	{
		mbs->GetElementPtr(elset->GetElementNumber(i))->SetMaterialNum(matIdx);
	}
	
	return 1;// rv;
}

//$ AD 2014-08-11: command ComAssignLoad(elementset,loadnumber) - assigns a(n additional) load to  all the elements of the set
mystr CEDC_ComAssignLoad::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of elements\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: load number to be added or \"ClearAll\" to remove all loads\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComAssignLoad::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0)
	{
		edcParser->EDCError(mystr("Error in AssignLoad - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetElements)
	{
		edcParser->EDCError(mystr("Error in AssignLoad - requires set_type = set_of_elements"));
		return 0;
	}
	ElementSet* elset = (ElementSet*) mbs->GetSetPtr(setIdx);

	
	ElementData* ed_load = parameter_EDCs(2)->GetPtr(1); // input parameter "Load"
	// special input:   Load == "ClearAll"
	if(ed_load->IsText())
	{
		mystr txt = ed_load->GetText();
		if( txt.Compare(mystr("ClearAll")) )
		{
			for(int i=1; i<= elset->GetNrOfEntries(); i++)
			{
				while(mbs->GetElementPtr(elset->GetElementNumber(i))->NLoads())
				{
					mbs->GetElementPtr(elset->GetElementNumber(i))->DeleteLoad(1);	
				}
			}
			return 0;
		}
	}
	
	int loadIdx = ResolveParameterEntryToLoadNumber(mbs, edcParser, ed_load);
	if(loadIdx==0) // check existence
	{
		edcParser->EDCError(mystr("Error in AssignLoad - Parameter 2 could not be resolved into a Load"));
		return 0;
	}
// TO DO: REVISIT type check to limit the allowed load types
	int loadType = mbs->GetLoadPtr(loadIdx)->GetLoadType();
	if(! ((loadType == /*TMBSLoad::*/Tbodyload) ||
	      (loadType == /*TMBSLoad::*/Tgravity) ||
	      (loadType == /*TMBSLoad::*/TGCload) ||
	      (loadType == /*TMBSLoad::*/Tsurfacepressure) ||
	      (loadType == /*TMBSLoad::*/Tforcevector2D) ||
	      (loadType == /*TMBSLoad::*/Tforcevector3D) ||
	      (loadType == /*TMBSLoad::*/Tmomentvector2D) ||
	      (loadType == /*TMBSLoad::*/Tmomentvector3D) ||
	      (loadType == /*TMBSLoad::*/Tbodyloadspatial) ||
	      (loadType == /*TMBSLoad::*/Tsurfaceheatsource) ||
	      (loadType == /*TMBSLoad::*/Tvolumeheatsource) ||
	      (loadType == /*TMBSLoad::*/Tareaload)  ||
		  (loadType == /*TMBSLoad::*/Tareatraction)  ||
		  (loadType == /*TMBSLoad::*/Tcentrifugal) ||
		  (loadType == /*TMBSLoad::*/Tsurfaceload) ))
	{
		edcParser->EDCError(mystr("Error in AssignLoad - requires load_type = { BodyLoad, GravityLoad, GCLoad, SurfacePressure, ForceVector2D, ForceVector3D, MomentVector2D, MomentVector3D}"));
		return 0;	
	}
	MBSLoad* pLoad = mbs->GetLoadPtr(loadIdx);

	// actually change the material type of the elements
	mystr errstr; // for consistency checks
	for(int i=1; i<= elset->GetNrOfEntries(); i++)
	{
		//AD: consistency check always fails - element::Dim() = 3 but load::Dim() = 0
		////if( pLoad->CheckConsistency(mbs->GetElement(elset->GetElementNumber(i)), errstr) )
		////{
		////	edcParser->EDCError(mystr("Error in AssignLoad - ConsistencyCheck failed: ")+errstr);
		////	return 0;
		////}

		//mbs->GetElementPtr(elset->GetElementNumber(i))->AddLoad(*pLoad); // this would create an instance of the same MBSLoad for each element...
		TArray<int> el_loads = mbs->GetElementPtr(elset->GetElementNumber(i))->GetLoadNrs(); // while this simply adds the load number of an existing load to the elements list of loads
		el_loads.Add(loadIdx);
		mbs->GetElementPtr(elset->GetElementNumber(i))->SetLoadNrs(el_loads);
	}
	return 1;//rv;
}

//==================================================================================================================
//$ DR 2015-02-17: command AssignGeomElementToElement(elementset,geomelment_number) - assigns a(n additional) geomelement to all the elements of the set
mystr CEDC_ComAssignGeomElementToElement::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: an element number\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: a geom element number\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComAssignGeomElementToElement::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementDataContainer* edc_par = parameter_EDCs(1); // input parameter "Element"

	ElementData* edp = edc_par->GetPtr(1);
	int elNr = 0;
	if(edp->IsInt())
	{
		elNr = edp->GetInt(); // integer variabe
	}
	if(elNr <= 0 || elNr > mbs->NE())
	{
		edcParser->EDCError(mystr("Error in AssignGeomElementToElement - Parameter 1 is not a valid element number"));
		return 0;
	}
	
	edc_par = parameter_EDCs(2); // input parameter "GeomElement"
	edp = edc_par->GetPtr(1);
	int geomElNr = 0;
	if(edp->IsInt())
	{
		geomElNr = edp->GetInt(); // integer variabe
	}
	if(geomElNr <= 0 || geomElNr > mbs->NDrawElements())
	{
		edcParser->EDCError(mystr("Error in AssignGeomElementToElement - Parameter 2 is not a valid geom element number"));
		return 0;
	}

	// change the settings of the element
	TArray<int> el_geomEl = mbs->GetElementPtr(elNr)->GetGeomElementNrs();
	el_geomEl.Add(geomElNr);
	mbs->GetElementPtr(elNr)->SetAltShape(1);
	mbs->GetElementPtr(elNr)->SetGeomElementNrs(el_geomEl);

	// change the settings of the geomelement
	//Vector3D refPoint = mbs->GetElementPtr(elNr)->GetRefPosD();
	Vector3D refPoint = mbs->GetElementPtr(elNr)->GetRefPos_dc(TCD_reference_configuration);
	mbs->GetDrawElement(geomElNr)->SetElnum(elNr);
	//Matrix3D rotate = mbs->GetElementPtr(elNr)->GetRotMatrixD();
	Matrix3D rotate = mbs->GetElementPtr(elNr)->GetRotMatrix_dc(refPoint,TCD_reference_configuration);

	//mbs->UO()<< "AssignGeomElementToElement \n";
	//mbs->UO()<< "RefPos =  " << refPoint << " \n";
	//mbs->UO()<< "RotMat =  " << rotate << " \n";

	int success = rotate.Invert();	
	Matrix3D rotInv = rotate;

	//mbs->UO()<< "Compute RotMat^-1: success = " << success << " \n";
	//mbs->UO()<< "RotMat^-1 =  " << rotInv << " \n";

	mbs->GetDrawElement(geomElNr)->Rotate(rotInv);
	mbs->GetDrawElement(geomElNr)->Translate(-1*rotInv*refPoint);

	return 1;//rv;
}
//==================================================================================================================
//$ SW 2015-7-1: command SetInitialCondition(elementset/global node set, index, expression) - set initial value for all elements
mystr CEDC_ComSetInitialCondition::GetCommandTexParameterDescription() const
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of elements or global nodes\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: index of initial value\n");
	descr +=			mystr("	 \\item $3^{rd}$ parameter: expression of the value. Can contain: \n");
	descr +=			mystr("	 \\begin{itemize} \n");
	descr +=			mystr("	 \\item \"x\" for the global reference position of the node or element\n");
	descr +=			mystr("	 \\item \"y\" for the global reference position of the node or element\n");
	descr +=			mystr("	 \\item \"z\" for the global reference position of the node or element\n");
	descr +=			mystr("	\\end{itemize} \n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");
	return descr;
}

int CEDC_ComSetInitialCondition::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	
	// TODO ? Check if 3 parameters are provided
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in SetInitialCondition - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetElements &&
		 mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in SetInitialCondition - requires set_type = set_of_elements || set_of_global_nodes"));
		return 0;
	}

	int index = 0;
	ElementDataContainer* edc_index = parameter_EDCs(2);
	if (edc_index->GetType() != 1 /*->EDC*/)
	{
		return 0;
	}
	ElementData* edp = edc_index->GetPtr(1);
	if(edp->IsInt())
	{
		index = edp->GetInt(); // integer variabe
	}
	else if(edp->IsText())
	{
		mystr str = edp->GetText();
		if(str.IsValidNumber())
		{
			index = str.MakeInt(); // inline number
		}
		else
		{
			return 0;
		}
	}
	else if(edp->IsDouble())
	{
		index = (int) (edp->GetDouble() + 0.5);
	}
	else
	{
		return 0;
	}

	ElementDataContainer* edc_expression = parameter_EDCs(3);
	if (edc_expression->GetType() != 1 /*->EDC*/ || edc_expression->Get(1).GetType() != 8 /*->text*/ )
	{
		return 0;
	}
	
	char * expression = edc_expression->Get(1).GetText();


	CParsedFunction expr= CParsedFunction();

	expr.SetParsedFunction3D(&(mbs->MBSParser()), expression,"x,y,z");

	if(mbs->GetSetPtr(setIdx)->GetType() == /*TMBSSetType::*/TMBSSetElements)
	{
		ElementSet* elset = (ElementSet*) mbs->GetSetPtr(setIdx);

		// actually change the EDC of the elements
		for(int i=1; i<=elset->GetNrOfEntries(); i++)
		{
			Element* el = mbs->GetElementPtr(elset->GetElementNumber(i));
			if ( index <= el->SS() )
			{
				Vector x_init(el->GetXInit());
				Vector3D pos = el->GetRefPos();
				x_init(index) = expr.Evaluate(pos);
				el->SetInitConditions(x_init);
			}
			else
			{
				edcParser->EDCError("Error in SetInitialCondition: index out of bounds (index = " + mystr(index) + ", maximum possible element index: " + mystr(el->SS()) + ")\r\n");
			}
		}
	}
	else if(mbs->GetSetPtr(setIdx)->GetType() == /*TMBSSetType::*/TMBSSetGlobalNodes) 
	{
		GlobalNodeSet* gnset = (GlobalNodeSet*) mbs->GetSetPtr(setIdx);

		for(int i=1; i<=gnset->GetNrOfEntries(); i++)
		{
			Node* n = mbs->GetNodePtr(gnset->GetGlobalNodeNumber(i));
			int ss = n->SOS()*2+n->IS()+n->ES();
			if (index <= ss)
			{
				Vector & x_init = n->X_Init();
				Vector3D pos = n->RefConfPos();
				x_init(index) = expr.Evaluate(pos);
			}
			else
			{
				edcParser->EDCError("Error in SetInitialCondition: index out of bounds (index = " + mystr(index) + ", maximum possible node index: " + mystr(ss) + ")\r\n");
			}
		}
	}
	return 1;
}

//==================================================================================================================
//$ AD 2015-01-28: command ComChangeProperties(elementset,substitute) - change any properties for all elements of the set
mystr CEDC_ComChangeProperties::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: a set of elements or global nodes\n");
	descr +=			mystr("	 \\item $2^{nd}$ parameter: EDC containing substitute parameters EDC\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_ComChangeProperties::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;
	
	// first container holds a SET
	ElementData* ed_set = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set);
	if(setIdx==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in CEDC_ComChangeProperties - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if(mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetElements &&
		 mbs->GetSetPtr(setIdx)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in CEDC_ComChangeProperties - requires set_type = set_of_elements || set_of_global_nodes"));
		return 0;
	}

	// second container holds the replacement EDC
	ElementDataContainer* edc_replacement = parameter_EDCs(2); 

	mystr warn_str;
	if(mbs->GetSetPtr(setIdx)->GetType() == /*TMBSSetType::*/TMBSSetElements)
	{
		ElementSet* elset = (ElementSet*) mbs->GetSetPtr(setIdx);

		// actually change the EDC of the elements
		for(int i=1; i<=elset->GetNrOfEntries(); i++)
		{
			Element* el = mbs->GetElementPtr(elset->GetElementNumber(i));
			ElementDataContainer edc_el; 
			el->GetElementData(edc_el);

			edc_el.TreeReplaceEDCDataWith(edc_replacement,1,warn_str);
			
			el->SetElementData(edc_el);
		}
	}
	else if(mbs->GetSetPtr(setIdx)->GetType() == /*TMBSSetType::*/TMBSSetGlobalNodes) 
	{
		GlobalNodeSet* gnset = (GlobalNodeSet*) mbs->GetSetPtr(setIdx);

		for(int i=1; i<=gnset->GetNrOfEntries(); i++)
		{
			Node* n = mbs->GetNodePtr(gnset->GetGlobalNodeNumber(i));
			ElementDataContainer edc_n;
			n->GetElementData(edc_n);

			edc_n.TreeReplaceEDCDataWith(edc_replacement,1,warn_str);

			n->SetElementData(edc_n);
		}
	}
	else return 0;
	
	return 1;// rv;
}

int CEDC_ComSet_MatchSets::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv=1;

	// first container holds a SET
	ElementData* ed_set_order = parameter_EDCs(1)->GetPtr(1); // input parameter "SET"
	int setIdx_order = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set_order);
	if (setIdx_order==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in MatchSets - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if (mbs->GetSetPtr(setIdx_order)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in MatchSets - requires set_type = set_of_elements || set_of_global_nodes"));
		return 0;
	}

		// first container holds a SET
	ElementData* ed_set_toSort = parameter_EDCs(2)->GetPtr(1); // input parameter "SET"
	int setIdx_toSort = ResolveParameterEntryToSetNumber(mbs, edcParser, ed_set_toSort);
	if (setIdx_toSort==0) 	// check existence
	{
		edcParser->EDCError(mystr("Error in MatchSets - Parameter 1 could not be resolved into a Set"));
		return 0;
	}
	if (mbs->GetSetPtr(setIdx_toSort)->GetType() != /*TMBSSetType::*/TMBSSetGlobalNodes) 	// check type
	{
		edcParser->EDCError(mystr("Error in MatchSets - requires set_type =  set_of_global_nodes"));
		return 0;
	}

	mbs->MatchSets(mbs->GetSetPtr(setIdx_order),mbs->GetSetPtr(setIdx_toSort));

	return 1;// rv;
}

//$ AD 2014-08-11: command CEDC_DeleteElement(nr) 
mystr CEDC_DeleteElement::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: number of the element to delete from the MBS\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_DeleteElement::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ED_par1 = parameter_EDCs(1)->GetPtr(1);
  int par1;
	if(ED_par1->IsInt()) // entry is a variable of integer type
	{
		par1 = ED_par1->GetInt();
	}
	if(ED_par1->IsDouble()) // entry is a variable of double type
	{
		par1 = (int) ED_par1->GetDouble();
	}
	else if(ED_par1->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr p1 = ED_par1->GetText();
		par1 = p1.MakeInt();
	}
	else
	{
		edcParser->EDCError("Error while executing DeleteElement operation: first parameter is not an integer");
		return rv;
	}

	mbs->DeleteElement(par1);

	return 1;
}

//$ AD 2014-08-11: command CEDC_DeletNode(nr) 
mystr CEDC_DeleteNode::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: number of the node to delete from the MBS\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
}

int CEDC_DeleteNode::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int rv = 0;

	ElementData* ED_par1 = parameter_EDCs(1)->GetPtr(1);
  int par1;
	if(ED_par1->IsInt()) // entry is a variable of integer type
	{
		par1 = ED_par1->GetInt();
	}
	if(ED_par1->IsDouble()) // entry is a variable of double type
	{
		par1 = (int) ED_par1->GetDouble();
	}
	else if(ED_par1->IsText()) // entry is a number ( inline, gets recognized as a string )
	{
		mystr p1 = ED_par1->GetText();
		par1 = p1.MakeInt();
	}
	else
	{
		edcParser->EDCError("Error while executing DeleteNode operation: first parameter is not an integer");
		return rv;
	}

	mbs->DeleteNode(par1);

	return 1;
}


mystr CEDC_OpenCompiledModel::GetCommandTexParameterDescription() const 
{
	mystr descr = mystr("The parameters of this command are \n");
	descr +=			mystr("\\begin{enumerate} \n");
	descr +=			mystr("	 \\item $1^{st}$ parameter: string with the name of the model to load\n");
	descr +=			mystr("\\end{enumerate} \n");
	descr +=			mystr("-");	//DR hack, to avoid bugs in latex
	return descr;
};

int CEDC_OpenCompiledModel::ExecuteCommand(MBS * mbs, CEDCParser * edcParser, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	// first container contains the model name
	ElementDataContainer* edc_modelname = parameter_EDCs(1);
	ElementData* ed_modelname = edc_modelname->GetPtr(1);
	mystr modelname;
	if(ed_modelname!=NULL && ed_modelname->IsText()) { modelname = ed_modelname->GetText(); }
	else { edcParser->EDCError(mystr("Error in OpenCompiledModel - first entry (modelname) must be a string")); }	

	// compare to available compiled models
	int modelidx = 0;
	for(int i=1; i<= mbs->GetModelsLibrary()->GetModelsCount(); i++)
	{
		mystr availmodelname = mbs->GetModelsLibrary()->GetModelInterface(i)->GetMBSModelName();
		int one4match =  availmodelname.Compare(modelname);
		if(one4match) 
		{
			modelidx = i;
			break;
		}
	}
	if(modelidx==0)
	{
		edcParser->EDCError("Error while executing OpenCompiledModel operation: specified model not in List of compiled models");
		return 0; 1;
	}

// select the comliled model
	mbs->GetOptions()->GeneralOptions()->ModelFileInternalModelFunctionName() = modelname;

// mark as called from HID, 
	mystr filename = mbs->GetOptions()->GeneralOptions()->ModelFileHotintInputDataFilename();
	ElementData ed;
	ed.SetText(filename,"HidCalledCpp");
	mbs->GetModelDataContainer()->Add(ed);

	return 1;
};


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//void CEDCParser::EDC2TexTable(const ElementDataContainer& edc, mystr& str, const mystr& indent, const mystr& treename) //$ DR 2013-01-11 moved to models_auto_documentation

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int CEDCParser::GetParameterEDC(mystr text, TArray<ElementDataContainer*>& parameter_EDCs)
{
	int parPos = 0;
	int parCount = 1;
	mystr EDCtext;

	text.GetUntil(parPos, ',', EDCtext, 0);
	while(1)		// handle all parameters
	{
		bool success = false;
		//int future: this should be either an EDC name, or a structure defining an EDC {...}
		//at the moment, only an EDC-name is possible
		EDCtext.EraseSpacesHeadTail();

		//the EDC must be either global or local
		ElementDataContainer* edc = mbsParser->GetLocalEDC(); //local
		ElementDataContainer* edc2 = mbsParser->GetLocalEDC2(); //global
		ElementDataContainer* edcMBS = mbs->GetModelDataContainer();	//$ DR 2013-07-25 added for the case of multiple input files

		//$ RL 2012-7-19:[ new code with ExecuteCommand planned to be possible only with arguments
		ElementData* ed = 0; 
		if(edc)
		{
			ed = edc->TreeFind(EDCtext); // search only if edc exists!!
		}
		ElementData* ed2 = 0;
		if(edc2)
		{
			ed2 = edc2->TreeFind(EDCtext); // search only if edc exists!!
		}
		ElementData* edMBS = 0;
		if(edcMBS)
		{
			edMBS = edcMBS->TreeFind(EDCtext); //$ DR 2013-07-25 added for the case of multiple input files
		}

		if (!ed && ed2) {ed = ed2;}
		else if (!ed && !ed2 && edMBS) {ed = edMBS;}	//$ DR 2013-07-25 added for the case of multiple input files
		//else if (!ed && !ed2) 
		else if (!ed && !ed2 && !edMBS)
		{
			// no parameter edc found -> add parameter string to parameter_EDCs
			int pos=0;
			//if (text.PosPeek(pos) == doublequote)
			if (EDCtext.PosPeek(pos) == doublequote)
			{
				//pos = text.Length()-1;
				pos = EDCtext.Length()-1;
				//if(text.PosPeek(pos) == doublequote)
				if(EDCtext.PosPeek(pos) == doublequote)
				{
					// text is within doublequotes
					pos=0;
					//text = text.GetStringInBrackets(pos, doublequote, doublequote); // remove doublequotes
					EDCtext = EDCtext.GetStringInBrackets(pos, doublequote, doublequote); // remove doublequotes				
				}
			}
			ElementData ed3;
			mystr parName = "parameter"+mystr(parCount);
			//ed3.SetText(text,parName); // store parameter text
			ed3.SetText(EDCtext,parName); // store parameter text
			ElementDataContainer edc_text;
			edc_text.Add(ed3);
			parameter_EDCs.Add(edc_text.GetCopy()); //parameter_EDCs are deleted outside this function 
			//return 1;
			success = true;
		}

		if(!success)
		{
			//$ DR 2012-11-29: [ also use element data, e.g. for CEDC_ComPrint
			if (!ed->IsEDC() && (ed->IsBool() || ed->IsInt() || ed->IsDouble() || ed->IsMatrix() || ed->IsText() || ed->IsVector()))
			{
				ElementDataContainer edc_ed;
				edc_ed.Add(*ed->GetCopy());
				parameter_EDCs.Add(edc_ed.GetCopy()); //parameter_EDCs are deleted outside this function 
				//return 1;
				success = true;
			}
			//$ DR 2012-11-29: ]
		}

		if(!success)
		{
			//always use ed, ed2 overrides ed:
			//now the parameter should indicate an EDC
			if (ed->IsEDC())
			{
				parameter_EDCs.Add((ed->GetEDC())->GetCopy());  //parameter_EDCs are deleted outside this function 
			}
			else
			{
				//failed
				EDCError(mystr("ERROR: Expected a parameter set (ElementDataContainer) for function parameters"));
				return 0;
			}
		}

		if(parPos == -1) return 1;	// last parameter
		parPos++;
		text.GetUntil(parPos, ',', EDCtext, 0);
		parCount++;
	}
	return 1;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//this function is already adapted for TreeEDCs!
//$ DR 2013-01-31 empty subEDCs are not written at all, see log 384
void CEDCParser::EDC2String(const ElementDataContainer& edc, mystr& str, const mystr& indent)
{
//symbols already defined in CEDCParser:
//	mystr el = "\n"; //end of line
//	mystr ob = "{";  //open brace
//	mystr cb = "}";  //closing brace
//	mystr osb = "[";  //open square bracket
//	mystr csb = "]";  //closing square bracket
//	mystr ds = ": "; //double stop
//	mystr semicolon = ";"; //semicolon
//	char commentc = '%'; //comment
//	mystr doublequote = mystr('"'); //for strings or text

//symbols with additional whitespaces:
	mystr eq = "= "; //equal sign
	mystr comma = ", "; //comma
	mystr comment = mystr("  ")+mystr(commentc); //comment

	mystr matrixsep = el; //may be changed to semicolon ...
	str = "";
	mystr num;

	for (int i=1; i <= edc.Length(); i++)
	{
		const ElementData& ed = edc.Get(i);

		if (!ed.IsLocked() && !ed.IsElementInfo() && !ed.IsCompAction() && !ed.IsWCDriverAction())
		{
			//str += indent + ed.GetDataName();
			if (ed.IsEDC())
			{ //hierarchically
				mystr substr;
				const ElementDataContainer* edcptr = ed.GetEDC();
				if (edcptr)
				{
					EDC2String(*edcptr, substr, indent+mystr("  "));
				}
				if(substr.Length())
				{
					str += indent + ed.GetDataName();
					str += el + indent + ob + el;
					str += substr;
					str += indent + cb + el;
				}
			}
			else
			{ //just this list
				str += indent + ed.GetDataName();//
				str += eq;
				if (ed.IsBool())
				{
					str += mystr(ed.GetBool());
				}
				else if (ed.IsInt())
				{
					str += mystr(ed.GetInt());
				}
				else if (ed.IsDouble())
				{
					num.SmartDouble2String(ed.GetDouble());
					str += num;
				}
				else if (ed.IsText())
				{
					str += mystr(doublequote) + mystr(ed.GetText()) + mystr(doublequote);
				}
				else if (ed.IsVector())
				{
					str += osb;
					for (int j = 1; j <= ed.GetVectorLen(); j++)
					{
						num.SmartDouble2String(ed.GetVectorVal(j));
						str += num;
						if (j < ed.GetVectorLen()) str += comma;
					}
					str += csb;
				}
				else if (ed.IsMatrix())
				{
					str += osb;
					for (int j1 = 1; j1 <= ed.GetMatrixRows(); j1++)
					{
						for (int j2 = 1; j2 <= ed.GetMatrixCols(); j2++)
						{
							num.SmartDouble2String(ed.GetMatrixVal(j1, j2));
							str += num;
							if (j2 < ed.GetMatrixCols()) str += comma;
						}
						if (j1 < ed.GetMatrixRows()) str += matrixsep + indent + mystr("    ");
					}
					str += csb;
				}
				if (ed.HasToolTip())
				{
					str += comment + ed.GetToolTipText();
				}
				str += el;
			}
		}
	}
}

//old version for saving MBS data, do not use!
//$JG2012-01-27: mbs is available in CEDCParser
int CEDCParser::String2EDC(/*MBS* mbs, */mystr& str, ElementDataContainer& edc, ElementDataContainer& testedc)
{
	char osb = '[';  //open square bracket
	char csb = ']';  //closing square bracket
	char ds = ':'; //double stop
	char eq = '='; //equal sign
	char comma = ','; //comma sign
	char semicolon = ';'; //semicolon sign

	int oo = 0;
	if (oo) mbs->UO() << "String2EDC\n";

	mystr dataname, data, data2, data3;
	int endstr = 0;
	int pos = 0;
	int posold;
	TArray<double> datalist;


	while (!endstr)
	{
		posold = pos;
		dataname = str.GetWord(pos, 0);
		if (pos != -1 && str[pos] == ds)
		{
			str = str.Right(str.Length()-posold); return 1;
		}
		if (pos != -1)
		{
			if (oo) mbs->UO() << "Dataname='" << dataname << "'\n";

			if (!str.GoUntil(pos, eq)) {return 0;}; //write error

			int x = testedc.Find(dataname.c_str());
			if (x)
			{
				ElementData& ed = testedc.Get(x);
				if (ed.IsVector())
				{
					//read [ ... ]
					if (!str.GoUntil(pos, osb)) {LoadError("Vector '[...]' expected!"); return 0;}

					if (!str.GetUntil(pos, csb, data, 0)) {LoadError(mystr("Vector '[...]' expected, but received '")+data+mystr("'")); return 0;};

					int len = ed.GetVectorLen();
					datalist.SetLen(0);

					int pos2 = 0;
					while (pos2 != -1 && (datalist.Length() <= len || ed.IsVariableLength()))
					{
						data.GetUntil(pos2, comma, data2, 0);

						datalist.Add(data2.MakeDouble());
						if (pos2 != -1) data.GoUntil(pos2, comma);
					}

					if (datalist.Length() != len && !ed.IsVariableLength()) {LoadError(mystr(ed.GetDataName())+mystr(": Vector of length ")+mystr(len)+mystr(" expected, but received length=")+mystr(datalist.Length())); return 0;};

					len = datalist.Length();
					Vector v(len);
					for (int i=1; i <= len; i++) v(i) = datalist(i);

					ed.SetVector(v.GetVecPtr(), len, dataname.c_str());
					edc.Add(ed);
					if (oo) mbs->UO() << v << "\n";

					str.GoUntil(pos, csb); //read final square bracket


				}
				else if (ed.IsMatrix())
				{
					//read [ ... ]
					if (!str.GoUntil(pos, osb)) {LoadError("Matrix '[...]' expected!"); return 0;}

					if (!str.GetUntil(pos, csb, data, 0)) {LoadError(mystr("Matrix '[...]' expected, but received '")+data+mystr("'")); return 0;};

					int rows = ed.GetMatrixRows();
					int cols = ed.GetMatrixCols();


					TArray<double> doublelist;

					//++++++++++++++++++++++++
					//same as in CustomEditDialog!
					char cstr[258];
					int limit = 256;
					int startpos = 0;

					int end = 0;
					int endpos, endpos2;
					int maxcol = 0;
					int error = 0;
					int nrows = 0;

					data = data+mystr("\n\n");

					while (!end)
					{
						int emptystring = 1;
						int ncol = 0;
						int endcol = 0;

						while (!endcol)
						{
							endpos = data.Find(startpos,',');
							endpos2 = data.FindEOL(startpos);

							if (endpos2 < endpos || endpos == -1) 
							{
								endpos = endpos2;
								endcol = 1;
							}

							if (endpos != -1 && endpos != 0 && startpos < endpos && endpos+1 <= data.Length()-1 && data.Length() != 0)
							{
								int len = endpos-startpos;
								data.CopySubStringNoSpaces(cstr, startpos, endpos-1, limit);

								if (strlen(cstr) != 0)
								{
									doublelist.Add(atof(cstr));
									ncol++;

									startpos = endpos+1;
									emptystring = 0;
								}
								else
								{
									startpos = endpos+1;
									doublelist.Add(0);
								}
							}
							else 
							{
								end = 1;
								endcol = 1;
							}
						}
						if (!end && !emptystring) nrows++;
						if (ncol != maxcol) 
						{
							if (maxcol != 0 && ncol != 0)
							{
								error = 1;
							}

							if (ncol > maxcol) maxcol = ncol;
						}
					}



					//++++++++++++++++++++++++


					/*
					int pos2 = 0;
					int foundcols = 0;
					TArray<double> doublelist;

					while (pos2 != -1)
					{
					data.GetUntil(pos2, semicolon, data3, 0);

					int pos3 = 0;
					while (pos3 != -1)
					{
					data3.GetUntil(pos3, comma, data2, 0);
					doublelist.Add(data2.MakeDouble());
					if (pos3 != -1) data3.GoUntil(pos3, comma);
					}
					if (!foundcols) foundcols = doublelist.Length(); //take first number of columns for matrix

					data.GoUntil(pos2, semicolon);
					}
					*/

					str.GoUntil(pos, csb); //read final square bracket

					if (cols == 0 || ed.IsVariableLength()) cols = maxcol;
					else if (cols != maxcol || error) {LoadError(mystr("Matrix should contain ")+mystr(rows)+mystr("rows and ")+mystr(cols)+mystr("columns !!!")); return 0;}

					Matrix m;
					if (cols != 0) 
					{
						if (rows == 0 || ed.IsVariableLength()) rows = doublelist.Length()/cols;
						else if (doublelist.Length()/cols != rows) {LoadError(mystr("Matrix should contain ")+mystr(rows)+mystr("rows and ")+mystr(cols)+mystr("columns !!!")); return 0;}

						m.SetSize(rows, cols);
						for (int i1=1; i1 <= rows; i1++)
						{
							for (int i2=1; i2 <= cols; i2++)
							{
								int ind = (i1-1)*cols+i2;
								if (ind <= doublelist.Length())
								{
									m(i1, i2) = doublelist(ind);
								}
							}
						}
					}
					else rows = 0;

					ed.SetMatrix(m.GetMatPtr(), rows, cols, dataname.c_str());
					edc.Add(ed);

					if (oo) mbs->UO() << m << "\n";

				}
				else
				{
					//data = str.GetWord(pos, 0);
					str.GetUntil(pos, '\n', data);
					data.EraseSpaces();
					if (oo) mbs->UO() << "single data='" << data << "'\n";
					if (ed.IsBool())
					{
						ed.SetBool(data.MakeInt()); edc.Add(ed);
					} 
					else if (ed.IsInt())
					{
						ed.SetInt(data.MakeInt()); edc.Add(ed);
					} 
					else if (ed.IsDouble())
					{
						ed.SetDouble(data.MakeDouble()); edc.Add(ed);
						//if (oo) mbs->UO() << "double data='" << data.MakeDouble() << ", " << ed.GetDouble() << "'\n";
					} 
					else if (ed.IsText())
					{
						ed.SetText(data.c_str()); edc.Add(ed);
					} 
					else
					{
						//error should not happen
					}
				}
			}
			else
			{
				LoadError(mystr("Specifier '")+dataname+mystr("' is unknown"));
				str.GetUntil(pos, '\n', data); //one line unknown keywords are no problem!!!
				//return 0;
			}
		}
		else
		{
			str = str.Right(str.Length()-pos); return 1;
		}


		if (pos == -1) endstr = 1;

	}
	return 1;
}

void CEDCParser::Initialize() 
{
	//here, things can be done during initialization of the CEDCparser
	el = "\n"; //end of line
	elc = '\n';
	obc = '{';  //open brace, for sub structure
	cbc = '}';  //closing brace
	ob = mystr(obc);
	cb = mystr(cbc);
	ofc = '(';  //open function (round '(') bracket
	cfc = ')';  //close function (round ')') bracket
	of = mystr(ofc);
	cf = mystr(cfc);
	osbc = '[';  //open square bracket, for vector or matrix
	csbc = ']';  //closing square bracket
	osb = osbc;  //open square bracket, for vector or matrix
	csb = csbc;  //closing square bracket
	ds = ": ";	//double stop - not used any more
	eq = "=";		//equal sign
	semicolon = ";"; //double stop
	semicolonc = ';'; //double stop
	commac = ','; //comma
	comma = commac; //comma
	doublequote = '"'; //for strings or text
	dotc = '.'; //dot
	dot = dotc; //dot
	space = ' '; //space
	spacec = ' ';
	tabc = '\t'; //tab
	tab = '\t';
	bool_yes = "yes";
	bool_no = "no";
	commentc = '%'; //comment
	comment = commentc; //comment
	matrixsep = el; //may be changed to semicolon ...

	//intialize switches and others:
	//oo = 0; //$ DR 2015-07-01 removed for simplification of code
	line_cnt = 0;
	total_lines = 0;

	version_check_performed = 0;
	
	InitializeCommands();
}

//read spaces, then comment sign, parse comment string into Tooltiptext of ed
//return 1 if success, 0 if fail
int CEDCParser::ReadComment(mystr& str, int& pos, ElementData& ed, const mystr& dataname)
{
	mystr data;
	int rv = str.GetUntilEOL(pos, commentc, data);
	if (rv == 2) //comment found
	{
		//mbs->UO() << "comment='" << data << "'+";
		if (str.PosPeek(pos) != commentc) 
		{
			mystr name = dataname;
			mbs->UO(UO_LVL_warn) << "Problem with comment in entry " << name <<". Check Textfile!\n";
			rv = 0;
		}

		// comment --> ToolTipText (double)
		mystr dummycomment;
		str.GetUntilEOL(pos, (char)0, dummycomment);
		mystr comment = dummycomment.SubString(1, dummycomment.Length()-1);
		ed.SetToolTipText(comment);
		//mbs->UO() << "'" << comment << "'\n";
	}

	return rv;
}

int CEDCParser::ReadBoolIntDoubleExpression(mystr& text, int& pos, ElementData& ed, const mystr& elemname)
{
	int rv = 1;
	text.EraseSpaces();
	if (PrintOutput()) mbs->UO() << "single text='" << text << "'\n";
	if (text[text.Length()-1] == ';') //accept ";" at end of statement
	{
		text.EraseChar(text.Length());
	}

	if(text.IsValidNumber())
	{
		ed.SetDouble(text.MakeDouble(), elemname); 
	}
	else // "expression"
	{
		if (PrintOutput()) mbs->UO() << "text='" << text << "' is no valid number\n";
		//bool variable
		if (bool_no.Compare(text))
		{
			ed.SetBool(0, elemname);
		}
		else if (bool_yes.Compare(text))
		{
			ed.SetBool(1, elemname);
		}
		else 
		{
			// DR 2012-08-02
			// AD 2013-10-16: collected calls to parse command
//AD: new procedure for commands
			mystr word = text; // "left hand side expression", not entire "str" and "pos" as for right hand side call
			int pos_tmp=0;

			mystr dataname = word.GetIdentifier(pos_tmp, 0); //name of data structure or data variable
			int com_id = IdentifyAndExecuteCommand(dataname, word, pos_tmp, ed, 1/*right hand side call*/);
			ed.SetDataName(elemname);

			if (com_id == -1)
			{ 
				rv = 0; // Parsing error occurred
			}
			else if (com_id > 0)
			{
				//rv = 1; // command executed
			}
			else //if (com_id == 0)
// NOT A COMMAND
			{
//!AD: 2012-12-11 assign entire EDCs: [
				// check if text is an EDC
// this would be already available in 
// CParseObj* CMBSParser::GetVariable(mystr& str)
// ElementDataContainer* CMBSParser::GetElementDataContainers(int i)
				int found_as_edc = 0;
				int found_as_entry = 0;
				////while (i <= NElementDataContainers() && ! found)
				////{
				////	ElementDataContainer* edc = GetElementDataContainers(i);
				
				//ElementDataContainer* edc_existing = this->GetMBS()->GetModelDataContainer();  // GLOBAL
				//if (edc_existing && edc_existing->TreeFind(text))
				//{
				//	ElementData* ed_other = edc_existing->TreeFind(text);
				//	if (ed_other->IsEDC())
				//	{
				//		found_as_edc = 1;
				//		ElementDataContainer* edc_new = new ElementDataContainer();
				//		edc_new->TreeReplaceEDCDataWith(ed_other->GetEDC());  // use TreeReplaceEDCDataWith to add the entries not clear the old ones
				//		ed.SetEDC(edc_new,elemname);
				//	}
				//}

				if (!(found_as_edc || found_as_entry))	
				{
			// check if RightHandSide exists as entry in LOCAL EDC first
					ElementDataContainer* localEDC = this->GetMBSParser()->GetLocalEDC();  // local context of the existing EDC
					if(localEDC && localEDC->TreeFind(text))
					{
						ElementData* ed_other = localEDC->TreeFind(text);
						if(ed_other->IsEDC())
						{
							found_as_edc = 1;
							ElementDataContainer* edc_new = new ElementDataContainer();
							edc_new->TreeReplaceEDCDataWith(ed_other->GetEDC());  // use TreeReplaceEDCDataWith to add the entries not clear the old ones
							ed.SetEDC(edc_new,elemname);
						}
						else
						{
							found_as_entry = 1;
							ed.CopyFrom(*ed_other);
							ed.SetDataName(elemname);
						}
					}
				}
				if (!(found_as_edc || found_as_entry))
				{
			// check if RightHandSide exists as entry in GLOBAL EDC first
					ElementDataContainer* globalEDC = this->GetMBSParser()->GetLocalEDC2();  // ROOT of the existing EDC 
					if(globalEDC && globalEDC->TreeFind(text))
					{
						ElementData* ed_other = globalEDC->TreeFind(text);
						if(ed_other->IsEDC())
						{
							found_as_edc = 1;
							ElementDataContainer* edc_new = new ElementDataContainer();
							edc_new->TreeReplaceEDCDataWith(ed_other->GetEDC());  // use TreeReplaceEDCDataWith to add the entries not clear the old ones
							ed.SetEDC(edc_new,elemname);
						}
						else
						{
							found_as_entry = 1;
							ed.CopyFrom(*ed_other);
							ed.SetDataName(elemname);
						}
					}
				}

				////	i++;
				////}
//!AD: ]
				if (!(found_as_edc || found_as_entry))
				{
					int errorflag;
					//////JG 2011-03-24
					//////Expression parser, using MBS_EDCs (edc_modeldata, mbs_edc_options, mbs_edc_global_variables) as variables
					////double value = ExpressionToDouble(mbs, text, errorflag, line_cnt);
					////	if (errorflag)
					////{
					////	ed.SetDouble(0, elemname); 
					////	rv = 0;
					////}
					////else
					////{
					////	ed.SetDouble(value, elemname); 
					////	rv = 1;
					////}	

//AD 2012-12-10
					//Expression parser extended to Vector and Matrix variables
					double d_value = 0;
					Vector v_value(0.);
					Matrix m_value(0.);
					int2 rv_rowscols = mbsParser->EvaluateExpression(text, errorflag, d_value, v_value, m_value, line_cnt);

					if (rv_rowscols == int2(1,1)) // scalar
					{
						if (errorflag) { ed.SetDouble(0, elemname); rv = 0; }
						else { ed.SetDouble(d_value, elemname); rv = 1; }			
					}
					else if (rv_rowscols(1)==1 || rv_rowscols(2)==1) // vector
					{
						if (errorflag) { ed.SetVector(v_value.GetVecPtr(), 1, elemname); rv = 0; }
						else { ed.SetVector(v_value.GetVecPtr(), v_value.Length(), elemname); rv = 1;}
					}
					else // matrix
					{
						if (errorflag) { ed.SetMatrix(m_value.GetMatPtr(), 1, 1, elemname); rv = 0; }
						else { ed.SetMatrix(m_value.GetMatPtr(), m_value.Getrows(), m_value.Getcols(), elemname); rv = 1; }
					}
				}
			}
		}
	}
	return rv;
}

//read a matrix or vector into ElementData
int CEDCParser::ReadMatrixVector(const mystr& text, ElementData& ed, const mystr& elemname, const mystr& dataname)
{
	int rv = 1;

	// col seperators : " "     ","
	// row seperators : "\n"    ";"
	TArray<double> ta_vec;
	mystr string_value;       // string containing the next value
	mystr string_entirerow;   // string containing the processed row 
	int pos_value = 0;        // position of next value to read
	int pos_in_string = 0;    // position to which the string is read
	int pos_in_row = 0;				// position to which the substring (row) is read
	int found_rowsep = 0;			// position of next row separator
	int found_colsep = 0;			// position of next column separator
	int rows=0; int cols=0; 	// number of rows, columns of the matrix 

	// check which separation chars are used
	text.ReadLeadingSpacesAndCountLines(pos_in_string, line_cnt);  // !skip spaces at the beginning of text to prevent pos_space==0 (for the case that ' ' is column separator)
	int pos_comma = text.Find(pos_in_string,commac);
	int pos_space = text.Find(pos_in_string,spacec);
	int pos_tab = text.Find(pos_in_string,tabc);
	int pos_semi = text.Find(pos_in_string,semicolonc);
	int pos_newline = text.Find(pos_in_string,elc);

	char colsep, rowsep;
	if ((pos_comma==-1)&&(pos_space==-1)) { colsep = commac; } // separation will not be found since this is a column vector
	if ((pos_comma > 0)&&(pos_space==-1)) { colsep = commac; } // separation with ','
	if ((pos_comma==-1)&&((pos_space > 0) || (pos_tab > 0))) { colsep = spacec; } // separation with ' '
	if ((pos_comma > 0)&&((pos_space > 0) || (pos_tab > 0))) { colsep = commac; } //mbs->UO(UO_LVL_warn) << "String2TreeEDC found column separators \",\" and \" \" in entry " << dataname <<". Check Textfile!\n"; }

	if ((pos_semi==-1)&&(pos_newline==-1)) { rowsep = semicolonc; } // separation will not be found since this is a row vector
	if ((pos_semi > 0)&&(pos_newline==-1)) { rowsep = semicolonc; } // separation with ';'
	if ((pos_semi==-1)&&(pos_newline > 0)) { rowsep = elc; }        // separation with '\n'
	if ((pos_semi > 0)&&(pos_newline > 0)) { rowsep = semicolonc; } // mbs->UO(UO_LVL_warn) << "String2TreeEDC found row separators \";\" and \"\\n\" in entry " << dataname <<". Check Textfile!\n"; }      

	while(pos_in_string != -1) // get -1 at end of text
	{

		rows++;
		found_rowsep = text.GetUntil(pos_in_string, rowsep, string_entirerow);
		string_entirerow.Replace(tab, space);

		int ccols = 0;	
		pos_in_row = 0;
		string_entirerow.EraseSpacesHeadTail();
		string_entirerow.ReadLeadingSpacesAndCountLines(pos_in_row, line_cnt); // !skip spaces at the beginning of a new row (for the case that ' ' is column separator)

		while(pos_in_row != -1 && rv) // get -1 at end of row
		{
			ccols++;
			string_entirerow.ReadLeadingSpaces(pos_in_row);
			found_colsep = string_entirerow.GetUntil(pos_in_row, colsep, string_value);

			if(string_value.IsValidNumber())
			{
				ta_vec.Add(string_value.MakeDouble());
			}
			else
			{				

				int errorflag;
				double value = mbsParser->ExpressionToDouble(string_value, errorflag, line_cnt);

				if (errorflag)
				{
					ta_vec.Add(0);
					rv = 0;
				}
				else
				{
					ta_vec.Add(value);
					rv = 1;
				}
			}

			if(found_colsep) pos_in_row++; // next value
			else // check consistency
			{
				if(rows == 1) 
				{
					cols = ccols;
				}
				else 
				{
					mystr name = dataname;
					if (cols!=ccols) 
					{
						mbs->UO(UO_LVL_warn) << "String2TreeEDC found mismatch in number of columns in entry " << name <<". Check Textfile!\n"; 
						rv = 0;
					}
				} 
			}
		}
		if(found_rowsep) {pos_in_string++;} // next row

	}
	Vector vec(ta_vec);
	if (rows <= 1) //if rows == 0, then assume vector
	{
		ed.SetVector(vec.GetVecPtr(), vec.Length(), elemname.c_str());  // vector
	}
	else 
	{
		ed.SetMatrix(vec.GetVecPtr(), rows, cols, elemname.c_str()); // matrix
	}

	return rv;
}


//####################################################################
//b: STANDARD VERSION of ParseAndExecuteString (f.k.a. String2TreeEDC)
// new version for parsing a string into an ElementDataContainer inclusive ToolTipText
// 6.9.10 added parameter treestr: name of tree structure for recursive calls (parent history...)
// 11.5.11 RL, AD: String2TreeEDC calls following recursive function "String2TreeEDC_recursive"
//                 Please use only "String2TreeEDC"!!!
// **.10.13 AD: functionality now only in ParseAndExecuteString
//              !! removed all calls to String2TreeEDC, removed implementation !!
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//definition of EDC-structure:
// '%' | [identifier												 //comment or identifier appears
// if identifier is not a function:
//   ['{' | '=' | '(' | '%']                 //after identifier, either a structure opens '{' or a variable is assigned '=', a function is called '(' or a comment appears '%'
//   '{' ==> ['{ recursive_EDC }']
//   '=' ==> ['"' | '[...]' | [+-.1234567890] | mathobject/variable/function                        //read assignement: either string starts '"', vector/matrix '[...]', number, sin(2*pi*x)
//   ['%']																	 //comment might appear
//   ']'                                     //end of identifier structure
// if identifier is a function/command
//   '(function_parameters)'                 //read function parameters in paranthesis ()
//   '%'																		 //comment might appear
//   '{' ==> ['{ recursive_EDC }']           //if function supports a statement block (e.g. if/for/while/etc), then recursively read edc and execute function e.g. for-loop
// %
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// AD 2013-09-24: these functions parses a string ans applies it ENTRY BY ENTRY to the target element data container
// this function should replace the function StringToTreeEDC for the Commands "include", "for", etc...
// code is basically copied from function StringToTreeEDC
int CEDCParser::ParseAndExecuteString(mystr str, ElementDataContainer& edc)
{
// statistics on the string
	line_cnt = 0; //line counter in EDC
	total_lines = str.CountLines();
	mbs->UO(UO_LVL_dbg1) << "EDC total lines=" << total_lines << "\n";

// remember previous local EDC
	ElementDataContainer* prevlocaledc = mbsParser->GetLocalEDC();

// Set local variable context for the parser
	mbsParser->SetLocalEDC(&edc); //parsed (root-)edc is stored in local edc 

// cal the recursive part - the actual parsing
	int rv = ParseAndExecuteString_recursive(str, edc); 
	
// set local EDC back to previous 
	mbsParser->SetLocalEDC(prevlocaledc);

	return rv;
}

int CEDCParser::ParseAndExecuteString_recursive(mystr str, ElementDataContainer& edc)
{
	int rv = 1; //success
	int endstr = 0;
	int pos = 0;
	mystr dataname; // full name, may contain (sub-)tree
	mystr treename; // tree only
	mystr elemname; // varname only
	char ch;
	ElementData ed, return_value;
	mystr dummycomment; //dummy comment string
	
	
	int is_first_command = 1;
	TArray<ElementDataContainer*> parameter_EDCs;

// main loop
	while (!endstr && rv != 0 && pos != -1)
	{
		ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);
		
// skip lines that begin with comments  (e.g. file header, ...)
		while (ch == commentc && pos != -1)                               
		{
			str.GetUntilEOL(pos, (char)0, dummycomment);
			ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);
		}

		int command_found = 0;
		if (pos == -1) //regular end of structure!
		{
			endstr = 1;
		}

//$AD 10-2013: new procedure for commands
		else
		{
			dataname = str.GetIdentifier(pos, 0); //name of data structure or data variable
			int com_id = IdentifyAndExecuteCommand(dataname, str, pos, ed, 0/*left hand side call*/);

			if (com_id == -1)
			{ 
				rv = 0; // Parsing error occurred
			}
			else if (com_id > 0)
			{
				command_found = 1;
				//rv = 1; // command executed
			}
			else //if (com_id == 0)
// NOT A COMMAND		
			{
				// split into tree name and element name for assignment without full tree structure
				edc.SplitIntoTreeAndElementName(dataname,treename,elemname); 
			}
		}
//$AD: end new procedure for commands

		if (pos == -1) //regular end of structure!
		{
			endstr = 1;
		}
		else if (command_found)
		{
			//do nothing, command already processed
		}

// BEGIN PARSING BEYOND DATANAME
		else
		{
			if (PrintOutput()) mbs->UO() << "Dataname='" << dataname << "'\n";

			ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);

			while (ch == commentc && pos != -1)
			{
				str.GetUntilEOL(pos, (char)0, dummycomment);
				ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);
			}

// FOUND "=":  ASSIGNMENT of various types
			if (ch == eq)
			{
				//read data:
				pos++; //go after eq sign
				ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);
				
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				if (ch == osb) //vector/matrix
				{
					mystr text = str.GetStringInBrackets(pos, osbc, csbc);					
					ReadMatrixVector(text, ed, elemname, dataname);
					ReadComment(str, pos, ed, dataname);
					edc.TreeSet(treename,ed);//$JG2012-01-27: old if/else cases replaced //$ RL 2011-6-28: 
				}
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				//string expected
				else if (ch == doublequote)
				{
					mystr text = str.GetStringInBrackets(pos, doublequote, doublequote);
					ed.SetText(text, elemname);
					ReadComment(str, pos, ed, dataname);
					edc.TreeSet(treename,ed);//$JG2012-01-27: old if/else cases replaced //$ RL 2011-6-28: 
				}
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				//int, double, bool or regular expression
				else //must be double or integer number or "expression"
				{
					mystr text;
					int rv2 = str.GetUntilEOL(pos, commentc, text);
					if (rv2 == 2) {pos--;} //go back one character in order to find comment character in ReadComment(...) function hereafter
// EVALUATE RIGHT HAND SIDE OF THE ASSIGNMENT ( could include COMMANDS )
					rv = ReadBoolIntDoubleExpression(text, pos, ed, elemname);

					if ( pos!=-1 )
					{
						ReadComment(str, pos, ed, dataname);
					}
// ADD OR REPLACE ENTRYs					
					edc.TreeSet(treename,ed); 
				}
			//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// LINE IS PARSED - APPLY TO THE MODEL EDC
			}
			else if (ch == ob)
// FOUND "{": SUB TREE - RECURSION 
			{
				//read sub item
				mystr substr = str.GetStringInBrackets(pos, obc, cbc);
				if (pos != -1)
				{
					// check if tree already exists
					int i = edc.Find(elemname);
					if ((i != 0)&&(edc.Get(i).IsEDC())) // already exists as sub-edc
					{
						mbsParser->SetLocalEDC(edc.Get(i).GetEDC()); //$ RL 2011-5-11: //$ AD 2011-5-11: set local edc.
						rv = ParseAndExecuteString_recursive(substr, *(edc.Get(i).GetEDC()));
						mbsParser->SetLocalEDC(&edc); //$ RL 2011-5-11: //$ AD 2011-5-11: set local context edc back to value before recursion
					}
					else
					{
						ElementDataContainer* edc_sub = new ElementDataContainer(); //this is deleted when new EDC is set or EDC is deleted (old entry is always deleted)
						ed.SetBool(0,"dummy");
						int n = edc.Add(ed);
						edc.Get(n).SetEDCno_copy(edc_sub, elemname);
						mbsParser->SetLocalEDC(edc_sub); //$ RL 2011-5-11: //$ AD 2011-5-11: store sub-edc in local parser edc.
						rv = ParseAndExecuteString_recursive(substr, *edc_sub);
						mbsParser->SetLocalEDC(&edc); //$ RL 2011-5-11: //$ AD 2011-5-11: set local context edc back to value before recursion
					}
				}
				else
				{
					mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: could not read item '" << dataname << "' !\n";
					return 0; //end of file reached, not valid!
				}
			}

// ??? ADD LEFT SIDE LIST ACCESS RIGHT HERE ???
// Allow ASSIGNMENT to A SINGLE ENTRY of a vector or matrix
			else if (ch == osb)
// FOUND "[": LEFT HAND SIDE LIST ACCESS
			{
				mystr indices = str.GetStringInBrackets(pos, osbc, csbc);
			
				// check 1: variable must already exists 
				ElementData* ed_existing = NULL;
				ElementDataContainer* localEDC = this->GetMBSParser()->GetLocalEDC();  // local context of the existing EDC
				ElementDataContainer* globalEDC = this->GetMBSParser()->GetLocalEDC2();  // ROOT of the existing EDC 
				if(localEDC && localEDC->TreeFind(dataname))
				{
					ed_existing = localEDC->TreeFind(dataname);
				}
				else if(globalEDC && globalEDC->TreeFind(dataname))
				{
					ed_existing = globalEDC->TreeFind(dataname);
				}
				if(ed_existing == NULL)
				{
					mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: trying to accesss a component of variable '" << dataname << "' which is not defined in this context!\n";
					return 0; 
				}
				
				// check 2: indices must not exceed limits
				mystr fullstr = dataname + "[" + indices + "]";
				int i1=0; int i2=0;
				mystr ci1,ci2;
				int nr_indices = fullstr.Trim_And_Get_Indices(ci1,ci2);  // use string version to allow variables as index
				if(ci1.IsValidNumber())
				{
					i1 = ci1.MakeInt();
				}
				else
				{
					rv = ReadBoolIntDoubleExpression(ci1, pos, ed, elemname); // evaluate variable for index 
					i1 = (int)ed.GetDouble();
				}
				if(ci2.IsValidNumber())
				{
					i2 = ci2.MakeInt();
				}
				else
				{
					rv = ReadBoolIntDoubleExpression(ci2, pos, ed, elemname);
					i2 = (int)ed.GetDouble();
				}


				if (ed_existing->IsVector() && nr_indices == 1)
				{
				// existing vector with single index
					if(i1>0 && i1<=ed_existing->GetVectorLen())
					{
						// valid vector
					}
					else
					{
						mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: index out of bound when acessing component of vector-variable '" << dataname << "' !\n";
						return 0; 
					}
				}
				else if (ed_existing->IsMatrix() && nr_indices == 2)
				{
				// existing vector with single index
					if(i1>0 && i1<=ed_existing->GetMatrixRows() && i2>0 && i2<=ed_existing->GetMatrixCols())
					{
						// valid matrix
					}
					else
					{
						mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: index out of bound when accessing component of matrix-variable '" << dataname << "' !\n";
						return 0; 
					}
				}
				else 
				{
					mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: number of dimensions wrong for variable '" << dataname << "' !\n";
					return 0; 
				}
	 
				// check 3: followed by a '='
				ch = str.ReadLeadingSpacesAndCountLines(pos, line_cnt);
				if (ch == eq)
				{
					// read right hand side into double
					pos++; //go after eq sign
					mystr text;
					int rv2 = str.GetUntilEOL(pos, commentc, text);
					rv = ReadBoolIntDoubleExpression(text, pos, ed, elemname);
				}
				else 
				{
					mbs->UO(UO_LVL_err) << "ParseAndExecuteString_recursive: left hand side access of variable '" << dataname << "' must be followed by a '=' for assignment!\n";
					return 0; 
				}
				
				// assign value to entry
				//double val = ed.GetDouble();
				double val = ed.GetSingleValAsDouble();
				if (nr_indices==1)
				{
					ed_existing->SetVectorVal(i1,val);
				}
				else if (nr_indices==2)
				{
					ed_existing->SetMatrixVal(i1,i2,val);
				}
				else
				{
					// already stated as error
				}

			}

// DATANAME WAS FOLLOWED BY INVALID CHARACTER
			else
			{
				mbs->UO(UO_LVL_warn) << "did not expect '" << mystr(ch) << "' at line " << line_cnt+1 << " !\n";
				pos++;
				return 0; //end of file reached, not valid!
			}


		}
	} // end main loop
	return rv;
}

//e: STANDARD VERSION of ParseAndExecuteString
//############################################


// AD 2013-10-16: single call for commands - collect code from String2TreeEDC and ReadBoolIntDoubleExpression - returns "com_id" (including 0 for no command) or "-1" on error 
int CEDCParser::IdentifyAndExecuteCommand(mystr& dataname, mystr& str, int& pos, ElementData& return_value, int flag_RHS)
{
	int n_params, option, has_return_value;
 	int com_id = IdentifyCommand(dataname, n_params, has_return_value, option, flag_RHS);

	if (com_id > 0) // command found
	{
		if(!version_check_performed)
		{
			ElementDataContainer* edc_local = mbsParser->GetLocalEDC();
			HotintDataFileVersionCheck(version_check_performed, *edc_local);
		}

		TArray<ElementDataContainer*> parameter_EDCs;
		int parse_success = ParseCommandParameter(parameter_EDCs, dataname, n_params, str, pos, option);

		if (parse_success)
		{
// COMMAND IS IMMEDIATELY EXECUTED
			int rv = ExecuteCommand(com_id, parameter_EDCs, return_value, option);
			if(rv == 0)
			{
				com_id = -1;    // error occurred during execute command
			}
		}
		else
		{
			com_id = -1;      // error occurred during parsing of the command parameters
		}

		for(int np=1;np<=parameter_EDCs.Length();np++)
		{
			delete parameter_EDCs(np); // delete edc's
		}
	}
	return com_id;
}

// AD 2013-10-16: identifies command type and checks restrictions - returns "com_id" (including 0 for no command) or "-1" on error
int CEDCParser::IdentifyCommand(mystr& dataname, int& n_params, int& has_return_value, int& option, int flag_RHS )
{
	mystr handlename;				// treename
	mystr commandname;			// elementname
	ElementDataContainer* edc_local = mbsParser->GetLocalEDC();
	edc_local->SplitIntoTreeAndElementName(dataname, handlename, commandname);

// find registered command
  int com_id = GetCommandType(commandname, n_params, has_return_value, option);

	if(com_id > 0) 
// perform consistency checks
	{
// check 1: some commands are left hand side only (IF, FOR, ..)
		if ((option & TCO_LeftHandSideOnly) && flag_RHS)
		{
			EDCError(mystr("Right hand side call of Function '")+commandname+mystr("' is not allowed!"));
			return -1;
		}

// check 2: commands requiring a handle
		else if (option & TCO_RequiresHandle)
		{
	// call of a handle command is "HANDLE.COMMAND" so HANDLE must be registered in the current EDC
			if (handlename.Length())
			{
				int i = edc_local	->Find(handlename);
				if ((i != 0) && edc_local->Get(i).IsEDC())
				{
	// check if the handle is indeed a handle-object
					ElementDataContainer* edc_handle = edc_local->Get(i).GetEDC();
					int j = edc_handle->Find("isHandle");
					if (j==0)
					{
						EDCError(mystr("Handle '")+handlename+mystr("' for command '")+commandname+mystr("' is not a valid handle '"));
						return -1;
					}
					else 
					{
						j = edc_handle->Find("Functions"); // this container name is set by GenerateNewMesh
						if ((j!=0) && edc_handle->Get(j).IsEDC())
						{
	// check if the command is registered to the handle
							ElementDataContainer* edc_registered = edc_handle->Get(j).GetEDC();
							int k = edc_registered->Find(commandname);
							if (k!=0)
							{
								;
							}
							else
							{
								EDCError(mystr("Function '")+commandname+mystr("' is not registered for Handle '")+handlename);
								return -1;
							}
						}
						else
						{
							EDCError(mystr("Handle '")+handlename+mystr("' has no allowed functions'"));
							return -1;
						}
					}
				}
			}
		}
// check 2b: when a handle is present and 
		else //(if !(option & TCO_RequiresHandle))
		{
			if (handlename.Length())
			{
				int i = edc_local	->Find(handlename);
				if ((i != 0) && edc_local->Get(i).IsEDC())
				{
	// check if the handle is indeed a handle-object
					ElementDataContainer* edc_handle = edc_local->Get(i).GetEDC();
					int j = edc_handle->Find("isHandle");
					if (j !=0)
					{
						EDCError(mystr("Function '")+commandname+mystr("' does not require a handle but is associated with one ('")+handlename+mystr("')"));
						return -1;
					}
				}
			}
		}
	}
	return com_id;
}

// AD 2013-10-16: parses parameters into EDCs - returns "1" on success and "0" on error
int CEDCParser::ParseCommandParameter(TArray<ElementDataContainer*>& parameter_EDCs, mystr& dataname, int n_params, mystr& str, int& pos, int option)
{
	int rv = 1;
	parameter_EDCs.SetLen(0);

//read parameter text
	mystr param_text = str.GetStringInBrackets(pos, ofc, cfc);
	if (param_text == mystr("") && n_params != 0)
	{
		EDCError(mystr("Expected ")+mystr(n_params)+mystr(" function parameters but received zero parameters in function '")+dataname+mystr("'"));
		rv = 0;
	}
	else 
	{	
		int prv = GetParameterEDC(param_text, parameter_EDCs); // fill parameter_EDCs with NEW edc's
		if (!prv)
		{
			EDCError(mystr("Function parameters '")+param_text+("' of function '")+dataname+mystr("' are not valid or incomplete"));
			rv = 0;
		}
		if (parameter_EDCs.Length() < n_params)
		{
			EDCError(mystr("Received ")+mystr(parameter_EDCs.Length())+(" function parameters, but function ")+dataname+(" should have ")+mystr(n_params)+mystr(" parameters"));
			rv = 0;
		}
		if (parameter_EDCs.Length() > n_params)	//$ DR 2014-09-25 added this functionality
		{
			EDCWarning(mystr("Received ")+mystr(parameter_EDCs.Length())+(" function parameters, but function ")+dataname+(" has only ")+mystr(n_params)+mystr(" parameters as minimum. Maybe you are using an optional argument, then please ignore this warning."));
		}
		else
		{
// additional parameter EDC contains (code-)container		
			if (option & TCO_FollowedByContainer)
			{
				while (str[pos] != obc) pos++;
				mystr codeincontainer = str.GetStringInBrackets(pos,obc,cbc);
				ElementData code;
				code.SetText(codeincontainer,"code");
				ElementDataContainer* codecontainer = new ElementDataContainer;
				codecontainer->Add(code);
				parameter_EDCs.Add(codecontainer);
			}
// additional parameter EDC contains the handle
			if (option & TCO_RequiresHandle)
			{
				ElementDataContainer* local = mbsParser->GetLocalEDC();
				mystr handlename, commandname;
				local->SplitIntoTreeAndElementName(dataname, handlename, commandname);

				//validity check only here
				int j = local->Find(handlename);
				if(j != 0 && local->Get(j).IsEDC())
				{
					//ElementDataContainer* edc_handle = local->Get(j).GetEDC();
					ElementData handle;
					handle.SetText(handlename,"handle");
					ElementDataContainer* handlecontainer = new ElementDataContainer;
					handlecontainer->Add(handle);
					parameter_EDCs.Add(handlecontainer);
				}
				else
				{
					EDCError(mystr("Handle '")+handlename+mystr("' could not be linked to Command '")+commandname+mystr("'. call only in local context!"));
					rv = 0;
				}
			}
		}
	}
	return rv;
}

//execute a command with type_id, a list of EDCs for command parameters and an option
int CEDCParser::ExecuteCommand(int command_type_id, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	return commands(command_type_id)->ExecuteCommand(GetMBS(), this, parameter_EDCs, return_value, option);
}

// AD 2013-10-16: perform a check for the datafile version (once)	
int CEDCParser::HotintDataFileVersionCheck(int& version_check_performed, ElementDataContainer& edc)
{
	version_check_performed = 1;
	mystr tmp_str = edc.TreeGetString("HOTINT_data_file_version");
	if(tmp_str.Length())
	{
		hid_version_number = HotintVersionInfo(tmp_str);		// version in the input file
		HotintVersionInfo ver = mbs->GetHotintVersion();	
		int ver_log = ver.GetLogNumber();
		int ver_file_log = hid_version_number.GetLogNumber();
		if((ver > hid_version_number) || ((ver == hid_version_number) && (ver_log > ver_file_log)))
		{
			mbs->UO(UO_LVL_warn) << "WARNING: The input file was created with an old version of HOTINT. This may be the reason for errors. The current version of HOTINT is '" << mbs->GetHotintVersion().GetString() << "'. To avoid this warning, check the functionality of the input file and correct the HOTINT_data_file_version in the input file afterwards.\n";
		}
	}
	else
	{
		hid_version_number = HotintVersionInfo(0,0,0);
		mbs->UO(UO_LVL_warn) << "WARNING: The input file does not contain the variable HOTINT_data_file_version which indicates, that it has probably been created with an old version of HOTINT. This may be the reason for errors. The current version of HOTINT is '" << mbs->GetHotintVersion().GetString() << "'. To avoid this warning, check the functionality of the input file and set the variable HOTINT_data_file_version in the input file right at the beginning.\n";
	}
	return 1;
}

// DR 2013-01-14: this function evaluates a command of the parser, using the parameter_EDC as additional input data
// AD 2013-10-16: remark: can be called from .cpp model for a) testing new script function, b) use existing script routines
int CEDCParser::ExecuteParserCommand(mystr & command, TArray<ElementDataContainer*>& parameter_EDCs, ElementData& return_value, int option)
{
	int pos = 0;
	mystr elemname = "element_name";
	ElementDataContainer *edc;
	ElementData ed;

	if(parameter_EDCs.Length())
	{
		edc = parameter_EDCs(1);
		ed.SetEDC(edc, "parameter_EDC");
		mbsParser->SetLocalEDC(edc);
	}
	else
	{
		ed.SetBool(1,"empty");
	}

	int rv = ReadBoolIntDoubleExpression(command, pos, ed, elemname);

	return rv;
}

//$ DR 2013-07-04:
// checks if filename has relative path, and if yes, converts it into an absolute one
// relative means relative to last_file_name and this is the file in which e.g. the command "Include" is used
//$ SW 2014-7-1: this function had problems if the last_file_name was also relative and if the path included / rather than \ 
int CEDCParser::MakeFilenameAbsolute(mystr& filename)
{
	int rv = 0; //1=success
	int pos = 1;
	int isRelativePath = filename.PosPeek(pos) != ':';
	if(isRelativePath)
	{		
		mystr rel_path(""); 
		mystr last_file_name = mbs->GetModelDataContainer()->TreeGetString("last_file_name");
		//mystr last_file_name = mbs->GetOptions()->GeneralOptions()->PathsHotintInputDataPath();

		if(last_file_name[1] != ':') //last file name is relative itself -> it is relative to the current working directory
		{
			char last_file_name_char[MAX_PATH];
			_fullpath(last_file_name_char, last_file_name.c_str(), MAX_PATH);
			if (last_file_name_char == NULL) //fullpath function did not succeed
			{
				rv = 0;
				return rv;
			}
			last_file_name = mystr(last_file_name_char);
		}

		int until;
		for(until = last_file_name.Length()-1;until>=0;until--)
		{
			if(last_file_name.PosPeek(until) == '\\' || last_file_name.PosPeek(until) == '/')
				break; // until <=> last backslash
		}
		for(int i=0;i<=until;i++)
		{
			rel_path += last_file_name.PosPeek(i); // path until last backslash
		}
		
		// remove ..\ 
		int pos=rel_path.Length()-1;
		if(rel_path.PosPeek(pos)=='\\' || rel_path.PosPeek(pos)=='/')
		{
			rel_path.EraseChar(pos+1);
		}

		int replaced = filename.Replace(mystr("..\\"), mystr("")); //find string searchstr, replace with string replacestr; this is done repeatedly; the return value counts the number of replacements
		replaced += filename.Replace(mystr("../"), mystr(""));

		for(int pos=rel_path.Length();pos>0;pos--)
		{
			if(!replaced){break;}
			int prev = pos-1;
			if(rel_path.PosPeek(prev)=='\\' || rel_path.PosPeek(prev)=='/')
			{
				replaced --;
			}
			rel_path.EraseChar(pos);
		}
		filename = rel_path + mystr("\\") + filename;
		rv = MakeFilenameAbsolute(filename);	// recursive check
	}
	else
	{
		rv = 1;	// no change --> success
	}
	if(rv<1)
	{
		//edcParser->EDCError(mystr("Error happened during CEDCParser::MakeFilenameAbsolute ");
	}
	return rv;
}