//#***************************************************************************************
//#
//# filename:             mesh.cpp
//#
//# author:               Dorninger Alexander
//#
//# generated:						September 2014
//# description:          implementation file for mesh_aux
//#                       contains tools for meshing - 
//# remarks:						  
//#
//# Copyright (c) 2003-2018 Johannes Gerstmayr, Linz Center of Mechatronics GmbH, Austrian
//# Center of Competence in Mechatronics GmbH, Institute of Technical Mechanics at the 
//# Johannes Kepler Universitaet Linz, Austria. All rights reserved.
//#
//# This file is part of HotInt.
//# HotInt is free software: you can redistribute it and/or modify it under the terms of 
//# the HOTINT license. See folder 'licenses' for more details.
//#
//# bug reports are welcome!!!
//# WWW:		www.hotint.org
//# email:	bug_reports@hotint.org or support@hotint.org
//#***************************************************************************************
#include "mesh.h"
#include "myfile.h"
#include <sstream>
#include <fstream>
#include <map>
#include <vector>


Vector3D MeshElement::InterpolateVector3D(Vector3D ploc, const TArray<Vector3D>& nodalvalues)
{
	Vector3D retval(0.,0.,0.);
	for(int i=1; i<= nodalvalues.Length(); i++)
	{
		retval = retval + GetS0(ploc,i)*nodalvalues(i);
	}
	return retval;
}
double MeshElement::InterpolateScalar(Vector3D ploc, TArray<double>& nodalvalues)
{
	double retval = 0.;
	for(int i=1; i<= nodalvalues.Length(); i++)
	{
		retval = retval + GetS0(ploc,i)*nodalvalues(i);
	}
	return retval;
}

Vector3D MeshElement::GetLocPosOfGlobalPos(Vector3D pglob, TArray<Vector3D>& nodalposition)
{
	// newton parameter
	double eps = 1.e-7;
	int niter = 3;

	Vector3D ploc(0.),ptmp;
	Matrix3D jac, jacinv;

	for (int i = 1; i <= niter; ++i)
	{
		ptmp = InterpolateVector3D(ploc,nodalposition);
		jac = GetJacobi(ploc,nodalposition);
		jac.GetInverse(jacinv);
		ploc = ploc - jacinv*(ptmp - pglob);
	}
	return ploc;
}
Matrix3D MeshElement::GetJacobi(Vector3D ploc,TArray<Vector3D>& nodalposition)
{
	Matrix dsmat = GetDSMatrix0(ploc);
	Matrix3D jac(0.,0.,0., 0.,0.,0., 0.,0.,0.);
	for(int j=1; j<=Dim(); j++)
	{
		for(int k=1; k<=NNodes(); k++)
		{
			Vector3D npos = nodalposition(k);
			for(int i=1; i<=Dim(); i++)
			{
				jac(i,j) += dsmat(j,k)*npos(i);
			}
		}
	}
	if (Dim()==2)
	{
		jac(3,3) = 1.;
	}

	return jac;
}
int MeshElement::ContainsNode(int globalnode)
{
	for(int nidx=1; nidx<=NNodes(); nidx++)
	{
		if(nodes(nidx)==globalnode) { return nidx; }
	}
	return 0;
}
int MeshElement::IsSameBasicType(MeshElement& other)
{
	if(IsMeshHex() && other.IsMeshHex()) return 1;
	if(IsMeshTet() && other.IsMeshTet()) return 1;
	if(IsMeshPri() && other.IsMeshPri()) return 1;
	if(IsMeshPyr() && other.IsMeshPyr()) return 1;
	if(IsMeshQuad() && other.IsMeshQuad()) return 1;
	if(IsMeshTrig() && other.IsMeshTrig()) return 1;
	if(IsMeshLine() && other.IsMeshLine()) return 1;
	return 0;
}

int MeshElement::IsCyclicEqual(MeshElement& other) 
{
	if(! IsSameBasicType(other)) { return 0; } 
	
	TArray<int> pattern;
	TArray<int> inversepattern;

	for(int orientation=1; orientation<=NOrientations(); orientation++)
	{
		int isequal = 1;
		GetNodeListOriented(orientation,pattern,inversepattern);
		for(int nidx=1; nidx<=pattern.Length(); nidx++)
		{
			if (GetNode(nidx) != other.GetNode(pattern(nidx))) 
			{
				isequal = 0;  // mismatch for current orientation
				break;
			}
		}
		if(isequal==1) { return orientation; } // matching orientation
	}
	return 0; // no match in any orientation
}


//****************************************************************************************
// 3D MeshElements
//****************************************************************************************
// HEXAHEDRAL
//****************************************************************************************
int MeshHex::Invert()
{
	if(NNodes()==8||NNodes()==20) // corners
	{
		SwapNodeNumberEntries(1,5);
		SwapNodeNumberEntries(2,6);
		SwapNodeNumberEntries(3,7);
		SwapNodeNumberEntries(4,8);
		if(NNodes()==20)
		{
			SwapNodeNumberEntries( 9,11);
			SwapNodeNumberEntries(10,12);
			SwapNodeNumberEntries(13,15);
			SwapNodeNumberEntries(14,16);
			return 20;
		}
		return 8;
	}
	return 0;
}
//
int MeshHex::GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse)
{
//    orientation 0      orientation e.g.10
//                       0->(2xRX)->(2xRZ)
//       7-------8		        4-------3
//      /|      /|		       /|      /|
//     5-X---Y-6 |		      2-------1 |
//     | |     | |		      | |     | |
//     | 3- - -|-4		      | 8- - -|-7
//     |/      |/		      |/      |/
//     1-------2		      6-Y---X-5

	nodelist.Flush(); inverse.Flush();
	switch(orientation)
	{           
	case  0: nodelist.SetXN(8, 1,2,3,4, 5,6,7,8); inverse.SetXN(8, 1,2,3,4, 5,6,7,8); break; // no rotation        // inv:O0
	case  1: nodelist.SetXN(8, 3,1,4,2, 7,5,8,6); inverse.SetXN(8, 2,4,1,3, 6,8,5,7); break; // Rz                 // inv:O3
	case  2: nodelist.SetXN(8, 4,3,2,1, 8,7,6,5); inverse.SetXN(8, 4,3,2,1, 8,7,6,5); break; // Rz,Rz              // inv:O2
	case  3: nodelist.SetXN(8, 2,4,1,3, 6,8,5,7); inverse.SetXN(8, 3,1,4,2, 7,5,8,6); break; // Rz,Rz,Rz           // inv:O1         
	
	case  4: nodelist.SetXN(8, 5,6,1,2, 7,8,3,4); inverse.SetXN(8, 3,4,7,8, 1,2,5,6); break; // Rx                 // inv:O12     
	case  5: nodelist.SetXN(8, 1,5,2,6, 3,7,4,8); inverse.SetXN(8, 1,3,5,7, 2,4,6,8); break; // Rx,Rz              // inv:O23
	case  6: nodelist.SetXN(8, 2,1,6,5, 4,3,8,7); inverse.SetXN(8, 2,1,6,5, 4,3,8,7); break; // Rx,Rz,Rz           // inv:O6     
	case  7: nodelist.SetXN(8, 6,2,5,1, 8,4,7,3); inverse.SetXN(8, 4,2,8,6, 3,1,7,5); break; // Rx,Rz,Rz,Rz        // inv:O17     

	case  8: nodelist.SetXN(8, 7,8,5,6, 3,4,1,2); inverse.SetXN(8, 7,8,5,6, 3,4,1,2); break; // Rx,Rx              // inv:O8         
	case  9: nodelist.SetXN(8, 5,7,6,8, 1,3,2,4); inverse.SetXN(8, 5,7,6,8, 1,3,2,4); break; // Rx,Rx,Rz           // inv:O9      
	case 10: nodelist.SetXN(8, 6,5,8,7, 2,1,4,3); inverse.SetXN(8, 6,5,8,7, 2,1,4,3); break; // Rx,Rx,Rz,Rz        // inv:O10        
	case 11: nodelist.SetXN(8, 8,6,7,5, 4,2,3,1); inverse.SetXN(8, 8,6,7,5, 4,2,3,1); break; // Rx,Rx,Rz,Rz,Rz     // inv:O11      

	case 12: nodelist.SetXN(8, 3,4,7,8, 1,2,5,6); inverse.SetXN(8, 1,5,2,6, 3,7,4,8); break; // Rx,Rx,Rx           // inv:O5         
	case 13: nodelist.SetXN(8, 7,3,8,4, 5,1,6,2); inverse.SetXN(8, 6,8,2,4, 5,7,1,3); break; // Rx,Rx,Rx,Rz        // inv:O19                  
	case 14: nodelist.SetXN(8, 8,7,4,3, 6,5,2,1); inverse.SetXN(8, 8,7,4,3, 6,5,2,1); break; // Rx,Rx,Rx,Rz,Rz     // inv:O15                  
	case 15: nodelist.SetXN(8, 4,8,3,7, 2,6,1,5); inverse.SetXN(8, 7,5,3,1, 8,6,4,2); break; // Rx,Rx,Rx,Rz,Rz,Rz  // inv:O21                  

	case 16: nodelist.SetXN(8, 2,6,4,8, 1,5,3,7); inverse.SetXN(8, 5,1,7,3, 6,2,8,4); break; // Ry                 // inv:O20
	case 17: nodelist.SetXN(8, 4,2,8,6, 3,1,7,5); inverse.SetXN(8, 6,2,5,1, 8,4,7,3); break; // Ry,Rz              // inv:O7
	case 18: nodelist.SetXN(8, 8,4,6,2, 7,3,5,1); inverse.SetXN(8, 8,4,6,2, 7,3,5,1); break; // Ry,Rz,Rz           // inv:O18
	case 19: nodelist.SetXN(8, 6,8,2,4, 5,7,1,3); inverse.SetXN(8, 7,3,8,4, 5,1,6,2); break; // Ry,Rz,Rz,Rz        // inv:O13

	case 20: nodelist.SetXN(8, 5,1,7,3, 6,2,8,4); inverse.SetXN(8, 2,6,4,8, 1,5,3,7); break; // Ry,Ry,Ry           // inv:O16
	case 21: nodelist.SetXN(8, 7,5,3,1, 8,6,4,2); inverse.SetXN(8, 4,8,3,7, 2,6,1,5); break; // Ry,Ry,Ry,Rz        // inv:O15
	case 22: nodelist.SetXN(8, 3,7,1,5, 4,8,2,6); inverse.SetXN(8, 3,7,1,5, 4,8,2,6); break; // Ry,Ry,Ry,Rz,Rz     // inv:O22
	case 23: nodelist.SetXN(8, 1,3,5,7, 2,4,6,8); inverse.SetXN(8, 1,5,2,6, 3,7,4,8); break; // Ry,Ry,Ry,Rz,Rz,Rz  // inv:O5
	}
	return nodelist.Length();
}
//
MeshElement MeshHex::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshQuad(1,5,7,3),
		MeshQuad(2,4,8,6),
		MeshQuad(1,2,6,5),
		MeshQuad(3,7,8,4),
		MeshQuad(3,4,2,1),
		MeshQuad(5,6,8,7),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshQuad(1,5,7,3, 17,15,19,13),
		MeshQuad(2,4,8,6, 14,20,16,18),
		MeshQuad(1,2,6,5,  9,18,11,17),
		MeshQuad(3,7,8,4, 19,12,20,10),
		MeshQuad(3,4,2,1, 10,14, 9,13),
		MeshQuad(5,6,8,7, 11,16,12,15)
	};
	if(faceidx < 1 || faceidx > 6) { assert(0 && "MeshHex::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 8) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()==20) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshHex::Wrong number of Nodes available!"); return localFacesLinear[0]; } 
}

int2 MeshHex::GetInterPair(int i)
{
	if(NNodes()==8)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(3,4); break;
		case  3: return int2(5,6); break;
		case  4: return int2(7,8); break;
		case  5: return int2(1,3); break;
		case  6: return int2(2,4); break;
		case  7: return int2(5,7); break;
		case  8: return int2(6,8); break;
		case  9: return int2(1,5); break;
		case 10: return int2(2,6); break;
		case 11: return int2(3,7); break;
		case 12: return int2(4,8); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
//
Vector3D MeshHex::GetLocPosOfIthNode(int i)
{
// !!! ATTENTION : Compare with HexahedralGeneric::GetNodeLocPos !!!
	switch(i)
	{
	//linear:
	case  1: return Vector3D(-1.,-1.,-1.); break;
	case  2: return Vector3D( 1.,-1.,-1.); break;
	case  3: return Vector3D(-1., 1.,-1.); break;
	case  4: return Vector3D( 1., 1.,-1.); break;
	case  5: return Vector3D(-1.,-1., 1.); break;
	case  6: return Vector3D( 1.,-1., 1.); break;
	case  7: return Vector3D(-1., 1., 1.); break;
	case  8: return Vector3D( 1., 1., 1.); break;
	//quadratic:
	case  9: return Vector3D( 0.,-1.,-1.); break;
	case 10: return Vector3D( 0., 1.,-1.); break;
	case 11: return Vector3D( 0.,-1., 1.); break;
	case 12: return Vector3D( 0., 1., 1.); break;
	case 13: return Vector3D(-1., 0.,-1.); break;
	case 14: return Vector3D( 1., 0.,-1.); break;
	case 15: return Vector3D(-1., 0., 1.); break;
	case 16: return Vector3D( 1., 0., 1.); break;
	case 17: return Vector3D(-1.,-1., 0.); break;
	case 18: return Vector3D( 1.,-1., 0.); break;
	case 19: return Vector3D(-1., 1., 0.); break;
	case 20: return Vector3D( 1., 1., 0.); break;
	}
	return Vector3D(0);
}
double MeshHex::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();

	if (NNodes() == 8)
	{
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)*(1.0-t)/8.0; break;
		case 2: return (1.0+r)*(1.0-s)*(1.0-t)/8.0; break;
		case 3: return (1.0-r)*(1.0+s)*(1.0-t)/8.0; break;
		case 4: return (1.0+r)*(1.0+s)*(1.0-t)/8.0; break;
		case 5: return (1.0-r)*(1.0-s)*(1.0+t)/8.0; break;
		case 6: return (1.0+r)*(1.0-s)*(1.0+t)/8.0;	break;
		case 7: return (1.0-r)*(1.0+s)*(1.0+t)/8.0; break;
		case 8: return (1.0+r)*(1.0+s)*(1.0+t)/8.0;	break;
		}
	}
	else if(NNodes() == 20)
	{
		switch(shape)
		{
		case 1: return (1.0-r)*(1.0-s)*(1.0-t)*(-r-s-t-2.0)/8.0; break;
		case 2: return (1.0+r)*(1.0-s)*(1.0-t)*(r-s-t-2.0)/8.0; break;
		case 3: return (1.0-r)*(1.0+s)*(1.0-t)*(-r+s-t-2.0)/8.0; break;
		case 4: return (1.0+r)*(1.0+s)*(1.0-t)*(r+s-t-2.0)/8.0; break;
		case 5: return (1.0-r)*(1.0-s)*(1.0+t)*(-r-s+t-2.0)/8.0; break;
		case 6: return (1.0+r)*(1.0-s)*(1.0+t)*(r-s+t-2.0)/8.0;	break;
		case 7: return (1.0-r)*(1.0+s)*(1.0+t)*(-r+s+t-2.0)/8.0; break;
		case 8: return (1.0+r)*(1.0+s)*(1.0+t)*(r+s+t-2.0)/8.0;	break;
		case 9: return (1.0-r*r)*(1.0-s)*(1.0-t)/4.0; break;
		case 10: return (1.0-r*r)*(1.0+s)*(1.0-t)/4.0; break;
		case 11: return (1.0-r*r)*(1.0-s)*(1.0+t)/4.0; break;
		case 12: return (1.0-r*r)*(1.0+s)*(1.0+t)/4.0; break;
		case 13: return (1.0-s*s)*(1.0-r)*(1.0-t)/4.0; break;
		case 14: return (1.0-s*s)*(1.0+r)*(1.0-t)/4.0; break;
		case 15: return (1.0-s*s)*(1.0-r)*(1.0+t)/4.0; break;
		case 16: return (1.0-s*s)*(1.0+r)*(1.0+t)/4.0; break;
		case 17: return (1.0-t*t)*(1.0-r)*(1.0-s)/4.0; break;
		case 18: return (1.0-t*t)*(1.0+r)*(1.0-s)/4.0; break;
		case 19: return (1.0-t*t)*(1.0-r)*(1.0+s)/4.0; break;
		case 20: return (1.0-t*t)*(1.0+r)*(1.0+s)/4.0; break;
		}
	}
	assert(0 && "Wrong number of nodes or wrong index.");
	return 0;
}
Matrix MeshHex::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if (NNodes() == 8)
	{
		sm(1,1) = -(1.0-s)*(1.0-t)/8.0;
		sm(1,2) = (1.0-s)*(1.0-t)/8.0;
		sm(1,3) = -(1.0+s)*(1.0-t)/8.0;
		sm(1,4) = (1.0+s)*(1.0-t)/8.0;
		sm(1,5) = -(1.0-s)*(1.0+t)/8.0;
		sm(1,6) = (1.0-s)*(1.0+t)/8.0;
		sm(1,7) = -(1.0+s)*(1.0+t)/8.0;
		sm(1,8) = (1.0+s)*(1.0+t)/8.0;
		sm(2,1) = -(1.0-r)*(1.0-t)/8.0;
		sm(2,2) = -(1.0+r)*(1.0-t)/8.0;
		sm(2,3) = (1.0-r)*(1.0-t)/8.0;
		sm(2,4) = (1.0+r)*(1.0-t)/8.0;
		sm(2,5) = -(1.0-r)*(1.0+t)/8.0;
		sm(2,6) = -(1.0+r)*(1.0+t)/8.0;
		sm(2,7) = (1.0-r)*(1.0+t)/8.0;
		sm(2,8) = (1.0+r)*(1.0+t)/8.0;
		sm(3,1) = -(1.0-r)*(1.0-s)/8.0;
		sm(3,2) = -(1.0+r)*(1.0-s)/8.0;
		sm(3,3) = -(1.0-r)*(1.0+s)/8.0;
		sm(3,4) = -(1.0+r)*(1.0+s)/8.0;
		sm(3,5) = (1.0-r)*(1.0-s)/8.0;
		sm(3,6) = (1.0+r)*(1.0-s)/8.0;
		sm(3,7) = (1.0-r)*(1.0+s)/8.0;
		sm(3,8) = (1.0+r)*(1.0+s)/8.0;
	}
	else if(NNodes() == 20)
	{
		sm(1,1) = -(1.0-s)*(1.0-t)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(1,2) = (1.0-s)*(1.0-t)*(r-s-t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(1,3) = -(1.0+s)*(1.0-t)*(-r+s-t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(1,4) = (1.0+s)*(1.0-t)*(r+s-t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(1,5) = -(1.0-s)*(1.0+t)*(-r-s+t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(1,6) = (1.0-s)*(1.0+t)*(r-s+t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(1,7) = -(1.0+s)*(1.0+t)*(-r+s+t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(1,8) = (1.0+s)*(1.0+t)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(1,9) = -r*(1.0-s)*(1.0-t)/2.0;
		sm(1,10) = -r*(1.0+s)*(1.0-t)/2.0;
		sm(1,11) = -r*(1.0-s)*(1.0+t)/2.0;
		sm(1,12) = -r*(1.0+s)*(1.0+t)/2.0;
		sm(1,13) = -(1.0-s*s)*(1.0-t)/4.0;
		sm(1,14) = (1.0-s*s)*(1.0-t)/4.0;
		sm(1,15) = -(1.0-s*s)*(1.0+t)/4.0;
		sm(1,16) = (1.0-s*s)*(1.0+t)/4.0;
		sm(1,17) = -(1.0-t*t)*(1.0-s)/4.0;
		sm(1,18) = (1.0-t*t)*(1.0-s)/4.0;
		sm(1,19) = -(1.0-t*t)*(1.0+s)/4.0;
		sm(1,20) = (1.0-t*t)*(1.0+s)/4.0;
		sm(2,1) = -(1.0-r)*(1.0-t)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(2,2) = -(1.0+r)*(1.0-t)*(r-s-t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(2,3) = (1.0-r)*(1.0-t)*(-r+s-t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(2,4) = (1.0+r)*(1.0-t)*(r+s-t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(2,5) = -(1.0-r)*(1.0+t)*(-r-s+t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(2,6) = -(1.0+r)*(1.0+t)*(r-s+t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(2,7) = (1.0-r)*(1.0+t)*(-r+s+t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(2,8) = (1.0+r)*(1.0+t)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(2,9) = -(1.0-r*r)*(1.0-t)/4.0;
		sm(2,10) = (1.0-r*r)*(1.0-t)/4.0;
		sm(2,11) = -(1.0-r*r)*(1.0+t)/4.0;
		sm(2,12) = (1.0-r*r)*(1.0+t)/4.0;
		sm(2,13) = -s*(1.0-r)*(1.0-t)/2.0;
		sm(2,14) = -s*(1.0+r)*(1.0-t)/2.0;
		sm(2,15) = -s*(1.0-r)*(1.0+t)/2.0;
		sm(2,16) = -s*(1.0+r)*(1.0+t)/2.0;
		sm(2,17) = -(1.0-t*t)*(1.0-r)/4.0;
		sm(2,18) = -(1.0-t*t)*(1.0+r)/4.0;
		sm(2,19) = (1.0-t*t)*(1.0-r)/4.0;
		sm(2,20) = (1.0-t*t)*(1.0+r)/4.0;
		sm(3,1) = -(1.0-r)*(1.0-s)*(-r-s-t-2.0)/8.0-(1.0-r)*(1.0-s)*(1.0-t)/8.0;
		sm(3,2) = -(1.0+r)*(1.0-s)*(r-s-t-2.0)/8.0-(1.0+r)*(1.0-s)*(1.0-t)/8.0;
		sm(3,3) = -(1.0-r)*(1.0+s)*(-r+s-t-2.0)/8.0-(1.0-r)*(1.0+s)*(1.0-t)/8.0;
		sm(3,4) = -(1.0+r)*(1.0+s)*(r+s-t-2.0)/8.0-(1.0+r)*(1.0+s)*(1.0-t)/8.0;
		sm(3,5) = (1.0-r)*(1.0-s)*(-r-s+t-2.0)/8.0+(1.0-r)*(1.0-s)*(1.0+t)/8.0;
		sm(3,6) = (1.0+r)*(1.0-s)*(r-s+t-2.0)/8.0+(1.0+r)*(1.0-s)*(1.0+t)/8.0;
		sm(3,7) = (1.0-r)*(1.0+s)*(-r+s+t-2.0)/8.0+(1.0-r)*(1.0+s)*(1.0+t)/8.0;
		sm(3,8) = (1.0+r)*(1.0+s)*(r+s+t-2.0)/8.0+(1.0+r)*(1.0+s)*(1.0+t)/8.0;
		sm(3,9) = -(1.0-r*r)*(1.0-s)/4.0;
		sm(3,10) = -(1.0-r*r)*(1.0+s)/4.0;
		sm(3,11) = (1.0-r*r)*(1.0-s)/4.0;
		sm(3,12) = (1.0-r*r)*(1.0+s)/4.0;
		sm(3,13) = -(1.0-s*s)*(1.0-r)/4.0;
		sm(3,14) = -(1.0-s*s)*(1.0+r)/4.0;
		sm(3,15) = (1.0-s*s)*(1.0-r)/4.0;
		sm(3,16) = (1.0-s*s)*(1.0+r)/4.0;
		sm(3,17) = -t*(1.0-r)*(1.0-s)/2.0;
		sm(3,18) = -t*(1.0+r)*(1.0-s)/2.0;
		sm(3,19) = -t*(1.0-r)*(1.0+s)/2.0;
		sm(3,20) = -t*(1.0+r)*(1.0+s)/2.0;
	}
	return sm;
}
// 
int3 MeshHex::IdentifyRequiredTemplate(TArray<int>& lvl, int method)
{
	
// NOTES: precomputed list - permutaiton of bitpattern with nodelist 
// table has 256 entries 2^8 combinations, each node to be refined or not
// entries in the int3:  (1) Number of 1s 
//                       (2) shape - relative positioning of the 1s
//                       (3) orientation := index such that the selected nodes are in exact template position
//                           if e.g. nodes 2 and 4 are selected (bitvalue 2+8=10) gives 2 nodes in shape 0 (edge)
//                           the orientation where they match the (2,0,?) prototype is require node2 at corner1, node4 at corner2, thats orientation#3      bv 10 --> (2,0,3)
// description for the prototype in terms of "lines" with involved nodes
// (0,0,0) -> 1 entry
// (1,0,Y) -> 8 entries - 8 corners
// (2,X,Y) -> 28 entries from 3 shapes: 12 edges(1-2), 12 facediag(1-6), 4 volumediag(1-8)
// (3,X,Y) -> 56 entries from 3 shapes: 24 edgeedge(1-2-3), 24 edgefacediag(1,2,7), 8 facediagfacediag(1,6,7)
// (4,X,Y) -> 70 entries from 7 shapes: 6 entire face(1-2-5-6), 8 tripods(1,2,3,5), 12 around lefthanded(1,5,7,8), 12 around righthanded(1,2,4,8), 24 N.N. (1,2,5,8), 6 prisms(1,4,5,8), 2 twisted(1,4,6,7)
// (5,X,Y) -> inverse nodes to 3xy - difficult to describe in terms of lines...
// ...
// NOTE: all degenerated options must be tested
// Tested configurations: 0**, 1**, 20*, 40*, 8**
// NOT YET TESTED ( element patterns not implemented yet)
//  
// templates for method 0 - all directions require all 256 bitvalues
	static const int3 pattern2template[] = {                                                                        // bitvalue
		int3(0,0, 0),int3(1,0, 0),int3(1,0, 3),int3(2,0, 0),  int3(1,0, 1),int3(2,0, 1),int3(2,1, 6),int3(3,0, 0),  //   0..  7
		int3(1,0, 2),int3(2,1,17),int3(2,0, 3),int3(3,0, 3),  int3(2,0, 2),int3(3,0, 1),int3(3,0, 2),int3(4,0, 6),  //   8.. 15
		int3(1,0, 4),int3(2,0, 5),int3(2,1,16),int3(3,0, 5),  int3(2,1, 1),int3(3,0,23),int3(3,2, 9),int3(4,1, 0),  //  16.. 23
		int3(2,2, 2),int3(3,1, 5),int3(3,1, 3),int3(4,2, 9),  int3(3,1,12),int3(4,3, 2),int3(4,4,17),int3(5,0,11),  //  24.. 31
		int3(1,0, 7),int3(2,1, 0),int3(2,0,16),int3(3,0, 6),  int3(2,2, 1),int3(3,1,23),int3(3,1,16),int3(4,3, 1),  //  32.. 39
		int3(2,1, 7),int3(3,2,10),int3(3,0,16),int3(4,1, 3),  int3(3,1, 2),int3(4,4,12),int3(4,2,10),int3(5,0, 8),  //  40.. 47
		int3(2,0, 4),int3(3,0, 4),int3(3,0, 7),int3(4,0, 0),  int3(3,1, 4),int3(4,2,12),int3(4,4,10),int3(5,0,14),  //  48.. 55
		int3(3,1,10),int3(4,4,20),int3(4,3, 4),int3(5,0,13),  int3(4,5,17),int3(5,1,14),int3(5,1, 8),int3(6,0, 8),  //  56.. 63

		int3(1,0,21),int3(2,1, 5),int3(2,2, 3),int3(3,1, 0),  int3(2,0,22),int3(3,0,22),int3(3,1,22),int3(4,2, 8),  //  64.. 71
		int3(2,1, 8),int3(3,2, 5),int3(3,1,17),int3(4,4, 6),  int3(3,0,12),int3(4,1, 1),int3(4,3, 3),int3(5,0,10),  //  72.. 79
		int3(2,0, 9),int3(3,0,20),int3(3,1, 9),int3(4,3, 6),  int3(3,0,21),int3(4,0, 1),int3(4,4,13),int3(5,0,18),  //  80.. 87
		int3(3,1,21),int3(4,4, 9),int3(4,5, 4),int3(5,1,11),  int3(4,2, 4),int3(5,0,19),int3(5,1,19),int3(6,0,11),  //  88.. 95
		int3(2,1,19),int3(3,2, 0),int3(3,1, 7),int3(4,4,16),  int3(3,1,13),int3(4,4, 1),int3(4,5, 1),int3(5,1,18),  //  96..103
		int3(3,2, 2),int3(4,6, 0),int3(4,4, 3),int3(5,2, 1),  int3(4,4,22),int3(5,2, 3),int3(5,1,20),int3(6,1, 4),  // 104..111
		int3(3,0, 9),int3(4,1, 4),int3(4,2, 3),int3(5,0,15),  int3(4,3,10),int3(5,0,17),int3(5,1,15),int3(6,0,15),  // 112..119
		int3(4,4, 4),int3(5,2, 6),int3(5,1, 1),int3(6,1,18),  int3(5,1, 6),int3(6,1, 3),int3(6,2, 0),int3(7,0,11),  // 120..127 

		int3(1,0,11),int3(2,2, 0),int3(2,1, 3),int3(3,1, 6),  int3(2,1,18),int3(3,1, 1),int3(3,2, 6),int3(4,4,23),  // 128..135
		int3(2,0,15),int3(3,1,15),int3(3,0,17),int3(4,3, 0),  int3(3,0,15),int3(4,2, 5),int3(4,1, 2),int3(5,0, 9),  // 136..143
		int3(2,1, 4),int3(3,1,20),int3(3,2, 3),int3(4,4, 0),  int3(3,2, 1),int3(4,4, 5),int3(4,6, 1),int3(5,2, 2),  // 144..151
		int3(3,1,18),int3(4,5, 0),int3(4,4,15),int3(5,1,13),  int3(4,4, 2),int3(5,1, 7),int3(5,2, 0),int3(6,1,19),  // 152..159
		int3(2,0,11),int3(3,1,19),int3(3,0,19),int3(4,2,14),  int3(3,1,11),int3(4,5, 4),int3(4,4, 5),int3(5,1,21),  // 160..167
		int3(3,0,18),int3(4,4,11),int3(4,0, 3),int3(5,0,21),  int3(4,3,12),int3(5,1, 9),int3(5,0,20),int3(6,0, 9),  // 168..175
		int3(3,0,10),int3(4,3, 5),int3(4,1,10),int3(5,0,12),  int3(4,4,19),int3(5,1,17),int3(5,2, 5),int3(6,1, 2),  // 176..183
		int3(4,2, 2),int3(5,1,22),int3(5,0,22),int3(6,0,13),  int3(5,1, 0),int3(6,2, 3),int3(6,1, 5),int3(7,0, 8),  // 184..191

		int3(2,0, 8),int3(3,1, 8),int3(3,1,14),int3(4,5,16),  int3(3,0,13),int3(4,3,14),int3(4,4, 8),int3(5,1,10),  // 192..199
		int3(3,0,14),int3(4,4,18),int3(4,2, 6),int3(5,1, 4),  int3(4,0, 2),int3(5,0, 7),int3(5,0, 5),int3(6,0, 4),  // 200..207
		int3(3,0, 8),int3(4,2, 0),int3(4,4,21),int3(5,1, 2),  int3(4,1, 8),int3(5,0,16),int3(5,2,10),int3(6,1, 7),  // 208..215
		int3(4,3, 9),int3(5,1,16),int3(5,1,23),int3(6,2, 1),  int3(5,0, 6),int3(6,0, 7),int3(6,1, 0),int3(7,0,10),  // 216..223
		int3(3,0,11),int3(4,4,14),int3(4,3, 8),int3(5,1,12),  int3(4,2, 6),int3(5,1, 3),int3(5,1, 5),int3(6,2, 2),  // 224..231
		int3(4,1,11),int3(5,2, 9),int3(5,0,23),int3(6,1, 1),  int3(5,0, 5),int3(6,1,16),int3(6,0, 5),int3(7,0, 4),  // 232..239
		int3(4,0, 4),int3(5,0, 2),int3(5,0, 1),int3(6,0, 2),  int3(5,0, 3),int3(6,0, 4),int3(6,1,17),int3(7,0, 2),  // 240..247
		int3(5,0, 0),int3(6,1, 6),int3(6,0, 1),int3(7,0, 1),  int3(6,0, 0),int3(7,0, 3),int3(7,0, 0),int3(8,0, 0)   // 248..255	
	};
// templates "2D" for methods 1,2,3 - templates for a single face - NOT identical to MeshQuad, as node order is different 
//  FACE      QUAD               F1     F3     F5    N=0 -> BV 0         -> 000                N=4 -> BV 15         -> 400
// L3--L4     4--3              3--7   5--6   2--4   N=1 -> BV 1,2,8,4   -> 100,101,102,103    N=3 -> BV 11,14,13,7 -> 300,301,302,303
//  |  |  vs  |  |              |  |   |  |   |  |   N=2 -> BV 3,10,12,5 -> 200,201,202,203
// L1--L2     1--2              1--5   1--2   1--3       -> BV 9,6       -> 210,211
// 
//   HEX O0:       M1 O0:    M1 01:    M1 O2:    M1 O3:         
//    7----8       4----8    2----4    6----2    8----6     M2 O0 -> HEX O0     M3 O0 -> HEX O23
//   5----6|	  3----7|   1----3|   5----1|   7----5|     M2 O1 -> HEX O16    M3 O1 -> HEX O12
//   ||   ||	  ||   ||   ||   ||   ||   ||   ||   ||     M2 O2 -> HEX O10    M3 O2 -> HEX O17
//   |3---|4	  |2---|6   |6---|8   |8---|4   |4---|2		M2 O3 -> HEX O20    M3 O3 -> HEX O6
//   1----2		  1----5    5----7    7----3    3----1
//                HEX O5    HEX O9    HEX O13   HEX O1
	static const int3 pattern2template2DM1[] = {
		int3(0,0,/*0*/ 5),int3(1,0,/*0*/ 5),int3(1,0,/*1*/ 9),int3(2,0,/*0*/ 5), //  0.. 3  QUAD:int3(0,0,0),int3(1,0,0),int3(1,0,1),int3(2,0,0),
		int3(1,0,/*3*/ 1),int3(2,0,/*3*/ 1),int3(2,1,/*1*/ 9),int3(3,0,/*3*/ 1), //  4.. 7  QUAD:int3(1,0,2),int3(2,1,0),int3(2,0,1),int3(3,0,0),
		int3(1,0,/*2*/13),int3(2,1,/*0*/ 5),int3(2,0,/*1*/ 9),int3(3,0,/*0*/ 5), //  8..11  QUAD:int3(1,0,3),int3(2,0,3),int3(2,1,1),int3(3,0,3),
		int3(2,0,/*2*/13),int3(3,0,/*2*/13),int3(3,0,/*1*/ 9),int3(4,0,/*0*/ 5)  // 12..15  QUAD:int3(2,0,2),int3(3,0,2),int3(3,0,1),int3(4,0,0) 
	};
	static const int3 pattern2template2DM2[] = {
		int3(0,0,/*0*/ 0),int3(1,0,/*0*/ 0),int3(1,0,/*1*/16),int3(2,0,/*0*/ 0), 
		int3(1,0,/*3*/20),int3(2,0,/*3*/20),int3(2,1,/*1*/16),int3(3,0,/*3*/20), 
		int3(1,0,/*2*/10),int3(2,1,/*0*/ 0),int3(2,0,/*1*/16),int3(3,0,/*0*/ 0), 
		int3(2,0,/*2*/10),int3(3,0,/*2*/10),int3(3,0,/*1*/16),int3(4,0,/*0*/ 0)  
	};
	static const int3 pattern2template2DM3[] = {
		int3(0,0,/*0*/23),int3(1,0,/*0*/23),int3(1,0,/*1*/12),int3(2,0,/*0*/23), 
		int3(1,0,/*3*/ 6),int3(2,0,/*3*/ 6),int3(2,1,/*1*/12),int3(3,0,/*3*/ 6), 
		int3(1,0,/*2*/17),int3(2,1,/*0*/23),int3(2,0,/*1*/12),int3(3,0,/*0*/23), 
		int3(2,0,/*2*/17),int3(3,0,/*2*/17),int3(3,0,/*1*/12),int3(4,0,/*0*/23)  
	};

	if(method==0)
	{	
		int nactive = 0;
		int patternbinary = 0;
		if(lvl(1)>0) { nactive++; patternbinary +=  1; }
		if(lvl(2)>0) { nactive++; patternbinary +=  2; }
		if(lvl(3)>0) { nactive++; patternbinary +=  4; }
		if(lvl(4)>0) { nactive++; patternbinary +=  8; }
		if(lvl(5)>0) { nactive++; patternbinary += 16; }
		if(lvl(6)>0) { nactive++; patternbinary += 32; }
		if(lvl(7)>0) { nactive++; patternbinary += 64; }
		if(lvl(8)>0) { nactive++; patternbinary +=128; }
		return pattern2template[patternbinary];
	}
	else if(method==1) // relevant faces: 1->2
	{
		int nactive1 = 0; int nactive2 = 0;        
		int face1pattern=0; int face2pattern=0;      
		if(lvl(1)>0) { nactive1++; face1pattern += 1; }		if(lvl(2)>0) { nactive2++; face2pattern += 1; }
		if(lvl(5)>0) { nactive1++; face1pattern += 2; }		if(lvl(6)>0) { nactive2++; face2pattern += 2; }
		if(lvl(3)>0) { nactive1++; face1pattern += 4; }		if(lvl(4)>0) { nactive2++; face2pattern += 4; }
		if(lvl(7)>0) { nactive1++; face1pattern += 8; }		if(lvl(8)>0) { nactive2++; face2pattern += 8; }
		return pattern2template2DM1[face1pattern];
	}
	else if(method==2) // relevant faces: 3->4
	{
		int nactive3 = 0; int nactive4 = 0;          
		int face3pattern=0; int face4pattern=0;       	                          
		if(lvl(1)>0) { nactive3++; face3pattern += 1; }		if(lvl(3)>0) { nactive4++; face4pattern += 1; }
		if(lvl(2)>0) { nactive3++; face3pattern += 2; }		if(lvl(4)>0) { nactive4++; face4pattern += 2; }
		if(lvl(5)>0) { nactive3++; face3pattern += 4; }		if(lvl(7)>0) { nactive4++; face4pattern += 4; }
		if(lvl(6)>0) { nactive3++; face3pattern += 8; }		if(lvl(8)>0) { nactive4++; face4pattern += 8; }
		return pattern2template2DM2[face3pattern];
	}
	else if(method==3) // relevant faces: 5->6
	{
		int nactive5 = 0; int nactive6 = 0;         
		int face5pattern=0; int face6pattern=0;       	                          
		if(lvl(1)>0) { nactive5++; face5pattern += 1; }		if(lvl(5)>0) { nactive6++; face6pattern += 1; }
		if(lvl(3)>0) { nactive5++; face5pattern += 2; }		if(lvl(7)>0) { nactive6++; face6pattern += 2; }
		if(lvl(2)>0) { nactive5++; face5pattern += 4; }		if(lvl(6)>0) { nactive6++; face6pattern += 4; }
		if(lvl(4)>0) { nactive5++; face5pattern += 8; }		if(lvl(8)>0) { nactive6++; face6pattern += 8; }
		return pattern2template2DM3[face5pattern];
	}
	return int3(0,0,0);
}
int MeshHex::IsTemplateImplemented(TArray<int>& lvl, int method)
{
	int3 templatepattern = IdentifyRequiredTemplate(lvl,method);
	if(method==0)
	{
// these are the templates that are actually implemented
		if( templatepattern(1) == 0 ) return 1;
		if( templatepattern(1) == 1 ) return 1;
		if( templatepattern(1) == 2 && templatepattern(2) == 0 ) return 1;
		if( templatepattern(1) == 4 && templatepattern(2) == 0 ) return 1;
		if( templatepattern(1) == 8 ) return 1;

// templates mapped to (4,0,*)
		if(( templatepattern(1) == 2 && templatepattern(2) == 1) ||  // node 1 and 6 marked -> also mark node 2 and 5  (2,1,0)->bv33
		   ( templatepattern(1) == 3 && templatepattern(2) == 0) )   // node 1, 2 and 3 marked -> also mark node 4     (3,0,0)->bv7
		{
			int orientation = templatepattern(3);
			TArray<int> nodelist;
			TArray<int> inverse;
			GetNodeListOriented(orientation,nodelist,inverse);
			if(templatepattern(1) == 2) 
			{ // node 1 and 6 of orientaiton 0 are marked,  node 2 and 3 must be marked in addition
				lvl(nodelist(2)) = 1;
				lvl(nodelist(5)) = 1;
			}
			else if(templatepattern(1) == 3)
			{ // node 1,2 and 3 of orientaiton 0 are marked,  node 4 must be marked in addition
				lvl(nodelist(4)) = 1;
			}
			return 0;
		}

// templates mapped to (8,0,0)
		for(int i=1; i<=lvl.Length(); i++)
		{
			if(lvl(i)==0) lvl(i)=1;
		}
		return 0;
	}
	else
		return 1; // all templates for method 1,2,3 are implemented
}

Vector3D MeshHex::ApplyOrientation(Vector3D locpos, int orientation)
{
// do transformation such that the local position of the new node is correctly oriented
// orientation verified with Corner-configurations and colored additional nodes
	Matrix3D A = GetOrientationMatrix(orientation);
	Vector3D locposoriented = A * locpos;                                                                                      
	return locposoriented;
}
Matrix3D MeshHex::GetOrientationMatrix(int orientation)
{
	static const double rotations[] = { 1., 0., 0.,   0., 1., 0.,   0., 0., 1.,   // O 0
		                                0., 1., 0.,  -1., 0., 0.,   0., 0., 1.,   // O 1
                                       -1., 0., 0.,   0.,-1., 0.,   0., 0., 1.,   // O 2
		                                0.,-1., 0.,   1., 0., 0.,   0., 0., 1.,   // O 3 

										1., 0., 0.,   0., 0., 1.,   0.,-1., 0.,   // O 4
										0., 1., 0.,   0., 0., 1.,   1., 0., 0.,   // O 5
									   -1., 0., 0.,   0., 0., 1.,   0., 1., 0.,   // O 6
										0.,-1., 0.,   0., 0., 1.,  -1., 0., 0.,   // O 7

                                        1., 0., 0.,   0.,-1., 0.,   0., 0.,-1.,   // O 8
                                        0., 1., 0.,   1., 0., 0.,   0., 0.,-1.,   // O 9
                                       -1., 0., 0.,   0., 1., 0.,   0., 0.,-1.,   // O10
                                        0.,-1., 0.,  -1., 0., 0.,   0., 0.,-1.,   // O11

										1., 0., 0.,   0., 0.,-1.,   0., 1., 0.,   // O12
										0., 1., 0.,   0., 0.,-1.,  -1., 0., 0.,   // O13
									   -1., 0., 0.,   0., 0.,-1.,   0.,-1., 0.,   // O14
										0.,-1., 0.,   0., 0.,-1.,   1., 0., 0.,   // O15

										0., 0.,-1.,   0., 1., 0.,   1., 0., 0.,   // O16
										0., 0.,-1.,  -1., 0., 0.,   0., 1., 0.,   // O17
										0., 0.,-1.,   0.,-1., 0.,  -1., 0., 0.,   // O18
										0., 0.,-1.,   1., 0., 0.,   0.,-1., 0.,   // O19

										0., 0., 1.,   0., 1., 0.,  -1., 0., 0.,   // O20
										0., 0., 1.,  -1., 0., 0.,   0.,-1., 0.,   // O21
										0., 0., 1.,   0.,-1., 0.,   1., 0., 0.,   // O22
										0., 0., 1.,   1., 0., 0.,   0., 1., 0. }; // O23

	double* rv0 = (double*) &rotations[orientation*9];
	Matrix3D orientationmatrix( rv0[0], rv0[1], rv0[2],
		                       rv0[3], rv0[4], rv0[5],
							   rv0[6], rv0[7], rv0[8] );
	return orientationmatrix;
}
// 
int MeshHex::GetNumberOfAdditionalNodesForRefine(int3 pattern, int method)
{
	if(method==0)
	{
		if(pattern(1)==0) return 0;
		if(pattern(1)==1) return 7; //3+4
		if(pattern(1)==2 && pattern(2)==0) return 20; //8+8+4
		//2b
		//2c
		//3a
		//3b
		//3c
		if(pattern(1)==4 && pattern(2)==0) return 40; //8+10+10+8+4
        //4b
		//4c
		//4d
		//4e
		//4f
		//4g
		//5a
		//5b
		//5c
		//6a
		//6b
		//6c
		//7
		if(pattern(1)==8) return 56; //12+16+16+12
	}
	else if (method==1 || method==2 || method==3) // twice as for MeshQuad refinement
	{
		if(pattern(1)==0) return 2* 0; 
		if(pattern(1)==1) return 2* 3; 
		if(pattern(1)==2) return 2* 8; 
		if(pattern(1)==3) return 2*10; 
		if(pattern(1)==4) return 2*12; 
	}
	return -1;
}
int MeshHex::GetGridNumberOfIthAdditionalNode(int i, int3 pattern, int method)
{
// lookup tables - grid index of the required nodes for the template   
// LOTS of grigpoints are NOT shown...
//	       O7-G55-G56--O8
//        /	  /   /   /|
//      G51-G52-G53-G54|
//      /   /   /   / G44
//    G47-G48-G49-G50  |
//    /   /   /   /    |
//   O5-G45-G46--O6   G28
//   |           |     |
//   |          ...    |
//  G29... O3-G11-G12--O4
//   |    /   /   /   /
//   |   G7--G8--G9-G10
//  G13...  /   /   /
//   | G3--G4--G5--G6
//   |/   /   /   /
//   O1--G1--G2--O2
//
// lookup tables:  for all templates the additional nodes in ascending order of gridnumbers and layer by layer
	static const int NType1[] =  { 1,    3, 4,                                                                
		                          13,14,      17,18 };                                                        
	static const int NType2a[] = { 1, 2, 3, 4, 5, 6,    8, 9,                                                 
		                          13,14,15,16,17,18,19,20,                                                    
	                                 30,31,                  38,39};     
	/*not implemented yet*/static const int NType2b[] = {0};
	/*not implemented yet*/static const int NType2c[] = {0};
	/*not implemented yet*/static const int NType3a[] = {0};
	/*not implemented yet*/static const int NType3b[] = {0};
	/*not implemented yet*/static const int NType3c[] = {0};
	static const int NType4a[] = { 1, 2, 3, 4, 5, 6,    8, 9,                                                 
		                          13,14,15,16,17,18,19,20,21,      24,                                       
								  29,30,31,32,33,34,35,36,37,      40,
                                  45,46,47,48,49,50,   52,53,
								  1418,1519,3034,3135 }; // additional nodes NOT on the 1/3 grid
	/*not implemented yet*/static const int NType4b[] = {0};
	/*not implemented yet*/static const int NType4c[] = {0};
	/*not implemented yet*/static const int NType4d[] = {0};
	/*not implemented yet*/static const int NType4e[] = {0};
	/*not implemented yet*/static const int NType4f[] = {0};
	/*not implemented yet*/static const int NType5a[] = {0};
	/*not implemented yet*/static const int NType5b[] = {0};
	/*not implemented yet*/static const int NType5c[] = {0};
	/*not implemented yet*/static const int NType6a[] = {0};
	/*not implemented yet*/static const int NType6b[] = {0};
	/*not implemented yet*/static const int NType6c[] = {0};
	/*not implemented yet*/static const int NType7[] = {0};
	static const int NType8[]  = { 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12, 
	                              13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,
								  29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,
								  45,46,47,48,49,50,51,52,53,54,55,56 };
// lookup tables reduced:  for all templates the additional nodes on foregroud face, then background face
	static const int NType12D[] =  { 1,     13, 14,
	                                11,     25, 26 };
	static const int NType2a2D[] = { 1,  2, 13, 14, 15, 16,     30, 31,
	                                11, 12, 25, 26, 27, 28,     42, 43};
	static const int NType2b2D[] = { 1,     13, 14, 15,         30, 31, 32,     46,
	                                11,     25, 26, 27,         42, 43, 44,     56};
	static const int NType32D[] =  { 1,  2, 13, 14, 15, 16,     30, 31, 32,     46,
	                                11, 12, 25, 26, 27, 28,     42, 43, 44,     56};
	static const int NType42D[] =  { 1,  2, 13, 14, 15, 16, 29, 30, 31, 32, 45, 46,
	                                11, 12, 25, 26, 27, 28, 41, 42, 43, 44, 55, 56};

	int ncorners =    pattern(1);
	int subtype =	  pattern(2);
  	int orientation = pattern(3);

	int gridnr = 0;
	if(method==0)
	{
		switch(ncorners)
		{
		case 0: break;
		case 1: if(i>0&&i<= 7) gridnr= NType1[i-1]; break;
		case 2: if(subtype==0) { if(i>0&&i<=20) gridnr= NType2a[i-1]; }
				break;
		case 3: break;
		case 4: if(subtype==0) { if(i>0&&i<=40) gridnr= NType4a[i-1]; }
				break;
		case 5: break;
		case 6: break;
		case 7: break;
		case 8: if(i>0&&i<=56) gridnr= NType8[i-1]; break;
		}
	}
	else if (method==1 || method==2 || method==3)
	{
		switch(ncorners)
		{
		case 0: break;
		case 1: if(i>0&&i<= 6) gridnr= NType12D[i-1]; break;
		case 2: if(subtype==0) { if(i>0&&i<=16) gridnr= NType2a2D[i-1]; }
				else { if(i>0&&i<=16) gridnr= NType2b2D[i-1]; } break;
		case 3: if(i>0&&i<=20) gridnr= NType32D[i-1]; break;
		case 4: if(i>0&&i<=24) gridnr= NType42D[i-1]; break;
		}
	}
	return gridnr;
}
Vector3D MeshHex::GetLocPosOfIthAdditionalNode(int i, int3 pattern, int method)
{
// get the gridnumber of the requested additional node
	int gridnr = GetGridNumberOfIthAdditionalNode(i,pattern,method);

// get local position of the grid node
	Vector3D locpos(0.,0.,0.);
	if(gridnr== 3 || gridnr== 7 || gridnr==13 || gridnr==17 || gridnr==21 || gridnr==25 || 
	   gridnr==29 || gridnr==33 || gridnr==37 || gridnr==41 || gridnr==47 || gridnr==51) locpos.X()= -1.;
	if(gridnr== 1 || gridnr== 4 || gridnr== 8 || gridnr==11 || gridnr==14 || gridnr==18 || gridnr==22 || gridnr==26 || 
	   gridnr==30 || gridnr==34 || gridnr==38 || gridnr==42 || gridnr==45 || gridnr==48 || gridnr==52 || gridnr==55) locpos.X()= -1./3.;
	if(gridnr== 2 || gridnr== 5 || gridnr== 9 || gridnr==12 || gridnr==15 || gridnr==19 || gridnr==23 || gridnr==27 || 
	   gridnr==31 || gridnr==35 || gridnr==39 || gridnr==43 || gridnr==46 || gridnr==49 || gridnr==53 || gridnr==56) locpos.X()= +1./3.;
	if(gridnr== 6 || gridnr==10 || gridnr==16 || gridnr==20 || gridnr==24 || gridnr==28 || 
	   gridnr==32 || gridnr==36 || gridnr==40 || gridnr==44 || gridnr==50 || gridnr==54) locpos.X()= +1.;

	if(gridnr== 1 || gridnr== 2 || gridnr==13 || gridnr==14 || gridnr==15 || gridnr==16 || 
	   gridnr==29 || gridnr==30 || gridnr==31 || gridnr==32 || gridnr==45 || gridnr==46) locpos.Y()= -1.;
	if(gridnr== 3 || gridnr== 4 || gridnr== 5 || gridnr== 6 || gridnr==17 || gridnr==18 || gridnr==19 || gridnr==20 || 
	   gridnr==33 || gridnr==34 || gridnr==35 || gridnr==36 || gridnr==47 || gridnr==48 || gridnr==49 || gridnr==50) locpos.Y()= -1./3.;
	if(gridnr== 7 || gridnr== 8 || gridnr== 9 || gridnr==10 || gridnr==21 || gridnr==22 || gridnr==23 || gridnr==24 || 
	   gridnr==37 || gridnr==38 || gridnr==39 || gridnr==40 || gridnr==51 || gridnr==52 || gridnr==53 || gridnr==54) locpos.Y()= +1./3.;
	if(gridnr==11 || gridnr==12 || gridnr==25 || gridnr==26 || gridnr==27 || gridnr==28 || 
	   gridnr==41 || gridnr==42 || gridnr==43 || gridnr==44 || gridnr==55 || gridnr==56) locpos.Y()= +1.;

	if(gridnr== 1 || gridnr== 2 || gridnr== 3 || gridnr== 4 || gridnr== 5 || gridnr== 6 || 
	   gridnr== 7 || gridnr== 8 || gridnr== 9 || gridnr==10 || gridnr==11 || gridnr==12) locpos.Z()= -1.;
	if(gridnr==13 || gridnr==14 || gridnr==15 || gridnr==16 || gridnr==17 || gridnr==18 || gridnr==19 || gridnr==20 || 
	   gridnr==21 || gridnr==22 || gridnr==23 || gridnr==24 || gridnr==25 || gridnr==26 || gridnr==27 || gridnr==28) locpos.Z()= -1./3.;
	if(gridnr==29 || gridnr==30 || gridnr==31 || gridnr==32 || gridnr==33 || gridnr==34 || gridnr==35 || gridnr==36 || 
	   gridnr==37 || gridnr==38 || gridnr==39 || gridnr==40 || gridnr==41 || gridnr==42 || gridnr==43 || gridnr==44) locpos.Z()= +1./3.;
	if(gridnr==45 || gridnr==46 || gridnr==47 || gridnr==48 || gridnr==49 || gridnr==50 || 
	   gridnr==51 || gridnr==52 || gridnr==53 || gridnr==54 || gridnr==55 || gridnr==56) locpos.Z()= +1.;
// offgrid nodes
	if(gridnr==1418) locpos=Vector3D(-1./3.,0.,-1./3.);
	if(gridnr==1519) locpos=Vector3D(+1./3.,0.,-1./3.);
	if(gridnr==3034) locpos=Vector3D(-1./3.,0.,+1./3.);
	if(gridnr==3135) locpos=Vector3D(+1./3.,0.,+1./3.);

// loc pos is now the local position for orientation 0
	int orientation = pattern(3);
	return ApplyOrientation(locpos,orientation);
}
int MeshHex::GetSubdivisionLevelOfIthAdditionalNode(int i, int3 pattern, TArray<int>& cornersdlvls, int method)
{
//NOTE: general rule: if on an edge, subdivisionlevel is equal to minimum of edge,
//                    if in middle of a face, subdivision level is equal to minimum of all 4 corners	
//                    if in middle of volume, subdivision level is equal to minimum of all 8 corners

	int orientation = pattern(3);
	TArray<int> orientedNL;
	TArray<int> inverseNL;
	GetNodeListOriented(orientation,orientedNL,inverseNL);

	int weight = 0;
	int gridnr = GetGridNumberOfIthAdditionalNode(i,pattern,method);

// node on edge: 12 edges
	if(gridnr== 1 || gridnr== 2) // on edge O1 - O2 ( == E12 prototype)
	{ 
		weight = cornersdlvls(orientedNL(1));
		if(weight > cornersdlvls(orientedNL(2))) weight = cornersdlvls(orientedNL(2));
	}                                                                                                                                                           //E12
	else if(gridnr== 3 || gridnr== 7) { weight = cornersdlvls(orientedNL(1)); if(weight > cornersdlvls(orientedNL(3))) weight = cornersdlvls(orientedNL(3)); }  //E13
	else if(gridnr==11 || gridnr==12) { weight = cornersdlvls(orientedNL(3)); if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4)); }  //E34
	else if(gridnr== 6 || gridnr==10) { weight = cornersdlvls(orientedNL(2)); if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4)); }  //E24

	else if(gridnr==45 || gridnr==46) { weight = cornersdlvls(orientedNL(5)); if(weight > cornersdlvls(orientedNL(6))) weight = cornersdlvls(orientedNL(6)); }  //E56
	else if(gridnr==47 || gridnr==51) { weight = cornersdlvls(orientedNL(5)); if(weight > cornersdlvls(orientedNL(7))) weight = cornersdlvls(orientedNL(7)); }  //E57
	else if(gridnr==55 || gridnr==56) { weight = cornersdlvls(orientedNL(7)); if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8)); }  //E78
	else if(gridnr==50 || gridnr==54) { weight = cornersdlvls(orientedNL(6)); if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8)); }  //E68

	else if(gridnr==13 || gridnr==29) { weight = cornersdlvls(orientedNL(1)); if(weight > cornersdlvls(orientedNL(5))) weight = cornersdlvls(orientedNL(5)); }  //E15
	else if(gridnr==16 || gridnr==32) { weight = cornersdlvls(orientedNL(2)); if(weight > cornersdlvls(orientedNL(6))) weight = cornersdlvls(orientedNL(6)); }  //E26
	else if(gridnr==25 || gridnr==41) { weight = cornersdlvls(orientedNL(3)); if(weight > cornersdlvls(orientedNL(7))) weight = cornersdlvls(orientedNL(7)); }  //E37
	else if(gridnr==28 || gridnr==44) { weight = cornersdlvls(orientedNL(4)); if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8)); }  //E48

// node on face: 6 faces
	else if(gridnr==14 || gridnr==15 || gridnr==30 || gridnr==31)                      // F1256                    
	{
		weight = cornersdlvls(orientedNL(1));
		if(weight > cornersdlvls(orientedNL(2))) weight = cornersdlvls(orientedNL(2));
		if(weight > cornersdlvls(orientedNL(5))) weight = cornersdlvls(orientedNL(5));
		if(weight > cornersdlvls(orientedNL(6))) weight = cornersdlvls(orientedNL(6));
	}                                                                                         
	else if (gridnr==17 || gridnr==21 || gridnr==33 || gridnr==37)                     // F1357
	{
		weight = cornersdlvls(orientedNL(1));
		if(weight > cornersdlvls(orientedNL(3))) weight = cornersdlvls(orientedNL(3));
		if(weight > cornersdlvls(orientedNL(5))) weight = cornersdlvls(orientedNL(5));
		if(weight > cornersdlvls(orientedNL(7))) weight = cornersdlvls(orientedNL(7));
	}                                                                                          
	else if (gridnr==26 || gridnr==27 || gridnr==42 || gridnr==43)                     // F3478
	{
		weight = cornersdlvls(orientedNL(3));
		if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4));
		if(weight > cornersdlvls(orientedNL(7))) weight = cornersdlvls(orientedNL(7));
		if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8));
	}                                                                              
	else if (gridnr==20 || gridnr==24 || gridnr==36 || gridnr==40)                     // F2468
	{
		weight = cornersdlvls(orientedNL(2));
		if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4));
		if(weight > cornersdlvls(orientedNL(6))) weight = cornersdlvls(orientedNL(6));
		if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8));
	}                                                                                          
	else if (gridnr== 4 || gridnr== 5 || gridnr== 8 || gridnr== 9)                     // F1234
	{
		weight = cornersdlvls(orientedNL(1));
		if(weight > cornersdlvls(orientedNL(2))) weight = cornersdlvls(orientedNL(2));
		if(weight > cornersdlvls(orientedNL(3))) weight = cornersdlvls(orientedNL(3));
		if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4));
	}                                                                                         
	else if (gridnr==48 || gridnr==49 || gridnr==50 || gridnr==51)                     // F5678
	{
		weight = cornersdlvls(orientedNL(5));
		if(weight > cornersdlvls(orientedNL(6))) weight = cornersdlvls(orientedNL(6));
		if(weight > cornersdlvls(orientedNL(7))) weight = cornersdlvls(orientedNL(7));
		if(weight > cornersdlvls(orientedNL(8))) weight = cornersdlvls(orientedNL(8));
	}                                                                                       
	else                                                                               // bulk
	{
		weight = cornersdlvls(1);                              if(weight > cornersdlvls(2)) weight = cornersdlvls(2);
		if(weight > cornersdlvls(3)) weight = cornersdlvls(3); if(weight > cornersdlvls(4)) weight = cornersdlvls(4);
		if(weight > cornersdlvls(5)) weight = cornersdlvls(5); if(weight > cornersdlvls(6)) weight = cornersdlvls(6);
		if(weight > cornersdlvls(7)) weight = cornersdlvls(7); if(weight > cornersdlvls(8)) weight = cornersdlvls(8);
	}
	return weight;
} 
//
int MeshHex::GetNumberOfElementsForRefinement(int3 pattern, int method)
{
	if(method==0)
	{
		if(pattern(1)==0) return 1;
		if(pattern(1)==1) return 4;
		if(pattern(1)==2 && pattern(2)==0) return 11; 
		//2b
		//2c
		//3a
		//3b
		//3c
		if(pattern(1)==4 && pattern(2)==0) return 22; 
        //4b
		//4c
		//4d
		//4e
		//4f
		//4g
		//5a
		//5b
		//5c
		//6a
		//6b
		//6c
		//7
		if(pattern(1)==8) return 27;
	}
	else if(method==1 || method==2 || method==3) // same as in MeshQad
	{
		if(pattern(1)==0) return  1; 
		if(pattern(1)==1) return  3; 
		if(pattern(1)==2) return  7; 
		if(pattern(1)==3) return  8; 
		if(pattern(1)==4) return  9; 
	}
	return -1;
}
int MeshHex::GetNodeListOfIthRefinedElement(int i, int3 pattern, TArray<int>& nodelist, int method)
{
// lookup tables: 
// !!! all templates are implemented for orientation '0' !!!
// nodes from unrefined have negative number
// additional nodes have positive numbers  
	static const int NLType0[]  = { -1, -2, -3, -4,  -5, -6, -7, -8 };
	static const int NLType1[]  = { -1,  1,  2,  3,   4,  5,  6,  7,    // small cube @ C1
		                             1, -2,  3, -4,   5, -6,  7, -8,    // right frustum "->x"
									 2,  3, -3, -4,   6,  7, -7, -8,    // back frustum  "->y"
									 4,  5,  6,  7,  -5, -6, -7, -8 };  // top frustum   "->z"
	static const int NLType2a[] = { -1,  1,  3,  4,   9, 10, 13, 14, 
		                             1,  2,  4,  5,  10, 11, 14, 15,
									 2, -2,  5,  6,  11, 12, 15, 16,
									10, 11, 14, 15,  17, 18, 19, 20, // elements of "front"								
									 9, 10, 13, 14,  -5, 17, -7, 19,
								    17, 18, 19, 20,  -5, -6, -7, -8,
									11, 12, 15, 16,  18, -6, 20, -8,
									 4,  5,  7,  8,  14, 15, 19, 20, // elements of "bottom"	   
									 3,  4, -3,  7,  13, 14, -7, 19,
									 7,  8, -3, -4,  19, 20, -7, -8,
									 5,  6,  8, -4,  15, 16, 20, -8 };
	static const int NLType4a[] = { -1,  1,  3,  4,   9, 10, 13, 14, // elements of "front"
		                             1,  2,  4,  5,  10, 11, 14, 15,
									 2, -2,  5,  6,  11, 12, 15, 16,
									 9, 10, 13, 14,  19, 20, 23, 24,
								    10, 11, 14, 15,  20, 21, 24, 25,
									11, 12, 15, 16,  21, 22, 25, 26,
									19, 20, 23, 24,  -5, 29, 31, 32,
									20, 21, 24, 25,  29, 30, 32, 33,
									21, 22, 25, 26,  30, -6, 33, 34,
									 3,  4, -3,  7,  13, 14, 17, 37, // elements of left slice
								    13, 14, 17, 37,  23, 24, 27, 39,
									17, 37, -3,  7,  27, 39, -7, 35,
									23, 24, 27, 39,  31, 32, -7, 35,
									 5,  6,  8, -4,  15, 16, 38, 18, // elements of right slice
								    15, 16, 38, 18,  25, 26, 40, 28,
									38, 18,  8, -4,  40, 28, 36, -8,
									25, 26, 40, 28,  33, 34, 36, -8,
									 4,  5,  7,  8,  14, 15, 37, 38, // elements of center slice
									14, 15, 37, 38,  24, 25, 39, 40,
									37, 38,  7,  8,  39, 40, 35, 36,
									24, 25, 39, 40,  32, 33, 35, 36,
									7,  8, -3, -4,  35, 36, -7, -8 };// element at back
	static const int NLType8[]  = { -1,  1,  3,  4,  13, 14, 17, 18, // first layer
		                             1,  2,  4,  5,  14, 15, 18, 19,
									 2, -2,  5,  6,  15, 16, 19, 20,
									 3,  4,  7,  8,  17, 18, 21, 22,
									 4,  5,  8,  9,  18, 19, 22, 23,
									 5,  6,  9, 10,  19, 20, 23, 24,
									 7,  8, -3, 11,  21, 22, 25, 26,
									 8,  9, 11, 12,  22, 23, 26, 27,
									 9, 10, 12, -4,  23, 24, 27, 28,
                                    13, 14, 17, 18,  29, 30, 33, 34, // second layer
									14, 15, 18, 19,  30, 31, 34, 35,
									15, 16, 19, 20,  31, 32, 35, 36,
									17, 18, 21, 22,  33, 34, 37, 38, 
									18, 19, 22, 23,  34, 35, 38, 39,
									19, 20, 23, 24,  35, 36, 39, 40,
									21, 22, 25, 26,  37, 38, 41, 42,
									22, 23, 26, 27,  38, 39, 42, 43,
									23, 24, 27, 28,  39, 40, 43, 44,
									29, 30, 33, 34,  -5, 45, 47, 48, // third layer
									30, 31, 34, 35,  45, 46, 48, 49,
									31, 32, 35, 36,  46, -6, 49, 50,
									33, 34, 37, 38,  47, 48, 51, 52,
									34, 35, 38, 39,  48, 49, 52, 53,
									35, 36, 39, 40,  49, 50, 53, 54,
									37, 38, 41, 42,  51, 52, -7, 55,
									38, 39, 42, 43,  52, 53, 55, 56,
									39, 40, 43, 44,  53, 54, 56, -8 };

	static const int NLType02D[]  = { -1, -2, -3, -4,  -5, -6, -7, -8 };
	static const int NLType12D[]  = { -1,  1, -3,  4,   2,  3,  5,  6,    // Quad -1,1,3,2 -> Front -1,1,*,* 2,3,*,*   *:= +3(for additional) or -2(for original)
	                                   1, -2,  4, -4,   3, -6,  6, -8,    // Quad 1,-2,-3,3 
									   2,  3,  5,  6,  -5, -6, -7, -8 };              
	static const int NLType2a2D[] = { -1,  1, -3,  9,   3,  4, 11, 12,    // Quad -1,1,4,3 -> Front -1,1,*,* 3,4,*,*   *:= +8(for additional) or -2(for original)
		                               1,  2,  9, 10,   4,  5, 12, 13,
									   2, -2, 10, -4,   5,  6, 13, 14,
									   3,  4, 11, 12,  -5,  7, -7, 15,
									   4,  5, 12, 13,   7,  8, 15, 16,
									   5,  6, 13, 14,   8, -6, 16, -8,
									   7,  8, 15, 16,  -5, -6, -7, -8 };
	static const int NLType2b2D[] = { -1,  1, -3,  9,   2,  3, 10, 11,    // Quad -1,1,3,2 -> Front -1,1,*,* 2,3,*,*   *:= +8(for additional) or -2(for original)
		                               1, -2,  9, -4,   3,  4, 11, 12,
									   2,  3, 10, 11,  -5,  5, -7, 13,
									   3,  4, 11, 12,   5,  6, 13, 14,
									   4, -2, 12, -4,   6,  7, 14, 15,
									   5,  6, 13, 14,  -5,  8, -7, 16,
									   6,  7, 14, 15,   8, -6, 16, -8 };
	static const int NLType32D[]  = { -1,  1, -3, 11,   3,  4, 13, 14,    // Quad -1,1,4,3 -> Front -1,1,*,* 3,4,*,*   *:= +10(for additional) or -2(for original)
		                               1,  2, 11, 12,   4,  5, 14, 15,
									   2, -2, 12, -4,   5,  6, 15, 16,
									   3,  4, 13, 14,  -5,  7, -7, 17,
									   4,  5, 14, 15,   7,  8, 17, 18,
									   5,  6, 15, 16,   8,  9, 18, 19,
									   7,  8, 17, 18,  -5, 10, -7, 20,
									   8,  9, 18, 19,  10, -6, 20, -8 };
	static const int NLType42D[]  = { -1,  1, -3, 13,   3,  4, 15, 16,    // Quad -1,1,4,3 -> Front -1,1,*,* 3,4,*,*   *:= +12(for additional) or -2(for original)
		                               1,  2, 13, 14,   4,  5, 16, 17, 
									   2, -2, 14, -4,   5,  6, 17, 18,
									   3,  4, 15, 16,   7,  8, 19, 20,
									   4,  5, 16, 17,   8,  9, 20, 21,
									   5,  6, 17, 18,   9, 10, 21, 22,
									   7,  8, 19, 20,  -5, 11, -7, 23,
									   8,  9, 20, 21,  11, 12, 23, 24,
									   9, 10, 21, 22,  12, -6, 24, -8 };

	int ncorners =    pattern(1);
	int subtype =     pattern(2);
  	int orientation = pattern(3);
	int* rv0 = NULL;
	int ith = 8*(i-1); // offset of 8 - hex has 8 nodes
	if(method==0)
	{
// get correct 8 nodenumbers in orientation 0
		if(ncorners==0) { rv0 = (int*) &NLType0; }
		if(ncorners==1) { rv0 = (int*) &NLType1 + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (int*) &NLType2a + ith; }
		//if(ncorners==2 && subtype==1) { rv0 = (int*) &NLType2b + ith; }
		//if(ncorners==2 && subtype==2) { rv0 = (int*) &NLType2c + ith; }
		//if(ncorners==3 && subtype==0) { rv0 = (int*) &NLType3a + ith; }
		//if(ncorners==3 && subtype==1) { rv0 = (int*) &NLType3b + ith; }
		//if(ncorners==3 && subtype==2) { rv0 = (int*) &NLType3c + ith; }
		if(ncorners==4 && subtype==0) { rv0 = (int*) &NLType4a + ith; }
		//if(ncorners==4 && subtype==1) { rv0 = (int*) &NLType4b + ith; }
		//if(ncorners==4 && subtype==2) { rv0 = (int*) &NLType4c + ith; }
		//if(ncorners==4 && subtype==3) { rv0 = (int*) &NLType4d + ith; }
		//if(ncorners==4 && subtype==4) { rv0 = (int*) &NLType4e + ith; }
		//if(ncorners==4 && subtype==5) { rv0 = (int*) &NLType4f + ith; }
		//if(ncorners==4 && subtype==6) { rv0 = (int*) &NLType4g + ith; }
		//if(ncorners==5 && subtype==0) { rv0 = (int*) &NLType5a + ith; }
		//if(ncorners==5 && subtype==1) { rv0 = (int*) &NLType5b + ith; }
		//if(ncorners==5 && subtype==2) { rv0 = (int*) &NLType5c + ith; }
		//if(ncorners==6 && subtype==0) { rv0 = (int*) &NLType6a + ith; }
		//if(ncorners==6 && subtype==1) { rv0 = (int*) &NLType6b + ith; }
		//if(ncorners==6 && subtype==2) { rv0 = (int*) &NLType6c + ith; }
		//if(ncorners==7) { rv0 = (int*) &NLType7 + ith; }
		if(ncorners==8) { rv0 = (int*) &NLType8 + ith; }
	}
	else if(method==1 || method==2 || method==3)
	{
		if(ncorners==0) { rv0 = (int*) &NLType02D; }
		if(ncorners==1) { rv0 = (int*) &NLType12D + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (int*) &NLType2a2D + ith; }
		if(ncorners==2 && subtype==1) { rv0 = (int*) &NLType2b2D + ith; }
		if(ncorners==3 && subtype==0) { rv0 = (int*) &NLType32D + ith; }
		if(ncorners==4 && subtype==0) { rv0 = (int*) &NLType42D + ith; }
	}

// account for cyclic position of the pattern 
// all additional nodes have been rotated into positon as they were generated
// all nodes of the original hex are still in template orientation - apply permutaion "nodelist"
	nodelist.SetLen(8); nodelist.SetAll(0);
	TArray<int> orientedNL;
	TArray<int> inverseNL;
	TArray<int> intermediateNL; intermediateNL.SetLen(8);
	GetNodeListOriented(orientation,orientedNL,inverseNL);
// 1) insert the correct original nodes into the elements nodelist -> correct nodelist, 
	for(int j=0; j<8; j++)
	{
		if(rv0[j]>0)
		{
			intermediateNL(j+1) = rv0[j];
		}
		else
		{
			intermediateNL(j+1) = -orientedNL(-rv0[j]); // negate input and output - still a negative number is required to mark a node of the unrefined element
		}
	}
// 2) reorient new element to match the orientaiton of the unrefined element
	for(int j=1; j<=8; j++)
	{
		nodelist(j) = intermediateNL(inverseNL(j));
	}

	return 0;
}
Vector3D MeshHex::GetLocPosCenterOfIthRefinedElement(int i, int3 pattern, int method)
{
// lookup tables: 
// NOTE: all self similar child elements have only -2/3, 0 +2/3 entries
//       all different shapes also have -1/3,+1/3,-1/6,+1/6 entries, they are not refined further by the algorithm
// TODO: compute accurate element center for the od shaped elements
	static const double LPCType0[]  = {    0.,    0.,    0. };
	static const double LPCType1[]  = {-2./3.,-2./3.,-2./3.,  // self-similar
		                                1./3.,-2./3.,-2./3.,  // frustum - ? TODO: insert accurate values ? 
							  	 	   -2./3., 1./3.,-2./3.,
							 		   -2./3.,-2./3., 1./3. };
	static const double LPCType2a[] = {-2./3.,-2./3.,-2./3., // s
		                                   0.,-2./3.,-2./3., // s
										2./3.,-2./3.,-2./3., // s
		                                   0.,-2./3.,    0., // s
									   -2./3.,-1./3., 1./3.,
									       0.,-1./3., 2./3.,
									    2./3.,-1./3., 1./3.,
										   0.,    0.,-2./3., // s
									   -2./3., 1./3.,-1./3.,
									       0., 2./3.,-1./3.,
										2./3., 1./3.,-1./3. };
	static const double LPCType4a[] = {-2./3.,-2./3.,-2./3., // s // front 
		                                   0.,-2./3.,-2./3., // s
										2./3.,-2./3.,-2./3., // s 
                                       -2./3.,-2./3.,    0., // s
		                                   0.,-2./3.,    0., // s
										2./3.,-2./3.,    0., // s 
                                       -2./3.,-2./3., 2./3., // s
		                                   0.,-2./3., 2./3., // s
										2./3.,-2./3., 2./3., // s 
									   -2./3.,    0.,-2./3.,      // left slice
									   -2./3.,    0.,    0.,
									   -2./3., 1./3.,    0., 
									   -2./3.,    0., 2./3.,
									    2./3.,    0.,-2./3.,      // right slice
									    2./3.,    0.,    0.,
									    2./3., 1./3.,    0., 
									    2./3.,    0., 2./3.,
										   0.,    0.,-2./3.,     // middle slice
										   0.,-1./6.,    0.,
										   0.,+1./6.,    0.,
										   0.,    0.,+2./3.,  
										   0., 2./3.,    0. };
	static const double LPCType8[] =  {-2./3.,-2./3.,-2./3., // s // front 
		                                   0.,-2./3.,-2./3., // s
										2./3.,-2./3.,-2./3., // s 
                                       -2./3.,    0.,-2./3., // s 
		                                   0.,    0.,-2./3., // s
										2./3.,    0.,-2./3., // s 
                                       -2./3., 2./3.,-2./3., // s 
		                                   0., 2./3.,-2./3., // s
										2./3., 2./3.,-2./3., // s 
                                       -2./3.,-2./3.,    0., // s // middle
		                                   0.,-2./3.,    0., // s
										2./3.,-2./3.,    0., // s 
                                       -2./3.,    0.,    0., // s 
		                                   0.,    0.,    0., // s
										2./3.,    0.,    0., // s 
                                       -2./3., 2./3.,    0., // s 
		                                   0., 2./3.,    0., // s
										2./3., 2./3.,    0., // s 
                                       -2./3.,-2./3., 2./3., // s // back
		                                   0.,-2./3., 2./3., // s
										2./3.,-2./3., 2./3., // s 
                                       -2./3.,    0., 2./3., // s 
		                                   0.,    0., 2./3., // s
										2./3.,    0., 2./3., // s 
                                       -2./3., 2./3., 2./3., // s 
		                                   0., 2./3., 2./3., // s
										   2./3., 2./3., 2./3. }; // s 
// all at y=0
	static const double LPCType02D[]  = {    0., 0.,    0. };
	static const double LPCType12D[]  = {-2./3., 0.,-2./3.,
                                          1./3., 0.,-1./3.,
									     -1./3., 0.,    0. };
	static const double LPCType2a2D[] = {-2./3., 0.,-2./3.,
		                                     0., 0.,-2./3.,
									      2./3., 0.,-2./3.,
									     -2./3., 0., 1./6.,
									         0., 0.,    0.,
									      2./3., 0., 1./6.,
										     0., 0., 2./3. };
	static const double LPCType2b2D[] = {-2./3., 0.,-2./3.,
		                                  1./6., 0.,-2./3.,
									     -2./3., 0., 1./6.,
									         0., 0.,    0.,
									      2./3., 0.,-1./6.,
									     -1./6., 0., 2./3.,
									      2./3., 0., 2./3. };
	static const double LPCType32D[]  = {-2./3., 0.,-2./3.,
		                                     0., 0.,-2./3.,
									      2./3., 0.,-2./3.,
									     -2./3., 0., 1./6.,
									         0., 0.,    0.,
									      2./3., 0.,    0.,
									     -1./6., 0., 2./3.,
									      2./3., 0., 2./3. };
	static const double LPCType42D[]  = {-2./3., 0.,-2./3.,
		                                     0., 0.,-2./3.,
									      2./3., 0.,-2./3.,
									     -2./3., 0.,    0.,
		                                     0., 0.,    0.,
									      2./3., 0.,    0.,
									     -2./3., 0., 2./3.,
		                                     0., 0., 2./3.,
										  2./3., 0., 2./3. };
	int ncorners =    pattern(1);
	int subtype =     pattern(2);
	double* rv0 = NULL;
	int ith = 3*(i-1); // offset of 3 (coordinates)
	if(method==0)
	{
// get correct center coordinates
		if(ncorners==0) { rv0 = (double*) &LPCType0; }
		if(ncorners==1) { rv0 = (double*) &LPCType1 + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (double*) &LPCType2a + ith; }
		//if(ncorners==2 && subtype==1) { rv0 = (double*) &LPCType2b + ith; }
		//if(ncorners==2 && subtype==2) { rv0 = (double*) &LPCType2c + ith; }
		//if(ncorners==3 && subtype==0) { rv0 = (double*) &LPCType3a + ith; }
		//if(ncorners==3 && subtype==1) { rv0 = (double*) &LPCType3b + ith; }
		//if(ncorners==3 && subtype==2) { rv0 = (double*) &LPCType3c + ith; }
		if(ncorners==4 && subtype==0) { rv0 = (double*) &LPCType4a + ith; }
		//if(ncorners==4 && subtype==1) { rv0 = (double*) &LPCType4b + ith; }
		//if(ncorners==4 && subtype==2) { rv0 = (double*) &LPCType4c + ith; }
		//if(ncorners==4 && subtype==3) { rv0 = (double*) &LPCType4d + ith; }
		//if(ncorners==4 && subtype==4) { rv0 = (double*) &LPCType4e + ith; }
		//if(ncorners==4 && subtype==5) { rv0 = (double*) &LPCType4f + ith; }
		//if(ncorners==4 && subtype==6) { rv0 = (double*) &LPCType4g + ith; }
		//if(ncorners==5 && subtype==0) { rv0 = (double*) &LPCType5a + ith; }
		//if(ncorners==5 && subtype==1) { rv0 = (double*) &LPCType5b + ith; }
		//if(ncorners==5 && subtype==2) { rv0 = (double*) &LPCType5c + ith; }
		//if(ncorners==6 && subtype==0) { rv0 = (double*) &LPCType6a + ith; }
		//if(ncorners==6 && subtype==1) { rv0 = (double*) &LPCType6b + ith; }
		//if(ncorners==6 && subtype==2) { rv0 = (double*) &LPCType6c + ith; }
		//if(ncorners==7) { rv0 = (double*) &LPCType7 + ith; }
		if(ncorners==8) { rv0 = (double*) &LPCType8 + ith; }
	}
	else if (method==1 || method==2 || method==3)
	{
		if(ncorners==0) { rv0 = (double*) &LPCType02D; }
		if(ncorners==1) { rv0 = (double*) &LPCType12D + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (double*) &LPCType2a2D + ith; }
		if(ncorners==2 && subtype==1) { rv0 = (double*) &LPCType2b2D + ith; }
		if(ncorners==3) { rv0 = (double*) &LPCType32D + ith; }
		if(ncorners==4) { rv0 = (double*) &LPCType42D + ith; }
	}
	Vector3D ploc0(rv0[0],rv0[1],rv0[2]);
  	int orientation = pattern(3);
	return ApplyOrientation(ploc0,orientation);
}

int MeshHex::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<(-1.-eps) || ploc.X()>(1.+eps)) return 0;
	if(ploc.Y()<(-1.-eps) || ploc.Y()>(1.+eps)) return 0;
	if(ploc.Z()<(-1.-eps) || ploc.Z()>(1.+eps)) return 0;
	return 1;
}
//****************************************************************************************
int MeshTet::Invert()
{
	if(NNodes()==4||NNodes()==10) // corners
	{
		SwapNodeNumberEntries(2,3);
		if(NNodes()==10)
		{
			SwapNodeNumberEntries(5,7);
			SwapNodeNumberEntries(8,9);
			return 10;
		}
		return 4;
	}
	return 0;
}
int MeshTet::GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse)
{ 
	assert(0 && "TODO" );
	return 0;
}
MeshElement MeshTet::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshTrig(1,3,2),
		MeshTrig(1,2,4),
		MeshTrig(1,4,3),
		MeshTrig(2,3,4),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshTrig(1,3,2, 7,6,5),
		MeshTrig(1,2,4, 5,8,10),
		MeshTrig(1,4,3, 10,9,7),
		MeshTrig(2,3,4, 6,9,8),
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 4) { assert(0 && "MeshTet::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 4) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()==10) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshTet::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshTet::GetInterPair(int i)
{
	if(NNodes()==4)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(2,3); break;
		case  3: return int2(3,1); break;
		case  4: return int2(2,4); break;
		case  5: return int2(3,4); break;
		case  6: return int2(1,4); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshTet::GetLocPosOfIthNode(int i)
{
// !!! ATTENTION : Compare with TetrahedralGeneric::GetNodeLocPos !!!
	switch(i)
	{
	//linear:
	case 1: return Vector3D(0., 0., 0.); break;
	case 2: return Vector3D(1., 0., 0.); break;
	case 3: return Vector3D(0., 1., 0.); break;
	case 4: return Vector3D(0., 0., 1.); break;
	//quadratic:
	case 5: return Vector3D(0.5, 0. , 0. ); break;
	case 6: return Vector3D(0.5, 0.5, 0. ); break;
	case 7: return Vector3D(0. , 0.5, 0. ); break;
	case 8: return Vector3D(0.5, 0. , 0.5); break;
	case 9: return Vector3D(0. , 0.5, 0.5); break;
	case 10:return Vector3D(0. , 0. , 0.5); break;
	}
	return Vector3D(0.);
}
double MeshTet::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();

	if (NNodes() == 4)
	{
		switch(shape)
		{
		case 1: return 1. - r - s - t; break;
		case 2: return r; break;
		case 3: return s; break;
		case 4: return t; break;
		default: return 0;
		}
	}
	else if(NNodes() == 10)
	{

		switch(shape)
		{
			case 1: return 2.0*(1.0-r-s-t)*(0.5-r-s-t); break;
			case 2: return 2.0*r*(r-0.5); break;
			case 3: return 2.0*s*(s-0.5); break;
			case 4: return 2.0*t*(t-0.5);
			case 5: return 4.0*r*(1.0-r-s-t); break;
			case 6: return 4.0*r*s; break;
			case 7: return 4.0*s*(1.0-r-s-t); break;
			case 8: return 4.0*r*t; break;
			case 9: return 4.0*s*t; break;
			case 10: return 4.0*t*(1.0-r-s-t); break; 
			default: return 0;
		}

	}
	assert(0 && "Wrong number of nodes or wrong index.");
	return 0;
}
Matrix MeshTet::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if (NNodes() == 4)
	{
		sm(1,1) = -1.0;
		sm(1,2) = 1.0;
		sm(1,3) = 0.0;
		sm(1,4) = 0.0;
		sm(2,1) = -1.0;
		sm(2,2) = 0.0;
		sm(2,3) = 1.0;
		sm(2,4) = 0.0;
		sm(3,1) = -1.0;
		sm(3,2) = 0.0;
		sm(3,3) = 0.0;
		sm(3,4) = 1.0;
	}
	else if(NNodes() == 10)
	{
		sm(1,1) = 4.0*(-0.75+r+s+t);
		sm(1,2) = 4*(r-0.25);
		sm(1,3) = 0.;
		sm(1,4) = 0.;
		sm(1,5) = 4.0-8.0*r-4.0*s-4.0*t;
		sm(1,6) = 4.0*s;
		sm(1,7) = -4.0*s;
		sm(1,8) = 4.0*t;
		sm(1,9) = 0.0;
		sm(1,10) = -4.0*t;
		sm(2,1) = 4.0*(-0.75+r+s+t);
		sm(2,2) = 0.;
		sm(2,3) = 4*(s-0.25);
		sm(2,4) = 0.;
		sm(2,5) = -4.0*r;
		sm(2,6) = 4.0*r;
		sm(2,7) = 4.0-4.0*r-8.0*s-4.0*t;
		sm(2,8) = 0.0;
		sm(2,9) = 4.0*t;
		sm(2,10) = -4.0*t;
		sm(3,1) = 4.0*(-0.75+r+s+t);
		sm(3,2) = 0.;
		sm(3,3) = 0.;
		sm(3,4) = 4*(t-0.25);
		sm(3,5) = -4.0*r;
		sm(3,6) = 0.0;
		sm(3,7) = -4.0*s;
		sm(3,8) = 4.0*r;
		sm(3,9) = 4.0*s;
		sm(3,10) = 4.0-4.0*r-4.0*s-8.0*t;
	}
	return sm;
}

int MeshTet::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<-eps || ploc.X()>(1.+eps)) return 0;
	if(ploc.Y()<-eps || ploc.Y()>(1.+eps)) return 0;
	if(ploc.Z()<-eps || ploc.Z()>(1.+eps)) return 0;
	if( (ploc.X()+ploc.Y()+ploc.Z())>(1.+eps) ) return 0; //test face {N2,N3,N4}
	return 1;
}
//****************************************************************************************
int MeshPyr::Invert()
{
	if(NNodes()==5||NNodes()==13) // corners
	{
		SwapNodeNumberEntries(1,4);
		if(NNodes()==13)
		{
			SwapNodeNumberEntries(10,13);
			SwapNodeNumberEntries( 7, 8);
			SwapNodeNumberEntries( 9, 6);
			return 13;
		}
		return 5;
	}
	return 0;
}
int MeshPyr::GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse)
{ 
	assert(0 && "TODO" );
	return 0;
}
MeshElement MeshPyr::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshQuad(1,3,4,2),
		MeshTrig(1,2,5),
		MeshTrig(2,4,5),
		MeshTrig(3,5,4),
		MeshTrig(1,5,3),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshQuad(1,3,4,2, 8,7,9,6),
		MeshTrig(1,2,5,   6,11,10),
		MeshTrig(2,4,5,   9,13,11),
		MeshTrig(3,5,4,  12,13,7), 
		MeshTrig(1,5,3,  10,12,8), 
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 5) { assert(0 && "MeshPyr::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 5) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()==13) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshPyr::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshPyr::GetInterPair(int i)
{
	if(NNodes()==5)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(3,4); break; //return int2(2,4); break;
		case  3: return int2(1,3); break; //return int2(4,3); break;
		case  4: return int2(2,4); break; return int2(3,1); break;
		case  5: return int2(1,5); break;
		case  6: return int2(2,5); break;
		case  7: return int2(3,5); break;
		case  8: return int2(4,5); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshPyr::GetLocPosOfIthNode(int i)
{
// !!! ATTENTION : Compare with PyramidGeneric::GetNodeLocPos !!!
	switch(i)
	{
	//linear:
	case 1: return Vector3D(0., 0., 0.); break;
	case 2: return Vector3D(1., 0., 0.); break;
	case 3: return Vector3D(0., 1., 0.); break;
	case 4: return Vector3D(1., 1., 0.); break;
	case 5: return Vector3D(0., 0., 1.); break;
	//quadratic:
	case 6: return Vector3D(0.5, 0., 0.); break;
	case 7: return Vector3D(0.5, 1. , 0. ); break;
	case 8: return Vector3D(0., 0.5, 0. ); break;
	case 9: return Vector3D(1., 0.5, 0. ); break;
	case 10: return Vector3D(0., 0., 0.5); break;
	case 11: return Vector3D(0.5, 0., 0.5); break;
	case 12: return Vector3D(0., 0.5, 0.5); break;
	case 13: return Vector3D(0.5, 0.5, 0.5); break;
	}
	return Vector3D(0.);
}
double MeshPyr::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	if (NNodes() == 5)
	{
		switch(shape)
		{
			case 1: return (1-r)*(1-s) + t*(min(r,s)-1); break;
			case 2: return r*(1-s) - t*min(r,s); break;
			case 3: return (1-r)*s - t*min(r,s); break;
			case 4: return r*s + t*min(r,s); break;
			case 5: return t; break;
			default: return 0;
		}
	}
	else if (NNodes() == 13) 
	{
		switch(shape)
		{
			case 1: return ((1-r)*(1-s) + t*(min(r,s) - 1))*(1 - 2*r - 2*s - 2*t); break;
			case 2: return (r*(1 - s) - t*min(r,s))*(2*r - 2*s - 1); break;
			case 3: return ((1-r)*s - t*min(r,s))*(2*s - 2*r - 1); break;
			case 4: return (r*s + t*min(r,s))*(2*r + 2*s + 2*t - 3); break;		
			case 5: return t*(2*t-1); break;
			case 6:
				if (r > s)
				{	return 4*r*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1-r-t); }
				else
				{ return 4*r*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1-s-t); }
				break;
			case 7: 
				if (r > s)
				{ return 4*r*((1-r)*s - t*s) + 2*s*t*(1-r-t); }
				else
				{ return 4*r*((1-r)*s - t*r) + 2*r*t*(1-s-t); }
				break;
			case 8:
				if (r > s) 
				{ return 4*s*((1-r)*(1-s) + t*(s-1)) - 2*s*t*(1 - r - t); }
				else
				{ return 4*s*((1-r)*(1-s) + t*(r-1)) - 2*r*t*(1- s - t); }
				break;
			case 9: 
				if (r > s)
				{ return 4*s*(r*(1-s) - t*s) + 2*s*t*(1 - r - t); }
				else 
				{ return 4*s*(r*(1-s) - t*r) + 2*r*t*(1 - s - t); }
				break;
			case 10: return 4*t*((1-r)*(1-s) + t*(min(r,s) - 1)); break;
			case 11: return 4*t*(r*(1-s) - t*min(r,s)) ; break;
			case 12: return 4*t*((1 - r)*s - t*min(r,s)); break;
			case 13: return 4*t*(r*s + t*min(r,s)) ; break;
			default: return 0;
		}		
	}
	return 0;
}
Matrix MeshPyr::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if (NNodes() == 5)
	{
		if (r > s)
		{
			//(1-r)*(1-s) + t*(min(r,s)-1)
			sm(1,1) = s - 1;
			sm(2,1) = r + t - 1;
			sm(3,1) = s - 1;
			//r*(1-s) - t*min(r,s)
			sm(1,2) = 1 - s;
			sm(2,2) = -r - t;
			sm(3,2) = -s;
			//(1-r)*s - t*min(r,s)
			sm(1,3) = -s;
			sm(2,3) = 1 - r - t;
			sm(3,3) = -s;
			//r*s + t*min(r,s)
			sm(1,4) = s;
			sm(2,4) = r + t;
			sm(3,4) = s;
		}
		else
		{
			//(1-r)*(1-s) + t*(min(r,s)-1)
			sm(1,1) = s + t - 1;
			sm(2,1) = r - 1;
			sm(3,1) = r - 1;
			//r*(1-s) - t*min(r,s)
			sm(1,2) = 1 - s - t;
			sm(2,2) = -r;
			sm(3,2) = -r;
			//(1-r)*s - t*min(r,s)
			sm(1,3) = -s - t;
			sm(2,3) = 1 - r;
			sm(3,3) = -r;
			//r*s + t*min(r,s)
			sm(1,4) = s + t;
			sm(2,4) = r;
			sm(3,4) = r;
		}
		//t
		sm(1,5) = 0.;
		sm(2,5) = 0.;
		sm(3,5) = 1.;
	}
	else if(NNodes() == 13)
	{
		if (r > s)
		{
			sm(1,1) = (1 - s)*(-3 + 4*r + 2*s + 4*t) ;
			sm(2,1) = (1 - r - t)*(-3 + 2*r + 4*s + 2*t);

			sm(1,2) = 2*(r*(1 - s) - t*min(r, s)) + (1 - 2*r + 2*s)*(-1 + s);
			sm(2,2) = r*(4*s - 1 - 2*r) + 2*t*min(r, s) + (1 - 2*r + 2*s)*t ;
			
			sm(1,3) = 2*((r - 1)*s + t*min(r, s)) + (1 + 2*r - 2*s)*s;
			sm(2,3) = 2*((1 - r)*s - t*min(r, s)) + (1 + 2*r - 2*s)*(r + t - 1);

			sm(1,4) = s*(4*r + 2*s + 4*t - 3);
			sm(2,4) = (r + t)*(2*r + 4*s + 2*t - 3);

			sm(1,6) = 8*r*(s - 1) + 6*s*t - 4*(s + t - 1);
			sm(2,6) = 2*(r + t - 1)*(2*r + t);
			sm(3,6) = 4*s*t - 4*r - 2*s + 6*r*s;

			sm(1,7) = 2*s*(2 - 4*r - 3*t);
			sm(2,7) = 2*(1 - r - t)*(2*r + t);
			sm(3,7) = 2*s*(1 - 3*r - 2*t);

			sm(1,8) = 2*s*(2*s + t - 2);
			sm(2,8) = 2*(r + t - 1)*(4*s + t - 2);
			sm(3,8) = 2*s*(r + 2*s + 2*t - 3);

			sm(1,9) = 2*s*(2 - 2*s - t);
			sm(2,9) = 2*(r*(2 - 4*s - t) + t*(1 - 4*s - t));
			sm(3,9) = 2*s*(1 - r - 2*s - 2*t);

			sm(1,10) = 4*t*(s - 1);
			sm(2,10) = 4*t*(r + t - 1);

			sm(1,11) = 4*t*(1 - s);
			sm(2,11) = -4*t*(r + t);

			sm(1,12) = -4*t*s;
			sm(2,12) = 4*t*(1 - r - t);

			sm(1,13) = 4*t*s;
			sm(2,13) = 4*t*(r + t);
		}
		else
		{
			sm(1,1) = (1 - s - t)*(-3 + 4*r + 2*s + 2*t);
			sm(2,1) = (1 - r)*(-3 + 2*r + 4*s + 4*t);

			sm(1,2) = 2*(r*(1 - s) - t*min(r, s)) + (1 - 2*r + 2*s)*(-1 + s + t);
			sm(2,2) = r*(4*s - 1 - 2*r) + 2*t*min(r, s);

			sm(1,3) = 2*((r - 1)*s + t*min(r, s)) + (1 + 2*r - 2*s)*(s + t);
			sm(2,3) = 2*((1 - r)*s - t*min(r, s)) + (1 + 2*r - 2*s)*(r - 1);

			sm(1,4) = (s + t)*(4*r + 2*s + 2*t - 3);
			sm(2,4) = r*(2*r + 4*s + 4*t - 3);

			sm(1,6) = 2*(4*r + t - 2)*(s + t - 1);
			sm(2,6) = 2*r*(2*r + t - 2);
			sm(3,6) = 2*r*(2*r + s + 2*t - 3);

			sm(1,7) = 2*(s*(2 - 4*r - t) + t*(1 - 4*r - t));
			sm(2,7) = 2*r*(2 - 2*r - t);
			sm(3,7) = 2*r*(1 - 2*r - s - 2*t);

			sm(1,8) =  2*(s + t - 1)*(2*s + t);
			sm(2,8) = 4*(1 - 2*s - t) + r*(8*s + 6*t - 4);
			sm(3,8) = r*(6*s + 4*t - 2) - 4*s;

			sm(1,9) = 2*(1 - s - t)*(2*s + t);
			sm(2,9) = 2*r*(2 - 4*s - 3*t);
			sm(3,9) = 2*r*(1 - 3*s - 2*t);

			sm(1,10) = 4*t*(s + t - 1);
			sm(2,10) = 4*t*(r - 1);

			sm(1,11) = 4*t*(1 - s - t);
			sm(2,11) = -4*t*r;

			sm(1,12) = -4*t*(s + t);
			sm(2,12) = 4*t*(1 - r);

			sm(1,13) = 4*t*(s + t);
			sm(2,13) = 4*t*r;
		}
		double minrs = min(r, s);

		sm(3,1) = -3 - 2*r*(s - 2) + 4*s + 4*t + (1 - 2*r - 2*s - 4*t)*minrs;
		sm(3,2) = (1 - 2*r + 2*s)*minrs;
		sm(3,3) = (1 + 2*r - 2*s)*minrs;
		sm(3,4) = 2*r*s + (2*r + 2*s + 4*t - 3)*minrs;
		
		sm(1,5) = 0.;
		sm(2,5) = 0.;
		sm(3,5) = 4*t - 1;

		sm(3,10) = 4*(r - 1)*(s - 1) + 8*t*(minrs - 1);
		sm(3,11) = 4*r*(1 - s) - 8*t*minrs;
		
		sm(3,12) = 4*s*(1 - r) - 8*t*minrs;
		sm(3,13) = 4*r*s + 8*t*minrs;
	}
	return sm;
}

int MeshPyr::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<-eps || ploc.X()>(1+eps)) return 0;
	if(ploc.Y()<-eps || ploc.Y()>(1+eps)) return 0;
	if(ploc.Z()<-eps || ploc.Z()>(1+eps)) return 0;
	if( (ploc.X()+ploc.Z())>(1.+eps) ) return 0; // test plane containing face {2,4,5} & cutting {3,4,5}
	if( (ploc.Y()+ploc.Z())>(1.+eps) ) return 0; // test plane containing face {3,4,5} & cutting {2,4,5}
	return 1;
}
//****************************************************************************************
int MeshPri::Invert()
{
	if(NNodes()==6||NNodes()==15) // corners
	{
		SwapNodeNumberEntries(1,4);
		SwapNodeNumberEntries(2,5);
		SwapNodeNumberEntries(3,6);
		if(NNodes()==15)
		{
			SwapNodeNumberEntries( 7,13);
			SwapNodeNumberEntries( 8,14);
			SwapNodeNumberEntries( 9,15);
			return 15;
		}
		return 6;
	}
	return 0;
}
int MeshPri::GetNodeListOriented(int orientation, TArray<int>& nodelist, TArray<int>& inverse)
{ 
	assert(0 && "TODO" );
	return 0;
}
MeshElement MeshPri::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshQuad(1,2,5,4),
		MeshQuad(1,4,6,3),
		MeshQuad(2,3,6,5),
		MeshTrig(1,3,2),
		MeshTrig(4,5,6),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshQuad(1,2,5,4, 7,11,13,10),
		MeshQuad(1,4,6,3, 10,15,12,9),
		MeshQuad(2,3,6,5, 8,12,14,11),
		MeshTrig(1,3,2,   9,8,7),
		MeshTrig(4,5,6,   13,14,15),
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 5) { assert(0 && "MeshPri::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 6) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()==15) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshPri::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshPri::GetInterPair(int i)
{
	if(NNodes()==6)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(2,3); break;
		case  3: return int2(3,1); break;
		case  4: return int2(1,4); break;
		case  5: return int2(2,5); break;
		case  6: return int2(3,6); break;
		case  7: return int2(4,5); break;
		case  8: return int2(5,6); break;
		case  9: return int2(6,4); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshPri::GetLocPosOfIthNode(int i)
{
// !!! ATTENTION : Compare with PrismGeneric::GetNodeLocPos !!!
	switch(i)
	{
	//linear:
	case 1: return Vector3D(0., 0., 0.); break;
	case 2: return Vector3D(1., 0., 0.); break;
	case 3: return Vector3D(0., 1., 0.); break;
	case 4: return Vector3D(0., 0., 1.); break;
	case 5: return Vector3D(1., 0., 1.); break;
	case 6: return Vector3D(0., 1., 1.); break;
	//quadratic:
	case 7: return Vector3D(0.5, 0. , 0. ); break;
	case 8: return Vector3D(0.5, 0.5, 0. ); break;
	case 9: return Vector3D(0., 0.5, 0. ); break;
	case 10: return Vector3D(0., 0., 0.5); break;
	case 11: return Vector3D(1., 0., 0.5); break;
	case 12: return Vector3D(0., 1., 0.5); break;
	case 13: return Vector3D(0.5, 0., 1.); break;
	case 14: return Vector3D(0.5, 0.5, 1.); break;
	case 15: return Vector3D(0., 0.5, 1.); break;
	}
	return Vector3D(0.);
}
double MeshPri::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	if (NNodes() == 6) //linear
	{
		switch(shape)
		{
		case 1: return (1. - t)*(1. - r - s); break;
		case 2: return (1. - t)*r; break;
		case 3: return (1. - t)*s; break;
		case 4: return t*(1. - r - s); break;
		case 5: return t*r; break;
		case 6: return t*s; break;
		}
	}
	else if (NNodes() == 15) 
	{
		switch(shape)
		{
		case 1: return (1 - r - s)*(1 - 2*r - 2*s)*(1-t)*(1-2*t); break;
		case 2: return (2*r-1)*r*(1-t)*(1-2*t); break;
		case 3: return (2*s-1)*s*(1-t)*(1-2*t); break;
		case 4: return (1-r-s)*(1-2*r-2*s)*t*(2*t-1); break;
		case 5: return (2*r-1)*r*t*(2*t-1); break;
		case 6: return (2*s-1)*s*t*(2*t-1); break;
		case 7: return 4 * r * (1 - r - s) * (1 - t)*(1-2*t); //node 7
		case 8: return 4 * r*s*(1 - t)*(1-2*t); //node 8
		case 9: return 4 * s * (1 - r - s) * (1 - t)*(1-2*t); //node 9
		case 10: return 4*(1-r-s)*t*(1-t); //node 10
		case 11: return 4*r*t*(1-t); //node 11
		case 12: return 4*s*t*(1-t); //node 12
		case 13: return 4 * (1-r-s)*r*t*(2*t-1); //node 13
		case 14: return 4 * r*s*t*(2*t-1); //node 14
		case 15: return 4 * (1-r-s)*s*t*(2*t-1); //node 15
		}
	}
	//$EK 2017-01-11: error handling added
	assert(0 && "MeshPri::GetS0 - Wrong number of nodes or wrong index.");
	return 0;
}
Matrix MeshPri::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	double s = ploc.Y();
	double t = ploc.Z();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if (NNodes() == 6)
	{
		sm(1,1) = t - 1.;
		sm(1,2) = 1. - t;
		sm(1,3) = 0.;
		sm(1,4) = -t;
		sm(1,5) = t;
		sm(1,6) = 0.;
		sm(2,1) = t - 1.;
		sm(2,2) = 0.0;
		sm(2,3) = 1. - t;
		sm(2,4) = -t;
		sm(2,5) = 0.;
		sm(2,6) = t;
		sm(3,1) = r + s - 1.;
		sm(3,2) = -r;
		sm(3,3) = -s;
		sm(3,4) = 1. - r - s;
		sm(3,5) = r;
		sm(3,6) = s;
	}
	else if(NNodes() == 15)
	{
		//case 1: return (1 - r - s)*(1 - 2*r - 2*s)*(1-t)*(1-2*t); break;
		sm(1,1) = (-3 + 4*r + 4*s)*(t-1)*(2*t-1);
		sm(2,1) = (-3 + 4*r + 4*s)*(t-1)*(2*t-1);
		sm(3,1) = (-1 + r + s)*(-1+2*r+2*s)*(4*t-3);
		//case 2: return (2*r-1)*r*(1-t)*(1-2*t); break;
		sm(1,2) = (4*r-1)*(1-t)*(1-2*t); 
		sm(2,2) = 0;
		sm(3,2) = (2*r-1)*r*(4*t-3);
		//case 3: return (2*s-1)*s*(1-t)*(1-2*t); break;
		sm(1,3) = 0.0;
		sm(2,3) = (4*s-1)*(1-t)*(1-2*t);
		sm(3,3) = (2*s-1)*s*(4*t-3);
		//case 4: return (1-r-s)*(1-2*r-2*s)*t*(2*t-1); break;
		sm(1,4) = (-3+4*r+4*s)*t*(2*t-1);
		sm(2,4) = (-3+4*r+4*s)*t*(2*t-1);
		sm(3,4) = (1-r-s)*(1-2*r-2*s)*(4*t-1);
		//case 5: return (2*r-1)*r*t*(2*t-1); break;
		sm(1,5) = (4*r-1)*t*(2*t-1);
		sm(2,5) = 0;
		sm(3,5) = (2*r-1)*r*(4*t-1);
		//case 6: return (2*s-1)*s*t*(2*t-1); break;
		sm(1,6) = 0;
		sm(2,6) = (4*s-1)*t*(2*t-1);
		sm(3,6) = (2*s-1)*s*(4*t-1);
		//case 7: return 4 * r * (1 - r - s) * (1 - t)*(1-2*t); //node 7
		sm(1,7) = 4*(1-2*r-s)*(1-t)*(1-2*t);
		sm(2,7) = -4*r*(1-t)*(1-2*t);
		sm(3,7) = 4*(1-r-s)*r*(4*t-3);	
		//case 8: return 4 * r*s*(1 - t)*(1-2*t); //node 8
		sm(1,8) = 4*s*(1-t)*(1-2*t);
		sm(2,8) = 4*r*(1-t)*(1-2*t);
		sm(3,8) = 4*r*s*(4*t-3);
		//case 9: return 4 * s * (1 - r - s) * (1 - t)*(1-2*t); //node 9
		sm(1,9) = -4*s*(1-t)*(1-2*t);
		sm(2,9) = 4*(1-r-2*s)*(1-t)*(1-2*t);
		sm(3,9) = 4*(1-r-s)*s*(4*t-3);
		//case 10: return 4*(1-r-s)*t*(1-t); //node 10
		sm(1,10) = -4*t*(1-t);
		sm(2,10) = -4*t*(1-t);
		sm(3,10) = 4*(1-r-s)*(1-2*t);
		// case 11: return 4*r*t*(1-t); //node 11
		sm(1,11) = 4*t*(1-t);
		sm(2,11) = 0;
		sm(3,11) = 4*r*(1-2*t);
		//case 12: return 4*s*t*(1-t); //node 12
		sm(1,12) = 0; 
		sm(2,12) = 4*t*(1-t); 
		sm(3,12) = 4*s*(1-2*t); 
		//case 13: return 4 * (1-r-s)*r*t*(2*t-1); //node 13
		sm(1,13) = 4*(1-2*r-s)*t*(2*t-1);
		sm(2,13) = -4*r*t*(2*t-1);
		sm(3,13) = 4*(1-r-s)*r*(4*t-1);
		//case 14: return 4 * r*s*t*(2*t-1); //node 14
		sm(1,14) = 4*s*t*(2*t-1);
		sm(2,14) =4*r*t*(2*t-1);
		sm(3,14) = 4*r*s*(4*t-1);	
		//case 15: return 4 * (1-r-s)*s*t*(2*t-1); //node 15
		sm(1,15) = -4*s*t*(2*t-1);
		sm(2,15) = 4*(1-r-2*s)*t*(2*t-1);
		sm(3,15) = 4*(1-r-s)*s*(4*t-1);	
	}
	//$EK 2017-01-11: error handling added
	else
	{
		assert(0 && "MeshPri::GetDSMatrix0 - Wrong number of nodes.");
	}
	return sm;
}

int MeshPri::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<-eps || ploc.X()>(1.+eps)) return 0;
	if(ploc.Y()<-eps || ploc.Y()>(1.+eps)) return 0;
	if(ploc.Z()<-eps || ploc.Z()>(1.+eps)) return 0;
	if( (ploc.X()+ploc.Y())>(1.+eps) ) return 0; // test face {2,3,6,5}
	return 1;
}
//****************************************************************************************
// 2D Mesh Elements
//****************************************************************************************
int MeshQuad::Invert()
{
	if(NNodes()==4||NNodes()==9) // corners
	{
		SwapNodeNumberEntries(2,4);
		if(NNodes()==9)
		{
			SwapNodeNumberEntries( 5, 8);
			SwapNodeNumberEntries( 6, 7);
			return 9;
		}
		return 4;
	}
	return 0;
}
int MeshQuad::GetNodeListOriented(int orientation,TArray<int>& nodelist, TArray<int>& inverse)
{
	nodelist.Flush(); inverse.Flush();
	switch(orientation)
	{
		case 0: nodelist.Set4(1,2,3,4); inverse.Set4(1,2,3,4); break;
		case 1: nodelist.Set4(2,3,4,1); inverse.Set4(4,1,2,3); break;
		case 2: nodelist.Set4(3,4,1,2); inverse.Set4(3,4,1,2); break;
		case 3: nodelist.Set4(4,1,2,3); inverse.Set4(2,3,4,1); break;
	}
	return nodelist.Length();
}
MeshElement MeshQuad::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshLine(1,2),
		MeshLine(2,3),
		MeshLine(3,4),
		MeshLine(4,1),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshLine(1,2,5),
		MeshLine(2,3,6),
		MeshLine(3,4,7),
		MeshLine(4,1,8),
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 4) { assert(0 && "MeshQuad::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 4) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()== 8) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshQuad::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshQuad::GetInterPair(int i)
{
	if(NNodes()==4)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(2,3); break;
		case  3: return int2(3,4); break;
		case  4: return int2(4,1); break;
		case  5: return int2(1,3); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshQuad::GetLocPosOfIthNode(int i)
{
	// compare to QuadrilateralGeneric::GetNodePosLoc2D
	//$SW 25.03.2020: Bugfix, wrong order of nodes
	switch(i)
	{
	//linear:
	case 3: return Vector3D(-1.,-1., 0.);
	case 4: return Vector3D( 1.,-1., 0.);
	case 1: return Vector3D( 1., 1., 0.);
	case 2: return Vector3D(-1., 1., 0.);
	case 7: return Vector3D( 0.,-1., 0.);
	case 8: return Vector3D( 1., 0., 0.);
	case 5: return Vector3D( 0., 1., 0.);
	case 6: return Vector3D(-1., 0., 0.);
	case 9: return Vector3D( 0., 0., 0.);
	}
	return Vector3D(0.);
}
double MeshQuad::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();
	double s = ploc.Y();

	if(NNodes() == 4)
	{
		switch(shape)
		{
		case 1:	return 0.25*(1+r)*(1+s);
		case 2:	return 0.25*(1-r)*(1+s);			
		case 3:	return 0.25*(1-r)*(1-s);
		case 4:	return 0.25*(1+r)*(1-s);
		}
	}
	else
	{
		double r2 = r*r;
		double s2 = s*s;

		switch (shape)
		{
		case 1: return (1.0 + r)*(1.0 + s)*(r + s - 1.0) / 4.0; break;
		case 2: return (1.0 - r)*(1.0 + s)*(-r + s - 1.0) / 4.0; break;
		case 3: return (1.0 - r)*(1.0 - s)*(-r - s - 1.0) / 4.0; break;
		case 4: return (1.0 + r)*(1.0 - s)*(r - s - 1.0) / 4.0; break;
		case 5: return (1.0 - r * r)*(1.0 + s) / 2.0; break;
		case 6: return (1.0 - s * s)*(1.0 - r) / 2.0; break;
		case 7: return (1.0 - r * r)*(1.0 - s) / 2.0; break;
		case 8: return (1.0 - s * s)*(1.0 + r) / 2.0; break;
		}
	}
	assert(0);
	return 0;
}
Matrix MeshQuad::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	double s = ploc.Y();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if(NNodes() == 4)
	{
		sm(1,1) = (1.0+s)/4.0;
		sm(1,2) = -(1.0+s)/4.0;
		sm(1,3) = -(1.0-s)/4.0;
		sm(1,4) = (1.0-s)/4.0;
		sm(2,1) = (1.0+r)/4.0;
		sm(2,2) = (1.0-r)/4.0;
		sm(2,3) = -(1.0-r)/4.0;
		sm(2,4) = -(1.0+r)/4.0;
	}
	else if(NNodes() == 8)
	{
		sm(1,1) =  r*(0.5 + 0.5*s) + (0.25 + 0.25*s)*s;
		sm(1,2) =  r*(0.5 + 0.5*s) + (-0.25 - 0.25*s)*s;
		sm(1,3) =  r*(0.5 - 0.5*s) + (0.25 - 0.25*s)*s;
		sm(1,4) =  r*(0.5 - 0.5*s) + (-0.25 + 0.25*s)*s;
		sm(1,5) = -r*(1.0+s);
		sm(1,6) =  (-1.0+s*s)/2.0;  
		sm(1,7) =  r*(s-1.0);
		sm(1,8) =  (1.0-s*s)/2.0;
		sm(2,1) =  r*( 0.25 + 0.25*r + 0.5*s) + 0.5*s;
		sm(2,2) =  r*(-0.25 + 0.25*r - 0.5*s) + 0.5*s;
		sm(2,3) =  r*( 0.25 - 0.25*r - 0.5*s) + 0.5*s;
		sm(2,4) =  r*(-0.25 - 0.25*r + 0.5*s) + 0.5*s;
		sm(2,5) = (1.0-r*r)/2.0;
		sm(2,6) = -s*(1.0-r);
		sm(2,7) = (-1.0+r*r)/2.0;
		sm(2,8) = -s*(1.0+r);
	}
	else if(NNodes() == 9)
	{
		double r2 = r*r;
		double s2 = s*s;
		//D_sf/D_r:
		sm(1,1) =  s/4.0+r*s/2.0+s2*r/2.0+s2/4.0;
		sm(1,2) =  -s/4.0+r*s/2.0+s2*r/2.0-s2/4.0;
		sm(1,3) =  s/4.0-s2/4.0+s2*r/2.0-r*s/2.0;
		sm(1,4) =  -s/4.0-r*s/2.0+s2*r/2.0+s2/4.0;
		sm(1,5) =  -r*s-s2*r;
		sm(1,6) =  -1.0/2.0+s2/2.0+r-s2*r;
		sm(1,7) =  r*s-s2*r;
		sm(1,8) =  1.0/2.0-s2/2.0+r-s2*r;
		sm(1,9) =  2.0*r*(-1.0+s2);
		//D_sf/D_s:
		sm(2,1) =  r/4.0+r2/4.0+r2*s/2.0+r*s/2.0;
		sm(2,2) =  -r/4.0+r2/4.0+r2*s/2.0-r*s/2.0;
		sm(2,3) =  r/4.0-r*s/2.0+r2*s/2.0-r2/4.0;
		sm(2,4) =  -r/4.0-r2/4.0+r2*s/2.0+r*s/2.0;
		sm(2,5) =  1.0/2.0-r2/2.0+s-r2*s;
		sm(2,6) =  r*s-r2*s;
		sm(2,7) =  -1.0/2.0+r2/2.0+s-r2*s;
		sm(2,8) =  -r*s-r2*s;
		sm(2,9) =  2.0*(-1.0+r2)*s;
	}
	return sm;
}


int3 MeshQuad::IdentifyRequiredTemplate(TArray<int>& lvl, int method)
{
// PB   0     // PB   1      2      4      8     // PB   3      6     12      9      5     10
//    O---O	  //    O---O  O---O  O---*  *---O	 //    O---O  O---*  *---*  *---O  O---*  *---O 
//    |   |	  //    |   |  |   |  |   |  |   |	 //    |   |  |   |	 |   |  |   |  |   |  |   |	
//    O---O	  //    *---O  O---*  O---O  O---O	 //    *---*  O---*  O---O  *---O  *---O  O---*	
// RV 0,0,0	  // RV 1,0,0  1,0,1  1,0,2  1,0,3	 // RV 2,0,0  2,0,1  2,0,2  2,0,3  2,1,0  2,1,1
//
// PB   7     14     13     11     // PB  15
//    O---*  *---*  *---*  *---O   //    *---*
//    |   |  |   |	|   |  |   |   //    |   |
//    *---*  O---*	*---O  *---*   //    *---*
// RV 3,0,0  3,0,1  3,0,2  3,0,3   // RV 4,0,0


// templates for method 0 - 
	static const int3 pattern2template[] = {
		int3(0,0,0),int3(1,0,0),int3(1,0,1),int3(2,0,0), //  0.. 3
		int3(1,0,2),int3(2,1,0),int3(2,0,1),int3(3,0,0), //  4.. 7
		int3(1,0,3),int3(2,0,3),int3(2,1,1),int3(3,0,3), //  8..11
		int3(2,0,2),int3(3,0,2),int3(3,0,1),int3(4,0,0)  // 12..15
	};

//	if(method==0)
//	{
		int nactive = 0;
		int patternbinary = 0;
		if(lvl(1)>0) { nactive++; patternbinary += 1; }
		if(lvl(2)>0) { nactive++; patternbinary += 2; }
		if(lvl(3)>0) { nactive++; patternbinary += 4; }
		if(lvl(4)>0) { nactive++; patternbinary += 8; }
//	}
	return pattern2template[patternbinary];
}
Vector3D MeshQuad::ApplyOrientation(Vector3D vec0, int orientation)
{
// do transformation such that the local position of the new node is correctly oriented
// orientation verified with Corner-configurations and colored additional nodes
	Matrix3D A = GetOrientationMatrix(orientation);
	Vector3D locposoriented = A * vec0;                                                                                      
	return locposoriented;
}
Matrix3D MeshQuad::GetOrientationMatrix(int orientation)
{
	static const double rotations[] = { 1., 0., 0.,   0., 1., 0.,   0., 0., 1.,   // O 0
		                                0.,-1., 0.,   1., 0., 0.,   0., 0., 1.,   // O 1
                                       -1., 0., 0.,   0.,-1., 0.,   0., 0., 1.,   // O 2
									    0., 1., 0.,  -1., 0., 0.,   0., 0., 1. }; // O 3
	double* rv0 = (double*) &rotations[orientation*9];
	Matrix3D orientaionmatrix( rv0[0], rv0[1], rv0[2],
		                       rv0[3], rv0[4], rv0[5],
							   rv0[6], rv0[7], rv0[8] );
	return orientaionmatrix;
}

int MeshQuad::GetNumberOfAdditionalNodesForRefine(int3 pattern, int method)
{
	if(method==0)
	{
		if(pattern(1)==0) return  0; // total  4 nodes
		if(pattern(1)==1) return  3; // total  7 nodes
		if(pattern(1)==2) return  8; // total 12 nodes
		if(pattern(1)==3) return 10; // total 14 nodes
		if(pattern(1)==4) return 12; // total 16 nodes
	}
	return -1;
}
int MeshQuad::GetGridNumberOfIthAdditionalNode(int i, int3 pattern, int method)
{
// lookup tables - grid index of the required nodes for the template  
//  O4-G11-G12--O3 
//  |   |   |   | 
//  |   |   |   | 
//  G7--G8--G9-G10
//  |   |   |   | 
//  |   |   |   | 
//  G3--G4--G5--G6 
//  |   |   |   | 
//  |   |   |   | 
//  O1--G1--G2--O2 
//
// lookup tables:   for all templates the additional nodes in ascending order of gridnumbers                                              
	static const int NType1[]  = {1,  3,4                   };
	static const int NType2a[] = {1,2,3,4,5,6,  8,9         };
	static const int NType2b[] = {1,  3,4,5,    8,9,10,   12};
	static const int NType3[]  = {1,2,3,4,5,6,  8,9,10,   12};
	static const int NType4[]  = {1,2,3,4,5,6,7,8,9,10,11,12};
	
	int ncorners =    pattern(1);   
	int subtype =	  pattern(2);
	int orientation = pattern(3);	

	int gridnr = 0;		
	if(method==0)
	{
		switch(ncorners)										         
		{														         
		case 0:	break;											         
		case 1: if(i>0&&i<= 3) gridnr= NType1[i-1]; break;
		case 2: if(i>0&&i<= 8) { if(subtype==0) { gridnr= NType2a[i-1]; } if(subtype==1) { gridnr= NType2b[i-1]; } } break;
		case 3: if(i>0&&i<=10) gridnr= NType3[i-1]; break;
		case 4: if(i>0&&i<=12) gridnr= NType4[i-1]; break;
		}
	}
	return gridnr;
}
Vector3D MeshQuad::GetLocPosOfIthAdditionalNode(int i, int3 pattern, int method)
{
// get the gridnumber of the requested additional node
	int gridnr = GetGridNumberOfIthAdditionalNode(i,pattern,method);
	
// get local position of the grid node
	Vector3D locpos(0.,0.,0.);
	if(gridnr== 3 || gridnr== 7) locpos.X()= -1.;
	if(gridnr== 1 || gridnr== 4 || gridnr== 8 || gridnr==11) locpos.X() = -1./3.;
	if(gridnr== 2 || gridnr== 5 || gridnr== 9 || gridnr==12) locpos.X() = +1./3.;
	if(gridnr== 6 || gridnr==10) locpos.X()= +1.;
	if(gridnr== 1 || gridnr== 2) locpos.Y()= -1.;
	if(gridnr== 3 || gridnr== 4 || gridnr== 5 || gridnr== 6) locpos.Y() = -1./3.;
	if(gridnr== 7 || gridnr== 8 || gridnr== 9 || gridnr==10) locpos.Y() = +1./3.;
	if(gridnr==11 || gridnr==12) locpos.Y()= +1.;	
	
// loc pos is now the local position for orientation 0
// do transformation such that the local position of the new node is correctly oriented    
	int orientation = pattern(3);
	return ApplyOrientation(locpos,orientation);
}
int MeshQuad::GetSubdivisionLevelOfIthAdditionalNode(int i, int3 pattern, TArray<int>& cornersdlvls, int method)
{
//NOTE: general rule: if on an edge, subdivisionlevel is equal to minimum of edge,
//                    if in middle, subdivision level is equal to minimum of all 4 corners
	int orientation = pattern(3);
	TArray<int> orientedNL;
	TArray<int> inverseNL;
	GetNodeListOriented(orientation,orientedNL,inverseNL);

	int weight = 0;
	int gridnr = GetGridNumberOfIthAdditionalNode(i,pattern,method);

// node on edge: 4 edges
	if(gridnr== 1 || gridnr== 2) // on edge O1-O2 ( == E12 prototype)
	{
		weight = cornersdlvls(orientedNL(1));
		if(weight > cornersdlvls(orientedNL(2))) weight = cornersdlvls(orientedNL(2));
	}
	else if(gridnr== 6 || gridnr==10) { weight = cornersdlvls(orientedNL(2)); if(weight > cornersdlvls(orientedNL(3))) weight = cornersdlvls(orientedNL(3)); }  //E23
	else if(gridnr==11 || gridnr==12) { weight = cornersdlvls(orientedNL(3)); if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4)); }  //E34
	else if(gridnr== 3 || gridnr== 7) { weight = cornersdlvls(orientedNL(1)); if(weight > cornersdlvls(orientedNL(4))) weight = cornersdlvls(orientedNL(4)); }  //E14
	else                                                                               // bulk
	{
		weight = cornersdlvls(1);                              if(weight > cornersdlvls(2)) weight = cornersdlvls(2);
		if(weight > cornersdlvls(3)) weight = cornersdlvls(3); if(weight > cornersdlvls(4)) weight = cornersdlvls(4);
	}
	return weight;
}
int MeshQuad::GetNumberOfElementsForRefinement(int3 pattern, int method)
{
	if(method==0)
	{
		if(pattern(1)==0) return  1; 
		if(pattern(1)==1) return  3; 
		if(pattern(1)==2) return  7; 
		if(pattern(1)==3) return  8; 
		if(pattern(1)==4) return  9; 
	}
	return -1;
}
int MeshQuad::GetNodeListOfIthRefinedElement(int i, int3 pattern, TArray<int>& nodelist, int method)
{
// lookup tables: 
// !!! all templates are implemented for orientation '0' !!!
// nodes from unrefined have negative numbers               //     type1                          // 90 type1 
// additional nodes have positice numbers                   //  4--.--.--3                        // 4--.--.--3
	static const int NLType0[]  = { -1,-2,-3,-4 };	        //  |      / |							 | \      |
	static const int NLType1[]  = { -1, 1, 3, 2,		    //  .  .  .  .							 .  .  .  .
		                             1,-2,-3, 3,		    //  |   /    |							 |   \    |
						             2, 3,-3,-4 };	        // A2-A3  .  .	 //     type2a			 .  . A3-A1
	static const int NLType2a[] = { -1, 1, 4, 3,		    //  |  |     |	 //  4--.--.--3			 |     |  |
	                                 1, 2, 5, 4,		    // O1-A1--.--2	 //  | \    / |			 1--.-A2-O2
							         2,-2, 6, 5, 						     //  . A7-A8  .
							         3, 4, 7,-4,							 //  |  |  |  |
							         4, 5, 8, 7,							 // A3-A4-A5-A6
							         5, 6,-3, 8,							 //  |  |  |  |
							         7, 8,-3,-4 };	 	    //     type2b    // O1-A1-A2-O2
	static const int NLType2b[] = { -1, 1, 3, 2,            //  4--.-A8-O3
		                             1,-2, 4, 3,			//  | \   |  |
							         2, 3, 5,-4,			//  . A5-A6-A7
							         3, 4, 6, 5,			//  |  |  |  |
							         4,-2, 7, 6,			// A2-A3-A4  .
							         5, 6, 8,-4,			//  |  |   \ |
							         6, 7,-3, 8 };		    // O1-A1--.--2   
	static const int NLType3[]  = { -1, 1, 4, 3,                             //      type3
		                             1, 2, 5, 4,							 //  4--.A10-O3
							         2,-2, 6, 5,							 //  | \   |  |
							         3, 4, 7,-4,							 //  . A7-A8-A9
							         4, 5, 8, 7,							 //  |  |  |  |
							         5, 6, 9, 8,							 // A3-A4-A5-A6
							         7, 8,10,-4,							 //  |  |  |  |
							         8, 9,-3,10 };						     // O1-A1-A2--2
	static const int NLType4[]  = { -1, 1, 4, 3,            //      type4
		                             1, 2, 5, 4,			// O4-11-12-O3
							         2,-2, 6, 5,			//  |  |  |  |
							         3, 4, 8, 7,			//  7--8--9-10
							         4, 5, 9, 8,			//  |  |  |  |
							         5, 6,10, 9,			//  3--4--5--6
							         7, 8,11,-4,			//  |  |  |  |
							         8, 9,12,11,			// O1--1--2-O2
							         9,10,-3,12 };

	int ncorners =    pattern(1);
	int subtype =     pattern(2);
	int orientation = pattern(3);
	int* rv0 = NULL;
	int ith = 4*(i-1); // offset of 4 - quad has 4 nodes
	if(method==0)
	{
// get correct 4 nodenumbers in orientation 0
		if(ncorners==0) { rv0 = (int*) &NLType0; }
		if(ncorners==1) { rv0 = (int*) &NLType1 + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (int*) &NLType2a + ith; }
		if(ncorners==2 && subtype==1) { rv0 = (int*) &NLType2b + ith; }
		if(ncorners==3) { rv0 = (int*) &NLType3 + ith; }
		if(ncorners==4) { rv0 = (int*) &NLType4 + ith; }
	}

// account for cyclic position of the pattern 
// new nodes have rotated position so just the nodes of the unrefined element must be corrected
	nodelist.Set4(0,0,0,0);
	TArray<int> orientedNL;
	TArray<int> inverseNL;
	TArray<int> intermediateNL; intermediateNL.SetLen(4);
	GetNodeListOriented(orientation,orientedNL,inverseNL);
// 1) insert the correct original nodes into the elements nodelist -> correct nodelist, 
	for(int j=0; j<4; j++)
	{
		if(rv0[j]>0)
		{ 
			intermediateNL(j+1) = rv0[j]; 
		}
		else
		{
			intermediateNL(j+1) = -orientedNL(-rv0[j]); // negate input and output - still a negative number is required to mark a node of the unrefined element
		}
	}
// 2) reorient new element to match the orientaiton of the unrefined element
	for(int j=1; j<=4; j++)
	{
		nodelist(j) = intermediateNL(inverseNL(j));
	}
	return 0;
}
Vector3D MeshQuad::GetLocPosCenterOfIthRefinedElement(int i, int3 pattern, int method)
{
// lookup tables: 
// NOTE: all self similar child elements have only -2/3, 0 +2/3 entries
//       all different shapes also have -1/3,+1/3,-1/6,+1/6 entries, they are not refined further by the algorithm
	static const double LPCType0[]  = {    0.,    0., 0. };
	static const double LPCType1[]  = {-2./3.,-2./3., 0.,  // self-similar
                                        1./3.,-1./3., 0.,  // frustum
									   -1./3., 1./3., 0. };
	static const double LPCType2a[] = {-2./3.,-2./3., 0.,  // s
		                                   0.,-2./3., 0.,  // s
									    2./3.,-2./3., 0.,  // s
									   -2./3., 1./6., 0.,
									       0.,    0., 0.,  // s
									    2./3., 1./6., 0.,
										   0., 2./3., 0. };
	static const double LPCType2b[] = {-2./3.,-2./3., 0.,  // s
		                                1./6.,-2./3., 0.,
									   -2./3., 1./6., 0.,
									       0.,    0., 0.,  // s
									    2./3.,-1./6., 0.,
									   -1./6., 2./3., 0.,
									    2./3., 2./3., 0. };// s
	static const double LPCType3[]  = {-2./3.,-2./3., 0.,  // s
		                                   0.,-2./3., 0.,  // s
									    2./3.,-2./3., 0.,  // s 
									   -2./3., 1./6., 0.,
									       0.,    0., 0.,  // s
									    2./3.,    0., 0.,  // s
									   -1./6., 2./3., 0.,
									    2./3., 2./3., 0. };// s
	static const double LPCType4[]  = {-2./3.,-2./3., 0.,  // s
		                                   0.,-2./3., 0.,  // s
									    2./3.,-2./3., 0.,  // s 
									   -2./3.,    0., 0.,  // s
		                                   0.,    0., 0.,  // s
									    2./3.,    0., 0.,  // s 
									   -2./3., 2./3., 0.,  // s
		                                   0., 2./3., 0.,  // s
										2./3., 2./3., 0. };// s
	int ncorners =    pattern(1);
	int subtype =     pattern(2);
	double* rv0 = NULL;
	int ith = 3*(i-1); // offset of 3 (coordinates)
	if(method==0)
	{
// get correct center coordinates
		if(ncorners==0) { rv0 = (double*) &LPCType0; }
		if(ncorners==1) { rv0 = (double*) &LPCType1 + ith; }
		if(ncorners==2 && subtype==0) { rv0 = (double*) &LPCType2a + ith; }
		if(ncorners==2 && subtype==1) { rv0 = (double*) &LPCType2b + ith; }
		if(ncorners==3) { rv0 = (double*) &LPCType3 + ith; }
		if(ncorners==4) { rv0 = (double*) &LPCType4; }
	}
	Vector3D ploc0(rv0[0],rv0[1],rv0[2]);
  	int orientation = pattern(3);
	return ApplyOrientation(ploc0,orientation);
}

int MeshQuad::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<(-1.-eps) || ploc.X()>(1.+eps)) return 0;
	if(ploc.Y()<(-1.-eps) || ploc.Y()>(1.+eps)) return 0;
	//if(ploc.Z()<0. || ploc.Z()>1.) return 0; // what to do with 3rd coordinate ?
	return 1;
}
//****************************************************************************************
int MeshTrig::Invert()
{
	if(NNodes()==3||NNodes()==6) // corners
	{
		SwapNodeNumberEntries(2,3);
		if(NNodes()==6)
		{
			SwapNodeNumberEntries( 4, 6);
			return 6;
		}
		return 3;
	}
	return 0;
}
int MeshTrig::GetNodeListOriented(int orientation,TArray<int>& nodelist, TArray<int>& inverse)
{
	nodelist.Flush(); inverse.Flush();
	switch(orientation)
	{
		case 0: nodelist.Set3(1,2,3); inverse.Set3(1,2,3); break;
		case 1: nodelist.Set3(2,3,1); inverse.Set3(3,1,2); break;
		case 2: nodelist.Set3(3,1,2); inverse.Set3(2,3,1); break;
	}
	return nodelist.Length();
}
MeshElement MeshTrig::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshLine(1,2),
		MeshLine(2,3),
		MeshLine(3,1),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshLine(1,2,4),
		MeshLine(2,3,5),
		MeshLine(3,4,6),
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 3) { assert(0 && "MeshTrig::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 3) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()== 6) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshTrig::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshTrig::GetInterPair(int i)
{
	if(NNodes()==3)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		case  2: return int2(2,3); break;
		case  3: return int2(3,1); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshTrig::GetLocPosOfIthNode(int i)
{
	switch(i)
	{
	//linear:
	case 1: return Vector3D( 0., 0., 0.); break;          //
	case 2: return Vector3D( 1., 0., 0.); break;          //    3
	case 3: return Vector3D( 0., 1., 0.); break;          //    | \
	//quadratic:                                          //    6  5
	case 4: return Vector3D(0.5, 0., 0.); break;          //    |    \     //
	case 5: return Vector3D(0.5,0.5, 0.); break;          //    1--4--2
	case 6: return Vector3D( 0.,0.5, 0.); break;          // 
	}
	return Vector3D(0.);
}
double MeshTrig::GetS0(Vector3D ploc, int shape)
{
// compare Trig2Dlin::GetS0 - reoriented
//         Trig2DQuad::GetS0 - reoriented
	double r = ploc.X();
	double s = ploc.Y();
	if(NNodes() == 3)
	{
		switch(shape)
		{
		case 1: return 1.-r-s;
		case 2: return r;
		case 3: return s;
		default: return 0.;
		}
	}
	else if(NNodes() == 6)
	{
		switch(shape)
		{
		case 1: return (1.-r-s)*(2.*(1.-r-s)-1.);  // 2t^2-t = 2r^2+2s^2+4rs-3r-3s+1
		case 2: return r*(2.*r-1.);                // 2r^2-r       
		case 3: return s*(2.*s-1.);                // 2s^2-s
		case 4: return 4*r*(1.-r-s);               //-4r^2-4rs+4
		case 5: return 4*r*s;                      // 4rs
		case 6: return 4*s*(1.-r-s);               //-4s^2-4rs+4
		default: return 0.;
		}
	}
	return 0;
}
Matrix MeshTrig::GetDSMatrix0(Vector3D ploc)
{

	double r = ploc.X();
	double s = ploc.Y();
	Matrix sf(Dim(),NNodes());

	if(NNodes() == 3)
	{
		sf(1,1) = -1.;
		sf(1,2) = 1.;
		sf(1,3) = 0.;
		sf(2,1) = -1;
		sf(2,2) = 0;
		sf(2,3) = 1;
	}
	else if (NNodes() == 6)
	{
		sf(1,1) =  4*r + 4*s - 3;
		sf(1,2) =  4*r - 1;
		sf(1,3) =  0;
		sf(1,4) =  -8*r - 4*s + 4;
		sf(1,5) = 4*s;
		sf(1,6) =  -4*s;  

		sf(2,1) =  4*r + 4*s - 3;
		sf(2,2) =  0;
		sf(2,3) =  4*s - 1;
		sf(2,4) =  -4*r;
		sf(2,5) =  4*r;
		sf(2,6) =  -4*r - 8*s + 4;
	}
	return sf;
}

int MeshTrig::TestPLocOnElem(Vector3D ploc, double eps)
{
	if(ploc.X()<-eps || ploc.X()>(1.+eps)) return 0;
	if(ploc.Y()<-eps || ploc.Y()>(1.+eps)) return 0;
	//if(ploc.Z()<0. || ploc.Z()>1.) return 0; // ? what to do with 3rd coordinate
	if( (ploc.X()+ploc.Y())>(1.+eps) ) return 0; //test edge {N2,N3}
	return 1;
}
//****************************************************************************************
int MeshLine::Invert()
{
	if(NNodes()==2||NNodes()==3) // corners
	{
	//	SwapNodeNumberEntries(1,2);
		if(NNodes()==3)
		{
			return 3;
		}
		return 2;
	}
	return 0;
}
int MeshLine::GetNodeListOriented(int orientation,TArray<int>& nodelist, TArray<int>& inverse)
{
	nodelist.Flush(); inverse.Flush();
	switch(orientation)
	{
		case 0: nodelist.Set2(1,2); inverse.Set2(1,2); break;
		case 1: nodelist.Set2(2,1); inverse.Set2(2,1); break;
	}
	return nodelist.Length();
}
MeshElement MeshLine::GetLocalFace(int faceidx)
{
	static MeshElement localFacesLinear[] = {
		MeshLine(1,0),
		MeshLine(2,0),
		MeshQuad(0,0,0,0)
	};
	static MeshElement localFacesQuadratic[] = {
		MeshLine(1,0,0),
		MeshLine(2,0,0),
		MeshQuad(0,0,0,0)
	};
	if(faceidx < 1 || faceidx > 3) { assert(0 && "MeshLine::GetLocalFace called with a wrong index."); return localFacesLinear[0]; } 
	else if(NNodes()== 2) { return localFacesLinear[faceidx-1]; }
	else if(NNodes()== 3) { return localFacesQuadratic[faceidx-1]; }
	//$EK 2017-01-11: error handling added
	else { assert(0 && "MeshLine::GetLocalFace called with a wrong # of nodes."); return localFacesLinear[0]; } 
}
int2 MeshLine::GetInterPair(int i)
{
	if(NNodes()==2)
	{
		switch(i)
		{
		case  1: return int2(1,2); break;
		default: return int2(0,0);
		}
	}
	return int2(0,0);
}
Vector3D MeshLine::GetLocPosOfIthNode(int i)
{
	switch(i)
	{
	//linear:
	case 1: return Vector3D(-1., 0., 0.); break;
	case 2: return Vector3D( 1., 0., 0.); break;
	//quadratic:
	case 3: return Vector3D( 0., 0., 0.); break;
	}
	return Vector3D(0.);
}
double MeshLine::GetS0(Vector3D ploc, int shape)
{
	double r = ploc.X();

	if(NNodes() == 2)
	{
		switch(shape)
		{
		case 1: return 0.5*(1-r); 
		case 2: return 0.5*(1+r);
		default: return 0;
		}
	}
	else if(NNodes() == 3)
	{
		switch(shape)
		{
		case 1: return 0.5*r*(r-1.); 
		case 2: return -(r-1.)*(r+1.);
		case 3: return 0.5*r*(r+1.);
		default: return 0;
		}
	}
	assert(0);
	return 0;
}
Matrix MeshLine::GetDSMatrix0(Vector3D ploc)
{
	double r = ploc.X();
	Matrix sm(Dim(),NNodes());
	sm.SetAll(0.);

	if(NNodes() == 2)
	{
		sm(1,1) = -0.5;
		sm(2,1) =  0.5; 
	}
	else if(NNodes() == 3)
	{
		sm(1,1) = r-0.5;
		sm(2,1) =-2.*r; 
		sm(3,1) = r+0.5;
	}
	return sm;
}

int MeshLine::TestPLocOnElem(Vector3D ploc, double eps)
{
// ATTENTION: can not test for interval -length/2..+length/2, no global positions available at this point
	//Vector3D axial = npos(2)-npos(1);
	//double halflength = axial.Norm()*0.5;
	//if(ploc.X()<halflength || ploc.X()>halflength) return 0;
	//if(ploc.Y()<-1. || ploc.Y()>1.) return 0; // what to do with 2nd coordinate ? cross section position
	//if(ploc.Z()<0. || ploc.Z()>1.) return 0; // what to do with 3rd coordinate ? cross section position
	return 1;
}


// Functions to generate Nodes, Elements, ... for given Component types

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshComponent
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MeshComponent::MapNodeListComponentToMesh(TArray<int> &nodelist)
{
	for(int i=1; i<=nodelist.Length(); i++)
	{
		nodelist(i) = GetNodeMeshNumber(nodelist(i));
	}
}

void MeshComponent::GetElementData(ElementDataContainer& edc) 
{
	ElementData ed;
	ed.SetText(GetName(),"component_name"); ed.SetToolTipText("the name of the mesh component"); edc.Add(ed);
	ed.SetText(GetComponentSpecification().c_str(),"component_type"); ed.SetToolTipText("type of the mesh component"); edc.Add(ed);

/*
	Vector nodenums(NNodes()); for(int i=1; i<=NNodes(); i++) { nodenums(i) = this->GetNodeMeshNumber(i); } // nodes for this component in mesh numbering
	ed.SetVector(nodenums.GetVecPtr(), nodenums.Length(), "list_of_nodes", 1 , NNodes()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS node numbers of the component"); edc.Add(ed);
*/
	
	TArray<int> nodenumstarray(NNodes()); for(int i=1; i<=NNodes(); i++) { nodenumstarray(i) = this->GetNodeMeshNumber(i); } // nodes for this component in mesh numbering
	mesh->MapNodeListToMBSNodes(nodenumstarray); // convert to MBS numbering
	Vector nodenumsvector(NNodes()); for(int i=1; i<=NNodes(); i++) { nodenumsvector(i) = nodenumstarray(i); } // nodes for this component in MBS numbering
	ed.SetVector(nodenumsvector.GetVecPtr(), nodenumsvector.Length(), "list_of_nodes", 1 , NNodes()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS node numbers of the component - entries '-1' or '0' hint to nodes not actually used in mesh"); edc.Add(ed);

/*
	Vector elemnums(NElems()); for(int i=1; i<=NElems(); i++) { elemnums(i) = this->GetElementMeshNumber(i); } // elems for this component in mesh numbering
	ed.SetVector(elemnums.GetVecPtr(), elemnums.Length(), "list_of_elements", 1 , NElems()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the component"); edc.Add(ed);
*/

	TArray<int> elemnumstarray(NElems()); for(int i=1; i<=NElems(); i++) { elemnumstarray(i) = this->GetElementMeshNumber(i); } // elems for this component in mesh numbering
	mesh->MapElemListToMBSElems(elemnumstarray); // convert to MBS numbering
	Vector elemnumsvector(NElems()); for(int i=1; i<=NElems(); i++) { elemnumsvector(i) = elemnumstarray(i); } // elems for this component in MBS numbering
	ed.SetVector(elemnumsvector.GetVecPtr(), elemnumsvector.Length(), "list_of_elements", 1 , NElems()); ed.SetValuesInt(); ed.SetVariableLength(); ed.SetLocked(1); ed.SetToolTipText("MBS element numbers of the component - entries '-1' or '0' hint to elements not actually used in mesh"); edc.Add(ed);

	ed.SetInt(GetBodyIndex(),"body_index"); ed.SetToolTipText("the body index / domain number for the entire mesh component"); edc.TreeAdd("Generation",ed);
	ed.SetInt(GetMaterialNumber(),"material_number"); ed.SetToolTipText("the material number for the entire mesh component"); edc.TreeAdd("Generation",ed);
	ed.SetInt(GetGeometricNonlinearityStatus(),"GeometricNonlinearityStatus"); ed.SetToolTipText("-1..use element defaults, 0..GNS_Linear, 1..GNS_NonlinearSmallStrain, 2..GNS_NonlinearLargeStrain"); edc.TreeAdd("Generation",ed);
	ed.SetVector3D(GetColor().X(),GetColor().Y(),GetColor().Z(),"RGB_color"); ed.SetToolTipText("[red,green,blue] color of element, range = 0..1, use default color:[-1,-1,-1]"); edc.TreeAdd("Graphics",ed);
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshComponentSource
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
TArray<int> MeshComponentSource::GetElementNodeList(int i)
{
	TArray<int> nodelist = GetElemPtr(i)->GetNodeList(); // component index node list
	return nodelist;                                     
}
TArray<int> MeshComponentSource::GetElementFaceNodeList(int i, int f)
{
	TArray<int> nodelist = GetElemPtr(i)->GetFaceNodeList(f); // component index node list
	return nodelist;
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshComponentPrimitive
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshComponentPrimitive::MeshYourself()
{
	GenerateOwnNodes();
	GenerateOwnElements();
	InitNodeIdx();
	InitElemIdx();
	return 1;
}

int MeshComponentPrimitive::ComputeBoundingBox()
{
	boundingbox.Clear();
	for(int i=1; i<=points.Length(); i++)
	{
		Vector3D pos = points(i);
		boundingbox.Add(pos);
	}
	return points.Length();
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshBlock
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshBlock::GenerateOwnNodes() 
{
	nodes.Flush();
	nodes.SetLen(0);

	Vector3D pos;
	Vector3D bilinear_lower;
	Vector3D bilinear_upper;
	MeshNode n;
	
	int typenr = mesh->GetComponentNodeTypeIdx(this);	// interface to ObjectFactory via Mesh
	n.SetTypeNr(typenr); // define node type (ObjectFactory type number) - all nodes of component must have same node type
	int bodyidx = this->GetBodyIndex();                
	n.SetBodyIdx(bodyidx); // each node gets assigned the body index of the component

	for(int k=0; k<=Discretization(3); k++)
	{
		double fz = (double)k / (double)Discretization(3);
		for(int j=0; j<=Discretization(2); j++)
		{
			double fy = (double)j / (double)Discretization(2);
			for(int i=0; i<=Discretization(1); i++)
			{
				double fx = (double)i / (double)Discretization(1);
				bilinear_lower = Point(1) + fx*(Point(2)-Point(1)) + fy*(Point(3)-Point(1)) + fx*fy*(Point(1)-Point(2)-Point(3)+Point(4));
				bilinear_upper = Point(5) + fx*(Point(6)-Point(5)) + fy*(Point(7)-Point(5)) + fx*fy*(Point(5)-Point(6)-Point(7)+Point(8)); 
				pos = bilinear_lower + fz*(bilinear_upper-bilinear_lower);

				n.SetNodePos(pos); // define node position

				AddNode(n); // register node in component
			}
		}
	}
	return NNodes();
}
int MeshBlock::GenerateOwnElements() 
{
	elems.Flush();
	elems.SetLen(0);

	TArray<int> nodelist;
	MeshHex hex;

	int typenr = mesh->GetComponentElemTypeIdx(this); // interface to ObjectFactory via Mesh
	hex.SetTypeNr(typenr); // define element type (ObjectFactory type number) - all elements of component must have same element type
	hex.SetMatNr(matnr); // define material number - all elements of component are generated with same material type 

	for(int iz=1; iz<=Discretization(3); iz++)
	{
		for(int iy=1; iy<=Discretization(2); iy++)
		{
			for(int ix=1; ix<=Discretization(1); ix++)
			{
				int node1 = (iz-1)*(Discretization(2)+1)*(Discretization(1)+1) + (iy-1)*(Discretization(1)+1) + (ix-1)*1 + 1;
				GetNodesForHex(node1, nodelist);
				
				hex.SetNodeList(nodelist); // defines the nodelist for the element
				
				AddElem(hex); // register element in component
			}
		}
	}
	return NElems();
}
int MeshBlock::GetElemAlternateFlag(int i)
{
	int i0 = i-1;
	int ix0 = i0%Discretization(1);
	int iy0 = i0/Discretization(1)%Discretization(2);
	int iz0 = i0/(Discretization(1)*Discretization(2))%Discretization(3);

	return (ix0+iy0+iz0)%2;
}

int MeshBlock::GetNodesForHex(int node1idx, TArray<int>& nodelist)
{
	nodelist.SetLen(8);
	nodelist(1) = node1idx                                                                           ;
	nodelist(2) = node1idx                                                                       + 1 ;
	nodelist(3) = node1idx                                               + (Discretization(1)+1)     ;
	nodelist(4) = node1idx                                               + (Discretization(1)+1) + 1 ;
	nodelist(5) = node1idx + (Discretization(2)+1)*(Discretization(1)+1)                             ;				
	nodelist(6) = node1idx + (Discretization(2)+1)*(Discretization(1)+1)                         + 1 ;
	nodelist(7) = node1idx + (Discretization(2)+1)*(Discretization(1)+1) + (Discretization(1)+1)     ;
	nodelist(8) = node1idx + (Discretization(2)+1)*(Discretization(1)+1) + (Discretization(1)+1) + 1 ;
	return 8;
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshCylinder
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshCylinder::GenerateOwnNodes() 
{
	nodes.Flush();
	nodes.SetLen(0);

	for(int k=0; k<=Discretization(3); k++)
	{
		Vector3D discPinner( Point(1) + (Point(3)-Point(1))*((double)k/(double)Discretization(3)) );
		Vector3D discPouter( Point(2) + (Point(4)-Point(2))*((double)k/(double)Discretization(3)) );
		if(IsFullCylinder())
		{
			GenerateNodesDisc(discPinner,discPouter,0.5);
		}
		else
		{
			GenerateNodesShim(discPinner,discPouter);
		}
	}
	return NNodes();
}
int MeshCylinder::GenerateNodesDisc(Vector3D pi, Vector3D pa, double inner_fraction)
{
	int d = Discretization(2);         // angular divisions
	double ra = pa.X();                // outer radius
	double rs = ra * inner_fraction;   // "small" radius for squares
	
	double rza = pa.X();    // z-interpolation   
	double rz0 = pi.X();    // z-interpolation
	double za = pa.Z();     // z-interpolation
	double z0 = pi.Z();     // z-interpolation

	Vector3D pos(0.0,0.0,0.0);
	MeshNode n;	

	int typenr = mesh->GetComponentNodeTypeIdx(this);
	n.SetTypeNr(typenr);
	int bodyidx = this->GetBodyIndex();                
	n.SetBodyIdx(bodyidx); 

	//inner part: squares
	double a=rs/(double)d;
	for(int i=0; i<=d; i++)
	{
		pos.Y() = i*a;
		for(int j=0; j<=d; j++)
		{
			pos.X() = j*a;
			double rij = sqrt(pos.X()*pos.X()+pos.Y()*pos.Y());
			pos.Z() = z0 + (za-z0)*(rz0+rij)/(rza-rz0);
			
			n.SetNodePos(pos);
				
			AddNode(n);
		}
	}

	// outer part: 
	double r0_tmp,r1_tmp,r2_tmp;
	double x0;
	double r=ra;
	double r0 = rs*sqrt(2.);
	double r1=rs;

	double sqrt2=sqrt(2.);
	double alpha_max, alpha_tmp;
	double tmp1, tmp2;
	int j;//,m=1;

	for(int i=1;i<=d; i++)
	{
		r0_tmp=r0+((r-r0)/d)*i;
		r1_tmp=r1+((r-r1)/d)*i;
		tmp1 =-r0_tmp * (sqrt2 * r1_tmp - r0_tmp) / (-r0_tmp*r0_tmp + r0_tmp * sqrt2 * r1_tmp - r1_tmp*r1_tmp);
		tmp2=-(r0_tmp * sqrt2 - r1_tmp) * r1_tmp / (-r0_tmp*r0_tmp + r0_tmp * sqrt2 * r1_tmp - r1_tmp*r1_tmp);
		alpha_max=atan2(tmp1,tmp2);
		r2_tmp=r0_tmp*sin(MY_PI/4)/sin(alpha_max);
		x0=r2_tmp-r1_tmp;
		for (j=0; j<=d; j++)
		{
			alpha_tmp=j*alpha_max/d;
			pos.X()=r2_tmp*cos(alpha_tmp)-x0;
			pos.Y()=r2_tmp*sin(alpha_tmp);
			double rij = sqrt(pos.X()*pos.X()+pos.Y()*pos.Y());
			pos.Z() = z0 + (za-z0)*(rz0+rij)/(rza-rz0);
			
			n.SetNodePos(pos);
				
			AddNode(n);
//			m++;
		}
		j=d;
		while (j>0)
		{
			j=j-1;
			alpha_tmp=j*alpha_max/d;
			pos.X()=r2_tmp*sin(alpha_tmp);
			pos.Y()=r2_tmp*cos(alpha_tmp)-x0;
			double rij = sqrt(pos.X()*pos.X()+pos.Y()*pos.Y());
			pos.Z() = z0 + (za-z0)*(rz0+rij)/(rza-rz0);
			
			n.SetNodePos(pos);
		
			AddNode(n);
		}
	}
	return NNodes();
}
int MeshCylinder::GenerateNodesShim(Vector3D pi, Vector3D pa)
{
	int d = Discretization(2);
	double r2 = pa.X();
	double r1 = pi.X();
	double phi_tmp;
	double za = pa.Z();
	double z0 = pi.Z();

	Vector3D pos(0.0,0.0,0.0);
	MeshNode n;

	int typenr = mesh->GetComponentNodeTypeIdx(this);
	n.SetTypeNr(typenr);
	int bodyidx = this->GetBodyIndex();                
	n.SetBodyIdx(bodyidx); // each node gets assigned the body index of the component
	
	for(int i=0; i<=Discretization(1); i++)
	{
		double r_tmp = r1 + (r2-r1)*((double)i/(double)Discretization(1));
		for (int j=0; j<=2*d; j++)
		{
			phi_tmp=j*MY_PI/(4*d);
			
			pos.X()=r_tmp*cos(phi_tmp);
			pos.Y()=r_tmp*sin(phi_tmp);
			double rij = r_tmp; //sqrt(pos.X()*pos.X()+pos.Y()*pos.Y());
			pos.Z() = z0 + (za-z0)*(r1+rij)/(r2-r1);
			
			n.SetNodePos(pos);

			AddNode(n);
		}
	}
	return NNodes();
}
int MeshCylinder::GenerateOwnElements()
{
	elems.Flush();
	elems.SetLen(0);

	TArray<int> nodelist;
	nodelist.SetLen(8);
	MeshHex hex;

	int typenr = mesh->GetComponentElemTypeIdx(this);
	hex.SetTypeNr(typenr);
	hex.SetMatNr(matnr);

	if(IsFullCylinder()) 
	{
//  7--__ 
//  |    6
//  |   / \         H1 (1,2,3,4, 8,9,10,11)
//  3--4   |        H2 (2,5,4,6, 8,12,11,13)
//  |  |    \       H3 (4,6,3,7, 11,13,10,14)
//  1--2----5
		int d2 = Discretization(2);
		int npl = (d2+1)*(d2+1) + d2*(2*d2+1); // nodes_per_layer 
		for(int h=0; h<Discretization(3); h++)
		{
		// inner square
			for(int j=0; j<d2; j++) 
			{
				for(int i=1; i<=d2; i++)
				{
					nodelist(1) = i          + j*(d2+1) +    h *npl ;
					nodelist(2) = i      + 1 + j*(d2+1) +    h *npl ;
					nodelist(3) = i + d2 + 1 + j*(d2+1) +    h *npl ;
					nodelist(4) = i + d2 + 2 + j*(d2+1) +    h *npl ;
					nodelist(5) = i          + j*(d2+1) + (h+1)*npl ;
					nodelist(6) = i      + 1 + j*(d2+1) + (h+1)*npl ;
					nodelist(7) = i + d2 + 1 + j*(d2+1) + (h+1)*npl ;
					nodelist(8) = i + d2 + 2 + j*(d2+1) + (h+1)*npl ;
					
					hex.SetNodeList(nodelist);

					AddElem(hex);
				}
			}
		// transition zone
			for(int a=1;a<=d2; a++) 
			{
				nodelist(1) =  0  +     a *(d2+1)      +    h *npl ;
				nodelist(2) =  a  + (d2+1)*(d2+1)      +    h *npl ;
				nodelist(3) =  0  +  (a+1)*(d2+1)      +    h *npl ;
				nodelist(4) = a+1 + (d2+1)*(d2+1) 	   +    h *npl ;
				nodelist(5) =  0  +     a *(d2+1)      + (h+1)*npl ;
				nodelist(6) =  a  + (d2+1)*(d2+1)      + (h+1)*npl ;
				nodelist(7) =  0  +  (a+1)*(d2+1)      + (h+1)*npl ;
				nodelist(8) = a+1 + (d2+1)*(d2+1) 	   + (h+1)*npl ;
				
				hex.SetNodeList(nodelist);

				AddElem(hex);
			}
		// transition zone
			for(int a=0; a<d2; a++)
			{
				nodelist(1) = -a         + (d2+1)*(d2+1)      +    h *npl ;
				nodelist(2) = a+(d2+1)   + (d2+1)*(d2+1)      +    h *npl ;
				nodelist(3) = -(a+1)     + (d2+1)*(d2+1)      +    h *npl ;
				nodelist(4) = a+(d2+1)+1 + (d2+1)*(d2+1)      +    h *npl ;
				nodelist(5) = -a         + (d2+1)*(d2+1)      + (h+1)*npl ;
				nodelist(6) = a+(d2+1)   + (d2+1)*(d2+1)      + (h+1)*npl ;
				nodelist(7) = -(a+1)     + (d2+1)*(d2+1)      + (h+1)*npl ;
				nodelist(8) = a+(d2+1)+1 + (d2+1)*(d2+1)      + (h+1)*npl ;							
				
				hex.SetNodeList(nodelist);

				AddElem(hex);
			}
		// outer circles
			int d1 = Discretization(1);
			int npc = (2*d2+1); // nodes_per_circle
			for(int r=1; r<d1; r++)
			{
				for(int a=1; a<=2*d2; a++)
				{
					nodelist(1) = a           + (r-1)*npc + (d2+1)*(d2+1) +    h *npl ;
					nodelist(2) = a + npc     + (r-1)*npc + (d2+1)*(d2+1) +    h *npl ;
					nodelist(3) = a       + 1 + (r-1)*npc + (d2+1)*(d2+1) +    h *npl ;
					nodelist(4) = a + npc + 1 + (r-1)*npc + (d2+1)*(d2+1) +    h *npl ;
					nodelist(5) = a           + (r-1)*npc + (d2+1)*(d2+1) + (h+1)*npl ;
					nodelist(6) = a + npc     + (r-1)*npc + (d2+1)*(d2+1) + (h+1)*npl ;
					nodelist(7) = a       + 1 + (r-1)*npc + (d2+1)*(d2+1) + (h+1)*npl ;
					nodelist(8) = a + npc + 1 + (r-1)*npc + (d2+1)*(d2+1) + (h+1)*npl ;						
					
					hex.SetNodeList(nodelist);

					AddElem(hex);
				}
			}
		}
	}
	else
	{
//  6----____
//  |        5
//  |      /  |
//  3-_  /    |
//      2      |     H1 (1,4,2,5, 7,10,8,11)
//       |     |     H2 (2,5,3,6, 8,11,9,12)
//       1-----4
		int d1 = Discretization(1);
		int d2 = Discretization(2);
		int npl = (d1+1)*(2*d2+1); // nodes_per_layer
		for(int h=0; h<Discretization(3); h++) 
		{
			int npc = (2*d2+1); // nodes_per_circle
			for(int r=0; r<d1; r++)
			{
				for(int a=1; a<=2*d2; a++) 
				{
					nodelist(1) = a           + r*npc +    h *npl ;
					nodelist(2) = a + npc     + r*npc +    h *npl ;
					nodelist(3) = a       + 1 + r*npc +    h *npl ;
					nodelist(4) = a + npc + 1 + r*npc +    h *npl ;
					nodelist(5) = a           + r*npc + (h+1)*npl ;
					nodelist(6) = a + npc     + r*npc + (h+1)*npl ;
					nodelist(7) = a       + 1 + r*npc + (h+1)*npl ;
					nodelist(8) = a + npc + 1 + r*npc + (h+1)*npl ;
					
					hex.SetNodeList(nodelist);

					AddElem(hex);
				}
			}
		}
	}
	return NElems();
}

int MeshCylinder::GetElemAlternateFlag(int i)
{
	if(IsFullCylinder())
		return 0;

	int i0 = i-1;
	int ix0 = i0%(2*Discretization(1));
	int iy0 = i0/(2*Discretization(1))%Discretization(2);
	int iz0 = i0/(2*Discretization(1)*Discretization(2))%Discretization(3);

	return (ix0+iy0+iz0)%2;
}

int MeshCylinder::ComputeBoundingBox() 
{
	double zmin = min( min(Point(1).Z(),Point(2).Z()), min(Point(3).Z(),Point(4).Z()) );
	double zmax = max( max(Point(1).Z(),Point(2).Z()), max(Point(3).Z(),Point(4).Z()) );
	double rmax = max( max(sqrt(sqr(Point(1).X())+sqr(Point(1).Y())), sqrt(sqr(Point(2).X())+sqr(Point(2).Y()))),
                     max(sqrt(sqr(Point(3).X())+sqr(Point(3).Y())), sqrt(sqr(Point(4).X())+sqr(Point(4).Y()))) );

	boundingbox.Clear();
	Vector3D centerbase(0.,0.,zmin);
	boundingbox.Add(centerbase);
	Vector3D centertops(0.,0.,zmax);
	boundingbox.Add(centertops);
	boundingbox.Increase(rmax);

	return points.Length();
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshQuadrilateral
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshQuadrilateral::GenerateOwnNodes()
{
	nodes.Flush();
	nodes.SetLen(0);

	Vector3D pos;
	MeshNode n;
	
	int typenr = mesh->GetComponentNodeTypeIdx(this);
	n.SetTypeNr(typenr);
	int bodyidx = this->GetBodyIndex();                
	n.SetBodyIdx(bodyidx); // each node gets assigned the body index of the component

	for(int j=0; j<=Discretization(2); j++)
	{
		double fy = (double)j / (double)Discretization(2);
		for(int i=0; i<=Discretization(1); i++)
		{
			double fx = (double)i / (double)Discretization(1);
			pos = Point(1) + fx*(Point(2)-Point(1)) + fy*(Point(3)-Point(1)) + fx*fy*(Point(1)-Point(2)-Point(3)+Point(4));
			
			n.SetNodePos(pos);

			AddNode(n);
		}
	}
	return NNodes();
}
int MeshQuadrilateral::GenerateOwnElements()
{
	elems.Flush();
	elems.SetLen(0);

	TArray<int> nodelist;
	MeshQuad quad;

	int typenr = mesh->GetComponentElemTypeIdx(this);
	quad.SetTypeNr(typenr);
	quad.SetMatNr(matnr);

	for(int iy=1; iy<=Discretization(2); iy++)
	{
		for(int ix=1; ix<=Discretization(1); ix++)
		{
			int node1 = (iy-1)*(Discretization(1)+1) + (ix-1)*1 + 1;
			GetNodesForQuad(node1,nodelist);
			
			quad.SetNodeList(nodelist);

			AddElem(quad);
		}
	}
	return NElems();
}
int MeshQuadrilateral::GetNodesForQuad(int node1idx, TArray<int>& nodelist)
{
	nodelist.SetLen(4);
	nodelist(1) = node1idx                             ;
	nodelist(2) = node1idx                         + 1 ;
	nodelist(3) = node1idx + (Discretization(1)+1) + 1 ;
	nodelist(4) = node1idx + (Discretization(1)+1)     ;
	return 4;
}
int MeshQuadrilateral::ComputeOrientation(Vector3D& pos)
{
	if(orientationcomputed/*&&!posdependent*/) 
	{
		return 0;
	}
	else
	{
	// no explicit position dependency for a flat component
		Vector3D p12 = Point(2)-Point(1); 
		Vector3D p13 = Point(3)-Point(1);
		Vector3D n(p12.Cross(13));
		p12.Normalize();
		p13.Normalize();

		refslope1 = p12;
		refslope2 = p13;
		
		refslope3 = Vector3D(0.,0.,1.);
		refangles = Vector3D(0.,0.,0.);

		orientationcomputed = 1;
		return 1;
	}
}
int MeshQuadrilateral::ComputeElementSize(int i)
{
	if(sizecomputed/*&&constantelementsize*/)
	{
		return 0;
	}
	else
	{
	// size constant for all elements of the component
		Vector3D p12 = Point(2)-Point(1); 
		Vector3D p13 = Point(3)-Point(1);

		size1 = p12.Norm() / Discretization(1);
		size2 = p13.Norm() / Discretization(2);

		sizecomputed = 1;
		return 1;
	}
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshCurve
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshCurve::GenerateOwnNodes()
{
	nodes.Flush();
	nodes.SetLen(0);
	
	MeshNode n;

	int elem_typenr = mesh->GetComponentElemTypeIdx(this);
	int typenr = mesh->GetComponentNodeTypeIdx(this);
	n.SetTypeNr(typenr);
	int bodyidx = this->GetBodyIndex();                
	n.SetBodyIdx(bodyidx); // each node gets assigned the body index of the component

	
	int discretization = Discretization(1);
	if (IsQuadraticElementType())
		discretization *= 2;
	Vector3D vector_to_next_node(Point(2)-Point(1));
	vector_to_next_node /= discretization;

	Vector3D pos = Point(1);
	for(int i=0; i<=discretization; i++)
	{
		n.SetNodePos(pos);
		AddNode(n);
		pos += vector_to_next_node;
	}
	return NNodes();
}

int MeshCurve::GenerateOwnElements()
{
	elems.Flush();
	elems.SetLen(0);

	TArray<int> nodelist;
	MeshLine line;

	int typenr = mesh->GetComponentElemTypeIdx(this);
	line.SetTypeNr(typenr);
	line.SetMatNr(matnr);

	bool is_quadratic_element_type = IsQuadraticElementType();
	int nr_of_nodes = is_quadratic_element_type ? Discretization(1)*2 + 1 : Discretization(1) + 1;
	
	int node1 = 1;
	while (node1 < nr_of_nodes)
	{		
		if (is_quadratic_element_type)
		{
			GetNodesForQuadratic(node1++, nodelist);
			node1++;
		}
		else
		{
			GetNodesForLinear(node1++, nodelist);
		}
		
		line.SetNodeList(nodelist);		
		AddElem(line);
	}
	return NElems();
}
int MeshCurve::GetNodesForLinear(int node1idx, TArray<int>& nodelist)
{
	nodelist.SetLen(2);
	nodelist(1) = node1idx     ;
	nodelist(2) = node1idx + 1 ;
	return 2;
}
int MeshCurve::GetNodesForQuadratic(int node1idx, TArray<int>& nodelist)
{
	nodelist.SetLen(2);
	nodelist(1) = node1idx     ;
	nodelist(2) = node1idx + 2 ;
	nodelist(3) = node1idx + 1 ;
	return 3;
}
int MeshCurve::ComputeOrientation(Vector3D& pos)
{
	if(orientationcomputed/*&&!posdependent*/)
	{
		return 0;
	}
	else
	{
		Vector3D p12 = Point(2) - Point(1);
		double pyth_xy = sqrt( p12.X()*p12.X() + p12.Y()*p12.Y() );
		
		double roll = 0.; // torsion goes here ?!?
		double pitch = atan2(p12.Y(), p12.X());
		double yaw = atan2(p12.Z(), pyth_xy);

		refangles = Vector3D( roll, -yaw, pitch );

		//	Matrix3D mat = ComputeRotMatrix3DEuler( )

		double sinYaw = sin(-yaw);     double cosYaw = cos(-yaw);
		double sinPitch = sin(pitch); double cosPitch = cos(pitch);
		double sinRoll = sin(roll);   double cosRoll = cos(roll);
		refslope1 = Vector3D( cosYaw*cosPitch,
		                      cosYaw*sinPitch,
													sinYaw );
		refslope2 = Vector3D( sinRoll*sinYaw*cosPitch - cosRoll*sinPitch,
		                      sinRoll*sinYaw*sinPitch + cosRoll*cosPitch,
													sinRoll*cosYaw );
		refslope3 = Vector3D( cosRoll*sinYaw*cosPitch + sinRoll*sinPitch,
			                    cosRoll*sinYaw*sinPitch - sinRoll*cosPitch,
													cosRoll*cosYaw );
		orientationcomputed = 1;
		return 1;
	}
}
int MeshCurve::ComputeElementSize(int i)
{
	if(sizecomputed/*&&constantelementsize*/)
	{
		return 0;
	}
	else
	{
	// size constant for all elements of the component
		Vector3D p12 = Point(2)-Point(1); 
		
		size1 = p12.Norm() / Discretization(1);
		
		sizecomputed = 1;
		return 1;
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshSourceExtended
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// 
int MeshSourceExtended::GetElementGeometricNonlinearityStatus(int i)
{
	if(component_gnls!=-1)
		return component_gnls;
	/*else*/
	if(subordinate!=NULL)
	{
		int nesub = GetSubordinateComponent()->NElems();
		int iorig = (i-1)%(nesub) +1;
		return GetSubordinateComponent()->GetElementGeometricNonlinearityStatus(iorig);
	}
	else
		return -1;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshLinearToQuadratic
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// introduces additional nodes and fully replaces the elements
MeshNode* MeshLinearToQuadratic::GetNodePtr(int i)
{
	if(NSubs()==0) { return NULL; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodePtr(i); }
	else { return nodes(i-nnsub); }
}
int MeshLinearToQuadratic::InitializeLookupTables()
{	
	int maxnodetypes = 20;  // TODO: get accuracte number from ObjectFactory
	// policy:  NODETYPE DOES NOT CHANGE
	nodetypelist = NaturalNumbers(maxnodetypes);
	
	int maxelemtypes = 100; // TODO: get accurate number from ObjectFactory
	// policy: DISABLE ALL BUT MANUALLY ENTERED 
	elementtypelist.SetLen(maxelemtypes); 
	elementtypelist.SetAll(0);

//  these elements do not change type (but get additional nodes), return value is '-1' if the Element type is not registered
	int objectfactoryelementtypeHex = mesh->GetComponentElemTypeIdx(mystr("Hexahedral"));	 
	if (objectfactoryelementtypeHex>=0) { elementtypelist(objectfactoryelementtypeHex) = objectfactoryelementtypeHex; }// Hexahedral stays Hexahedral
	int objectfactoryelementtypeTet = mesh->GetComponentElemTypeIdx(mystr("Tetrahedral"));
	if (objectfactoryelementtypeTet>=0) { elementtypelist(objectfactoryelementtypeTet) = objectfactoryelementtypeTet; }// Tetradehdral stays Tetrahedral
	int objectfactoryelementtypePri = mesh->GetComponentElemTypeIdx(mystr("Prism"));
	if (objectfactoryelementtypePri>=0) { elementtypelist(objectfactoryelementtypePri) = objectfactoryelementtypePri; }
	int objectfactoryelementtypePyr = mesh->GetComponentElemTypeIdx(mystr("Pyramid"));
	if (objectfactoryelementtypePyr>=0) { elementtypelist(objectfactoryelementtypePyr) = objectfactoryelementtypePyr; }
	int objectfactoryelementtypeHexTM = mesh->GetComponentElemTypeIdx(mystr("HexahedralThermoMechanic"));
	if (objectfactoryelementtypeHexTM>=0) { elementtypelist(objectfactoryelementtypeHexTM) = objectfactoryelementtypeHexTM; } 		
	int objectfactoryelementtypeHexT = mesh->GetComponentElemTypeIdx(mystr("HexahedralThermo"));
	if (objectfactoryelementtypeHexT>=0) { elementtypelist(objectfactoryelementtypeHexT) = objectfactoryelementtypeHexT; }
#ifdef stacked //$AD 2017-04-21: put restriced elements in ifdefs !
	int objectfactoryelementtypeHexS = mesh->GetComponentElemTypeIdx(mystr("HexahedralStacked"));
	if (objectfactoryelementtypeHexS>=0) { elementtypelist(objectfactoryelementtypeHexS) = objectfactoryelementtypeHexS; }
    //$EK 2017-01-11: add possible TetrahedralStacked to split-routine		
	int objectfactoryelementtypeTetS = mesh->GetComponentElemTypeIdx(mystr("TetrahedralStacked"));
	if (objectfactoryelementtypeTetS>=0) { elementtypelist(objectfactoryelementtypeTetS) = objectfactoryelementtypeTetS; }
	//$EK 2017-01-11: add possible PrismStacked to split-routine		
	int objectfactoryelementtypePrismS = mesh->GetComponentElemTypeIdx(mystr("PrismStacked"));
	if (objectfactoryelementtypePrismS>=0) { elementtypelist(objectfactoryelementtypePrismS) = objectfactoryelementtypePrismS; }
	//$EK 2017-01-11: add possible PyramidStacked to split-routine		
	int objectfactoryelementtypePyramidS = mesh->GetComponentElemTypeIdx(mystr("PyramidStacked"));
	if (objectfactoryelementtypePyramidS>=0) { elementtypelist(objectfactoryelementtypePyramidS) = objectfactoryelementtypePyramidS; }
#endif

	int objectfactoryelementtypeQuad = mesh->GetComponentElemTypeIdx(mystr("Quadrilateral"));	
	if (objectfactoryelementtypeQuad>=0) {elementtypelist(objectfactoryelementtypeQuad) = objectfactoryelementtypeQuad; }
	int objectfactoryelementtypeTrig = mesh->GetComponentElemTypeIdx(mystr("Triangular"));	
	if (objectfactoryelementtypeTrig>=0) {elementtypelist(objectfactoryelementtypeTrig) = objectfactoryelementtypeTrig; }

// TODO: better error handling
	int objectfactoryelementtypeANCFBeam3DLinear = mesh->GetComponentElemTypeIdx(mystr("ANCFBeamShear3DLinear"));
	int objectfactoryelementtypeANCFBeam3DQuadratic = mesh->GetComponentElemTypeIdx(mystr("ANCFBeamShear3DQuadratic"));
	int objectfactoryelementtypeANCFBeamFE2DLinear = mesh->GetComponentElemTypeIdx(mystr("ANCFBeamShearFE2DLinear"));
	int objectfactoryelementtypeANCFBeamFE2DQuadratic = mesh->GetComponentElemTypeIdx(mystr("ANCFBeamShearFE2DQuadratic"));
	int objectfactoryelementtypeLinearBeam3D = mesh->GetComponentElemTypeIdx(mystr("LinearBeam3D"));
// these elements change type
	elementtypelist(objectfactoryelementtypeANCFBeam3DLinear) = objectfactoryelementtypeANCFBeam3DQuadratic;                 // ANCFBeamShear3DLinear becomes ANCFBeamShear3DQuadratic
	elementtypelist(objectfactoryelementtypeANCFBeamFE2DLinear) = objectfactoryelementtypeANCFBeamFE2DQuadratic;             // ANCFBeamShearFE2DLinear becomes ANCFBeamShearFE2DQuadratic
// these elements are already 'quadratic'
	elementtypelist(objectfactoryelementtypeANCFBeam3DQuadratic) = -objectfactoryelementtypeANCFBeam3DQuadratic;             // ANCFBeamShear3DQuadratic is already quadratic - change sign to mark
	elementtypelist(objectfactoryelementtypeANCFBeamFE2DQuadratic) = -objectfactoryelementtypeANCFBeamFE2DQuadratic;         // ANCFBeamShearFE2DQuadratic is already quadratic - change sign to mark
// these elements have no matching 'quadratic' element
	elementtypelist(objectfactoryelementtypeLinearBeam3D) = 0;                                                               // LinearBeam3D has no matching 'quadratic' - set 0 so it can be skipped
	return 0;
}
int MeshLinearToQuadratic::GetNodeTypeLookup(int lineartype)
{
	return nodetypelist(lineartype);
}
int MeshLinearToQuadratic::GetElemTypeLookup(int lineartype)
{ 
	return elementtypelist(lineartype);
}

int MeshLinearToQuadratic::MeshYourself()
{
	InitializeLookupTables();
	if(NSubs()==0) { return 0; }
	GetSubordinateComponent()->MeshYourself();

	int nn = GetSubordinateComponent()->NNodes();
	TArray<TArray<int2>> inter; // buffer for newly created intermediate nodes
	inter.SetLen(nn);
	inter.SetAll(TArray<int2>(0));
	parentnodes.Flush();

	int nprocessed=0;
	int nalreadyquadratic=0;
	int nnotquadraticmatch=0;
	int ne = GetSubordinateComponent()->NElems();
	for(int i=1; i<=ne; i++) // i ... element number i
	{
		MeshElement* el = GetSubordinateComponent()->GetElemPtr(i);                 
		TArray<int> nodelist = GetSubordinateComponent()->GetElementNodeList(i);

		int newtype = GetElemTypeLookup(el->GetTypeNr());  // look up the objectfactory type for the quadratic element
		if(newtype<0)
		{
			assert("element already is quadratic");
			// instead continue and just skip this element ?
			nalreadyquadratic++;
			AddElem(*el);
			continue;
		}
		if(newtype==0)
		{
			assert("no quadratic type defined for this element");
			// instead continue and just skip this element ?
			nnotquadraticmatch++;
			AddElem(*el);
			continue;
		}

		int ninter = el->NInter();
		for(int j=1; j<=ninter; j++) // j ... intermediate node pair j
		{
			int2 pair = el->GetInterPair(j);
			int nnr1 = nodelist(pair(1));
			int nnr2 = nodelist(pair(2));

			TArray<int2>& lookup = inter(nnr1);
			int nnrinter = 0;
			for(int k=1; k<=lookup.Length(); k++) // k ... component node number k  
			{
				if(lookup(k)(1)==nnr2) 
				{
					nnrinter = lookup(k)(2); // intermediate node alreay created previously
					break;
				}
			}
			if(nnrinter==0)
			{
				MeshNode node;
				node.SetTypeNr(GetSubordinateComponent()->GetNodePtr(nnr1)->GetTypeNr()); // copy type
				node.SetBodyIdx(GetSubordinateComponent()->GetNodePtr(nnr1)->GetBodyIdx()); // copy body index
				Vector3D p1 = GetSubordinateComponent()->GetNodePosition(nnr1);
				Vector3D p2 = GetSubordinateComponent()->GetNodePosition(nnr2);
				Vector3D pos = 0.5*(p1+p2); // compute intermediate position
				node.SetNodePos(pos); 
				nnrinter = AddNode(node) + nn;  // index in list of additional nodes (THIS component) plus number of nodes in the linear component (SUBORDINATE component)
				
				inter(nnr1).Add(int2(nnr2,nnrinter));
				inter(nnr2).Add(int2(nnr1,nnrinter));
				parentnodes.Add(int2(nnr1,nnr2));
			}
			nodelist.Add(nnrinter);
		}

		MeshElement* newel = el->GetCopy();
		newel->SetNodeList(nodelist);
		newel->SetTypeNr(newtype);
		nprocessed++;
		AddElem(*newel);
	}
	InitNodeIdx();
	InitElemIdx();
	this->GetMesh()->GetMBS()->UO() << "MeshLinearToQuadratic: successfully processed " << nprocessed << " elements. " << nalreadyquadratic << " elements were already quadratic and " << nnotquadraticmatch << " elements could not be converted. \n";
	return 1;
}

int MeshLinearToQuadratic::GetNodeObjectFactoryType(int i)
{
	if(NSubs()==0) { return 0; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeObjectFactoryType(i); }
	else { return nodes(i-nnsub)->GetTypeNr(); }
}
void MeshLinearToQuadratic::SetNodeMeshNumber(int i, int idx)
{
	if(NSubs()==0) { return; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { GetSubordinateComponent()->SetNodeMeshNumber(i,idx); }
	else { nodeIdx(i-nnsub) = idx; }
}
int MeshLinearToQuadratic::GetNodeMeshNumber(int i)
{
	if(NSubs()==0) { return 0; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeMeshNumber(i); }
	else { return nodeIdx(i-nnsub); }
}
int MeshLinearToQuadratic::ComputeOrientation(int i)
{
	if(NSubs()==0) { return 0; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->ComputeOrientation(i); }
	else { return GetSubordinateComponent()->ComputeOrientation(GetNodePosition(i)); } // use the position from the additional node to let subordinate compute orientation ?
}
Vector3D MeshLinearToQuadratic::GetNodePosition(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodePosition(i); }
	else { return nodes(i-nnsub)->GetNodePos(); }
}

Vector3D MeshLinearToQuadratic::GetNodeRefAngles(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefAngles(i); }
	else { return 0.5*( GetSubordinateComponent()->GetNodeRefAngles(parentnodes(i-nnsub)(1)) + GetSubordinateComponent()->GetNodeRefAngles(parentnodes(i-nnsub)(2)) ); } // linear interpolation for the intermediate nodes
}
Vector3D MeshLinearToQuadratic::GetNodeRefSlope1(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope1(i); }
	else { return 0.5*( GetSubordinateComponent()->GetNodeRefSlope1(parentnodes(i-nnsub)(1)) + GetSubordinateComponent()->GetNodeRefSlope1(parentnodes(i-nnsub)(2)) ); }
}
Vector3D MeshLinearToQuadratic::GetNodeRefSlope2(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope2(i); }
	else { return 0.5*( GetSubordinateComponent()->GetNodeRefSlope2(parentnodes(i-nnsub)(1)) + GetSubordinateComponent()->GetNodeRefSlope2(parentnodes(i-nnsub)(2)) ); }
}
Vector3D MeshLinearToQuadratic::GetNodeRefSlope3(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope3(i); }
	else { return 0.5*( GetSubordinateComponent()->GetNodeRefSlope3(parentnodes(i-nnsub)(1)) + GetSubordinateComponent()->GetNodeRefSlope3(parentnodes(i-nnsub)(2)) ); }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshSplitHexes
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshSplitHexes::GetNodesFor2Prisms(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate)
{
	for(int i=1; i<=2; i++) { nodelist(i).SetLen(6); }
	
	//if(flag_alternate == 0)
	//{
		nodelist(1).Set6( nodelist_hex(1), nodelist_hex(2), nodelist_hex(3), nodelist_hex(5), nodelist_hex(6), nodelist_hex(7) );
		nodelist(2).Set6( nodelist_hex(4), nodelist_hex(3), nodelist_hex(2), nodelist_hex(8), nodelist_hex(7), nodelist_hex(6) );
	//}
	//else if(flag_alternate == 1)
	//{

	//}
	return 12;
}
int MeshSplitHexes::GetNodesFor3Pyramids(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate)
{
	for(int i=1; i<=3; i++) { nodelist(i).SetLen(5); }
	
	if(flag_alternate == 0)
	{
		nodelist(1).SetXN(5, nodelist_hex(1), nodelist_hex(2), nodelist_hex(3), nodelist_hex(4), nodelist_hex(5));
		nodelist(2).SetXN(5, nodelist_hex(6), nodelist_hex(8), nodelist_hex(2), nodelist_hex(4), nodelist_hex(5)); 
		nodelist(3).SetXN(5, nodelist_hex(7), nodelist_hex(3), nodelist_hex(8), nodelist_hex(4), nodelist_hex(5)); 
	}
	else if(flag_alternate == 1)
	{
		nodelist(1).SetXN(5, nodelist_hex(8), nodelist_hex(6), nodelist_hex(7), nodelist_hex(5), nodelist_hex(4));
		nodelist(2).SetXN(5, nodelist_hex(3), nodelist_hex(7), nodelist_hex(1), nodelist_hex(5), nodelist_hex(4)); 
		nodelist(3).SetXN(5, nodelist_hex(2), nodelist_hex(1), nodelist_hex(6), nodelist_hex(5), nodelist_hex(4)); 
	}
	//else if(flag_alternate == 1)
	//{
	//	nodelist(1).SetXN(5, nodelist_hex(8), nodelist_hex(7), nodelist_hex(6), nodelist_hex(5), nodelist_hex(4));
	//	nodelist(2).SetXN(5, nodelist_hex(3), nodelist_hex(1), nodelist_hex(7), nodelist_hex(5), nodelist_hex(4)); 
	//	nodelist(3).SetXN(5, nodelist_hex(2), nodelist_hex(6), nodelist_hex(1), nodelist_hex(5), nodelist_hex(4)); 
	//}
	return 15;
}
int MeshSplitHexes::GetNodesFor5Tets(TArray<int>& nodelist_hex, TArray<TArray<int>>& nodelist, int flag_alternate)
{
	for(int i=1; i<=5; i++) { nodelist(i).SetLen(4); }
//revisited 2016-01-11
	if(flag_alternate == 0)
	{
		nodelist(1).Set4( nodelist_hex(1), nodelist_hex(2), nodelist_hex(3), nodelist_hex(5) );
		nodelist(2).Set4( nodelist_hex(4), nodelist_hex(3), nodelist_hex(2), nodelist_hex(8) );
		nodelist(3).Set4( nodelist_hex(6), nodelist_hex(5), nodelist_hex(8), nodelist_hex(2) );
		nodelist(4).Set4( nodelist_hex(7), nodelist_hex(8), nodelist_hex(5), nodelist_hex(3) ); // C
		nodelist(5).Set4( nodelist_hex(2), nodelist_hex(5), nodelist_hex(8), nodelist_hex(3) ); // C
	}
	else if(flag_alternate == 1)
	{
		nodelist(1).Set4( nodelist_hex(8), nodelist_hex(6), nodelist_hex(7), nodelist_hex(4) ); // CC
		nodelist(2).Set4( nodelist_hex(5), nodelist_hex(7), nodelist_hex(6), nodelist_hex(1) ); // CC
		nodelist(3).Set4( nodelist_hex(3), nodelist_hex(1), nodelist_hex(4), nodelist_hex(7) ); // CC
		nodelist(4).Set4( nodelist_hex(2), nodelist_hex(4), nodelist_hex(1), nodelist_hex(6) ); // CC
		nodelist(5).Set4( nodelist_hex(7), nodelist_hex(4), nodelist_hex(1), nodelist_hex(6) ); // CC
	}
	return 20;
}

int MeshSplitHexes::MeshYourself()
{
	if(NSubs()==0) { return 0; }
	GetSubordinateComponent()->MeshYourself();	
	nodes.Flush();
	elems.Flush();

	int ne = GetSubordinateComponent()->NElems();
//TODO: get discretization from subordinate component to compute the correct alternate flag for a consistent tetrahedral mesh
	for(int i=1; i<=ne; i++) // i ... element number i
	{
		MeshElement* el = GetSubordinateComponent()->GetElemPtr(i);
		TArray<int> nodelist = GetSubordinateComponent()->GetElementNodeList(i);
	
		int split=0;
		//$EK 2017-01-11: add possible TetrahedralStacked to split-routine     //$AD 2018-08-14 added TetrahedralThermo
		if(GetElemType().Compare("Tetrahedral")==1 || GetElemType().Compare("TetrahedralStacked")==1 || 
			GetElemType().Compare("TetrahedralThermo")==1 || GetElemType().Compare("TetrahedralThermoMechanic")==1) { split = 5; }
		else if(GetElemType().Compare("Pyramid")==1 || GetElemType().Compare("PyramidStacked")==1) { split = 3; }
		else if(GetElemType().Compare("Prism")==1 || GetElemType().Compare("PrismStacked")==1) { split = 2; }
		int typenr = mesh->GetComponentElemTypeIdx(this);

		int alternate=0;
		alternate = GetSubordinateComponent()->GetElemAlternateFlag(i);
		
		TArray<TArray<int>> splitlist;
		splitlist.SetLen(split);

		if(el->NNodes()==8) // hexes
		{
			if(split==2)
			{
				GetNodesFor2Prisms(nodelist, splitlist,alternate);
				for(int j=1; j<=split; j++)
				{
					MeshPri elem;
					elem.SetMatNr(el->GetMatNr());
					elem.SetTypeNr(typenr);
					elem.SetNodeList(splitlist(j));
					AddElem(elem);
				}
			}
			else if(split==3)
			{
				GetNodesFor3Pyramids(nodelist, splitlist,alternate);
				for(int j=1; j<=split; j++)
				{
					MeshPyr elem;
					elem.SetMatNr(el->GetMatNr());
					elem.SetTypeNr(typenr);
					elem.SetNodeList(splitlist(j));
					AddElem(elem);
				}
			}
			else if(split==5)
			{
				GetNodesFor5Tets(nodelist, splitlist,alternate);
				for(int j=1; j<=split; j++)
				{
					MeshTet elem;
					elem.SetMatNr(el->GetMatNr());
					elem.SetTypeNr(typenr);
					elem.SetNodeList(splitlist(j));
					AddElem(elem);
				}
			}
		}
	}
	InitNodeIdx();
	InitElemIdx();
	return 1;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshRefine
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MeshNode* MeshRefine::GetNodePtr(int i)
{
	if(NSubs()==0) { return NULL; }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodePtr(i); }
	else { return nodes(i-nnsub); }
}
MeshElement* MeshRefine::GetElemPtr(int i)
{
	if(NSubs()==0) { return NULL; }
	int p = ParentOfIthElement(i);
	int c = ChildNrOfIthElement(i);
	if(c==0) { return GetSubordinateComponent()->GetElemPtr(p); } // unchanged from original mesh
	else { return elems(i); }                                     // element is a result of a split
}
TArray<int> MeshRefine::GetElementNodeList(int i)
{
	if(NSubs()==0) { return NULL; }
	int p = ParentOfIthElement(i);
	int c = ChildNrOfIthElement(i);
	if(c==0) { return GetSubordinateComponent()->GetElementNodeList(i); }
	else { return elems(i)->GetNodeList(); }
}

int MeshRefine::MeshYourself()
{
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();
	sub->MeshYourself();
	nodes.Flush();
	elems.Flush();

// initialize node subdivision level - this should be done by SetElementData when using script or SetSDlvl when using cpp
// TODO: feedback in case the lengths of the sdlvl does not match the number of nodes in the subordinate component
	nodesdlvl.CopyFrom(nodesdlvl_seed);

// initialize mapping of additional elements to elements of subordinate and list of elements
// NOTE: initialize list of elements with NULLS because: (1) GetElemPtr(i) will give correct object,
//                                                       (2) a pointer to the object in the subordinate component causes problems in destructor
//                                                       (3) copying the objects is waste of memory
	int nesub = sub->NElems();
	elementmapping_R2PC.SetLen(nesub);
	for(int i=1; i<=nesub; i++)
	{
		elementmapping_R2PC(i) = int2(i,0);
	}
	elems.SetLen(nesub);
	elems.SetAll(NULL);

// initialize mapping of additional nodes to elements of subordinate
	int nnsub = sub->NNodes();
	unrefinedelementofnode.SetLen(nnsub); unrefinedelementofnode.SetAll(0);
	locposnodeinunrefinedelement.SetLen(nnsub); 
	locposelemcenterunrefinedelement.SetLen(nesub); 
	elementshrinkingfactor.SetLen(nesub); elementshrinkingfactor.SetAll(1.);

	for(int i=1; i<=nesub; i++)
	{
		TArray<int> elemnodes = GetElementNodeList(i); // AD: includes mapping to deeper subordinate components (e.g. in case of Loader)
		for(int j=1; j<=elemnodes.Length(); j++)
		{
			int nnrsub = elemnodes(j);
			if(unrefinedelementofnode(nnrsub)==0)
			{
				unrefinedelementofnode(nnrsub) = i;
				locposnodeinunrefinedelement(nnrsub) = sub->GetElemPtr(i)->GetLocPosOfIthNode(j);
			}
		}
		locposelemcenterunrefinedelement(i) = sub->GetElemPtr(i)->GetLocPosOfCenter();
	}
	
	Refine(this->method); // enter recursive

	InitNodeIdx();
	InitElemIdx();
	return 1;
}

int MeshRefine::Refine(int method)
{
// recursion stop
	int maxsdlvl = 0;
	for(int i=1; i<=nodesdlvl.Length(); i++)
	{
		if(nodesdlvl(i)>maxsdlvl) maxsdlvl = nodesdlvl(i);
	}
	if(maxsdlvl == 0) return 0; // recursion ends when all nodes have subdivision level 0
	
// count the number of child elements for each original element (subordinate)
	int nnsub = GetSubordinateComponent()->NNodes();
	int nesub = GetSubordinateComponent()->NElems();
	TArray<int> childcount(nesub); childcount.SetAll(0);
	for(int i=1; i<=nesub; i++)
	{
		if(elementmapping_R2PC(i)(2) > childcount(i)) childcount(i)=elementmapping_R2PC(i)(2);  
	}
// count number of nodes and elements before refining step
	int nodesunrefined = nodesdlvl.Length();
	int elemsunrefined = elementmapping_R2PC.Length();     

// fix subdivision levels such that only the implemented templates are used
	FixSDLVLList();

// begin with single refinement step: all elements present from previos cycle     
	for(int i=1; i<=elemsunrefined; i++)
	{
		TArray<int> elemnodes = GetElementNodeList(i);
		TArray<int> cornersubdivisionlevel(elemnodes.Length());              // subdivisionlevel for all corner nodes of the current element
		TArray<Vector3D> noderefposlist(elemnodes.Length());                 // "global" positions for all cornernodes considering all subordinate components

		for(int j=1; j<=elemnodes.Length(); j++)
		{                
			int nodenr = elemnodes(j);                                 // nodenumber of the element node in component
			cornersubdivisionlevel(j) = nodesdlvl(nodenr);             // associated subdivisionlevel for the node 
			noderefposlist(j) = GetNodePosition(nodenr);               // global position of node
		}	
	
// NOTE: dont use the nodes subdivision levels for determining if the element is to be split - in case of 1 to 2 refinement this is misleading
// NOTE: for identification of a 1 to 2 refinment pattern the neighboring elements must be considered too !
		MeshElement* el = GetElemPtr(i);                               // GetElemPtr(i) already checks if the element is result of a previous refinement or from subordinate component
		int3 templatepattern = el->IdentifyRequiredTemplate(cornersubdivisionlevel,method);               // look up the template depending on the subdivision level of the corner nodes ( only valid for 1 to 3 refinement ) 
		
		if( !(el->GetMEType()==MEQuad || el->GetMEType()==MEHex))
		{
			mesh->GetMBS()->UO(UO_LVL_warn) << "WARNING: refinement is not implemented for this element type (so far only 4Node-Quads and 8Node-Hexes ! Refinement operation stopped\n";
			return 1;
		}

		if(el->GetMEType()==MEHex && // for the hexahedral: catch 
			   method==0 && (        // in case of full 3d refinement
		   !( templatepattern(1)==0 ||                              // (0**)
			  templatepattern(1)==1 ||                              // (1**)
			 (templatepattern(1)==2 && templatepattern(2)==0) ||    // (20*)
			 (templatepattern(1)==4 && templatepattern(2)==0) ||    // (40*)
			  templatepattern(1)==8 )))                             // (8**)
		{
			mesh->GetMBS()->UO(UO_LVL_warn) << "WARNING: refinement pattern (" << templatepattern(1) << "," << templatepattern(2) << "," << templatepattern(3) << ") not implemented for hexahedral yet! Refinement operation stopped\n";
			return 1;
		}
			
		int nsplitelements = el->GetNumberOfElementsForRefinement(templatepattern,method);	       // determine how many child elements are required
		if(nsplitelements==1)
		{
// no splitting for this element required, do nothing ...
		}
		else
		{
// for the current element: generate all additional nodes
			int nodesbeforeadditional = NNodes();                                                      // nodes before this element is refined (inludes additional from previous elements)
			int nadditionalnodes = el->GetNumberOfAdditionalNodesForRefine(templatepattern,method);    // determine how many additional nodes are required
			TArray<int> additionalnodes;
			for(int j=1; j<=nadditionalnodes; j++)
			{
				Vector3D ploc = el->GetLocPosOfIthAdditionalNode(j,templatepattern,method);            // this is the local position on the template - orientation is already accounted for
				int sdlvl = el->GetSubdivisionLevelOfIthAdditionalNode(j,templatepattern,cornersubdivisionlevel,method);
				Vector3D pglob = el->InterpolateVector3D(ploc, noderefposlist); 
// check if the node has already been added by a previous element in this cycle of refinement
// start searching with newest additional node
				int idx = 0;
				for(int k=nodesbeforeadditional; k>nodesunrefined; k--)
				{
					Vector3D distvec = GetNodePosition(k)-pglob; 
					double dist = abs(distvec.Norm());
					if(dist<=1E-8)
					{
						idx = k;
						break; // required node already generated by a previous element in the same refinement iteration
					}
				}
				if(idx==0)
				{
					MeshNode node;
					node.SetTypeNr(GetNodePtr(1)->GetTypeNr());            // same node type as all the others
					node.SetBodyIdx(GetNodePtr(el->GetNode(1))->GetBodyIdx()); // same body index as all the other nodes of the element
					node.SetNodePos(pglob);
					idx = AddNode(node) + nnsub;                           // index in list of additional nodes (THIS component) plus number of nodes in the original component (SUBORDINATE component)
					nodesdlvl.Add(sdlvl);                                  // assign subdivisionlevel for the new node
		
// track node local positions to unrefined element				
// NOTE:
// the 3-Refinement algorithm ensures that, for multiple iterations, the element that gets refined is always the same shape as the original and NOT a deformed 
// e.g Quad with sdlvl2 in node 3:
//     1st iteration generates 3 elements, the one containing node 3 has local posiitons {(1/3,1/3),(1/3,1),(1,1),(1,1/3)},
//     2nd iteration also generates 3 elements from the previous element
					int unrefinedelem = ParentOfIthElement(i);
					unrefinedelementofnode.Add(unrefinedelem);                // the unrefined element for the additional node
					
					Vector3D locposonchild = ploc;					
					Vector3D locposonoriginal;
					locposonoriginal.X() = locposelemcenterunrefinedelement(i).X() + locposonchild.X() * elementshrinkingfactor(i).X();
					locposonoriginal.Y() = locposelemcenterunrefinedelement(i).Y() + locposonchild.Y() * elementshrinkingfactor(i).Y();
					locposonoriginal.Z() = locposelemcenterunrefinedelement(i).Z() + locposonchild.Z() * elementshrinkingfactor(i).Z();
					locposnodeinunrefinedelement.Add(locposonoriginal);       // the local position on the unrefined element for the additional node				
				}
				additionalnodes.Add(idx);
			} // end loop additinalnodes

// for the current element: generate all child elements
			int parent = elementmapping_R2PC(i)(1);
			int child  = elementmapping_R2PC(i)(2);
			MeshElement* firstchild;  
			if(child==0)
			{
// case: the element to split is an element of the subordinate component
				firstchild = el->GetCopy();                       // make a copy of the object
				firstchild->SetNodeList( GetElementNodeList(i) ); // in the copy use the mapped node list
				elems(i) = firstchild;                            // entry in elementlist of refine component
				elementmapping_R2PC(i)(2) = 1;                    // this is now a first child
				childcount(parent)++;
			}
			else
			{
				firstchild = el;                        // just relabel
			}

// generate child elemenets in sequence   2nd...nth child, then 1st child 
// NOTE: this way no additional blueprint element has to be allocated and deleted, 1st child can be used as lookup
			TArray<int> unrefinednodes = GetElementNodeList(i);
			TArray<int> templatenodelist;
			IVector permutated_nodelist(8), inverse_permutation(8);
			for(int j=2; j<=nsplitelements; j++)
			{
				//firstchild->GetNodeListOfIthRefinedElement(j,templatepattern,templatenodelist); //why doesnt that work? getcopy should keep the class valid
				el->GetNodeListOfIthRefinedElement(j,templatepattern,templatenodelist,method);

				MapTemplateNodeListToActualNodes(templatenodelist, unrefinednodes, additionalnodes); // AD: 2015-10-06: element is now in same orientaiton as the original element

				MeshElement* child = firstchild->GetCopy();

				child->SetNodeList(templatenodelist);

				elems.Add(child);
				childcount(parent)++;
				elementmapping_R2PC.Add(int2(parent,childcount(parent)));
// track node local positions to unrefined element				
				Vector3D locposcenterofchildoriented = el->GetLocPosCenterOfIthRefinedElement(j,templatepattern,method); // orientation of parent already applied
				Vector3D locposcenterofparent = locposelemcenterunrefinedelement(i);
				Vector3D elementshrinkfactorofparent = elementshrinkingfactor(i);
// center of child element in local coordinates of original element:
// take local coordinates of child center on parent
// apply rotation matrix of parent element 
// shrink the vector to relative element size
				Vector3D locposcenterofchildonoriginal;
				locposcenterofchildonoriginal.X() = locposcenterofparent.X() + elementshrinkfactorofparent.X()*locposcenterofchildoriented.X();  // center of child element in local coordinates of unrefined element
				locposcenterofchildonoriginal.Y() = locposcenterofparent.Y() + elementshrinkfactorofparent.Y()*locposcenterofchildoriented.Y();
				locposcenterofchildonoriginal.Z() = locposcenterofparent.Z() + elementshrinkfactorofparent.Z()*locposcenterofchildoriented.Z();
				locposelemcenterunrefinedelement.Add(locposcenterofchildonoriginal);

				Vector3D elementshrinkfactorchild = elementshrinkfactorofparent*(1./3.);  // valid for 3-Refinement
				if (method == 1)
				{
					elementshrinkfactorchild.X() *= 3;
				}
				else if (method == 2)
				{
					elementshrinkfactorchild.Y() *= 3;
				}
				else if (method == 3)
				{
					elementshrinkfactorchild.Z() *= 3;
				}

				elementshrinkingfactor.Add(elementshrinkfactorchild);
int dbgdummy = 1;
			} // end loop splitelements 2..n
// first child changes the nodelist			
			firstchild->GetNodeListOfIthRefinedElement(1,templatepattern,templatenodelist,method);

			MapTemplateNodeListToActualNodes(templatenodelist, unrefinednodes, additionalnodes); // AD: 2015-10-06: element is now in same orientaiton as the original element

			firstchild->SetNodeList(templatenodelist);

			Vector3D locposcenterofchildoriented = el->GetLocPosCenterOfIthRefinedElement(1,templatepattern,method); // orientation of parent already applied
			Vector3D locposcenterofparent = locposelemcenterunrefinedelement(i);
			Vector3D elementshrinkfactorofparent = elementshrinkingfactor(i);

			Vector3D locposcenterofchildonoriginal;  // center of child element in local coordinates of unrefined element
			locposcenterofchildonoriginal.X() = locposcenterofparent.X() + elementshrinkfactorofparent.X()*locposcenterofchildoriented.X();
			locposcenterofchildonoriginal.Y() = locposcenterofparent.Y() + elementshrinkfactorofparent.Y()*locposcenterofchildoriented.Y();
			locposcenterofchildonoriginal.Z() = locposcenterofparent.Z() + elementshrinkfactorofparent.Z()*locposcenterofchildoriented.Z();
			Vector3D elementshrinkfactorchild = elementshrinkfactorofparent*(1./3.); // valid for 3-Refinement
			if (method == 1)
			{
				elementshrinkfactorchild.X() *= 3;
			}
			else if (method == 2)
			{
				elementshrinkfactorchild.Y() *= 3;
			}
			else if (method == 3)
			{
				elementshrinkfactorchild.Z() *= 3;
			}
		// do not add, replace	
			locposelemcenterunrefinedelement(i) = locposcenterofchildonoriginal;
			elementshrinkingfactor(i) = elementshrinkfactorchild;
		} 
	} // end loop unrefinedelements

// reduce the subdivisionlevel of all nodes by one (no less then 0)
	for(int i=1; i<=nodesdlvl.Length(); i++)
	{
		if(nodesdlvl(i)>0) nodesdlvl(i)--;
	}

	return 1+Refine(method); // enter next recursion
}

int MeshRefine::FixSDLVLList()
{
	int elemsunrefined = elementmapping_R2PC.Length();    

	int iterations = 0;

	int flag_hadtofix = 1;
	while(flag_hadtofix==1)
	{
		iterations++;
		flag_hadtofix=0;
		for(int i=1; i<=elemsunrefined; i++)
		{
			MeshElement* el = GetElemPtr(i);                               // GetElemPtr(i) already checks if the element is result of a previous refinement or from subordinate component
			TArray<int> cornersubdivisionlevel(el->NNodes());              // subdivisionlevel for all corner nodes of the current element

			for(int j=1; j<=el->NNodes(); j++)
			{
				int nodenr = el->GetNode(j);                               // nodenumber of the element node in component
				cornersubdivisionlevel(j) = nodesdlvl(nodenr);             // associated subdivisionlevel for the node 
			}

			if(method==0) // tridirectional refinement (Task: all elements must have subdivision level to fit an implemented template)
			{
				if(!el->IsTemplateImplemented(cornersubdivisionlevel))
				{
					flag_hadtofix=1;
					for(int j=1; j<=el->NNodes(); j++)
					{
						int nodenr = el->GetNode(j);                           // nodenumber of the element node in component
						nodesdlvl(nodenr) = cornersubdivisionlevel(j);         // correct the global node subdivision level due to missing templates
					}
				}
			}
			else if(method==1) // bidirectional refinement y and z (Task: match face 1 and 2 - all templates are implemented)
			{
			// force subdivision level of node 1 and 2 both zero or both nonzero
				if(cornersubdivisionlevel(1)>0 && cornersubdivisionlevel(2)==0) { nodesdlvl(el->GetNode(2))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(1)==0 && cornersubdivisionlevel(2)>0) { nodesdlvl(el->GetNode(1))++; flag_hadtofix=1; }
			// same for pairs 3-4, 5-6, 7-8
				if(cornersubdivisionlevel(3)>0 && cornersubdivisionlevel(4)==0) { nodesdlvl(el->GetNode(4))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(3)==0 && cornersubdivisionlevel(4)>0) { nodesdlvl(el->GetNode(3))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(5)>0 && cornersubdivisionlevel(6)==0) { nodesdlvl(el->GetNode(6))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(5)==0 && cornersubdivisionlevel(6)>0) { nodesdlvl(el->GetNode(5))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(7)>0 && cornersubdivisionlevel(8)==0) { nodesdlvl(el->GetNode(8))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(7)==0 && cornersubdivisionlevel(8)>0) { nodesdlvl(el->GetNode(7))++; flag_hadtofix=1; }
			}
			else if(method==2) // bidirectional refinement x and z (Task: match face 3 and 4 - all templates are implemented)
			{
			// force subdivision level of node 1 and 3 both zero or both nonzero
				if(cornersubdivisionlevel(1)>0 && cornersubdivisionlevel(3)==0) { nodesdlvl(el->GetNode(3))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(1)==0 && cornersubdivisionlevel(3)>0) { nodesdlvl(el->GetNode(1))++; flag_hadtofix=1; }
			// same for pairs 2-4, 5-7, 6-8
				if(cornersubdivisionlevel(2)>0 && cornersubdivisionlevel(4)==0) { nodesdlvl(el->GetNode(4))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(2)==0 && cornersubdivisionlevel(4)>0) { nodesdlvl(el->GetNode(2))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(5)>0 && cornersubdivisionlevel(7)==0) { nodesdlvl(el->GetNode(7))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(5)==0 && cornersubdivisionlevel(7)>0) { nodesdlvl(el->GetNode(5))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(6)>0 && cornersubdivisionlevel(8)==0) { nodesdlvl(el->GetNode(8))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(6)==0 && cornersubdivisionlevel(8)>0) { nodesdlvl(el->GetNode(6))++; flag_hadtofix=1; }
			}
			else if(method==3) // bidirectional refinement x and y (Task: match face 5 and 6 - all templates are implemented)
			{
			// force subdivision level of node 1 and 5 both zero or both nonzero
				if(cornersubdivisionlevel(1)>0 && cornersubdivisionlevel(5)==0) { nodesdlvl(el->GetNode(5))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(1)==0 && cornersubdivisionlevel(5)>0) { nodesdlvl(el->GetNode(1))++; flag_hadtofix=1; }
			// same for pairs 2-6, 3-7, 4-8
				if(cornersubdivisionlevel(2)>0 && cornersubdivisionlevel(6)==0) { nodesdlvl(el->GetNode(6))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(2)==0 && cornersubdivisionlevel(6)>0) { nodesdlvl(el->GetNode(2))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(3)>0 && cornersubdivisionlevel(7)==0) { nodesdlvl(el->GetNode(7))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(3)==0 && cornersubdivisionlevel(7)>0) { nodesdlvl(el->GetNode(3))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(4)>0 && cornersubdivisionlevel(8)==0) { nodesdlvl(el->GetNode(8))++; flag_hadtofix=1; }
				if(cornersubdivisionlevel(4)==0 && cornersubdivisionlevel(8)>0) { nodesdlvl(el->GetNode(4))++; flag_hadtofix=1; }
			}			
		}
	}

	return iterations;
}

int MeshRefine::MapTemplateNodeListToActualNodes(TArray<int>& templatenodelist, TArray<int>& unrefinedelementnodes, TArray<int>& additionalnodes)
{
	for(int i=1; i<=templatenodelist.Length(); i++)
	{
		int entry = templatenodelist(i);
		if(entry<0) 
		{ 
			templatenodelist(i) = unrefinedelementnodes(-entry); // this is a node of the unrefined element
		}
		else 
		{ 
			templatenodelist(i) = additionalnodes(entry); // this is an additional node
		}
	}
 return 0;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshExtrude
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshExtrude::MeshYourself()
{
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();
	sub->MeshYourself();	

	GenerateOwnNodes();
	GenerateOwnElements();
	InitNodeIdx();
	InitElemIdx();
	return 1;
}
int MeshExtrude::GenerateOwnNodes()
{
	nodes.Flush();
	nodes.SetLen(0);
	nodemapping.Flush(); 

	MeshComponent* sub = GetSubordinateComponent();
	Vector3D pos;
	MeshNode n;

	Vector3D vec(0.,0.,0.);
	double dist = ExtrusionDistance()/ExtrusionDiscretization();
	if(ExtrusionAxis()==1) { vec.X()=dist; }
	else if (ExtrusionAxis()==2) { vec.Y()=dist; }
	else if (ExtrusionAxis()==3) { vec.Z()=dist; }
	else return 0;

	TArray<int> sourcenodedim;
	sourcenodedim.SetLen(sub->NNodes());
	sourcenodedim.SetAll(0);
	for(int i=1; i<=sub->NElems(); i++)
	{
		int eldim = sub->GetElemPtr(i)->Dim();
		if(eldim>2) { GetMesh()->GetMBS()->UO() << "ERROR MeshExtrude::GenerateOwnNodes - can not extrude a Mesh containing a 3D element"; return 0; }
		TArray<int> nodelist = sub->GetElementNodeList(i);
		for(int j=1; j<=nodelist.Length(); j++)
		{
			int nnr = nodelist(j);
			if(sourcenodedim(nnr) == 0) { sourcenodedim(nnr) = eldim; } // node not yet visited
			else if(sourcenodedim(nnr) == eldim) { ; } // in agreement with previous value
			else { GetMesh()->GetMBS()->UO() << "ERROR MeshExtrude::GenerateOwnNodes - at least one node is incuded in elements of different dimensionality"; return 0;}
		}
	}
	int typenr2D = mesh->GetComponentNodeTypeIdx(nodetype2D);
	int typenr3D = mesh->GetComponentNodeTypeIdx(nodetype3D);

	// generate slices along the extrusion axis
	for(int j=0; j<=extrusion_discretization; j++)
	{
		Vector3D shift = vec*j;
		for(int i=1; i<=sub->NNodes(); i++)
		{
			if(sourcenodedim(i)==1) { n.SetTypeNr(typenr2D); }
			else if(sourcenodedim(i)==2) { n.SetTypeNr(typenr3D); }
			else { GetMesh()->GetMBS()->UO() << "ERROR MeshExtrude::GenerateOwnNodes source node is not of dimension 1 or 2"; return 0; }
			
			pos = sub->GetNodePosition(i);
			pos = pos + shift;

			n.SetBodyIdx(sub->GetNodePtr(i)->GetBodyIdx());  // same domain as the original node
			n.SetNodePos(pos);
			
			int nnr = AddNode(n);
			nodemapping.Add(nnr);
		}
	}
	return NNodes();
}
int MeshExtrude::GenerateOwnElements()
{
	elems.Flush();
	elems.SetLen(0);

	MeshComponent* sub = GetSubordinateComponent();
	int typenr2D = mesh->GetComponentElemTypeIdx(elemtype2D);
	int typenr3DHex = mesh->GetComponentElemTypeIdx(elemtype3DHex);
	int typenr3DPri = mesh->GetComponentElemTypeIdx(elemtype3DPri);

	for(int j=1; j<=extrusion_discretization; j++)
	{
		int off1 = sub->NNodes()*(j-1);   // offset for the 1st slice
		int off2 = sub->NNodes()*(j);     // offset for the 2nd slice
		for(int i=1; i<=sub->NElems(); i++)
		{
			MeshElement* el = sub->GetElemPtr(i);
			TArray<int> nodelistsource = sub->GetElementNodeList(i); // use nodelistsource instead of el->GetNode() - MeshGlue requires mapping

			if(el->Dim()==1)
			{
				MeshQuad quad;
				TArray<int> nodelist;
				nodelist(1) = nodemapping(nodelistsource(1)+off1); nodelist(2) = nodemapping(nodelistsource(2)+off1);
				nodelist(4) = nodemapping(nodelistsource(1)+off2); nodelist(3) = nodemapping(nodelistsource(2)+off2);
				quad.SetTypeNr(typenr2D);
				quad.SetMatNr(el->GetMatNr());
				quad.SetNodeList(nodelist);
				AddElem(quad);
			}
			else if(el->Dim()==2)
			{
				if(el->NNodes()==3)
				{
					MeshPri prism;
					TArray<int> nodelist;
					nodelist(1) = nodemapping(nodelistsource(1)+off1); nodelist(2) = nodemapping(nodelistsource(2)+off1); nodelist(3) = nodemapping(nodelistsource(3)+off1);
					nodelist(4) = nodemapping(nodelistsource(1)+off2); nodelist(5) = nodemapping(nodelistsource(2)+off2); nodelist(6) = nodemapping(nodelistsource(3)+off2);						
					prism.SetTypeNr(typenr3DPri);
					prism.SetMatNr(el->GetMatNr());
					prism.SetNodeList(nodelist);
					AddElem(prism);						
				}
				else // if(el->NNodes()==4)
				{
					MeshHex hex;
					TArray<int> nodelist;
					nodelist(1) = nodemapping(nodelistsource(1)+off1); nodelist(2) = nodemapping(nodelistsource(2)+off1); nodelist(4) = nodemapping(nodelistsource(3)+off1); nodelist(3) = nodemapping(nodelistsource(4)+off1);
					nodelist(5) = nodemapping(nodelistsource(1)+off2); nodelist(6) = nodemapping(nodelistsource(2)+off2); nodelist(8) = nodemapping(nodelistsource(3)+off2); nodelist(7) = nodemapping(nodelistsource(4)+off2);						
					hex.SetTypeNr(typenr3DHex);
					hex.SetMatNr(el->GetMatNr());
					hex.SetNodeList(nodelist);
					AddElem(hex);						
				}	
			}
		}
	}
	return NElems();
}
Vector3D MeshExtrude::GetNodeRefSlope1(int i)
{
	if(subordinate!=NULL)
	{
		MeshComponent* sub = GetSubordinateComponent();
		int nnrSource = (i-1)%(sub->NNodes())+1;
		sub->ComputeOrientation(nnrSource);
// Node3DS1S2 from Node3DRxyz
		Vector3D refangles = sub->GetNodeRefAngles(nnrSource);
		double roll =  refangles.X();
		double yaw =  -refangles.Y();
		double pitch = refangles.Z();
		Vector3D dir = Vector3D( cos(yaw)*cos(pitch),
														 cos(yaw)*sin(pitch),
														 sin(yaw) );
		return dir;
	}
	else { return Vector3D(0.,0.,0.); }
}
Vector3D MeshExtrude::GetNodeRefSlope2(int i)
{
// Node3DS1S2 from Node3DRxyz
	if(ExtrusionAxis()==1) { return Vector3D(1.,0.,0.); }
	else if (ExtrusionAxis()==2) { return Vector3D(0.,1.,0.); }
	else if (ExtrusionAxis()==3) { return Vector3D(0.,0.,1.); }
	else return Vector3D(0.,0.,0.);
}
Vector3D MeshExtrude::GetElementSizeVector(int i)
{
	if(subordinate!=NULL)
	{
		MeshComponent* sub = GetSubordinateComponent();
		int enrSource = (i-1)%(sub->NElems())+1;
// ANCFThinPlate3D from LinearBeam3D
		double thick = ExtrusionPlateThickness();
		double beamsize = sub->GetElementSizeVector(enrSource).Y();
		double dist = ExtrusionDistance()/ExtrusionDiscretization(); 
		return Vector3D(thick,beamsize,dist);
	}
	else { return Vector3D(0.,0.,0.); }
}
mystr MeshExtrude::GetElemType(int i)
{
	if(GetElemPtr(i)->Dim()==2) return elemtype2D;
	else return mystr("");
}
int MeshExtrude::NCoincident()
{
	if(subordinate!=NULL) 
	{
		return GetSubordinateComponent()->NCoincident()*(ExtrusionDiscretization()+1); 
	}
	else { return 0; }
}
int2 MeshExtrude::GetCoincidentNodeNumbers(int i)
{
	if(subordinate!=NULL)
	{
		int pairnrSource = ((i-1)%GetSubordinateComponent()->NCoincident())+1;
		int2 nnrsSource = GetSubordinateComponent()->GetCoincidentNodeNumbers( pairnrSource );

		int slice0 = ((i-1)/GetSubordinateComponent()->NCoincident());
		int offset = GetSubordinateComponent()->NNodes()*slice0;

		int2 nnrsExtrude = int2( nodemapping( nnrsSource(1)+offset ),
			                       nodemapping( nnrsSource(2)+offset ));
		return nnrsExtrude;
	}
	else { return int2(0,0); }	
}
int MeshExtrude::GetCoincidentPenaltyFlag(int i)
{
	if(subordinate!=NULL)
	{
		int pairnrSource = ((i-1)%GetSubordinateComponent()->NCoincident())+1;
		return GetSubordinateComponent()->GetCoincidentPenaltyFlag(pairnrSource);
	}
	else { return 0; }	
}
void MeshExtrude::SetAutoConstraintNumber(int i, int cnr)
{
	if(subordinate!=NULL)
	{
		GetSubordinateComponent()->SetAutoConstraintNumber(i, cnr);
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshRotational
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshRotational::MeshYourself()
{
// NOTE: Triangular Source Mesh is assumed	
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();
	sub->MeshYourself();	

	GenerateOwnNodes();
	GenerateOwnElements();
	InitNodeIdx();
	InitElemIdx();
	return 1;
}

int MeshRotational::GenerateOwnNodes()
{
	nodes.Flush();
	nodes.SetLen(0);
	nodemapping.Flush(); 

	MeshComponent* sub = GetSubordinateComponent();
	Vector3D pos;
	MeshNode n;

	int typenr = mesh->GetComponentNodeTypeIdx(this);	// interface to ObjectFactory via Mesh     the type is ALWAYS Node3D
	n.SetTypeNr(typenr); // define node type (ObjectFactory type number) - all nodes of component must have same node type
	//AD: moved line: body number is now same as for 0 node

// check if node is on rotation axis
	TArray<int> onaxis;
	for(int i=1; i<=sub->NNodes(); i++)
	{
		pos = sub->GetNodePosition(i);
		if (rotation_axis_number == 1)      { if( abs(pos.Y())<=1e-10 && abs(pos.Z())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else if (rotation_axis_number == 2) { if( abs(pos.X())<=1e-10 && abs(pos.Z())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else if (rotation_axis_number == 3) { if( abs(pos.X())<=1e-10 && abs(pos.Y())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else return 0;
	}

// generate the 3d nodes from the 2d source - slices of
	for(int j=0; j<=angular_segments; j++)
	{
		double angle = (MY_PI/180.) * (final_angle_deg/angular_segments) * j;
		Matrix3D rot;
		if (rotation_axis_number == 1) { rot = RotMatrix1(angle); }
		else if (rotation_axis_number == 2) { rot = RotMatrix2(angle); }
		else if (rotation_axis_number == 3) { rot = RotMatrix3(angle); }
		else { return 0; }

		if(j==angular_segments && abs(final_angle_deg-360.)<=1e-10 ) // last slice for a full rotation - use nodes from first slice
		{
			for(int i=1; i<=sub->NNodes(); i++)
			{
				nodemapping.Add(i);		
			}
		}
		else 
		{
			for(int i=1; i<=sub->NNodes(); i++) 	// node is on axis - use number from first slice
			{
				if(j!=0 && onaxis(i)) 
				{
					nodemapping.Add(i);
				}
				else
				{
					pos = sub->GetNodePosition(i);
					pos = rot*pos;
				
					n.SetNodePos(pos);
					n.SetBodyIdx(sub->GetBodyIndex(i));  // same body index as original node

					int nnr = AddNode(n);
					nodemapping.Add(nnr);
				}
			}
		}
	}
	return NNodes();
}
int MeshRotational::GenerateOwnElements()
{	
	elems.Flush();
	elems.SetLen(0);
	
	MeshComponent* sub = GetSubordinateComponent();
	Vector3D pos;
	int typenrpri = mesh->GetComponentElemTypeIdx(typepri);
	int typenrpyr = mesh->GetComponentElemTypeIdx(typepyr);
	int typenrtet = mesh->GetComponentElemTypeIdx(typetet);
	int typenrhex = mesh->GetComponentElemTypeIdx(typehex);

// check if node is on rotation axis
	TArray<int> onaxis;
	for(int i=1; i<=sub->NNodes(); i++)
	{
		pos = sub->GetNodePosition(i);
		if (rotation_axis_number == 1)      { if( abs(pos.Y())<=1e-10 && abs(pos.Z())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else if (rotation_axis_number == 2) { if( abs(pos.X())<=1e-10 && abs(pos.Z())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else if (rotation_axis_number == 3) { if( abs(pos.X())<=1e-10 && abs(pos.Y())<=1e-10 ) { onaxis.Add(1); } else { onaxis.Add(0); } }
		else return 0;
	}
	// count and identify nodes on axis for all elements 
	TArray<int2> elaxis;
	elaxis.SetLen(sub->NElems());
	elaxis.SetAll(int2(0,0));
	for(int j=1; j<=sub->NElems(); j++)
	{
		TArray<int> nl = sub->GetElementNodeList(j);

//		MeshElement* el = sub->GetElemPtr(j);
		int pow = 1;
		for(int i=1; i<=nl.Length(); i++)
		{
			int nnr=nl(i);
			if(onaxis(nnr)!=0)
			{
				elaxis(j)(1)++;                   // counter
				elaxis(j)(2)+=pow;                // bitcode
			}
			pow*=2;
		}
	}
	
	for(int j=1; j<=angular_segments; j++)
	{
		int off1 = sub->NNodes()*(j-1);   // offset for the 1st slice
		int off2 = sub->NNodes()*(j);     // offset for the 2nd slice
		for(int i=1; i<=sub->NElems(); i++)
		{
			MeshElement* el = sub->GetElemPtr(i);
			
			if(el->NNodes()==4) // original element is quad
			{
				TArray<int> nl4 = sub->GetElementNodeList(i);

				if(elaxis(i)(1)==0) // no node on axis -> hex , !!! mind Quad to Hex 3<->4 node order !!!
				{
					MeshHex hex;
					TArray<int> nodelist;
					nodelist(1) = nodemapping(nl4(1)+off1); nodelist(2) = nodemapping(nl4(2)+off1); nodelist(3) = nodemapping(nl4(4)+off1); nodelist(4) = nodemapping(nl4(3)+off1);
					nodelist(5) = nodemapping(nl4(1)+off2); nodelist(6) = nodemapping(nl4(2)+off2); nodelist(7) = nodemapping(nl4(4)+off2); nodelist(8) = nodemapping(nl4(3)+off2);
					hex.SetTypeNr(typenrhex);
					hex.SetMatNr(el->GetMatNr());
					hex.SetNodeList(nodelist);
					AddElem(hex);
				}
				if(elaxis(i)(1)==1) // one nodes on axis -> pyramid and prism
				{
					MeshPyr pyramid;
					TArray<int> nodelistpyr;
					MeshPri prism;
					TArray<int> nodelistpri;
					if(elaxis(i)(2) & 1)      {  nodelistpyr(1) = nodemapping(nl4(4)+off1); nodelistpyr(2) = nodemapping(nl4(2)+off1);
					                             nodelistpyr(3) = nodemapping(nl4(4)+off2); nodelistpyr(4) = nodemapping(nl4(2)+off2); nodelistpyr(5) = nodemapping(nl4(1)+off1); 
												 nodelistpri(1) = nodemapping(nl4(3)+off1); nodelistpri(2) = nodemapping(nl4(4)+off1); nodelistpri(3) = nodemapping(nl4(2)+off1);
												 nodelistpri(4) = nodemapping(nl4(3)+off2); nodelistpri(5) = nodemapping(nl4(4)+off2); nodelistpri(6) = nodemapping(nl4(2)+off2); }
					else if(elaxis(i)(2) & 2) {  nodelistpyr(1) = nodemapping(nl4(1)+off1); nodelistpyr(2) = nodemapping(nl4(3)+off1);
					                             nodelistpyr(3) = nodemapping(nl4(1)+off2); nodelistpyr(4) = nodemapping(nl4(3)+off2); nodelistpyr(5) = nodemapping(nl4(2)+off1); 
												 nodelistpri(1) = nodemapping(nl4(4)+off1); nodelistpri(2) = nodemapping(nl4(1)+off1); nodelistpri(3) = nodemapping(nl4(3)+off1);
												 nodelistpri(4) = nodemapping(nl4(4)+off2); nodelistpri(5) = nodemapping(nl4(1)+off2); nodelistpri(6) = nodemapping(nl4(3)+off2); }
					else if(elaxis(i)(2) & 4) {  nodelistpyr(1) = nodemapping(nl4(2)+off1); nodelistpyr(2) = nodemapping(nl4(4)+off1);
					                             nodelistpyr(3) = nodemapping(nl4(2)+off2); nodelistpyr(4) = nodemapping(nl4(4)+off2); nodelistpyr(5) = nodemapping(nl4(3)+off1); 
												 nodelistpri(1) = nodemapping(nl4(1)+off1); nodelistpri(2) = nodemapping(nl4(2)+off1); nodelistpri(3) = nodemapping(nl4(4)+off1);
												 nodelistpri(4) = nodemapping(nl4(1)+off2); nodelistpri(5) = nodemapping(nl4(2)+off2); nodelistpri(6) = nodemapping(nl4(4)+off2); }
					else if(elaxis(i)(2) & 8) {  nodelistpyr(1) = nodemapping(nl4(3)+off1); nodelistpyr(2) = nodemapping(nl4(1)+off1);
					                             nodelistpyr(3) = nodemapping(nl4(3)+off2); nodelistpyr(4) = nodemapping(nl4(1)+off2); nodelistpyr(5) = nodemapping(nl4(4)+off1); 
												 nodelistpri(1) = nodemapping(nl4(2)+off1); nodelistpri(2) = nodemapping(nl4(3)+off1); nodelistpri(3) = nodemapping(nl4(1)+off1);
												 nodelistpri(4) = nodemapping(nl4(2)+off2); nodelistpri(5) = nodemapping(nl4(3)+off2); nodelistpri(6) = nodemapping(nl4(1)+off2); }
					pyramid.SetTypeNr(typenrpyr);
					pyramid.SetMatNr(el->GetMatNr());
					pyramid.SetNodeList(nodelistpyr);
					AddElem(pyramid);
					prism.SetTypeNr(typenrpri);
					prism.SetMatNr(el->GetMatNr());
					prism.SetNodeList(nodelistpri);
					AddElem(prism);					
				}
				if(elaxis(i)(1)==2) // two nodes on axis -> prism
				{
					MeshPri prism;
					TArray<int> nodelist;
					if(elaxis(i)(2) == 3)      { nodelist(1) = nodemapping(nl4(1)+off1); nodelist(2) = nodemapping(nl4(4)+off1); nodelist(3) = nodemapping(nl4(4)+off2);
					                             nodelist(4) = nodemapping(nl4(2)+off1); nodelist(5) = nodemapping(nl4(3)+off1); nodelist(6) = nodemapping(nl4(3)+off2); }
					else if(elaxis(i)(2) == 6) { nodelist(1) = nodemapping(nl4(2)+off1); nodelist(2) = nodemapping(nl4(1)+off1); nodelist(3) = nodemapping(nl4(1)+off2);
					                             nodelist(4) = nodemapping(nl4(3)+off1); nodelist(5) = nodemapping(nl4(4)+off1); nodelist(6) = nodemapping(nl4(4)+off2); }
					else if(elaxis(i)(2) ==12) { nodelist(1) = nodemapping(nl4(3)+off1); nodelist(2) = nodemapping(nl4(2)+off1); nodelist(3) = nodemapping(nl4(2)+off2);
					                             nodelist(4) = nodemapping(nl4(4)+off1); nodelist(5) = nodemapping(nl4(1)+off1); nodelist(6) = nodemapping(nl4(1)+off2); }
					else if(elaxis(i)(2) == 9) { nodelist(1) = nodemapping(nl4(4)+off1); nodelist(2) = nodemapping(nl4(3)+off1); nodelist(3) = nodemapping(nl4(3)+off2);
					                             nodelist(4) = nodemapping(nl4(1)+off1); nodelist(5) = nodemapping(nl4(2)+off1); nodelist(6) = nodemapping(nl4(2)+off2); }
					prism.SetTypeNr(typenrpri);
					prism.SetMatNr(el->GetMatNr());
					prism.SetNodeList(nodelist);
					AddElem(prism);
				}
			}
			else if(el->NNodes()==3) // original element is trig
			{
				TArray<int> nl3 = sub->GetElementNodeList(i);

				if(elaxis(i)(1)==0) // no node on axis -> prism
				{
					MeshPri prism;
					TArray<int> nodelist;
					nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(2)+off1); nodelist(3) = nodemapping(nl3(3)+off1);
					nodelist(4) = nodemapping(nl3(1)+off2); nodelist(5) = nodemapping(nl3(2)+off2); nodelist(6) = nodemapping(nl3(3)+off2);
					prism.SetTypeNr(typenrpri);
					prism.SetMatNr(el->GetMatNr());
					prism.SetNodeList(nodelist);
					AddElem(prism);
				}
				else if(elaxis(i)(1)==1) // one node on axis -> pyramid
				{
					MeshPyr pyramid;
					TArray<int> nodelist;
					// to do permutaion -> one side is always drawn in brighter color....   
					if(elaxis(i)(2) & 1) 
					{      
						//nodelist(1) = nodemapping(nl3(2)+off1); nodelist(2) = nodemapping(nl3(3)+off1);                               
						//nodelist(4) = nodemapping(nl3(2)+off2); nodelist(3) = nodemapping(nl3(3)+off2); nodelist(5) = nodemapping(nl3(1)+off1); 
						nodelist(1) = nodemapping(nl3(3)+off1); nodelist(2) = nodemapping(nl3(2)+off1);                               
						nodelist(3) = nodemapping(nl3(3)+off2); nodelist(4) = nodemapping(nl3(2)+off2); nodelist(5) = nodemapping(nl3(1)+off1);
					}
					else if(elaxis(i)(2) & 2) 
					{ 
						//nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(3)+off1);
						//nodelist(4) = nodemapping(nl3(1)+off2); nodelist(3) = nodemapping(nl3(3)+off2); nodelist(5) = nodemapping(nl3(2)+off1); 
						nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(3)+off1);
						nodelist(3) = nodemapping(nl3(1)+off2); nodelist(4) = nodemapping(nl3(3)+off2); nodelist(5) = nodemapping(nl3(2)+off1); 				
					}
					else if(elaxis(i)(2) & 4) 
					{ 
						//nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(2)+off1);
					 //   nodelist(4) = nodemapping(nl3(1)+off2); nodelist(3) = nodemapping(nl3(2)+off2); nodelist(5) = nodemapping(nl3(3)+off1); 
						nodelist(1) = nodemapping(nl3(2)+off1); nodelist(2) = nodemapping(nl3(1)+off1);
					    nodelist(3) = nodemapping(nl3(2)+off2); nodelist(4) = nodemapping(nl3(1)+off2); nodelist(5) = nodemapping(nl3(3)+off1); 
					}
					else { ; }
					pyramid.SetTypeNr(typenrpyr);
					pyramid.SetMatNr(el->GetMatNr());
					pyramid.SetNodeList(nodelist);
					AddElem(pyramid);
				}
				else if(elaxis(i)(1)==2) // two nodes on axis -> tet
				{
					MeshTet tetraeder;
					TArray<int> nodelist;
					if(elaxis(i)(2) == 6) {      nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(2)+off1); nodelist(3) = nodemapping(nl3(3)+off1); nodelist(4) = nodemapping(nl3(1)+off2); }
					else if(elaxis(i)(2) == 5) { nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(2)+off1); nodelist(3) = nodemapping(nl3(3)+off1); nodelist(4) = nodemapping(nl3(2)+off2); }
					else if(elaxis(i)(2) == 3) { nodelist(1) = nodemapping(nl3(1)+off1); nodelist(2) = nodemapping(nl3(2)+off1); nodelist(3) = nodemapping(nl3(3)+off1); nodelist(4) = nodemapping(nl3(3)+off2); }
					else { ; }
					tetraeder.SetTypeNr(typenrtet);
					tetraeder.SetMatNr(el->GetMatNr());
					tetraeder.SetNodeList(nodelist);
					AddElem(tetraeder);
				}
			}	
		}
	}
	return NElems();
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshMirror
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MeshNode* MeshMirror::GetNodePtr(int i)
{
	if(NSubs()==0) { return NULL; }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodePtr(i); }
	else { return nodes(i-nnsub); }
}
MeshElement* MeshMirror::GetElemPtr(int i)
{
	if(NSubs()==0) { return NULL; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElemPtr(i); }
	else { return elems(i-nesub); }
}
int MeshMirror::MeshYourself()
{
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();
	sub->MeshYourself();	

	GenerateOwnNodes();
	GenerateOwnElements();
	InitNodeIdx();
	InitElemIdx();
	return 1;
}
int MeshMirror::GenerateOwnNodes()
{
	nodes.Flush();
	nodes.SetLen(0);
	nodemapping.Flush(); 
	nodemapping = NaturalNumbers(GetSubordinateComponent()->NNodes());

	int nodetypenrNode3D = this->mesh->GetComponentNodeTypeIdx(mystr("Node3D"));
	int nodetypenrNode2D = this->mesh->GetComponentNodeTypeIdx(mystr("Node2D"));

	for(int i=1; i<=GetSubordinateComponent()->NNodes(); i++)
	{
		Vector3D pos = GetSubordinateComponent()->GetNodePosition(i);

		if(mirror_plane_number==1 || mirror_plane_number==2 || mirror_plane_number==3)
		{
			double d = pos(mirror_plane_number)-mirror_distance;
			int nodetypenr = GetSubordinateComponent()->GetNodePtr(i)->GetTypeNr();

// generate duplicate of node ?
// NOT when on plane and non-oriented node type
			if(abs(d) < 1e-8 && 
				 (nodetypenr == nodetypenrNode3D || nodetypenr == nodetypenrNode2D) )
			{
				nodemapping.Add(i); // this happens ONLY for non-oriented nodes on the mirror plane
			}
			else
			{
				pos(mirror_plane_number) = mirror_distance - d;

				MeshNode node;
				node.SetTypeNr(GetSubordinateComponent()->GetNodePtr(i)->GetTypeNr());
				node.SetBodyIdx(GetSubordinateComponent()->GetNodePtr(i)->GetBodyIdx());
				node.SetNodePos(pos);

				int nnr = AddNode(node)+GetSubordinateComponent()->NNodes();
				nodemapping.Add(nnr);
			}
		}
	}
	return NNodes();
}

int MeshMirror::GenerateOwnElements()
{
	elems.Flush();
	elems.SetLen(GetSubordinateComponent()->NElems());

	for(int i=1; i<=GetSubordinateComponent()->NElems(); i++)
	{
		elems(i) = GetSubordinateComponent()->GetElemPtr(i)->GetCopy(); // clone the element from the existing - THIS KEEPS THE TYPE - Adding to the list is also done (important)
		MeshElement* clone = elems(i);

		TArray<int> nodelist = GetSubordinateComponent()->GetElementNodeList(i); // get the nodelist via the subordinate nevertheless - any mapping (e.g.Glue) would be lost otherwise
// WRONG: TArray<int> nodelist = clone->GetNodeList();
		int nnsub = GetSubordinateComponent()->NNodes();
		for(int j=1; j<=nodelist.Length(); j++)
		{
			nodelist(j) = nodemapping(nodelist(j)+nnsub);
		}
		clone->SetNodeList(nodelist);
		clone->Invert();               // no invert for plates ?
		
		// already added ...
	}
	return NElems();
}
int MeshMirror::GetNodeObjectFactoryType(int i)
{
	if(NSubs()==0) { return 0; }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeObjectFactoryType(i); }
	else { return nodes(i-nnsub)->GetTypeNr(); }
}
void MeshMirror::SetNodeMeshNumber(int i, int idx)
{
	if(NSubs()==0) { return; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { GetSubordinateComponent()->SetNodeMeshNumber(i,idx); }
	else { nodeIdx(i-nnsub) = idx; }
}
int MeshMirror::GetNodeMeshNumber(int i)
{
	if(NSubs()==0) { return 0; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeMeshNumber(i); }
	else { return nodeIdx(i-nnsub); }
}
int MeshMirror::ComputeOrientation(int i)
{
//NOTE AD: how it should work: compute the orientation for the "original node" 
//                             when accessing the orientation vectors
	if(NSubs()==0) { return 0; } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->ComputeOrientation(i); }
	else { return GetSubordinateComponent()->ComputeOrientation(i-nnsub); }
}
Vector3D MeshMirror::GetNodePosition(int i)
{
	if(NSubs()==0) { return Vector3D(0.,0.,0.); } 
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodePosition(i); }
	else { return nodes(i-nnsub)->GetNodePos(); }
}
Vector3D MeshMirror::GetNodeRefSlope1(int i)
{
	if(NSubs()==0) { return Vector3D(0,0,0); }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope1(i); }
	else
	{
		Vector3D ref = GetSubordinateComponent()->GetNodeRefSlope1(i-nnsub);
// TODO: TEST this with a plate...
		// refslope1 computed as (cosYaw*cosPitch,cosYaw*sinPitch,sinYaw)
		if(mirror_plane_number==1)      { ref.X() *= -1.; }			
		else if(mirror_plane_number==2) { ref.Y() *= -1.; }				
		else if(mirror_plane_number==3) { ref.Z() *= -1.; }			
		else { return Vector3D(0.,0.,0.); }
		return ref;
	}
}
Vector3D MeshMirror::GetNodeRefSlope2(int i)
{
	if(NSubs()==0) { return Vector3D(0,0,0); }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope2(i); }
	else
	{
		Vector3D ref = GetSubordinateComponent()->GetNodeRefSlope2(i-nnsub);
		// assume roll==0.: refslope2 computed as (-sinPitch, cosPitch, 0)
		if(mirror_plane_number==1)			{ ref.X() *= -1.; }				
		else if(mirror_plane_number==2) { ref.Y() *= -1.; }				
		else if(mirror_plane_number==3) { ref.Z() *= -1.; }											
		else { return Vector3D(0.,0.,0.); }
		return ref;
	}
}
Vector3D MeshMirror::GetNodeRefSlope3(int i)
{
	if(NSubs()==0) { return Vector3D(0,0,0); }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefSlope3(i); }
	else
	{
		Vector3D ref = GetSubordinateComponent()->GetNodeRefSlope3(i-nnsub);
		// assume roll==0.: refslope3 computed as (sinYaw*cosPitch, sinYaw*sinPitch, cosYaw)
		if(mirror_plane_number==1)			{ ref.X() *= -1.; }				
		else if(mirror_plane_number==2) { ref.Y() *= -1.; }				
		else if(mirror_plane_number==3) { ref.Z() *= -1.;	}				
		else { return Vector3D(0.,0.,0.); }
		return ref;
	}
}
Vector3D MeshMirror::GetNodeRefAngles(int i)
{
	if(NSubs()==0) { return Vector3D(0,0,0); }
	int nnsub = GetSubordinateComponent()->NNodes();
	if(i<=nnsub) { return GetSubordinateComponent()->GetNodeRefAngles(i); }
	else
	{
		Vector3D ref = GetSubordinateComponent()->GetNodeRefAngles(i-nnsub);
// TODO: TEST this with a beam...
		// assume roll==0.
		if(mirror_plane_number==1)			{ ref.Y() *= -1.; ref.Z() *= -1.;}				// invert X() -> invert YAW and PITCH
		else if(mirror_plane_number==2) { ref.Z() *= -1.; }												// invert Y() -> invert PITCH -> invert angle.Z
		else if(mirror_plane_number==3) { ref.Y() *= -1.; }				                // invert Z() -> invert  YAW  -> invert angle.Y
		else { return Vector3D(0.,0.,0.); }
		return ref;
	}
}
int MeshMirror::GetElementObjectFactoryType(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementObjectFactoryType(i); }
	else { return elems(i-nesub)->GetTypeNr(); }
}
int MeshMirror::GetElementMaterialNumber(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementMaterialNumber(i); }
	else { return elems(i-nesub)->GetMatNr(); }
}
void MeshMirror::SetElementMeshNumber(int i, int idx)
{
	if(NSubs()==0) { return; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { GetSubordinateComponent()->SetElementMeshNumber(i,idx); }
	else { elemIdx(i-nesub) = idx; }
}
int MeshMirror::GetElementMeshNumber(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementMeshNumber(i); }
	else { return elemIdx(i-nesub); }
}
TArray<int> MeshMirror::GetElementNodeList(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementNodeList(i); }
	else { return elems(i-nesub)->GetNodeList(); }	
}
TArray<int> MeshMirror::GetElementFaceNodeList(int i,int f)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementFaceNodeList(i,f); }
	else { return elems(i-nesub)->GetFaceNodeList(f); }
}
int MeshMirror::ComputeElementSize(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->ComputeElementSize(i); }
	else { return GetSubordinateComponent()->ComputeElementSize(i-nesub); }
}
Vector3D MeshMirror::GetElementSizeVector(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElementSizeVector(i); }
	else { return GetSubordinateComponent()->GetElementSizeVector(i-nesub); }
}
int MeshMirror::GetElemAlternateFlag(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElemAlternateFlag(i); }
	else { return GetSubordinateComponent()->GetElemAlternateFlag(i-nesub); }
}
mystr MeshMirror::GetElemType(int i)
{
	if(NSubs()==0) { return 0; }
	int nesub = GetSubordinateComponent()->NElems();
	if(i<=nesub) { return GetSubordinateComponent()->GetElemType(i); }
	else { return GetSubordinateComponent()->GetElemType(i-nesub); }
}
int MeshMirror::NCoincident()
{
	if(NSubs()==0) { return 0; }
	return 2*GetSubordinateComponent()->NCoincident();
}
int2 MeshMirror::GetCoincidentNodeNumbers(int i)
{
	if(NSubs()==0) { return 0; }
	int ncsub = GetSubordinateComponent()->NCoincident();
	if(i<=ncsub) { return GetSubordinateComponent()->GetCoincidentNodeNumbers(i); }
	else
	{
		int2 pair = GetSubordinateComponent()->GetCoincidentNodeNumbers(i-ncsub);
		for(int j=GetSubordinateComponent()->NNodes()+1; j<=NNodes(); j++)
		{
			if(pair(1)==nodemapping(j)) { pair(1)=j; break; }
		}
		for(int j=GetSubordinateComponent()->NNodes()+1; j<=NNodes(); j++)
		{
			if(pair(2)==nodemapping(j)) { pair(2)=j; break; }
		}
		return pair;
	}
}
int MeshMirror::GetCoincidentPenaltyFlag(int i)
{
	if(NSubs()==0) { return 0; }
	int ncsub = GetSubordinateComponent()->NCoincident();
	if(i<=ncsub) { return GetSubordinateComponent()->GetCoincidentPenaltyFlag(i); }
	else { return GetCoincidentPenaltyFlag(i-ncsub); }
}
void MeshMirror::SetAutoConstraintNumber(int i, int cnr)
{
	if(subordinate!=NULL)
	{
		GetSubordinateComponent()->SetAutoConstraintNumber(i, cnr);
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshProcessor
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshTransformation
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshTransformation::MeshYourself()
{ 
	if(subordinate!=NULL) 
	{
		int rv = GetSubordinateComponent()->MeshYourself(); 
		ComputeBoundingBox();

// processing steps:
		// no actual manipulation of the mesh
		// positions and orientations of the nodes are computed on access
		InitNodeIdx();
		InitElemIdx();
		return 1;
	}
	return 0;
}
int MeshTransformation::ScaleXYZ(Vector3D& scale)
{
	Matrix3D scalemat;
	scalemat.Set44( scale.X(),        0.,        0., 0., 
		                     0., scale.Y(),        0., 0.,
									       0.,        0., scale.Z(), 0.,
									       0.,        0.,        0., 1.);

	transformation4D = scalemat * transformation4D;
	return 1;
}
int MeshTransformation::RotateXYZ(Vector3D& rotate)
{
	if(abs(rotate.X())>1e-8)
	{
		Matrix3D rotX;
		double x=rotate.X();
		rotX.Set44( 1.,      0.,      0., 0.,
			          0.,  cos(x), -sin(x), 0.,
								0.,  sin(x),  cos(x), 0.,
								0.,      0.,      0., 1.);
		transformation4D = rotX * transformation4D;
	}
	if(abs(rotate.Y())>1e-8)
	{
		Matrix3D rotY;
		double y=rotate.Y();
		rotY.Set44(  cos(y), 0.,  sin(y), 0.,
			               0., 1.,      0., 0.,
								-sin(y), 0.,  cos(y), 0.,
								     0., 0.,      0., 1.);
		transformation4D = rotY * transformation4D;
	}
	if(abs(rotate.Z())>1e-8)
	{
		Matrix3D rotZ;
		double z=rotate.Z();
		rotZ.Set44(  cos(z), -sin(z), 0., 0.,
			           sin(z),  cos(z), 0., 0.,
								     0.,      0., 1., 0.,
								     0.,      0., 0., 1.);
		transformation4D = rotZ * transformation4D;
	}
	return 1;
}
int MeshTransformation::TranslateXYZ(Vector3D& translate)
{
	Matrix3D translation;
	translation.Set44( 1., 0., 0., translate.X(),
		                 0., 1., 0., translate.Y(),
									   0., 0., 1., translate.Z(),
									   0., 0., 0.,            1.);
	transformation4D = translation * transformation4D ;
	return 1;
}
void MeshTransformation::TransformPositionVector(Vector3D& position)
{
		Vector3D transformedPosition;
		transformedPosition.X() = transformation4D(1,1)*position(1) + transformation4D(1,2)*position(2) + transformation4D(1,3)*position(3) + transformation4D(1,4);
		transformedPosition.Y() = transformation4D(2,1)*position(1) + transformation4D(2,2)*position(2) + transformation4D(2,3)*position(3) + transformation4D(2,4);
		transformedPosition.Z() = transformation4D(3,1)*position(1) + transformation4D(3,2)*position(2) + transformation4D(3,3)*position(3) + transformation4D(3,4);

		position = transformedPosition;
}
void MeshTransformation::TransformOrientationVector(Vector3D& orientation)
{
	Vector3D transformedOrientation;
	transformedOrientation.X() = transformation4D(1,1)*orientation(1) + transformation4D(1,2)*orientation(2) + transformation4D(1,3)*orientation(3);
	transformedOrientation.Y() = transformation4D(2,1)*orientation(1) + transformation4D(2,2)*orientation(2) + transformation4D(2,3)*orientation(3);
	transformedOrientation.Z() = transformation4D(3,1)*orientation(1) + transformation4D(3,2)*orientation(2) + transformation4D(3,3)*orientation(3);

	orientation = transformedOrientation;
}
int MeshTransformation::ComputeBoundingBox()
{
	Box3D transformed;
	for(int i=0; i<=7; i++)
	{
		Vector3D pos;
		if(i&1) { pos.X() = boundingbox.PMin().X(); } else { pos.X() = boundingbox.PMax().X(); }
		if(i&2) { pos.Y() = boundingbox.PMin().Y(); } else { pos.Y() = boundingbox.PMax().Y(); }
		if(i&4) { pos.Z() = boundingbox.PMin().Z(); } else { pos.Z() = boundingbox.PMax().Z(); }

		TransformPositionVector(pos);
		transformed.Add(pos);
	}
	boundingbox = transformed;
	return 8;
}

Vector3D MeshTransformation::GetNodePosition(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D position = GetSubordinateComponent()->GetNodePosition(i);
	TransformPositionVector(position);
	return position;
}
Vector3D MeshTransformation::GetNodeRefSlope1(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D orientation = GetSubordinateComponent()->GetNodeRefSlope1(i);
	TransformOrientationVector(orientation);
	return orientation;
}
Vector3D MeshTransformation::GetNodeRefSlope2(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D orientation = GetSubordinateComponent()->GetNodeRefSlope2(i);
	TransformOrientationVector(orientation);
	return orientation;
}
Vector3D MeshTransformation::GetNodeRefSlope3(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D orientation = GetSubordinateComponent()->GetNodeRefSlope3(i);
	TransformOrientationVector(orientation);
	return orientation;
}
Vector3D MeshTransformation::GetNodeRefAngles(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D orientation = GetSubordinateComponent()->GetNodeRefAngles(i);
	TransformOrientationVector(orientation);
	return orientation;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshDistortion
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MeshDistortion::Reset()
{
	MeshProcessor::Reset();
	
	PXofXYZ.SetParsedFunction3D(pParser,"x","x,y,z");
	PYofXYZ.SetParsedFunction3D(pParser,"y","x,y,z");
	PZofXYZ.SetParsedFunction3D(pParser,"z","x,y,z");
	OXofXYZ.SetParsedFunction3D(pParser,"0","x,y,z");
	OYofXYZ.SetParsedFunction3D(pParser,"0","x,y,z");
	OZofXYZ.SetParsedFunction3D(pParser,"0","x,y,z");
}
int MeshDistortion::MeshYourself()
{ 
	if(subordinate!=NULL) 
	{
		int rv = GetSubordinateComponent()->MeshYourself(); 
		ComputeBoundingBox();

// processing steps:
		// no actual manipulation of the mesh
		// positions and orientations of the nodes are computed on access
		InitNodeIdx();
		InitElemIdx();
		return 1;
	}
	return 0;
}

void MeshDistortion::TransformPositionVector(Vector3D& position)
{
		Vector3D transformedPosition;
		
		transformedPosition.X() = PXofXYZ.Evaluate(position.X(),position.Y(),position.Z());
		transformedPosition.Y() = PYofXYZ.Evaluate(position.X(),position.Y(),position.Z());
		transformedPosition.Z() = PZofXYZ.Evaluate(position.X(),position.Y(),position.Z());

		position = transformedPosition;
}
void MeshDistortion::TransformOrientationVector(Vector3D& orientation) // in posiiton, out orientation
{
	Vector3D transformedOrientation;

	transformedOrientation.X() = OXofXYZ.Evaluate(orientation.X(),orientation.Y(),orientation.Z());
	transformedOrientation.Y() = OYofXYZ.Evaluate(orientation.X(),orientation.Y(),orientation.Z());
	transformedOrientation.Z() = OZofXYZ.Evaluate(orientation.X(),orientation.Y(),orientation.Z());

	double len = transformedOrientation.Norm();
	if(abs(len)>1e-4)
		orientation = transformedOrientation;  
}

int MeshDistortion::ComputeBoundingBox()
{
	Box3D transformed;
	for(int i=1; i<=NNodes(); i++)
	{
		Vector3D pos = GetNodePosition(i);
		transformed.Add(pos);
	}
	boundingbox = transformed;
	return NNodes();
}

Vector3D MeshDistortion::GetNodePosition(int i)  
{
	if(subordinate==NULL) { return Vector3D(0.,0.,0.); }
	Vector3D position = GetSubordinateComponent()->GetNodePosition(i);
	TransformPositionVector(position);
	return position;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshModifier
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshModifier::MeshYourself()
{
	applymask.Flush();
	if(subordinate!=NULL)
	{
		subordinate->MeshYourself();

// entity specific length for mask
		int failcounter=0;
		if(entity & WOE_elemtype)
		{
			ElementSet* elemset = (ElementSet*) filter;
			applymask.SetLen(subordinate->NElems()); 
			if(elemset==NULL || elemset->GetNrOfEntries()==0) { applymask.SetAll(1); } // no filter set -> apply to all subordinate
			else 
			{ 
				applymask.SetAll(0); 
				for(int i=1; i<=elemset->GetNrOfEntries(); i++)
				{
					if(elemset->GetElementNumber(i) <= applymask.Length())
					{
						applymask(elemset->GetElementNumber(i)) = 1; // mark this element as 'to be changed'
					}
					else { failcounter++; }
				}
			}
		}
		else
		if(entity & WOE_nodetype)
		{
			GlobalNodeSet* nodeset = (GlobalNodeSet*) filter;
			applymask.SetLen(subordinate->NNodes()); 
			if(nodeset==NULL || nodeset->GetNrOfEntries()==0) { applymask.SetAll(1); } // no filter set -> apply to all subordinate
			else 
			{ 
				applymask.SetAll(0);
				for(int i=1; i<=nodeset->GetNrOfEntries(); i++)
				{
					if(nodeset->GetGlobalNodeNumber(i) <= applymask.Length())
					{
						applymask(nodeset->GetGlobalNodeNumber(i)) = 1; // mark this element as 'to be changed'
					}
					else { failcounter++; }
				}
			}
		}
		else
		if(entity & WOE_materialnumber)
		{
			ElementSet* elemset = (ElementSet*) filter;
			applymask.SetLen(subordinate->NElems()); 
			if(elemset==NULL || elemset->GetNrOfEntries()==0) { applymask.SetAll(1); } // no filter set -> apply to all subordinate
			else 
			{ 
				applymask.SetAll(0);
				for(int i=1; i<=elemset->GetNrOfEntries(); i++)
				{
					if(elemset->GetElementNumber(i) <= applymask.Length())
					{
						applymask(elemset->GetElementNumber(i)) = 1; // mark this element as 'to be changed'
					}
					else { failcounter++; }
				}
			}
		}
		else
		if(entity & WOE_nonlinearity)
		{
			ElementSet* elemset = (ElementSet*) filter;
			applymask.SetLen(subordinate->NElems()); 
			if(elemset==NULL || elemset->GetNrOfEntries()==0) { applymask.SetAll(1); } // no filter set -> apply to all subordinate
			else 
			{ 
				applymask.SetAll(0);
				for(int i=1; i<=elemset->GetNrOfEntries(); i++)
				{
					if(elemset->GetElementNumber(i) <= applymask.Length())
					{
						applymask(elemset->GetElementNumber(i)) = 1; // mark this element as 'to be changed'
					}
					else { failcounter++; }
				}
			}
		}
		if(failcounter>0)
		{
			GetMesh()->GetMBS()->UO() << "WARNING MeshModify::filter set exceeds range in " << failcounter << " cases."; 
		}
	}
	return applymask.Length();
}

int MeshModifier::GetNodeObjectFactoryType(int i)
{
	if(subordinate!=NULL)
	{	
		int oldtype = subordinate->GetNodeObjectFactoryType(i);

		if(entity & WOE_nodetype)
		{
			//   no oldVal    matching oldVal       in filter set
			if( (oldInt<=0 || oldInt==oldtype) && (applymask(i)!=0) ) 
			{
				return newInt; // return the replacement node type
			}
			else return oldtype; // no change, the node is not included in the filter set
		}
		return oldtype; // no change, this component does not change the nodetype
	}
	return 0; // no subordinate for the modifier node
}
int MeshModifier::GetElementObjectFactoryType(int i)
{
	if(subordinate!=NULL)
	{
		int oldtype = subordinate->GetElementObjectFactoryType(i);
		if(entity & WOE_elemtype)
		{
			//   no oldVal    matching oldVal       in filter set
			if( (oldInt<=0 || oldInt==oldtype) && (applymask(i)!=0) ) 
			{
				return newInt; // return the replacement element type
			}
			else return oldtype; // no change, the element is not included in the filter set
		}
		return oldtype; // no change, this component does not change the elementtype
	}
	return 0; // no subordinate for the modifier node
}
int MeshModifier::GetElementMaterialNumber(int i)
{
	if(subordinate!=NULL)
	{
		int oldtype = subordinate->GetElementMaterialNumber(i);
		if(entity & WOE_materialnumber)
		{
			//   no oldVal    matching oldVal       in filter set
			if( (oldInt<=0 || oldInt==oldtype) && (applymask(i)!=0) ) 
			{
				return newInt; // return the replacement element material number
			}
			else return oldtype; // no change, the element is not included in the filter set
		}
		return oldtype; // no change, this component does not change the elementtype
	}
	return 0; // no subordinate for the modifier node
}
int MeshModifier::GetElementGeometricNonlinearityStatus(int i)
{
	if(subordinate!=NULL)
	{
		int oldtype = subordinate->GetElementGeometricNonlinearityStatus(i);
		if(entity & WOE_nonlinearity)
		{
			//   no oldVal    matching oldVal       in filter set
			if( (oldInt<=0 || oldInt==oldtype) && (applymask(i)!=0) ) 
			{
				return newInt; // return the replacement nonlinearity status
			}
			else return oldtype; // no change, the element is not included in the filter set
		}
		return oldtype; // no change, this component does not change the nonlinearity status
	}
	return 0; // no subordinate for the modifier node
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshComponentWriter
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshComponentWriter::MeshYourself()
{
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();
	sub->MeshYourself();	

	return WriteToFile();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshWriterNeutral3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshWriterNeutral3D::PermuteNodeList(TArray<int>& nodelist)
{
	// TODO: !!! complete this list !!! 
	// implementation for HOTINT node numbering to default ANSYS node numbering - taken from FEMesh::MapNodesAnsys2Hotint
	int dummy;
	if(nodelist.Length()==8 || nodelist.Length()==20)
	{
		dummy = nodelist(4); nodelist(4) = nodelist(3); nodelist(3)= dummy;
		dummy = nodelist(8); nodelist(8) = nodelist(7); nodelist(7)= dummy;
		if(nodelist.Length()==20)
		{
			dummy = nodelist(11); nodelist(11) = nodelist(13); nodelist(13) = nodelist(12); nodelist(12) = nodelist(15);
			nodelist(15) = nodelist(16); nodelist(16) = nodelist(14); nodelist(14) = nodelist(10); nodelist(10) = dummy;
			dummy = nodelist(20); nodelist(20) = nodelist(19); nodelist(19) = dummy;	
			return 13; // 4+9 = 13 nodes changed place
		}
		return 4; // 4 nodes changed place
	}
	if(nodelist.Length()==4 || nodelist.Length()==10)
	{
		if(nodelist.Length()==10)
		{
			dummy = nodelist(9); nodelist(9) = nodelist(10); nodelist(10) = nodelist(8); nodelist(8) = dummy; 
			return 3; 
		}
		return 0;
	}
	return -1;
}

int MeshWriterNeutral3D::WriteToFile()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();
	if(NSubs()==0) { return 0; }
	MeshComponent* sub = GetSubordinateComponent();

	// open target file

	int opt = ios::out;
	ofstream file;
	file.open(filename.c_str(), opt); //with read sharing allowed!

	if(!file.is_open())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file " << filename << "\n";
		return 0;
	}

	int npoints = sub->NNodes();
	file << npoints << "\t";
	mbs->UO() << "npoints=" << npoints << "\n";
	int nelems = sub->NElems();;
	file << nelems << "\n";
	mbs->UO() << "nelems=" << nelems << "\n";

// write the points N x {idx, px, py, pz}
	Vector3D p;
	for(int i=1; i<=npoints; i++) 
	{
		file << i << "\t";
		p=sub->GetNodePosition(i);
		file << p.X() << "\t";
		file << p.Y() << "\t";
		file << p.Z() << "\n";
	}

// write the elements N x {idx, type, M x {nodenum}}
	int dimension = 3;
	int elemnodes;
	for(int i=1; i<=nelems; i++)
	{
		file << i << "\t";
		dimension = sub->GetElemPtr(i)->Dim();
		elemnodes = sub->GetElemPtr(i)->NNodes();
		file << (100*dimension+elemnodes);
		TArray<int> pointlist = sub->GetElementNodeList(i);
		PermuteNodeList(pointlist);
		for(int j=1; j<=elemnodes; j++)
		{
			file << "\t" << pointlist(j);
		}
		file << "\n";
	}
	return 1;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshMultiProcessor
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MeshNode* MeshMultiProcessor::GetNodePtr(int i) // forward to the first occurrence of the node in a subordinate
{ 
	if(NSubs()>0)
	{
		int nci = NCofI(i);
		int ni = NofI(i);
		return GetSubordinateComponent(nci) -> GetNodePtr(ni); 
	//	return GetSubordinateComponent(NCofI(i)) -> GetNodePtr(NofI(i)); 
	}
	else return NULL; 
}		 

MeshElement* MeshMultiProcessor::GetElemPtr(int i) // forward to the element in a subordinate
{ 
	if(NSubs()>0)
	{
		int eci = ECofI(i);
		int ei = EofI(i);
		return GetSubordinateComponent(eci) -> GetElemPtr(ei); 		
	//	return GetSubordinateComponent(ECofI(i)) -> GetElemPtr(EofI(i)); 
	}
	else return NULL; 
}

void MeshMultiProcessor::SetNodeMeshNumber(int i,int idx)
{
	nodeIdx(i) = idx; // entry in MeshMultiProcessor 

	if(NSubs()>0)
	{
		// check all subordinate components IF the current nodes is present there - for redundant nodes there MAY be multiple entries 
		int first_component = NCofI(i);
		int first_component_node = NofI(i);
		GetSubordinateComponent(first_component) -> SetNodeMeshNumber(first_component_node, idx); 	// FIRST occurrence

		for(int nc=first_component+1; nc<NSubs(); nc++) // loop through the rest of the subordinates
		{
			int n = nodemapping_CN2I(nc).Find(i);
			if(n>0)
			{
				GetSubordinateComponent(nc) -> SetNodeMeshNumber(n, idx); // 
			}
		}
	}
}

void MeshMultiProcessor::SetElementMeshNumber(int i,int idx)
{
	elemIdx(i) = idx; // entry in MeshMultiProcessor

	if(NSubs()>0) 
	{
	// each elemenet is unique to a subordinate element, just forward to ONE subordinate component
		GetSubordinateComponent(ECofI(i)) -> SetElementMeshNumber(EofI(i), idx); 
	}
}

TArray<int> MeshMultiProcessor::GetElementNodeList(int elnr)
{
	TArray<int> nodelist;
// aquire the reference element nodelist 
	int cnSource = ECofI(elnr);
	int enSource = EofI(elnr);
	nodelist = GetSubordinateComponent(ECofI(elnr)) -> GetElementNodeList(EofI(elnr));

// adjust node numbers from subordinate components to node numbers of MeshGlue then to node index in current component
	for(int i=1; i<=nodelist.Length(); i++)
	{
		int nnSource = nodelist(i);
		int nnGlue = nodemapping_CN2I(cnSource)(nnSource);
		nodelist(i) = nnGlue;
	}
	return nodelist;
}

TArray<int> MeshMultiProcessor::GetElementFaceNodeList(int elnr, int fnr)
{
	TArray<int> nodelist;
// aquire the reference element nodelist 
	int cnSource = ECofI(elnr);
	int enSource = EofI(elnr);
	nodelist = GetSubordinateComponent(ECofI(elnr)) -> GetElementFaceNodeList(EofI(elnr),fnr);

// adjust node numbers from subordinate components to node numbers of MeshGlue then to node index in current component
	for(int i=1; i<=nodelist.Length(); i++)
	{
		int nnSource = nodelist(i);
		int nnGlue = nodemapping_CN2I(cnSource)(nnSource);
		nodelist(i) = nnGlue;
	}
	return nodelist;
}

void MeshMultiProcessor::MapNodeListComponentToMesh(TArray<int>& nodelist)
{
// NOTE: in MeshGlue::GetElementNodeList the numbering is revised to fit the order in the Glued component, index shift to 
	for(int i=1; i<=nodelist.Length(); i++)
	{
		nodelist(i) = nodeIdx(nodelist(i));
	}
	return;
}

int MeshMultiProcessor::NCoincident()
{
	int sum = this->coincident_pairs.Length();
	for(int i=1; i<=NSubs(); i++)
	{
		sum += GetSubordinateComponent(i)->NCoincident();
	}
	return sum;
}

int2 MeshMultiProcessor::GetCoincidentNodeNumbers(int cpnr)
{
	int2 pair2;
	if(cpnr<=this->coincident_pairs.Length())
	{ // coincident pair with node numbering from in THIS (== origin) MeshMultiProcessor
		int4 pair4 = coincident_pairs(cpnr);
		pair2 = int2( nodemapping_CN2I(pair4(1))(pair4(2)),
		              nodemapping_CN2I(pair4(3))(pair4(4)) );
		// map the node numbering to the numbering of the root mesh - this is done only ONCE
		pair2(1) = GetNodeMeshNumber(pair2(1));
		pair2(2) = GetNodeMeshNumber(pair2(2));
		return pair2;
	}
	else
	{ // coincident pair registered in a subordinate MeshMultiProcessor (fttb only GlueMesh)
		cpnr -= coincident_pairs.Length();
		int cnSource;
		for(cnSource=1; cnSource<=NSubs(); cnSource++)
		{
			int ncp = GetSubordinateComponent(cnSource)->NCoincident();
			if(cpnr>ncp)
			{
				cpnr -= ncp;
			}
			else
			{
				pair2 = GetSubordinateComponent(cnSource)->GetCoincidentNodeNumbers(cpnr);
				// found the pair in a subordinate mesh component, numbering alredy adjusted to root mesh by subordinate
				return pair2;
			}
		}
	}
	return pair2;
}

int MeshMultiProcessor::ComputeBoundingBox()
{
	boundingbox.Empty();
	for(int i=1; i<=NSubs(); i++)
	{
		GetSubordinateComponent(i)->ComputeBoundingBox();
		boundingbox.Add(GetSubordinateComponent(i)->GetBoundingBox());
	}
	return NSubs();
}

//$EK 2017-01-11: pass alternate-flag of subordinates, e.g. in case of MeshGlue
int MeshMultiProcessor::GetElemAlternateFlag(int i) // forward to the element in a subordinate
{ 
	if(NSubs()>0)
	{
		int eci = ECofI(i);
		int ei = EofI(i);
		return GetSubordinateComponent(eci) -> GetElemAlternateFlag(ei); 		
	//	return GetSubordinateComponent(ECofI(i)) -> GetElemPtr(EofI(i)); 
	}
	else return NULL; 
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class MeshGlue
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// special case: glue with a single subordinate Loader 
int MeshGlue::NSubsWLoaders()
{
	int count=0;
	for(int i=1; i<=subordinates.Length(); i++)
	{
		if(subordinates(i)->IsLoader()) { count+= subordinates(i)->NSubs(); }
		else { count++; }
	}
	return count;
}

MeshComponent* MeshGlue::GetSubordinateComponentWLoaders(int subnr)
{
	for(int i=1; i<=subordinates.Length(); i++)
	{
		if(subordinates(i)->IsLoader()) 
		{ 
			int subNSub = subordinates(i)->NSubs(); 
			if(subnr<=subNSub) { return subordinates(i)->GetSubordinateComponent(subnr); }
			else { subnr -= subNSub; }
		}
		else
		{
			if(subnr==1) { return subordinates(i); }
			else { subnr--; }
		}	
	}
	return NULL;
}

// find redundant nodes and coincident nodes
int MeshGlue::IdentifyNodePairs()
{
	redundant_pairs.Flush();
	coincident_pairs.Flush();

// exception: single subordinate of type Loader, 
//            when no manipulation of the Dimoain containers is done use the coincident buffer entries
	if(NSubs()==1 && GetSubordinateComponent(1)->IsLoader())
	{
		MeshComponentLoader* theLoader = (MeshComponentLoader*) GetSubordinateComponent(1);
		for(int i=1; i<= theLoader->NSubs(); i++)
		{
			MeshComponent* sub = theLoader->GetSubordinateComponent(i);
			if(!sub->IsSource()) 
			{ 
// the subordinate of the Loader component is NOT a source, entries in the coincident buffer can not be used savely
				GetMesh()->GetMBS()->UO(UO_LVL_warn) << "WARNING: Glue operation on Loader is SKIPPED, only allowed when no additional Processing on single Domains is done.\n";
				return 0; 
			} 
		}


		for(int i=1; i<= theLoader->NCoincidentBuffer(); i++)
		{
			int2 pair = theLoader->GetNodeNumbersFromCoincidentBuffer(i);
			int4 entry = int4(1,pair(1),1,pair(2));
			coincident_pairs.Add(entry);
		}
		return coincident_pairs.Length();
	}

// default case: multiple subordinates
//               only test for node pairs in separate subordinates

	for(int i=1; i<=NSubs(); i++) // loop component I
	{
		MeshComponent* compi = GetSubordinateComponent(i);
		int nni = compi->NNodes();
		for(int j=i+1; j<=NSubs(); j++) // loop comonent J
		{
			MeshComponent* compj = GetSubordinateComponent(j);

			int nnj = compj->NNodes();

			for(int ii=1; ii<=nni; ii++) // loop over nodes of component I
			{
				Vector3D posii = compi->GetNodePosition(ii);

				for(int jj=1; jj<=nnj; jj++) // loop over nodes of component J
				{
					Vector3D posjj = compj->GetNodePosition(jj);
					Vector3D dist = posjj - posii; // go through components to accumulate Transformations

					if(dist.Norm() <= 1e-10) 
					{ // same position

						if( compi->GetNodeObjectFactoryType(ii) == compj->GetNodeObjectFactoryType(jj) )
						{ // same nodetype

							int domii = compi->GetNodeBodyIndex(ii);
							int domjj = compj->GetNodeBodyIndex(jj);
			
							if(domii==domjj)
							{ // same domain

								compi->ComputeOrientation(ii);
								compj->ComputeOrientation(jj);
								int redundant = 1;
							
								Vector3D deltaRS1 = compj->GetNodeRefSlope1(jj) - compi->GetNodeRefSlope1(ii);
								if(deltaRS1.Norm() >= 1e-10)
								{
									redundant = 0;
								} 
								else 
								{
									Vector3D deltaRS2 = compj->GetNodeRefSlope2(jj) - compi->GetNodeRefSlope2(ii);
									if(deltaRS2.Norm() >= 1e-10)
									{
										redundant = 0;
									}
									else
									{
										Vector3D deltaRS3 = compj->GetNodeRefSlope3(jj) - compi->GetNodeRefSlope3(ii);
										if(deltaRS3.Norm() >= 1e-10)
										{
											redundant = 0;
										}								
										else
										{
											Vector3D deltaAng = compj->GetNodeRefAngles(jj) - compi->GetNodeRefAngles(ii);
											if(deltaAng.Norm() >= 1e-10)
											{
												redundant = 0;
											}
										}
									}
								}
								if(redundant)
								{
									redundant_pairs.Add(int4(i,ii,j,jj));  // same position, same nodetype, same domain, same orientation
								}
								else
								{
									coincident_pairs.Add(int4(i,ii,j,jj)); // same position, same nodetypr, same domain, different orientation
								}
							}// end if domii==domjj
							else
							{
								coincident_pairs.Add(int4(i,ii,j,jj)); // same position, same nodetype, different domain
							}
						}// end if nodetype == nodetype
						else
						{
							coincident_pairs.Add(int4(i,ii,j,jj)); // same position, different nodetype
						}
					}// end if dist <= 1e-10
					// different position...
				}
			}
		}
	}
	return redundant_pairs.Length()+coincident_pairs.Length();
}
int MeshGlue::GenerateNodeMapping()
{
	nodemapping_CN2I.Flush();
	nodemapping_CN2I.SetLen(NSubs());
	for(int i=1; i<=NSubs(); i++) 
	{
		nodemapping_CN2I(i) = TArray<int>(0);
	}
	nodemapping_I2CN.Flush();
	nodemapping_I2CN.SetLen(0);

	// loop through all subordinate components and all their nodes
	// check if the component/node pair is registered in redundant nodes
	// if so: find the (first) node the current node is redundant to and copy its nodeindex  
	for(int i=1; i<=NSubs(); i++)
	{
		MeshComponent* compi = GetSubordinateComponent(i);
		for(int ii=1; ii<=compi->NNodes(); ii++)
		{	
			for(int j=1; j<=redundant_pairs.Length(); j++)
			{
				if(redundant_pairs(j)(3)==i && redundant_pairs(j)(4)==ii)
				{
					int cnr = redundant_pairs(j)(1);
					int nnr = redundant_pairs(j)(2);
					int glueidx = nodemapping_CN2I(cnr)(nnr);
					// no new entry in nodemapping_I2CN
					nodemapping_CN2I(i).Add(glueidx);
					break; // for j
				}
			}
			if(nodemapping_CN2I(i).Length()<ii)
			{ // node has not been identified as redundant
				nodemapping_I2CN.Add(int2(i,ii));
				nodemapping_CN2I(i).Add(nodemapping_I2CN.Length()); 
			}
		}
	}

	// reduce the list of coincident nodes from tail to head
	for(int i=coincident_pairs.Length(); i>1; i--)
	{
		int nnr1_i = nodemapping_CN2I(coincident_pairs(i)(1))(coincident_pairs(i)(2));
		int nnr2_i = nodemapping_CN2I(coincident_pairs(i)(3))(coincident_pairs(i)(4)); 
	
		for(int j=i-1; j>0; j--) 
		{
			int nnr1_j = nodemapping_CN2I(coincident_pairs(j)(1))(coincident_pairs(j)(2));
			int nnr2_j = nodemapping_CN2I(coincident_pairs(j)(3))(coincident_pairs(j)(4)); 
		
			if( (nnr1_i==nnr1_j) && ( nnr2_i==nnr2_j) )
			{
				coincident_pairs.Erase(i); // found a redundant coincident_pair, erase the one with higher index
				break; // for j
			}
		}
	}
	return NNodes();
}

int MeshGlue::GenerateElementMapping()
{
	elementmapping_CE2I.Flush();
	elementmapping_CE2I.SetLen(NSubs());
	for(int i=1; i<=NSubs(); i++) 
	{
		elementmapping_CE2I(i) = TArray<int>(0);
	}
	elementmapping_I2CE.Flush();
	elementmapping_I2CE.SetLen(0);

	// straightforward sequential mapping
	for(int i=1; i<=NSubs(); i++)
	{
		MeshComponent* compi = GetSubordinateComponent(i);
		for(int ii=1; ii<=compi->NElems(); ii++)
		{
			elementmapping_I2CE.Add(int2(i,ii));
			elementmapping_CE2I(i).Add(elementmapping_I2CE.Length());
		}
	}
	return NElems();
}

void MeshGlue::SetAutoConstraintNumber(int i, int cnr)
{
// assume that the numbers displayed are always all constraints that are added in this glue mesh plus all from the subordinates
// for the time being only one array is used.     splitting up requires changes in MeshGlue lifecycle...
	if(i<=coincident_pairs.Length())
	{
		//OWN
	autoconstraintnrs(i) = cnr; 
	}
	else
	{
		// always add nontheless
	autoconstraintnrs(i) = cnr; 
		// from a subordinate
		i -= coincident_pairs.Length();
		int cnSource;
		for(cnSource=1; cnSource<=NSubs(); cnSource++)
		{
			int ncp = GetSubordinateComponent(cnSource)->NCoincident();
			if(i>ncp)
			{
				i -= ncp;
			}
			else
			{
				GetSubordinateComponent(cnSource)->SetAutoConstraintNumber(i,cnr);
			}
		}	
	}
}

int MeshGlue::MeshYourself()
{
	for(int i=1; i<=NSubs(); i++)
	{
		MeshComponent* subi = GetSubordinateComponent(i);
		subi->MeshYourself();
	}
	ComputeBoundingBox();

// processing steps:
	IdentifyNodePairs(); 
	GenerateNodeMapping();
	GenerateElementMapping();
	InitNodeIdx();
	InitElemIdx();
	return 1;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshComponentLoader
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshComponentLoader::MeshYourself()
{
	LoadFromFile();
	ComputeBoundingBox();
	
	InitNodeIdx();
	InitElemIdx();
	return 1;
}

int MeshComponentLoader::RemoveUnusedNodes()
{
// in all elements, register all the nodes that are actually used, 
// remove all nodes that are not used in any element
	pnesource = int2(npoints,nelements);

	TArray<int> nodemapping;  // also node useage counter
	nodemapping.SetLen(npoints);
	nodemapping.SetAll(0);

// count how often nodes are used
	for(int i=1; i<=elementbuffer.Length(); i++)
	{
		TArray<int>& curelemnodelist = elementbuffer(i)->NodeList();
		for(int j=1; j<=curelemnodelist.Length(); j++)
		{
			nodemapping( curelemnodelist(j) )++;
		}
	}

	int unused = 0;
// determine nodenumber in revised list
	for(int i=1; i<=nodemapping.Length(); i++)
	{
		if(nodemapping(i) != 0) 
		{
			nodemapping(i) = i-unused; 
		}
		else 
		{ 
			unused++; 
		}
	}

// update nodenumbers in all elements
	for(int i=1; i<=elementbuffer.Length(); i++)
	{
		TArray<int>& curelemnodelist = elementbuffer(i)->NodeList();
		for(int j=1; j<=curelemnodelist.Length(); j++)
		{
			curelemnodelist(j) = nodemapping( curelemnodelist(j) );
		}
	}

// update nodenumbers in all boundary conditions
	for(int i=1; i<=bc_s.Length(); i++)
	{
		TArray<int>& bc = bc_s(i);
		for(int j=2; j<=bc.Length(); j++) // node numbers start at index 2, first entry is the BC identifier
		{
			bc(j) = nodemapping( bc(j) );
		}
	}

// update nodenumbers in all Faces/Edges
	for(int i=1; i<surfacesbuffer.Length(); i++)
	{
		FaceInfo& face = surfacesbuffer(i);
		for(int j=1; j<=face.globalNodeNrs.Length(); j++)
		{
			face.globalNodeNrs(j) = nodemapping( face.globalNodeNrs(j) );
		}
	}

// update nodenumbers in all identifications
	for(int i=1; i<=idents.Length(); i++)
	{
		int2& ident = idents(i);
		ident(1) = nodemapping(ident(1));
		ident(2) = nodemapping(ident(2));
	}

// remove all unused nodes from pointbuffer
	for(int i=nodemapping.Length(); 0<i; i--)
	{
		if (nodemapping(i)==0)
		{
			pointbuffer.Erase(i);
		}
	}

	npoints = pointbuffer.Length();
	return unused;
}

void MeshComponentLoader::CopyFrom(const MeshComponentLoader& other)
{
	MeshMultiProcessor::CopyFrom(other);

	const MeshComponentLoader& o = (const MeshComponentLoader&) other;
// specific for derived
	filename = other.filename;
	usedtypelist = other.usedtypelist;
	nodetypelist.CopyFrom(other.nodetypelist);
	elementtypelist.CopyFrom(other.elementtypelist);
	
	npoints = other.npoints; 
	nelements = other.nelements;
	pnesource = other.pnesource;
	pointbuffer.CopyFrom(other.pointbuffer);
	elementbuffer.CopyFrom(other.elementbuffer);
	useddomainsbuffer.CopyFrom(other.useddomainsbuffer);

	nborder = other.nborder;
	identifications.CopyFrom(other.identifications);
	surfacesbuffer.CopyFrom(other.surfacesbuffer);
	bc_s.CopyFrom(other.bc_s);
	idents.CopyFrom(other.idents);

	nodedomainmultiplicity.CopyFrom(other.nodedomainmultiplicity);
	nodenumberindomain.CopyFrom(other.nodenumberindomain);
	nodedomain.CopyFrom(other.nodedomain);
	coincidentbuffer.CopyFrom(other.coincidentbuffer);
	collectedBCs = other.collectedBCs;
	bcmapping = other.bcmapping;
}

void MeshComponentLoader::Reset()
{
	MeshMultiProcessor::Reset();

	filename = mystr("");
	usedtypelist = 1;
	nodetypelist.Flush();
	elementtypelist.Flush();

	npoints = 0; 
	nelements = 0;
	pnesource = int2(0,0);
	pointbuffer.Flush();
	elementbuffer.Flush();
	useddomainsbuffer.Flush();

	nborder = 0;
	identifications.Flush();
	surfacesbuffer.Flush();
	bc_s.Flush();
	idents.Flush();

	nodedomainmultiplicity.Flush();
	nodenumberindomain.Flush();
	nodedomain.Flush();
	coincidentbuffer.Flush();
	collectedBCs.clear();
	bcmapping.clear();
}

void MeshComponentLoader::GetElementData(ElementDataContainer& edc)
{
	MeshMultiProcessor::GetElementData(edc);

	ElementData ed;
	ed.SetText(filename.c_str(),"source_file"); ed.SetToolTipText("external source file for the mesh"); edc.TreeAdd("Source",ed);
	ed.SetInt(npoints,"source_nodes"); ed.SetToolTipText("number of nodes read from external source file"); ed.SetLocked(1); edc.TreeAdd("Source",ed);
	ed.SetInt(nelements,"source_elems"); ed.SetToolTipText("number of elements read from external source file"); ed.SetLocked(1); edc.TreeAdd("Source",ed);
	ed.SetInt(NNodes(),"generated_nodes"); ed.SetToolTipText("number of nodes after domain splitting"); ed.SetLocked(1); edc.TreeAdd("Source",ed);

// interfaces... compute from coincident nodes
	for(int i=1; i<NSubs(); i++)
	{
		for(int j=i+1; j<=NSubs(); j++)
		{
			TArray<int2> pairs2;
			for(int k=1; k<=coincidentbuffer.Length(); k++)
			{	
				int4 pair4 = coincidentbuffer(k);	
				if( (pair4(1)==i && pair4(3)==j) ||	(pair4(1)==j && pair4(3)==i) )
				{
					pairs2.Add( int2(nodemapping_CN2I(pair4(1))(pair4(2)), nodemapping_CN2I(pair4(3))(pair4(4))) );
				}
			}
			if(pairs2.Length()>0)
			{
				Vector vv(pairs2.Length()); 
				for(int ii=1; ii<=pairs2.Length(); ii++) { vv(ii) = GetNodeMeshNumber(pairs2(ii)(1)); } 
				ed.SetVector(vv.GetVecPtr(), vv.Length(), "global_nodes_1"); 		
				ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("nodes on side 1 of interface"); 
				edc.TreeAdd( mystr("Interfaces.Interface_")+mystr(i)+mystr("_")+mystr(j),ed );		
				
				for(int ii=1; ii<=pairs2.Length(); ii++) { vv(ii) = GetNodeMeshNumber(pairs2(ii)(2)); } 
				ed.SetVector(vv.GetVecPtr(), vv.Length(), "global_nodes_2"); 		
				ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("nodes on side 2 of interface"); 
				edc.TreeAdd( mystr("Interfaces.Interface_")+mystr(i)+mystr("_")+mystr(j),ed );	
			}
		}
	}

// boundary conditions... 
	ed.SetInt(collectedBCs.size()+bc_s.Length(),"number_of_boundary_conditions"); ed.SetToolTipText("number of boundary conditions");
	//ed.SetInt(bc_s.Length(),"number_of_boundary_conditions"); ed.SetToolTipText("number of boundary conditions");
	edc.TreeAdd( mystr("BoundaryConditions"),ed);

	int useddomainnumbers = nodenumberindomain(1).Length();
	TArray<int> counter; counter.SetLen(useddomainnumbers); 

// display BCMapping as matrix #bcNr #domIn #domOut #setNrIn #setNrOut
	Matrix bcm(bcmapping.size(),5);
	bcm.SetAll(0);
	int row=0;
	for(std::map<int,BoundaryMapping>::iterator curBC = bcmapping.begin(); curBC!=bcmapping.end(); ++curBC)
	{
		row++;
		bcm(row,1) = curBC->second.bcTypeNr;		
		bcm(row,2) = curBC->second.domin;
		bcm(row,3) = curBC->second.domout;
		bcm(row,4) = curBC->second.setNrin;
		bcm(row,5) = curBC->second.setNrout;
	}
	ed.SetMatrix(bcm.GetMatPtr(),bcmapping.size(),5,"bcmapping"); ed.SetValuesInt(); ed.SetVariableLength(); 
	ed.SetLocked(1); ed.SetToolTipText("mapping of loaded bc to internal set numbers (bcNr, dom1, dom2, setNr1, setNr2)");
	edc.TreeAdd( mystr("BoundaryConditions"),ed);

	for (int i=1; i<=bc_s.Length(); i++)
	{
		const TArray<int>& nodes_bc = bc_s(i); 
		int nNodes_bc = nodes_bc.Length()-1; // first entry of this array is the assigned (~domain)number from Netgen
		
// step 1: identify the domains for the boundary condition 
//         loop over all nodes in the bc and look for those domains where ALL of the nodes have an entry in mapping "nodenumberindomain"
// TODO: TODO:
//         ATTENTION: exceptions for single edge bcs must be treated (both nodes could be in three domains if two domains are embedded in a large one 
//         if(nNodes_bc == 2) ...
		Vector bc_original_nodes(nNodes_bc); // node numbers as loaded from the file
		counter.SetAll(0);
		for(int j=1; j<=nNodes_bc; j++)
		{	
			int node_number = nodes_bc(j+1);
			bc_original_nodes(j) = node_number;
			TArray<int> & nodedomains = nodenumberindomain(node_number);
			for(int k=1; k<=useddomainnumbers; k++)
			{
				if(nodedomains(k)!=0)
				{
					counter(k)++;
				}
			}
		}
		int domainnr1 = 0;
		int domainnr2 = 0;
		for(int k=1; k<=useddomainnumbers; k++)
		{
			if(counter(k)==nNodes_bc)
			{
				if(domainnr1==0) { domainnr1 = k; }
				else if(domainnr2==0) { domainnr2 = k; }
				else { /* fail - single edge border between two small domains surrounded ba large domain */ }
			}
		}
// step 2: make lists with the nodenumbers after split 
//         loop over all nodes in the bc and lookin the nodenumberindomain mapping for the actual nodenumber for the given domain
		Vector bc_global_nodes_1(nNodes_bc); // node numbers in first domain (after splitting)
		Vector bc_global_nodes_2(nNodes_bc); // node numbers in second domain (after splitting)
		for(int j=1; j<=nNodes_bc; j++)
		{	
			int node_number = nodes_bc(j+1);
			TArray<int> & nodedomains = nodenumberindomain(node_number);
			if(domainnr1>0) { bc_global_nodes_1(j) = nodedomains(domainnr1); } else { bc_global_nodes_1(j) = 0; }
			if(domainnr2>0) { bc_global_nodes_2(j) = nodedomains(domainnr2); } else { bc_global_nodes_2(j) = 0; }
		}
// step 3: write to EDC
		ed.SetVector(bc_original_nodes.GetVecPtr(), bc_original_nodes.Length(), "original nodes"); 		
		ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("nodes of boundary condition as in source file"); 
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);
		ed.SetInt(nNodes_bc,"number_of_nodes"); ed.SetToolTipText("number of nodes on each side");
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);
		ed.SetInt(nodes_bc(1),"type"); ed.SetToolTipText("boundary condition type");
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);

		ed.SetInt(domainnr1,"Domain1"); ed.SetToolTipText("domain number of side 1");
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);
		ed.SetVector(bc_global_nodes_1.GetVecPtr(), bc_global_nodes_1.Length(), "global_nodes_1"); 		
		ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("nodes on side 1 of boundary condition"); 
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);

		ed.SetInt(domainnr2,"Domain2"); ed.SetToolTipText("domain number of side 2");
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);
		ed.SetVector(bc_global_nodes_2.GetVecPtr(), bc_global_nodes_2.Length(), "global_nodes_2"); 		
		ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("nodes on side 2 of boundary condition"); 
		edc.TreeAdd( mystr("BoundaryConditions.BC_")+mystr(i),ed);		

	}

	// idents
	int nIdents = idents.Length();
	Vector idents_global_nodes_1(nIdents); // node numbers in first domain (after splitting)
	Vector idents_global_nodes_2(nIdents); // node numbers in second domain (after splitting)
	for(int j=1; j<=nIdents; j++)
	{	
		idents_global_nodes_1(j)=GetNodeMeshNumber(idents(j)(1));
		idents_global_nodes_2(j)=GetNodeMeshNumber(idents(j)(2));
	}
	ed.SetVector(idents_global_nodes_1.GetVecPtr(), idents_global_nodes_1.Length(), "nodes_1"); 		
	ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("identical nodes side 1 (periodic boundary conditions)"); 
	edc.TreeAdd( mystr("PeriodicBoundaryConditions"),ed);

	ed.SetVector(idents_global_nodes_2.GetVecPtr(), idents_global_nodes_2.Length(), "nodes_2"); 		
	ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("identical nodes side 2 (periodic boundary conditions)"); 
	edc.TreeAdd( mystr("PeriodicBoundaryConditions"),ed);
}





int MeshComponentLoader::GenerateDomainContainers()
{
// generate the subordinate components - domaincontainers

// count the required number of containers, sort in ascending domain order
	for(int i=1; i<=useddomainsbuffer.Length(); i++)
	{
		for(int j=i+1; j<=useddomainsbuffer.Length(); j++)
		{
			if(useddomainsbuffer(j)<useddomainsbuffer(i))
			{
				int tmp = useddomainsbuffer(j);
				useddomainsbuffer(j) = useddomainsbuffer(i);
				useddomainsbuffer(i) = tmp;
			}
		}
	}
		
	int nsubs= NSubs();  // check for already present processors on domains
	for(int i=1; i<=useddomainsbuffer.Length(); i++)
	{
		MeshDomainContainer* domcon = new MeshDomainContainer();
		mystr conname = mystr("Domain") + mystr(useddomainsbuffer(i));
		domcon->SetName(conname);
		domcon->SetMesh(this->mesh);
		domcon->SetMaterialNumber(useddomainsbuffer(i));
		domcon->SetBodyIndex(useddomainsbuffer(i));
		if (nsubs>=i)
		{
			if (GetSubordinateComponent(i) != NULL)
			{
				GetSubordinateComponent(i)->SetSubordinate(domcon,1);
			}
			else
			{
				SetSubordinate(domcon,i);
			}
		}
		else
		{
			AddSubordinate(domcon);
		}
	}
 
// initialize the mappings main to subordinates
	nodemapping_CN2I.Flush();
	nodemapping_CN2I.SetLen(NSubs());
	for(int i=1; i<=NSubs(); i++) {	nodemapping_CN2I(i) = TArray<int>(0); }
	nodemapping_I2CN.Flush();
	nodemapping_I2CN.SetLen(npoints); // miminal length is known
	elementmapping_CE2I.Flush();
	elementmapping_CE2I.SetLen(NSubs());
	for(int i=1; i<=NSubs(); i++) {	elementmapping_CE2I(i) = TArray<int>(0); }
	elementmapping_I2CE.Flush();
	elementmapping_I2CE.SetLen(nelements); // actual length is known

	return NSubs();
}

int MeshComponentLoader::SplitIntoDomains()
{
// note: order of nodes and elements does not really matter but it is easier to find errors when the numbering is consistend with the source file ( given additional nodes generated... )

// identify nodes that must be generated multiple times ( because they are used in elements of more than one domain )			

// initialize buffers with entries for all known nodes ( from file ) such that any 
// additional nodes generated by the splitting are added with indices higher than the original nodes
	
	nodedomainmultiplicity.SetLen(npoints);				// fixed size - keep track of in how many domain the node is present
	nodedomainmultiplicity.SetAll(0);	
	nodenumberindomain.SetLen(npoints);					// fixed size - keep track of the listnumber of the original node in all domains
	for(int j=1; j<=npoints; j++)
	{
		nodenumberindomain(j).SetLen(useddomainsbuffer.Length());
		nodenumberindomain(j).SetAll(0);
	} 
	nodedomain.SetLen(npoints);                         // can grow in size - is appended when a node is duplicated or a new domain
	nodedomain.SetAll(0);
	coincidentbuffer.Flush();

	nelements = elementbuffer.Length();
	for(int i=1; i<=nelements; i++)
	{
		MeshElement* curelem = elementbuffer(i);
		int domainnr = curelem->GetMatNr();
		int domainidx = useddomainsbuffer.Find(domainnr);

	// on the fly GenerateElementMapping()
		elementmapping_CE2I(domainidx).Add(i);
		elementmapping_I2CE(i) = int2(domainidx,elementmapping_CE2I(domainidx).Length());

		TArray<int>& curelemnodelist = curelem->GetNodeList();
		for(int j=1; j<=curelemnodelist.Length(); j++)
		{
			int nnrsource = curelemnodelist(j);						    // source file node number 
			int nnrunique = nodenumberindomain(nnrsource)(domainidx);	// unique node number associated with the combination of source nodenumber and element domain
	
			if(nnrunique==0) 
			{
			// first time the combination of original node number and domain occurs 
				if(nodedomainmultiplicity(nnrsource)==0)
				{		
				// first time the original node number is used
				// assign the primary domainindex to the node
					nodedomain(nnrsource) = domainidx;

				// register node/domain combination as 'in use' in the loader 
					nodenumberindomain(nnrsource)(domainidx) = nnrsource; 
					nodedomainmultiplicity(nnrsource) = 1;
					nnrunique = nodenumberindomain(nnrsource)(domainidx);

				// on the fly GenerateNodeMapping()
					nodemapping_CN2I(domainidx).Add(nnrunique);
					nodemapping_I2CN(nnrsource) = int2(domainidx,nodemapping_CN2I(domainidx).Length());
				}
				else
				{
				// first time used for a new domain:
				// additional node must be generated in new domain
					nodedomain.Add(domainidx);
					Vector3D pos = pointbuffer(nnrsource);
					pointbuffer.Add(pos);

				// register node/domain combination as 'in use'
					nodenumberindomain(nnrsource)(domainidx) = nodedomain.Length(); // == nnrunique
					nodedomainmultiplicity(nnrsource) += 1;
					nnrunique = nodenumberindomain(nnrsource)(domainidx);

				// on the fly GenerateNodeMapping()
					nodemapping_CN2I(domainidx).Add(nnrunique);
					nodemapping_I2CN.Add(int2(domainidx,nodemapping_CN2I(domainidx).Length()));
					                           // domain 1st/2nd               domain-nodenr 1st/2nd
					coincidentbuffer.Add(int4( nodemapping_I2CN(nnrsource)(1), nodemapping_I2CN(nnrsource)(2), 
					                           nodemapping_I2CN.Last()(1),     nodemapping_I2CN.Last()(2) ));

				// alter entry in element nodelist to unique node number ( Loader node number, not yet container node number )
					curelem->SetNode(j,nnrunique);
				}
			}
			else
			{
			// node is already in use, either in primaty domain or as duplicate in other domain
				if(nnrsource != nnrunique)
				{
				// alter entry in element nodelist to unique node number ( Loader node number, not yet container node number )
					curelem->SetNode(j,nnrunique);
				}		
			}
		}
	}

	// update (normal) boundary conditions:
	for(int i=1; i<=bc_s.Length(); i++)
	{
		TArray<int>& curbc = bc_s(i);
		for(int j=1; j<=curbc.Length(); j++) 
		{
			continue; // this CANNOT work - no domain info directly in bc_s - replace with 2D Facees a.s.a.p.
		}
	}
	// update identified faces
	for(int i=1; i<=surfacesbuffer.Length(); i++)
	{
		FaceInfo& face = surfacesbuffer(i);
		for(int j=1; j<=face.globalNodeNrs.Length(); j++)
		{
			int nnrsource = face.globalNodeNrs(j);
			int domainidx = elementbuffer(face.elNr)->GetMatNr();
			int nnrunique = nodenumberindomain(nnrsource)(domainidx);	// unique node number associated
			face.globalNodeNrs(j) = nnrunique;
		}
	}

	// update periodic boundary conditions:
	// add boundary conditions for updated elements
	int nIdents = idents.Length();
	for (int i = 1; i <= nIdents; i++)
	{
		int& n1 = idents(i)(1);
		int& n2 = idents(i)(2);

		assert(nodedomainmultiplicity(n1) == nodedomainmultiplicity(n2));

		if (nodedomainmultiplicity(n1) > 1)
		{
			const TArray<int>& domainNrs1 =  nodenumberindomain(n1);
			const TArray<int>& domainNrs2 =  nodenumberindomain(n2);
			TArray<bool> usedDomains2;
			usedDomains2.SetLen(domainNrs1.Length());
			usedDomains2.SetAll(false);
			for (int k = 1; k<= domainNrs1.Length(); k++)
			{
				if (domainNrs1(k) != 0)
				{
					if (domainNrs1(k) == n1)
					{
						if (domainNrs2(k) != 0)
						{
							usedDomains2(k) = true;
							n2 = domainNrs2(k);
						}
						else
						{
							//search the first unused domain, but do not use domains which have both
							for (int m = 1; m <= domainNrs1.Length(); m++)
							{
								if (domainNrs2(m) != 0 && !usedDomains2(m) && domainNrs1(m) == 0)
								{
									usedDomains2(m) = true;
									n2 = domainNrs2(m);
								}
							}
						}
					}
					else// (domainNrs1(k) != n1) // generate a new entry
					{
						if (domainNrs2(k) != 0)
						{
							usedDomains2(k) = true;
							idents.AddIfNotExists(int2(domainNrs1(k),domainNrs2(k)));
						}
						else
						{
							//search the first unused domain, but do not use domains which have both
							for (int m = 1; m <= domainNrs1.Length(); m++)
							{
								if (domainNrs2(m) != 0 && !usedDomains2(m) && domainNrs1(m) == 0)
								{
									usedDomains2(m) = true;
									idents.AddIfNotExists(int2(domainNrs1(k), domainNrs2(m)));
								}
							}
						}
					}
				}
			}
		}
	}
	return NSubs();
}

int MeshComponentLoader::SplitBoundaries()
{
// process the TArray<FaceInfo> that holds all surface elements:
// split by "bcTypeNr" but also by element domain in case of an interface
// ASSUMPTIONS:  * all surface elements are read from file in order	
//               * interfaces generate an additional entry at the tail of the list
//               * order for both sides of the interface is thus the same
// container:
// std::vector<                                   > keeps original BC Numbers
//             std::map<                         >  one or two (interface) entires - key holds domain number
//                      int,std::vector<FaceInfo>   all FaceInfo with same original type and side of interface
	
	for(int i=1,n=surfacesbuffer.Length(); i<=n; i++)
	{
		FaceInfo& face = surfacesbuffer(i);
		
		int bcNr = face.bcTypeNr;
		int domNr = elementbuffer(face.elNr)->GetMatNr();

		std::map<int,std::vector<FaceInfo>>& curBC = collectedBCs[bcNr];
		std::vector<FaceInfo>& curBCDom = curBC[domNr];
		curBCDom.push_back(face);

	}
	return NSubs();
}

int MeshComponentLoader::GenerateOwnNodes()
{
// generate nodes and elements in subordinate domain containers (use nodemapping for correct sequence)
	MeshNode n;
	n.SetTypeNr( GetNodeTypeLookup() );
	for(int i=1; i<=NSubs(); i++)
	{
		MeshComponent* sub = GetSubordinateComponent(i);
		while (sub->NSubs() > 0)
		{			
			sub = sub->GetSubordinateComponent(1);
		} 
		MeshDomainContainer* domaincontainer = (MeshDomainContainer*) sub;
		// all the nodes
		for(int j=1; j<=nodemapping_CN2I(i).Length(); j++)
		{
			int nnrunique = nodemapping_CN2I(i)(j);
			n.SetBodyIdx(i);
			n.SetNodePos(pointbuffer(nnrunique));
			domaincontainer->AddNode(n);
		}
	}
	return NNodes();
}

int MeshComponentLoader::GenerateOwnElements()
{
	for(int i=1; i<=NSubs(); i++)
	{
		MeshComponent* sub = sub = GetSubordinateComponent(i);
		while (sub->NSubs() > 0)
		{
			sub = sub->GetSubordinateComponent(1);
		} 

		MeshDomainContainer* domaincontainer = (MeshDomainContainer*) sub;
		// all the elements
		for(int j=1; j<=elementmapping_CE2I(i).Length(); j++)
		{
			int enrunique = elementmapping_CE2I(i)(j);
			MeshElement* curelem = elementbuffer(enrunique);
			for(int k=1; k<=curelem->NNodes(); k++)
			{
				int nnrunique = curelem->GetNode(k);
				int nnrdomain = nodemapping_I2CN(nnrunique)(2);
				curelem->SetNode(k,nnrdomain);
			}
			domaincontainer->AddElem(*curelem);
		}
	}
	return NElems();
}

int MeshComponentLoader::GenerateOwnSets()
{
	BaseMesh* mesh = GetMesh();
	MBS* mbs = (MBS*) GetMesh()->GetMBS();
	
// add all buffered sets to the mesh 
	for(std::map< int, std::map< int, std::vector<FaceInfo>>>::iterator curBC = collectedBCs.begin(); curBC!=collectedBCs.end(); ++curBC) 
	{
		for(std::map<int, std::vector<FaceInfo>>::iterator curBCDom = curBC->second.begin(); curBCDom!=curBC->second.end(); ++curBCDom)
		{
			FaceSetA curBCSet(mbs, "BC" + mystr(curBC->first) + "_D" + mystr(curBCDom->first) );
			for(std::vector<FaceInfo>::iterator curFace = curBCDom->second.begin(); curFace!=curBCDom->second.end(); ++curFace)
			{
				//compute face area
				Vector3D n1,n2,n3,a2,a3;
				int nNodes = curFace->globalNodeNrs.Length();
				
				n1 = mesh->GetNodePtr(curFace->globalNodeNrs(1))->GetNodePos();
				n2 = mesh->GetNodePtr(curFace->globalNodeNrs(2))->GetNodePos();
				
				if(nNodes == 2) // edge
				{
					curFace->area = (n2-n1).Norm();
				}
				else // trig or quad
				{
					n3 = mesh->GetNodePtr(nNodes)->GetNodePos();
					a2 = n2-n1; a3 = n3-n1;
					curFace->area = (a2.Cross(a3)).Norm();
					if(nNodes==3) curFace->area *= .5;
				}

				curBCSet.AddEntry(curFace->elNr, curFace->faceNr, curFace->globalNodeNrs.Length(), curFace->globalNodeNrs, curFace->localNodeNrs, curFace->area); 
			}
			int nMesh = mesh->AddSet(&curBCSet);
		
			BoundaryMapping& curBCM = bcmapping[curBC->first];
			curBCM.bcTypeNr = curBC->first;
			if(curBCDom==curBC->second.begin())
			{
				curBCM.domin = curBCDom->first;
				curBCM.setNrin = nMesh;
			}
			else
			{
				curBCM.domout = curBCDom->first;
				curBCM.setNrout = nMesh;			
			}
		}
	}
	return 1;//NSets();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderSTL
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderSTL::InitializeLookupTables()
{
	nodetypelist.Flush(); 
	int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node2D"));
	nodetypelist.Add(defaultnodetype);

	elementtypelist.Flush();
	int defaultelementtype = mesh->GetComponentElemTypeIdx(mystr("Triangular"));  
	elementtypelist.Add(defaultelementtype);

	// NOTE AD: Triangular requires Node2D thus it flattens the STL
	// * option 1) use triangle <-- limits to X-Y plane          THIS IS CURRENTLY IMPLEMENTED
	// * option 2) always extrude/rotate                         THIS WAS PREVIOUSLY IMPLEMENTED

	return 1;
}
int MeshLoaderSTL::GetNodeTypeLookup(int i)
{
	return nodetypelist(1); // always returns default node type
}
int MeshLoaderSTL::GetElemTypeLookup(int nnodes, int i)
{
	if(nnodes==3) return elementtypelist(1); // always return triangle type
	return 0;
}
int MeshLoaderSTL::LoadFromFile()
{
	InitializeLookupTables();

	LoadFromFileToBuffers();
	RemoveUnusedNodes();
	GenerateDomainContainers();
	SplitIntoDomains();
	GenerateOwnNodes();
	GenerateOwnElements();	

	return 1;
}

int MeshLoaderSTL::LoadFromFileToBuffers()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();

// open source file
	ElementDataContainer edc;
	mbs->File2EDC(filename,&edc);

// additional ASSUMPTIONS:
// 1) all nodes are used at least once
// 2) all elements are 3-node trigs
// 3) all elements are material and domain 1 ( not specified in source file )
	int mat = 1;
	useddomainsbuffer.Flush();
	useddomainsbuffer.Add(1);

	Matrix m;
// read points - do NOT generate Mesh nodes right away
	GetElemDataMatrix(mbs,edc,"points",m);
	npoints = m.Getrows();
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints); // initialize at known size, most of the time only a single ReSize() operations on the TArray required

	Vector3D p;
	for(int i=1; i<=npoints;i++)
	{
		p = Vector3D(m(i,1),m(i,2),m(i,3));
		pointbuffer.Add(p);
	}

// read trigs - do NOT genreate MeshTrigs right away
	GetElemDataMatrix(mbs,edc,"triangles",m);
	nelements = m.Getrows();
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements); // initialize at known size, no more ReSize() operations on the TArray required
	
	MeshTrig trig;
	trig.SetTypeNr( GetElemTypeLookup(3) );
	trig.SetMatNr(mat);
	for(int i=1; i<=m.Getrows();i++)
	{
		trig.SetNode(1,(int)m(i,1));
		trig.SetNode(2,(int)m(i,2));
		trig.SetNode(3,(int)m(i,3));
		elementbuffer.Add(trig.GetCopy());
	}
	return useddomainsbuffer.Length();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderNetgen2D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderNetgen2D::InitializeLookupTables()
{
	nodetypelist.Flush(); 
	int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node2D"));
	nodetypelist.Add(defaultnodetype);

	elementtypelist.Flush();
	int triangularelementtype = mesh->GetComponentElemTypeIdx(mystr("Triangular"));  
	elementtypelist.Add(triangularelementtype);
	int quadrilateralelementtype = mesh->GetComponentElemTypeIdx(mystr("Quadrilateral"));  
	elementtypelist.Add(quadrilateralelementtype);

	return 1;
}
int MeshLoaderNetgen2D::GetNodeTypeLookup(int i)
{
	return nodetypelist(1); // always returns default node type
}
int MeshLoaderNetgen2D::GetElemTypeLookup(int nnodes, int i)
{
	if(nnodes==3 || nnodes==6) return elementtypelist(1); 
	if(nnodes==4 || nnodes==8) return elementtypelist(2);
	return 0;
}
int MeshLoaderNetgen2D::LoadFromFile()
{
	InitializeLookupTables();

	LoadFromFileToBuffers();
	RemoveUnusedNodes();

	GenerateDomainContainers();
	SplitIntoDomains();
	
	GenerateOwnNodes();
	GenerateOwnElements();	
		
	return 1;
}

int MeshLoaderNetgen2D::LoadFromFileToBuffers()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();

// open source file
	CMFile meshfile(filename.c_str(), TFMread);
	if(!meshfile.IsGood())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file " << filename << "\n";
		return 0;
	}
	
// additional ASSUMPTIONS:
// 1) all nodes are used at least once

	mystr file; int pos_file = 0;
	mystr line; int pos_line = 0; 
	mystr entry;

	meshfile.RWF(file);
// first line contains number of points as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	entry = line.GetWord(pos_line,1);
	npoints = entry.MakeInt();
	mbs->UO() << "npoints=" << npoints << " (file)\n";
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints); // initialize at known size, most of the time only a single ReSize() operations on the TArray required

// a known number of lines contain pairs of floats
	for(int i=1; i<=npoints; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		Vector3D p;
		entry = line.GetWord(pos_line,1);
		p.X() = entry.MakeDouble();
		entry = line.GetWord(pos_line,1);
		p.Y() = entry.MakeDouble();
		pointbuffer.Add(p);
	}

// next line contains number of elements as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	pos_line = 0;
	entry = line.GetWord(pos_line,1);
	nelements = entry.MakeInt();
	mbs->UO() << "npoints=" << npoints << " (file)\n";
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements); // initialize at known size, no more ReSize() operations on the TArray required

// a known number of lines contain an unknown number of integer entries
	useddomainsbuffer.Flush();
	for(int i=1; i<=nelements; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		entry = line.GetWord(pos_line,1);
		int domain = entry.MakeInt();
		if(useddomainsbuffer.Find(domain)<=0)
		{
			useddomainsbuffer.Add(domain);
		}		
	
		TArray<int> pointlist;
		entry = line.GetWord(pos_line,1);
		while(pos_line!=-1)
		{
			int nodenr = entry.MakeInt();
			pointlist.Add(nodenr);
			entry = line.GetWord(pos_line,1);
		}
		if(pointlist.Length()==3 || pointlist.Length()==6)
		{
			MeshTrig trig;
			trig.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
			trig.SetNodeList(pointlist);
			trig.SetMatNr(domain);
			elementbuffer.Add(trig.GetCopy());
		}
		else if(pointlist.Length()==4 || pointlist.Length()==8)
		{
			MeshQuad quad;
			quad.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
			quad.SetNodeList(pointlist);
			quad.SetMatNr(domain);
			elementbuffer.Add(quad.GetCopy());		
		}
		else
		{
			mbs->UO() << "nodelist for element" << i << "does not have the correct number of nodes for neither triangle(3,6) nor quadrilateral(4,8) \n";
		}
	}

// next line contains number of (domain)border-elements as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	pos_line = 0;
	entry = line.GetWord(pos_line,1);
	int nBCelems = entry.MakeInt();
	mbs->UO() << "nBCelements=" << nBCelems << "\n";

// a known number of lines contain triplets of integers, 
// when the first entry is >10000 then process further ( MagOpt specific policy )
	int bc_num_cur = -1;
	int number_of_bc = 0;
	int cur_index = 1;
	for(int i=1; i<=nBCelems; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		entry = line.GetWord(pos_line,1);
		int bc_num = entry.MakeInt();

		if(bc_num < 10000) { continue; } // skip these entries
		
		if(bc_num!=bc_num_cur)
		{
			bc_num_cur = bc_num;
			number_of_bc++;
			bc_s.Add(TArray<int>(2));
			bc_s(number_of_bc).Add(bc_num);
			cur_index=2;
		}

		entry = line.GetWord(pos_line,1);
		int nodenum1 = entry.MakeInt();
		bc_s(number_of_bc)(cur_index++) = nodenum1;

		entry = line.GetWord(pos_line,1);
		int nodenum2 = entry.MakeInt();
		bc_s(number_of_bc).Add(nodenum2);

// new: FaceInfo entries
		FaceInfo face;
		face.bcTypeNr = bc_num-10000;
		face.globalNodeNrs = TArray<int>(nodenum1,nodenum2);
		surfacesbuffer.Add(face);




	}

	return useddomainsbuffer.Length();
}

int MeshLoaderNetgen2D::LoadFromFileToBuffers_old()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();
	
// format specific assumptions
	int type = 1; //TODO: member variable of MeshLoaderNetgen2D
	int order = 1; //TODO: member variable of MeshLoaderNetgen2D

// open source file
	ifstream file(filename.c_str());
	if(!file.is_open())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file " << filename << "\n";
		return 0;
	}

// additional ASSUMPTIONS:
// 1) all nodes are used at least once
// 2) all elements are 3-node trigs or all elements are 6-node

// read points - do NOT generate Mesh nodes right away
	file >> npoints;
	mbs->UO() << "npoints=" << npoints << "\n";
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints); // initialize at known size, most of the time only a single ReSize() operations on the TArray required
 
	Vector3D p;
	for(int i=1; i<=npoints; i++) 
	{
		file >> p.X();
		file >> p.Y();
		pointbuffer.Add(p);
	}

// read trigs - do NOT genreate MeshTrigs right away
// identify domain numbers
	file >> nelements;
	mbs->UO() << "nelements=" << nelements << "\n";
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements); // initialize at known size, no more ReSize() operations on the TArray required
	
	TArray<int> pointlist;
	int pointspertrig = 3;
	if(order==2) pointspertrig = 6;
	pointlist.SetLen(pointspertrig);
	MeshTrig trig;
	trig.SetTypeNr( GetElemTypeLookup(pointspertrig) );

	int domain;
	useddomainsbuffer.Flush();
	for(int i=1; i<=nelements; i++)
	{
		file >> domain;
		if(useddomainsbuffer.Find(domain)<=0)
		{
			useddomainsbuffer.Add(domain);
		}
	
		for(int j=1; j<=pointspertrig; j++)
		{
			file >> pointlist(j);
		}
		trig.SetNodeList(pointlist);
		trig.SetMatNr(domain);
		elementbuffer.Add(trig.GetCopy());
	}

	int nBC_lines;
	file >> nBC_lines;

	int bc_num_old = -1;
	int number_of_bc = 0;
	int curr_index = 1;
	for (int i=1; i<=nBC_lines; i++)
	{
		int bc_num;
		file >> bc_num;
		if (bc_num < 10000)
		{
			file >> bc_num;
			file >> bc_num;
			continue;
		}
		if (bc_num != bc_num_old)
		{
			bc_num_old = bc_num;
			number_of_bc++;
			bc_s.Add(TArray<int>(2));
			bc_s(number_of_bc).Add(bc_num);
			curr_index=2;
		}

		int nodenum;
		file >> nodenum;
		bc_s(number_of_bc)(curr_index++) = nodenum;
		file >> nodenum;
		bc_s(number_of_bc).Add(nodenum);
	}


	return useddomainsbuffer.Length();
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderNetgen3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderNetgen3D::InitializeLookupTables()
{
	nodetypelist.Flush(); 
	int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node3D"));
	nodetypelist.Add(defaultnodetype);

	elementtypelist.Flush();
	int elementtype = mesh->GetComponentElemTypeIdx(mystr("Tetrahedral"));  
	elementtypelist.Add(elementtype);
	//$ PG 2017-10-9: MIXEDMESH_TODO: Mixed hexahedral/tetrahedral meshes yet to be implemented! Netgen (5.0.0) has problems with exporting mixed meshes.
	//elementtype = mesh->GetComponentElemTypeIdx(mystr("Quadrilateral"));  
	//elementtypelist.Add(elementtype);

	return 1;
}
int MeshLoaderNetgen3D::GetNodeTypeLookup(int i)
{
	return nodetypelist(1); // always returns default node type
}
int MeshLoaderNetgen3D::GetElemTypeLookup(int nnodes, int i)
{
	if(nnodes==4 || nnodes==10) return elementtypelist(1); 
	//if(nnodes==8 || nnodes==20) return elementtypelist(2);  //$ PG 2017-10-9: MIXEDMESH_TODO
	return 0;
}
int MeshLoaderNetgen3D::LoadFromFile()
{
	InitializeLookupTables();

	LoadFromFileToBuffers();
	RemoveUnusedNodes();
	GenerateDomainContainers();
	SplitIntoDomains();
	GenerateOwnNodes();
	GenerateOwnElements();	
		
	return 1;
}

int MeshLoaderNetgen3D::LoadFromFileToBuffers()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();

// open source file
	CMFile meshfile(filename.c_str(), TFMread);
	if(!meshfile.IsGood())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file " << filename << "\n";
		return 0;
	}
	
// additional ASSUMPTIONS:
// 1) all nodes are u sed at least once

	mystr file; int pos_file = 0;
	mystr line; int pos_line = 0; 
	mystr entry;

	meshfile.RWF(file);
// first line contains number of points as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	entry = line.GetWord(pos_line,1);
	npoints = entry.MakeInt();
	mbs->UO() << "npoints=" << npoints << " (file)\n";
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints); // initialize at known size, most of the time only a single ReSize() operations on the TArray required

// a known number of lines contain triplets of floats
	for(int i=1; i<=npoints; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		Vector3D p;
		entry = line.GetWord(pos_line,1);
		p.X() = entry.MakeDouble();
		entry = line.GetWord(pos_line,1);
		p.Y() = entry.MakeDouble();
		entry = line.GetWord(pos_line,1);
		p.Z() = entry.MakeDouble();
		pointbuffer.Add(p);
	}

// next line contains number of elements as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	pos_line = 0;
	entry = line.GetWord(pos_line,1);
	nelements = entry.MakeInt();
	mbs->UO() << "nelements=" << nelements << " (file)\n";
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements); // initialize at known size, no more ReSize() operations on the TArray required

// a known number of lines contain an unknown number of integer entries
	useddomainsbuffer.Flush();
	for(int i=1; i<=nelements; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		entry = line.GetWord(pos_line,1);
		int domain = entry.MakeInt();
		if(useddomainsbuffer.Find(domain)<=0)
		{
			useddomainsbuffer.Add(domain);
		}		
	
		TArray<int> pointlist;
		entry = line.GetWord(pos_line,1);
		while(pos_line!=-1)
		{
			int nodenr = entry.MakeInt();
			pointlist.Add(nodenr);
			entry = line.GetWord(pos_line,1);
		}
		if(pointlist.Length()==4 || pointlist.Length()==10)
		{
			MeshTet tet;
			tet.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
			tet.SetNodeList(pointlist);
			tet.SetMatNr(domain);
			elementbuffer.Add(tet.GetCopy());
		}
		//else if(pointlist.Length()==8 || pointlist.Length()==20)  //$ PG 2017-10-9: MIXEDMESH_TODO
		//{
		//	MeshQuad quad;
		//	quad.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
		//	quad.SetNodeList(pointlist);
		//	quad.SetMatNr(domain);
		//	elementbuffer.Add(quad.GetCopy());		
		//}
		else
		{
			mbs->UO() << "nodelist for element" << i << "does not have the correct number of nodes for neither tetrahedral(4,10) \n";
		}
	}

// next line contains number of (domain)border-elements as single entry
	file.GetUntilEOL(pos_file, '\n', line, 1);
	pos_line = 0;
	entry = line.GetWord(pos_line,1);
	int nBCelems = entry.MakeInt();
	mbs->UO() << "nBCelements=" << nBCelems << "\n";


#if 0 // do we need marking of boundaries in 3D? and if yes, how to generalize from 2D case, as done in the commented source
// a known number of lines contain triplets of integers, 
// when the first entry is >10000 then process further ( MagOpt specific policy )
	int bc_num_cur = -1;
	int number_of_bc = 0;
	int cur_index = 1;
	for(int i=1; i<=nBCelems; i++)
	{
		file.GetUntilEOL(pos_file, '\n', line, 1);
		pos_line = 0;
		entry = line.GetWord(pos_line,1);
		int bc_num = entry.MakeInt();

		if(bc_num < 10000) { continue; } // skip these entries
		
		if(bc_num!=bc_num_cur)
		{
			bc_num_cur = bc_num;
			number_of_bc++;
			bc_s.Add(TArray<int>(2));
			bc_s(number_of_bc).Add(bc_num);
			cur_index=2;
		}

		entry = line.GetWord(pos_line,1);
		int nodenum = entry.MakeInt();
		bc_s(number_of_bc)(cur_index++) = nodenum;

		entry = line.GetWord(pos_line,1);
		nodenum = entry.MakeInt();
		bc_s(number_of_bc).Add(nodenum);
	}
#endif

	return useddomainsbuffer.Length();
}





//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderNeutral3D
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderNeutral3D::InitializeLookupTables()
{
	nodetypelist.Flush(); 
	int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node3D"));
	nodetypelist.Add(defaultnodetype);

	elementtypelist.Flush();
	int objectfactoryelementtypeHex = mesh->GetComponentElemTypeIdx(mystr("Hexahedral"));	 
	int objectfactoryelementtypeTet = mesh->GetComponentElemTypeIdx(mystr("Tetrahedral"));
	int objectfactoryelementtypePri = mesh->GetComponentElemTypeIdx(mystr("Prism"));
	int objectfactoryelementtypePyr = mesh->GetComponentElemTypeIdx(mystr("Pyramid"));
	elementtypelist.Add(objectfactoryelementtypeHex);
	elementtypelist.Add(objectfactoryelementtypeTet);
	elementtypelist.Add(objectfactoryelementtypePri);
	elementtypelist.Add(objectfactoryelementtypePyr);

	return 4;
}
int MeshLoaderNeutral3D::GetNodeTypeLookup(int i)
{
	return nodetypelist(1); // always returns default node type
}
int MeshLoaderNeutral3D::GetElemTypeLookup(int nnodes, int i)
{
	if(nnodes==8 || nnodes==20) return elementtypelist(1); // Hexahedral type
	if(nnodes==4 || nnodes==10) return elementtypelist(2); // Tetrahedral type
	if(nnodes==6 || nnodes==15) return elementtypelist(3); // Prism type
	if(nnodes==5 || nnodes==13) return elementtypelist(4); // Pyramid type
	return 0;
}
int MeshLoaderNeutral3D::PermuteNodeList(TArray<int>& nodelist)
{
	// TODO: !!! complete this list !!! 
	// implementation for default ANSYS node numbering to HOTINT node numbering - taken from FEMesh::MapNodesAnsys2Hotint
	int dummy;
	if(nodelist.Length()==8 || nodelist.Length()==20)
	{
		dummy = nodelist(3); nodelist(3) = nodelist(4); nodelist(4)= dummy;
		dummy = nodelist(7); nodelist(7) = nodelist(8); nodelist(8)= dummy;
		if(nodelist.Length()==20)
		{

//			dummy = nodelist(10); nodelist(10) = nodelist(14); nodelist(14) = nodelist(16); nodelist(16) = nodelist(15);
//			nodelist(15) = nodelist(12); nodelist(12) = nodelist(13); nodelist(13) = nodelist(11); nodelist(11) = dummy;
//			dummy = nodelist(19); nodelist(19) = nodelist(20); nodelist(20) = dummy;

			dummy = nodelist(10); nodelist(10) = nodelist(11); nodelist(11) = nodelist(13); nodelist(13) = nodelist(12); 
			nodelist(12) = nodelist(15); nodelist(15) = nodelist(16); nodelist(16) = nodelist(14); nodelist(14) = dummy;
			dummy = nodelist(19); nodelist(19) = nodelist(20); nodelist(20) = dummy;

			return 13; // 4+9 = 13 nodes changed place
		}
		return 4; // 4 nodes changed place
	}
	if(nodelist.Length()==4 || nodelist.Length()==10)
	{
		if(nodelist.Length()==10)
		{
			dummy = nodelist(8); nodelist(8) = nodelist(10); nodelist(10) = nodelist(9); nodelist(9) = dummy; 
			return 3; 
		}
		return 0;
	}
	if(nodelist.Length()==6 || nodelist.Length()==15)
	{
		if(nodelist.Length()==15)
		{
			dummy = nodelist(10); nodelist(10) = nodelist(13); nodelist(13) = dummy;
			dummy = nodelist(11); nodelist(11) = nodelist(14); nodelist(14) = dummy;
			dummy = nodelist(12); nodelist(12) = nodelist(15); nodelist(15) = dummy;
			return 6;
		}
		return 0;
	}
	return -1;
}


int MeshLoaderNeutral3D::LoadFromFile()
{
	InitializeLookupTables();

	LoadFromFileToBuffers();
	RemoveUnusedNodes();
	GenerateDomainContainers();
	SplitIntoDomains();
	GenerateOwnNodes();
	GenerateOwnElements();	

	return 1;
}

int MeshLoaderNeutral3D::LoadFromFileToBuffers()
{	
	MBS* mbs = (MBS*) GetMesh()->GetMBS();

// open source file
	ifstream file(filename.c_str());
	if(!file.is_open())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file " << filename << "\n";
		return 0;
	}

// additional ASSUMPTIONS:
// 1) all nodes are used at least once
// 2) all elements are Hexes(8/20), Prisms(6/15), Pyramids(5/13), Tetra(4/10)
// 3) all elements are material and domain 1 ( not specified in source file )
	int mat = 1;
	useddomainsbuffer.Flush();
	useddomainsbuffer.Add(1);

// read first line  1 x {number of points, nr of elements}
	file >> npoints;
	mbs->UO() << "npoints=" << npoints << " (file)\n";
	file >> nelements;
	mbs->UO() << "nelements=" << nelements << " (file)\n"; // number of elements, ATTENTION thats the total value for 1D,2D,3D
	int dummy;

// read the points N x {idx, px, py, pz}
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints);

	Vector3D p;
	for(int i=1; i<=npoints; i++) 
	{	
		file >> dummy; // DO NOT PROCESS: Node Number
		file >> p.X();
		file >> p.Y();
		file >> p.Z();
		pointbuffer.Add(p);
	}

// read the elements N x {idx, type, M x {nodenum}}
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements);

// read the elements N x {idx, type, M x {nodenum}}
	int elemtype,dimension,elemnodes;
	for(int i=1; i<=nelements; i++) 
	{
		file >> dummy; // DO NOT PROCESS: Element Number
		file >> elemtype; // elemtype = 100*dimension + nodes
		dimension = elemtype/100;
		elemnodes = elemtype%100;
		
		TArray<int> pointlist;
		for(int j=1; j<=elemnodes; j++)
		{
			file >> pointlist(j);
		}

		if(dimension == 3)
		{
			PermuteNodeList(pointlist); // rearrange in HOTINT order
			if(elemnodes==4||elemnodes==10)
			{
				MeshTet tet;
				tet.SetNodeList(pointlist);
				tet.SetMatNr(mat);
				tet.SetTypeNr( GetElemTypeLookup(elemnodes) );
				elementbuffer.Add(tet.GetCopy());
			}
			if(elemnodes==5||elemnodes==13)
			{
				MeshPyr pyr;
				pyr.SetNodeList(pointlist);
				pyr.SetMatNr(mat);
				pyr.SetTypeNr( GetElemTypeLookup(elemnodes) );
				elementbuffer.Add(pyr.GetCopy());
			}
			if(elemnodes==6||elemnodes==15)
			{
				MeshPri pri;
				pri.SetNodeList(pointlist);
				pri.SetMatNr(mat);
				pri.SetTypeNr( GetElemTypeLookup(elemnodes) );
				elementbuffer.Add(pri.GetCopy());
			}
			if(elemnodes==8||elemnodes==20)
			{
				MeshHex hex;
				hex.SetNodeList(pointlist);
				hex.SetMatNr(mat);
				hex.SetTypeNr( GetElemTypeLookup(elemnodes) );
				elementbuffer.Add(hex.GetCopy());
			}
		}
	}
	nelements = elementbuffer.Length();
	mbs->UO() << "npoints=" << npoints << " (buffer)\n";
	file >> nelements;
	mbs->UO() << "nelements=" << nelements << " (buffer)\n"; 

	return useddomainsbuffer.Length();
}

namespace
{
	class uniform_series
	{
	public:
		uniform_series():
		  idx(-4),
		  values(0.5),
		  next_values(2)
		  {}
		double next()
		{
			idx++;
			double val = 0.;
			if (idx == -3)
			{
				val = 0.;
			}
			else if (idx == -2)
			{
				val =  1.;
			}
			else if (idx == -1)
			{
				val = 0.5;
			}
			else
			{
				if (idx % 2 == 0)
				{
					val =  0.5*values(idx/2+1);
				}
				else
				{
					val =  0.5*(1. + values(idx/2+1));
				}
				next_values(idx+1) = val;
			}
			if ((idx+1)/2 == values.Length())
			{
				values = next_values;
				idx = -1;
				next_values = Vector(2*values.Length());
			}
			return val;
		}
	private:
		Vector values;
		Vector next_values;
		int idx;
	};

	Vector3D rgb_from_hsl(Vector3D hsl)
	{
		double h = hsl(1);
		double s = hsl(2);
		double l = hsl(3);

		if (s == 0)
		{
			return Vector3D(l);
		}

		double t1;
		double t2;

		if (l<0.5)
		{
			t1 = l*(1+s);
		}
		else
		{
			t1 = l+s-l*s;
		}

		t2= 2*l-t1;

		Vector3D rgb;

		for (int i=1;i<=3;i++)
		{
			double temp = h+2./3.-i/3.;
			if (temp < 0)
				temp += 1;
			if (temp > 1)
				temp -= 1;

			if (6*temp<1)
			{
				rgb(i) = temp+6*temp*(t1-t2);
			}
			else if (2*temp<1)
			{
				rgb(i) = t1;
			}
			else if (3*temp < 2)
			{
				rgb(i) = t2 + (t1-t2)*(2./3.-temp)*6;
			}
			else
			{
				rgb(i)= t2;
			}
		}

		return rgb;
	}

	void ignoreLine(std::ifstream& file)
	{
		std::string line;
		std::getline(file, line);
	}

	bool checkLine(std::ifstream& file, const char* expected, MBS* mbs)
	{
		std::string line;
		std::getline(file, line);
		if (line.find(expected) == std::string::npos)
		{
			mbs->UO(UO_LVL_err) << "ERROR: file format wrong: \n";
			mbs->UO(UO_LVL_err) << "Line must be \""<< expected <<"\" but is " << line.c_str() << " \n";
			return false;
		}
		return true;
	}

	std::istringstream getLineStream(std::ifstream& file)
	{
		std::string line;
		std::getline(file, line);
		return std::istringstream(line);
	}

}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderVOL
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderVOL::InitializeLookupTables()
{
	nodetypelist.Flush(); 
	if (dim==3)
	{
		int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node3D"));
		nodetypelist.Add(defaultnodetype);
	}
	else
	{
		int defaultnodetype = mesh->GetComponentNodeTypeIdx(mystr("Node2D"));
		nodetypelist.Add(defaultnodetype);
	}

	elementtypelist.Flush();
	if (dim==3)
	{
		int elementtype = mesh->GetComponentElemTypeIdx(mystr("Tetrahedral"));  
		elementtypelist.Add(elementtype);
	}
	else
	{
		int elementtype = mesh->GetComponentElemTypeIdx(mystr("Triangular"));  
		elementtypelist.Add(elementtype);
		elementtype = mesh->GetComponentElemTypeIdx(mystr("Quadrilateral"));  
		elementtypelist.Add(elementtype);
	}
	return 1;
}
int MeshLoaderVOL::GetNodeTypeLookup(int i)
{
	return nodetypelist(1); // always returns default node type
}
int MeshLoaderVOL::GetElemTypeLookup(int nnodes, int i)
{
	if ( dim==3 && (nnodes==4 || nnodes==10))
	{
		return mesh->GetComponentElemTypeIdx(mystr("Tetrahedral")); 
	}
	else if ( dim==2 && (nnodes==3 || nnodes==6))
	{
		return mesh->GetComponentElemTypeIdx(mystr("Triangular")); 
	}
	else if ( dim==2 && (nnodes==4 || nnodes==8))
	{
		return mesh->GetComponentElemTypeIdx(mystr("Quadrilateral")); 
	}
	return 0;
}
int MeshLoaderVOL::LoadFromFile()
{
	InitializeLookupTables();
	
	LoadFromFileToBuffers();
	RemoveUnusedNodes();
	MatchSurfaceElements();
	
	GenerateDomainContainers();
	SplitIntoDomains();
	UpdateSymmetryNodes();
	SplitBoundaries();

	GenerateOwnNodes();
	GenerateOwnElements();
	GenerateOwnSets();

	return 1;
}

int MeshLoaderVOL::LoadFromFileToBuffers()
{
	MBS* mbs = (MBS*) GetMesh()->GetMBS();
	std::map<int, std::vector<std::vector<int>>> boundary_conditions_faces;
	//std::map<int, std::pair<int,int>> boundary_conditions_domains;

	mbs->UO(UO_LVL_dbg1) << "loading vol file \"" << filename << "\"\n";

	std::ifstream volFile(filename.c_str());
	std::string line;

	// file could be opened ?
	if(!volFile.is_open())
	{
		mbs->UO(UO_LVL_err) << "ERROR: could not open file: " << filename << "\n";	
		return 0;
	}

	// name
	ignoreLine(volFile); // get name (ignore)

	// "dimension"
	checkLine(volFile, "dimension", mbs);

	// dimension
	dim = 0;
	if (!(getLineStream(volFile) >> dim))
	{
		mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
		mbs->UO(UO_LVL_err) << "Wrong dimension: " << line.c_str() << "\n";
		return 0;
	}

	mbs->UO(UO_LVL_dbg1) << "VOL file dimension is " << dim << "\n"; 

	// "geomtype"
	checkLine(volFile, "geomtype", mbs);

	// geomtype
	int geomtype = 0;
	if (!(getLineStream(volFile) >> geomtype))
	{
		mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
		mbs->UO(UO_LVL_err) << "Wrong geomtype: " << line.c_str() << "\n";
		return 0;
	}

	// empty line
	ignoreLine(volFile);
	
	// header line
	ignoreLine(volFile);

// ************************************
// * BLOCK: load the surface elements *
// ************************************

	// "geomtype"
	checkLine(volFile, "surfaceelements", mbs);

	// number of surface elements
	int nSurf = 0;
	if (!(getLineStream(volFile) >> nSurf))
	{
		mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
		mbs->UO(UO_LVL_err) << "Wrong nSurf: " << line.c_str() << "\n";
		return 0;
	}

	if (dim == 2)
	{
		for (int i = 0; i<nSurf; ++i)
		{
			std::istringstream iss = getLineStream(volFile);
		
			int domain=0;
			int bcn=0;
			int nNodes=0;
			int domin = 0;
			int domout = 0;

			if (!(iss >> domain >> bcn >> domin >> domout >> nNodes))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "Wrong element node list " << line.c_str() << "\n";
				return 0;
			}
			useddomainsbuffer.AddIfNotExists(domain-1); // for 2D case the surface elements are the main elements

			TArray<int> pointlist = TArray<int>(nNodes);

			for (int i=1; i<=nNodes; ++i)
			{
				if (!(iss >> pointlist(i)))
				{
					mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
					mbs->UO(UO_LVL_err) << "Wrong element node list " << line.c_str() << "\n";
					return 0;
				}
			}

			if(pointlist.Length()==3 || pointlist.Length()==6)
			{
				MeshTrig trig;
				trig.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
				trig.SetNodeList(pointlist);
				trig.SetMatNr(domain-1);
				elementbuffer.Add(trig.GetCopy());
			}
			else if(pointlist.Length()==4 || pointlist.Length()==8)
			{
				MeshQuad quad;
				quad.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
				quad.SetNodeList(pointlist);
				quad.SetMatNr(domain-1);
				elementbuffer.Add(quad.GetCopy());
			}
			else
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "MeshLoaderVOL can only read quad or triangle meshes " << line.c_str() << "\n";
				return 0;
			}
		}
	}
	else // dim == 3: these are 2D surfaces of 3D elements
	{
		nborder = nSurf;
		surfacesbuffer.ReSize(2*nborder); // all loaded surfaceelements could split into two faces
		// get boundary conditions
		for (int i = 0; i<nSurf; ++i)
		{
			int surfnr, bcnr, domin, domout, np;
			std::istringstream lineStream = getLineStream(volFile);
			if (!(lineStream >> surfnr >> bcnr >> domin >> domout >> np))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				return 0;
			}

			// store domain numbers for the BC
			// NOTE: this overwrites the entry every time
		//	std::pair<int,int> domains; domains.first=domin; domains.second=domout;
//			boundary_conditions_domains[bcnr]=domains;

			std::vector<int> nodeNrs(np);
			for (int i = 0; i < np; ++i)
			{
				if (!(lineStream >> nodeNrs[i]))
				{
					mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
					return 0;
				}
			}
			//boundary_conditions_faces[bcnr].push_back(nodeNrs);	

			
			FaceInfo face;
//			face.entryNr = surfacesbuffer.Length()+1;
			face.bcTypeNr = bcnr;
			face.domin = domin;
			face.domout = domout;
			face.globalNodeNrs.ReSize(np);
			for (int i = 0; i < np; ++i) { face.globalNodeNrs.Add(nodeNrs[i]); }
			surfacesbuffer.Add(face);
		}
	}

	// ignore 3 lines
	ignoreLine(volFile);
	ignoreLine(volFile);
	ignoreLine(volFile);

// ***********************************
// * BLOCK: load the volume elements *
// ***********************************

	checkLine(volFile, "volumeelements", mbs);

	// number of elements
	int nEl = 0;
	if (!(getLineStream(volFile) >> nEl))
	{
		mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
		mbs->UO(UO_LVL_err) << "Wrong nEl: " << line.c_str() << "\n";
		return 0;
	}
	//read elements

	if (dim==2)
	{
		for (int i = 0; i<nEl; ++i)
		{
			ignoreLine(volFile);
		}
	}
	else
	{
/*
		std::map<int, std::vector<FaceInfo>> boundary_conditions_FaceInfo;
*/
		for (int iE = 1; iE<=nEl; ++iE)
		{
			std::istringstream iss = getLineStream(volFile);
		
			int domain=0;
			int nNodes=0;

			if (!(iss >> domain >> nNodes))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "Wrong element node list " << line.c_str() << "\n";
				return 0;
			}
			useddomainsbuffer.AddIfNotExists(domain);

			TArray<int> pointlist = TArray<int>(nNodes);

			for (int i=1; i<=nNodes; ++i)
			{
				if (!(iss >> pointlist(i)))
				{
					mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
					mbs->UO(UO_LVL_err) << "Wrong element node list " << line.c_str() << "\n";
					return 0;
				}
			}
			if(pointlist.Length()==4 || pointlist.Length()==10)
			{
				MeshTet tet;
				tet.SetTypeNr( GetElemTypeLookup(pointlist.Length()) );
				tet.SetNodeList(pointlist);
				tet.SetMatNr(domain);
/*				
				// check if one of the element has a boundary condition face
				for (std::map<int, std::vector<std::vector<int>>>::const_iterator it = boundary_conditions_faces.begin(); it != boundary_conditions_faces.end(); ++it)
				{
					const std::vector<std::vector<int>> & nodelists = it->second;
					for (int j = 1, n=nodelists.size(); j<=n; ++j)
					{
						const std::vector<int> & nodelist = nodelists[j-1];
						bool allFound = true;
						for (int k = 1, n= nodelist.size(); k <= n && allFound ; ++k)
						{
							allFound = (pointlist.Find(nodelist[k-1]) > 0);
						}
						if (allFound)
						{
							int locFaceNr = -1;
							for (int k = 1, n=tet.NFaces(); k<=n; ++k)
							{
								const TArray<int> & facenodes = tet.GetFaceNodeList(k);
								bool allFound = true;
								for (int m = 1, n = nodelist.size(); m<=n; ++m)
								{
									allFound = (facenodes.Find(nodelist[m-1]) > 0);
								}
								if (allFound)
								{
									locFaceNr = k;
									break;
								}
							}
							assert(locFaceNr > 0);
							FaceInfo face;
							face.elNr = iE;
							face.faceNr = locFaceNr;
							face.localNodeNrs = tet.GetLocalFace(locFaceNr).GetNodeList();
							face.globalNodeNrs = tet.GetFaceNodeList(locFaceNr);
							face.area = 0.;
							boundary_conditions_FaceInfo[it->first].push_back(face);
						}
					}
				}
*/				
				elementbuffer.Add(tet.GetCopy());
			}
			else
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "MeshLoaderVOL can only read tet-meshes " << line.c_str() << "\n";
				return 0;
			}
		}
/*
		bc_faceSetsNumbers.Flush();
		for (std::map<int, std::vector<FaceInfo>>::const_iterator it = boundary_conditions_FaceInfo.begin(); it != boundary_conditions_FaceInfo.end(); ++it)
		{
			int bcNr = it->first;
//			FaceSetA bc(mbs, "bc" + mystr(bcNr) + "_A");
	
			int domin = boundary_conditions_domains[bcNr].first;
			int domout = boundary_conditions_domains[bcNr].second;
			FaceSetA bcInner(mbs, "bc" + mystr(bcNr) + "_D" + mystr(domin));
			FaceSetA bcOuter(mbs, "bc" + mystr(bcNr) + "_D" + mystr(domout));

			int nFaces = it->second.size();
			for (int i = 1; i<nFaces; ++i)
			{
				const FaceInfo & face = it->second[i-1];
//				bc.AddEntry(face.elNr, face.faceNr, face.globalNodeNrs.Length(), face.globalNodeNrs, face.localNodeNrs, face.area);
				
				int eldom = elementbuffer(face.elNr)->GetMatNr();
				if(eldom==domin)  { bcInner.AddEntry(face.elNr, face.faceNr, face.globalNodeNrs.Length(), face.globalNodeNrs, face.localNodeNrs, face.area); }
				if(eldom==domout) { bcOuter.AddEntry(face.elNr, face.faceNr, face.globalNodeNrs.Length(), face.globalNodeNrs, face.localNodeNrs, face.area); }
			}
			//int nMesh = mesh->AddSet(&bc);
			//bc_faceSetsNumbers.Add(int2(bcNr, nMesh));
			if(bcInner.GetNrOfEntries()>0) { int nMesh = mesh->AddSet(&bcInner); bc_faceSetsNumbers.Add(int2(bcNr, nMesh)); }
			if(bcOuter.GetNrOfEntries()>0) { int nMesh = mesh->AddSet(&bcOuter); bc_faceSetsNumbers.Add(int2(bcNr, nMesh)); }
		}
		*/
	}


	// ignore 3 lines
	ignoreLine(volFile);
	ignoreLine(volFile);
	ignoreLine(volFile);

	checkLine(volFile, "edgesegmentsgi2", mbs);

	// number of elements
	int edgesegmentsgi2 = 0;
	{
		if (!(getLineStream(volFile) >> edgesegmentsgi2))
		{
			mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
			mbs->UO(UO_LVL_err) << "Wrong edgesegmentsgi2: " << line.c_str() << "\n";
			return 0;
		}
	}

	//ignore
	int bc_num_cur = -1;
	int number_of_bc = 0;
	int cur_index = 1;


	if(dim==2) 
	{
		nborder = edgesegmentsgi2;
		surfacesbuffer.ReSize(2*nborder); // all loaded surfaceelements could split into two faces
	}
	for (int i = 0; i<edgesegmentsgi2; ++i)
	{
		std::istringstream iss = getLineStream(volFile);

		int bc_num = 0;
		int dummy = 0;
		int p1 = 0;
		int p2 = 0;
		int trignum1 = 0;
		int trignum2 = 0;
		int dominsurfnr1 = 0;
		int domoutsurfnr2 = 0;
		int ednr1 = 0;
		double dist1 = 0.;
		int ednr2 = 0;
		double dist2 = 0.;
		
		if (!(iss >> bc_num >> dummy >> p1 >> p2 >> trignum1 >> trignum2 >> dominsurfnr1 >> domoutsurfnr2 >> ednr1 >> dist1 >> ednr2 >> dist2))
		{
			mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
			mbs->UO(UO_LVL_err) << "Wrong boundary conditions " << line.c_str() << "\n";
			return 0;
		}

		if (bc_num == 1007) // SyMSpace specific: 1007: symmetry boudary condition
		{
			symmetricNodes.AddIfNotExists(p1);
			symmetricNodes.AddIfNotExists(p2);
		}

		if(bc_num < 10000) { continue; } // skip these entries

		//if(bc_num != bc_num_cur)
		//{
		//	bc_num_cur = bc_num;
		//	number_of_bc++;
		//	bc_s.Add(TArray<int>(2));
		//	bc_s(number_of_bc).Add(bc_num);
		//	cur_index=2;
		//}

		//bc_s(number_of_bc)(cur_index++) = p1;
		//bc_s(number_of_bc).Add(p2);

		FaceInfo face;
		face.bcTypeNr = bc_num;
		face.domin = dominsurfnr1;
		face.domout = domoutsurfnr2;
		face.globalNodeNrs.Flush();
		face.globalNodeNrs.Add(p1);
		face.globalNodeNrs.Add(p2);
		surfacesbuffer.Add(face);
	}

	// ignore 3 lines
	ignoreLine(volFile);
	ignoreLine(volFile);
	ignoreLine(volFile);

	checkLine(volFile, "points", mbs);

	// number of nodes
	int nNodes = 0;
	{
		if (!(getLineStream(volFile) >> nNodes))
		{
			mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
			mbs->UO(UO_LVL_err) << "Wrong nNodes: " << line.c_str() << "\n";
			return 0;
		}
	}

	//read nodes
	for (int i = 0; i<nNodes; ++i)
	{
		double x;
		double y;
		double z;
		if (!(getLineStream(volFile) >> x >> y >> z))
		{
			mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
			mbs->UO(UO_LVL_err) << "Wrong node position " << line.c_str() << "\n";
			return 0;
		}
		pointbuffer.Add(Vector3D(x,y,z));
	}

	std::getline(volFile, line);
	if (line.find("identifications") != std::string::npos)
	{
		mbs->UO(UO_LVL_dbg1) << "INFO: Periodic boundary conditions encountered in " << filename << "\n";
		int nIdent = 0;
		{
			if (!(getLineStream(volFile) >> nIdent))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "Wrong nIdentifications: " << line.c_str() << "\n";
				return 0;
			}
		}

		for (int i = 0; i<nIdent; ++i)
		{
			int n1;
			int n2;
			int dummy;
			if (!(getLineStream(volFile) >> n1 >> n2 >> dummy))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "Wrong identification " << line.c_str() << "\n";
				return 0;
			}

			//idents.AddIfNotExists(int2(n1,n2));  
			// AD: this causes redundant constraints if the 1st and second side are swapped for a segmented periodic boundary ( has happened for e.g. multiple ForceFits )
			int findforward = idents.Find(int2(n1, n2));
			int findreverse = idents.Find(int2(n2, n1));
			if (!findforward && !findreverse)
			{
				idents.Add(int2(n1, n2));
			}
		}
	}

// AD: unsure about sequence -> TODO: while not EOF check keywords ...
// volfile may also contain the following entries:
// no proper processing yet

	std::getline(volFile, line);
	if (line.find("materials") != std::string::npos)
	{
		mbs->UO(UO_LVL_dbg1) << "INFO: material names encoutered in  " << filename << "\n";
		int nMatNames = 0;
		{
			if (!(getLineStream(volFile) >> nMatNames))
			{
				mbs->UO(UO_LVL_err) << "ERROR: file format wrong: " << filename << "\n";
				mbs->UO(UO_LVL_err) << "Wrong nIdentifications: " << line.c_str() << "\n";
				return 0;
			}
		}
	}

	std::getline(volFile, line);
	if (line.find("bcnames") != std::string::npos)
	{
		mbs->UO(UO_LVL_dbg1) << "INFO: boundary condition names encoutered in  " << filename << "\n";
	}



	//ignore rest of the file for the moment

	return useddomainsbuffer.Length();
}

int MeshLoaderVOL::MatchSurfaceElements()
{
	for(int i=1, n=surfacesbuffer.Length(); i<=n; i++) // loop surfaceelements
	{
		FaceInfo& face = surfacesbuffer(i);
		bool checkDomin = (face.domin !=0);
		bool checkDomout = (face.domout !=0);
		int matched = 0;
		for(int j=1, n=elementbuffer.Length(); j<=n && (checkDomin || checkDomout); j++) // loop meshelements
		{
			MeshElement* elem = elementbuffer(j);

			if( (checkDomin && face.domin !=0 && face.domin == elem->GetMatNr()) ||    // filter 1: domains that match and still have to be checked
				(checkDomout && face.domout != 0 && face.domout == elem->GetMatNr() )) 
			{
				bool allFound = true;                                                 
				for(int k=1, n=face.globalNodeNrs.Length(); k<=n && allFound; k++)    // filter 2: all nodes from Facet/Edge in Element
				{
					allFound = elem->ContainsNode(face.globalNodeNrs(k));
				}
				if(allFound)
				{
					for(int k=1, n=elem->NFaces(); k<=n; k++) // loop meshelemenfaces
					{
						TArray<int>& elemfacenodes = elem->GetFaceNodeList(k);
						if(face.globalNodeNrs.Length()==elemfacenodes.Length())      // filter 3: same number of nodes for border and elementfacet
						{
							bool allFound = true;
							for(int l=1, n=face.globalNodeNrs.Length(); l<=n && allFound; l++)
							{
								allFound =(elemfacenodes.Find(face.globalNodeNrs(l)) > 0);
							}
							if(allFound)
							{
							// all nodes of Face(border) found Face(element)
								if(face.domin == elem->GetMatNr())  { checkDomin = false;  }  // no longer check inner domain
								if(face.domout == elem->GetMatNr()) { checkDomout = false; }  // no longer check outer domain
							
								if(matched==0) 
								{
									face.elNr = j;
									face.faceNr = k;
									face.localNodeNrs = TArray<int>(face.globalNodeNrs.Length());
									for (int i=1; i<= face.globalNodeNrs.Length(); ++i)
									{
										face.localNodeNrs(i) = elem->NodeList().Find(face.globalNodeNrs(i));
									}
									
								}
								else
								{
									FaceInfo second;
									second.bcTypeNr = face.bcTypeNr;									
									second.domin = face.domin; second.domout = face.domout;
									second.globalNodeNrs.CopyFrom(face.globalNodeNrs);

									second.elNr = j;
									second.faceNr = k;
									second.localNodeNrs = TArray<int>(second.globalNodeNrs.Length());
									for (int i=1; i<= second.globalNodeNrs.Length(); ++i)
									{
										second.localNodeNrs(i) = elem->NodeList().Find(second.globalNodeNrs(i));
									}
									surfacesbuffer.Add(second);
								}		
								matched++;
							}
						}
					}
				}
			}
		}
		if (matched == 0)
		{
			MBS* mbs = (MBS*)GetMesh()->GetMBS();
			mbs->UO(UO_LVL_warn) << "WARNING: face with nodes (" << face.globalNodeNrs(1) << ", " << face.globalNodeNrs(1) << ") is not part of any element\n";
			surfacesbuffer.Erase(i);
			n -= 1;
			i -= 1;
		}
	}
	return surfacesbuffer.Length();
}

//void MeshLoaderVOL::IdentifySymmetricNodes()
//{
//	for (int i = 1, n = symmetricNodes.Length(); i <= n; i++) // loop surfaceelements
//	{
//		int3& symmetricNode = symmetricNodes(i);
//		bool matched = false;
//		for (int j = 1, n = elementbuffer.Length(); j <= n; j++) // loop meshelements
//		{
//			MeshElement* elem = elementbuffer(j);
//			bool found = elem->ContainsNode(symmetricNode(1));
//			if (found)
//			{
//				matched = true;
//				symmetricNode(2) = j;
//				symmetricNode(3) = elem->NodeList().Find(symmetricNode(1));
//				break;
//			}
//		}
//		//release_assert(matched)
//	}
//}

void MeshLoaderVOL::UpdateSymmetryNodes()
{
	//update symmetric boundary condition
	// add boundary conditions for updated elements
	for (int i = 1, nSymm = symmetricNodes.Length(); i <= nSymm; i++)
	{
		int& globalNodeNumber = symmetricNodes(i);

		if (nodedomainmultiplicity(globalNodeNumber) > 1)
		{
			const TArray<int>& domainNrs = nodenumberindomain(globalNodeNumber);
			for (int k = 1; k <= domainNrs.Length(); k++)
			{
				if (domainNrs(k) != 0 && domainNrs(k) != globalNodeNumber)
				{
					// generate a new entry
					symmetricNodes.Add(domainNrs(k));
				}
			}
		}
	}
}


Vector3D MeshLoaderVOL::GetColor(int i)
{
	if (i == 0 || elementbuffer.Length()<i)
	{
		return Vector3D(0,0,0);
	}
	else
	{
		int matnr = elementbuffer(i)->GetMatNr();
		double s = 1.0;
		double l = 0.5;
		while (matnr > 48)
		{
			matnr -= 48;
		}
		while (matnr > 8)
		{
			matnr -= 8.;
			s *= 0.80;
		}
		if (matnr > 4)
		{
			matnr -= 4.;
			l = 0.25;
		}
		uniform_series gen;
		double h = 0.;
		for (int i = 0; i<= matnr; i++)
		{
			h = gen.next();
		}
		return rgb_from_hsl(Vector3D(h, s, l));
	}
}

void MeshLoaderVOL::GetElementData(ElementDataContainer & edc)
{
	MeshComponentLoader::GetElementData(edc);

	ElementData ed;
	Vector symmetricNodesDoubleVector(symmetricNodes.Length());
	for (size_t i = 1, n = symmetricNodes.Length(); i <= n; ++i)
	{
		symmetricNodesDoubleVector(i) = double(symmetricNodes(i));
	}
	ed.SetVector(symmetricNodesDoubleVector.GetVecPtr(), symmetricNodesDoubleVector.Length(), "nodes");
	ed.SetValuesInt(); ed.SetVariableLength(); ed.SetToolTipText("symmetric nodes"); ed.SetLocked(1);
	edc.TreeAdd(mystr("SymmetryBoundaryConditions"), ed);
}

TArray<int2> MeshLoaderVOL::GetPeriodicBCNodePairs() const
{
	return idents;
}

TArray<int> MeshLoaderVOL::GetSymmetricNodes() const
{
	return symmetricNodes;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// class  MeshLoaderDataArray
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MeshLoaderDataArrays::LoadNodesFromArrays(int npointsI, int dofsperpoint, double* pointdofs)
{ 
// assume that exactly one DomainContainer exists - generated in Reset()
	if(NSubs()==0) return 0;

// load routine
	npoints = npointsI;
	pointbuffer.Flush();
	pointbuffer.ReSize(npoints); 

	Vector3D p;
	int idx0;
	for(int i=1; i<=npoints; i++)
	{
		idx0 = dofsperpoint*(i-1);
		p = Vector3D(pointdofs[idx0],pointdofs[idx0+1],pointdofs[idx0+2]);
		pointbuffer.Add(p);
	}
// mapping routine
	nodemapping_CN2I.Flush();
	nodemapping_CN2I.SetLen(1);
	nodemapping_CN2I(1) = NaturalNumbers(npoints);
	nodemapping_I2CN.Flush();
	nodemapping_I2CN.ReSize(npoints);
	nodemapping_I2CN.SetLen(0);
	for(int i=1; i<=npoints; i++) { nodemapping_I2CN.Add( int2(1,i) ); }
// generate routine
	MeshDomainContainer* domaincontainer = (MeshDomainContainer*) GetSubordinateComponent(1);
	MeshNode n;
	int nodetypenumber = mesh->GetComponentNodeTypeIdx(this->GetNodeType());
	n.SetTypeNr(nodetypenumber);
	n.SetBodyIdx(1/*this->GetBodyIndex()*/);
	for(int j=1; j<=nodemapping_CN2I(1).Length(); j++)
	{
		int nnrunique = nodemapping_CN2I(1)(j);
		n.SetNodePos(pointbuffer(nnrunique));
		domaincontainer->AddNode(n);		
	}
	return NNodes();
}
int MeshLoaderDataArrays::LoadElementsFromArrays(int nelemsI, int nodesperelement, int* elementnodelists)
{
// assume that exactly one DomainContainer exists - generated in Reset()
	if(NSubs()==0) return 0;

// load routine
	nelements = nelemsI;
	elementbuffer.Flush();
	elementbuffer.ReSize(nelements); 

	int elemtypenumber = mesh->GetComponentElemTypeIdx(elemtype);
	int idx0;
	for(int i=1; i<=nelements; i++)
	{
		int idx0 = nodesperelement*(i-1);
		TArray<int> nodelist(nodesperelement);
		for(int j=0; j<nodesperelement; j++)
		{
			nodelist.Add(elementnodelists[idx0+j]);
		}
		if(nodesperelement==8) // replace this with mystr->Elementtype->NNodes()
		{
			MeshHex hex;
			hex.SetNodeList(nodelist);
			hex.SetMatNr(1);
			hex.SetTypeNr(elemtypenumber);
			elementbuffer.Add(hex.GetCopy());
		}
	}
// mapping routine
	elementmapping_CE2I.Flush();
	elementmapping_CE2I.SetLen(1);
	elementmapping_CE2I(1) = NaturalNumbers(nelements);
	elementmapping_I2CE.Flush();
	elementmapping_I2CE.ReSize(nelements);
	elementmapping_I2CE.SetLen(0);
	for(int i=1; i<=nelements; i++) { elementmapping_I2CE.Add( int2(1,i) ); }
// generate routine
	MeshDomainContainer* domaincontainer = (MeshDomainContainer*) GetSubordinateComponent(1);
	for(int j=1; j<=elementmapping_CE2I(1).Length(); j++)
	{
		int enrunique = elementmapping_CE2I(1)(j);
		MeshElement* curelem = elementbuffer(enrunique);
		for(int k=1; k<=curelem->NNodes(); k++)
		{
			int nnrunique = curelem->GetNode(k);
			int nnrdomain = nodemapping_I2CN(nnrunique)(2);
			curelem->SetNode(k,nnrdomain);
		}
		domaincontainer->AddElem(*curelem);
	}
	return NElems();
}
